@echo off
REM
REM   $Id: SETDEB.BAT 1.46 2003/08/25 15:17:14Z thomd Exp $
REM
REM   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
REM
REM   Copyright (c) 1993 - 2002 Datalight, Inc.
REM   All Rights Reserved
REM
REM   Datalight, Inc. is a Washington corporation with an address at
REM   21520 30TH DR SE MS110 Bothell, WA, 98021
REM   Tel: 800.221.6630  Fax: 425.951.8091
REM   Http://www.datalight.com
REM
REM   This software, including without limitation all source code and
REM   documentation, is the confidential, trade secret property of
REM   Datalight, Inc., and is protected under the copyright laws of
REM   the United States and other jurisdictions. Portions of the
REM   software may also be subject to one or more the following US
REM   patents: US#5860082, US#6260156.
REM
REM   In addition to civil penalties for infringement of copyright
REM   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
REM   penalties for violation of (a) the restrictions on circumvention
REM   of copyright protection systems found in 17 U.S.C. 1201 and
REM   (b) the protections for the integrity of copyright management
REM   information found in 17 U.S.C. 1202.
REM
REM   U.S. Government Restricted Rights. Use, duplication,
REM   reproduction, or transfer of this commercial product and
REM   accompanying documentation is restricted in accordance with
REM   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
REM
REM   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
REM   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
REM   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
REM   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
REM   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
REM   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
REM   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
REM   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
REM   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
REM   THIS SOFTWARE IN ERROR.
REM
REM
REM  SETDEB.BAT
REM
REM  Copyright (c) 1989-2002 Datalight, Inc.
REM  All Rights Reserved.
REM  Datalight Confidential and Proprietary
REM
REM  SETDEB.BAT sets the ROMDOS build environment switches.
REM
REM  USAGE:
REM
REM  setdeb mode [32] [7] [BUILDNUM]
REM
REM  modes: RELEASE   Sets normal switches for release (SDKs)
REM         XT        Sets for XT (8088/8086) compilation (default is 186)
REM         OFF       For use with SoftIce, symbols in C files
REM         ON        For use with SoftIce, symbols in C and ASM files
REM         TD        Sets -v for debugging utilities with Turbo Debug
REM
REM  BUILDNUM  Enables build numbers in the sign on of the utilities
REM        32  Enables FAT32 support (changes version number to 7.1)
REM         7  Enables Long File Name support (in kernel only)
REM       SUV  Enables Single User Version
REM       REP  Enables REPlication Version
REM
REM
REM  HISTORY:
REM
REM  3-22-2002 td
REM  - added "CANON_186" environment setting for 0x186 build
REM
REM  11-22-2000 bt
REM  - Rewrote the parsing algorithms to allow options to appear out of order
REM  - Added BUILDNUM switch to allow build numbers in the utilities
REM
REM  02-11-1999 BT
REM  - Added a switch to enable FAT32 support
REM  - Updated the usage screen
REM  - Added more command line parameter checking
REM  - Added a check for the CPATH enviornment variable
REM

SET CANON_186=FALSE

REM  Test for SETX in current path
REM  (This should be located in the BIN subdirectory of the Source Tree)
REM
setx rd_setdeb=FOUND
if "%rd_setdeb%"=="" goto setx_error
goto Clear_Environment

:setx_error
echo.
echo Error! SETX is not in your current path.
echo.
goto Exit

:Clear_Environment
REM  Clear the old environment
REM
setx rd_setdeb=
setx rd_sw=
setx rd_sw_a=
setx rd_sw_l=
setx vdos=
setx rd_lev=
setx buildnum=
setx suv=
setx rep=


REM  Make sure there is at least one command line argument
REM
if "%1"=="" goto Usage
if "%1"=="?" goto Current
if "%1"=="/?" goto Current


REM make sure environment is right
REM
if %BCROOT%!==! goto BadEnv
if %TASM4%!==!  goto BadEnv


REM  Set default options
REM
set _FAT32=DISABLED
set _B386=DISABLED
set _LFNS=DISABLED
set _BUILDNUM=DISABLED
set _SUV=DISABLED
set _REP=DISABLED

REM  Handle the different types of command line arguments
REM
:HdlArg


REM  Determine if we are done processing the command line
REM
if "%1"=="" goto SetCfg


REM  Convert known options to uppercase
REM
set _OPTION=%1
if "%_OPTION%"=="release" set _OPTION=RELEASE
if "%_OPTION%"=="xt" set _OPTION=XT
if "%_OPTION%"=="td" set _OPTION=TD
if "%_OPTION%"=="on" set _OPTION=ON
if "%_OPTION%"=="off" set _OPTION=OFF
if "%_OPTION%"=="buildnum" set _OPTION=BUILDNUM
if "%_OPTION%"=="suv" set _OPTION=SUV
if "%_OPTION%"=="rep" set _OPTION=REP


REM  Make sure we have not already set a build mode
REM
if not "%_BUILDMODE%"=="" goto GotMode


REM  Determine if this is a valid mode
REM
if "%_OPTION%"=="RELEASE" goto SetMode
if "%_OPTION%"=="XT" goto SetMode
if "%_OPTION%"=="TD" goto SetMode
if "%_OPTION%"=="ON" goto SetMode
if "%_OPTION%"=="OFF" goto SetMode


REM  This option is not a build mode
REM
goto GotMode


REM  Save the build mode
REM
:SetMode
set _BUILDMODE=%_OPTION%
goto NextArg


REM  Build mode alread set.  There can be only one build mode
REM
:GotMode


REM  Determine if FAT32 is enabled
REM
if "%_OPTION%"=="32" if "%_FAT32%"=="ENABLED" goto ConfOption
if "%_OPTION%"=="32" set _FAT32=ENABLED
if "%_OPTION%"=="32" goto NextArg


REM Removed -- some unknown number of the "C" instructions we currently use
REM to handle interrupts do NOT preserve the upper word of large registers.
REM
REM REM  Determine if 386 compile is enabled
REM REM
REM if "%_OPTION%"=="386" if "%_B386%"=="ENABLED" goto ConfOption
REM if "%_OPTION%"=="386" set _B386=ENABLED
REM if "%_OPTION%"=="386" goto NextArg


REM  Determine if LFN is enabled
REM
if "%_OPTION%"=="7" if "%_LFNS%"=="ENABLED" goto ConfOption
if "%_OPTION%"=="7" set _LFNS=ENABLED
if "%_OPTION%"=="7" goto NextArg


REM  Determine if build numbers will be stamped in the utilities
REM
if "%_OPTION%"=="BUILDNUM" if "%_BUILDNUM%"=="ENABLED" goto ConfOption
if "%_OPTION%"=="BUILDNUM" set _BUILDNUM=ENABLED
if "%_OPTION%"=="BUILDNUM" goto NextArg

REM Determine if Single User Version will be stamped in the utilites
REM
if "%_OPTION%"=="SUV" if "%_SUV%"=="ENABLED" goto ConfOption
if "%_OPTION%"=="SUV" set _SUV=ENABLED
if "%_OPTION%"=="SUV" goto NextArg

REM Determine if REPlication Version will be stamped in the utilites
REM
if "%_OPTION%"=="REP" if "%_REP%"=="ENABLED" goto ConfOption
if "%_OPTION%"=="REP" set _REP=ENABLED
if "%_OPTION%"=="REP" goto NextArg


REM  Display an error message for the bad option
REM
goto BadOption


REM  Go to the next argument
REM
:NextArg
shift
goto HdlArg


REM  Display a message for a bad command line option
REM
:BadOption
echo.
echo Error! Incorrect command line option: %_OPTION%
echo.
goto end


REM  Display a message for a conflicting command line option
REM
:ConfOption
echo.
echo Error! Conflicting command line option: %_OPTION%
echo.
goto end


REM  Display a message indicating that no build mode was supplied
REM
:NoMode
echo.
echo Error! No build mode specified.
echo.
goto end


:BadEnv
echo You must set some environment variables before the project will
echo compile:
echo.
echo BCROOT   defines where to find BorlandC (example C:\BC2)
echo TASM4    defines where to find tasm4    (example C:\TASM4\BIN\TASM4.EXE)
goto end


REM  Display the usage screen
REM
:Usage
echo SETDEB.BAT sets the ROMDOS build environment switches.
echo.
echo setdeb mode [32] [7] [BUILDNUM]
echo.
echo modes: RELEASE  Sets normal switches for release (SDKs)
echo        XT       Sets for XT (8088/8086) compilation (default is 186)
echo.       OFF      For use with SoftIce, symbols in C files
echo.       ON       For use with SoftIce, symbols in C and ASM files
echo        TD       Sets -v for debugging utilities with Turbo Debug
echo.
echo  BUILDNUM  Enables build numbers in the sign on of the utilities
echo        32  Enables FAT32 support (changes version number to 7.1)
echo         7  Enables Long File Name support (in kernel only)
REM echo       386  Enables a 386 build of ROM-DOS 7.1
echo       SUV  Enables Single User Version
echo       REP  Enables REPlication Version
echo.
:Current
if NOT "%RDBuild%"=="" echo  Your current build environment is:  %RDBuild%
echo.
goto End


REM  Set the build environment
REM
:SetCfg


REM  Make sure that we know what build mode to set
REM
if "%_BUILDMODE%"=="" goto NoMode


REM  This variable is used not to enable a FAT32 build but to
REM  enable the FAT32 functionality is some of the utilities.
REM
setx ENABLEF32=


REM  Set the LFN environment
REM
if not "%_LFNS%"=="ENABLED" setx VDOS=6
if not "%_LFNS%"=="ENABLED" setx RD_LEV=VDOS=6
if "%_LFNS%"=="ENABLED"     setx VDOS=7
if "%_LFNS%"=="ENABLED"     setx RD_LEV=VDOS=7


if "%_BUILDNUM%"=="ENABLED" setx BUILDNUM=-DBUILDNUM=1
if "%_SUV%"=="ENABLED" setx SUV=-DRD_SUV=1
if "%_REP%"=="ENABLED" setx REP=-DRD_REP=1


REM  Set the environment for the appropriate configuration
REM
goto %_BUILDMODE%


:RELEASE
setx rd_sw=-1 %BUILDNUM% %SUV% %REP%
setx rd_sw_l=
setx rd_sw_a=/Mx/W2
if "%_SUV%"=="ENABLED" setx rd_sw_a=%rd_sw_a% /DRD_SUV=1
if "%_REP%"=="ENABLED" setx rd_sw_a=%rd_sw_a% /DRD_REP=1
goto SetFAT32


:XT
setx rd_sw=-DXTVER="(XT)"
setx rd_sw_l=
setx rd_sw_a=/Mx/W2
goto SetFAT32


:TD
setx rd_sw=-1 -v
setx rd_sw_l=/v/s
setx rd_sw_a=/Zi/Zd/Mx/W2
goto SetFAT32


:ON
setx rd_sw=-1 -y
setx rd_sw_l=/l /s
setx rd_sw_a=/Zi/Zd/Mx/W2
goto SetFAT32


:OFF
setx rd_sw=-1 -y
setx rd_sw_l=/s
setx rd_sw_a=/Mx/W2
goto SetFAT32


:SetFAT32
if not "%_FAT32%"=="ENABLED" goto Set386Build
setx rd_sw=%rd_sw% -DUSING_FAT32=1
setx rd_sw_a=%rd_sw_a% /DFAT32=1
REM  This variable is used not to enable a FAT32 build but to
REM  enable the FAT32 functionality is some of the utilities.
REM
setx ENABLEF32=-DENABLEF32=1


REM  It is only valid to enable a 386 build when building in FAT32 mode
REM
:Set386Build

if "%CANON_186%"=="TRUE" GOTO done

if not "%_FAT32%"=="ENABLED" set _B386=DISABLED
if not "%_B386%"=="ENABLED" goto DONE
if not exist %BCROOT%\include\_nfile.h set _B386=DISABLED (Not using DEVTOOLS)
if not "%_B386%"=="ENABLED" goto DONE
setx rd_sw=%rd_sw% -3


:done
REM  Display the configuration
REM
echo.
echo    Build Mode: %_BUILDMODE%
echo         FAT32: %_FAT32%
REM echo     386 Build: %_B386%
echo          LFNS: %_LFNS%
echo Build Numbers: %_BUILDNUM%
echo   Single User: %_SUV%
echo   Replication: %_REP%
echo.
REM Store the configuration for later display
                            setx RDBuild=%_BUILDMODE%
if "%_FAT32%"=="ENABLED"    setx RDBuild=%RDBuild% 32
REM if "%_B386%"=="ENABLED"     setx RDBuild=%RDBuild% 386
if "%_LFNS%"=="ENABLED"     setx RDBuild=%RDBuild% 7
if "%_BUILDNUM%"=="ENABLED" setx RDBuild=%RDBuild% BUILDNUM
if "%_SUV%"=="ENABLED" setx RDBuild=%RDBuild% SUV
if "%_REP%"=="ENABLED" setx RDBuild=%RDBuild% REP


:End
setx noehs=
setx noehm=
setx noehc=
setx noehl=
setx bc5cargs=
setx bc5largs=
setx ssetargv=
setx sargv=
if exist %BCROOT%\lib\noehc.lib setx sargv=TRUE
if exist %BCROOT%\lib\noehc.lib setx NOEHS=noehs.lib
if exist %BCROOT%\lib\noehc.lib setx NOEHC=noehc.lib
if exist %BCROOT%\lib\noehc.lib setx NOEHM=noehm.lib
if exist %BCROOT%\lib\noehc.lib setx NOEHL=noehl.lib
if exist %BCROOT%\lib\noehc.lib setx BC5CARGS=-n. -w -f- -x-
if exist %BCROOT%\lib\noehc.lib setx BC5LARGS=/P /n
:Exit
set _OPTION=
set _BUILDMODE=
set _FAT32=
set _B386=
set _LFNS=
set _BUILDNUM=
set _SUV=
set _REP=
