//
//   $Id: DOSSUPP.C 1.3 2002/10/16 21:35:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSCVT.C  created 11/29/90  by Drew Gislason

   Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains routines needed to convert one data
	structure to another.  None of these routines talk to
	the outside world directly; that is no disk access, no
	memory allocation, etc.

**********************************************************/
#define __FILENUM__ 37

#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

/*
	-------------------------------------------------------

					Miscellaneous Support Routines

	-------------------------------------------------------
*/

/*
	Forward string search.
*/
char	* PASCAL strchr(char *s,char c)
{
	ENTER_FUNCTION();

	// validate pointer
	ASSERT(s);

	while(*s)
	{
		if(*s == c)
		{
			EXIT_FUNCTION();
			return s;
		}
		++s;
	}

	EXIT_FUNCTION();

	return NULL;
}


/*
	Reverse string search.
*/
char	* PASCAL strrchr(register char *p,char c)
{
	char *pstart = p;

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(p);

	p += strlen(p) - 1;		/* point to just before terminating 0	*/
	while (p >= pstart)
	{
		if(*p == c)
		{
			EXIT_FUNCTION();
			return p;
		}
		--p;
	}

	EXIT_FUNCTION();

	return NULL;
}


/*

	Skip the whitespace of a string.
*/
char	* PASCAL skipwhite(char *s)
{
	ENTER_FUNCTION();

	// validate pointer
	ASSERT(s);

	while(*s && *s<=' ')
		++s;

	EXIT_FUNCTION();

	return s;
}

/*
	Skip the whitespace of a far string.
*/
char far * PASCAL fskipwhite(char far *s)
{
	ENTER_FUNCTION();

	// validate pointer
	ASSERT(s);

	while(*s && *s<=' ')
		++s;

	EXIT_FUNCTION();

	return s;
}


/*
	Convert a string to an integer (unsigned actually)
*/
int	PASCAL fatoi(char far *s)
{
	int i=0;

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(s);

	while(isdigit(*s))
	{
		i = i*10 + (*s-'0');
		++s;
	}

	EXIT_FUNCTION();

	return i;
}



/*
	Convert a string to an integer (unsigned actually)
*/
int	PASCAL atoi(char *s)
{
	int i=0;

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(s);

	while(isdigit(*s))
	{
		i = i*10 + (*s-'0');
		++s;
	}

	EXIT_FUNCTION();

	return i;
}

/*
	Change a string to uppercase
*/
void	PASCAL strupr(char *p)
{
	ENTER_FUNCTION();

	// validate pointer
	ASSERT(p);

	uppercase((void far *)p,-1);

	EXIT_FUNCTION();
}

/*
	Change a far string to uppercase
*/
void	PASCAL fstrupr(char far *p)
{
	ENTER_FUNCTION();

	// validate pointer
	ASSERT(p);

	uppercase(p,-1);

	EXIT_FUNCTION();
}

/*
	Concatinate strings.
*/
void PASCAL strcat(char *s1, char *s2)
{
	ENTER_FUNCTION();

	// validate parameters
	ASSERT(s1);
	ASSERT(s2);

	s1 += strlen(s1);
	memmove(s1, s2, strlen(s2)+1);

	EXIT_FUNCTION();
}

/*
	Run Only On Bootup:

	Used only internally, allocate a number of bytes from
	the memory pool.  Also, zero the allocated memory.
	"bytes" may be 0 to create a 0 length MCB.

	Returns the segment of this memory.  If the memory
	cannot be allocated, this halts the machine with an
	error message.
*/
unsigned	PASCAL mem_z_alloc(unsigned bytes, char id)
{
	extern int error_no;
	unsigned seg,paras;

	/* convert bytes to paragraphs */
	paras = bytes/16 + ((bytes & 0xF)?1:0);

	/* allocate the memory */
	error_no = 0;
	seg = mem_alloc(paras);
	PERMASSERT(!error_no);

	/* zero the memory */
	fmemset(MK_FP(seg,0),0,paras*16);

	/* assign the ID */
	((struct MCB far *)MK_FP(seg-1, 0))->init_id = id;

	return seg;
}

/*
	This function is called if an assert fails.  It does not
	actually return at all, but must have a return type so
	that the assert expression can be resolved.

	Does not return.
*/
void	PASCAL krn_assert(unsigned filenum, unsigned linenum)
{
   printf( "KRN ASSERT file# %u line# %u\n", filenum, linenum );
}
