//
//   $Id: CHKCPYRT.C 1.4 2002/10/15 19:58:14Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  CHKCPYRT.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//
//  This file checks to make sure that the copyright information is
//  present in the supplied text file.  It checks to make sure that
//  the romdos revision and version reflect the current information
//  as well as that the current year is in the copyright.
//
//  This program returns a non-zero error code if there are any
//  problems.
//
//  The purpose of this utility was to help keep all the version,
//  revision, and copyright information present and current.
//
//  USAGE:
//
//  CHKCPYRT <filename>
//
//  <filename> is a file (text format) which contains copyright
//             information.  Such as a usage screen.
//
//  HISTORY:
//
//  07-07-99  BT  Created
//


//  Borland libraries
//
#include <stdio.h>
#include <mem.h>
#include <string.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"
#include "fatworld.h"
#include "dosver.h"


//  Prototypes
//
#include "chkcpyrt.pt"


//  Constants for return values (error codes)
//
#define NO_ERR				0
#define OPEN_ERR			1
#define READ_ERR			2
#define CLOSE_ERR			3
#define COPYRIGHT_ERR	4
#define YEAR_ERR			5
#define VERSION_ERR		6
#define REVISION_ERR		7

//  Constant for max number of bytes to search for a copyright
//
#define BUFFER_SIZE		2048

bool fIsFat32World;


//  Start of the program
//
uchar main( int argc, char * argv[] )
{
	uchar ucReturnValue;



	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	//  Display the sign-on information
	//
	DisplaySignon();

	//  Check if there are two many or too few arguments
	//
	if( argc != 2 )
	{
		//  Display the user usage screen
		//
		DisplayUsage();
	}

	//  Check that a copyright is present, that it contains the
	//  current year, and the appropiate version and revision
	//  information.
	//
	ucReturnValue = CheckInformation( argv[1] );

	//  Set the appropriate error code and exit
	//
	return ucReturnValue;
}


//  Display the sign-on information
//
void DisplaySignon( void )
{
	char szSignOn[] = "CHKCPYRT %s %s\n%s\n\n";


	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

//  Display the user usage screen
//
void DisplayUsage( void )
{
	char szUsage[] =
	"CHKCPYRT <filename>\n"
	"\n"
	"  <filename> is a file (text format) which contains copyright\n"
	"             information.  Such as a usage screen.\n";


	printf( szUsage );
}

//  Check that a copyright is present, that it contains the
//  current year, and the appropiate version and revision
//  information.
//
uchar CheckInformation( char * szFileName )
{
	static char rgucBuffer[BUFFER_SIZE];
	unsigned uOpenMode = O_RDONLY;
	unsigned uAttributes = A_NORMAL;
	unsigned uAction = ACT_EXIST_OPEN;
	unsigned uByteCount;
	int iHandle;
	struct DOSDate Date;
	bool fSuccessful;
	char szYear[5];


	//  Open the file
	//
	fSuccessful = dlOpenOrCreateFile( uOpenMode, uAttributes,
			     &uAction, szFileName, &iHandle);
	if( !fSuccessful )
	{
		//  If there is an error we cannot proceed.
		//
		return OPEN_ERR;
	}

	//  Read the first 2K of the file
	//
	uByteCount = BUFFER_SIZE;
	fSuccessful = dlReadFile( iHandle, &uByteCount, rgucBuffer );
	if( !fSuccessful )
	{
		//  If there is an error we cannot proceed.
		//
		dlCloseFile(iHandle);
		return READ_ERR;
	}

	//  Close the file
	//
	fSuccessful = dlCloseFile(iHandle);
	if( !fSuccessful )
	{
		//  If there is an error we cannot proceed.
		//
		return CLOSE_ERR;
	}

	//  Make sure that there is a copyright
	//
	fSuccessful = SearchBuffer( rgucBuffer, uByteCount, "Copyright" );
	if(!fSuccessful )
	{
		return COPYRIGHT_ERR;
	}
	fSuccessful = SearchBuffer( rgucBuffer, uByteCount, "Datalight, Inc." );
	if(!fSuccessful )
	{
		return COPYRIGHT_ERR;
	}

	//  Make sure that the year in the copyright is current
	//
	fSuccessful = dlGetDOSDate( &Date );
	if( !fSuccessful )
	{
		return YEAR_ERR;
	}
	sprintf( szYear, "%u", Date.year );
	fSuccessful = SearchBuffer( rgucBuffer, uByteCount, szYear );
	if( !fSuccessful )
	{
		return YEAR_ERR;
	}

	//  Make sure that the correct version is being shown
	//
	if( fIsFat32World )
	{
		fSuccessful = SearchBuffer( rgucBuffer, uByteCount, SHORTVERSION );
	}
	else
	{
		fSuccessful = SearchBuffer( rgucBuffer, uByteCount, SHORTVERSION622 );
	}
	if( !fSuccessful )
	{
		return VERSION_ERR;
	}

	//  Make sure that the correct revision is being shown
	//
	fSuccessful = SearchBuffer( rgucBuffer, uByteCount, REVISION );
	if( !fSuccessful )
	{
		return REVISION_ERR;
	}

	return NO_ERR;
}

//  Search for a string in a fixed sized buffer.  Return TRUE
//  if the string was found
//
bool SearchBuffer( char * rgucBuffer, unsigned uBufferSize,
		char * szString )
{
	bool fDone = FALSE;
	unsigned uStringSize;
	unsigned uBufferIndex = 0;
	int iReturnValue;


	uStringSize = strlen( szString );
	while( !fDone )
	{
		if( uStringSize + uBufferIndex > uBufferSize )
		{
			fDone = TRUE;
		}
		else
		{
			iReturnValue = memcmp( &rgucBuffer[uBufferIndex], szString,
					uStringSize );
			if( iReturnValue == 0 )
			{
				return TRUE;
			}
			else
			{
				uBufferIndex++;
			}
		}
	}

	return FALSE;
}
