//
//   $Id: DV.CPP 1.3 2002/10/15 20:06:44Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	DV.CPP  created  09/28/91

   Copyright (c) 1991-2001 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

*********************************************************/
#include <dos.h>
#include <stdlib.h>
#include <string.h>

#if defined(__BORLANDC__)
	#if(__BORLANDC__ < 0x500)
typedef int bool;
	#endif
#endif

#include "dv.h"
#include "dvkey.h"
#include "getkey.h"

#define MODETEXT 0x0003
#define MODEMONO 0x0007

#define SETCURTYPE 0x0100
#define SETCURSOR 0x0200
#define READCURSOR 0x0300
#define GETMODE 0x0F00
#define EGACHECK 0x1130

#define VIDINT 0x10

unsigned	dv_base		= DV_COLORBASE;
unsigned	dv_row;
unsigned	dv_col;
unsigned	dv_rows;		/* current # of screen rows */
unsigned	dv_cols;		/* current # of screen cols */
unsigned dv_ega;
unsigned	dv_mode;
unsigned dv_curattr	= 7;
unsigned dv_curtype;

static	unsigned	dv_buffer[DV_MAXCOL];
static	union REGS r;

// static functions
static void	video_io(unsigned ax,unsigned bx);

/*
	dv_open

	Initialize the direct-video display package.  May be called repeatedly,
	but should generally only be called once.  Sets the following:

	dv_ega		1=EGA active, 0=EGA not active
	dv_rows		current number of rows
	dv_cols		current number of columns
	dv_row		current cursor row
	dv_col		current cursor column
	dv_mode		current video mode
	dv_base		segment of video RAM
	dv_curtype	initial cursor type
*/
void	dv_open(void)
{
	r.x.dx = 0;							/* determine if EGA active */
	video_io(EGACHECK,0);
	if(r.h.dl == 0)
	{
		dv_ega = 0;
		dv_rows = 25;
	}
	else
	{
		dv_ega = 1;
		dv_rows = r.x.dx + 1;	/* r is global */
	}

	video_io(GETMODE,0);			/* check current mode */
	dv_cols = r.h.ah;
	dv_mode = r.h.al;
	if(dv_mode == MODEMONO)
		dv_base = DV_MONOBASE;
	else
		dv_base = DV_COLORBASE;

	video_io(READCURSOR,0);		/* get current cursor position and type */
	dv_row = r.h.dh;
	dv_col = r.h.dl;
	dv_curtype = r.x.cx;
}

/*
	dv_goto

	Move the hardware and virtual cursors to a specific row and column.
	If out of range, will force to be within the legal screen area.
*/
void	dv_goto(unsigned row,unsigned col)
{
	dv_move(row,col);
	dv_flush();
}

void	dv_getcursor(int *row, int *col)
{
	video_io(READCURSOR,0);		/* get current cursor position and type */
	*row = r.h.dh;
	*col = r.h.dl;
}

/*
	dv_move

	Move the virtual cursor to a specific row and column (both zero-based).
	If out of range, will force to be within the legal screen area.
*/
void	dv_move(unsigned row,unsigned col)
{
	dv_row = row % dv_rows;
	dv_col = col % dv_cols;
}

/*
	dv_attr

	Set the current display attribute.  This is used for subsequent dv_putc,
	dv_puts, and dv_textputs calls.  It is also used to fill areas cleared
	by dv_eeol, dv_blank, dv_cls, dv_scrollup and dv_scrolldn.
*/
void	dv_attr(int attr)
{
	dv_curattr = attr;
}

/*
	dv_putc

	Display a single character on the screen, at the virtual cursor position.
	Newline (\n) does perform a CR/LF motion, and scrolls the screen if
	necessary.  Every other character is simply placed on the screen.
*/
void	dv_putc(unsigned char c)
{
	unsigned char s[2];

	s[0] = c;
	s[1] = '\0';
	dv_textputs(s);
}

/*
	dv_textputs

	Displays a null-terminated string on the screen, at the virtual cursor
	position.  Newlines perform CR/LF motion.  All other characters are
	placed on the screen.  If the string extends past the right edge of the
	screen, it will be truncated at that point.
*/
void	dv_textputs(unsigned char *s)
{
	char *p;

	p = strchr((char *)s,'\n');
	while(p)
		{
		*p = '\0';
		dv_puts(s);
		if(dv_row < dv_rows-1)
			dv_move(dv_row+1,0);
		else
			{
			dv_scrollup(0,dv_rows,1);		/* xxx */
			dv_move(dv_row,0);
			}
		s = (unsigned char *)p+1;
		p = strchr((char *)s,'\n');
		}
	dv_puts(s);
}

/*
	dv_putsn

	Display the first "n" characters of a null-terminated string at the
	virtual cursor position.  No translations are performed.  If the string
	extends beyond the right edge of the screen, it is truncated at that
	point.
*/
void	dv_putsn(unsigned char *s,int len)
{
	int n,col;

	n = 0;
	col = dv_col;
	while(*s && col < dv_cols && n < len)
		{
		dv_buffer[n++] = (unsigned)(*s++) + (dv_curattr << 8);
		++col;
		}
	dv_write(dv_row,dv_col,n,dv_buffer);
	dv_move(dv_row,dv_col+n);
}

/*
	dv_puts

	Displays a null-terminated string at the virtual cursor position.
	No translations are performed.  If the string extends beyond the
	right edge of the screen, it is truncated at that point.
*/
void	dv_puts(unsigned char *s)
{
	dv_putsn(s,DV_MAXCOL);
}

/*
	dv_cls

	Clears the screen, and moves the virtual cursor to the upper left corner.
	The entire screen is set to the current attribute.
*/
void	dv_cls(void)
{
	dv_clear(0,dv_rows);
	dv_move(0,0);
}

/*
	dv_clear

	Clears one or more screen rows, setting them to the current attribute.
*/
void	dv_clear(int top,int numrows)
{
	int oldrow,oldcol;

	oldrow = dv_row;
	oldcol = dv_col;
	while(numrows--)
		{
		dv_move(top,0);
		dv_eeol();
		++top;
		}
	dv_move(oldrow,oldcol);
}

/*
	dv_blank

	Clears a portion of one screen line, setting it to the current attribute.
	Beginning at the virtual cursor position, "len" columns are blanked.
*/
void	dv_blank(int len)
{
	int n;
	unsigned word,*ip;

	word = ' ' + (dv_curattr << 8);
	for(ip=dv_buffer,n=0;n<len;++ip,++n)
		*ip = word;
	dv_write(dv_row,dv_col,len,dv_buffer);
}

/*
	dv_eeol

	Erase to end of the line, beginning at the virtual cursor position.
	This area is set to the current attribute.
*/
void	dv_eeol(void)
{
	dv_blank(dv_cols-dv_col);
}

/*
	dv_gettext

	Direct replacement for the Turbo C gettext function.  Recommend using
	dv_getblock instead.
*/
void	dv_gettext(int left,int top,int right,int bottom,unsigned char *buffer)
{
	dv_getblock((unsigned *)buffer,top-1,bottom-top+1,left-1,right-left+1);
}

/*
	dv_getblock

	Read a rectangle from the screen into a buffer.  Characters and
	attributes are read from the screen into a passed buffer.
*/
void	dv_getblock(unsigned *buffer,int top,int numr,int left,int numc)
{
	while(numr--)
		{
		movedata(dv_base,(top*dv_cols+left)*2,FP_SEG(buffer),FP_OFF(buffer),
				numc*2);
		buffer += numc;
		++top;
		}
}

/*
	dv_puttext

	Direct replacement for the Turbo C puttext function.  Recommend using
	dv_putblock instead.
*/
void	dv_puttext(int left,int top,int right,int bottom,unsigned char *buffer)
{
	dv_putblock((unsigned *)buffer,top-1,bottom-top+1,left-1,right-left+1);
}

/*
	dv_putblock

	Write a rectangle to the screen from a buffer.  Characters and
	attributes are written.
*/
void	dv_putblock(unsigned *buffer,int top,int numr,int left,int numc)
{
	while(numr--)
		{
		movedata(FP_SEG(buffer),FP_OFF(buffer),dv_base,(top*dv_cols+left)*2,
				numc*2);
		buffer += numc;
		++top;
		}
}

/*
	dv_write

	Low-level video write function.  Displays a character/attribute filled
	buffer at a particular row and column on the screen.
*/
void	dv_write(int row,int col,int len,unsigned *text)
{
	movedata(FP_SEG(text),FP_OFF(text),dv_base,(row*dv_cols+col)*2,len*2);
}

/*
	dv_scrollup

	Scroll a portion of the screen up numrows lines.  The area affected is
	described by a top line, and a number of lines.
*/
void	dv_scrollup(int top,int wsize,int numrows)
{
	int i,j;

	for(i=top,j=0;j<wsize-numrows;++i,++j)
		movedata(dv_base,(i+1)*dv_cols*2,dv_base,i*dv_cols*2,dv_cols*2);
	dv_clear(top+wsize-numrows,numrows);
}

/*
	dv_scrolldn

	Scroll a portion of the screen up numrows lines.  The area affected is
	described by a top line, and a number of lines.
*/
void	dv_scrolldn(int top,int wsize,int numrows)
{
	int i,j;

	for(i=top+wsize,j=0;j<wsize-numrows;--i,++j)
		movedata(dv_base,(i-1)*dv_cols*2,dv_base,i*dv_cols*2,dv_cols*2);
	dv_clear(top,numrows);
}

/*
	dv_flush

	Flushes any output which has not yet been written to the screen.  Also
	updates the hardware cursor to the virtual cursor position.  Should
	always be called before performing input.
*/
void	dv_flush(void)
{
	r.x.dx = (dv_row << 8) + dv_col;
	video_io(SETCURSOR,0);
}

/*
	dv_close

	Close the direct-video display package.  Need not be called.  Updates
	the hardware cursor to the virtual cursor position, restores the
	original cursor type.
*/
void	dv_close(void)
{
	dv_flush();
	dv_setcurtype(dv_curtype);
}

/*
	dv_morelines

	Enables the 43 line mode under EGA, or the 50 line mode under VGA.
	This function has no effect if the video adapter does not support
	these modes.  Also, will have no effect if the adapter is already
	in one of these modes.
*/
int	dv_morelines(void)
{
	if(!dv_ega || dv_rows != 25)
		return dv_rows;

	dv_flush();
	if(dv_mode != MODEMONO)
	{
		video_io(MODETEXT,0);
		video_io(0x1112,0);				/* characters 8x8 */
	}
	dv_open();
	return dv_rows;
}

/*
	dv_25lines

	Restores the screen to its normal 25 line mode.  This function has
	no effect if the screen is already in 25 line mode.
*/
int	dv_25lines(void)
{
	if(!dv_ega || dv_rows<=25)
		return dv_rows;

	video_io(MODETEXT,0);
	dv_open();
	return dv_rows;
}

/*
	video_io

	For internal use only
*/
static void	video_io(unsigned ax,unsigned bx)
{
	r.x.ax = ax;
	r.x.bx = bx;
	int86(VIDINT,&r,&r);
}

/*
	dv_getc

	Get a character from the keyboard.  This function replaces the getch
	function of Turbo C.
*/
int	dv_getc(void)
{
	return StdinGetKey();
}

/*
	dv_getkey

	Get a "keystroke" from the keyboard.  If a non-ASCII key is pressed,
	converts the scan code into a single-byte value and returns it.
*/
int	dv_getkey(void)
{
	int c;

	do
	{
		c = dv_pollkey();
	} while(c < 0);

	return c;
}

/*
	Returns -1 if no key, returns the key otherwise
*/
int PollDosForKey(void)
{
	unsigned char key;
	#define ZF 0x40		/* zero flag */

	_AH = 6;
	_DL = 0xFF;
	geninterrupt(0x21);
	key = _AL;
	if(_FLAGS & ZF)
		return -1;
	return key;
}

/*
	dv_pollkey

	Poll for a key.  If no key is waiting, return -1, else
	return the key and pull it from the DOS buffer.  This
	one is extended to include pressing and releasing Alt
	as a key (for menus).
*/
int	dv_pollkey(void)
{
	#define ALTSTAT	0x08
	#define CTRLSTAT	0x04
	#define SHIFTSTAT	0x03
	static int alt = 0;
	static unsigned char altstat;
	static unsigned char keybetween;

	/*
		"cursors" and "shiftkeys" are two matching arrays.
		If we get a "cursor" key, with the shift key
		depressed, translate it to a "shiftkey"
	*/
	static unsigned char cursors[] =
	{
		keyIns, '0', keyDel, '.',
		keyLeft, '4', keyRight, '6',
		keyEnd, '1', keyHome, '7',
		keyDown, '2', keyUp, '8',
		keyPgDn, '3', keyPgUp, '9',
		keyCtrlLeft, keyCtrlRight, keyCtrlEnd, keyCtrlHome,
		keyCtrlDown, keyCtrlUp, keyCtrlPgDn, keyCtrlPgUp,
		'5'
	};
	static int shiftkeys[] =
	{
		keyShiftIns, keyShiftIns, keyShiftDel, keyShiftDel,
		keyShiftLeft, keyShiftLeft, keyShiftRight, keyShiftRight,
		keyShiftEnd, keyShiftEnd, keyShiftHome, keyShiftHome,
		keyShiftDown, keyShiftDown, keyShiftUp, keyShiftUp,
		keyShiftPgDn, keyShiftPgDn, keyShiftPgUp, keyShiftPgUp,
		keyShiftCtrlLeft, keyShiftCtrlRight, keyShiftCtrlEnd, keyShiftCtrlHome,
		keyShiftCtrlDown, keyShiftCtrlUp, keyShiftCtrlPgDn, keyShiftCtrlPgUp,
		keyShiftCenter
	};

	int c;
	unsigned char status;
	unsigned char * p;

	// get the shift/alt status of the keyboard
	_AH = 2;
	geninterrupt(0x16);
	status = _AL;

	// get a key from DOS if available
	c = PollDosForKey();
	if(c == 0)
	{
		c = PollDosForKey() + 112;

		// special, backtab conflicts with ctrl-BS
		if(c == 127)
			c = keyBackTab;
	}

	// make Tab different from ^I, CR different from ^M
	if(c < ' ' && (status&CTRLSTAT)==0)
	{
		if(c == keyCtrl('H')) c = keyBS;
		if(c == keyCtrl('I')) c = keyTab;
		if(c == keyCtrl('M')) c = keyCR;
	}

	// translate cursor and shift status into a shift-cursor key
	if((status&SHIFTSTAT) && c>0)
	{
		if((p = (unsigned char *)memchr(cursors, c, sizeof(cursors))) != 0)
			c = shiftkeys[(int)(p-cursors)];
	}

	// we've pressed a key
	if(c >= 0)
		keybetween = 1;

	// we have not pressed a key, see if Alt has been toggled
	else
	{
		altstat = (status & ALTSTAT);

		// if alt has been depressed and released with no intervening key
		if(alt>200 && !altstat && !keybetween)
			c = keyAlt;		// we have an "ALT" key

		// count how long the alt key is depressed
		if(altstat)
			++alt;
		else
			alt = keybetween = 0;
	}

	return c;
}

/*
	dv_setcurtype

	Sets the hardware cursor type.  The call is formatted as:
	dv_setcurtype(TOPROW*256+BOTROW)
*/
void	dv_setcurtype(unsigned type)
{
	r.x.cx = type;
	video_io(SETCURTYPE,0);
}