//
//   $Id: W.CPP 1.3 2002/10/15 20:06:44Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
	Window library for Turbo C
   Copyright (c) 1991-2001 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary
*/

/* Internal code */

#include <assert.h>
#include <mem.h>
#include <alloc.h>
#include <stdlib.h>

#include "dv.h"
#include "w.h"

bool	w_init(void)
{
	assert(heapcheck() == _HEAPOK);
	if(w_buffer)
		return TRUE;

	dv_open();
	w_origrows = dv_rows;

	// allocate a buffer for the screen
	if( (w_buffer = (word*)w_malloc(W_MAXROWS*W_MAXCOLS*WORDSIZE)) == 0 )
		return FALSE;

	// allocate a buffer for the backdrop
	if( (w_backdrop = (WINDOW*)w_open(0,dv_rows,0,dv_cols,W_VARIABLE)) == 0 )
	{
		assert(heapchecknode(w_buffer) == _USEDENTRY);
		w_free(w_buffer);
		w_buffer = 0;
		return FALSE;
	}

	// set up the backdrop
	w_getbackdrop();
	w_goto(dv_row,dv_col);

	assert(heapcheck() == _HEAPOK);
	return TRUE;
}

void * w_open(int top,int rows,int left,int cols,wtype type)
{
	WINDOW *w;

	w_init();	// doesn't do anything if INIT already happened

	if( (w = (WINDOW*)w_malloc(sizeof(WINDOW))) == 0 )
		return 0;

	/* initialize the actual structure here: */
	w->winid = WINID;
	w->type = type;
	w->frame = W_NOFRAME;
	w->shadow = FALSE;
	w->maxrows = (type==W_FIXED) ? rows : W_MAXROWS;
	w->maxcols = (type==W_FIXED) ? cols : W_MAXCOLS;

	w->top = top;
	w->rows = rows;
	w->crow = 0;

	w->left = left;
	w->cols = cols;
	w->ccol = 0;

	w->textattr = w_deftextattr;
	w->borderattr = w_defborderattr;

	w->dirty = TRUE;

	assert(heapcheck() == _HEAPOK);
	if((w->buffer = (word*)w_malloc(w->maxrows * w->maxcols * WORDSIZE)) == 0)
	{
		assert(heapchecknode(w) == _USEDENTRY);
		w_free(w);
		return 0;
	}
	/* end of structure initialization */

	w_add(w);
	w_pick(w);
	w_legalsize();		/* make sure size is legal */
	w_clrall();			/* make sure it's blank to start with */

	assert(heapcheck() == _HEAPOK);
	return w;
}

/* return the # of rows in current video mode (25/43/50/etc) */
int	w_screenrows(void)
{
	extern unsigned dv_rows;
	return dv_rows;
}

/* return the # of cols in current video mode (80/132/ect) */
int	w_screencols(void)
{
	extern unsigned dv_cols;
	return dv_cols;
}

/* return the current video mode */
int w_getmode(void)
{
	return dv_mode;
}

/* change into VGA/EGA text mode (43/50 lines) */
int	w_morelines(void)
{
	dv_morelines();
	WINDOW * old_win;
	old_win = w_current;
	w_pickbackdrop();
	int old_rows = w_current->rows;
	w_setsize(dv_rows,dv_cols);
	w_clrpart(old_rows,dv_rows-old_rows,0,dv_cols);
	w_pick(old_win);
	return dv_rows;
}

/* change into CGA/MONO text mode (25 lines) */
int	w_25lines(void)
{
	dv_25lines();
	WINDOW *old_win = w_current;
	w_pickbackdrop();
	w_setsize(dv_rows,dv_cols);
	w_pick(old_win);
	return dv_rows;
}

/* return the # of rows for this current window */
int	w_getrows(void)
{
	return w_current->rows;
}

/* return the # of cols for this current window */
int	w_getcols(void)
{
	return w_current->cols;
}

void	w_closeall(void)
{
	// the window package has already been closed
	assert(heapcheck() == _HEAPOK);
	if(!w_buffer)
		return;

	if(w_origrows > 25)
		dv_morelines();
	else
		dv_25lines();

	// put the backdrop to the forefront and reset the cursor
	w_pickbackdrop();
	w_front();
	w_updatecursor();
	w_flush();

	// free up the memory for all the windows

	// free up the backdrop and screen buffer
	assert(heapcheck() == _HEAPOK);
	assert(heapchecknode(w_buffer) == _USEDENTRY);
	w_free(w_buffer);
	assert(heapchecknode(w_backdrop) == _USEDENTRY);
	w_free(w_backdrop);
	w_buffer = 0;
	assert(heapcheck() == _HEAPOK);
}

void	w_close(void *wp)
{
	WINDOW *win = (WINDOW *)wp;

	if(!w_iswin(win))
		return;

	assert(heapcheck() == _HEAPOK);
	assert(heapchecknode(win) == _USEDENTRY);
	if(win->buffer)
	{
		assert(heapchecknode(win->buffer) == _USEDENTRY);
		w_free(win->buffer);
	}
	w_del(win);
	w_free(win);
	assert(heapcheck() == _HEAPOK);
}

void	w_pick(void *wp)
{
	WINDOW *win = (WINDOW *)wp;

	if(w_iswin(win))
		w_current = win;
	w_legalsize();
}

void	w_pickbackdrop(void)
{
	w_current = w_backdrop;
}

/*
	Get the current screen into the backdrop
*/
void	w_getbackdrop(void)
{
	dv_getblock(w_backdrop->buffer,0,dv_rows,0,dv_cols);
}

void	*w_getpicked(void)
{
	return w_current;
}

void	w_setsize(int rows,int cols)
{
	if(w_current->type == W_FIXED)
		return;
	w_current->rows = max(1,min(w_current->maxrows,rows));
	w_current->cols = max(1,min(w_current->maxcols,cols));
	w_legalsize();
	w_current->dirty = TRUE;
}

static	void	w_legalsize(void)
{
	if(w_current->top >= dv_rows)
		w_current->top = 0;
	if(w_current->left >= dv_cols)
		w_current->left = 0;
	w_current->rows = min(w_current->rows,dv_rows-w_current->top);
	w_current->cols = min(w_current->cols,dv_cols-w_current->left);
}

void	w_move(int top,int left)
{
	w_current->top = top;
	w_current->left = left;
	w_legalsize();
	w_backdrop->dirty = TRUE;
}

void	w_front(void)
{
	w_del(w_current);
	w_add(w_current);
}

/*
	Update the visible screen to match the current state
	of the windows.
*/
void	w_flush(void)
{
	WINDOW *win;

	if(w_backdrop->dirty)
	{
		w_stuff(w_buffer,w_backdrop);
		w_backdrop->dirty = FALSE;
	}

	win = w_list;
	while(win)
	{
		w_stuff(w_buffer,win);
		win->dirty = FALSE;
		win = win->next;
	}
	dv_putblock(w_buffer,0,dv_rows,0,dv_cols);
}

/*
	Stuff a single window onto the visible screen buffer.
	Presumably, the visible screen buffer will then be
	displayed to the user.

	This function takes care of window frames and shadows.
*/
static	void	w_stuff(word *buffer, WINDOW *win)
{
	register int i;
	register int j;
	int pRow, lRow;	// physical and logical rows
	int pCol;			// physical columns

	// stuff in the top of the frame
	if(win->frame && win->top)
	{
		i = (win->top-1) * W_MAXCOLS + win->left;
		if(win->left)
			buffer[i-1] = win->upperLeft;
		for(j=0; j<win->cols; ++j)
			buffer[i++] = win->horizontal;
		if(win->left + win->cols + 1 < W_MAXCOLS)
			buffer[i] = win->upperRight;
	}

	for(pRow=win->top,lRow=0; lRow<win->rows && pRow<dv_rows; ++pRow,++lRow)
	{
		// put the left part of the frame on the window
		if(win->frame)
		{
			i = win->left - 1;
			if(i >= 0 && i < W_MAXCOLS)
				buffer[pRow * W_MAXCOLS + i] = win->vertical;
		}

		// put the window itself
		memcpy(	buffer + (pRow * W_MAXCOLS + win->left),
					win->buffer + lRow * win->maxcols,
					win->cols * WORDSIZE);

		// put the right part of the frame and possible shadow
		pCol = win->left + win->cols;
		if(win->frame && (pCol < W_MAXCOLS))
		{
			buffer[pRow * W_MAXCOLS + pCol] = win->vertical;

			// place the shadow if neccessary
			if(win->shadow && (pCol+1 < W_MAXCOLS))
			{
				j = pRow * W_MAXCOLS + pCol + 1;
				buffer[j] = mkword(buffer[j],GRAY);
				if(pCol+2 < W_MAXCOLS)
					buffer[j+1] = mkword(buffer[j+1],GRAY);
			}
		}
	}

	// stuff in the bottom of the frame
	if(win->frame && (pRow < dv_rows))
	{
		i = pRow * W_MAXCOLS + win->left;
		if(win->left)
			buffer[i-1] = win->lowerLeft;
		for(j=0; j<win->cols; ++j)
			buffer[i++] = win->horizontal;
		pCol = win->left + win->cols;
		if(pCol + 1 < W_MAXCOLS)
			buffer[i++] = win->lowerRight;

		// place the shadow if neccessary
		if(win->shadow)
		{
			if(pCol + 1 < W_MAXCOLS)
				buffer[i] = mkword(buffer[i],GRAY);
			if(pCol + 2 < W_MAXCOLS)
				buffer[i+1] = mkword(buffer[i+1],GRAY);
			if(pRow + 1 < dv_rows)
			{
				pCol = win->left + 1;
				i = (pRow+1) * W_MAXCOLS + pCol;
				for(j=0; j<win->cols+2 && pCol<W_MAXCOLS; ++j, ++pCol)
				{
					buffer[i] = mkword(buffer[i],GRAY);
					++i;
				}
			}
		}

	}
}

void	w_goto(int row,int col)
{
	w_current->crow = max(0,min(row,w_current->maxrows));
	w_current->ccol = max(0,min(col,w_current->maxcols));
}

int	w_getrow(void)
{
	return w_current->crow;
}

int	w_getcol(void)
{
	return w_current->ccol;
}

void	w_updatecursor(void)
{
	dv_goto(w_current->top+w_current->crow,
				w_current->left+w_current->ccol);
}

void	w_puts(char *s)
{
	while(*s)
		w_putc(*s++);
	w_current->dirty = TRUE;
}

void	w_putc(char c)
{
	word *p;

	if(w_current->ccol >= w_current->cols)
		return;

	p = w_current->buffer +
			w_current->crow*w_current->maxcols +
			w_current->ccol;
	*p++ = mkword(c,w_current->textattr);
	++w_current->ccol;

	w_current->dirty = TRUE;
}

void	w_putblock(void *win,void *buffer,int top,int rows,int left,int cols)
{
	int i,j;

	for(i = top,j=0;j < rows;++i,++j)
	{
		memcpy(	((WINDOW *)win)->buffer + i*((WINDOW *)win)->maxcols + left,
					(word *)buffer + j*cols,
					cols * WORDSIZE);
	}
}

void	w_clrall(void)
{
	w_clrpart(0,w_current->maxrows,0,w_current->maxcols);
}

void	w_clrpart(int top,int rows,int left,int cols)
{
	w_fillpart(top,rows,left,cols,' ');
}

void	w_clreol(void)
{
	// if past edge of window, don't clear anything
	if (w_current->ccol < w_current->cols)
		w_clrpart(	w_current->crow,1,
					w_current->ccol,(w_current->maxcols-w_current->ccol));
}

int	w_getattr(void)
{
	return w_current->textattr;
}

int	w_setattr(int newattr)
{
	w_current->textattr = newattr;
	return 0;
}

int	w_getborderattr(void)
{
	return w_current->borderattr;
//	return 0;
}

int	w_setborderattr(int newattr)
{
	w_current->borderattr = newattr;
	w_setframe(w_current->frame);
	return 0;
}

void w_setshadow(bool set)
{
	w_current->shadow = set;
	w_current->dirty = TRUE;
}

int	w_setframe(wframe frame)
{
	w_current->frame = frame;

	// set up the frame type (if any)
	if(frame)
	{
		int attr = w_current->borderattr;
		w_current->vertical	= mkword(frame == W_SFRAME ? VT1 : VT2,attr);
		w_current->horizontal= mkword(frame == W_SFRAME ? HZ1 : HZ2,attr);
		w_current->upperLeft	= mkword(frame == W_SFRAME ? UL1 : UL2,attr);
		w_current->upperRight= mkword(frame == W_SFRAME ? UR1 : UR2,attr);
		w_current->lowerLeft	= mkword(frame == W_SFRAME ? LL1 : LL2,attr);
		w_current->lowerRight= mkword(frame == W_SFRAME ? LR1 : LR2,attr);
	}
	w_current->dirty = TRUE;
	return 0;
}

int	w_getkey()
{
	w_flush();
	return dv_getkey();
}

int	w_pollkey()
{
	w_flush();
	return -1;
}

void	w_fillall(char c)
{
	w_fillpart(0,w_current->maxrows,0,w_current->maxcols,c);
}

void	w_fillpart(int top,int rows,int left,int cols,char c)
{
	int nr,nc;
	word *p;

	for(nr=0;nr<rows;++nr,++top)
	{
		p = w_current->buffer + top*w_current->maxcols + left;
		for(nc=0;nc<cols;++nc,++p)
			*p = mkword(c,w_current->textattr);
	}
	w_current->dirty = TRUE;
}

void	w_attrall(int newattr)
{
	w_attrpart(0,w_current->maxrows,0,w_current->maxcols,newattr);
}

void	w_attrpart(int top,int rows,int left,int cols,int newattr)
{
	int nr,nc;
	word *p;

	for(nr=0;nr<rows;++nr,++top)
	{
		p = w_current->buffer + top*w_current->maxcols + left;
		for(nc=0;nc<cols;++nc,++p)
			*p = mkword(charof(*p),newattr);
	}
	w_current->dirty = TRUE;
}

void	w_scrollall(bool down)
{
	w_scrollpart(0,w_current->maxrows,0,w_current->maxcols,down);
}

void	w_scrollpart(int top,int rows,int left,int cols,bool down)
{
	int nr;
	word *dest;

	if(down)
	{
		top += rows-1;
		for(nr=0;nr<rows-1;++nr,--top)
		{
			dest = w_current->buffer + top*w_current->maxcols + left;
			memcpy(dest,dest-w_current->maxcols,cols*WORDSIZE);
		}
	}
	else
	{
		for(nr=0;nr<rows-1;++nr,++top)
		{
			dest = w_current->buffer + top*w_current->maxcols + left;
			memcpy(dest,dest+w_current->maxcols,cols*WORDSIZE);
		}
	}
	w_current->dirty = TRUE;
}

static	void	w_add(WINDOW *item)
{
	WINDOW *win;

	item->next = NULL;
	if(!w_list)
	{
		w_list = item;
		item->prev = NULL;
	}
	else
	{
		win = w_list;
		while(win->next)
			win = win->next;
		win->next = item;
		item->prev = win;
	}
}

static	void	w_del(void *item)
{
	WINDOW *win;

	if(!w_list)
		return;
	win = w_list;
	while(win != item && win->next)
		win = win->next;
	if(win == item)
	{
		if(win == w_list)
			w_list = win->next;
		else
			(win->prev)->next = win->next;

		if(win->next)
			(win->next)->prev = win->prev;
	}
}

// get the position of the physical cursor
void	w_getcursor(int *row, int *col)
{
	dv_getcursor(row,col);
}

void	w_setcursor(int row, int col)
{
	dv_goto(row,col);
}

void	*w_malloc(int size)
{ void *p;

	assert(heapcheck() == _HEAPOK);
	p = malloc(size);
	assert(heapcheck() == _HEAPOK);
	assert((heapchecknode(p) == _USEDENTRY) && (p));
	return p;
}

void	w_free(void *p)
{
	assert(p);
	assert(heapcheck() == _HEAPOK);
	assert(heapchecknode(p) == _USEDENTRY);

	free(p);
	assert(heapcheck() == _HEAPOK);
}

