//
//   $Id: W.H 1.4 2002/10/15 20:06:44Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
	Window library for Turbo C
   Copyright (c) 1995-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary
*/

/* Internal definitions */

#include <stdio.h>
#include "win.h"

#ifdef __cplusplus
#define max(a,b)    (((a) > (b)) ? (a) : (b))
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif

#define WINID 2338
const int W_MAXROWS = 80;
const int W_MAXCOLS = 80;

#define WORDSIZE sizeof(word)

#define charof(i)			(i & 0xff)
#define attrof(i)			(i >> 8)
#define w_iswin(w)		(w->winid == WINID)

typedef struct awindow
{
	int	winid;
	wtype	type;
	wframe frame;
	bool  shadow;
	int	maxrows;
	int	maxcols;

	int	top;			//
	int	rows;			// # of rows (at least 1)
	int	crow;			// cursor row

	int	left;			// left
	int	cols;			// # of columns
	int	ccol;			// cursor column

	int	textattr;
	int	borderattr;

	word	vertical, horizontal;
	word	upperLeft, upperRight;
	word	lowerLeft, lowerRight;

	bool	dirty;

	word	*buffer;

	struct awindow *next;
	struct awindow *prev;
} WINDOW;

/*
	Single Border Characters
*/
#define UL1 218			/* 	Upper Left */
#define UR1 191			/* 	Upper Right */
#define LL1 192			/* 	Lower Left */
#define LR1 217			/* 	Lower Right */
#define HZ1 196			/* 	Horizontal */
#define VT1 179			/* 	Vertical */
#define LM1 195			/* 	Left Middle */
#define RM1 180			/* 	Right Middle */
#define TM1 194			/* 	Top Middle */
#define BM1 193			/* 	Bottom Middle */
#define CE1 197			/* 	Center */

/* Double Border Characters */
#define UL2 201			/* 	Upper Left */
#define UR2 187			/* 	Upper Right */
#define LL2 200			/* 	Lower Left */
#define LR2 188			/* 	Lower Right */
#define HZ2 205			/* 	Horizontal */
#define VT2 186			/* 	Vertical */
#define LM2 204			/* 	Left Middle */
#define RM2 185			/* 	Right Middle */
#define TM2 203			/* 	Top Middle */
#define BM2 202			/* 	Bottom Middle */
#define CE2 206			/* 	Center */

/* Mixed Border Characters 1 */
#define UL3 213			/* 	Upper Left */
#define UR3 184			/* 	Upper Right */
#define LL3 212			/* 	Lower Left */
#define LR3 190			/* 	Lower Right */
#define HZ3 205			/* 	Horizontal */
#define VT3 179			/* 	Vertical */
#define LM3 198			/* 	Left Middle */
#define RM3 181			/* 	Right Middle */
#define TM3 209			/* 	Top Middle */
#define BM3 207			/* 	Bottom Middle */
#define CE3 216			/* 	Center */

/* Mixed Border Characters 2 */
#define UL4 214			/* 	Upper Left */
#define UR4 183			/* 	Upper Right */
#define LL4 211			/* 	Lower Left */
#define LR4 189			/* 	Lower Right */
#define HZ4 196			/* 	Horizontal */
#define VT4 186			/* 	Vertical */
#define LM4 199			/* 	Left Middle */
#define RM4 182			/* 	Right Middle */
#define TM4 210			/* 	Top Middle */
#define BM4 208			/* 	Bottom Middle */
#define CE4 215			/* 	Center */


static	WINDOW	*w_list				= NULL;
static	WINDOW	*w_backdrop			= NULL;
static	WINDOW	*w_current			= NULL;
static	word		*w_buffer			= NULL;
#if 0
static	word		*w_save				= NULL;
#endif
// static	int		w_origrow;
// static	int		w_origcol;
static	int		w_origrows			= 25;
static	int		w_deftextattr		= mkattr(WHITE,BLACK);
static	int		w_defborderattr	= mkattr(WHITE,BLACK);

static	void	w_legalsize(void);
static	void	w_stuff(word *buffer,WINDOW *win);
static	void	w_add(WINDOW *item);
static	void	w_del(void *item);


