//
//   $Id: COMCMD.C 1.26 2002/10/15 20:08:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:COMCMD.C

	Copyright (c) 1989-2002 Datalight, Inc.
	Datalight Confidential and Proprietary.
   All Rights Reserved.

	This module contains the all the internal commands such as DIR,
	ERASE, VER, etc...  They are in alphabetical order.

	$DOCHISTORY
	01/25/2002 pb  Removed the extra blank line that was being displayed
						before the usage screens for internal commands.
						This was done because MS does not display this extra
						line.
	03/11/1999 dg  Updated all functions to LFN versions
	               Moved all non do_xx functions to other modules
	08/20/1998 dg  Added Super-Boot to VER/R
	08/20/1998 dg  Added LFN support to command
	07/14/1998 dg  Added BIOS version to VER/R
	07/08/1998 dg  2 digit year 01-01-00 goes to 2000
   01/28/1998 de  Modified do_date strings to use 4-digit year
   08/20/1997 de  SPR 1361: revised to fix batch file processing bug
	11/10/1988 dg  Created by Drew Gislason
*********************************************************************/
#define __FILENUM__ 3

// ansi includes
#include <dir.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// datalight includes
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dl64.h"
#include "command.h"
#include "comprot.h"
#include "dosver.h"

// file/directory parsing variables. used by any commands that use
// a filename or path. see determine_dir()
char szSplitOldCwd[MAXDIR];				// old-style CWD, can be smaller
char szSplitDrive[MAXDRIVE];				// drive size didn't change
char szSplitDirectory[DL_MAXPATH];		// must be big enough to hold LFN
char szSplitFileName[DL_MAXPATH];		// must be big enough to hold LFN

/*
	BREAK		["ON" | "OFF"]
*/
void PASCAL do_break(char *szArgs)
{
	int set;

	set = INT_setbreak(FALSE,0);
	if(set_on_off(szArgs,pszCmdName,&set))
		INT_setbreak(TRUE,set);
	else
		PrintError(EONOFF);
}

/*
	CALL	[drive:][path]filename [batch-parameters]

	Call a batch file, EXE or COM file.
*/
void PASCAL do_call(char *szArgs)
{
	fIsCall = TRUE;
	strcpy(szCmdLine, szArgs);
	OneCommand();
}

/*
	CD			[drive:][path]
	CHDIR		[drive:][path]

	Change to another directory on the specified drive. Does not change
	current drive, only the directory on the drive specified.
*/
void PASCAL do_chdir(char *szArgs)
{
	char szCurrentDirectory[DL_MAXPATH];
	int  iDrive;
	unsigned uLength;

	/* allow up to 1 parameter */
	if(ParmCount(szArgs) > 1)
	{
		PrintError(ENUMPARM);
		return;
	}
   else if( strchr(szArgs,'/'))
   {
		PrintError(EBADSWITCH);
		return;
   }
	LFNStripArgQuotes(szArgs);

	/* make sure drive is valid */
	if((iDrive = cmdgetdrive(szArgs)) < 0)
		return;

	/* no path part, simply display the path */
	if(*szArgs==0 || strcmp(szArgs+1,":")==0)
	{
		if(SmartGetCurrentDirectory(iDrive+1, szCurrentDirectory))
		{
			uLength = strlen(szCurrentDirectory);
			if(szCurrentDirectory[0] == '\\' )
			{
				memmove( szCurrentDirectory+2, szCurrentDirectory,
						uLength+1 );
			}
			else
			{
				memmove( szCurrentDirectory+3, szCurrentDirectory,
						uLength+1 );
				szCurrentDirectory[2] = '\\';
			}
			szCurrentDirectory[0] = iDrive+'A';
			szCurrentDirectory[1] = ':';
			if(fLFNPresent)
			{
				GetLongName(szCurrentDirectory);
			}
			printf("%s\n",szCurrentDirectory);
		}
	}

	/* attempt to set this directory */
	else
	{
		if(!SmartChangeDirectory(szArgs))
			PrintError(EINVDIR);
	}
}

/*
	CLS
*/
#pragma argsused
void PASCAL do_cls(char *szArgs)
{
	clear_screen();
}

/*
	COPY	<source[+]...> [destination] [/A | /B] [/H] [/V] [/Y]

	Copy one or many files to a destination file or directory. This can
	also concatinate files if the '+' is used.

	/A		Copy the file(s) in ASCII format, stopping on Ctrl-Z
	/B		Copy the file(s) in Binary format
	/H		Copy hidden file(s) as well
	/V		Copy the file(s) with Verify
	/Y		Auto answer "Yes" to overwrite prompt

	Environment variables:
	COPYCMD	allows "default" switches to be specified.

	Examples:
	COPY   src.*         dst     concatinate src.* into dst
	COPY   src.* + src2  dst     concatinate into file dst
	COPY   src.*         dst.*   copy files, while changing the name
	COPY   .             C:\TEMP copy *.* from current directory to c:\temp
*/
void PASCAL do_copy(char *szArgs)
{
	extern char cmdstr_copied[];

	bool	fAscii;					/* /A will stop copying on a ^Z */
	bool	fBinary;					/* /B will copy in binary mode */
	bool	fVerify;					/* /V will verify copies */
	int	fHidden;					/* /H will copy hidden files too */
	bool	fNoPrompt;				/* /Y will no prompt on overwrite */
	int	iAttr;					/* search attribute */
	bool	fConcatinate;			/* concatinate files? */
	int	iArgCount;					/* how many files did we copy? */
	FIND	dp;
	int 	iCopied = 0;
	int   iFileCount = 0;
	bool	bFound;
	char	cLastCharacter;
	char * szNamePart;
	char * szTemp2NamePart;
	bool  fShowName;
//	bool  fShowFileCount = FALSE;


	// initialize all the options
	fVerify = fHidden = FALSE;						// assume no on these options
	fNoPrompt = (c_in_batch || c_one_only);	// assume yes, unless in batch
	fAscii = fBinary = INVALID;					// assume not set (invalid)

	//  Allow no error message to be displayed for invalid
	//  COPYCMD arguments.  Also this will not interprete
	//  conflicting switches in COPYCMD and command line.
	//
#if 0
	// allow for "default" options from COPYCMD environment variable
	stuff_env_var(szArgs, "COPYCMD");

	// extract and set all options
	GetSwitch(szArgs, 'A', &fAscii);
	GetSwitch(szArgs, 'B', &fBinary);
	GetSwitch(szArgs, 'H', &fHidden);
	GetSwitch(szArgs, 'V', &fVerify);
	GetSwitch(szArgs, 'Y', &fNoPrompt);
#else
	// allow for "default" options from COPYCMD environment variable
	stuff_env_var(szTemp3, "COPYCMD");

	// extract and set all options
	GetSwitch(szTemp3, 'A', &fAscii);
	GetSwitch(szTemp3, 'B', &fBinary);
	GetSwitch(szTemp3, 'H', &fHidden);
	GetSwitch(szTemp3, 'V', &fVerify);
	GetSwitch(szTemp3, 'Y', &fNoPrompt);

	// extract and set all options
	GetSwitch(szArgs, 'A', &fAscii);
	GetSwitch(szArgs, 'B', &fBinary);
	GetSwitch(szArgs, 'H', &fHidden);
	GetSwitch(szArgs, 'V', &fVerify);
	GetSwitch(szArgs, 'Y', &fNoPrompt);
#endif
	if(switchesleft(szArgs))
		return;

	// attribute allows hidden or not
	iAttr = (fHidden ? (A_SYSTEM | A_HIDDEN) : A_NORMAL);

	// do we want to prompt on overwrite?
	fCopyOverwritePrompt = !fNoPrompt;

	// binary copy mode takes precedence if both /A/B are specified
	if(fBinary != INVALID)
		fAscii = FALSE;
	else if(fAscii != INVALID)
		fAscii = TRUE;
	// fBinary not used after this point!

	// determine if we are concatinating, and if # args OK
	if(!ProcessCopyArgs(szArgs, &iArgCount, &fConcatinate))
		return;

	// if not overridden by user, default to fAscii for concatinate
	if(fAscii == INVALID)
		fAscii = fConcatinate;

	/* loop around copying files to destination */
	iFileCount = 0;
	while(iArgCount)
	{
		/* get the src filespec */
		GetCopyArg(szTemp, szArgs);

		/* check for invalid drive or directory */
		if(!ValidateDirectory(szTemp))
			break;

		//  We need to determine the exact length of the path
		//  that will be displayed
		//
		szNamePart = NamePart(szTemp);
		*szNamePart = 0;
		cLastCharacter = lastchar(szTemp);
		if( (*szTemp != 0) && (cLastCharacter != ':') &&
				(cLastCharacter != '\\') )
		{
			strcat( szTemp, "\\" );
		}
		strcpy( szTemp2, szTemp );
		szNamePart = NamePart(szTemp);
		szTemp2NamePart = NamePart(szTemp2);

		/* make abspath, name may have wildcards */
		MakePath(szTemp, szSplitFileName);

		//  We only show the file names of copied files if there is
		//  more than one file copied.
		//
		fShowName = CheckWild(szTemp);


		//  We will show the copied file count if the source is wild
		//
//		fShowFileCount = fShowName;


		/* find the first file */
		bFound = SmartFindFirst(iAttr, szTemp, &dp);
		if( !bFound )
		{
			PrintError(ENOFILE);
		}

		/* copy all the file(s) in this filespec */
		while(bFound)
		{
			/* a file was found, use it's name in abspath */
			// We should not expand the path for the user to see.  The path
			// is already validated so use the users line
			//
			*szNamePart = 0;
			MakePath(szTemp, dp.name);
			GetLongName(szTemp);
			szNamePart = NamePart(szTemp);
			strcat(szTemp2, szNamePart);

			// actually copy the file
			// Returns  -1  = some kind of error error
			//           0  = didn't copy the file (0 length)
			//           1  = copied the file
			iCopied = CopyFile(szTemp2, fAscii, fConcatinate, !iFileCount,
					fShowName );
			if(iCopied < 0)		// error!
			{
				//  If we ran out of disk space we still want to show the
				//  number of files copied.
				//
				if( iCopied == -2 )
				{
					//  Because command is fragil, change the iCopied variable
					//  back to the "command global error value" (-1)
					//
					iCopied = -1;
//					fShowFileCount = TRUE;
				}
				break;
			}
			if( !fConcatinate )
			{
		  		iFileCount += iCopied;
			}
			else
			{
				iFileCount = 1;
			}

			// find the next item in the file spec
			bFound = SmartFindNext(&dp);

			//  Get the path ready for the next string
			//
			*szTemp2NamePart = 0;
		}

		// remember to close the search
		SmartFindClose(&dp);

		// there was an error, quit
		if(iCopied < 0)
			break;

		// on to next filespec
		szArgs = NextCopyArg(szArgs);
		--iArgCount;
	}

//  BT 08-18-00
//
//  Don't know why this was placed here.  I disabled it to see
//  what other failures removing this causes in the test suite.
//  Failures are showing up in the test suite now because it is
//  not there.
//
#if 0
	if( (iFileCount) || (fShowFileCount) )
#endif
	{
		// display the # of files copied
		printf(cmdstr_copied, iFileCount);
	}
}

/*
	CTTY	device
*/
void PASCAL do_ctty(char *szArgs)
{
	unsigned char hval;
	int fd;

	/* ptr to a device name (not too long) */
	szArgs = skipequal(szArgs);
	szArgs[15] = '\0';

	/* make sure it's a legal char device */
	fd = CTTYOpen(szArgs, RDWR);
	if(fd <= 0)
	{
		PrintError(EBADDEV);
		return;
	}
	dos_close(fd);

	/* OK, now use it for STDIN, STDOUT, STDERR */
	strcpy(c_ctty, szArgs);
	Rehandle();

	/* for batch files, remember values to restore */
	if(c_in_batch)
	{
		c_bat_redirect[0] = GetHandleVal(0);
		c_bat_redirect[1] = GetHandleVal(1);
	}
}

/*
	DATE	[date]

	Get or set the date.  For USA MM/DD/YYYY, Europe DD/MM/YYYY,
	Japan YYYY/MM/DD.
*/
void PASCAL do_date(char *szArgs)
{
	extern struct rd_country c_country;
	extern char date_id[];

	union REGS regs;
	unsigned year, month, day;
	char str[LINELEN+4];
	char cntry_fmt[20];	/* format of country date (mm-dd-yy) */
	int ret;
	char *date_types[] =
	{
		"mm%sdd%syyyy",
		"dd%smm%syyyy",
		"yyyy%smm%sdd",
	};
	static char sep[4] = "-/";	/* 3 possible separators (plus NULL) */
	int	i;
	char * szTemp;


	/* determine date separator */
	sep[2] = *c_country.dtsep;

	/* determine date format for current country */
	i = c_country.datefmt;
	if(i < 0 || i > 2)
		i = 0;
	sprintf(str,date_types[i], sep+2, sep+2);
	sprintf(cntry_fmt," (%s)",str);

	if(*szArgs)
		strcpy(str+2, szArgs);
	else
	{
		memset( str, 0, LINELEN+4 );
		get_datetime_str(str,date_id, showdate, cntry_fmt);
	}

again:

#if 1

	szTemp = str+2;
	while( ( *szTemp == ' ' ) || ( *szTemp == 0x09 ) )
	{
		szTemp++;
	}

	if( ( *szTemp == 0x0d ) || ( *szTemp == 0x00 ) )
	{
		return;
	}

#else
	if(*(str+2)==0)
		return;
#endif

	/* determine date separator (assume user uses the same one) */
	/* i will be left on 2 if not found (country specific separator) */
	for(i=0; i<2; ++i)
		if(strchr(str+2, sep[i]))
			break;

	/* extract date fields from the string */
	switch(c_country.datefmt)
	{
		case 0:ret=get_field(str+2, 3, sep[i], &month, &day, &year); break;
		case 1:ret=get_field(str+2, 3, sep[i], &day, &month, &year); break;
		case 2:ret=get_field(str+2, 3, sep[i], &year, &month, &day); break;
	}
	if(!ret)
		goto err;

	/* if customer enters 2 digit year, be year 2000 compliant */
	if(year < 100)
	{
		if(year < 80)		/* less than 80? must mean year 2000+ */
			year += 2000;
		else
			year += 1900;	/* greater than 80? must mean year 1900+ */
	}

	/* set the date */
	regs.x.ax = 0x2B00;
	regs.x.cx = year;
	regs.x.dx = (month << 8) | day;
	int86(0x21, (void *)&regs, (void *)&regs);
	if(regs.x.ax & 0xFF)
	{
err:
		PrintMessage(EDATE);		// MSDOS sends this one to STDOUT vs. STDERR
		printf("\n");
		memset( str, 0, LINELEN+4 );
		get_datetime_str(str, date_id, 0, cntry_fmt);
		goto again;
	}
}

/*
	DEL	[drive:][path]filespec [/P]
	ERASE	[drive:][path]filespec [/P]

	The /P option causes DEL to prompt before deleting a file
	and ask the user if the file should be deleted!

	Delete one or many files.
*/
void PASCAL do_del(char *szArgs)
{
	extern char cmdstr_promptdel[];
	FIND	dp;
	bool	fPromptOnDelete = FALSE;
	bool	bFound;
	char	*szName;
	int	iErrorCode;


	// get /P switch if it's there
	GetSwitch(szArgs, 'P', &fPromptOnDelete);

	// should only be 1 parameter
	if(ParmCount(szArgs) != 1)
	{
		PrintError(ENUMPARM);
		return;
	}
	LFNStripArgQuotes(szArgs);

	// split into pieces to see if name is "*.*"
	// spits out error on invalid drives or directories
	if(!ValidateDirectory(szArgs))
		return;

	// create an absolute path
	MakePath(szDirPath, "");

	// user wants to be prompted for each file?
	if(fPromptOnDelete)
	{
		bFound = SmartFindFirst(A_NORMAL, szArgs, &dp);
		if(!bFound)
		{
			PrintError(ENOFILE);
			return;
		}

		while(bFound)
		{
			add_path(szDirPath, dp.name);

			// found a name, convert to longname if possible
			szName = GetLongFileNameOnly(szDirPath);
			printf(cmdstr_promptdel, szName);
			if(GetYesNo())
				DeleteWithErrorMsg(szDirPath);
			del_path(szDirPath);
			bFound = SmartFindNext(&dp);
		}

		// make sure to close the find
		SmartFindClose(&dp);
		return;
	}

	// ask on deleting all files
	if(strcmp(szSplitFileName, "*")==0 || strcmp(szSplitFileName, "*.*")==0)
	{
		PrintMessage(MSG_DEL_ALL_FILES);
		if(!GetYesNo())
			return;
	}

	// delete the file(s)
	MakePath(szDirPath, szSplitFileName);
	DeleteWithErrorMsg(szDirPath);
}

/*
	DIR	[drive:][path][filespec] [/A:?] [/B] [/L] [/O:?] [/P] [/S] [/V] [/W] [/4]

	/A?  Display only files with a matching attribute as show below
	     A Archive   D Directory    H Hidden     R Read Only
	     S System    X Show Attrs   - Prefix for not
	/B   Display in bare format
	/L   Display in lowercase
	/O?  Sort directory listing by item as shown below
	     N Name      E Extension           D Date and time
	     S Size      G Group directories   - Prefix for reverse
	     A Last access date
	/P   Display one screen full and pause
	/S   Display files in subdirectories as well
	/V   Verbose directory information
	/W   Display in wide format without date, time, or size
	/4   Displays year with 4 digits (default)

	Environment variables:
	DIRCMD	allows "default" switches to be specified.
	DIRSIZE	rows [cols] allows odd screen sizes.
*/
void PASCAL do_dir(char *szArgs)
{
	extern char cmdstr_dirlisted[]; 		// Total files listed:
	extern int fDir4DigitDate;
   char * invalidfunction;             //TD special case
   char * szArgCopy;                   //TD special case
	int	drive;


	//  Display 2 digit year by default
	//
	fDir4DigitDate = FALSE;


	// look for switch options in cmdline
	if(!do_dir_args(szArgs))
		return;
	if(switchesleft(szArgs))
		return;

   //
   // Handle a MS special case - dir ".x."
   //  (x is a "redirection" and . is any amount of whitespace)
   // gives a response of: Invalid function
   // Note:
   //  dir ""x gives approprate response (file not found or creation error)
   //  dir x"" ignores the redirection character completely!!!
   //
   invalidfunction=NULL;
   if( (szArgCopy = strchr(szArgs,'"')) != NULL )
   {
      if(   ( (invalidfunction = strchr(szArgCopy,'<')) != NULL )
         || ( (invalidfunction = strchr(szArgCopy,'|')) != NULL )
         || ( (invalidfunction = strchr(szArgCopy,'>')) != NULL ) )
      {
         //
         // "short circuit" will pass value into statement if true
         //
         invalidfunction = strchr(invalidfunction,'"');
      }
   }

	// should be no switches left. if there are, complain
	LFNStripArgQuotes(szArgs);

	//  There could be spaces following the name so get rid
	//  of them
	//
	StripEndingSpaces(szArgs);

	// display volume for all DIR calls but bare
	if(!IsDirBare())
	{
		// display the volume and allow pause
		drive = DisplayVolume(szArgs, FALSE);
      //
      // report error for special case, then return
      //
      if( invalidfunction )
      {
         printf("Invalid function\n");
         return;
      }
      if( drive == ERR )
         return;
	}

	// reset directory total reporting
	DirZeroTotals();

	// get disk freespace and size. must happen before RecurseIntoDirectory()
	// to set up size of cluster, etc..
	if(IsDirBare())
	{
		drive = cmdgetdrive(szArgs);
	}
	DirGetFreeSpace(drive);

	// display the directory (and possibly all its subdirs)
	if(!RecurseIntoDirectory(szArgs))
		return;

	// no more display for bare format
	if(IsDirBare())
		return;

	// no files, nothing to report
	if(!IsDirFiles())
	{
		DirDisplayLF();
		PrintError(ENOFILE);
		if(IsDirVerbose())
			DirDisplayTotalDiskSpace();
		else
			DirDisplayFreeBytes(FALSE);
		return;
	}

	// sum up for user
	DirDisplaySummary();
}

/*
	ECHO ["ON" | "OFF" | <string>]

	Turns echo on and off, displays the value of echo, or
	prints a string to the console.

   Note: Microsoft has the following functionality.

      cmd: ECHO
      rsp: ECHO is on

      cmd: ECHO ON
      rsp: (none)

      cmd: ECHO on lowercase
      rsp: ON

*/
void PASCAL do_echo(char *szArgs)
{
	int c_echoecho;

	c_echoecho = c_echo;
	if(!set_on_off(LFNSkipWhite(szArgs), pszCmdName, &c_echoecho))
		printf("%s\n",szArgs+1);
	if(c_in_batch)
		pCurBCB->batEcho = c_echoecho;
	else
		c_no_batch_echo = c_echoecho;
}


/*
	EXIT

	Do exit if this is not the first COMMAND.EXE
*/
#pragma argsused
void PASCAL do_exit(char *szArgs)
{
	extern int c_no_exit;

	if(!c_no_exit)
	{
		PerformExit();
	}
}

/*
	FOR batch file command

	FOR %v IN (set) DO command

	The "set" may be space or comma delimited
	Examples
		FOR %%f IN (A:,a:,B:,b:,C:,c:) DO if %1.==%%f. goto GOODDST
		FOR %%X IN (*.TXT) DO type %%X
	Longname Example
		FOR %%l IN ("For Data(*)",) DO echo %%l
*/
void PASCAL do_for(char *szArgs)
{
	static char cForVariable;
	static char szForLine[BATCH_LINELEN];
	static char szSetItem[DL_MAXPATH];		// possibly relative path
	static char szForName[DL_MAXPATH];		// name after it's converted

	char *pszInArg,*pszSetArg,*pszDoArg;
	char *psz;
	FIND	dp;
	bool	fIsWild;			// wildcards mean use files
	bool	fWorked;			// did the findfirst/next work?
	bool	fUseSmart;		// are we using the smart findfirst/next?
	bool	prev_c_one_only;
	bool	prev_c_in_batch;


	// determine the 'variable' character
	if(*szArgs != '%' || DL_IS_SPACE(szArgs[1]) || !DL_IS_SPACE(szArgs[2]))
	{
		PrintError(ESYNTAX);
		return;
	}
	cForVariable = szArgs[1];

	// verify the 'IN' argument is OK
	pszInArg = LFNNextArg(szArgs);
	LFNArgCopy(szTemp, pszInArg, 5);
	if(stricmp(szTemp, "IN"))
	{
		PrintError(ESYNTAX);
		return;
	}

	// verify that the set is OK
	pszSetArg = LFNNextArg(pszInArg);
	if(*pszSetArg != '(')
	{
		PrintError(ESYNTAX);
		return;
	}

	// copy the set and cmdline into our own string (so szCmdLine can
	// be used for the DO command). We know it won't overflow.
	strcpy(szForLine, pszSetArg+1);
	pszSetArg = szForLine;

	// find the end of the set
	pszDoArg = pszSetArg;
	while(*pszDoArg != ')')
	{
		pszDoArg = LFNEndForArg(pszDoArg);
		pszDoArg = LFNSkipSpecialSpaces(pszDoArg, ",");

		// hit end of string before end of set!
		if(*pszDoArg == 0)
		{
			PrintError(ESYNTAX);
			return;
		}
	}
	*pszDoArg = 0;		// null terminate the set

	// verify the DO argument is OK
	pszDoArg = LFNSkipWhite(pszDoArg+1);
	LFNArgCopy(szTemp, pszDoArg, 5);
	if(stricmp(szTemp, "DO"))
	{
		PrintError(ESYNTAX);
		return;
	}

	// move on to the actual command
	pszDoArg = LFNNextArg(pszDoArg);

	// now, loop through entire set
	fIsWild = FALSE;
	while(1)
	{
		// not currently in a findfirst/next
		if(!fIsWild)
		{
			// all done with the set?
			if(*pszSetArg == 0)
				break;

			// copy the name of this set into it's own space
			psz = LFNEndForArg(pszSetArg);
			LFNArgCopy(szSetItem, pszSetArg, (psz-pszSetArg)+1);

			// move on to next argument
			pszSetArg = LFNNextForArg(pszSetArg);

			// any wildcards? then treat it as a file list
			if(CheckWild(szSetItem))
			{
				// use long or short name depending on LFNFOR setting
				fUseSmart = c_lfnfor;
				if(fUseSmart)
					fWorked = SmartFindFirst(A_NORMAL, szSetItem, &dp);
				else
				{
					dlSetDTA(&dp);
					fWorked = dlFindFirstFile(A_NORMAL, szSetItem);
				}


				// if no matching filespecs, don't process this item in the set
				if(!fWorked)
					continue;

				// in a findfirst/next block
				fIsWild = TRUE;
			}
		}

		// get the string for this one set item into
		strcpy(szForName, szSetItem);

		// may manipulate filename
		if(fIsWild)
		{
			// convert the file mask to the actual found filename
			// by appending name to the path as typed by the user
			psz = NamePart(szForName);
			strcpy(psz, dp.name);

			// copy the long name to the path as typed by the user
			if(fUseSmart)
			{
				strcpy(szTemp, szForName);
				if(GetLongName(szTemp))
					strcpy(psz, NamePart(szTemp));
			}
		}

		// expand the cmdline into the FOR format
		strcpy(szCmdLine, pszDoArg);
		ForExpand(szCmdLine, szForName, cForVariable);

		// display the system prompt and command
		TempCloseRedirection();
		prompt_string();
		if(!c_in_batch || c_echo)
			printf("%s\n", szCmdLine);
		TempOpenRedirection();

		// perform the command
		prev_c_in_batch = c_in_batch;
		prev_c_one_only = c_one_only;
		c_one_only = TRUE;
		c_in_batch = FALSE;
		c_in_batch_for = TRUE;
		do{
			OneCommand();
		}while(GetCommand(szCmdLine));
		c_one_only = prev_c_one_only;
		c_in_batch = prev_c_in_batch;

		// if we did the command, add in an extra line
		if(!c_in_batch || c_echo)
		{
			TempCloseRedirection();
			printf("\n");
			TempOpenRedirection();
		}

		// filespec, do a findnext
		if(fIsWild)
		{
			// any files left for findnext?
			if(fUseSmart)
				fWorked = SmartFindNext(&dp);
			else
			{
				dlSetDTA(&dp);
				fWorked = dlFindNextFile();
			}

			// no more files, close the find
			if(!fWorked)
			{
				fIsWild = FALSE;
				if(fUseSmart)
					SmartFindClose(&dp);
			}
		}

	}	// while loop

	c_in_batch_for = FALSE;
}

/*
	GOTO	label

	This command only works in batch files
*/
void PASCAL do_goto(char *szArgs)
{
	extern unsigned c_resseg;

	int iLen;
	char *szTestLabel, *psz;
	char *szLabel = szArgs;
	int fd;

	// goto at cmdline is ignored
	if( !c_in_batch)
	{
		//  (SHIT ASS CASE!) making command work since the design is
		//  flawed
		//
		if( !c_in_batch_for )
		{
			return;
		}
	}

	/* open the batch file */
	fd = dos_faropen(pCurBCB->batName,RD);
	if(fd < 0)
	{
		PrintError(ENOBATCH);
		EndAllBatches();			// must definately end nested batch processes
		return;
	}

	/* make a copy of the label, we'll be using szCmdLine */
	if(*szLabel == ':')
		++szLabel;
	LFNArgCopy(szTemp, szLabel, DL_MAXPATH);
	szLabel = szTemp;

	/* find the label on a line (if it exists) */
	c_bat_pos = 0;
	while(1)
	{
		/* read in the line */
		dos_seek(fd, c_bat_pos, SEEK_SET);
		iLen = ReadBatchLine(fd, szCmdLine);

		// label not found? must definately end nested batch processes
		if(iLen < 0)
		{
			PrintError(ENOLABEL);
			EndAllBatches();
			break;
		}

		// remove the CR/LF
		c_bat_pos += iLen + 2;

		// is this line a label?
		szTestLabel = LFNSkipWhite(szCmdLine);
		if(*szTestLabel == ':')
		{
			// remove quotes for comparison
			++szTestLabel;
			psz = LFNEndArg(szTestLabel);
			*psz = 0;
			LFNStripArgQuotes(szTestLabel);

			// are they identical? (ignores case)
			if(stricmp(szTestLabel, szLabel) == 0)
				break;
		}
	}

	// done with batch file
	dos_close(fd);
}


/*
	HELP [cmdname]

	Displays the help for an internal command if COMMAND.HLP can be found.
	Displays help for external commands by calling them with /?.
*/
void PASCAL do_help(char *szArgs)
{
	extern char cmdstr_nohelp[];
	extern char cmdstr_badhelp[];

	char *szHelpFile;
	char *szHelpRootName = "\\COMMAND.HLP";
	int	iCmdNum, iOffset, iLen, iHandle;
	int	iStartHelpInfo;
	bool	fFound = FALSE;
	bool  fFirstTime = TRUE;

	/* if no arguments, assume they want help on help */
	if(strlen(szArgs) == 0)
		szArgs = "help";

	/* is this command an internal command? */
	for(iCmdNum=0; iCmdNum < MAX_CMD; ++iCmdNum)
	{
		// found the internal command
		if(stricmp(szArgs, cmds[iCmdNum].cmd) == 0)
		{
			fFound = TRUE;
			break;
		}
	}

	// not an internal command. call external program with /?
	if(!fFound)
	{
		memmove(szCmdLine, szArgs, strlen(szArgs)+1);
		strcat(szCmdLine, " /?");
		MakeCommand();
		if(!ExecuteCommand())
			PrintError(ENOHELP);
		return;
	}

	// open the helpfile in the same directory as COMMAND.COM
	szHelpFile = GetEnvironmentVariable("COMSPEC");
	if(szHelpFile != NULL)
		strcpy(ExtPart(szHelpFile), ".HLP");

	// if no COMSPEC, assume root of current drive
	else
		szHelpFile = szHelpRootName;

	// open COMMAND.HLP
openagain:
	iHandle = dos_open(szHelpFile, RD);
	if(iHandle == -1)
	{
		// try to open in the current directory
		if(fFirstTime)
		{
			szHelpFile = szHelpRootName+1;
			fFirstTime = FALSE;
			goto openagain;
		}

		// no help file!
		printf(cmdstr_nohelp, szHelpFile);
		return;
	}

	// verify this is the right version
	// should start with a string (copyright), followed by a long (helpver)
	szTemp[LINELEN] = 0;
	iOffset = dos_read(iHandle, szTemp, LINELEN-1);
	iLen = strlen(szTemp)+1;
	if(iLen >= iOffset || *(ulong *)(szTemp+iLen) != HELPVER)
	{
		printf(cmdstr_badhelp, szHelpFile);
		return;
	}
	// skip past ID information
	iStartHelpInfo = iLen + sizeof(ulong);
	dos_seek(iHandle, iStartHelpInfo, SEEK_SET);

	// find the entry we are looking for
	dos_read(iHandle, szTemp, MAX_CMD * sizeof(char *));
	iOffset = *(unsigned *)(szTemp+(iCmdNum * sizeof(int)));
	dos_seek(iHandle, iOffset, SEEK_SET);

	// display the entry
	do
	{
		dos_read(iHandle, szTemp, BATCH_LINELEN-1);
		szTemp[BATCH_LINELEN-1] = '\0';
		printf(szTemp);
	} while(strlen(szTemp) == BATCH_LINELEN-1);

	// done, close the handle
	dos_close(iHandle);
}


/*
	IF		["NOT"] "ERRORLEVEL"[=]number | s1==s2 | "EXIST" fname [command]
*/
void PASCAL do_if(char *szArgs)
{
	extern char cmdstr_askprompt[];

	char	c;
	char	szCondition[OLD_NAMESIZE];
	bool	fTrue = FALSE;
	int	n;
	char *p, *szArg2, *p2, *pTemp;
	FIND	dp;

	/* check for the NOT operation */
	LFNArgCopy(szCondition,szArgs,12);
	if(strcmpi(szCondition,"NOT") == 0)
	{
		szArgs  = LFNNextArg(szArgs);
		fTrue = 1;
	}

	/* handle ERRORLEVEL */
	LFNArgCopy(szCondition, szArgs, 12);      // ERRORLEVEL + char + \0
   if( szCondition[10]=='=' )                // if the char is '=', truncate
      szCondition[10]='\0';
	if(strcmpi(szCondition, "ERRORLEVEL") == 0)
	{
		szArgs = LFNNextEqualArg(szArgs);
		n = atoi(szArgs);
		if((unsigned char)c_errorlevel >= (unsigned char) n)
			fTrue = !fTrue;
		szArgs = LFNNextArg(szArgs);
	}

	/* handle EXIST */
	else if(strcmpi(szCondition, "EXIST") == 0)
	{
		szArgs = LFNNextArg(szArgs);
		LFNArgCopy(szTemp, szArgs, DL_MAXPATH);
		szArgs = LFNNextArg(szArgs);

		/* does the file exist */
		if(SmartFindFirst(A_ALLFILES, szTemp, &dp))
		{
			fTrue = !fTrue;
			SmartFindClose(&dp);
		}
	}

	/* string1==string2 */
	else
	{

//  This was the old code which did not work (wasn't compatible)
//
#if 0
		p = LFNEndEqualArg(szArgs);

		// must have 2 equals '=='
		szArg2 = LFNSkipWhite(p);
		if(*szArg2 != '=' || *(szArg2+1) != '=')
		{
			PrintError(ESYNTAX);
			fTrue = FALSE;
		}

		// compare the 2 strings
		else
		{
			*p = '\0';
			szArg2 = LFNSkipWhite(szArg2+2);
			p = LFNEndArg(szArg2);
			p2 = LFNNextArg(szArg2);
			*p = '\0';
			if(strcmp(szArgs,szArg2)==0)
				fTrue = !fTrue;
			szArgs = p2;
		}
#endif // if 0

		//  Find the '==' in the statement
		//
		p = szArgs;
		while( (*p != '=') && (*p != 0) )
		{
			p++;
		}

		if( (*p != '=') || (*(p+1) != '=') || (p == szArgs) )
		{
			PrintError(ESYNTAX);
			fTrue = FALSE;
		}
		else
		{
			//  Make an absolute end to string1
			//
			*p = 0;


			//  This will handle the following case for LFN (windows)
			//  compatibility.
			//
			//  if THIS IS==THIS echo TRUE  (true)
			//
			{
				pTemp = szArgs;
				while( !DL_IS_SPACE( *pTemp ) && (*pTemp != 0) )
				{
					pTemp++;
				}
				*pTemp = 0;
			}


			//  Find the start of string2
			//
			p += 2;
			while( DL_IS_SPACE( *p ))
			{
				p++;
			}

			if( *p == 0 )
			{
				PrintError(ESYNTAX);
				fTrue = FALSE;
			}
			else
			{
				//  Save the start of string2
				//
				szArg2 = p;

				//  Find the end of string2 and the start of the command
				//  Note the following condition:
				//
				//  if "=="=="==" echo 4
				//
				//  This should produce bad command or file name because
				//  string1 is '"', the condition is '==', string2 is '"'
				//  and '=="==" echo 4' is the command
				//
				while( !DL_IS_SPACE( *p ) && (*p != '=') && (*p != 0) )
				{
					p++;
				}

				if( *p == 0 )
				{
					PrintError(ESYNTAX);
					fTrue = FALSE;
				}
				else
				{
					//  Mark the end of string2
					//
					*p = 0;

					//  Save the location of the command for the TRUE condition
					//
					p2 = p+1;


					//  The start of string1 is szArgs.  Find the end of
					//  string1.  Note the following conditions for MSDOS
					//  compatibility:
					//
					//  if "hello bye"=="hello"  echo 12  (FALSE)
					//  if "hello bye"=="hello echo 13    (TRUE)
					//
					p = szArgs;
					while( !DL_IS_SPACE( *p ) && (*p != '=') && (*p != 0) )
					{
						p++;
					}
					*p = 0;


					//  Compare the strings
					//
					if(strcmp(szArgs,szArg2)==0)
					{
						fTrue = !fTrue;

					}

					//  Save the location of the command since this is TRUE
					//
					szArgs = LFNSkipWhite(p2);
				}
			}
		}
	}

	/* check if the "IF" condition is TRUE, do the command */
	if(fTrue)
	{
		strcpy(szCmdLine, szArgs);

		if(c_ask_autoexec)
		{
			printf(cmdstr_askprompt, szCmdLine);
			c = GetKey("YNA");
			if(c == 'A')
				c_ask_autoexec = FALSE;
			if(c == 'N')
				fTrue = FALSE;
		}

		if(fTrue)
			OneCommand();
	}
}

/*
	LFNFOR [On | OFF]
*/
void PASCAL do_lfnfor(char *szArgs)
{
	if(!set_on_off(szArgs, pszCmdName, &c_lfnfor))
		PrintError(EONOFF);
}

/*
	LOADHIGH [drive:][path]filename [parameters]
	LH [drive:][path]filename [parameters]

	Call a batch file, EXE or COM file.
*/
void PASCAL do_loadhigh(char *szArgs)
{
	extern int c_umb_link;

	c_umb_link = get_umb_link();
	set_umb_link(0x180);				// set link, UMBs first

	//  At the moment we do not process LH arguments /L and /S so
	//  skip over these arguments
	//
	while( *szArgs == '/' )
	{
		while( !DL_IS_SPACE( *szArgs ))
		{
			szArgs++;
		}

		while( DL_IS_SPACE( *szArgs ))
		{
			szArgs++;
		}
	}

	do_call(szArgs);
	set_umb_link(c_umb_link);
}

/*
	$DOCPROC:do_mkdir

	MD		[drive:]path
	MKDIR	[drive:]path
*/
void PASCAL do_mkdir(char *szArgs)
{
	int	iErrorCode;
	FIND	dp;
	bool	bFound = FALSE;

	// only 1 arg allowed
	if(ParmCount(szArgs) != 1)
	{
		PrintError(ENUMPARM);
		return;
	}

	/* strip the quotes (if long filenames) */
	LFNStripArgQuotes(szArgs);

	/* make sure drive is valid */
	if(cmdgetdrive(szArgs) < 0)
		return;

	/* make the directory */
	if( !SmartMakeDirectory( szArgs ) )
	{
		iErrorCode = dlCheckDOSError();

      // MS has some special exceptions not handled by MakeDirectory,
      // so we'll handle them here.  Note: only in the error case.
      //
      bFound = SmartFindFirst( A_SUBDIR, szArgs, &dp );
      if (bFound)
      {
   		// remember to close the search
   		SmartFindClose(&dp);
      }
      SmartExpandPath( FULLPATH_LONGNAME, szArgs, szSplitFileName );

      if (   ( strlen( szSplitFileName ) == 3 )    // exception: root dir
          || ( !bFound )                           // exception: hidden|system
          || ( iErrorCode != 5 )
         )
      {
			PrintError(EMKDIR);        // Unable to create directory
		}
		else
		{
			PrintError(EDIREXISTS);    // Directory already exists
		}
	}
}

/*
	PATH	[[drive:]path[;...]]

	Get or set the path. The path is searched when finding
	executables or batch files (external commands).
*/
void PASCAL do_path(char *szArgs)
{
	extern char cmdstr_nopath[];

	char *pname = "PATH=";
	char *p;
	bool fLastPath = FALSE;
	char * szStart;
	char * szPathLoc;
	char * szSemi;
	bool fFirstPass = TRUE;


	if(*szArgs)
	{
		strcpy(szTemp, pname);
		szArgs = skipequal(szArgs);

		// "path=;" means delete path
		if(*szArgs != ';')
		{
			// only 1 parameter allowed
			if(ParmCount(szArgs) != 1)
			{
				//  Show the path if no arguments were specified
				//
				if(ParmCount(szArgs) == 0)
				{
					goto show_path;
				}


				//  Display an error message
				//
				PrintError(ETOOMANY);
				return;
			}

			LFNStripArgQuotes(szArgs);


			//  Determine if we are in long file name mode
			//
			if( fLFNPresent )
			{
				//
				//  To be like WINDOWS, we need to attempt to replace the
				//  short name version of a long file name directory supplied
				//  in the path but only if it has a drive letter and an
				//  absolute starting point.
				//


				//  Save the location for the start of our temporary path
				//  string
				//
				*szTemp3 = 0;
				szPathLoc = szTemp3;


				//  Save the start of the user suplied path
				//
				szStart = szArgs;


				//  Expand each part of the path to retrieve the short name
				//  for the directory.
				//
				do
				{
					//  Determine the location of the next semicolon
					//
					szSemi = strchr( szStart, ';' );
					if( *szSemi == ';' )
					{
						//  Terminate the string at the semicolon
						//
						*szSemi = 0;
					}
					else
					{
						//  Set a flag indicating that this is the
						//  last part of the path
						//
						fLastPath = TRUE;
					}


					//  Append a semicolon on the end of the short name
					//  path that we are building then set the starting
					//  place for the next path entry
					//
					if( !fFirstPass )
					{
						szPathLoc = szPathLoc + strlen( szPathLoc );
						*szPathLoc = ';';
						szPathLoc++;
						*szPathLoc = 0;
					}
					else
					{
						fFirstPass = FALSE;
					}


					//  Only when a drive letter is given will WINDOWS attempt
					//  to shorten the path.
					//
					if( (szStart[1] == ':') && (szStart[2] == '\\') )
					{
						//  If this is a valid path and we can retrieve the short
						//  name version, then add the short name version to our
						//  path.  Otherwise, copy the name verbatim.
						//
						if( !SmartExpandPath(FULLPATH_SHORTNAME, szStart, szPathLoc) )
						{
							strcpy( szPathLoc, szStart );
						}
					}
					else
					{
						strcpy( szPathLoc, szStart );
					}


					//  Return the semicolon to the szArgs variable and
					//  set the start of the next path part
					//
					if( !fLastPath )
					{
						*szSemi = ';';
						szStart = szSemi + 1;
					}

				}while( !fLastPath );


				strcat(szTemp,szTemp3);
			}
			else
			{
				//  There is no need to look for short name versions
				//  of long file name paths if we are not in LFN mode
				//
				strcat(szTemp,szArgs);
			}
		}
		do_set(szTemp);
	}
	else
	{
		// no args, display the path
show_path:
		p = GetEnvironmentVariable("PATH");
		if(p)
			printf("%s%s\n",pname,p);
		else
			printf(cmdstr_nopath);
	}
}


/*
	PAUSE

	Pause in a batch file with message
*/
#pragma argsused
void PASCAL do_pause(char *szArgs)
{
	extern char cmdstr_pause[];
	printf(cmdstr_pause);
	bdos(0x0C,0,8);
	printf("\n");
}

/*
	PROMPT	[text]
*/
void PASCAL do_prompt(char *szArgs)
{
	// no argument means remove prompt, otherwise what user says
	szArgs = skipequal(szArgs);
	strcpy(szTemp,"PROMPT=");
	if(*szArgs)
		StrNCat(szTemp,szArgs,sizeof(szTemp));
	do_set(szTemp);
}


/*
	REM	[comment]

	REMarkable, isn't it?
*/
#pragma argsused
void PASCAL do_rem(char *szArgs)
{
	/* nothing to do here */
}

/*
	REN		[drive:][path]filespec newspec
	RENAME	[drive:][path]filespec newspec

	Rename one more more files or directories. No path or drive allowed
	in 2nd parameter.
*/
void PASCAL do_rename(char *szArgs)
{
	char *szNewName;
	char *szOldName;
	char *psz;
	int iOrgDrive;
	int iErrorCode;
	extern bool fAssumeWild_szSplitFileName;
	bool fSuccessful;


	/* rename has exactly 2 parameters */
	if(ParmCount(szArgs) != 2)
	{
		PrintError(ENUMPARM);
		return;
	}

	// get 2nd name
	szNewName = LFNNextArg(szArgs);
	LFNStripArgQuotes(szNewName);

	// make sure 2nd part doesn't have a directory or drive
	if(NamePart(szNewName) != szNewName)
	{
		printf("%s - %s\n",messages[EPARMS], szNewName);
		return;
	}

	// null terminate 1st argument
	psz = LFNEndArg(szArgs);
	*psz = 0;
	LFNStripArgQuotes(szArgs);

	// make sure drive and directory is valid
	if(!determine_dir(szArgs))
		return;

	szOldName = NamePart(szArgs);
	iOrgDrive = INT_get_cur_dr();
	INT_set_cur_dr(*szSplitDrive - 'A');
	if( stricmp( szSplitFileName, szOldName ) != 0 )
	{
		INT_set_cwd("..");
	}

	//  Don't assume wildcard names
	//
	if( !fIsFat32World )
	{
		if( fAssumeWild_szSplitFileName && (!fLFNPresent) )
		{
			// change back to old directory
			INT_set_cur_dr(iOrgDrive);
	 		PrintError(ERENAME);
			return;
		}
	}

	if(!WildcardRename(szOldName, szNewName))
	{
		iErrorCode = dlCheckDOSError();
		if( iErrorCode == 3 )
		{
			PrintError(ENOPATH);
		}
		else
		{
			PrintError(ERENAME);
		}
	}

	// change back to old directory
	INT_set_cur_dr(iOrgDrive);
	INT_set_cwd(szSplitOldCwd);
}

/*
	$DOCPROC:do_rmdir

	Remove a directory.  The directory must be empty.

	RD		[drive:]path
	RMDIR	[drive:]path
*/
void PASCAL do_rmdir(char *szArgs)
{
	/* only 1 argument allowed */
	if(ParmCount(szArgs) != 1)
	{
		PrintError(ENUMPARM);
		return;
	}

	/* strip the quotes (if long filenames) */
	LFNStripArgQuotes(szArgs);

	/* make sure drive is valid */
	if(cmdgetdrive(szArgs) < 0)
		return;

	/* remove the directory */
	if(!SmartRemoveDirectory(szArgs))
	{
		// don't remove current directory
		if(dlCheckDOSError() == 0x10)
			PrintError(ECURDIR);

		// path not found, or not a directory
		else
			PrintError(ERMDIR);
	}
}

/*
	SET	[variable=[string]]

	Add or delete an environment variable, or show the whole environment.
	If the variable is "PATH", then it needs to be uppercased.

	Examples:
	SET  var=value		set VAR to value (always UPPERCASES the variable name)
	SET					display the entire environment
	SET  var				display a single variable
	SET  var=			delete the variable
*/
void PASCAL do_set(char *szArgs)
{
	extern char szEnvComSpec[];
	extern char szEnvPath[];

	char far *lpszEnv;
	char far *lpsz;
	unsigned uEnvSeg, uEnvSize;
	char *p;
	bool	fShowOneOnly = FALSE;

	/* find the environment */
	uEnvSeg  = *(unsigned far *) MK_FP(_psp, 0x2C);
	uEnvSize = *(unsigned far *) MK_FP(uEnvSeg-1,3);
	uEnvSize *= 16;
	lpszEnv = (void far *) MK_FP(uEnvSeg, 0);

	/* display the entire environment */
	if(*szArgs == '\0')
	{
		env_show(lpszEnv);
		return;
	}

	// verify line doesn't start with an equal
	p = intl_strchr(szArgs,'=');
	if(*szArgs == '=')
	{
		PrintError(ESYNTAX);
		return;
	}

	// null terminate the left hand side of equation
	if(p)
		*p = 0;

	// no right hand side, display one variable
	else
		fShowOneOnly = TRUE;

	// find the environment string (must be UPPERCASE)
	intl_strupr(szArgs);
	lpsz = env_match(lpszEnv, szArgs);

	// display just the one variable
	if(fShowOneOnly)
	{
		if(lpsz)
			ShowOneEnvironmentVariable(lpsz);
		return;
	}

	// first remove the variable (may add a new one later)
	if(lpsz)
		env_remove(lpsz);

	// put the equal sign back into the string
	*p++= '=';

	// just delete the variable? (nothing to add?)
	if(*p==0)
		return;

	// if we are setting the path, we need to uppercase
	//  the whole string.  The 5 below is the length of "PATH="
	if( strncmp( szArgs, szEnvPath, 5) == NULL )
		intl_strupr(szArgs);

	// add it to the environment
	env_add(lpszEnv, szArgs, uEnvSize);

	// COMSPEC may change while running a batch file. Keep the resident
	// portion up to date.
	if( strncmp( szArgs, szEnvComSpec, 8) == NULL )
		StrNCopy(c_cmdname, GetEnvironmentVariable("COMSPEC"), OLD_NAMESIZE);
}

/*
	SHIFT

	Shift the %0-%9 arguments over one, so as to access more
	than 10 arguments in a batch file.  Does nothing if not
	in a batch file.
*/
#pragma argsused
void PASCAL do_shift(char *szArgs)
{
	char far * lpBatArgs;
	char far * lpNextBatArg;

	lpBatArgs    = FarLFNSkipWhite(pCurBCB->batArgs);
	lpNextBatArg = FarLFNNextArg(lpBatArgs);
	if(lpBatArgs != lpNextBatArg)
		farmemcpy(pCurBCB->batArgs, lpNextBatArg, farstrlen(lpNextBatArg)+1);
}

/*
	TIME	[time]

	Get or set the system time.  In the form of HH:MM:SS.
	This is intelligent about different country separators.
*/
void PASCAL do_time(char *szArgs)
{
	extern struct rd_country c_country;
	extern char time_id[];

	union REGS regs;
	char str[LINELEN+4];
	char *pm_check;
	unsigned  hour=0, seconds=0, minutes=0;
	int pm_hour=0;		// 0=24hr, -1=am, 1=pm
	char * szTemp;

	if(*szArgs)
		strcpy(str+2,szArgs);
	else
	{
		memset( str, 0, LINELEN+4 );
		get_datetime_str(str,time_id, showtime, "");
	}

again:
#if 1

	szTemp = str+2;
	while( ( *szTemp == ' ' ) || ( *szTemp == 0x09 ) )
	{
		szTemp++;
	}

	if( ( *szTemp == 0x0d ) || ( *szTemp == 0x00 ) )
	{
		return;
	}

#else

	if(*(str+2)==0)
		return;

#endif

	/* check for 12 Hour Clock */
	for(pm_check=str+2; *pm_check; ++pm_check)
	{
		if(isalpha(*pm_check))
		{
			/* the first letter had better be 'P' or 'A' */
			if(toupper(*pm_check) == 'P')
				++pm_hour;
			else if(toupper(*pm_check) == 'A')
				--pm_hour;
			else
				goto err;
			*pm_check = 0;
			break;
		}
	}

	/* get hours, minutes, seconds */
	if(!get_field(str+2, 3, *c_country.tmsep, &hour, &minutes, &seconds))
		goto err;

	/* allow for someone typing 24:00pm, or 0:23 */
	if(pm_hour > 0)		// PM
	{
		if(hour < 12)
			hour += 12;
	}
	if(pm_hour < 0)		// AM
	{
		if(hour > 12)
			goto err;
		if(hour == 12)
			hour -= 12;
	}
	if(hour == 24)
		hour = 0;

	regs.x.ax = 0x2D00;
	regs.x.cx = ((hour) << 8) | minutes;
	regs.x.dx = (seconds << 8);				// we do not DO hundredths
	int86(0x21, (void *)&regs, (void *)&regs);
	if(regs.x.ax & 0xFF)
	{
err:
		PrintError(ETIME);
		memset( str, 0, LINELEN+4 );
		get_datetime_str(str,time_id, 0, "");
		goto again;
	}
}


//  TRUENAME [drive:][path][filename]
//
//  Display the path, expanding substed drives and network drives,
//  with a prepared filename if one exists
//
void do_truename(char *szArgs)
{
	char * szPath = ".";
	char * szNextArg;
	char * szNextArg2;
	unsigned uParmCount;
//	unsigned uLength;
//	LPCDS lpCDS;
//	unsigned uIndex;
	bool fSuccessful;


	szNextArg = strchr(szArgs,'/');
	if( szNextArg )
   {
		szNextArg2 = LFNNextArg(szArgs);
		*szNextArg2 = 0;
		intl_strupr(szNextArg);
		szNextArg2 = strchr(szArgs+1,'/');
		if( *szNextArg2 == '/' )
		{
			*szNextArg2 = 0;
		}
		eprintf("%s - %s\n",messages[EBADSWITCH], szNextArg);
		return;
   }


	uParmCount = ParmCount(szArgs);
	switch( uParmCount )
	{
		case 0:
			break;

		case 1:
			szPath = szArgs;
			break;

		default:
			szNextArg = LFNNextArg(szArgs);
			eprintf("%s - %s\n",messages[ETOOMANY], szNextArg);
			return;
	}


	fSuccessful = dlExpandPath( szPath, szSplitFileName);
	if( (fSuccessful) && (szSplitFileName[1] == '\\') )
	{
		//  Do nothing more for network drives
	}
	else
	{
		fSuccessful = SmartExpandPath(_FULLPATH_DEFNAME | FULLPATH_NOSUBST,
				szPath, szSplitFileName);
	}

	if( fSuccessful )
	{
		intl_strupr(szSplitFileName);
		printf("\n%s\n", szSplitFileName );
	}
	else
	{
		PrintError(EDRIVE);
	}
}


/*
	TYPE	[drive:][path]filename [/P]

	/P   Display one screen full and pause

	Type a file to the console.
*/
void PASCAL do_type(char *szArgs)
{
	char huge * lpBuff;
	long lSize;
	ulong ulCount;
	int iHandle, iActionTaken;
	bool	fPause = FALSE;

	// do we need to pause?
	GetSwitch(szArgs, 'P', &fPause);

	/* type requires exactly 1 parameter */
	if(ParmCount(szArgs) != 1)
	{
		PrintError(ENUMPARM);
		return;
	}
	LFNStripArgQuotes(szArgs);

	/* display error if file can't be opened */
	if(!SmartCreateOpenFile(OPEN_ACCESS_READONLY, A_NORMAL, FILE_OPEN,
					szArgs, &iHandle, &iActionTaken))
	{
		PrintError(ENOFILE);
		return;
	}

	/* allocate as much memory as possible */
	lpBuff = UseBigBuffer(&lSize);

	/* put the file to stdout */
	while((ulCount = HugeRead(iHandle, lpBuff, lSize)) > 0L)
	{
		if(fPause)
			DisplayWithPause(lpBuff, ulCount);
		else
			HugeWrite(STDOUT, lpBuff, ulCount);
	}

	dos_close(iHandle);
}

/*
	VER	[n.nn]

	Displays the ROM-DOS version banner.  Also allows
	user to change the reported internal DOS version.

	Datalight ROM-DOS Version 7.1
	Copyright (c) 1989-1998 Datalight, Inc.

	Kernel Reports Version			5.0
	Kernel Resides in             HMA
	Kernel Revision               2.60.15
	Command Revision              2.60.15
	Super-Boot Kernel             ID(98h), Order(first), Boot
	Datalight BIOS Revision       0.96
*/
void PASCAL do_ver(char *szArgs)
{
	/* from DOSMSG.C */
	extern char cmdstr_dosmem[];
	extern char cmdstr_hmamem[];
	extern char cmdstr_rommem[];
	extern char cmdstr_rammem[];
	extern char cmdstr_kernelver[];
	extern char cmdstr_kernelrev[];
	extern char cmdstr_dosmem[];
	extern char cmdstr_cmdrev[];
	extern char cmdstr_biosrev[];
	extern char cmdstr_superboot[];
	extern char cmdstr_revcol[];
	extern char cmdstr_revcol2[];
	extern char cmdstr_rdver[];

	static char szPromptVersion[]  ="%s%s%s";
	static char szDisplayVersion[] = "\n%s%s%s\n%s\n";
	static char szSuperInfo[]  = "ID(%02Xh), Order(%s), Boot(%s)";
	static char *aSuperOrder[] = { "first", "middle", "last" };
	static char *aSuperBoot[]  = { "normal", "super" };
	char * szPosition;
	unsigned uVersion, uValue;
	int fFull = FALSE, fInPrompt=FALSE;
	uchar ucSuperID, ucSuperOrder, ucSuperBoot;
	char szTemp[34];

	/* prompt string shows simpler version display */
	if(*szArgs == '$')
	{
		fInPrompt = TRUE;
		++szArgs;
	}

	/* does user want full information? */
	GetSwitch(szArgs, 'R', &fFull);

	/* does user wish to SET the version? (ROM-DOS only) */
	if(isdigit(*szArgs))
	{
		// major version
		uVersion = atoi(szArgs);
		if( uVersion > 255 )
		{
			//  Version value is too high
			//
			return;
		}

		// minor version
		szArgs = intl_strchr(szArgs,'.');
		if(szArgs)
		{
			if( isdigit(szArgs[1]) )
			{
				if( isdigit(szArgs[2]) )
				{
					if( isdigit(szArgs[3]) )
					{
						//  Ignore anything past two digits
						//
						szArgs[3] = 0;
					}
				}
			}

			uValue = atoi(szArgs+1);

			// single digit minor versions are really 2 digits (7.1 to 7.10)
			if(!isdigit(szArgs[2]))
				uValue *= 10;

			// minor version goes in high byte
			uVersion |= (uValue << 8);
		}

		// only works for ROM-DOS. ignored with MS kernel
		set_dos_ver(uVersion);
	}

	/* get standard DOS version into string form */
	uVersion = GetDosVersion();
	uValue = uVersion>>8;
	if((uValue%10) == 0)
		uValue /= 10;
	sprintf(szTemp,"%d.%d",(uchar)uVersion, uValue);

	/* do a simplified version if in prompt string */
	if(fInPrompt)
	{
		printf(szPromptVersion, cmdstr_rdver, szTemp, SUV_LABEL);
		return;
	}

	/* do the standard version, with copyright */
	printf(szDisplayVersion, cmdstr_rdver, szTemp, SUV_LABEL, COPYRIGHT);

	/* does user want details as well? */
	if(fFull)
	{
		/* put a line between version and detailed info */
		printf("\n");

		/* display reported kernel version */
		if( fLFNPresent )
		{
			printf(cmdstr_revcol2, cmdstr_kernelver, szTemp,
					"(With LFN support)");
		}
		else
		{
			printf(cmdstr_revcol, cmdstr_kernelver, szTemp );
		}

		/* display where kernel is in memory (HMA, ROM, etc..) */
		switch(get_dos_mem_position())
		{
			case 0x10: szPosition = cmdstr_hmamem; break;
			case 0x8:  szPosition = cmdstr_rommem; break;
			default:   szPosition = cmdstr_rammem; break;
		}
		printf(cmdstr_revcol, cmdstr_dosmem, szPosition, "");

		/* display ROM-DOS kernel revision */
		get_internal_ver(szTemp);
		printf(cmdstr_revcol, cmdstr_kernelrev, szTemp, "");

		/* display command revision */
		if( fLFNPresent )
		{
			printf(cmdstr_revcol, cmdstr_cmdrev, REVISION, "(With LFN Support)");
		}
		else
		{
			printf(cmdstr_revcol, cmdstr_cmdrev, REVISION, "");
		}

		/* display super-boot info if super boot kernel */
		if(GetSuperBoot(&ucSuperID, &ucSuperOrder, &ucSuperBoot))
		{
			/* make sure arrays are in-bounds */
			if(ucSuperOrder > sizeof(aSuperOrder)/sizeof(char *))
				ucSuperOrder = 0;
			if(ucSuperBoot > sizeof(aSuperBoot)/sizeof(char *))
				ucSuperBoot = 0;

			/* display super-boot options */
			sprintf(szTemp, szSuperInfo, ucSuperID, aSuperOrder[ucSuperOrder],
				aSuperBoot[ucSuperBoot]);
			printf(cmdstr_revcol, cmdstr_superboot, szTemp);
		}

		/* display BIOS info if Datalight BIOS */
		if(GetDLBIOSVer(szTemp))
			printf(cmdstr_revcol, cmdstr_biosrev, szTemp);
	}
}


/*
	VERIFY	["ON" | "OFF"]
*/
void PASCAL do_verify(char *szArgs)
{
	int set = INT_setverify(FALSE,0);
	if(set_on_off(szArgs,pszCmdName,&set))
		INT_setverify(TRUE,set);
	else
		PrintError(EONOFF);
}

/*
	VOL [drive]

	Display the volume label
*/
void PASCAL do_vol(char *szArgs)
{
	// don't allow bizaare pathnames
	if(*szArgs && (strcmp(szArgs+1, ":") != 0))
		PrintError(EDRIVE);

	// make sure that the "pause" message doesn't appear during
	// DisplayVolume(), since that function is shared by DIR.
	DisplayVolume(szArgs, TRUE);
}
