//
//   $Id: COMMAND.C 1.30 2003/04/08 16:00:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
	$DOCMODULE:COMMAND.C

	Copyright (c) 1989-2002 Datalight, Inc.
	Datalight Confidential and Proprietary.
   All Rights Reserved.

	This is the COMMAND processor for Datalight ROM-DOS 7.1. It handles
	all internal commands such as "DIR" and "ECHO". COMMAND provides the
	standard DOS prompt, and includes the abilities to run batch files.

	A small (~4K) portion of this code, called RESIDENT code, stays
	in memory at all times. The rest of command, the TRANSIENT code,
	is freed up when running programs. This way, COMMAND essentially
	takes up only about 4K of memory from a program's perspective.

	A000 +---------------+
	     | TRANSIENT CMD |
	     +---------------+ -+
	     |               |  |
	     |  free space   |  |----- used for batch files, DIRs, etc..
	     |               |  |
	     +---------------+ -+
	     | RESIDENT CMD  |
	     +---------------+
	     | Dos, Drivers  |
        | and TSRs      |
	0000 +---------------+

	$DOCHISTORY
	04/08/2003 pb  added code to CTTYOpen to check if there is a device
						attached to the place we are attempting to move the
						console to.
	08/22/2002 jmb added ptr to resident variable for ignoring Abort, Retry,
                  Fail errors
  	03/01/1999 dg  Added 1K line length
	02/16/1999 dg  Added LFN support to piping
	08/20/1997 de  SPR 1361: revised to fix batch file processing bug
	11/10/1988 dg  Created by Drew Gislason
*********************************************************************/
#define __FILENUM__ 2

// standard includes
#include <dir.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// datalight includes
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dllfn.h"
#include "dlcntry.h"
#include "command.h"
#include "comprot.h"
#include "fatworld.h"

int eprintf(const char *fmt, ...);

int errno;

/*
	The RESIDENT (ASM) portion of COMMAND sets "resseg" so the TRANSIENT (C)
	portion can find it.
	Note: "resseg" must be in data segment so it doesn't get zeroed by
	startup code.
*/
unsigned resseg=0;					// segment of resident portion
bool far *c_res_slashf_flag_ptr;    // pointer to resident flag for Int24 Failure
char szCmdLine[BATCH_LINELEN];	// current line
char *pszCmdName;						// cmd name after parsed by MakeCommand()
char *pszCmdArgs;						// cmd args after parsed by MakeCommand()
char szTemp[BATCH_LINELEN];		// temporary storage
char szTemp2[BATCH_LINELEN];		// temporary storage
char szTemp3[BATCH_LINELEN];		// temporary storage
char szInternalCmd[CMDSIZE];		// for internal commands only
bool fDontShowLine;					// Flag reflecting '@' in batch line
int  errno;								// set error
char osmajor;							// Major version of DOS
char osminor;							// Minor version of DOS
bool fLFNPresent;			//  Using an LFN kernel
bool fIsFat32World;		//  In a 7.1 enviornment
bool fAssumeWild_szSplitFileName;

char autoexec[] = "AUTOEXEC.BAT";	// startup batch file name


/*
	Process command line and perform all setup that only
	happens upon startup.  main() will never be executed
	again after startup.

	/E:xxx	Set the environment size
	/C	      Spawn command for a "single" command then exit
	/D       Even though permanent, DONT run autoexec.bat
	/K       Like /C, but don't exit
	/N, /P	Make this the permanent COMMAND
	/Y       Step through each command in the batch file
	/?	Help

	The cmdline is in the form of "#string\r" where # is the number of
	bytes in the string.
*/
void main(char *szCmdArgs)
{
	extern unsigned c_uPrevPSP;
	extern bool c_fFailInt24;
	extern bool c_fDisplayErrorlevel;
	extern char szStatusErrorLevel[];
	extern char szStatusReloaded[];

	int	iLen;
	char	cOpt;
	unsigned uEnvParas;
	unsigned uEnvSeg;
	unsigned far * lpu;
	unsigned far * lpu2;
	int fStayUntilExit;
	int fAskAutoexec;
	struct PSP far * lpPSP;


	/* point our PSP to ourselves so programs KNOW we are COMMAND */
	lpPSP = MK_FP( _psp, 0 );
	c_uPrevPSP = lpPSP->last_psp;
	lpPSP->last_psp = _psp;

	//  Determine if we have LFNs
	//
	fLFNPresent = LFNPresent();

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* set up the country information */
	InitCountryInfo();

	/* initialize switches to off */
	fAskAutoexec = fStayUntilExit = FALSE;

	/* echo starts on */
	c_no_batch_echo = TRUE;

	/* no redirection to start with */
	c_redirect[0] = c_redirect[1] = -1;

	/* user's can now press ctrl-C */
	ctrlc_ok_now();

	/* null terminate the cmdline */
	iLen = *szCmdArgs++;
	*(szCmdArgs+iLen) = 0;

	/* process each line */
	while(*szCmdArgs)
	{
		if(*szCmdArgs++ != '/')
			continue;

		/* determine the command line option */
		cOpt = toupper(*szCmdArgs);
		++szCmdArgs;

		/* do the option */
		switch(cOpt)
		{
			/* this COMMAND.COM cannot exit - "permanent" */
			case 'N':	/* DL specific: no exit */
			case 'P':	/* permanent */

				/* tell all we are permanent */
				c_no_exit |= S_NOEXIT;

				/* create the COMSPEC */
				SetComspec();
				break;

			/* don't perform autoexec.bat */
			case 'D':
				c_no_exit |= S_NOAUTOEXEC;
				break;

			/* perform a single command, but don't exit until user says */
			case 'K':
				c_one_only = fStayUntilExit = TRUE;
				break;

			/* ask on each line in AUTOEXEC.BAT */
			case 'Y':
				fAskAutoexec = TRUE;
				break;

			/* modify environment size */
			case 'E':

				/* skip past the /E: switch */
				++szCmdArgs;
				iLen = atoi(szCmdArgs);
				while(isdigit(*szCmdArgs))
					++szCmdArgs;

				//  ETR ? 9-13-00 BT
				//
				//  MS-DOS programs have at least 256 bytes of
				//  environment space.
				//
				/* allow env of 160 bytes to 32K bytes */
				if(iLen < 256)
					iLen = 256;
				if(iLen > 32760)
					iLen = 32760;
				uEnvParas = ((iLen+15)/16);		/* get number of paras */

				/* allocate space for the new enviroment */
				farcall(res_openhole);
				uEnvSeg = dos_malloc(uEnvParas);
				farcall(res_closehole);

				/* copy the old environment to the new */
				if(uEnvSeg)
					env_make_new_copy(uEnvSeg, uEnvParas);
				break;

			/* perform a single command and exit */
			case 'C':
				c_one_only = TRUE;
				break;

			//  Display error level after exec
			//
			case 'Z':
				c_fDisplayErrorlevel = TRUE;
				break;

			//  Fail all int 24h requests
			//
			case 'F':
				c_fFailInt24 = TRUE;
				*c_res_slashf_flag_ptr = TRUE;
#if DEBUG_EXE
				{
					extern unsigned far res_c_fFailInt24;
					res_c_fFailInt24 = TRUE;
				}
#endif
				break;

			/* Display COMMAND Help */
			case '?':
				do_ver("");
				PrintMessage(EUSAGE);
				PerformExit();

		} /* end of switch */

		if(c_one_only)
			break;

	} /* end of while */

	/* only one command */
	if(c_one_only)
	{
		/* if redirection failed, don't do anything */
		strcpy(szCmdLine, szCmdArgs);
		if(OpenRedirection(szCmdLine) == ERR)
		{
			PerformExit();
		}

		/* actually process the command */
		OneCommand();

		/* for batch files, continue processing them */
		if(fStayUntilExit || c_in_batch || c_pipe_ptr)
		{
			/* stay until exit is /K switch */
			if(fStayUntilExit)
				c_one_only = FALSE;
			CommandLoop();
		}

		/* done, exit COMMAND */
		PerformExit();
	}

	/* all this stuff is done for a "permanent" command */
	if(c_no_exit)
	{
		/* copy all INT 22H, 23H, 24H to our PSP */
		lpu = MK_FP(_psp, 0xA);
		lpu2 = MK_FP(0, 0x22*4);
		for(iLen=0; iLen<6; ++iLen)
			*lpu++ = *lpu2++;

		/* check for AUTOEXEC.BAT */
		if(!(c_no_exit&S_NOAUTOEXEC) && FileExists(autoexec))
		{
			/* user pressed F8, ask during AUTOEXEC */
			c_ask_autoexec = fAskAutoexec;

			strcpy(szCmdLine, autoexec);
			OneCommand();
			CommandLoop();
		}
		else
		{
			if( !fIsFat32World )
			{
				if( !(c_no_exit & S_NOAUTOEXEC) )
				{
					/* no AUTOEXEC.BAT, do DATE and TIME */
					printf("\n");
					do_date("");
					printf("\n");
					do_time("");
					printf("\n");
				}
			}
		}
	}

	/* display the version for all to see */
	do_ver("");

#if 0
	if( c_fDisplayErrorlevel )
	{
		printf(szStatusErrorLevel, c_errorlevel );
		if( TRUE )
		{
			printf(szStatusReloaded);
		}
	}
#endif

	/* go off into normal command loop */
	CommandLoop();
}


//  Perform any on exit code and then exit
//
void PASCAL PerformExit( void )
{
	extern void res_exit();
	extern unsigned _psp;
	extern unsigned c_uPrevPSP;
	struct PSP far * lpPSP;


	//  Restore the previous PSP address.  (this had to be changed to
	//  point to ourselves otherwise 3rd party programs could not
	//  find command's environment
	//
	lpPSP = MK_FP( _psp, 0 );
	lpPSP->last_psp = c_uPrevPSP;


	//  exit
	//
	farcall(res_exit);
}


/*
	Create a COMSPEC from the name found at the end of the environment
	(passed to TRANSIENT from RESIDENT portion in "c_exe_name"). Done
	once upon startup.
*/
void PASCAL SetComspec(void)
{
	strcpy(szTemp, "COMSPEC=");
	strcat(szTemp, c_exe_name);
	do_set(szTemp);
}

/*
	"CommandLoop" never returns. The ONLY time it can exit is if the
	"one_only" variable is set. It's job is to:

	1. print the prompt
	2. get a command
	3. perform the command
	4. back to step 1
*/
void CommandLoop(void)
{
	extern char cmdstr_askprompt[];

	int	iDosVer;
	bool	fCrLf;
	bool	fDoCmd;
	int	c;

	/* debugging */
#if HISTORY
	int fd;
	char * szDebugName = "C:\\CMD.HST";
#endif

	// just in case we're reloading, set DOS version here
	iDosVer = GetDosVersion();
	osmajor = (char)iDosVer;
	osminor = iDosVer >> 8;

	// just in case we're reloading, Determine if we have LFNs
	//
	fLFNPresent = LFNPresent();

	// just in case we're reloading, Determine if we are operating
	// in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	// just in case we are reloading, close any previous redirection
	CloseRedirection();

	// if the user changed the COMSPEC environment variable as the result
	// of an error during a reload of the transient part of COMMAND then
	// we need to adjust COMSPEC to that new value passed in c_exe_name.
	//
	if(c_new_COMSPEC)
	{
		SetComspec();				// set COMSPEC=c_exe_name
		c_new_COMSPEC = 0;		// it is official now
	}

	// pick up any open batch files
	if(c_in_batch)
		pCurBCB = MK_FP(c_curBCBSeg, 0);

	fCrLf = !c_in_batch;

	while(1)
	{
		errno = 0;

		/* if COMMAND called with /C, exit once done w/ batch */
		if(c_one_only && !c_in_batch && !c_pipe_ptr)
			break;

		if(fCrLf)
			printf("\n");

		// LFN handle count MUST be 0
		assert(SmartFindAreAllClosed());

		/* typically, we both want to do the command and print CR/LF */
		fDoCmd = fCrLf = TRUE;
		if(GetCommand(szCmdLine))
		{

#if HISTORY
			/* debugging - display command history */
			fd = dos_open(szDebugName, WR);
			if(fd < 0)
				fd = dos_creat(szDebugName);
			if(fd > 0)
			{
				dos_seek(fd, 0L, SEEK_END);			/* seek to the end */
				dos_write(fd, (char far *)szCmdLine, strlen(szCmdLine));
				dos_write(fd, (char far *)"\r\n", 2);
			}
			dos_close(fd);
#endif

			/* turn off asking after AUTOEXEC is done */
			if(c_ask_autoexec && !c_in_batch)
				c_ask_autoexec = FALSE;

			/* user pressed F8, ask on each line */
			if(c_ask_autoexec)
			{
				// don't ask on blank or comment lines
				if(!IsCommentLine(szCmdLine))
				{
					printf(cmdstr_askprompt, szCmdLine);
					c = GetKey("YNA");
					if(c == 'A')
						c_ask_autoexec = FALSE;
					if(c == 'N')
						fDoCmd = FALSE;
				}
			}

			/* Check for redirection.  Remove "<redirection" string(s). */
			if(fDoCmd && (OpenRedirection(szCmdLine) == ERR))
				fDoCmd = FALSE;

			if(fDoCmd)
			{
#if FALSE

				//  For some reason we seem to loose the state of these
				//  variables.  Assert them now.
				//
				if( fLFNPresent != LFNPresent() )
				{
					_assert( __LINE__ );
				}

				if( fIsFat32World != dlIsFat32World() )
				{
					_assert( __LINE__ );
				}

#endif // #if TRUE

				/* Split the command up into "pszCmdName" and "pszCmdArgs" */
				MakeCommand();

				/* actually go perform the command */
				/* no crlf if in batch or no echo */
				if(!PerformCommand() || (!c_echo || fDontShowLine))
				{
					fCrLf = FALSE;
				}
				else
				{
					if( c_in_batch )
					{
						fCrLf = c_echo;
					}
					else
					{
						fCrLf = TRUE;
					}
				}
#if FALSE

				//  For some reason we seem to loose the state of these
				//  variables.  Assert them now.
				//
				if( fLFNPresent != LFNPresent() )
				{
					_assert( __LINE__ );
				}

				if( fIsFat32World != dlIsFat32World() )
				{
					_assert( __LINE__ );
				}

#endif // #if TRUE

			}
		}
		CloseRedirection();		/* if any redirection */
	}

	/* needed for "COMMAND /C COMMAND /C COPY X.Y X.Z >NUL" */
	PerformExit();
}


/*
	Execute a single command.  This function expects the
	command to be in the "cmdline" character array.
*/
bool PASCAL OneCommand(void)
{
#if FALSE

	//  For some reason we seem to loose the state of these
	//  variables.  Assert them now.
	//
	if( fLFNPresent != LFNPresent() )
	{
		_assert( __LINE__ );
	}

	if( fIsFat32World != dlIsFat32World() )
	{
		_assert( __LINE__ );
	}

#endif // #if TRUE

	MakeCommand();
	return PerformCommand();
}


/*
	Parse the cmdline into 2 parts: pszCmdName and pszCmdArgs
*/
void PASCAL MakeCommand(void)
{
	int  i, iLen;
	char *psz;
	char *szEndCommand;
	static char szSeparatorChars[] = ".\\:";

	// Remove preceding whitespace and/or equal signs.
	psz = skipequal(szCmdLine);
	iLen = strlen(psz);
	memmove(szCmdLine, psz, iLen+1);
	psz = szCmdLine;

	/*
		internal commands are special - they don't need whitespace between
		command and arguments.
		Examples:
			DIR\SUBDIR	will do a directory of \SUBDIR
			DIR.EXE		will list all EXEs
			ECHO.			will echo a blank line
	*/
	szEndCommand = LFNEndArg(psz);
	for(i=0; filechar(*psz) && !intl_strchr(szSeparatorChars, *psz) &&
			psz < szEndCommand && i<CMDSIZE-1; ++i)
	{
		if(IsDoubleByteChar(*psz))
			szInternalCmd[i++] = *psz++;
		szInternalCmd[i] = *psz++;
	}
	szInternalCmd[i] = 0;

	// not and internal command, must be external (.bat, .exe or .com)
	if(IsInternalCommand(szInternalCmd) < 0)
	{
      // ETR 3615 fix probably goes here :)

		// NOT an internal command
		*szInternalCmd = 0;

		// set pointer to end of the command
		psz = szEndCommand;
	}

	// null terminate the cmdname (must move up args by 1)
	memmove(psz+1, psz, sizeof(szCmdLine) - (psz-szCmdLine) - 1);
	if( c_pipe_ptr )
	{
		c_pipe_ptr++;
	}
	*psz++ = 0;

	// set up pointers to command name and args
	pszCmdArgs = psz;
	pszCmdName = szCmdLine;

	// strip quotes (if any) off longname commands
	LFNStripArgQuotes(pszCmdName);
}

/*
	Actually do what the command says.
*/
bool PASCAL PerformCommand(void)
{
	extern struct cmds cmds[];
	extern int errno;
	extern bool c_fDisplayErrorlevel;
	extern char szStatusErrorLevel[];
	extern char szStatusReloaded[];

	int i;
	char *p;


	/* if no command (just CR) */
	if(*pszCmdName == 0)
		return FALSE;

	/* Check for Invalid Drive */
	if(cmdgetdrive(pszCmdName) == ERR)
		return TRUE;

	/* check for select drive (ie A:) */
	if( (strcmp(pszCmdName+1, ":") == 0) || (strcmp(pszCmdName+1, ":\\") == 0) )
	{
		INT_set_cur_dr(toupper(pszCmdName[0]) - 'A');
		return TRUE;
	}

	/* check for internal command */
	i = IsInternalCommand(pszCmdName);
	if(i >= 0)
	{
		/* Did the user request help? */
		p = LFNSkipWhite(pszCmdArgs);
		if(strncmp(p, "/?", 2) == 0)
		{
			do_help(pszCmdName);
			return TRUE;
		}

		/* skip preceding whitespace for all internal commands but echo */
		if(cmds[i].func != do_echo)
			strcpy(pszCmdArgs, LFNSkipWhite(pszCmdArgs));

		/* skip following spaces if not echo or prompt */
		if(cmds[i].func != do_echo && cmds[i].func != do_prompt)
			StripEndingSpaces(pszCmdArgs);

		/* perform the internal command */
		(*cmds[i].func)(pszCmdArgs);
		return TRUE;
	}

	// execute the command (.COM, .EXE, or .BAT)
	if(!ExecuteCommand())
	{
		errno = -c_iExecError;
		if(errno == 8)					/* Not Enough Memory */
			PrintError(ENOMEM);
		else if(errno == 0xB)		/* Error in EXE file */
			PrintError(EBADEXE);
		else
			PrintError(EBADCMD);			/* Bad Command or Filename */
	}
	else
	{
#if 0
		if( c_fDisplayErrorlevel )
		{
			printf(szStatusErrorLevel, c_errorlevel );
			if( TRUE )
			{
				printf(szStatusReloaded);
			}
		}
#endif
	}

	if(c_iExecError == 256)
		CancelBatchJobOnRequest();

	return TRUE;
}


/*
	is this an internal command?, Returns # or ERR if not
*/
int PASCAL IsInternalCommand(char *szCommand)
{
	extern int MAX_CMD;
	extern struct cmds cmds[];

	int i;
	for(i=0; i<MAX_CMD; ++i)
		if(stricmp(szCommand, cmds[i].cmd) == 0)
			return i;
	return ERR;
}

/*
	Execute an EXE, COM or BAT file. Uses "pszCmdName" and "pszCmdArgs".

	1) search current dir w/ extensions (.COM, .EXE, .BAT)
	2) if no path in name, search PATH w/ extensions (.COM, .EXE, .BAT)
*/
bool PASCAL ExecuteCommand(void)
{
	int iLen;
	int iFileType;			/* can be ERR=not found, COM, EXE, BAT */
	unsigned uIndex;
	unsigned uLength;


	// start with no error
	c_iExecError = 0;

	//  Reparse the command line to catch cases like the following:
	//
	//  test/a/b/c
	//
	//  where:
	//		argv[0] = "test"
	//		argv[0] = "/a/b/c"
	//
	uIndex = 0;
	while( pszCmdName[uIndex] != 0 )
	{
		if( pszCmdName[uIndex] == '/' )
		{
			break;
		}

		uIndex++;
	}

	if( pszCmdName[uIndex] == '/' )
	{
		uLength = strlen( &(pszCmdName[uIndex]) );
		memmove( &(pszCmdArgs[uLength]), pszCmdArgs,
				strlen( pszCmdArgs ) + 1 );
		memmove( pszCmdArgs, &(pszCmdName[uIndex]), uLength );
		pszCmdName[uIndex] = 0;
	}

	// look for the program (.COM, .EXE or .BAT)
	if(!FindInPath(c_exe_name, pszCmdName, &iFileType))
		return FALSE;

	// open the BATch file for input
	if(iFileType == BAT)
	{
		return BatchOpen(c_exe_name, pszCmdArgs);
	}

	// must be .EXE or .COM

	// set up the command line
	iLen = strlen(pszCmdArgs);
	if(iLen > LINELEN-2)
		iLen = LINELEN-2;
	c_exe_cmdline[0] = iLen;
	memmove(c_exe_cmdline+1, pszCmdArgs, iLen);
	c_exe_cmdline[1+iLen] = '\r';

	// make a copy of the COMMAND line in environment space
#if LFN
	strcpy(szTemp,"CMDLINE=");
	StrNCat(szTemp, pszCmdName, sizeof(szTemp));
	StrNCat(szTemp, pszCmdArgs, sizeof(szTemp));
	do_set(szTemp);
#endif

	// set up the FCBs
	dos1_parse((void *)c_exe_fcb1, pszCmdArgs);
	dos1_parse((void *)c_exe_fcb2, LFNNextArg(LFNSkipWhite(pszCmdArgs)));

	// execute the subprocess
try_in_low_mem:
	farcall(res_spawn);

	// check if spawn failed
	if(c_iExecError < 0)
	{
		// if we tried do LOADHIGH and it failed, try it low
		if(get_umb_link())
		{
			c_iExecError = 0;
			set_umb_link(0);
			goto try_in_low_mem;
		}
		return FALSE;
	}

	// program executed
	return TRUE;
}


/*
	Find a program (.COM, .EXE or .BAT) in the current directory or path.
	Uses the following rules:

	1. Search first in current directory.
	2. If not found, search using each item in the PATH.
	3. If user specifies an extension, use only that extension
	4. If user specifies a path, don't search items in the PATH.

	szDest	guaranteed to be a shortname that is found.
	szName	input name to try. may have path already.

	Also uses globals:
	szTemp[]
	szDirPath[]

	Returns TRUE if found, FALSE if not.
*/
bool PASCAL FindInPath(char szDest[], char *szName, int *piFileType)
{
	static char * szExt[] = { ".COM", ".EXE", ".BAT" };

	int	iFileType, iUseFileType;
	char	*pszExt, *psz;
	int	i;
	bool	fFirstTime;
	char * szLocationInDest;
	char * szLocationInTemp;

	// determine if user wants only a specific file extension
	pszExt = ExtPart(szName);
	iFileType = iUseFileType = ERR;
	if(*pszExt)
	{
		for(i=0; szExt[i] && iUseFileType==ERR; ++i)
		{
			if(stricmp(pszExt, szExt[i]) == 0)
				iUseFileType = i;
		}

		// only allow valid file extensions
		if(iUseFileType == ERR)
			return FALSE;
	}

	// 1) search current dir w/ extensions (.COM, .EXE, .BAT)
	// 2) search PATH w/ extensions (.COM, .EXE, .BAT)
	fFirstTime = TRUE;
	iFileType  = 0;
	*szDirPath = 0;
	while(1)
	{
		// first, build the path
		StrNCopy(szTemp, szDirPath, DL_MAXPATH);
		StrNCat(szTemp, szName, DL_MAXPATH);

		// extension not specified. try next one.
		if(iUseFileType == ERR)
		{
			// if length is too long, reduce name size by ".EXT" (5) chars
			pszExt = ExtPart(szTemp);
			if(pszExt - szTemp > DL_MAXPATH - 5)
				pszExt = szTemp + (DL_MAXPATH - 5);

			// copy in new extension
			strcpy(pszExt, szExt[iFileType]);
		}

		/* does it exist? great! run it! */
		if(FileExists(szTemp))
		{
			//  Uppercase the only the name part if this is not
			//  a batch file
			//
			if( iFileType != BAT )
			{
				intl_strupr( NamePart(szTemp) );
			}


			//  Build a path that would be like MS-DOS 6.22, MS-DOS 7.1,
			//  or windows
			//
			if( fLFNPresent )
			{
				// get the short (absolute) name. this HAS to work because
				// we already found the name. just in case, make no name.
				*szDest = 0;
				GetShortName(szDest, szTemp);
				intl_strupr( szDest );
			}
			else
			{
				szLocationInDest = szDest;
				szLocationInTemp = szTemp;

				if( szLocationInTemp[1] == ':' )
				{
					szLocationInDest[0] = szLocationInTemp[0];
					szLocationInDest[1] = szLocationInTemp[1];
					szLocationInDest += 2;
					szLocationInTemp += 2;
				}

				if( *szLocationInTemp == '\\' )
				{
					strcpy( szLocationInDest, szLocationInTemp );
				}
				else
				{
					szLocationInDest[0] = '.';
					szLocationInDest[1] = 0;
					SmartExpandPath( FULLPATH_DEFNAME,szDest, szDest );
					szLocationInDest = szDest + strlen(szDest);
					if( szLocationInDest != szDest )
					{
						szLocationInDest--;
						if( *szLocationInDest != '\\' )
						{
							szLocationInDest[1] = '\\';
							szLocationInDest[2] = 0;
						}
					}
					strcat( szDest, szLocationInTemp );
				}
			}


			// return file type
			if(iUseFileType != ERR)
				*piFileType = iUseFileType;
			else
				*piFileType = iFileType;

			// we found it!
			return TRUE;
		}

		// not found, try the next extension (unless specified)
		if(iUseFileType == ERR)
		{
			++iFileType;
			if(iFileType < sizeof(szExt) / sizeof(char *))
				continue;
			iFileType = 0;
		}

		// not found, try the next path (unless specified)
		if(intl_strchr(szName,'\\'))
			break;

		// remove drive letter if there is one
		psz = intl_strchr(szName, ':');
		if(psz == szName+1)
			szName += 2;

		// find the next path item
		if(!GetNextPathItem(szDirPath, fFirstTime))
			break;

		// no longer first time for GetNextPathItem()
		fFirstTime = FALSE;
	}

	// not found
	return FALSE;
}

/*
	Get an item from the path.
*/
bool PASCAL GetNextPathItem(char *szPath, bool fFirst)
{
	static char * pszThisItem;
	char	* psz;
	int	iLen;

again:

	// point to an item in the path
	if(fFirst)
		pszThisItem = GetEnvironmentVariable("PATH");

	// skip the ';'
	else if(pszThisItem && *pszThisItem)
		++pszThisItem;

	// skip any whitespace in front of path item
	if(pszThisItem)
		pszThisItem = LFNSkipWhite(pszThisItem);

	// no more path items
	if(pszThisItem == NULL || *pszThisItem == 0)
	{
		*szPath = 0;
		return FALSE;
	}

	// copy the path into string
	psz = intl_strchr(pszThisItem,';');
	if(psz == NULL)
		psz = pszThisItem+strlen(pszThisItem);

	// copy this path item to user space
	iLen = min(DL_MAXPATH-1, psz-pszThisItem);
	StrNCopy(szPath, pszThisItem, iLen+1);
	pszThisItem = psz;

	// append a \ if needed
	if(iLen)
	{
		if(szPath[iLen-1] != '\\' && szPath[iLen-1] != ':')
			strcat(szPath, "\\");
	}

	// if the drive is invalid, skip this item in the path
	// but display an error message first
	if(cmdgetdrive(szPath) == ERR)
	{
		PrintError(EINVDR);			/* invalid drive in path */
		fFirst = FALSE;
		goto again;
	}

	// got the next path
	return TRUE;
}

/*
	Display an error message to stderr. Displayed by number.
*/
void PASCAL PrintError(int num)
{
	// eprintf printfs to stderr
	if(num < maxerr)
		eprintf("%s\n",messages[num]);
	assert(num < maxerr);
}

/*
	Display a message to stdout. Displayed by number.
*/
void PASCAL PrintMessage(int num)
{
	if(num < maxerr)
		printf(messages[num]);				// like fprintf(stdout,....)
	assert(num < maxerr);
}


/*************************************

/*************************************

				Redirection

	|               (piping)
	>  file/device  (creat)
	>> file/device  (append)
	<  file/device  (read)

	0 = input
	1 = output

**************************************/
char rchar[2] = { '<','>' };
char *pipe, 								// a pointer to a pipe's pathname
		temp_path[DL_MAXPATH], 			// the complete path of the pipe
		pipe_name[] = "\\PIPE0.$$$";	// the pipe's file name
#define PIPE_NUM_OFFSET 5				// where the PIPEn.$$$ number is

// used by IOPipe
#define PIPE_IN	  0
#define PIPE_OUT	  1
#define PIPE_APPEND 2


/*
	Open redirection for piping. This also removes the actual piping
	string from the cmdline.

	Example
		echo hello > x.y

	will remove the "> x.y"

	If redirecting a .BAT file this will parse the .BAT file command line
	and redirect STDIN/STDOUT as needed. Later, PerformCommand() will call
	ExecCommand() which calls BatchOpen() which (if c_in_batch is FALSE)
	will remember the .BAT files STDIN & STDOUT handles set up here and
	return with c_in_batch == TRUE. When CloseRedirection() is called and
	c_in_batch is TRUE, the handles saved by batch_open() will be restored.
	Once the .BAT file has finished c_in_batch will be set to FALSE and
	CloseRedirection() will restore STDIN and STDOUT to CON:. This way, as
	with MS-DOS, a .BAT file's output may be redirected and, additionally,
	each line of the .BAT may be redirected to a different file but
	redirected a CALLed or chained .BAT file has no effect, the original
	redirection is used.

	Returns ERR if failed, OK if worked.
*/
int PASCAL OpenRedirection(char *str)
{
	char	*p, *p2, *p3, *ev;
	char	file[DL_MAXPATH];
	int	i,append=FALSE;
	struct ffblk dp;
	int	tried = 0;
	bool	bFound;
	int	iError;
	char * pCurPos;
	char * pCurPos2;
	bool fInQuote;


	/* don't open redirect on a FOR until executing the actual DO statement */
	str = LFNSkipWhite(str);
	if(strncmpi(str, "FOR", 3) == 0 && DL_IS_SPACE(str[4]))
		return OK;

	// If user does not have write access to the current dir then
	// a pipe will surely fail if we try to open/create a file and
	// write to it. Check the TEMP environment variable and if it
	// is defined use that path, rather than the current dir as the
	// path to the pipe files.
   //
   ev = GetEnvironmentVariable("TEMP");
   if(ev)
	   StrNCopy(temp_path, ev, DL_MAXPATH);

	else
	{
TRY_WITHOUT_TEMP_PATH:
		temp_path[0]= '\0';
	}

	// set up the pipe
   strcat(temp_path, pipe_name);
   pipe = temp_path;

	/* if we're done with piping, delete all old pipe files */
	if(c_pipe_del)
	{
		// pipe files will be numbered 0-9, so use a '?' where the
		// number would be to delete all the pipes.
		pipe[strlen(temp_path)+PIPE_NUM_OFFSET] = '?';
		strcpy(file, pipe);
		SmartWildcardDelete(file);
		c_pipe_num = c_pipe_del = 0;
	}

	/* check for piping input */
	if(c_pipe_ptr)
	{
		pipe[strlen(temp_path)+5] = c_pipe_num+'0';
		if(IOPipe(pipe, PIPE_IN) == ERR)
		{
			/*
				it's possible that TEMP points to an invalid path or one
				in which we don't have write priveleges, so try it in the
				current directory too
			*/
			if((pipe[1] == ':') && !tried++)
				goto TRY_WITHOUT_TEMP_PATH;

			// pipe failed, tell user
			errno = ENOPIPE;
			PrintError(errno);
			return ERR;
		}
		tried = 0;
		++c_pipe_num;
	}


	//  See these two cases for handling pipes in a quote
	//
	//  echo "|" >bt3.out
	//  echo "| >bt3.out
	//
	strcpy( szTemp3, str );
	if( (!fIsFat32World) && (!fLFNPresent) )
	{
		pCurPos = szTemp3;
		fInQuote = FALSE;
		while( *pCurPos )
		{
			if( *pCurPos == '"' )
			{
				fInQuote = !fInQuote;
			}
			else
			{
				if( *pCurPos == '|' )
				{
					if( fInQuote )
					{
						pCurPos2 = pCurPos + 1;
						while( *pCurPos2 )
						{
							if( *pCurPos2 == '"' )
							{
								break;
							}
							pCurPos2++;
						}

						if( *pCurPos2 == '"' )
						{
							*pCurPos = ' ';
						}
					}
				}
			}

			pCurPos++;
		}
	}

	/* check for piping input AND output */
	//
	//	 p = intl_inpipestr(str,'|');
	//
	p = intl_inpipestr(szTemp3,'|');
	if( p )
	{
		p = str + (p - szTemp3);
	}

	if(p)
	{
		pipe[strlen(temp_path)+PIPE_NUM_OFFSET] = c_pipe_num+'0';
		if(IOPipe(pipe, PIPE_OUT) == ERR)
		{
			/*
				it's possible that TEMP points to an invalid path or one
				in which we don't have write priveleges, so try it in the
				current directory too
			*/
			if((pipe[1] == ':') && !tried++)
				goto TRY_WITHOUT_TEMP_PATH;

			/* failed to open pipe. let user know */
			errno = ENOPIPE;
			PrintError(errno);
			return ERR;
		}

		tried = 0;
		c_pipe_ptr = p+1;
		*p = 0;
	}

	else if(c_pipe_ptr)
	{
		c_pipe_ptr = NULL;
		c_pipe_del = 1;
	}


	/*
		[]-------------------------[]
		 |		REDIRECT SECTION		|
		[]-------------------------[]
	*/


	/* find any redirection in the string */
	for(i=0; i<2; ++i)
	{
		//  Ignore any redirection in quoted arguments
		//
		strcpy( szTemp3, str );

		//  BT 3-5-01
		//
		//  Some tests show failures in the test suite because MS 6.22 command
		//  ignores redirection or pipes in quotes in curtain instances.
		//  Re-enabled this to find the failures where it does not ignore
		//  quotes.
		//
		//  An example of where it does ignore redirection or pipes
		//  inside of quotes from the FIND tests in 6.22 non-lfn mode
		//
		//  echo "#include <stdio.h>" >hello.c
		//
		//  Here are instances where it does not:
		//
		//  echo "hello >bt2.out
		//  if "1"=="1 echo passed>bt.out
		//
		//  It seems that if redirection was found inside a block that
		//  had a starting quote but no end quote, the starting quote
		//  was ignored, thus performing the redirection.
		//
		if( (!fIsFat32World) && (!fLFNPresent) )
		{
			pCurPos = szTemp3;
			fInQuote = FALSE;
			while( *pCurPos )
			{
				if( *pCurPos == '"' )
				{
					fInQuote = !fInQuote;
				}
				else
				{
					if( (*pCurPos == '<') || (*pCurPos == '>') )
					{
						if( fInQuote )
						{
							pCurPos2 = pCurPos + 1;
							while( *pCurPos2 )
							{
								if( *pCurPos2 == '"' )
								{
									break;
								}
								pCurPos2++;
							}

							if( *pCurPos2 == '"' )
							{
								*pCurPos = ' ';
							}
						}
					}
				}

				pCurPos++;
			}
		}

      p2 = NULL;
		p = intl_inpipestr(szTemp3,rchar[i]);

      //
      // we either have no redirect characters, or else we have one or more
      // in the "one or more" case, we need to go to the last of them
      //
      while( p )
      {
         p2 = p;
         p = intl_inpipestr(p2+1,rchar[i]);
      }
      //
      // if p2 has a value, then we are in the "one or more" case
      // and we have walked p off the end, so p2 is the last viable one
      //
      if (p2)
      {
         p = p2;
      }

		if( p )
		{
			p = str + (p - szTemp3);
		}


		if(p)
		{
			if( p != szTemp3 )
			{
				if( *(p-1) == '>' )
				{
					p--;
				}
			}

			if(*p=='>' && *(p+1)=='>')		/* check for append */
			{
				append=TRUE;
				p2=LFNSkipWhite(p+2);			/* skip leading spaces */
			}
			else
				p2=LFNSkipWhite(p+1);			/* skip leading spaces */

			/* remove redirection from command line */
			p3 = LFNNextArg(p2);

#if 0
			while((*p3) && (!DL_IS_SPACE(*p3)) && (*p3 != '<') && (*p3 != '>') )
				++p3;
#endif
			LFNArgCopy(file, p2, DL_MAXPATH-2);

			// remove colon from a "lpt1:" type pipe name
			if(strlen(file) > 2 && lastchar(file) == ':')
				file[strlen(file)-1] = 0;

			//  Correct spacing issues for things like:
			//
			//  if NOT "1"=="1>b echo TRUE
			//
			if( *p3 != 0 )
			{
				if( *(p3-1) == ' ' )
				{
					p3--;
				}
			}

			strcpy(p,p3);

			if(i)
				iError = IOPipe(file, (append ? PIPE_APPEND : PIPE_OUT));
			else
				iError = IOPipe(file, PIPE_IN);
			if(iError)
			{
				PrintError(errno);
				return ERR;
			}
		}
	}
	return OK;
}

/* Return the handle value */
unsigned char PASCAL GetHandleVal(int handle)
{
	unsigned char far * hptr;
	hptr = *(void far * far *)MK_FP(_psp, 0x34);
	return hptr[handle];
}

/* Set a handle value */
void PASCAL SetHandleVal(int handle, unsigned char hval)
{
	unsigned char far * hptr;
	hptr = *(void far * far *)MK_FP(_psp, 0x34);
	hptr[handle] = hval;
}

/*
	Open/Create a file for piping. Understands whether this file

	iMode can be PIPE_IN, PIPE_OUT, or PIPE_APPEND

	Returns ERR on failure, or
*/
int PASCAL IOPipe(char *szPipeFileName, int iPipeMode)
{
	extern int c_redirect[2];
	extern unsigned _psp;

	int	i;
	bool	bWorked;
	int	iModeFlags, iAction;
	int	iHandle;
	unsigned char hval;

	// remove any quotes surrounding this filename
	LFNStripArgQuotes(szPipeFileName);

	// create (pipe char '>')
	if(iPipeMode == PIPE_OUT)
	{
		iAction = FILE_CREATE | FILE_TRUNCATE;
		iModeFlags = OPEN_ACCESS_WRITEONLY;
	}

	// open (pipe char '<')
	else if(iPipeMode == PIPE_IN)
	{
		iAction = FILE_OPEN;
		iModeFlags = OPEN_ACCESS_READONLY;
	}

	// append (pipe char '>>')
	else
	{
		iAction = FILE_CREATE | FILE_OPEN;
		iModeFlags = OPEN_ACCESS_READWRITE;
	}

	// can we open/create the file?
	bWorked = SmartCreateOpenFile(iModeFlags, A_NORMAL, iAction,
					szPipeFileName, &iHandle, &iAction);

	// failed - no redirection possible, abort command
	if(!bWorked)
	{
		CloseRedirection();
		if(iPipeMode != PIPE_IN)
			errno = ECREAT;
		else
			errno = (dlCheckDOSError()==2) ? ENOFILE : ENOPATH;
		return ERR;
	}

	/* append, need to seek to end of file */
	if(iPipeMode == PIPE_APPEND)
		dos_seek(iHandle, 0L, SEEK_END);

	// determine which STD thing to trap (STDIN or STDOUT)
	i = (iPipeMode == PIPE_IN ? STDIN : STDOUT);

	// remember this redirection (batch files are handled differently)
	if(!c_in_batch)
		c_redirect[i] = iHandle;

	// force dup manually (some device drivers require it)
	hval = GetHandleVal(iHandle);
	SetHandleVal(i, hval);
	SetHandleVal(iHandle, 0xFF);

	// worked
	return OK;
}

static unsigned char tmp_redir[2] = {1,1};

/*
	Make stdin/stdout point to whatever file/device they might have been
	redirected to.
*/
void PASCAL TempOpenRedirection(void)
{
	int i;
	for(i=0; i<2; ++i)
		SetHandleVal(i, tmp_redir[i]);
}

/*
	Make stdin and stdout point back to stderr.
	Remember current redirection.
*/
void PASCAL TempCloseRedirection(void)
{
	int i;
	unsigned char hval;

	/* record the current stdin/stdout handles */
	for(i=0;i<2;++i)
		tmp_redir[i] = GetHandleVal(i);

	/* set stdin/stdout to stderr */
	hval = GetHandleVal(STDERR);
	for(i=0;i<2;++i)
		SetHandleVal(i, hval);
}

/*
	close any redirection that might be open
*/
void PASCAL CloseRedirection(void)
{
	int i;

	// See comments above OpenRedirection()...
	if(c_in_batch)
	{
		for(i=0; i<2; ++i)
		{
			/* did the batch file redirect something? */
			if(GetHandleVal(i) != c_bat_redirect[i])
			{
				// OK, then close the redirection, and restore original
				dos_close(i);
				SetHandleVal(i, c_bat_redirect[i]);
			}
		}
	}

	// not in a batch file
	else
	{
		for(i=0;i<2;++i)
		{
			if(c_redirect[i] != ERR)
			{
				/* close the redirected (or piped) file */
				dos_close(i);

				/* restore to stderr (CON:) handle which was never redirected */
				SetHandleVal(i, GetHandleVal(STDERR));

				/* no longer redirecting */
				c_redirect[i] = ERR;
			}

			/* CON: is handle value 1??? */
			c_bat_redirect[i] = 1;
		}
	}
}


/*
	1) Open the name sans path/drive
	2) Make sure its ioctl says it's a device
	3) Return fd on device, -1 on error
*/
int PASCAL CTTYOpen(char *name, int mode)
{
	int  fd;
	unsigned udata;
	int  iTemp;
	unsigned uLen = 2;

	/* remove a trailing colon (if any) */
	if(lastchar(name) == ':')
		name[strlen(name)-1] = '\0';

	/* try to open it as a device */
	fd = dos_open(name, mode);
	if(fd >= 0)
	{
		udata = dos_ioctl(fd,0,0);

		if(udata & 0x80)			/* device */
		{
			//  Write to the device to see if there is
			//  a device attached.
			//
			iTemp = dos_write( fd,"\n", uLen );
			if( iTemp == uLen )
			{
				/* set to Console (ie, change C0 into C3) */
				dos_ioctl(fd, 1, (udata|3) & 0xFF);
				return fd;
			}
		}

		dos_close(fd);
	}
	return ERR;
}

/*
	This procedure is called when a ^C (INT 23H) or critical error
	(INT 24H) happens.
*/
void BreakHandler(void)
{
	int c;

	// in case of redirection, restore standard handle state
	Rehandle();

//  Adding too many cr/lf to the screen
//
//	printf("\n");

	/* Current drive no longer valid> */
	if(c_crit_drive != 0xFF && c_ignore &&
			INT_get_cur_dr()==c_crit_drive && c_crit_error)
	{
		/* give user a chance to change drive */
		PrintMessage(EINVDRV);
		c = GetKey(0);
		if(c>='A' && c<='Z')
			INT_set_cur_dr(c-'A');
	}

	/* reset the ignore flag */
	c_ignore = 0;

	/* cancel batch job */
	CancelBatchJobOnRequest();

	/* back to normal command processing */
	CommandLoop();
}

/*
	User hit ctrl-break. Might be canceling a batch job.
*/
void PASCAL CancelBatchJobOnRequest(void)
{
	int c;

	if(c_in_batch)
	{
		if( c_crit_error == -1 )
		{
			PrintMessage(ETERMBAT);		/* Terminate batch job? */
			if(GetYesNo())
			{
				EndAllBatches();			/* End nested batch Calls */
			}
		}
	}
}

/*
	Similar to force_dup, this sets a handle to point to
	a particular SFT.  Does no error checking.
*/
void PASCAL SetHandleSFT(int handle, unsigned char *pHandle)
{
	extern unsigned c_resseg;

	unsigned char far * hptr;
	unsigned char hval;

	/* get handle value from our resident portion */
	hptr = MK_FP(c_resseg, (unsigned)pHandle);
	hval = *hptr;

	/* now, restore our handle values here */
	SetHandleVal(handle, hval);
}

/*
	Close all handles and re-open the standard ones.
*/
void PASCAL Rehandle(void)
{
	extern char c_ctty[];
	extern unsigned char far res_con_sft;
	extern unsigned char far res_aux_sft;
	extern unsigned char far res_prn_sft;

	int fd,i;

	/* close all handles */
	for(i=19; i>=0; i--)
		dos_close(i);

	/* restore normal handles */
	if(stricmp(c_ctty, "CON") == 0)
	{
		SetHandleSFT(STDIN,  (void *)&res_con_sft);
		SetHandleSFT(STDOUT, (void *)&res_con_sft);
		SetHandleSFT(STDERR, (void *)&res_con_sft);
	}

	/* we are not pointing to CON, make CTTY raw */
	else
	{
		unsigned char hval;

		/* point stdout/stderr to stdin */
		fd = CTTYOpen(c_ctty, RDWR);

		/* note: fd WILL be 0 at this point, since we closed all handles */
		hval = GetHandleVal(fd);
		SetHandleVal(STDOUT, hval);
		SetHandleVal(STDERR, hval);
	}

	/* restore AUX and PRN */
	SetHandleSFT(STDAUX, (void *)&res_aux_sft);
	SetHandleSFT(STDPRN, (void *)&res_prn_sft);

	/* close any search handles */
	SmartFindCloseAll();
}
