//
//   $Id: ATA.C 1.7 2002/10/15 20:31:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/****************************************************************

	ATA.C						Created 8-1-95 PKG
								Updated 12-04-96 PKG

	ATA read and write routines

	Copyright (c) 1995-2002 Datalight, Inc.
	All rights reserved.
   Datalight Confidential and Proprietary

 ****************************************************************/
#define __FILENUM__ 6

#include <dos.h>
#include "datalght.h"
#include "common.h"
#include "pcic.h"
#include "ataglue.h"
#include "ata.h"

/* Publics... */
struct _tag_ATACardInfo ATACardInfo[MAX_PCIC_SOCKETS];

/* Used within the ATA module only */
static void ATASeekSet(uchar socket, ulong start, unsigned len);
static unsigned pascal ATAGetDataWord(void);
static unsigned pascal ATAPutDataWord(unsigned word);
static unsigned ATAReadyWait(void);

/*
	Read for ATA drives
	Given a pointer to a data buffer, read in from offset start
	length len into the user buffer.  start and len will always
	be a multiple of SECSIZE (512).  Return TRUE upon success
	FALSE if there is a failure or time-out.
*/
unsigned ATABlockRead(uchar socket, void far *dataPtr,
								ulong start, unsigned len)
{
	unsigned far * data;
	unsigned cmdPort;
	unsigned statusPort;
	unsigned count, word;
	uchar status;

	ASSERT(socket<numSockets);
	ASSERT(dataPtr);
	ASSERT(start % SECSIZE == 0);
	ASSERT(len);
	ASSERT(len % SECSIZE == 0);

	/* Make sure that we got valid card info from the socket
	*/
	if(!ATACardInfo[socket].isValid)
		return FALSE;

	/* Open the IO memory window while we need it */
	OpenIOWindow(socket);

	/* Setup the port addresses and the data pointer */
	data = (unsigned far *) dataPtr;
	statusPort = ioStartAddr + ATA_PORT_STATUS;
	cmdPort = ioStartAddr + ATA_PORT_CMD;

	/* Write the CHS registers and the count (length) */
	ATASeekSet(socket, start, len);

	/* Send the read command */
	outp(cmdPort, ATA_CMD_READ);

	/* Initialize our counters */
	len /= 2;
	count = 0;

	/* Wait till the busy flag is reset */
	if(!ATAReadyWait())
		goto Error;

	/* Get all the data we can... */
	do {
		/* Check that we have data from the ATA device */
		status = inp(statusPort);
		if(status & ATA_STATUS_DRQ)
		{
#if ATA_READ_SECTOR
         ata_get_sector( data, SECSIZE/2, ioStartAddr + ATA_PORT_DATA );
         count += SECSIZE/2;
         data += SECSIZE/2;
#else
         data[count++] = ATAGetDataWord();
#endif  // #if ATA_READ_SECTOR

			/* Wait till the busy flag is reset */
			if(!ATAReadyWait())
				goto Error;
		}

		if((status & ATA_STATUS_ERR) || (status & ATA_STATUS_WRITE_FAULT))
			break;
	} while((count < len));

	/* get ATA status once more, to signify an end of interrupt */
	status = inp(statusPort);

	/* If we didn't read all the data then it failed */

Error:
	CloseIOWindow(socket);

	if(count == len)
		return TRUE;

	/* Return failure, timeout, or ECC error */
	return FALSE;

}


/*
	Write for ATA drives
	Given a pointer to a data buffer, write to offset start
	length len from the user buffer.  start and len will always
	be a multiple of SECSIZE (512).  Return TRUE upon success
	FALSE if there is a failure or time-out.
*/
unsigned ATABlockWrite(uchar socket, void far *dataPtr,
								ulong start, unsigned len)
{
	unsigned far * data = dataPtr;
	unsigned cmdPort;
	unsigned statusPort;
	unsigned count, word;
	uchar status;

	ASSERT(socket<numSockets);
	ASSERT(dataPtr);
	ASSERT(start % SECSIZE == 0);
	ASSERT(len);
	ASSERT(len % SECSIZE == 0);

	/* Make sure that we got valid card info from the socket
	*/
	if(!ATACardInfo[socket].isValid)
		return FALSE;

	/* IO window open! */
	OpenIOWindow(socket);

	statusPort = ioStartAddr + ATA_PORT_STATUS;
	cmdPort = ioStartAddr + ATA_PORT_CMD;

	/* Write the CHS registers and the count (length) */
	ATASeekSet(socket, start, len);

	/* Send the write command */
	outp(cmdPort, ATA_CMD_WRITE);

	/* Initialize our counters */
	len /= 2;
	count = 0;

	/* Wait till the busy flag is reset */
	if(!ATAReadyWait())
		goto Error;

	/* Get all the data we can... */
	do {
		/* Check that we have data from the ATA device */
		status = inp(statusPort);
		if(status & ATA_STATUS_DRQ)
		{
#if ATA_READ_SECTOR
         ata_put_sector( data, SECSIZE/2, ioStartAddr + ATA_PORT_DATA );
         count += SECSIZE/2;
         data += SECSIZE/2;
#else
         ATAPutDataWord(data[count++]);
#endif  // #if ATA_READ_SECTOR

			/* Wait till the busy flag is reset */
			if(!ATAReadyWait())
				goto Error;
		}

		if((status & ATA_STATUS_ERR) || (status & ATA_STATUS_WRITE_FAULT))
			break;
	} while((count < len));

	/* get ATA status once more to signify an end of interrupt */
	status = inp(statusPort);

Error:
	CloseIOWindow(socket);

	/* If we didn't read all the data then it failed */
	if(count == len)
		return TRUE;

	return FALSE;

}


/*
	Initialize the ATA card and get info about it
	Reset the drive, program it for 8bit or 16bit
	Send the identify command to the PCMCIA ATA card
	to determine the size of the ATA disk.  Update the
	disk info spt, heads, and cylinder counts
	and return the total size in bytes of the disk.
 */
ulong SetupATACard(uchar socket)
{
	ulong size;
	unsigned maxCount;
	unsigned statusPort = ioStartAddr + ATA_PORT_STATUS;
	unsigned cmdPort = ioStartAddr + ATA_PORT_CMD;
	unsigned devicePort = ioStartAddr + ATA_PORT_DEV_CTRL;

	ASSERT(socket<numSockets);

	/* Open the IO data path to the card, be sure to close it! */
	OpenIOWindow(socket);

	/* Need to initialize the ATA device so it does not */
	/* attempt to generate an IRQ on DataReQuest Ready  */
	outp(devicePort, ATA_DEV_CONTROL);

	/* Just delay for a 30ms! */
	internal_delay(DELAY_TIME);

	outp(devicePort, ATA_DEV_CONTROL & 0x0a);
	ATAReadyWait();

	/* If we should be using an 8 bit data path enable it */
#if !USE_STANDARD_DATA
	outp(ioStartAddr+ATA_PORT_FEATURES, SET_8BIT_MODE);
	outp(cmdPort, ATA_CMD_SET_FEAT);
	ATAReadyWait();
#endif

	/* Wait till the busy flag is reset */
	if(!ATAReadyWait())
	{
		size = 0L;
		goto Error;
	}
	outp(cmdPort, ATA_CMD_IDENTIFY);

	/* Wait till the busy flag is reset */
	if(!ATAReadyWait())
	{
		size = 0L;
		goto Error;
	}

	while(TRUE)
	{
		uchar status;

		/* get ATA status */
		status = inp(statusPort);

		/* if ATA error, abort */
		if (status & ATA_STATUS_ERR)
		{
			size = 0L;
			goto Error;
		}

		/* if ATA data ready, stop waiting */
		if (status & ATA_STATUS_DRQ)
			break;
	}

	/* skip general configuration word */
	ATAGetDataWord();

	/* get # cylinders */
	ATACardInfo[socket].cylinders = ATAGetDataWord();

	/* skip reserved */
	ATAGetDataWord();

	/* get # heads */
	ATACardInfo[socket].heads = ATAGetDataWord();

	/* skip unformatted bpt, unformatted bps */
	ATAGetDataWord();
	ATAGetDataWord();

	/* get # sectors per track */
	ATACardInfo[socket].sectors = ATAGetDataWord();

	/* skip remaining words (but avoid endless loop) */
	maxCount = SECSIZE;
	while( (inp(statusPort) & ATA_STATUS_DRQ) && maxCount--)
		ATAGetDataWord();

	/* compute media size */
	size = (ulong)SECSIZE * (ulong)ATACardInfo[socket].sectors
					* (ulong)ATACardInfo[socket].heads
					* (ulong)ATACardInfo[socket].cylinders;

Error:
	CloseIOWindow(socket);

	return size;

}


/*
	Change the sector (LBA) into a CHS and write the
	values to the ATA registers
*/
static void ATASeekSet(uchar socket, ulong start, unsigned len)
{
	unsigned devicePort;
	uchar sector, head, count;
	unsigned track;
	ulong lba;
	unsigned ataPort;

	/* Make sure that we got valid card info from the socket
	*/
	ASSERT(ATACardInfo[socket].isValid);
	ASSERT(ATACardInfo[socket].cylinders);
	ASSERT(ATACardInfo[socket].heads);
	ASSERT(ATACardInfo[socket].sectors);
	ASSERT(socket<numSockets);
	ASSERT(start % SECSIZE == 0);
	ASSERT(len);
	ASSERT(len % SECSIZE == 0);


	/* Calculate the IO ports we will be using */
	ataPort = ioStartAddr;

	/*
		Need to initialize the ATA device so it does not
		attempt to generate an IRQ on DataReQuest Ready
		SET THIS BIT BEFORE EACH ACCESS! IT IS A WRITE ONLY REG
	*/
	outp(ataPort + ATA_PORT_DEV_CTRL, ATA_DEV_CONTROL & 0x0a);

	count = len / SECSIZE;
	lba = (ulong) start / SECSIZE;

	/* Calculate the CHS */
	sector = (uchar) (lba % ATACardInfo[socket].sectors) + 1;
	head = (uchar) ((lba / ATACardInfo[socket].sectors)
								% ATACardInfo[socket].heads);
	track = (unsigned) (lba / (ATACardInfo[socket].sectors
								* ATACardInfo[socket].heads));

	ASSERT(track < 1024);
	ASSERT(sector < 64);
	ASSERT(head < 16);

	/* Write the sector count */
	outp(ataPort+ATA_PORT_SEC_CNT, count);

	/* Write the sector number */
	outp(ataPort+ATA_PORT_SEC_NUM, sector);

	/* Write the track to the track high and low bytes */
	outp(ataPort+ATA_PORT_CYL_LOW, (uchar)track);
	outp(ataPort+ATA_PORT_CYL_HIG, (uchar)(track>>8));

	/* Write the data to the ATA registers */
	outp(ataPort+ATA_PORT_DRV_HD, (head | ATA_DH_CHS));

}



/*
	Read a single word from the ATA data port
*/
static unsigned pascal ATAGetDataWord(void)
{
	unsigned uWord = 0;

#if USE_STANDARD_DATA

	uWord = inport(ioStartAddr + ATA_PORT_DATA);

#else

	uWord = inp(ioStartAddr + ATA_ALT_DATA_EVEN);
	uWord |= inp(ioStartAddr + ATA_ALT_DATA_ODD) << 8;

#endif

	return uWord;

}


/*
	Write a single word to the ATA data port
*/
static unsigned pascal ATAPutDataWord(unsigned uWord)
{
#if USE_STANDARD_DATA

	outport(ioStartAddr+ATA_PORT_DATA, uWord);

#else

	outp(ioStartAddr + ATA_ALT_DATA_EVEN, uWord);
	outp(ioStartAddr + ATA_ALT_DATA_ODD, uWord >> 8);

#endif

	return uWord;

}


/*
	Wait till the device is ready to continue.  Return FALSE
	if we time out, TRUE once the device is ready to continue.
*/
static unsigned ATAReadyWait(void)
{
	ulong far * pTicks;
	ulong stopTicks;

	pTicks = (ulong far *) MK_FP(0x40, 0x6c);
	stopTicks = *pTicks + MAX_BIOS_TICKS;

	/* Wait till the device is ready for a command */
	while((inp(ioStartAddr + ATA_PORT_STATUS) & ATA_STATUS_BSY))
		if(*pTicks > stopTicks)
			return FALSE;

	/* The device is ready! */
	return TRUE;
}
