//
//   $Id: ATAGLUE.C 1.5 2002/10/15 20:31:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/************************************************************************
 *  Source   : ATAGLUE.C
 *  Desc     : Provides the glue between the high level read/write
 *             routines, the hardware interface, and the CIS parsing.
 *  Notes    : none
 *
 *     Created: 6-8-95 PKG
 *     Updated: 12-04-96 PKG
 *
 *  Copyright (c) 1995-2002 Datalight, Inc.
 *  All right reserved.
 *  Datalight Confidential and Proprietary
 ***********************************************************************/
#define __FILENUM__ 5

#include <stdio.h>
#include "datalght.h"
#include "common.h"
#include "pcic.h"
#include "ataglue.h"
#include "cis.h"
#include "ata.h"

/* The number of drives we know about */
unsigned numSockets;

/* The memory segment to use.  Can be set by upper OEM layer */
ulong memStartAddr   = MEM_START;
unsigned ioStartAddr = IO_START;
bool bSlightDelay = FALSE;


/* ------------------------------------------------------------ *
 * Init365
 *
 * Initialize the PCIC module interface.
 *  1. Determine the number of hardware sockets available
 *  2. Reset the socket(s)
 *  3. Setup our Callback handler
 *  4. Enable any/all change IRQ lines
 *
 * On Entry:
 *    cmdline far pointer to command line, may be NULL
 *
 * On Exit:
 * 	returns FALSE if no hardware or problems occur, TRUE else
 * ------------------------------------------------------------ */
#ifdef __BORLANDC__
#pragma argsused
#endif
unsigned Init365(uchar far *cmdline)
{
	uchar socket;

	/* Don't parse a NULL pointed string */
	while (*cmdline && cmdline)
	{
		if(*cmdline == '/')
		{
			++cmdline;
			switch(*cmdline)
			{
				/* The default IO address */
				case	'I':
				case	'i':
					++cmdline;
					/* Set the new number of sockets */
					ioStartAddr = atox(cmdline);
					break;

				/* The default memory address */
				case	'A':
				case	'a':
					++cmdline;
					/* Set the new number of sockets */
					memStartAddr = atox(cmdline);
					memStartAddr <<= 16;
					break;

				default:
					++cmdline;
					break;
			}
		}
		++cmdline;
	}

	/* Be sure the IO and memory addresses are good */
	/* It must be aligned on a 16 byte boundary */
	if(ioStartAddr & 0xF)
	{
		bputs("Bad IO address specified!\n");
		return FALSE;
	}

	/* Must be on a 4K segment boundary */
	if(memStartAddr & 0xFFFFFFL)
	{
		bputs("Bad memory address specified!\n");
		return FALSE;
	}

	/* Determine number of slots available */
	for(numSockets=0; numSockets<MAX_PCIC_SOCKETS; ++numSockets)
	{
		if(CheckHardware(numSockets))
		{
			/* Be sure the reset signal is OFF */
			ResetSocket(numSockets);
		}
		else
			break;
	}

	/* If we don't have any then we have no hardware */
	if(numSockets == 0)
		return FALSE;

	/* Do this signon after we determine we have hardware */
	bputs("  Memory address: ");
	bputh(memStartAddr>>16);
	bputc('\n');
	bputs("  IO address: ");
	bputh(ioStartAddr);
	bputc('\n');

	/*
		Here we will have to set up the interrupt
		for the PCIC so we can keep track of insert
		and removals
	*/
	for(socket=0; socket<numSockets; ++socket)
		EnableChangeIRQ(socket, CHANGE_IRQ);

	/*
		Attempt to activate each socket if a card is detected
	*/
	for(socket=0; socket<numSockets; ++socket)
		if(CardDetect(socket))
			ActivateSocket(socket);
		else
			DeactivateSocket(socket);

	return TRUE;

}


/* ------------------------------------------------------------ *
 * Uninit365
 *
 * UnInitialize the PCIC module interface.
 *  1. Disable any change request IRQs
 *  2. Disable all sockets
 *
 * On Entry:
 *    nothing
 *
 * On Exit:
 * 	nothing
 * ------------------------------------------------------------ */
void Uninit365(void)
{
	uchar socket;

	/* Release ALL IRQs assigned to the sockets */
	for(socket=0; socket<numSockets; ++socket)
		DisableChangeIRQ(socket, CHANGE_IRQ);

	/* Turn all the sockets off! */
	for(socket=0; socket<numSockets; ++socket)
		DeactivateSocket(socket);

}


/* ------------------------------------------------------------ *
 * OpenWindow
 *
 *  Allocate memory window and set it up.
 *
 * On Entry:
 *    nothing
 *
 * On Exit:
 * 	returns TRUE upon success, FALSE upon failure
 * ------------------------------------------------------------ */
unsigned OpenWindow(uchar socket)
{
	ASSERT(socket<numSockets);

	if(!SetupMemWindow(socket, MEM_COMMON, MEM_WIN_NUM,
								memStartAddr, WINDOW_SIZE))
	{
		return FALSE;
	}

	/* Set it up for Memory device */
	ConfigureSocket(socket, SOCKET_MEM);

   if(!WindowEnable(socket, MEM_WIN_NUM))
	{
		return FALSE;
	}
   else
   {
      if( bSlightDelay )
      {
      	internal_delay(DELAY_TIME);
      }
   	return TRUE;
   }

}


/* ------------------------------------------------------------ *
 *  MapWindow
 *
 *  Maps a given address of the PCMCIA card memory into the
 *  a window that has been opend with OpenWindow.
 *
 * On Entry:
 *    Starting address of the media to map to.
 *    type of memory to map to real mode address space
 *
 * On Exit:
 *    Pointer to data requested!
 *
 * ------------------------------------------------------------ */
unsigned far * MapWindow(uchar socket, ulong start, unsigned type)
{
	ulong addr;

	ASSERT(socket<numSockets);
	ASSERT(type == MEM_ATTRIB || type == MEM_COMMON);

	WindowDisable(socket, MEM_WIN_NUM);

	/* Move the window to the new address */
	if(!SlideMemWindow(socket, type, MEM_WIN_NUM,
							start & ~(WINDOW_SIZE-1)))
	{
		return(0L);
	}

	WindowEnable(socket, MEM_WIN_NUM);
   if( bSlightDelay )
   {
      internal_delay(DELAY_TIME);
      bSlightDelay = FALSE;
   }

	/* Calculate the offset to the data requested */
	addr = (memStartAddr + (start & (WINDOW_SIZE-1)));
	ASSERT(addr);

	return (unsigned far*)(addr);

} /* END-MapPage */


/* ------------------------------------------------------------ *
 * CloseWindow
 *
 *  Release the memory window opened with OpenWindow
 *
 * On Entry:
 *    socket number for the window allocated
 *
 * On Exit:
 * 	returns TRUE upon success, FALSE upon failure
 * ------------------------------------------------------------ */
unsigned CloseWindow(uchar socket)
{
	ASSERT(socket<numSockets);

	/* Disable and release the resources */
	if(!WindowDisable(socket, MEM_WIN_NUM))
	{
		return FALSE;
	}

	/* It worked */
	return TRUE;

}


/* ------------------------------------------------------------ *
 * OpenIOWindow
 *
 *  Allocate an IO window for the given socket and enable it
 *
 * On Entry:
 *    socket to allocate an IO window for
 *
 * On Exit:
 * 	returns TRUE upon success, FALSE upon failure
 * ------------------------------------------------------------ */
unsigned OpenIOWindow(uchar socket)
{
	ASSERT(socket<numSockets);

	/* Setup the window and enable it */
	if(!SetupIoWindow(socket, IO_WIN_NUM, ioStartAddr))
	{
		return FALSE;
	}

	/* Must turn on the PCIC IO Address Multiplexer */
	ConfigureSocket(socket, SOCKET_IO);

	if(!WindowEnable(socket, IO_WIN_NUM))
	{
		return FALSE;
	}
   else
	{
   	return TRUE;
	}
}


/* ------------------------------------------------------------ *
 * CloseIOWindow
 *
 *  Release the IO window allocated to a socket.
 *
 * On Entry:
 *    socket that had been allocated an IO window
 *
 * On Exit:
 * 	returns TRUE upon success, FALSE upon failure
 * ------------------------------------------------------------ */
unsigned CloseIOWindow(uchar socket)
{
	ASSERT(socket<numSockets);

	/* Disable and release the resources */
	if(!WindowDisable(socket, IO_WIN_NUM))
	{
		return FALSE;
	}

	/* It worked! */
	return TRUE;

}


/* ------------------------------------------------------------ *
 *  DeactivateSocket
 *
 *  DeactivateSocket will release any resources allocated to
 *  that socket. DeactivateSocket may be called for sockets
 *  that are not fully active, but have been partially
 *  configured; it needs to release any resources that were
 *  allocated.
 *
 * NOTE:  DeactivateSocket is called during a Callback!
 *
 * On Entry:
 *    socket to deallocate resources from
 *
 * On Exit:
 * 	returns TRUE upon success, FALSE upon failure
 * ------------------------------------------------------------ */
void DeactivateSocket(uchar socket)
{
	ASSERT(socket<numSockets);

	/* Turn Vpp off... May have been turned on */
	PowerDownSocket(socket);

	ATACardInfo[socket].isValid = FALSE;

}


/* ------------------------------------------------------------ *
 *  ActivateSocket
 *
 *  ActivateSocket will determine if the card in the given
 *  socket is an ATA card and if so power the socket up and
 *  reset it.
 *
 * On Entry:
 *    socket to check and enable
 *
 * On Exit:
 *    nothing
 * ------------------------------------------------------------ */
void ActivateSocket(uchar socket)
{
	unsigned far * data;
	uchar mfgCode;
	uchar devCode;
	ulong configReg;
	unsigned configValue;

	ASSERT(socket<numSockets);

	/* Enable the socket and turn on Vpp in case its needed */
	PowerUpSocket(socket, VPP_OFF);
   bSlightDelay = TRUE;

	/* Parse the CIS to determine if its an ATA card */
	if(!IsATACard(socket, &mfgCode, &devCode, &configReg, &configValue))
	{
		DeactivateSocket(socket);
		return;
	}

	/* Power up the Vpp pins if needed */
	if(!(devCode & ATA_VPP_ON))
		PowerUpSocket(socket, VPP_ON);

	/* Write the config value into the config register in ATTRIB memory */
	{
		/* Set the config register value to what the CIS listed */
		OpenWindow(socket);

		/* Write the configuration register */
		data = MapWindow(socket, configReg, MEM_ATTRIB);
		ASSERT(data);
		*data = configValue;

		/* Get rid of the window we just used */
		CloseWindow(socket);
	}

	/* Reset and configure the ATA card */
	if(SetupATACard(socket) == 0L)
	{
		DeactivateSocket(socket);
		return;
	}

#if DEBUG
	{
		ulong size;

		size = SetupATACard(socket);
		bputs("\nATA card activated!\nATA card is ");
		bputh(size>>16);
		bputh(size);
		bputs(" bytes.\n");
	}
#endif

	/* Everything worked the card is ready for use! */
	ATACardInfo[socket].isValid = TRUE;
	return;

}


/* ------------------------------------------------------------ *
 *  CheckForChange
 *
 *  Checks the card status sources bit of the PCIC to determine
 *  if the PC Card has changed.  If any changes have occurred
 *  process the change, possibly enabling the card.
 *
 * On Entry:
 *    nothing
 *
 * On Exit:
 *    TRUE, Yes it has changed, and has been processed
 *		FALSE, otherwise
 *
 * ------------------------------------------------------------ */
unsigned CheckForChange(uchar socket)
{
	ASSERT(socket<numSockets);

	/* If a change occurrs, we need to process and */
	/* update all our globals. */
	if(AckCDIRQ(socket))
	{
#if DEBUG
		bputs("\nCard Change Detected!");
#endif
		if(CardDetect(socket))
			ActivateSocket(socket);
		else
			DeactivateSocket(socket);
		return TRUE;
	}
	else
		return FALSE;

}


/* ------------------------------------------------------------ *
 * IsOurATAConfigTuple
 *
 * Given the config tuple data, check it out to see it it is
 * for us to service.
 *
 * On Entry:
 *   pointer to the CISTPL_CONFIG data
 *
 * On Exit:
 *   returns TRUE if tuple is for us, FALSE otherwise
 *
 * ------------------------------------------------------------ */
unsigned IsOurATAConfigTuple(uchar * tuple)
{
	uchar * ptr = tuple;
	unsigned features;
	unsigned powerEntries;

	ASSERT(tuple);

	/* if there's no interface-definition byte, it's not ours */
	if ( !(*ptr & INTERFACE_DEF) )
		return FALSE;

	/* advance to interface-definition byte */
	++ptr;

	/* if it's not an I/O entry, it's not ours */
	if ( (*ptr & IO_ENTRY) != 1)
		return FALSE;

	/* advance to feature-selection byte */
	++ptr;

	/* memorize it */
	features = *ptr;

	/* if there is no I/O structure, it's not ours */
	if (! (features & IO_STRUCT) )
		return FALSE;

	/* determine number of power entries */
	powerEntries = features & PWR_FEATURES;

	/* skip all power entries */
	while (powerEntries--)
	{
		unsigned numParams;
		int i;

		/* advance to first byte of power entry */
		++ptr;

		/* determine number of parameters in this power entry  */
		/* each bit (0..6) represents a power entry, we simply */
		/* need to skip each on of them! */
		numParams = 0;
		for (i = 0; i < 7; ++i)
			if (*ptr & (1<<i))
				++numParams;

		while (numParams--)
		{
			/* advance to first parameter */
			++ptr;

			/* as long as high bit is set, skip bytes */
			while (*ptr & 0x80)
				++ptr;

			/* we are now pointed at the last byte of the parameter */
		}

		/* we are now pointed at the last byte of the last power parameter */
	}

	/* if there is a timing structure, skip it */
	if (features & 0x04)
	{
		unsigned timing;

		/* advance to first byte of timing structure */
		++ptr;

		timing = *ptr;

		/* if there are wait byte(s), skip them */
		if ( (timing & WAIT_BYTES) != WAIT_BYTES)
		{
			/* advance to first wait byte */
			++ptr;

			/* as long as high bit is set, skip bytes */
			while (*ptr & 0x80)
				++ptr;
		}

		/* if there are rdy/bsy byte(s), skip them */
		if ( (timing & READY_SCALE) != READY_SCALE)
		{
			/* advance to first rdy/bsy byte */
			++ptr;

			/* was long as high bit is set, skip bytes */
			while (*ptr & 0x80)
				++ptr;
		}

		/* if there are reserved-time byte(s), skip them */
		if ( (timing & RESERVED_SCALE) != RESERVED_SCALE)
		{
			/* advance to first rdy/bsy byte */
			++ptr;

			/* as long as high bit is set, skip bytes */
			while (*ptr & 0x80)
				++ptr;
		}

		/* we now point to the last byte in the timing structure */
	}

	/* now we are to the I/O structure; parse it */
	/* advance to first byte of the I/O structure */
	++ptr;

	/* if there's a range, it's not ours */
	if (*ptr & 0x80)
		return FALSE;

	/* if there aren't exactly 16 decoded lines, it's not ours */
	/* (note that a value of 4 means 2^4 which is 16) */
	if ( (*ptr & IO_LINES) != 4)
		return FALSE;

	/* if we made through all that, its good and ours */
	return TRUE;

}


/* ------------------------------------------------------------ *
 *  IsATACard
 *
 *  IsATACard will attempt to parse the CIS to determine whether
 *  the device is one we can handle.  The mtdFlash list will
 *  provide the manufactor ID and device ID to determine the
 *  flash type.
 *
 * On Entry:
 *    nothing
 *
 * On Exit:
 *    TRUE if the card is an ATA card fur us, FALSE otherwise
 * ------------------------------------------------------------ */
unsigned IsATACard(uchar socket, uchar *mfgCode, uchar * devCode,
							ulong *configReg, unsigned *configValue)
{
	unsigned tuplesProcessed;
	int tupleCode;
	uchar * tuple;
	unsigned foundATA;

	ASSERT(socket<numSockets);
	ASSERT(mfgCode);
	ASSERT(devCode);
	ASSERT(configReg);
	ASSERT(configValue);

	/* We have not found an ATA device yet */
	foundATA = FALSE;

	/* Need to init these before use! */
	*mfgCode = 0;
	*devCode = 0;
	*configReg = 0L;
	*configValue = 0;

	/* Parse the tuple chain */
	tuplesProcessed = 0;
	tupleCode = GetFirstTuple(socket);

#if DEBUGTUPLE
   bputs("\nT:");
   bputh(tupleCode);
   bputs("  ");
#endif
   bSlightDelay = TRUE;

	while(tupleCode >= 0 && tuplesProcessed++ < MAX_TUPLES)
	{
		if(tupleCode == TUPLE_JEDEC_C)
		{
			int i;

			/* Get the manufactor code and ID */
			tuple = GetTupleData(socket);

			*mfgCode = tuple[0];
			*devCode = tuple[1];
		}
		else if (tupleCode == TUPLE_FUNC_ID)
		{
			tuple = GetTupleData(socket);
			if (tuple[0] == TPL_FID_FUNCTION_DISK)
			{
				tupleCode = GetNextTuple(socket);
#if DEBUGTUPLE
            bputs("F:");
            bputh(tupleCode);
            bputs("  ");
#endif
				if (tupleCode == TUPLE_FUNC_E)
				{
					tuple = GetTupleData(socket);
					if (tuple[0] == TPL_FE_TYPE_DISK &&
							tuple[1] == TPL_FE_DATA_ATA)
					{
						foundATA = TRUE;
					}
				}
			}
		}
		else if(tupleCode == TUPLE_CONFIG)
		{
			unsigned size;
			int i;

			/* Get the tuple data */
			tuple = GetTupleData(socket);

			/* size is the number of bytes in the value */
			/* of the config register */
			size = tuple[0] + 1;

			/* tuple[2]..tuple[2+size] contain the base */
			/* config register address, LSB first */
			*configReg = tuple[2];
			for(i = 1; i < size; ++i)
				*configReg += (ulong)(tuple[2 + i] << (i * 8) );
		}
		else if(tupleCode == TUPLE_CFTABLE)
		{
			tuple = GetTupleData(socket);

			/*
				if it's a tuple that describes the mandatory,
				16-contiguous-I/O-ports configuration, then
				memorize it's config entry value
			*/
			if(IsOurATAConfigTuple(tuple))
				*configValue = tuple[0] & 0x0F;
		}

		tupleCode = GetNextTuple(socket);
#if DEBUGTUPLE
      bputs("T:");
      bputh(tupleCode);
      bputs("  ");
#endif
	}

#if DEBUGTUPLE
   bputc('\n');
#endif

	/* Be sure everything is in order */
	if((foundATA) && (*mfgCode == MFG_ATA) && (*configReg != 0L))
	{
#if DEBUG
		bputs("\nATA card detected.");
#endif
		return TRUE;
	}
	else
	{
#if DEBUG
		bputs("\nATA card NOT detected.");
#endif
		return FALSE;
	}

}
