//
//   $Id: CIS.C 1.5 2002/10/15 20:31:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/**********************************************************

	CIS.C  				Created: 6/13/95 PKG
							Updated: 12/04/96 PKG

	Contains Tuple parsing functions that use the PCIC
	direct hardware access functions.

   Copyright (c) 1995-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

***********************************************************/
#define __FILENUM__ 4

#include "datalght.h"
#include "common.h"
#include "pcic.h"
#include "ataglue.h"
#include "cis.h"

typedef struct _tuple_data_tag
{
	uchar tuple_code;		/* Current tuple code */
	uchar tuple_len;		/* Tuple data length */
	uchar * tuple_data;	/* Pointer to tuple data */
} TupleInfo;

typedef struct _cis_tag
{
	ulong currTupleAddr;	/* Current tuple address */
	ulong nextTupleAddr;	/* Next tuple address */
	unsigned incBytes;				/* Number to add to ++ address */
											/* 1 for COMMON and 2 for ATTRIBUTE */
	unsigned memType;					/* MEM_COMMON or MEM_ATTRIB */
	unsigned no_more_tuples;		/* Are we done yet? */

	ulong longLinkAddr;
	unsigned longLinkMem;
} CISInfo;

/* Need a couple tuple structures */
static TupleInfo tupleInfo[MAX_PCIC_SOCKETS];

/* Need a CIS place holder */
static CISInfo cisInfo[MAX_PCIC_SOCKETS];

/* a place to hold our tuple data */
static uchar tupleData[MAX_TUPLE_LEN];

/* Prototypes for this module */
static unsigned IsTargetTuple(uchar socket, ulong addr, unsigned type);
static unsigned DoLongLink(uchar socket);

#define CISWindowMap(s,t) MapPage(s, t)

/* ------------------------------------------------------------ *
 * GetFirstTuple
 *
 * Return the first tuple code in the tuple chain.
 *
 * On Entry:
 *    nothing
 *
 * On Exit:
 * 	returns the tuple code, -1 if no more tuples available
 * ------------------------------------------------------------ */
int GetFirstTuple(uchar socket)
{
	ASSERT(socket<numSockets);

	/* Set the start address to 0 */
	cisInfo[socket].currTupleAddr = 0L;
	cisInfo[socket].nextTupleAddr = 0L;

	/* Attribute memory you must skip one byte */
	cisInfo[socket].incBytes = 2;

	/* Start in attribute memory */
	cisInfo[socket].memType = MEM_ATTRIB;

	/* Clear the done flag */
	cisInfo[socket].no_more_tuples = FALSE;

	/* Clear the longlink addresses */
	/* Default to long link to common 0 */
	/* NOTE: this will be set before a long long reguardless */
	cisInfo[socket].longLinkMem = MEM_COMMON;
	cisInfo[socket].longLinkAddr = 0L;

	/* Get that tuple */
	return GetNextTuple(socket);

} /* END-GetFirstTuple */


/* ------------------------------------------------------------ *
 * GetNextTuple
 *
 * Return the next tuple code in the tuple chain.
 *
 * On Entry:
 *    nothing
 *
 * On Exit:
 * 	returns the tuple code, -1 if no more tuples available
 * ------------------------------------------------------------ */
int GetNextTuple(uchar socket)
{
	uchar far * data;
	uchar * tupleData;

	ASSERT(socket<numSockets);
	ASSERT(!cisInfo[socket].no_more_tuples);

	/* If nothing more to do */
	if(cisInfo[socket].no_more_tuples)
		return(-1);

	/* Open a window for use */
	if(!OpenWindow(socket))
	{
		return(-1);
	}

	/* Set the new tuple address */
	cisInfo[socket].currTupleAddr =
					cisInfo[socket].nextTupleAddr;

	/* Get the tuple code */
	data = (uchar far*) MapWindow(socket, cisInfo[socket].currTupleAddr,
									cisInfo[socket].memType);
	ASSERT(data);

	tupleInfo[socket].tuple_code = *data;

	switch(tupleInfo[socket].tuple_code)
	{
			/* CASE: End Of List (0xff) in attribute */
			/* CASE: End Of List (0xff) in common */
			case (uchar)CISTPL_END:
				if(cisInfo[socket].memType==MEM_ATTRIB)
				{
					if(!cisInfo[socket].longLinkAddr)
					{
						cisInfo[socket].nextTupleAddr = 0L;
						cisInfo[socket].memType = MEM_COMMON;

						/* Perform the long link if possible */
						if(!DoLongLink(socket))
						{
							cisInfo[socket].no_more_tuples = TRUE;
							tupleInfo[socket].tuple_code = CISTPL_END;
						}
					}
					else

					/* LONG_LINK and ATTRIBUTE, No Target */
					{
						if(!DoLongLink(socket))
						{
							cisInfo[socket].no_more_tuples = TRUE;
							tupleInfo[socket].tuple_code = CISTPL_END;
						}
					}
				}
				else
				{
					if(!cisInfo[socket].longLinkAddr)
					{
						cisInfo[socket].no_more_tuples = TRUE;
						tupleInfo[socket].tuple_code = CISTPL_END;
					}
					else
					{
						if(!DoLongLink(socket))
						{
							cisInfo[socket].no_more_tuples = TRUE;
							tupleInfo[socket].tuple_code = CISTPL_END;
						}
					}
				}
				break;

			/* CASE: LongLink_A */
			case (uchar)CISTPL_LONGLINK_A:
				cisInfo[socket].longLinkMem = MEM_ATTRIB;
				tupleData = GetTupleData(socket);
				cisInfo[socket].longLinkAddr =
												(ulong) *tupleData;
				break;

			/* CASE: LongLink_C */
			case (uchar)CISTPL_LONGLINK_C:
				cisInfo[socket].longLinkMem = MEM_COMMON;
				tupleData = GetTupleData(socket);
				cisInfo[socket].longLinkAddr =
												(ulong) *tupleData;
				break;

			/* CASE: NO_LINK */
			case (uchar)CISTPL_NO_LINK:

				/* We had better NEVER Link to Offset 0 */
				/* CIS always starts at Attribute 0 AND */
				/* Common 0 */
				cisInfo[socket].longLinkAddr = 0L;
				break;

			/* CASE: LongLink_MFC Currently un-supported*/
			case (uchar)CISTPL_LONGLINK_MFC:
				ERROR();
	}

	/* Next byte... */
	cisInfo[socket].currTupleAddr
					+= cisInfo[socket].incBytes;

	if(tupleInfo[socket].tuple_code!=CISTPL_END)
	{
		/* Get the tuple data length */
		data = (uchar far*) MapWindow(socket, cisInfo[socket].currTupleAddr,
									cisInfo[socket].memType);
		ASSERT(data);
		tupleInfo[socket].tuple_len	= *data;

		/* Get the next address */
		cisInfo[socket].nextTupleAddr
				+= (*data + 2) * cisInfo[socket].incBytes;
	}


	/* Close the window */
	if(!CloseWindow(socket))
		return(-1);

	/* Return the tuple_code */
	if(tupleInfo[socket].tuple_code != CISTPL_END)
		return tupleInfo[socket].tuple_code;
	else
		return (-1);

} /* END-GetNextTuple */


/* ------------------------------------------------------------ *
 * GetTupleLen
 *
 * Return the length of the current tuple
 *
 * On Entry:
 *    nothing
 *
 * On Exit:
 * 	returns the data length of
 * ------------------------------------------------------------ */
unsigned	GetTupleLen(uchar socket)
{
	ASSERT(socket<numSockets);
	ASSERT(!cisInfo[socket].no_more_tuples);

	/* Return the current tuple_len */
	if(!cisInfo[socket].no_more_tuples)
		return tupleInfo[socket].tuple_len;
	else
		return 0;

} /* END-GetTupleLen */


/* ------------------------------------------------------------ *
 * GetTupleData
 *
 * Returns a pointer to the current tuple data.
 *
 * On Entry:
 *    nothing
 *
 * On Exit:
 * 	returns pointer to char of the tuple data
 * ------------------------------------------------------------ */
uchar *	GetTupleData(uchar socket)
{
	unsigned	len,
				index,
				addr=cisInfo[socket].currTupleAddr;
	uchar far * data;

	ASSERT(socket<numSockets);
	/* Better have something to do */
	ASSERT(!cisInfo[socket].no_more_tuples)

	/* Open a window for use */
	if(!OpenWindow(socket))
		return FALSE;

	/* Get the length of the current tuple */
	len = GetTupleLen(socket);

	ASSERT(len);

	/* Skip the first length byte */
	addr += cisInfo[socket].incBytes;

	/* Get each byte of data into our buffer */
	for(index=0; index<len; ++index)
	{
		/* Get the tuple data byte */
		data = (uchar far*) MapWindow(socket, addr, cisInfo[socket].memType);
		ASSERT(data);
		tupleData[index] = (uchar) *data;
		addr += cisInfo[socket].incBytes;
	}

	/* Close the window */
	if(!CloseWindow(socket))
		return FALSE;

	/* Return a pointer to the tuple data */
	return (uchar *) &tupleData;

} /* END-GetTupleData */


/* ------------------------------------------------------------ *
 * DoLongLink
 *
 * Perform a long link to the address and memory type specified
 * in the cis structure.  DoLongLink will evaluate and verify
 * the target tuple and adjust the tuple pointers to point to
 * the next tuple.
 *
 * On Entry:
 *    Assumes a window is already open
 *    nothing
 *
 * On Exit:
 * 	returns TRUE if long link is performed, FALSE otherwise
 * ------------------------------------------------------------ */
static unsigned DoLongLink(uchar socket)
{
	uchar far* data;

	ASSERT(socket<numSockets);

	/* Go to long link tuple */
	cisInfo[socket].nextTupleAddr
			= cisInfo[socket].currTupleAddr
			= cisInfo[socket].longLinkAddr;
	cisInfo[socket].memType
			= cisInfo[socket].longLinkMem;

	/* Check out that tuple code */
	if(!IsTargetTuple(socket, cisInfo[socket].nextTupleAddr,
							cisInfo[socket].memType))
		return FALSE;

	/* Go to next tuple */
	cisInfo[socket].currTupleAddr
		= cisInfo[socket].nextTupleAddr;

	/* Set the possibly new memory type */
	if(cisInfo[socket].memType==MEM_ATTRIB)
		cisInfo[socket].incBytes = 2;
	else
		cisInfo[socket].incBytes = 1;

	/* Be sure to reset the tuple link */
	cisInfo[socket].longLinkAddr = 0L;

	/* Get the tuple length */
	data = (uchar far*) MapWindow(socket, cisInfo[socket].currTupleAddr
				+ cisInfo[socket].incBytes, cisInfo[socket].memType);

	/* Target tuple with no length, its invalid */
	ASSERT(*data>2);

	/* Currently pointing to tuple code, skip code and length */
	cisInfo[socket].nextTupleAddr
		+= (*data + 2)
			* cisInfo[socket].incBytes;

	data = (uchar far*) MapWindow(socket, cisInfo[socket].nextTupleAddr,
								cisInfo[socket].memType);

	ASSERT(data);

	/* If we have no more tuples we are done */
	if(*data == (uchar)CISTPL_END)
		return FALSE;

	/* Get the NEW tuple code and length */
	tupleInfo[socket].tuple_code = *data;

	data = (uchar far*) MapWindow(socket, cisInfo[socket].nextTupleAddr
								+cisInfo[socket].incBytes,
								cisInfo[socket].memType);

	tupleInfo[socket].tuple_len = *data;

	/* And update the current tuple pointer */
	cisInfo[socket].currTupleAddr =
		cisInfo[socket].nextTupleAddr;

	return TRUE;

} /* END-DoLongLink */


/* ------------------------------------------------------------ *
 * IsTargetTuple
 *
 * Determines if the address is a target tuple
 *
 * NOTE: This function could check for the signature of
 * "CIS", which is defined in the PCIC spec.  Instead we
 * will just check for the proper code and assume its
 * good.
 *
 * On Entry:
 *    addr long address to target
 *    type MEM_COMMON or MEM_ATTRIB
 *
 * On Exit:
 * 	returns (1) on target tuple existance or failure (0)
 * ------------------------------------------------------------ */
static unsigned IsTargetTuple(uchar socket, ulong addr, unsigned type)
{
	uchar far *data;

	ASSERT(socket<numSockets);
	ASSERT(type==MEM_COMMON || type==MEM_ATTRIB);

	data = (uchar far*) MapWindow(socket, addr, type);

	ASSERT(data);

	if(*data != 0x13)
	{
		cisInfo[socket].no_more_tuples = TRUE;
		return FALSE;
	}

	return TRUE;

} /* END-IsTargetTuple */


