//
//   $Id: FAT.H 1.5 2002/10/15 20:31:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   FAT.H  created 07/20/94 DG
          updated 12/04/96 PKG

	Copyright (c) 1994-2002 Datalight, Inc.
	All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains the structures necessary to deal
	with a FAT style disk, whether Flash, RAM or ROM.

**********************************************************/


#define SECSIZE		512	/* default sector size */
#define START_BPB		 11	/* offset to the BPB in the boot sector */
#define SECID		0xAA55	/* all valid boot sectors have this */
#define IDOFF		 0x1FE	/* where is the ID located */
#define DEF_SEC_SIZE	512	/* default sector size */
#define FAT_ID		  0xF0	/* default FAT id byte */
#define SECTOR_ID 0xAA55	/* sector ID offset is determined by BPB */
#define BPBOFF		  0x0B	/* where do we find a BPB? */
#define SERIALOFF	  0x27	/* where is serial number */
#define MEDIABYTE   0xF8	/* what media byte do we use? */
#define MAX12CLUS	 0xFF0	/* maximum clusters for 12 bit FATs */
#define MAX16CLUS  0xFFF0	/* maximus clusters for 16 bit FATs */


/*
 *  Standard BIOS Parameter Block.  Every formatted disk
 *  has one of these, whether floppy, hard or PCMCIA disk.
 */
struct BPB
{
	unsigned	bytesPerSector;	/* 00 */
	uchar		secPerCluster;		/* 02 */
	unsigned	reserved;			/* 03 */
	uchar		numFats;				/* 05 */
	unsigned	maxDirs;				/* 06 */
	unsigned	totalSecs;			/* 08   this will be 0 of 32M+ disks */
	uchar		mediaDescrip;		/* 0A */
	unsigned	secPerFAT;			/* 0B */
	unsigned	secPerTrak;			/* 0D */
	unsigned	sides;				/* 0F */
	ulong	hidden;				/* 11 */
	ulong	lTotalSecs;			/* 15   used on 32M+ disks */
};


/* 12-bit FAT entry decode functions */
#define fat0(u) ((u)&0xFFF)
#define fat1(u) ((u)>>4)


/*
 *  Standard Boot Sector.  Sector 0 on every floppy or
 *  hard disk partition.
 */
struct boot_sector
{
	uchar		jmp[3];				/* 00H jump instruction */
	uchar		id[8];				/* 02H ascii id, like "DLDOS5.0" */
	struct BPB bpb;				/* 0BH Bios Parameter Block */
	uchar		drive;				/* 19H Physical drive number */
	uchar		reserved;			/* 1AH Reserved, set to zero */
	uchar		signature;			/* 1BH Signature Xtnd BPB Always 0x29 */
	ulong		serialNumber;		/* 1CH Volume serial number */
	uchar		volumeLable[11];	/* 20H Volume lable string */
	uchar		idString[8];		/* 2BH ID string Ex: "FAT12" */
	uchar		unknown[0x1cb];	/* Boot sector code... */
	unsigned	secid;				/* 1FEH AA55 */
};


/* used for Master Boot Records */
#define PARTOFF	0x1BE		/* offset to first partition struct */
#define SIZEOFPT	0x10		/* size of one partition record */
#define NUMPART	4			/* 4 partitions per boot sector */

#define PDOS12		1			/* DOS partition with 12 bit FAT */
#define PDOS16		4			/* DOS partition up to 32 Meg */
#define PDOSEXT	5			/* extended partition */
#define PDOSBIG	6			/* DOS partition with more than 64K sectors */
#define PDOSLBA16 0x0E     // DOS partition, LBA addressable


/*
 *  Partion table entry.  Begins at offset 1BEH in Master Boot
 *  Record.  There are exactly 4 partition table entries per
 *  harddisk.
 */
struct parttable
{
	uchar isboot;		/* 00H=non, 80H=bootable          */
	uchar shead;		/* starting head of partition     */
	uchar ssect;		/* starting sector of partition   */
	uchar strak;		/* starting cylinder of partition */
	uchar sysid;		/* 0=unused, 1,4,5,6=DOS          */
	uchar ehead;		/* ending head of partition       */
	uchar esect;		/* ending sector of partition     */
	uchar etrak;		/* ending cylinder of partition   */
	ulong pstart;		/* logical start of partition     */
	ulong psize;		/* size of partition in sectors   */
};
