//
//   $Id: MEMATA.C 1.7 2002/10/15 20:31:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	MEMATA.C     created 12/03/96 PKG
                updated 12/04/96 PKG

   Copyright (c) 1996-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This module provides the MEMDISK style interface to the
	ROM-DOS MEMDISK. to build a TSR, Device driver, or even
	a built-in device driver into ROM-DOS that supports
	PCMCIA ATA cards.

	A major performance boost can be made by using the pre-erase
	feature or write without erase function provided by most
	flash ATA vendors.

	ATA cards do NOT have a write protect switch.

	This module defines the following:
	------------------------------------
   int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv);
   bool memread(long offset, unsigned len, char far *buffer);
   bool memwrite(long offset, unsigned len, char far *buffer);
   int memchanged(struct BPB *bpb);
	void memuninit(void);
   bool memioctl(unsigned category, char far *buffer);

History:
  12-17-96 PKG
    Fixed IOCTL track read/write functions.
    Added Get/Set Media ID, DOS format needs it.
    No longer place invalid data into the BPB.
  12-12-96 PKG
    Fixed drive letter display.
    Added memchanged into IOCTL request BPB, drive info
    Added drive existance check into IOCTL

**********************************************************/
#define __FILENUM__ 2

#include "datalght.h"
#include "common.h"
#include "pcic.h"
#include "ata.h"
#include "ataglue.h"
#include "fat.h"
#include "memata.h"
#include "dosver.h"
#include "fatworld.h"

bool fIsFat32World = FALSE;

/* Changed states, disk start info */
struct _tag_DriveInfo driveInfo[MAX_DRIVES];

/* Used for reading the BPB, MBR, etc... */
uchar scratchBuffer[DEF_SEC_SIZE];

/* a simple semi-valid BPB, used as a default */
struct BPB std_bpb = { 512,2,1,2,112, 720,0xFD,2, 9,2,0 };

void bputs(char *s);
void bputc(char c);
void bputi(unsigned val);
void bputh(unsigned val);

/*
   Initialize the device.  This is called only once.  It
   fills in the Bios Parameter Block structure, and returns
   the amount of memory used by the device.

   bpb       The BPB must be filled in for any device.
   cmdline   This is the command line passed to a device
             that is installed via CONFIG.SYS.  This
             parameter is meaningless for built-in devices.
   *endseg   This unsigned is the paragraph just beyond
             the code and data of the device.  Update this
             only if the device requires more RAM from DOS
             (ala a RAM disk in conventional memory).
   drv       This is the starting drive # (0=A, 1=B).  Can
             be used to display starting drive on screen.

   Returns # of units if worked, 0 if failed.
*/
#ifdef __BORLANDC__
#pragma argsused
#endif
int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv)
{
	unsigned i;

	ASSERT(bpb);

	fIsFat32World = dlIsFat32World();

	/* endseg is not used! */
	/* ASSERT(endseg); */

	/* Sign-on message */
	bputs("ATA.SYS");
	bputs(" ");
	if(fIsFat32World)
	{
		bputs(SHORTVERSION);
	}
	else
	{
		bputs(SHORTVERSION622);
	}
	bputs(" " BUILDVERSION "\nPCMCIA ATA disk driver\n" COPYRIGHT "\n\n" );

	/*
		Detect and setup the hardware
		In a built-in device driver this should be hardcoded here
		to provide any command line options.
	*/
	if(!Init365(cmdline))
	{
		bputs("  i82365 compatible controller not detected.\n");
		return 0;
	}

	/* Mark all drives as changed, and read the BPB */
	for(i = 0; i < numSockets; ++i)
	{
		/* Must default to changed and set start to zero. */
		driveInfo[i].changed = TRUE;
		driveInfo[i].start = 0L;

		//  IR 3848 BT 12-13-01
		//
		//  Let the devices be initialized the first time they are
		//  accessed rather than at init time.
		//
		//  This old code below also had a problem where on init, it
		//  would zero the BPB for a failed call but never set the
		//  rebuild flag.  The new code that follows it corrects
		//  this problem.  But it seems redundent to have this type
		//  of code in two places.  Thus it has been left disabled
		//  in the event a good reason is found that warrents it
		//  being re-enabled.
		//
		//  Old code:
		//
		//		/* Get the new BPB for each of them */
		//		memchanged(&bpb[i]);
		//
		//  New code:
		//
		//		if(memchanged(&bpb[i]))
		//		{
		//			driveInfo[i].changed = TRUE;
		//		}
	}

	/* List the drives allocated for us! */
	bputs("  Drive(s) ");
	bputc('A' + drv);
	if(numSockets > 1)
	{
		bputs(" - ");
		bputc('A' + drv + numSockets - 1);
	}
	bputc('\n');

	/* Return the number of drives we have */
	return numSockets;
}


/*
   This function reads data (usually 1 or more sectors)
   from the Generic Disk into the buffer specified by DOS.

   offset   The offset (in bytes) into the disk where the
            read should start.
   len      The length of the read in bytes.  The length
            will not exceed 62K (62*1024 bytes).
   buffer   The buffer, in segment:offset form.

   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
bool memread(long offset, unsigned len, char far *buffer)
{

	ASSERT(offset % DEF_SEC_SIZE == 0);
	ASSERT(len);
	ASSERT(buffer);

	/* Check for and possibly record a disk change */
	if(CheckForChange(memunit))
		driveInfo[memunit].changed = TRUE;

	/* Be sure the card is there */
	if(!CardDetect(memunit))
	{
		memerr = DEVERR_NOTREADY;
		return FALSE;
	}

	/* Skip the MBR */
	offset += driveInfo[memunit].start;

	/* Assume general failure */
	memerr = DEVERR_GENERALFAIL;
	return ATABlockRead(memunit, buffer, offset, len);

}


/*
   (Optional)
   This function is only needed if the disk can be written
   to (that is, not a ROMDISK).

   This function writes data (usually 1 or more sectors)
   to the Generic Disk from the buffer specified by DOS.

   offset   The offset (in bytes) into the disk where the
            write should start.
   len      The length of the write in bytes.  The length
            will not exceed 62K (62*1024 bytes).
   buffer   The buffer, in segment:offset form.

   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
bool memwrite(long offset, unsigned len, char far *buffer)
{

	ASSERT(offset % DEF_SEC_SIZE == 0);
	ASSERT(len);
	ASSERT(buffer);

	/* Check for and possibly record a disk change */
	if(CheckForChange(memunit))
		driveInfo[memunit].changed = TRUE;

	/* Be sure the card is there */
	if(!CardDetect(memunit))
	{
		memerr = DEVERR_NOTREADY;
		return FALSE;
	}

	/* Skip the MBR */
	offset += driveInfo[memunit].start;

	/* Assume general failure */
	memerr = DEVERR_GENERALFAIL;
	return ATABlockWrite(memunit, buffer, offset, len);

}


/*
   (Optional)
   This function is only needed if the disk can be removed
   or changed after bootup.  If the REMOVABLE option is set,
   this function will be called.

   This routine determines if the disk has changed.  Update
   the BPB structure if the device HAS changed or MAY have
   changed.

   bpb       This points to the BPB for THIS unit.  Fill
             the structure in any time the disk has changed.

	Returns CHANGED_YES, CHANGED_MAYBE, or CHANGED_NO
*/
int memchanged(struct BPB *bpb)
{
	unsigned i;
	struct parttable * pt;

	ASSERT(bpb);

	/* First be sure we have a card, if not its changed! */
	if(!CardDetect(memunit))
	{
		/* No drive in the socket, tell DOS it has changed */
		memerr = DEVERR_NOTREADY;
		return CHANGED_YES;
	}

	/* Check for a change, a change can be forced! */
	if(CheckForChange(memunit) || driveInfo[memunit].changed)
	{
		/* Avoid re-entrancy */
		driveInfo[memunit].changed = FALSE;
		driveInfo[memunit].start = 0L;

		/* May be the MBR */
		memread(0L, DEF_SEC_SIZE, scratchBuffer);

		/*
			check each partition table entry, Award v2.5 PCPREP does
			not use the first two entries!???
		*/
		for(i = PARTOFF; i < DEF_SEC_SIZE; i += SIZEOFPT)
		{
			/* look at each table entry! */
			pt = (struct parttable *)(scratchBuffer + i);
			if(pt->isboot && pt->isboot != 0x80)
				continue;
			if(!pt->pstart || !pt->psize)
				continue;

         // 10/15/01 TD IR 3806
         // Intel iFD005P2SA card uses partition ID 0x0E.
         // Added error code also, display error then return "not changed"
         //
			if(   pt->sysid != PDOS12
            && pt->sysid != PDOS16
            && pt->sysid != PDOSBIG
            && pt->sysid != PDOSLBA16
           )
         {
            bputs( "Unknown Partition ID: 0x" );
            bputh( pt->sysid );
            bputc('\n');

            // 10/15/01 TD IR 3806
            // Error processing needs to be improved,
            // but that might happen in a rewrite instead.
            //
            return CHANGED_NO;
         }

			/* if we get to here it looks good! quit! */
			break;
		}

		/*
			if we searched the entire table and did not find one
			i will index past the end of the media.
		*/
		if(i >= SECSIZE)
			driveInfo[memunit].start = 0L;
		else
		{
			driveInfo[memunit].start = pt->pstart * SECSIZE;
		}

		/*
			Now read the BPB, use offset 0 since the read and write
			routines are now adding driveInfo[].start to it!
		*/
		memread(0L, DEF_SEC_SIZE, scratchBuffer);
		fmemmove((uchar far*)bpb, (uchar far *)scratchBuffer
									+ START_BPB, sizeof(struct BPB));

		return CHANGED_YES;
	}
	else
		return CHANGED_NO;
}


/*
   (Optional)
	This function need only be defined if TSRDEV is defined
	in MEMDISK.ASM.  TSRDEV allows the Generic Memory Disk
	to be installed via the command line (as well as from
	CONFIG.SYS).  If a user asks to uninstall the driver
	this routine is called to "close down" any hardware or
	system resources used by this driver.
*/
void memuninit(void)
{
	/* Normally never called! */
	Uninit365();
}


/*
   (Optional)
	This function need only be defined if GENIOCTL is
	defined in MEMDISK.ASM.  It processes generic IOCTL
	calls, allowing the outside world to communicate
	directly with the Generic Disk Driver through DOS calls.

	An Example of this can be found in VDISK.C, where the
	DOS FORMAT interface is supported (so FORMAT.COM works
	on the VDISK just like it does on floppies and hard
	disks).

	category   This has one of 3 meanings, depending on
	           which DOS call is asking for IOCTL.

	           For INT 21H, 440DH, Generic IOCTL, the category
	           will be the value passed in CX to DOS (usually
	           something like 08xxH).

	           For INT 21H, 4404H, Block IOCTL Read, the
	           category will be "BLOCK_IOCTL_READ", a define
	           found in MEMDISK.H.

	           For INT 21H, 4405H, Block IOCTL Write, the
	           category will be "BLOCK_IOCTL_WRITE", a define
	           found in MEMDISK.H.

   buffer     The buffer, in segment:offset form.

   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
#ifdef __BORLANDC__
#pragma argsused
#endif
bool memioctl(unsigned category, char far *buffer)
{
	ulong cylinders;
	unsigned length;

	ASSERT(buffer);

	if(CheckForChange(memunit))
		driveInfo[memunit].changed = TRUE;

	if(!CardDetect(memunit))
	{
		memerr = DEVERR_NOTREADY;
		return FALSE;
	}

	switch(category)
	{
		/*
			These functions are stubbed since there is no need to do anything
			else when formatting a RAM disk. Add additional functionality here
			as might be required.
		*/
		case 0x842:	/* format track */
		case 0x862: /* verify track */
			return TRUE;

		/* set device parmaters */
		case 0x840:
			/*
				Do nothing!  Shouldn't be using an invalid BPB since
				we always use the BPB thats already on the media.
				We already have it unless the media changed.
			*/
			return TRUE;

		/* get device parameters */
		case 0x860:
		{
			ulong numSectors;
			struct DeviceParams far* lpDevParams =
				(struct DeviceParams far*)(buffer);

			/* Check for and possibly record a disk change */
			if(memchanged(&mem_bpb[memunit]))
				driveInfo[memunit].changed = TRUE;

			/*
				FORMAT will take care of all neccessary re-initializing of the
				disk by writing directly to the FAT and ROOT DIR sectors.
			*/
			/* Use our BPB and MS-DOS Prog Ref v6.0 page 38 says bit 2 is on */
			lpDevParams->dpSpecFunc = 5; /* Shouldn't be changed */
			lpDevParams->dpDevType = 5; /* unknown block device */
			lpDevParams->dpDevAttr = 2; /* removable and change line */
			lpDevParams->dpCylinders = (ATACardInfo[memunit].cylinders - 1);

			/* user also wants a copy of the BPB */
			fmemmove((char far *)(&(lpDevParams->bpb)),
					(char far *)(&mem_bpb[memunit]), sizeof(struct BPB));

			return(TRUE);
		}

		/* Read a physical track */
		case 0x861:
		{
			ulong offset;
			struct RWBlock far * lpRWBlock =
				(struct RWBlock far *) (buffer);

			offset = (lpRWBlock->Head + 1) * (lpRWBlock->Cylinder + 1)
							* (mem_bpb[memunit].secPerTrak)
							+ (lpRWBlock->FirstSector);
			length = mem_bpb[memunit].bytesPerSector * lpRWBlock->Sectors;

			ASSERT(length);

			/*
				Read the real offset, DOS FORMAT may be looking for the
				MBR which DOS normally does not want or need.
			*/
			return memread(offset, length, lpRWBlock->lpDataBuffer);
		}

		/* Write a physical track */
		case 0x841:
		{
			ulong offset;
			struct RWBlock far * lpRWBlock =
				(struct RWBlock far *) (buffer);

			offset = (lpRWBlock->Head + 1) * (lpRWBlock->Cylinder + 1)
							* (mem_bpb[memunit].secPerTrak)
							+ (lpRWBlock->FirstSector);
			length = mem_bpb[memunit].bytesPerSector * lpRWBlock->Sectors;

			ASSERT(length);

			/*
				Read the real offset, DOS FORMAT may be looking for the
				MBR which DOS, during normal operations does not want!
			*/
			return memwrite(offset, length, lpRWBlock->lpDataBuffer);
		}

		/* Get Media ID */
		case 0x866:
		{
			struct MediaID far * pMediaID;
			struct boot_sector * pBootSector;

			/* Callee's media id buffer */
			pMediaID = (struct MediaID far *)buffer;
			pBootSector = (struct boot_sector far *)scratchBuffer;

			/* Read the BPB */
			if(!memread(0L, SECSIZE, scratchBuffer))
				return FALSE;

			/* Save the serial number, volume label, and ID string */
			pMediaID->infoLevel = 0; /* DOS prog ref says this is zero */
			pMediaID->serialNumber = pBootSector->serialNumber;
			fmemmove((uchar far*)pMediaID->volumeLable,
						(uchar far*)pBootSector->volumeLable, 11);
			fmemmove((uchar far*)pMediaID->idString,
						(uchar far*)pBootSector->idString, 8);

			return TRUE;
		}


		/* Set media ID */
		case 0x846:
		{
			struct MediaID far * pMediaID;
			struct boot_sector * pBootSector;

			/* Callee's media id buffer */
			pMediaID = (struct MediaID far *)buffer;
			pBootSector = (struct boot_sector far *)scratchBuffer;

			/* Read the BPB */
			if(!memread(0L, SECSIZE, scratchBuffer))
				return FALSE;

			pBootSector->serialNumber = pMediaID->serialNumber;
			fmemmove((uchar far*)pBootSector->volumeLable,
						(uchar far*)pMediaID->volumeLable, 11);
			fmemmove((uchar far*)pBootSector->idString,
						(uchar far*)pMediaID->idString, 8);

			/* Write the boot sector back into the media */
			return memwrite(0L, SECSIZE, scratchBuffer);
		}

		/*
			get access flag, sub-sub-function 0 returns 0 if
			the disk is not formatted else 1 if it is
		*/
		case 0x867:
		{
			struct SpecialFunc far * lpSpecFuncs =
				(struct SpecialFunc far*)(buffer);

			/*
				documentation says special function should ALWAYS be 0
				see pg 38 DOS Prog Ref
			*/
			ASSERT(lpSpecFuncs->sfFunction == sfFUNC_NUM0);

			lpSpecFuncs->sfAccessFlg = sfACCESS_OK;
			return(TRUE);
		}

		default:
			ERROR();
			break;

	} /* END switch statement */

	memerr = DEVERR_BADCMD;
	return FALSE;
}



/*
	Display routine.
*/
void bputs(char *s)
{
	ASSERT(s);

	while(*s)
		bputc(*s++);
}


/*
	Display a character.  Is smart about the C newline.
*/
void bputc(char c)
{
	if(c == '\n')
		bputc('\r');

	asm mov AL, c
	asm mov AH, 0EH
	asm mov BX, 0
	asm int 10H

}

/*
	Displays a 16-bit hex number in 0000 form.
*/
void bputh(unsigned val)
{
	char buff[5];
	int i;
	char c;

	for(i=3; i>=0; --i)
	{
		c = val % 0x10;
		c = (c >= 10 ? c-10+'A' : c+'0');
		buff[i] = c;
		val /= 0x10;
	}

	buff[4] = 0;
	bputs(buff);

}


/*
	Given a pointer to a string, evaluate the hex characters
	and translate them to an integer.  Return zero on error
	this will translate characters until a non-hex character
	found.
*/
unsigned atox(char far *s)
{
	unsigned temp = 0;
	char c;

	ASSERT(s);

	/* Continue until end of string or break */
	while(*s)
	{
		/* Convert to an upper case */
		if(*s >= 'a')
			c = *s - 0x20;
		else
			c = *s;

		/* Change the character into a number 0..F hex */
		if((c >= '0') && (c <= '9'))
		{
			temp *= 0x10;
			temp += c - '0';
		}
		else
			if((c >= 'A') && (c <= 'F'))
			{
				temp *= 0x10;
				temp += c - 'A' + 10;
			}
			else
				break;	/* Invlid character, we are done */

		/* Increment our string pointer */
		++s;
	}

	/* Return our number, zero on error */
	return temp;

}
