//
//   $Id: API.C 1.9 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  api.c		created mm/dd/yy  by Roy Sherrill
 *				updated 06/14/96	by John Burnham
 *
 *  Copyright (c) 1996-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *  The interface routines to the CDFS library
 *
 *  revision history
 *  11/14/96 jmb	Changed a call from cd_read_an to ReadCDData
 *  11/20/96 jmb	Updated cd_fread to be more efficient and to
 *					take advantage of the fact that ReadCDData will
 *					return all of the requested bytes.
 *  12/04/96 jmb	Changed CD_GFIRST to check for the root director
 *					and to remove the path from the passed string.
 *  01/20/97 jmb	Added code to check return code more specifically
 *					on return from ReadCDData.
 *  02/13/97 jmb Changed a the loop in cd_u_open that finds the
 *					directory we are trying to open a file in from
 *					using the length of the name as a comparison
 *					and then comparing the strings to just
 *					comparing the strings.  I am also using a new
 *					routine called NameMatch that will take into
 *					account a wierd feature on some CDROMs that put
 *					a period after a filename without an extension.
 *  04/08/02 jmb Added code to check the path of the
 *             file passed into the open/extended open
 *             against the data from the last FindFirst/Next
 *             and if it matches use that data.
 *  05/20/2002 jmb   Added further check around open code to see
 *                   if the FindFirst/Next is closed.
 *********************************************************/


/*
**************************************************************************
^ CD_CLOSE -    Close the file indicated. Deallocate memory and data structures
^        allocated when opened and remove file from list of open files
^        for the file's owner drive data structure.
^
^ Summary -
^        #include <cdstruct.h>
^         VOID cd_close(fileno)
^         CDFD fileno; <-------index to file control structure
^         {
^
^ Returns -    nothing
^
**************************************************************************
*/
#define __FILENUM__ 10

#include "datalght.h"

#include "cdstruct.h"
#include <stdio.h>
#include <string.h>
#include "cdrom.h"


//
// prototype includes
//
#include "api.pt"
#include "cdutil.pt"
#include "cdmemory.pt"
#include "lowl.pt"
#include "util.pt"
#include "cddev.pt"

//
// external data
//
extern char firstCDROMDriveLetter;
extern int numDrives;
extern DSTAT findFirstNextStatus;
extern uchar far *lpSDAFileName1; 		// File Name array 1 in DOS's SDA
extern TEXT findFirstMatch[];

//
// external procedures
//
extern BOOL PASCAL NameMatch(char far *lpDOSName, char far *lpCDName, int iMaxSrchlen);


VOID PASCAL cd_close(CDFD fileno)                                        /*__fn__*/
{
	IMPORT CDFILE *file_array[];
	CDFILE *file,*temp;
	CDDRIVE *drive;

	// reset value for cd_errno
	cd_errno = 0;

	//
	// test fileno for range and open. If error, just return
	//
	if ((fileno < 0) || (fileno >= MAXFILES))
		return;

	if (file_array[fileno] == NULL)
		return;

	//
	// initialize some pointers
	//
	file=file_array[fileno];
	drive=file->owner_drive;


	//
	// test first file in drive structure. If none, just return.
	// If this file should be closed, do it. Otherwise, initialize loop
	// to find the file to be closed.
	//
	if (drive->first_file == NULL)
		return;

	else if (drive->first_file == file)
	{
		drive->first_file = file->next_file;
		cd_mem_file(file, 0);
		file_array[fileno] = NULL;
		return;
	}
	else
		temp = drive->first_file;


	//
	// loop through file list of drive looking for this file. When found,
	// close it. If not found, just return.
	//
	while (temp != NULL)
	{
		if (temp->next_file == file)
		{
			temp->next_file = file->next_file;
			cd_mem_file(file, 0);
			file_array[fileno] = NULL;
			return;
		}
		else
			temp=temp->next_file;
	}
	return;
}

/*
**************************************************************************
^ CDGCWD -    Build a path specifier for the current working directory.
^        The path string passed in should include the drive spec
^        and should be large enough to hold the path string built by
^        this function. The returned string will be null terminated.
^        If a drive spec is not passed in, the default drive for the
^        user will be used.
^        The function retruns "YES" if the build is successful,
^        otherwise "NO".
^
^ Summary -
^        #include <cdstruct.h>
^
^         BOOL cd_gcwd(path)
^         TEXT *path; <-------path specifier. Drive passed in, CWD
^                    path passed back.
^         {
^
^        or...
^
^         BOOL cd_u_gcwd(userid,path)
^         COUNT userid; <-------user id
^         TEXT *path;   <-------path specifier. Drive passed in, CWD
^                    path passed back.
^         {
^
^
^ Returns -    YES if path build is successful
^        NO if path build is unsuccessful
^
**************************************************************************
*/


//
// routine to build path spec by recursive calls to get a records parent
//   until the root dir record is found.
//

//TEXT *build_path(TEXT *path, UCOUNT recno, CDFILE *file, GEN_PATH *gen_rec) /*__fn__*/
//{
//	UCOUNT i;
//	UCOUNT depth;
//	UCOUNT rec_id[8];
//	TEXT *ret_path;
//
//	ASSERT((path != NULL) && (file != NULL) && (gen_rec != NULL));	// validate parameters
//
//	ret_path = path;
//	depth = 0;
//	while (recno!=1)
//	{
//		rec_id[depth++] = recno;
//		if (get_abs_path(recno,file,gen_rec) < 1L)
//			break;
//
//		recno = gen_rec->parent_rec;
//	}
//	*path++ = BACKSLASH;
//
//	while (depth>0)
//	{
//		if (get_abs_path(rec_id[--depth],file,gen_rec) < 1L)
//			break;
//
//		for (i=0;i<gen_rec->len_id;i++)
//		{
//			*path++ = gen_rec->dir_id[i];
//		}
//
//		if (depth)
//			*path++ = BACKSLASH;
//
//	}
//	*path = '\0';
//	return(ret_path);
//}
//
//
////
//// build path specifier string for the specified user/drive.
////
//
//BOOL cd_u_gcwd( COUNT userid, TEXT *path)                       /*__fn__*/
//{
//	TEXT *curpath;
//	TEXT term;
//	COUNT length;
//	LONG n;
//	COUNT driveno;
//	CDDRIVE *drive;
//	UCOUNT recno;
//
//
//	//
//	// remove any leading spaces and convert path to upper case
//	//
//
//	curpath = path;
//	str2upper(curpath,path);
//	curpath = cd_dblank(curpath);
//
//	//
//	// begin parsing path. If a drive is not specified, use the default
//	// drive for the user. Check that the drive is valid and mounted for
//	// the user.
//	//
//
//	cd_parsep(curpath,&length,&term);
//	if (term == COLON)
//	{
//	    driveno = (COUNT) (*curpath - firstCDROMDriveLetter);
//	}
//	else
//	    driveno = cd_u_getdefltdrvno(userid);
//
//	drive=cd_u_drvno2drv(driveno,userid);
//	if (drive == NULL) return(NO);
//
//	//
//	// now that we have the drive spec, begin building the path
//	// string with the drive spec.
//	//
//
//	*path++ = (TEXT) (driveno + firstCDROMDriveLetter);
//	*path++ = COLON; 							// colon after drive
//	*path = '\0';
//
//	//
//	// initialize current path record to the default directory by
//	// matching CWD with a path table record
//	//
//
//	n=get_match_path(&drive->dflt_dir,drive->path_file,&drive->cur_path);
//	if (n < 1) return(NO);
//
//	//
//	// Build the path spec with a recursive path table fetch. Each time
//	// the build routine is called it fetches a higher level path spec
//	// until the root dir spec is found. As it unravels, it builds the
//	// path spec in correct order.
//	//
//
//	recno=drive->cur_path.cur_rec; 		// start from current path rec
//
//	path = build_path(path,recno,drive->path_file,&drive->cur_path);
//
//	return(YES);
//}
//
////
//// build path specifier string for the default user(0) and specified drive
////
//
//BOOL cd_gcwd(TEXT *path)                                        /*__fn__*/
//{
//	ASSERT(path != NULL);					// validate parameter
//	return (cd_u_gcwd(0,path));
//}

/*
**************************************************************************
^ CDGDONE -    Finish a directory retrieval operation by closing the file
^        allocated for the operation.
^
^ Summary -
^        #include <cdstruct.h>
^
^         VOID cd_gdone(stat)
^         DSTAT *stat;  <-----directory entry statistic returned if
^                     match found
^         {
^
^Returns -    nothing
^
**************************************************************************
*/

//
//  complete directory match operation by closing the directory file
//

VOID PASCAL cd_gdone(DSTAT *stat)                                      /*__fn__*/
{
	ASSERT(stat != NULL);					// validate parameter

	if (stat->fileno != -1)
	{
		cd_close(stat->fileno);
		stat->fileno = -1;
	}
}

/*
**************************************************************************
^ CDGFIRST -     Attempt to retrieve the first directory entry that matches a
^        pattern. The path/pattern is parsed to the lowest directory
^        level. A file is opened for the directory using one of the
^        global file slots. The directory pointed to by this file is
^        then searched for the match pattern. The file pointer is
^        updated for subsequent calls. Since the drive must be open
^        by a user before any file open operations are performed,
^        a user specific version of this routine is provided. If the
^        default user (0) is assumed but the default user (0) has not
^        opened the drive, the match will return false.
^
^ Summary -
^        #include <cdstruct.h>
^
^         BOOL cd_gfirst(stat,path)
^         DSTAT *stat;  <-----directory statistic entry returned if
^                     match found
^         TEXT *path;   <-----path/match specifier
^         {
^
^    or...
^
^         BOOL cd_u_gfirst(stat,path,userid)
^         DSTAT *stat;  <-----directory statistic entry returned if
^                     match found
^         TEXT *path;   <-----path/match specifier
^         COUNT userid; <-----user id for file open
^         {
^
^
^Returns -    YES - if match found,
^            file index of directory file in arg "stat.fileno"
^            and generic format directory entry in arg "stat.dir".
^        NO  - if match not found or some error encountered.
^
**************************************************************************
*/

//
//  get first directory entry directory for a file matching 'path'
//

//BOOL cd_gfirst(DSTAT *stat, TEXT *path)                               /*__fn__*/
//{
//
//	ASSERT((stat != NULL) && (path != NULL));			// validate parameter
//
//	stat->fileno=cd_dopen(path);
//	return(cd_gnext(stat,path));
//}


//
//  get first directory entry directory for a file matching 'path' for
//  the specified user
//

BOOL PASCAL cd_u_gfirst(DSTAT *stat, TEXT *path, COUNT userid)              /*__fn__*/
{
	TEXT *curpath;
	TEXT *newpath;
	TEXT term;
	COUNT length,i;

	ASSERT((stat != NULL) && (path != NULL));			// validate parameters

	stat->fileno=cd_u_dopen(userid,path);

	//
	// we are counting the number of "\"'s that are contained in the path
	// to see if we are in the root or not
	// We are also finding the search pattern.
	//
	i = 0;
	stat->rootDirectorySearch = YES;
	curpath = path;
	newpath = cd_parsep(curpath,&length,&term);
	while (term != 0x00)
	{
		if (term == '\\')
			i++;
		curpath = newpath;
		newpath = cd_parsep(curpath,&length,&term);
	}

	//
	// we found more than 1 backslash
	//
	if (i > 1)
		stat->rootDirectorySearch = NO;




	return(cd_gnext(stat,curpath));
}

/*
**************************************************************************
^ CDGNEXT -     Attempt to retrieve the next directory entry that matches a
^        pattern. File pointer is updated for subsequent calls.
^
^ Summary -
^        #include <cdstruct.h>
^
^         BOOL cd_gnext(stat,path)
^         DSTAT *stat;  <-----directory entry statistic returned if
^                     match found
^         TEXT *path;   <-----path/match specifier
^         {
^
^Returns -    YES - if match found
^            updated file pointer for directory file
^            and generic format directory entry in arg 'stat.dir'.
^        NO  - if match not found or some error encountered.
^
**************************************************************************
*/

//
//  get next directory entry in directory file matching 'path'
//

BOOL PASCAL cd_gnext(DSTAT *stat, TEXT *path)                                /*__fn__*/
{
	TEXT *curpath;
	TEXT term;
	CDFILE *file;
	LONG n;
	COUNT length, i;



	ASSERT((stat != NULL) && (path != NULL));			// validate parameters



	//
	// convert file index to data structure pointer and test it
	//
	file = cd_fileno2file(stat->fileno);

	if (!file)
		return(NO);

	if(!file->owner_drive->cur_vtoc)
		return(NO);

	// obtain length of path
	curpath = path;
	cd_parsep(curpath,&length,&term);


	//
	// fetch directory records until a match is found
	//

	do
	{

		n=get_next_dir(file,&stat->dir);

		if (n <= 0)
			return(NO);

		if (stat->rootDirectorySearch)
		{
			//
			// we have just read the current directory directory record
			// now we must discard it and read the parent directory directory record
			// and discard it.  Finally we read the first directory record
			// that we want to check for comparison against the user's request
			//
			if (stat->dir.file_id[0] == 0x00)
			{
				n=get_next_dir(file, &stat->dir);
				if (n <= 0)
					return NO;

				n=get_next_dir(file, &stat->dir);
				if (n <= 0)
					return NO;
			}
		}



		i = stat->dir.len_fi;
	}
	while (!cd_pathcp(&stat->dir.file_id[0],i,path,length));


	if (stat->dir.file_id[0] == 0x00 || stat->dir.file_id[0] == 0x01)
	{
		//
		// this is the current directory listing
		// if we aren't in the root, substitute the appropriate filename
		// otherwise, ignore it
		//


		if (stat->dir.file_id[0] == 0x00 && !stat->rootDirectorySearch)
		{
			stat->dir.file_id[0] = '.';
			stat->dir.file_id[1] = '\0';
		}


		//
		// this is the parent directory listing
		// if we aren't inthe root, substiture the appopriate filename
		// otherwise, ignore it
		//
		if (stat->dir.file_id[0] == 0x01 && !stat->rootDirectorySearch)
		{
			stat->dir.file_id[0] = '.';
			stat->dir.file_id[1] = '.';
			stat->dir.file_id[2] = '\0';
		}
	}

	return(YES);
}


/*
**************************************************************************
^ CDLSEEK -    Move the file pointer 'offset' bytes from the origin
^        specified in argument origin. Three origins are defined
^        as follows:
^            PSEEK_SET - offset from beginning of file
^            PSEEK_CUR - offset from current file position
^            PSEEK_END - offset from end of file(backwards)
^        The new file position is returned after the seek operation.
^        If the seek attempts to seek past the end of the file, the
^        file length is returned and the pointer is set to 1 byte
^        beyond the end of the file. If the seek attempts to seek before
^        the beginning of the file, 0 is returned and the pointer is
^        set to the beginning of the file. File position pointers are
^        zero based.
^
^ Summary -
^        #include <cdstruct.h>
^
^        LONG cd_lseek(fileno,offset,origin)
^        CDFD fileno;   <--index to global file descriptor array
^        LONG offset;    <--# bytes to seek from origin
^        COUNT origin;   <--start of seek
^        {
^
^    or...
^
^        LONG cd_lfseek(file,offset,origin)
^        CDFILE *file;   <-- file control block pointer
^        LONG offset;    <--# bytes to seek from origin
^        COUNT origin;   <--start of seek
^        {
^
^
^Returns -  -1 on error
^           or
^           updated file position pointer
^           If error is returned cd_errno will contain
^             CDE_VTOC  - if VTOC not present or not recognized
^             CDE_INDEX - if illegal file index
^
**************************************************************************
*/

//
// seek to file position for the indicated file
//

LONG PASCAL cd_lfseek(CDFILE *file, LONG offset, COUNT origin)              /*__fn__*/
{
	LONG n;

	//
	// test file pointer
	//
	cd_errno = 0;

	if (file == NULL)
	{
		cd_errno = CDE_INDEX;
		return(-1L);
	}

	if (file->owner_drive->cur_vtoc == NULL)
	{
		cd_errno = CDE_VTOC;
		return(-1L);
	}

	//
	// position pointer offset bytes from beginning of file
	//
	if (origin == PSEEK_SET)
		n=offset;

	//
	// position pointer offset bytes from current file position
 	//
	else if (origin == PSEEK_CUR)
		n=offset + file->pos;

	//
	// position pointer offset bytes from end of file
	//
	else if (origin == PSEEK_END)
		n=(LONG)(file->dir.data_len) - offset;

	//
	// return current position unchanged if origin not recognized
	//
	else
		n=file->pos;

//	if (n < 0)
//		n=0;

//	if (n > (LONG)file->dir.data_len)
//		n= (LONG)file->dir.data_len;

	file->pos = n;
	return(n);
}

//
// seek to file position for the indicated file number
//

//LONG cd_lseek(CDFD fileno, LONG offset, COUNT origin)               /*__fn__*/
//{
//	//
//	// convert file index to data structure pointer
//	//
//	return (cd_lfseek(cd_fileno2file(fileno),offset,origin));
//}

/*
**************************************************************************
^ CDOPEN -     Attempt to open and initialize a file control data structure
^        and return a file index number into the global drive array.
^        Two open routines are included, one for the default user (0)
^        and one for a specified user.
^        If the routine fails a negative index is returned. The
^        path specifier passed in as an argument defines the file to
^        be opened. If a drive is not specified, the default drive
^        for the user. If the path does not include a root specifier,
^        the path is considered relative to the default directory for
^        the drive.
^
^
^ Summary -
^        #include <cdstruct.h>
^
^         CDFD cd_open(path)
^         TEXT *path; <-------path specifier
^         {
^
^        or...
^
^         CDFD cd_u_open(userid,path)
^         COUNT userid; <-------user id
^         TEXT *path;   <-------path specifier
^         {
^
^
^ Returns -
^        Index number to file control structure in global file
^        array if open successful or -1.
^        If open failed cd_errno will contain one of the following
^
^            CDE_EMFILE  - if file block init unsuccessful
^            CDE_DRIVENO - if drive number invalid or not mounted
^            CDE_ENOENT  - if path entry not found
^            CDE_INVAL   - if invalid path format
^
**************************************************************************
*/

//
// open specified file for specified user.
//

CDFD PASCAL cd_u_open(COUNT userid, TEXT *path)                            /*__fn__*/
{
	TEXT buf[257];
   TEXT *filename;
	TEXT *curpath;
	TEXT *newpath;
	TEXT term;
	COUNT length,i;
	LONG n;
	COUNT driveno;
	CDDRIVE *drive;
	UCOUNT recno;
	CDFD fd;

	ASSERT(path != NULL);					// validate parameter


	//
	// remove any leading spaces and convert path to upper case
	//

	cd_errno = 0;
	curpath = &buf[0];
   strcpy(curpath, path);


   // only use the FindFirst/Next structure if it is still open
   if (findFirstNextStatus.fileno != -1)
   {
      filename = strrchr(curpath, BACKSLASH);

      if (filename)
      {
         filename++;
      
         //
         // we need to check the directory somehow...
         //
         // this will cause problems if the following path was found in 
         // the last findfirst/next:
         //
         //  d:\path\a\file.txt
         //
         // and the path sent to open was
         //
         //  d:\path\b\file.txt
         //

         if (strcmp(curpath, findFirstMatch) == 0)
   //      if (strcmp(findFirstNextStatus.dir.file_id, filename) == 0)
         {
            driveno = GetInternalDriveNum(*curpath - 'A');
   	      drive=cd_u_drvno2drv(driveno,userid);

            memcpy(&(drive->cur_dir), &findFirstNextStatus.dir, sizeof(DIR_REC));
            fd = cd_fopen_cur(drive);

   	      if (fd < 0)
	      	   cd_errno = CDE_EMFILE;

      	   return(fd);
         }
      }
   }

#if 0
   str2upper((UTINY *)curpath, (UTINY *)path);
	curpath = cd_dblank(curpath);
#endif


	//
	// begin parsing path. If a drive is not specified, use the default
	// drive for the user. Check that the drive is valid and mounted for
	// the user.
	//

	newpath=cd_parsep(curpath,&length,&term);
	if (term == COLON)
	{
		driveno = GetInternalDriveNum(*curpath - 'A');
		curpath=newpath;
		newpath=cd_parsep(curpath,&length,&term);
	}
	else
		driveno = cd_u_getdefltdrvno(userid);

	drive=cd_u_drvno2drv(driveno,userid);
	if (drive == NULL)
	{
		cd_errno = CDE_DRIVENO;
		goto ret_error;
	}

	//
	// initialize current path record to the default directory in case
	// no "\" are encountered.
	//

	drive->cur_path.loc_extent=drive->dflt_dir.loc_extent;
	drive->cur_path.XAR_len=drive->dflt_dir.XAR_len;
	drive->cur_path.len_id=drive->dflt_dir.len_fi;

	for (i=0;i<33;i++)
		drive->cur_path.dir_id[i]=drive->dflt_dir.file_id[i];

	//
	// position path table pointer for first access if current parse
	// field is "\" terminated. If length is 0, use root. Otherwise use
	// default directory.
	//

	if (term == BACKSLASH)
	{
		if (length == 0)
		{
			n = get_first_path(drive->path_file, &drive->cur_path);
			if (n <= 0)
			{
				cd_errno = CDE_ENOENT;
				goto ret_error;
			}
			curpath=newpath;
			newpath=cd_parsep(curpath,&length,&term);
		}
		else
		{
			n = get_match_path(&drive->dflt_dir,drive->path_file,&drive->cur_path);
			if (n <= 0)
			{
				cd_errno = CDE_ENOENT;
				goto ret_error;
			}
		}
	}


	//
	// continue parsing until all "\" terminated path fields are
	// exhausted
	//

	while (term == BACKSLASH)
	{
		if ((length == 1) && (curpath[0] == '.'))
		   n = drive->cur_path.cur_rec;
		else if ((curpath[0] == '.') && (curpath[1] == '.'))
		{

			n = get_abs_path(drive->cur_path.parent_rec,drive->path_file,&drive->cur_path);
			if (n <= 0)
			{
				cd_errno = CDE_ENOENT;
				goto ret_error;
			}

		}
		else
		{
			recno=drive->cur_path.cur_rec;
			do
			{
				n = get_next_path(drive->path_file,&drive->cur_path);
				if (n <= 0)
				{
					cd_errno = CDE_ENOENT;
					goto ret_error;
				}
			}
			while ( (drive->cur_path.parent_rec != recno) ||
					(!cd_pathcp(drive->cur_path.dir_id,
					  		drive->cur_path.len_id, curpath, length ) ));

//			        ((COUNT)drive->cur_path.len_id != length) ||
//			        !(cmpbuf(&drive->cur_path.dir_id[0], curpath,length)));
		}
		curpath=newpath;
		newpath=cd_parsep(curpath,&length,&term);
	}


	//
	// the only valid entry at this point is a null terminated file
	// specification. If no file spec is given at this point, use the
	// current file specified in the current directory record of the
	// drive data structure. Otherwise, open the directory pointed to
	// by the current path record search for the file spec.
	//

	if (term)
	{
		cd_errno = CDE_INVAL;
		goto ret_error;
	}

	if (length > 0)
	{
		drive->dir_file->dir.XAR_len=drive->cur_path.XAR_len;
		drive->dir_file->dir.loc_extent=drive->cur_path.loc_extent;
		drive->dir_file->dir.data_len=256;
		drive->dir_file->dir.file_flags=2;
		n=get_first_dir(drive->dir_file,&drive->cur_dir);

		if (n <= 0)
		{
			cd_errno = CDE_ENOENT;
			goto ret_error;
		}

		//  Don't use N_SCOPY
		// drive->dir_file->dir = drive->cur_dir;
		memcpy( &drive->dir_file->dir, &drive->cur_dir, sizeof(drive->cur_dir) );



   	n=get_match_dir(curpath,length,drive->dir_file,&drive->cur_dir);

   	if (n <= 0)
   	{
		   cd_errno = CDE_ENOENT;
		   goto ret_error;
	   }

   }
   
   fd = cd_fopen_cur(drive);

	if (fd < 0)
		cd_errno = CDE_EMFILE;

	return(fd);
ret_error:
    return(-1);
}

//
// open specified file for default user (0).
//

//CDFD cd_open(TEXT *path)                                            /*__fn__*/
//{
//    return (cd_u_open(0,path));
//}

/*
**************************************************************************
^ CDREAD -    Attempt to read a specified number of bytes from a file
^        into the specified buffer. If the read is successful, the
^        number of bytes read is returned. If the end of file is
^        encountered before the request is complete, the number of
^        bytes returned will be less than the request. If an error
^        occurs, the returned value will be less than 0 and will
^        contain an error code. The file position pointer is
^        updated.
^
^ Summary -
^        #include <cdstruct.h>
^
^
^        UCOUNT cd_read(fileno,buf,count)
^        CDFD fileno;   <--index to global file descriptor array
^        BUF_TYPE *buf;    <--read buffer
^        UCOUNT count;    <--length of read request
^        {
^
^    or...
^
^        UCOUNT cd_fread(fileno,buf,count)
^        CDFILE *file;   <--file descriptor pointer
^        BUF_TYPE  *buf;    <--read buffer
^        UCOUNT count;    <--length of read request
^        {
^
^
^ Returns -  length of read if successful ~0 if error
^        If error cd_errno will be set to one of the following.
^             CDE_VTOC  - if VTOC not present or not recognized
^             CDE_INDEX - if illegal file index
^             CDE_IOERR   - if read error
^             CDE_CHANGE  - if drive status error (media change)
^             CDE_EVOL    - if read past end of volume detected
^
**************************************************************************
*/


//
// read bytes from specified file
//

UCOUNT PASCAL cd_fread(CDFILE *file,  BUF_TYPE *buf,  UCOUNT count)             /*__fn__*/
{
	LONG n;


	ASSERT(buf != NULL);						// validate buffer

	//
	// test file structure pointer
	//

	if (file == NULL)
	{
		cd_errno = CDE_INDEX;
		goto ret_error;
	}

	if (file->owner_drive->cur_vtoc == NULL)
	{
		cd_errno = CDE_VTOC;
		goto ret_error;
	}

	n=ReadCDData(file, buf, count, file->pos);

	if (n < 0)
		return (COUNT) n;

	if (n == 0)
	{
		// if we just hit end of file, return now
		if (cd_errno == CDE_OKAY)
			return 0;

		// check the status of the drive -- determine if it was a bad disk
		// or if the disk was removed.
		if (CDChanged(file->owner_drive->drive) == YES)
			cd_errno = CDE_CHANGE;				// the drive was not ready
		else
			cd_errno = CDE_IOERR;				// the read just failed
		goto ret_error;
	}

	file->pos += n;
	return (COUNT) n;

ret_error:
	return(~0);
}

//
// read bytes from specified file number
//

//COUNT cd_read(COUNT fileno, BUF_TYPE *buf, UCOUNT count)                  /*__fn__*/
//{
//	ASSERT(buf != NULL);						// validate parameter
//	return(cd_fread(cd_fileno2file(fileno),buf,count));
//}

/*
**************************************************************************
^ CDSCWD -    Set current working directory to the directory specified
^        by the path specifier. If a drive specifier is included in
^        the path, the cwd for that drive is set, otherwise the cwd
^        for the default drive for the user is set. This routine does
^        not set default drive.
^        Two routines are included, one for the default user (0)
^        and one for a specified user.
^
^ Summary -
^        #include <cdstruct.h>
^
^         BOOL cd_scwd(path)
^         TEXT *path; <-------path specifier
^         {
^
^        or...
^
^         BOOL cd_u_scwd(userid,path)
^         COUNT userid; <-------user id
^         TEXT *path;   <-------path specifier
^         {
^
^
^ Returns -    NO - if current working directory not set.
^        YES - if current working directory set.
^
**************************************************************************
*/


//
// set current working directory for specified user.
//

BOOL PASCAL cd_u_scwd(COUNT userid, TEXT *path)                            /*__fn__*/
{
	TEXT buf[257];
	TEXT *curpath;
	TEXT *newpath;
	TEXT term;
	COUNT length;
	LONG n;
	COUNT driveno;
	CDDRIVE *drive;
	UCOUNT recno;


	ASSERT(path != NULL);					// validate parameter


	//
	// remove any leading spaces and convert path to upper case
	//

	curpath = &buf[0];
	str2upper(curpath,path);
	curpath = cd_dblank(curpath);

	//
	// begin parsing path. If a drive is not specified, use the default
	// drive for the user. Check that the drive is valid and mounted for
	// the user.
	//

	newpath=cd_parsep(curpath,&length,&term);
	if (term == COLON)
	{
		driveno = GetInternalDriveNum(*curpath - 'A');
		curpath=newpath;
		newpath=cd_parsep(curpath,&length,&term);
	}
	else
		driveno = cd_u_getdefltdrvno(userid);

	drive=cd_u_drvno2drv(driveno,userid);
	if (!drive)
		return(NO);

	//
	// position path table pointer for first access:
	// if current parse field is "\" terminated and length is 0, use root.
	// Otherwise use default directory.
	//

	if ((term == BACKSLASH) && (length == 0))
	{
		n = get_first_path(drive->path_file,&drive->cur_path);
		if (n <= 0)
			return(NO);

		curpath=newpath;
		newpath=cd_parsep(curpath,&length,&term);
	}

	else
	{
		n = get_match_path(&drive->dflt_dir, drive->path_file, &drive->cur_path);
		if (n <= 0)
			return(NO);
	}

	//
	// continue parsing until all "\" terminated and nul terminated path
	// fields are exhausted
	//

	while ((term == BACKSLASH) || (!term))
	{
		if ((length == 1) && (curpath[0] == '.'))
			n = drive->cur_path.cur_rec;

		else if ((curpath[0] == '.') && (curpath[1] == '.'))
		{
			n = get_abs_path(drive->cur_path.parent_rec,drive->path_file,&drive->cur_path);
			if (n <= 0) return(NO);
		}

		else if(length > 0)
		{
			recno=drive->cur_path.cur_rec;
			do
			{
				n = get_next_path(drive->path_file,&drive->cur_path);
   			if (n <= 0) return(NO);
			}
			while ( (drive->cur_path.parent_rec != recno) ||
			        ((COUNT) drive->cur_path.len_id != length) ||
//			       !(cmpbuf((UTINY *)&drive->cur_path.dir_id[0],(UTINY *)curpath,length)));
			       !(cmpbuf(&drive->cur_path.dir_id[0], curpath, length)));
		}

		//
		// last path spec found when null terminator reached.
		//
		if (!term)
			break;

		curpath=newpath;
		newpath=cd_parsep(curpath,&length,&term);
	}

	//
	// the only valid entry at this point is a null terminated path entry
	// specification.
	//

	if (term)
		return(NO);

	drive->dir_file->dir.XAR_len=drive->cur_path.XAR_len;
	drive->dir_file->dir.loc_extent=drive->cur_path.loc_extent;
	drive->dir_file->dir.data_len=256;
	drive->dir_file->dir.file_flags=2;
	n=get_first_dir(drive->dir_file,&drive->dflt_dir);

	if(n <= 0)
		return(NO);

	return(YES);
}

//
// set current working directory for default user (0).
//

//BOOL cd_scwd(TEXT *path)                                            /*__fn__*/
//{
//	ASSERT(path != NULL);					// validate parameter
//	return (cd_u_scwd(0,path));
//}

/*
**************************************************************************
^ DSKABORT -     Close the indicated CD-ROM device for all users
^        and free all resources allocated. A second function
^        is included to close all drives and then deallocate all
^        resources including block buffer space.
^
^ Summary -
^        #include <cdstruct.h>
^         VOID cd_dskabort(path)
^         TEXT *path;
^         {
^
^    or...
^
^        #include <cdstruct.h>
^         VOID cd_dskaborta()
^         {
^
^ Returns -     nothing
^
**************************************************************************
*/

//
// close all cd drives for all users and deallocate buffers
//

//VOID cd_dskaborta()                                                 /*__fn__*/
//{
//	COUNT driveno;
//	CDDRIVE *drive;
//	COUNT userid;
//	IMPORT BUFCTL *broot;
//
//	for (driveno=0;driveno<MAXDRIVES;driveno++)
//	{
//		for (userid=0;userid<MAXUSERS;userid++)
//		{
//			drive = cd_u_drvno2drv(driveno,userid);
//			while(drive !=NULL)
//			{
//				cd_cddismount(driveno,userid);
//				drive = cd_u_drvno2drv(driveno,userid);
//			}
//		}
//	}
//	cd_blkfree(broot);
//}
//
////
//// close cd drive indicated for all users
////
//
//VOID cd_dskabort(TEXT *path)                                          /*__fn__*/
//{
//	TEXT buf[257];
//	TEXT *curpath;
//	TEXT term;
//	COUNT length;
//	COUNT driveno;
//	CDDRIVE *drive;
//	COUNT userid;
//
//
//
//	ASSERT(path != NULL);					// validate parameter
//
//
//
//	//
//	// remove any leading spaces from path and convert to upper case
//	//
//
//	curpath = &buf[0];
//	str2upper(curpath,path);
//	curpath = cd_dblank(curpath);
//
//
//	//
//	// parse first field. If it is a drive spec, close the drive.
//	// otherwise, just return
//	//
//
//	cd_parsep(curpath,&length,&term);
//	if ((term == COLON) && (length > 0))
//	{
//		driveno = (COUNT) (*curpath - firstCDROMDriveLetter);
//		for (userid=0;userid<MAXUSERS;userid++)
//		{
//			drive = cd_u_drvno2drv(driveno,userid);
//			while(drive !=NULL)
//			{
//				cd_cddismount(driveno,userid);
//				drive = cd_u_drvno2drv(driveno,userid);
//			}
//		}
//	}
//}

/*
**************************************************************************
^ DSKCLOSE -     Close the indicated CD-ROM device for the default
^        user id (0) or the indicated user. If the user count is still
^        greater than 0 for the user/drive combination, the count is
^        decremented and all resources allocated to the user are freed.
^        If no more users open the drive, all drive resources are freed.
^        Parse the indicated path and extract the drive letter, then
^        call the lower level cd_dismount routine.
^
^ Summary -
^        #include <cdstruct.h>
^         VOID cd_dskclose(path)
^         TEXT *path;
^         {
^
^    or...
^
^        #include <cdstruct.h>
^         VOID cd_u_dskclose(userid,path)
^         TEXT *path;
^         {
^
^ Returns -    NO if mount unsuccessful
^        YES if mount successful
^
**************************************************************************
*/

//
// close specified cd drive for specified user
//

//VOID cd_u_dskclose( COUNT userid, TEXT *path)
//{
//	TEXT buf[257];
//	TEXT *curpath;
//	TEXT term;
//	COUNT length;
//	COUNT driveno;
//
//
//
//	ASSERT(path != NULL);					// validate parameter
//
//
//	//
//	// remove any leading spaces from path and convert to upper case
//	//
//
//	curpath = &buf[0];
//	str2upper(curpath,path);
//	curpath = cd_dblank(curpath);
//
//	//
//	// parse first field. If it is a drive spec, close the drive.
//	// otherwise, just return
//	//
//
//	cd_parsep(curpath,&length,&term);
//	if ((term == COLON) && (length > 0))
//	{
//		driveno = (COUNT) (*curpath - firstCDROMDriveLetter);
//		cd_cddismount(driveno,userid);
//	}
//}
//
////
//// close specified cd drive for default user (0).
////
//
//VOID cd_dskclose(TEXT *path)                        /*__fn__*/
//{
//	ASSERT(path != NULL);					// validate parameter
//	cd_u_dskclose(0,path);
//	return;
//}
//

/*
**************************************************************************
^ DSKOPEN -      Attempt to open the indicated CD-ROM device for the default
^        user id (0) or the indicated user with the standard VTOC.
^        Parse the indicated path and extract the drive letter, then
^        call the lower level cdmount routine.
^
^ Summary -
^        #include <cdstruct.h>
^         BOOL cd_dskopen(path)
^         TEXT *path;
^         {
^
^    or...
^
^        #include <cdstruct.h>
^         BOOL cd_u_dskopen(userid,path)
^         TEXT *path;
^         {
^
^ Returns -    NO if mount unsuccessful
^        YES if mount successful
^
**************************************************************************
*/


//
// open specified cd drive for specified user
//

BOOL PASCAL cd_u_dskopen(COUNT userid, TEXT *path)                 /*__fn__*/
{
	TEXT buf[257];
	TEXT *curpath;
	TEXT term;
	COUNT length;
	COUNT driveno;
	UTINY vtoc_type = 1;    				// open for standard VTOC


	ASSERT(path != NULL);					// validate parameter


	//
	// remove any leading spaces from path and convert to upper case
	//

	curpath = &buf[0];
	str2upper(curpath,path);
	curpath = cd_dblank(curpath);

	//
	// parse first field. If it is a drive spec, mount the drive.
	// otherwise, return error
	//

	cd_parsep(curpath,&length,&term);
	if ( (term == COLON) && (length > 0) )
	{
		driveno = GetInternalDriveNum((*curpath - 'A'));
		if (cd_cdmount(driveno,vtoc_type,userid) &&
			 cd_u_setdefltdrvno(driveno,userid) )
			return (YES);
	}
	return NO;
}

//
// open specified cd drive for default user (0).
//

//BOOL cd_dskopen(TEXT *path)                                        /*__fn__*/
//{
//	ASSERT(path != NULL);					// validate parameter
//	return(cd_u_dskopen(0,path));
//}



/*
**************************************************************************
^ CDDRIVES -    Various CD rom drives functions including:
^           ->get default drive number for default (0) user
^           ->get default drive number for specified user
^           ->set default drive number for default (0) user
^           ->set default drive number for specified user
^           ->convert drive number to drive data structure pointer
^             for default (0) user
^           ->convert drive number to drive data structure pointer
^             for specified user
^           ->convert file number to file data structure pointer
^           ->convert drive/path specifier to a drive number
^
^ Summary -
^        #include <cdstruct.h>
^
^
^        COUNT cd_getdefltdrvno()
^        {
^        Returns - default drive number for default (0) user.
^
^        or...
^
^        COUNT cd_u_getdefltdrvno(userid)
^        COUNT userid;
^        {
^        Returns - drive number for specified user, or 0 if user is
^              not defined.
^
^        or...
^
^        VOID cd_setdefltdrvno(drive)
^        COUNT drive;
^        {
^
^        or...
^
^        VOID cd_u_setdefltdrvno(drive,userid)
^        COUNT drive;
^        COUNT userid;
^        {
^
^        or...
^
^        CDDRIVE *cd_drvno2drv(drive)
^        COUNT drive;
^        {
^        Returns - Pointer to drive structure for specified drive
^              number for default (0) user.
^              Return NULL if drive number invalid or drive is
^              not mounted.
^
^        or...
^
^        CDDRIVE *cd_u_drvno2drv(userid,drive)
^        COUNT userid;
^        COUNT drive;
^        {
^        Returns - Pointer to drive structure for specified drive
^              number for default (0) user.
^              Return NULL if drive number invalid or drive is
^              not mounted.
^
^        or...
^
^        CDFILE *cd_fileno2file(fileno)
^        COUNT fileno;
^        {
^        Returns - Pointer to file structure for specified file
^              number.
^              Return NULL if file number invalid or not open.
^
^        or...
^
^        COUNT cd_drv2drvno(path)
^        TEXT *path;
^        {
^        Returns - Drive number specified by path
^
**************************************************************************
*/

//
// return drive number for default (0) user.
//

//COUNT cd_getdefltdrvno()      /*__fn__*/
//{
//	IMPORT COUNT dfltdrv[];
//	return(dfltdrv[0]);
//}

//
// return drive number for specified user.
//

COUNT PASCAL cd_u_getdefltdrvno(COUNT userid)      /*__fn__*/
{
	IMPORT COUNT dfltdrv[];

	if (userid >= MAXUSERS)
		return(dfltdrv[0]);
	else
		return(dfltdrv[userid]);
}


//BOOL cd_set_default_drive(TEXT *path)                       /*__fn__*/
//{
//	ASSERT(path != NULL);					// validate parameter
//	return(cd_u_set_default_drive(0, path));
//}


//BOOL cd_u_set_default_drive(COUNT userid,TEXT *path)          /*__fn__*/
//{
//	COUNT drive;
//	TEXT *curpath;
//
//	ASSERT(path != NULL);					// validate parameter
//
//	curpath = path;
//	str2upper(curpath,path);
//	curpath = cd_dblank(curpath);
//
//	if (path[1] == ':')
//	{
//		if (path[0] <= 'z' && path[0] >= 'a')
//			drive = (COUNT) (path[0] - 'a');
//		else if (path[0] <= 'Z' && path[0] >= 'A')
//			drive = (COUNT) (path[0] - firstCDROMDriveLetter);
//		else
//			return(NO);
//	}
//
//	return(cd_u_setdefltdrvno(drive, userid));
//}


//
// set drive number for specified user.
//
BOOL PASCAL cd_u_setdefltdrvno(COUNT drive, COUNT userid)      /*__fn__*/
{
	IMPORT COUNT dfltdrv[];


	if (!cd_u_drvno2drv(drive, userid))
		return(NO);
	if (userid >= MAXUSERS)
		return(NO);;
//	if ((drive < 0) || (drive >=MAXDRIVES))
	if ((drive < 0) || (drive >= numDrives))
		return(NO);
	dfltdrv[userid] = drive;


	return(YES);
}


//
// set drive number for default (0) user.
//
//BOOL cd_setdefltdrvno(COUNT drive)      /*__fn__*/
//{
//	return(cd_u_setdefltdrvno(drive,0));
//}

//
// return pointer to drive data structure for the specified user,drive.
//
CDDRIVE * PASCAL cd_u_drvno2drv(COUNT drive, COUNT userid)      /*__fn__*/
{
//	IMPORT CDDRIVE *drv_array[];
	CDDRIVE *temp;

	//
	// test if drive number is valid
	//
//	if ((drive < 0) || (drive >=MAXDRIVES))
	if ((drive < 0) || (drive >= numDrives))
		return(NULL);

	//
	// scan drive list for this user
	//
	temp=drv_array[drive];
	while (temp != NULL)
		if (temp->userid == userid)
			return(temp);

	//
	// drive not mounted, return null
	//
	return(NULL);
}

//
// return pointer to drive data structure for the specified drive
// for default (0) user.
//

//CDDRIVE *cd_drvno2drv(COUNT drive)      /*__fn__*/
//{
//	return (cd_u_drvno2drv(drive,0));
//}

//
// return pointer to file data structure for the specified file
//

CDFILE * PASCAL cd_fileno2file(COUNT file)      /*__fn__*/
{
	IMPORT CDFILE *file_array[];

	//
	// test if file number is valid
	//
	if (( file< 0) || (file >= MAXFILES))
		return(NULL);

	return(file_array[file]);
}

//
// convert the drive/path text string to a drive number
//
//COUNT cd_drv2drvno(TEXT *path)      /*__fn__*/
//{
//	TEXT buf[257];
//	TEXT *curpath;
//	TEXT term;
//	COUNT length;
//
//
//	ASSERT(path != NULL);					// validate parameter
//
//	//
//	// remove any leading spaces from path and convert to upper case
//	//
//
//	curpath = &buf[0];
//	str2upper(curpath,path);
//	curpath = cd_dblank(curpath);
//
//	//
//	// parse first field. If it is a drive spec, mount the drive.
//	// otherwise, return error
//	//
//
//	cd_parsep(curpath,&length,&term);
//	if ((term == COLON) && (length > 0))
//	{
//		return((COUNT) (*curpath - firstCDROMDriveLetter) );
//	}
//	else
//		return (-1);
//}
