//
//   $Id: CDBOOT.C 1.6 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  CDBOOT.C	created	07/24/96	by John Burnham
 *
 *  Copyright (c) 1996-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *	CD ROM Driver Initialization file.  This file
 *	contains the initialzaition.
 *
 *	Revision History
 *
 * 10/08/98	bt	Make sure we set the drive letter for the cdrom
 *					in the device header.
 * 03/26/98 jb	Fixed a bug in the CDSAvailable routine that was using a
 *					global pointer instead of the local pointer to get the first
 *					CDS to start the counting.
 * 08/27/97 jb	Added a check for no command line parameters
 *					and if none are found, then usage information
 *					is displayed and we exit MSCDEX.
 * 12/04/96 jb	Changed code to conditionally check the
 *					mounting calls
 * 11/26/96 jb	Created a global variable in CDMAIN.C called
 *					lpOurCDS to make the CDS pointer a global
 *					variable.
 * 11/20/96 jb	Fixed some problems in the parameter
 *					processing code.
 *********************************************************/

#define __FILENUM__	13			// for assert macro

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <dir.h>
#include <ctype.h>
#include <string.h>
#include <mem.h>
#include <errno.h>

#include "datalght.h"

#include "cdrom.h"
#include "cddos.h"
#include "sft.h"
#include "doserr.h"
#include "cdstruct.h"
#include "xms.h"
#include "dosver.h"
#include "fatworld.h"

//
// prototype files
//
#include "cdboot.pt"
#include "cdasm.pt"
#include "cderr.pt"
#include "cdutil.pt"
#include "cdmemory.pt"
#include "cddev.pt"
#include "util.pt"
#include "xms.pt"


//
// external data
//
extern	int numDrives;
extern	char szDeviceName[];
extern  int cdOpened;
extern 	int currentDrive;
extern	char firstCDROMDriveLetter;
extern	int firstCDROMDriveNumber;
extern	int numBuffers;
extern	int bUseExtended;
extern  CDDRIVE *drive_core;
extern	struct CD_ROM_DEVHDR far *lpCDROMDevice;		// Pointer to CD-ROM Device Driver
extern char * Usage;


#if UNSUPPORTED
extern	int displayMemStats;
extern	int cdROMSharing;
extern	unsigned specialDebugSwitch;
extern	int supportKanji;
#endif

extern  unsigned localStackArray[];

//
// external procedures
//
extern void *AllocateBytes(unsigned bytes);
extern long far Next2fLabel;
extern void interrupt Int2FHandler();
extern unsigned GetCS(void);
extern unsigned GetDS(void);
extern void PASCAL ReleaseMemory(unsigned segmentOfBlock);

extern void interrupt(*next_2f)();

extern struct CDS far *lpOurCDS;

bool fIsFat32World;

void DisplaySignOnMessage( void )
{
	extern char * signon;

	if( fIsFat32World )
	{
		printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/////////////////////////////////////////////////////////////////////
// CD-ROM Driver main entry point.  Initializes for CD-ROM operations
// and becomes a TSR if conditions ok to run as a CD-ROM Drive.
//
unsigned main (int argc, char *argv[])
{
	unsigned  cdexLoadedFlag;
	struct PSP far *lpCDEXPSP;

	SetStackOverFlowCheck();

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* display sign-on */
	DisplaySignOnMessage();

	// correct dos version?
	if(DosMajorVersion() < 5)
		FatalError(cdrReqDOS5);

	// initialize the number of drives currently installed
	numDrives = 0;

	// if no command line parameters, then just print usage information
	if (argc <= 1)
	{
		printf("%s", Usage);
		exit(1);
	}



	if (argc > 1)
		ParseCommandLine(argc, argv);

	GetSDA();											// Get and setup the DOS SDA
	GetListOfListsPtr();

	if (CDSAvailable() < numDrives)
		FatalError(cdrNoDriveLetters);

	if (!cd_memory_init())
		FatalError(cdrInsufficientMEM);

	cdexLoadedFlag = CDIsInstalled();
	if (cdexLoadedFlag == CDEX_SIGNATURE)
		FatalError(cdrDriverAlreadyInst);

	else if (cdexLoadedFlag)
		FatalError(cdrMSCDEXLoaded);


	pLocalStack = (uchar *)&localStackArray[0] + STACK_SIZE - 6;


	/* Install the CD-ROM Driver */
	Install();


	for (currentDrive = 0; currentDrive < numDrives; currentDrive++)
	{
		drive_core[currentDrive].cdOpened = FALSE;
		if (CDChanged(currentDrive) == NO)
		{
			if ( CDOpenDrive(currentDrive))
				drive_core[currentDrive].cdOpened = TRUE;
		}
	}



	currentDrive = -1;

	next_2f = GetVector(0x2F);
	Next2fLabel = (long) next_2f;
	SetVector(0x2F, Int2FHandler);


	lpCDEXPSP = MK_FP(_psp,0);
	ReleaseMemory(lpCDEXPSP->pspEnvironment);

	CheckStackOverFlow();

	return TSR_NOW;
}


void far * SetVector( uchar ucInterrupt, void far * lpHandler )
{
	ulong far * lpulVector;
	void far * lpOldVector;

	lpulVector = MK_FP( 0, ((unsigned)ucInterrupt) * 4 );
	lpOldVector = (void far *)(*lpulVector);
	*lpulVector = (ulong)(lpHandler);
	return lpOldVector;
}


void far * GetVector( uchar ucInterrupt )
{
	ulong far * lpulVector;
	void far * lpOldVector;

	lpulVector = MK_FP( 0, ((unsigned)ucInterrupt) * 4 );
	lpOldVector = (void far *)(*lpulVector);
	return lpOldVector;
}


//
// ParseCommandLine processes the arguments passed to CDEX.
//
void PASCAL ParseCommandLine(int argc, char *cmd_line[])
{
	int i, j;
	char localbuf[5];
	char *currentParameter;
#if DEBUG
	int Flag;
#endif

	numBuffers = 4;


	for (i = 1; i < argc; i++)
	{
		currentParameter = cmd_line[i];
		while (*currentParameter)
		{
			if(*currentParameter == ' ' || *currentParameter == '\t')
			{
				currentParameter++;
				continue;
			}

			if(*currentParameter == '?')
				goto do_usage;

			if(*currentParameter  == '/' || *currentParameter == '-')
			{
				++currentParameter;
				switch (*currentParameter)
				{
					case 'd':								// Name Device Option
					case 'D':
					{
						int numUnits;
						CDDRIVE *pTempDrive;

						++currentParameter;
						if (*currentParameter != ':')
							FatalError(cdrInvalidOption, *(--currentParameter));
						else
							++currentParameter;

						for (j=0; j<8; ++j)
						{
							// exit the loop when we hit a null character or
							// any whitespace character
							if (!*currentParameter || isspace(*currentParameter))
								break;
							szDeviceName[j] = *currentParameter;
							currentParameter++;
						}

						// null terminate the device name
						szDeviceName[j] = '\0';

						// Initialize CD-ROM Drives
//						numUnits = CDDeviceInit(szDeviceName, lpDriverStrategy, lpDriverInterrupt);
						numUnits = CDDeviceInit(szDeviceName);

						// if no drives installed, fatal error and exit
						if (numUnits == 0)
							FatalError(cdrDeviceNotFound, szDeviceName);

						// allocate the number of units we have found so far
						pTempDrive = (CDDRIVE *)AllocateBytes(numUnits * sizeof(CDDRIVE));

						// if unable to allocate the space, print a message and exit
						if (!pTempDrive)
							FatalError(cdrNoDriveLetters);

						// clear the memory to nulls
						cd_memfill(pTempDrive, numUnits * sizeof(CDDRIVE), 0);


						// check to see if this is our first allocation
						if (numDrives == 0)
							// if so, then assign this memory to the root pointer
							drive_core = pTempDrive;

						// increase the installed drive number count
						numDrives += numUnits;

						// save pertinent information for each unit
						for (j = 0; j < numUnits; j++)
						{
							pTempDrive->unitNumber = j;
							pTempDrive->lpCDROMDevice = lpCDROMDevice;
							pTempDrive->lpDriverStrategy =
										MK_FP(FP_SEG(lpCDROMDevice), lpCDROMDevice->dhStrat);
							pTempDrive->lpDriverInterrupt =
										MK_FP(FP_SEG(lpCDROMDevice), lpCDROMDevice->dhInter);
							strcpy(pTempDrive->szDevice, szDeviceName);
							(char *)pTempDrive += sizeof(CDDRIVE);
						}
						break;
					}


#if UNSUPPORTED
					case 'f':
					case 'F':
						for (j=0; j<2 && isdigit(*currentParameter); ++j)
						{
							localbuf[j] = *currentParameter;
							currentParameter++;
						}

						localbuf[2] = '\0';
						maxOpenFiles = atoi(localbuf);

						if (maxOpenFiles == 0)
							FatalError(cdrInvalidMOptNum);

						break;
#endif

					case '?':								// Help command
					case 'h':
					case 'H':
					do_usage:
						printf("%s", Usage);
						exit(1);

#if UNSUPPORTED
					case 'k':								// Support Kanji volumes
					case 'K':
						supportKanji = TRUE;
						Warning(cdrNotYetImplemented,currentParameter);				// Notify user "NOT YET IMPLEMENTED! "
						break;
#endif

					case 'l':
					case 'L':
						++currentParameter;
						if (*currentParameter == ':')
								++currentParameter;
						else
							FatalError(cdrCDEXNotLoaded);

						firstCDROMDriveLetter = ToUpperCase(*currentParameter);
						firstCDROMDriveNumber = ToUpperCase(*currentParameter) - 'A';
						++currentParameter;
						break;

					case 'm':								// Specify number of sector buffers
					case 'M':
						++currentParameter;
						if (*currentParameter == ':')
							++currentParameter;
						else
							FatalError(cdrCDEXNotLoaded);

						j = 0;
						while (isdigit(*currentParameter) && j < sizeof(localbuf)-1)
//						for (j=0; j<2 && isdigit(*currentParameter); j++)
						{
							localbuf[j] = *currentParameter;
							currentParameter++;
							j++;
						}

						localbuf[j] = '\0';
						numBuffers = atoi(localbuf);

						if (numBuffers < 4)
							numBuffers = 4;

						break;

#if UNSUPPORTED
					case 's':
					case 'S':
						cdROMSharing = TRUE;
						Warning(cdrNotYetImplemented,currentParameter);				// Notify user "NOT YET IMPLEMENTED! "
						break;

					case 'v':
					case 'V':
						displayMemStats = TRUE;
						Warning(cdrNotYetImplemented,currentParameter);				// Notify user "NOT YET IMPLEMENTED! "
						break;
#endif

#if DEBUG
					case 'u':								// Unload CDEX
					case 'U':
						asm	mov	ax, 0x1599
						asm	int	0x2f
						asm	mov	Flag, ax
						if(Flag == 0)
							Message(cdrCDEXUnloaded);
						else
							Message(cdrCDEXNotLoaded);
						exit(1);
						break;

#endif

					case 'x':								// Use Extended memory for Buffers
					case 'X':
						if (HimemInstalled())
							bUseExtended = TRUE;

						else
						{
							bUseExtended = FALSE;
							Message(cdrNoXMSLoaded);
						}
					 	++currentParameter;
						break;



					default:
						goto bad_arg;


				}	// end switch
			} // end if

			else
			{
				bad_arg:
				FatalError(cdrInvalidOption, *currentParameter);
				exit(1);
			}
		} // end while
	}
}



//
// Install
// This function checks if there is enough room for one
// or more CD-ROM Drives and if so, then it sets the drive
// attributes for the drives, and changes the interrupt 2F vector.
//
void PASCAL Install(void)
{
	int i, numInstalled;

	// get a pointer to the list of current directory structures
	lpOurCDS = lpListOfLists->cds;

	// if the user requested a starting drive letter, start at that CDS
	if (firstCDROMDriveNumber > 0)
		(char far *)lpOurCDS += CDS_5_SIZE * firstCDROMDriveNumber;

	i = firstCDROMDriveNumber;
	numInstalled = 0;

	// search for free cds's to assign to cddrives.
	while (i <= lpListOfLists->lastDrive && numInstalled < numDrives)
	{
		// if this cds is not available, don't use it
		if (lpOurCDS->drvAttrib != 0)
		{
			firstCDROMDriveLetter++;
			firstCDROMDriveNumber++;
		}

		else
		{
			// install as our drive
			InstallDrive(lpOurCDS);

//			// if we need to modify the drive letter and number, do it here
//			if (firstCDROMDriveNumber  == 0 || (i > firstCDROMDriveNumber))
//			{
//				firstCDROMDriveLetter += i - firstCDROMDriveNumber;
//				firstCDROMDriveNumber = i;
//			}
//			printf("[%u] ", numInstalled);

			// store physical drive number
			drive_core[numInstalled].physDriveNumber = i;


			//  Drive letter(number) in device header is not
			//  zero based. (A=1,B=2,C=3,...)  Hence the +1.
			//
			lpCDROMDevice->dhCDDriveLtr = i+1;


			// tell the user that we just installed the drive
			Message(cdrInstalledAsDrive, firstCDROMDriveLetter+(i-firstCDROMDriveNumber),
						strupr(drive_core[numInstalled].szDevice),
						drive_core[numInstalled].unitNumber);

#if DEBUG
			printf("\n");
#endif
			// increment the counter that tells us how many drives we have
			// installed
			numInstalled++;
		}

		// move next cds
		i++;
		(char far *)lpOurCDS += CDS_5_SIZE;
	}

	// set the current drive to drive #0
	currentDrive = 0;


	// asserting for now.  We should really handle this case and not allow
	// installation to take place since we cannot get "drive letters" for
	// each drive unit

	ASSERT(numInstalled == numDrives);

}

//
// Install Drive actually modifies the DOS CDS structure to
// make the CD-ROM drives accessable!
//
int PASCAL InstallDrives(struct CDS far *cds)
{
	int i;


	for (i=0; i < numDrives; ++i)
	{
		InstallDrive(cds);
		(char far *)cds += CDS_5_SIZE;
	}
	return (TRUE);
}


//
// Install a specified drive!
// This function modifies DOS structures to make a drive a
// network drive.
//
void PASCAL InstallDrive(struct CDS far *cds)
{
	++lpListOfLists->blkDevCount;
	cds->drvAttrib = CDS_NETWORK | CDS_PHYSICAL;
	cds->cwd[0] = '\0';			// Init CWD to "E:\"
}


//
// this routine is used to check for a CDROM driver in memory.
//
unsigned PASCAL CDIsInstalled(void)
{
	unsigned RetValue;

	asm 	mov	ax, 0x150C;			// Get Version
	asm	mov	bx, 0
	asm	int	0x2f
	asm	cmp	cx, CDEX_SIGNATURE
	asm	jne	use_bx

	asm	mov	RetValue, cx
	asm	jmp	exit_now

	use_bx:
	asm	mov	RetValue, bx;

	exit_now:
	return (RetValue);
}


////
//// CD-ROM Driver Initalzation - Perform the initialization required
//// for the CD-ROM Driver.
////
//int PASCAL CDDriverInit(char *pDeviceName)
//{
//	int numDrives;
//
//	numDrives = CDDeviceInit(pDeviceName);
//
//	if(numDrives == 0)
//		FatalError(cdrDeviceNotFound, pDeviceName);
//
//	if (!cd_memory_init())
//		FatalError(cdrNoDriveLetters);
//
//	return(numDrives);
//}

//
// DosMajorVersion returns the major version of the currently running
// system
//
int PASCAL DosMajorVersion(void)
{
	// need to add code to correctly get the version from DOS
	return 5;
}


int CDSAvailable(void)
{
	int i;
	int numAvail;
	struct CDS far *lpCountCDS;

	// get a pointer to the list of current directory structures
	lpCountCDS = lpListOfLists->cds;

	// if the user requested a starting drive letter, start at that CDS
	if (firstCDROMDriveNumber > 0)
		(char far *)lpCountCDS += CDS_5_SIZE * firstCDROMDriveNumber;

	i = firstCDROMDriveNumber;

	// check to see if there are enough CDS's available to use as CDROM drives
	numAvail = 0;

	while (i <= lpListOfLists->lastDrive)
	{
		// if this CDS is available, count it
		if (lpCountCDS->drvAttrib == 0)
			numAvail++;

		// move next cds
		i++;
		(char far *)lpCountCDS += CDS_5_SIZE;

	}

 	// return number available CDSs
 	return numAvail;

}
