//
//   $Id: CDMAIN.C 1.12 2003/07/09 16:09:50Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  CDMAIN.C	created 	??/??/??	by Roy Sherrill
 *				updated	06/18/96	by John Burnham
 *
 *  Copyright (c) 1996-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *	CD ROM Driver Initialization file.  This file
 *	contains the initialzaition and other basic routines
 *	to handle the CD-ROM Driver.
 *
 *	Revision history:
 *
 * 10/12/98 jb	Removed checking for unsupported functions to actual loop
 *					so that the functions that COULD modify a cdrom could
 *					be properly checked for critical errors. I also changed the
 *					default Critical error to CRIT_DRIVE_NOT_READY and the
 *					critical errors caused by failed reads are changed to
 *					CRIT_NOT_READY.
 * 07/15/98 jb	Modified Stat2SDB and SDB2Stat to
 *					use a newly created bitmapped field in the SDB
 *					to allow for more than directory extents
 *					larger than 64K
 * 07/01/98 jb	Fixed a bug in int2fh function 0x1100
 *					to return immediately rather than allow
 *					some processing to occur before returning.
 * 02/11/98 jb	Added a routine, ClearError, to clear the
 *					carry flag before returning from the
 *					int 2f function.
 * 09/19/97 jb	Changed the code in the INT2F Function 15h
 *					code to give a not ready or a read fault error
 *					rather than generating a critical error.
 * 08/27/97 jb	Upon entry to the INT2Fh function 15f
 *					subfunction processor, I removed subfunction
 *					0Bh, Drive Check, from the list of functions
 *					checking for critical errors and validating
 *					drive numbers.  This fixes a bug with windows
 *					file manager where all of the drive icons
 *					show up as cdroms because we say we support
 *					them!
 * 01/20/97 jb	Added code to do some eror code checking
 *					on a failure from a read and to give a
 *					critical error if the read failed because
 *					of a problem with the data on the CDROM.
 * 12/04/96 jb  Did some more cleanup on the critical error
 *					handling code to make sure mounting happened
 *					at the appropriate time.  Also changed
 *					code to conditionally check the mounting calls
 * 12/16/96 jb	Removed unneeded fields to save memory
 * 11/25/96 jb	Reworked critical error handling to clean up
 *					when critical errors are generated and
 *					how they are handled.
 * 11/19/96 jb	Removed extra calls to CD_LFSEEK to help
 *					increase performance.  Replaced them with
 *					direct assignment of values.
 *********************************************************/


#define __FILENUM__	1			// for assert macro

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <errno.h>

#include "datalght.h"

#include "cdrom.h"
#include "cddos.h"
#include "sft.h"
#include "doserr.h"
#include "cdstruct.h"

//
// prototype files
//
#include "cdmain.pt"
#include "cdutil.pt"
#include "cddev.pt"
#include "api.pt"
#include "lowl.pt"
#include "util.pt"
#include "cdmemory.pt"


//
// external
//

extern long far Next2fLabel;
extern void interrupt Int2FHandler();
extern CDDRIVE *drive_core;
//extern void far *lpDriverInterrupt;					// Pointer to Device Driver
																	//  interrupt routine
//extern void far *lpDriverStrategy;						// Pointer to Device Driver
																	//  Strategy Routine
extern unsigned GetCS(void);
extern unsigned GetDS(void);
extern unsigned PASCAL DoCriticalError(unsigned errCode, void far *lpDevice, unsigned critAX);

extern unsigned fakeDevice[];
extern int in2FFlag;


#if DEBUG
extern unsigned _psp;
#endif

//
// local declarations
//

#if CACHE_DIR_ONLY

int fIsFileData = FALSE;

#endif

void interrupt(*next_2f)();

#if DEBUG
char *int2FCmd[] =
{
	"zero", 						// 0x00
	"rmdir",
	"unsp 2",
	"mkdir",
	"unsp 4",
	"chdir",
	"Close",
	"Commit",
	"Read",  					// 0x08
	"Write",
	"Lock",
	"Unlock",
	"Get Disk Space",
	"unsp 0xD",
	"Set F attrib",
	"Get F attrib",
	"unsp 0x10",				// 0x10
	"Rename",
	"unsp 0x12",
	"Delete",
	"unsp 0x14",
	"unsp 0x15",
	"Open",
	"Creat",
	"unsp 0x18", 				// 0x18
	"unsp 0x19",
	"unsp 0x1A",
	"FindFirst",
	"FindNext",
	"Close for Process",
	"Do Redirection",
	"Printer Setup",
	"Flush All Buffers",		// 0x20
	"Lseek",
	"Process Termination",
	"Parse Path",
	"Unsp 24",
	"Redir Printer Mode",
	" ",
	" ",
	" ",							// 0x28
	" ",
	" ",
	" ",
	" ",
	" ",
	"Extended Open",
	};
char call[35];

#endif

int cdOpened = FALSE; 						// detect whether or not the CDROM
													// interface was fully initialized.

int bUseExtended = FALSE;					// Use Extended memory for Buffers


#if UNSUPPORTED

int supportKanji = FALSE;					// Support Kanji for Volume Labels
													// Not yet supported

int displayMemStats = FALSE;				// Display memory statics at Start
													// Not yet supported

int cdROMSharing = FALSE;					// Support Sharing of CD-ROM with MS-NET
													// Not yet supported
#endif

int firstCDROMDriveNumber = 0;			// First Drive number of CDROM drive
int firstCDROMDriveLetter = 'A';		// first drive letter of CDROM drive

unsigned openFileFlag;						// flag to tell a routine whether to
													// return file handle 0 or 1

unsigned countOfFindFirsts = 0;

int numDrives;								// Number of of CD-ROM drives
int currentDrive = 0;						// drive number of currently active CDROM drive


char szDeviceName[9] = DEFAULT_DEVICE;// Device Name


int numBuffers = 4;							// Number of buffers for CD_ROM Drive

struct CDEX_SFT far *sft;

char findFirstString[SDA_FILENAMESIZE];
char findFirstMatch[SDA_FILENAMESIZE];

char localString[SDA_FILENAMESIZE];
uchar localStackArray[STACK_SIZE];		// Memory used for Local 2F Stack

struct CDS far *lpOurCDS;					// far pointer to our CDS structure


#if DEBUG
//
// UnInstall the CDEX driver.  This function restores the DOS CDS's
// and restores the Int 2F vector, and frees the memory of the program!
//
void PASCAL UnInstall(void)
{
	struct PSP far *lpCDEXPSP;


	printf("\nUninstalling CD-ROM Drive %c: to %c:\n", firstCDROMDriveLetter, firstCDROMDriveLetter);
	lpOurCDS->drvAttrib = 0;
	lpListOfLists->blkDevCount--;


	SetVector(0x2F, next_2f);				// Restore 2f Interrupt

	//
	// call dos function 49h to free the memory.  you can use the PSP
	// to do this, but first you must release the environment buy using
	// the far pointer in the psp structure as the address.
	//
	lpCDEXPSP = MK_FP(_psp,0);
	ReleaseMemory(lpCDEXPSP->pspEnvironment);
	ReleaseMemory(_psp);

}
#endif

void PASCAL ReleaseMemory(unsigned segmentOfBlock)
{
	ASSERT(segmentOfBlock != 0);
	asm	mov	ax, segmentOfBlock;
	asm	mov	es, ax;
	asm	mov	ax, 0x4900;
	asm	int	0x21;
}


//
// this procedure handles the INT 2Fh, function 15h subfunctions
//
unsigned Int2F15xxH(STACK far * lpUserStack)
{
	int iFunction;
	int iDriveNumber;
	CDDRIVE *cdromDriveControl;

	//
	// reset the error flag
	//
	errno = 0;

#ifndef bud

	iFunction = lpUserStack->h.al;

	switch (iFunction)
	{
		case 6:							// turn debugging on
		case 7:							// turn debugging off
		case 9:							// absolute eraseable CDROM disk write -- reserved and nonfunctional
		case 0x0A:						// reserved
			DoError(lpUserStack, INVALID_FUNCTION);
			return RETURN_FROM_INT2FH;


		case 0x02:
		case 0x03:
		case 0x04:
		case 0x05:
		case 0x08:
		case 0x0F:

			// get the drive number (0 based)
			iDriveNumber = GetInternalDriveNum(lpUserStack->h.cl);

			// validate the drive number
			if (iDriveNumber < 0 || iDriveNumber >= numDrives)
			{
				DoError(lpUserStack, INVALID_DRIVE);
				return RETURN_FROM_INT2FH;
			}

			//
			// check the status of the drive and execute a critical err
			// if it is not ready.
			//
			if (CDChanged(iDriveNumber) == YES)
			{
				if (drive_core[iDriveNumber].cdOpened == TRUE)
				{
					//
					// get the current drive flavor so that
					// we can reset the vtoc type so that the
					// underlying routines will read a new one
					//
					cdromDriveControl = cd_u_drvno2drv(iDriveNumber, 0);
					cd_cddismount(iDriveNumber,0);
					if (cdromDriveControl)
						cd_memfill(cdromDriveControl, sizeof(CDDRIVE)-30, 0);

					// tell the system that we need to remount the cdrom.
					drive_core[iDriveNumber].cdOpened = FALSE;
				}
			}

			// check to see if this drive has been mounted
			if (!drive_core[iDriveNumber].cdOpened)
			{

				// mount the cd, initialize the path and let the system know that
				// the cd has been mounted
				if (CDChanged(iDriveNumber) == NO && CDOpenDrive(iDriveNumber))
				{
					// mount the cd, initialize the path and set the flag telling the
					// system that the cd has been mounted
					lpOurCDS->cwd[0] = '\0';		// Init CWD to root of drive
					drive_core[iDriveNumber].cdOpened = TRUE;
				}

				// unable to mount the cd
				else
				{
					switch (iFunction)
					{
						case 0x02:
						case 0x03:
						case 0x04:
							DoError(lpUserStack,ERR_READFAULT);
							return RETURN_FROM_INT2FH;

						case 0x05:
						case 0x08:
						case 0x0F:
						default:
							DoError(lpUserStack,ERR_NOTREADY);
							return RETURN_FROM_INT2FH;
					}
				}
			}
	}



	switch (lpUserStack->h.al)
	{
		char far *lpUserBuffer;		// pointer to user passed buffer
											// for subfunction 01h


											VTOC far *lpUserVTOC;		// pointer to user data space where the
											// specified vtoc will be read

		CDDRIVE *pUserDriveRecord;	// drive record to get the
											// various filenames back to the user

		ULONG startingSector;		// used for read absolute function to
											// determine the sector number that the
											// user wants to start the read.

		//
		// determine if CDROM extensions are installed and how many
		// drives are CDROMs
		//
		case 0:
			lpUserStack->x.bx = numDrives;
			lpUserStack->x.cx = firstCDROMDriveNumber;
			break;


		//
		// Get CD-ROM Device driver list
		//
		case 1:
		{
			int i;


			//  Datalight cache only, replace driver in cx:dx with
			//  driver in si:di
			//
			for (i = 0; i < numDrives; i++)
			{
				if( drive_core[i].lpCDROMDevice ==
					MK_FP(lpUserStack->x.cx, lpUserStack->x.dx) )
				{
					drive_core[i].lpCDROMDevice = MK_FP(lpUserStack->x.si,
							lpUserStack->x.di);
					drive_core[i].lpDriverStrategy =
							MK_FP( FP_SEG(drive_core[i].lpCDROMDevice),
							((LPDEV_HDR)(drive_core[i].lpCDROMDevice))->dhStrat);
					drive_core[i].lpDriverInterrupt =
							MK_FP( FP_SEG(drive_core[i].lpCDROMDevice),
							((LPDEV_HDR)(drive_core[i].lpCDROMDevice))->dhInter);
				}
			}


			lpUserBuffer = MK_FP(lpUserStack->x.es, lpUserStack->x.bx);
			for (i = 0; i < numDrives; i++)
			{
				//
				// fill in subunit in device list, at user's ES:BX
				//
				*(char far *)lpUserBuffer = drive_core[i].unitNumber;
				lpUserBuffer++;

				//
				// fill in address of lpCDROMDevice header in user space
				//
				*(long far *)lpUserBuffer = (long)drive_core[i].lpCDROMDevice;
				lpUserBuffer += 4;
			}
		 	break;
		}


		//
		// Get Copyright, Abstract and Biblio filenames
		//
		case 2:
		case 3:
		case 4:

			//
			// get a pointer to the record that contains the filenames
			//
			lpUserBuffer = MK_FP(lpUserStack->x.es, lpUserStack->x.bx);
			pUserDriveRecord = cd_u_drvno2drv(iDriveNumber, 0);

 			//
			// get the copyright filename
			//
			if (lpUserStack->h.al == 2)
				_fmemcpy(lpUserBuffer, pUserDriveRecord->cur_vtoc->copy_file, CDROM_INFOFILE_SIZE);
			if (lpUserStack->h.al == 3)
				_fmemcpy(lpUserBuffer, pUserDriveRecord->cur_vtoc->abstr_file, CDROM_INFOFILE_SIZE);
			if (lpUserStack->h.al == 4)
				_fmemcpy(lpUserBuffer, pUserDriveRecord->cur_vtoc->biblio_file, CDROM_INFOFILE_SIZE);
			break;

		//
		// Read VTOC
		//
		case 5:

			//
			// get a pointer to the user area
			//
			lpUserVTOC = MK_FP(lpUserStack->x.es, lpUserStack->x.bx);

			if (!CD_CDRead(iDriveNumber, 1, lpUserStack->x.dx+FIRSTVTOC, (UTINY far *)lpUserVTOC))
				errno = ERR_NOTREADY;
			else
				lpUserStack->x.ax = lpUserVTOC->desc_type;

			break;

		//
		// CDROM Absolute Read
		//
		case 8:

			//
			// get a pointer to the user area
			//
			lpUserBuffer = MK_FP(lpUserStack->x.es, lpUserStack->x.bx);

			//
			// calculate the starting sector
			//
			startingSector = ((ULONG)lpUserStack->x.si << 16) + lpUserStack->x.di;

			//
			// get the data and return an error if unable to read
			//
			if (!CD_CDRead(iDriveNumber, lpUserStack->x.dx, startingSector,
								(UTINY far *)lpUserBuffer))
				errno = ERR_NOTREADY;

			break;


		//
		// Get CDEX drive number
		//
		case 0x0B:

			// get the drive number (0 based)
			iDriveNumber = GetInternalDriveNum(lpUserStack->h.cl);

			//
			// store Magic number for CDEX installation ID
			//
			lpUserStack->x.bx = cdexMagicNumber;

			//
			// check to see if drive supported
			//
			if (iDriveNumber >= 0 && iDriveNumber < numDrives)
//			if (ValidDriveNumber(iDriveNumber))
				lpUserStack->x.ax = driveSupported;
			else
				lpUserStack->x.ax = driveNotSupported;
			break;

		//
		// Get CDEX version number
		//
		case 0x0C:
			lpUserStack->x.bx = CDEX_VERSION;
			lpUserStack->x.cx = CDEX_SIGNATURE;
			break;


		//
		// Get CDROM drive letters
		//
		case 0x0D:
		{
			int i;

			lpUserBuffer = MK_FP(lpUserStack->x.es, lpUserStack->x.bx);
			for (i = 0; i < numDrives; i++)
			{
				*lpUserBuffer++ = drive_core[i].physDriveNumber;
			}
			break;
		}

		//
		// Get/Set Volume Descriptor Preference
		//
		case 0x0E:
			if (lpUserStack->x.bx == 1)
			{
				errno = ERR_BADFUNC;
				break;
			}

			pUserDriveRecord = cd_u_drvno2drv(iDriveNumber, 0);
			lpUserStack->h.dh = pUserDriveRecord->cur_vtoc->desc_type;
			lpUserStack->h.dl = 0;
			break;


		//
		// Send CDROM device driver request
		//
		case 0x10:
		{
			struct REQHDR far *lpReqHdr;
			void far *lpDriverStrat;
			void far *lpDriverInt;

			// get the drive number (0 based)
			iDriveNumber = GetInternalDriveNum(lpUserStack->h.cl);

			// validate the drive number
			if (iDriveNumber < 0 || iDriveNumber >= numDrives)
			{
				DoError(lpUserStack, INVALID_DRIVE);
				return RETURN_FROM_INT2FH;
			}
			// get a pointer to the request header
			lpReqHdr = MK_FP(lpUserStack->x.es, lpUserStack->x.bx);

			// put the unit number of this drive in the request header
			lpReqHdr->unit = drive_core[iDriveNumber].unitNumber;


			lpDriverStrat = drive_core[iDriveNumber].lpDriverStrategy;
			lpDriverInt   = drive_core[iDriveNumber].lpDriverInterrupt;

			asm	int 	3
			asm	push	es
			asm	push	bx
			asm	mov  	bx, word ptr [lpReqHdr+2]
			asm	mov	es, bx
			asm	mov	bx, word ptr [lpReqHdr]
			asm	call 	dword ptr [lpDriverStrat]
			asm	call 	dword ptr [lpDriverInt]
			asm	pop	bx
			asm	pop	es

			break;
		}



#if DEBUG
		//
		// Datalight CDEX specific call -- Uninstall
		//
		case 0x99:
		{
			extern long cdStatCount;
			extern long cdReadCount;
			in2FFlag = FALSE;				// Allow printing in this call
			printf("Dev Stats = %lu  Dev Reads = %lu\n", cdStatCount, cdReadCount);

			UnInstall();
			lpUserStack->x.ax = 0;		// notify that it worked!
			break;
		}
#endif


		//
		// functions currently not supported
		//
		case 0x0F:						// Get directory entry
			errno = INVALID_FUNCTION;
			break;


		//
		// default case - hit a function we don't support
		//
		default:
#if DEBUG
			printf("Invalid 2Fh-15h call AX='%x'\n", lpUserStack->x.ax);
#else
			errno = INVALID_FUNCTION;
			DoError(lpUserStack,errno);
#endif
	}
	if (errno)
		DoError(lpUserStack,errno);
	else
		ClearError(lpUserStack);

	return RETURN_FROM_INT2FH;

#else
	errno = INVALID_FUNCTION;
	DoError(lpUserStack, errno);
	return RETURN_FROM_INT2FH;
#endif
}		// end of Int21h15xxh procedure




DSTAT findFirstNextStatus = {-1,};
unsigned Int2F11xxH(STACK far * lpUserStack)
{
	//
	// [ local data ]
	//
	unsigned func, gotFile, changeDirectory;
	DSTAT stat;
	int handle;
	char *pFileName;
	char szEName[SIZE_ENAME];
	CDDRIVE *cdromDriveControl;
	LONG posFromEnd;
	bool fCDChanged;






	//
	// initialize the error code variable
	//
	errno = 0;


	//  Get the CDS of DOS's current CDS pointer
	//
	lpOurCDS = *((struct CDS far * far *)MK_FP(FP_SEG(lpSDAFileName1), 0x5A2 ));


	//
	// get a pointer to the caller passed System File Table (SFT)
	//
	sft = MK_FP(lpUserStack->x.es, lpUserStack->x.di);


	//  Put the function number in a local variable
	//
	func = lpUserStack->h.al;


   // do installation check
	//
   if( func == 0 )
   {
      if (lpUserStack->x.attrib == 0xDADA)
		{
			// set the appropriate return value
			//
         lpUserStack->x.attrib=cdexMagicNumber;
		}

		// assume we are installed
		//
		lpUserStack->x.ax = 0;


		// we are installed
		//
		lpUserStack->h.al = 0xFF;


		return RETURN_FROM_INT2FH;
   }


	// determine if this call is for CDEX
	//
	if( !IsOurDrive(func) )
	{
		return CHAIN_PRIOR_INT2FH;
	}


	//
	// check the status of the drive and execute a critical err
	// if it is not ready.
	//
	fCDChanged = CDChanged(currentDrive);
	if( fCDChanged == YES )
	{
		if (drive_core[currentDrive].cdOpened == TRUE)
		{
			//
			// get the current drive flavor so that
			// we can reset the vtoc type so that the
			// underlying routines will read a new one
			//
			cdromDriveControl = cd_u_drvno2drv(currentDrive, 0);
			cd_cddismount(currentDrive,0);
			if (cdromDriveControl)
				cd_memfill(cdromDriveControl, sizeof(CDDRIVE)-30, 0);

			// tell the system that we need to remount the cdrom.
			drive_core[currentDrive].cdOpened = FALSE;
		}
	}


	// has the cd been mounted?
	if (drive_core[currentDrive].cdOpened == FALSE)
	{
		// can we mount the drive?
		fCDChanged = CDChanged(currentDrive);
		if( (fCDChanged == NO) && (CDOpenDrive(currentDrive)) )
		{
			// mount the cd, initialize the path and set the flag telling the
			// system that the cd has been mounted
			lpOurCDS->cwd[0] = '\0';		// Init CWD to root of drive
			drive_core[currentDrive].cdOpened = TRUE;
		}

		// unable to mount the drive
		else
		{
			// generate a critical error only for specific functions.  Not all
			// cdrom functions generate critical errors.  If the user 'Fails'
			// the critical error return a function specific return code.
			switch (func)
			{
				case 1:					// Remove Directory
				case 3:					// Make Directory
				case 5:					// Change Directory
				case 8:					// Read from file
				case 9:					// file write
				case 0xC:				// Get Disk space
				case 0xE:				// Set File Attributes
				case 0xF:				// get file attributes
				case 0x11:				// Rename File
				case 0x13:				// Delete File
				case 0x17:				// Create File
				case 0x16:				// open file
				case 0x1B:				// find first file
				case 0x1C:				// find next file
				case 0x2e:				// open file - extended
					if (GenerateCriticalError(CRIT_DRIVE_NOT_READY) == CDROM_NOTREADY)
					{
						switch (func)
						{
							case 1:
							case 3:
							case 5:
								errno = ERR_NOTPATH;
								break;

							case 8:
							case 9:
								lpUserStack->x.cx = 0;
								break;

							case 0xE:
							case 0xF:
								errno = ERR_NOTFOUND;
								break;

							case 0x1B:
								errno = ERR_LASTFILE;
								break;

							case 0x0C:
							case 0x11:
							case 0x13:
							case 0x16:
							case 0x17:
							case 0x1C:
							case 0x2E:
								errno = ERR_DENIED;
								break;
						}
						if (errno != 0)
							DoError(lpUserStack, errno);

						return RETURN_FROM_INT2FH;
					}

					// the drive became ready again was remounted.
					// now reset the directory and the mounting flag
					else
					{
						lpOurCDS->cwd[0] = '\0';		// Init CWD to root of drive
						drive_core[currentDrive].cdOpened = TRUE;
					}
			}
		}
	}


#if DEBUG
	//
	// clear destination field and copy string to display to data area
	//
	cd_memfill(call, sizeof(call), 0);
	if (func < 0x2f)
		strcpy(call, int2FCmd[func]);
	if (func != 0x2f)
		printf("Int2F11xxH, where xx is %x, call %s\n", func, call);
#endif



	lpUserStack->x.flags &= ~CARRY;			// reset the error(carry) flag
	errno = 0;										// reset errno to NO error
	lpSDA->errno = 0;								// reset DOS errno
	GetSDAValues();								// get the Dynamic values of the SDA

	switch(func)
	{
		case 1:				// remove dir
		case 3:				// create dir
		case 9:				// write to file
		case 0x0E:			// set file attribute
		case 0x11:			// rename file
		case 0x13:			// delete file
		case 0x17:			// create/truncate file
				errno = ERR_DENIED;
				break;

		// unsupported functions
		case 7:										// Commit file
		case 0xA:									// Lock
		case 0xB:									// UnLock
		case 0x2d:									// unknown
				errno = ERR_DENIED;
				break;


		case 5:												// Change Directory
			_fmemcpy(localString, lpSDAFileName1, SDA_FILENAMESIZE);
			if (!(ValidateFileName(localString)) &&  (strlen(localString) > 3))
			{
				errno = ERR_NOTPATH;
				break;
			}
			cd_errno = 0;
			changeDirectory = cd_u_scwd(0, localString);
			if (changeDirectory == NO)
				errno = ERR_NOTPATH;
			break;

		case 6:										// Close file
			cd_close(sft->cdexHandle - CDEX_FILEHANDLE_MAGIC_NUMBER);
			if (cd_errno != 0)
			{
				ASSERT(cd_errno == 0);
			}
			else
			{
				//
				//	more than one reference to this handle, we can safely
				// decrement it to show that we have closed it.
				//
				if (sft->handles > 1)
					--sft->handles;

				//
				// only one reference to the handle, so we need to check
				// to see if this user is the owner of the handle.
				//
				else if (sft->handles == 1)
				{
					if (sft->owner == lpSDA->currentPSP)
						--sft->handles;

					else
						errno = ERR_DENIED;

				}

				//
				// the reference count is zero or less.  Assert here now
				//
				else
					ASSERT(sft->handles > 0);
			}

				break;

		case 8:										// Read File
		{
			CDFILE *pHandle;
			int uReadResult;

			pHandle = cd_fileno2file(sft->cdexHandle-CDEX_FILEHANDLE_MAGIC_NUMBER);

			ASSERT(pHandle);


			openFileFlag = OPENFILE;

			// DOS may have changed file position without telling us.
			// make sure we're at the right place in the file.
			pHandle->pos = sft->fPos;


			BuildCDFILE(sft, pHandle);

#if !CACHE_DIR_ONLY

			uReadResult = cd_fread(pHandle, lpSDADTA, lpUserStack->x.cx);

#else

			fIsFileData = TRUE;
			uReadResult = cd_fread(pHandle, lpSDADTA, lpUserStack->x.cx);
			fIsFileData = FALSE;

#endif

			// update the file position if there was no error
			if (!cd_errno)
			{
				sft->fPos = pHandle->pos;
				lpUserStack->x.cx = uReadResult;
			}

			//
			// handle the errors from the read function
			//
			else
			{

				//
				// update the file position
				//
				if (uReadResult > 0)
				{
					lpUserStack->x.cx = uReadResult;
					sft->fPos = pHandle->pos;
				}

				else
				{

					if (cd_errno == CDE_IOERR)
					{
						unsigned criticalErrorResponse;
						unsigned criticalErrorRetry = 0;


						// generate a critical error
						criticalErrorResponse = DoCriticalError(CRIT_READ_ERR, &fakeDevice,
										0x1E00 | (currentDrive + firstCDROMDriveNumber));

						while (criticalErrorResponse == RETRY_OPERATION &&
								 cd_errno != CDE_OKAY)
						{
							while (criticalErrorRetry < 3)
							{
								cd_errno = CDE_OKAY;
								uReadResult = cd_fread(pHandle, lpSDADTA, lpUserStack->x.cx);
								if (cd_errno == CDE_OKAY)
								{
									errno = 0;
									sft->fPos = pHandle->pos;
									lpUserStack->x.cx = uReadResult;
									criticalErrorRetry = 3;
								}
								// increment the retry count
								criticalErrorRetry++;
							}

							criticalErrorResponse = DoCriticalError(CRIT_READ_ERR, &fakeDevice,
											0x1E00 | (currentDrive + firstCDROMDriveNumber));

							criticalErrorRetry = 0;
						}

						// specific case for failing the operation
						if (criticalErrorResponse == FAIL_IN_PROGRESS)
						{
							// we need to return the current pos as 0
							sft->fPos = 0;
							pHandle->pos = 0;

							// we need to return the file length as 0
							sft->fSize = 0;
							pHandle->dir.data_len = 0;

							// return the number of bytes read as 0
							lpUserStack->x.cx = 0;

						}
					}

					// if the cdrom drive door was opened during the read
					else if (cd_errno == CDE_CHANGE)
					{
						errno = 0x15;				// drive not ready!
						lpUserStack->x.cx = 0;	// no data read
					}



					else
					{
						// if we reached the end of the file tell the user
						errno = 0;
						if (uReadResult < 0)
							lpUserStack->x.cx = 0;
					}
				}

				//
				// we need to handle errors here.  There possible errors are:
				//
				//
				//	User or data errors
				//  	Illegal file handle
				//  	Read error
				//	 	read past end of volume
				//
				// Critical Errors
				//	 	Drive status error such as media changed
				//  	VTOC missing or invalid
				//
			}

			break;
		}


		case 0x0C:											// Get Disk Space

			//
			// the following values were the values that were
			// returned from this call and a popular CDEX driver.
			//
			lpUserStack->x.ax = 0x0001;				// sectors per cluster
			lpUserStack->x.bx = 0xFFFF;				// total clusters
			lpUserStack->x.cx = 0x0800;				// bytes per sector
			lpUserStack->x.dx = 0x0000;				// number of available clusters
			break;


		case 0x0F:											// Get file attributes
			_fmemcpy(localString, lpSDAFileName1, SDA_FILENAMESIZE);
			cd_memfill(&stat, sizeof(stat), 0);

			gotFile = cd_u_gfirst(&stat, localString, 0);

			if (gotFile == NO)
			{
				errno = ERR_NOTFOUND;
			}
			else
			{
				lpUserStack->x.ax = FA_RDONLY;
				if (stat.dir.file_flags & FF_DIRECTORY)
					lpUserStack->x.ax |= FA_DIREC;

				if (stat.dir.file_flags & FF_EXISTENCE)
					lpUserStack->x.ax |= FA_HIDDEN;

				lpUserStack->x.bx = (unsigned)(stat.dir.data_len >> 16);
				lpUserStack->x.di = (unsigned)stat.dir.data_len;

			}
			if (stat.fileno != -1)
			{
				cd_close(stat.fileno);
				stat.fileno = -1;
			}
			cd_gdone(&stat);
			break;



		case 0x16:									// Open File
		case 0x2E:									// Extended Open File

			//
			// try to open the file.  Does it exist?
			//
			openFileFlag = OPENFILE;
			_fmemcpy(localString, lpSDAFileName1, SDA_FILENAMESIZE);

			//
			// let's make sure we are trying to open a file
			//
			if (!ValidateFileName(localString))
			{
				errno = ERR_ACCESS_DENIED;
				break;
			}


			handle = cd_u_open(0, localString);

			// file wasn't found
			if (cd_errno)
			{
				//
				// this is a regular open
				//
				if (func != 0x2E)
					// NOTE: may need to handle critical error here
					errno = ERR_NOTFOUND;

				//
				// this is an extended open -- need to check actions
				//
				else
				{
					//
					// user wants to create a non-existent file
					//
					if (lpUserStack->h.dl & FILE_NOT_EXIST_ACTION_CREATE)
						errno = NET_ACCESS_DENIED;

					//
					// user wants call to fail
					//
					else
						errno = ERR_NOTFOUND;

				}
			}

			//
			// no error, let's fill in the SFT fields.
			//
			else
			{

				if (IsDirectory(handle))
				{
					cd_close(handle);
					errno = ERR_ACCESS_DENIED;
					break;
				}

				//
				// we need to check the open actions
				//
				if (func == 0x2e)
				{

					//
					// truncate the file on opening is not allowed
					//
					if (lpSDA->spopAct & FILE_EXIST_ACTION_TRUNCATE)
//					if (lpUserStack->h.dl & FILE_EXIST_ACTION_TRUNCATE)
					{
						cd_close(handle);
						errno = NET_ACCESS_DENIED;
						break;
					}

					//
					// fail after opening
					//
					if (lpSDA->spopAct & 0xF0)
//					if (!(lpUserStack->h.dl & 0xF0))
					{
						cd_close(handle);
						errno = ERR_ACCESS_DENIED;
						break;
					}

					//
					// if the user action is to open the file normally, then we
					// will just fall through to the following code which is where
					// the subfunction 16h will fall through to as well.
					//
				}


				//
				// normal code for opening the file and setting up the SFT
				//
				{
					COUNT foundPos;
					char *pSearchString, *pSaveFound;
					char foundChar;
					char expandedName[ENAMESIZE];
					CDFILE *pHandle;

					// get a pointer to the CDFILE structure for this handle
					pHandle = cd_fileno2file(handle);
					ASSERT(pHandle);

					//
					// set up the CD file information into unused fields in the
					// SFT so that when we go to read the file, we can rebuild
					// the file information.
					//
					BuildSFT(pHandle, sft);

					//
					// Set the SFT name.  First, find the name in the path string
					//
					pSearchString = cd_parsep(localString, &foundPos, &foundChar);

					ASSERT(pSearchString);

					//
					// name should already be in qualified form (C:\PATH\NAME.EXT)
					//
					ASSERT(foundChar != NULL);

					//
					// keep searching until end of string.  last backslash
					// should be the filename part of the path string.
					///
					while (foundChar != NULL)
					{
						pSaveFound = pSearchString;
						pSearchString = cd_parsep(pSearchString, &foundPos, &foundChar);
					}
					pSearchString = pSaveFound;

					//
					// Now we have the name, expand to "NAME    EXT" form and
					// put into SFT.
					//
					NameExpand(expandedName, pSearchString);
					_fmemcpy(sft->fName, expandedName, ENAMESIZE);

					//
					// set up System File Table fields
					//
					sft->openMode = sdaOpenMode;

					//
					// get and save the current file position
					//
//					sft->fPos = cd_lfseek(pHandle, 0L, PSEEK_CUR);
					sft->fPos = pHandle->pos;

					//
					//	get the file size by seeking to the end of the file
					//	and then move the file pointer position back to the beginning
					//	of the file
					//
					sft->fSize = pHandle->dir.data_len;
//					sft->fSize = cd_lfseek(pHandle, 0L, PSEEK_END);
//					cd_lfseek(pHandle, 0L, PSEEK_SET);

					//
					// get and save the current drive
					//
					sft->devInfo = 0x8040 | currentDrive + firstCDROMDriveNumber;

					//
					// CDEX specific: handle for opening, closing and find firsts
					//
					sft->cdexHandle = handle+CDEX_FILEHANDLE_MAGIC_NUMBER;


					//
					// set up the correct PSP who owns this handle
					//
					sft->owner = lpSDA->currentPSP;


					sft->attr = GetFileAttributes(handle);
					sft->fDate = GetFileDate(handle);
					sft->fTime = GetFileTime(handle);
				}
			}

			break;



		case 0x1B:									// Find First

			openFileFlag = FINDFIRST;

			//
			// get search patter and default SDB
			//
			_fmemcpy(findFirstString, lpSDAFileName1, SDA_FILENAMESIZE);

			//
			// before initialization of the status structure,
			// we should make a called cd_gdone just to make sure
			// that the resources from previous find first/next's
			// are cleaned up.  They get allocated in cd_fopen_cur().
			//
			cd_gdone(&findFirstNextStatus);


			//
			// initialize the status structure
			//
			cd_memfill(&findFirstNextStatus, sizeof(findFirstNextStatus), 0);

			//
			// move past drive, colon and backslash and only grab the
			// search pattern.  Null terminate the search pattern once
			// we have grabbed it.
			//
			pFileName = strrchr(findFirstString, '\\') + 1;		// Move past '/'
			NameExpand(szEName, pFileName);
			szEName[11] = '\0';

			//
			// check to see if this is a LABEL
			//
			if(sdaSrchAttr == FA_LABEL)
			{

				//
				// get CDROM Drive Control Structure
				//
				cdromDriveControl = cd_u_drvno2drv(currentDrive, 0);

				//
				// set up default status values
				//

				// set file size to zero
				findFirstNextStatus.dir.data_len = 0;

				// store the volume name
				NameToDot(findFirstNextStatus.dir.file_id, cdromDriveControl->cur_vtoc->volume_id);

				// there is no file flag for volume labels, so I created a
				// temporary value that will be checked in the Stat2SDB
				// procedure so that we can set the proper DOS file flag: FA_LABEL
				findFirstNextStatus.dir.file_flags = FF_CDLABEL;

				//
				// Since this isn't a physical file, it has no date or time stamp
				// the statement above that sets all of the structure 'findFirstNextStatus'
				// to zero will set the date and time to zero.  Since
				// MSCDEX returns these values as zero, we are as well.
				//

				// call routine to transfer data to DOS's SDB.
				Stat2SDB(&findFirstNextStatus, lpSDASDB, szEName);

				break;
			}

			//
			//	at this point, we are searching for a file or a directory or
			// both.
			//

			if (!(ValidateFileName(findFirstString)) &&  (strlen(findFirstString) > 3))
			{
				errno = ERR_LASTFILE;
				break;
			}

			gotFile = cd_u_gfirst(&findFirstNextStatus, findFirstString, 0);


			if(gotFile == NO)
			{
            errno = ERR_LASTFILE;
				cd_gdone(&findFirstNextStatus);
			}
			else
			{
            //
            // save the matching path/file in a global
            //
            _fmemcpy(findFirstMatch, findFirstString, SDA_FILENAMESIZE);

				//
				// we found a file, now we need to see if it matches the
				// the type the user is looking for!
				//


				// first get one without the path, since we don't need it anymore
				NameToDot(findFirstString, szEName);

				while (!MatchFileAttributeType(findFirstNextStatus.dir.file_flags,sdaSrchAttr) && gotFile)
					gotFile = cd_gnext(&findFirstNextStatus, findFirstString);


				//
				// we got out of the loop, check to see if we ran out of files.  If we
				// did then, the file types we were looking for could not be found.
				//
				if (gotFile == NO)
				{
               errno = ERR_LASTFILE;
					cd_gdone(&findFirstNextStatus);
				}
				else
				{
					//
					// we found a file that matched -- use it
					Stat2SDB(&findFirstNextStatus, lpSDASDB, szEName);


               //
               // save the filename
               FFN_BuildMatch(findFirstMatch, findFirstNextStatus.dir.file_id);

				}




			}

			break;

		case 0x1C:									// Find Next


			if(lpSDASDB->srchAttr == FA_LABEL)
			{
				errno = ERR_LASTFILE;
				cd_gdone(&findFirstNextStatus);
				break;
			}


			// get the values back out of the SDB
			SDB2Stat(&findFirstNextStatus, lpSDASDB, szEName);

			NameToDot(findFirstString, szEName);

			//
			// get the correct value for the search attribute out of the
			// search data block
			//
			sdaSrchAttr = lpSDASDB->srchAttr;

			//
			// move past drive, colon and backslash and only grab the
			// search pattern.  Null terminate the search pattern once
			// we have grabbed it.
			//
//			pFileName = strrchr(findFirstString, '\\') + 1;		// Move past '/'
//			NameExpand(szEName, pFileName);
//			szEName[11] = '\0';


			gotFile = cd_gnext(&findFirstNextStatus, findFirstString);
			if(gotFile == NO)
			{
				errno = ERR_LASTFILE;
				cd_gdone(&findFirstNextStatus);
			}
			else
			{
				//
				// we found a file, now we need to see if it matches the
				// the type the user is looking for!
				//
				while (!MatchFileAttributeType(findFirstNextStatus.dir.file_flags,sdaSrchAttr) && gotFile)
					gotFile = cd_gnext(&findFirstNextStatus, findFirstString);


				//
				// we got out of the loop, check to see if we ran out of files.  If we
				// did then, the file types we were looking for could not be found.
				//
				if (gotFile == NO)
				{
					errno = ERR_LASTFILE;
					cd_gdone(&findFirstNextStatus);
				}
				else
            {
					//
					// we found a file that matched -- use it
					//
					Stat2SDB(&findFirstNextStatus, lpSDASDB, szEName);

               // save the filename
               FFN_BuildMatch(findFirstMatch, findFirstNextStatus.dir.file_id);
            }
			}
			break;


		case 0x21:									// Lseek
		{
			CDFILE *pHandle;

			//
			// get the position the user wants to set
			//
			posFromEnd = (((LONG)(lpUserStack->x.cx)) << 16) + (LONG)(lpUserStack->x.dx);

			//
			// so that we move the file pointer in the correct direction
			//
			posFromEnd = 0 - posFromEnd;


			//
			// set the position that the user specified
			//
			pHandle = cd_fileno2file(sft->cdexHandle-CDEX_FILEHANDLE_MAGIC_NUMBER);
			ASSERT(pHandle);

			// initialize the CDFILE structure with the data from the sft
			BuildCDFILE(sft, pHandle);

			// seek to the specified position
			posFromEnd = cd_lfseek(pHandle, posFromEnd, PSEEK_END);


			//
			// set the new position for this file handle
			//
			sft->fPos = posFromEnd;


			//
			// set up the return registers
			//
			lpUserStack->x.dx = (unsigned)(posFromEnd >> 16);
			lpUserStack->x.ax = (unsigned)posFromEnd;

			break;
		}

		default:
			errno = INVALID_FUNCTION;
			break;
	}

	in2FFlag = FALSE;

	if(errno)
		DoError(lpUserStack,errno);
	return RETURN_FROM_INT2FH;

}




//
// Checks if the current Interrrupt 2F call is for
// the Datalight CD-ROM Drive
//
int PASCAL IsOurDrive(unsigned Function)
{
	char far *lpCDSPath;

	//
 	// lpSDAFileName1 shows me who !
	//
	switch(Function)
	{
		case 1 :									// Remove Directory
		case 3:									// Make Directory
		case 5:									// Change Directory
		case 0xE:								// Set File Attributes
		case 0xF:								// Get File Attributes
		case 0x11:								// Rename File
		case 0x13:								// Delete File
		case 0x16:								// Open File
		case 0x17:								// Creat File
		case 0x1B:								// Find First
		case 0x2E:								// Extended Open File
			// get the drive number (0 based)
			currentDrive = GetInternalDriveNum(lpSDAFileName1[0] - 'A');
			return ((currentDrive >= 0) && (currentDrive < numDrives));

		case 0x1C:								// Find Next
			// get the drive number from the DTA
			currentDrive = lpSDA->sdb.drive;
			currentDrive = GetInternalDriveNum((currentDrive & 0x003F) - 1);
			return ((currentDrive >= 0) && (currentDrive < numDrives));


		case 0x23:								// Qualify Path and Filename
			lpCDSPath = MK_FP(lpDosStack->x.ds, lpDosStack->x.si);
			currentDrive = GetInternalDriveNum(lpCDSPath[0] - 'A');
			return ((currentDrive >= 0) && (currentDrive < numDrives));


		case 0xC:								// Get Disk Space
		 	//
			// es:di Points to the CDS
			//
			lpCDSPath = MK_FP(lpDosStack->x.es, lpDosStack->x.di);
			currentDrive = GetInternalDriveNum(lpCDSPath[0] - 'A');
			return ((currentDrive >= 0) && (currentDrive < numDrives));

		//
		// sft is valid
		//
		case 6:									// close file
		case 7:									// Commit file
		case 8:									// file read
		case 9:									// file write
		case 0xA:								// Lock region write
		case 0xB:								// Unlock region write
		case 0x21:								// Lseek
			//
			// this call is not meant for us
			//
			if (! (sft->devInfo & 0x8000) )
				return FALSE;

			//
			// this is a network drive, we need to validate the
			// the drive number and the file handle
			else
			{

				currentDrive = (sft->devInfo & 0x003f) - firstCDROMDriveNumber;

				//
				// this drive is not our cdrom
				//
				if ((currentDrive < 0) || (currentDrive >= numDrives))
					return FALSE;


				//
				// this drive is our cdrom, now we need to validate the
				// file handle.  We will assert if the file handle is
				// not valid.
				//
				else
				{
				  	//
				  	// invalid handle, but yet the drive says it is the CDROM drive
				  	// something has gone wrong.  ASSERT now!
				  	//
					ASSERT (((sft->cdexHandle - CDEX_FILEHANDLE_MAGIC_NUMBER) >= 0) &&
						     ((sft->cdexHandle - CDEX_FILEHANDLE_MAGIC_NUMBER) < MAXFILES) );

					//
					// this file handle does fit within the limits of our file
					// handles.  we are okay.
					return TRUE;
				}
			}


		default:
			return (FALSE);
	}
}



//
// ClearError setups a good return from the DOS 2F call without an error
//
void PASCAL ClearError(STACK far *lpUserStack)
{
	lpUserStack->x.flags &= (~CARRY);			// clear the carry flag
}

//
// DoError sets up a return from the DOS 2F call with an error.
//
void PASCAL DoError(STACK far *lpUserStack, unsigned errno)
{
	lpUserStack->x.ax = errno;
	lpUserStack->x.flags |= CARRY;
}

void PASCAL Stat2SDB(DSTAT *stat, struct SDB far *sdb, char *ename)
{

	//
	// local data
	//
	char ename_buf[13];
	CDFILE *pFile;



	// The 80 makes this drive an IFS Drive
	sdb->drive = (currentDrive + firstCDROMDriveNumber + 1) | 0x80;

	if(ename != NULL)
		_fmemcpy(sdb->szNameMask, ename, 12);

	// get a pointer to the file structure
	pFile = cd_fileno2file(stat->fileno);


	//
	// set up some default values for the search data block
	//
	sdb->srchAttr    = sdaSrchAttr;

	// if we have a valid file name (not a label) save the information
	if (pFile)
	{
      unsigned dirSizeInSectors;

		// get the number of logical sectors in the this directory extent
		// if any bytes spill over into the next sector, include that sector
		// as well
		dirSizeInSectors = (unsigned)(pFile->dir.data_len / LS_SIZE);
		if ((pFile->dir.data_len % LS_SIZE) > 0)
			dirSizeInSectors++;

		sdb->entryOffset = pFile->pos;
		sdb->parSector   = (pFile->dir.loc_extent & 0x0007FFFFl) |
                        ((ULONG)dirSizeInSectors << 20) |
                        ((ULONG)stat->rootDirectorySearch << 19);


	}
	else
	{
		// set up some defaults for the volume label
		sdb->entryOffset = 0;
		sdb->parSector   = 0x10;
	}


	//
	// get the size of the file, but set the size to zero if it is a
	// directory
	//
	if(stat->dir.file_flags & FF_DIRECTORY)
		sdb->size = 0l;
	else
		sdb->size = stat->dir.data_len;


	// get the expanded version of the filename
	// and copy it to the sdb.
	NameExpand(ename_buf, stat->dir.file_id);

	_fmemcpy(sdb->szFname, ename_buf, 11);


#if 0

	sdb->date_time = 0x12345678L; //stat->dir.record_time;

#else

	// save the date and time to the search data block
	sdb->time = ((stat->dir.record_time.hour & 0x1F) << 11);
	sdb->time |= (stat->dir.record_time.min & 0x3F) << 5;
	sdb->time |= (stat->dir.record_time.sec & 0x1F) >> 1;


	// if the month and the day of the record time are invalid
	// set the whole date to an invalid amount.
	if (stat->dir.record_time.month == 0 &&
	    stat->dir.record_time.day   == 0)
		 sdb->date = 0;

	else
	{
		sdb->date =  (stat->dir.record_time.year - 80) << 9;
		sdb->date |=  (stat->dir.record_time.month & 0xF) << 5;
		sdb->date |=  (stat->dir.record_time.day & 0x1F) << 0;
	}

#if DEBUG
	printf("Year = %u month %u day %u\n",
		stat->dir.record_time.year,
		stat->dir.record_time.month,
		stat->dir.record_time.day);
#endif


#endif

	// determine what kind of entry that we found.
	if(stat->dir.file_flags & FF_DIRECTORY)
		sdb->fAttr = FA_DIREC;

	else
	{
		if (stat->dir.file_flags == FF_CDLABEL)
			sdb->fAttr = FA_LABEL;

		else
			sdb->fAttr = FA_RDONLY;
	}

	if (stat->dir.file_flags & FF_EXISTENCE)
		sdb->fAttr |= FA_HIDDEN;

}



//
// checks the string for a backslash ending the path.  If one is found,
// then no valid filename is on the path. -1 is returned.
//
BOOL PASCAL ValidateFileName(char *pUserFileName)
{
	char *pLastBackSlash;

	//
	// check the last character of the string
	//

	pLastBackSlash = strrchr(pUserFileName, '\\');

	//
	// return FALSE if the string length is equal to the length of the last
	// backslash in the file name from the beginning of the string -- we
	// know that our file specification now ends in a path seperator and
	// this is not a valid filename.  Return TRUE otherwie
	//

	return !((unsigned)strlen(pUserFileName) == (unsigned)(pLastBackSlash - pUserFileName + 1));
//	if ((unsigned)strlen(pUserFileName) == (unsigned)(pLastBackSlash - pUserFileName + 1))
//		return -1;
//
//	// okay, the length was different from the index of the last backslash in
//	// the filename, so we must assume that we have some sort of filename.
//	else
//		return TRUE;
}

/*
	BuildSFT takes data from the CDFILE structure and places it into the
	sft assigned to this file for safekeeping.  This procedure should only
	be called when a file is opened on the CD.
*/
void PASCAL BuildSFT(CDFILE *file, struct CDEX_SFT far *sft)
{
	// store the signature to make this SFT look unique
	sft->cdexSignature = SFT_SIGNATURE;

	// save the volume set sequence number
	sft->VSSN = file->dir.VSSN;

	// save the length of the extended attribute record
	sft->extendedAttrRecordLength = file->dir.XAR_len;

	// save the logical block number signifying the start of file
	sft->lbnOfFileStart = file->dir.loc_extent;

	// save the interleave size value
	sft->interleaveSize = file->dir.il_size;

	// save the interlave gap value
	sft->interleaveSkip = file->dir.il_skip;

	// save the version of the file
	sft->fileVersion = file->dir.file_version;

	// save number of logical sector read ahead
//	sft->numLogicalSectors = file->nls;

	// save file position pointer
//	sft->cdFilePosition = file->pos;
}


/*
	BuildCDFILE takes data from the SFT and recreates the necessary data
	found inside of CDFILE structure.  Not all of the data in the CDFILE
	structure was saved initially, so only the saved parts are restored.
*/
void PASCAL BuildCDFILE(struct CDEX_SFT far *sft, CDFILE *file)
{
	char szSFTFilename[ENAMESIZE];

	// restore the length of the extended attribute record
	file->dir.XAR_len = sft->extendedAttrRecordLength;

	// restore the logical block number of the start of file
	file->dir.loc_extent = sft->lbnOfFileStart;

	// restore the size of the file
	file->dir.data_len = sft->fSize;

	// restore the date and time stamp of the file
	SFTDateTimeToCDROMDateTime(sft->fDate, sft->fTime, &file->dir.record_time);

	// restore the file flags
	SFTFlagsToCDROMFlags(sft->attr, &file->dir.file_flags);

	// restore the interleave size
	file->dir.il_size = sft->interleaveSize;

	// restore the interleave gap
	file->dir.il_skip = sft->interleaveSkip;

	// restore the volume sequence number
	file->dir.VSSN = sft->VSSN;

	// change the filename from expanded to dot format
	_fmemcpy(szSFTFilename, sft->fName, ENAMESIZE);
	NameToDot(file->dir.file_id, szSFTFilename);

	// determine the length of the filename in dot format
	file->dir.len_fi = strlen(file->dir.file_id);

	// restore the version of the file
	file->dir.file_version = sft->fileVersion;

	// dummy up the system area length value
//	file->dir.len_su = 0;

//	// restore number of logical sector read ahead
//	file->nls = sft->numLogicalSectors;

//	// restore file position pointer
// 	file->pos = sft->cdFilePosition;
}

//
// this routine takes the passed DOS formatted date and time stamp, and
// converts it to the format used on the CDROM and saves it appropriately
//
void PASCAL SFTDateTimeToCDROMDateTime(unsigned date, unsigned time, DIR_DATE *record_time)
{
	//
	// the cdrom's date stamp is offset from 1900
	// the PC's date stamp is offset from 1980, so we need to
	// add 80 years from the pc's year storage and move it to
	// year field of the file's record date
	//
	record_time->year  = (date >> 9) + 80;
	date = date & 0x01FF;

	//
	// get the saved value of the month, convert it back to the correct
	// value for the CDROM and store it.
	//
	record_time->month = date >> 5;
	date = date & 0x001F;

	//
	// get the saved value of the day, convert it back to the correct value
	// for the CDROM file and store it.
	//
	record_time->day = date;

	//
	// get the hour, convert it to the appopriate value and save it
	//
	record_time->hour = time >> 11;
	time = time & 0x07FF;

	//
	// get the saved value for the minutes, convert it a valid value for
	// the CDROM and save it.
	//
	record_time->min = time >> 5;
	time = time & 0x001F;

	//
	// get the seconds, and add it to the hour and minutes result in the packed
	// field
	//
	record_time->sec = time * 2;
}

//
// SFTFlagsToCDROMFlags converts the PC format flags to the format
// used by the file system on the CDROM
//
void PASCAL SFTFlagsToCDROMFlags(unsigned attr, uchar *cdFileFlags)
{
	*cdFileFlags = 0;
	if (attr & FA_DIREC)
		*cdFileFlags = FF_DIRECTORY;

	if (attr & FF_EXISTENCE)
		*cdFileFlags |= FA_HIDDEN;
}

void PASCAL SDB2Stat(DSTAT *stat, struct SDB far *sdb, char *ename)
{

	//
	// local data
	//
	CDFILE *pFile;

	// default file structure for find firsts.  Make sure this value is setup
	stat->fileno = FINDFIRST_FILENO;

	// get a pointer to the file structure
	pFile = cd_fileno2file(stat->fileno);

	Assert(pFile != NULL);


	//
	// set up some default values for the search data block
	//
	pFile->pos            = sdb->entryOffset;
	pFile->dir.loc_extent = sdb->parSector & 0x0007FFFFl;
	pFile->owner_drive    = cd_u_drvno2drv(currentDrive, 0);

	// get the size of the directory extent.  This will not be the exact
	// original amount, but an approximation in sectors.
   pFile->dir.data_len   = ((sdb->parSector & 0xFFF00000l) >> 20) * 2048l;
   stat->rootDirectorySearch = (unsigned)((sdb->parSector & 0x00080000l) >> 19);

	_fmemcpy(ename, sdb->szNameMask, 11);
	ename[11] = '\0';

}

void FFN_BuildMatch(TEXT *pDest, TEXT *pFilename)
{
   TEXT *pTemp = strrchr(pDest, BACKSLASH);

   if (pTemp)
   {
      pTemp[1] = 0x00;
      strcat(pDest, pFilename);

   }
   else
      strcpy(pDest, pFilename);
}
