//
//   $Id: CDMEMORY.C 1.5 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  CDMEMORY.C
 *
 *  Copyright (c) 1996-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *	Contains the code to allocate the necessary memory for
 *	the internal structures.
 *
 *  Revision History:
 *
 * 12/23/96 jmb	Updated buffer allocation scheme to allocate
 *					far buffers.  Also added capability to
 *					allocate the buffers in XMS.
 *********************************************************/
#define __FILENUM__ 12

#include "datalght.h"

#include "cdrom.h"
#include "cddos.h"
#include "cdstruct.h"
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include "xms.h"

//
// prototype includes
//
#include "cdmemory.pt"
#include "cdutil.pt"
#include "xms.pt"
#include "cderr.pt"


//
// external procedures
//
extern void *AllocateBytes(unsigned bytes);
extern void far *AllocateFarMemory(unsigned paragraphs);

//
// external data
//
extern	int numDrives;
extern	int numBuffers;
extern  int bUseExtended;
extern  unsigned uXMSBufferHandle;


//
// public data
//
UTINY   *buffer_core = 0;
PBUFFER_CONTROL pBufferControl = 0;
PSECTOR_BLOCK   pSectorList = 0;
RAW_BUF far *lpBuffers = 0;
CDDRIVE *drive_core  = 0;
CDFILE  *file_core   = 0;
VTOC    *vtoc_core   = 0;


//
// Fill a string
//
VOID PASCAL cd_memfill(VOID *vto, INT size, UTINY c)                       /* __fn__*/
{
	UTINY *to = (UTINY *) vto;

	ASSERT(vto != NULL);						// validate paramter

 	while (size--)
  	*to++ = c;
}


BOOL PASCAL cd_memory_init(void)                                               /*__fn__*/
{
	// UCOUNT n_drive_structures;
	UCOUNT size;


	//
	// allocating memory for the Sector Cache Control structure
	//

#if DEBUG
	printf("Getting BUFFER_CONTROL\n");
#endif
	size = sizeof(BUFFER_CONTROL);
	pBufferControl = (PBUFFER_CONTROL)AllocateBytes(size);
	if (!pBufferControl)
		goto mem_error;


	//
	// allocating memory for the sector list
	//

#if DEBUG
	printf("Getting SECTOR_BLOCK\n");
#endif
	size = numBuffers * sizeof(SECTOR_BLOCK);
	pSectorList = (PSECTOR_BLOCK)AllocateBytes(size);
	if (!pSectorList)
		goto mem_error;

	// allocating memory for the drive structure pointer array
	drv_array = (PCDDRIVE *)AllocateBytes(numDrives * sizeof(PCDDRIVE));
	cd_memfill(drv_array, numDrives * sizeof(PCDDRIVE), 0);

//	//
//	// Allocating memory for the drive structures
//	//
//
//#if DEBUG
//	 printf("Get Drive Structures\n");
//#endif
//
//
//	n_drive_structures = MAXUSERS * numDrives;
//	drive_core = (CDDRIVE *)AllocateBytes(n_drive_structures * sizeof(CDDRIVE));
//
//	if (!drive_core)
//		goto mem_error;
//
//	cd_memfill(drive_core, (n_drive_structures * sizeof(CDDRIVE)), (UTINY) 0);


	//
	// Allocating memory for the file structures
	//
#if DEBUG
	printf("Get CDFILE\n");
#endif
	file_core = (CDFILE *)AllocateBytes(MAXFILES * sizeof(CDFILE));
	if (!file_core)
		goto mem_error;
	cd_memfill(file_core,MAXFILES*sizeof(CDFILE),(UTINY) 0);


		//
		// Allocating memory for the Volume Table of Contents (VTOC) stuctures
		//
#if DEBUG
		printf("Get VTOC\n");
#endif
		vtoc_core = (VTOC *)AllocateBytes(2 * numDrives * sizeof(VTOC));
		if (!vtoc_core)
			goto mem_error;

		cd_memfill(vtoc_core,(2 * numDrives * sizeof(VTOC)), (UTINY) 0);

	//
	// allocating the buffer list in far memory
	//
#if DEBUG
	printf("Get RAW_BUF\n");
#endif
AllocateBuffers:
	// if extended memory is not requested, allocate the buffers in
	// conventional memory
	if (!bUseExtended)
	{
		size = ((numBuffers * sizeof(RAW_BUF)) + 0x0F) >> 4;
		lpBuffers = (RAW_BUF far *)AllocateFarMemory(size);
		if (!lpBuffers)
			goto mem_error;
	}

	// use extended memory to allocate the buffers.  We will still need
	// to allocate at least 1 buffer in conventional memory to transfer
	// data to and from extended memory.
	else
	{
		// determine number of K to allocate
		size = (UCOUNT)(((ULONG)numBuffers * sizeof(RAW_BUF)) / ONE_KILOBYTE);

		if (size > TotalXMS())
		{
			bUseExtended = FALSE;
			Message(cdrInsufficientXMS);
			goto AllocateBuffers;
		}
		else
		{
			uXMSBufferHandle = AllocateXMS(size);
			// get a pointer to the single conventional memory buffer
			lpBuffers = (RAW_BUF far *)AllocateFarMemory(sizeof(RAW_BUF) >> 4);
		}


}



	return(YES);

mem_error:
    cd_memory_close();
    return(NO);
}




VOID PASCAL cd_memory_close(void)                                               /*__fn__*/
{
}


//
// cd_mem_bufs() -  Allocate or free core for the buffer pool.
//
// This routine allocates the core needed for buffer management.
// The number of bytes needed equals the size of the control structure
// plus one rawbuf (512 bytes and one bufmng structure per buffer allocated.
// If the incomming pointer is NULL we are allocating. If NON NULL we are
// freeing.
//
// This may be replaced with static allocation if desired.
//

UTINY * PASCAL cd_mem_bufs(UTINY *p)                                          /*__fn__*/
{
	if (p)
  {
		//
		// (you could do this)  free(p);
		//
  	return(NULL);
	}
	else
	{
		//
		// You could do this:
		//
		// . size=NLSBUF * (sizeof(BUFMNG) + sizeof(RAW_BUF))
		// . size += sizeof(BUFCTL);
		// . buffers = (UTINY *)malloc(size);
		// . return(buffers);
		//
		return(buffer_core);
	}
}

//
// CDMemBufCtrl will return a pointer to the memory that
// will contain the cache control structure.  If the passed
// pointer contains a non-NULL pointer NULL will be returned
// as if the data was "released".  If NULL is passed, then
// the pointer to the memory space will be returned.
//
PBUFFER_CONTROL CDMemBufCtrl(PBUFFER_CONTROL pBufCtrl)
{
	// if a valid pointer, simulate a deallocation
	if (pBufCtrl)
		return NULL;

	// the pointer is null, return a pointer to memory allocated
	// for the cache control structure
	else
		return pBufferControl;
}

//
// CDMemSectorBlock will return a pointer to the memory that
// will contain a list of sector structures.  If the passed
// pointer contains a non-NULL pointer NULL will be returned
// as if the data was "released".  If NULL is passed, then
// the pointer to the memory space will be returned.
//
PSECTOR_BLOCK CDMemSectorBlock(PSECTOR_BLOCK pSectLst)
{
	// if a valid pointer, simulate a deallocation
	if (pSectLst)
		return NULL;

	// the pointer is null, return a pointer to memory allocated
	// for the sector block list.
	else
		return pSectorList;
}

//
// CDMemRawBuffers will return a pointer to the memory that
// will contain the cached data from the CDROM.  If the passed
// pointer contains a non-NULL pointer NULL will be returned
// as if the data was "released".  If NULL is passed, then
// the pointer to the memory space will be returned.
//
// Notice that the pointers used with this routine are far pointers.
// This is to allow the buffers to take up more than 64K of memory.
// They could feasibly take up much of conventional memory.
//
RAW_BUF far * CDMemRawBuffers(RAW_BUF far *lpBufferList)
{
	// if a valid pointer, simulate a deallocation
	if (lpBufferList)
		return NULL;

	// the passed pointer is null, so return a pointer to the allocated
	// data space for the actual data read from a sector on the CDROM
	else
		return lpBuffers;
}


CDDRIVE * PASCAL cd_mem_drive(CDDRIVE *p, COUNT drive, COUNT user)           /*__fn__*/
{
	COUNT i;

  if (p)
  {
		return(NULL);
  }
  else
	{
		if (drive >= numDrives || user >= MAXUSERS)
	   	return(NULL);

		i = (drive * MAXUSERS) + user;
		return &drive_core[i];
	}
}

CDFILE * PASCAL cd_mem_file(CDFILE *pfile, COUNT fileno)                    /*__fn__*/
{
    if (pfile)
        return(NULL);

    else
    {
        if (fileno >= MAXFILES)
            return(NULL);

        else
            return(file_core+fileno);
    }
}

VTOC * PASCAL cd_mem_vtoc(VTOC *pvtoc, COUNT drive,COUNT vtoc_type)         /*__fn__*/
{
COUNT i;

    if (pvtoc)
	  {
        /* (you could do this) free(pvtov) */
        return(NULL);
	  }
    else
    {
        /* (you could do this) return( (VTOC *) malloc(sizeof(VTOC)); */
        if (drive >= numDrives || (vtoc_type != 1 && vtoc_type != 2) )
	            return(NULL);
        i = (2 * drive) + (vtoc_type -1);
        return(vtoc_core + i);
    }
}
