//
//   $Id: CDUTIL.C 1.5 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  Copyright (c) 1989-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *********************************************************
 *
 * CD ROM Driver Initialization file.  This file
 * contains the initialzaition and other basic routines
 * to handle the CD-ROM Driver.
 *
 * Revision History
 *
 * 10/12/98	jb	removed call to ToUpperCase so that we return
 *					filenames and such as stored on the disk.
 * 02/13/97 jb	removed NameMatch function
 * 01/20/97 jb	removed static keyword from declaration of
 *					fakeDevice
 * 12/04/96 jb	Added a call to remount the drive after
 *					critical error has corrected itself
 * 11/25/96 jb	Changed the interface to CDOpenDrive.
 * 				Removed parameter, numDrives and reduced the
 *					code to the necessary amount to mount the
 *					specified drive.
 *********************************************************/

#define __FILENUM__ 5

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <dir.h>
#include <ctype.h>
#include <string.h>
#include <mem.h>

#include "datalght.h"

#include "cdrom.h"
#include "cddos.h"
#include "cdstruct.h"

//
// prototype include files
//
#include "cdutil.pt"
#include "cddev.pt"
#include "cderr.pt"
#include "api.pt"
#include "util.pt"


//
// external data
//
extern int numDrives;
extern int currentDrive;
extern int firstCDROMDriveNumber;
extern char firstCDROMDriveLetter;
extern CDDRIVE *drive_core;

//
extern unsigned PASCAL DoCriticalError(unsigned errCode, void far *lpDevice, unsigned critAX);


#if DEBUG
void PASCAL FarDump(void far *lpBufferToDump, unsigned count)
{
  //
	// [ LOCALS ]
	//
	unsigned char far *lpBufferIndexPointer = (unsigned char far *)lpBufferToDump;
	int i, j;


	for(j=0; j<count/4; ++j)
	{
		printf("%lp  ", lpBufferIndexPointer);

		for(i=0; i<16; ++i)
			printf("%x ", lpBufferIndexPointer[i]);

		printf(" - ");

		for(i=0; i<16; ++i)
			printf("%c", lpBufferIndexPointer[i]);

		lpBufferIndexPointer += 16;
		printf("\n");
	}
}
#endif


//
// CD-ROM Open Drive for operation.
//
// 	drive is the internal representation of the drive a 0-numDrives
// number.  We need to convert it to the actual physical drive number
// in the system.  We do that by adding the value from the drive_core
// array entry that corresponds with it.
//
BOOL PASCAL CDOpenDrive(int drive)
{
	char *pPath = "A:\\";

	//lpCDROMDevice->dhCDDriveLtr = firstCDROMDriveNumber+1;
	//CDSetDeviceLtr(drive+firstCDROMDriveNumber);


	pPath[0] = drive_core[drive].physDriveNumber + 'A';
	return cd_u_dskopen(0, pPath);
//	++*pPath;
}


//
// New CD-ROM Drive Assert Function.  display File number and line
// instead of file name.  Modified message 4 to display two decimals
// instead of a string and a decimal.
// Will also just hang the machine rather than trying to exit.
//
void PASCAL AssertFail(unsigned fileNumber, unsigned lineNumber)
{
	// switch back to text mode
//	asm mov	ax, 0x0003
//	asm int	0x10

	Message(cdrInternalError, fileNumber, lineNumber);
	while(1);		// hang
}


//
// Get the changing values from DOS's SDA for this 2F call.
//
void PASCAL GetSDAValues(void)
{
	lpSDADTA = lpSDA->dta;
	sdaOpenMode = lpSDA->openMode;
	sdaSrchAttr = lpSDA->srchAttr;
	sdaSDBDrive = lpSDA->sdb.drive;
}


//
// Get the attributes of the CDROM file and return them as DOS file
// attributes
//

uchar PASCAL GetFileAttributes(unsigned handle)
{
  CDFILE *pCDFile;
	uchar cdFileAttributes = 0;

	pCDFile = cd_fileno2file(handle);

	if (pCDFile->dir.file_flags & FF_DIRECTORY)
		cdFileAttributes = FA_DIREC;

	if (pCDFile->dir.file_flags & FF_EXISTENCE)
		cdFileAttributes |= FA_HIDDEN;

	return cdFileAttributes;
}

//
// Get the date stamp of the CDROM File and return it as a DOS file date stamp
//
unsigned PASCAL GetFileDate(unsigned handle)
{
  CDFILE *pCDFile;


	pCDFile = cd_fileno2file(handle);

	//
	// if the cdrom shows that the storage date of the file is prior to
	// 1980, we cannot convert it.
	//
	ASSERT(pCDFile->dir.record_time.year >= 80);

	//
	// the cdrom's date stamp is offset from 1900
	// the PC's date stamp is offset from 1980, so we need to
	// subtract 80 years from the cdrom's year storage and move it to
	// to the proper storage bits in the packed field.  Also get the
	// month and the day fields, shift them to their fields and add them
	// all together
	//
	return ((unsigned)(pCDFile->dir.record_time.year - 80) << 9) +
         ((unsigned)(pCDFile->dir.record_time.month)     << 5) +
         ((unsigned)pCDFile->dir.record_time.day);
}



//
// Get the time stamp of the CDROM File and return it as a DOS file time stamp
//
unsigned PASCAL GetFileTime(unsigned handle)
{
  CDFILE *pCDFile;

	pCDFile = cd_fileno2file(handle);

	//
	// get the hour, minutes and seconds.  Shift them to the appropriate bit
	// fields and add them all together
	//
	return ((unsigned)(pCDFile->dir.record_time.hour) << 11) +
         ((unsigned)(pCDFile->dir.record_time.min)  <<  5) +
	       ((unsigned)(pCDFile->dir.record_time.sec)  >>  1);

}

//
// Change a name from "name.ext" form to "name    ext"
// Expands wildcards into ?s.
//
char * PASCAL NameExpand(char *pEName, char *pName)
{
	uchar name1 = *(pName+1);

	//
	// for special subdirs "." and ".."
	//
	if(*pName == '.' && (!name1 || strchr(".\\/",name1)))
	{
		NameFillTo(pEName,pName,11);
		*pEName = '.';

		if(name1 == '.')
			*(pEName+1) = '.';

		return pName;
	}

	pName = NameFillTo(pEName,pName,8);

	if(*pName == '.')
		++pName;

	return NameFillTo(pEName+8,pName,3);
}



//
// match expanded filenames (ONLY name1 may contain wildcards)
//
//int PASCAL NameMatch(char *pName1, char *pName2)
//{
//	int i;
//
//	if (*(unsigned char *)pName2==0xE5)			// deleted
//		return -1;
//
//	for (i = 0; i < ENAMESIZE; i++)
//	{
//		if (*pName1 != '?' && *pName1 != *pName2)
//			return (pName1 > pName2 ? 1 : -1);
//
//		pName1++;
//		pName2++;
//	}
//	return 0;
//}

//
// Change an expanded name into the "name.ext" form.
// Does not understand wildcards.
//
void PASCAL NameToDot(char *pName, char *pEName)
{
	int i;

	for (i = 0; i < 8 && pEName[i] != ' '; ++i)
		*pName++ = pEName[i];

	if (pEName[8] != ' ')
		*pName++ = '.';

	for (i = 8; i < 11 && pEName[i] != ' '; ++i)
		*pName++ = pEName[i];

	*pName = 0;
}

char separators[] = " .:;,=+/\"[]\\<>|\t\n\r";

//
// Expand a portion of a filename padding with
// spaces and or question marks (for wildcards).
// Stops when it encounters a file separator.
//
char * PASCAL NameFillTo(char *pEName, char *pName, int num)
{
	int i;
	char fill;

	//
	// parse this section of the filename, and understand wildcards
	//
	for (i = 0; i < num; ++i)
	{
		if (*pName == '*')
			fill = '?';

		else if (*pName == '\0' || strchr(separators, *pName))
			fill = ' ';

		else
		{
			fill = *pName;
			pName++;
		}
		pEName[i] = fill;
	}

	//
	// skip past extra on name
	//
	while (*pName && !strchr(separators+1, *pName))
		pName++;

	return pName;
}




unsigned fakeDevice[] = {-1, -1, 0x1000, 0, 0, 0, 0, 0};

//
// Crit_Err Generates a Critical Error from with in a IFS call.
//
unsigned PASCAL GenerateCriticalError(unsigned critErrorNumber)
{

	unsigned criticalErrorResponse;
	unsigned criticalErrorRetry = 0;


	criticalErrorResponse = DoCriticalError(critErrorNumber, &fakeDevice,
	                             0x1E00 | (currentDrive + firstCDROMDriveNumber));

	while (criticalErrorResponse == RETRY_OPERATION)
	{
		while (criticalErrorRetry < 3)
		{
			// if the cdrom disk has not been changed
			if (CDChanged(currentDrive) == NO)
			{
				// attempt to remount the drive as the drive became ready
				if (CDOpenDrive(currentDrive))
					return CDROM_READY;
			}

			// increment the retry count
			criticalErrorRetry++;
		}

		criticalErrorResponse = DoCriticalError(critErrorNumber, &fakeDevice,
								0x1E00 | (currentDrive + firstCDROMDriveNumber));

		criticalErrorRetry = 0;
	}
	return CDROM_NOTREADY;
}



//
// MatchFileAttributeType will take the passed CD file attributes, convert
// them to DOS file attributes and then compare them to see if the converted
// attributes match the passed DOS attributes.  If they match, then TRUE
// is returned, otherwise FALSE is returned.
//
BOOL PASCAL MatchFileAttributeType(UTINY cdFileAttributes, uchar userDOSFileAttributes)
{
	uchar convertedCDFileAttributes = 0x00;

	//
	// convert CD File Attributes to DOS FAT File Attributes
	//
	if (cdFileAttributes & FF_DIRECTORY)
		convertedCDFileAttributes = FA_DIREC;				// set directory flag

	if (cdFileAttributes & FF_EXISTENCE)
		convertedCDFileAttributes |= FA_HIDDEN;			// set hidden flag


	//
	// check to see if the attributes match up
	//
	return !((userDOSFileAttributes & convertedCDFileAttributes) ^ convertedCDFileAttributes);

}


//
// IsDirectory takes the specified handle and tells us whether or not
// this directory entry is a directory specifier.  If it is, it returns
// TRUE otherwise it returns FALSE.
//
BOOL PASCAL IsDirectory(int handleToCheck)
{
  CDFILE *pCDFile;

	// get a pointer to the file record
	pCDFile = cd_fileno2file(handleToCheck);

	// return whether or not this file record points to a directory or a file
	return (pCDFile->dir.file_flags & FF_DIRECTORY);
}


//
//// ValidDriveNumber takes physical number of the specified drive and
//// makes sure that it is one of the drives that are installed.
//// If it is one of the installed, this routine returns true.  Otherwise
//// it returns false.
////
//int PASCAL ValidDriveNumber(int driveNumberToCheck)
//{
//	int i;
//
//	i = 0;
//
//	while (i < numDrives)
//	{
//		if (driveNumberToCheck == drive_core[i].physDriveNumber)
//			return TRUE;
//		i++;
//	}
//	return FALSE;
////	return (driveNumberToCheck == currentDrive);
//}


//
// GetInternalDriveNum will accept a drive number that is
// representative of the physical drive within the system.
//
// it will return the internal version of that driver number as it
// maps into the drive core array.  If a match is not found, -1
// will be returned as the error value.
//
COUNT PASCAL GetInternalDriveNum(TINY physicalDriveNum)
{
	int i;

	// initialize the index
	i = 0;

	// while the there are drives left to search
	while (i < numDrives)
	{
		// if this is the one we are looking for, return it's index
		if (drive_core[i].physDriveNumber == physicalDriveNum)
			return i;

		// we need to look at the next one.
		i++;

	}

	// return drive not found!
	return -1;
}
