//
//   $Id: STRIPCPI.C 1.2 2002/10/15 20:43:38Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
   Copyright (c) 1992-1997 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "display.h"


void ExitWithMessage(void)
{
		printf("\nUsage: STRIPCPI infile outfile\n"
				 "Creates a new .CPI with a the single \"ROM\" font\n"
				 "(code page 437) from the input file.\n");
		exit(-1);
}

CPIHeader header;
CodePageList outCodePgs = {0}, codePgs;
void *baseFontPtr, *highFontPtr;

int main(int argc, char **argv)
{ FILE *fin, *fout;
  int i, font = -1;
  long offset;
  char c;

	printf("STRIPCPI	Copyright (c) 1996 Datalight, Inc.\n"
			 "All rights reserved.\n\n");

	// check args
	//
	if(argc < 3)
		ExitWithMessage();

	// open input and output files
	//
	fin = fopen(argv[1],"rb");
	if(fin < 0)
	{
		perror(argv[0]);
		ExitWithMessage();
	}
	rewind(fin);
	fout= fopen(argv[2],"rb");		// verify file overwrite
	if(fout > 0)
	{
		printf("Output file %s already exists. Continue [Y/N]?",argv[2]);
		c = toupper(getchar());
		if(c != 'Y')
			exit(0);
		fclose(fout);
		printf("\n");
	}
	fout= fopen(argv[2],"wb+");
	if(fout < 0)
	{
		perror(argv[0]);
		ExitWithMessage();
	}

	// read the font file header and verify the file ID
	//
	fread(&header,sizeof(CPIHeader),1,fin);
	if(strcmp(header.id,"CPIKBS"))
	{
		printf("File %s is not a valid .CPI file, aborting...\n",argv[2]);
		ExitWithMessage();
	}

	// read the code page list and make sure 437 is there
	//
	fread(codePgs,sizeof(CodePageList),1,fin);
	for(i=0; i < header.maxCodePages; i++)
	{
		if(codePgs[i] == 437)
		{
			font = i;
			break;
		}
	}
	if(font < 0)
	{
		printf("Can't find code page 437 in %s, aborting...\n",argv[2]);
		ExitWithMessage();
	}

	// Read in the lower 128 bit patterns
	//
	baseFontPtr = calloc(header.chunkSize,1);
	if(baseFontPtr == 0)
	{
		perror(argv[0]);
		ExitWithMessage();
	}
	fread(baseFontPtr,header.chunkSize,1,fin);


	// Read in the upper 128 bit patterns
	//
	highFontPtr = calloc(header.chunkSize,1);
	if(highFontPtr == 0)
	{
		perror(argv[0]);
		ExitWithMessage();
	}
	offset = header.firstCodePage + (font * header.chunkSize);
	fseek(fin, offset, SEEK_SET);
	fread(highFontPtr,header.chunkSize,1,fin);

	fclose(fin);


	// set the number of codepages to 1 and write the header
	//
	header.maxCodePages = 1;
	fwrite(&header,sizeof(CPIHeader),1,fout);

	// set the codepage list to only support 437 and write
	// out the codepage list
	//
	*outCodePgs = 437;
	fwrite(outCodePgs,sizeof(CodePageList),1,fout);

	// write out code page 437
	//
	fwrite(baseFontPtr,header.chunkSize,1,fout);
	fwrite(highFontPtr,header.chunkSize,1,fout);

	// we're done!
	//
	fclose(fout);
	printf("New file %s created.\n",argv[2]);
	return 0;
}
