//
//   $Id: KEYB.H 1.5 2002/10/15 20:48:06Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary
*/
#define KEYBDATAFILE "KEYBOARD.SYS"
#define KEYBDATAFILE2 "KEYBRD2.SYS"

#define MAX_CAN_PREFIX 35

#define NO_KEY 0xFFFF

#define NORMAL 0
#define SHIFT 1
#define CTRL 2
#define ALTGR 3
#define ALT 4

#define NORMGET 0
#define NORMCHECK 1
#define NORMSHIFT 2
#define EXTGET 0x10
#define EXTCHECK 0x11
#define EXTSHIFT 0x12

#define KEYB_F1 0x68
#define KEYB_F2 (KEYB_F1+1)


#define ALL_RAW_SHIFTS	0xF
#define RAW_ALT_LSHIFT  0xA
#define RAW_ALT_RSHIFT  0x9

/*

A "prefix key" is a key that, when pressed, may change the code resulting
from an immediately subsequent keypress. If the subsequent key's meaning
is changed then that second key is "prefixable" by the former.

Typically, prefix keys have accent characters of some sort and prefixable
keys are usually vowels. A key may be a prefix key in some combination
of shift states and not in others. A key may be prefixable by a prefix
key in one combination of shift states and not in another. A key may
be prefixable by multiple prefix keys and vice versa.

*/
typedef struct
{
	unsigned char   scanCode;
	unsigned char   shifted;
	unsigned char   value[MAX_CAN_PREFIX];
} Prefix;

/*

A normal key is not prefixed.

There are three types of normal keys in KEYB:

	1) Normal normal keys. There meaning changes with the key state
	   just as in the US -- though the extended scan codes may be different.

	2) Numbers. The keys on the numberic keypad where the Shift key overrides
		the NumLock state.

	3) Letters. A letter is a key that is affected by the Caps Lock key.
		The Shift key can override the CapsLock state.


*/
typedef struct
{
	unsigned char scanCode;
	unsigned char shifted;
	unsigned char value;
} NormalKey;

typedef struct
{
	char            country[4];
	int				 codePage;
	int				 euroCapsLock;
	unsigned char * canPrefixList;
	Prefix *        prefixes;
	NormalKey *	    normalKeys;
	NormalKey *		 letters;
	NormalKey *		 numbers;
} KeyMapping;

// NOTE:
// if top row of numbers must be shifted to get numbers (vs. accented chars)
//	and if you get numbers when CAPS LOCK is on then euroCapsLock should
// be set to 1. If you still get accented chars when CAPS LOCK is on then
// euroCapsLock should be 0.
//
typedef struct
{
	char	country[4];
	int codePage;
	int euroCapsLock;
	int totalLen;
	int numPrefixes;
	int numNormalKeys;
	int numLetters;
	int numNumbers;
} Header;

typedef struct
{
	char copyright[0x30];
	int numMappings;
	int maxPrefixes;
	int maxNormalKeys;
	int maxLetters;
	int maxNumbers;
	int hotkeytype; // 0 = keyboard.sys: <Ctrl><Alt><F1> | <Ctrl><Alt><F2>
} GlobalHeader;	 // 1 = keybrd2.sys:  <Alt><LShift>   | <Alt><Rshift>
