;  $Id: DATALGHT.INC 1.6 2003/03/20 17:11:04Z thomd Exp $
;
;  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;  Copyright (c) 1996-2003 Datalight, Inc.
;  All Rights Reserved
;
;  Datalight, Inc. is a Washington corporation located at:
;
;  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
;  Bothell, WA  98021                    Fax: +1.425.951.8094
;  USA                                   Http://www.datalight.com
;
;  This software, including without limitation all source code and
;  documentation, is the confidential, trade secret property of
;  Datalight, Inc., and is protected under the copyright laws of
;  the United States and other jurisdictions.  Portions of the
;  software may also be subject to one or more the following US
;  patents: US#5860082, US#6260156.
;
;  In addition to civil penalties for infringement of copyright
;  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;  penalties for violation of (a) the restrictions on circumvention
;  of copyright protection systems found in 17 U.S.C. 1201 and
;  (b) the protections for the integrity of copyright management
;  information found in 17 U.S.C. 1202.
;
;  U.S. Government Restricted Rights.  Use, duplication,
;  reproduction, or transfer of this commercial product and
;  accompanying documentation is restricted in accordance with
;  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
;  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
;  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;  THIS SOFTWARE IN ERROR.
;
.xlist
;------------------------------------------------------------------------
;       $DOCFILE:DATALGHT.INC
;
;       Copyright (C) Datalight, Inc. 1996-2002
;       All Rights Reserved.
;       Datalight Confidential and Proprietary.
;
;       This include contains Datalight standard defines and macros.
;
;       This include should be included in all Datalight assembly
;       language modules.
;
;	Naming Conventions:
;
;	Macros are divided into three categories:
;	1) "Function" style macros.  These macros are generally large
;	   scale macros that are used to do significant amounts of work.
;	   Examples are the ProcBegin and ProcEnd macros.  These macros
;	   are named similar to C functions using mixed case, but no
;	   leading underscores.
;
;	2) "Inline" style macros.  These macros are designed to perform
;	   relatively simple functions and are often used to create
;	   pseudo assembly instructions -- hence the name inline.  These
;	   macros are named in all lowercase and generally the names
;	   have seven characters or less so that the macros can be easily
;	   used in a sequence of assembly instructions without interrupting
;	   the code flow.
;
;	3) "Helper" macros.  These macros are used by higher level macros
;	   and are not generally called from mainline assembly code.  The
;	   macros are named using mixed case, but are preceded with a '?'.
;	   Really low level helper macros may be preceded by multiple '?'s.
;
;	Data symbols used within macros have a standard naming convention
;	as well.
;
;	Permanent symbols that exist throughout the scope of a module and
;	are designed to be used by multiple macros, consist a '$' sign,
;	followed by a mixed case name, with the first character being
;	lower case.
;
;	Symbols that exist throughout the scope of the project, but are
;	designed to be used internally (by only one macro, or multiple
;	macros), use use the same convention as permanent symbols, but
;	are prefixed with a '?' instead.  Generally any symbols such as
;	these that are only used by a single macro should be declared
;	LOCAL, however this does not apparently work in all cases.
;
;	Local symbols are those that are used internally by a single
;	macro and are declared as LOCAL.  There is no standard naming
;	convention for these symbols, however, a mixed case name with
;	no leading special character is currently being used.
;
;	Global Symbols Usage For This Module
;
;	Global symbols that are initialized and generally remain unchanged
;	throughout the module are prefixed with "$m".  For example, the
;	InitModel macro initializes the "$mModel" symbol to be the memory
;	model to use for the module.
;
;	Global symbols that are re-used by each sequence of "Proc..."
;	macros are prefixed by "$p".  For example, the symbol "$pLang" is
;	initialized to the language to use for each procedure in the
;	module.
;
;       $DOCTODO
;       - Test with USE32.
;       - Test more thoroughly with MASM.
;       - Identify all cross assembler issues between TASM and MASM
;
;       $DOCHISTORY
;       02/17/97 gp     Created from WINLIGHT.INC
;       02/26/97 gp     First working version
;       02/27/97 gp     Modified InitModel to take a language parameter
;       03/20/97 gp     First published version.  Major cleanup and docs.
;       04/30/97 gp     Fixed a problem with MASM adding it's own underscore.
;       07/22/97 gp     Improved parameter validation and better ensure proper
;                       macro sequencing (ProcUses must precede ProcSetup etc).
;       08/26/97 gp     Moved generic macros to DLMACROS.INC
;	10/03/97 gp	Added the STACKWIDTH option (not completed yet)
;	10/08/97 gp	Added the MANUALFRAME option (and made it the default)
;	12/11/97 de	Added the :PTR type option to the ProcArgs macro.
;			This resolves to wptr in near data models and
;			dptr in far data models (COMPACT and LARGE).
;	12/11/97 de	Added the lptr_ds, lptr_es and assign macros that
;			work with :PTR type ProcArgs. These only load
;			a segment register in far data models. The assign
;			data model puts a value at the ptr location without
;			changing any registers.
;       12/18/97 gp     Updated macro headers and names according to new std.
;       03/30/98 gp     Updated ProcBegin to confirm that InitModel has been
;                       run first.
;       06/25/98 de	Updated ProcBegin to upper case PASCAL proc names
;			and to prepend an '_' char to C proc names.
;	06/25/98 de	Added the "NOWARN MCP" directive for TASM so that the
;			"MASM compatibility pass" warning is disabled. This is
;			needed when the UpperCaseName macro is used. The macro
;			is safe as used, the warning is inconsequential.
;       07/02/98 de	Fixed bug in ProcBegin where defining 'C' language
;			for both the ProcBegin and the InitModel macros would
;			generate 2 (vs the desired 1) prepended '_' chars on
;			the proc name.
;-------------------------------------------------------------------------
FALSE           = 0
TRUE            = 1

;; Set MANUALFRAME to FALSE to use the assembler's high level directives
;; for generating stack frames.  This does not work for VMM with it's
;; STACK32 needs, so the MANUALFRAME ability is currently the default.
MANUALFRAME = TRUE

DATALGHT_INC_INCLUDED = TRUE

$pName          equ <>          ;; create symbol so ProcBegin nest check works

;; This appears to be an effective way to tell if TASM or MASM
ifdef ??version
  ASM equ <TASM>

  LOCALS                        ;; enable @@ prefixed symbols to be local
  NOWARN MCP

  include TASM.INC		;; include TASM specific stuff
else
  ASM equ <MASM>

  OPTION M510                   ;; MASM 5.1 compatibility mode
  OPTION PROC:PRIVATE           ;; procs are private by default
  OPTION PROLOGUE:MASMPROLOGUE  ;; specify Prologue macro
  OPTION EPILOGUE:MASMEPILOGUE  ;; specify Epilogue macro

  include MASM.INC              ;; include MASM specific stuff
endif

; Bitflags used by the .TYPE directive
OPTYPE_CODE     equ <0001h>
OPTYPE_DATA     equ <0002h>
OPTYPE_CONST    equ <0004h>
OPTYPE_DIRECT   equ <0008h>
OPTYPE_REG      equ <0010h>
OPTYPE_DEFINED  equ <0020h>
;OPTYPE_SSBASED equ <0040h>     ; not supported by TASM?
OPTYPE_EXTERN   equ <0080h>

; Bitflag assembly mode equates for use with the @Cpu symbol
AMODE_8086      equ <0001h>
AMODE_80186     equ <0002h>
AMODE_80286     equ <0004h>
AMODE_80386     equ <0008h>
AMODE_80486     equ <0010h>
AMODE_80586     equ <0020h>
AMODE_RESERVED1 equ <0040h>
AMODE_PROT      equ <0080h>
AMODE_8087      equ <0100h>
AMODE_RESERVED2 equ <0200h>
AMODE_80287     equ <0400h>
AMODE_80387     equ <0800h>

; Bitflags used by the ?ProcSetup macro
PROCATTR_PUBLIC		equ <0001h>	;;
PROCATTR_EXPORT		equ <0002h>	;; requires LOADDS and PUBLIC
PROCATTR_FAR   		equ <0004h>	;;
PROCATTR_LOADDS		equ <0008h>	;; requires STACKFRAME
PROCATTR_INCBP 		equ <0010h>	;; requires STACKFRAME and FAR
PROCATTR_STACKFRAME	equ <0020h>	;; proc needs a stack frame
PROCATTR_CARGS     	equ <0040h>	;; uses C's right-to-left arguments
PROCATTR_CALLERPOP	equ <0080h>	;; caller cleans up the stack

; common equates for readability
bptr            equ   <byte ptr>
wptr            equ   <word ptr>
dptr            equ   <dword ptr>
fptr            equ   <far ptr>
nptr            equ   <near ptr>


; include generic macros
include DLMACROS.INC


;---------------------------------------------------------------------------
;       $DOCMACRO:InitModel
;
;       Syntax:  InitModel  [model] [,language] [,codewidth] [,stackwidth]
;
;       Where:  "model"      is Small, Compact, Medium, or Large
;       	"language"   is C, PASCAL, or NOLANGUAGE
;               "codewidth"  is USE16 or USE32
;		"stackwidth" is STACK16, STACK32, or STACK32LOW
;
;       All parameters are optional.  The defaults for any unspecified
;       values are Small model, PASCAL language, USE16 code width, and
;	STACK16 stack width.
;
;       Parameters are NOT positional, they may be specified in any order,
;	separated by commas.
;
;	STACK32LOW is a special option that denotes a 32 bit stack where
;	only the low 64 is used.  The macros will use EBP and ESP to set
;	up and remove the stack frame, but arguments and locals will be
;	addressed using BP.  This basically is a hack to avoid a design
;	flaw in the x86 architecture where the ENTER and LEAVE instructions
;	do not work properly in a USE16 code segment that has a 32 bit
;	stack.
;
;	Note:  Under MASM, NOLANGUAGE is the same as PASCAL.
;
;       Model may be specified via this macro, or on the assembler command
;       line using /DMMOD=s|c|m|l, but not both.
;--------------------------------------------------------------------------
InitModel       MACRO   p1,p2,p3,p4,ovr
local           tmpCodeWidth
.errnb <ovr>                    ;; not too many params

;; default module wide global symbols

$mCodeWidth     equ <>	;; code segment width
$mStackWidth	equ <>	;; stack segment width
$mLanguage      equ <>  ;; language
$mModel         equ <>  ;; memory model

;; loop through all the parameters looking for keywords we recognize
irp var,<p1,p2,p3,p4>
  ifb <&var>
    exitm                       ;; exit IRP if no more parameters
  endif

  ifidni        	<&var>,<USE16>
    ifnb $mCodeWidth
      %out "InitModel: CodeWidth specified more than once"
      .err
    endif

    $mCodeWidth equ <var>

  elseifidni    	<&var>,<USE32>
    ifnb $mCodeWidth
      %out "InitModel: CodeWidth specified more than once"
      .err
    endif

    $mCodeWidth equ <var>

  elseifidni    	<&var>,<STACK16>
    ifnb $mStackWidth
      %out "InitModel: StackWidth specified more than once"
      .err
    endif

    $mStackWidth equ <var>

  elseifidni    	<&var>,<STACK32>
    ifnb $mStackWidth
      %out "InitModel: StackWidth specified more than once"
      .err
    endif

    $mStackWidth equ <var>

  elseifidni    	<&var>,<STACK32LOW>
    ifnb $mStackWidth
      %out "InitModel: StackWidth specified more than once"
      .err
    endif

    $mStackWidth equ <var>

  elseifidni    	<&var>,<C>
    ifnb $mLanguage
      %out "InitModel: Language specified more than once"
      .err
    endif

    $mLanguage equ <var>

  elseifidni    	<&var>,<PASCAL>
    ifnb $mLanguage
      %out "InitModel: Language specified more than once"
      .err
    endif

    $mLanguage equ <var>

  elseifidni    	<&var>,<NOLANGUAGE>
    ifnb $mLanguage
      %out "InitModel: Language specified more than once"
      .err
    endif

    ifidni ASM,<TASM>       ;; if using TASM
      $mLanguage equ <var>
    else
      $mLanguage equ <PASCAL> ;; MASM doesn't support NOLANGUAGE, so use PASCAL
    endif

  elseifidni    	<&var>,<SMALL>
    ifnb $mModel
      %out "InitModel: Model specified more than once"
      .err
    endif

    $mModel equ <var>

  elseifidni    	<&var>,<COMPACT>
    ifnb $mModel
      %out "InitModel: Model specified more than once"
      .err
    endif

    $mModel equ <var>

  elseifidni    	<&var>,<MEDIUM>
    ifnb $mModel
      %out "InitModel: Model specified more than once"
      .err
    endif

    $mModel equ <var>

  elseifidni    	<&var>,<LARGE>
    ifnb $mModel
      %out "InitModel: Model specified more than once"
      .err
    endif

    $mModel equ <var>

  else
    %out "InitModel: Unrecognized parameter '&var'"
    .err
  endif
endm		;; end of IRP

;; use default values if none specified

ifb $mCodeWidth
  $mCodeWidth equ <USE16>	;; use default CodeWidth
endif

ifb $mStackWidth
  $mStackWidth equ <STACK16>	;; use default StackWidth
endif

ifb $mLanguage
  $mLanguage equ <PASCAL>	;; use default language
endif


;;------------------------------------------------------------------------
;; Determine desired memory model.  Model should be specified on the
;; assembler command line, or on the InitModel command line, but not both.
;;------------------------------------------------------------------------

;; if it was specified in the source code

ifnb $mModel
  ifdef MMOD
    %out "ERROR: InitModel - Model specified in source file and on command line"
    .err
  endif
else			;; else it was not specified in the source code
  ifndef MMOD
    %out "ERROR: InitModel - Model not specified"
    .err
  endif

  ;; convert abbreviated s, c, m, or l parameters to full names
  ifidni MMOD,<s>
    $mModel equ <Small>
  elseifidni MMOD,<c>
    $mModel equ <Compact>
  elseifidni MMOD,<m>
    $mModel equ <Medium>
  elseifidni MMOD,<l>
    $mModel equ <Large>
  else
    %out "ERROR: InitModel - Model '&MMOD' unrecognized. Must be /DMMOD=s|c|m|s"
    .err
  endif
endif

?SetTextSymbol tmpCodeWidth,%$mCodeWidth

ifidni ASM,<TASM>       ;; if using TASM

  ;; USE16 does not appear to work for Tasm 2.5, assume it works for 3.0
  if ??version LT 300h
    tmpCodeWidth equ <>
  endif

else                    ;; else using MASM

  ;; MASM specifies segment size using OPTION, not MODEL

  % OPTION SEGMENT:$mCodeWidth

  tmpCodeWidth equ <>

endif

;; invoke the desired .MODEL statement for this module
% .model tmpCodeWidth $mModel, $mLanguage

;; init default proc attributes
InitProcBehavior

ENDM


;-----------------------------------------------------------------------------
;       $DOCMACRO:InitProcBehavior
;
;       NOTE:   This macro is a work in progess.  It has been tested
;               primarily under the scenario of initializing the
;               default behaviors.  Override testing has not been
;               completed, nor have all the MASM/TASM compatibility
;               issues been determined and resolved.  To be specific,
;               don't use this macro with any parameters yet!
;
;       This macro initializes the PROC behavior for the ProcBegin/Setup/End
;       macros.  Generally this macro should be used once at the top of each
;       module, though it may be used multiple times if needed.
;
;       This macro is used inside the InitModel macro to set up the default
;       behaviors.  It is not necessary to use this macro unless the
;       default behaviors need to be modified.
;
;       Syntax:  InitProcBehavior [LOADDS=xxx][,INCBP=xxx][,FORCEFRAME=xxx]
;
;       Where "xxx" is one of the following:
;
;               NEVER   - The behavior is never enabled
;               EXPORT  - The behavior is enabled for EXPORT functions
;               DBG     - The behavior is enabled if DEBUG NE 0
;               ALWAYS  - The behavior is always enabled
;
;       Any options that are not specified on the command line are set to
;       the default behaviors which are:
;
;               LOADDS      = EXPORT
;               INCBP       = NEVER
;               FORCEFRAME  = NEVER
;
;       The parameters are not positional, and may be specified in any order.
;-----------------------------------------------------------------------------
InitProcBehavior MACRO   p1,p2,p3,ovr
local   varName,varVal,eqOffset,exprLen
.errnb <ovr>                    ;; not too many params

PROCINIT_NEVER  equ <0>         ;; feature is never used
PROCINIT_EXPORT equ <1>         ;; feature used only if function is EXPORT
PROCINIT_DBG    equ <2>         ;; feature used only if DEBUG NE 0
PROCINIT_ALWAYS equ <3>         ;; feature is always used

; default behaviors for the ProcBegin/Setup/End macros
PROC_LOADDS     = PROCINIT_EXPORT
PROC_INCBP      = PROCINIT_NEVER
PROC_FORCEFRAME = PROCINIT_NEVER

;; loop through all the parameters looking for keywords we recognize
irp var,<p1,p2,p3>
  ifb <&var>
    exitm                       ;; exit if no more parameters
  endif

  exprLen SIZESTR <&var>        ;; calc expression length

  eqOffset INSTR <&var>,<=>     ;; find position of the = sign
  if eqOffset LE 1
    .err
    %out ERROR: InitProcBehavior: Invalid parameter &var
  endif

  ;; get the keyword name and value
  varName SUBSTR <&var>,1,eqOffset-1
  varVal  SUBSTR <&var>,eqOffset+1,exprLen-eqOffset

  ;; look for valid keywords
  ifidn         varName,<LOADDS>
  elseifidn     varName,<INCBP>
  elseifidn     varName,<FORCEFRAME>
  else
    .err
    %out ERROR: InitProcBehavior: Invalid variable name &var ?
  endif

  ;; look for valid data values
  ifidn         varVal,<NEVER>
  elseifidn     varVal,<EXPORT>
  elseifidn     varVal,<DBG>
  elseifidn     varVal,<ALWAYS>
  else
    .err
    %out ERROR: InitProcBehavior: Invalid variable value &var ?
  endif

  ;; create the real keyword and data value symbols
  varName CATSTR <PROC_>,varName
  varVal  CATSTR <PROCINIT_>,varVal

  ;; set the symbol as specified
% varName = varVal

endm
ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:ProcBegin
;
;       Syntax:  ProcBegin  name, [language], [distance], [visiblity]
;
;       Valid "language" values are C and PASCAL.  If not specified, then
;       the InitModel default will be used.
;
;       Valid "distance" values are NEAR and FAR.  If a distance is not
;       specified the MODEL default will be used.
;
;       Valid "visibility" values are PRIVATE, PUBLIC, and EXPORT.  PRIVATE
;       is the default if none are specified.  EXPORT implies PUBLIC and
;       also causes the generation of prologue and epilogue code needed.
;
;       The optional parameters are not position sensitive.  This macro
;       is designed to be used in combination with the ProcUses, ProcArgs,
;       ProcLocals, ProcSetup, and ProcEnd macros.
;
;       $DOCTODO
;       Add support for behavioral override options:  LOADDS, NOLOADDS,
;       INCBP, NOINCBP, FORCEFRAME, NOFORCEFRAME, STACK16, and STACK32.
;---------------------------------------------------------------------------
?PasteName 	MACRO Was, Next
   $pName catstr <Was>, <Next>
endm
$pos		=   0
$cLowerCase	equ <abcdefghijklmnopqrstuvwxyz>
$cUpperCase	equ <ABCDEFGHIJKLMNOPQRSTUVWXYZ>
?UpperCaseName	MACRO Name
  irpc c, Name
%   $pos instr 1, <$cLowerCase>, <c>
    if $pos
      char substr $cUpperCase, $pos, 1
%     ?PasteName <&$pName>, <&char>
    else
%     ?PasteName <&$pName>, <&c>
    endif
  endm
endm
ProcBegin       MACRO pName,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,ovr
local           tempStr
.errb  <pname>          ;; enough parameters
.errnb <ovr>            ;; but not too many

.errnb $pName           ;; no nesting allowed

ifndef $mModel
  .err
  %out "ERROR: ProcBegin: Model not defined -- has InitModel been used"
endif

;; Init everything to defaults
$pName		equ <>
$pLang          equ <>
$pDist          equ <>
$pVis           equ <>
$pUses          equ <>
$pArgs          equ <>
$pLocals        equ <>
$pMasmName      equ <>
$pSetupOK       equ <>
$pAttr		=   0
$pLocalLen	=   0
$pArgLen	=   0

;; iterate through optional parameters looking for
;; valid LANGUAGE, DISTANCE, and VISIBILITY options
irp var,<p1,p2,p3,p4,p5,p6,p7,p8,p9,p10>
  ifb <&var>
    exitm                       ;; exit if no more parameters
  endif

  ifidni	<&var>,<PASCAL>
    ifnb $pLang
      .err
      %out "ERROR: ProcBegin: Multiple languages specified"
    endif
    $pLang equ <PASCAL>

  elseifidni	<&var>,<C>
    ifnb $pLang
      .err
      %out "ERROR: ProcBegin: Multiple languages specified"
    endif
    $pLang equ <C>

  elseifidni	<&var>,<NOLANGUAGE>
    ifnb $pLang
      .err
      %out "ERROR: ProcBegin: Multiple languages specified"
    endif
    $pLang equ <NOLANGUAGE>

  elseifidni	<&var>,<NEAR>
    ifnb $pDist
      .err
      %out "ERROR: ProcBegin: Multiple distances specified"
    endif
    $pDist equ <NEAR>

  elseifidni	<&var>,<FAR>
    ifnb $pDist
      .err
      %out "ERROR: ProcBegin: Multiple distances specified"
    endif
    $pDist equ <FAR>

  elseifidni 	<&var>,<PRIVATE>
    ifnb $pVis
      .err
      %out "ERROR: ProcBegin: Multiple visibilities specified"
    endif
    $pVis equ <PRIVATE>

  elseifidni 	<&var>,<PUBLIC>
    ifnb $pVis
      .err
      %out "ERROR: ProcBegin: Multiple visibilities specified"
    endif
    $pVis equ <PUBLIC>

  elseifidni 	<&var>,<EXPORT>
    ifnb $pVis
      .err
      %out "ERROR: ProcBegin: Multiple visibilities specified"
    endif
    $pVis equ <EXPORT>

  else
    .err
    %out "ERROR: ProcBegin: Unrecognized parameter: &var"
  endif
endm

;;------------------------------------------------------------
;; Use the InitModel default values for anything not specified
;;------------------------------------------------------------
ifb $pVis
  ;; default to PRIVATE if no visibility specified
  $pVis equ <PRIVATE>
endif

ifb $pLang
  ;; default to module default language
  ?SetTextSymbol $pLang,%$mLanguage
endif

;; Upper case all PASCAL names.
;;
ifidni $pLang,<PASCAL>
  ?UpperCaseName <pName>

;; Prepend an '_' to all C names unless .Model x,c is also stated
;; by the user (because that, will cause the assembler to prepend an
;; '_' to the name as well... cool, huh?)
;;
elseifidni $pLang,<C>
  ifidni $mLanguage,<C>
    ?PasteName <>, <pName>
  else
    ?PasteName <_>, <pName>
  endif

;; Names in any other language stay just as they are.
;;
else
  ?PasteName <>, <pName>
endif

ifb $pDist
  ifidni $mModel,<SMALL>
    $pDist equ <NEAR>
  elseifidni $mModel,<COMPACT>
    $pDist equ <NEAR>
  elseifidni $mModel,<MEDIUM>
    $pDist equ <FAR>
  elseifidni $mModel,<LARGE>
    $pDist equ <FAR>
  endif
endif

;;------------------------------------------------------------
;; Initalized $pAttr according to specified options
;;------------------------------------------------------------
ifidni $pLang,<C>
  $pAttr = $pAttr OR PROCATTR_CARGS OR PROCATTR_CALLERPOP
endif

ifidni $pVis,<PUBLIC>
  $pAttr = $pAttr OR PROCATTR_PUBLIC
endif

ifidni $pVis,<EXPORT>
  $pAttr = $pAttr OR PROCATTR_PUBLIC OR PROCATTR_EXPORT OR PROCATTR_LOADDS
endif

ifidni $pDist,<FAR>
  $pAttr = $pAttr OR PROCATTR_FAR
endif

ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:ProcUses
;
;       This macro specifies registers that must be saved and restored.
;       The use of this macro is optional.  This macro must follow
;       ProcBegin and be placed prior to ProcSetup.
;
;       Syntax:  ProcUses  reg, reg, reg, etc
;---------------------------------------------------------------------------
ProcUses        MACRO r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11,r12,r13,r14,r15,r16,r17,r18,r19,r20,ovr
.errb  <r1>                     ;; enough parameters
.errnb <ovr>                    ;; but not too many

.errb  $pName                   ;; must follow a ProcBegin
.errnb $pSetupOK                ;; must precede ProcSetup
.errnb $pUses                   ;; shouldn't have done this already

;; different syntaxes for TASM and MASM

;; loop through all parameters
irp reg,<r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11,r12,r13,r14,r15,r16,r17,r18,r19,r20>
  ifb <&reg>
    exitm
  endif

  ;; build a string that we'll eventually pass to TASM or MASM
  ifidni ASM,<TASM>
    ;; TASM uses a comma seperated list
    $pUses CATSTR $pUses,<&reg&,>
  else
    ;; while MASM uses a space seperated list
    $pUses CATSTR $pUses,<&reg& >
  endif
endm

ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:ProcArgs
;
;       Syntax:  ProcArgs  name:type, name:type, [...], [...], etc
;
;       Regardless whether the function is C or PASCAL, the parameter
;       list should be as specified in the callers parameter list.
;
;	The :type substring may be a standard assembler type such as
;	:BYTE, :WORD, or :DWORD, or it may be :PTR which resolves to
;	either :WORD or :DWORD depending on the memory model.
;
;       Note that ALL arguments must be specified on this line and that
;       a maxiumum of 20 arguments is supported.
;
;       The syntax used for each argument is assembler specific.  Care
;       must be taken to use a syntax that is cross assembler compatible.
;
;       Argument names should begin with a '@@' to allow TASM's local
;       scoping to be enabled.
;---------------------------------------------------------------------------
ProcArgs        MACRO p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,ovr
.errb  <p1>                     ;; enough parameters
.errnb <ovr>                    ;; but not too many

.errb  $pName                   ;; must follow a ProcBegin
.errnb $pSetupOK                ;; must precede ProcSetup
.errnb $pArgs                   ;; shouldn't have done this already

$pAttr = $pAttr OR PROCATTR_STACKFRAME

;; handle first argument since we want no trailing comma when complete
$pArgs  equ <p1>

;; handle remaining arguments
irp reg,<p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20>
  ifb <&reg>
    exitm
  endif

  ;; Concatenate a comma and the next argument, building a string that
  ;; we'll eventually pass to TASM or MASM
  $pArgs CATSTR $pArgs,<,&reg>
endm

ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:ProcLocals
;
;       Syntax:  ProcLocals  name:type, name:type, [...], [...], etc
;
;       Note that ALL arguments must be specified on this line and that
;       a maxiumum of 20 arguments are supported.  Use line continuation
;       to avoid lines longer than 80 characters.
;
;	The :type substring may be a standard assembler type such as
;	:BYTE, :WORD, or :DWORD, or it may be :PTR which resolves to
;	either :WORD or :DWORD depending on the memory model.
;
;       The syntax used for each argument is assembler specific.  Care
;       must be taken to use a syntax that is cross assembler compatible.
;
;       Argument names should begin with a '@@' to allow TASM's local
;       scoping to be enabled.
;
;---------------------------------------------------------------------------
ProcLocals      MACRO p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,ovr
.errb  <p1>                     ;; enough parameters
.errnb <ovr>                    ;; but not too many

.errb  $pName                   ;; must follow a ProcBegin
.errnb $pSetupOK                ;; must precede ProcSetup
.errnb $pLocals                 ;; shouldn't have done this already

$pAttr = $pAttr OR PROCATTR_STACKFRAME

;; handle first parameter outside the loop
$pLocals equ <p1>

;; loop through remaining parameters
irp reg,<p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20>
  ifb <&reg>
    ;; exit if no more parameters
    exitm
  endif

  ;; concatenate remaining parameters into a comma seperated list
  $pLocals CATSTR $pLocals,<,&reg>
endm

ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:ProcSetup
;
;       This macro does the real work of setting up a proc for use.  It
;       takes information built by ProcBegin, ProcUses, ProcArgs, and
;       ProcLocals and generates the appropriate PROC statement for
;       TASM or MASM.
;
;       Syntax:  ProcSetup
;---------------------------------------------------------------------------
ProcSetup       MACRO ovr
.errnb <ovr>                    ;; no parameters allowed

.errb  $pName                   ;; must follow a ProcBegin
.errnb $pSetupOK                ;; can only do this once

$pSetupOK equ <TRUE>

if MANUALFRAME
  ifidni ASM,<MASM>             ;; if this is MASM
    ;; default name to use for MASM
    $pMasmName equ $pName
  endif

  ?ProcSetup

  exitm
endif

ifidni ASM,<TASM>               ;; if this is TASM

  ?TasmProcSetup

else                            ;; else, must be MASM

  ?MasmProcSetup

endif

ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:ProcEnd
;
;       This macro is used to close a procedure.  Essentially the assembler
;       does all the work for us, generating the proper cleanup code when
;       it hits the "ret" instruction.
;
;       Syntax:  ProcEnd
;---------------------------------------------------------------------------
ProcEnd         MACRO ovr
.errnb <ovr>                    ;; no parameters allowed

.errb $pName                    ;; must follow a ProcBegin
.errb $pSetupOK                 ;; must follow a ProcSetup

  if MANUALFRAME
    ;; if using MANUALFRAME, we need to do all the work ourselves
    ?ProcEnd
  else
    ;; if not using MANUALFRAME, just let the assembler do all the
    ;; work when it encounters this RET instruction
    ret     	                ;; return NEAR or FAR (the assembler decides)
  endif

  ifidni ASM,<TASM>           ;; if this is TASM
    $pName  endp              ;; end the procedure
  else
    $pMasmName endp
  endif

  ;; Reset all externally used vars to ensure they're not used elsewhere
  $pName        equ <>
  $pLang        equ <>
  $pDist        equ <>
  $pVis         equ <>
  $pUses        equ <>
  $pArgs        equ <>
  $pLocals      equ <>
  $pMasmName    equ <>
  $pSetupOK     equ <>
  $pAttr	=   0
  $pLocalLen	=   0
  $pArgLen	=   0

ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:?ProcSetup
;
;	This helper macro is used to generate the correct prologue code when
;	the MANUALFRAME option is in effect.
;
;	In reality, this is one of the few macros in this file that actually
;	generates any code.
;---------------------------------------------------------------------------
?ProcSetup      MACRO ovr
local		dist
.errnb <ovr>                    ;; no overflow

  dist equ <NEAR>

  if $pAttr AND PROCATTR_FAR
    dist equ <FAR>
  endif

  if $pAttr AND PROCATTR_PUBLIC
    % PUBLIC $pName		;; declare the procedure PUBLIC if need be
  endif

  % $pName PROC	dist		;; execute the actual PROC directive

  ifnb $pLocals
    %  ?DeclareLocals $pLocals
  endif

  ifnb $pArgs
    %  ?DeclareArgs $pArgs
  endif

  ife $pAttr AND PROCATTR_STACKFRAME

    ;; just because we don't need a frame, we may still want DS loaded
    if $pAttr AND PROCATTR_LOADDS

      if $pAttr AND PROCATTR_EXPORT
        push    ds              ;; Put DS into AX -- we will place back
        pop     ax              ;; in DS later.  This sequence may be
        nop                     ;; altered by the OS at fixup time
        push    ds
        mov     ds,ax
      else
        push    ds		;; not EXPORT, so do a normal DS load
        mov     ax,DGROUP
        mov     ds,ax
      endif
    endif

  else
    ;;--------------------------------------------
    ;; At this point we know we need a stack frame
    ;;--------------------------------------------

    if $pAttr AND PROCATTR_LOADDS   ;; if LOADDS and EXPORT, generate first
      if $pAttr AND PROCATTR_EXPORT ;; half of DS init code.
        push    ds                ;; Put DS into AX -- we will place back
        pop     ax                ;; in DS later.  This sequence may be
        nop                       ;; altered by the OS at fixup time.
      endif
    endif

    ifidni $mStackWidth,<STACK16>
      if $pAttr AND PROCATTR_INCBP
        inc     bp              ;; mark as a far procedure for stack walking
      endif

        push    bp              ;; create the frame
        mov     bp,sp

      if $pLocalLen NE 0
        sub     sp,$pLocalLen   ;; make space on the stack for locals
      endif

    else

      if $pAttr AND PROCATTR_INCBP
        inc     ebp             ;; mark as a far procedure for stack walking
      endif

        push    ebp             ;; create the frame
        mov     ebp,esp

      if $pLocalLen NE 0
        sub     esp,$pLocalLen  ;; make space on the stack for locals
      endif
    endif

    if $pAttr AND PROCATTR_LOADDS   ;; Load up DS with the value in AX
      if $pAttr AND PROCATTR_EXPORT
        push    ds              ;;
        mov     ds,ax           ;; EXPORT functions just get patched in code
      else
        push    ds              ;;
        mov     ax,DGROUP       ;; otherwise just use DGROUP
        mov     ds,ax
      endif
    endif
  endif

  ;;--------------------------------------------
  ;; Save any registers as specified
  ;;--------------------------------------------
  ifnb $pUses

    ;; Anytime LOADDS is used, AX will get trashed and DS will be
    ;; preserved automatically, so provide a polite warning if
    ;; either of these registers are specified with ProcUses.

    if $pAttr AND PROCATTR_LOADDS
      % irp reg,<$pUses>
        ifidni <&reg>,<AX>
          %out "WARNING: ?ProcSetup:  AX has already been trashed"

        elseifidni <&reg>,<EAX>
          %out "WARNING: ?ProcSetup:  EAX has already been trashed"

        elseifidni <&reg>,<DS>
          %out "WARNING: ?ProcSetup:  DS has already been saved"

        endif
      endm
    endif
      %	pusher  $pUses               ; save regs specified with USES
  endif

ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:?DeclareArgs
;
;	This helper macro does the real work of ProcArgs.
;---------------------------------------------------------------------------
?DeclareArgs	MACRO	p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,ovr
local		reg
.errnb <ovr>

  $pArgLen = 0

  ifidni $mStackWidth,<STACK16>
    ?bpAdjust = 2              ;; saved BP is really BP
  else
    ?bpAdjust = 4              ;; saved BP is really EBP
  endif

  if $pAttr AND PROCATTR_FAR
    ?bpAdjust = ?bpAdjust + 4 ;; adjust for FAR return address
  else
    ?bpAdjust = ?bpAdjust + 2 ;; adjust for NEAR return address
  endif

  if $pAttr AND PROCATTR_CARGS

    ;; use this IRP for parameters pushed right-to-left (like C)

    irp reg,<p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20>
      ifnb <&reg>
        ??ProcessArg &reg,%?bpAdjust
      endif
    endm

  else

    ;; use this IRP for parameters pushed left-to-right (like PASCAL)

    irp reg,<p20,p19,p18,p17,p16,p15,p14,p13,p12,p11,p10,p9,p8,p7,p6,p5,p4,p3,p2,p1>
      ifnb <&reg>
        ??ProcessArg &reg,%?bpAdjust
      endif
    endm

  endif

ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:?DeclareLocals
;
;	This helper macro does the real work of validating and properly
;	identifying the type of the given local variable.
;---------------------------------------------------------------------------
?DeclareLocals	MACRO	p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,ovr
.errnb <ovr>

  $pLocalLen = 0

  irp reg,<p20,p19,p18,p17,p16,p15,p14,p13,p12,p11,p10,p9,p8,p7,p6,p5,p4,p3,p2,p1>
  local dpos, dwidth, dnumber, dbrack1, dbrack2, lenStr, typ
  ifnb <&reg>

    dwidth = 0
    lenStr SIZESTR <&reg>
    dpos INSTR <&reg>,<:>            ;; is the string present
    if dpos
      typ SUBSTR <&reg>,dpos,lenStr-dpos+1
      ifidni typ,<:BYTE>
        dwidth  = 1
        ?ptrType equ <bptr>
      elseifidni typ,<:WORD>
        dwidth  = 2
        ?ptrType equ <wptr>
      elseifidni typ,<:DWORD>
        dwidth  = 4
        ?ptrType equ <dptr>
      elseifidni typ,<:PTR>
        ifidni     $mModel,<COMPACT>
       	  dwidth  = 4
          ?ptrType equ <dptr>
        elseifidni $mModel,<LARGE>
          dwidth  = 4
          ?ptrType equ <dptr>
        else
          dwidth  = 2
          ?ptrType equ <wptr>
        endif
      else			;; else assume its a structure
        lenStr SIZESTR typ	;; lop off the leading colon
        typ SUBSTR typ,2,lenStr-1
        ?ptrType equ <>		;; not used for structures
        dwidth = TYPE typ	;; get the size of the structure
      endif
    endif

    if dwidth EQ 0
      .err
       %OUT "ERROR:  ProcLocals:  Data type expected '&reg'"
      exitm
    endif

    ;; now look for arrays in the style varname[23]:byte

    dbrack1 INSTR <&reg>,<[>	;; is the left bracket present
    if dbrack1
      dbrack2 INSTR <&reg>,<]>	;; find the right bracket
      ife dbrack2
        .err
        %out "ERROR: ProcLocals:  Ill formed local variable '&reg'"
        exitm
      endif
      dnumber SUBSTR <&reg>,dbrack1+1,(dbrack2-dbrack1)-1
      dwidth = dwidth * dnumber	;; multiply element count by type width
      dpos = dbrack1		;; position to end of the var name
    endif

    ?locName SUBSTR <&reg>,1,dpos-1
    $pLocalLen = $pLocalLen + dwidth

    ;; define the symbol to the assembler
    ??DefineLocal %?locName,%?ptrType,%$pLocalLen

  endif
  endm

  if $pLocalLen and 0001h
    $pLocalLen = $pLocalLen + 1    ;; make sure the number is even
  endif

ENDM


;---------------------------------------------------------------------------
;       $DOCMACRO:?ProcEnd
;
;	This helper macro is the back-end of ProcEnd.  This is one of the
;	few macros that generates actual code.
;
;	This macro is only used if MANUALFRAME is in effect (the default).
;---------------------------------------------------------------------------
?ProcEnd	MACRO	ovr
.errnb <ovr>

  ifnb $pUses
      %	popper	$pUses
  endif

  if $pAttr AND PROCATTR_LOADDS
	pop	ds
  endif

  if $pAttr AND PROCATTR_STACKFRAME
    ifidni $mStackWidth,<STACK16>
	mov	sp,bp
	pop	bp
      if $pAttr AND PROCATTR_INCBP
	dec	bp		;; Remove the increment of BP if necessary
      endif
    else
	mov	esp,ebp
	pop	ebp
      if $pAttr AND PROCATTR_INCBP
	dec	ebp		;; Remove the increment of BP if necessary
      endif
    endif
  endif

  if $pAttr AND PROCATTR_CALLERPOP	;; Caller pops stack arguments
    if $pAttr AND PROCATTR_FAR		;; FAR proc?
        retf
    else
        retn
    endif
  else                    	        ;; Callee pops args
    if $pAttr AND PROCATTR_FAR		;; FAR proc?
        retf	$pArgLen
    else
        retn	$pArgLen
    endif
  endif

ENDM



;---------------------------------------------------------------
;	$DOCMACRO:??ProcessArg
;
;	This helper macro does the meat of the work in properly
;	validating and determining the type of the given
;	argument.
;---------------------------------------------------------------
??ProcessArg	MACRO arg,bpOffset,ovr
local	dpos,dwidth,typ,lenStr
.errnb <ovr>
.errb <bpOffset>

  dwidth = 0
  lenStr SIZESTR <&arg>
  dpos INSTR <&arg>,<:>            ;; is the string present
  if dpos
    typ SUBSTR <&arg>,dpos,lenStr-dpos+1
    ifidni typ,<:BYTE>
      dwidth  = 1
    elseifidni typ,<:WORD>
      dwidth  = 2
    elseifidni typ,<:DWORD>
      dwidth  = 4
    elseifidni typ,<:PTR>
      ifidni     $mModel,<COMPACT>
      	dwidth  = 4
      elseifidni $mModel,<LARGE>
      	dwidth  = 4
      else
      	dwidth  = 2
      endif
    endif
  endif

  if dwidth EQ 0
    .err
     %OUT "ERROR:  ProcArgs:  Data type expected '&arg'"
    exitm
  endif

  ?locName SUBSTR <&arg>,1,dpos-1

  ?argAdjust = $pArgLen + bpOffset

  if dwidth EQ 1
    ??DefineArg %?locName,bptr,%?argAdjust
    $pArgLen = $pArgLen + 2

  elseif dwidth EQ 2
    ??DefineArg %?locName,wptr,%?argAdjust
    $pArgLen = $pArgLen + 2

  elseif dwidth EQ 4
    ??DefineArg %?locName,dptr,%?argAdjust
    $pArgLen = $pArgLen + 4
  endif

ENDM


;-----------------------------------------------------------
;	$DOCMACRO:??DefineArg
;
;	This is the lowest level helper macro that does the
;	actual definition of arguments to a proc.
;-----------------------------------------------------------
??DefineArg	MACRO varname,typ,loc

  ifidni $mStackWidth,<STACK32>
    varname equ <typ ss:[ebp+loc]>

  elseifidni $mStackWidth,<STACK32LOW>
    varname equ <typ ss:[bp+loc]>

  else
    varname equ <typ ss:[bp+loc]>

  endif

ENDM


;-----------------------------------------------------------
;	$DOCMACRO:??DefineLocal
;
;	This is the lowest level helper macro that does the
;	actual definition of local variables for a proc.
;-----------------------------------------------------------
??DefineLocal	MACRO varname,typ,loc

  ifidni $mStackWidth,<STACK32>
    varname equ <typ ss:[ebp-loc]>

  elseifidni $mStackWidth,<STACK32LOW>
    varname equ <typ ss:[bp-loc]>

  else
    varname equ <typ ss:[bp-loc]>

  endif

ENDM


;---------------------------------------------------------------
;	$DOCMACRO:lptr_ds
;
;	Loads a ptr value into DS:<reg>, altering both registers
;---------------------------------------------------------------
;
lptr_ds MACRO reg, ptrName
.errb <reg>
.errb <ptrName>
    ifidni     $mModel,<COMPACT>
	lds	reg, ptrName
    elseifidni $mModel,<LARGE>
	lds	reg, ptrName
    else
	mov	reg, ptrName
    endif

ENDM


;---------------------------------------------------------------
;	$DOCMACRO:lptr_es
;
;	Loads a ptr value into ES:<reg>, altering both registers
;---------------------------------------------------------------
;
lptr_es MACRO reg, ptrName
.errb <reg>
.errb <ptrName>
    ifidni     $mModel,<COMPACT>
	les	reg, ptrName
    elseifidni $mModel,<LARGE>
	les	reg, ptrName
    else
	mov	reg, ds
	mov	es,  reg
	mov	reg, ptrName
    endif

ENDM


;----------------------------------------------------------------------------
;	$DOCMACRO:assign
;
;	Derefernces the pointer variable <ptrName> and stores <value> there
;	using ds:<reg>. Assumes <value> is immediate or in a register.
;	Changes no registers.
;----------------------------------------------------------------------------
;
assign	MACRO	reg, ptrName, value
.errb <reg>
.errb <ptrName>
.errb <value>
    ifidni     $mModel,<COMPACT>
	push	ds
    elseifidni $mModel,<LARGE>
	push	ds
    endif
	push	reg
	lptr_ds	reg, ptrName
	mov	[reg], value
   	pop	reg
    ifidni     $mModel,<COMPACT>
	pop	ds
    elseifidni $mModel,<LARGE>
	pop	ds
    endif

ENDM


;---------------------------------------------------------------
;	$DOCMACRO:ProcBegin_f
;
;	Prepend an _f to the name argument of the ProcBegin
;       macro if the module is being built in large model but
;       not if the large model library is being built.  Whether
;       the large library is being built is determined by the
;       define of LARGELIBS being non zero.
;
;       An example of use is the strchr() and _fstrchr() standard
;       C functions that are built from the same source file
;---------------------------------------------------------------
;
ProcBegin_f MACRO pName,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,ovr

IFNDEF LARGELIBS
LARGELIBS equ 0
ENDIF

IF LARGELIBS
	ProcBegin <pName>,<p1>,<p2>,<p3>,<p4>,<p5>,<p6>,<p7>,<p8>,<p9>,<p10>,<ovr>
ELSE
IFIDNI $mModel,<Large>
	ProcBegin <_f&pName>,<p1>,<p2>,<p3>,<p4>,<p5>,<p6>,<p7>,<p8>,<p9>,<p10>,<ovr>
ELSE
	ProcBegin <pName>,<p1>,<p2>,<p3>,<p4>,<p5>,<p6>,<p7>,<p8>,<p9>,<p10>,<ovr>
ENDIF
ENDIF

ENDM


.list
