//
//   $Id: DEVICE.H 1.4 2002/10/15 21:41:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DEVICE.H
	Created 05-22-98 by Dennis Edwards

   Copyright (c) 1998-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains equates that are relevant to DOS device drivers.


*********************************************************/


#if !defined(__DEVICE_H__)
#define	__DEVICE_H__	1


/*
   DEVICE HEADER
   Description: The beginning of a valid device driver

   Reference: Adv MS-DOS, p 264  -or-  DOS Ref, p 340

   Device Attribute Char:
   1xxx xxxx xxxx xxxx    Character Device (1=char)
   x1xx xxxx xxxx xxxx    IOCTL (4402H-4405H)
   xx1x xxxx xxxx xxxx    Output Until Busy
   xxxx 1xxx xxxx xxxx    Open/Close
   xxxx xxxx 1xxx xxxx    Query IOCTL (4410H)
   xxxx xxxx x1xx xxxx    Generic IOCTL (440CH)
   xxxx xxxx xxx1 xxxx    FAST I/O (INT 29H)
   xxxx xxxx xxxx 1xxx    Clock Device
   xxxx xxxx xxxx x1xx    NUL Device
   xxxx xxxx xxxx xx1x    Stdout Device
   xxxx xxxx xxxx xxx1    Stdin Device

   Device Attribute Disk:
   0xxx xxxx xxxx xxxx    Disk Device (0=disk)
   x1xx xxxx xxxx xxxx    IOCTL (4402H-4405H)
   xx1x xxxx xxxx xxxx    IBM block format (0=FAT, 1=scratch?)
   xxx1 xxxx xxxx xxxx    Network device??
   xxxx 1xxx xxxx xxxx    Removable Media (4408H)
   xxxx xxxx 1xxx xxxx    Query IOCTL (4411H)
   xxxx xxxx x1xx xxxx    Generic IOCTL (440DH)
   xxxx xxxx xxxx xx1x    Driver Supports 32 bit sectors

struct dev_hdr
{
   struct dev_hdr far *next; // 00 ptr to next device
   unsigned attr;            // 04 attribute of device
   unsigned strategy;        // 06 ptr to strat routine
   unsigned service;         // 08 ptr to serv routine
   char name[8];             // 0A name of char device
};
*/

// device header attributes (see above)
//
#if !defined(DEVA_CHAR)
#define DEVA_CHAR		0x8000	/* 1=char, 0=disk device */
#define DEVA_IOCTL	0x4000	/* device supports ioctl calls */
#define DEVA_IBMBLK	0x2000	/* 0=fat, 1=scratch */
#define DEVA_REMOVE	0x0800	/* Removeable media call supported */
#define DEVA_OPNCLS	0x0800	/* Open/Close device call supported */
#define DEVA_QUERY	0x0080	/* query generic IOCTL */
#define DEVA_GENIO	0x0040	/* generic IOCTL */
#define DEVA_FAST		0x0010	/* INT 29H is supported */
#define DEVA_CLOCK	0x0008	/* this is the CLOCK driver */
#define DEVA_NUL		0x0004	/* this is the NUL driver */
#define DEVA_CON		0x0003	/* this is the CONsole driver */
#define DEVA_32BIT	0x0002	/* device supports 32 bit sectors */
#endif


/*

  Device Request Header Fields

  When DOS calls the device driver, the request
  header (pointed to by ES:BX) may contain some
  or all of the following fields:

*/
#if !defined(REQHDR_CMDLEN)
#define REQHDR_CMDLEN		0x00  // length of command
#define REQHDR_UNIT		   0x01  // unit # for disks
#define REQHDR_CMD		   0x02  // the command itself
#define REQHDR_STATUS	   0x03  // status word (0 on start)
#define REQHDR_MEDIA		   0x0D  // media byte (disks)
#define REQHDR_TRANS		   0x0E  // (far *) transfer address
#define REQHDR_COUNT		   0x12  // count for read/write
#define REQHDR_IOCTL		   0x13  // (far *) buffer for generic IOCTL (440DH)
#define REQHDR_START		   0x14  // start for read/write
#define REQHDR_START32	   0x1A  // start for read/write on 32 bit sectors
#endif


/*

	device driver command codes

	These are the numbers that may appear in the REQHDR_CMD field of the
	the device driver header.

*/
#if !defined(DEV_INIT)
#define DEV_INIT		0x00	/* initialize the device */
#define DEV_MEDIA		0x01	/* has disk changed?     */
#define DEV_BPB		0x02	/* get the BPB           */
#define DEV_IOREAD	0x03	/* ioctl read            */
#define DEV_READ		0x04	/* read from the device  */
#define DEV_NDREAD	0x05	/* non destructive read  */
#define DEV_INSTAT	0x06	/* input status          */
#define DEV_FLUSH		0x07	/* flush input buffers   */
#define DEV_WRITE		0x08	/* write to device       */
#define DEV_VWRITE	0x09	/* write with verify     */
#define DEV_OUTSTAT	0x0A	/* output status         */
#define DEV_FLUSHO	0x0B	/* flush output buffers  */
#define DEV_IOWRITE	0x0C	/* ioctl write           */
#define DEV_OPEN		0x0D	/* ioctl open            */
#define DEV_CLOSE		0x0E	/* ioctl close           */
#define DEV_REMOVE	0x0F	/* is device removeable? */
#define DEV_OUTBUSY	0x10	/* output until busy     */
#define DEV_GENIO		0x13	/* generic ioctl         */
#define DEV_GETDEV	0x17	/* get logical device    */
#define DEV_SETDEV	0x18	/* set logical device    */
#define DEV_QUERY		0x19	/* query generic ioctl   */
#endif


/*
	Device Error Codes

  If the device must return an error, the following
  are possible error conditions. These values are placed
  in the low byte of the device header status field word.
  The high byte of status field word contains the completion
  code bit mask bits defined immediately below.

  NOTE: These are the same errors identified by MSDOS v6.22 COMMAND.COM.
        Microsoft only documents errors 0-C in their prog. ref.
        See PC Interrupts (2nd Ed.) Pg. 524.

*/
#if !defined(DEVERR_WRITEPROT)
#define DEVERR_WRITEPROT	0x00	// write protected media
#define DEVERR_BADUNIT		0x01	// invalid unit #
#define DEVERR_NOTREADY		0x02	// drive not ready (door open?)
#define DEVERR_BADCMD		0x03	// unknown command
#define DEVERR_CRC			0x04	// data error (CRC)
#define DEVERR_BADSTRUCT	0x05	// bad request header length
#define DEVERR_SEEK			0x06  // seek error
#define DEVERR_BADMEDIA		0x07	// unknown media type
#define DEVERR_SECNOTFOUND	0x08	// sector not found
#define DEVERR_NOPAPER		0x09	// printer out of paper
#define DEVERR_WRITEFAULT	0x0A  // write fault
#define DEVERR_READFAULT	0x0B	// read error
#define DEVERR_GENERALFAIL	0x0C  // general failure
#define DEVERR_SHARING  	0x0D	// Sharing violation
#define DEVERR_LOCKING  	0x0E	// Lock violation
#define DEVERR_DISKCHANGE  0x0F	// Invalid disk change
#define DEVERR_NOFCB  		0x10	// FCB Unavailable
#define DEVERR_SHAREBUFF	0x11	// Sharing buffer overflow
#define DEVERR_BADCODEPG	0x12	// Code page mismatch
#define DEVERR_NOINPUT  	0x13  // Out of input
#define DEVERR_NODISKSPACE	0x14  // Insufficient disk space
#endif


// The possible values returned in the high byte of the req.hdr.
// status word (see above)
//
#if !defined(DEVSTAT_DONE)
#define DEVSTAT_DONE			0x0100
#define DEVSTAT_BUSY			0x0200
#define DEVSTAT_ERROR		0x8000
#endif


#if defined(DDTDISK)

/*
	This is the struct passed by reference in the REQHDR_IOCTL
	field on an int 440Dh call to issue IOCTL function 0x8D2 (The
	proprietary request for failure debugging call provided by
	Datalight's DDTDISK).  The meaning of the first two fields is
	always the same, the last four fields may changed depending
	on the device call being exercised by the test app. Actual
	field definitions will be noted in the comments in the
	FakingError() function in VDISK.C.

	The struc is not copied but must exist in the App's memory space
	until the "completion" of the specified DOS call.

*/
struct FailureRequest
{
	unsigned char	cCmdCode;		// 0: the command we want to fail
	unsigned char	cErrCode;		// 1: the error code we want to be returned
	unsigned       uRetData;		// 2: ?, on a write, number of sectors written
	unsigned long	ulXData1;		// 4: ?, on a write, first qualifying offset
	unsigned long	ulXData2;		// 8: ?, on a write, last qualifying offset
	unsigned long	ulXData3;		// C: ?, on a write, number times to error.
};
typedef struct FailureRequest FailureRequest;
typedef struct FailureRequest far * lpFR;

#define sectors_done(lpFR) 	(*(unsigned far *)((char far *)(lpFR)+2))
#define first_sector(lpFR)		(*(unsigned long far *)((char far *)(lpFR)+4))
#define last_sector(lpFR)		(*(unsigned long far *)((char far *)(lpFR)+8))
#define error_count(lpFR)		(*(unsigned long far *)((char far *)(lpFR)+0xC))

#define starting_sector(lpRH) ( (*(unsigned far *)     ((char far *)(lpRH)+REQHDR_START) == 0xFFFF)?\
										   *(unsigned long far *)((char far *)(lpRH)+REQHDR_START32) :\
										   *(unsigned far *)     ((char far *)(lpRH)+REQHDR_START)  )
#define sector_count(lpRH) ( *(unsigned far *)((char far *)(lpRH)+REQHDR_COUNT) )

#endif

#endif // #if !defined(__DEVICE_H__)
