//
//   $Id: DLAPP.H 1.6 2002/10/15 21:41:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   DLAPP.H  created 04/24/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Datalight ROM-DOS Tools Library:

   General application tools, for startup and exit.



	Interface:
		int ParseCmdline(int argCount, char *argStrings[], ARGDEF argDefs[]);

	Parameters:
		argCount			# of cmdline string(s)
		argStrings		cmdline string(s)
		argDefs			definitions for expected arguments (NULL terminated)

	Return Value:
		# of parameters left after removing switches

	Example:
		#include "..\include\dlapp.h"

		if(ParseCmdline(argc-1, argv+1, moveArgs) != 2)
			Error("Invalid # of parameters!");

	Notes:
		- Does NOT handle double-byte international strings!

	Description:
		Scans a list of C *argv[] command-line arguments for strings
		and switches. This module also depends on the external, user
		supplied routines (described below):

			void Usage(void);
			void OptError(char *pMsg);

		In a DEBUG version you will also need to satisfy ASSERT()
		with:

		void InternalError(int fileNum, int lineNum);

	The set of valid command line arguments is defined by the
	contents of the null terminated list of ARGDEF structs
	specified in the call to ParseCmdline(). Each ARGDEF entry
	defines a single unique argument; the properties specified
	in each ARGDEF are the text of the command line argument,
	the type and location of the destination variable associated
	with that argument and, in the case of strings, the
	maximum length of the destination array. The routine calls
	ABORT() for undefined argument types.

	Options are processed as they are encountered in the actual
	argument list supplied by the user at runtime. When a valid
	option is encountered, its type is evaluated and the variable
	referenced by the ARGDEF is filled in with the value
	specified by the user. This process is destructive to the
	argStrings array.

	If no arguments appear on the command line then Usage() is
	called. If an argument is encountered that is not present in
	the list of ARGDEFs or if the value specified by the user
	cannot be evaluated then OptError() is called with a pointer
	to the invalid (NUL terminated) sub-string.

	Revision History
	03/16/99	jmb	Added DATE_SWITCH, LFN_STRING_SWITCH, LFN_STRING_ARG
						and GetLFNCommandLine
********************************************************************/

#ifndef __DLAPP_H
#define __DLAPP_H

//
// Currently in a batch file?
//
bool AreInBatch(void);


// The routine to used to install the default int 23h, int 24h
// and signal handlers.  This function calls atexit() to provide
// automatic remove of the interrupt handlers.
//
void InstallExitPackage(void);

// The type of an int 23h user routine and the access routine
// used to establish the user routine. The access routine returns
// the previously installed handler. The user handler must return
// IGNORE_CTRL_C to continue the program otherwise the program is
// terminated using the low byte of the handlers return code as
// the argument to exit().
//
// return code requred by user ^C handler to continue program operation
//
#define IGNORE_CTRL_C	-1

typedef int far (*CtrlBrkHandlerType)(void);
CtrlBrkHandlerType SetCtrlBrkHandler(CtrlBrkHandlerType handler);

// The type of an int 24h user routine and the access routine
// used to establish the user routine. The access routine returns
// the previouslsy installed handler. See ..\DOSTRUCT.H for
// the definintion of the DEV_HDR (device driver header) struct.
// The user handler need not return but if it does, it must
// must return one of the INT24_xxx codes defined below.
//
// allowable return values for a user int 24h handler
//
#define INT24_IGNORE		0
#define INT24_RETRY	  	1
#define INT24_ABORT	   2
#define INT24_FAIL	   3

typedef int far (*CritErrHandlerType)(unsigned ax, unsigned di, LPDEV_HDR bpsi);
CritErrHandlerType SetCritErrHandler(CritErrHandlerType handler);

//
// this defines a single cmdline argument (switch or otherwise)
//
typedef struct
{
	int	type;
	void *pAddress;
	char	switchChar;		// used only for switch args
	int	maxLen;			// used only for string args
	int	fChanged;
} ARGDEF;

//
// argument types
//
#define BOOL_SWITCH		1		// first switch type
#define INT_SWITCH		10
#define LONG_SWITCH		11
#define HEXINT_SWITCH	12
#define LFN_STRING_SWITCH 47
#define DATE_SWITCH		48
#define STRING_SWITCH	49		// last switch type
#define INT_ARG			50		// first arg type
#define LONG_ARG			51
#define LFN_STRING_ARG 97
#define STRING_ARG		99		// last arg type

//
// exit code used by ParseCmdline
//
#define EXITCODE_BADOPT	99

//
//  flags used by SetParseFlags()
//
//  This is are bit flags so they are stackable
//
#define PARSEF_DEFAULT		0x0000
#define PARSEF_NOARGSOK		0x0001
#define PARSEF_NOSAMEARGS	0x0002
#define PARSEF_MULTIPASS	0x0004

//
// parse cmdline function
//
int ParseCmdline(int argCount, char *argStrings[], ARGDEF argDefs[]);
void SetParseFlags(int iFlags);
void GetLFNCommandLine(char * szLFNCmdLine, unsigned uMaxLen);

//
//  User supplied error functions
//
void OptError(char *pMsg);
void Usage(void);

#endif
