//  $Id: DLINT13.H 1.10 2003/03/20 16:48:12Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1999-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/***************************************************************************
	$DOCMODULE:DLINT13.H

	Copyright (c) 1999-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	This module defines C functions and structures to interaction with the
	BIOS Int 13h disk functions.

	$DOCHISTORY
	03/30/1999 dg  Created by Drew Gislason
****************************************************************************/

// dli13f00.obj
//
bool PASCAL DlBiosResetDisk(int iDrive);

// dli13f01.obj
//
int PASCAL DlBiosGetDiskStatus(int iDrive);

// maximum # of sectors that can be transfered (note: floppies cannot pass
// track boundaries on a transfer)
#define MAX_TRANSFER_SECTORS	0x7F

/*
	Used for the BIOS disk functions read, write and verify.
*/
typedef struct BiosDiskTransferPacket BDTP;
typedef BDTP * PBDTP;
typedef BDTP far * LPBDTP;
struct BiosDiskTransferPacket
{
	int	iTrack;
	uchar	ucSector;
	uchar	ucHead;
	int	iNumSectors;
	void far * lpBuffer;
};

// dli13f02.asm
//
bool PASCAL DlBiosReadSectors(int iDrive, PBDTP pbdtpPacket);

// dli13f03.asm
//
bool PASCAL DlBiosWriteSectors(int iDrive, PBDTP pbdtpPacket);

// dli13f04.asm
//
bool PASCAL DlBiosVerifySectors(int iDrive, PBDTP pbdtpPacket);


/*
	Int 1Eh, this structure is also returned by Int 13h fn 08h.
*/
typedef struct DiskBaseTable DBT;
typedef DBT * PDBT;
typedef DBT far * LPDBT;
struct DiskBaseTable
{
	uchar	uc1stSpecifyByte;
	uchar	uc2ndSpecifyByte;
	uchar	ucMotorTicks;			// time from last access to turn off motor
	uchar	ucBpsCode;				// bytes per sector (00=128, 1=256, 2=512)
	uchar	ucSectorsPerTrack;
	uchar	ucGapLength;
	uchar	ucDataLength;
	uchar	ucFormatFillByte;		// usually F6h
	uchar	ucHeadSettleTime;		// in milliseconds
	uchar	ucMotorStartTime;		// in 1/8th seconds
};

/*
	Returned by Int 13h, fn 08h.
*/
typedef struct BiosDriveParameters BDP;
typedef BDP * PBDP;
typedef BDP far * LPBDP;
struct BiosDriveParameters
{
	int	iTracks;					// highest track # (0 based)
	uchar	ucSectorsPerTrack;	// highest sector # (1 based)
	uchar	ucHeads;					// highest head # (0 based)
	uchar	ucNumDrives;			// returned in DL
	LPDBT	lpDbt;					// ptr to Disk Base Table (floppies only)
	uchar	ucDriveType;			// drive type (floppies only)
};


//	dli13f08.asm
//
bool PASCAL DlBiosGetDriveParameters(int iDrive, PBDP pbdpParameters);


// Dennis: what's this? a sector size? something else?
#define LBA_BLOCK_SIZE	512

#define MAX_CHS_SECTORS	0x0EC4000L	// 15,482,880 sectors MAX with CHS

/*
	EBIOS flags returned by int 13h fn 41h in CX. These are stored in the
	DDT_EBIOSAttributes byte of the DDT for the specified drive. Each BIT is
	set to one if the capability is supplied by the BIOS.
*/
#define EBIOS_HAS_LBA_DISK_SUPPORT		0x01
#define EBIOS_HAS_LOCK_EJECT_SUPPORT	0x02
#define EBIOS_HAS_ENHANCED_SUPPORT		0x04


// dli13f41.asm
//
bool PASCAL DlLbaCheckIfPresent(int iDrive, uchar *pucVersion, unsigned *puEFlags);


/*
	This is parameter packet passed to the EBIOS disk functions with
	int 13 fns 42h, 430xh, 44h and 47h. It uses large Linear Block Address
	sector numbers to overcome the int 13h interface limits for CHS geometries.
*/
typedef struct DeviceAddressPacket DAP;
typedef DAP * PDAP;
typedef DAP far * LPDAP;
struct DeviceAddressPacket
{
	uchar ucPacketSize;
	uchar ucReserved1;
	uchar ucBlockCount;			// limited to MAX_TRANSFER_SECTORS
	uchar ucReserved2;
	void far * lpBuffer;			// Seg:Offs of transfer buffer (-1 = use flat address)
	ulong ulStartingLBALow;		// Starting Linear Block Address (low)
	ulong ulStartingLBAHigh;	// high 32-bits of starting LBA

   // TD 11/14/2001 (IR 3854)
   // These are defined in a Bios specification, but only having to do
   // with flat model and only valid if lpBuffer = FFFF:FFFF.
   // They also bumped the PacketSize to 0x18, which caused problems
   //
   //	ulong qpBufferLow;			// Quad ptr to buffer (low 32-bits)
   //	ulong qpBufferHigh;			// high 32-bits of 64-bit linear addr
};

// dli13f42.asm
//
bool PASCAL DlLbaReadSectors(int iDrive, PDAP pdapAddressPacket);

// dli13f43.asm
//
bool PASCAL DlLbaWriteSectors(int iDrive, PDAP pdapAddressPacket);

// dli13f44.asm
//
bool PASCAL DlLbaVerifySectors(int iDrive, PDAP pdapAddressPacket);

/*
	This structure is returned by EBIOS int 13h fn 48h, Get Drive Parameters.
	The call will only return meaningful information if int 13h fn 41h, Check
	Extensions Present, returns successfully. The information in this structure
	is defined in the Phoenix BIOS Enhanced Disk Drive Specification v3.0.

	NOTES:
	(1) For logical sector information use int 13h fn 8h which reflects
		CHS value translation if required.
	(2) If this value is > MAX_CHS_SECTORS, B1 of EBDPT_uInfoFlags is
		clear and the int 13h interface for CHS geometry is insufficient
		to describe the disk.
	(3) Not currently supported.
*/
typedef struct EBiosDeviceParameterTable EBDPT;
typedef EBDPT * PEBDPT;
typedef EBDPT far * LPEBDPT;
struct EBiosDeviceParameterTable
{
	unsigned	uBufferSize;
	unsigned	uInfoFlags;					// see EBDPT_INFO flags
	ulong		ulPhysicalCylinders;		// 0 based
	ulong		ulPhysicalHeads;			// 0 based
	ulong		ulPhysicalSPT;				// 1 based
	ulong		ulBigPhysicalSectors;	// above 3 parms multiplied together
	ulong		ulBigPhysicalSectorsHi;
	unsigned	uBytesPerSector;			// usually 512
};


/*
	Bit values for flags in the EBDPT_uInfoFlags field above. B7-B15 are
	reserved and 0.
*/
#define EBDPT_INFO_HANDLES_DMA_ERRORS		0x01
#define EBDPT_INFO_CHS_GEOMETRY_VALID		0x02
#define EBDPT_INFO_CAN_REMOVE_MEDIA			0x04
#define EBDPT_INFO_SUPPORTS_VERIFY			0x08
#define EBDPT_INFO_REPORTS_MEDIA_CHANGE	0x10
#define EBDPT_INFO_CAN_LOCK_MEDIA			0x20
#define EBDPT_INFO_NO_MEDIA_PRESENT			0x40


// dli13f48.asm
//
bool PASCAL DlLbaGetDriveParameters(int iDrive, PEBDPT pebdptParameters);

/*
	This Datalight Library structure is used like a "handle" to access
	a bios disk through the DlSmartLbaXxx() routines. This structure
	is returned by DlSmartLbaGetDriveParameters(). The Smart routines
	then take care of whether the BIOS can handle LBA mode or not.
*/
typedef struct LbaSmartInfoPacket LSIP;
typedef LSIP * PLSIP;
typedef LSIP far * LPLSIP;
struct LbaSmartInfoPacket
{
	int	iDrive;			// remember drive number
	unsigned	uEFlags;		// flags returned by Int 13h fn 41h
	EBDPT	ebdpt;			// remember drive geometry
};


//  This function is used to set a flag which will limit the accessable
//  disk to only what can be accessed with CHS calls.
//
bool PASCAL DlSmartSetCHSLimitFlag( bool fEnableCHSLimit );


// dllbaf48.asm
//
bool PASCAL DlSmartLbaGetDriveParameters(int iDrive, PLSIP plsipInfo);

/*
	Used by the DlSmartLbaXxx() transfer functions. A convenient way
	to group fields.
*/
typedef struct LbaSmartTransferPacket LSTP;
typedef LSTP * PLSTP;
typedef LSTP far * LPLSTP;
struct LbaSmartTransferPacket
{
	ulong ulStartSector;
	int   iNumSectors;
	void far * lpBuffer;
};


//  dls13f4x.c
//
bool PASCAL LbaToCHS(PLSIP plsipInfo, ulong ulSector, int *piTrack,
					uchar *pucHead, uchar *pucSector);

// dllbaf42.asm
//
bool PASCAL DlSmartLbaReadSectors(PLSIP plsipInfo, PLSTP plstpTransfer);

// dllbaf43.asm
//
bool PASCAL DlSmartLbaWriteSectors(PLSIP plsipInfo, PLSTP plstpTransfer);

// dllbaf44.asm
//
bool PASCAL DlSmartLbaVerifySectors(PLSIP plsipInfo, PLSTP plstpTransfer);

// dlgeterr.asm
//
int PASCAL DlGetBiosError( void );

// dlseterr.asm
//
int PASCAL DlSetBiosError( int );

// handy functions
//
#define DlSmartLbaGetCylinders(plsipInfo) ((plsipInfo)->ebdpt.ulPhysicalCylinders)
#define DlSmartLbaGetHeads(plsipInfo) ((plsipInfo)->ebdpt.ulPhysicalHeads)
#define DlSmartLbaGetSPT(plsipInfo) ((plsipInfo)->ebdpt.ulPhysicalSPT)
