;  $Id: DLMACROS.INC 1.5 2003/03/20 17:11:58Z thomd Exp $
;
;  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;  Copyright (c) 1996-2003 Datalight, Inc.
;  All Rights Reserved
;
;  Datalight, Inc. is a Washington corporation located at:
;
;  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
;  Bothell, WA  98021                    Fax: +1.425.951.8094
;  USA                                   Http://www.datalight.com
;
;  This software, including without limitation all source code and
;  documentation, is the confidential, trade secret property of
;  Datalight, Inc., and is protected under the copyright laws of
;  the United States and other jurisdictions.  Portions of the
;  software may also be subject to one or more the following US
;  patents: US#5860082, US#6260156.
;
;  In addition to civil penalties for infringement of copyright
;  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;  penalties for violation of (a) the restrictions on circumvention
;  of copyright protection systems found in 17 U.S.C. 1201 and
;  (b) the protections for the integrity of copyright management
;  information found in 17 U.S.C. 1202.
;
;  U.S. Government Restricted Rights.  Use, duplication,
;  reproduction, or transfer of this commercial product and
;  accompanying documentation is restricted in accordance with
;  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
;  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
;  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;  THIS SOFTWARE IN ERROR.
;
.xlist
;------------------------------------------------------------------------
;       $DOCFILE:DLMACROS.INC
;
;       Copyright (C) Datalight, Inc. 1996-2002
;       All Rights Reserved.
;       Datalight Confidential and Proprietary.
;
;       This include contains Datalight standard defines and macros.
;
;       This file is usually included automatically by DATALGHT.INC.
;
;       $DOCHISTORY
;       08/26/97 gp     Created from DATALGHT.INC
;       10/13/97 gp     Added the strcpy macro
;       12/16/97 gp     Added the ASSERT macro
;       03/30/98 gp     Added the jcxnz pseudo instruction
;       04/15/98 gp     Added the TERMINATE macro
;       04/15/98 gp     Added the AND and NAND operators to ASSERT.  Also
;                       supports the "flags" operand.  Modified ASSERT so
;                       that no flags are modified.
;------------------------------------------------------------------------


;------------------------------------------------------------------------
;       $DOCMACRO:ASSERT_FAILED
;
;       This macro is called by the ASSERT macro when it fails.  It is
;       provided in template form only.
;
;       If the ASSERT macro is to be used, this macro must be defined
;       to do something that is appropriate for the specific module
;       or project in question.
;
;       Note that regardless what functionality may be performed by this
;       macro, if it is possible to return from this macro back to the
;       ASSERT macro, all registers and critical flags (DF and IF) must
;       be preserved.
;------------------------------------------------------------------------
;ASSERT_FAILED    MACRO   linenum,ovr
;.errnb <ovr>
;.errb  <linenum>
;
;       ; fill this in with code that is appropriate for your project
;
;                ENDM



;------------------------------------------------------------------------
;       $DOCMACRO:TERMINATE     ; LineNumber,[PRODUCTION]
;
;       This macro calls the ASSERT_FAILED macro with the specified
;       line number.
;
;       If the optional PRODUCTION flag is used, this macro will generate
;       code even if DEBUG=0.
;
;       We'd rather call this macro "ABORT" but that word is already
;       defined in WINDOWS.INC.
;------------------------------------------------------------------------
TERMINATE       MACRO lineNum,prod,ovr
.errnb <ovr>
.errb  <lineNum>

  ifb <prod>
    ife DEBUG
      exitm                     ; if RELEASE, generate no code
    endif
  else
    ifdifi <prod>,<PRODUCTION>  ; generate only if DEBUG=TRUE?
      .err "TERMINATE: Unrecognized parameter: &prod"
      exitm
    endif
  endif

        ASSERT_FAILED lineNum

ENDM


;------------------------------------------------------------------------
;       $DOCMACRO:ASSERT        ; <expression>,LineNumber,[PRODUCTION]
;
;       Assert that the specified expression is TRUE.  Valid operators are
;       listed below.  Registers, memory variables, and immediate values are
;       allowed, though they must be the same type - bytes, words, or dwords.
;       Type overrides (byte ptr, word ptr, dword ptr) are allowed.  If two
;       memory variables are specified, the first operator will be moved into
;       a register for the comparison.
;
;       All comparisons are considered to be unsigned, all registers are
;       preserved, and the flags are unchanged.
;
;       Values in segment registers cannot be tested at this time.
;
;       Valid operators are:    EQ - equal
;                               NE - not equal
;                               GT - greater than
;                               GE - greater than or equal
;                               LT - less than
;                               LE - less than or equal
;                              AND - bitwise AND
;                             NAND - bitwise NOT-AND
;
;       The operator must be specified in UPPER CASE, and separated from
;       each operand by at least one white space.
;
;       Examples:
;               ASSERT <ax NE 0>,123,PRODUCTION        ; assert AX != 0
;               ASSERT <bptr wordVar GT 1>,456         ; assert lobyte(wVar) > 1
;               ASSERT <flags NAND CPUFLAG_CARRY>,1    ; verify that CF is clear
;               ASSERT <flags AND CPUFLAG_INTERRUPT>,1 ; verify that IF is set
;
;       If the ASSERT fails, the ASSERT_FAILED macro will be executed with
;       the specified line number.
;
;       If the optional PRODUCTION flag is used, this macro will generate
;       code even if DEBUG=0.
;------------------------------------------------------------------------
ASSERT          MACRO expression,lineNum,prod,ovr
local           asrtOp1,asrtOp2,asrtOperator,assertPassed,workReg
.errnb <ovr>
.errb  <lineNum>

  ifb <prod>
    ife DEBUG
      exitm                     ; if RELEASE, generate no code
    endif
  else
    ifdifi <prod>,<PRODUCTION>  ; generate only if DEBUG=TRUE?
      .err "ASSERT: Unrecognized parameter: &prod"
      exitm
    endif
  endif

        pushf                   ; preserve the flags

  asrtOp1       equ <>  ; the first operand
  asrtOp2       equ <>  ; the second operand
  asrtOperator  equ <>  ; the comparison operator

  ; break the expression into its three components
  ?ParseExpression <expression>,asrtOp1,asrtOp2,asrtOperator
  ifb &asrtOperator
    .err "ASSERT: Invalid expression: &expression"
    exitm
  endif

  ifidni &asrtOperator,<AND>            ;; bitwise AND
    ifidni &asrtOp1,<flags>
      workReg equ <ax>
      ifidni &asrtOp2,<ax>
        workReg equ <bx>
      endif
        push    workReg
        pushf
        pop     workReg
        test    workReg,asrtOp2
        pop     workReg
        jnz     assertPassed
    else
        test    asrtOp1,asrtOp2
        jnz     assertPassed
    endif

  elseifidni &asrtOperator,<NAND>       ;; bitwise NOT-AND
    ifidni &asrtOp1,<flags>
      workReg equ <ax>
      ifidni &asrtOp2,<ax>
        workReg equ <bx>
      endif
        push    workReg
        pushf
        pop     workReg
        test    workReg,asrtOp2
        pop     workReg
        jz      assertPassed
    else
        test    asrtOp1,asrtOp2
        jz      assertPassed
    endif

  else                  ;; wasn't AND or NAND, rest handled as a group

    ;; Use our handy-dandy extended compare macro to allow BYTE,
    ;; WORD, or DWORD comparisons on any x86 processor.

        cmpx    <asrtOp1>,<asrtOp2>

    ; we could save a several bytes here by checking if op2 is 0 and op1 is
    ; a register, if so, we could use an OR instruction instead of CMP.

    ; validate the results according to the specified operator
    ifidni &asrtOperator,<EQ>
        je short assertPassed
    elseifidni &asrtOperator,<NE>
        jne short assertPassed
    elseifidni &asrtOperator,<GT>
        ja short assertPassed
    elseifidni &asrtOperator,<GE>
        jae short assertPassed
    elseifidni &asrtOperator,<LT>
        jb short assertPassed
    elseifidni &asrtOperator,<LE>
        jbe short assertPassed
    else
        .err "ASSERT: Unrecognized operator: &asrtOperator"
        exitm
    endif
  endif

        ASSERT_FAILED lineNum

  assertPassed:
        popf                    ; restore original flags

ENDM


;-------------------------------------------------------------------------
;       $DOCMACRO:pushall
;
;       Push all 16 bit registers and flags (except for CS and IP).  If
;       the target processor is a 386 or higher, FS and GS will be pushed
;       as well.
;
;       If the target processor is a 286 or higher PUSHA will be used
;       rather than pushing individual registers.  Do not rely on the
;       pushed SP value regardless what processor is used.
;
;       Use popall to restore all these registers.
;-------------------------------------------------------------------------
pushall         MACRO   ovr
.errnb <ovr>    ;; no parameters allowed

  if (@Cpu and AMODE_80286)
        pusha
  else
        pusher  ax,cx,dx,bx,sp,bp,si,di
  endif

        pushf
        push    ds
        push    es

  if (@Cpu and AMODE_80386)
        push    fs
        push    gs
  endif

ENDM


;-------------------------------------------------------------------------
;       $DOCMACRO:popall
;
;       Pop all 16 bit registers and flags that were pushed with pushall.
;       The SP value is thrown away.
;-------------------------------------------------------------------------
popall          MACRO   ovr
.errnb <ovr>    ;; no parameters allowed

  if (@Cpu and AMODE_80386)
        pop     gs
        pop     fs
  endif

        pop     es
        pop     ds
        popf

  if (@Cpu and AMODE_80286)
        popa
  else
        ;; this is NOT a type, BX is listed twice because the first BX
        ;; popped is really the original SP value that we don't care about
        popper  ax,cx,dx,bx,bx,bp,si,di
  endif

ENDM


;-------------------------------------------------------------------------
;       $DOCMACRO:pushalld
;
;       Push all 32 bit registers and flags, including all the segment
;       registers.  This macro is only legal in .386 code.
;-------------------------------------------------------------------------
pushalld        MACRO   ovr
.errnb <ovr>    ;; no parameters allowed
        pushad
        pushfd
        push    ds
        push    es
        push    fs
        push    gs
ENDM


;-------------------------------------------------------------------------
;       $DOCMACRO:popalld
;
;       Pop all 32 bit registers and flags that were pushed with pushalld.
;       This macro is only legal in .386 code.
;-------------------------------------------------------------------------
popalld         MACRO   ovr
.errnb <ovr>    ;; no parameters allowed
        pop     gs
        pop     fs
        pop     es
        pop     ds
        popfd
        popad
ENDM


;------------------------------------------------------------------------------
;       $DOCMACRO:pushx
;
;       This is the core macro for PUSH operations.  Specify "flags" or
;       "eflags" to push the 16 or 32 bit flags.  The pseudo values MAX,
;       MBX, MCX, MDX, MSI, MDI, MBP, MSP, and MIP are supported.  These
;       values cause either a 16 or 32 bit value to be pushed depending
;       on whether the assembly mode .386 or higher.
;
;       This macro has no effect for normally specified registers or
;       locations, whether 16 or 32 bit, so the assembler will generate
;       standard error messages for improperly specified values.
;
;       $DOCTODO:  Update this macro to handle immediate values intelligently
;                  if running on an 8086.
;------------------------------------------------------------------------------
pushx           MACRO pval,ovr
local           regVal,pvalLen,mpos
.errb  <pval>   ;; enough parameters
.errnb <ovr>    ;; but not too many

  ifidni <pval>,<flags>
        pushf
        exitm
  endif

  ifidni <pval>,<eflags>
        pushfd
        exitm
  endif

  pvalLen SIZESTR <pval>        ; calc parameter len
  mpos INSTR <pval>,<m>

  ; if first character is an 'm' and the string length is 3
  if (mpos EQ 1) and (pvalLen EQ 3)

    regVal SUBSTR <pval>,2,2    ; calc real register name

    if (@Cpu and AMODE_80386)
        regVal CATSTR <e>,regVal
        push    regVal          ; prepend an 'e' to the register name
    else
        push    regVal          ; just push the register if not .386
    endif

  else                          ; else, not one of our pseudo regs
        push    pval            ; just push the value
  endif

ENDM


;------------------------------------------------------------------------------
;       $DOCMACRO:popx
;
;       This is the core macro for POP operations.  Specify "flags" or
;       "eflags" to pop the 16 or 32 bit flags.  The pseudo values MAX,
;       MBX, MCX, MDX, MSI, MDI, MBP, MSP, and MIP are supported.  These
;       values cause either a 16 or 32 bit value to be popped depending
;       on whether the assembly mode .386 or higher.
;
;       This macro has no effect for normally specified registers or
;       locations, whether 16 or 32 bit, so the assembler will generate
;       standard error messages for improperly specified values.
;------------------------------------------------------------------------------
popx            MACRO pval,ovr
local           regVal,pvalLen,mpos
.errb  <pval>
.errnb <ovr>

  ifidni <pval>,<flags>
        popf
        exitm
  endif

  ifidni <pval>,<eflags>
        popfd
        exitm
  endif

  pvalLen SIZESTR <pval>        ; calc parameter len
  mpos INSTR <pval>,<m>

  ; if first character is an 'm' and the string length is 3
  if (mpos EQ 1) and (pvalLen EQ 3)

    regVal SUBSTR <pval>,2,2    ; calc real register name

    if (@Cpu and AMODE_80386)
        regVal CATSTR <e>,regVal
        pop     regVal          ; prepend an 'e' to the register name
    else
        pop     regVal          ; just pop the register if not .386
    endif

  else                          ; else, not one of our pseudo regs
        pop     pval            ; just pop the value
  endif

ENDM


;------------------------------------------------------------------------------
;       $DOCMACRO:popper
;
;       Pop the specified registers in reverse order.  Specify "flags" or
;       "eflags" to pop the 16 or 32 bit flags.
;
;       This macro supports the pseudo values MAX, MBX, MCX, MDX, MSI, MDI,
;       MBP, MSP, and MIP.  These values cause either a 16 or 32 bit value
;       to be popped depending on the assembly mode .8086 or .386.
;------------------------------------------------------------------------------
popper          MACRO w1,w2,w3,w4,w5,w6,w7,w8,w9,w10,w11,w12,w13,w14,w15,w16,ovr
.errb  <w1>
.errnb <ovr>

  irp var,<w16,w15,w14,w13,w12,w11,w10,w9,w8,w7,w6,w5,w4,w3,w2,w1>
  ifnb <&var>
        popx    <&var>
  endif
  endm
ENDM


;-----------------------------------------------------------------------------
;       $DOCMACRO:pusher
;
;       Push the specified registers in the specified order.  Use popper to
;       pop the registers in reverse order.  Specify "flags" or "eflags" to
;       pop the 16 or 32 bit flags.
;
;       This macro supports the pseudo values MAX, MBX, MCX, MDX, MSI, MDI,
;       MBP, MSP, and MIP.  These values cause either a 16 or 32 bit value
;       to be pushed depending on the assembly mode .8086 or .386.
;-----------------------------------------------------------------------------
pusher          MACRO w1,w2,w3,w4,w5,w6,w7,w8,w9,w10,w11,w12,w13,w14,w15,w16,ovr
.errb  <w1>
.errnb <ovr>

  irp var,<w1,w2,w3,w4,w5,w6,w7,w8,w9,w10,w11,w12,w13,w14,w15,w16>
  ifnb <&var>
        pushx   <&var>                  ; general register push
  endif
  endm
ENDM


;-----------------------------------------------------------------------------
;       $DOCMACRO:datamov
;
;       Syntax:  datamov len, [SHORT | EVEN | WORDS]
;
;       This macro encodes an optimal "rep movsb" coding sequence for the
;       length given.  Source and destination operands must be already
;       initialized in DS:SI and ES:DI.
;
;       The SHORT, EVEN, and WORDS options are not valid when "len" is
;       a constant.
;
;       SHORT - Use this option when the "len" value is known to be relatively
;               short.  This causes a single "rep movsb" to be generated with
;               no further ado.
;
;       EVEN  - Use this option when the "len" value is known to be even.
;               The allows more optimal code to be generated.
;
;       WORDS - Use this option when the "len" value contains a count of words
;               to move rather than bytes.  This causes a single "rep movsw"
;               to be generated.
;
;       Upon exit, SI and DI will point to their respective positions
;       following the data moved.  Note that CX <may> be modified -- it is
;       NOT guaranteed to be zero upon exit!
;
;       NOTE:  This macro cannot be used when the direction flag is set
;       with the STD instruction.  DF should always be cleared prior to
;       using this macro.
;-----------------------------------------------------------------------------
datamov         MACRO len,p1,ovr
.errb <len>
.errnb <ovr>

  if (.TYPE len) and OPTYPE_CONST
    ifnb <p1>
      .err "datamov: Invalid parameter when a constant is specified: &p1"
      exitm
    endif

        ;; for small constant values, just use explicit MOVS instructions

    if     len EQ 1
        movsb
    elseif len EQ 2
        movsw
    elseif len EQ 3
        movsw
        movsb
    elseif len EQ 4
        movsw
        movsw
    elseif len EQ 5
        movsw
        movsw
        movsb
    elseif len EQ 6
        movsw
        movsw
        movsw
    elseif len EQ 7
        movsw
        movsw
        movsw
        movsb
    elseif len EQ 8
        movsw
        movsw
        movsw
        movsw
    elseif len EQ 9
        movsw
        movsw
        movsw
        movsw
        movsb
    else

        ;; not so small, so use a REP MOVS instruction

        mov     cx,(len)/2
        rep movsw
      if (((len)/2)*2) NE len
        movsb                           ;; if len is odd
      endif
    endif
  else
    ifdifi <len>,<cx>
        mov     cx,len                  ;; put len in CX if not alread there
    endif

    ifidni <p1>,<EVEN>
        shr     cx,1                    ;; we know it's even so split and halve
        rep movsw                       ;; words instead
    elseifidni <p1>,<WORDS>
        rep movsw                       ;; CX refers to words, so just do it
    elseifidni <p1>,<SHORT>
        rep movsb                       ;; CX is real short, so just do it
    else
      ifnb <p1>
       .err "datamov: Bad parameter: &p1"
       exitm
      endif

        shr     cx,1                    ;; if nothing else cut it, just do the
        rep movsw                       ;; typical move of words first, then
        rcl     cx,1                    ;; any remaining byte
        rep movsb
    endif
  endif
ENDM


;-----------------------------------------------------------------------------
;       $DOCMACRO:jcxzl
;
;       Syntax:         jcxzl   label
;
;       This is a pseudo instruction that allows the JCXZ instruction to be
;       used with labels that are normally out of range.
;-----------------------------------------------------------------------------
jcxzl           MACRO   lbl,ovr
.errb   <lbl>
.errnb  <ovr>
        or      cx,cx
        jz      lbl
ENDM


;-----------------------------------------------------------------------------
;       $DOCMACRO:jcxnz
;
;       Syntax:         jcxnz   label
;
;       This is a pseudo instruction that tests CX and jumps if it is
;       non-zero.  Note that the flags ARE modified by this macro.
;-----------------------------------------------------------------------------
jcxnz           MACRO   lbl,ovr
.errb   <lbl>
.errnb  <ovr>
        or      cx,cx
        jnz     lbl
ENDM


;----------------------------------------------------------------------------
;       DOCMACRO:cmpx
;
;       This is a pseudo instruction that provides an extended compare
;       ability.  It generates optimal code based on the CPU model being
;       targeted.  It allows memory-to-memory comparisons, and DWORD
;       comparisons on the older intel processors.  (DWORDs are supported
;       on pre-386 processors for memory-to-memory comparisons only since
;       32 bit registers and constants are not available.)
;
;       There are six legal combinations of the three basic types of
;       operands.  These are:
;
;               cmpx    reg,reg
;               cmpx    reg,mem
;               cmpx    reg,const
;               cmpx    mem,reg
;               cmpx    mem,mem
;               cmpx    mem,const
;
;       Operands must be the same size, though size overrides (bptr, wptr,
;       dptr, etc) are allowed.  Note that register-indirect operands are
;       considered the same as memory operands.
;
;       When memory-to-memory comparisons are performed (and DWORD compares
;       in pre-386 code), one or more operands will be moved into registers
;       for the comparison.
;
;       No registers are changed, and the flags reflect the state of the
;       compare exactly as if the CMP instruction were used.
;----------------------------------------------------------------------------
cmpx            MACRO   p1,p2,ovr
local           cmpx_done
.errb  <p2>
.errnb <ovr>

; This validation would be nice, but it does not work if
; the first operand is something line "wptr DS:[QUE_SIG]".
; Is there another way to do this?
;
; if (.TYPE p1) and OPTYPE_CONST
;   %out "cmpx: First operand cannot be a constant"
;   .err
;   exitm
; endif

  ;; First let's whip out the easy cases

  if (.TYPE p1) and OPTYPE_REG
        cmp     p1,p2
        exitm
  elseif (.TYPE p2) and OPTYPE_REG
        cmp     p1,p2
        exitm
  elseif (.TYPE p2) and OPTYPE_CONST
        cmp     p1,p2
        exitm
  endif

  ;; If we get to this point, the only variation we have
  ;; left is a memory-to-memory comparison.

  if (TYPE p1) EQ 1             ;; is it a byte comparison?
        push    ax
        mov     al,p2
        cmp     p1,al
        pop     ax
  elseif (TYPE p1) EQ 2         ;; is it a word comparison?
        push    ax
        mov     ax,p2
        cmp     p1,ax
        pop     ax
  elseif (TYPE p1) EQ 4         ;; is it a dword comparison?
    if (@Cpu and AMODE_80386)
        push    eax             ;; we've got a 386 or later so do it the easy way
        mov     eax,p2
        cmp     p1,eax
        pop     eax
    else
        pusher  ax,bx
        sdw     ax,bx,p2        ;; get operand 2 into AX,BX
        cmp     ax,wptr p1+2    ;; compare the high word first
        jne     cmpx_done       ;; we're finished now if it doesn't match
        cmp     bx,wptr p1      ;; only if high word is the same do we do the low word
    cmpx_done:
        popper  ax,bx
    endif
  else
      %out "cmpx: Operands must be BYTEs, WORDs, or DWORDs"
      .err
      exitm
  endif

ENDM


;-----------------------------------------------------------------------------
;       $DOCMACRO:address
;
;       This macro allows register based address manipulation using optimal
;       code sequences.
;
;       Syntax:  address <DS:SI EQ ES:DI>        ; move ES:DI into DS:SI
;                address <DS:SI XCHG ES:DI>      ; exchange DS:SI and ES:DI
;-----------------------------------------------------------------------------
address         MACRO expr,ovr
local   colonOffset,addrOp1,addrOp2,addrOperator,destSeg,destOffs,srcSeg,srcOffs,xReg
.errb   <expr>
.errnb  <ovr>

  addrOp1       equ <>  ; the first operand
  addrOp2       equ <>  ; the second operand
  addrOperator  equ <>  ; the comparison operator

  ; break the expression into its three components
  ?ParseExpression <expr>,addrOp1,addrOp2,addrOperator
  ifb &addrOperator
    .err "address: Invalid expression: &expr"
    exitm
  endif

  ;; must be a colon in the third position
  colonOffset INSTR addrOp1,<:>
  if colonOffset NE 3
    .err "address: Invalid operand 1: &addrOp1"
    exitm
  endif

  ;; determine destination segment and offset registers
  destSeg SUBSTR addrOp1,1,2
  destOffs SUBSTR addrOp1,4

  ;; must be a colon in the third position
  colonOffset INSTR addrOp2,<:>
  if colonOffset NE 3
    .err "address: Invalid operand 2: &addrOp2"
    exitm
  endif

  ;; determine source segment and offset registers
  srcSeg SUBSTR addrOp2,1,2
  srcOffs SUBSTR addrOp2,4

  ifidni addrOperator,<EQ>
    if (TYPE destOffs) EQ 4       ;; is it a 32 bit register?
        xReg SUBSTR destOffs,2,2  ;; determine the name of low 16 bits of
        mov     xReg,srcSeg       ;; the reg since we can't move the segreg
        mov     destSeg,xReg      ;; into a DWORD register
        mov     destOffs,srcOffs
    else
        mov     destOffs,srcSeg   ;; simple case, both 16 bit regs
        mov     destSeg,destOffs
        mov     destOffs,srcOffs
    endif
  elseifidni addrOperator,<XCHG>
        push    srcSeg
        pop     destSeg
        xchg    srcOffs,destOffs
  else
    .err "address: Unsupported operator: &addrOperator"
    exitm
  endif

ENDM


;-----------------------------------------------------------------------------
;       $DOCMACRO:ldw
;
;       Syntax:         ldw     reg16hi,reg16lo,dwVar
;
;       This is the pseudo instruction "Load from DWord".  It loads the two
;       16 bit registers from a 32 bit location.  The source location must
;       be a memory variable.
;
;       For example:
;               .data
;               dwordVar                dd  0
;
;               .code
;                       ldw     dx,ax,dwordvar
;-----------------------------------------------------------------------------
ldw             MACRO   hi,lo,loc,ovr
.errb   <loc>
.errnb  <ovr>

        mov     hi,wptr loc+2
        mov     lo,wptr loc
ENDM


;-----------------------------------------------------------------------------
;       $DOCMACRO:sdw
;
;       Syntax:         sdw     dwVar,reg16hi,reg16lo
;
;       This is the pseudo instruction "Store to DWord".  It stores the
;       two 16 bit register values in a 32 bit location.  The destination
;       location must be a memory variable.
;
;       For example:
;               .data
;               dwordVar                dd  0
;
;               .code
;                       sdw     dwordvar,dx,ax
;-----------------------------------------------------------------------------
sdw             MACRO   loc,hi,lo,ovr
.errb   <lo>
.errnb  <ovr>

        mov     wptr loc+2,hi
        mov     wptr loc,lo
ENDM


;------------------------------------------------------------------------
;       $DOCMACRO:lso
;
;       Syntax: lso     segreg,reg,data
;
;       This macro loads "segreg:reg" with the address of "data" when
;       data is a variable residing on the stack.
;
;       Example:
;               ProcArgs        lParam:DWORD
;               ProcLocals      localBuff[32]:BYTE
;
;               lso     es,di,localBuff ; get address of localBuff into ES:DI
;               lso     ax,ebx,lParam   ; get address of lParam into AX:EBX
;
;       Note that "segreg" does not have to literally be a segment
;       register, though typically this is the case.
;------------------------------------------------------------------------
lso             MACRO   sreg,oreg,dat,ovr
.errb   <dat>
.errnb  <ovr>

  ifidni <sreg>,<es>
        mov     oreg,ss         ;; special handling if dest is a segreg
        mov     es,oreg
  elseifidni <sreg>,<ds>
        mov     oreg,ss
        mov     ds,oreg
  elseifidni <sreg>,<fs>
        mov     oreg,ss
        mov     fs,oreg
  elseifidni <sreg>,<gs>
        mov     oreg,ss
        mov     gs,oreg
  else
        mov     sreg,ss         ;; not a segreg, so move directly
  endif

        lea     oreg,dat        ;; offset into the specified register

ENDM


;------------------------------------------------------------------------
;       $DOCMACRO:?ParseExpression      ; <expression>,op1,op2,operator
;
;       This helper macro separates an expression into its component
;       operands and operator.  The two operands are separated and
;       recognized solely by the valid operator.  Valid operators are:
;
;               EQ      ; equal
;               NE      ; not equal
;               GT      ; greater than
;               GE      ; greater than or equal
;               LT      ; less than
;               LE      ; less than or equal
;               XCHG    ; exchange
;               AND     ; bitwise AND
;               NAND    ; bitwise NOT-AND
;
;       The operator must be all UPPER CASE and must be separated from
;       each operand by at least one white space.
;
;       The operands must simply be non-null to be considered valid.  On
;       exit, operator will equate to <> if an error occurs.
;------------------------------------------------------------------------
?ParseExpression MACRO expression,op1,op2,oper,ovr
local           operOffset, operLen
.errnb <ovr>
.errb  <oper>

operLen = 0

  ; determine the operator
  operOffset INSTR <expression>,< EQ >
  if operOffset
    operLen = 2
  else
    operOffset INSTR <expression>,< NE >
    if operOffset
      operLen = 2
    else
      operOffset INSTR <expression>,< GT >
      if operOffset
        operLen = 2
      else
        operOffset INSTR <expression>,< GE >
        if operOffset
          operLen = 2
        else
          operOffset INSTR <expression>,< LT >
          if operOffset
            operLen = 2
          else
            operOffset INSTR <expression>,< LE >
            if operOffset
              operLen = 2
            else
              operOffset INSTR <expression>,< XCHG >
              if operOffset
                operLen = 4
              else
                operOffset INSTR <expression>,< AND >
                if operOffset
                  operLen = 3
                else
                  operOffset INSTR <expression>,< NAND >
                  if operOffset
                    operLen = 4
                  else
                    .err "?ParseExpression:  Invalid Expression:  &expression"
                    exitm
                  endif
                endif
              endif
            endif
          endif
        endif
      endif
    endif
  endif

  ; break expression into its components
  op1  SUBSTR <expression>,1,operOffset-1
  op2  SUBSTR <expression>,operOffset+operLen+2
  oper SUBSTR <expression>,operOffset+1,operLen

ENDM


;------------------------------------------------------------------------
;       $DOCMACRO:?RegLength    ; reg,returnVal
;
;       This is a helper macro that returns the width of the specified
;       register - either 1, 2, or 4 bytes, or 0 if a bad register
;       name was specified.
;------------------------------------------------------------------------
?RegLength      MACRO reg,retval,ovr
local           plen,returnVal,chr1,chr2
.errnb <ovr>
.errb  <oper>

  returnVal     = 0

; ifidni ASM,<TASM>       ;; if using TASM
    plen SIZESTR <reg>

    if plen EQ 2
      chr1 SUBSTR <reg>,2,1
      ifidni chr1,<H>
        returnVal = 1
      elseifidni chr1,<L>
        returnVal = 1
      else
        returnVal = 2
      endif
    elseif plen EQ 3
      chr1 SUBSTR <reg>,1,1
      chr2 SUBSTR <reg>,3,1
      ifidni chr1,<E>
        ifidni chr2,<X>         ;; EAX, EBX, ECX, EDX
          returnVal = 4
        elseifidni chr2,<I>     ;; ESI, EDI
          returnVal = 4
        elseifidni chr2,<P>     ;; EBP, ESP, EIP
          returnVal = 4
        endif
      endif
    endif

; else
;
;   MASM doesn't like this for SS,CS,FS,and GS !?!?!?!
;   returnVal = TYPE reg                ;; sure is easier in MASM!
;
; endif

  retval        equ <returnVal>

ENDM


;------------------------------------------------------------------------
;       $DOCMACRO:?DisplaySymbols       ; sym1,sym2,sym3,etc
;
;       This helper macro displays the specified values using the
;       OUT directive.  It is typically used for macro debugging.
;------------------------------------------------------------------------
?DisplaySymbols  MACRO p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,ovr
.errnb <ovr>

  %out "&p1, &p2, &p3, &p4, &p5, &p6, &p7, &p8, &p9, &p10"

ENDM


;------------------------------------------------------------------------
;       $DOCMACRO:?SetTextSymbol        ; name,value
;
;       Set the specified symbol to the specified value.  This macro
;       is primarily of use when one symbol needs to be set to the value
;       of another.
;
;       For example, to set SYMBOL1 to the value of SYMBOL2, you can
;       use code such as the following:
;
;               SYMBOL1 equ <>
;               SYMBOL2 equ <value>
;               ?SetTextSymbol SYMBOL1,%SYMBOL2
;------------------------------------------------------------------------
?SetTextSymbol  MACRO nam,val,ovr
.errnb <ovr>
.errb  <val>

  nam equ <&val>

ENDM






.list
