//  $Id: DOSTRUCT.H 1.29 2003/09/09 16:42:30Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1989-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSTRUCT.H    created 04/14/89  by Drew Gislason

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This file contains all structures used within ROM-DOS 'C'
   language modules, most of which are DOS 5.0 compatible. The
   following list of texts can describe some of these structures
   in more detail.


   1. DOS Internals
      by Geoff Chappell, Addison-Wesley, 1994

   2. Undocumented DOS, 2nd Edition
      by Andrew Schulman, Addison-Wesley, 1993

   3. MS-DOS Programmer's Reference 6.0
      Microsoft Press, 1993

   4. DOS Programmer's Reference, 3rd Edition
      by Terry R. Dettmann,  Que Corporation 1992

   5. Microsoft Developer Network On-line documentation.


   Conditionals that can affect definitions within this file and
   hence the files in which it is included:

   USING_BIG_DISKS:
         if >0 then EBIOS LBA disk accesses are supported

   USING_FAT32:
         if >0 then FAT32 data structures are defined otherwise
         all data structures are the FAT12/FAT16 subset. We
         assume all FAT32 media have the HARD_MEDIA type.

   LSEEK_ENHANCEMENT:
         if >0 then we turn on an enhancement that uses two
         fields from the SFT to keep track of a pointer to the
         current "contiguous" block of clusters.  When seeking
         in reverse, we can do simple math (when appropriate)
         instead of hitting the FAT table stored on disk.

   USING_SWAPPABLE_HARD_DISKS:

         CURRENTLY UNAVAILABLE!!!

         if >0 then support for hot swappable HARD_MEDIA media
         types (hard disks) is available.

**********************************************************/

#ifndef DOSSTRUCT         /* include this file only once */
#define DOSSTRUCT 1

//
//		Common Types
//
#if !defined(COMMON_TYPES)
#define COMMON_TYPES 1
typedef int bool;
typedef unsigned char uchar;
typedef unsigned long ulong;
typedef unsigned char byte;
typedef unsigned int word;
#endif

//
//		Common Defines
//
#define FALSE	0
#define TRUE	!FALSE

typedef int (*funcptr)();
typedef char far * lpchar;
typedef lpchar * plpchar;

#if USING_FAT32
typedef ulong CLUSTER;
typedef ulong NCLUSTER;
#define USING_BIG_DISKS  1
#else
typedef unsigned CLUSTER;
typedef unsigned NCLUSTER;
#endif

// This define will use a specific "undocumented" portion of the SFT
// (see below) to keep track of contiguous blocks; specifically, the
// address of the beginning of current contiguous block. This allows
// ROM-DOS to do simple math to calculate cluster addresses, instead
// of FAT entry lookups.
//
#define LSEEK_ENHANCEMENT 0

// This define will allow the ROM-DOS kernel to update the LFN field
// for Last Accessed Date within the directory entry.
//
// Enabling this will cause a performance degradation, and our
// research has shown that very few utilities can use this field.
//
// The ROM-DOS kernel is unaware of the write protect status of a
// diskette or drive entry from moment to moment.  If this is
// enabled, the ROM-DOS kernel will attempt to update this LFN
// field on all disks, and will then generate a critical error
// ("Abort, Retry, or Fail") on any that are not writable.
// This flaw will be corrected when we can find a way that does
// not add a further performance hit.
//
#define UPDATE_LFN_LAST_ACCESSED_DATE 0

// DON'T CHANGE THIS!!!!
//
// This is a primarily a mechanism for disabling hard disk code in
// change_of_disk() without removing it form the source. EBIOS
// support for removable media will be required for the option to
// be available but those functions are not yet supported by any
// version of ROM-DOS.
//
#define	USING_SWAPPABLE_HARD_DISKS 	0


/*
	Absolute Read/Write Info
   Description: Int 25 and Int 26H (Absolute Read / Write)
   use this structure if CX == -1

   Reference: DOS Ref, p 753

	Note: This same structure is used by int 21h fn 7305h to read/write
	sectors on FAT32 drives. 4G*sectors * 512bytes/sector = 2TBytes, the
	limit in FAT32 drive size.

	AX		=	7305h
	DS:BX	=	offset of AbsRdWr data struc as defined below.
	CX		=	-1
	DL		=  drive number: 0=default, 1=A, 2=B...
	SI		=  "type info"
				B14	B13  B1
								?	1 = write, 0 = read
				0		0		x	Other/unknown
				0		1		x	FAT  data
				1		0		x  DIR  data
				1		1		x	FILE data
				all other bits are reserved and must be 0

	source: MSDN online documentation

*/
struct AbsoluteRdWr
{
	ulong		ulStartSector;	// starting sector
	unsigned uCount;			// # of sectors to read/write
	void far *lpBuffer;		// buffer pointing to sector
};
typedef struct AbsoluteRdWr ABS_RDWR;
typedef ABS_RDWR far * LPABS_RDWR;


#if USING_FAT32
/*
 File System Information Structure (aka BIGFATBOOTFSINFO (FAT32) by MS).

 Contains information about the file system on a FAT32 volume. This
 structure is implemented in Windows OEM Service Release 2 and later.

 ulSignature
 	The signature of the file system information sector. The value
 	in this field is FSINFOSIG (0x61417272L).
 ulFreeClusters
 	The count of free clusters on the drive. Set to -1 when the
 	count is unknown.
 ulNextFreeCluster
 	The cluster number of the cluster that was most recently
   allocated.
 Reserved.
 	Reserved field. Do not use.
 uSignature2
  	End of "boot sector" marker (0x55AA)

*/
#define	FSINFO_SECTOR_SIGNATURE		0x41615252L		// "RRaA"
#define	FSINFO_STRUC_OFFSET			0x01E4
#define	FSINFO_SIGNATURE_VALUE		0x61417272L		// "rrAa"
#define	BOOT_SECTOR_SIGNATURE		0xAA55

struct FileSystemInfo
{
	ulong		ulSignature;									// 0   (1E4)
	ulong		ulFreeClusters;                        // 4   (1E8)
	ulong		ulNextFreeCluster;                     // 8   (1EC)
	ulong		ulReserved[3]; 								// 0C  (1F0 - 1F8)
	unsigned	uReserved2;										// 18  (1FC)
	unsigned	uSignature2;									// 1A  (1FE)
};
typedef struct FileSystemInfo FSINFO ;
typedef FSINFO * PFSINFO;
typedef FSINFO far * LPFSINFO;

#endif // USING_FAT32

/*
   BIOS PARAMETER BLOCK (BPB)
   Description: This structure is stored on disks in
   sector 0, AT OFFSET 11d.

   Reference: Adv MS-DOS, p 269  -or-  DOS Ref, p 222
				  MSDN online documentation.

	Note:	The FAT32 (DOS v7.1) implementation of this data structue
			has additional fields.

*/
typedef struct BiosParameterBlock BPB;
typedef BPB     *PBPB;
typedef BPB far *LPBPB;
struct BiosParameterBlock
{
   	unsigned    sec_size;          	   // 00 bytes per sector
   	uchar    	sec_per_clus;      	   // 02 sectors per cluster
   	unsigned    res_sec;           	   // 03 reserved sector count (before FAT)
   	uchar    	num_fats;         	   // 05 count of FATs
#if USING_FAT32
   	unsigned    uDirEntries;         	// 06 count of root DIR entries (0 on FAT32)
#else
		unsigned    dir_ent;						// 06 count of root DIR entries (FAT16 name)
#endif
   	unsigned    num_sec;           	   // 08 count of drive sectors (may be 0)
   	uchar    	media;             	   // 0A media ID byte
#if USING_FAT32
   	unsigned    uFatSectors;          	// 0B count of sectors per fat (0 on FAT32)
#else
   	unsigned    fat_size;	          	// 0B count of sectors per fat (FAT16 name)
#endif
														// the next 4 are for device use only
   	unsigned    sec_per_trak;      	   // 0D count of sectors per track
   	unsigned    sides;             	   // 0F count of disk heads (sides)
   	ulong		   hidden;            	   // 11 count of hidden sectors (before boot sector)
   	ulong			lnumsec;           	   // 15 count of sectors (if num_sec==0)

#if	USING_FAT32
   	ulong		   ulSectorsPerFat;    	   // 19 FAT32 sectors per FAT
   	unsigned    uExtendedFlags;        	// 1D FAT32 extended flags (defined below)
   	unsigned    uFileSystemVersion;    	// 1F FAT32 File system version number
   	ulong    	ulRootDirStartCluster;	// 21 FAT32 Root dir start (can move)
   	unsigned    uFSInfoSector;         	// 25 FAT32 sector # of FSINFO struc  (-1 if none)
   	unsigned    uBackUpBootSector;      // 27 FAT32 sector # of backup boot sector (-1 if none)
   	unsigned    Reserved[6];           	// 29 FAT32 ???
#endif

};


// Definitions for the FAT32 BPB uExtendedFlags field.
// B4-B6 and B8-B15 are reserved and zero, B0-B3 are only significant if
// B7 is set. FAT mirroring may be turned off when the 1st (or any subsequent
// mirror FAT) is damaged. FAT mirroring could conceivably be turned off when
// resizing a partition (or whatever) as well.
//
#define	BPB_NO_FAT_MIRRORING 0x80			// B7 set if NOT mirroring FAT
#define	BPB_ACTIVE_FAT_MASK	0x0F			// B0-B3: active FAT index


#if USING_FAT32

//  Do not use this structure.  It is only for use by the boot sector
//  structures.  Implementation of FAT32 defines that you not use this
//  structure.
//
typedef struct BootSectorBPB16 BPB622;
typedef BPB622 * PBPB622;
typedef BPB622 far * LPBPB622;
struct BootSectorBPB16
{
   	unsigned    sec_size;          	   // 00 bytes per sector
   	uchar    	sec_per_clus;      	   // 02 sectors per cluster
   	unsigned    res_sec;           	   // 03 reserved sector count (before FAT)
   	uchar    	num_fats;         	   // 05 count of FATs
   	unsigned    uDirEntries;         	// 06 count of root DIR entries (0 on FAT32)
   	unsigned    num_sec;           	   // 08 count of drive sectors (may be 0)
   	uchar    	media;             	   // 0A media ID byte
   	unsigned    uFatSectors;          	// 0B count of sectors per fat (0 on FAT32)
														// the next 4 are for device use only
   	unsigned    sec_per_trak;      	   // 0D count of sectors per track
   	unsigned    sides;             	   // 0F count of disk heads (sides)
   	ulong		   hidden;            	   // 11 count of hidden sectors (before boot sector)
   	ulong			lnumsec;           	   // 15 count of sectors (if num_sec==0)
};

#endif // #if USING_FAT32


/*
   Critical Error

   Description: Upon INT 24H, a control word is passed to
   the handler in AX.  AH describes the error, AL is the
   disk on which the critical error occurred (0=A, 1=B).

   Control Byte (AH)
   1xxx xxxx   Char Device (0=disk)
   xx1x xxxx   Ignore allowed
   xxx1 xxxx   Retry allowed
   xxxx 1xxx   Fail allowed
   xxxx x11x   Disk Area (00=boot, 01=fat, 10=dir, 11=data)
   xxxx xxx1   Write Attempt (0=read)
*/

/*
   DEVICE HEADER
   Description: The beginning of a valid device driver

   Reference: Adv MS-DOS, p 264  -or-  DOS Ref, p 340

   Device Attribute Char:
   1xxx xxxx xxxx xxxx    Character Device (1=char)
   x1xx xxxx xxxx xxxx    IOCTL (4402H-4405H)
   xx1x xxxx xxxx xxxx    Output Until Busy
   xxxx 1xxx xxxx xxxx    Open/Close
   xxxx xxxx 1xxx xxxx    Query IOCTL (4410H)
   xxxx xxxx x1xx xxxx    Generic IOCTL (440CH)
   xxxx xxxx xxx1 xxxx    FAST I/O (INT 29H)
   xxxx xxxx xxxx 1xxx    Clock Device
   xxxx xxxx xxxx x1xx    NUL Device
   xxxx xxxx xxxx xx1x    Stdout Device
   xxxx xxxx xxxx xxx1    Stdin Device

   Device Attribute Disk:
   0xxx xxxx xxxx xxxx    Disk Device (0=disk)
   x1xx xxxx xxxx xxxx    IOCTL (4402H-4405H)
   xx1x xxxx xxxx xxxx    IBM block format (0=FAT, 1=scratch?)
   xxx1 xxxx xxxx xxxx    Network device??
   xxxx 1xxx xxxx xxxx    Removable Media (4408H)
   xxxx xxxx 1xxx xxxx    Query IOCTL (4411H)
   xxxx xxxx x1xx xxxx    Generic IOCTL (440DH)
   xxxx xxxx xxxx xx1x    Driver Supports 32 bit sectors
*/
typedef struct DeviceHeader DEV_HDR;
typedef DEV_HDR * PDEV_HDR;
typedef DEV_HDR far * LPDEV_HDR;
struct DeviceHeader /*()*/
{
   LPDEV_HDR next; 				/* 00 ptr to next device   */
   unsigned  attr;            /* 04 attribute of device  */
   unsigned  strategy;        /* 06 ptr to strat routine */
   unsigned  service;         /* 08 ptr to serv routine  */
   char      name[8];         /* 0A name of char device  */
};

/*
   DEVICE REQUEST HEADER
   Description: Passed to a device (a device call)

   Device Commands
      0   Init
      1   Media Check
      2   Build BPB
      3   IOCTL Read        (bit 14 required)
      4   Read
      5   Nondestructive Read
      6   Input Status
      7   Input Flush
      8   Write
      9   Write w/ Verify
     0A   Output Status
     0B   Output Flush
     0C   IOCTL Write       (bit 14 required)
     0D   Device Open       (bit 11 required)
     0E   Device Close      (bit 11 required)
     0F   Disk Removable?   (bit 11 required)
     10   Output Until Busy (bit 13 required)
     13   Generic IOCTL     (bit  6 required)

   Device Status
   15           9 8 7             0
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |E|         |B|D|               |
   |r|         |u|o|               |
   |r|         |s|n|               |
   | |         |y|e|               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

   Reference: Adv MS-DOS, p 265  -or-  DOS Ref, p 345
*/
typedef struct dev_req_hdr DEVREQHDR;
typedef DEVREQHDR * PDEVREQHDR;
typedef DEVREQHDR far * LPDEVREQHDR;
struct dev_req_hdr /*()*/
{
   uchar      length;        /* 00 length of the command    */
   uchar      unit;          /* 01 unit # for disk access   */
   uchar      command;       /* 02 the command itself       */
   unsigned   status;        /* 03 returned status          */
   uchar      reserved[8];   /* 05 */
   uchar      media;         /* 0D media ID for disks       */
   void far * trans;         /* 0E data transfer buffer     */
   unsigned   count;         /* 12 length of data           */
   unsigned   start;         /* 14 starting sector for data */
   uchar      drive;         /* 16 starting drive #         */
   uchar      res2[3];       /* 17 */
   ulong      start32;       /* 1A long starting sector     */
};



/*
   DRIVE PARAMETER BLOCK (DPB)

   Description: Each FAT drive has one of these and one
   Current Directory Structure.  The DPB keeps track of
   the various LOGICAL areas on the disk.  It does not
   understand anything PHYSICAL about the disk.

   Reference: DOS Ref, p 637
*/
typedef struct DriveParameterBlock DPB;
typedef DPB * PDPB;
typedef DPB far * LPDPB;
struct DriveParameterBlock
{
	uchar    	drive;           	// 00 0=A, 1=B, 2=C, etc
	uchar       unit;            	// 01 unit # to pass to device
	unsigned    bytes_per_sec;  	// 02 bytes per sector (128,256,512,1024)
	uchar       ms_spc;          	// 04 sectors per cluster-1
	uchar       shift;           	// 05 shift factor to get spc
	unsigned    fat_start;       	// 06 # of "reserved" sectors
	uchar       num_fats;	     	// 08 copies of fat
	unsigned    dir_ent;			  	// 09 # of root dir entries (FAT32 drive=0)
	unsigned    data_start;      	// 0B (low word) of sector # for start of data
	unsigned    hiclus;          	// 0D highest cluster # (FAT32 drive=0)
	unsigned    fat_len;         	// 0F sectors per fat (FAT32 drive=0)
	unsigned    dir_start;       	// 11 SECTOR where root dir starts (FAT32 drive=-1)
	LPDEV_HDR   device;        	// 13 pointer to the device driver
	uchar       media;           	// 17 disk ID byte (360K = 0xFD, h.d. = 0xF8)
   uchar			rebuild;				// 18 0xFF means rebuild drive, 0 means no need, 1 means FSINFO needs to be committed
   LPDPB    	next;            	// 19 pointer to next DPB
   unsigned		newclus;				// 1D start here when allocating (FAT32 undefined)
   unsigned		freeclus;         // 1F (low word) of free clusters, -1 = unknown

#if USING_FAT32
   unsigned		uFreeClustersHigh;// 21 FAT32 high word of free clusters (-1 if not FAT32 drive, see FSINFO)
	unsigned		uExtendedFlags;   // 23 FAT32 extended flags from BPB (undefined? if not FAT32 drive)
   unsigned    uFSInfoSector;    // 25 FAT32 sector # of FSINFO struc  (-1 if none, undefined? if not FAT32 drive)
   unsigned    uBackUpBootSector;// 27 FAT32 sector # of backup boot sector (-1 if none, undefined? if not FAT32 drive)
   ulong			ulFirstDataSector;// 29 FAT32 sector # of first cluster (low word also stored in data_start, if not FAT32 drive then = data_start)
   ulong			ulMaxCluster;		// 2D FAT32 count of drive clusters + 1 (if not FAT32 drive then = hiclus)
   ulong			ulSectorsPerFAT;	// 31 FAT32 count of sectors per FAT (if not FAT32 drive then = fat_len)
   ulong			ulRootDirStartCluster; // 35 FAT32, may move (undefined? if not FAT32 drive)
   ulong			ulNextFreeCluster;// 39 FAT32 newclus (low word also stored in newclus if FAT32 drive else this field is 0, see FSINFO)
#endif
};

// DPB.media value for a "hard disk", used to determine if disk media can
// physically be changed ("hot swapped", actually) or if it just needs to
// be "rebuilt" at init.
//
#define	HARD_MEDIA	0xF8


/*
   BUFFER LIST STRUCTURE, DOS 5

   Description: Simple disk cache mechanism through sector
   oriented buffering scheme.  Based on most recently used
   method.

   Reference: (DOS Ref, p 965)

   Bit mapped flags:
   x1xx xxxx   Buffer is Dirty
   xx1x xxxx   Buffer is Valid
   xxxx 1xxx   Buffer contains Data sector
   xxxx x1xx   Buffer contains Root Dir sector
   xxxx xx1x   Buffer contains FAT sector
   xxxx xxx1   Buffer contains Boot sector
*/
typedef struct BufferListStructure BUFLIST;
typedef BUFLIST far * LPBUFLIST;
struct BufferListStructure
{
   BUFLIST *  prev;          /* 00 least recently used   */
   BUFLIST *  next;          /* 02 most recently used    */
   uchar      drive;         /* 04 drive, 0xFF = invalid */
   uchar      flags;         /* 05 bit mapped flags      */
   ulong      sector;        /* 06 what sector           */
   uchar      copies;        /* 0A copies of this sector */
#if USING_FAT32
   ulong      ulDistance;    /* 0B distance to next copy */
#else
   unsigned   distance;      /* 0B distance to next copy */
#endif
   LPDPB 	  lpDPB;      	  /* 0D (0F) pointer to DPB        */
	uchar      store_drive;   /* 11 (13) "remember" during crit errors */
   uchar      store_flags;   /* 12 (14)"remember" during crit errors */
   uchar      unknown;       /* 13 (15) */
#if USING_FAT32
	uchar      ucUnknown2;    /*    (16) */
	uchar      ucUnknown3;    /*    (17) */
#endif
   uchar      buffer[1];     /* 14 (18) the actual buffer     */
};


/*
   CURRENT DIRECTORY STRUCTURE (CDS)

   Description: Each drive has one of these.  All FAT
   drives have in addition a Drive Parameter Block.  The
   CDS contains the current directory and status for the
   drive.

   Reference: (DOS Ref, p 966)

   Status bit map:
   1xxx xxxx xxxx xxxx    Network drive
   x1xx xxxx xxxx xxxx    Valid drive (must be set)
   xx1x xxxx xxxx xxxx    JOINed drive
   xxx1 xxxx xxxx xxxx    SUBST drive
   xxxx xxx1 xxxx xxxx    Drive is Remote
*/
typedef struct CurrentDirectoryStructure CDS;
typedef CDS far *LPCDS;
struct CurrentDirectoryStructure /*()*/
{
   char       cwd[67];    					/* 00 as in "C:\" or "Network.node\" */
   unsigned   status;     					/* 43 bit mapped status, see above   */
   LPDPB 	  lpDPB;      					/* 45 network drives have no DPB     */
#if USING_FAT32
	ulong		  ulSubDirStartingCluster; // 49 subdir starting cluster, never 0
#else
   unsigned   subdir;						/* 49 subdir cluster # (0000=root)   */
   unsigned   unknown1;  				 	/* 4B */
#endif
   unsigned   unknown2;   					/* 4D */
   unsigned   skip;       					/* 4F chars to skip to get past root */
   uchar      unknown3;   					/* 51 */
   void far   *ifsptr;    					/* 52 ??? */
   unsigned   unknown4;   					/* 56 */
};



/*
   Internal structure for CONFIG.SYS processing.

   flags:
   xxxx xxx1           convert parameters to uppercase
   nnnn xxxx           pass on which to process
*/
struct config /*()*/
{
   char *command;
   void (*func)(char far *);
   char flags;
};

#define CFG_UPPERCASE (1)
#define PASS1         (1)
#define PASS2         (2)
#define PASS3         (4)
#define PASS1_FLAGS   (PASS1 << 4)
#define PASS2_FLAGS   (PASS2 << 4)
#define PASS3_FLAGS   (PASS3 << 4)


/*
   COUNTRY INFORMATION

   Description: The current country information

   Reference: DOS Ref, p 587
*/
#define DATESTART 7
typedef struct COUNTRY_INFO COUNTRY_INFO;
typedef COUNTRY_INFO * PCOUNTRY_INFO;
typedef COUNTRY_INFO far * LPCOUNTRY_INFO;
struct COUNTRY_INFO /*()*/
{
   uchar info_id;         /* 00 info id, always 1 */
   unsigned size;         /* 01 size 26H */
   unsigned country_code; /* 03 1 = USA */
   unsigned code_page;    /* 05 437 = US standard */
   unsigned datefmt;      /* 07 date format (0=USA) */
   char currency[5];      /* 09 currency symbol */
   char thousand[2];      /* 0E thousands separator */
   char decimal[2];       /* 10 decimal separator */
   char datesep[2];       /* 12 date separator */
   char timesep[2];       /* 14 time separator */
   uchar currfmt;         /* 16 currency format */
   uchar digits;          /* 17 # of digits after decimal */
   uchar timefmt;         /* 18 time format */
   void far * casemap;    /* 19 address of casemap routine */
   uchar datalist[2];     /* 1D data list separator */
};


/*
   DIRECTORY ENTRY (DIR)
   Description: Each item in the directory looks like this

   Directory attribute:
   xxxx xxxx xxxx xxx1    Read only
   xxxx xxxx xxxx xx1x    Hidden
   xxxx xxxx xxxx x1xx    System
   xxxx xxxx xxxx 1xxx    Volume label
   xxxx xxxx xxx1 xxxx    Subdirectory
   xxxx xxxx xx1x xxxx    Archive

   Reference: Adv MS-DOS, p 184  -or-  DOS Ref, p 248
*/
typedef struct DirectoryEntry DIR;
typedef DIR * PDIR;
typedef DIR far * LPDIR;
typedef LPDIR * PLPDIR;
struct DirectoryEntry /*()*/
{
   char     name[8];         		// 00 file name
   char     ext[3];          		// 08
   uchar    attr;            		// 0B file attribute
#if USING_FAT32
   uchar    reserved;				// 0Ch ???
	uchar		bCentiSeconds;			// 0Dh centiseconds of creation time
   unsigned uCreationTime;			// 0Eh file creation date/time
   unsigned uCreationDate;			// 10h
   unsigned uLastAccessDate;		// 12h date file was last accessed (opened)
   unsigned uStartingClusterHi;	// 14h high word of the starting cluster
#else
   uchar    reserved[10];    		// 0C */
#endif
   unsigned time;		            // 16 file date
   unsigned date;    		      // 18 file time
#if USING_FAT32
	unsigned uStartingClusterLo;	// 1A starting cluster of file
#else
	unsigned start;           		// 1A starting cluster of file
#endif
   ulong    size;            		// 1C file size in bytes
};


/*
   ROM-DOS USE ONLY!!

   This structure not used by that (we won't mention names)
   other DOS.  It keeps track of a directory entry.
   Note: this limits sectors to 4K in size maximum (80H
   directory entries, 20H bytes each = 4K)
*/
struct DIRENTRY /*()*/
{
   ulong dirsec;    /* sector # for this sector */
   uchar dirent;    /* 0-7FH */
   uchar eps;       /* entries per sector (10H at 512 byte sectors) */
	CLUSTER PhysicalCluster; // Current physical cluster for this dirent
	CLUSTER LogicalCluster; // Current logical cluster for this dirent
};


/*
   DOS REGISTERS

   Description: This is the standard user stack as seen by
   ROM-DOS.  All the user registers are saved in this order
   during the majority of DOS calls so DOS may modify them.

   Reference: (Adv MS-DOS, p 149  -or-  DOS Ref, p 74)
*/

union dos_regs /*()*/
{
   struct wordregs /*()*/
   {
      unsigned   ax;    /* 00 last on stack */
      unsigned   bx;    /* 02 */
      unsigned   cx;    /* 04 */
      unsigned   dx;    /* 06 */
      unsigned   si;    /* 08 */
      unsigned   di;    /* 0A */
      unsigned   bp;    /* 0C */
      unsigned   ds;    /* 0E */
      unsigned   es;    /* 10 */
      unsigned   ip;    /* 12 return address after INT 21 call */
      unsigned   cs;    /* 14 */
      unsigned   flags; /* 16 */
   } x;

   struct byteregs /*()*/
   {
      uchar      al;
      uchar      ah;
      uchar      bl;
      uchar      bh;
      uchar      cl;
      uchar      ch;
      uchar      dl;
      uchar      dh;
   } h;
};


/* PARTSTART

	Used only inside the DDT below, this structure
	contains DOS 4+ information for hard drives only.
	Floppy drives have a ULONG record of the last tick
	they were accessed on.
*/
struct PARTSTART
{
	int	ishd;						// set to 1 for Hard Disks
	int	strak;					// Starting trak of this partition
};

/*
   DRIVE DATA TABLE (DDT)

   Description: One of these is allocated to each internal
   drive unit (floppy and hard disks) during startup.
   This is used by smartdrv, to determine what the floppies
   and harddisks really are.

   Reference: DOS Internals, p 609 and Undocumented DOS 2nd Ed, p 765
*/
#define DATTR_NOTREMOVABLE							0x0001
#define DATTR_SUPPORTS_CHANGELINE				0x0002
#define DATTR_DISK_BPB_LOCKED						0x0004
#define DATTR_SECTORS_EQUALLY_SIZED				0x0008
#define DATTR_SHARES_PHYSICAL_DEVICE			0x0010
#define DATTR_ACTIVE_UNIT_FOR_SHARED_DEVICE	0x0020
#define DATTR_DISK_CHANGE_DETECTED				0x0040
#define DATTR_SET_DASD_BEFORE_FORMAT     		0x0080
#define DATTR_DISK_REFORMAT_DETECTED    		0x0100
#define DATTR_INVALID_MEDIA						0x0200 // Read/Write disabled
#define DATTR_LBA_REQUIRED							0x0400
typedef struct DriveDataTable DDT;
typedef DDT far * LPDDT;
struct DriveDataTable /*()*/
{
   LPDDT 	nextptr;  			// 00H  next one in the list
   uchar    bdrive;           // 04H  BIOS drive #
   uchar    dos_drive;        // 05H  dos_drive #
   BPB 		bpb;          		// 06H  the "current" BPB
   uchar    FatType;          // 1FH (3Bh) 0=12 bit, 40H=16 bit, 20h=32 bit
   int      opencnt;          // 20H (3Ch)  open count
   uchar    dtype;            // 22H (3Eh)  0=360,1=1.2,2=720,5=hard,7=1.44
   int      dattr;            // 23H (3Fh)  1=not-removable, 2=chgline status
   unsigned dcyl;             // 25H (41h)  # of cylinders
   BPB 		defbpb;       		// 27H (43h)  if drive is 1.44M, this BPB will be
#if (!USING_FAT32)
   char     unknown[6];    	// 40H        ???
#endif
   uchar    curtrack;         // 46H (78h)  for floppy HST adjustment
	union
	{
	   ulong  ticks;           // 47H (79h)  last tick count
		struct PARTSTART ps;		// 47H (79h)  start info for this partition
	} ddt_info;
   char     volume[12];       // 4BH (7Dh)  Volume Label from boot sector
   ulong    serial;           // 57H (89h)  Serial # from boot sector
   char     fatstr[9];        // 5BH (8Dh)  "FAT12" from boot sector
#if USING_BIG_DISKS
   unsigned	uEBiosDriveInfoFlags;  // 64h (96h)
   unsigned uEBiosDriveAttributes; // 66h (98h)
#endif
};


#if USING_FAT32

/*
   DRIVE DATA TABLE v6.22 (DDT622)

   Description: One of these is allocated to each internal
   drive unit (floppy and hard disks) during startup.
   This is used by smartdrv, to determine what the floppies
   and harddisks really are.

   Reference: DOS Internals, p 609 and Undocumented DOS 2nd Ed, p 765
*/
typedef struct DriveDataTable622 DDT622;
typedef DDT622 far * LPDDT622;
struct DriveDataTable622 /*()*/
{
   LPDDT 	nextptr;  			// 00H  next one in the list
   uchar    bdrive;           // 04H  BIOS drive #
   uchar    dos_drive;        // 05H  dos_drive #
   BPB622	bpb;          		// 06H  the "current" BPB
   uchar    FatType;          // 1FH (3Bh) 0=12 bit, 40H=16 bit, 20h=32 bit
   int      opencnt;          // 20H (3Ch)  open count
   uchar    dtype;            // 22H (3Eh)  0=360,1=1.2,2=720,5=hard,7=1.44
   int      dattr;            // 23H (3Fh)  1=not-removable, 2=chgline status
   unsigned dcyl;             // 25H (41h)  # of cylinders
   BPB622	defbpb;       		// 27H (43h)  if drive is 1.44M, this BPB will be
   char     unknown[6];    	// 40H        ???
   uchar    curtrack;         // 46H (78h)  for floppy HST adjustment
	union
	{
	   ulong  ticks;           // 47H (79h)  last tick count
		struct PARTSTART ps;		// 47H (79h)  start info for this partition
	} ddt_info;
   char     volume[12];       // 4BH (7Dh)  Volume Label from boot sector
   ulong    serial;           // 57H (89h)  Serial # from boot sector
   char     fatstr[9];        // 5BH (8Dh)  "FAT12" from boot sector
};

#endif // #if USING_FAT32


/*
	DEVICE PARMS

	Description: Used for calls 440DH - 0860H and 0840H.

   Reference: MS Prog Ref, p38
*/
struct DEVICE_PARMS
{
	uchar			specFunc;
	uchar			devType;
	unsigned		devAttr;
	unsigned		cylinders;
	uchar			mediaType;
	BPB         bpb;
};


/*
   EXE HEADER
   Description: Every EXE begins with this structure

   Reference: Adv MS-DOS, p 32  -or-  DOS Ref, p 62
*/
typedef struct EXEHDR EXEHDR;
typedef EXEHDR * PEXEHDR;
typedef EXEHDR far * LPEXEHDR;
struct EXEHDR /*()*/
{
   unsigned sig;     /* 00  Mark Zimbowski does it again...  */
   unsigned imgmod;  /* 02  image len mod 512 bytes (a page) */
   unsigned imglen;  /* 04  image len in 512-byte pages      */
   unsigned nreloc;  /* 06  number of relocatable items      */
   unsigned hdrlen;  /* 08  size of header in paragraphs     */
   unsigned minpar;  /* 0A  min paragraphs of RAM required   */
   unsigned maxpar;  /* 0C  max paragraphs of RAM needed     */
   unsigned ssdsp;   /* 0E  SS = this + startseg             */
   unsigned spofs;   /* 10  value placed in SP upon execute  */
   unsigned cksum;   /* 12  ignored by operating system      */
   unsigned ipofs;   /* 14  value placed in IP upon execute  */
   unsigned csdsp;   /* 16  CS = this + startseg             */
   unsigned reldsp;  /* 18  location of 1st relocatable item */
   unsigned ovlnum;  /* 1A  Overlay no. (normally 0)         */
};


/*
   EXEC PARAMETER BLOCK
   Description: Used by INT 21H call 4BH, spawn a program

   Reference: Adv MS-DOS, p   -or-  DOS Ref, p 667
*/
union EXEPARM /*()*/
{
   struct
   {
      unsigned envseg;        /* 00 */
      void far *command_line; /* 02 */
      void far *fcb1;         /* 06 */
      void far *fcb2;         /* 0A */
      unsigned spofs;         /* 0E  ------------------------ */
      unsigned ssdsp;         /* 10  these are set by DOS and */
      unsigned ipofs;         /* 12  returned on a 4B01H call */
      unsigned csdsp;         /* 14  ------------------------ */
   } e;

   struct
   {
      unsigned ovrseg;        /* overlay segment */
      unsigned reloc;         /* relocation factor */
   } o;

   struct
   {
      unsigned res_0;         /* reserved, should be 0 */
      unsigned type;          /* bit 0 = EXE, bit 1 = overlay */
		unsigned long filespec; /* seg:ofs to filename, may include path */
		unsigned pspseg;        /* PSP seg of new program */
		unsigned long newcsip;  /* cs:ip of new program */
		unsigned long exelen;   /* program size, including PSP */
   } es;
};


/*
   FILE CONTROL BLOCK (FCB)
   Description: DOS 1 file I/O uses this structure.  Note
   that the file time and date are in reverse order from
   all other DOS structures.  Also, a DOS1 findfirst/next
   does not use this structure (findfirst returns a DIR).

   Reference: (DOS Ref, p 273)
*/
typedef struct FileControlBlock FCB;
typedef FCB * PFCB;
typedef FCB far * LPFCB;
struct FileControlBlock /*()*/
{
   uchar      drive;             // 00  drive of file
   uchar      name[8];           // 01  file name
   uchar      ext[3];            // 09  and extension
   unsigned   cur_blk;           // 0C  current block
   unsigned   record_size;       // 0E  record size
   ulong      file_size;         // 10  file size
   unsigned   date;              // 14  date of file
   unsigned   time;              // 16  time of file
   uchar      sft;               // 18  sft # for file
   uchar      net_ioctl;         // 19  lo byte ioctl for netFCBs
   unsigned   reserved1;         // 1A  unknown
#if USING_FAT32
   ulong		  ulStartingCluster;	// 1C
#else
   unsigned   net_stclus;        // 1C  starting cluster
   unsigned   reserved2;         // 1E  unknown
#endif
   uchar      cur_record;        // 20  current record
   ulong      rand_record;       // 21  random record #
};


/*
   EXTENDED FILE CONTROL BLOCK (XFCB)
   Description: DOS 1 file I/O uses this structure.  Note
   that the file time and date are in reverse order from
   all other DOS structures.  Also, a DOS1 findfirst/next
   does not use this structure (findfirst returns a DIR).

   Reference: (DOS Ref, p 273)
*/
typedef struct XFileControlBlock XFCB;
typedef XFCB * PXFCB;
typedef XFCB far * LPXFCB;
struct XFileControlBlock /*()*/
{
	uchar			uXFCBIndicator;		// 00 should be FFh
	uchar		 	reserved3[5];			// 01
	uchar			uAttributes;			// 06 attributes file/dir
   uchar			drive;               // 07  drive of file (0=default,1=A)
   uchar       name[8];             // 08  file name
   uchar       ext[3];              // 10  and extension
   unsigned    cur_blk;             // 13  current block
   unsigned    record_size;         // 15  record size
   ulong       file_size;           // 17  file size
   unsigned    date;                // 1b  date of file
   unsigned    time;                // 1d  time of file
   uchar       sft;                 // 1f  sft # for file
   uchar       net_ioctl;           // 20  lo byte ioctl for netFCBs
   unsigned    reserved1;           // 21  unknown
#if USING_FAT32
   ulong		   ulStartingCluster;	// 23
#else
   unsigned    net_stclus;          // 23  starting cluster
   unsigned    reserved2;           // 25  unknown
#endif
   uchar       cur_record;          // 27  current record
   ulong       rand_record;         // 28  random record #
};


/*
   DOS2 FINDFIRST/FINDNEXT
   Description: Format of a findfirst/findnext placed in
   the user's memory pointed to by the DTA.

   Reference: Adv MS-DOS, p 445  -or-  DOS Ref, p 671
*/
typedef struct FindFirstNext FIND;
typedef FIND * PFIND;
typedef FIND far * LPFIND;
struct FindFirstNext /*()*/
{
   uchar      	drive;       			// 00  1=A, 2=B, 3=C, etc
   char       	emask[11];   			// 01  expanded name
   uchar      	srchattr;    			// 0C  search attribute
   unsigned   	dirnum;      			// 0D  directory entry number
#if USING_FAT32
	ulong			ulStartingCluster;	// 0F  cluster # of subdir
	uchar		 	res[2];					// 13  reserved (always 0)
#else
   unsigned   	subdir;      			// 0F  cluster # of subdir
   uchar      	res[4];      			// 11  reserved (always 0)
#endif
   uchar      	attr;        			// 15  attribute found
   unsigned   	time;        			// 16  file's time
   unsigned   	date;        			// 18  file's date
   ulong      	size;        			// 1A  file's size
   char       	name[13];    			// 1E  filename in dot form
};


/*
   ROM-DOS USE ONLY!!

   This structure not used by that (we won't mention names)
   other DOS.  It keeps track of a directory entry.
*/
struct LOCK /*()*/
{
   ulong start;    /* start of lock/unlock region */
   ulong len;      /* length of lock/unlock region */
};


/*
   SYSTEM FILE TABLE (SFT)

   Description: Each file handle points to one of these
   structures.  There is one list of SFTs for DOS2 calls
   and a separate list of SFTs for DOS1 FCB calls.

   Reference: DOS Ref, p 812

   Field Descriptions:

   RWMODE bit map
   -------------------
   x1xx xxxx xxxx xxxx      All writes happen immediately
   xx1x xxxx xxxx xxxx      Do not generate Critical Error
   xxxx xxxx 1xxx xxxx      Handle not passed to child

   IOCTL bit map
   -------------------
   1xxx xxxx xxxx xxxx      File is on Network device
   xxxx xxxx 1xxx xxxx      1=device, 0=file

   ioctl bit map if file
   x1xx xxxx 0xxx xxxx      Don't change date/time on close
   xxx1 xxxx 0xxx xxxx      Child process won't inherit file
   xxxx xxxx 01xx xxxx      File has not been written to
   xxxx xxxx 0x11 1111      Drive # (0=A, 2=C)

   ioctl bit map if device
   x1xx xxxx 1xxx xxxx      Device can process ioctl strings
   xxxx xxxx 1xxx xxxx      Always 1 (specifies char device)
   xxxx xxxx 11xx xxxx      Not EOF
   xxxx xxxx 1x1x xxxx      Raw (or binary) mode
   xxxx xxxx 1xx1 xxxx      Use Fast I/O (INT 29H)
   xxxx xxxx 1xxx 1xxx      CLOCK$ device
   xxxx xxxx 1xxx x1xx      NUL device
   xxxx xxxx 1xxx xx1x      CON output
   xxxx xxxx 1xxx xxx1      CON input
*/
typedef struct SystemFileTable SFT;
typedef SFT far * LPSFT;
typedef LPSFT * PLPSFT;
struct SystemFileTable /*()*/
{
   unsigned     handles;         // 00 # of handles using this SFT
   unsigned     rwmode;          // 02 0=rd_only, 1=wr_only, 2=rw
   uchar        attr;            // 04 file attribute
   unsigned     ioctl;           // 05 bit map for ioctl (device)
   LPDEV_HDR    device;          // 07 ptr to device or DPB
#if USING_FAT32
   unsigned     uRelClusterHigh; // 0B relative cluster of file
#else
   unsigned     start;           // 0B starting cluster of file
#endif  // #if USING_FAT32
   unsigned     time;            // 0D date and time of file entry
   unsigned     date;            // 0F
   ulong        size;            // 11 the file size
   ulong        offset;          // 15 offset for next read/write
#if USING_FAT32
   unsigned     uRelClusterLow;  // 19 relative cluster of file
#else
   unsigned     relclus;         // 19 logical clus last accessed
#endif  // #if USING_FAT32
   ulong        dirsec;          // 1B sector containing dir entry
   uchar        dirent;          // 1F entry # within above sector
   char         name[8];         // 20 filename */
   char         ext[3];
#if USING_FAT32
   ulong        ulStartingCluster; // 2B starting cluster of file
#else
   void far    *shareptr;        // 2B ??? share pointer
#endif  // #if USING_FAT32
   unsigned     shareid;         // 2F ??? network id of process
   unsigned     owner;           // 31 psp of owner of this SFT

#if USING_FAT32 && LSEEK_ENHANCEMENT
   unsigned     uRDlseekHigh;    // 33 ???
#else
   unsigned     shareoff;        // 33 ??? sharing record offset
#endif  // #if USING_FAT32 && LSEEK_ENHANCEMENT

#if USING_FAT32
   ulong        ulAbsCluster;    // 35 as above, 0000 = invalid
#if LSEEK_ENHANCEMENT
   unsigned     uRDlseekLow;     // 39 ???
#else
   unsigned     unknown;         // 39 ???
#endif  // #if LSEEK_ENHANCEMENT
#else
   unsigned     absclus;         // 35 as above, 0000 = invalid
   void far    *ifsptr;          // 37 ???
#endif  // #if USING_FAT32
};

typedef struct SystemFileTable622 SFT622;
typedef SFT622 far * LPSFT622;
typedef LPSFT622 * PLPSFT622;
struct SystemFileTable622 /*()*/
{
   unsigned   	handles;         	// 00 # of handles using this SFT
   unsigned   	rwmode;          	// 02 0=rd_only, 1=wr_only, 2=rw
   uchar      	attr;            	// 04 file attribute
   unsigned   	ioctl;           	// 05 bit map for ioctl (device)
   LPDEV_HDR 	device; 				// 07 ptr to device or DPB
   unsigned   	start;           	// 0B starting cluster of file
   unsigned   	time;            	// 0D date and time of file entry
   unsigned   	date;            	// 0F
   ulong      	size;            	// 11 the file size
   ulong      	offset;          	// 15 offset for next read/write
   unsigned   	relclus;         	// 19 logical clus last accessed
   ulong      	dirsec;          	// 1B sector containing dir entry
   uchar      	dirent;          	// 1F entry # within above sector
   char       	name[8];         	// 20 filename */
   char       	ext[3];
   void far    *shareptr;        // 2B ??? share pointer
   unsigned   	shareid;         	// 2F ??? network id of process
   unsigned   	owner;           	// 31 psp of owner of this SFT
   unsigned   	shareoff;        	// 33 ??? sharing record offset
   unsigned   	absclus;         	// 35 as above, 0000 = invalid
   void far 	*ifsptr;				// 37 ???
};


/*
   DEVICE CONTROL BLOCK (DCB)

   Description: A high-brid of a linked list and an
   array of System File Tables.  The numsft field says
   how many SFTs follow this structure.

   Reference: DOS REF 2nd, p 810
*/
typedef struct DeviceControlBlock DCB;
typedef DCB far *LPDCB;
typedef LPDCB * PLPDCB;
struct DeviceControlBlock /*()*/
{
   LPDCB 		next;  		// 00 ptr next array of SFTs
   unsigned   	numsft;     // 04 number of SFTs in this array
   SFT 		  	sft[1];     // 06 the array of SFTs for this DCB
};

/*
   LIST OF LISTS (LOL),  DOS 5

   Description: Also called SysVars this structure provides
   access to DOS internal structures. It is mainly static,
   and is initialized at boot time.  See INT 21H, function
   52H.

   Reference: (DOS Ref, p 964)
*/

typedef struct LOL LOL;
typedef LOL * PLOL;
typedef LOL far * LPLOL;
struct LOL /*()*/
{
   char       	unknown1[4];        	/* 00 ??? */
   char       	dos5;               	/* 04 DOS 3 or 5 structures?    */
   char       	unknown2[9];        	/* 05 ??? */
   int        	netbios_id;        	/* 0E returned via func 5EH     */
   char       	unknown3[10];       	/* 10 ??? */
   int        	share_retry;        	/* 1A sharing retry             */
   int        	share_delay;        	/* 1C sharing delay             */
   void far 	*cur_buf;            /* 1E pointer to current buffer */
   unsigned   	bufoff;             	/* 22 offset within cur buf     */
   unsigned   	first_mcb;          	/* 24 segment of 1st MCB        */

   /* ----- address returned by INT 21H func 52H (DGROUP:26) ---- */

   LPDPB 		first_dpb;     		/* 26 linked list of DPBs       */
   LPDCB 		first_dcb;    		 	/* 2A linked list of SFTs       */
   LPDEV_HDR 	clkdev;    				/* 2E ptr to CLOCK device       */
   LPDEV_HDR 	condev;    				/* 32 ptr to CON device         */
   unsigned   	bufsize;            	/* 36 max sect size on any disk */
   void far * far *bufinfo;       	/* 38 linked list of disk buffs */
   LPCDS 	  	first_cds;  	   	/* 3C ptr to array of CDSs      */
   LPDCB 		first_fcb;	   	  	/* 40 list of SFTs for FCBs     */
   unsigned   	numfcbs;            	/* 44 # SFTs for use w/ FCBs    */
   uchar      	numdrives;          	/* 46 Number of Logical Drives  */
   uchar      	lastdrive;          	/* 47 Value of lastdrive        */
   DEV_HDR     first_dev;      		/* 48 actual hdr of NULL device */
   uchar      	nulbyte;            	/* 5A just a 0                  */
   unsigned   	specprogs;          	/* 5B ??? */
   void far 	*utilfuncs;          /* 5D ??? */
   void far 	*ifschain;           /* 61 ??? */
   unsigned   	numbufs;            	/* 65 Total BUFFERS */
   unsigned   	lookahead;          	/* 67 ??? */
   uchar      	bootdrive;          	/* 69 A=1, B=2, C=3, etc */
   uchar      	unknown;            	/* 6A ??? */
   unsigned   	extmemsize;         	/* 6B Extended memory size in K */
};


/*
   MEMORY CONTROL BLOCK (MCB)

   Description: each block of memory is owned by a process

   Reference: (Adv MS-DOS, p --  -or-  DOS Ref, p 77)

	The ROM-DOS specific init_id field serves two purposes.

	1) It temporarily holds the ID of the allocated block
		during CONFIG.SYS processing so that, if the process
		succeeds, we can update the block ID field in those
		MCBs that are contained insid the System Data (SD)
		memory block.

	2) In the special case of the lol.first_mcb MCB, the init_id
		field is set to zero during CONFIG.SYS processing so that
		we perform all allocations using an alternate MCB chain
		that is contained entirely inside the SD memory block.
		At the end of config sys processing the variable is
		incremented to indicate that the lol.first_mcb MCB is
		to be treated as a single MCB containing all of the
		system data. See the comments on pFirstMCB() (DOSMEM.C)
		as well as mem_init() and mem_combine() (DOSMAIN.C).
*/
typedef struct MCB MCB;
typedef MCB * PMCB;
typedef MCB far * LPMCB;
struct MCB /*()*/
{
   uchar      id;        /* 00 'M' for normal, 'Z' for last MCB */
   unsigned   owner;     /* 01 psp segment of owner        */
   unsigned   para_cnt;  /* 03 # of paragraphs in this MCB */
   char       unused[2]; /* 05                             */
   char       init_id;   /* 07 ROM-DOS specific            */
   char       name[8];   /* 08 name of program or device   */
};


/*
   MASTER BOOT RECORD PARTITION
   Description: Used by the hard disk driver to determine
   hard disk partition sizes.

   Reference: Adv MS-DOS, p   -or-  DOS Ref, p 217
*/
struct PARTITION /*()*/
{
   uchar boot;             /* 00 is this partition bootable? (80H=true) */
   uchar shead;            /* 01 starting head   */
   uchar ssect;            /* 02 starting sector */
   uchar strak;            /* 03 starting track  */
   uchar id;               /* 04 id, 1=DOS12, 4=DOS16, 5=DOS extended */
   uchar ehead;            /* 05 ending head   */
   uchar esect;            /* 06 ending sector */
   uchar etrak;            /* 07 ending track  */
   ulong hidden;           /* 08 start of partition from part sector */
   ulong size;             /* 0C size of partition in sectors */
};

// Hard Disk Partition IDs
//
#define DOS12				0x01
#define DOS16				0x04
#define DOSEXT				0x05
#define DOSBIG				0x06
#define DOS32				0x0B
#define DOS32_LBA			0x0C
#define DOSBIG_LBA		0x0E
#define DOSEXT_LBA		0x0F
#define RELIANCE        0xCC

// Bootable partition flag (DOS)
//
#define BOOTABLE_PART	0x80

#define PARTOFF			0x1BE	// offset of the partition information



/*
   PROGRAM SEGMENT PREFIX (PSP)
   Description: DOS builds 100H bytes of special data
   before every loaded program (COM or EXE).

   Reference: Adv MS-DOS, p 23  -or-  DOS Ref, p --
*/
typedef struct PSP PSP;
typedef PSP * PPSP;
typedef PSP far * LPPSP;
struct PSP /*()*/
{
   unsigned   int_20;             /* 00 INT 20H instruction  */
   unsigned   top_of_memory;      /* 02 the size of this MCB */
   uchar      res1;               /* 04 unused               */
   uchar      dos_long_call[5];   /* 05 CPM style DOS call   */
   void far * term_adr;           /* 0A prev INT 22H handler */
   void far * ctrl_brk_adr;       /* 0E prev INT 23H handler */
   void far * critical_error_adr; /* 12 prev INT 24H handler */
   unsigned   last_psp;           /* 16 prev psp segment     */
   uchar      hptr[20];           /* 18 the handle table     */
   unsigned   env_seg;            /* 2C environment segment  */
   void far * user_stack;         /* 2E ptr to stack on terminate */
   unsigned   hcount;             /* 32 (20 handles default) */
   void far * htbl_addr;          /* 34 ptr to handle table  */
   long       share_psp;          /* 38 share only, always -1 */
   uchar      res2[4];            /* 3C unused               */
   unsigned   dosver;             /* 40 DOS version          */
   uchar      res3[0x0E];         /* 42 unused               */
   uchar      doscall[3];         /* 50 not supported        */
   uchar      res4[9];            /* 53 unused               */
   uchar      fcb1[16];           /* 5C a parsed FCB         */
   uchar      fcb2[16];           /* 6C a 2nd parsed FCB     */
   uchar      res5[4];            /* 7C unused               */
   uchar      count;              /* 80 size of cmdline      */
   uchar      command_line[0x7f]; /* 81 */
};


/*
   RELOCATABLE ITEM IN EXE
   Description: One of these entries exist for every relocatable
   item in an EXE.  DOS fixes them up at load time so an EXE can
   run anywhere in memory.
*/
struct rel_item /*()*/
{
   unsigned offset;
   unsigned segment;
};


/*
   SEARCH DATA BLOCK (SDB)

   Description: This structure is the search parameters
   for a FindFirst/FindNext.  It is found in the SDA.
   Network redirectors use this structure.  See Found
   File Block.

	MSDOS 6.22 sets the DTA to the sda_sdb prior to doing an IFS
	find first/next. The IFS will fill in the SDB drive variable
	with the number of the IFS drive. MSDOS then stores this SDB
	value in the byte just after the SDB in the caller's FCB on exit
	from an int 21h fn 11h/12h. On entry to subsequent int 21h fn
	12h calls, the SDB.drive value is restored from the caller's
	this byte following the SDB. We update this byte in sda_fcb
	prior to copying the local FCB data into the user's space.

*/
typedef struct SearchDataBlock SDB;
struct SearchDataBlock
{
   uchar      	drive;       			// 00  0=default, 1=A, 2=B...
   char       	emask[11];   			// 01  name mask in expanded form
   uchar      	srchattr;   		 	// 0C  attribute search mask
   unsigned   	dirnum;		      	// 0D  directory entry number
#if USING_FAT32
   ulong	   	ulStartingCluster;   // 0F  cluster # of subdirectory
   uchar      	res1[2];     			// 13  unused
#else
   unsigned   	subdir;     		 	// 0F  cluster # of subdirectory
   uchar      	res1[4];		     		// 11  unused
#endif
};


#if USING_FAT32
/*
	Extended Get Free Space Structure

	This is structure used for getting the free space for a
	high capacity drive when using windows 95 OSR2 or above
	using int 21h fn 7303h.

*/
typedef struct GetDiskFreeSpace32 GDF32;
typedef GDF32 * PGDF32;
typedef GDF32 far * LPGDF32;
struct GetDiskFreeSpace32 /*()*/
{
	unsigned uSize;								// 00
	unsigned uLevel;								// 02
	ulong 	ulSectorsPerCluster;				// 04
	ulong 	ulBytesPerSector;					// 08
	ulong	 	ulAvailableClusters;				// 0c
	ulong 	ulTotalClusters;					// 10
	ulong 	ulAvailablePhysicalSectors;	// 14
	ulong 	ulTotalPhysicalSectors;			// 18
	ulong 	ulAvailableAllocationUnits;	// 1c
	ulong 	ulTotalAllocationUnits;			// 20
	ulong		ulReserved[2];						// 24
};

#define	GDF32_LEVEL_0	0

#endif	// USING_FAT32


#if USING_FAT32

//  Boot Sector for FAT16/FAT32 drives
//
//  This structure is constucted using a union of the FAT16 and FAT32
//  boot sector information.  You should use the typedefs below, not
//  the structures directly.
//
typedef struct BootSector BOOTSECTOR;
typedef BOOTSECTOR *      pBOOTSECTOR;
typedef BOOTSECTOR far *  lpBOOTSECTOR;

// Lengths of fields in the Boot Sector
//
#define BOOT_VOL_LEN				11
#define BOOT_FATID_LEN			0x08
#define BOOT_FATID_DIGITS_NDX 0x03
#define BOOT_OEM_NAME_LEN		0x08
#define BOOT_CODE_STARTS_16	0x5a
#define BOOT_CODE_STARTS_32	0x3e
#define XBOOT_REC_SIG			0x29


struct BootInformation32
{
    struct BiosParameterBlock BootSectorBPB;								 // 0B
    unsigned	uPhysicalDriveNumber;										 // 40
    uchar		ucExtendedBootRecordSig;									 // 42
    ulong		ulVolumeSerialNumber;										 // 43
    char			rgucVolumeLabel[BOOT_VOL_LEN];							 // 47
    char			rgucFileSystemID[BOOT_FATID_LEN];						 // 52
    char			rgucBootCodeStarts[(510-(11+BOOT_CODE_STARTS_32))];	// 5A
};

struct BootInformation16
{
    struct BootSectorBPB16	BootSectorBPB;									 // 0B
    unsigned	uPhysicalDriveNumber;										 // 24
    uchar		ucExtendedBootRecordSig;									 // 26
    ulong		ulVolumeSerialNumber;										 // 27
    char			rgucVolumeLabel[BOOT_VOL_LEN];							 // 2B
    char			rgucFileSystemID[BOOT_FATID_LEN];						 // 36
    char			rgucBootCodeStarts[(510-(11+BOOT_CODE_STARTS_16))];	// 3E
};

union BootInformation
{
    struct BootInformation16 BootInfo16;
    struct BootInformation32 BootInfo32;
};

struct BootSector
{
    char		rgucBS_JumpInstruction[3];								// 000
    char		rgucBS_FileSystemOEMName[BOOT_OEM_NAME_LEN];		// 003
    union BootInformation BootSectorInfo;							// 00B
    unsigned uBootSectorSignature;									// 1FE
};

#define BOOT_BPB				0x0B

#else	// #if USING_FAT32

//  These constants are dangerious because field offsets in structures
//  can change.  Use the structures in dostruct.h to get the offsets

// Offsets in Boot Sector
#define BOOT_BPB				0x0B
#define BOOT_XBOOTSIG		0x26		// location of value below
#define XBOOT_REC_SIG		0x29		// value at above spot in boot sector
#define BOOT_SERIAL			0x27
#define BOOT_VOL_STR			0x2B
#define BOOT_FATID_STR		0x36
#define BOOT_VOL_LEN			11
#define BOOT_FATID_LEN		0x08
#define BOOT_SERIAL_LEN	(4 + BOOT_VOL_LEN + BOOT_FATID_LEN)

#endif // #if USING_FAT32

#if USING_FAT32
typedef struct SDPDFormatStruct SFS;  // type tag: sfs (special format struc)
typedef SFS * PSFS;
typedef SFS far * LPSFS;
struct SDPDFormatStruct
{
    unsigned uSize;
    unsigned uLevel;
    ulong    ulFunction;
    ulong    ulValue1;
    ulong    ulValue2;
    ulong    ulValue3;
    ulong    ulValue4;
};
#define SETDPB_SET_ALLOC_INFO           0
#define SETDPB_SET_DPB_FROM_BPB         1
#define SETDPB_FORCE_MEDIA_CHANGE       2
#define SETDPB_GET_SET_FAT_MIRROR       3
#define SETDPB_GET_SET_ROOT_DIR_CLUS    4
#endif  // USING_FAT32

#endif	// #ifndef	DOSTRUCT