;  $Id: USE186.INC 1.31 2003/08/25 15:17:44Z thomd Exp $
;
;  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;  Copyright (c) 2001-2003 Datalight, Inc.
;  All Rights Reserved
;
;  Datalight, Inc. is a Washington corporation located at:
;
;  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
;  Bothell, WA  98021                    Fax: +1.425.951.8094
;  USA                                   Http://www.datalight.com
;
;  This software, including without limitation all source code and
;  documentation, is the confidential, trade secret property of
;  Datalight, Inc., and is protected under the copyright laws of
;  the United States and other jurisdictions.  Portions of the
;  software may also be subject to one or more the following US
;  patents: US#5860082, US#6260156.
;
;  In addition to civil penalties for infringement of copyright
;  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;  penalties for violation of (a) the restrictions on circumvention
;  of copyright protection systems found in 17 U.S.C. 1201 and
;  (b) the protections for the integrity of copyright management
;  information found in 17 U.S.C. 1202.
;
;  U.S. Government Restricted Rights.  Use, duplication,
;  reproduction, or transfer of this commercial product and
;  accompanying documentation is restricted in accordance with
;  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
;  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
;  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;  THIS SOFTWARE IN ERROR.
;
;----------------------------------------------------------
; $DOCFILE:USE186.ASM
;
; Copyright (c) 2001 Datalight, Inc.
; All Rights Reserved.
; Datalight Confidential and Proprietary.
;


;  Enable this constant to compile ROMDOS 7.1 without 386 instructions
;
USE_186		equ	0


; when using .TYPE operator -- this the bit set for REGISTER
TYPE_REG EQU 10h

; macros for shifting and rotating 32-bit values.
;
; When USE_186 is set to 0, then these macros will use 386 registers
; and instructions to perform the specified shift or rotate.
; When USE_186 is set to 1, only 16 bit registers are used to perform
; the specified shift or rotate.
;
; below is a description of each macro.  In each macro, ovr is used
; to limit the number of parameters sent to the macro.
;
; ***********************************************************************
; * NOTE:  In 186 mode, the CX register cannot be passed to the macros  *
; *        as it contains the number of bits to shift as a loop counter.*
; *        An error message *will* be displayed if the CX register is   *
; *        detected as a parameter.                                     *
; ***********************************************************************
;
; M32_SHL
;       This macro shifts a 32-bit value left for the specified number
;       of bits.
;
;   386 mode:
;       reg     the register with the value to shift can be a 32 bit
;               register, but it is not required
;       num     the number of bit places to shift, between 1 and 31,
;               inclusive.
;
;   186 mode:
;       reghi   the 16 bit register containing the high word of the
;               32-bit value
;       reglo   the 16 bit register containing the low word of the
;               32-bit value
;       num     the number of bit places to shift, between 1 and 31,
;               inclusive.
;
; M32_SHR
;       This macro shifts a 32-bit value right for the specified number
;       of bits
;
;   386 mode:
;       reg     the register with the value to shift can be a 32 bit
;               register, but it is not required
;       num     the number of bit places to shift, between 1 and 31,
;               inclusive.
;
;   186 mode:
;       reghi   the 16 bit register containing the high word of the
;               32-bit value
;       reglo   the 16 bit register containing the low word of the
;               32-bit value
;       num     the number of bit places to shift, between 1 and 31,
;               inclusive.
;
; M32_SAR
;       This macro shifts a 32-bit value right for the specified number
;       of bits and retains the sign of the value (most significant
;       bit value is retained)
;
;   386 mode:
;       reg     the register with the value to shift can be a 32 bit
;               register, but it is not required
;       num     the number of bit places to shift, between 1 and 31,
;               inclusive.
;
;   186 mode:
;       reghi   the 16 bit register containing the high word of the
;               32-bit value
;       reglo   the 16 bit register containing the low word of the
;               32-bit value
;       num     the number of bit places to shift, between 1 and 31,
;               inclusive.
;
; M32_ROL
;       This macro rotates a 32-bit value left for the specified number
;       of bits
;
;   386 mode:
;       reg     the register with the value to rotate can be a 32 bit
;               register, but it is not required
;       num     the number of bit places to rotoate, between 1 and 31,
;               inclusive.
;
;   186 mode:
;       reghi   the 16 bit register containing the high word of the
;               32-bit value
;       reglo   the 16 bit register containing the low word of the
;               32-bit value
;       num     the number of bit places to rotate, between 1 and 31,
;               inclusive.
;
; M32_ROR
;       This macro rotate a 32-bit value right for the specified number
;       of bits
;
;   386 mode:
;       reg     the register with the value to rotate can be a 32 bit
;               register, but it is not required
;       num     the number of bit places to rotate, between 1 and 31,
;               inclusive.
;
;   186 mode:
;       reghi   the 16 bit register containing the high word of the
;               32-bit value
;       reglo   the 16 bit register containing the low word of the
;               32-bit value
;       num     the number of bit places to rotate, between 1 and 31,
;               inclusive.
;
; M32_RCL
;       This macro rotates a 32-bit value left for the specified number
;       of bits through the carry flag.
;
;   386 mode:
;       reg     the register with the value to rotate can be a 32 bit
;               register, but it is not required
;       num     the number of bit places to rotoate, between 1 and 31,
;               inclusive.
;
;   186 mode:
;       reghi   the 16 bit register containing the high word of the
;               32-bit value
;       reglo   the 16 bit register containing the low word of the
;               32-bit value
;       num     the number of bit places to rotate, between 1 and 31,
;               inclusive.
;
; M32_RCR
;       This macro rotate a 32-bit value right for the specified number
;       of bits through the carry flag.
;
;   386 mode:
;       reg     the register with the value to rotate can be a 32 bit
;               register, but it is not required
;       num     the number of bit places to rotate, between 1 and 31,
;               inclusive.
;
;   186 mode:
;       reghi   the 16 bit register containing the high word of the
;               32-bit value
;       reglo   the 16 bit register containing the low word of the
;               32-bit value
;       num     the number of bit places to rotate, between 1 and 31,
;               inclusive.
;

if USE_186 EQ 0



;***********************************************************************
M32_SHL MACRO reg, num, ovr
.errb <reg>
.errb <num>
.errnb <ovr>

if (.TYPE reg) AND TYPE_REG
   if num EQ 0
      .err "M32_SHL: shift size must be 1-31: &num"
      exitm
    elseif num GT 31
      .err "M32_SHL: shift size must be 1-31: &num"
      exitm
    else
      shl   reg, num
    endif
else
   .err "M32_SHL: bad parameter: &reg"
  exitm
endif
ENDM

;***********************************************************************
M32_SHR MACRO reg, num, ovr
.errb <reg>
.errb <num>
.errnb <ovr>

if (.TYPE reg) AND TYPE_REG
   if num EQ 0
      .err "M32_SHR: shift size must be 1-31: &num"
      exitm
    elseif num GT 31
      .err "M32_SHR: shift size must be 1-31: &num"
      exitm
    else
      shr   reg, num
    endif
else
   .err "M32_SHR: bad parameter: &reg"
  exitm
endif
ENDM

;***********************************************************************
M32_SAL MACRO reg, num, ovr
.errb <reg>
.errb <num>
.errnb <ovr>

if (.TYPE reg) AND TYPE_REG
   if num EQ 0
      .err "M32_SAL: shift size must be 1-31: &num"
      exitm
    elseif num GT 31
      .err "M32_SAL: shift size must be 1-31: &num"
      exitm
    else
      sal   reg, num
    endif
else
   .err "M32_SAL: bad parameter: &reg"
  exitm
endif
ENDM

;***********************************************************************
M32_SAR MACRO reg, num, ovr
.errb <reg>
.errb <num>
.errnb <ovr>

if (.TYPE reg) AND TYPE_REG
   if num EQ 0
      .err "M32_SAR: shift size must be 1-31: &num"
      exitm
    elseif num GT 31
      .err "M32_SAR: shift size must be 1-31: &num"
      exitm
    else
      sar   reg, num
    endif
else
   .err "M32_SAR: bad parameter: &reg"
  exitm
endif
ENDM

;***********************************************************************
M32_ROL MACRO reg, num, ovr
.errb <reg>
.errb <num>
.errnb <ovr>

if (.TYPE reg) AND TYPE_REG
   if num EQ 0
      .err "M32_ROL: shift size must be 1-31: &num"
      exitm
    elseif num GT 31
      .err "M32_ROL: shift size must be 1-31: &num"
      exitm
    else
      rol   reg, num
    endif
else
   .err "M32_ROL: bad parameter: &reg"
  exitm
endif
ENDM

;***********************************************************************
M32_ROR MACRO reg, num, ovr
.errb <reg>
.errb <num>
.errnb <ovr>

if (.TYPE reg) AND TYPE_REG
   if num EQ 0
      .err "M32_ROR: shift size must be 1-31: &num"
      exitm
    elseif num GT 31
      .err "M32_ROR: shift size must be 1-31: &num"
      exitm
    else
      ror   reg, num
    endif
else
   .err "M32_ROR: bad parameter: &reg"
  exitm
endif
ENDM


;***********************************************************************
M32_RCL MACRO reg, num, ovr
.errb <reg>
.errb <num>
.errnb <ovr>

if (.TYPE reg) AND TYPE_REG
   if num EQ 0
      .err "M32_RCL: shift size must be 1-31: &num"
      exitm
    elseif num GT 31
      .err "M32_RCL: shift size must be 1-31: &num"
      exitm
    else
      rcl   reg, num
    endif
else
   .err "M32_RCL: bad parameter: &reg"
  exitm
endif
ENDM

;***********************************************************************
M32_RCR MACRO reg, num, ovr
.errb <reg>
.errb <num>
.errnb <ovr>

if (.TYPE reg) AND TYPE_REG
   if num EQ 0
      .err "M32_RCR: shift size must be 1-31: &num"
      exitm
    elseif num GT 31
      .err "M32_RCRL: shift size must be 1-31: &num"
      exitm
    else
      rcr   reg, num
    endif
else
   .err "M32_RCR: bad parameter: &reg"
  exitm
endif
ENDM

else    ;; USE_186

;***********************************************************************
M32_SHL MACRO reghi, reglo, num, ovr
LOCAL SHL_loop
.errb <reghi>
.errb <reglo>
.errb <num>
.errnb <ovr>

if (.TYPE reghi) AND TYPE_REG
   if (.TYPE reglo) AND TYPE_REG
      if num EQ 0
         .err "M32_SHL: shift size must be 1-31: &num"
         exitm
      elseif num GT 31
         .err "M32_SHL: shift size must be 1-31: &num"
         exitm
      else
         ifidni <reghi>, <cx>
            .err "M32_SHL: cx is not a valid register!"
            exitm
         endif
         ifidni <reglo>, <cx>
            .err "M32_SHL: cx is not a valid register!"
            exitm
         endif

         push  cx
         mov   cx, num

       SHL_loop:
         shl   reglo, 1
         rcl   reghi, 1
         loop  SHL_loop
         pop   cx
      endif
   else
   .err "M32_SHL: bad parameter: &reglo"
   endif
else
   .err "M32_SHL: bad parameter: &reghi"
  exitm
endif
ENDM

;***********************************************************************
M32_SHR MACRO reghi, reglo, num, ovr
LOCAL SHR_loop
.errb <reghi>
.errb <reglo>
.errb <num>
.errnb <ovr>

if (.TYPE reghi) AND TYPE_REG
   if (.TYPE reglo) AND TYPE_REG
      if num EQ 0
         .err "M32_SHR: shift size must be 1-31: &num"
        exitm
      elseif num GT 31
         .err "M32_SHR: shift size must be 1-31: &num"
        exitm
      else
         ifidni <reghi>, <cx>
            .err "M32_SHR: cx is not a valid register!"
           exitm
         endif
         ifidni <reglo>, <cx>
            .err "M32_SHR: cx is not a valid register!"
           exitm
         endif

         push  cx
         mov   cx, num

       SHR_loop:
         shr   reghi, 1
         rcr   reglo, 1
         loop  SHR_loop
         pop   cx
      endif
   else
      .err "M32_SHR: bad parameter: &reglo"
      exitm
   endif
else
   .err "M32_SHR: bad parameter: &reghi"
   exitm
endif
ENDM

;***********************************************************************
M32_SAL MACRO reghi, reglo, num, ovr
LOCAL SAL_loop
.errb <reghi>
.errb <reglo>
.errb <num>
.errnb <ovr>

if (.TYPE reghi) AND TYPE_REG
   if (.TYPE reglo) AND TYPE_REG
      if num EQ 0
         .err "M32_SAL: shift size must be 1-31: &num"
        exitm
      elseif num GT 31
         .err "M32_SAL: shift size must be 1-31: &num"
        exitm
      else
         ifidni <reghi>, <cx>
            .err "M32_SAL: cx is not a valid register!"
           exitm
         endif
         ifidni <reglo>, <cx>
            .err "M32_SAL: cx is not a valid register!"
           exitm
         endif

         push  cx
         mov   cx, num

       SAL_loop:
         shl   reglo, 1
         rcl   reghi, 1
         loop  SAL_loop
         pop   cx
      endif
   else
      .err "M32_SAL: bad parameter: &reglo"
      exitm
   endif
else
   .err "M32_SAL: bad parameter: &reghi"
   exitm
endif
ENDM

;***********************************************************************
M32_SAR MACRO reghi, reglo, num, ovr
LOCAL SAL_loop
.errb <reghi>
.errb <reglo>
.errb <num>
.errnb <ovr>

if (.TYPE reghi) AND TYPE_REG
   if (.TYPE reglo) AND TYPE_REG
      if num EQ 0
         .err "M32_SAR: shift size must be 1-31: &num"
        exitm
      elseif num GT 31
         .err "M32_SAR: shift size must be 1-31: &num"
        exitm
      else
         ifidni <reghi>, <cx>
            .err "M32_SAR: cx is not a valid register!"
           exitm
         endif
         ifidni <reglo>, <cx>
            .err "M32_SAR: cx is not a valid register!"
           exitm
         endif

         push  cx
         mov   cx, num

       SAL_loop:
         sar   reghi, 1
         rcr   reglo, 1
         loop  SAL_loop
         pop   cx
      endif
   else
      .err "M32_SAR: bad parameter: &reglo"
      exitm
   endif
else
   .err "M32_SAR: bad parameter: &reghi"
   exitm
endif
ENDM

;***********************************************************************
M32_ROL MACRO reghi, reglo, num, ovr
LOCAL ROL_loop, ROL_endloop
.errb <reghi>
.errb <reglo>
.errb <num>
.errnb <ovr>

if (.TYPE reghi) AND TYPE_REG
   if (.TYPE reglo) AND TYPE_REG
      if num EQ 0
         .err "M32_ROL: shift size must be 1-31: &num"
        exitm
      elseif num GT 31
         .err "M32_ROL: shift size must be 1-31: &num"
        exitm
      else
         ifidni <reghi>, <cx>
            .err "M32_ROL: cx is not a valid register!"
           exitm
         endif
         ifidni <reglo>, <cx>
            .err "M32_ROL: cx is not a valid register!"
           exitm
         endif

         push  cx
         mov   cx, num

       ROL_loop:
         shl   reglo, 1
         rcl   reghi, 1
         jnc   ROL_endloop
         or    reglo, 0001h
       ROL_endloop:
         loop  ROL_loop
         pop   cx
      endif
   else
      .err "M32_ROL: bad parameter: &reglo"
      exitm
   endif
else
   .err "M32_ROL: bad parameter: &reghi"
   exitm
endif
ENDM

;***********************************************************************
M32_ROR MACRO reghi, reglo, num, ovr
LOCAL ROR_loop, ROR_endloop
.errb <reghi>
.errb <reglo>
.errb <num>
.errnb <ovr>

if (.TYPE reghi) AND TYPE_REG
   if (.TYPE reglo) AND TYPE_REG
      if num EQ 0
         .err "M32_ROR: shift size must be 1-31: &num"
        exitm
      elseif num GT 31
         .err "M32_ROR: shift size must be 1-31: &num"
        exitm
      else
         ifidni <reghi>, <cx>
            .err "M32_ROR: cx is not a valid register!"
           exitm
         endif
         ifidni <reglo>, <cx>
            .err "M32_ROR: cx is not a valid register!"
           exitm
         endif

         push  cx
         mov   cx, num

       ROR_loop:
         shr   reghi, 1
         rcr   reglo, 1
         jnc   ROR_endloop
         or    reghi, 8000h
       ROR_endloop:
         loop  ROR_loop
         pop   cx
      endif
   else
      .err "M32_ROR: bad parameter: &reglo"
      exitm
   endif
else
   .err "M32_ROR: bad parameter: &reghi"
   exitm
endif
ENDM


;***********************************************************************
M32_RCL MACRO reghi, reglo, num, ovr
LOCAL RCL_loop, RCL_endloop
.errb <reghi>
.errb <reglo>
.errb <num>
.errnb <ovr>

if (.TYPE reghi) AND TYPE_REG
   if (.TYPE reglo) AND TYPE_REG
      if num EQ 0
         .err "M32_RCL: shift size must be 1-31: &num"
        exitm
      elseif num GT 31
         .err "M32_RCL: shift size must be 1-31: &num"
        exitm
      else
         ifidni <reghi>, <cx>
            .err "M32_RCL: cx is not a valid register!"
           exitm
         endif
         ifidni <reglo>, <cx>
            .err "M32_RCL: cx is not a valid register!"
           exitm
         endif

         ; save original cx and flags
         push  cx
         pushf

         ; get number of bits to shift
         mov   cx, num

       RCL_loop:
         ; save loop counter
         push  cx

         ; save flags before shift and rotate
         pushf
         shl   reglo, 1
         rcl   reghi, 1

         ; save this carry flag as well
         pushf
         pop   cx

         ; restore flags prior to shift/rotate
         popf

         ; check the carry flag
         jnc   RCL_endloop
         or    reglo, 0001h
       RCL_endloop:
         ; restore rotated/shifted flags
         push  cx
         popf

         ; restore loop counter
         pop   cx

         loop  RCL_loop

         ; restore original flags and cx
         popf
         pop   cx
      endif
   else
      .err "M32_RCL: bad parameter: &reglo"
      exitm
   endif
else
   .err "M32_RCL: bad parameter: &reghi"
   exitm
endif
ENDM

;***********************************************************************
M32_RCR MACRO reghi, reglo, num, ovr
LOCAL RCR_loop, RCR_endloop
.errb <reghi>
.errb <reglo>
.errb <num>
.errnb <ovr>

if (.TYPE reghi) AND TYPE_REG
   if (.TYPE reglo) AND TYPE_REG
      if num EQ 0
         .err "M32_RCR: shift size must be 1-31: &num"
        exitm
      elseif num GT 31
         .err "M32_RCR: shift size must be 1-31: &num"
        exitm
      else
         ifidni <reghi>, <cx>
            .err "M32_RCR: cx is not a valid register!"
           exitm
         endif
         ifidni <reglo>, <cx>
            .err "M32_RCR: cx is not a valid register!"
           exitm
         endif

         ; save original cx and flags
         push  cx
         pushf

         ; get number of bits to shift
         mov   cx, num

       RCR_loop:
         ; save loop counter
         push  cx

         ; save flags before shift and rotate
         pushf
         shr   reghi, 1
         rcr   reglo, 1

         ; save this carry flag as well
         pushf
         pop   cx

         ; restore flags prior to shift/rotate
         popf

         ; check the carry flag
         jnc   RCR_endloop
         or    reghi, 8000h
       RCR_endloop:
         ; restore rotated/shifted flags
         push  cx
         popf

         ; restore loop counter
         pop   cx

         loop  RCR_loop

         ; restore original flags and cx
         popf
         pop   cx
      endif
   else
      .err "M32_RCR: bad parameter: &reglo"
      exitm
   endif
else
   .err "M32_RCR: bad parameter: &reghi"
   exitm
endif
ENDM


endif   ;; USE_186



