;  $Id: DEVCLK.ASM 1.8 2003/03/20 17:03:32Z thomd Exp $
;
;  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;  Copyright (c) 1989-2003 Datalight, Inc.
;  All Rights Reserved
;
;  Datalight, Inc. is a Washington corporation located at:
;
;  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
;  Bothell, WA  98021                    Fax: +1.425.951.8094
;  USA                                   Http://www.datalight.com
;
;  This software, including without limitation all source code and
;  documentation, is the confidential, trade secret property of
;  Datalight, Inc., and is protected under the copyright laws of
;  the United States and other jurisdictions.  Portions of the
;  software may also be subject to one or more the following US
;  patents: US#5860082, US#6260156.
;
;  In addition to civil penalties for infringement of copyright
;  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;  penalties for violation of (a) the restrictions on circumvention
;  of copyright protection systems found in 17 U.S.C. 1201 and
;  (b) the protections for the integrity of copyright management
;  information found in 17 U.S.C. 1202.
;
;  U.S. Government Restricted Rights.  Use, duplication,
;  reproduction, or transfer of this commercial product and
;  accompanying documentation is restricted in accordance with
;  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
;  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
;  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;  THIS SOFTWARE IN ERROR.
;
COMMENT ~
----------------------------------------------------------

   DEVCLK.ASM  created 04/10/89  by Drew Gislason

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This file contains the internal (built-in) clock device
   driver for ROM-DOS.  It is a character device with the
   name of "CLOCK$"

   This CLOCK device relies on BIOS interrupt 1AH for all
   date/time functions.  It uses functions 02H-05H (use
   the Real Time Clock) if possible and will fall back on
   functions 00H-01H if the RTC is not available.

---------------------------------------------------------- ~
include macros.inc
include devhdr.inc
include profile.inc

PUBLIC	_clkx


;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;
;	BIOS RESOURCES USED
;
;	This area defines all the BIOS calls that this
;	device driver makes.

; BIOS interrupts used
BIOSCLK	EQU	1AH	; BIOS "time" interrupt

; BIOS 1AH functions used
GETTICK	EQU	0	; no Real Time Clock required
SETTICK	EQU	1	; no RTC required
GETTIME	EQU	2	; uses RTC
SETTIME	EQU	3	; uses RTC
GETDATE	EQU	4	; uses RTC
SETDATE	EQU	5	; uses RTC


cBegCode
	EXTRN	dev_serv:near
	EXTRN	dev_cmd_err:near, dev_err_exit:near
	EXTRN	dev_busy_exit:near, dev_err_cnt:near
	EXTRN	dev_exit:near

cmdtbl:
	dw	clk_init	; 0 - init
	dw	dev_cmd_err	; 1
	dw	dev_cmd_err	; 2
	dw	dev_cmd_err	; 3
	dw	clk_read	; 4 - read
	dw	dev_cmd_err	; 5
	dw	dev_cmd_err	; 6
	dw	dev_cmd_err	; 7
	dw	clk_write	; 8 - write
cmdend:

cEndCode


_DEVDATA segment byte public 'DEVDATA'
	ASSUME CS:_DEVDATA,DS:_DEVDATA
	EXTRN	dev_fstrat:NEAR, cmd_ptr:DWORD


;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;
;		DEVICE HEADER
;
;  This is the actual device header.  It must be in RAM
;  so the pointer to the next device driver can change.
;
_clkx:
	; Device Attributes (8008H)
	; 1xxx xxxx xxxx xxxx	Character device
	; xxxx xxxx xxxx 1xxx	Clock device
	dw 	-1,-1
	dw	8008H
	dw	offset _DEVDATA:dev_fstrat
	dw	offset clk_serv
	db	"CLOCK$  ",0

	public	clk_serv
clk_serv:
	push	AX
	push	SI

	; calc # of commands and where to find cmdtbl
	mov	AH, (OFFSET cmdend - OFFSET cmdtbl)/2
	mov	SI, OFFSET _TEXT:cmdtbl

	; jump to the REAL service routine in ROM (CODE SEG)
	jmp	DWORD PTR CS:[cmd_ptr]

_DEVDATA ENDS


cBegUdata
;	EXTRN	_day_count:word
	EXTRN	_clk_use_rtc:byte

	extrn	daylight:byte	; is it daylight savings?

	_day_count	DW	?
cEndUdata


cBegCode
	c_extrn	bcd2days,near, bcd2time,near
	c_extrn time2bcd,near, days2bcd,near



;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;   CLOCK INIT (function 00H)
;
;   Initialize the clock driver (set tick count)
;
	public	clk_init
clk_init:

	ENTER_FUNCTION

	; set the BIOS tick count based on the RTC
	call	get_rtc_time	; CX:DX = "normal" time
	jc	init_no_rtc
	call	time2ticks	; convert into ticks (CX:DX -> CX:DX)
	mov	AH, SETTICK
	int	BIOSCLK

	; set DATE based on RTC
	call	get_rtc_date

	; we're done
init_no_rtc:

	EXIT_FUNCTION

	jmp	dev_exit


;---------------------------------------------------------
;  Get ticks from BIOS, rollover date if needed.
;
	public	clk_get_ticks
clk_get_ticks:

	ENTER_FUNCTION

	mov	AH, GETTICK	; fn 0
	int	BIOSCLK		; int 1Ah

	; update daycount by roll-over value
	xor	AH, AH
	add	[_day_count], AX
	or	ax, ax
	jz	ticks_chk_range

	; Check for the year 2000. Many RTCs/BIOSs do not increment
	; the century count at midnight on 12/31/99. Once the RTC
	; century is set to 20h, however, everything is OK.
	;
	; (2000-1980)*365 + 5 = 7305
	;
	cmp	[_day_count], 7305d
	jne	ticks_chk_range

	push	cx
	push	dx

	mov	ah, SETDATE
	mov	cx, 2000h
	mov	dx, 101h
	int	BIOSCLK

	pop	dx
	pop	cx

	; make sure tick value is in a reasonable range
    ticks_chk_range:
	cmp	CX, 18H
	jb	ticks_ok
	ja	ticks_bad
	cmp	DX, 0B2H	; 1800B2H is highest tick value
	jbe	ticks_ok

    ticks_bad:
	xor	CX, CX		; bad ticks! bad, bad ticks...
	xor	DX, DX		; set to midnight!
	mov	AH, SETTICK
	int	BIOSCLK

    ticks_ok:

	EXIT_FUNCTION

	ret


;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;   CLOCK READ (function 04H)
;
;   Read date and time from RTC or BIOS
;
;   On Entry
;	DS    = DGROUP
;	ES:DI = transfer area
;
;   On Exit
;	00-01H 	day_count since 1-1-1980
;	02H	minutes  (0-59)
;	03H	hours    (0-23)
;	04H	hundreds (0-99)
;	05H	seconds  (0-59)
;
	PUBLIC	clk_read
clk_read:

	ENTER_FUNCTION

	; use the RTC or the clock ticks?
	cmp	[_clk_use_rtc], 0
	jz	no_rtc_time

	; save "normal" time on the stack (hrs/min/sec/hun)
	call	get_rtc_time
	jc	no_rtc_time
	push	CX
	push	DX
	jmp	short chk_date

	; read and save the BIOS tick count
	; BIOS should update rollover (AL) once for each day
	; that has passed since gettick was last called.
no_rtc_time:
	call	clk_get_ticks

	; convert tick count (CX:DX) into time (AX:BX, hrs/min/sec/hun)
	call	ticks2time
	push	AX
	push	BX	; save "normal" time on stack

	; use the RTC for date or the DOS date?
chk_date:
	cmp	[_clk_use_rtc], 0
	jz	no_rtc_date
	call	get_rtc_date

	; assume date has already been updated by clk_get_ticks
no_rtc_date:

	; store date in transfer buffer (ES:DI)
got_date:
	mov	AX, [_day_count]
	stosw

	; store time in transfer buffer (ES:DI)
	pop	DX	; sec | hun
	pop	AX	; hrs | min
	stosw
	xchg	DX, AX
	stosw

	EXIT_FUNCTION

	jmp	dev_exit


;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;   CLOCK WRITE (function 08H)
;
;   Write date and time to BIOS
;
;   On Entry
;	DS    = DGROUP
;	ES:DI = transfer area
;	CX    = count
;
	PUBLIC	clk_write
clk_write:

	ENTER_FUNCTION

	; save ES to restore it later
	mov	SI, ES

	; store the day count in the device driver
	mov	AX, ES:[DI]
	mov	[_day_count], AX

	; tell the Real Time Clock about the date
	push	AX
	call	days2bcd	; returns BCD date in DX:AX
	pop	CX
	mov	CX, AX		; CX = year in BCD (ie 1992H)
	mov	AH, SETDATE	; DX = mn | dy in BCD (ie 0831H - aug 31st)
	int	BIOSCLK

	; tell the Real Time Clock about the time
	mov	ES, SI
	push	ES:[DI+2]	; hour| min (not in BCD yet)
	push	ES:[DI+4]	; sec | hun (not in BCD yet)
	call	time2bcd	; returns values in DX:AX
	add	SP, 4
	mov	CX, AX		; CX = hr | mn (in BCD, ie 1504H = 3:04pm)
	mov	DL, [daylight]	; DH = sec
	mov	AH, SETTIME
	int	BIOSCLK

	; convert values into ticks
	mov	ES, SI
	mov	CX, ES:[DI+2]
	mov	DX, ES:[DI+4]
	call	time2ticks

	; set the BIOS tick count as well
	mov	AH, SETTICK
	int	BIOSCLK

	EXIT_FUNCTION

	jmp	dev_exit


;----------------------------------------------------------
;  Get Real Time Clock date into AX
;  Returns carry clear, AX = day count
;  - or -
;  Carry set on error
;
get_rtc_date:

	ENTER_FUNCTION

	; attempt to get the DATE from Real Time Clock
	mov	AH, GETDATE
	xor	CX, CX
	clc
	int	BIOSCLK
	jc	rtc_err

	; Verify that the DATE is reasonable
	; Check for the year 2000. Many RTCs/BIOSs do not increment
	; the century count at midnight on 12/31/99. Once the RTC
	; century is set to 20h, however, everything is OK.
	;
	cmp	CX, 1900h	; bad century, no RTC
	je	fix_rtc_date	; century returned must be either
	jb	rtc_err
	cmp	CH, 20H		; 19 or 20 (bcd)
	ja	rtc_err

	; convert bcd DATE to day count
	;
    yes_rtc_date:
	push	CX
	push	DX
	call	bcd2days	; AX = day count
	add	SP, 4

	; save the day count
	mov	[_day_count], AX
	clc

	EXIT_FUNCTION

	ret

    fix_rtc_date:
	mov	ch, 20h
	mov	ah, SETDATE
	int	BIOSCLK
	mov	AH, GETDATE
	int	BIOSCLK

	cmp	CH, 20H		; make sure the date took
	je	yes_rtc_date

	EXIT_FUNCTION

	jmp	rtc_err


;----------------------------------------------------------
;  Get Real Time Clock time into CX:DX
;  Returns carry clear
;	CX = hour << 8 | mins (binary, not BCD, ie 0F04H = 3:04pm)
;	DX = secs << 8 | huns (binary, not BCD, ie 2302H = 3:04:35.02pm)
;  - or -
;  Carry set on error
;
get_rtc_time:

	ENTER_FUNCTION

	; attempt to get the TIME from the Real Time Clock
	mov	AH, GETTIME
	mov	CX, 0FFFFH
	clc
	int	BIOSCLK
	jc	rtc_err		; carry set, no RTC

	; verify that the time is reasonable
	cmp	CH, 23H		; hours 0-23 (bcd)
	ja	rtc_err
	cmp	CL, 59H		; mins  0-59 (bcd)
	ja	rtc_err

	; save daylight savings flag
	mov	[daylight], DL
	xor	DL, DL		; no hundredths for RTC

	; convert BCD time into "normal" binary time
	push	DX
	push	CX
	call	bcd2time	; Returns	DX = sec | hun
	add	SP, 4		;		AX = hrs | min

	; return CX:DX hrs | min || sec | hun
	mov	CX, AX
	clc

	EXIT_FUNCTION

	ret

rtc_err:
	stc

	EXIT_FUNCTION

	ret


comment " ticks2time{}; "
;----------------------------------------------------------
;   Convert BIOS clock ticks into "normal" time
;
;   CX:DX = ticks
;
;   Returns
;      AX = hour << 8 | mins (not BCD)
;      BX = secs << 8 | huns (not BCD)
;----------------------------------------------------------
	public	ticks2time
ticks2time	proc near

	ENTER_FUNCTION

	mov	AX, CX
	mov	BX, DX		; AX:BX = ticks
	shl	DX, 1
	rcl	CX, 1
	shl	DX, 1
	rcl	CX, 1		; multiply ticks by 4
	add	DX, BX
	adc	AX, CX   	; 5*ticks
	xchg	DX, AX
	mov	CX, 0E90BH	; divided by ticks per
	div	CX
	mov	BX, AX
	xor	AX, AX
	div	CX
	mov	DX, BX
	mov	CX, 00C8H
	div	CX
	cmp	DL, 64H
	jb	l1
	sub	DL, 64H
l1:	cmc
	mov	BL, DL
	rcl	AX, 1
	mov	DL, 0
	rcl	DX, 1
	mov	CX, 003CH
	div	CX
	mov	BH, DL
	div	CL
	xchg	AL, AH

	EXIT_FUNCTION

	ret
ticks2time	endp ;

COMMENT " time2ticks() {}; "
;
;----------------------------------------------------------
;   Convert "normal" time into BIOS ticks
;
;   On Entry
;	CX = hour << 8 | mins (binary, not BCD, ie 0F04H = 3:04pm)
;	DX = secs << 8 | huns (binary, not BCD, ie 2302H = 3:04:35.02pm)
;
;   Returns
;      ticks CX:DX
;---------------------------------------------------------- ~
	public	time2ticks
time2ticks	proc near

	ENTER_FUNCTION

	mov	AL, 60		; 60 minutes an hour
	mul	CH		; AX = hours expressed in minutes
	mov	CH, 0
	add	AX, CX		; AX = total minutes
	mov	CX, 6000
	mov	BX, DX		; save seconds and hundredths
	mul	CX		; DX:AX = minutes expressed in hundredths
	mov	CX, AX		; save low part
	mov	AL, 100
	mul	BH		; express seconds in hundredths
	add	CX, AX
	adc	DX, 0		; DX:CX = hrs,mins,secs in hundredths
	mov	BH, 0
	add	CX, BX
	adc	DX, 0		; DX:CX = total hundredths today
	xchg	DX, AX
	xchg	CX, AX		; CX:AX = total hundredths today
	mov	BX, 0E90BH
	mul	BX
	xchg	DX, CX
	xchg	DX, AX
	mul	BX
	add	AX, CX
	adc	DX, 0
	xchg	DX, AX
	mov	BX, 5
	div	BL
	mov	CL, AL
	mov	CH, 0
	mov	AL, AH
	cbw
	xchg	DX, AX
	div	BX
	mov	DX, AX		; DX:AX = ticks

	EXIT_FUNCTION

	ret
time2ticks	endp

cEndCode

	END
