;  $Id: DEVROM.ASM 1.8 2003/03/20 17:06:08Z thomd Exp $
;
;  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;  Copyright (c) 1988-2003 Datalight, Inc.
;  All Rights Reserved
;
;  Datalight, Inc. is a Washington corporation located at:
;
;  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
;  Bothell, WA  98021                    Fax: +1.425.951.8094
;  USA                                   Http://www.datalight.com
;
;  This software, including without limitation all source code and
;  documentation, is the confidential, trade secret property of
;  Datalight, Inc., and is protected under the copyright laws of
;  the United States and other jurisdictions.  Portions of the
;  software may also be subject to one or more the following US
;  patents: US#5860082, US#6260156.
;
;  In addition to civil penalties for infringement of copyright
;  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;  penalties for violation of (a) the restrictions on circumvention
;  of copyright protection systems found in 17 U.S.C. 1201 and
;  (b) the protections for the integrity of copyright management
;  information found in 17 U.S.C. 1202.
;
;  U.S. Government Restricted Rights.  Use, duplication,
;  reproduction, or transfer of this commercial product and
;  accompanying documentation is restricted in accordance with
;  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
;  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
;  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;  THIS SOFTWARE IN ERROR.
;
COMMENT ~
----------------------------------------------------------

   DEVROM.ASM  created 04/08/88  by Drew Gislason

   Copyright (c) 1988-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This disk device allows ROM-DOS to boot completely
   from ROM.  It is a ROM-DISK, which is effectively a
   read-only RAM disk.

---------------------------------------------------------- ~
include macros.inc
include dostruct.inc
include devhdr.inc
include profile.inc

MAX_ROM_DISKS	= 2

PUBLIC	_romx

cBegData
	extrn	_romdisk:word
cEndData


;----------------------------------------------------------------------
;
;		DEVICE HEADER
;
cBegCode
	EXTRN	dev_serv:NEAR

cmdtbl:
	dw	rom_init	; 0 - init
	dw	rom_media	; 1 - media check
	dw	rom_bpb		; 2 - build a bpb
	dw	dev_cmd_err	; 3 - ioctl input (has none)
	dw	rom_read	; 4 - device read
	dw	dev_cmd_err	; 5 - non-destructive read (none)
	dw	dev_cmd_err	; 6 - input status (none)
	dw	dev_cmd_err	; 7 - input flush (none)
	dw	rom_write	; 8 - write (always write protected)
	dw	rom_write	; 9 - write w/ verify (always write protected)
cmdend:

cEndCode

_DEVDATA segment byte public 'DEVDATA'

	ASSUME CS:_DEVDATA,DS:_DEVDATA
	EXTRN	dev_fstrat:NEAR, cmd_ptr:DWORD

_romx:
	dw	-1,-1
	dw	0800H		; block, non-removable
	dw	dev_fstrat	; strategy routine
	dw	rom_serv	; service routine
numdrives label byte
	dw	1		; 1 drive
	dw	0
	dw	0
	dw	0

	public	rom_serv
rom_serv:
	push	AX
	push	SI

	; calc # of commands and where to find cmdtbl
	mov	AH,(OFFSET cmdend - OFFSET cmdtbl)/2
	mov	SI,OFFSET cmdtbl

	; Jump to the REAL service routine in ROM (CODE SEG)
	jmp	DWORD PTR CS:cmd_ptr
_DEVDATA ENDS


;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;
;	DATA SEGMENT
;
cBegUdata
	; how many ROM-DISKs were found?
	disks_found	dw ?

	; start of each disk
	DRVSTART	dw MAX_ROM_DISKS dup (?)

	; BPBs for each disk
	DRVBPB		BiosParameterBlock MAX_ROM_DISKS dup (<?>)

	; ptrs to each BPB
	initptr		dw MAX_ROM_DISKS dup (?)
cEndUdata

cBegData
	EXTRN	_ptrsav:DWORD
cEndData

cBegCode
	EXTRN	dev_serv:near
	EXTRN	dev_cmd_err:near, dev_err_exit:near
	EXTRN	dev_busy_exit:near, dev_err_cnt:near
	EXTRN	dev_exit:near


;----------------------------------------------------------------------
;
;	MEDIA CHECK (function 01H)
;	Has disk been changed?
;
;   On Entry
;	DS	= DGROUP
;
rom_media:

	ENTER_FUNCTION

	lds	BX, _ptrsav
	mov	byte ptr [BX+TRANS], 1	; romdisk never changes

	EXIT_FUNCTION

	jmp	dev_exit

;----------------------------------------------------------------------
;
;	BUILD A BPB (function 02H)
;
;   On Entry
;	AL	= unit #
;	ES:DI	= scratch buffer
;	DS	= DGROUP
;
rom_bpb:
	ENTER_FUNCTION

	; get pointer to this BPB (DX:SI)
	call	get_bpb
	mov	DX, DS

	; tell request header about it
	lds	BX, _ptrsav		   ; DS:BX ptr to i/o packet
	mov	WORD PTR [BX+COUNT], SI
	mov	WORD PTR [BX+COUNT+2], DX  ; DX:SI = bpb

	EXIT_FUNCTION

	jmp	dev_exit

;   Return the BPB for this unit (in SI)
;   On Entry
;	AL	= unit #
;   On Exit
;	SI	= offset of BPB
;
	PUBLIC get_bpb
get_bpb:

	ENTER_FUNCTION

	mov	SI, GOFFSET DRVBPB
	mov	AH, SIZE BiosParameterBlock
	mul	AH
	add	SI, AX

	EXIT_FUNCTION

	ret

;----------------------------------------------------------------------
;
;	READ FUNCTION (function 04H)
;
; note: DOS will never ask for more than 64K worth of sectors
;       at a time.
; warning: this routine does not check for read past end of disk.
;
; Upon Entry:
;	AL	= unit #
;	DS	= ROM-DOS data segment
;	ES:DI	= buffer to place data
;	CX	= # of sectors to read
;	DX	= starting sector
;
	public rom_read
rom_read:

	ENTER_FUNCTION

	; determine start of this ROM-DISK
	mov	BL, AL
	xor	BH, BH
	shl	BX, 1
	add	BX, GOFFSET DRVSTART
	mov	BX, WORD PTR [BX]
	push	BX

	; convert unit # (AL) into ptr to BPB (SI)
	call	get_BPB

	; determine starting paragraph to read within this ROM-DISK
	push	CX			; save # of sectors
	mov	AX, WORD PTR [SI.BPB_sec_size]  ; sector size
	mov	CL, 4
	shr	AX, CL			; paragraphs/sector
	mul	DX			; start * para_per_sec
	mov	BX, AX			; paragraphs from start (in BX)
	pop	CX

	; determine # of bytes to read
	mov	AX, WORD PTR [SI.BPB_sec_size]  ; sector size
	mul	CX			; count * sec_size
	mov	CX, AX			; # bytes (in CX)

	; determine start of this ROM-DISK
	pop	AX			; start of disk
	add	AX, BX			; add starting offset
	mov	DS, AX
	xor	SI, SI			; DS:SI = romdisk address

;	ES:DI	- destination buffer
;	DS:SI	- rom disk sector
;	CX	- count in bytes

	cld
	rep	MOVSB

	EXIT_FUNCTION

	jmp	dev_exit


;----------------------------------------------------------------------
;
;		WRITE FUNCTION (always fails)
;
rom_write:

	ENTER_FUNCTION

	mov	AL, 0		; write protect violation

	EXIT_FUNCTION

	jmp	dev_err_cnt


;----------------------------------------------------------------------
;
;	INIT FUNCTION (function 00H)
;
; On Exit:
;	+-------------------------+
;	| 13-Byte header          |
;	+-------------------------+
;	| BYTE   Number of units  |
;	+-------------------------+
;	| DWORD  End Address      |	* ignored on built-in devices
;	+-------------------------+
;	| DWORD  Ptr to BPB array |
;	+-------------------------+
;

; signature checks for a valid ROM-DISK
rd_label DW	3
	 DB	7,"DLRDISK"
rd_jmp   DW	0
	 DB	3,0EBH,28H,90H

;  Checks ES:DI for a string in the above form
;  SI destroyed
	PUBLIC rd_check
rd_check:

	ENTER_FUNCTION

	push	DS
	push	DI

	; set DS = CS
	push	CS
	pop	DS

	; DS:SI = signature, ES:DI = sig to check
	mov	DI, [BX]
	xor	CH, CH
	mov	CL, [BX+2]
	add	BX, 3
	mov	SI, BX

	cld
	repz	CMPSB		; ZF if worked, NZ if not

	pop	DI
	pop	DS

	EXIT_FUNCTION

	ret

	public rom_init
rom_init:

	ENTER_FUNCTION

	; set ptrs to each BPB
	mov	CX, MAX_ROM_DISKS
	mov	AX, GOFFSET DRVBPB
	mov	BX, GOFFSET initptr
set_bpb_ptr:
	mov	[BX], AX
	inc	BX
	inc	BX
	add	AX, SIZE BiosParameterBlock
	loop	set_bpb_ptr

	; check memory for a valid ROMDISK in memory
	mov	DI, [_romdisk]	; start checking at this segment

	; get as many ROM-DISKs as user requests
look_again:
	cmp	byte ptr disks_found, MAX_ROM_DISKS
	jae	look_done

RI0:	mov	ES, DI
	mov	BX, OFFSET rd_label	; check for ROM DISK signature
	call	rd_check
	jnz	RI1

	mov	BX, OFFSET rd_jmp	; check for the JMP instruction
	call	rd_check
	jz	RI2

	; not found, look again
RI1:	inc	DI
	cmp	DI, 0FFFFH
	jz	look_done
	jmp	RI0

	; found! save paragraph of this ROM-DISK
RI2:	mov	BX, disks_found
	shl	BX, 1
	add	BX, GOFFSET DRVSTART
	mov	[BX], ES	; ES:0 = romdisk start

	; ES now pts to ROM-DISK, DS = DGROUP
	; move ROM BPB into RAM structure
	mov	AH, byte ptr disks_found
	mov	AL, SIZE BiosParameterBlock
	mul	AH
	add	AX, GOFFSET DRVBPB	; AX = offset to BPB in DGROUP
	push	ES			; swap ES and DS
	push	DS
	pop	ES
	pop	DS
	mov	SI, 11			; DS:SI = BPB in ROMDISK
	mov	DI, AX			; ES:DI = RAM BPB
	mov	CX, SIZE BiosParameterBlock ; copy all bytes of BPB
	cld
	rep	MOVSB
	mov	DI, DS			; remember search segment
	push	ES
	pop	DS			; restore DGROUP

	; check for another disk
	inc	DI
	inc	disks_found
	jmp	look_again

	; we've reached our limit (either top of mem or max disks)
look_done:
	mov	AL, byte ptr disks_found
	or	AL, AL
	jz	init_err	; no units, error!

	; set up the i/o packet
	lds	BX, _ptrsav
	mov	BYTE PTR [BX+MEDIA], AL		; # of units (ROM-DISKs)
	mov	WORD PTR [BX+COUNT], GOFFSET initptr
	mov	WORD PTR [BX+COUNT+2], DS	; point to BPB list

	EXIT_FUNCTION

	jmp	dev_exit

init_err:
	mov	AL, 0CH		; general failure error

	EXIT_FUNCTION

	jmp	dev_err_exit

cEndCode


	END
