//
//   $Id: DOSCNTRY.C 1.8 2002/10/15 21:44:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSCNTRY.C  created 01/15/91  by Drew Gislason

   Copyright (c) 1991-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file includes the DOS country handling routines.

**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "dosver.h"
#include "profile.h"

#define __FILENUM__ 3

/* include the country data here */
#define ROMDOS	1
#include "dosmsg.c"

unsigned romdos_rev = VERSIONVAL;		/* ROM-DOS internal revision */
char revision_str[] = REVISION;			/* ROM-DOS internal revision string */
unsigned dos5_flags = 0;		/* HI byte, bit mask xxxH Rxxx (H-HMA, R-ROM) */


/*
	The default country information, USA standard.
*/
extern struct COUNTRY_INFO country_info;

#if 0
//
//  THIS DATA IS NOW IN ASSEMBLY IN DOSASM.ASM.
//  THE DATA WAS MOVED TO PROVIDE A MORE COMPATIBLE DOS
//  DATA AREA.
//
struct COUNTRY_INFO country_info =
{
	1,
	38,
	COUNTRY_ID,
	CODE_PAGE,
	DATE_TIME_FMT,
	CURRENCY,
	THOUSANDS,
	DECIMAL,
	DATE_SEPARATOR,
	TIME_SEPARATOR,
	CURRENCY_FORMAT,
	DEC_DIGITS,
	TIME_FORMAT,
	(void far *)0,
	DATA_LIST
};

/* Filename Uppercase Table */
uchar patch_uppercase[0x24] =   /* patch starts at offset 1 */
{
         0x9A, 0x45, 0x41, 0x8E, 0x41, 0x8F, 0x80,
   0x45, 0x45, 0x45, 0x49, 0x49, 0x49, 0x8E, 0x8F,
   0x90, 0x92, 0x92, 0x4F, 0x99, 0x4F, 0x55, 0x55,
   0x59, 0x99, 0x9A, 0x9B, 0x9C, 0x9D, 0x9E, 0x9F,
   0x41, 0x49, 0x4F, 0x55, 0xA5
};

/* Country Collating Table */
uchar patch_6506[0x30] =   /* patch starts at offset 0x80 */
{
   0x43, 0x55, 0x45, 0x41, 0x41, 0x41, 0x41, 0x43,
   0x45, 0x45, 0x45, 0x49, 0x49, 0x49, 0x41, 0x41,
   0x45, 0x41, 0x41, 0x4F, 0x4F, 0x4F, 0x55, 0x55,
   0x59, 0x4F, 0x55, 0x24, 0x24, 0x24, 0x24, 0x24,
   0x41, 0x49, 0x4F, 0x55, 0x4E, 0x4E, 0xA6, 0xA7,
   0x3F, 0xA9, 0xAA, 0xAB, 0xAC, 0x21, 0x22, 0x22
};

#endif

/* Filename Terminators */
extern uchar term_table[];
unsigned char *terminators = term_table+11;


/*
	-------------------------------------------------------

					Country Information

	-------------------------------------------------------
*/

#if 0
//
//  THIS CODE IS NOW IN ASSEMBLY IN DOSINIT.ASM.
//  THE CODE WAS MOVED TO ALLOW IT TO BE THROWN AWAY AFTER
//  BOOTING ROM-DOS.
//
/*
	Called once on startup
*/
void	init_country_info(void)
{
	/*
		[ externals ]
	*/
	extern struct COUNTRY_INFO country_info;
	extern uchar uppercase_table[];
	extern uchar collating_table[];
	extern uchar patch_uppercase[0x24];
	extern uchar patch_6506[0x30];

	int i;

	/* this code is simply to reduce initialized DATA size */
	*(int *)uppercase_table = 0x80;
	for(i=0x80;i<=0xFF;++i)
		uppercase_table[i-0x80+2] = i;
	memmove(uppercase_table+1+2, patch_uppercase, sizeof(patch_uppercase));

	/* this code is simply to reduce initialized DATA size */
	*(int *)collating_table = 0x100;
	for(i=0;i<=0xFF;++i)
		collating_table[i+2] = (i<0x61 || i>0x7A ? i : i-0x20);
	memmove(collating_table+0x80+2, patch_6506,sizeof(patch_6506));
	collating_table[0xE1+2] = 0x53;
}
#endif


/*
	DOS call 38H

	Get/Set Country Information.  Currently we only allow
	the getting of information.  Maybe, when we support
	countries better, we can recode this.
*/
void	get_country_info(void)
{
	extern struct COUNTRY_INFO country_info;
	extern int error_no;
	int cntry;

	ENTER_FUNCTION();

	/* get country information */
	if(R_DX != 0xFFFF)
	{
		cntry = R_AL;
		if(cntry == 0xFF)
			cntry = R_BX;
		if(!cntry || cntry==country_info.country_code)
		{
			fmemcpy(MK_FP(R_DS, R_DX), (uchar far *)(&country_info)+DATESTART,
				sizeof(struct COUNTRY_INFO)-DATESTART);
			R_AX = R_BX = country_info.country_code;

			EXIT_FUNCTION();

			return;
		}
	}

	/* set country information */
	error_no = ERR_BADFUNC;

	EXIT_FUNCTION();
}


/*
	DOS call 65xxH

	Not only does this call get Extended Country Information,
	but it performs a variety of duties like converting to
	UPPERCASe.
*/
void	get_ext_country_info(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern int int21_ax;
	extern uchar int21_al;
	extern uchar uppercase_table[];
	extern uchar collating_table[];
	extern uchar term_table[];
	extern uchar lead_byte_table[];
	extern struct COUNTRY_INFO country_info;
	extern char fatal_unsupp[];


	/*
		[ locals ]
	*/
	uchar far * userbuff;
	unsigned offset;
	int len;
	unsigned copysize;		/* for get country info */
	uchar c;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	userbuff = MK_FP(R_ES,R_DI);
	copysize = R_CX;
	c = R_DL;
	switch(int21_al)
	{
		case 0x01:
			if(copysize > 0x29)
				copysize = 0x29;
			fmemset(userbuff, 0, copysize);
			len = sizeof(struct COUNTRY_INFO);
			if(len > copysize)
				len = copysize;
			fmemcpy(userbuff, (void far *)(&country_info), len);
			R_CX = copysize;
			break;

		/* 6502,6504  Country Uppercase Table */
		/* 6505       Extended Filename Character Table */
		/* 6506       Country Collating Table */
		/* 6507       DBCS Lead Byte Table */
		case 0x02:
		case 0x04:
		case 0x05:
		case 0x06:
		case 0x07:
			copysize = 5;
			*userbuff = int21_al;
			switch(int21_al)
			{
				case 7: offset = (unsigned)lead_byte_table; break;
				case 6: offset = (unsigned)collating_table; break;
				case 5: offset = (unsigned)term_table; break;
				default: offset = (unsigned)uppercase_table; break;
			}
			*(unsigned far *)(userbuff+1) = offset;
			*(unsigned far *)(userbuff+3) = getDS();
			R_CX = copysize;
			break;

		/* convert character */
		case 0x20:
			if(int21_al == 0x20)
				R_AL = R_DL = toupper(c);
			break;

		/* convert string */
		case 0x21:
		case 0x22:
			/* convert the string using the uppercase table */
			userbuff = MK_FP(R_DS,R_DX);
			len = (int21_al == 0x21 ? R_CX : -1);
			R_AL = uppercase(userbuff, len);
			break;

		/* Yes(1), No(0) or BadValue(2) */
		case 0x23:
			len = 2;
			if(toupper(c)=='Y')
				len = 1;
			if(toupper(c)=='N')
				len = 0;
			R_AX = len;
			break;

		default:
			panic(fatal_unsupp, 0x21, int21_ax);
			break;
	}


	EXIT_FUNCTION();
}


/*
	DOS call DBxxH

	ROM-DOS provides a few internal calls specific to our
	DOS.  These calls provide access to internal variables
	and structures.

	There are 2 ways to make an internal ROM-DOS call.
	Datalight recommends the first approach.

	1) int 21H, AX=30DBH, SI=B2D2H, DI=function #
	   returns: CX=B2D2H, variable
	2) int 21H, AX=DBxxH  where xx=function #
		returns: variable

	Functions:
	02H Set DOS Version, CX = new version, returns nothing
	03H Get ROM-DOS Revision, returns AX = rev, ES:BX = rev string, CX=cookie,
		 DX=Compatibility version (of structures)
	04H Set Country Info, BX=cntry code, CX=code page, DS:DX=cntry info
	05H Get Config, BX=cfg_seg, CX=cfg_size
	06H Get Super-Boot, CH=order, CL=ID, DH=boot drive, DL=boot SB partition?

	Support some internal, ROM-DOS specific DOS calls.
*/
void	PASCAL debug_call(uchar func)
{
	extern char sda_retry;
	extern unsigned romdos_rev;
	extern struct COUNTRY_INFO country_info;
	extern unsigned ROM_DS;
	extern unsigned current_psp;
	extern char revision_str[];
	extern char no_ctrl_c;
	extern unsigned cfg_seg;
	extern unsigned cfg_size;
	extern unsigned SuperID;		// lo byte=ID, hi byte=order
	extern unsigned SuperBoot;		// always boot from Super-Boot drive?


	ENTER_FUNCTION();

	/* turn off the ctrl-C delay.. no ctrl-Cs during bootup! */
	if(func == 1)
		no_ctrl_c = 0;

	/* Set DOS version (returned by func 30H, Get DOS Ver) */
	else if(func == 2)
	{
		((struct PSP far *)MK_FP(current_psp,0))->dosver = R_CX;
	}

	/* get ROM-DOS revision string */
	else if(func == 3)
	{
		R_AX = romdos_rev;
		R_ES = ROM_DS;
		R_BX = (unsigned)revision_str;
		R_CX = 0xB2D2;		/* Magic Cookie for ROM-DOS identification */
		R_DX = COMPATIBILITY;
	}

	/* set country information */
	else if(func == 4)
	{
		country_info.country_code = R_BX;
		country_info.code_page    = R_CX;
		fmemcpy((uchar far *)(&country_info)+DATESTART,
			MK_FP(R_DS, R_DX), sizeof(struct COUNTRY_INFO)-DATESTART);
	}

	/* return config.sys location and size */
	else if(func == 5)
	{
		R_AX = 0xB2D2;
		R_BX = cfg_seg;
		R_CX = cfg_size;
	}

	/* return Super-Boot status */
	else if(func == 6)
	{
		R_CX = SuperID;
		R_DX = SuperBoot;
	}

	else
	{
		// not a valid option!
		ABORT();
	}

	EXIT_FUNCTION();
}


/*
	-----------------------------------------

		Failure Routines, Assert, etc...

	-----------------------------------------
*/

/*
	This function is called if an assert fails.  It does not
	actually return at all, but must have a return type so
	that the assert expression can be resolved.

	Does not return.
*/
void	PASCAL krn_assert(unsigned filenum, unsigned linenum)
{
	extern char fatal_interr[];
	panic(fatal_interr,filenum,linenum);
}


/*
	This is the default divide by zero handler.  There is an
	assembly counterpart "int_00()" which will call this
	routine, and after this routine will terminate the
	current process.
*/
void	divide_by_zero(unsigned cs,unsigned ip)
{
	extern char fatal_zero[];

	/* let user know where this div by zero happened */
	printf(fatal_zero,cs,ip);

	/* if it happened internally, we have a big problem... */
	PERMASSERT(cs != getCS());
}


/*
	The stack has overflowed during this DOS call.
*/
void	stk_overflow(void)
{
	extern char fatal_overflow[];
	panic(fatal_overflow,R_CS,R_IP);
}

#if DEBUG > 1
/*
	We reentered 1 too many times.
*/
void	bad_reenter(void)
{
	ABORT();
}
#endif
