//
//   $Id: DOSCVT.C 1.20 2003/06/12 20:57:14Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSCVT.C  created 11/29/90  by Drew Gislason

   Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains routines needed to convert one data
	structure to another.  None of these routines talk to
	the outside world directly; that is no disk access, no
	memory allocation, etc.

**********************************************************/

#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#define __FILENUM__ 2

/*
	Initialize a CDS.  This is used only on startup (as
	we install disk devices) and during int 2FH 12xxH calls.
	Cannot fail as drive # has already been validated.
*/
void PASCAL setup_cds(LPCDS cds, uchar drive)
{
	extern struct LOL lol;
	static char root[] = "A:\\";

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(cds);

	root[0] = 'A' + drive;
	fmemcpy(cds->cwd, (char far *)root, sizeof(root));
#if USING_FAT32
	cds->ulSubDirStartingCluster = 0;	// fat16 drives have root of 0
	cds->unknown2 = 0xFFFF;
#else
	cds->subdir = 0;
	cds->unknown1 = cds->unknown2 = 0xFFFF;
#endif
	if(drive < lol.numdrives)
	{
		cds->lpDPB = drv_to_dpb(drive);
		cds->status = C_VALID;
	}
	cds->skip = 2;

	EXIT_FUNCTION();
}

/*
	DOS Call 29H

	Convert an ASCII filename into an FCB.  The filename may
	include a drive specifier but not a path.  The function
	stops parsing when it hits a terminator.

	Sets R_SI and R_AL (FF=error).
*/
void	file_parse(void)
{
	/*
		[ externals ]
	*/
	extern uchar int21_al;
	extern FCB sda_fcb;
	extern char sda_ename1[];


	/*
		[ locals ]
	*/
	uchar retcode;
	uchar drive = 0;
	int len;
	char far * name;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* get the old FCB into our local one, we only need the 1st 10H bytes */
	len = retcode = 0;
	fmemcpy((void far *)&sda_fcb,MK_FP(R_ES,R_DI),0x10);

	/* ignore leading whitespace */
	name = MK_FP(R_DS,R_SI);
	while(*name == ' ' || *name == '\t')
		++name;

	/* if there is a drive part, use it */
	if(*name && *(name+1)==':')
	{
		drive = *name;
		drive = toupper(drive) - 'A' + 1;
		sda_fcb.drive = drive;
		name += 2;
		len += 2;
	}

	/* Bit 1, If no drive, 1=Leave unchanged, 0=Set DR# to 0 */
	else if(!(int21_al&2))
		sda_fcb.drive = 0;

	/* get the name into expanded form */
	len += dot_to_ename(sda_ename1,name,TRUE);
	if(wild_name())
		retcode = 1;

	/* Bit 2, If no filename  1=Leave unchanged, 0=Set filename to blanks */
	if(sda_ename1[0] != ' ' || !(int21_al&4))
		memmove(sda_fcb.name,sda_ename1,8);

	/* Bit 3, If no extension, 1=Leave unchanged, 0=Set ext to blanks */
	if(sda_ename1[8] != ' ' || !(int21_al&8))
		memmove(sda_fcb.ext,sda_ename1+8,3);

	/* zero out current block and record size */
	sda_fcb.cur_blk = sda_fcb.record_size = 0;

	/* only move 10H bytes into the user FCB area */
	fmemcpy(MK_FP(R_ES,R_DI),(void far *)&sda_fcb,0x10);

	/* if it's an invalid drive, the call fails */
	if(drive && !drive_to_sda(drive-1))
		retcode = 0xFF;

	/* Point DS:SI to just after the string we've parsed */
	R_SI += len;
	R_AL = retcode;

	EXIT_FUNCTION();
}


/*
	Convert an SFT # into an SFT pointer.  An "SFT #" is
	found in the user's PSP handle table.

	Returns a pointer to the SFT or (void *)NULLF if the SFT
	doesn't exist.
*/
LPSFT PASCAL h_to_sft(int h)
{
	extern struct LOL lol;
	extern unsigned sda_callflags;
	LPDCB dcb;
	LPSFT sft;
	PLPDCB first;

	ENTER_FUNCTION();

	first = callf_fcb() ? &lol.first_fcb : &lol.first_dcb;
	dcb = *first;
	(void *)sft = NULLF;
	while(1)
	{
		/* we have found it */
		if(h < dcb->numsft)
		{
			sft = MK_FP(FP_SEG(dcb),dcb->sft+h);
			break;
		}

		/* check if we are at the end of the list */
		if((void *)dcb->next == NULLF)
			break;

		/* check the next device control block */
		h -= dcb->numsft;
		dcb = dcb->next;
	}

	EXIT_FUNCTION();

	return sft;
}


/*
	Convert a drv # into a Drive Parameter Block pointer.
	Assumes the drive is valid.
*/
LPDPB PASCAL drv_to_dpb(uchar drv)
{
	extern struct LOL lol;
	LPDPB dpb = lol.first_dpb;

	ENTER_FUNCTION();

	while(drv--)
	{
		dpb = dpb->next;
	}

	EXIT_FUNCTION();

	return dpb;
}


/*
	Convert a drive # into a CDS.

	Returns NULL on error, ptr to the CDS if worked.
*/
LPCDS PASCAL drv_to_cds(uchar drive)
{
	LPCDS cds;

	ENTER_FUNCTION();

	// must be VALID for this cds check
	cds = drv_to_cds_no_check(drive);
	if((void *)cds == NULLF || (cds->status & C_VALID) == 0)
	{
		EXIT_FUNCTION();
		return NULLF;
	}

	EXIT_FUNCTION();

	return cds;
}

/*
	Return a pointer to the CDS.  Only the drive # is
	checked.  Doesn't care whether the CDS is valid or
	not.
*/
LPCDS drv_to_cds_no_check(uchar drive)
{
	extern struct LOL lol;
	extern bool temp_cds;
	uchar maxdrive;

	ENTER_FUNCTION();

	/* we MUST have a CDS for file operations */
	ASSERT(FP_SEG(lol.first_cds));

	/* check in the array for valid CDS */
	maxdrive = (temp_cds ? lol.numdrives : lol.lastdrive);
	if(drive < maxdrive)
	{
		EXIT_FUNCTION();
		return (lol.first_cds+drive);
	}

	EXIT_FUNCTION();

	/* operation failed, not a real drive */
	return NULLF;
}

#if USING_FAT32

/*
	This function was duplicated because of the changes of
	USING_FAT32 caused

	Convert a BIOS Parameter Block (found on the disk)
	into a Drive Parameter Block (internal to DOS, 1 for
	each FAT disk).  Sets things like FAT ID, sectors per
	cluster, etc...
*/
bool PASCAL bpb_to_dpb(LPBPB bpb, LPDPB dpb)
{
	unsigned bps;
	unsigned res;
	ulong    dirstart;
	ulong    datastart;
	ulong    ulFreeClusters;
	ulong		ulNextFreeCluster;
	ulong    ulMaxCluster;
	unsigned entries;
	ulong    sectors;
	uchar    spc;
	uchar    num_fats;
	uchar    shift;


	ENTER_FUNCTION();

	// validate pointers to structures
	//
	ASSERT(bpb);
	ASSERT(dpb);

	//  Make local copies of some of the variables in the BPB
	//
	bps      = bpb->sec_size;
	res      = bpb->res_sec;
	spc      = bpb->sec_per_clus;
	entries  = bpb->uDirEntries;
	num_fats = bpb->num_fats;

	// make sure BPB looks reasonable
	//
	if(!bps || bps&0x7F || !spc)
	{
		EXIT_FUNCTION();

		return FALSE;
	}

	//  Determine the shift factor for the cluster
	//
	spc = spc >> 1;
	for(shift=0; spc; ++shift)
	{
		spc >>= 1;
	}

	//
	//  These calculations are the same for both FAT16 and FAT32 drives
	//

	sectors      	= (bpb->num_sec ? (long)bpb->num_sec : bpb->lnumsec);
	ulFreeClusters	= 0xFFFFFFFFL;

	// FAT32 disk? then set the fields up for FAT32
	if( !bpb->uFatSectors )
	{
		datastart    = res + _lmulu(bpb->ulSectorsPerFat, num_fats);
		ulMaxCluster = _lshru(sectors - datastart, shift) + 1;

		//  These fields are assigned zero for a FAT32 DPB
		//
		dpb->dir_ent   = 0;
		dpb->fat_len   = 0;
		dpb->hiclus    = 0;

		//  This field is not used for FAT32 drives and should be
		//  assigned -1
		//
		dpb->dir_start	= Negative1;

		//  Set the FAT32 fields of the DPB
		//
		dpb->uExtendedFlags        = bpb->uExtendedFlags;
		dpb->uFSInfoSector         = bpb->uFSInfoSector;
		dpb->uBackUpBootSector     = bpb->uBackUpBootSector;
		dpb->ulSectorsPerFAT       = bpb->ulSectorsPerFat;
		dpb->ulRootDirStartCluster = bpb->ulRootDirStartCluster;
		dpb->ulFirstDataSector     = datastart;
		dpb->ulMaxCluster          = ulMaxCluster;
	}
	else
	{
		dirstart     = res + _lmulu(bpb->uFatSectors, num_fats);
		datastart    = dirstart + ((entries + ((bps>>5) - 1)) << 5) / bps;
		ulMaxCluster = _lshru(sectors - datastart, shift) + 1;

		dpb->dir_ent	= entries;
		dpb->dir_start	= LOWORD(dirstart);
		dpb->hiclus    = LOWORD(ulMaxCluster);
		dpb->fat_len   = bpb->uFatSectors;
	   dpb->newclus   = 0;
	}

	//
	//  These fields get set the same for FAT32 or FAT16
	//

	dpb->bytes_per_sec	= bps;
	dpb->ms_spc				= bpb->sec_per_clus - 1;
	dpb->fat_start			= res;
	dpb->num_fats			= num_fats;
	dpb->media				= bpb->media;
	dpb->shift           = shift;

	//  This field gets set with with the low word of the same variable
	//  for both FAT32 and FAT16 drives.  See the notes in the DPB
	//  structure before changing these assignments.
	//
	dpb->data_start  = LOWORD(datastart);

	//  For a FAT32 drive we need to retrieve the number of free clusters
	//  and the next available cluster number from the disk.  The DPB needs
	//  to be fully initilized and the sda_dpbptr needs to point to the
	//  dpb.  For romdos init this is done in dev_init.
	//
	if( !bpb->uFatSectors )
	{
		// determine freespace and next free cluster
		//
		ulFreeClusters = 0xffffffffL;
		ulNextFreeCluster = 0;
		dpb->uFreeClustersHigh = HIWORD(ulFreeClusters);
		dpb->ulNextFreeCluster = ulNextFreeCluster;
		dpb->newclus           = LOWORD(ulNextFreeCluster);
	}
	else
	{
		dpb->rebuild			  = 0xFF;
	}

	//  This field gets set with with the low word of the same variable
	//  for both FAT32 and FAT16 drives.  See the notes in the DPB
	//  structure before changing these assignments.
	//
	dpb->freeclus    = LOWORD(ulFreeClusters);

	EXIT_FUNCTION();

	return TRUE;
}

#else // #if USING_FAT32

/*
	Convert a BIOS Parameter Block (found on the disk)
	into a Drive Parameter Block (internal to DOS, 1 for
	each FAT disk).  Sets things like FAT ID, sectors per
	cluster, etc...
*/
  bool PASCAL bpb_to_dpb(LPBPB bpb, LPDPB dpb)
{
	unsigned bps      = bpb->sec_size;
	unsigned res      = bpb->res_sec;
	uchar    spc      = bpb->sec_per_clus;
	unsigned spf      = bpb->fat_size;
	uchar    num_fats = bpb->num_fats;
	unsigned entries   = bpb->dir_ent;
	long     sectors;
	unsigned dirstart;
	unsigned datastart;
	uchar    shift;

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(bpb);
	ASSERT(dpb);

	/* make sure BPB looks reasonable */
	if(!bps || bps&0x7F || !spc)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	dirstart = res + (num_fats * spf);
	datastart= dirstart + ((entries + ((bps>>5) - 1)) << 5) / bps;

	dpb->bytes_per_sec	= bps;
	dpb->ms_spc				= spc - 1;
	dpb->fat_start			= res;
	dpb->dir_start			= dirstart;
	dpb->num_fats			= num_fats;
	dpb->dir_ent			= entries;
	dpb->fat_len			= spf;
	dpb->data_start		= datastart;
	dpb->media				= bpb->media;
	dpb->rebuild			= 0xFF;
	dpb->newclus			= 0;
	dpb->freeclus			= 0xFFFF;

	/* figure out the shift factor for the cluster */
	spc = spc >> 1;
	for(shift=0; spc; ++shift)
		spc = spc >> 1;
	dpb->shift = shift;

	sectors = (bpb->num_sec ? (long)bpb->num_sec : bpb->lnumsec);
	dpb->hiclus = (unsigned)_lshru(sectors - datastart, shift) + 1;

	EXIT_FUNCTION();

	return TRUE;
}

#endif  // #if USING_FAT32

#if USING_FAT32
/*
 * This handles int 21h fn 7304h: SetDPB for FORMAT, it has 5 subfunctions described below.
 */

 bool PASCAL GetSetFat32FormatInformation(LPSFS lpsfsFormatStruct)
{
	extern int error_no;
	extern LPDPB sda_dpbptr;

	unsigned far *lpuFreeClusters;

	ENTER_FUNCTION();

	ASSERT(lpsfsFormatStruct);

#if 0
// This function does not use CX as the structure size...
//
	if(/*(lpsfsFormatStruct->uSize  != sizeof(struct SDPDFormatStruct)) ||*/
   	(R_CX                      != sizeof(struct SDPDFormatStruct)) ||
   	(lpsfsFormatStruct->uLevel != 0))
#else
	if((lpsfsFormatStruct->uSize  != sizeof(struct SDPDFormatStruct)) ||
   	(lpsfsFormatStruct->uLevel != 0))
#endif
	{
BAD_DATA_ERROR:
   	error_no = ERR_BADDATA;
		EXIT_FUNCTION();
   	return FALSE;
	}
	if(!defdrive_to_sda(R_DL) || (sda_cdsptr->status&C_NETWORK))
   	goto BAD_DATA_ERROR;

	switch(lpsfsFormatStruct->ulFunction)
	{
	case SETDPB_SET_ALLOC_INFO:
	  	{
      	switch(lpsfsFormatStruct->ulValue1)
      	{
      	case 0:                     // no change to free count
            	break;
      	case 0xFFFFFFFFL:           // set free count to "unknown"
      	default:                    // specify non-zero free count for drive
            	lpuFreeClusters = MK_FP(FP_SEG(&lpsfsFormatStruct->ulValue1),
                                    	FP_OFF(&lpsfsFormatStruct->ulValue1));
            	sda_dpbptr->freeclus          = *lpuFreeClusters++;
            	sda_dpbptr->uFreeClustersHigh = *lpuFreeClusters;
      	}
      	switch(lpsfsFormatStruct->ulValue2)
      	{
      	case 0:                     // no change to next free cluster
            	break;
      	case 0xFFFFFFFFL:           // set next free cluster to the start of the FAT
      	default:                    // specify next free cluster search starting point
            	sda_dpbptr->ulNextFreeCluster = lpsfsFormatStruct->ulValue2;
      	}
			buf_getsector(0);

// The effect of these lines is to erase the changes made above.
// Hence, they are not useful...
//
//			lpbpb = (LPBPB)((char far *)sda_cur_buf->buffer + BOOT_BPB);
//			sda_dpbptr->rebuild = 0xFF;
//			bpb_to_dpb(lpbpb, sda_dpbptr);
			sda_dpbptr->rebuild = 1;
			SetFSInfoFields();
   	}
   	break;
	case SETDPB_SET_DPB_FROM_BPB:
   	{
			EXIT_FUNCTION();
      	return bpb_to_dpb((LPBPB)lpsfsFormatStruct->ulValue2, sda_dpbptr);
   	}
	case SETDPB_FORCE_MEDIA_CHANGE:
   	{
      	sda_dpbptr->rebuild = 0xFF;  // force rebuilding DPB from BPB on drive
   	}
   	break;
	case SETDPB_GET_SET_FAT_MIRROR:
   	{
      	if(!IsFat32Drive(sda_dpbptr))
            	goto BAD_FUNCTION_ERROR;

      	// return current FAT mirroring state and current FAT index in value 2
      	lpsfsFormatStruct->ulValue2 = (ulong)sda_dpbptr->uExtendedFlags;

      	// set current FAT mirroring state and current index based on value 1
      	if(lpsfsFormatStruct->ulValue1 != 0xFFFFFFFFL)  // just return info if -1
      	{
            	if(lpsfsFormatStruct->ulValue1 & 0xFFFFFF70L) // only x000xxxxb allowed
               	goto BAD_DATA_ERROR;

            	if(lpsfsFormatStruct->ulValue1 & 0x80L)   // if turning off FAT mirroring
            	{
               	if((lpsfsFormatStruct->ulValue1 & 0x0FL) >= sda_dpbptr->num_fats)
                  	goto BAD_DATA_ERROR;    // index must be in range
            	}
            	else if (lpsfsFormatStruct->ulValue1 != 0L)
               	goto BAD_DATA_ERROR;        // index must be zero if FAT is mirrored

            	// We don't allow the FAT mirror state to be changed. To do it correctly
            	// we need to:
            	// 1) update the sda_dpbptr->uExtendedFlags from lpsfsFormatStruct->ulValue1
            	// 2) Set the free space and next free cluster to "unknown", updating FSInfo.
            	// 3) update the BPB.uExtendedFlags in the boot sector of the drive
            	// 4) if FAT mirroring was off and is being turned on then need to copy
            	//    the current FAT contents into all mirrored copies of the FAT.
            	//
            	//  The user is responsible for subsequently updating the device by
            	//  making a SetDeviceParamaters call.
            	//
            	if ((lpsfsFormatStruct->ulValue1 != 0L) &&
               	(lpsfsFormatStruct->ulValue1 != (ulong)sda_dpbptr->uExtendedFlags))
               	goto BAD_DATA_ERROR;
      	}
   	}
   	break;
	case SETDPB_GET_SET_ROOT_DIR_CLUS:
		{
   		if(!IsFat32Drive(sda_dpbptr))
   	  		goto BAD_FUNCTION_ERROR;

   		// return current root dir starting cluster in value 2
   		lpsfsFormatStruct->ulValue2 = sda_dpbptr->ulRootDirStartCluster;
   		if(lpsfsFormatStruct->ulValue1 != 0xFFFFFFFFL)  // just return info if -1
   		{
         	// We don't allow changing the location of the root directory. To support
         	// this call we need to:
         	//  1) update the BPB.ulRootDirStartCluster field in the disks boot sector.
         	//
         	//  The user is responsible for subsequently updating the device by
         	//  making a SetDeviceParamaters call.
         	//
         	if(lpsfsFormatStruct->ulValue1 != sda_dpbptr->ulRootDirStartCluster)
            	goto BAD_DATA_ERROR;
   		}
		}
		break;
   default:
	BAD_FUNCTION_ERROR:
		error_no = ERR_BADFUNC;
		EXIT_FUNCTION();
		return FALSE;
	}

	EXIT_FUNCTION();

	return TRUE;
}
#endif // USING_FAT32

/*
	Convert a "FILENAME.EXT" to an expanded name form.  That
	is "FILE.C" to "FILE    C  ".  Will stop processing at
	any terminating character, filling out the rest of the
	expanded string with spaces.  It also expands wildcards.

	"Test.c"					-> "TEST    C  "  should work
	"TEST   .C  "			-> "TEST    C  "  should work
	"tEsT my PROGRAM.C"	-> "TEST MY C  "  should fail to open
	"TEST.C XYZ"			-> "TEST    C X"  should fail to open
	"INTERACT.ABC"			-> "INTERACTABC"  should work
	"interact here.abc"	-> "INTERACTABC"  should work
	"INTERACT IT^A.ABC"	-> "INTERACT   "  should stop on ^A

	Returns the length of the string used.
*/
int	PASCAL dot_to_ename(char *ename,char far *dotname, bool termspace)
{
	extern uchar * terminators;

	int dlen;			/* count the # of character used */
	int elen;			/* current length is expanded name */
	bool founddot;		/* we can only have one "." */
	bool leadbyte;
	uchar c;

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(ename);
	ASSERT(dotname);

	dlen = elen = 0;
	founddot = leadbyte = FALSE;
	while(1)
	{
		c = *dotname;

		/* lead bytes are special */
		if(leadbyte)
			goto nextchar;

		/* if we have hit a terminator, we are done */
		/* "file_parse" terminates on space as well */
		if(c < ' ' || strchr((char *)terminators,(char)c) ||
				(termspace && c==' '))
			break;

		/* check for the dot separator */
		if(c == '.')
		{
			/* two dots are not allowed, 2nd dot becomes a terminator */
			if(founddot)
				break;
			else
			{
				founddot = TRUE;
				while(elen < 8)		/* fill to end of "name" */
				{
					*ename++ = ' ';
					++elen;
				}
			}
		}

		/* check for wildcards */
		else if(c == '*')
		{
			while(elen < (founddot ? 11 : 8))
			{
				*ename++ = '?';
				++elen;
			}
		}

		/* must be a valid character */
		else
		{
nextchar:
			if(elen < (founddot ? 11 : 8))
			{
				if(!leadbyte && is_lead_byte(c))
					leadbyte = 2;
				*ename++ = (leadbyte ? c : toupper(c));
				++elen;
			}
		}

		++dlen;
		++dotname;
		if(leadbyte)
			--leadbyte;
	}

	/* fill out the rest of the name with spaces */
	while(elen++ < 11)
		*ename++ = ' ';

	EXIT_FUNCTION();

	/* return length of string used to make ename */
	return dlen;
}


/*
	Convert an expanded name form to a "FILE.EXT" form.
	That is "FILE    C  " to "FILE.C".  Does no case
	conversion.
*/
void	PASCAL ename_to_dot(char *dotname, char *ename)
{
	uchar len;
	uchar end = 8;

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(dotname);
	ASSERT(ename);

	/* place the name part into the dotname string */
again:
	for(len = end; len > 0 && ename[len-1] == ' '; --len)
		/* just count */;
	if(len)
	{
		if(end == 3)
			*dotname++ = '.';
		memmove(dotname,ename,len);
		dotname += len;
	}
	if(end == 8)
	{
		ename += 8;
		end = 3;
		goto again;
	}

	*dotname = 0;

	EXIT_FUNCTION();
}


/*
	Convert a relative path into a validated absolute path.
	Assumes the drive has already been stripped off.  Makes
	no checks as to whether the path exists or not, but only
	manipulates the string.

	Inputs:
		relpath	- relative unvalidated pathname
	Outputs:
		abspath	- absolute validated pathname
		ename    - the last name in the given path

	Rules:
		1) No invalid characters anywhere in the string.
		2) Convert all lowercase letters to UPPERCASE.
		3) The input string may be any length.
		4) All forslashes are converted to backslashes.
		5) Except for the first (or root) set of slashes,
		   any number of slashes is converted to 1 backslash.
		6) The first set of slashes are special in that if
		   there are two or more, they are left and the drive
		   part is omitted.
		7) The maximum length of the output string is 127
		   characters.

	Returns the absolute path and TRUE if worked.
	Returns FALSE if it failed.
*/
bool	PASCAL rel_to_abspath(char *abspath, char far *relpath, char *ename)
{
	/*
		[ externals ]
	*/
	extern LPCDS sda_cdsptr;
	extern unsigned char sda_srchattr;

	/*
		[ locals ]
	*/
	#define MAXLEN	126
	char *path;
	int len;
	uchar c;
	char name[14];
	int dotfound;
	char far * cwd;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate pointers
	ASSERT(abspath);
	ASSERT(relpath);
	ASSERT(ename);

	/* there must be a path */
	c = *relpath;
	if(!c)
   {
      EXIT_FUNCTION();
		return FALSE;
   }

	/* if this path starts at the root, say so */
	if(isslash(c))
	{
		/* skip all preceding backslashes */
		path = abspath;
		while(isslash(c))
		{
			*path++ = '\\';
			c = *(++relpath);

			/* skip past root \ so we can't backup .. too far */
			++abspath;
		}

		/* just root is ok */
		if(!c)
			goto done;
	}

	/* path is relative, use cwd */
	else
	{
		cwd = sda_cdsptr->cwd + sda_cdsptr->skip;
		len = fstrlen(cwd);
		fmemcpy((void far *)abspath, cwd, len);

		/* "C:\", "C:\TEMP" or "Net.\A.\" are all valid cwds. */
		if(len == 1 && isslash(*abspath))
			len = 0;

		/* abspath ALWAYS begins with a '\' */
		path = abspath+len;
		*path++ = '\\';

		/* skip past root \ so we can't backup .. too far */
		++abspath;
	}

	/* there MUST be a name part */
	dot_to_ename(ename, (void far *)"", FALSE);

	/* go through the whole of the relative path */
	while(1)
	{
		/* check for the special names "." and ".." */
		dotfound = len = countdots(relpath);

		/* if name begins with dots */
		if(dotfound)
		{
			if(len > 2)
			{
				EXIT_FUNCTION();
				return FALSE;
			}

			/* the ".." entry means to go back one level */
			if(len == 2)
			{
				/* we are just after a slash, move to just BEFORE it */
				path -= 2;

				/* "root" is as far back as we go */
				if(path < abspath)
            {
               EXIT_FUNCTION();
					return FALSE;
            }

				/* move to just AFTER preceding slash */
				*path = 0;
				while(path > abspath  && lastchar(abspath) != '\\')
					*(--path) = 0;
			}
		}

		/* parse the name */
		else
		{
			len = dot_to_ename(ename,relpath, FALSE);
			if(!len)
			{
				EXIT_FUNCTION();
				return FALSE;
			}

#if USING_FAT32

			//  Names can have spaces in as part of the name when
			//  in DOS 7.1 truncated mode
			//

#else  // #if USING_FAT32

//  BT 6-17-01
//
//  Names can contain spaces and be to long in which case they
//  will be truncated.  Command is filtering these from the prompt
//  but no holds bard when calling the interrupts directly
//
//			// Volume labels may contain spaces. Nothing else can.
//			//
//			if( !(sda_srchattr & A_VOLUME) )
//				if(!ok_ename(ename))
//					return FALSE;

#endif // #if USING_FAT32

		}

		/* on to the next part */
		relpath += len;
		c = *relpath;

		/* skip following spaces */
		if(c == ' ')
		{
			while((c = *relpath) == ' ')
				++relpath;

			/* spaces not allowed in path part (name part only) */
			if(isslash(c))
			{
				EXIT_FUNCTION();
				return FALSE;
			}
		}

		/* copy the name into abspath */
		if(!dotfound)
		{
			ename_to_dot(name,ename);

			len = strlen(name);
			memmove(path,name,len);
			path += len;
		}

		/* we are done with the string (found the name part) */
		if(!c)
		{
			/* don't include trailing "\" */
			if(dotfound && path>abspath)
				--path;
done:
			/* fill ename with blanks at end of string */
			if(path == abspath || dotfound)
				dot_to_ename(ename, relpath, FALSE);
			break;
		}

		/* now we must be in a path part - now wildcards allowed */
		ename[11] = 0;
		if(strchr(ename,'?'))
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		/* if the next part is not a slash, it's not a valid path */
		if(!isslash(c))
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		/* skip multiple slashes */
		while(isslash(c))
			c = *(++relpath);

		if(!dotfound)
			*path++ = '\\';
	}

	/* SystemSoft driver's require a WORD of 0s after path */
	*(unsigned *)path = 0;

	EXIT_FUNCTION();

	/* path OK */
	return TRUE;
}


/*
	Count the # of contiguous dots beginning at str.
*/
int	PASCAL countdots(char far *str)
{
	int num = 0;

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(str);

	while(*str++ == '.')
		++num;

	EXIT_FUNCTION();

	return num;
}

/*
	enames must not have spaces in the middle - they can
	have spaces at the end, but not in the middle.
*/
int ok_ename(char *ename)
{
	int i,j;
	int spacemode;
	int limit = 8;

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(ename);

	for(j=0; j<11; j+= 8)
	{
		spacemode = FALSE;
		for(i=j; i<limit; ++i)
		{
			if(ename[i] == ' ')
				spacemode = TRUE;
			else if(spacemode)
			{
				EXIT_FUNCTION();
				return FALSE;
			}
		}
		limit += 3;
	}

	EXIT_FUNCTION();

	return TRUE;
}


/*
	Convert a directory structure into an SFT.  Assumes the
	SFT has already been initialized.
*/
void	dir_to_sft(void)
{
	/*
		[ externals ]
	*/
	extern uchar sda_drive;
	extern DIR sda_ffb;
	extern SDB sda_sdb;
	extern LPSFT sda_sftptr;


	/*
		[ locals ]
	*/
	struct DIRENTRY dirent;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* set up the SFT */
#if USING_FAT32
	sda_sftptr->ulStartingCluster = Fat32Start(sda_ffb);
#if LSEEK_ENHANCEMENT
   sda_sftptr->uRDlseekHigh      = HIWORD(sda_sftptr->ulStartingCluster);
   sda_sftptr->uRDlseekLow       = LOWORD(sda_sftptr->ulStartingCluster);
#endif  // #if LSEEK_ENHANCEMENT
#else
	sda_sftptr->start  = sda_ffb.start;
#endif
	sda_sftptr->size   = sda_ffb.size;
	either_to_sft(IO_NOTWRITTEN | sda_drive);

	/* get the directory sector of this file */
	dirent.PhysicalCluster = 0;
	dir_sector(FALSE, &dirent);
	sda_sftptr->dirsec = dirent.dirsec;
	sda_sftptr->dirent = dirent.dirent;

	EXIT_FUNCTION();
}


/*
	Convert a device pointer into an SFT.  Assumes the SFT
	has already been initialized.
*/
void	dev_to_sft(void)
{
	extern DIR sda_ffb;
	extern LPDEV_HDR sda_devptr;
	extern LPSFT sda_sftptr;

	ENTER_FUNCTION();

	/* fake a file found block */
	get_fdate();
	fmemcpy((char far *)sda_ffb.name,sda_devptr->name,8);
	memset(sda_ffb.ext,' ',3);
	sda_ffb.attr = 0x20;

	/* set up the SFT */
	either_to_sft((IO_DEVICE | IO_NOTEOF) | (sda_devptr->attr&0x7FFF));

	EXIT_FUNCTION();
}


/*
	Used by dev_to_sft and dir_to_sft to set things in
	common.  This routine is simply a code size saver.
*/
void	PASCAL either_to_sft(unsigned ioctl)
{
	extern DIR sda_ffb;
	extern LPSFT sda_sftptr;

	ENTER_FUNCTION();

	sda_sftptr->ioctl  = ioctl;
	sda_sftptr->attr   = sda_ffb.attr;
	sda_sftptr->time   = sda_ffb.time;
	sda_sftptr->date   = sda_ffb.date;
	fmemcpy(sda_sftptr->name,(char far *)sda_ffb.name,11);

	EXIT_FUNCTION();
}


/*
	Convert an SFT entry to a DIR entry.  Updates the date
	and time to the current time if need be.  This function
	is used when a file has changed and we need to update
	the entry on disk.
*/
void	sft_to_dir(void)
{
	extern DIR sda_ffb;
	extern LPSFT sda_sftptr;

	ENTER_FUNCTION();

	fmemcpy((void far *)sda_ffb.name, sda_sftptr->name, 11);
	sda_ffb.attr = sda_sftptr->attr | A_ARCHIVE;
	if(sda_sftptr->ioctl & IO_SETDATE)
	{
		sda_ffb.date = sda_sftptr->date;
		sda_ffb.time = sda_sftptr->time;
	}
	else
		get_fdate();
#if USING_FAT32
	sda_ffb.uStartingClusterLo = LOWORD(sda_sftptr->ulStartingCluster);
	sda_ffb.uStartingClusterHi = HIWORD(sda_sftptr->ulStartingCluster);
#else
	sda_ffb.start = sda_sftptr->start;
#endif
	sda_ffb.size  = sda_sftptr->size;

	EXIT_FUNCTION();
}

/*
	DOS 1 File Found Block to DTA.  Used by the DOS1 find
	first/next calls to place the data in the SDA back into
	user space.
*/
void	dos1_ffb_to_dta(void)
{
	/*
		[ externals ]
	*/
	extern uchar sda_drive;
	extern uchar sda_extfcb;
	extern uchar sda_srchattr;
	extern uchar far * DTA;
	extern DIR sda_ffb;


	/*
		[ locals ]
	*/
	static char ext[7];
	uchar far * dta;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* skip past extended part of FCB */
	dta = DTA;
	if(sda_extfcb)
	{
		*ext = 0xFF;
		*(ext+6) = sda_srchattr;
		fmemcpy(dta, ext, sizeof(ext));
		dta += 7;
	}

	/* copy the drive and File Found Block to the DTA */
	*dta = sda_drive+1;
	fmemcpy(dta+1, (void far *)&sda_ffb, sizeof(DIR));

	EXIT_FUNCTION();
}


/*
	File Found Block to DTA.  Used by findfirst/next to
	place the data in the SDA back into user space.
*/
void	dos2_ffb_to_dta(void)
{
	/*
		[ externals ]
	*/
	extern uchar far * DTA;
	extern DIR sda_ffb;
	extern SDB sda_sdb;


	/*
		[ locals ]
	*/
	LPFIND ff;
	char name[13];


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	fmemcpy(DTA, (void far *)&sda_sdb, sizeof(SDB));
	ff = (void far *)DTA;
	ff->attr  = sda_ffb.attr;
	ff->time  = sda_ffb.time;
	ff->date  = sda_ffb.date;
	ff->size  = sda_ffb.size;
	ename_to_dot(name, sda_ffb.name);
	fmemcpy(ff->name, (char far *)name, strlen(name)+1);

	EXIT_FUNCTION();
}

/*
	-------------------------------------------------------

					Miscellaneous Support Routines

	-------------------------------------------------------
*/

/*
	Forward string search.
*/
char	* PASCAL strchr(char *s,char c)
{
	ENTER_FUNCTION();

	// validate pointer
	ASSERT(s);

	while(*s)
	{
		if(*s == c)
		{
			EXIT_FUNCTION();
			return s;
		}
		++s;
	}

	EXIT_FUNCTION();

	return NULL;
}


/*
	Reverse string search.
*/
char	* PASCAL strrchr(register char *p,char c)
{
	char *pstart = p;

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(p);

	p += strlen(p) - 1;		/* point to just before terminating 0	*/
	while (p >= pstart)
	{
		if(*p == c)
		{
			EXIT_FUNCTION();
			return p;
		}
		--p;
	}

	EXIT_FUNCTION();

	return NULL;
}


/*

	Skip the whitespace of a string.
*/
char	* PASCAL skipwhite(char *s)
{
	ENTER_FUNCTION();

	// validate pointer
	ASSERT(s);

	while(*s && DL_IS_SPACE(*s))
		++s;

	EXIT_FUNCTION();

	return s;
}


/*
   Skip the whitespace of a string.
*/
char far * PASCAL fskipwhite(char far *fs)
{
   ENTER_FUNCTION();

   // validate pointer
   ASSERT(fs);

	while(*fs && DL_IS_SPACE(*fs))
      ++fs;

   EXIT_FUNCTION();

   return fs;
}


/*
   Convert a far string to an integer (unsigned actually)
*/
int   PASCAL fatoi(char far *fs)
{
   int i=0;

   ENTER_FUNCTION();

   // validate pointer
   ASSERT(fs);

   while(isdigit(*fs))
   {
      i = i*10 + (*fs-'0');
      ++fs;
   }

   EXIT_FUNCTION();

   return i;
}

/*
   Convert a string to an integer (unsigned actually)
*/
int   PASCAL atoi(char *s)
{
   return( fatoi( (char far *)s ) );
}


/*
	Change a string to uppercase
*/
void	PASCAL strupr(char *p)
{
	ENTER_FUNCTION();

	// validate pointer
	ASSERT(p);

	uppercase((void far *)p,-1);

	EXIT_FUNCTION();
}


/*
	Concatinate strings.
*/
void PASCAL strcat(char *s1, char *s2)
{
	ENTER_FUNCTION();

	// validate parameters
	ASSERT(s1);
	ASSERT(s2);

	s1 += strlen(s1);
	memmove(s1, s2, strlen(s2)+1);

	EXIT_FUNCTION();
}
