//
//   $Id: DOSDIR.C 1.25 2003/08/21 14:32:46Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSDIR.C  created 12/01/90  by Drew Gislason

   Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains the directory functions and all
	those DOS calls that deal with a disk as a whole (such
	as get disk free space).  It contains both DOS level
	calls and low level directory management.

	$DOCHISTORY
	03-03-99 (DE) Changed remove_dir_entry() to work with FAT32 when
				freeing the FAT chain for the entry.


**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#define __FILENUM__ 5

/*
	-------------------------------------------------------

					Directory/Drive DOS Calls

	-------------------------------------------------------
*/

/*
	DOS call 0EH

	Set the default drive number.  This can be to any valid
	drive including a network drive.

	Returns lol.lastdrive.
*/
uchar	PASCAL set_current_drive(uchar drive)
{
	extern struct LOL lol;
	extern uchar CUR_DRIVE;

	ENTER_FUNCTION();

	if(drive_to_sda(drive))
		CUR_DRIVE = drive;

	EXIT_FUNCTION();

	return lol.lastdrive;
}


/*
	DOS calls 11H and 4EH

	Find the first occurence of a filespec in a given
	directory.  Will find a character device anywhere on
	the disk.

	Sets error_no on error.
*/
void	find_first(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_drive;
	extern unsigned sda_callflags;
	extern uchar sda_chardev;
	extern uchar sda_srchattr;
	extern DIR sda_ffb;
	extern SDB sda_sdb;
	extern char sda_ename1[];


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* set up the search data block */
	sda_sdb.srchattr = sda_srchattr;
	sda_sdb.dirnum   = 0;
	sda_sdb.drive    = sda_drive + 1;

	/* volume labels are ALWAYS in the root directory */
	if(sda_sdb.srchattr == A_VOLUME)
#if USING_FAT32
		sda_sdb.ulStartingCluster = 0;
#else
		sda_sdb.subdir = 0;
#endif

	/* a character device (like CON) is found in every directory */
	if(sda_chardev)
	{
		memset(&sda_ffb,0,sizeof(DIR));
		sda_ffb.attr = A_DEVICE;
		get_fdate();
		memmove(sda_ffb.name,sda_ename1,11);
		memmove(sda_sdb.emask,sda_ename1,11);
		sda_sdb.dirnum = 0xFFFF;
	}

	/* look for the file on the disk */
	else if(!dir_findfirst())
		error_no = ERR_LASTFILE;

	EXIT_FUNCTION();
}


/*
	DOS calls 12H and 4FH

	Sets error_no on error.
*/
void	find_next(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern unsigned sda_callflags;
	extern SDB sda_sdb;
	extern uchar far * DTA;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* let the SDA know about this search */
	if(sda_sdb.dirnum == 0xFFFF || !defdrive_to_sda(sda_sdb.drive)
		|| !dir_findnext())
		error_no = ERR_LASTFILE;

	EXIT_FUNCTION();
}

/*
	For FAT12/16 drives, determine the # of free clusters.
	Cannot fail. If a critical error occurs, will never return here.
*/
void CalcFreeClusters(unsigned *pFreeClusters)
{
	unsigned uFreeClusters;
	unsigned uCluster;
	unsigned uHiCluster;

	ENTER_FUNCTION();

	/* check if we must determine the freespace */
	uHiCluster = sda_dpbptr->hiclus;
	uFreeClusters = sda_dpbptr->freeclus;
	if(uFreeClusters > uHiCluster)
	{
		uFreeClusters = 0;
		for(uCluster = 2; uCluster <= uHiCluster; ++uCluster)
		{
         ASSERT_ValidCluster( uCluster );
			if(!get_fat_entry(uCluster))
				++uFreeClusters;
		}
		sda_dpbptr->freeclus = uFreeClusters;
	}
	// return the free clusters
	*pFreeClusters = uFreeClusters;

	EXIT_FUNCTION();
}


/*
	DOS calls 1BH, 1CH and 36H

	This routine will get freespace and other information
	about a disk.  It does a change of disk call.  If
	calcfree is TRUE, it may recalculate the freespace.

	Sets the registers AL, BX, CX, DX and maybe DS.
*/
void	PASCAL disk_freespace(bool calcfree)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;


	/*
		[ locals ]
	*/
	unsigned uTotalClusters;
	unsigned uFreeClusters;
	uchar		spc;

	ENTER_FUNCTION();

	/* always check if the disk has changed */
	change_of_disk();

	//
	spc = (1 << sda_dpbptr->shift);
	uTotalClusters  = sda_dpbptr->hiclus - 1;

#if USING_FAT32
	// FAT32 drives return strange values because 16-bits for uTotalClusters
	// is not enough bits.
	if(IsFat32Drive(sda_dpbptr))
	{
		spc = MAXSPC;					// assume maximum sectors per cluster
		uTotalClusters = HICLUS16;
	}
#endif

	/* we may have to calculate disk freespace */
	if(calcfree)
	{

#if USING_FAT32

		// FAT32 drives use a different freespace calculation
		if(IsFat32Drive(sda_dpbptr))
		{
			ulong ulFreeClusters;
			unsigned uAdjust = MAXSPC / (1 << sda_dpbptr->shift);

			if( Fat32FreeClusters() == 0xffffffffL )
			{
				GetFSInfoFields();
			}

			// FAT32 drives already have freespace in DPB
			ulFreeClusters = Fat32FreeClusters();

			// convert to clusters of 40h sec/clus
			ulFreeClusters = _ldivu(ulFreeClusters, uAdjust);
			if(ulFreeClusters >= HICLUS16)
				uFreeClusters = HICLUS16;
			else
				uFreeClusters = (unsigned)ulFreeClusters;
		}
		else
		{
			// FAT16/12 drive. Go ahead and calc the free if needed
			//
			if( sda_dpbptr->freeclus == 0xffff )
			{
				CalcFreeClusters(&uFreeClusters);
			}
			else
			{
				uFreeClusters = sda_dpbptr->freeclus;
			}
		}

		// inform user of the freespace
		R_BX = uFreeClusters;

		// compiled for FAT16 only
#else
		CalcFreeClusters(&uFreeClusters);
		R_BX = uFreeClusters;
#endif
	}

	/* otherwise point to media byte */
	else
	{
		R_DS = FP_SEG(sda_dpbptr);
		R_BX = FP_OFF(&sda_dpbptr->media);
	}

	/* set the rest of the return registers */
	R_CX = sda_dpbptr->bytes_per_sec;
	R_DX = uTotalClusters;

	/* return the sectors per cluster (goes in AL) */
	R_AL = spc;

	EXIT_FUNCTION();
}

#if USING_FAT32
/*
	DOS call 7303h

	This routine will get freespace and other information about a disk for
	FAT12/FAT16/FAT32 drives. It does a change of disk call. If calcfree is
	TRUE, it may recalculate the freespace.

	Sets the results in the GetDiskFreeSpace32 structure passed in by the
	user.

	On Entry		cx = buffer size (must be at least xx bytes)
					ds:dx = drive string (eg "C:\")
					es:di = buffer
*/
void GetExtendedFreeSpace(void)
{
	char far * lpPath;
	uchar 	drive;
	LPGDF32	lpGDF32;
	unsigned	spc;
	unsigned uFreeClusters;

	ENTER_FUNCTION();

	// verify the drive is correct
	lpPath = MK_FP(R_DS, R_DX);
	drive = toupper(*lpPath) - 'A';		// 0=A, 1=B and so on
	if(!drive_to_sda(drive))
	{
		error_no = ERR_NOTDRIVE;
		EXIT_FUNCTION();
		return;
	}

	// make sure user buffer is big enough
	if(R_CX < sizeof(GDF32))
	{
		error_no = ERR_BUFFSIZE;
		EXIT_FUNCTION();
		return;
	}

	// we only accept "level" 0
	lpGDF32 = MK_FP(R_ES, R_DI);
	if(lpGDF32->uLevel != GDF32_LEVEL_0)
	{
		error_no = ERR_BADLEVEL;
		EXIT_FUNCTION();
		return;
	}

	// verify its an OK disk
	change_of_disk();

	// set up common fields
	//
	spc = 1 << sda_dpbptr->shift;

	lpGDF32->ulSectorsPerCluster = spc;
	lpGDF32->ulBytesPerSector    = sda_dpbptr->bytes_per_sec;

	// set up fields specific to FAT32 drives
	if(IsFat32Drive(sda_dpbptr))
	{
		//  Get the values if they are not known
		//
		if( Fat32FreeClusters() == 0xffffffffL )
		{
			GetFSInfoFields();
		}

		lpGDF32->ulAvailableClusters = Fat32FreeClusters();
		lpGDF32->ulTotalClusters     = sda_dpbptr->ulMaxCluster-1;
	}

	// set up fields specific to FAT16/12 drives
	else
	{
		if( sda_dpbptr->freeclus == 0xffff )
		{
			CalcFreeClusters(&uFreeClusters);
		}
		else
		{
			uFreeClusters = sda_dpbptr->freeclus;
		}

		lpGDF32->ulAvailableClusters = uFreeClusters;
		lpGDF32->ulTotalClusters     = sda_dpbptr->hiclus - 1;
	}

	// we don't support compressed drives, so all these fields
	// just agree with Avail and Total Clusters.
	//
	lpGDF32->ulAvailableAllocationUnits = lpGDF32->ulAvailableClusters;
	lpGDF32->ulTotalAllocationUnits     = lpGDF32->ulTotalClusters;

	// Convert Free and Total cluster values to Free and Total sector counts.
	//
	lpGDF32->ulAvailablePhysicalSectors =
			_lshlu(lpGDF32->ulAvailableClusters, sda_dpbptr->shift);
	lpGDF32->ulTotalPhysicalSectors     =
			_lshlu(lpGDF32->ulTotalClusters, sda_dpbptr->shift);

	// indicate it worked by returning the size in ax
	//
	lpGDF32->uLevel = GDF32_LEVEL_0;
	lpGDF32->uSize  = sizeof(GDF32);

	lpGDF32->ulReserved[0] = 0;		// reserved fields may as well be 0
	lpGDF32->ulReserved[1] = 0;

	// ?????
	R_AX = sizeof(GDF32);

	EXIT_FUNCTION();
}

/*
	Int 21h fn 7302h GetExtendedDPB

	On Entry		cx = buffer size (must be at least 3Fh)
					dl = drive
					es:di = buffer
	On Exit		buffer filled with data. Returns CY on failure.
*/
void GetExtendedDPB(void)
{
	int far * lpBuffer;

	ENTER_FUNCTION();

	// make sure drive is valid
	if(!defdrive_to_sda(R_DL))
	{
		error_no = ERR_NOTDRIVE;
		EXIT_FUNCTION();
		return;
	}

	// make sure user buffer is big enough
	if(R_CX < sizeof(DPB)+sizeof(int))
	{
		error_no = ERR_BUFFSIZE;
		EXIT_FUNCTION();
		return;
	}

	// fill in the buffer with a size and the DPB itself
	lpBuffer = MK_FP(R_ES,R_DI);
	*lpBuffer++ = sizeof(DPB);
	fmemcpy(lpBuffer,sda_dpbptr,sizeof(DPB));

	// indicate passed
	R_AX = 0;

	EXIT_FUNCTION();
}

#endif // #if USING_FAT32

/*
	DOS calls 1FH and 32H

	Make sure the drive is valid.  If so, return a pointer
	to the DPB for that drive.

	Sets error_no if failed.
*/
uchar	PASCAL get_dpb(uchar drive)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// make sure the drive is valid, FAIL if not
	// Fat32 drives not allowed with this function
#if USING_FAT32
	if(!defdrive_to_sda(drive) || IsFat32Drive(sda_dpbptr))
	{
		EXIT_FUNCTION();
		return 0xFF;
	}
#else
	if(!defdrive_to_sda(drive))
	{
		EXIT_FUNCTION();
		return 0xFF;
	}
#endif

	/* check for a change of disk here */
	change_of_disk();

	/* record the address of this dpb */
	R_DS = FP_SEG(sda_dpbptr);
	R_BX = FP_OFF(sda_dpbptr);

	EXIT_FUNCTION();

	return 0;
}

/*
	DOS call 39H

	Create a new subdirectory.

	Sets error_no on error.  Returns nothng.

	SPR 1638 (de) Completely changed the way we make a directory.
			The old way was really slow and lost clusters on a
			critical error during a DIR update.

*/
void	mkdir(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern DIR sda_ffb;
	extern SDB sda_sdb;
	extern char sda_ename1[];
	extern char sda_fname1[];
	extern uchar sda_openmode;
	extern LPBUFLIST sda_cur_buf;


	/*
		[ locals ]
	*/
	CLUSTER    start;
	int        blanks = 0;
	LPDIR      lpDir;

	/*
		[ code]
	*/

	ENTER_FUNCTION();

	/* don't allow pathnames that are too long */
	if(strlen(sda_fname1) > PATHLEN)
		goto errout;

	/* check if all of sda_ename1 is blank */
	if( (sda_ename1[0] == ' ') && (sda_ename1[8] == ' ') )
	{
		blanks = 11;
	}

	/* entry cannot exist, contain wildcards or be a device name */
	if(blanks==11 || wild_name() || dir_findfirst() || find_device(sda_ename1))
	{
errout:
		error_no = ERR_DENIED;
		EXIT_FUNCTION();
		return;
	}

	// Find a free dir entry in the parent directory. This will
	// set up the sda_sdb and sda_ffb structures to update the
	// sector containing the subdirectory entry in the parent
	// directory.
	//
	if(!dir_findblank())
		goto errout;

	// Update the creation time for the subdir entry in the
	// parent. This time will be used for the "." and ".."
	// entries in the new subdir. Set the SDA open mode.
	//
	get_fdate();
	sda_openmode = A_SUBDIR;

	// Allocate the starting cluster so we know where it will be.
	// We mark the entry as free so that we don't end up with a
	// lost cluster in case either of the DIR writes fails.
	//
	start = alloc_next_free();
#if USING_FAT32
	if(start == END_OF_FAT)
		goto errout;
#else
	if(start == 0xFFFF)
		goto errout;
#endif

   ASSERT_ValidCluster( start );
	set_fat_entry(start,0);

	// When this routine returns, sda_cur_buf will point to the
	// buffer entry for the first sector of the new subdirectory.
	// That buffer will be marked valid and dirty. We build the
	// "." and ".." entries here to minimise disk accesses.
	//
	zero_cluster(start);
	lpDir = (LPDIR)buf_dataptr(sda_cur_buf);

	// Create the "." entry in the new subdirectory, size is already 0.
	//
	memset((void*)&sda_ffb.name,' ',11);
	sda_ffb.name[0] = '.';
	sda_ffb.attr	 = A_SUBDIR;
#if USING_FAT32
	sda_ffb.uStartingClusterLo = LOWORD(start);
	sda_ffb.uStartingClusterHi = HIWORD(start);
#else
	sda_ffb.start   = start;
#endif
	sda_ffb.size    = 0L;
	fmemcpy((void far*)lpDir,(void far*)&sda_ffb,sizeof(DIR));

	// Create the ".." entry in the new subdirectory, size is already 0.
	//
	++lpDir;
	fmemcpy((void far*)lpDir,(void far*)&sda_ffb,sizeof(DIR));
	lpDir->name[1]  = '.';
#if USING_FAT32
	lpDir->uStartingClusterLo = LOWORD(sda_sdb.ulStartingCluster);
	lpDir->uStartingClusterHi = HIWORD(sda_sdb.ulStartingCluster);
#else
	lpDir->start = sda_sdb.subdir;
#endif

	// The new subdirectory has been created. Write it out to
	// disk. If that works, update and write the subirectory's
	// entry in the parent dir. Finally, if that works too
	// then update the FAT. This write order is required to avoid
	// losing clusters when a write to a directory sector fails.
	// Note the set_fat_entry() also changes the sda_cur_buf pointer
	// to the buffer containing the newly modified fat sector.
	// create_dir_entry() then moves the root dir sector back to the
	// front of the buffer list and then flushes ALL buffers for the
	// drive.
	//
	// NOTE: SPR 1832 (de)
	//		We MUST update the FAT before calling create_dir_entry()
	// 	or else when it calls buf_flush_all() (with the buffered but
	// 	not updated FAT sectors) then Flash FX's FAT monitor will throw
	//		away our sub-directory sectors.
	//
	buf_flush(sda_cur_buf);

   ASSERT_ValidCluster( start );
#if USING_FAT32
 	set_fat_entry(start, END_OF_FAT);
#else
	set_fat_entry(start,0xFFFF);
#endif
	create_dir_entry(start);

	EXIT_FUNCTION();
}

/*
	DOS call 3AH

	Remove directory.

	Sets error_no on error.  Returns nothing.
*/
void	rmdir(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern DIR sda_ffb;
	extern SDB sda_sdb;
	extern LPCDS sda_cdsptr;


	/*
		[ locals ]
	*/
	DIR oldffb;
	SDB oldsdb;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* make sure the entry is a directory */
	if(!(sda_ffb.attr & A_SUBDIR))
		error_no = ERR_DENIED;

	/* make sure that the entry is not the current directory */
#if USING_FAT32
	else if(Fat32Start(sda_ffb) == sda_cdsptr->ulSubDirStartingCluster)
		error_no = ERR_CURDIR;
#else
	else if(sda_ffb.start == sda_cdsptr->subdir)
		error_no = ERR_CURDIR;
#endif

	/* make sure the directory is empty */
	else
	{
		/* save the old search data block */
		memmove(&oldsdb,&sda_sdb,sizeof(SDB));
		memmove(&oldffb,&sda_ffb,sizeof(DIR));

		/* set up a new one */
#if USING_FAT32
		sda_sdb.ulStartingCluster = Fat32Start(sda_ffb);
#else
		sda_sdb.subdir = sda_ffb.start;
#endif
		sda_sdb.dirnum = 1;
		memset(sda_sdb.emask,'?',11);

		if(dir_findnext())
			error_no = ERR_DENIED;

		/* restore the old search data block */
		memmove(&sda_sdb,&oldsdb,sizeof(SDB));
		memmove(&sda_ffb,&oldffb,sizeof(DIR));
	}

	/* actually remove the entry */
	if(!error_no)
		remove_dir_entry();

	EXIT_FUNCTION();
}


/*
	DOS call 3BH

	Change directory.  Changes the current working directory
	for a given drive.

	Sets error_no on error.  Returns nothing.
*/
void	chdir(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern DIR sda_ffb;
	extern SDB sda_sdb;
	extern LPCDS sda_cdsptr;
	extern char sda_fname1[];
	extern char sda_ename1[];


	/*
		[ locals ]
	*/
	CLUSTER subdir;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* special case: check for root (both name & ext are blank) */
	if(*sda_ename1 ==' ' && sda_ename1[8] == ' ')
		subdir = 0;

	/* verify that the entry exists and is a directory */
	else
	{
		sda_sdb.srchattr = A_DIRCHK;
		if(!dir_findfirst() || !(sda_ffb.attr&A_SUBDIR))
		{
			error_no = ERR_NOTPATH;
			EXIT_FUNCTION();
			return;
		}
#if USING_FAT32
		subdir = Fat32Start(sda_ffb);
#else
		subdir = sda_ffb.start;
#endif
	}

	/* update the directory cluster # */
#if USING_FAT32
	sda_cdsptr->ulSubDirStartingCluster = subdir;
#else
	sda_cdsptr->subdir = subdir;
#endif

	EXIT_FUNCTION();
}

void change_to_root(void)
{
	extern LPCDS sda_cdsptr;
	char far * path;

	ENTER_FUNCTION();

	/*
		a normal root is "C:\", a SUBST root is "C:\TEMP"
		if there is a path part to the root, it CANNOT end
		in a backslash.
	*/
	path = sda_cdsptr->cwd + sda_cdsptr->skip;
	if(sda_cdsptr->skip == 2)
		*path++ = '\\';
	*path = 0;

	EXIT_FUNCTION();
}

/*
	DOS call 47H

	Get the current working directory in the form of
	"PATH\MOREPATH\LAST".  There will never be a preceding
	or trailing backlash.

	Sets error_no on error.  Returns nothing.
*/
void	PASCAL get_cwd(uchar drive)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern LPCDS sda_cdsptr;


	/*
		[ locals ]
	*/
	int len;
	char far * root;
	char cwdtest[67];


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	if(!defdrive_to_sda(drive))
	{
		error_no = ERR_NOTDRIVE;
		EXIT_FUNCTION();
		return;
	}

	/* if not a network drive, check for change of disk */
	root = sda_cdsptr->cwd + sda_cdsptr->skip;
	if(!(sda_cdsptr->status & C_NETWORK))
	{
		/*
			make sure to do a change of disk and "read" check
			in case this drive has never been accessed, but only
			on the root.  If we have already changed directories
			it has DEFINETLY been accessed.
		*/
		if(*root==0 || *(root+1)==0)		/* "" or "\" */
		{
			change_of_disk();

			//  You might ask yourself why this is here.  It is here because
			//  change of disk does not generate critical errors.  Basically
			//  we are utilizing the critical error handling of the
			//  get_fat_entry code.  This also marks all the variables to say
			//  that this drive is invalid.
			//
			//  If you don't have this here, you will not generate a critical
			//  error when requesting the cwd for the a: drive even when there
			//  is no disk in the drive.
			//
			get_fat_entry(2);
		}

		/* check subdirectory by name */
		else
		{
			/*
				dir_valid, requires ending filename and no starting
				"\".  any name will do, since only the path part is
				checked, so just use X.
			*/
			fmemcpy(cwdtest, sda_cdsptr->cwd + SKIPVAL + 1, sizeof(cwdtest)-1);
			memmove(cwdtest + strlen(cwdtest), "\\X", 3);

			/* change to root if dir not valid */
			if(!dir_valid(cwdtest))
				change_to_root();
		}
	}

	/* skip leading '\' on drive's CWD */
	len = fstrlen(root);
	if(len)
		++root;

	/* "network" drives might not have a '\' for root dir */
	else
		++len;

	/* copy the Current Working Directory into user space */
	fmemcpy(MK_FP(R_DS,R_SI),root,len);

	EXIT_FUNCTION();
}


/*
	-------------------------------------------------------

						Directory Management

	-------------------------------------------------------
*/


/*
	Create a new directory entry.  This gets the name from
	"sda_ename1" and the attribute from "sda_openmode".
	It also gives the entry the current date/time.  All
	other fields are set to 0.  Assumes that the entry has
	already been "found" so that it's location is known by
	"sda_sdb".
*/
void	PASCAL create_dir_entry(CLUSTER start)
{
	/*
		[ externals ]
	*/
	extern char sda_ename1[];
	extern uchar sda_openmode;
	extern DIR sda_ffb;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* fill out the new entry */
	memset(&sda_ffb,0,sizeof(DIR));
	memmove(sda_ffb.name,sda_ename1,11);

   // special character E5 handling
   //
   if( (uchar)sda_ffb.name[0] == (uchar)0xE5 )
   {
      // translate this character to a 0x05 for storage on the disk
      //
      sda_ffb.name[0] = 0x05;
   }
#if USING_FAT32
	sda_ffb.uStartingClusterLo = LOWORD(start);
	sda_ffb.uStartingClusterHi = HIWORD(start);
#else
	sda_ffb.start = start;
#endif
	sda_ffb.attr  = sda_openmode;

	/* and write it to disk */
	update_dir_entry(TRUE);

	EXIT_FUNCTION();
}


/*
	Remove a directory entry.  This is used both rmdir()
	and delete file(s).  The entry must have already been
	"found" and must be in the SDA file found block.
*/
void	remove_dir_entry(void)
{
	/*
		[ externals ]
	*/
	extern DIR sda_ffb;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* delete the clusters of this file/directory */
#if USING_FAT32
	free_the_fat(Fat32Start(sda_ffb));
#else
	free_the_fat(sda_ffb.start);
#endif

	/* mark the entry as deleted */
	(uchar)(*sda_ffb.name) = DELETED;
	update_dir_entry(FALSE);

	EXIT_FUNCTION();
}


/*
	Update the directory entry on the disk.  The directory
	entry to be updated must have been previously "found"
	so that its location is known to the SDA structure
	"sda_sdb".  The structure "sda_ffb" defines the entry.

	The buffer is flushed to disk!

*/
void	PASCAL update_dir_entry(bool setdate)
{
	/*
		[ locals ]
	*/
	struct DIRENTRY dirent;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* set the current date and time */
	if(setdate)
		get_fdate();

	/* get the sector into the current buffer */
	dirent.PhysicalCluster = 0;
	dir_sector(FALSE, &dirent);

	/* and save it out */
	save_dir_entry(dirent.dirent);

	EXIT_FUNCTION();
}


/*
	Given that the directory sector is in the current
	buffer, and given the entry within that sector, save
	the current DIR entry out to the disk.
*/
void	PASCAL save_dir_entry(uchar dirent)
{
	/*
		[ externals ]
	*/
	extern struct LOL lol;
	extern DIR sda_ffb;
	extern uchar int21_ah;
	extern LPBUFLIST sda_cur_buf;

	/* locals */
	LPDIR dir;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// assuming that sectors will never be larger than 2K,
	// make sure that the directory number is in range.
	ASSERT(dirent < 2048/sizeof(DIR));

	/* move the new data into the sector, mark it as dirty */
	dir = (void far *)sda_cur_buf->buffer;
	fmemcpy(dir+dirent, (void far *)&sda_ffb, sizeof(DIR));
	buf_set_modified(sda_cur_buf);

#if 0
	/*
		For speed purposes, DOS1 delete does its own flushing.
		Ouch! but this code is ugly!
	*/
	if(int21_ah != 0x13)
		buf_flush_all();		/* write entry and everything else to disk */
#else
	//  BT 6-12-02
	//
	if(int21_ah == 0x13)
	{
		//
		//  For speed purposes, DOS1 delete does its own flushing.
		//  So ... do nothing
		//
	}
	else if( int21_ah == 0x6c )
	{
		//  Only flush the current buffer.  A buf_flush_all will cause the
		//  FSINFO sector to be committed and in the case of truncating an
		//  existing file, this is not how MS does it.
		//
		buf_flush(sda_cur_buf);
	}
	else
	{
		//  This is where the buf_flush_all happens for directory
		//  operations, file closes, and deletes.  These things change
		//  the free space (FSINFO commit stuff).
		//
		buf_flush_all();		/* write entry and everything else to disk */
	}
#endif

	EXIT_FUNCTION();
}


/*
	Validate the whole directory path to make sure it really
	does exist on the drive in question.  The drive must be
	previously parsed into the SDA.  Assumes path is in the
	form of "\path\morepath\name".  Does no validating of
	the name part.  This routine also does a change of disk
	if needed.

	Returns FALSE if failed, TRUE if worked.
*/
bool	PASCAL dir_valid(char *path)
{
	/*
		[ externals ]
	*/
	extern SDB sda_sdb;
	extern DIR sda_ffb;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// make sure string exists
	ASSERT(path);

	/* make sure the disk that we want is really in here */
	change_of_disk();

	/* start searching at the root */
#if USING_FAT32
	sda_sdb.ulStartingCluster = 0;
#else
	sda_sdb.subdir   = 0;
#endif
	sda_sdb.srchattr = A_DIRCHK;
	while(1)
	{
		/* get the entry name from the path string */
		path = getename(sda_sdb.emask, path);

      // special character E5 handling
      //
      if( (uchar)sda_sdb.emask[0] == (uchar)0xE5 )
      {
         // Special translation to allow 0xE5 character in directory entry
         //
         sda_sdb.emask[0] = 0x05;
      }
		if(*path == 0)		/* we are on the name part, stop validating */
			break;
		++path;				/* skip the '\' in the string */

		/* the entry must exist */
		if(!dir_findfirst())
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		/* the entry must be a directory */
		if(! (sda_ffb.attr & A_SUBDIR) )
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		/* search the child directory */
#if USING_FAT32
		sda_sdb.ulStartingCluster = Fat32Start(sda_ffb);
#else
		sda_sdb.subdir = sda_ffb.start;
#endif
	}

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Create an absolute path from a relative one.
*/
int abspath(char far *dst, char far *src)
{
	extern LPCDS sda_cdsptr;
	extern char sda_fname1[];
	extern char sda_fname2[];
	extern int error_no;

	char c;
	int  iStrLen, iSkip, bSrcWasRoot;

	ENTER_FUNCTION();

	// make sure pointers exist
	ASSERT(dst);
	ASSERT(src);

	// 8-27-98 (de) limit copy to length of src path string.
	// If we don't we'll GP fault if (FP_OFF + MAXNAMELEN >= 0x10000L)
	// See test i21f60.c
	//
	iStrLen = 1 + fstrlen(src);
	if(iStrLen > MAXNAMELEN)
		iStrLen = MAXNAMELEN;

	// if the path ends in a backslash, we must fake out our
	// routines or we will get an error.  All this rigamarol
	// is for that one purpose.
	//
	fmemcpy((void far *)sda_fname2, src, iStrLen);
	c = sda_fname2[MAXNAMELEN] = 0;
	if(lastchar(sda_fname2) == '\\')
		*(unsigned *)(sda_fname2+strlen(sda_fname2)) = c = 'X';

	// can the path be made absolute??
	// if so, it will end up in sda_fname1
	//
	if(path_to_sda((void far *)sda_fname2,TRUE))
	{
		// more rigamarol for ending backslashes...
		// 08-27-98 (DE) w/ MS-DOS v6.22 if src ends in a '\'
		// 	then return a trailing '\'.
		//
		if(c)
		{
			bSrcWasRoot = ((--iStrLen == 3) && (src[1] == ':'));
			iStrLen     = strlen(sda_fname1);
			iSkip       = sda_cdsptr->skip;

			sda_fname1[iStrLen-1] = 0;	// get rid of the damned X

			// Special case of network and SUBST drives. DON'T
			// return a slash in these cases.
			//
			if(bSrcWasRoot && (iSkip > 2) && (sda_fname1[iStrLen-2] == '\\'))
			{
				sda_fname1[iStrLen-2] = 0;
			}
		}

		// move absolute string to user space
		fmemcpy(dst, (void far *)sda_fname1,strlen(sda_fname1)+1);
	}

	// error code depends upon starting character
	else
	{
		error_no = *sda_fname2 > ' ' ? ERR_NOTPATH : ERR_NOTFOUND;
		EXIT_FUNCTION();
		return error_no;
	}

	EXIT_FUNCTION();

	return 0;
}


/*
	Find the first occurence of a filespec defined by the
	"sda_sdb" Search Data Block.

	Returns TRUE if found, FALSE if not.
*/
bool	dir_findfirst(void)
{
	/*
		[ externals ]
	*/
	extern SDB sda_sdb;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	sda_sdb.dirnum = 0;

	EXIT_FUNCTION();

	return dir_findloop(FALSE);
}


/*
	Returns TRUE if found, FALSE if not.
*/
bool	dir_findnext(void)
{
	/*
		[ externals ]
	*/
	extern SDB sda_sdb;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	++sda_sdb.dirnum;

	EXIT_FUNCTION();

	return dir_findloop(FALSE);
}


/*
	Find a blank (or deleted) entry.  This will even get a
	new cluster if the subdirectory is full.

	Returns TRUE if it found a blank entry (the sda_sdb
	contains the entry #).  Returns FALSE if failed.
*/
bool	dir_findblank(void)
{
	/*
		[ externals ]
	*/
	extern SDB sda_sdb;
	extern uchar sda_drive;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	sda_sdb.dirnum = 0;

	EXIT_FUNCTION();

	return dir_findloop(TRUE);
}

/*
	optimization for create: find a blank entry using
	knowledge from previous findfirst
*/
bool	dir_findblank_create(void)
{
	extern unsigned sda_1st_blank;	/* optimization for create */
	extern SDB sda_sdb;

	ENTER_FUNCTION();

	/* if not invalid, 1st_blank has starting dirnum for speed */
	sda_sdb.dirnum = (sda_1st_blank == 0xFFFF) ? 0 : sda_1st_blank;

	EXIT_FUNCTION();

	return dir_findloop(TRUE);
}

/*
	Using the Search Data Block (sda_sdb) which can be found
	in the SDA, find a matching entry.  The "dirnum" field
	is updated and so is the File Found Block (sda_ffb).
	This routine can also search for next blank entry.  When
	searching for a blank entry, it can get new clusters for
	the directory.

	Returns TRUE if found, FALSE if not.
*/
bool	PASCAL dir_findloop(bool getblank)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;
	extern DIR sda_ffb;
	extern SDB sda_sdb;
	extern struct LOL lol;
	extern unsigned sda_1st_blank;
	extern LPBUFLIST sda_cur_buf;


	/*
		[ locals ]
	*/
	uchar  ent;
	uchar  firstchar;
	LPDIR  dir;
	struct DIRENTRY dirent;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* optimization for create - no blank entry found yet */
	if(!getblank)
		sda_1st_blank = 0xFFFF;

	dirent.PhysicalCluster = 0;

	while(1)
	{
		/* find the starting sector for this search */
		if(!dir_sector(getblank, &dirent))
			break;

		/* search the sector for a matching entry */
		dir = (void far *)buf_dataptr(sda_cur_buf);
		ent = dirent.dirent;
		dir += ent;
		for(; ent < dirent.eps; ent++)
		{
			/* we are at the end of the list, quit looking */
			firstchar = *dir->name;
			if(firstchar == 0)
			{
				/* optimization for create: remember 1st blank entry */
				if(sda_1st_blank == 0xFFFF)
					sda_1st_blank = sda_sdb.dirnum;

				EXIT_FUNCTION();
				return getblank;
			}

			/* check for deleted entries */
			if(IsDeleted(firstchar))
			{
				/* optimization for create: remember 1st blank entry */
				if(sda_1st_blank == 0xFFFF)
					sda_1st_blank = sda_sdb.dirnum;

				/* if we want a blank one anyway, we've got it */
				if(getblank)
				{
					EXIT_FUNCTION();
					return TRUE;
				}
			}
			/* was this the entry we were looking for? */
			else
			{
				//  Compare the name and attributs for a match
				//
				if( name_match(dir->name, sda_sdb.emask) &&
						attr_match(dir->attr,sda_sdb.srchattr) )
				{
					//  Get the directory entry into the internal structures
					//
					fmemcpy((void far *)&sda_ffb,dir,sizeof(DIR));

               // special character E5 handling
               //
               if( firstchar == 0x05 )
               {
                  // translate this character back to an 0xE5 for matching
                  //
                  sda_ffb.name[0] = 0xE5;
               }

					EXIT_FUNCTION();
					return TRUE;
				}
			}

			/* look at the next entry */
			++dir;
			++sda_sdb.dirnum;
		}

	}

	EXIT_FUNCTION();

	return FALSE;
}


/*
	Convert the "sda_sdb" fields into an absolute sector #
	on the disk.  This uses the "subdir" and "dirnum"
	fields.  Also counts on the DPB for current drive.

	This function can also allocate new clusters to the
	directory chain (only for subdirs) if "getnew" is true
	and .

	Returns TRUE if worked, FALSE if failed.
*/
bool	PASCAL dir_sector(bool getnew, struct DIRENTRY *dirent)
{
	/*
		[ externals ]
	*/
	extern SDB sda_sdb;
	extern LPDPB sda_dpbptr;


	/*
		[ locals ]
	*/
	unsigned lsect;	/* assumes <64K sectors in any given directory */
	CLUSTER lclus;
	CLUSTER pclus;
	CLUSTER clusleft;
	unsigned spc;
#if USING_FAT32
	ulong ulSubDirectoryCluster;
#endif

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// make sure pointer is ok
	ASSERT(dirent);

	/* calculate the logical sector that contains this dirnum */
	dirent->eps = (sda_dpbptr->bytes_per_sec / sizeof(DIR));
	dirent->dirent = sda_sdb.dirnum % dirent->eps;
	lsect	= sda_sdb.dirnum / dirent->eps;

	/* this is the root directory, treat it special */
#if USING_FAT32

	//  If this is the root directory on a FAT32 drive then pretent
	//  it is a subdirectory for a moment.
	//
	if( (IsFat32Drive(sda_dpbptr)) && (sda_sdb.ulStartingCluster == 0) )
	{
		ulSubDirectoryCluster = sda_dpbptr->ulRootDirStartCluster;
	}
	else
	{
		ulSubDirectoryCluster = sda_sdb.ulStartingCluster;
	}

	if(ulSubDirectoryCluster == 0)
	{
		if(sda_sdb.dirnum >= sda_dpbptr->dir_ent)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		/* assume root directory is less than 64K worth of sectors */
		dirent->dirsec = (ulong)(sda_dpbptr->dir_start + lsect);
		buf_get_dir_sector(dirent->dirsec);
		EXIT_FUNCTION();
		return TRUE;
	}

	/* convert logical sector to logical cluster */
	spc	= 1 << sda_dpbptr->shift;
	lclus	= lsect / spc;

	if( (!(dirent->PhysicalCluster)) || (lclus < dirent->LogicalCluster) )
	{
		dirent->PhysicalCluster = ulSubDirectoryCluster;
		dirent->LogicalCluster = 0;
	}
	clusleft = lclus - dirent->LogicalCluster;

#else  // #if USING_FAT32

	if(sda_sdb.subdir == 0)
	{
		if(sda_sdb.dirnum >= sda_dpbptr->dir_ent)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		/* assume root directory is less than 64K worth of sectors */
		dirent->dirsec = (ulong)(sda_dpbptr->dir_start + lsect);
		buf_get_dir_sector(dirent->dirsec);
		EXIT_FUNCTION();
		return TRUE;
	}

	/* convert logical sector to logical cluster */
	spc	= 1 << sda_dpbptr->shift;
	lclus	= lsect / spc;

	if( (!(dirent->PhysicalCluster)) || (lclus < dirent->LogicalCluster) )
	{
		dirent->PhysicalCluster = sda_sdb.subdir;
		dirent->LogicalCluster = 0;
	}
	clusleft = lclus - dirent->LogicalCluster;

#endif // #if USING_FAT32

	/* get the physical sector containing dir entry */
	if( clusleft )
	{
		pclus = lclus_to_pclus(dirent->PhysicalCluster, &clusleft);
	}
	else
	{
		pclus = dirent->PhysicalCluster;
	}

	if( !clusleft )
	{
		dirent->PhysicalCluster = pclus;
		dirent->LogicalCluster = lclus;
	}


	/* we are past the end of the directory */
	if(clusleft)
	{
		PERMASSERT(clusleft <= 1);

		/* try to get a new cluster if we need to */
		if(!getnew || get_new_clusters(&pclus,clusleft) != clusleft)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		/* zero out the new cluster */
		pclus = lclus_to_pclus(pclus, &clusleft);
		if( !clusleft )
		{
			dirent->PhysicalCluster = pclus;
			dirent->LogicalCluster = lclus;
		}
		zero_cluster(pclus);
	}

	/* determine physical sector from physical cluster */
	dirent->dirsec = clus_to_sec(pclus) + (lsect % spc);

	/* get the sector into currentbuf */
	buf_get_dir_sector(dirent->dirsec);

	EXIT_FUNCTION();

	// SPR1638 (de) If we are doing an extended open with NO_I24,
	// (in which case dev_crit_call() sets error_no = 3) we could
	// end up here with garbage in the buffer and think we had
	// something. If we don't, say we don't.
	//
	return (error_no == 0);
}


/*
	Zero out the cluster in question.  Only needed for
	subdirectories.
*/
void	PASCAL zero_cluster(CLUSTER clus)
{
	/*
		[ externals ]
	*/
	extern struct LOL lol;
	extern LPDPB sda_dpbptr;
	extern LPBUFLIST sda_cur_buf;


	/*
		[ locals ]
	*/
	long sector;
	uchar spc;
	uchar i;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// make sure cluster is valid.  A valid cluster is always 2 or
	// greater.
   ASSERT_ValidCluster( clus );

	sector = clus_to_sec(clus);
	spc = 1 << sda_dpbptr->shift;

	// SPR 1638 (de)
	// Zero out each sector in the new cluster. Start at the top so
	// that when we are done, sda_cur_buf points to the first sector
	// in the cluster. This allows us to build the DIR "." and ".."
	// entries in mkdir() without going through any other routines.
	//
	// SPR 1832 (de)
	// We must flush these entries so that if the subdir lies over a
	// bad spot in the disk that we fail the i21h call without losing
	// clusters in the process.
	//
	for( i = 1; i < spc; i++ )
	{
		buf_getnoread(sector+i);
		buf_mark_as_dir(sda_cur_buf);
		fmemset(buf_dataptr(sda_cur_buf),0,lol.bufsize);
		buf_set_modified(sda_cur_buf);
	}

	//  BT 6-13-02
	//
	//  The buffers should not be flushed at this level.
	//
	//	 buf_flush_all();

	buf_getnoread(sector);
	buf_mark_as_dir(sda_cur_buf);
	fmemset(buf_dataptr(sda_cur_buf),0,lol.bufsize);
	buf_set_modified(sda_cur_buf);

	EXIT_FUNCTION();
}


/*
	Compare two expanded names.  The "emask" name can have
	wildcards, but the other cannot.

	Returns TRUE if they match, FALSE if not.
*/
bool PASCAL name_match(char far *ename,char far *emask )
{
	int i;

	ENTER_FUNCTION();

	// make sure pointers exist
	ASSERT(ename);
	ASSERT(emask);

	for(i=0;i<11;++i)
	{
		if(*emask!='?' && *emask!=*ename)
		{
			EXIT_FUNCTION();
			return FALSE;
		}
		++ename;
		++emask;
	}

	EXIT_FUNCTION();

	return TRUE;
}


/*
	Modify a name based on the mask.  For example:
	ename		TESTING 123
	emask		??XX????N??

	result	TEXXING N23
*/
void	PASCAL name_change(char *ename, char *emask)
{
	int i;

	ENTER_FUNCTION();

	// make sure pointers exist
	ASSERT(ename);
	ASSERT(emask);

	for(i=0; i<11; ++i)
	{
		if(*emask != '?')
			*ename = *emask;
		++ename;
		++emask;
	}

	EXIT_FUNCTION();
}


/*
	If the mask has at least as many bits as the attribute
	then it matches.  Volume is special case.
	Returns TRUE if they match, FALSE if not.
*/
bool	PASCAL attr_match(uchar attr, uchar mask)
{
	ENTER_FUNCTION();

	//  BT 2-1-2000 ETR 3271
	//
	//  Short name find first calls should not see LFN entries
	//
	if( attr == 0x0f )
	{
		//  Don't match LFN entries
		//
      EXIT_FUNCTION();
		return FALSE;
	}

	attr &= A_ALL;
	mask &= A_ALL;

	if(mask == A_VOLUME)
	{
		if(attr == A_VOLUME)
      {
         EXIT_FUNCTION();
			return TRUE;
      }
	}
	else if( ! ((mask & attr) ^ attr) )
	{
		EXIT_FUNCTION();
		return TRUE;
	}

	EXIT_FUNCTION();

	return FALSE;
}


/*
	Extract and expand a portion of a valid path.  The
	path must be prevalidated.   Does no disk access, but
	only simple memory manipulation.

	Returns a pointer to the '\' or the NULL following the
	name.
*/
char	* PASCAL getename(char *ename,char *path)
{
	register char * p = path;

	ENTER_FUNCTION();

	// make sure pointers are valid
	ASSERT(ename);
	ASSERT(path);

	// expand the current name part
	dot_to_ename(ename,(char far *)path, FALSE);

	// search for a '\' or the end of the string
	while(*p && *p != '\\')
	{
		if(is_lead_byte(*p))
			++p;
		++p;
	}

	EXIT_FUNCTION();

	return p;
}

/*
	All volume labels are in the root directory.
*/
void prepare_volume_search(void)
{
	extern SDB sda_sdb;

	ENTER_FUNCTION();

#if USING_FAT32
	sda_sdb.ulStartingCluster = 0;
#else
	sda_sdb.subdir = 0;
#endif
	sda_sdb.srchattr = A_VOLUME;
	memset(sda_sdb.emask,'?',11);

	EXIT_FUNCTION();
}
