//
//   $Id: DOSENTRY.C 1.20 2002/10/15 21:44:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	DOSENTRY.C  created 01/14/91  by Drew Gislason

   Copyright (c) 1991-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains the big switch table for INT 21H
	calls.

**********************************************************/
#define __FILENUM__ 7

#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#include "farlfn.h"

#if USING_FAT32

bool PASCAL GetSetFat32FormatInformation(LPSFS lpsfsFormatStruct);

#endif

#if DEBUG > 2
int do_debug;
#endif

#if 0

// handy debugging routines for dumping the user's registers

/*
	Display routine.
*/
void bputs(char *s)
{
	ENTER_FUNCTION();

	while(*s)
		bios_putc(*s++);

	EXIT_FUNCTION();
}


/*
	Displays a 16-bit hex number in 0000 form.
*/
void bputh(unsigned val)
{
	char buff[5];
	int i;
	char c;

	ENTER_FUNCTION();

	for(i=3; i>=0; --i)
	{
		c = val % 0x10;
		c = (c >= 10 ? c-10+'A' : c+'0');
		buff[i] = c;
		val /= 0x10;
	}
	buff[4] = 0;
	bputs(buff);

	EXIT_FUNCTION();
}

void dump_regs(union dos_regs far *p)
{
	ENTER_FUNCTION();

	bios_putc('\n'); bios_putc('\r');
	bputs("AX   : "); bputh(p->x.ax   ); bios_putc('\n'); bios_putc('\r');
   bputs("BX   : "); bputh(p->x.bx   ); bios_putc('\n'); bios_putc('\r');
	bputs("CX   : "); bputh(p->x.cx   ); bios_putc('\n'); bios_putc('\r');
	bputs("DX   : "); bputh(p->x.dx   ); bios_putc('\n'); bios_putc('\r');
	bputs("SI   : "); bputh(p->x.si   ); bios_putc('\n'); bios_putc('\r');
	bputs("DI   : "); bputh(p->x.di   ); bios_putc('\n'); bios_putc('\r');
	bputs("BP   : "); bputh(p->x.bp   ); bios_putc('\n'); bios_putc('\r');
	bputs("DS   : "); bputh(p->x.ds   ); bios_putc('\n'); bios_putc('\r');
	bputs("ES   : "); bputh(p->x.es   ); bios_putc('\n'); bios_putc('\r');
	bputs("IP   : "); bputh(p->x.ip   ); bios_putc('\n'); bios_putc('\r');
	bputs("CS   : "); bputh(p->x.cs   ); bios_putc('\n'); bios_putc('\r');
	bputs("FLAGS: "); bputh(p->x.flags); bios_putc('\n'); bios_putc('\r');

	EXIT_FUNCTION();
}

void dump_userregs(void)
{
	ENTER_FUNCTION();

	dump_regs(userregs);

	EXIT_FUNCTION();
}

#endif // stack dumping debugging rountines


/*
	Every INT 21H call is passed through this function...
	well, every INT 21H call that can't be handled by the
	fully reentrant routines anyway.
*/
void	dos_entry(void)
{
	/*
		[ externals ]
	*/
	extern uchar int21_ah;
	extern char sda_isnet;
	extern char in_windows;
	extern uchar InDOS;
	extern int error_no;
	extern uchar BREAK_FLAG;
	extern uchar CritErr;
	extern unsigned dos_win_ver;
	extern unsigned dos_ver;
	extern unsigned machine_id;
	extern uchar sda_in_5D;
	extern unsigned sda_handle;
	extern void far * fpFarLFNCall;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

#if DEBUG > 2
 if(do_debug)
 {
 printf("%p ",R_AX);
 if(InDOS > 1)
	printf("in%d ",InDOS);
 con_getkey();
 }
#endif

	/* reset the error # */
	if(!in_windows)
		machine_id = 0;			// kludge for Windows 3.1
	dos_win_ver = dos_ver;		// kludge for Windows 3.1
	sda_isnet = error_no = 0;

	/* when BREAK is ON, check for ^C */
	if(BREAK_FLAG && !CritErr && R_AH>0x0C)
		ctrlc_check(0);

	/* long filename call? go do it, and we're done */
	if(LONGNAME_FUNCTIONS())
	{
		EXIT_FUNCTION();
		dos_exit();
	}

	/* convert the user registers into SDA variables */
	regs_to_sda();

	if(!error_no)
	{
		/* an Installable File System may take this call */
		if(ifs_call())
			/* the damage is already done */;

		/* FAT style calls */
		else
		{
			/* go out to the disk and verify paths, etc */
			validate_sda();

			/* if we had no problem, do the DOS call */
			if(!error_no)
				normal_call();
		}
	}

	if(error_no)
	{
		// FINDFIRST:
		// If the find first fails with the DOS style file name,
		// try it again with the \\SERVER\PATH style name.
		// Only try this if the user requested that style path.
		if(int21_ah==0x4E)
		{ if(*(unsigned far *)MK_FP(R_DS,R_DX)==0x5C5C)
		  {
				if(!ifs_findfirst())
					error_no = ERR_LASTFILE;
		  }
		  else if(sda_in_5D)
		  {
				// MS network/Win95 have to have this error level
				// instead of "path not found" or they will not log
				// into the root
				//
				error_no = ERR_LASTFILE;
		  }
		}
		else if(int21_ah==0x3E)
		{
			// even if there was an error closing the file
			// free the handle for DOS2 calls (even IFS)
			//
			set_handle_val(sda_handle, 0xFF);
		}
	}


	/* send the data back to the user */
	if(!error_no)
	{
		sda_to_regs();
	}

	EXIT_FUNCTION();
	dos_exit();
}


/*
	Get out of DOS and back to user program.  Checks for
	"error_no" and acts appropriately.
*/
void	dos_exit(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar int21_ah;
	extern unsigned sda_callflags;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* Enable interrupts */
	R_FLAGS |= F_INT;



//	/* Clear all flags except I, Z, C, D. */
//	R_FLAGS &= (F_ZERO | F_INT | F_DIR | F_CARRY );
//
// What app demanded that we clear these other flags?
// Be sure not to change any of the 8086 reserved flags. Changing
// these flags will hang the NEC V25 chips.
//
	R_FLAGS &= ~(F_OVERFLOW | F_TRAP | F_SIGN | F_AUXCARRY | F_PARITY);


	/* mark all dirty buffers as "old" dirtness */
	buf_old_dirty();

	/* hey, no problem, man!  like, clear the carry and return... */
	if(!error_no)
	{
		/* DOS1 style error */
		if(callf_fcb())
			R_AL = 0;

		R_FLAGS &= ~F_CARRY;
	}

	/* whoa dude!  we gotta let the user know about this problem... */
	else
	{
		/* set the extended error if need be */
		default_error_codes();

		/* DOS1 style error */
		if(int21_ah <= DOS1CALL)
		{
			R_AL = 0xFF;
			if(callf_rdwr())
				R_AL = error_no;
			if(int21_ah == 0x36)
				R_AX = 0xFFFF;
		}

		/* DOS2 style error */
		else
		{
			R_FLAGS |= F_CARRY;
			R_AX     = error_no;
		}
	}


	EXIT_FUNCTION();

	dos_return();
}

/*
	Set extended error code from passed parameters
*/
void set_extended_error( unsigned error, unsigned extended, unsigned action, uchar locus)
{
	extern int error_no;
	extern int err_extended;
	extern uchar err_locus;
	extern unsigned err_action;
	extern uchar err_already_set;

	ENTER_FUNCTION();

	// Set error values
	error_no = error;
	err_extended = extended;
	err_action = action;
	err_locus = locus;

	// don't just use DEFAULT error codes
	err_already_set = TRUE;

	EXIT_FUNCTION();
}


/*
	Set extended error code based on error_no.
*/
void	default_error_codes(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern int err_extended;
	extern uchar err_locus;
	extern unsigned err_action;
	extern uchar err_already_set;


	/*
		[ locals ]
	*/
	int err;
	static unsigned action[] =		/* and class */
	{
		0x0000,0x0704,0x0803,0x0803,0x0104,0x0303,0x0704,0x0704,
		0x0104,0x0704,0x0704,0x0903,0x0704,0x0704,0x0704,0x0803,
		0x0303,0x0D03,0x0803,0x0507,0x0000,0x0000,0x0000,0x0000,
		0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0D04
	};

	static uchar locus[]  =
	{
		0x00,0x01,0x02,0x02,0x01,0x02,0x01,0x01,
		0x05,0x05,0x05,0x01,0x01,0x01,0x01,0x02,
		0x02,0x02,0x02,0x02,0x00,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x02
	};


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// set extended error already called
	if(err_already_set)
	{
		err_already_set = FALSE;
		EXIT_FUNCTION();
		return;
	}

	// Check for invalid ERR codes
	if(error_no >= ERR_WPROTECT)
		err = ERR_DENIED;
	else
		err = error_no;

	err_extended = error_no;
	err_locus  = locus[err];
	err_action = action[err];

	EXIT_FUNCTION();
}

void	compute_FCB_records(void)
{
	/*
		[ externals ]
	*/
	extern FCB sda_fcb;
	extern DIR sda_ffb;

	/*
		[ locals ]
	*/
	ulong ltmp;
	unsigned tmp;

	ENTER_FUNCTION();

	tmp = sda_fcb.record_size;
	if(!tmp)
		tmp = 0x80;
	ltmp = _ldivu(sda_ffb.size, tmp);
	if((unsigned)(sda_ffb.size) % tmp)
		++ltmp;
	sda_fcb.rand_record = ltmp;

	EXIT_FUNCTION();
}


/*
	This is the BIG SWITCH inside of ROM-DOS.  Every command
	conceivable to the common DOS user can be found here.

	Sets "error_no" if there was a problem.
*/
void	normal_call(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern struct COUNTRY_INFO country_info;
	extern struct LOL lol;
	extern uchar far * DTA;
	extern unsigned ROM_DS;
	extern unsigned return_code;
	extern unsigned dos_vars;
	extern uchar int21_ah;
	extern uchar int21_al;
	extern unsigned int21_ax;
	extern unsigned end_sda;
	extern uchar InDOS;
	extern uchar CritErr;
	extern uchar hitbrk;
	extern uchar mem_fit;
	extern uchar umb_link;
	extern uchar verify_flag;
	extern uchar CUR_DRIVE;
	extern uchar switch_char;
	extern uchar sda_sft_h;
	extern char	sda_fname1[];
	extern char	sda_fname2[];
	extern char	machine_name[];
	extern FCB sda_fcb;
	extern DIR sda_ffb;
	extern LPSFT sda_sftptr;
	extern uchar sda_openmode;
	extern unsigned current_psp;
	extern char fatal_unsupp[];
	extern int use_stdout;
	extern uchar lead_byte_table[];
	extern uchar sda_chardev;
	extern uchar sda_srchattr;
	extern SDB sda_sdb;
	extern unsigned dos_ver;

	/*
		[ locals ]
	*/
	static char cfuncs[] = { 1,6,7,8,0xA, 0 };
	char far *fptr;
	uchar c;
	uchar doscall;
#if NOT_IN_DOSINT
	ulong ltmp;
#endif
	unsigned tmp;
	union EXEPARM far *pEP;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* actually do the thing */
	doscall = int21_ah;
again:
	switch(doscall)
	{
		/* Terminate Program */
		case 0x00:
         EXIT_FUNCTION();
			terminate(0,0);
			break;


		/* Keybd Input w/ Echo (^C) */
		case 0x01:
			c = readchk();
			use_stdout = TRUE;
			write_w_tabs(c);
			use_stdout = FALSE;
			R_AL = c;
			break;


		/* Auxiliary Input (^C) */
		case 0x03:
			R_AL = readchk();
			break;


		/* Display Output (^C) */
		case 0x02:

		/* Auxiliary Output (^C) */
		case 0x04:

		/* Printer Output (^C) */
		case 0x05:
			con_ctrlc_check();
			write_w_tabs(R_DL);
			break;


		/* Direct Console I/O */
		case 0x06:
			c = R_DL;
			if(c == 0xFF)
			{
				do_int28();
				if(ndread1byte() >= 0)
				{
					R_FLAGS &= ~F_ZERO;
					R_AL = read1byte();
				}
				else
				{
					R_FLAGS |= F_ZERO;
				}
			}
			else
			{
				write1byte(c);
				R_AL = c;
			}
			break;


		/* Direct Console Input */
		case 0x07:
			R_AL = read1byte();
			break;


		/* Keybd Input, No Echo (^C) */
		case 0x08:
			R_AL = readchk();
			break;


		/* Print String (^C) */
		case 0x09:
			/* the string is terminated by a '$' and cannot wrap */
			fptr = MK_FP(R_DS, R_DX);
			while(*fptr != '$' && FP_OFF(fptr)<0xFFFF)
			{
				ctrlc_check(0);
				write_w_tabs(*fptr++);
			}
			break;


		/* String Input (^C) */
		case 0x0A:
			use_stdout = TRUE;
			string_input(MK_FP(R_DS, R_DX));
			use_stdout = FALSE;
			break;


		/* Check Keybd Status (^C) */
		case 0x0B:
			if((ndread1byte() < 0) && !hitbrk)
			{
				R_AL = 0;
			}
			else
			{
				ctrlc_check(0);
				R_AL = (ndread1byte() >= 0 ? 0xFF : 0);
			}
			break;


		/* Flush Keybd, Then I/O (^C) */
		case 0x0C:
			flushbytes();
			if(int21_al && strchr(cfuncs,int21_al))
			{
				doscall = int21_al;
				goto again;
			}
			break;


		/* Disk Reset */
		case 0x0D:
			buf_flush_all();
			break;


		/* Set Current Drive */
		case 0x0E:
			R_AL = set_current_drive(R_DL);
			break;


		/* DOS1 (FCB) Open */
		case 0x0F:
			file_open();
			break;


#if 0	/* at this level, this is identical to 0x3E (below) */
		/* DOS1 (FCB) Close */
		case 0x10:
			file_close();
			decr_handle_count();
			break;
#endif


		/* DOS1 (FCB) Find First */
		case 0x11:
			find_first();
			break;


		/* DOS1 (FCB) Find Next */
		case 0x12:
			find_next();
			break;


		/* DOS1 (FCB) Delete File(s) */
		case 0x13:
			file_delete();
			break;


		/* DOS1 (FCB) Sequential Read */
		/* all file reads are identical at this point */
		case 0x14:
		case 0x21:
		case 0x27:
		case 0x3F:
			file_read();
			break;


		/* DOS1 (FCB) Sequential Write */
		/* all file writes are identical at this point */
		case 0x15:
		case 0x22:
		case 0x28:
		case 0x40:
			file_write();
			break;


		/* DOS1 (FCB) Create File */
		case 0x16:
			file_create();
			break;


		/* DOS1 (FCB) & DOS2+ Rename File(s).  Wild cards are allowed
			with function 56h only if invoked via function 5Dxxh when
			the CL register will contain file attributes to match.
		*/
		case 0x17:
		case 0x56:
			file_rename();
			break;


#if NOT_IN_DOSINT
		/* Get Current Drive */
		case 0x19:
			R_AL = CUR_DRIVE;
			break;
#endif

#if NOT_IN_DOSINT
		/* Set DTA */
		case 0x1A:
			DTA = MK_FP(R_DS, R_DX);
			break;
#endif

		/* Get Disk Info (Default Drive) */
		case 0x1B:
			R_DL = 0;	/* and fall thru... */


		/* Get Disk Info (Any Drive) */
		case 0x1C:
			disk_freespace(FALSE);
			break;


		/* Get Default DPB */
		case 0x1F:
			R_DL = 0;
			goto func32;


#if 0		/* these are identical to 0x14 (coded above) */

		/* DOS1 (FCB) Random Read */
		case 0x21:
			file_read();
			break;


		/* DOS1 (FCB) Random Write */
		case 0x22:
			file_write();
			break;
#endif


		/* Get File Size (in records) */
		case 0x23:
			compute_FCB_records();
			break;


		/* DOS1 (FCB) Set Random Record */
		case 0x24:
			sda_fcb.rand_record = (sda_fcb.cur_blk * 128) + sda_fcb.cur_record;
			break;


#if NOT_IN_DOSINT

		/* Set Interrupt Vector */
		case 0x25:
			set_vector(int21_al, MK_FP(R_DS, R_DX));
			break;

#endif


		/* Create New PSP */
		case 0x26:
			create_new_psp(R_DX,FALSE,FALSE);
			break;


#if 0	/* at this level, these are identical to 0x14 and 0x15 */

		/* DOS1 (FCB) Random Blk Read */
		case 0x27:
			file_read();
			break;


		/* DOS1 (FCB) Random Blk Write */
		case 0x28:
			file_write();
			break;

#endif


		/* DOS1 (FCB) Parse Filename */
		case 0x29:
			file_parse();
			break;


		/* Get Date */
		case 0x2A:
			R_AL = get_date();
			break;


		/* Set Date */
		case 0x2B:
			R_AL = set_date(R_CX, R_DX);
			break;


		/* Get Time */
		case 0x2C:
			get_time();
			break;


		/* Set Time */
		case 0x2D:
			R_AL = set_time(R_CX, R_DX);
			break;


#if NOT_IN_DOSINT
		/* Set Verify Flag */
		case 0x2E:
			verify_flag = int21_al;
			break;
#endif


#if NOT_IN_DOSINT
		/* Get DTA */
		case 0x2F:
			R_ES = FP_SEG(DTA);
			R_BX = FP_OFF(DTA);
			break;
#endif



		/* Terminate, Stay Resident */
		case 0x31:
			if( R_DX < 6 )
			{
				//  ETR 3626  BT 12-14-2000
				//
				//  The minimum amount of paragraphs you can release
				//  is 6.  Verified with MS-DOS 6.22.  This was
				//  discovered while debugging ETR 3622.
				//
				EXIT_FUNCTION();
				terminate(int21_al, 6 );
			}
			else
			{
				EXIT_FUNCTION();
				terminate(int21_al, R_DX);
			}
			break;


		/* Get Any DPB */
		case 0x32:
func32:
			/* does a change of disk if it needs to */
			R_AX = get_dpb(R_DL);
			break;


#if NOT_IN_DOSINT

		/* Get/Set Break Flag */
		case 0x33:
			if(int21_al)
				BREAK_FLAG = R_DL;
			R_DL = BREAK_FLAG;
			break;

#endif

#if NOT_IN_DOSINT
		/* Get InDOS Flag Address */
		case 0x34:
			R_ES = ROM_DS;
			R_BX = (unsigned)&InDOS;
			break;
#endif

#if NOT_IN_DOSINT

		/* Get Interrupt Vector */
		case 0x35:
			fp = get_vector(int21_al);
			R_BX = FP_OFF(fp);
			R_ES = FP_SEG(fp);
			break;

#endif

		/* Get Disk Free Space */
		case 0x36:
			disk_freespace(TRUE);
			R_AX = (unsigned char)R_AL;
			break;


		/* Get/Set Switch Character */
		case 0x37:
			if(int21_al == 0)
				R_DL = switch_char;
			else if(int21_al == 2)
				R_DL = 0xFF;
			else if(int21_al > 3)
				error_no = ERR_BADFUNC;
			break;


		/* Get/Set Country Info */
		case 0x38:
			get_country_info();
			break;


		/* Create Directory */
		case 0x39:
			mkdir();
			break;


		/* Remote Directory */
		case 0x3A:
			rmdir();
			break;


		/* Change Current Directory */
		case 0x3B:
			chdir();
			break;


		/* Create File */
		case 0x3C:
			file_create();
			break;


		/* Open File */
		case 0x3D:
			file_open();
			break;


		case 0x10:           // DOS1 (FCB) Close
		case 0x3E:				// Close File Handle
			//
			// SPR 1638 (de) MS-DOS decs SFT handle count prior to calling
			// the device (to update FAT, DIR, data). See I24DATA.C.
			//
			decr_handle_count();
			//
			// fall through
			//
		case 0x68:				// Flush File Buffers (commit)
			file_close();		// updates dir entry and flushes data to disk
			break;


#if 0	/* at this level, these are identical to 0x14 and 0x15 */

		/* Read from File */
		case 0x3F:
			file_read();
			break;

		/* Write to File */
		case 0x40:
			file_write();
			break;
#endif


		/* Delete a File */
		case 0x41:
			file_delete();
			break;


#if NOT_IN_DOSINT
		/* Seek Within File */
		case 0x42:
			ltmp = file_seek((ulong)MK_FP(R_CX,R_DX), int21_al);
			R_AX = (unsigned)ltmp;
			R_DX = (unsigned)(ltmp>>16);
			break;
#endif


		/* Get/Set File Attribute */
		case 0x43:
			if(int21_al == 0)
				R_AX = R_CX = sda_ffb.attr;
			else if(int21_al == 1)
				file_setattr(R_CL);
			else
				error_no = ERR_BADFUNC;
			break;


		/* I/O Control */
		case 0x44:
			ioctl(int21_al, R_CX);
			break;


		/* Duplicate File Handle */
		case 0x45:
			R_AX = dup_handle(sda_sft_h);
			break;


		/* Force Duplicate Handle */
		case 0x46:
			file_force_dup(R_BX,R_CX);
			break;


		/* Get Current Directory */
		case 0x47:
			get_cwd(R_DL);
			if(!error_no)
				R_AX = 0x100;
			break;


		/* Allocate Memory Block */
		case 0x48:
			tmp = mem_alloc(R_BX);
			if(error_no)
				R_BX = tmp;		/* avail */
			else
				R_AX = tmp;		/* segment */
			break;


		/* Free Memory Block */
		case 0x49:
			if(R_ES != 0)								// allow asserts to be used
				R_AX = mem_free(R_ES);
			else
				error_no = ERR_BADMEMSEG;
			break;


		/* Resize Memory Block */
		case 0x4A:
			if(R_ES == 0)								// allow asserts to be used
			{
				error_no = ERR_BADMEMSEG;
				break;
			}
			tmp = mem_resize(R_ES, R_BX);
			if(error_no == ERR_NOMEM)
				R_BX = tmp; 	/* avail */
			else
				R_AX = tmp;		/* new size */
			break;


		/* Execute Process */
		case 0x4B:

			/* Set execution state */
			if( int21_al == 5 )
			{
				/* Parameter Block passed in DS:DX */
				pEP = MK_FP(R_DS,R_DX);
				tmp = pEP->es.pspseg;
				fptr = MK_FP( tmp, 0x40 );
				*(unsigned far *)fptr = dos_ver;
				R_AX = 0;
			}

			// normal exec call
			else
				exec(MK_FP(R_ES,R_BX), MK_FP(R_CS,R_IP));
			break;


		/* Terminate Process */
		case 0x4C:
			EXIT_FUNCTION();
			terminate(int21_al, 0);
			break;


#if NOT_IN_DOSINT
		/* Get Return Code */
		case 0x4D:
			R_AX = return_code;
			return_code = 0;
			break;
#endif

		/* Find First */
		case 0x4E:
			find_first();
			if(sda_chardev && (sda_srchattr&A_VOLUME))
				error_no = ERR_LASTFILE;
			break;


		/* Find Next */
		case 0x4F:
			find_next();
			break;


#if NOT_IN_DOSINT
		/* Set PSP */
		case 0x50:
			current_psp = R_BX;
			break;


		/* Get PSP */
		case 0x51:
			R_BX = current_psp;
			break;
#endif


#if NOT_IN_DOSINT
		/* Get List of Lists */
		case 0x52:
			R_ES= ROM_DS;
			R_BX= (unsigned)(&dos_vars);
			break;
#endif


		/* Translate BPB to DPB */
		case 0x53:
			bpb_to_dpb(MK_FP(R_DS,R_SI), MK_FP(R_ES,R_BP));
			break;


#if NOT_IN_DOSINT
		/* Get Verify Flag */
		case 0x54:
			R_AL = verify_flag;
			break;
#endif


		/* Create Child PSP */
		case 0x55:
			create_new_psp(R_DX,FALSE,TRUE);
			current_psp = R_DX;
			break;


		/* Get/Set File Date/Time */
		case 0x57:
			if(int21_al == 0)
			{
				R_CX = sda_sftptr->time;
				R_DX = sda_sftptr->date;
			}
			else if(int21_al == 1)
				file_setdate(R_CX,R_DX);
			else
				error_no = ERR_BADFUNC;
			break;


		/* Get/Set Allocation Strategy */
		case 0x58:

			switch(int21_al)
			{
				/* Get Allocation Strategy */
				case 0:	R_AX = mem_fit; break;

				/* Get UMB link Status */
				case 2:	R_AL = umb_link; break;

				/* Set Allocation Strategy */
				case 1:
					c = R_BL;
					if((c&0x3F) <= 2)
					{
						mem_fit = c;
						break;
					}
					/* else fall thru (bad func) */

				default: goto badfunc;
			}
			break;


#if NOT_IN_DOSINT
		/* Get Extended Error */
		case 0x59:
			break;
#endif


		/* Create Unique File */
		case 0x5A:
			file_create_unique();
			break;


		/* Create New File */
		case 0x5B:
			file_create_new();
			break;


		/* Network Calls (Get/Set Machine Name, etc) */
		case 0x5E:

			/* Get/Set Machine Name */
			if(int21_al <= 1)
			{
				R_AX = 0;
				fptr = MK_FP(R_DS,R_DX);
				if(int21_al)
				{
					lol.netbios_id = R_CX;
					fmemcpy((void far *)machine_name, fptr, 0x10);
				}
				else
				{
					R_CX = lol.netbios_id;
					fmemcpy(fptr, (void far *)machine_name, 0x10);
				}
				break;
			}
			/* <<< We DON'T support 5E02H, 5E03H (Get/Set Printer) >>> */
			/* fall thru to bad func... */

		/* Lock/Unlock File Access (network) */
		case 0x5C:
badfunc:
			error_no = ERR_BADFUNC;
			break;


		/* Multifunction Network Calls */
		case 0x5D:

			/* Get Critical Error Flag Address (which happens to begin SDA) */
			if(int21_al == 6)
			{
				R_DS = ROM_DS;
				R_SI = (unsigned)(&CritErr);
				R_CX = (unsigned)(&end_sda)  - (unsigned)(&CritErr);
				R_DX = (unsigned)(&int21_ax) - (unsigned)(&CritErr);
			}
#if NOT_IN_DOSINT
			/*	Set Error Data Values */
			else if(int21_al == 0xA)
				/* implemented in ASSEMBLY */;
#endif

			/* all SHARE functions return error */
			else if(int21_al && int21_al < 0xA)
				error_no = ERR_BADFUNC;
			else
				goto panic_out;
			break;

#if 0
		/* Redirector Functions (handled by IFS call) */
		case 0x5F:
			break;
#endif


		/* Absolute Path */
		case 0x60:
			abspath(MK_FP(R_ES,R_DI), MK_FP(R_DS,R_SI));
			break;


#if NOT_IN_DOSINT
		/* Get PSP */
		case 0x62:
			R_BX = current_psp;
			break;
#endif

		/* Kanjii Support */
		case 0x63:
			R_SI = (unsigned)lead_byte_table;
			R_DS = ROM_DS;
			break;


		/* Get Extended Country Info */
		case 0x65:
			get_ext_country_info();
			break;


		/* Get/Set Global Code Page */
		case 0x66:
			if(int21_al == 1)
				R_BX = R_DX = country_info.code_page;
			else
				error_no = ERR_BADFUNC;
			break;


		/* Set Handle Count */
		case 0x67:
			set_handle_count(R_BX);
			break;


#if 0	/* this is the same action (except for adjusting the handle count)
		   that is performed by 0x10 (FCB close) and 0x3E (handle close)
			which are coded above.
		 */
		/* Flush File Buffers */
		case 0x68:
			file_close();			/* updates dir entry and dumps data */
			break;
#endif

		/* Get/Set Media Type */
		case 0x69:
			tmp = 0x800;
			if(int21_al == 0)	/* Get */
				tmp |= 0x66;
			if(int21_al == 1)	/* Set */
				tmp |= 0x46;
			ioctl(0x0D, tmp);		/* just like 440DH, CX=tmp */
			break;


		/* Extended Open */
		case 0x6C:
			file_special_open();
			break;

#if USING_FAT32

		// FAT32 functions
		//
		case 0x73:
			if(int21_al == 2)
				GetExtendedDPB();
			else if(int21_al == 3)
			{
				GetExtendedFreeSpace();


				//  BT 6-12-02
				//
				//  Commit the FSINFO sector when user is requesting free space
				//  counts and there has been a change to the free space since
				//  the last FSINFO sector was committed.
				//
				SetFSInfoFields();
	      }
			else if(int21_al == 4)
	            GetSetFat32FormatInformation(MK_FP(R_ES, R_DI));
			else if(int21_al == 5)
				ExtendedAbsReadWrite();
			else
				error_no = ERR_BADFUNC;
			break;
#endif

		/* !! ROM-DOS unique functions !! */
		case 0xDB:
			debug_call(int21_al);
			break;

		/* unsupported call */
		default:
			R_AL = 0;
			EXIT_FUNCTION();
			dos_return();

		/*
			ROM-DOS doesn't support this call.  Inquiring
			minds like mine want to know!
		*/
panic_out:
			panic(fatal_unsupp, 0x21, int21_ax);

	} /* end of the BIG SWITCH table */

	EXIT_FUNCTION();
}


/*
	This is the Installable File System dispatcher.  It
	detects that DOS calls are for the IFS (networks, CD ROM
	drives, etc) and performs the INT 2FH.

	Returns FALSE if this DOS call should be handled by the
	FAT file sytem instead.  Processes the call and returns
	TRUE if the call was for the IFS.

	0F -- FCB Open 			1116	ES:DI=sft, stk open
	10 -- FCB Close 			1106	ES:DI=sft
	11 -- FCB Find First		111B	ES:DI=cds
	12 -- FCB Find Next		111C	ES:DI=cds
	13 -- FCB Delete 			1113	ES:DI=cds
	14 -- FCB seq. read		1108	ES:DI=sft, CX=bytes DTA, ret CX
	15 -- FCB seq. write		1109	ES:DI=sft, CX=bytes DTA, ret CX
	16 -- FCB Create 			1117	ES:DI=sft, stk attr
	17 -- FCB Rename 			1111	ES:DI=cds
	21 -- FCB rand. read		1108	ES:DI=sft, CX=bytes DTA, ret CX
	22 -- FCB rand. write	1109	ES:DI=sft, CX=bytes DTA, ret CX
	27 -- FCB block read		1108	ES:DI=sft, CX=bytes DTA, ret CX
	28 -- FCB block write	1109	ES:DI=sft, CX=bytes DTA, ret CX
	36 -- Get disk space		110C	ES:DI=cds, ret AX,BX,CX,DX
	39 -- MKDIR					1103	ES:DI=cds
	3A -- RMDIR					1101	ES:DI=cds
	3B -- CHDIR					1105	ES:DI=cds
	3C -- Create 				1117	ES:DI=sft, stk attr
	3D -- Open 					1116	ES:DI=sft, stk open
	3E -- Close 				1106	ES:DI=sft
	3F -- File Read 			1108	ES:DI=sft, CX=bytes DTA, ret CX
	40 -- File Write 			1109	ES:DI=sft, CX=bytes DTA, ret CX
	41 -- Delete File 		1113	ES:DI=cds
	43 -- Get File Attr		110F	ES:DI=cds, ret AX
	43 -- Set File Attr		110E	ES:DI=cds, stk attr
	4E -- Find First 			111B	ES:DI=cds
	4F -- Find Next 			111C	ES:DI=cds
	56 -- Rename 				1111	ES:DI=cds
	5C -- Lock File			110A	ES:DI=sft, BL=0 DS:DX=buff w/ start,len
	5C -- Unlock File			110A	ES:DI=sft, BL=1 DS:DX=buff w/ start,len
	68 -- Commit File			1107	ES:DI=sft
	6C -- Extended Open     112E  ES:DI=sft, sopen_mode,sopen_attr,sopen_action,sda_fname1

	1123H	- qualify a filename

*/
bool	ifs_call(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern char sda_isnet;
	extern unsigned sda_rdwrlen;
	extern char sda_fname1[];
	extern char sda_fname2[];
	extern uchar int21_al;
	extern uchar int21_ah;
	extern unsigned int21_ax;
	extern uchar sda_chardev;
	extern uchar sda_openmode;
	extern uchar far * DTA;
	extern uchar far * sda_rdwrbuff;
	extern FCB sda_fcb;
	extern DIR sda_ffb;
	extern SDB sda_sdb;
	extern LPCDS sda_cdsptr;
	extern LPSFT sda_sftptr;
	extern unsigned int2f_ax;
	extern unsigned int2f_bx;
	extern unsigned int2f_cx;
	extern unsigned int2f_dx;
	extern unsigned int2f_di;
	extern unsigned sda_handle;
	extern unsigned sda_rdwrlen;
	extern unsigned sda_callflags;
	extern unsigned doscallflag[];
	extern uchar sda_srchattr;
	extern ulong sda_rdwroff;
	extern struct LOCK lock_buff;
	extern unsigned sopen_action;
	extern char sda_ename1[];
	extern uchar sda_fcbattr;
	extern uchar sda_sft_h2;
	extern unsigned sopen_mode;


	/*
		[ locals ]
	*/
	void far * esdi;
	uchar ifscall;
	uchar far * olddta;
	unsigned status;
	bool sftused;
	unsigned call_cx;
	unsigned i2FstackWord = 0;
#if USING_FAT32
	LPGDF32 lpgdf32;
	char far* lpPath;
	uchar drive;
#endif

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

#if USING_FAT32
	if(R_AX == 0x7303)
	{
		// verify the drive is correct
		lpPath = MK_FP(R_DS, R_DX);
		drive = toupper(*lpPath) - 'A';		// 0=A, 1=B and so on
		if(!drive_to_sda(drive))
		{
			error_no = ERR_NOTDRIVE;
			EXIT_FUNCTION();
			return TRUE;
		}
		sda_callflags = doscallflag[0x36];	// same normal get drive free space
	}
#endif

	/* get call # (this is just a MACRO, see DOSDEF.H) */
	ifscall = callf_xlatifs();

	/* not an IFS call for open/create CON, etc.. */
	if(callf_path() && sda_chardev)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	/* make sure the call really IS for the IFS */
	if(callf_handle() && !callf_path())
	{
		status = sda_sftptr->ioctl;
		if((status & IO_NETWORK) != IO_NETWORK)
		{
			EXIT_FUNCTION();
			return FALSE;
		}
	}
	else if(sda_cdsptr != NULLF)
	{
		status = sda_cdsptr->status;
	}

	/* is FindNext (DOS1 or 2) to be passed through to IFS? */
	if(ifscall == 0x1C)
	{
		/* drive is not IFS, no go */
		if((sda_sdb.drive&0x80) == 0)
		{
			EXIT_FUNCTION();
			return FALSE;
		}
	}

	/* 5FH calls are passed through */
	else if(int21_ah != 0x5F && !sda_isnet)
	{
		/* check all other calls for validity */
		if(!ifscall || ((sda_cdsptr != NULLF) && !(status&C_NETWORK)))
		{
			EXIT_FUNCTION();
			return FALSE;
		}
	}

	/*
		[ at this point, we KNOW it's an IFS call ]
	*/

	// Special case for "create/open file w/out CDS" (a network device).
	//
	// NOTE: Unlike MSDOS, we don't set the SDA CDS ptr offset to FFFF
	//			when a call is made to open "NET:\DEV\LPT1:" when that
	//			device has been redirected by a Novel network.
	//
	if((ifscall == 0x17) && (FP_OFF(sda_cdsptr) == 0xFFFF))
		++ifscall;

	/* Set File Attr (4301H) is a special case */
	if(ifscall == 0x0f )
	{
		if( find_device(sda_ename1) )
		{
			error_no = ERR_NOTFOUND;
			EXIT_FUNCTION();
			return TRUE;
		}

		if(int21_ax == 0x4301)
		{
			--ifscall;
			sda_openmode = R_CL;
			i2FstackWord = R_CL;
		}
	}

	// POWERLan needs this to prevent deleting read-only files.
	//
	if(ifscall == 0x13)
		sda_srchattr = A_FILECHK;

	/* for FindFirst/Next DTA must point to SDB, else RDWR area */
	olddta = DTA;
	if(ifscall == 0x1B || ifscall == 0x1C)
		DTA = (void far *)(&sda_sdb);
	else
		DTA = sda_rdwrbuff;

	/* we have not modified the SFT yet */
	sftused = FALSE;

	/* actually call INT 2FH, func 11xxH */
	if(callf_handle())
	{
		/* read/write calls must know offset */
		sda_rdwroff = sda_sftptr->offset;
		esdi = (void far *)sda_sftptr;

		/*
			mark handle as temporarily used (so reentering via 5D00H
			doesn't use it!
		*/
		if(sda_sftptr->handles == 0)
		{
			sftused = TRUE;
			sda_sftptr->handles = 0xFFFF;
		}

		// set rdwr mode for open/create
		// MSDOS does not record the "don't inherit" bit in the rwmode
		// field but puts it in B12 of the ioctl field of the SFT.
		// All other file access bits are stored in rwmode.
		//
		if(ifscall == 0x16)
		{
			// clear the O_NOCHILD bit, but set the IO_NOCHILD
			// depending on O_NOCHILD.  Weird, huh?
			if(sda_openmode & O_NOCHILD)
				sda_sftptr->ioctl |= IO_NOCHILD;
			sda_openmode &= ~O_NOCHILD;
			sda_sftptr->rwmode = sda_openmode;
			i2FstackWord = sda_openmode;

			if( int21_ah == 0x3d )
			{
				sda_sft_h2 = 0;
				sda_srchattr = A_HIDDEN | A_SYSTEM | A_SUBDIR;
			}
		}

		// create
		else if(ifscall == 0x17)
		{
			sda_sftptr->rwmode = O_RDWR;
			if( int21_ah == 0x16 )
			{
				i2FstackWord = sda_fcbattr | A_ARCHIVE;
			}
			else
			{
				i2FstackWord = R_CL | A_ARCHIVE;
				if( int21_ah != 0x3c )
				{
					//  For int 21h function 5ah and 5bh set the "create
					//  brand new file or fail" bit in the high byte
					//
					i2FstackWord |= 0x100;
				}

				sda_sft_h2 = 0;
				sda_srchattr = A_SYSTEM | A_HIDDEN;
			}
		}
	}
	else
		esdi = (void far *)sda_cdsptr;


	// File read/write need to set CX to sda_rdwrlen
	// Everything else should pass in the user's CX
	if( (ifscall == 0x08) || (ifscall == 0x09) )
		call_cx = sda_rdwrlen;
	else
		call_cx = R_CX;

	// Don't call IFS if reading zero bytes.
	//
	if((ifscall == 0x08) && (call_cx == 0))
	{
		EXIT_FUNCTION();
		return TRUE;
	}

	// Create new file functions must fail if the file already exists!
	// This is signified by B8 of the word passed to the IFS on the stack.
	//
	//  ETR 3592 3-15-01
	//
	//  The below line was not correct.  The low byte of the stack
	//  word should be the attributes for the file.
	//
	//	 i2FstackWord = sda_openmode;
	//
	if( int21_ah == 0x6C )
	{
 		sda_sftptr->rwmode = sopen_mode;
		if( sopen_action == ACT_NEXIST_CREATE )
		{
				//  Set the attributes for the file
				//
				if( R_CL & (~(A_READONLY | A_HIDDEN | A_SYSTEM | A_ARCHIVE)))
				{
					//  Why??? Like MS.
					//
					i2FstackWord = A_ARCHIVE | A_VOLUME;
				}
				else
				{
					i2FstackWord = R_CL | A_ARCHIVE;
				}


				// set the "create brand new file or fail" bit
				//
				i2FstackWord |= 0x100;
		}
		else
		{
			if( sopen_action & ACT_EXIST_REPLACE )
			{
				//  Set the attributes for the file
				//
				if( R_CL & (~(A_READONLY | A_HIDDEN | A_SYSTEM | A_ARCHIVE)))
				{
					//  Why??? Like MS.
					//
					i2FstackWord = A_ARCHIVE | A_VOLUME;
				}
				else
				{
					i2FstackWord = R_CL | A_ARCHIVE;
				}
			}
			else
			{
				//  Set the open mode in the stack word
				//
				i2FstackWord = sda_openmode;
			}
		}


		//  Set the sda_srchattr variable
		//
		if( sopen_action & ACT_EXIST_OPEN )
		{
			sda_srchattr = A_SUBDIR | A_SYSTEM | A_HIDDEN;
		}
		else
		{
			sda_srchattr = A_SYSTEM | A_HIDDEN;
		}
		sda_sft_h2 = 0;
	}


	/* Lock/Unlock Region (5Cxx) is a special case */
	if(int21_ah == 0x5C)
	{
		/* Lock = 0AH, Unlock = 0BH */
		lock_buff.start = (long)MK_FP(R_CX,R_DX);
		lock_buff.len   = (long)MK_FP(R_SI,R_DI);
		lock_file(int21_al, esdi, &lock_buff);
	}

	/* int 21H, 5FxxH */
	else if (ifscall == 0x1E)
	{
		generate_int2F5F();
	}
	else
	{
		generate_int2F(ifscall, i2FstackWord, call_cx, esdi, (void far *)sda_fname2);
	}

	/* we are no longer in a network call */
	sda_isnet = 0;

	/* restore the DTA to original value */
	DTA = olddta;

	/*
		we are back, restore handle count to 0 (it will get updated
		later).  see sftused above.
	*/
	if(sftused && sda_sftptr->handles == 0xFFFF)
	{
		sda_sftptr->handles = 0;
	}

	// Surprisingly, Novel returns the documented indication of
	// O_NOCHILD in the SFT rwmode field. But MSDOS v6.22a changes
	// it to be IO_NOCHILD in the ioctl field. Lovely, hey.
	//
	if(!error_no && (ifscall == 0x16) && (sda_openmode & O_NOCHILD))
	{
		sda_sftptr->rwmode &= ~O_NOCHILD;
		sda_sftptr->ioctl  |= IO_NOCHILD;
	}

	/* return from 21/5FxxH */
	if(ifscall == 0x1E)
		R_AX = int2f_ax;

	/* error, what to do */
	else if(error_no)
	{
		/* always return 12H for error on findnext */
		if(ifscall == 0x1C)
			error_no = ERR_LASTFILE;
	}

	/* possible return registers, but not for INT 21H, 5FH (IFS 1E) */
	else
	{
		if(callf_rdwr())
			sda_rdwrlen = int2f_cx;

		if(ifscall == 0x0C)
		{
			int2f_ax &= 0xFF;
#if USING_FAT32
			if(R_AX == 0x7303)
			{
				lpgdf32 = MK_FP(R_ES, R_DI);
				lpgdf32->ulSectorsPerCluster         = (ulong)int2f_ax;
				lpgdf32->ulBytesPerSector            = (ulong)int2f_cx;
				lpgdf32->ulAvailableClusters         = (ulong)int2f_dx;
				lpgdf32->ulTotalClusters             = (ulong)int2f_bx;
				lpgdf32->ulAvailablePhysicalSectors  = _lmulu((ulong)int2f_dx, int2f_ax);
				lpgdf32->ulTotalPhysicalSectors      = _lmulu((ulong)int2f_bx, int2f_ax);
				lpgdf32->ulAvailableAllocationUnits  = (ulong)int2f_dx;
				lpgdf32->ulTotalAllocationUnits      = (ulong)int2f_bx;
			}
			else
			{
				R_AX = int2f_ax;
				R_BX = int2f_dx;	/* BX and DX are switched */
				R_CX = int2f_cx;
				R_DX = int2f_bx;
			}
#else
			R_AX = int2f_ax;
			R_BX = int2f_dx;	/* BX and DX are switched */
			R_CX = int2f_cx;
			R_DX = int2f_bx;
#endif
		}
		if(ifscall == 0x0F)
		{
			if(int21_ah == 0x23)
			{

				// need to returned file size by record size
				// to get record count
				//
				sda_ffb.size = (ulong)(MK_FP(int2f_bx,int2f_di));
				compute_FCB_records();
			}
			R_AX = R_CX = int2f_ax;
		}
	}

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Returns TRUE if worked, FALSE if failed.
*/
bool ifs_findfirst(void)
{
	extern int error_no;
	extern uchar far * DTA;
	extern char sda_fname1[128];
	extern uchar sda_srchattr;
	extern LPCDS sda_cdsptr;
	extern SDB sda_sdb;

	void far * fptr;

	ENTER_FUNCTION();

	// set up dta and cdsptr
	fptr = DTA;
	DTA = (void far *)(&sda_sdb);
	sda_cdsptr = MK_FP(FP_SEG(sda_cdsptr),0xFFFF);

	// If we got here and this is not a X:\DIR\FILE request,
	// then the file really doesn't exist, so don't try an
	// IFS findfirst
	if(sda_fname1[1] != ':')
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// adjust the filename
	*sda_fname1 = '\\';
	memmove(sda_fname1+1,sda_fname1+2,sizeof(sda_fname1)-2);

	// does IFS listen?
	error_no = 0;
	generate_int2F(0x19, 0, sda_srchattr, sda_cdsptr, (void far *)sda_fname1);

	// restore dta pointer
	DTA = fptr;

	EXIT_FUNCTION();

	return (!error_no);
}
