//
//   $Id: DOSLOAD.C 1.12 2002/10/15 21:44:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	DOSLOAD.C  created 01/11/89  by Drew Gislason

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains the necessary routines to load an
	.EXE or COM file into memory and execute it.  It also
	contains the routines to terminate a process.

**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#include "farlfn.h"

#define __FILENUM__ 10


extern unsigned dos_psp;	/* PSP (segment) of ROM-DOS itself */
extern unsigned root_psp;	/* PSP (segment) of the first program ever run */
extern uchar booting;		/* set to 1 on boot (this is NOT a boolean #) */

/*
	This function calculates memory requirements for COM and
	EXE programs, as well as returning the offset inside an
	EXE for both the image and the fixup table.

	In the case of an error, error_no will be non-zero on
	exit of this function and internal_close will be called to
	close the file if opened before returning an error code.

	Currently we are using the following formula for determining
	least and most memory allocation values (PSPSIZE + 10h must
	be added by the calling program, because overlays will not
	need those values and execs will):

	COM files
		Min = imglen + PSPSIZE + 10h
		Max = FFFFh
	EXE files
		Min = imglen - hdrlen + minparas + PSPSIZE + 10h
		Max = imglen - hdrlen + maxparas + PSPSIZE + 10h
	EXE files where maxpara is 0
		Min = as above
		Max = 0 (Flag value to be changed to FFFFh later)
*/
void open_program_for_exec( unsigned *pLeastParas, unsigned *pMostParas,
	struct EXEHDR *pExeHdr, bool *pIsExe )
{
	/*
		[ externals ]
	*/
	extern LPSFT sda_sftptr;
	extern int error_no;

	/*
		[ locals ]
	*/
	long least=0, most=0;
	unsigned register len;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(pLeastParas);
	ASSERT(pMostParas);
	ASSERT(pExeHdr);
	ASSERT(pIsExe);

	/* Because we may or may not set error_no on exit,
		we need to be sure that it isn't already set. */
	PERMASSERT( !error_no );

	/* Open the file mentioned in the SDA so we can read it in */
	if(internal_open((char far *)NULL, FALSE) == 0xFF)
	{
		if( error_no != ERR_DENIED )
		{
 			error_no = ERR_NOTFOUND;
		}
		EXIT_FUNCTION();
		return;
	}

	/* determine what kind of file we have (read in the .EXE header) */
	len = internal_read((void far *)pExeHdr, sizeof(struct EXEHDR));

	/* can't run a program with no stuff in it */
	if(len == 0)
	{
		error_no = ERR_NOMEM;
		EXIT_FUNCTION();
		return;
	}

	/* is this an EXE or COM style program? */
	*pIsExe = (len == sizeof(struct EXEHDR) && pExeHdr->sig == EXESIG);

	/* EXE program: determine length, start, and registers */
	if(*pIsExe)
	{
		/* determine length of EXE from its header */
		least = (pExeHdr->imglen * 0x20) - pExeHdr->hdrlen;
		if(_lshlu((long)least, 4) >= sda_sftptr->size+0x200)
		{
			internal_close(FALSE);
			error_no = ERR_INVALID_FMT;
			EXIT_FUNCTION();
			return;
		}

		/* how much memory does this thing want? (FFFF means all) */
		if( pExeHdr->maxpar )
			most = least + pExeHdr->maxpar;

		/* how much memory does this thing need (we MUST have this much) */
		least += pExeHdr->minpar;

		/* skip past header (use long because it may be larger than 64K) */
		internal_seek(_lshlu((long)pExeHdr->hdrlen, 4), SK_BEG);
	}

	/* COM program: determine length, start, and registers */
	else
	{
		/* COM files are allocated all of available memory */
		most = 0xFFFF;

		/* find the length */
		least = internal_seek(0L, SK_END);
		least = _lshru((least + 15), 4);

		/* seek to beginning of file for reading */
		internal_seek(0L,0);
	}

	if( least < 0xFFFFL )
		*pLeastParas = (unsigned)least;
	else
		*pLeastParas = 0xFFFF;
	if( most < 0xFFFFL )
		*pMostParas = (unsigned)most;
	else
		*pMostParas = 0xFFFF;

	EXIT_FUNCTION();
}


/*
	This function reads a binary image into memory.  Assumes
	the file is already open and the file pointer points past
	any header information.

	Returns FALSE on failure, with error_no set.
*/
bool PASCAL load_image(unsigned seg, unsigned img_paras)
{
	/*
		[ externals ]
	*/
	extern int error_no;

	/*
		[ locals ]
	*/
	#define PARAS (0xFE0)	/* almost 64K size chunks to read in programs */
	unsigned readlen;

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(seg);
	ASSERT(img_paras);

	/* read in the file */
	while(img_paras)
	{
		/* get readlength of read */
		if(img_paras > PARAS)
			readlen = PARAS;
		else
			readlen = img_paras;
		if( internal_read(MK_FP(seg, 0), (readlen << 4)) != (readlen << 4) )
			readlen = img_paras = 0;
		if (error_no)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		/* on to next "block" */
		seg += PARAS;
		img_paras -= readlen;
	}

	EXIT_FUNCTION();

	return TRUE;
}

/*
	This function loads a COM or EXE file into memory and
	does the fixups if this in an EXE.  Assumes the filename
	has already been parsed into the SDA.

	Returns FALSE on failure, with error_no set.
*/
bool PASCAL load_overlay(unsigned max_size, unsigned loadseg, unsigned fixseg)
{
	/*
		[ externals ]
	*/
	extern int error_no;

	/*
		[ locals ]
	*/
	unsigned most_paras;		/* paragraphs wanted to load the process	*/
	unsigned least_paras; 	/* paragraphs needed to load the process  */
	struct EXEHDR exehdr;
	bool is_exe;
	int tmp_error;				// If an error occurs, we need to close
									// the file, which zeroes error_no

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(loadseg);
	ASSERT(fixseg);

	/* Open program, set up allocation lengths, etc. */
	open_program_for_exec( &least_paras, &most_paras, &exehdr, &is_exe );

	/* Make sure that we are allowed to read least_paras of memory */
	if( least_paras > max_size )
	{
		internal_close(FALSE);
		error_no = ERR_NOMEM;
		EXIT_FUNCTION();
		return FALSE;
	}

	/* Load the program into memory.  If this fails, we haven't
		claimed ownership of any memory, so we don't need to. */
	if( !load_image(loadseg, least_paras) )
	{
		tmp_error = error_no;
		internal_close(FALSE);
		error_no = tmp_error;
		EXIT_FUNCTION();
		return FALSE;
	}

	/* relocate the fixups in the EXE */
	if(is_exe)
	{
		internal_seek((long)exehdr.reldsp, SK_BEG);
		reloc(fixseg, exehdr.nreloc);
	}

	/* close the file */
	internal_close(FALSE);

	EXIT_FUNCTION();

	return TRUE;
}


/*
	Relocate the items described in the EXE's header.  These
	are called fixups.
*/
void	reloc(unsigned loadseg, unsigned count)
{
	/*
		[ locals ]
	*/
	#define FIXBLK		32
	unsigned far *fptr;
	unsigned i, j;
	struct rel_item reloc_item[FIXBLK];


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(loadseg);

	for(i=0; i<count; i+= FIXBLK)
	{
		internal_read((void far *)reloc_item, sizeof(reloc_item));
		for(j=0; j < FIXBLK && i+j < count; ++j)
		{
			fptr = MK_FP(reloc_item[j].segment+loadseg, reloc_item[j].offset);
			*fptr += loadseg;
		}
	}

	EXIT_FUNCTION();
}


/*
	This function loads a COM or EXE file into memory and	does
	the fixups if this in an EXE.  Assumes the filename has
	already been parsed into the SDA.

	This includes allocating the memory for the process, building
	a new environment and a PSP, and loading the program.  This
	is used by the exec function.

	Returns FALSE on failure, with error_no set.
*/
bool PASCAL load_process(union EXEPARM far *parm, void far *retadr, bool go)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern unsigned current_psp;
	extern uchar BREAK_FLAG;
	extern uchar mem_fit;
	extern int sda_no24;

	/*
		[ locals ]
	*/
	#define PSP_SIZE	0x10	/* 10h paragraphs for PSP at 100h bytes   */
	#define PSP_ALLOW	0x20	/* kludge for extra 10h paras used by MS-DOS */
	unsigned envseg;			/* where the environment is placed        */
	unsigned pspseg;			/* where the PSP is created               */
	unsigned loadseg;			/* where the load segment is placed       */
	unsigned avail;			/* memory available to the process        */
	unsigned most_paras;		/* paragraphs wanted to load the process	*/
	unsigned least_paras; 	/* paragraphs needed to load the process  */
	struct EXEHDR exehdr;
	bool load_at_end = FALSE;
	bool is_exe;
	int tmp_error;				// If an error occurs, we need to close
									// the file, which zeroes error_no

	unsigned childCS, childIP, childSS, childSP, childDS;
	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(parm);
	ASSERT(retadr);

	/* Open program, set up allocation lengths, etc. */
	open_program_for_exec( &least_paras, &most_paras, &exehdr, &is_exe );

	/* most_paras can store a "magic cookie" flag value of 0.
		When that happens, the program needs to be loaded at the
		end of the block */
	if( !most_paras )
	{
		load_at_end = TRUE;
		most_paras = 0xFFFF;
	}

	/* save room for the psp, but don't overflow. */
	if( least_paras < 0xFFFF - PSP_SIZE )
		least_paras += PSP_SIZE;
	else
		least_paras = 0xFFFF;
	if( most_paras < 0xFFFF - PSP_SIZE )
		most_paras += PSP_SIZE;
	else
		most_paras = 0xFFFF;

	/* make sure we can build the environment */
	envseg = build_env(parm->e.envseg);
	if(envseg == 0xFFFF)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	/* allocate memory to our process, but keep it "free" */
	avail = mem_alloc_best(least_paras, most_paras, &pspseg);
	if( error_no )
	{
		tmp_error = error_no;
		internal_close(FALSE);
		error_no = tmp_error;
		EXIT_FUNCTION();
		return FALSE;
	}

	/* Free the environment AFTER allocating program!!! */
	mem_new_owner(envseg, FREEMEM);

	/* Adjust allocation size to allow for the PSP */
	least_paras -= PSP_SIZE;
	if( load_at_end )
	{
		loadseg = pspseg + avail - least_paras;
		avail = least_paras;
	} else
	{
		loadseg = pspseg + PSP_SIZE;
		avail  -= PSP_SIZE;
	}

	/* Load the program into memory.  If this fails, we haven't
		claimed ownership of any memory, so we don't need to. */
	if( !load_image(loadseg,least_paras) )
	{
		tmp_error = error_no;
		internal_close(FALSE);
		error_no = tmp_error;
		EXIT_FUNCTION();
		return FALSE;
	}

	/* relocate the fixups in the EXE */
	if(is_exe)
	{
		internal_seek((long)exehdr.reldsp, SK_BEG);
		reloc(loadseg, exehdr.nreloc);
	}

	/* close the file */
	internal_close(FALSE);

	/* We need to set user registers for executing EXEs
		different then COMs */
	if( is_exe )
	{
		// set stack and program counter
		//
		childCS = exehdr.csdsp + loadseg;
		childIP = exehdr.ipofs;
		childSS = exehdr.ssdsp + loadseg;
		childSP = exehdr.spofs;
		childDS = pspseg;

		// If no stack segment explicit in exe header then
		// handle the .EXE's stack as if it were a .COM file.
		// Even though it may be that CODE+DATA > 64K and
		// avail > IMAGE > 64K, MSDOS will still set SS=CS
		// and SP=0. So that's what that warning is for...
		//
		if((childCS == childSS) && (childSP == 0))
		{
			if(avail < 0x1000)
				childSP = (avail << 4);
		}
	}
	else
	{
		if(avail >= 0x1000)
			childSP = 0;
		else
			childSP = (avail << 4);
		childSS = pspseg;
		childCS = pspseg;
		childDS = pspseg;
		childIP = 0x100;
	}

	/* the memory we got above can now be used.  "allocate" it. */
	mem_new_owner(envseg, pspseg);
	mem_new_owner(pspseg, pspseg);

	/* actually fill in the PSP */
	build_psp(pspseg, envseg, parm, retadr);

	/* increment handle count on all open files */
	exec_open(TRUE);

	/* push a 0 on the stack */
	childSP -= 2;
	*(int far *)MK_FP(childSS,childSP) = 0;

	/* place program name in last 8 bytes of MCB */
	set_mcb_prog_name(current_psp);

	/* load but don't go */
	if( !go )
	{
		/* let caller know where the process was loaded */
		parm->e.spofs= childSP;
		parm->e.ssdsp= childSS;
		parm->e.ipofs= childIP;
		parm->e.csdsp= childCS;
		R_AX = childIP;
	}

	/* load and execute */
	else
	{
		/* if user hit break while file was loading, break out now */
		if(BREAK_FLAG)
			ctrlc_check(0);

		/* record the first program ever executed */
		if(!root_psp && !booting)
			root_psp = current_psp;

		/* set the interrupts */
		set_process_ints(MK_FP(current_psp, 0xA));

		/* Allow for the SP difference like MS-DOS does */
		if( is_exe)
			childSP += 2;

		/* in case this was turned on for exec, turn it off now */
		sda_no24 = FALSE;

		//  BT 6-13-02
		//
		//  MS does not flush all buffers before exec'ing a program
		//
		//  /* flush all buffers, all drives */
		//  buf_flush_all();

		EXIT_FUNCTION();

		/* We will NOT return from here! */
	   go_asm(childIP, childCS, childSP, childSS, childDS);
	}

	EXIT_FUNCTION();

	return TRUE;
}


/*
	Create a new PSP at the specified paragraph, setting the
	terminate address either to the parent's terminate
	address, or the current CS and IP.
*/
void	PASCAL create_new_psp(unsigned seg, int child,
		int fIncrementOpenHandles )
{
	/*
		[ externals ]
	*/
	extern unsigned current_psp;


	/*
		[ locals ]
	*/
	void far * ret_addr;
	struct PSP far * psp;
	union EXEPARM parm;
	unsigned oldpsp;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameter
	ASSERT(seg);

	oldpsp = current_psp;
	psp = MK_FP(current_psp,0);
	parm.e.fcb1 = &psp->fcb1;
	parm.e.fcb2 = &psp->fcb1;
	parm.e.command_line = &psp->count;

	/* return address is different, depending */
	if(child)
		ret_addr= MK_FP(R_CS,R_IP);
	else
		ret_addr= get_vector(0x22);

	/* actually make the thing */
	build_psp(seg, psp->env_seg, (void far *)(&parm), ret_addr);

	if( fIncrementOpenHandles )
	{
		/* increment handle count on all open files */
		exec_open(TRUE);
	}

	/* set current back to original PSP for "new", but not "child" PSPs */
	if(!child)
	{
		current_psp = oldpsp;
	}

#if 0
 printf("create psp %p, old %p, stack %lp\n",seg,oldpsp,psp->user_stack);
 con_getkey();
#endif

	EXIT_FUNCTION();
}

/*
	Don't let offsets wrap around, causing protection
	violations.
*/
unsigned nowrap(unsigned off, unsigned len)
{
	ENTER_FUNCTION();

	if(len+off < off)
	{
		EXIT_FUNCTION();
		return 0 - off;
	}

	EXIT_FUNCTION();

	return len;
}

/*
	Build a psp at the specified address.  Sets the
	current_psp to this address.
*/
void	PASCAL build_psp(unsigned seg,unsigned envseg,union EXEPARM far *parm,void far *retadr)
{
	/*
		[ externals ]
	*/
	extern void far * DTA;
	extern unsigned current_psp;
	extern unsigned dos_ver;
	extern union dos_regs far * userregs;


	/*
		[ locals ]
	*/
	#define MAXFILES	20
	static uchar dispatch[5] = { 0x9A, 0xF0, 0xFE, 0x1D, 0xF0 };
	static uchar doscall[3]  = { 0xCD, 0x21, 0xCB };
	struct PSP far *psp;
	unsigned memtop;
	struct MCB far * mcb;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(seg);
	ASSERT(envseg);
	ASSERT(parm);

	/* get the top of memory for this PSP */
	mcb = MK_FP(seg-1,0);
	memtop = seg + mcb->para_cnt;

	/* point the DTA into this PSP */
	psp = MK_FP(seg, 0);
	DTA = MK_FP(seg, 0x80);

	/* a PSP inherits all its parent handles, etc... */
	if(current_psp == SYSTEM_OWNER)
	{
		fmemset(psp, 0, 0x100);
		psp->dosver  = dos_ver;
	}
	else
	{
		fmemcpy(psp,MK_FP(current_psp,0),0x80);

		if(psp->htbl_addr != psp->hptr)
			fmemcpy(psp->hptr, psp->htbl_addr, MAXFILES);

		/* copy command line */
		fmemcpy(psp->fcb1,parm->e.fcb1,nowrap(FP_OFF(parm->e.fcb1),0x10));
		fmemcpy(psp->fcb2,parm->e.fcb2,nowrap(FP_OFF(parm->e.fcb2),0x10));
		fmemcpy(&psp->count,parm->e.command_line,
			nowrap(FP_OFF(parm->e.command_line),0x80));
	}
	psp->env_seg = envseg;

	/* a psp has all sorts of fields */
	psp->int_20             = 0x20CD;			/* INT 20 Instruction */
	psp->top_of_memory      = memtop;
	fmemcpy(psp->dos_long_call, (void far *)dispatch, sizeof(dispatch));
	fmemcpy(psp->doscall, (void far *)doscall, sizeof(doscall));
	psp->term_adr = retadr;
	psp->ctrl_brk_adr = get_vector(0x23);
	psp->critical_error_adr = get_vector(0x24);
	psp->hcount = MAXFILES;
	psp->htbl_addr = psp->hptr;
	psp->last_psp = current_psp;
	psp->share_psp = -1L;

	/* set up pointer to user stack in parent PSP */
	psp = MK_FP(current_psp, 0);
	psp->user_stack = (void far *)userregs;

	current_psp = seg;

	//  05-14-99 BT
	//
	//  Calling this again increaments the SFT open file count a
	//  second time and you will soon run out  of handles (available
	//  SFTs).  Calling this here will break fn 26 create PSP which
	//  should not increment the open handle count in the SFTs for
	//  files which are open and are copied to a new PSP. Fn 26 does
	//  not work like fn 55 in this respect.
	//
//	exec_open(TRUE);

	EXIT_FUNCTION();
}


/*
	Build a new environment by making a copy of the old one
	and appending the new path.

	Returns the environment segment, or 0xFFFF on failure.
*/
unsigned build_env(unsigned parmenv)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern char sda_fname1[];
	extern unsigned current_psp;
	extern uchar sda_drive;


	/*
		[ locals ]
	*/
	unsigned envseg;
	unsigned oldenv;
	unsigned envblks;
	unsigned envsize;
	struct PSP far *psp;
	char far *ep;
	int namelen;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* find the old environment */
	if(parmenv == 0)
	{
		psp = MK_FP(current_psp,0);
		oldenv = psp->env_seg;
	}
	else
		oldenv = parmenv;

	/* calculate the new size (including filename) in paragraphs */
	envsize = env_size(oldenv);
	if(error_no)
	{
		EXIT_FUNCTION();
		return 0xFFFF;
	}

	//  ETR 3483 BT 8-8-00
	//
	//  To compute the size of the new environment, we were using the
	//  string length of the normalized exec name and copying in
	//  (correctly) the exec name supplied with the 4bh call.
	//
	//  namelen = strlen(sda_fname1) + 1;
	//
	namelen = fstrlen(MK_FP(R_DS,R_DX))+1;

	/* special case: do not create an env for root, use current */
	if(!root_psp && !booting)
		envseg = oldenv;

	/* allocate memory for new env and copy in old */
	else
	{
		envblks = (envsize + sizeof(int) + namelen + 15)/0x10;

		//  ETR ? 9-13-00 BT
		//
		//  MS-DOS allocates a minimum of 256 bytes for a program
		//  environment.  Make sure that the program size is at
		//  least 256 bytes (10h paragraphs).
		//
		if( envblks < 0x10 )
		{
			envblks = 0x10;
		}

		envseg = mem_alloc(envblks);
		if(error_no)
		{
			EXIT_FUNCTION();
			return 0xFFFF;
		}
		fmemcpy(MK_FP(envseg, 0),MK_FP(oldenv,0),envsize);
	}

	/* tack on the name given to 4BH */
	ep = MK_FP(envseg,envsize);
	*(int far *)ep = 1;
	fmemcpy(ep+2, MK_FP(R_DS,R_DX), fstrlen(MK_FP(R_DS,R_DX))+1);

	EXIT_FUNCTION();

	return envseg;
}


/*
	Returns the size (in bytes) of this environment.
*/
unsigned env_size(unsigned envseg)
{
	extern int error_no;

	char far * ep;
	unsigned len=0;
	unsigned itemlen;

	ENTER_FUNCTION();

	// validate parameter
	ASSERT(envseg);

	/* find the length of the environment */
	ep = MK_FP(envseg,0);
	do
	{
		itemlen = 0;
		while(*ep && itemlen < 0x8000)
			++ep, ++len, ++itemlen;

		// no items bigger than 32K
		if(itemlen >= 0x8000)
		{
			error_no = ERR_INVALID_ENV;
			EXIT_FUNCTION();
			return 0;
		}

		++ep, ++len;	// skip ending 0
	} while(*ep);
	++len;				// add 1 for double 0

	EXIT_FUNCTION();

	return len;
}

/*
	Copy the program name into the MCB.  This way, memory
	dumpers can tell which program's own which MCBs.
*/
void set_mcb_prog_name(unsigned seg)
{
	extern char sda_ename1[];
	char name[8];
	int i;

	ENTER_FUNCTION();

	// validate parameter
	ASSERT(seg);

	for(i=0; i < 8 && sda_ename1[i] != ' '; ++i)
		name[i] = sda_ename1[i];
	if(i < 8)
		name[i] = 0;
	fmemcpy(MK_FP(seg-1,8), (void far *)name, 8);

	EXIT_FUNCTION();
}

/*
	The execute function of DOS.  Has options to load
	overlays, load and don't go, and load and execute.  It
	gives the child process access to all presently opened
	handles, and gives all of free memory over to the new
	process.

	Does not return.
*/
void	exec(union EXEPARM far *parm, void far *return_adr)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar int21_al;

	/*
		[ locals ]
	*/
	#define PROC_GO	TRUE
	#define PROC_DONT	FALSE

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(parm);
	ASSERT(return_adr);

	/* check to see that function is valid (0,1,3) */
	switch( int21_al )
	{
		case 0 :
			load_process(parm, return_adr, PROC_GO);
			break;
		case 1 :
			load_process(parm, return_adr, PROC_DONT);
			break;
		case 3 :
			load_overlay(0xFFFF, parm->o.ovrseg, parm->o.reloc);
			break;
		default :
			error_no = ERR_BADFUNC;
			break;
	}

	EXIT_FUNCTION();

	return;
}


/*
	Terminate a process, passing control back to the parent
	(or whoever is pointed to by the terminate address in
	the PSP).  All memory and all SFTs owned by this process
	are freed.

	"keep" is # of paras to keep. 0 means free all.

	Note: DOES NOT RETURN!
*/
void	PASCAL terminate(unsigned retcode, unsigned keep)
{
	/*
		[ externals ]
	*/
	extern uchar InDOS;
	extern uchar CritErr;
	extern unsigned current_psp;
	extern unsigned return_code;
	extern int error_no;
	extern unsigned special_ret;	/* if INT 23H or 24H terminate */


	/*
		[ locals ]
	*/
	struct PSP far * psp;
	void far * return_adr;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* check memory for damage */
	mem_chk();

	/* let network know we are terminating */
	generate_int2Fx(0x22);
	error_no = 0;				/* note: testing only */


	/* get return code code */
	return_code = special_ret ? special_ret : retcode;
	special_ret = 0;

	/* allow no critical errors to occur while terminating */
	CritErr = TRUE;

	/* free all memory OWNED by process */
	// set up ptr to current_psp in case proc is going resident
	psp = MK_FP(current_psp, 0);
	if(!keep)
	{
		/*
			COMMAND.COM points its own PSP to itself, so when you
			A)bort it doesn't free handles and memory.  This should
			probably be recorded to be handled in COMMAND.
		*/
		if(psp->last_psp != current_psp)
		{
			// NOTE: Novel VLMs depend on the handles already being
			// closed BEFORE the "Close all Remote Handles" call is
			// made. Otherwise all pending data is lost.
			//
			/* close all OPEN handles OWNED by process */
			term_close();

	 		/* close all OPEN handles on remote IFS devices */
			generate_int2Fx(0x1D);

			mem_term_free();
		}
	}

	/* We are keeping the process in memory */
	else
	{
		mem_resize(current_psp, keep);
		error_no = 0;		/* we don't care if the resize worked */

		// some TSRs depend on this field pointing to the end of
		// their "kept" memory block.
		//
		psp->top_of_memory = current_psp + keep;
		return_code |= 0x300;
	}

	/* flush all buffers, all drives */
	buf_flush_all();

	// why DID we invalidate? - dg 3/97
	//
	//  Because we may have called a program which touch the disk
	//  directly.  Thus what was in the buffers may be invalid
	//
	buf_invalidate();

	/* reset vectors to previous places from the terminating proc's psp */
	return_adr = psp->term_adr;
	set_process_ints(&psp->term_adr);

	/* return back to old PSP */
	current_psp = psp->last_psp;
	psp = MK_FP(current_psp, 0);
	userregs = psp->user_stack;	/* back to original stack */
	R_IP = FP_OFF(return_adr);		/* set return address */
	R_CS = FP_SEG(return_adr);
	R_FLAGS |= F_INT;					/* always enable interrupts */
	R_FLAGS &= ~F_DIR;				/* always disable direction */

	/* Reset DOS and exit */
	InDOS = 1;
	error_no = CritErr = 0;

	EXIT_FUNCTION();

	dos_exit();
}

/*
	Return the maximum # of handles for the current process.
*/
int max_handles(void)
{
	extern unsigned current_psp;
	struct PSP far * psp;

	ENTER_FUNCTION();

	psp = MK_FP(current_psp,0);

	EXIT_FUNCTION();

	return (psp->hcount);
}

/*
	Used only by EXEC. If the "open" variable is FALSE then this
	decrements the SFT open count for all valid handles in the
	current PSP.

	If the "open" variable is TRUE then this function may
	increment the SFT open count for a valid file handle in the
	current PSP. Since, when this is called, the current PSP has
	only been copied from the parent PSP, we need to check for NO
	INHERENT attributes in the SFT; when we find them we must
	invalidate the handle in the current PSP without incrementing
	the SFT open count.

*/
void	exec_open(bool open)
{
	/*
		[ externals ]
	*/
	extern LPSFT sda_sftptr;
	extern unsigned sda_callflags;
	extern unsigned current_psp;

	/*
		[ locals ]
	*/
	unsigned i;
	unsigned hcount;
	struct PSP far *psp_ptr;
	unsigned char far *psp_hptr;

	/*
		[ code ]
	*/


	ENTER_FUNCTION();

	/* determine the # of handles */
	hcount = max_handles();

	// get a ptr to the current JFT (PSP file handle array)
	//
	psp_ptr  = MK_FP(current_psp,0);
	psp_hptr = MK_FP(current_psp,FP_OFF(psp_ptr->hptr));

	/* make sure we are using DOS2 SFTs (not DOS1) */
	sda_callflags = 0;

	/* increment all that are open */
	for(i=0; i<hcount; ++i, psp_hptr++)
	{
		if(handle_to_sda(i))
		{
			if(open)
			{
				// if no inherit bit is set in SFT then don't let
				// child inherit the file else increment the handle count
				//
				if(sda_sftptr->ioctl & IO_NOCHILD)
					*psp_hptr = 0xFF;
				else
					++sda_sftptr->handles;
			}
			else
				internal_close(TRUE);
		}
	}

	EXIT_FUNCTION();
}


/*
	Used only by TERMINATE, this function will close and
	free all SFTs owned by this process.
*/
void	term_close(void)
{
	/*
		[ externals ]
	*/
	extern int sda_handle;
	extern LPSFT sda_sftptr;
	extern unsigned current_psp;
	extern unsigned sda_callflags;
	extern void far * fpFarLFNCall;

	void FreeProcessSearches(void);


	/*
		[ locals ]
	*/
	unsigned h;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* make sure we are on the DOS2 SFT list */
	sda_callflags = 0;

	/* go through all handles and make sure they are closed */
	exec_open(FALSE);

	/* go through stranded SFTs and make sure they are free */
	h = sda_handle = 0;
	while((void *)h_to_sft(h) != NULLF)
	{
		if(h_to_sda(h) && sda_sftptr->owner == current_psp)
		{
			internal_close(TRUE);
			sda_sftptr->handles = 0;
		}
		++h;
	}

	/* now, also free any LFN search handles allocated */
	FREE_PROCESS_SEARCHES();

	EXIT_FUNCTION();
}
