//
//   $Id: DOSPRINT.C 1.8 2002/10/15 21:44:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSPRINT.C  created 01/17/90  by Drew Gislason

   Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains printf.

**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "profile.h"

#define __FILENUM__ 14

/* PROTOTYPES used */
int  PASCAL strlen(char *s);
bool PASCAL handle_to_sda(unsigned handle);
void PASCAL memmove(void *to,void *from,unsigned len);
void PASCAL fmemcpy(void far *dst,void far *src,unsigned len);
void secret_elephant_burial_grounds(void);


/* do not print anything over 80 columns in length */
#define PRBUFFSIZE 80
static char prbuff[PRBUFFSIZE];

void PASCAL bios_putc(uchar c);
void PASCAL (*prfunc)(uchar c) = bios_putc;


/*
	Used by functions outside of printf.

	Change a number from a word to a decimal or hexidecimal
	number.  Radix can be anything from 2 to 16.  A radix
	of 2 or 16 will fill with 0s.
*/
void xitoa(char *buffer, unsigned num, int radix)
{
	static char hex[] = "0123456789ABCDEF";
	char str[18];
	int size = 0;
	int fill;

	ENTER_FUNCTION();

	do
	{
		str[size] = hex[num % radix];
		num = num/radix;
		++size;
	} while(num);


	/* for radix of 2 or 16, fill with 0s */
	if(radix == 2 || radix == 16)
	{
		fill = (radix==16 ? 4 : 16) - size;
		while(fill--)
			*buffer++ = '0';
	}

	/* reverse the # */
	while(size-- > 0)
		*buffer++ = str[size];
	*buffer = 0;

	EXIT_FUNCTION();
}

/*
	Print a character through the CONsole.  This function
	does not actually print anything until it gets to the
	end of a string (a 0 byte).  The function uses INT 29H
	directly to print characters.
*/
 static void PASCAL outchar(char c)
{
	static char * out = prbuff;

	ENTER_FUNCTION();

	*out++ = c;
	if(!c || out >= prbuff+PRBUFFSIZE-1)
	{
		*out = 0;
		out = prbuff;
		while(*out)
			(*prfunc)(*out++);
		out = prbuff;
	}

	EXIT_FUNCTION();
}

/*
	Displays a string to the console.  See outchar().
*/
 static void PASCAL outstring(char *s)
{
	ENTER_FUNCTION();

	while(*s)
		outchar(*s++);

	EXIT_FUNCTION();
}


/*
	Don't accept cheap imitations!  This is the real printf
	specifically for ROM-DOS.  What a treat!  However,
	there are limitations: No # formatting at all (that is
	%-5.3s will fail, you must use just %s).

	%c			Print a character
	%d			Print an integer
	%s			Print a string
	%p			Print a near pointer (padded with 0s)
	%x			Print an integer in hex
	%ld		Print a long
	%ls		Print a far string
	%lp		Print a far pointer (padded with 0s)
	%lx		Print a long in hex
*/
void vprintf(char *fmt, unsigned *args)
{
	/*
		[ externals ]
	*/
	extern char prbuff[];


	/*
		[ locals ]
	*/
	unsigned segment, offset;
	char far *fp;
	char buff[16];


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* fill the string with our formatted text */
	while(*fmt)
	{
		/* general output */
		if(*fmt != '%')
		{
			if(*fmt == '\n')
				outchar('\r');			/* CR before LF */
			outchar(*fmt);
		}

		/* this is a formatted output */
		else
		{
			++fmt;
			switch(*fmt)
			{
				/* unknown format, do not translate */
				default:
badform:
					outchar('%');
					outchar(*fmt);
					break;

				/* character format */
				case 'c':
					outchar(*args);
					break;

				/* integer format */
				case 'd':
					xitoa(buff, *args, 10);
					outstring(buff);
					break;

				/* near pointer format */
				case 'p':
					/* same as below */

				/* hexadecimal output */
				case 'x':
					xitoa(buff, *args, 16);
					outstring(buff);
					break;

				/* near string format */
				case 's':
					outstring((char *)*args);
					break;

				/* the long formats */
				case 'l':
					++fmt;
					offset  = *args++;
					segment = *args;

					switch(*fmt)
					{
						default:
							--fmt;
							goto badform;

						case 'd':
						case 'p':
						case 'x':
							xitoa(buff, segment, 16);
							outstring(buff);
							if(*fmt != 'd')
								outchar(':');
							xitoa(buff, offset, 16);
							outstring(buff);
							break;
						case 's':
							fp = MK_FP(segment,offset);
							while(*fp)
								outchar(*fp++);
							break;
					}
					break;
			}

			++args;

		}		/* end of formatted output */

		++fmt;
	}

	/* actually display the output */
	outchar(0);

	EXIT_FUNCTION();
}


/*
	This is the entry for a printf.  It calls the real
	workhorse vprintf.  The function declaration is tabbed
	over, and a prototype is included in this comment for
	the use of my prototyping program.  That is how
	DOSPROTO.H is generated.

void printf(char *fmt, ...)
*/
	void	printf(char *fmt, unsigned arglist)
{
	ENTER_FUNCTION();

	vprintf(fmt,&arglist);

	EXIT_FUNCTION();
}


/*
	A fatal error has occurred.  Kill the system.

void panic(char *fmt, ...)
*/
	void	panic(char *msg,unsigned arglist)
{
	extern char fatal_halt[];
	extern char fatal_romdos[];

	ENTER_FUNCTION();

	/* display the fatal message */
	vprintf(fatal_romdos,&arglist);
	vprintf(msg,&arglist);
	vprintf(fatal_halt,&arglist);

	/*
		When ROM-DOS knows that the time to die has come, it
		inevitably makes its way to the...
	*/
	EXIT_FUNCTION();
	secret_elephant_burial_grounds();
}
