//
//   $Id: LONGNAME.H 1.8 2002/10/15 21:44:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   LONGNAME.H  created 12/31/96  by Drew Gislason
   Copyright (c) 1996-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This include file defines the structures and constants
	used by long filenames, but not by the rest of DOS.

**********************************************************/

// various name length maximums
#define MAXPATH  260		// allow room for lfn plus "C:\"
#define MAXLNAME 256		// up to 256 characters in a long filename
#define DOTNAMESIZE 14	// up to 14 characters in an 8.3 filename

// starting handle number for findfirst/next searches
#define START_HANDLE		0x238

// user defined alias hint
#define O_ALIAS_HINT 0x0400

// first long filename DIR entry must have sequence bit set
#define SEQUENCE_ID	0x40
#define SEQUENCE		0x3F

// wildcard characters
#define WILDCARD '*'
#define WILDCHAR '?'
#define IsWildCard(c) ((c)==WILDCARD || (c)==WILDCHAR)

// uniCode character
typedef unsigned uniChar;
typedef uniChar far * LPUNICHAR;
#define UNIFILL 0xFFFF			// a fill character
#define WANSI_CHARS   0			// use Windows ANSI character set
#define DOS_CHARS     1			// use current DOS character set
#define UNICODE_CHARS 2			// use UniCode character set

// used for Volume Info
#define FS_CASE_SENSITIVE		0x0001
#define FS_CASE_IS_PRESERVED	0x0002
#define FS_UNICODE_ON_DISK		0x0004
#define FS_LFN_APIS				0x4000
#define FS_VOLUME_COMPRESSED	0x8000
#define FS_FLAGS (FS_CASE_IS_PRESERVED | FS_UNICODE_ON_DISK | FS_LFN_APIS)

// don't modify last access date when opening
#define OPEN_ACCESS_NOMODLASTACCESS 0x0004

// # of 100-nanosecond intervals since 1-1-1901
// the lowDateTime may also be normal DOS date/time
struct FILETIME
{
	ulong lowDateTime;		// may also be normal DOS date/time
	ulong highDateTime;
};
typedef struct FILETIME FILETIME;
typedef FILETIME       * pFILETIME;
typedef FILETIME   far * lpFILETIME;
typedef FILETIME * far * plpFILETIME;

// how is the above filetime used? ala DOS, or Win32?
#define WIN32DATE 0
#define DOSDATE   1

// returned by 71A6h Get Longname File Info
struct BY_HANDLE_FILE_INFORMATION
{
   uchar	attr;								// 00h file attributes
   uchar	upperAttr[3];					// 01h not used (upper attribs)
   struct FILETIME creationTime;		// 04h only with long filenames
   struct FILETIME lastAccessDate;	// 0Ch only with long filenames
   struct FILETIME lastWriteTime;	// 14h standard DOS file time/date
   ulong volumeSerialNumber;			// 1Ch serial # of disk (if any)
   ulong fileSizeHigh;					// 20h
   ulong	fileSize;						// 24h file size
   ulong numLinks;						// 28h # of times file was opened
   ulong fileIndexHigh;					// 2Ch
   ulong fileIndex;						// 30h index to file
};
typedef struct BY_HANDLE_FILE_INFORMATION far *LP_BY_HANDLE_FILE_INFORMATION;

// returned by a FindFirst/Next
#define FIRST_PART_W32FD 0x2C
struct WIN32_FIND_DATA
{
   uchar	attr;								// 00h file attributes
   uchar	upperAttr[3];					// 01h not used (upper attribs)
   struct FILETIME creationTime;		// 04h only with long filenames
   struct FILETIME lastAccessDate;	// 0Ch only with long filenames
   struct FILETIME lastWriteTime;	// 14h standard DOS file time/date
	ulong	fileSizeHigh;					// 1Ch unused by FAT file system
   ulong	fileSize;						// 20h file size
	ulong reserved[2];					// 24h ???
	char	szFilename[MAXPATH];			// 2Ch long filename, or 8.3 if no lfn
	char	szAliasFilename[DOTNAMESIZE];	// 12Ch 8.3 alias
};
typedef struct WIN32_FIND_DATA * P_WIN32_FIND_DATA;
typedef struct WIN32_FIND_DATA far * LP_WIN32_FIND_DATA;

// ROM-DOS ONLY: More of that found data stuff the user doesn't
// need to know (yeah, right), but the OS certainly does. For
// things like deleting, creating and renaming files/dirs
struct WIN32_MORE_FIND_DATA
{
	CLUSTER	subdir;			// cluster # of subdir containing DIR entry
	unsigned	startDirNum;	// start of long filename entries
	int		numEntries;		// how many entries taken by this longname/8.3
	CLUSTER	startClus;		// starting cluster of file or dir
	bool		onFilename;		// on the filename part when failed
};

#define SPEEDUP_INVALID -1

// ROM-DOS ONLY: This structure is used internally and is NOT
// compatible.  Used as a SearchDataBlock (SDB), but with long
// filename extensions.
struct WIN32_SDB
{
	uchar		inUse;					// 00h is this SDB used
	uchar		drive;					// 01h which drive #
	CLUSTER	subdir;					// 02h start cluster of subdir for next search
	unsigned	dirnum;					// 06h directory entry # to start next search
	ulong		dirsec;					// 08h sector # currently being accessed
   uchar		eps;						// 0Ch DIRs per sector
	uchar		found;					// 0Dh anything found during current search?
	unsigned	startDirNum;			// 0Eh starting DIR # for last found
	int		numEntries;				// 10h num entries found in this longname
	CLUSTER	newSubDir;				// 12h new subdirectory (aka start clus)
	uchar		mustMatchAttr;			// 16h must match these attributes
	uchar		searchAttr;				// 17h standard DOS search attributes
	int		lastLogicalSector;	// 18h speedup: logical sector of "dirnum"
	unsigned	process;					// 1Ah process that opened this search
	char szFileMask[MAXPATH];		// 1Ch mask for long filename search
};
typedef struct WIN32_SDB * P_WIN32_SDB;
typedef struct WIN32_SDB far * LP_WIN32_SDB;
typedef LP_WIN32_SDB * PLP_WIN32_SDB;

// just like a struct DIR, but with the Win32 extensions
// this is the alias
struct WIN_DIR
{
   char     name[8];					// 00h file name and extension
   char     ext[3];					// 08h
   uchar    attr;						// 0Bh file attribute
   uchar    reserved;				// 0Ch
	uchar		milli;					// 0Dh milliseconds
   unsigned creationTime;			// 0Eh file creation date/time
   unsigned creationDate;			// 10h
   unsigned lastAccessDate;		// 12h date file was last accessed (opened)
   unsigned uStartingClusterHi;	// 14h (hi word) starting cluster
   unsigned lastModTime;			// 16h last date/time file was modified
   unsigned lastModDate;			// 18h
#if USING_FAT32
   unsigned	uStartingClusterLo;	// 1Ah (lo word) starting cluster
#else
	unsigned	start;					// 1Ah starting cluster
#endif
	ulong    fileSize;				// 1Ch file size in bytes
};

// long filename directory entry
#define NUM_NAME1 5
#define NUM_NAME2 6
#define NUM_NAME3 2
#define CHARS_IN_LFN_DIR 13
struct LFN_DIR
{
	uchar		sequence;		// 00h sequence number
	uniChar	fileName1[5];	// 01h first 5 characters of name
	uchar		attr;				// 0Bh always 0FH
	uchar		type;				// 0Ch what is this???
	uchar		checksum;		// 0Dh checksum byte
	uniChar	fileName2[6];	// 0Eh next 6 characters of name
	unsigned	reserved;		// 1Ah always 0
	uniChar	fileName3[2];	// 1Ch next 2 characters of name
};
typedef struct LFN_DIR far * LPLFNDIR;

// generic directory entry
union GEN_DIR
{
	DIR				dir;			// standard DIR entry
	struct WIN_DIR	winDir;		// alias
	struct LFN_DIR	lfnDir;		// part of long filename
};
typedef union GEN_DIR far * LPGENDIR;
typedef LPGENDIR * PLPGENDIR;

// in-line functions
//
//  01-04-01 BT
//
//  ETR 3635 ROMDOS LFN kernel would not treat directory entries that
//  were deleted by placing 00h instead of E5h in the first charcter
//  of the name as an available entry.  SCANDISK for DOS will create
//  entries as described above when ever valid entries follow an
//  entry that starts with 00h.  Note also that scandisk (GUI) in
//  windows places E5h in all available entries in the same scenario.
//
//#define IsEndOfDirList(lpDir) (*(lpDir)->dir.name==0 && (lpDir)->dir.attr==0)
#define IsEndOfDirList(lpDir) (*(lpDir)->dir.name==0)

/*
	Used with IOCTL Get/Set Media ID. Known as MID in the MS Prog Ref.

	See page 54 of Microsoft Programmer's Reference 6.0
*/
struct MEDIA_ID
{
	unsigned			infoLevel;
	unsigned long	serialNum;			// serial #
	char 				volLabel[11];		// volume label
	char				fileSysType[8];	// file system type "FAT16"
};

// flags to the DeleteLongFileOrDirectory() function
#define DEL_DIR		0x01	// delete DIRs instead of files?
#define DEL_WILDOK	0x02	// wild cards OK?
#define DEL_CXATTR	0x04	// DOS7 uses CX for attrs on deletes

// rename functions
#define REN_SHORTONLY	0x01	// short names only
#define REN_WILDOK		0x02	// rename multiple things

// SDA variable used with Long Filename Support
extern unsigned int21_ax;
extern LPCDS sda_cdsptr;
extern LPDPB sda_dpbptr;
extern LPSFT sda_sftptr;
extern LPBUFLIST sda_cur_buf;
extern unsigned sda_1st_blank;
extern uchar sda_drive;		// drive for this call only
extern uchar CUR_DRIVE;		// "current" drive from user's point of view
extern char sda_fname1[];	// used to put short name paths
extern char sda_ename1[];
extern unsigned sda_callflags;
extern unsigned current_psp;
extern unsigned machine_id;
extern unsigned sda_handle;
extern uchar sda_sft_h;
extern struct dev_req_hdr sda_reqhdr;
extern LPDEV_HDR sda_devptr;
extern SDB sda_sdb;
extern DIR sda_ffb;
extern FCB sda_fcb;
extern uchar sda_srchattr;
extern unsigned current_psp;
extern uchar sda_in_5D;

// required by special open (716CH)
extern unsigned sopen_act;		/* the special open action (DX) */
extern unsigned sopen_attr;	/* the special open attribute (CX) */
extern unsigned sopen_mode;	/* the special open mode (BX) */
