//
//   $Id: DLS13F41.C 1.8 2002/10/15 22:43:50Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
   $DOCMODULE:DLS13F41.C

	Copyright (c) 1999-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	DLS13F41:Datalight Library Smart Int 13 Function 41 Smart Lba Get
	Drive Parameters.

	This module contains the Smart "Init" function for bios disk calls.
	Call this function to get the Info Packet prior to calling any of the
	other Smart bios disk functions. See DLINT13.H for prototypes to all
	the funtions.

	bool PASCAL DlSmartLbaGetDriveParameters(int iDrive, PLSIP plsipInfo);

	$DOCHISTORY
	03/31/1999 dg  Created by Drew Gislason
****************************************************************************/
#include <string.h>

#include "datalght.h"
#include "dlint13.h"


bool _fEnableCHSLimit = FALSE;


//  This function is used to set a flag which will limit the accessable
//  disk to only what can be accessed with CHS calls.
//
bool PASCAL DlSmartSetCHSLimitFlag( bool fEnableCHSLimit )
{
	_fEnableCHSLimit = fEnableCHSLimit;
}


/*
	$DOCPROC:DlSmartLbaGetDriveParameters

	Get Drive Parameters, using LBA functions if possible. Returns a
	"handle" structure to be used with all other Smart LBA disk functions.

	On Entry
		iDrive				BIOS drive # (eg 0x80)
	On Exit
		plsipInfo			filled in with "handle" information and drive
								geometry. Pass this structure to other smart
								functions.

	Returns TRUE if worked, FALSE if failed.
*/
bool PASCAL DlSmartLbaGetDriveParameters(int iDrive, PLSIP plsipInfo)
{
	uchar		ucLbaVersion;
   //
   // Initialize uEFlags, because if DlLbaCheckIfPresent fails it won't be
   //
	unsigned	uEFlags = 0;
	BDP		bdpParameters;
	ulong		ulTotalSectors;
	bool		fGeometryModified = FALSE;


	// get the version
	if(DlLbaCheckIfPresent(iDrive, &ucLbaVersion, &uEFlags) &&
		DlLbaGetDriveParameters(iDrive, &plsipInfo->ebdpt))
	{
		if(DlBiosGetDriveParameters(iDrive, &bdpParameters))
		{
			//  If this size of drive is supported by the int 13h
			//  function 8h call then set the geometry to that which
			//  is returned by function 8h.  Scenario:  A drive which
			//  does not require LBA but being used on a system which
			//  has LBA support may return a different CHS geometry
			//  because CHS geometry doesn't mean anything in LBA.  It
			//  all gets converted to sectors anyway.
			//
			plsipInfo->ebdpt.ulPhysicalHeads = bdpParameters.ucHeads+1;
			plsipInfo->ebdpt.ulPhysicalSPT = bdpParameters.ucSectorsPerTrack;
			if( _fEnableCHSLimit )
			{
				plsipInfo->ebdpt.ulPhysicalCylinders =
						(unsigned)bdpParameters.iTracks+1;
			}
			else
			{
				plsipInfo->ebdpt.ulPhysicalCylinders =
						plsipInfo->ebdpt.ulBigPhysicalSectors;
				plsipInfo->ebdpt.ulPhysicalCylinders /=
						plsipInfo->ebdpt.ulPhysicalHeads *
						plsipInfo->ebdpt.ulPhysicalSPT;
				ulTotalSectors = plsipInfo->ebdpt.ulPhysicalCylinders *
						plsipInfo->ebdpt.ulPhysicalHeads *
						plsipInfo->ebdpt.ulPhysicalSPT;
				plsipInfo->ebdpt.ulBigPhysicalSectors = ulTotalSectors;
			}
			fGeometryModified = TRUE;
		}

		if( !fGeometryModified )
		{
			ulTotalSectors  = plsipInfo->ebdpt.ulPhysicalHeads;
			ulTotalSectors *= plsipInfo->ebdpt.ulPhysicalSPT;
			ulTotalSectors *= plsipInfo->ebdpt.ulPhysicalCylinders;
			if( ulTotalSectors != plsipInfo->ebdpt.ulBigPhysicalSectors )
			{
				//  Sometimes CHS geometry is bogas.  If it appears to
				//  be BS then calculate our own geometry because CHS
				//  geometry doesn't mean anything in LBA.  It all gets
				//  converted to sectors anyway.
				//
				ulTotalSectors = plsipInfo->ebdpt.ulBigPhysicalSectors;
				plsipInfo->ebdpt.ulPhysicalCylinders =
						ulTotalSectors / (255 * 63);
				plsipInfo->ebdpt.ulPhysicalHeads = 255;
				plsipInfo->ebdpt.ulPhysicalSPT = 63;
			}
		}

		plsipInfo->uEFlags = uEFlags;
	}

	// use old-style function
	else if(DlBiosGetDriveParameters(iDrive, &bdpParameters))
	{
		memset(plsipInfo, 0, sizeof(LSIP));

		plsipInfo->ebdpt.uInfoFlags = EBDPT_INFO_CHS_GEOMETRY_VALID;
		plsipInfo->ebdpt.ulPhysicalCylinders = bdpParameters.iTracks + 1;
		plsipInfo->ebdpt.ulPhysicalHeads     = bdpParameters.ucHeads + 1;
		plsipInfo->ebdpt.ulPhysicalSPT       = bdpParameters.ucSectorsPerTrack;
		plsipInfo->ebdpt.ulBigPhysicalSectors =
			plsipInfo->ebdpt.ulPhysicalCylinders *
			plsipInfo->ebdpt.ulPhysicalHeads *
			plsipInfo->ebdpt.ulPhysicalSPT;
		plsipInfo->ebdpt.uBytesPerSector = 512;
	}

	// not a valid drive
	else
		return FALSE;

	// record drive
	plsipInfo->iDrive = iDrive;

	return TRUE;
}

#if UNIT_TEST
#include <stdio.h>

int main(void)
{
	LSIP lsipInfo;

	// test
	if(!DlSmartLbaGetDriveParameters(0x80, &lsipInfo))
	{
		printf("Failed to get parms for drive 80\n");
		return 1;
	}

	// display results
	printf("lsipInfo.iDrive                     %X\n",lsipInfo.iDrive);
	printf("lsipInfo.uEFlags                    %X\n",lsipInfo.uEFlags);
	printf("lsipInfo.ebdpt.uInfoFlags           %X\n",lsipInfo.ebdpt.uInfoFlags);
	printf("lsipInfo.ebdpt.ulPhysicalCylinders  %lu\n",lsipInfo.ebdpt.ulPhysicalCylinders);
	printf("lsipInfo.ebdpt.ulPhysicalHeads      %lu\n",lsipInfo.ebdpt.ulPhysicalHeads);
	printf("lsipInfo.ebdpt.ulPhysicalSPT        %lu\n",lsipInfo.ebdpt.ulPhysicalSPT);
	printf("lsipInfo.ebdpt.ulBigPhysicalSectors %lu\n",lsipInfo.ebdpt.ulBigPhysicalSectors);
	printf("lsipInfo.ebdpt.uBytesPerSector      %u\n",lsipInfo.ebdpt.uBytesPerSector);

	printf("Everything worked\n");
	return 0;
}

#endif
