//
//   $Id: TSTINT13.C 1.5 2002/10/15 22:43:50Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
	$DOCMODULE:TSTINT13.C

   Copyright (c) 1999-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	This module is the unit test for the Int 13h BIOS Disk functions.

	$DOCHISTORY
	03/30/1999 dg  Created by Drew Gislason
***************************************************************************/
#include <string.h>
#include <stdio.h>
#include "datalght.h"
#include "dlint13.h"

char cBuffer[1024];
char cBuffer2[1024];

int main(void)
{
	int		iDrive    = 0x80;
	int		iFloppy   = 0;
	int		iBadDrive = 5;
	BDTP		bdtpTransferPacket;
	unsigned					uOrgValue;
	#define					CHANGED_OFFSET	0x1A0
	LBA_PACKET				lbaPacket;
	LBA_EXT_DRV_PARMS 	lbaExtDrvParms;

	// test reset
	if(!DlBiosResetDisk(iDrive))
	{
		printf("Reset failed!\n");
		return 1;
	}

	// test get disk status
	DlBiosGetDiskStatus(iDrive);

	// read a sector from floppy drive A:
	bdtpTransferPacket.iTrack		 = 0;
	bdtpTransferPacket.ucSector	 = 1;
	bdtpTransferPacket.ucHead		 = 0;
	bdtpTransferPacket.iNumSectors = 1;
	bdtpTransferPacket.lpBuffer	 = cBuffer;
	if(!DlBiosReadSectors(ucFloppy, &bdtpTransferPacket))
	{
		printf("Error! i13f02 drv %Xh\n",ucFloppy);
		return 2;
	}

	// write a sector to floppy drive A:
	if(!DlBiosWriteSectors(ucFloppy, &bdtpTransferPacket))
	{
		printf("Error! i13f03 drv %Xh\n",ucFloppy);
		return 3;
	}

	// verify a sector on floppy drive A:
	if(!DlBiosVerifySectors(ucFloppy, &bdtpTransferPacket))
	{
		printf("Error! i13f04 drv %Xh\n",ucFloppy);
		return 4;
	}

	// test dlLbaCheckIfPresent
	if(!dlLbaCheckIfPresent(ucDrive))
	{
		printf("Error! should be present for drive %Xh\n",ucDrive);
		return 1;
	}
	if(dlLbaCheckIfPresent(ucBadDrive))
	{
		printf("Error! should NOT be present for drive %Xh\n",ucBadDrive);
		return 2;
	}

	// test read/write/verify
	lbaPacket.iNumBlocks   = 1;
	lbaPacket.lpBuffer     = cBuffer;
	lbaPacket.ulSectorLow  = 0;
	lbaPacket.ulSectorHigh = 0;
	if(!dlLbaReadSectors(ucDrive, &lbaPacket))
	{
		printf("Error reading MBR!\n");
		return 3;
	}

	// test verify
	if(!dlLbaVerifySectors(ucDrive, &lbaPacket))
	{
		printf("Error verifying MBR!\n");
		return 4;
	}

	// change to new contents and read it back
	uOrgValue = *(unsigned *)(cBuffer+CHANGED_OFFSET);
	*(unsigned *)(cBuffer+CHANGED_OFFSET) = 0x1234;
	if(!dlLbaWriteSectors(ucDrive, &lbaPacket))
	{
		printf("Error writing MBR!\n");
		return 5;
	}
	lbaPacket.lpBuffer     = cBuffer2;
	if(!dlLbaReadSectors(ucDrive, &lbaPacket))
	{
		printf("Error reading MBR2!\n");
		return 6;
	}
	if(*(unsigned *)(cBuffer2+CHANGED_OFFSET) != 0x1234)
	{
		printf("Write didn't take to MBR!\n");
		return 7;
	}

	// change back to original
	*(unsigned *)(cBuffer+CHANGED_OFFSET) = uOrgValue;
	lbaPacket.lpBuffer     = cBuffer;
	if(!dlLbaWriteSectors(ucDrive, &lbaPacket))
	{
		printf("Error writing MBR2!\n");
		return 8;
	}

	// test get drive parameters
	if(!dlLbaGetDriveParameters(ucDrive, &lbaExtDrvParms))
	{
		printf("Failed to get Drive Parameters\n");
		return 9;
	}
	else
	{
		printf("iSize %u\n",lbaExtDrvParms.iSize );
		printf("iFlags %u\n",lbaExtDrvParms.iFlags );
		printf("ulCylinders %lu\n",lbaExtDrvParms.ulCylinders );
		printf("ulHeads %lu\n",lbaExtDrvParms.ulHeads );
		printf("ulSectors %lu\n",lbaExtDrvParms.ulSectors );
		printf("ulTotalSectorsLow %lu\n",lbaExtDrvParms.ulTotalSectorsLow );
		printf("ulTotalSectorsHigh %lu\n",lbaExtDrvParms.ulTotalSectorsHigh );
		printf("uBytesPerSector %u\n",lbaExtDrvParms.uBytesPerSector );
		printf("ulReserved %lX\n",lbaExtDrvParms.ulReserved );
	}

	printf("Everything worked\n");
	return 0;
}
