;
;   $Id: DLABSRD.ASM 1.6 2002/10/15 22:52:14Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;--------------------------------------------------------------------
; $DOCMODULE:DLABSRD.ASM
; Created 12-22-1997 by Dennis Edwards
;
; Copyright (c) 1997-2002 Datalight, Inc.
; All Rights Reserved.
; Datalight Confidential and Proprietary.
;
; This module is a wrapper for interrupt 25 absolute sector read
; and is part of the Datalight Library.
;
; bool PASCAL dlAbsRead(bool bIsBig, int iDrive, int iSects, long lStart,
;		void far * lpBuffer)
;
; $DOCHISTORY
; 02/01/1999 de  Another attempt to make it work
; 01/04/1998 bt  Made the call update the global error code variable
; 01/04/1998 bt  Fixed an error that would have occured if a
;                compact or larger memory model was used.
; 01/04/1998 bt  Commented the code
; 07/22/1998 bt  copied it from defrag and made it into library
;		 function.
; 07/22/1998 bt  made function return a boolean instead of 0 and 1
; 12/22/1997 de  Created.
;--------------------------------------------------------------------
include datalght.inc
InitModel %MEMMODEL, pascal, use16
include _dldosi2.inc


ABS_DISK_IO	struc
	startingSector	dd	?
	sectorCount	dw    	?
	buffer		dd	?
ABS_DISK_IO	ends


.code

;--------------------------------------------------------------------
; $DOCPROC: dlAbsRead
;
; bool PASCAL dlAbsRead(bool bIsBig, int iDrive, int iSects, long lStart,
;		void far * lpBuffer)
;
; Int 25h interface to read logical sectors (64Kbytes max).
; Returns TRUE if successful, FALSE if it failed.
;
; On Entry
;
;	none  (bIsBig)	boolean: drive has more than 65536 sectors
;	al    (iDrive)	Drive number (A=0,B=1,C=2,...)
;	cx    (iSects)	> 0 number of sectors to read
;			= -1 use extended format
;	dx    (lStart)	starting sector number
;	ds:bx (pBuffer)	pointer to DTA if cx > 0, pointer to parameters
;			if cx = -1
; On Exit
;	ax		TRUE if the function was successful, FALSE otherwise
;
;--------------------------------------------------------------------
ProcBegin 	dlAbsRead, public
ProcUses  	ds,si,di,bx
ProcArgs	@@isBig:word, @@drive:word, @@sects:word, @@start:dword, @@lpbuff:dword
ProcLocals	@@absReadData:ABS_DISK_IO
ProcSetup


	; Load the drive number, common for both "large" and "small" disks
	;
	mov	al, byte ptr @@drive

	; If this a "small" drive, jump to code that loads args into registers
	; otherwise we have to pass the args by reference to a data struct
	;
	mov	dx, word ptr @@isBig
	or	dx, dx
	jz	ABSR_is_small

	; this is a BIG (32M+) partition, pass data in a struc
	; pointed to by DS:BX, CX=-1 signals this, AL = drive
	; Set the flag to indicate we are using a parameter block
	;
	mov	cx, -1

	; Set the starting sector number
	;
	mov	dx, word ptr @@start
	mov	word ptr @@absReadData.startingSector, dx
	mov	dx, word ptr @@start[2]
	mov	word ptr @@absReadData.startingSector[2], dx

	; Set the number of sectors to read
	;
	mov	dx, @@sects
	mov	@@absReadData.sectorCount, dx

	; Set the destination buffer
	;
	mov	dx, word ptr @@lpBuff
	mov	word ptr @@absReadData.buffer, dx

	mov	dx, word ptr @@lpBuff[2]
	mov	word ptr @@absReadData.buffer[2], dx

	; Set the address of the parameter block in DS:BX
	; then perform the absolute sector read
	;
	push	ds
	mov	bx, ss
	mov	ds, bx
	lea	bx, @@absReadData
	jmp	short ABSR_do_int25

	; this is a SMALL (<=32M) partition, all args fit in registers
	; Set the number of sectors to read
	;
    ABSR_is_small:
	mov	cx, @@sects

	; Set the starting sector number
	;
	mov	dx, word ptr @@start

	; Set the address to the buffer
	;
	push	ds
	lds	bx, @@lpBuff

	; Perform the absolute sector read
	; int 25h/26h leave the flags on the top of the stack
	; get rid of them
	;
    ABSR_do_int25:
	int	25h
	pop	bx
	pop	ds

	; Set the return value and the internal error code variable
	;
	SET_ABSRDWR_ERROR

    ABSR_read_done:
ProcEnd

@curseg	ends
	end
