//
//   $Id: WILDDEL6.C 1.5 2002/10/15 22:49:52Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  $DOCMODULE:WILDDEL.C
//
//  Copyright (c) 1999-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary.
//
//  This module is a function to perform a wild card delete using a find
//  first, delete, and find next.  This is part of the Datalight Library.
//
//  bool dlWildcardDelete( char * szFileName, FIND * pFFBlock,
// 		char * szTmpNameAndPath );
//
//  $DOCHISTORY
//
//  09/23/1999 bt  Created.
//

//
//  [CONSTANTS]
//

//  Borland libraries
//
#include <string.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"
#include "dlpath.h"


//  $DOCPROC: dlWildcardDelete
//
//  bool dlWildcardDelete( char * szFileName, FIND * pFFBlock,
// 		char * szTmpNameAndPath );
//
//  This module is a function to perform a wild card delete using a find
//  first, delete, and find next.  This is part of the Datalight Library.
//
//  ON ENTRY:
//    szFileName - Name and path of the file.  May contain wild cards.
//		pFFBlock   - Find first block needed to search for wildcards
//    szTmpNameAndPath - Temp variable (char szTmpNameAndPath[129])
//
//  ON EXIT:
//    returns TRUE if at least one file was deleted and every found file
//    was deleted
//
bool dlWildcardDelete( char * szFileName, FIND * pFFBlock,
		char * szTmpNameAndPath )
{
	bool     fSuccessful;
	bool		fContainsWildcards = FALSE;
	unsigned uFileCount = 0;
	void far * fpDTAAddress;
	char * szLastPart;
	unsigned uAttributes = 0x07;
	unsigned uIndex = 0;
	int	iErrorCode = 0;


	//  Save the current DTA
	//
	dlGetDTA(&fpDTAAddress);

	//  Set the DTA to our find first block
	//
	dlSetDTA(pFFBlock);

	//  if there is no wild cards we should not skip read-only files
	//  thus generating an accessed denied error code
	//
	while( szFileName[uIndex] != 0 )
	{
		if( (szFileName[uIndex] == '?') || (szFileName[uIndex] == '*') )
		{
			fContainsWildcards = TRUE;
			break;
		}

		uIndex++;
	}

	//  Check for the existance of the file
	//
	fSuccessful = dlFindFirstFile( uAttributes, szFileName);
	if( fSuccessful )
	{
		//  We don't want to set and restore the CWD because in
		//  the event of an error our GLOBAL error code will be
		//  overwritten and will not return the cause of the
		//  error with dlCheckDOSError().  Instead we will create
		//  the path to the file and append the name of the files
		//  which are found.
		//
		strcpy( szTmpNameAndPath, szFileName );
		szLastPart = LastPart( szTmpNameAndPath );
		while( fSuccessful )
		{
			strcpy( szLastPart, pFFBlock->name );
			fSuccessful = dlDeleteFile( szTmpNameAndPath );
			if( fSuccessful )
			{
				//  Mark that we have deleted another file
				//
				uFileCount++;
			}

			if( fContainsWildcards )
			{
            // Save off the previous error state
            //
            iErrorCode = dlCheckDOSError();

				//  Look for the next file
				//
				fSuccessful = dlFindNextFile();
			}
		}

      if( fContainsWildcards && uFileCount==0 )
      {
         // dlWildcardDelete() only returns the boolean state of success.
         // If there's a failure, dlCheckDOSError() will help determine
         // what the problem was.  BUT now the error has the state of
         // the dlFindNextFile()!  So we need to reset it...
         //
         dlSetDOSError( iErrorCode );
      }
	}

	//  Restore the DTA
	//
	dlSetDTA(fpDTAAddress);

	//  Return whether the function was successful
	//
	return (uFileCount != 0);
}
