//
//   $Id: LFNARG.C 1.9 2002/10/15 22:54:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:LFNSMART.C

	Copyright (c) 1999-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	This module has generic LFN argument string manipulation.

	$DOCHISTORY
	03/09/1999 dg  Added findfirst/next and file functions
*********************************************************************/
#include <dir.h>
#include <ctype.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dllfn.h"
#include "dlcntry.h"
#include "fatworld.h"


/*
	$DOCPROC:LFNEndArg

	Find the end of this argument. Is "quote" and international aware.

	Returns a pointer to the first space after the argument, or the
	terminating null if no spaces after argument.

	Examples:
	this is four args
	"this is one arg"
	"this is also one ""arg"
	"this is two " "args"
*/
char * PASCAL LFNEndArg(char *szArg)
{
	bool bInQuote = FALSE;
	static bool fDOSSet = FALSE;
	static bool fIsFat32World = FALSE;
	static bool fLFNPresent = FALSE;


	if( !fDOSSet )
	{
		//  Determine if we have LFNs
		//
		fLFNPresent = LFNPresent();

		//  Determine if we are operating in a DOS 7.1 enviornment
		//
		fIsFat32World = dlIsFat32World();

		//  Never set these again
		//
		fDOSSet = TRUE;
	}

	// skip past all the text for this arg
	while(*szArg && (bInQuote || !DL_IS_SPACE(*szArg)))
	{
		if(IsDoubleByteChar(*szArg))
		{
			++szArg;
		}
		else if(*szArg == '"')
		{
			if( (fLFNPresent) || (fIsFat32World) )
			{
				bInQuote = !bInQuote;
			}
		}
		++szArg;
	}

	// return end of the argument
	return (szArg);
}

#if UNIT_TEST

char * pszEndArgTests[] =
{
	"this is four args", " is four args",
	"\"this is one arg\"", "",
	"\"this is also one \"\"arg\"", "",
	"\"this is two \" \"args\"", " \"args\"",
	0
};

int main(void)
{
	char *pszEndArg;
	int	i;

	// verify LFNEndArg
	i = 0;
	while(pszEndArgTests[i])
	{
		pszEndArg = LFNEndArg(pszEndArgTests[i]);
		if(strcmp(pszEndArg, pszEndArgTests[i+1]) != 0)
		{
			printf("Error in test <%s>\n",pszEndArgTests[i]);
			return 1;
		}
		i += 2;
	}

	// assumes that LFNNextArg works because it just does a skip white

	printf("Everything passed\n");
	return 0;
}

#endif // UNIT_TEST
