//
//   $Id: LFNSPLIT.C 1.6 2002/10/15 22:54:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
	$DOCMODULE:LFNSPLIT.C

	Copyright (c) 1999-2002 Datalight, Inc.
	Datalight Confidential and Proprietary.
   All Rights Reserved.

	Split a long filename into its components.

	$DOCHISTORY
	02/22/1999 dg  Created by Drew Gislason
***************************************************************************/

#if UNIT_TEST
#include <stdio.h>
#endif
#include <mem.h>
#include <dir.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dllfn.h"
#include "dlcntry.h"

#if UNIT_TEST

char szDrive[SPLIT_MAXDRIVE];
char szDirectory[DL_MAXPATH];
char szFileName [DL_MAXPATH];

typedef struct
{
	char *szSrcPath;
	char *szExpectedDrive;
	char *szExpectedDirectory;
	char *szExpectedFileName;
	int  iExpectedFlags;
} TestCase;

// unit test for intl_fnsplit
TestCase testCases[] =
{
	// shortname tests
	{ "",								// () () ()
	  "", "", "",
	  0 },

	{ " ",							// () () ()
	  "", "", "",
	  0 },

	{ ".",							// () (.) ()
	  "", ".", "",
	  SPLIT_DIRECTORY },

	{ "..",							// () (..) ()
	  "", "..", "",
	  SPLIT_DIRECTORY },

	{ ".EXT",						// () () (.EXT)
	  "", "", ".EXT",
	  SPLIT_FILENAME },

	{ ".\\NAME.",					// () (.\) (NAME.)
	  "", ".\\", "NAME.",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "\\.",							// () (\.) ()
	  "", "\\.", "",
	  SPLIT_DIRECTORY },

	{ "\\..",						// () (\..) ()
	  "", "\\..", "",
	  SPLIT_DIRECTORY },

	{ "\\.EXT",						// () (\) (.EXT)
	  "", "\\", ".EXT",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "\\..\\NAME.EXT",			// () (\..\) (NAME.EXT)
	  "", "\\..\\", "NAME.EXT",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "\\.\\NAME.EXT",			// () (\.\) (NAME.EXT)
	  "", "\\.\\", "NAME.EXT",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "\\..\\NA?ME",				// () (\..\) (NA?ME)
	  "", "\\..\\", "NA?ME",
	  SPLIT_DIRECTORY | SPLIT_FILENAME | SPLIT_WILDCARDS },

	{ "\\.\\NAME*",				// () (\.\) (NAME*)
	  "", "\\.\\", "NAME*",
	  SPLIT_DIRECTORY | SPLIT_FILENAME | SPLIT_WILDCARDS },

	{ "\\.\\NAME.",				// () (\.\) (NAME.)
	  "", "\\.\\", "NAME.",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "NAME",						// () () (NAME)
	  "", "", "NAME",
	  SPLIT_FILENAME },

	{ "*NAME.EXT",					// () () (*NAME.EXT)
	  "", "", "*NAME.EXT",
	  SPLIT_FILENAME | SPLIT_WILDCARDS },

	{ "NAME .EXT",					// () () (NAME .EXT)
	  "", "", "NAME .EXT",
	  SPLIT_FILENAME },

	{ "NAME .EX  ",				// () () (NAME .EX)
	  "", "", "NAME .EX",
	  SPLIT_FILENAME },

	{ "\\NAME",						// () (\) (NAME)
	  "", "\\", "NAME",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "\\.EXT",						// () (\) (.EXT)
	  "", "\\", ".EXT",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "\\NAME.EXT",				// () (\) (NAME.EXT)
	  "", "\\", "NAME.EXT",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "PATH\\NAME",				// () (PATH\) (NAME)
	  "", "PATH\\", "NAME",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "\\PATH\\NAME",				// () (\PATH\) (NAME)
	  "", "\\PATH\\", "NAME",
	  SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "A:",							// (A:) () ()
	  "A:", "", "",
	  SPLIT_DRIVE },

	{ "A:\\",						// (A:) (\) ()
	  "A:", "\\", "",
	  SPLIT_DRIVE | SPLIT_DIRECTORY },

	{ "A:*",							// (A:) () (*)
	  "A:", "", "*",
	  SPLIT_DRIVE | SPLIT_FILENAME | SPLIT_WILDCARDS },

	{ "A:\\NAME",					// (A:) (\) (NAME)
	  "A:", "\\", "NAME",
	  SPLIT_DRIVE | SPLIT_DIRECTORY | SPLIT_FILENAME },

	{ "A:\\PATH\\.NAME.EXT*",	// (A:) (\PATH\) (.NAME.EXT*)
	  "A:", "\\PATH\\", ".NAME.EXT*",
	  SPLIT_DRIVE | SPLIT_DIRECTORY | SPLIT_FILENAME | SPLIT_WILDCARDS },

	// longname tests
	{ "A:\\Long Subdir\\This is a long filename .exten?sion",
	  "A:", "\\Long Subdir\\", "This is a long filename .exten?sion",
	  SPLIT_DRIVE | SPLIT_DIRECTORY | SPLIT_FILENAME | SPLIT_WILDCARDS },

	// end of list
	{ 0 }
};

#endif // UNIT_TEST

/*
	FileName Split, with International (2-byte char) and LongName support.
	Does no checking for validity of characters. LongName quotes must be
	stripped prior to calling this function.

	On Entry		szSrcName		name of relative path (no quotes)
	On Exit		szDrive			drive letter filled in (if any)
					szDirectory		directory name filled in (if any)
					szFileName		filename filled in (if any)
					flags				SPLIT_WILDCARDS
										SPLIT_FILENAME
										SPLIT_DIRECTORY
										SPLIT_DRIVE

	Does no checking for validity of characters. Does not access the
	drive in question (string manipulation only). Cannot fail.
*/
int PASCAL LFNSplitFileName(char *szSrcName, char *szDrive,
				char *szDirectory, char *szFileName)
{
	int iFlags = 0;
	int iLen;
	char * psz;
	char * szEndDirPart = 0;
	char * szLastWhiteSpace = 0;
	char * szEndSrcName;

	// initialize each string
	*szDrive = *szDirectory = *szFileName = 0;

	// do we have a drive part?
	if(*szSrcName && !IsDoubleByteChar(*szSrcName) && *(szSrcName+1)==':')
	{
		*szDrive = *szSrcName;
		*(szDrive+1) = ':';
		*(szDrive+2) = 0;
		szSrcName += 2;

		iFlags |= SPLIT_DRIVE;
	}

	// scan string for a directory part and wildcards
	psz = szSrcName;
	while(*psz)
	{
		// handle directory part
		if(*psz == '\\')
			szEndDirPart = psz+1;

		// handle wildcards
		if(*psz == '*' || *psz == '?')
			iFlags |= SPLIT_WILDCARDS;

		// for handling trailing spaces
		if(*psz != ' ' && *psz != '\t')
			szLastWhiteSpace = psz+1;

		if(IsDoubleByteChar(*psz))
			++psz;
		++psz;
	}

	// determine end of string (but don't count trailing spaces)
	if(!szLastWhiteSpace)
		szLastWhiteSpace = szSrcName;
	if(AllSpaces(szLastWhiteSpace))
		szEndSrcName = szLastWhiteSpace;
	else
		szEndSrcName = psz;

	// special case: "." and ".." are really directories, not filenames
	if(szEndDirPart)
		psz = szEndDirPart;
	else
		psz = szSrcName;
	if(IsDotDirectory(psz))
		szEndDirPart = psz+IsDotDirectory(psz);

	// is there a directory part?
	if(szEndDirPart)
	{
		// determine length of string. don't allow too-long strings
		iLen = (int)(szEndDirPart - szSrcName);
		if(iLen > DL_MAXPATH-1)
			iLen = DL_MAXPATH-1;

		// copy in the directory part
		memmove(szDirectory, szSrcName, iLen);
		*(szDirectory+iLen) = 0;

		// on to next part
		szSrcName = szEndDirPart;

		// has a directory
		iFlags |= SPLIT_DIRECTORY;
	}

	// ok, the rest is the filename part, if any
	if(szSrcName < szEndSrcName)
	{
		iLen = (int)(szEndSrcName - szSrcName);
		memmove(szFileName, szSrcName, iLen);
		*(szFileName+iLen) = 0;

		iFlags |= SPLIT_FILENAME;
	}

	// return the flags that were found
	return iFlags;
}

#if UNIT_TEST

#define ERR_FLAGS			1
#define ERR_DRIVE			2
#define ERR_DIRECTORY	3
#define ERR_FILENAME		4

int main(void)
{
	int	iFlags;
	TestCase * pTestCase;

	// test all the cases
	pTestCase = testCases;
	while(pTestCase->szSrcPath)
	{
		iFlags = InternationalSplitFileName(pTestCase->szSrcPath, szDrive,
						szDirectory, szFileName);

		// verify flags
		if(iFlags != pTestCase->iExpectedFlags)
		{
			printf("Invalid flags on test case %s\n",pTestCase->szSrcPath);
			return ERR_FLAGS;
		}

		// verify drive string
		if(strcmp(szDrive, pTestCase->szExpectedDrive) != 0)
		{
			printf("Invalid drive %s on test case %s\n",
				pTestCase->szExpectedDrive, pTestCase->szSrcPath);
			return ERR_DRIVE;
		}

		// verify directory string
		if(strcmp(szDirectory, pTestCase->szExpectedDirectory) != 0)
		{
			printf("Invalid directory %s on test case %s\n",
				pTestCase->szExpectedDirectory, pTestCase->szSrcPath);
			return ERR_DIRECTORY;
		}

		// verify filename string
		if(strcmp(szFileName, pTestCase->szExpectedFileName) != 0)
		{
			printf("Invalid filename %s on test case %s\n",
				pTestCase->szExpectedFileName, pTestCase->szSrcPath);
			return ERR_FILENAME;
		}

		// on to next case
		++pTestCase;
	}

	printf("All tests passed\n");
	return 0;
}
#endif	// UNIT_TEST
