//
//   $Id: smtdel.c 1.3 2002/10/15 22:54:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************

	$DOCMODULE:SMTDEL.C

   Copyright (c) 1998-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This smart functions works whether long filenames are present or
	not. With long filenames present, the functions use the LFN
	equivalents. With no long filename support, the older, DOS2 style
	functions are used.

	The calling app need only make sure that all path strings conform
	to the longname string limits "LFN_MAX_PATH".

	$DOCHISTORY

   05-01-2001 td  Created this function to get away from
                  SmartWildCardDelete, which uses 5D calls.
                  These are not supported under Windows NT.

*********************************************************************/

#include <ctype.h>
#include <dir.h>

#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlerrcod.h"

/*
	$DOCPROC:SmartDelete

	Delete one file.  LFN support contingent upon DOS supporting
   long filenames

	Returns TRUE if file(s) deleted.
*/
bool PASCAL	SmartDelete(char * szFileName)
{
	bool	fSuccessful;
	int	iErrorCode = 0;

	uchar ucDrive;

	if( szFileName[1] == ':' )
	{
		ucDrive = toupper(szFileName[0]) - 'A' + 1;
	}
	else
	{
		ucDrive = 0;
	}

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
		fSuccessful = LFNDeleteFiles(0, szFileName, DEL_NOWILD);
		if( !fSuccessful )
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				fSuccessful = dlDeleteFile( szFileName );
			}
		}
	}
	else
	{
		//  Retrieve a set of temperary variables used by
		//  both the wildcard rename and wildcard delete
		//
   	// Network drives may not support LFNs
   	//
   	fSuccessful = dlDeleteFile( szFileName );
	}

	return fSuccessful;
}
