;
;   $Id: DL8250A.ASM 1.5 2002/10/15 22:57:48Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;
; DL8250A.ASM
;
;  Copyright (c) 1989-2002 Datalight, Inc.
;  All rights reserved.
;  Datalight Confidential and Proprietary
;
; Code to handle serial communications interrupts. Handles shared IRQs
; and arbitrary buffer sizes.
;
; Limitations of this module:
;	 - Small data models only (ISR code, near ptrs in ring_buffer strucs).
;	 - Only allows single COM port (ISR code, single comport struc extern).
;    - Only handles Rx Data Available interrupts (ISR code, only 1
;      ring_buffer struc in the comport struc definition).
;
include macros.inc


; A circular queue data structure.
;
ring_buffer struc
	head	dw	?	; the head (write ptr) of the ring buffer
	tail	dw	?	; the tail (read ptr) of the ring buffer
	bufsize	dw	?	; the size of the ring buffer
	buffer	dw	?	; a near ptr to the ring buffer data
ring_buffer ends

; Our COM port definition.
;
comport	struc
	base        dw	?	; I/O base address of COM port
	iIrqNumber  dw ?	; Irq number for the port
	old_vect    dd	?	; far pointer to previous ISR
	old_mask    db	?	; old 8259 Interrupt Mask Register value (1st PIC)
   old_mask2   db ?  ; old 8259 Interrupt Mask Register value (2nd PIC)
	Rx		ring_buffer <>	; the receive buffer
comport	ends



; Bit definitions for the INS8250 Interrupt ID Register (IIR) at Base+2
; For more 8250 info, see DL8250.H
;
IIR_ModemStat	=	0
IIR_NoInt    	=	1
IIR_TxData   	=	2
IIR_RxData      =	4
IIR_RxStat	=	6

; using the 3 bits above will mask off all but the int ID bits from the IIR
;
ANY_INT_MASK 	= 	7



; Bit definitions for the INS8250 Interrupt Enable Register (IER) at Base+1
;
IER_RxInt	=	1
IER_TxInt	=	2
IER_RxStat	=	4
IER_ModemStat	= 	8
SUPPORTED_INTS_MASK  =	IER_RxInt	; ONLY Rx Data availble supported


LINESTATUS   	=	5
LS_DRDY 	=	1


; A break interrupt is a special case of the RxStat interrupt
; (which also occurs for overrun, parity and framing errors)
; identified by a BrkInt bit in the Line Status Register (LSR)
;
LSR_BrkInt	=	10h


; Port/bit definitions for 8259 Programmable Interrupt Controller (PIC)
;
PIC_EOIVal	=	20h

PIC_EOIPort1	=	20h
PIC_MaskPort1	=	21h	; each bit: 0=interrupt enabled, 1=masked off

PIC_EOIPort2	=	0A0h
PIC_MaskPort2	=	0A1h	; each bit: 0=interrupt enabled, 1=masked off


cBegData
	extrn _cp:comport
cEndData



cBegCode

comment ~

 The COM port interrupt.

 NOTE: The _cp struct must have been initialized before this routine
 		ever gets control.


void far interrupt comm_int_handler(void)

end_comment ~

c_func comm_int_handler

		push	si
		push	ds                      ; save registers we will trash

		push	ax
		push	dx

		assume	ds:_DATA

		mov	si, seg	_cp		; set DS:SI to our port definition
		mov	ds, si
		mov	si, offset _cp
		mov	dx, [si]		; get COM port base I/O address

	CRCI_is_it_our_int:
		inc	dx			; point to IIR at Base+2
		inc	dx
		in	al, dx			; if no interrupt pending for our port
		and	al, ANY_INT_MASK
		cmp	al, IIR_NoInt
		je	CRCI_chain_interrupt	; then pass it on

	Drain:
IF 0
		cmp	al, IIR_RxData
		je	CRCI_get_comm_input	; handle Rx data available ints
ELSE
		in	al, dx			; if no interrupt pending for our port
		and	al, ANY_INT_MASK
		cmp	al, IIR_RxData
		je	CRCI_get_comm_input	; handle Rx data available ints
		cmp	al, IIR_NoInt
		jne	Drain
ENDIF

	; we could implement interrupt driven transmits, break detection
	; and hardware handshaking by adding code to support these
	; additional interrupt types below.
	;
	CRCI_not_Rx_data_interrupt:

	; We received an interrupt we couldn't handle.
	; Restore the proper value in the UART's IER and return.
	;
	CRCI_unsupported_interrupt:
		mov	al, SUPPORTED_INTS_MASK ; set bits for our ints only
		dec	dx			; IER is at Base+1
;		out	dx, al			; reprogram the 8259
		dec	dx			; Get dx back to BASE register
		jmp	short CRCI_done

	CRCI_get_comm_input:

		dec	dx    			; point to comm data register
		dec	dx

	CRCI_next_char:

		push	bx
		push	cx

		in	al, dx			; read the incoming data

	CRCI_buffer_new_char:
		mov	bx, [si].Rx.head
		mov	cx, bx
		add	bx, [si].Rx.buffer	; DS:BX is the current buffer slot
		mov	[bx], al       		; buffer the new char

	CRCI_advance_buf_head:
		inc	cx			; point to next spot in buffer
		cmp	cx, [si].Rx.bufsize	; need to wrap to buffer start?
		jb	CRCI_update_buff_position

		xor	cx, cx			; past end of buffer, wrap around

	CRCI_update_buff_position:
		mov	[si].Rx.head, cx	; update the head ptr

		pop	cx
		pop	bx

		inc	dx
		inc	dx
		jmp	Drain

	CRCI_done:
		add	dx, LINESTATUS
		in	al, dx			; read LSR
		sub	dx, LINESTATUS
		test	al, LS_DRDY		; see if more Rx data available
		jnz	CRCI_next_char		; if so, get it

		cmp	[si].iIrqNumber, 8
		jb	CRCI_1st_PIC

		mov	al, PIC_EOIVal			; sending non-specific EOI to the 2nd PIC
		out	PIC_EOIPort2, al		; enables this and lower priority IRQs

	CRCI_1st_PIC:
		mov	al, PIC_EOIVal			; sending non-specific EOI to the 1st PIC
		out	PIC_EOIPort1, al		; enables this and lower priority IRQs

		pop	dx
		pop	ax
		jmp	short CRCI_exit

	; We call rather than JMP FAR to be re-entrant, the cost is ten
	; extra bytes from the current stack.
	;
	CRCI_chain_interrupt:
		pop	dx
		pop	ax
		pushf
		call	dword ptr ds:[si].old_vect

		assume	ds:nothing

	CRCI_exit:
		pop	ds
		pop	si
		iret
c_endp comm_int_handler


cEndCode
	end
