//
//   $Id: DLATIME.C 1.6 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   DLATIME.C  created 04/30/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Convert file date/time to ascii versions, using foreign
	country date/time separators.

	Datalight ROM-DOS Tools Library:
	void InitCountryInfo(void);
	void TimeToAscii(unsigned time, char szStr[]);
	void DateToAscii(unsigned date, char szStr[]);
	bool AsciiToDate(unsigned *puDate, char szStr[]);

	Revision History
	03/16/99	jmb	added AsciiToDate

********************************************************************/
#define __FILENUM__ 511

// includes
#include <dos.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"
#include "dldate.h"

struct COUNTRY_INFO countryInfo;

/*
	Initialize the foreign country info.  Get it from DOS.
*/
void InitCountryInfo(void)
{
	union REGS r;

	// used for foreign country support with printf()
	extern char printf_thousands_sep;

	dlGetCountryInfo( 0x01, 0xFFFF, sizeof(countryInfo), 0xFFFF,
			(uchar *)(&countryInfo) );

	// set up the thousands separator
	printf_thousands_sep = *countryInfo.thousand;
}

#if UNIT_TEST
void main(void)
{
	char *szGoodDate[] =
							{
							  "1/1/80",   "12/31/2007",		// year boundary conditions
	                    "1/1/1980", "1/1/79",
							  "2/29/80",  "2/29/1980",		// valid leap years
							  "2/29/2004",
							  "1/31/99",  "2/28/99",		// valid day tests
							  "3/31/99",  "4/30/99",
							  ""
							};

	char *szBadDate[] =
							{
							  "1/1/179",   "12/31/108",		// just outside year boundaries
							  "1/1/1979", "12/31/2108",
							  "2/29/99",  "2/29/1999",		// not leap years
							  "5/0/99",   "6/31/99",		// invalid day tests
							  "7/32/99",  "8/99/99",
							  "0/1/99",   "13/31/99",		// invalid month tests
							  ""
							};
	int i;
	unsigned uDate, uCompareDate;
	char szDate[15];


	// initialize country information
	InitCountryInfo();

	// initalize counter
	i = 0;

	// do good tests
	while (*szGoodDate[i])
	{
		// get test date numeric value
		if (AsciiToDate(&uDate, szGoodDate[i]) == FALSE)
		{
			printf("FAIL %s failed to convert\n", szGoodDate[i]);
			goto next;
		}

		// convert it back to ascii
		DateToAscii(uDate, szDate);

		// get the value again - using the newly converted string
		if (AsciiToDate(&uCompareDate, szDate) == FALSE)
		{
			printf("FAIL Comparison date (%s) failed to convert\n", szDate);
			goto next;
		}

		// compare the dates
		if (uDate == uCompareDate)
			printf("PASS %s converted!\n", szGoodDate[i]);
		else
			printf("FAIL %s (%u) did not match %s (%u)\n",
						szGoodDate[i], uDate, szDate, uCompareDate);

next:
		i++;
	}

	// initalize counter
	i = 0;

	// do bad tests
	while (*szBadDate[i])
	{
		// get test date numeric value
		if (AsciiToDate(&uDate, szBadDate[i]) == TRUE)
			printf("FAIL %s converted okay.\n", szBadDate[i]);
		else
			printf("PASS %s did not convert.\n", szBadDate[i]);

		i++;
	}

}
#endif
