//
//   $Id: DLMKTREE.C 1.9 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   DLMKTREE.C  created 04/30/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Make a full directory tree.

	$DOCHISTORY
	06/21/2000 td  Bumped to use "Smart" functions

********************************************************************/
#define __FILENUM__	513

#include <ctype.h>
#include <dir.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlpath.h"

/*
	Make a complete directory tree.  The passed path can be
	in any form.

	Returns TRUE if worked, returns FALSE if failed.
*/
bool MakeDirTree(char szPath[])
{
	char * pSlash;
	char	oldChar;
	char	szOldPath[LFN_MAX_PATH];
	int	orgDrive;
	bool	worked;

	// path must be abs path
	ASSERT(szPath);

	// we'll be parsing the path
	pSlash = szPath;

	// determine drive letter
	if(szPath[1] == ':')
	{
		orgDrive = toupper(*szPath) - 'A';
		pSlash += 2;
	}
	else
	{
		dlGetCurrentDrive( &orgDrive );
	}

	// not a valid drive, give up!
	if(!IsValidDrive(orgDrive))
		return FALSE;

	// make directories until we are done
	do
	{
		// get next directory
		pSlash = strchr(pSlash+1, '\\');
		if(!pSlash)
			pSlash = szPath + strlen(szPath);

		// used to determine end of string
		oldChar = *pSlash;

		// make the directory by null terminating the substring
		// then putting it back to normal
		*pSlash = 0;
      SmartMakeDirectory( szPath );
		*pSlash = oldChar;

	} while(oldChar != 0);

	// get the current directory of that drive
	GetCurrentDir(szOldPath, orgDrive);

	// make sure we can set the directory
	worked = SmartChangeDirectory( szPath );

	// restore original drive
	SetCurrentDir(szOldPath);

	return worked;
}

#if UNIT_TEST

#include <stdio.h>
#include <stdlib.h>

void InternalError(int fileNum, int lineNum)
{
	printf("ASSERT %u.%u\n",fileNum,lineNum);
	exit(99);
}

int main(void)
{
	ASSERT(!MakeDirTree("E:\\TEMP"));
	ASSERT(MakeDirTree("C:\\TEMP\\JUNK\\NEW"));
	ASSERT(MakeDirTree("TEMP\\JUNK\\NEW"));
	ASSERT(MakeDirTree(".\\TEMP\\JUNK\\NEW"));

	printf("passed");
	return 0;
}

#endif
