//
//   $Id: DLPATH.C 1.11 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DLPATH.C  created 03/07/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Contains the routines needed to get a generic path.

	bool ExpandPath(char *pFullPath, char *pPartialPath);

	Todo:
	Make this internationally aware.

	Revision History
	03/16/99	jmb	Added ExandPath

**********************************************************/
#define __FILENUM__	507

// includes
#include <dir.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dlpath.h"
#include "dlcntry.h"

/*
	Converts a partial path into drive:\path\name(s).

	Examples (assuming current DIR is C:\TEMP):
	.						C:\TEMP\*.*
	C:						C:\TEMP\*.*
	C:*.C					C:\TEMP\*.C
	*.*  					C:\TEMP\*.*
	D:\TEMP				D:\TEMP\*.*
	\DEV\ROMDOS\*.c	C:\DEV\ROMDOS\*.C

	returns TRUE if worked, FALSE if failed (bad drive).
*/
bool ExpandPath(char *pFullPath, char *pPartialPath)
{
	int flags;
	int iTemp;
	char szOrgCwd[MAXPATH];		// original Current Working Dir for drive
	char szDrive[MAXDRIVE];		// eg "C:"
	char szDir[MAXDIR];			// eg "\TEMP"
	char szName[MAXFILE];		// eg "FILENAME"
	char szExt[MAXEXT];			// eg "EXT"

	// validate parameters
	ASSERT(pFullPath);
	ASSERT(pPartialPath);

	// split up the path into managable parts
	flags = IntlFileNameSplit(pPartialPath, szDrive, szDir, szName, szExt);

	// if name is similar to "CON:", remove the colon
	if(LastChar(szName) == ':')
		*(szName+strlen(szName)-1) = 0;

	// determine drive
	if(flags & DRIVE)
		IntlStrUpr(szDrive);
	else
	{
		strcpy(szDrive, "A:");
		dlGetCurrentDrive( &iTemp );
		*szDrive = iTemp + 'A';
	}

	// get current working DIR for the drive
	// returns FALSE if invalid drive
	if(!GetCurrentDir(szOrgCwd, *szDrive - 'A'))
		return FALSE;

	if (flags & DIRECTORY)
	{
		// it must be a subdir under something...
		if (*szDir != '\\')
		{
			if (LastChar(szOrgCwd) != '\\')
				strcat(szOrgCwd, "\\");
			strcat(szOrgCwd, szDir);
			strcpy(szDir, szOrgCwd+2);
		}
	}
	else
	{
		// no path specified...
		if (LastChar(szOrgCwd) != '\\')
			strcat(szOrgCwd, "\\");
		strcpy(szDir, szOrgCwd+2);
	}

	if (!(flags & FILENAME))
		strcpy(szName, "*");

	if (!(flags & EXTENSION))
		strcpy(szExt, ".*");

	strcpy(pFullPath, szDrive);
	strcat(pFullPath, szDir);
	if(LastChar(szDir) != '\\')
		strcat(pFullPath, "\\");
	strcat(pFullPath, szName);
	strcat(pFullPath, szExt);

	return TRUE;
}

#if UNIT_TEST	// {

char szVeryLongName[] =
	"1234567890123456789012345678901234567890123456789012345678901234567890"
	"1234567890123456789012345678901234567890123456789012345678901234567890"
	"1234567890123456789012345678901234567890123456789012345678901234567890"
	"1234567890123456789012345678901234567890";	// 250 character name

/*
	Test helping function
*/
void TestPathConvert(char szPath[], char szOrgPath[])
{
	// Test PartialToFullPath().  Assumes cwd is C:\TEMP
	GetCurrentDir(szOrgPath, MAXPATH);
	mkdir("C:\\TEMP");
	SetCurrentDir("C:\\TEMP");

	// *.*  					C:\TEMP\*.*
	if(!PartialToFullPath(szPath, "*.*"))
		ABORT();
	if(strcmp(szPath, "C:\\TEMP\\*.*") != 0)
		ABORT();

	// .  					C:\TEMP\*.*
	if(!PartialToFullPath(szPath, "."))
		ABORT();
	if(strcmp(szPath, "C:\\TEMP\\*.*") != 0)
		ABORT();

	// \TEMP  					C:\TEMP\*.*
	if(!PartialToFullPath(szPath, "\\TEMP"))
		ABORT();
	if(strcmp(szPath, "C:\\TEMP\\*.*") != 0)
		ABORT();

	// C:						C:\TEMP\*.*
	if(!PartialToFullPath(szPath, "C:"))
		ABORT();
	if(strcmp(szPath, "C:\\TEMP\\*.*") != 0)
		ABORT();

	// *.C					C:\TEMP\*.C
	if(!PartialToFullPath(szPath, "*.C"))
		ABORT();
	if(strcmp(szPath, "C:\\TEMP\\*.C") != 0)
		ABORT();

	// NOTHERE\JUNK	should fail
	if(PartialToFullPath(szPath, "NOTHERE\\JUNK"))
		ABORT();

	// A:\				A:\*.*
	if(!PartialToFullPath(szPath, "A:\\"))
		ABORT();
	if(strcmp(szPath, "A:\\*.*") != 0)
		ABORT();

	// \dev\romdos\*.c	C:\DEV\ROMDOS\*.c
	// path will always be uppercased, name will not
	if(!PartialToFullPath(szPath, "\\dev\\romdos\\*.c"))
		ABORT();
	if(strcmp(szPath, "C:\\DEV\\ROMDOS\\*.c") != 0)
		ABORT();

	// T:\NOTHERE\JUNK	should fail
	if(PartialToFullPath(szPath, "T:\\NOTHERE\\JUNK"))
		ABORT();

	// T:\APPS	should work
	if(!PartialToFullPath(szPath, "T:\\APPS"))
		ABORT();
	if(strcmp(szPath, "T:\\APPS\\*.*") != 0)
		ABORT();

	// done, change back to this directory
	SetCurrentDir(szOrgPath);
}

void TestFileNameSplit(void)
{
	int flags;
	char szDrive[MAXDRIVE];		// eg "C:"
	char szDir[MAXDIR];			// eg "\TEMP"
	char szName[MAXFILE];		// eg "FILENAME"
	char szExt[MAXEXT];			// eg "EXT"

	// test ".", should be dir only
	flags = IntlFileNameSplit(".", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == DIRECTORY);
	SHIPASSERT(strcmp(szDir, ".") == 0);
	SHIPASSERT(*szDrive == 0 && *szName == 0 && *szExt == 0);

	// test "..", should be dir only
	flags = IntlFileNameSplit("..", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == DIRECTORY);
	SHIPASSERT(strcmp(szDir, "..") == 0);
	SHIPASSERT(*szDrive == 0 && *szName == 0 && *szExt == 0);

	// test "\\", should be dir only
	flags = IntlFileNameSplit("\\", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == DIRECTORY);
	SHIPASSERT(strcmp(szDir, "\\") == 0);
	SHIPASSERT(*szDrive == 0 && *szName == 0 && *szExt == 0);

	// test "x", should be name only
	flags = IntlFileNameSplit("x", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == FILENAME);
	SHIPASSERT(strcmp(szName, "x") == 0);
	SHIPASSERT(*szDrive == 0 && *szDir == 0 && *szExt == 0);

	// test "x.", should be name/ext
	flags = IntlFileNameSplit("x.", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == (FILENAME | EXTENSION));
	SHIPASSERT(strcmp(szName, "x") == 0);
	SHIPASSERT(strcmp(szExt, ".") == 0);
	SHIPASSERT(*szDrive == 0 && *szDir == 0);

	// test ".x", should be name/ext
	flags = IntlFileNameSplit(".x", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == EXTENSION);
	SHIPASSERT(strcmp(szExt, ".x") == 0);
	SHIPASSERT(*szDrive == 0 && *szDir == 0 && *szName == 0);

	// test drive only
	flags = IntlFileNameSplit("C:", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == DRIVE);
	SHIPASSERT(strcmp(szDrive, "C:") == 0);
	SHIPASSERT(*szDir == 0 && *szName == 0 && *szExt == 0);

	// test path only
	flags = IntlFileNameSplit("\\TEMP\\", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == DIRECTORY);
	SHIPASSERT(strcmp(szDir, "\\TEMP\\") == 0);
	SHIPASSERT(*szDrive == 0 && *szName == 0 && *szExt == 0);

	// test name only
	flags = IntlFileNameSplit("FILENAMEX", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == FILENAME);
	SHIPASSERT(strcmp(szName, "FILENAME") == 0);
	SHIPASSERT(*szDrive == 0 && *szDir == 0 && *szExt == 0);

	// test ext only
	flags = IntlFileNameSplit(".EXT", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == EXTENSION);
	SHIPASSERT(strcmp(szExt, ".EXT") == 0);
	SHIPASSERT(*szDrive == 0 && *szDir == 0 && *szName == 0);

	// test full path
	flags = IntlFileNameSplit("c:x.y\\junk\\x.*", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == (WILDCARDS | EXTENSION | FILENAME | DIRECTORY | DRIVE));
	SHIPASSERT(strcmp(szDrive, "c:") == 0);
	SHIPASSERT(strcmp(szDir, "x.y\\junk\\") == 0);
	SHIPASSERT(strcmp(szName, "x") == 0);
	SHIPASSERT(strcmp(szExt, ".*") == 0);

	// test "x.*", file, ext
	flags = IntlFileNameSplit("x.*", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == (WILDCARDS | EXTENSION | FILENAME));
	SHIPASSERT(strcmp(szName, "x") == 0);
	SHIPASSERT(strcmp(szExt, ".*") == 0);
	SHIPASSERT(*szDrive == 0 && *szDir == 0);

	// test "..\\.*"
	flags = IntlFileNameSplit("..\\.*", szDrive, szDir, szName, szExt);
	SHIPASSERT(flags == (WILDCARDS | DIRECTORY | EXTENSION));
	SHIPASSERT(strcmp(szDir, "..\\") == 0);
	SHIPASSERT(strcmp(szExt, ".*") == 0);
	SHIPASSERT(*szDrive == 0 && *szName == 0);
}

/*
	Unit test assert.
*/
#pragma argsused
void InternalError(int fileNum, int lineNum)
{
	printf("BACPATH Unit Test Error, line %u\n",lineNum);
	exit(9);
}

int main(void)
{
	char szPath[MAXPATH];
	char szOrgPath[MAXPATH];

	// test file name split
	TestFileNameSplit();

	// test converting path
	TestPathConvert(szPath, szOrgPath);

	return 0;
}

#endif	// } UNIT_TEST
