//
//   $Id: DLPATH9.C 1.3 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DLPATH9.C  created 03/07/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Contains the routines needed to get a generic path.

	Datalight ROM-DOS Tools Library:
	int PASCAL IntlFileNameSplit(char *pStr, char *pDrive, char *pDir,
													char *pName, char *pExt);
	bool ExpandPath(char *pFullPath, char *pPartialPath);

	Todo:
	Make this internationally aware.

	Revision History
	03/16/99	jmb	Added ExandPath
	03-06-01  bt  Broke up dlpath.c for library granularity
	03-06-01  bt  Changed calls to borland libraries to DL libs

**********************************************************/
#define __FILENUM__	507

// includes
#include <dir.h>
//#include <dos.h>
//#include <stdio.h>
//#include <stdlib.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dlpath.h"
#include "dlcntry.h"

/*
	FileName Split, with International (2-byte char) support.
	Splits the name into 4 parts: drive, directory, name and
	extension.

	Examples:

	pStr               pDrive  pDir    pName   pExt
	C:\TEMP\*.C  --->  C:      \TEMP\  *       .C
	.                          .
	x.                                 x       .
	.x                                         .x

	Returns same flags as "fnsplit" from dir.h.
	(WILDCARDS, EXTENSION, FILENAME, DIRECTORY, DRIVE)
*/
int PASCAL IntlFileNameSplit(char *pStr, char *pDrive, char *pDir,
													char *pName, char *pExt)
{
	int flags = 0;		// nothing found yet
	int len;
	char * pEndStr;
	char * pLastSlash = 0;
	char * pLastDot = 0;

	// validate parameters
	ASSERT(pStr);
	ASSERT(pDrive);
	ASSERT(pDir);
	ASSERT(pName);
	ASSERT(pExt);

	// initialize each string, nothing in them
	*pDrive = *pDir = *pName = *pExt = 0;

	// do we have a drive part?
	if(*pStr && !IsDoubleByteChar(*pStr) && pStr[1]==':')
	{
		// copy both characters of the "A:"
		*(unsigned *)pDrive = *(unsigned *)pStr;
		pDrive[2] = 0;

		// skip drive letter
		pStr += 2;

		// mark as having a drive part
		flags |= DRIVE;
	}

	// stop on end of string or extension
	pEndStr = pStr;
	while(*pEndStr)
	{
		// found more to the path? use it!
		if(*pEndStr == '\\')
			pLastSlash = pEndStr;

		// found a possible file extension?
		if(*pEndStr == '.')
			pLastDot = pEndStr;

		// on to next character (handle double-byte chars too!)
		if(IsDoubleByteChar(*pEndStr))
			++pEndStr;
		++pEndStr;
	}

	// the string MAY end in "." or "..", if so, they are part of
	// the path, NOT the extension
	if(pLastDot && *(pLastDot+1) == 0)
	{
		// do we have a "." or ".." at beginning of string?
		if(!pLastSlash)
		{
			if( pLastDot == pStr ||
				(pLastDot == pStr+1 && *(pLastDot-1) == '.'))
			pLastSlash = pLastDot;
		}

		// do we have a "." or ".." after last slash?
		else
		{
			if( pLastDot == pLastSlash+1 ||
				(pLastDot == pLastSlash+2 && *(pLastDot - 1) == '.'))
			pLastSlash = pLastDot;
		}
	} // done fiddling with "." and ".." entries

	// is there a "pDir" part?
	if(pLastSlash && pLastSlash >= pStr)
	{
		// there is a directory part, copy it in
		len = min(MAXDIR - 1, (int)(pLastSlash - pStr) + 1);
		memmove(pDir, pStr, len);
		*(pDir+len) = 0;

		// mark as having a directory part
		flags |= DIRECTORY;

		// set pointer to "pName" part
		pStr = pLastSlash + 1;
	}

	// no extension? then just use end of string
	// last dot may have been inside the path string "..\whatever"
	if(!pLastDot || (pLastSlash && pLastDot <= pLastSlash))
		pLastDot = pEndStr;

	// is there a "pName" part?
	if(pStr < pLastDot)
	{
		// copy in name part
		len = min(MAXFILE-1, (int)(pLastDot-pStr));
		memmove(pName, pStr, len);
		*(pName+len) = 0;

		// mark as having a name
		flags |= FILENAME;
	}

	// is there an extension?
	if(pLastDot < pEndStr)
	{
		// copy in the file extension
		len = min(MAXEXT-1, (int)(pEndStr - pLastDot));
		memmove(pExt, pLastDot, len);
		*(pExt+len) = 0;

		// mark as having an extension
		flags |= EXTENSION;
	}

	// wildcards in the name?
	if(IsWild(pName) || IsWild(pExt))
		flags |= WILDCARDS;

	// return all the parts that were found
	return flags;
}
