//
//   $Id: DLXMS.C 1.5 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	DLXMS.C  created 12/06/96 by Drew Gislason

	Copyright (c) 1996-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Datalight ROM-DOS Tools Library

	Interface:

	This module does the XMS interaction for the Datalight
	Backup/Restore.

**********************************************************/
#define __FILENUM__ 505

// includes
#if DEBUG
#include <stdio.h>
#endif
#include <dos.h>
#include "datalght.h"
#include "dlmem.h"

//
// New types
//

typedef void (far * farfuncptr)(void);
farfuncptr hmaHandler;

struct XMSCopyInfo
{
	long     numBytes;
	unsigned srcHandle;
	long     srcOffset;
	unsigned dstHandle;
	long     dstOffset;
};

/*
	Is Himem (the XMS driver) installed?
	This routine MUST be called before any other XMS routines
	are used.
	Returns TRUE if installed, FALSE otherwise.
*/
int IsHimemInstalled(void)
{
	#define XMS_FOUND_ID 0x80
	byte rAL;
	unsigned rSeg, rOff;

	// already installed
	if(hmaHandler)
		return TRUE;

	// is an XMS driver installed?
	ASM mov AX, 4300H
	ASM int 2FH
	ASM mov rAL, AL

	// get address of HMA driver
	if(rAL == XMS_FOUND_ID)
	{
		ASM mov 	AX, 4310H
		ASM int 	2FH
		ASM mov	rOff, BX
		ASM mov	rSeg, ES
		if(!rSeg || !rOff)
			rAL = 0;
		else
			hmaHandler = (farfuncptr)MK_FP(rSeg, rOff);
	}

	// return whether we found the XMS driver or not
	return (rAL == XMS_FOUND_ID);
}

/*
	Return size of largest XMS block in K bytes.
*/
unsigned TotalXMS(void)
{
	unsigned kBytes;

	ASSERTLP(hmaHandler);

	// ask XMS how much memory is available
	// returned in ax.
	ASM mov AX, 800H
	(*hmaHandler)();
	ASM mov kBytes, AX

	// return number of K bytes available
	return kBytes;
}

/*
	Allocate some K bytes from XMS.
	Returns the handle if there was enough available memory.
	Use TotalXMS() to determine how much memory is available.
	Returns a handle of 0 if failed.
*/
unsigned AllocateXMS(unsigned sizeK)
{
	unsigned handle;
	unsigned success;

	ASSERT(sizeK);
	ASSERTLP(hmaHandler);

	/* allocate the XMS memory */
	ASM mov DX, sizeK
	ASM mov AX, 900H
	(*hmaHandler)();
	ASM mov success, AX
	ASM mov handle, DX

	// any error means failed
	if(!success)
		handle = 0;

	// got the memory, return a handle to it
	return handle;
}

/*
	Free XMS memory previously allocated.
*/
void FreeXMS(unsigned handle)
{
	word success;

	// make sure handle is valid
	ASSERT(handle);
	ASSERTLP(hmaHandler);

	// allocate the XMS memory
	ASM mov DX, handle
	ASM mov AX, 0A00H
	(*hmaHandler)();
	ASM mov success, AX

	// don't allow invalid handles!
	ASSERT(success);
}

/*
	Copy some memory from XMS to convention or vice versa.
	Cannot fail.  If any parameters to the CopyInfo are wrong,
	the function will ASSERT.
*/
static void CopyXMS(struct XMSCopyInfo * pCopyInfo)
{
	auto int worked;
#if DEBUG
	auto byte errorCode;
#endif
#if (__COMPACT__ || __LARGE__)
	auto	farSegment;
#endif
	auto	off;

	// make sure it at least points to something
	ASSERT(pCopyInfo);
	ASSERTLP(hmaHandler);

	// copy a block
	ASM	push	DS
#if (__COMPACT__ || __LARGE__)
	farSegment = FP_SEG(pCopyInfo);
	ASM	mov	DS, farSegment
#endif
	off = (unsigned)pCopyInfo;
	ASM	mov	SI, off				// DS:SI must point to block
	ASM	mov	AX, 0B00H
	(*hmaHandler)();
	ASM mov worked, AX
#if DEBUG
	ASM mov errorCode, BL
#endif
	ASM	pop	DS

	// something seriously went wrong if it failed to move
#if DEBUG
	if(!worked)
		printf("XMS Failed to Copy! (%x)\n",errorCode);
#endif

	// make sure it worked
	ASSERT(worked);
}

/*
	Copy from conventional memory to XMS memory.  Think of the
	parameters like strncpy, where the destination is first,
	followed by the source.

	The routine assumes a valid XMS handle returned by
	"AllocateXMS", and that all offsets are valid.  Will ASSERT
	if there are any failures.
*/
void CopyToXMS(unsigned hXMS, long offset, void far *lpMem, long numBytes)
{
	struct XMSCopyInfo xmsCopyInfo;

	// validate parameters
	ASSERT(hXMS);
	ASSERTLP(lpMem);
	ASSERT(numBytes);
	ASSERTLP(hmaHandler);

	// set up the copy structure
	xmsCopyInfo.numBytes  = numBytes;
	xmsCopyInfo.srcHandle = 0;
	xmsCopyInfo.srcOffset = (long)lpMem;
	xmsCopyInfo.dstHandle = hXMS;
	xmsCopyInfo.dstOffset = offset;

	// actually do the copy
	CopyXMS(&xmsCopyInfo);
}

/*
	Copy from XMS memory to conventional memory.  Think of the
	parameters like strncpy, where the destination is first,
	followed by the source.

	The routine assumes a valid XMS handle returned by
	"AllocateXMS", and that all offsets are valid.  Will ASSERT
	if there are any failures.
*/
void CopyFromXMS(void far *lpMem, unsigned hXMS, long offset, long numBytes)
{
	struct XMSCopyInfo xmsCopyInfo;

	// validate parameters
	ASSERT(hXMS);
	ASSERTLP(lpMem);
	ASSERT(numBytes);
	ASSERTLP(hmaHandler);

	// set up the copy structure
	xmsCopyInfo.numBytes  = numBytes;
	xmsCopyInfo.srcHandle = hXMS;
	xmsCopyInfo.srcOffset = offset;
	xmsCopyInfo.dstHandle = 0;
	xmsCopyInfo.dstOffset = (long)lpMem;

	// actually do the copy
	CopyXMS(&xmsCopyInfo);
}

#if UNIT_TEST

#include <stdio.h>
#include <stdlib.h>
#include <mem.h>

void InternalError(int fileNum, int lineNum)
{
	printf("ASSERT: File %u, line %u\n",fileNum, lineNum);
	exit(10);
}

// two buffers for testing xms copy
char buffer[1024];
char buffer2[1024];

int main(void)
{
	unsigned kBytes, k;
	unsigned handle;

	// is there XMS available?
	if(!IsHimemInstalled())
	{
		printf("No XMS driver!\n");
		return 1;
	}

	// determine total XMS memory
	kBytes = TotalXMS();
	if(!kBytes)
	{
		printf("No XMS memory!\n");
		return 2;
	}

	// allocate the memory
	handle = AllocateXMS(kBytes);
	if(!handle)
	{
		printf("Failed to allocate XMS memory!\n");
		return 3;
	}

	// now fill each K of XMS with a pattern
	for(k=0; k<kBytes; ++k)
	{
		// set up a unique pattern
		memset(buffer, k, sizeof(buffer));

		// fill XMS memory
		CopyToXMS(handle, 1024L*k, (void far *)buffer, 1024L);
	}

	// now, verify the memory looks OK
	for(k=0; k<kBytes; ++k)
	{
		// set up a unique pattern
		memset(buffer, k, sizeof(buffer));

		// get the memory from XMS
		CopyFromXMS((void far *)buffer2, handle, 1024L*k, 1024L);

		// make sure they compare OK
		if(memcmp(buffer, buffer2, 1024) != 0)
		{
			printf("Failed to compare on %uK boundary\n",k);
			return 4;
		}
	}

	// make sure to free the memory
	FreeXMS(handle);

	// everything passed
	printf("Passed\n");
	return 0;
}

#endif // UNIT_TEST

