//
//   $Id: DLCMDLN.C 1.7 2002/10/15 22:45:24Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   DLCMDLN.C  created 04/24/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Datalight ROM-DOS Tools Library

	Interface:
	int ParseCmdline(int argCount, char *argStrings[], ARGDEF argDefs[]);
	void SetParseFlags(int iFlags);
	void Usage(void);				// optionally supplied by user
	void OptError(char *pMsg);	// optionally supplied by user
	void GetLFNCommandLine(char szLFNCmdLine[], unsigned uMaxLen);

03/16/99	jmb	Added GetLFNCommandLine
08-12-97 DE fixed this so it would handle /p:n, as it should have
	    fixed bug in compacting arguments where argv[i] <- argv[i+1]
	    was a copy (vs a move, as it should have been)

********************************************************************/
#define __FILENUM__	500

// includes
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dlmath.h"
#include "dlapp.h"
#include "dldate.h"


#include "dlcmdln.h"
#include "dlcmdln.def"


char * SkipSwitch( char * pStartStr );

// in-line routines
#define IsSwitchType(type) ((type) >= BOOL_SWITCH && (type) <= STRING_SWITCH)
#define IsArgType(type) ((type) >= INT_ARG && (type) <= STRING_ARG)

static int iParseFlags;


#if CMDLN_MULTIPASS_PARSE
static bool fSkipSwitch;
#endif


/*
	We've found a switch.  See if it is in our list!
	By default, /? is ALWAYS in the list, which calls Usage();

	Displays error message and exits if switch is not in "argDefs"
	list.
*/
static void ProcessSwitch(char *pSwitch, ARGDEF argDefs[])
{
	ARGDEF *	pArgDef;
	bool		found = FALSE;
	int		switchLen = 2;
	char		switchChar;
	char   *	pStartStr;
	char   *	pEndStr;
	char   * pSrcPos;
	bool		inQuote;
	int		len;
	bool		trueVar = TRUE;
	long	   value;


#if CMDLN_MULTIPASS_PARSE
	//  Initialize the flag that indicates to the calling routine
	//  to skip this argument
	//
	fSkipSwitch = FALSE;
#endif


	// validate parameters
	ASSERT(pSwitch && *pSwitch);
	ASSERT(argDefs);

	// is this /? for help?
	switchChar = toupper(*(pSwitch+1));
	if(switchChar == '?')
	{
		Usage();
		goto RemoveSwitch;
	}

	// allow negative to be BEFORE switch char, example: /-Y
	if(switchChar == '-')
	{
		trueVar = FALSE;
		switchChar = toupper(*(pSwitch+2));
		++switchLen;
	}

	// does this match one of the switches?
	for(pArgDef = argDefs; pArgDef->type; ++pArgDef)
	{
		if((IsSwitchType(pArgDef->type)) &&
			(toupper(pArgDef->switchChar) == switchChar))
		{
			found = TRUE;
			break;
		}
	}

	// didn't find switch character? let user know it's bogus!
	if(!found)
	{
#if CMDLN_MULTIPASS_PARSE
		if( iParseFlags & PARSEF_MULTIPASS )
		{
			fSkipSwitch = TRUE;
			return;
		}
		else
		{
			OptError(pSwitch);
			goto RemoveSwitch;
		}
#else
		OptError(pSwitch);
		goto RemoveSwitch;
#endif
	}

	// make sure argdef is valid
	ASSERT(pArgDef->pAddress);

	switch(pArgDef->type)
	{

#if CMDLN_BOOL_SWITCH
		// Examples: /v, /a-, /-Y
		case BOOL_SWITCH:
			{
				if( iParseFlags & PARSEF_NOSAMEARGS )
				{
					if( pArgDef->fChanged )
					{
						OptError(pSwitch);
					}
					pArgDef->fChanged = TRUE;
				}
				*(bool *)(pArgDef->pAddress) = trueVar;
				if(*(pSwitch+2) == '-')
				{
					++switchLen;
					*(bool *)pArgDef->pAddress = FALSE;
				}
			}
			break;
#endif // #if CMDLN_BOOL_SWITCH


#if CMDLN_HEXINT_SWITCH
		case HEXINT_SWITCH:
			{
				// determine start of argument string
				pStartStr = SkipSwitch(pSwitch);
				if(!isxdigit(*pStartStr))
					OptError(pSwitch);

				// calculate value
				value = atoh(pStartStr);
				*(int *)(pArgDef->pAddress) = (int)value;

				// set switch length
				pEndStr = pStartStr;
				while(isxdigit(*pEndStr))
					++pEndStr;
				switchLen = (int)(pEndStr - pSwitch);
			}
			break;
#endif // #if CMDLN_HEXINT_SWITCH


#if CMDLN_INT_SWITCH
		case INT_SWITCH:
			//
			// Fall through
			//
#endif // #if CMDLN_INT_SWITCH


#if CMDLN_LONG_SWITCH | CMDLN_INT_SWITCH
		case LONG_SWITCH:
			{
				// determine start of argument string
				pStartStr = SkipSwitch(pSwitch);
				if(!isdigit(*pStartStr))
					OptError(pSwitch);

				// convert the string argument to a numeric value
				value = atolong(pStartStr);

				// set the value
				if(pArgDef->type == INT_SWITCH)
					*(int *)(pArgDef->pAddress) = (int)value;
				else
					*(long *)(pArgDef->pAddress) = value;

				pEndStr = pStartStr;
				while(isdigit(*pEndStr))
					++pEndStr;
				switchLen = (int)(pEndStr - pSwitch);
			}
			break;
#endif // #if CMDLN_LONG_SWITCH | CMDLN_INT_SWITCH


#if CMDLN_DATE_SWITCH
		// Examples: /D:12/12/99 where the / can be part of the date
		case DATE_SWITCH:
			{
				// determine start of argument string
				pStartStr = pSwitch+2;
				if(*pStartStr == ':')
					++pStartStr;

				// translate from date string to
				if (AsciiToDate((unsigned *)pArgDef->pAddress, pStartStr) == FALSE)
					OptError(pSwitch);

				// determine length of date string...
				pEndStr = strchr(pStartStr, '-');
				if (pEndStr)
				{
					pEndStr++;
					pEndStr = strchr(pEndStr,'-');

				}
				else
				{
					pEndStr = strchr(pStartStr, '/');
					if (pEndStr)
					{
						pEndStr++;
						pEndStr = strchr(pEndStr, '/');
					}
					else
					{
						pEndStr = strchr(pStartStr,'.');
						pEndStr++;
						pEndStr = strchr(pEndStr,'.');
					}
				}
				pEndStr++;
				while (isdigit(*pEndStr))
					++pEndStr;
				switchLen = (int)(pEndStr - pSwitch);
			}
			break;
#endif // #if CMDLN_DATE_SWITCH


#if CMDLN_STRING_SWITCH
		// Examples: /V:"Volume Name"  /vHello  /v:abc
		case STRING_SWITCH:
			{
				// determine start of argument string
				pStartStr = pSwitch+2;
				if(*pStartStr == ':')
					++pStartStr;


				// handle quotes around the string
				inQuote = FALSE;
				if(*pStartStr == '"')
				{
					inQuote = TRUE;
					++pStartStr;
				}

				// determine length of string
				pEndStr = pStartStr;
				while(*pEndStr)
				{
					if(inQuote && *pEndStr == '"')	// break on unquote
						break;

					if(!inQuote && *pEndStr == '/')		// break on next switch if
						break;									// not in quoted string

					++pEndStr;
				}

				// make sure argdef is valid
				ASSERT(pArgDef->maxLen > 0);

				// only copy in as much as user said to copy
				len = (int)(pEndStr - pStartStr);
				if(len > pArgDef->maxLen-1)
					len = pArgDef->maxLen-1;
				((char *)(pArgDef->pAddress))[0] = 0;  // null terminate the string
				memmove(pArgDef->pAddress, pStartStr, len);

				// determine full length of switch
				if(*pEndStr == '"')
					++pEndStr;
				switchLen = (int)(pEndStr - pSwitch);
			}
			break;
#endif // #if CMDLN_STRING_SWITCH


#if CMDLN_LFN_STRING_SWITCH
		// Examples: /i:THIS_IS_A_REALLY_LONG_FILE.TXT
		//           /i"This is a really long file.txt"
		case LFN_STRING_SWITCH:
			{
				// make sure argdef is valid
				ASSERT(pArgDef->maxLen > 0);

				// determine start of argument string
				pStartStr = pSwitch+2;
				if(*pStartStr == ':')
					++pStartStr;


				// handle quotes around the string
				inQuote = FALSE;
				if(*pStartStr == '"')
				{
					inQuote = TRUE;
					++pStartStr;
				}

				// determine length of string
				pEndStr = pStartStr;
				pSrcPos = pArgDef->pAddress;
				len = 0;

				while(*pEndStr)
				{

					if (*pEndStr == '"')
					{
						if (!inQuote)				// we just found the opening quote
							inQuote = TRUE;

						else
							inQuote = FALSE;
					}

					if (!inQuote && *pEndStr == ' ')		// break on next space if
						break;									// not in quoted string

					if(!inQuote && *pEndStr == '/')		// break on next switch if
						break;									// not in quoted string

					if (*pEndStr != '"')
					{
						if (len == pArgDef->maxLen-1)		// store no more characters
							break;								// than the target can hold
						else
						{
							*pSrcPos = *pEndStr;				// store the character
							pSrcPos++;							// next position, please
							len++;								// count it
						}
					}

					++pEndStr;
				}

				// null terminate
				*pSrcPos = '\0';

				// determine full length of switch
				if(*pEndStr == '"')
					++pEndStr;
				switchLen = (int)(pEndStr - pSwitch);
			}
			break;
#endif // #if CMDLN_LFN_STRING_SWITCH


		default:
			{
				ABORT();
			}
	}

RemoveSwitch:

	// remove switch from string
	strcpy(pSwitch, pSwitch+switchLen);
}


/*
	We have a single cmdline argument.  Process it.
*/
#pragma argsused
static void ProcessArg(char *pArg, ARGDEF * pArgDef)
{
	long	value;
	int	len;
	char *pStart,*pEnd, *pSrcPos;
#if CMDLN_LFN_STRING_ARG
	bool inQuote=FALSE;
#endif // #if CMDLN_LFN_STRING_ARG

	// validate parameters
	ASSERT(pArg);
	ASSERT(pArgDef && IsArgType(pArgDef->type));


	switch(pArgDef->type)
	{

#if CMDLN_INT_ARG
		case INT_ARG:
			//
			//  Fall through
			//
#endif // #if CMDLN_INT_ARG


#if CMDLN_LONG_ARG | CMDLN_INT_ARG
		case LONG_ARG:
			{
				// make sure this is an integer
				if(!isdigit(*pArg))
					OptError(pArg);

				// convert the string argument to a numeric value
				value = atolong(pArg);

				// set the value
				if(pArgDef->type == INT_ARG)
					*(int *)(pArgDef->pAddress) = (int)value;
				else
					*(long *)(pArgDef->pAddress) = value;
			}
			break;
#endif // #if CMDLN_LONG_ARG | CMDLN_INT_ARG


#if CMDLN_LFN_STRING_ARG
		// Examples: THIS_IS_A_REALLY_LONG_FILE.TXT
		//           "This is a really long file.txt"
		//           "This is a "really" long file".txt
		case LFN_STRING_ARG:
			{
				// make sure argdef is valid
				ASSERT(pArgDef->maxLen > 0);
				ASSERT(pArgDef->pAddress != NULL);

				// determine start of argument string
				pStart = pArg;

				// kill leading white space
				while (DL_IS_SPACE(*pStart))
				    pStart++;

				// handle quotes around the string
				inQuote = FALSE;
				if(*pStart == '"')
				{
					inQuote = TRUE;
					++pStart;
				}

				// determine length of string
				pEnd = pStart;
				pSrcPos = pArgDef->pAddress;
				len = 0;

				while(*pEnd)
				{
					if (*pEnd == '"')
					{
						if (inQuote == FALSE)				// we just found the opening quote
							inQuote = TRUE;
						else
							inQuote = FALSE;					// we just found the close quote
					}

					if (inQuote==FALSE && *pEnd == ' ')	// break on next space if
						break;									// not in quoted string

					if (*pEnd != '"')
					{
						if (len == pArgDef->maxLen-1)		// store no more characters
							break;								// than the target can hold
						else
						{
							*pSrcPos = *pEnd;					// store the character
							pSrcPos++;							// next position, please
							len++;								// count it
						}
					}
					++pEnd;
				}

				// null terminate
				((char *)pArgDef->pAddress)[len] = 0;

				// remove argument from string
				len = (unsigned)(pEnd - pArg);
				strcpy(pArg, pArg+len);
			}
			break;
#endif // #if CMDLN_LFN_STRING_ARG


#if CMDLN_STRING_ARG
		case STRING_ARG:
			{
				// validate argument definition
				ASSERT(pArgDef->maxLen > 0);

				// determine length to copy
				len = strlen(pArg);
				if(len > pArgDef->maxLen - 1)
					len = pArgDef->maxLen - 1;

				// copy the string, null terminate it
				memmove(pArgDef->pAddress, pArg, len);
				((char *)pArgDef->pAddress)[len] = 0;
			}
			break;
#endif // #if CMDLN_STRING_ARG
	}
}

void SetParseFlags(int iFlags)
{
	iParseFlags = iFlags;
}

/*
	Scan the command-line for strings and switches.  Fills in the
	variables defined by "argDefs".  Will call Usage() if no
	arguments.

	Note: this functions is destructive to "argStrings!"

	Parameters:
	argCount			# of cmdline string(s)
	argStrings		cmdline string(s)
	argDefs			definitions for expected arguments (NULL terminated)

	Example:
	if(ParseCmdline(argc-1, argv+1, moveArgs) != 2)
		Error("Invalid # of parameters!");

	Notes:
	- Does NOT handle double-byte international strings!
*/
int ParseCmdline(int argCount, char *argStrings[], ARGDEF argDefs[])
{
	int		i;
	char   *	pSwitch;
	ARGDEF *	pArgDef;

	// validate parameters
	ASSERT(argCount >= 0);
	ASSERT(argStrings);
	ASSERT(argDefs);


	// no arguments, just show help
	if(!(iParseFlags & PARSEF_NOARGSOK) && argCount==0)
		Usage();

	// loop through, looking for switches
	for(i=0; i < argCount; ++i)
	{
		// any switches in this cmdline argument?
		pSwitch = argStrings[i];
		while(pSwitch && *pSwitch)
		{
			// look for next switch
			pSwitch = strchr(pSwitch, '/');
			if(!pSwitch)
				break;

			// actually pull the switch out of the arg string
			ProcessSwitch(pSwitch, argDefs);

#if CMDLN_MULTIPASS_PARSE
			if( fSkipSwitch )
			{
				if( (*pSwitch != 0) && (*pSwitch != ' ') )
				{
					pSwitch++;
				}

				while((*pSwitch != 0) && (*pSwitch != ' ') && (*pSwitch != '/'))
				{
					pSwitch++;
				}
			}
#endif
		}
	}

	// loop through, looking for arguments
	i = 0;
	pArgDef = argDefs;
	while(i < argCount)
	{
		// empty parameter? move ptr down
		if(*argStrings[i] == 0)
		{
			// only move if argument ptrs to shuffle down
			if(i < argCount-1)
				memmove(argStrings+i, argStrings+i+1,
												sizeof(char *) * (argCount-i-1));

			// one less argument
			--argCount;
			argStrings[argCount]=0;
		}

		else
		{
			// find the next argument definition (if any)
			while(pArgDef->type)
			{
				if(IsArgType(pArgDef->type))
					break;
				++pArgDef;
			}

			// do we still have args left to process
			if(IsArgType(pArgDef->type))
			{
				ProcessArg(argStrings[i], pArgDef);
				++pArgDef;
			}

			// on to next argument string
			++i;
		}
	}	// end of while (i < argCount) loop

	// return # of arguments found
	return argCount;
}


#ifdef UNIT_TEST

void Usage(void)
{
	printf("Usage\n");
}

void InternalError(int fileNum, int lineNum)
{
	printf("ASSERT %u.%u\n",fileNum,lineNum);
	exit(99);
}

/*
	Any error in the cmdline will display here
*/
void OptError(char *pMsg)
{
	static char errMsgBadOpt[] = "Invalid command line option \"%s\"!\n";
	printf(errMsgBadOpt, pMsg);
}

// test variables (switches and arguments)
bool swBool1;
bool swBool2;
bool swBool3;
int  swInt1;
char swStr1[30];
char swStr2[5];
int  argInt1;
int  argInt2;
long argLong1;
long argLong2;
char argStr1[40];
char argStr2[4];

ARGDEF argTest1[] =
{
	{ BOOL_SWITCH, &swBool1, 'V' },
	{ 0 }
};
char *cmdLine1[] = { "/v" };

ARGDEF argTest2[] =
{
	{ BOOL_SWITCH, &swBool1, 'V' },
	{ INT_ARG, &argInt1 },
	{ 0 }
};
char *cmdLine2[] = { "/v-", "1234" };

ARGDEF argTest3[] =
{
	{ LONG_ARG, &argLong1 },
	{ STRING_ARG, &argStr2, 0, sizeof(argStr2) },
	{ BOOL_SWITCH, &swBool1, 'v' },
	{ BOOL_SWITCH, &swBool2, 'b' },
	{ BOOL_SWITCH, &swBool3, 'a' },
	{ STRING_SWITCH, &swStr1, 'm', sizeof(swStr1) },
	{ 0 }
};
char *cmdLine3[] = { "/v/b/a-129k", "Too Long String/z", "/m:Hello" };

ARGDEF argTest4[] =
{
	{ BOOL_SWITCH,    &swBool1, 'B' },
	{ HEXINT_SWITCH,  &swInt1,  'S' },
	{ 0 }
};

char *cmdLine4[] = { "81", "/s12", "/b" };

int main(void)
{
	// test a single switch
	if(ParseCmdline(1, cmdLine1, argTest1) != 0)
		ABORT();
	ASSERT(swBool1);

	// test both a switch and an arg, where the switch makes the 1st
	// arg empty, and switch is turned off
	if(ParseCmdline(2, cmdLine2, argTest2) != 1)
		ABORT();
	ASSERT(!swBool1);
	ASSERT(argInt1 == 1234);

	// test string args, and an invalid argument
	if(ParseCmdline(sizeof(cmdLine3) / sizeof(char *), cmdLine3, argTest3) != 2)
		ABORT();
	ASSERT(swBool1);
	ASSERT(swBool2);
	ASSERT(!swBool3);
	ASSERT(strcmp(swStr1, "Hello") == 0);
	ASSERT(strcmp(argStr2, "Too") == 0);
	ASSERT(argLong1 == 129L*1024);

	// bug when fixing FDISK: "FDISK 80 /S12 /B" wiped out the "80"
	if(ParseCmdline(sizeof(cmdLine4)/sizeof(char *), cmdLine4, argTest4) != 1)
		ABORT();
	ASSERT(swBool1);
	ASSERT(swInt1 == 0x12);
	ASSERT(strcmp(cmdLine4[0],"81")==0);

	// everything worked
	printf("passed");
	return 0;
}

#endif
