//
//   $Id: DLMEMSTO.C 1.4 2002/10/15 22:45:24Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DLMEMSTO.C  created 03/07/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This module contains all the memory routines.  The
   routines provided here make it easy for backup to use
   as much as possible, including both conventional and
   XMS memory.  These routines are used by the backup
   and restore file routines.

**********************************************************/
#define __FILENUM__	502

//#include <alloc.h>
#include <mem.h>
#include <stdio.h>
#include <stdlib.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dlmem.h"
#include "dlXMS.h"


#include "dlmemsto.h"
#include "dlmemsto.def"


//  Enable these to see more error information on failures
//
#define DEBUG_ERRORS			FALSE
#define DEBUG_CALLS			FALSE


//  Constants for READ/WRITE mode
//
#define MEM_READ				TRUE
#define MEM_WRITE				FALSE


// note: all the below count on being initialized to 0
#if MEMSTO_USE_XMS
static unsigned	uXMSHandle;
static ulong 		ulXMSSize;
#endif
static void far * lpMem;
static ulong 		ulConvMemSize;			// size of conventional memory
static ulong 		ulMemPos;


unsigned uConvSegment;

/*
	Initialize the memory storage system, used to speed up
	copying files to/from a backup.  Can only fail if no memory
	is available.
*/
bool InitMemStore(void)
{
	unsigned uConvMemSize = 0;
	bool fSuccessful;
#if MEMSTO_USE_XMS
	unsigned uXMSSizeK = 0;
	unsigned uTotalXMSK;
#endif

	//  Determine the amount of available conventional memory
	//
	dlAllocMem( 0xffff, &uConvMemSize );
	ulConvMemSize = (ulong)uConvMemSize << 4;

//  BT 2-25-2000
//
//  There should not be a limit imposed for the size of the memroy store
//
//	if(ulConvMemSize < MIN_STORE_SIZE)
//		return FALSE;


	if( uConvMemSize )
	{
		//  Allocate all the conventional memory we can
		//
		uConvSegment = 0;
		fSuccessful = dlAllocMem( uConvMemSize, &uConvSegment );
		if( !fSuccessful )
		{
			return FALSE;
		}
//		printf("mem store segment: %x  para: %x\n", uConvSegment, uConvSegment );

		lpMem = MK_FP(uConvSegment, 0);
		if(!lpMem)
			return FALSE;
	}

#if MEMSTO_USE_XMS
	// allocate all of XMS memory too
	//
	uXMSHandle = 0;
	ulXMSSize  = 0;
	if(CheckXMMInstalled() == XMS_SUCCESS)
	{
		if( GetXMSFree( &uXMSSizeK, &uTotalXMSK ) == XMS_SUCCESS)
		{
			if( AllocateXMSBlock( uXMSSizeK, &uXMSHandle ) == XMS_SUCCESS)
			{
				ulXMSSize = 1024L * uXMSSizeK;
			}
		}
	}
	if( (!ulConvMemSize) && (!ulXMSSize) )
	{
		//  there must be some memory for the memory store
		return FALSE;
	}

#else
	if( !ulConvMemSize )
	{
		//  there must be some memory for the memory store
		return FALSE;
	}
#endif

	// got the memory
	return TRUE;
}

/*
	Free all the memory we just allocated.  Does not hurt to
	BEFORE the memory has been allocated, or after it has been
	freed.
*/
void KillMemStore(void)
{
#if MEMSTO_USE_XMS
	// any XMS memory allocated? free it!
	if(uXMSHandle)
	{
		FreeXMSBlock(uXMSHandle);
		uXMSHandle = 0;
	}
#endif

	// any conventional memory allocated? free it!
	if(uConvSegment)
	{
		dlFreeMem( uConvSegment );
		uConvSegment = 0;
	}
}

/*
	Return size of total in-memory storage.
*/
ulong MemStoreSize(void)
{
	// return total size of store
#if MEMSTO_USE_XMS
	return ulConvMemSize + ulXMSSize;
#else
	return ulConvMemSize;
#endif
}

/*
	What is the current position in the in-memory storage
	stream?
*/
ulong MemStoreTell(void)
{
	return ulMemPos;
}


#if SUPPORT_GET_ADDRESS
void far * GetCurrentStoreAddress( void )
{
	return MK_FP( uConvSegment + (unsigned)(ulMemPos >> 4),
			((unsigned)ulMemPos) & 0xf );
}
#endif


/*
	How much memory is left in the store?
*/
ulong MemStoreLeft(void)
{
	return MemStoreSize() - MemStoreTell();
}

/*
	Seek to a given position in the in-memory store.  Unlike
	with a file seek position, do NOT seek out of bounds.
*/
void MemStoreSeek( ulong ulPosition, unsigned uWhence)
{
	// validate parameter
	ASSERT(uWhence == SEEK_SET || uWhence == SEEK_CUR || uWhence == SEEK_END);

	// seek to where user wants
	switch(uWhence)
	{
		case SEEK_SET:	ulMemPos = ulPosition; 	break;
		case SEEK_CUR:	ulMemPos += ulPosition;	break;
		case SEEK_END:	ulMemPos = MemStoreSize() + ulPosition; break;
		default: ABORT();
	}

	// should NEVER seek out of bounds
	ASSERT(ulMemPos >= 0L && ulMemPos <= MemStoreSize());
}

/*
	Read or write to/from Memory store to/from the caller's buffer.
*/
unsigned MemStoreReadWrite(void far * lpBytes, unsigned size, bool fDoRead )
{
	char far * lpBuff = lpBytes;
	unsigned uConvReadWriteSize;
#if MEMSTO_USE_XMS
	unsigned uXMSReadWriteSize;
	unsigned uNonDwordAlignedSize;
	unsigned uDwordAlignedSize;
	static XMS_MOVE_STRUCT XMSMoveStruct;
	static char cTempBuff[4];
	char huge * hpBuff;
	bool fReadAheadInXMS;
	unsigned ucTemp;
	int iError;
#endif


	ASSERT(lpBuff);

	// out of bounds, no reading possible
	if(!size)
		return 0;

	ASSERT((ulMemPos >= 0L) && (ulMemPos+size <= ulConvMemSize+ulXMSSize));
#if MEMSTO_USE_XMS
	if(ulMemPos < 0L || (ulMemPos >= ulConvMemSize+ulXMSSize))
		return 0;
#else
	if(ulMemPos < 0L || (ulMemPos >= ulConvMemSize))
		return 0;
#endif


#if MEMSTO_USE_XMS
	// must shorten read
	if((ulMemPos + size) > (ulConvMemSize+ulXMSSize))
		size = (unsigned)((ulConvMemSize+ulXMSSize)- ulMemPos);
#else
	// must shorten read
	if((ulMemPos + size) > ulConvMemSize )
		size = (unsigned)(ulConvMemSize - ulMemPos);
#endif


#if MEMSTO_USE_XMS
	//  Determine how much of the read will be done in conventional memory
	//
	if( ulMemPos > ulConvMemSize )
	{
		uConvReadWriteSize = 0;
	}
	else
	{
		if( (ulMemPos+size) > ulConvMemSize )
		{
			uConvReadWriteSize = (unsigned)(ulConvMemSize - ulMemPos);
		}
		else
		{
			uConvReadWriteSize = size;
		}
	}
#else
	uConvReadWriteSize = size;
#endif

#if MEMSTO_USE_XMS
	//  Determine how much of the read will be done in XMS memory
	//
	uXMSReadWriteSize = size - uConvReadWriteSize;
#endif


	//  Copy the part in conventional memory
	//
	if( uConvReadWriteSize )
	{
		if( fDoRead )
		{
			// let compiler do the math
			_fmemcpy(lpBuff, Normalize(lpMem,ulMemPos), uConvReadWriteSize);
		}
		else
		{
			// let compiler do the math
			_fmemcpy(Normalize(lpMem,ulMemPos), lpBuff, uConvReadWriteSize);
		}
	}

	// record our new position
	MemStoreSeek(uConvReadWriteSize, SEEK_CUR);

#if MEMSTO_USE_XMS
	//  Copy the part in XMS memory
	//
	if( uXMSReadWriteSize )
	{
		hpBuff = lpBuff;
		hpBuff += uConvReadWriteSize;
		lpBuff = (char far * )hpBuff;


		//  Perform any dword aligned part of the move
		//
		uDwordAlignedSize = uXMSReadWriteSize & 0xfffc;
		if( fDoRead )
		{
			XMSMoveStruct.length = uDwordAlignedSize;
			XMSMoveStruct.sourceHandle = uXMSHandle;
			XMSMoveStruct.sourceOffset.offset32 = ulMemPos - ulConvMemSize;
			XMSMoveStruct.destHandle = 0;
			XMSMoveStruct.destOffset.p.lpOff = FP_OFF(lpBuff);
			XMSMoveStruct.destOffset.p.lpSeg = FP_SEG(lpBuff);
		}
		else
		{
			XMSMoveStruct.length = uDwordAlignedSize;
			XMSMoveStruct.sourceHandle = 0;
			XMSMoveStruct.sourceOffset.p.lpOff = FP_OFF(lpBuff);
			XMSMoveStruct.sourceOffset.p.lpSeg = FP_SEG(lpBuff);
			XMSMoveStruct.destHandle = uXMSHandle;
			XMSMoveStruct.destOffset.offset32 = ulMemPos - ulConvMemSize;
		}
//  Debug output
//
#if DEBUG_CALLS
					printf("\n");
		printf("Line: %u\n", __LINE__ );
		printf("XMSMoveStruct.length: %lx\n",
				XMSMoveStruct.length );
		printf("XMSMoveStruct.sourceHandle: %x\n",
				XMSMoveStruct.sourceHandle );
		printf("XMSMoveStruct.sourceOffset.offset32: %lx\n",
				XMSMoveStruct.sourceOffset.offset32 );
		printf("XMSMoveStruct.destHandle: %x\n",
				XMSMoveStruct.destHandle );
		printf("XMSMoveStruct.destOffset.offset32: %lx\n",
				XMSMoveStruct.destOffset.offset32 );
		printf("ulConvMemSize: %lx\n", ulConvMemSize );
		printf("ulXMSSize: %lx\n", ulXMSSize );
		printf("MemStoreSize(): %lx\n", MemStoreSize() );
					printf("\n");
#endif
		iError = MoveXMSBlock(&XMSMoveStruct);
		if( iError != XMS_SUCCESS )
		{
			printf("XMS mem move error! error: %x\n", iError );

//  Debug output
//
#if DEBUG_ERRORS
					printf("\n");
			printf("Line: %u\n", __LINE__ );
			printf("XMSMoveStruct.length: %lx\n",
					XMSMoveStruct.length );
			printf("XMSMoveStruct.sourceHandle: %x\n",
					XMSMoveStruct.sourceHandle );
			printf("XMSMoveStruct.sourceOffset.offset32: %lx\n",
					XMSMoveStruct.sourceOffset.offset32 );
			printf("XMSMoveStruct.destHandle: %x\n",
					XMSMoveStruct.destHandle );
			printf("XMSMoveStruct.destOffset.offset32: %lx\n",
					XMSMoveStruct.destOffset.offset32 );
			printf("ulConvMemSize: %lx\n", ulConvMemSize );
			printf("ulXMSSize: %lx\n", ulXMSSize );
			printf("MemStoreSize(): %lx\n", MemStoreSize() );
					printf("\n");
#endif

			exit(255);
		}
//  Debug output
//
#if 0
		else
		{
			printf("XMSMoveStruct.length: %lx\n",
					XMSMoveStruct.length );
			printf("XMSMoveStruct.sourceHandle: %x\n",
					XMSMoveStruct.sourceHandle );
			printf("XMSMoveStruct.sourceOffset.offset32: %lx\n",
					XMSMoveStruct.sourceOffset.offset32 );
			printf("XMSMoveStruct.destHandle: %x\n",
					XMSMoveStruct.destHandle );
			printf("XMSMoveStruct.destOffset.offset32: %lx\n",
					XMSMoveStruct.destOffset.offset32 );
		}
#endif

		//  Perform any non-dword aligned move
		//
		uNonDwordAlignedSize = uXMSReadWriteSize & 3;
		if( uNonDwordAlignedSize )
		{
			hpBuff = lpBuff;
			hpBuff += uDwordAlignedSize;

			if( (ulMemPos-ulConvMemSize+uDwordAlignedSize+4) > ulXMSSize )
			{
				fReadAheadInXMS = FALSE;
			}
			else
			{
				fReadAheadInXMS = TRUE;
			}

			if( fReadAheadInXMS )
			{
				XMSMoveStruct.length = 4;
				XMSMoveStruct.sourceHandle = uXMSHandle;
				XMSMoveStruct.sourceOffset.offset32 = ulMemPos - ulConvMemSize +
						uDwordAlignedSize;
				XMSMoveStruct.destHandle = 0;
				XMSMoveStruct.destOffset.p.lpOff = FP_OFF(cTempBuff);
				XMSMoveStruct.destOffset.p.lpSeg = FP_SEG(cTempBuff);
			}
			else
			{
				XMSMoveStruct.length = 4;
				XMSMoveStruct.sourceHandle = uXMSHandle;
				XMSMoveStruct.sourceOffset.offset32 = ulMemPos - ulConvMemSize +
						uDwordAlignedSize - (4 - uNonDwordAlignedSize);
				XMSMoveStruct.destHandle = 0;
				XMSMoveStruct.destOffset.p.lpOff = FP_OFF(cTempBuff);
				XMSMoveStruct.destOffset.p.lpSeg = FP_SEG(cTempBuff);
			}

			if( fDoRead )
			{
//  Debug output
//
#if DEBUG_CALLS
				printf("Line: %u\n", __LINE__ );
				printf("XMSMoveStruct.length: %lx\n",
						XMSMoveStruct.length );
				printf("XMSMoveStruct.sourceHandle: %x\n",
						XMSMoveStruct.sourceHandle );
				printf("XMSMoveStruct.sourceOffset.offset32: %lx\n",
						XMSMoveStruct.sourceOffset.offset32 );
				printf("XMSMoveStruct.destHandle: %x\n",
						XMSMoveStruct.destHandle );
				printf("XMSMoveStruct.destOffset.offset32: %lx\n",
						XMSMoveStruct.destOffset.offset32 );
				printf("ulConvMemSize: %lx\n", ulConvMemSize );
				printf("ulXMSSize: %lx\n", ulXMSSize );
				printf("MemStoreSize(): %lx\n", MemStoreSize() );
#endif
				iError = MoveXMSBlock(&XMSMoveStruct);
				if( iError != XMS_SUCCESS )
				{
					printf("XMS mem move error! error: %x\n", iError );

//  Debug output
//
#if DEBUG_ERRORS
					printf("Line: %u\n", __LINE__ );
					printf("XMSMoveStruct.length: %lx\n",
							XMSMoveStruct.length );
					printf("XMSMoveStruct.sourceHandle: %x\n",
							XMSMoveStruct.sourceHandle );
					printf("XMSMoveStruct.sourceOffset.offset32: %lx\n",
							XMSMoveStruct.sourceOffset.offset32 );
					printf("XMSMoveStruct.destHandle: %x\n",
							XMSMoveStruct.destHandle );
					printf("XMSMoveStruct.destOffset.offset32: %lx\n",
							XMSMoveStruct.destOffset.offset32 );
					printf("ulConvMemSize: %lx\n", ulConvMemSize );
					printf("ulXMSSize: %lx\n", ulXMSSize );
					printf("MemStoreSize(): %lx\n", MemStoreSize() );
#endif
				}

				if( fReadAheadInXMS )
				{
					for( ucTemp = 0; ucTemp < uNonDwordAlignedSize; ucTemp++ )
					{
						hpBuff[ucTemp] = cTempBuff[ucTemp];
					}
				}
				else
				{
					for( ucTemp = 0; ucTemp < uNonDwordAlignedSize; ucTemp++ )
					{
						hpBuff[ucTemp] = cTempBuff[4-uNonDwordAlignedSize+ucTemp];
					}
				}
			}
			else
			{
//  Debug output
//
#if DEBUG_CALLS
					printf("\n");
				printf("Line: %u\n", __LINE__ );
				printf("XMSMoveStruct.length: %lx\n",
						XMSMoveStruct.length );
				printf("XMSMoveStruct.sourceHandle: %x\n",
						XMSMoveStruct.sourceHandle );
				printf("XMSMoveStruct.sourceOffset.offset32: %lx\n",
						XMSMoveStruct.sourceOffset.offset32 );
				printf("XMSMoveStruct.destHandle: %x\n",
						XMSMoveStruct.destHandle );
				printf("XMSMoveStruct.destOffset.offset32: %lx\n",
						XMSMoveStruct.destOffset.offset32 );
				printf("ulConvMemSize: %lx\n", ulConvMemSize );
				printf("ulXMSSize: %lx\n", ulXMSSize );
				printf("MemStoreSize(): %lx\n", MemStoreSize() );
					printf("\n");
#endif
				iError = MoveXMSBlock(&XMSMoveStruct);
				if( iError != XMS_SUCCESS )
				{
					printf("XMS mem move error! error: %x\n", iError );

//  Debug output
//
#if DEBUG_ERRORS
					printf("\n");
					printf("Line: %u\n", __LINE__ );
					printf("XMSMoveStruct.length: %lx\n",
							XMSMoveStruct.length );
					printf("XMSMoveStruct.sourceHandle: %x\n",
							XMSMoveStruct.sourceHandle );
					printf("XMSMoveStruct.sourceOffset.offset32: %lx\n",
							XMSMoveStruct.sourceOffset.offset32 );
					printf("XMSMoveStruct.destHandle: %x\n",
							XMSMoveStruct.destHandle );
					printf("XMSMoveStruct.destOffset.offset32: %lx\n",
							XMSMoveStruct.destOffset.offset32 );
					printf("ulConvMemSize: %lx\n", ulConvMemSize );
					printf("ulXMSSize: %lx\n", ulXMSSize );
					printf("MemStoreSize(): %lx\n", MemStoreSize() );
					printf("\n");
#endif
				}

				if( fReadAheadInXMS )
				{
					for( ucTemp = 0; ucTemp < uNonDwordAlignedSize; ucTemp++ )
					{
						cTempBuff[ucTemp] = hpBuff[ucTemp];
					}

					XMSMoveStruct.length = 4;
					XMSMoveStruct.sourceHandle = 0;
					XMSMoveStruct.sourceOffset.p.lpOff = FP_OFF(cTempBuff);
					XMSMoveStruct.sourceOffset.p.lpSeg = FP_SEG(cTempBuff);
					XMSMoveStruct.destHandle = uXMSHandle;
					XMSMoveStruct.destOffset.offset32 = ulMemPos - ulConvMemSize +
							uDwordAlignedSize;
				}
				else
				{
					for( ucTemp = 0; ucTemp < uNonDwordAlignedSize; ucTemp++ )
					{
						cTempBuff[4-uNonDwordAlignedSize+ucTemp] = hpBuff[ucTemp];
					}

					XMSMoveStruct.length = 4;
					XMSMoveStruct.sourceHandle = 0;
					XMSMoveStruct.sourceOffset.p.lpOff = FP_OFF(cTempBuff);
					XMSMoveStruct.sourceOffset.p.lpSeg = FP_SEG(cTempBuff);
					XMSMoveStruct.destHandle = uXMSHandle;
					XMSMoveStruct.destOffset.offset32 = ulMemPos - ulConvMemSize +
						uDwordAlignedSize - (4 - uNonDwordAlignedSize);
				}

//  Debug output
//
#if DEBUG_CALLS
					printf("\n");
				printf("Line: %u\n", __LINE__ );
				printf("XMSMoveStruct.length: %lx\n",
						XMSMoveStruct.length );
				printf("XMSMoveStruct.sourceHandle: %x\n",
						XMSMoveStruct.sourceHandle );
				printf("XMSMoveStruct.sourceOffset.offset32: %lx\n",
						XMSMoveStruct.sourceOffset.offset32 );
				printf("XMSMoveStruct.destHandle: %x\n",
						XMSMoveStruct.destHandle );
				printf("XMSMoveStruct.destOffset.offset32: %lx\n",
						XMSMoveStruct.destOffset.offset32 );
				printf("ulConvMemSize: %lx\n", ulConvMemSize );
				printf("ulXMSSize: %lx\n", ulXMSSize );
				printf("MemStoreSize(): %lx\n", MemStoreSize() );
					printf("\n");
#endif
				iError = MoveXMSBlock(&XMSMoveStruct);
				if( iError != XMS_SUCCESS )
				{
					printf("XMS mem move error! error: %x\n", iError );

//  Debug output
//
#if DEBUG_ERRORS
					printf("\n");
					printf("Line: %u\n", __LINE__ );
					printf("XMSMoveStruct.length: %lx\n",
							XMSMoveStruct.length );
					printf("XMSMoveStruct.sourceHandle: %x\n",
							XMSMoveStruct.sourceHandle );
					printf("XMSMoveStruct.sourceOffset.offset32: %lx\n",
							XMSMoveStruct.sourceOffset.offset32 );
					printf("XMSMoveStruct.destHandle: %x\n",
							XMSMoveStruct.destHandle );
					printf("XMSMoveStruct.destOffset.offset32: %lx\n",
							XMSMoveStruct.destOffset.offset32 );
					printf("ulConvMemSize: %lx\n", ulConvMemSize );
					printf("ulXMSSize: %lx\n", ulXMSSize );
					printf("MemStoreSize(): %lx\n", MemStoreSize() );
					printf("\n");
#endif
				}
			}
		}
	}

	// record our new position
	MemStoreSeek(uXMSReadWriteSize, SEEK_CUR);
#endif

	return size;
}

/*
	Read from Memory store into the caller's buffer.
*/
unsigned MemStoreRead(void far * lpBytes, unsigned size)
{
	return MemStoreReadWrite( lpBytes, size, MEM_READ );
}


/*
	Write to Memory store from the caller's buffer.
	note: not yet aware of XMS memory.
*/
unsigned MemStoreWrite(void far * lpBytes, unsigned size)
{
	return MemStoreReadWrite( lpBytes, size, MEM_WRITE );
}


#if UNIT_TEST

#include <stdlib.h>

void InternalError(int fileNum, int lineNum)
{
	printf("Error %d.%d\n",fileNum,lineNum);
	exit(99);
}

void main(void)
{
	// start up memory store
	if(!InitMemStore())
		ABORT();

	// done, exit
	KillMemStore();
}
#endif
