//
//   $Id: PRINTF.C 1.5 2002/10/15 22:45:24Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
	$DOCFILE:DLPRINTF.C

	Copyright (c) 1996-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Contains printf() in all its glory (well, some of it anyway).

	Supports
		%c    char
		%d    integer
		%i    same as %d
		%o    octal
		%u    unsigned
		%x    unsigned in hex form, UPPERCASE
		%s    char * (in ascii)
		%p    near pointer (same as %04x)
		%ld   long integer
		%li   same as %ld
		%lo   long octal
		%lu   unsigned long
		%lx   unsigned long in hex form, UPPERCASE
		%ls	char far * (in ascii)
		%lp	far pointer (same as %04X:%04X)
		%qd	quad integer (64-bits)

	Also support "count", "fill", "left justify" and "commas"
	modifiers:
		%11d		fill with up to 11 spaces
		%011d		fill with up to 11 0s
		%-5d		left justify with fill
		%,5d		use commas for thousands separator
		%*d		field width on stack

	$DOCHISTORY
	02/21/1998 dg  Added 64-bit %qd formatted output
						Added %*d variable width formatting
   08/12/1997 DE	fixed %c bug where c was a NUL (wasn't outputing NUL)
						now returns correct result (was 1 too few)
						now supports %o, %lo
	04/18/1996 dg  Created by Drew Gislason
***************************************************************************/
#define __FILENUM__	501


#include <string.h>


#include "datalght.h"
#include "dlprintf.h"
#include "dl64.h"

#include "_printf.h"


// can be changed for foreign country support
#if OPT_THOUSAND
extern char printf_thousands_sep;
extern char printf_sep;								// fill number separator
#endif


// various printf modifiers, set during a given call to printf
#if OPT_FILL
char printf_fill;		// fill character
int	printf_count;	// fill count
char printf_left;		// fill alignment
#endif




static int charCount;


// the output buffer
char printf_buffer[BUFSIZE];
unsigned printf_buffer_size = BUFSIZE;
int printf_position;


/*
	Flush the printf buffer to
*/
void _FlushBuffer(void)
{
	// nothing to do? just return
	if(!printf_position)
		return;

#if SPRINTF
	// flush the buffer to user space
	if(printf_pUserBuffer)
	{
		memmove(printf_pUserBuffer, printf_buffer, printf_position);
		printf_pUserBuffer += printf_position;
	}

	// flush the buffer to DOS handle
	else
#endif // SPRINTF
	{
		// workaround for BUG in BC 2.0 compiler that screws up
		// on inline assembly.  If a variable is on an odd boundary,
		// generates the instruction from the wrong address!
		unsigned r_bx = printfHandle;
		unsigned r_cx = printf_position;
		unsigned r_dx = (unsigned)(printf_buffer);

		// call DOS to display the string
		asm	mov	bx, r_bx
		asm	mov	cx, r_cx
		asm	mov	dx, r_dx
		asm	mov	ah, 40h
		asm	int	21h
	}


	// update # of characters flushed
	charCount += printf_position;


	// buffer is now empty
	printf_position = 0;
}


/*
	Just as with standard printf(), this function is the real
	work-horse.  It prints the string to a buffer, and flushes
	to screen or user space.  This function supports a subset
	of the standard C printf().  See module header for a list
	of what's supported.

	Note: to fool the Datalight prototyping program, the
	function header must be indented.
*/
 int vprintf(const char *fmt, void *xargs)
{
	int *args = (int *)xargs;
	char szTempStr[32];		// big enough to hold a 64-bit ascii number

	// we will be counting the # of character displayed.
	// start with none.
	charCount = 0;

	// keep going until the end of the format string
	while(*fmt)
	{
		// we can't ASSERT here, so just return if fmt string is bad
		if(!fmt)
			break;

		// general output
		if(*fmt != '%')
		{
			// special case: display a CR before a LF to screen,
			// but NOT when using sprintf().
			if(*fmt == '\n')
			{
#if SPRINTF
				if(!printf_pUserBuffer)
#endif
					_putc('\r');
			}

			// output the character
			_putc(*fmt);
		}

		/* this is a formatted output */
		else
		{
			// skip the "%" part of the formatted output
			++fmt;

#if OPT_FILL
			// determine fill alignment (left or right)
			printf_left = FALSE;
			if(*fmt == '-')
			{
				++fmt;
				printf_left = TRUE;
			}
#endif // OPT_FILL

#if OPT_THOUSAND
			// Use thousands separator?
			printf_sep = 0;
			if(*fmt == ',')
			{
				++fmt;
				printf_sep = printf_thousands_sep;
			}
#endif // OPT_THOUSAND

#if OPT_FILL
			// is there a fill count?
			printf_count = 0;

			// variable length fill
			if(*fmt == '*')
			{
				printf_fill = ' ';
				printf_count = *args++;
				++fmt;
			}

			else if(IsDigit(*fmt))
			{
				// should we fill with 0s instead of spaces?
				if(*fmt == '0')
					printf_fill='0', ++fmt;
				else
					printf_fill=' ';

				// determine count of bytes to fill
				while(IsDigit(*fmt))
				{
					printf_count *= 10;
					printf_count += *fmt++ - '0';
				}
			}
#endif // OPT_FILL

			// what kind of thing is this?
			switch(*fmt)
			{
#if FMT_C
				// character (%c)
				case 'c':
					// get the character
					szTempStr[0] = (char)(*args++);
					if(szTempStr[0] == 0)
					{
						_putc(0); // have to be able to %c a NUL for ZMODEM ZFILE
					}
					else
					{
						szTempStr[1] = 0;

						// we do it in string form so we can fill if needed
						_PutStringWithFill(szTempStr);
					}
					break;
#endif // FMT_C

#if FMT_D
				// integer (%d)
				case 'd':
				case 'i':

					// convert to number string
					itoa(*args++, szTempStr, 10);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
					break;
#endif // FMT_D

#if FMT_O
				case 'o':

					// convert octal number to string
					itoa(*args++, szTempStr, 8);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
					break;
#endif // FMT_O

#if FMT_U
				// unsigned (%u)
				case 'u':
					// convert to number unsigned string
					ltoa((unsigned)(*args++), szTempStr, 10);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
					break;
#endif // FMT_U

#if FMT_X
				// hex (%x)
				case 'x':
				case 'X':
					// convert to number unsigned string
					itoa(*args++, szTempStr, 16);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
					break;
#endif // FMT_X

#if FMT_S
				// string (%s)
				case 's':

					// simply display the string
#if (__SMALL__ || __TINY__ || __MEDIUM__)
					_PutStringWithFill((char *)(*args++));
#else
					{
						char far *lpStr;
						lpStr = (char far *)(*(((long *)args)++));
						_PutStringWithFill(lpStr);
					}
#endif
					break;
#endif // FMT_S

#if FMT_P
				// pointer in hex form (%p)
				case 'p':

					// set fill and count, as if user did %04x
					printf_fill='0'; printf_count=4;

					// get and display the near pointer
					itoa(*args++, szTempStr, 16);
					_PutStringWithFill(szTempStr);

					break;
#endif // FMT_P

				// some printf() things are longs
				case 'l':

					// skip the 'l', to get to the type 'd', 's', etc..
					++fmt;

#if FMT_LD
				// long integer (%ld)
				if(*fmt == 'd' || *fmt == 'i')
				{
					// convert to long number string
					ltoa(*(((long *)args)++), szTempStr, 10);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
				}
#endif // FMT_LD

#if FMT_LO
				// long octal integer (%lo)
				if(*fmt == 'o')
				{
					// convert to long number string
					ultoa(*(((long *)args)++), szTempStr, 8);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
				}
#endif // FMT_LO

#if FMT_LU
				// long integer (%lu)
				if(*fmt == 'u')
				{
					// convert to long number string
					ultoa(*(((long *)args)++), szTempStr, 10);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
				}
#endif // FMT_LU

#if FMT_LX
				// long hex number (%lx)
				if(*fmt == 'x' || *fmt == 'X')
				{
					// convert to long number string
					ltoa(*(((long *)args)++), szTempStr, 16);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
				}
#endif // FMT_LX

#if FMT_LS
				// far string (%ls)
				if(*fmt == 's')
				{
					char far *lpStr;

					// determine start of string
					lpStr = (char far *)(*(((long *)args)++));

					// display each character
					while(*lpStr)
					{
						_putc(*lpStr);
						++lpStr;
					}
				}
#endif // FMT_LS

#if FMT_LP
				// far pointer (%lp) (eg 0300:0FAE)
				if(*fmt == 'p')
				{
					// set fill and count, as if user did %04x
					printf_fill='0'; printf_count=4;

					// get and display segment part of number
					itoa(args[1], szTempStr, 16);
					_PutStringWithFill(szTempStr);

					// separate by a colon
					_putc(':');

					// get and display offset part of number
					itoa(args[0], szTempStr, 16);
					_PutStringWithFill(szTempStr);

					// skip the far pointer	argument
					args += 2;
				}
#endif // FMT_LP

					// done with case '%l'
					break;

#if FMT_QD
				// some printf() things are longs
				case 'q':

					++fmt;

					// only handle unsigned quads
					if(*fmt == 'd' || *fmt == 'u')
					{
						QuadToAscii((uquad *)args, szTempStr, 10);
						args += 4;
						_PutStringWithFill(szTempStr);
					}
					break;
#endif // FMT_QD

				// unknown formatting character, just spit it out
				default:
					_putc(*fmt);
			}
		}

		// try next character
		++fmt;
	}

	// flush the last of the output to stdout or wherever it's going
	_FlushBuffer();

	// return the # of characters printed
	return charCount;
}


/*
	Put a number or string, paying attention to fill bytes,
	count and alignment.
*/
void _PutStringWithFill(char *pStr)
{
#if OPT_FILL || OPT_THOUSAND
	int len;

	// determine length of string
	len = strlen(pStr);
#endif

#if OPT_THOUSAND
	// expand the string with "thousands" separators
	if(printf_sep)
	{
		int numLen;

		// only need a thousands separator every 3rd number
		numLen = len;
		while(numLen > 3)
		{
			numLen -= 3;

			// open up string and put in thousands ","
			memmove(pStr+numLen+1, pStr+numLen, len-numLen+1);
			pStr[numLen] = printf_sep;

			// string is now 1 byte longer
			++len;
		}
	}
#endif // OPT_THOUSAND

#if OPT_FILL
	// left aligned string? Then spit out the meat BEFORE the fill
	if(printf_left)
		_puts(pStr);

	// any characters to fill?
	while(len < printf_count)
	{
		_putc(printf_fill);
		--printf_count;
	}

	// right aligned string? Then spit out the meat AFTER the fill
	if(!printf_left)
		_puts(pStr);
#else
	_puts(pStr);
#endif
}
