//  $Id: DYNDRVR.C 1.6 2003/03/20 16:56:12Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1997-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/********************************************************************
   $DOCMODULE:DYNDRVR.C

	Copyright (c) 1998-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	This module shows how to use the the Datalight Dynamic Driver
	Interface (installing DOS Device Drivers based on hardware
	detection at boot time).

	This example simply detects how much memory is available via the
	XMS (HIMEM) interface, and creates 2 RAM disks: one that uses
	up 1/3 of the XMS memory, the other that uses up 2/3 of the XMS
	memory. Installs no RAM disks if HIMEM (XMS) is not detected.

	This technology can be used to detect CD-ROM drives, network
	cards, GPS units, or other possible add-on hardware in a system.
	Then the appropriate drivers can be installed.

	$DOCHISTORY
	08-12-1998 dg  Added VDISK disk example.
	10-07-1997 de  Created by Dennis Edwards.
*********************************************************************/
#include "dosver.h"
#include "compiler.h"
#include "memdisk.h"
#include "dyndrvr.h"

//#include "fatworld.h"
bool PASCAL dlIsFat32World( void );

#define DEF_DIRS  64			/* default # of root directory entries */
#define DEF_SIZEK 64			/* default size of VDISK in KBytes */
#define DEF_DIR_ENTRIES 	DEF_DIRS
struct BPB std_bpb =
{
	DEF_SEC_SIZE,1,1,2,DEF_DIR_ENTRIES,DEF_SIZEK*2,FAT_ID,6,9,2,0
};
char std_volume[] = "VDISK V6.22";
unsigned long std_serial = 0x12345678L;
char std_fattype[] = "FAT12    ";

bool fIsFat32World;

void DisplaySignOnMessage( void )
{
	char * szSignOn = "\nDatalight Dynamic Driver Loader %s %s\n%s\n";

	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/*
	The "main" routine. Modify this routine to
   Returns 0, will never install as a device.
*/
#ifdef __TURBOC__
#pragma argsused
#endif
int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv)
{
	char		cDriveLetter;
	int		iHandle;
	unsigned	uAction;
	int		iCurDrive;
	unsigned	uBytes;
	unsigned kBytesExtended=0;
	unsigned kBytesDrive1;
	unsigned kBytesDrive2;
	static char szVolume[] = "DL RAM D.ISK";	// volume label in 8.3 form
	static char szDynCfg[] = "A:\\DYNCFG.SYS";
	static char szInstallVdisk[] = "DEVICE=A:\\VDISK.SYS %d /e\r\n";
	static char szDriveStr[40];

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* display sign-on */
	DisplaySignOnMessage();

	// Find VDISK.SYS in root of current drive
	dlGetCurrentDrive(&iCurDrive);
	szInstallVdisk[7] = (char)iCurDrive + 'A';

	// determine sizes of VDISKs
	if(!IsHimemInstalled() || (kBytesExtended=TotalXMS()) == 0)
		printf("Error: No extended memory!\n");
	kBytesDrive1 = kBytesExtended / 3;
	kBytesDrive2 = kBytesExtended - kBytesDrive1;

#if DEBUG
	// Wake up the debugger...
	//
	ASM int 3
#endif

	// determine drive letter for the disk to contain the Dynamic CONFIG.SYS
	cDriveLetter = FindDiskVolume(szVolume);
	if(!cDriveLetter)
	{
		printf("Error: volume (%s) not found!\n",szVolume);
		return 0;
	}

	// found the drive letter! Update NEWFILE command in current CONFIG.SYS
	// to reflect that drive letter.
	if(!PatchConfigBufferNEWFILEDrive(szDynCfg, cDriveLetter))
	{
		printf("Error: NEWFILE=%s not found!\n",szDynCfg);
		return 0;
	}
	*szDynCfg = cDriveLetter;

	// now create that dynamic CONFIG.SYS (DYNCFG.SYS)
	uAction = ACT_NEXIST_CREATE;
	if(!dlOpenOrCreateFile(O_WRONLY, A_NORMAL, &uAction, szDynCfg, &iHandle))
	{
		printf("Error: couldn't create %s!\n",szDynCfg);
		return 0;
	}

	// only install VDISK if extended memory is available
	if(kBytesExtended)
	{
		// install VDISK drive 1 (1/3 of XMS memory)
		uBytes = sprintf(szDriveStr, szInstallVdisk, kBytesDrive1);
		dlWriteFile(iHandle, &uBytes, szDriveStr);

		// install VDISK drive 2 (2/3 of XMS memory)
		uBytes = sprintf(szDriveStr, szInstallVdisk, kBytesDrive2);
		dlWriteFile(iHandle, &uBytes, szDriveStr);
	}

	// done, close the file
	dlCloseFile(iHandle);

	// display where control is going
	printf("Passing control to... %s\n",szDynCfg);

	return 0;
}


/*
	Dummy function to make MEMDISK link.
*/
#ifdef __TURBOC__
#pragma argsused
#endif
bool memread(long offset, unsigned len, char far *buffer)
{
	return 0;
}
