//  $Id: INT15PAG.C 1.6 2003/03/20 16:55:50Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1998-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:INT15PAG.C

	Copyright (c) 1998-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This module is the paging mechanism for any system whose BIOS
	supports Int 15h, function 87h (move block) and 88h (get extended
	memory size). It is used both by the Int15 Paged Disk Driver
	and the Int15 Paged Loader.

   Modules:
   MEMDISK+MEMPAGED+INT15PAG   Makes the Int15 Paged Disk Driver.
   LDRBEXT+PAGEDLDR+INT15PAG   Makes the Int15 Paged Loader.

   A paged disk driver treats the memory as a disk (usually a read-
   only ROM-DISK) once the ROM-DOS kernel is loaded. Depending on
   the options set with MEMDISK, this paged disk driver can either
   be built-in to the ROM-DOS kernel, or installed via CONFIG.SYS.

   A paged loader uses the memory to load the ROM-DOS kernel itself
   (before disk drivers are loaded). It will find ROM-DOS.SYS on the
   ROM-DISK image in the paged memory.

	This particular example uses Int 15h to copy memory from extended
	(above 1 Meg) to conventional (below 640K) and vice versa. A
	small 2K window is used to copy the memory.

	This merely serves as an example, or skeleton for making your
	own Paged Loader or Paged Disk Driver.

	$DOCHISTORY
	07/23/1998 dg  Updated to Solid Code
   04/08/1998 dg  Created by Drew Gislason
*********************************************************************/
#include "compiler.h"
#include "memdisk.h"

/* specific to Int15Pag.c */
#define PAGESIZE 		1024 		/* 1K window */
static char window[PAGESIZE];
unsigned nMaxPages;

/* invalid page # */
#define INVALID_PAGE 0xFFFF

/* these variables MUST be defined for each paging mechansim */
char szPagedDiskIDString[] = {"(Int15 Version)"};
long lPagedDiskWindowSize = PAGESIZE;
unsigned char far *lpPagedDiskWindow;

/*
	set_page

	This sets the page to the given page. Will save off the old page
	into extended memory using Int 15.

	On Entry    page = page # (always numbered from 0-n)
	On Exit     Returns TRUE if worked, FALSE if failed.
*/
bool set_page(unsigned nPage)
{
	#define ONE_MEG 0x100000L

	static unsigned nOldPage = INVALID_PAGE;
	static unsigned long lOldOffset;
	unsigned long lOffset;
	unsigned long lDstAddr;
	unsigned long lSrcAddr;

	/* validate parameter */
	ASSERT(nPage < nMaxPages);

	/* determine linear address from the page # */
	lOffset = _lmulu(PAGESIZE, nPage);

	/* save off old page data into extended memory */
	if(nPage != nOldPage)
	{
		/* copy data out to old page */
		if(nOldPage != INVALID_PAGE)
		{
			lDstAddr = ONE_MEG + lOldOffset;
			if(!CopyToInt15(lDstAddr, lpPagedDiskWindow, PAGESIZE/2))
				return FALSE;
		}

		/* copy in data from new page */
		if(nPage != INVALID_PAGE)
		{
			lSrcAddr = ONE_MEG + lOffset;
			if(!CopyFromInt15(lpPagedDiskWindow, lSrcAddr, PAGESIZE/2))
				return FALSE;
		}

		/* remember new page */
		nOldPage   = nPage;
		lOldOffset = lOffset;
	}

	/* worked */
	return TRUE;
}

/*
	init_page

	The Int15 initpage actually initializes a RAM disk. The size is
	limited to 16 Meg.

	On Entry    nothing
	On Exit     returns TRUE (non-zero) if worked, FALSE (0) if failed.
*/
bool init_page(void)
{
	/* determine size of extended memory RAM disk */
	nMaxPages = Int15MemSize();
	if(!nMaxPages)
		return FALSE;

	/* limit the disk size to 16 Meg */
	if(nMaxPages > 16*1024)
		nMaxPages = 16*1024;

	/* set up window pointer */
	lpPagedDiskWindow = (unsigned char far *)window;

	/* initialize RAM-DISK structures */
	if(!FormatRAMDisk(DEF_SEC_SIZE, 64,
			nMaxPages * (PAGESIZE / DEF_SEC_SIZE)))
		return FALSE;

	/* everything worked! */
	return TRUE;
}

/*
	uninit_page

	On Entry   nothing.
	On Exit    final page is flushed to extended memory (in case of
	           battery backed RAM disk).
*/
void uninit_page(void)
{
	/* flush the final page */
	set_page(INVALID_PAGE);
}
