#  $Id: MEMBC.MAK 1.7 2003/03/20 17:22:10Z thomd Exp $
#
#  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
#
#  Copyright (c) 1989-2003 Datalight, Inc.
#  All Rights Reserved
#
#  Datalight, Inc. is a Washington corporation located at:
#
#  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
#  Bothell, WA  98021                    Fax: +1.425.951.8094
#  USA                                   Http://www.datalight.com
#
#  This software, including without limitation all source code and
#  documentation, is the confidential, trade secret property of
#  Datalight, Inc., and is protected under the copyright laws of
#  the United States and other jurisdictions.  Portions of the
#  software may also be subject to one or more the following US
#  patents: US#5860082, US#6260156.
#
#  In addition to civil penalties for infringement of copyright
#  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
#  penalties for violation of (a) the restrictions on circumvention
#  of copyright protection systems found in 17 U.S.C. 1201 and
#  (b) the protections for the integrity of copyright management
#  information found in 17 U.S.C. 1202.
#
#  U.S. Government Restricted Rights.  Use, duplication,
#  reproduction, or transfer of this commercial product and
#  accompanying documentation is restricted in accordance with
#  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
#
#  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
#  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
#  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
#  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
#  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
#  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
#  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
#  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
#  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
#  THIS SOFTWARE IN ERROR.
#
# -------------------------------------------------------------------
#  Makefile for the Datalight Generic Memory Disk Drivers.
#  using Borland C/C++ 2.0 and later.
#
#  ROM-DOS Copyright (c) 1989-2002 Datalight, Inc.
#  Datalight Proprietary and Confidential.
#  All Rights Reserved.

#  Targets (.SYS and .LIB):
#  VDISK      Conventional, Extended, XMS and EMS RAM disk.
#  MEMFIXED   Fixed location conventional memory RAM disk.
#  MEMROM     Fixed location extended memory ROM disk.
#  MEMEXT     Fixed location extended memory RAM disk.
#  MEMPAGED   SC400 paged ROM disk.
#  INT15PAG   Extended memory paged RAM disk.
#  DYNDRVR    Dynamic Driver Loader (standalone .SYS only)
#  note: see DEVROM.ASM in \DEVSRC for a conventional memory ROM disk.

#  Helper Modules:
#  MEMDISK  ASM   DOS interface code, included in all drivers.
#  MEMUTIL  LIB   Library of ROM-DOS Kernel compatible routines.
#  LDRBEXT  ASM   BIOS extension to load ROM-DOS.SYS from a paged disk.
#  PAGEDLDR C     Load ROM-DOS.SYS from a paged disk.

#  Client Modules:
#  VDISK    C     Conventional, Extended and EMS disk (with static option)
#  MEMFIXED C     Fixed location conventional RAM disk.
#  MEMROM   C     Fixed location extended ROM disk.
#  MEMEXT   C     Fixed location extended RAM disk.
#  MEMPAGED C     Paged memory RAM/ROM disk (requires paging mechanism).
#  SC400PAG C     Paging mechanism for AMD Elan SC400.
#  INT15PAG C     Paging mechanism for Int 15H.
#
#  All drivers can be built-in to the ROM-DOS Kernel (.LIB target),
#  or act as stand-alone disk drivers that are loaded from CONFIG.SYS
#  (.SYS target). In addition, all drivers have a Unit Test (.EXE)
#  for testing changes when modifying the driver.
#
#  RD_SW, RD_SW_A, and RD_SW_L are environment variables that can be
#  set to include debugging information.
# -------------------------------------------------------------------

# BORLAND C compiler/linker options
# ---------------------------------
CC       = BCC
CFLAGS   = -c -w -O -Z $(RD_SW) -I..\devsrc
ASM      = TASM
AFLAGS   = /z /mx /zd $(RD_SW_A) /I..\devsrc
LINK     = TLINK
LFLAGS   = /s /m /c /l $(RD_SW_L)
LINKLIB  = ..\lib\dls.lib+CS.LIB
LIBR     = TLIB
LIBEND   =

# Assembly Options
# ---------------------------------
ABUILTIN  = /DBUILTIN=1       # built-in disk driver.
AROMDISK  = /DROMDISK=1       # rom-disk driver (read-only).
AREMOVABLE= /DREMOVABLE=1     # removable disk driver.
ATSRDEV   = /DTSRDEV=1        # stand-alone driver can be run TSR.
AGENIOCTL = /DGENIOCTL=1      # Generic IOCTL calls.
AUNITTEST = /DUNIT_TEST=1     # Unit test.

# C Options
# ---------------------------------
CBUILTIN  = -DBUILTIN=1       # built-in disk driver.
CROMDISK  = -DROMDISK=1       # rom-disk driver (read-only).
CREMOVABLE= -DREMOVABLE=1     # removable disk driver.
CTSRDEV   = -DTSRDEV=1        # stand-alone driver can be run TSR.
CGENIOCTL = -DGENIOCTL=1      # Generic IOCTL calls.
CUNITTEST = -DUNIT_TEST=1     # Unit test.


# -------------------------------------------------------------------
#                           MEMORY DISKS
# -------------------------------------------------------------------
release : VDISK.SYS MEMFIXED.SYS MEMROM.SYS MEMEXT.SYS MEMPAGED.SYS INT15PAG.SYS DYNDRVR.SYS
   @echo .
   @echo .  All Generic Memory Disk Drivers made successfully!
   @echo .


# -------------------------------------------------------------------
# Target for VDISK.SYS (standalone CONFIG.SYS driver)
# Conventional, Extended, XMS and EMS RAM disk.
# -------------------------------------------------------------------
vdisk.sys : memdisk.asm vdisk.c
   $(ASM) $(AFLAGS) $(AGENIOCTL) memdisk.asm;
   $(CC) $(CFLAGS) vdisk.c
   $(LINK) $(LFLAGS) memdisk+vdisk,vdisk.sys,,memutil;

# -------------------------------------------------------------------
# Target for VDISK.LIB (Build-in to ROM-DOS Kernel)
# Conventional, Extended, XMS and EMS RAM disk.
# -------------------------------------------------------------------
vdisk.lib : memdisk.asm vdisk.c
   $(ASM) $(AFLAGS) $(ABUILTIN) memdisk.asm;
   $(CC) $(CFLAGS) $(CBUILTIN) vdisk.c
   $(LIBR) memutil *memmath *memptr *membuff *memctype$(LIBEND)
   $(LIBR) memutil *raminit *dlxms *dlext15 *dlint15$(LIBEND)
	$(LIBR) memutil *dlint15w *dlint15x *dlint15y *dlint15z $(LIBEND)
   $(LIBR) memutil *emsf00 *emsf41 *emsf42 *emsf43 *emsf44$(LIBEND)
   $(LIBR) ..\user -+memdisk -+vdisk -+memmath -+memptr$(LIBEND)
   $(LIBR) ..\user -+membuff -+memctype -+raminit$(LIBEND)
   $(LIBR) ..\user -+dlxms -+dlext15 -+dlint15$(LIBEND)
	$(LIBR) ..\user -+dlint15w -+dlint15x -+dlint15y -+dlint15z$(LIBEND)
   $(LIBR) ..\user -+emsf00 -+emsf41 -+emsf42 -+emsf43 -+emsf44$(LIBEND)

# -------------------------------------------------------------------
# Target for MEMFIXED.SYS (standalone CONFIG.SYS driver)
# Fixed location conventional memory RAM disk.
# -------------------------------------------------------------------
memfixed.sys : memdisk.asm memfixed.c
   $(ASM) $(AFLAGS) memdisk.asm;
   $(CC) $(CFLAGS) memfixed.c
   $(LINK) $(LFLAGS) memdisk+memfixed,memfixed.sys,,memutil;

# -------------------------------------------------------------------
# Target for MEMFIXED.LIB (Build-in to ROM-DOS Kernel)
# Fixed location conventional memory RAM disk.
# -------------------------------------------------------------------
# makes a built-in fixed-location RAM-DISK driver.
memfixed.lib : memdisk.asm  memfixed.c
   $(ASM) $(AFLAGS) $(ABUILTIN) memdisk.asm;
   $(CC) $(CFLAGS) $(CBUILTIN) memfixed.c
   $(LIBR) memutil *memmath *memptr *memprhlp *membuff *memctype$(LIBEND)
   $(LIBR) memutil *memvalid *raminit$(LIBEND)
   $(LIBR) ..\user -+memdisk -+memfixed -+memmath -+memptr$(LIBEND)
   $(LIBR) ..\user -+memprhlp -+membuff -+memctype -+raminit$(LIBEND)
   $(LIBR) ..\user -+memvalid$(LIBEND)



# -------------------------------------------------------------------
# Target for MEMROM.SYS (standalone CONFIG.SYS driver)
# Fixed location extended memory ROM disk.
# -------------------------------------------------------------------
memrom.sys : memdisk.asm memrom.c
   $(ASM) $(AFLAGS) $(AROMDISK) memdisk.asm;
   $(CC) $(CFLAGS) $(CROMDISK) memrom.c
   $(LINK) $(LFLAGS) memdisk+memrom,memrom.sys,,memutil;

# -------------------------------------------------------------------
# Target for MEMROM.LIB (Build-in to ROM-DOS Kernel)
# Fixed location extended memory ROM disk.
# -------------------------------------------------------------------
memrom.lib : memdisk.asm memrom.c
   $(ASM) $(AFLAGS) $(AROMDISK) $(ABUILTIN) memdisk.asm;
   $(CC) $(CFLAGS) $(CROMDISK) $(CBUILTIN) memrom.c
   $(LIBR) memutil *membuff *memctype *memword *memvalid$(LIBEND)
   $(LIBR) memutil *dlext15 *dlint15$(LIBEND)
	$(LIBR) memutil *dlint15w *dlint15x *dlint15y *dlint15z $(LIBEND)
   $(LIBR) ..\user -+memdisk -+memrom$(LIBEND)
   $(LIBR) ..\user -+membuff -+memctype -+memword -+memvalid$(LIBEND)
   $(LIBR) ..\user -+dlext15 -+dlint15$(LIBEND)
	$(LIBR) ..\user -+dlint15w -+dlint15x -+dlint15y -+dlint15z$(LIBEND)


# -------------------------------------------------------------------
# Target for MEMEXT.SYS (standalone CONFIG.SYS driver)
# Fixed location extended memory RAM disk.
# -------------------------------------------------------------------
memext.sys : memdisk.asm  memext.c
   $(ASM) $(AFLAGS) memdisk.asm;
   $(CC) $(CFLAGS) memext.c
   $(LINK) $(LFLAGS) memdisk+memext,memext.sys,,memutil;

# -------------------------------------------------------------------
# Target for MEMEXT.LIB (Build-in to ROM-DOS Kernel)
# Fixed location extended memory RAM disk.
# -------------------------------------------------------------------
memext.lib : memdisk.asm memext.c
   $(ASM) $(AFLAGS) $(ABUILTIN) memdisk.asm;
   $(CC) $(CFLAGS) $(CBUILTIN) memext.c
   $(LIBR) memutil *membuff *memctype *memword *raminit$(LIBEND)
   $(LIBR) memutil *memvalid *dlext15 *dlint15$(LIBEND)
	$(LIBR) memutil *dlint15w *dlint15x *dlint15y *dlint15z $(LIBEND)
   $(LIBR) ..\user -+memdisk -+memext $(LIBEND)
   $(LIBR) ..\user -+membuff -+memctype -+memword -+raminit$(LIBEND)
   $(LIBR) ..\user -+memvalid -+dlext15 -+dlint15$(LIBEND)
	$(LIBR) ..\user -+dlint15w -+dlint15x -+dlint15y -+dlint15z$(LIBEND)

# -------------------------------------------------------------------
# Target for MEMPAGED.SYS (standalone CONFIG.SYS driver)
# SC400 paged ROM disk.
# -------------------------------------------------------------------
mempaged.sys : memdisk.asm mempaged.c sc400pag.c
   $(ASM) $(AFLAGS) $(AROMDISK) memdisk.asm;
   $(CC) $(CFLAGS) $(CROMDISK) mempaged.c
   $(CC) $(CFLAGS) $(CROMDISK) sc400pag.c
   $(LINK) $(LFLAGS) memdisk+mempaged+sc400pag,mempaged.sys,,memutil;

# -------------------------------------------------------------------
# Target for MEMPAGED.LIB (Build-in to ROM-DOS Kernel)
# SC400 paged ROM disk.
# -------------------------------------------------------------------
mempaged.lib : memdisk.asm mempaged.c sc400pag.c
   $(ASM) $(AFLAGS) $(ABUILTIN) $(AROMDISK) memdisk.asm;
   $(CC) $(CFLAGS) $(CBUILTIN) $(CROMDISK) mempaged.c
   $(CC) $(CFLAGS) $(CBUILTIN) $(CROMDISK) sc400pag.c
   $(LIBR) memutil *membuff *memio *memvalid *memmath$(LIBEND)
   $(LIBR) ..\user -+memdisk -+mempaged -+sc400pag$(LIBEND)
   $(LIBR) ..\user -+membuff -+memio -+memvalid -+memmath$(LIBEND)

# -------------------------------------------------------------------
# Target for MEMPAGED.IMG (Paged Loader BIOS Extension)
# SC400 paged ROM disk.
# -------------------------------------------------------------------
mempaged.img : ldrbext.asm pagedldr.c sc400pag.c
   $(ASM) $(AFLAGS) ldrbext.asm
   $(CC) $(CFLAGS) $(CROMDISK) pagedldr.c
   $(CC) $(CFLAGS) $(CROMDISK) sc400pag.c
   $(LINK) $(LFLAGS) ldrbext+pagedldr+sc400pag,pagedldr,,memutil;
   loc @pagedldr
   if exist mempaged.img del mempaged.img
   ren pagedldr.img mempaged.img


# -------------------------------------------------------------------
# Target for INT15PAG.SYS (standalone CONFIG.SYS driver)
# Extended memory paged RAM disk.
# -------------------------------------------------------------------
int15pag.sys : memdisk.asm mempaged.c int15pag.c
   $(ASM) $(AFLAGS) memdisk.asm;
   $(CC) $(CFLAGS) mempaged.c
   $(CC) $(CFLAGS) int15pag.c
   $(LINK) $(LFLAGS) memdisk+mempaged+int15pag,int15pag.sys,,memutil;

# -------------------------------------------------------------------
# Target for INT15PAG.LIB (Build-in to ROM-DOS Kernel)
# Extended memory paged RAM disk.
# -------------------------------------------------------------------
int15pag.lib : memdisk.asm mempaged.c int15pag.c
   $(ASM) $(AFLAGS) $(ABUILTIN) memdisk.asm;
   $(CC) $(CFLAGS) $(CBUILTIN) mempaged.c
   $(CC) $(CFLAGS) $(CBUILTIN) int15pag.c
   $(LIBR) memutil *membuff *memctype *memword *raminit$(LIBEND)
   $(LIBR) memutil *memvalid *memmath *dlext15 *dlint15$(LIBEND)
	$(LIBR) memutil *dlint15w *dlint15x *dlint15y *dlint15z $(LIBEND)
   $(LIBR) ..\user -+memdisk -+mempaged -+int15pag $(LIBEND)
   $(LIBR) ..\user -+membuff -+memctype -+memword -+raminit$(LIBEND)
   $(LIBR) ..\user -+memvalid -+memmath -+dlext15 -+dlint15$(LIBEND)
	$(LIBR) ..\user -+dlint15w -+dlint15x -+dlint15y -+dlint15z$(LIBEND)

# -------------------------------------------------------------------
# Target for DYNDRVR (standalong CONFIG.SYS driver)
# Dynamic Driver Loader
# -------------------------------------------------------------------
dyndrvr.sys : dyndrvr.obj
	$(ASM) $(AFLAGS) $(AROMDISK) memdisk.asm;
	$(LINK) $(LFLAGS) memdisk+dyndrvr,dyndrvr.sys,,memutil.lib/m;



# A quick way to "clean" out the directory
# ----------------------------------------
clean :
   if exist *.obj del *.obj
   if exist *.lst del *.lst
   if exist *.map del *.map
   if exist *.bak del *.bak
   if exist *.abm del *.abm
   if exist *.exe del *.exe
   if exist *.img del *.img
   if exist *.tr del *.tr
   if exist *.td del *.td
   if exist *.cv4 del *.cv4
   if exist *.pdb del *.pdb
   if exist *.sts del *.sts
   if exist *.log del *.log
   if exist x.y del x.y


# GENERIC RULES
# -----------------
.c.obj:
   $(CC) $(CFLAGS) $*.c

.asm.obj:
   $(ASM) $(AFLAGS) $*.asm;

.c.pt:
   proto $*.c >$*.pt

.asm.pt:
   proto $*.asm >$*.pt
