//
//   $Id: MEMRAMUT.C 1.5 2002/10/17 14:55:20Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:MEMRAMUT.C

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	Unit Test code for RAM disks

	$DOCHISTORY
	07/22/1998 by Drew Gislason
*********************************************************************/
#define __FILENUM__	220
#include <dos.h>
#include "compiler.h"
#include "memdisk.h"

char	buffer[512];
char	buffer2[512];
struct BPB	testBPB;
unsigned		endseg = 0x1000;
unsigned		memtop = 0xA000;
uchar memerr;
uchar memunit;


/*
	$DOCPROC:WordFill

	Like memset, but with words. Assumes length is in 2 byte
	increments.

	On Entry
		pBuffer		buffer to fill
		nVal			word value to fill with
		nBuffLen		length of buffer in bytes
	On Exit
		Buffer filled with words of "nVal"
*/
void WordFill(void *pBuffer, int nVal, unsigned nBuffLen)
{
	int * pPtr = pBuffer;

	/* validate parameters */
	ASSERT((nBuffLen %2) == 0);
	ASSERT(pBuffer);

	/* fill the buffer with a word value */
	nBuffLen /= 2;
	while(nBuffLen--)
		*pPtr++ = nVal;
}

/*
	$DOCPROC:WordCompare

	Compare the words in the buffer for the value.

	On Entry
		pBuffer		buffer to compare
		nVal			word value to compare with
		nBuffLen		length of buffer in bytes (must be even #)

	Returns TRUE if identical, FALSE if not.
*/
bool WordCompare(void *pBuffer, int nVal, unsigned nBuffLen)
{
	int * pPtr = pBuffer;

	/* validate parameters */
	ASSERT((nBuffLen %2) == 0);
	ASSERT(pBuffer);

	/* check each word in the buffer */
	nBuffLen /= 2;
	while(nBuffLen--)
		if(*pPtr++ != nVal)
			return FALSE;

	/* buffers are identical */
	return TRUE;
}

/*
	$DOCPROC:DosPutc

	Send output of memdisk printf() through DOS (rather than BIOS) so
	it can be piped to a file.
*/
void PASCAL DosPutc(uchar c)
{
	bdos(2, c, 0);
}

/*
	$DOCPROC:RamDiskUnitTest

	Write to and read from every sector in the disk. The disk size is
	defined by the szCmdline. Endseg may or may not be used (depending
	on the meminit function).
*/
bool RamDiskUnitTest(char szCmdline[])
{
	extern void (PASCAL *prfunc)(uchar c);
	unsigned i;

	/* send output through STDOUT so it can be piped into a file */
	prfunc = DosPutc;

	/* initialize the RAM disk */
	if(!meminit(&testBPB, szCmdline, &endseg, 4))
	{
		printf("Failed to initialize the disk!\n");
		return FALSE;
	}

	/* BPB must be valid */
	if(testBPB.TotalSecs == 0)
	{
		printf("Invalid BPB!\n");
		return FALSE;
	}

	/* write to each sector */
	for(i=1; i<testBPB.TotalSecs; ++i)
	{
		WordFill(buffer, i, testBPB.BytesPerSector);
		if(!memwrite(_lmulu(i,testBPB.BytesPerSector),
										testBPB.BytesPerSector, buffer))
		{
			printf("Failed to write at sector %d!\n",i);
			return FALSE;
		}
	}

	/* read each sector back */
	for(i=1; i<testBPB.TotalSecs; ++i)
	{
		WordFill(buffer2, i, testBPB.BytesPerSector);
		if(!memread(_lmulu(i,testBPB.BytesPerSector),
										testBPB.BytesPerSector, buffer))
		{
			printf("Failed to read at sector %d!\n",i);
			return FALSE;
		}

		/* remember, THIS memcmp() returns TRUE or FALSE */
		if(memcmp(buffer, buffer2, testBPB.BytesPerSector) != 0)
		{
			printf("Failed to compare at sector %d!\n",i);
			return FALSE;
		}
	}

	/* verify checksum */

	printf("Unit Tests passed.\n");
	return TRUE;
}

