//
//   $Id: NEWFDRV.C 1.5 2002/10/17 14:59:00Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:NEWFDRV.C

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	This module:
	1) Locate a labeled disk volume.
	2) Patch the drive letter of a specified NEWFILE path name
	   in the ROM-DOS CONFIG processing buffer.

	See NEWFEXAM.C for a simple application that uses these routines.

	$DOCHISTORY
	10-09-1997 de  Created by Dennis Edwards
*********************************************************************/
#include "newfdrv.h"

/* Set this to !0 to use the standard C library isspace(), isalpha()
	and strlen() functions.
*/
#define USE_C_RTL 0
#if (USE_C_RTL == 0)

/* Throw away stuff that normally comes from the C RTL which we
   want to avoid linking with for the demo device driver...
*/
#define isspace(c) (((c) == ' ') || (((c) >= '\x08') && ((c) <= '\x0D')))
#define isalpha(c) ((((c) | 0x20) >= 'a') && (((c) | 0x20) <= 'z'))

#endif



/*
   DOS2 FINDFIRST/FINDNEXT
   Description: Format of a findfirst/findnext placed in
   the user's memory pointed to by the DTA.

   Reference: Adv MS-DOS, p 445  -or-  DOS Ref, p 671
*/
typedef	unsigned char uchar;
typedef  unsigned long ulong;
struct FIND   /*()*/
{
   uchar      drive;       /* 00  1=A, 2=B, 3=C, etc     */
   char       emask[11];   /* 01  expanded name          */
   uchar      srchattr;    /* 0C  search attribute       */
   unsigned   dirnum;      /* 0D  directory entry number */
   unsigned   subdir;      /* 0F  cluster # of subdir    */
   uchar      res[4];      /* 11  reserved (always 0)    */
   uchar      attr;        /* 15  attribute found        */
   unsigned   time;        /* 16  file's time            */
   unsigned   date;        /* 18  file's date            */
   ulong      size;        /* 1A  file's size            */
   char       name[13];    /* 1E  filename in dot form   */
};
#define A_VOLUME		0x08



/*
	$DOCPROC:StringSearch

	searches upto size chars in buffer for the target string
   returns a pointer to target in buffer on success else NULL
   Case matters.
*/
char far *StringSearch(char far*lpBuffer, unsigned size, char *pTarget)
{
	char far *lpResult = 0;

	asm {
		les 	di, lpBuffer
		mov 	cx, size
	}
SCANNER:
	asm {
		mov 	si, pTarget
		cld
		lodsb
		repne	scasb
		jne	DONE
		push	di
	}
COMPARER:
	asm {
		cld
		lodsb
		or		al, al
		je		FOUND
		cld
		scasb
		je		COMPARER
		pop	di
		jmp	SCANNER
	}
FOUND:
	asm {
		pop	ax
		dec	ax
		mov	word ptr lpResult[2],es
		mov	word ptr lpResult, ax
	}
DONE:
	return lpResult;
}


/*
	$DOCPROC:FindDiskVolume

	Searches drives {2..lastDrive} looking for the volume label
 	specified in szDriveName. The case of the label is significant.

  	Returns the drive letter of the sought drive if successful else 0.
*/
char FindDiskVolume(char *szDriveName)
{
	int	i, j;
	int	blockDeviceCount;
	struct FIND found;
	static char anyName[] = "C:\*.*";

	// get the number of installed block devices out of DOS' "List of Lists"
	asm {
		mov	ah, 52h
		int	21h
		add 	bx, 20h
		xor	ax, ax
		mov	al, es:[bx]
		mov	blockDeviceCount, ax
	}

	//	set DTA to point to our local find-first block
	asm {
		mov 	ah, 1Ah
		lea 	dx, found
		int 	21h
	}

	/* Look for a disk containing the volume label specified in
	   the driveName string. Start with C: to avoid crit errors on
	   floppies with no diskettes
	*/
	for(i=2; i < blockDeviceCount; i++)
	{
		/* read the volume label on the specified drive, starting
		   with drive 'C'. Continue if volume has no label.
		*/
		asm {
			mov 	ah, 4Eh
			lea 	dx, anyName
			mov 	cx, A_VOLUME
			int 	21h
			jc	 	NOT_FOUND
		}

		/* see if the drive is the one we want, continue if not
		*/
		for(j=0; szDriveName[j] != 0; j++)
		{
			if(found.name[j] != szDriveName[j])
				goto NOT_FOUND;
		}

		/* If we got here then i is the sought drive number and anyName[0]
		   is its letter, return the drive letter.
		*/
		return *anyName;


		/* this drive doesn't have the right label, look at the next drive
		*/
NOT_FOUND:
		(*anyName)++;
	}

	// not found
	return 0;
}


/*
	$DOCPROC:PatchConfigBufferNEWFILEDrive

	Looks for the command "NEWFILE=<szNewFileName>" in the ROM-DOS CONFIG
   processing buffer. If the line is found, patches the NEWFILE drive to
   be newDriveLetter. The ASCIIZ string pointed to by szNewFileName must
   be of the form:

  			D:[\path\]filename.ext

   The case of all characters in szNewFileName must match those in the
   the CONFIG.SYS file currently being processed by DOS and all "NEWFILE"
   commands in the file must be in upper case.

   Returns FALSE (0) if the entry wasn't found in the buffer,
           TRUE (!0) if successful.
*/
int PatchConfigBufferNEWFILEDrive(char *szNewFileName, char newDriveLetter)
{
  int i, consideringEntry;
  static char szNewFileCommand[] = "NEWFILE";
  char far *lpConfigBuffer = 0, far *lpConfigBufferStart = 0,
       far *lpConfigText = 0;
  unsigned  configBuffSz = 0;

	/* Call a ROM-DOS "back-door" function to determine the size and
	   location of the CONFIG.SYS file contents as it is in memory.
	*/
	asm {
		mov	ax, 30DBh
		mov	si, 0B2D2h
		mov	di, 5
		int	21h
		jc		ALL_DONE
		cmp	ax, 0B2D2h
		jne	ALL_DONE
		mov	word ptr lpConfigBufferStart[2], bx
		mov	word ptr lpConfigBufferStart, 0
		mov	word ptr configBuffSz, cx
	}
  	lpConfigBuffer = lpConfigBufferStart;

	/* process the whole buffer if need be
	*/
	while(1)
	{
		/* look for a "NEWFILE" command in the CONFIG buffer, if we can't
		   find it then fail.
		*/
		lpConfigText = StringSearch(lpConfigBuffer,configBuffSz,szNewFileCommand);
		if(lpConfigText == 0L)
			break;

		/* now that we have found a "NEWFILE=" in the buffer, move our
		   pointer into the buffer past the end of the line. Adjust the
		   number of chars left in the buffer.
		*/
		lpConfigText += sizeof(szNewFileCommand)-1;
		lpConfigBuffer = lpConfigText;
		while(*lpConfigBuffer++ != '\n')
			{;}
		configBuffSz = (unsigned)((unsigned long)(lpConfigBuffer)
	                         	 -(unsigned long)(lpConfigBufferStart));

		/* compare the argument of the NEWFILE command to szNewFileName and
		   if it doesn't match (or if the arg does not start with a DOS drive
		   letter) then look for another NEWFILE entry.

		   If the strings do match, patch the drive letter at the beginning of
		   the buffered string and return TRUE to indicate succes.
		*/
		consideringEntry = 1;
		while((lpConfigText < lpConfigBuffer) && consideringEntry)
		{
			// skip equals and spaces
			if((*lpConfigText == '=') || isspace(*lpConfigText))
			{
				++lpConfigText;
				continue;
			}

			// verify NEWFILE argument starts with a drive letter.
			if(!isalpha(*lpConfigText) || (lpConfigText[1] != ':'))
			{
				consideringEntry = 0;
			}
			else
			{
				i = 0;
				while(lpConfigText[i] != '\r')
				{
					if(lpConfigText[i] != szNewFileName[i])
					{
						consideringEntry = 0;
						break;
					}
					if(szNewFileName[++i] == 0)
					{
						*lpConfigText = newDriveLetter;
						return 1;
					}
				}
			}
		}
	}
ALL_DONE:
	return 0;
}

#if UNIT_TEST
#include <stdio.h>

int main(void)
{
	char cDriveLetter;

	cDriveLetter = FindDiskVolume("DL RAM D.ISK");
	if(!cDriveLetter)
		printf("Failed to find VDISK!\n");
	else
		printf("VDISK on drive %c\n",cDriveLetter);

	return 0;
}

#endif
