//  $Id: RAMINIT.C 1.7 2003/03/20 16:52:30Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1997-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/*******************************************************************
	$DOCMODULE:RAMINIT.C

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This module contains the code necessary to initialize a RAM
	disk. It uses memwrite() to actually do the work, and so should
	be compatible with any Datalight Generic Disk Driver.

	$DOCHISTORY
	06/25/1998 dg  Added GetRAMDiskBPB()
	11/27/1997 dg  Created by Drew Gislason
*******************************************************************/
#define __FILENUM__ 201
#include "compiler.h"
#include "memdisk.h"
#include "vdisk.h"

extern bool bStatic;

#define SectorOffset(x) (_lmulu(x, BytesPerSector))

/* volume label, with fixed date/time for simplicity */
struct DIR DirLabel =
{
	{ 'D','L',' ','R','A','M',' ','D' },	/* File Name */
	{ 'I','S','K' },								/* Extension */
	0x08,												/* Attribute */
	{ 0 },											/* Reserved  */
	0x6001,											/* Time 12:00pm */
	0x248E											/* Date 4-14-1998 */
};

/* initialization data */
unsigned BootID = 0xAA55;
char InitFAT[4] = { VDISK_ID, 0xFF, 0xFF, 0xFF };

/*
	$DOCPROC:FormatRAMDisk

	Create the system structures for the RAM disk by writing to the
	media. Uses memwrite() to create the structures of the RAM disk.

	On Entry
		iBytesPerSector   may be 128, 256 or 512
		iDirEntries       legal values are 4 - 2048
		nNumSectors       legal values are 16 - 65535
	On Exit
		The media is filled with a freshly formatted disk.
		Returns TRUE (non-zero) if worked, FALSE (0) if failed.
*/
bool FormatRAMDisk(int iBytesPerSector, int iDirEntries, unsigned nNumSectors)
{
	int iDirsPerSector;
	int iFatBytes;
	unsigned nNumClusters;
	unsigned nSector;
	unsigned long lOffset;
	unsigned nFatStart;
	unsigned nDirStart;
	struct BPB bpb;
	struct BPB bpb2;
	extern struct BPB std_bpb;

	memset( &bpb, 0, sizeof(struct BPB) );
	memset( &bpb2, 0, sizeof(struct BPB) );

	/* validate iBytesPerSector */
	if(iBytesPerSector != 128 && iBytesPerSector != 256 &&
		iBytesPerSector != 512)
		return FALSE;

	/* validate iDirEntries (rounds up if not exact fit) */
	iDirsPerSector = iBytesPerSector / sizeof(struct DIR);
	if(iDirEntries < iDirsPerSector)	/* min */
		iDirEntries = iDirsPerSector;
	if(iDirEntries > 1024)				/* max */
		iDirEntries = 1024;
	if(iDirEntries % iDirsPerSector)	/* exact fit in sectors */
		iDirEntries += iDirsPerSector - (iDirEntries % iDirsPerSector);

	/* start setting up the BPB */
	bpb.BytesPerSector = iBytesPerSector;
	bpb.MaxDirs = iDirEntries;
	bpb.TotalSecs = nNumSectors;

	/* use the old vdisk formula */
	bpb.SecPerCluster = 1;
	nNumClusters = nNumSectors;
	while(nNumClusters > 0xFF0)
	{
		bpb.SecPerCluster *= 2;
		nNumClusters = nNumSectors / bpb.SecPerCluster;
	}

	/* determine size of FAT */
	if(nNumClusters > 0xFF0)
	{
		iFatBytes = nNumClusters * 2;
		InitFAT[3] = 0xFF;	/* 4 byte ID for 16-bit FATs */
	}
	else
	{
		iFatBytes = (nNumClusters * 3) / 2;
		InitFAT[3] = 0;		/* 3 byte ID for 12-bit FATs */
	}
	bpb.SecPerFAT = (iFatBytes + (iBytesPerSector-1)) / iBytesPerSector;

	/* set the hard-coded fields */
	bpb.Reserved   = 1;
	bpb.NumbFats   = 2;
	bpb.MediaDescrip = VDISK_ID;
	bpb.SecPerTrak = 16;
	bpb.Sides      = 2;
	bpb.Hidden     = bpb.LNumSec = 0;

	//  Determine if the user is using a static ram drive
	//
	if( bStatic )
	{
		//  Make sure that the size of the ram drive that may
		//  be present contains the same BPB that we just built.
		//  if not, we must reinitialize the ram drive
		//
		bStatic = GetDiskBPB( &bpb2 );
		if( bStatic )
		{
			if( IsSameBPB(&bpb, &bpb2) )
			{
			}
			else
			{
				bStatic = FALSE;
			}
		}
	}


	/* initialize boot sector (write out BPB, and boot ID) */
	memset(aMemBuff, 0, iBytesPerSector);
	memmove(aMemBuff, DiskIntro, sizeof(DiskIntro));
	memmove(aMemBuff+START_BPB, &bpb, sizeof(struct BPB));
	*(unsigned *)(aMemBuff+iBytesPerSector-2) = BootID;

	//  If this is not a static ram drive then initialize (format) the
	//  ram drive.
	//
	if( !bStatic )
	{
		if(!memwrite(0L, iBytesPerSector, aMemBuff))
			return FALSE;

		/* todo: volume label should have current date/time */

		/* initialize through the end of the root directory */
		nFatStart   = bpb.Reserved;
		nDirStart   = nFatStart + bpb.NumbFats * bpb.SecPerFAT;
		nNumSectors = nDirStart + iDirEntries / iDirsPerSector;
		lOffset = iBytesPerSector;
		for(nSector=1; nSector<nNumSectors; ++nSector)
		{
			/* fill sector with 0s */
			memset(aMemBuff, 0, iBytesPerSector);

			/* root directory gets volume label */
			if(nSector == nDirStart)
				memmove(aMemBuff, &DirLabel, sizeof(DirLabel));

			/* 1st sector of each FAT gets FAT id */
			else if(nSector >= nFatStart && nSector < nDirStart)
			{
				if(((nSector - nFatStart) % bpb.SecPerFAT) == 0)
					memmove(aMemBuff, InitFAT, sizeof(InitFAT));
			}

			/* write the sector out */
			if(!memwrite(lOffset, iBytesPerSector, aMemBuff))
				return FALSE;

			/* on to next sector */
			lOffset += iBytesPerSector;
		}

		/* todo: update boot sector with checksum */
	}

	/* everything worked! */
	return TRUE;
}
