;
;   $Id: BIOSINIT.INC 1.4 2002/10/15 23:49:14Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
COMMENT ~ ***********************************************************

  BIOSINIT.INC  revision 3.1

  Copyright (C) 1994-2002 Datalight, Inc.
  All rights reserved
  Datalight Confidential and Proprietary

  LOW-LEVEL HARDWARE INIT ROUTINES FOR DATALIGHT'S generic miniBIOS
  It is intended that this file should contain all low-level system
  hardware initialization routines.  These routines can be modified
  when targeting the miniBIOS to a new hardware platform but should
  follow the framework layed out here.  This file is included by name
  within minibios.asm.

  This file contains all routines to initialize the system hardware
  the tasks to be completed here include:

        1) Initialize chip selects for ROM and RAM

        2) initialize DRAM and DRAM refresh systems if they are present

        3) initialize timer system (18.2 ticks/second)

        4) set up any other custom hardware


******************************************************************** ~

;-------------------------------------------------------------------
; EQUATES:
TARGET_NAME   equ       "Generic miniBIOS "   ;insert specific target name here


comment &
        The equate "dlytime" should be adjusted for the target processor to
        give approximately a 7 ms delay when used in the following loop:

        mov     cx,dlytime
dly_loop:
        loop    dly_loop

        A 25 MHz 386 requires a value of about 13000.
&
dlytime equ     13146

;The equates below are for a standard PC compatible system, adjust as required
;to fit the target platform.

; ----- DMA Controller ports -----

        dma     equ     00              ; base addr for reg-set (0 - F)

; ----- Timer ports (PIT) -----

        timer_0 equ     040h
        timer_1 equ     041h
        timer_2 equ     042h
        pit_cmd equ     043h

; ----- Interrupt Controller ports (PIC) -----

        eoi_cmd         equ     20h     ; eoi command to 8259 (20h)
        icmd_reg        equ     20h     ; int ctlr command reg addr (20h)
        imsk_reg        equ     21h     ; int ctlr mask reg         (21h)
        icmd_reg_slv    equ     0A0h    ; slv int ctlr command reg addr (A0h)
        imsk_reg_slv    equ     0A1h    ; slv int ctlr mask reg         (A1h)
;
; *** define interrupt mask (bit) for the BIOS tick timer source ***
; *** (for miniBIOS use only)
; >>>> CHANGE THIS TO ZERO TO DISABLE THE INTERNAL TIMER INTERRUPT <<<<

        TMRintmask      equ     0001h     ; std XT-config timer interrupt mask

        Console_Baud_Divisor    equ     COM_PORT_DIVIDEND/Console_Baud_Rate

if Console_Port EQ 1
        Con_Port        equ     3f8h
endif

if Console_Port EQ 2
        Con_Port        equ     2f8h
endif

if Console_Port EQ 3
        Con_Port        equ     3e8h
endif

if Console_Port EQ 4
        Con_Port        equ     2e8h
endif

if SupportPostCodes

Post_Init macro
;  Send POST codes to COM port specified as the console in minibios.inc

;       outportb(LINECONTROL,0x80);     /* set high bit, so we can set baud */
        mov     dx,Con_Port+3
        mov     al,80h
        out     dx,al
;       outportb(BAUDLOW,baud & 0xFF);          /* Set baud rate */
.386
;calculating constant below requires 386 directive even though answer is < 64K!
        mov     ax,Console_Baud_Divisor
.8086
        sub     dl,3                    ;make Con_Port
        out     dx,al
;       outportb(BAUDHIGH,baud >> 8);
        inc     dx              ;make Con_Port+1
        mov     al,ah           ; High byte of divisor
        out     dx,al
;       outportb(LINECONTROL,parms);    /* stop/data/parity bits */
        mov     al,Console_Parms
        add     dl,2                    ;make Con_Port+3
        out     dx,al
        endm

Post_Print macro  x
        local PP1
        local PP2
PP1:
        mov     dx,Con_Port+5
        in      al,dx
        test    al,20h
        jz      PP1               ;/* wait for trans buffer to be empty */

        mov     al,'&x'         ; print a 'x' character
        sub     dl,5            ;make Con_Port
        out     dx,al
        add     dl,5            ;make Con_Port + 5
PP2:
        in      al,dx
        test    al,20h
        jz      PP2
        endm
else    ; DO NOT SUPPORT_POST_CODES

Post_Init macro         ; Define NULL Macro
        endm

Post_Print macro  x     ; Define NULL Macro
        endm

endif   ; SUPPORT_POST CODES


; ** Define the PIT divisor for timer0 to generate 18.21 tic/sec rate **
        TICK_DIV        equ     65535   ; pit timer-0 divisor for clock


; ---------------------------------------------------------------------------
; >>>                     RESET VECTOR JUMP TO BIOS                       <<<
; ---------------------------------------------------------------------------
; DOES THE INITIAL UPPER MEM CHIP SELECT IF NECESSARY THEN JUMPS TO COLD_BOOT
; THIS SEGMENT IS LOCATED TO 0XFFFF

boot    segment para    public  'BOOT'
        assume cs:boot
reset:
.186
        cli

        ;PREPARE ROM FOR JUMP TO cold_boot IF REQUIRED.

        jmp far ptr cold_boot

        db BIOS_DATE

boot    ends

CbegCode
        db BIOS_DATE
        db TARGET_NAME

        public _astart
_astart:
cold_boot:
        Post_Init       ; Initialize the PostCode port
        Post_Print A    ; Let the user know the system is alive

;------------------------------------------------------------------------------
; >>>                    INITIALIZE SYSTEM DRAM                             <<<
;------------------------------------------------------------------------------
; PERFORM ALL TASKS NECESSARY TO INITIALIZE THE DRAM IF THERE IS ANY.  THIS
; INCLUDES THINGS LIKE CHIP SELECTS, AND INITIALIZATION OF REFRESH. THE DRAM
; SHOULD BE RUNNING AT THE END OF THIS SECTION


if TARGET_CPU GE 386
;Initialize CR0 as dictated by the equipment list
.386p
        mov     eax,cr0
        and     al,0f0h                 ;clear PE, TS, MP and EM bits
if (OriginalEquipmentList AND 2)        ;if 80387 is on the equipment list
        or      al,2                    ;set MP if coprocessor is present
endif
        mov     cr0,eax
.186
endif

;------------------------------------------------------------------------------
; >>>                    INITIALIZE REMAINING HARDWARE                      <<<
;------------------------------------------------------------------------------
; Initialize DMA controllers, interrupt contollers, etc.


comment &
        The code below assumes a standard 8259 compatible interrupt controller
        with a slave.  It is assumed to be AT compatible.  Adjust as required
        for the target platform.
&
; configure MASTER 8259 int controller
; for edge-triggered interrupts

        cli
        mov al,11h              ; icw1 (init cmd)- Dual PIC system
        out icmd_reg,al
        mov al,08h              ; icw2- int type 8
        out imsk_reg,al
        mov al,04h              ; icw3- Master/Slave connections. Setting a bit
;    indicates that a slave 82C59A is attached to corresponding IR signal.
;    High order interrupts set to 70h.   04h is proper for cascade interrupt.
;    slave is cascaded from the Master's IR2 signal, write 04H to this register

        out imsk_reg,al
        mov al,01h              ; icw4- norm nest mode, buf'd slave,
        out imsk_reg,al         ; normal EOI, 8088 mode
        mov al,0ffh             ; mask ALL ints
        out imsk_reg,al

        mov ax,0820h            ; issue non-specific eoi's in case of
clrints:                        ; any incidental requests pending
        out icmd_reg,al
        dec ah
        jnz clrints

; ---------------------------------------------
; configure SLAVE 8259 int controller address 0A0h port
; for edge-triggered interrupts

        mov al,11h              ; icw1 (init cmd)- Dual PIC system
        out icmd_reg_slv,al
        mov al,70h              ; icw2- int type 70
        out imsk_reg_slv,al
        mov al,02h              ; icw3- Slave connections. Since the internal
;    slave is cascaded from the Master's IR2 signal, write 02H to this register

        out imsk_reg_slv,al
        mov al,01h              ; icw4- norm nest mode, buf'd slave,
        out imsk_reg_slv,al         ; normal EOI, 8088 mode
        mov al,0ffh             ; mask ALL ints
        out imsk_reg_slv,al

        mov ax,0820h            ; issue non-specific eoi's in case of
clrints_slv:                    ; any incidental requests pending
        out icmd_reg_slv,al
        dec ah
        jnz clrints_slv

;eoi the master PIC
        mov     al,eoi_cmd
        out     icmd_reg,al

;initialize the DMA controllers if any.

        mov     al,04           ; disable dma controller, all channels
        out     dma+8,al

;------------------------------------------------------------------------------
; >>>                    JUMP BACK TO MINIBIOS.ASM                          <<<
;------------------------------------------------------------------------------

        Post_Print H    ; Post_Prints for diagnostics
        jmp     ENTRY

comment &
        The bios_vec_table contains a list of interrupt numbers and the
        corresponding interrupt service routines.  Code in minibios.asm will
        use this table to initialize the desired BIOS services as specified in
        the table.  To remove the services and the corresponding code
        associated with a particular BIOS interrupt, simply comment out the
        three lines associated with that interrupt.  Be certain to comment out
        all three lines, all data in the following table must be in pairs, a
        interrupt number specified as a byte value followed by the offset of
        the corresponding ISR specified as a word value.  If the extrn
        statement is not commented out, the linker will pull in the associated
        object module even though it is not actually used.  If you want to
        remove support for an interrupt, check the table, bios_init_funcs to
        see if that interrupt also has an initialization routine.  If so, it
        should be commented out also.
&
bios_vec_table:

extrn _INT07_isr:near          ;used for handling esc instructions
db      7
dw      offset _INT07_isr

extrn _INT08_isr:near          ;BIOS timer tick, requires int 1Ch
db      8
dw      offset _INT08_isr

;Required for ROM-DOS, see comments in int16.asm regarding int 9.
db      9
dw      offset _INT09_isr

extrn _INT10_isr:near           ;stdout via serial port, Required for ROM-DOS
db      10h
dw      offset _INT10_isr

; Required by ROM-DOS, get equipment configuration
db      11h
dw      offset _INT11_isr

; Required for ROM-DOS, get memory size
db      12h
dw      offset _INT12_isr

; Required by ROM-DOS, fail disk I/O to indicate no physical disks on system.
extrn _INT13_isr:near
db      13h
dw      offset _INT13_isr

extrn _INT14_isr:near           ;Optional COM port support.
db      14h
dw      offset _INT14_isr

extrn _INT15_isr:near           ;Required by ROM-DOS, get system model table
db      15h
dw      offset _INT15_isr

extrn _INT16_isr:near           ;Required by ROM0DOS, Keyboard driver
db      16h
dw      offset _INT16_isr

extrn _INT18_isr:near           ;handler in case ROM-DOS is not found
db      18h
dw      offset _INT18_isr

extrn _INT19_isr:near           ;Normally replaced by ROM-DOS
db      19h
dw      offset _INT19_isr

extrn _INT1A_isr:near           ;BIOS time of day
db      1Ah
dw      offset _INT1A_isr

;default BIOS timer tick handler, required if int 08 is enabled.
db      1Ch
dw      offset _INT1C_isr

;don't move this label, it marks the end of the table!
bios_vec_table_end      label byte

comment &
        The table below contains the addresses of several initialization
        functions.  These are mostly associated with initializing various
        optional BIOS isrs.  If the service is not required, comment out the
        lines associated with that interrupt in the table below.  Also be
        certain to comment out the corresponding lines in the bios_vec_table
        above.
&
bios_init_funcs:

;Function below is in this file, Optional timer tick support
dw      offset _Int08_Init

extrn _console_init:near        ;Console services are required by miniBIOS
dw      offset _console_init    ;associated with int 10h, stdout via serial port

extrn _Int14_Init:near
dw      offset _Int14_Init      ;Option BIOS COM port support.

;Don't move the label below, it marks the end of the table!
bios_init_funcs_end     label byte

;------------------------------------------------------------------------------
;>>>         PLACE ANY HARDWARE ACCESSING SUBFUNCTIONS HERE                 <<<
;------------------------------------------------------------------------------

c_extrn  console_putchar,proc   ;must be defined for minibios.asm

ifdef _INT08_isr        ;This is needed only if we are using int 08
_Int08_Init:
;------------------------------------------------------------------------------
; >>>                    INITIALIZE 18.2 HZ TIMER                           <<<
;------------------------------------------------------------------------------
;CONFIGURE TIMER HARDWARE TO GENERATE 18.2 HZ INTERRUPT

;  Assumes PC compatible 8253 timer chip
;  This code configures the timer that calls interrupt 08H to produce
;  the time of day clock.

        ; fire up pit timer 0 for time of day clk
        mov al,34h              ; rate mode =2, 2 byte cnt val
        out PIT_CMD,al
        mov ax,TICK_DIV         ; appx. 18.21 tics/sec
        out TIMER_0,al
        mov al,ah
        out TIMER_0,al
; tell int ctlr to enable int 8
        in      al,imsk_reg             ; 0x21 on PC compatibles
        and     al,NOT TMRintmask
        out     imsk_reg,al             ; 8 bit port
        ret

public int8_eoi
int8_eoi:       ;called by int08_isr, used to eoi the interrupt controller
        mov     al,eoi_cmd
        out     icmd_reg,al
        ret
endif

