//  $Id: CMDMAIN.C 1.5 2003/03/20 16:59:48Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1992-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/*********************************************************

   CMDMAIN.C  created October, 1992

   Copyright (c) 1992-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This file has the main function and the command loop the
   spawn routines and other utility routines.

**********************************************************/
#include "minicmd.h"


struct internal command_table[] =
{
	{"", do_disk_change},
#if CMD_CD
	{"CD", do_chdir},
#endif
#if CMD_COPY
	{"COPY", do_copy},
#endif
#if CMD_DEL
	{"DEL", do_del},
#endif
#if CMD_DIR
	{"DIR", do_dir},
#endif
#if CMD_ECHO
	{"ECHO", do_echo},
#endif
#if CMD_EXIT
	{"EXIT", do_exit},
#endif
#if CMD_GOTO
	{"GOTO", do_goto},
#endif
#if CMD_IF
	{"IF", do_if},
#endif
#if CMD_PAUSE
	{"PAUSE", do_pause},
#endif
#if CMD_REM
	{"REM", do_nothing},
#endif
#if CMD_SET
	{"SET", do_set},
#endif
#if CMD_TYPE
	{"TYPE", do_type},
#endif
#if CMD_VER
	{"VER", do_ver},
#endif
};


char *error_messages[] =
{
	"Duplicate file name or File not found",	/* 0 */
	"Not enough memory",
	"Bad command or file name",
	"File not found",
	"Invalid drive",
	"Program too large",						/* 5 */
	"Out of environment space",
	"Cannot open file",
	"Invalid directory"
};

int exit_allowed = TRUE;
char *AUTOEXEC = "A:\\AUTOEXEC.BAT";
struct FFBLK ffblk;

struct exec exec_block;
extern char command[82];

/*
 * main
 *
 * This is the C entry point
 *
 * Return
 * 	None
 */
void main(char far *args)
{
	char c;
	while(isspace(*args))
		++args;

	if(*args == '/')
	{
		c = *(++args);

		/* make MINICMD.COM permanent, which means process AUTOEXEC */
		if(toupper(c) == 'P')
		{
				exit_allowed = FALSE;
				psp->parent = FP_SEG(psp);
				if(findfirst(AUTOEXEC+2, &ffblk, 0xFF) == 0)
				{
					// patch the constant AUTOEXEC.BAT name string with
					// the current drive letter: "D:\AUTOEXEC.BAT"
					// Note that get_current_drive() returns {A=1,2,...}
					// NOT {A=0,1,...}
					//
					*AUTOEXEC = '@' + get_current_drive();
					strcpy(Batch_File, AUTOEXEC);
					In_Batch = TRUE;
				}
		}
	}

	main_loop();
}

/*
 * main_loop
 *
 * This is the get a command execute a command loop. It does not exit.
 * This may be interrupted via a Ctrl-C or a Critical error.
 *
 * Return
 * 	None
 */
void main_loop(void)
{
	while(1)
	{
		do_prompt(0);
		get_string();
		do_one_command();
#if OPT_OUT_PIPE
		unredirect();
#endif
	}
}


char cmd_name[128];
char cmd_args[128];

/*
 * do_one_command
 *
 * This functions executes one command as found in ..
 * The command may be from the console or a Batch file.
 *
 * Return
 *	None
 */

void do_one_command(void)
{
	int i;
	char name[9];

	strcpy(cmd_name, cmd);
	cmd_name[args-cmd - 1] = '\0';
	strcpy(cmd_args + 1, args);

	if(fcb0.drive != 0 || !memcmp(fcb0.name, "         ", 8))
		putc('\n');

	/* check for internal commands */
	if(strchr(cmd_name, '\\') == NULL)
	{
		for(i=0; i<(sizeof(command_table)/sizeof(command_table[1])); ++i)
		{
			parse_fcb(command_table[i].cmd, &internal_fcb);
			if(memcmp(fcb0.name, internal_fcb.name, 8))
			{
				if(command_table[i].func != NULL)
					(*command_table[i].func)(cmd_args + 1);
				return;
			}
		}
	}

	/* not internal, maybe external? */
	do_exe();
}


/*
 * memcmp
 *
 * This function is a standard library function memory compare fufnction
 * Return
 *	True of memory is same
 *	False if memory differes at any byte
 */

int memcmp(char *s1, char *s2, unsigned count)
{
	int i;
	for(i=0; i<count; ++i)
	{
		if(s1[i] != s2[i])
			return FALSE;
	}
	return TRUE;
}

/*
 * do_nothing
 *
 * This function is a null function
 * Return
 *	None
 */
#pragma argsused
void do_nothing(char *parms)
{
}


/*
 * do_disk_change
 *
 * Change drives, ie A:, B:, C:
 *
 * Returns
 *    None
 */
#pragma argsused
void do_disk_change(char *parms)
{
	args = parse_fcb(command+2, &fcb0);
	if(fcb0.drive != 0)
	{
		set_current_drive(fcb0.drive);
		if(fcb0.drive != get_current_drive())
			error(4);
	}
	else
		;			/* Blank command */
}


/*
 * do_prompt
 *
 * This function displays the miniCOMMAND Prompt.
 * The prompt will either be a hard coded string or a
 * standard $p$g prompt, depending on options.
 *
 * Return
 * 	None
 */
#pragma argsused
void do_prompt(char *parms)
{
#if OPT_FULL_PROMPT
	char path[66];
	int drive;

	if(!Echo)
		return;
	drive = get_current_drive();
	path[0] = 'A' + drive - 1;
	path[1] = ':';
	path[2] = '\\';
	get_cwd(drive, path+3);
	strcat(path, ">");
	puts(path);
#else
	static char * prompt = PROMPT;
	putc('\n');
	if(!Echo)
		return;
	prompt[0] = get_current_drive() + 'A' - 1;
	puts(prompt);
#endif
}


char path_exe_name[128];

extern struct FFBLK ffblk;

/*
 * do_exe
 *
 * This function will execute an external command(Program).
 * It searches current directory and the PATH for the program.
 * do_exe displays error messages as it goes.
 *
 * Return
 * 	None
 */
void do_exe(void)
{
	int exit_code;
	char far *path;
	char *s1;

	/* Prep args for Spawn */
	strcat(cmd_args + 1, "\r");
	cmd_args[0] = strlen(cmd_args + 1) -1;

	exit_code = do_exe_extension(cmd_name);

	if(exit_code == ERR_NOTFOUND)
	{
		path = get_env("PATH=");
		if(*path == '=')
			++path;
		while(*path)
		{
			s1 = path_exe_name;
			while(*path != ';' && *path)
				*s1++ = *path++;
			*s1 = '\0';
			strcat(path_exe_name, "\\");
			strcat(path_exe_name, cmd_name);
			exit_code = do_exe_extension(path_exe_name);
			if(exit_code == 0)
				break;
			if(*path == '\0')
				break;
			++path;
		}
	}
	if(exit_code == ERR_NOTFOUND)
		error(2);
	if(exit_code == ERR_NOMEM)
		error(1);
}

/*
 * do_exe_extension
 *
 * Given a path, this function will try to execute a COM,
 * then EXE then BAT file in that path.  This function is
 * a utility of function do_exe.
 *
 * Return
 * 	The return Code of the program or batch file.
 */
int do_exe_extension(char *cmd)
{
	char *ext;
	int exit_code;

	ext = strchr(cmd, '.');
	if(ext == NULL)
		ext = cmd+ strlen(cmd);

	strcpy(ext, ".EXE");
	exit_code = exec(cmd, cmd_args);
	if(exit_code != ERR_NOTFOUND)
		return (exit_code);

	strcpy(ext, ".COM");
	exit_code = exec(cmd, cmd_args);

	if(exit_code != ERR_NOTFOUND)
		return(exit_code);

	strcpy(ext, ".BAT");
	if(findfirst(cmd, &ffblk, 0x20) == 0)
	{
		do_batch(cmd);
		exit_code = 0;
	}

	return (exit_code);
}



char Batch_File[66];
int  In_Batch = FALSE;
long Batch_pos = 0;

char command[82];

struct fcb fcb0;
struct fcb internal_fcb;
int Got_Cntl_C = 0;

char *args, *cmd;

/*
 * get_string
 *
 * This function gets a string from either a Batch File
 * or the Console.  It understands piping.
 *
 * Return
 * 	None
 */
void get_string(void)
{
	/* set max size of input string: size of buff - 2 for max&len byte */
	command[0] = sizeof(command) - 2;

	if(In_Batch)
	{
		/* zero out old command so it can't appear as args to new one */
		memset( command+1, 0, sizeof(command) - 1 );

		if(Got_Cntl_C)
		{
			Got_Cntl_C = FALSE;
			while(1)
			{
				puts("\nTerminate Batch job(Y/N) ");
				switch(getc(1))
				{
					case 'Y':
					case 'y':
						In_Batch = FALSE;
						return;


					case 'N':
					case 'n':
						return;

				}
			}
		}
		get_batch_file_line(command);
		if(Echo)
			puts(command+2);
	}
	else
	{
		command[command[1]+2] = '\r';
		get_dos_bufferd_input(command);
	}

	make_cmd();
}

void make_cmd(void)
{
	char *p;

	/* null terminate the string */
	memset(command + command[1] + 2, 0, sizeof(command) - command[1]-2);

	cmd = command+2;

#if OPT_OUT_PIPE
	redirect(cmd);
#endif

	args = parse_fcb(command+2, &fcb0);
	cmd  = skipwhite(cmd);
	for(p=cmd; *p && !isspace(*p); ++p)
		;
	if(args < p)
		args = p;
	++args;
}

#if OPT_OUT_PIPE

#define OUT_PIPE_CHAR '>'
#define STDIN  0
#define STDOUT 1
#define STDERR 2

int  redir_out;
char fname[66];

void unredirect(void)
{
	unsigned char far * hptr;
	int fd;
	if(redir_out)
	{
		dosclose(STDOUT);
		hptr = psp->htbl_addr;
		hptr[STDOUT] = hptr[STDERR];
		redir_out = 0;
	}
}

/*
 *  Possibly redirect the output.  Currently supports ">"
 *  (create) and ">>" (append).
 */
void redirect(char *cmd)
{
	int fd, i;
	unsigned char far * hptr;
	char * p, * p2;
	int append = 0;
	int yes_redir = 0;

	/* pipe to a file? */
	for(p=cmd; *p; ++p)
	{
		if(*p == OUT_PIPE_CHAR)
		{
			p2 = p+1;
			if(*p2 == OUT_PIPE_CHAR)
				append = 1, ++p2;
			p2 = skipwhite(p2);
			for(i=0; *p2 && !isspace(*p2) && i+1<sizeof(fname); ++i)
				fname[i] = *p2++;
			fname[i] = 0;

			/* remove the name from the cmd string */
			do { *p++ = *p2++; } while(*(p2-1));

			yes_redir = 1;
			break;
		}
	}

	/* not found */
	if(!yes_redir)
		return;

	fd = -1;
	if(append)
	{
		fd = dosopen(fname, 1);		/* open for writing */
		if(fd >= 0)
			doslseek(fd, 0L, 2);
	}
	if(fd < 0)
		fd = doscreate(fname);

	/* did it work?  were we able to open the file? */
	if(fd < 0)
	{
		puts("\nFile creation error\n");
		*cmd = 0;
	}
	else
	{
		/* force dup manually (some device drivers require it) */
		hptr = psp->htbl_addr;
		hptr[STDOUT] = hptr[fd];
		hptr[fd] = 0xFF;
		redir_out = 1;
	}
}
#endif


/*
 * get_batch_file_line
 *
 * This function gets a single line from a Batch File.
 * It is a utility function for get_string.  It will
 * ignore any lines that are :labels.
 *
 * Return
 * 	None
 */
void get_batch_file_line(char *str)
{
	do
	{
		get_file_line(str);
	} while(str[2]==':');
}

/*
 * get_file_line
 *
 * Reads a line from a file.  The line will always be
 * terminated by a a '\0', with no CR/LF.
 */
void get_file_line(char *str)
{
	int handle, i;
	char c;

	handle = dosopen(Batch_File, 0);
	doslseek(handle, Batch_pos, 0);
	for(i=2; i<str[0]; ++i)
	{
		c = getc(handle);

		if(c == 0)					/* Hit end of file */
		{
			In_Batch = FALSE;
			Echo = TRUE;
			c = '\n';
		}
		str[i] = c;
		if(c == '\n')				/* Hit end of line */
			break;
	}

	Batch_pos = doslseek(handle, 0L, 1);
	str[1] = i;
	str[i] = '\0';

	dosclose(handle);
}

/*
 * getc
 *
 * This function gets a single character from a handle.  It is used by
 * get_string from the console or get_batch_file_line.
 * It is a utility function for get_string and get_batch_file_line.
 *
 * Return
 * 	None
 */
char getc(unsigned handle)
{
	char c = 0;

	dosread(handle, &c, 1);
	switch(c)
	{
		case '\r':
			c = getc(handle); break;
		case 26:			/* Cntl C */
			c = 0; break;
		default:
			break;
	}
	return(c);
}

/*
 * do_batch
 *
 * This function sets miniCOMMAND up for executing a Batch File.
 *
 * Return
 * 	None
 */
void do_batch(char *file)
{
	strcpy(Batch_File, file);
	Batch_pos = 0;
	Echo = In_Batch = TRUE;
}

/*
 * toupper
 *
 * This function converts a lower case char to an upper case char.
 *
 * Return
 * 	the uppercase character
 */
char toupper(char c)
{
	if(c >= 'a' && c <= 'z')
		c += 'A' - 'a';
	return (c);
}

int isspace(char c)
{
	return (c==' ' || c=='\t');
}

/*
 * memset
 *
 * This basic library function sets memory to a specified value.
 *
 * Return
 * 	None
 */
void memset(char *buf, char val, unsigned size)
{
	int i;
	for(i=0; i<size; ++i)
		buf[i] = val;
}

/*
 * far_memcpy
 *
 * This function copies a block of memory using far pointers.
 *
 * Return
 * 	None
 */
void far_memcpy(void far *p1, void far *p2, unsigned count)
{
	char far *c1 = p1;
	char far *c2 = p2;
	unsigned i;
	for(i=0; i<count; ++i)
		c1[i] = c2[i];
}

/*
 * strcat
 *
 * This basic library function concatiniates one string to another.
 *
 * Return
 * 	None
 */
void strcat(char *s1, char *s2)
{
	while(*s1)
		++s1;
	while(*s2)
		*s1++ = *s2++;
	*s1 = '\0';
}

/*
 * strcpy
 *
 * This basic library function copies a C string.
 *
 * Return
 * 	None
 */
void strcpy(char *s1, char *s2)
{
	*s1 = '\0';
	strcat(s1, s2);
}


/*
 * strlen
 *
 * This basic library function determines the length of a C string.
 *
 * Return
 * 	The length in bytes.
 */
int strlen(char *s)
{
	int i;
	for(i=0; s[i]; ++i)
		;
	return(i);
}


/*
 * strchr
 *
 * This basic library function return the pointer to a specific
 * char in a string.
 *
 * Return
 * 	The pointer or NULL.
 */
char *strchr(char *s, char c)
{
	while(*s)
		if(*s == c)
			return s;
		else
			++s;
	return(NULL);
}

/*
 * get_env
 *
 * This function returns a pointer to the environment.
 *
 * Return
 * 	The environment address.
 */
char far *get_env(char *var)
{
	char far *p = get_env_full(var);
	if(p != NULL)
		p += strchr(var, '=') - var + 1;
	return (p);
}

/*
 * get_env_full
 *
 * This function returns a pointer to the environment.
 *
 * Return
 * 	The environment address.
 */
char far *get_env_full(char *var)
{
	char far *env = MK_FP(psp->Env_Seg, 0);
	char *p;
	unsigned len;

	if((p=strchr(var, '=')) == NULL)
		return NULL;
	else
		len = p  - var;

	while(*env)
	{
		if(far_memcmp(env, var, len))
			return(env);
		while(*env)
			++env;
		++env;
	}
	return (NULL);
}

int exec(char *progname, char *args)
{
	exec_block.environment = 0;
	exec_block.args = (void far *)args;
	exec_block.fcb1 = 0;
	exec_block.fcb2 = 0;

	return exec_func(progname, &exec_block);
}


/*
 * puts
 *
 * This basic library function displays a string using putc.
 *
 * Return
 * 	None
 */
void puts(char *s)
{
	char * p = s;
	while(*p)
	{
		if(*p == '\n')
			putc(*p++);
		else
		{
			while(*p && *p != '\n')
				++p;
			doswrite(1, s, p-s);
		}
		s = p;
	}
}


/*
 * error
 *
 * This function displays an error message given the error number.
 *
 * Return
 * 	None
 */
void error(unsigned number)
{
	puts(error_messages[number]);
	puts("\n");
}

/*
 * far_memcmp
 *
 * This funcntion compares two memory buffers using far pointers.
 *
 * Return
 * 	True for compare the same
 * 	False for non-Compare
 */
int far_memcmp(char far *s1, char far *s2, unsigned count)
{
	int i;
	for(i=0; i<count; ++i)
	{
		if(*s1++ != *s2++)
			return (FALSE);
	}
	return (TRUE);
}

char * skipwhite(char *p)
{
	while(isspace(*p))
		++p;
	return(p);
}
