//  $Id: MINICMD.H 1.5 2003/03/20 17:00:18Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1992-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/*********************************************************

   MINICMD.H

   Copyright (c) 1992-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains all the options for the miniCOMMAND
	command processor.  Enable or Disable options at your
	convenience.

**********************************************************/
#include "cmdprot.h"

/*
 * The following #defines implement or remove optional commands
 * from the miniCOMMAND processor.  A "1" implements the command.
 * A "0" removes the command from the command_table and removes
 * its code, reducing the size of the miniCOMMAND processor.
 */

#define CMD_CD		1			/* Implement the CD    command */
#define CMD_COPY	1			/* Implement the COPY  command */
#define CMD_DEL	1			/* Implement the DEL   command */
#define CMD_DIR	1			/* Implement the DIR   command */
#define CMD_ECHO	1			/* Implement the ECHO  command */
#define CMD_EXIT	1			/* Implement the EXIT  command */
#define CMD_GOTO	0			/* Implement the GOTO  command */
#define CMD_IF		0			/* Implement the IF    command */
#define CMD_PAUSE	1			/* Implement the PAUSE command */
#define CMD_REM	1			/* Implement the REM   command */
#define CMD_SET	1			/* Implement the SET   command */
#define CMD_TYPE	1			/* Implement the TYPE  command */
#define CMD_VER 	1			/* Implement the VER   command */

#define OPT_OUT_PIPE	1		/* allow output piping, dir >x.y */
#define OPT_FULL_PROMPT	0	/* use the standard $p$g prompt */

#define MIN_ENV_SIZE		128		/* # of bytes for Environment */
#define PROMPT "A: miniCMD>"		/* system prompt */

#define TRUE 1
#define FALSE 0
#define NULL (0)


#if 1
#define MK_FP(seg,off) \
	((void far *)(((unsigned long)(seg)<<16) | (unsigned)(off)))
#define FP_SEG(farptr)	((unsigned)((unsigned long)(farptr) >> 16))
#define FP_OFF(farptr)	((unsigned)(farptr))
#endif

#define iswhite(c) (c == ' '|| c == '\t')

/* DOS errors */
#define ERR_BADFUNC		0
#define ERR_NOTFOUND		2
#define ERR_NOTPATH		3
#define ERR_NOHANDLE		4
#define ERR_DENIED		5
#define ERR_BADHANDLE	6
#define ERR_BADMEMBLK	7
#define ERR_NOMEM			8
#define ERR_BADMEMSEG	9
#define ERR_INVLAID_ENV	10
#define ERR_INVALID_FMT	10
#define ERR_BADCODE		12
#define ERR_BADDATA		13
#define ERR_NOTDRIVE		15
#define ERR_CURDIR		16
#define ERR_NOT_SAME_DV	17
#define ERR_LASTFILE		18
#define ERR_NOTNEW		0x80		/* file already exists call 5BH */

/*
	File Control Block for DOS1 style calls.
*/
struct fcb
{
	unsigned char drive;
	unsigned char name[8];
	unsigned char ext[3];
	unsigned cur_block;
	unsigned record_size;
	unsigned long 	Size;
	unsigned 		date;
	unsigned 		Time;
	unsigned char	reserved[8];
	unsigned char 	cur_record;
	unsigned long	rand_record;
};

/*
	Execute Program, 4BxxH
*/
struct exec
{
	unsigned environment;
	void far * args;
	struct fcb far *fcb1;
	struct fcb far *fcb2;
};

/*
	File Found Block
*/
struct FFBLK
{
	unsigned char reserved[21];
	unsigned char attrib;
	unsigned 	  time;
	unsigned 	  date;
	unsigned long size;
	unsigned char name[13];
};

#define A_DIR	0x10

/*
	Program Segment Prefix, header for all programs
*/
struct PSP
{
	char			reserved[0xA];			/* 00 */
	void far *	term_address;			/* 0A */
	char			reserved2[0x8];		/* 0E */
	unsigned		parent;					/* 16 */
	char			reserved3[0x14];		/* 18 */
	unsigned		Env_Seg;					/* 2C */
	char			reserved4[4];			/* 2E */
   unsigned		hcount;             	/* 32 */
   void far *	htbl_addr;				/* 34 */
};

/*
	Memory Control Block, DOS memory allocation unit
*/
struct MCB
{
	unsigned char		sig;
	unsigned		owner;
	unsigned		size;
};

/*
	Used to construct internal commands
*/
struct internal
{
	char *cmd;
	void (*func)(char *);
};


extern struct exec exec_block;
extern struct fcb fcb0;
extern struct fcb internal_fcb;
extern unsigned _errno;
extern struct FFBLK ffblk;
extern char *cmd, *args, cmd_name[128], cmd_args[128];
extern char *error_messages[];
extern char Batch_File[66];
extern int In_Batch;
extern long Batch_pos;
extern char Echo;
extern int exit_avail;
extern struct PSP far *psp;
