//
//   $Id: BACSTORE.C 1.7 2002/10/15 23:51:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   BACSTORE.C  created 04/02/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This module contains the common file storage routines
	for BACKUP/RESTORE, and other utilities such as XCOPY
	and MOVE.

**********************************************************/
#define __FILENUM__ 5

// includes
#include <alloc.h>
#include <stdio.h>
#include "backup.h"
#include "bacvol.h"

// prototypes
#include "bacvol.pt"

// pointer to a buffer for file I/O
static char * pFileBuff;

// position on current volume
int   curVolume;						// current volume #
bool  volumeIsRemovable;			// is the volume removable?
ulong volumePos;						// current position in this volume
ulong volumeSize;						// size of this volume
char  szVolumeName[MAXBAKNAME];	// eg "A:DL970331.001"
FILE *fpVolume;						// file pointer to volume

/*
	Return the address of the file buffer.
	(DEBUG version makes sure it has been initialized).
*/
char * FileStoreBuff(void)
{
	ASSERT(pFileBuff);
	return pFileBuff;
}

/*
	Initialize the File Storage SubSystem for Backup/Restore.

	This does the following:
	1) Allocate memory for a disk I/O buffer
	2) Initialize memory store subsystem
	3) Determine volume size, and whether removable

	Returns FALSE if not enough memory. Returns TRUE otherwise.
*/
bool InitFileStore(void)
{
	// get memory for a disk I/O buffer
	pFileBuff = malloc(DISKIO_BUFFSIZE);
	if(!pFileBuff)
		return FALSE;

	// get memory for the in-memory storage system
	if(!InitMemStore())
	{
		free(pFileBuff);
		return FALSE;
	}

	// everything worked
	return TRUE;
}

/*
	Shut down the File Storage SubSystem for Backup/Restore.
*/
void KillFileStore(void)
{
	// free the memory store
	KillMemStore();

	// free our disk I/O buffer
	if(pFileBuff)
	{
		free(pFileBuff);
		pFileBuff = 0;
	}
}

/*
	Return size of this open file.  Leaves the current
	file position where it was.
*/
long DetermineFileSize(FILE *fp)
{
	long fileSize, curPos;

	// remember old position
	curPos = ftell(fp);

	// determine file size
	fseek(fp, 0L, SEEK_END);
	fileSize = ftell(fp);

	// restore old position
	fseek(fp, curPos, SEEK_SET);

	return fileSize;
}

/*
	Just wait for the return key.  Will quit on EOF for those
	systems that are doing input via a file.
*/
void WaitForReturn(void)
{
	int c;

	flushall();
	while(1)
	{
		c = getchar();
		if(c == '\n' || c == EOF)
			break;
	}
}
