//
//   $Id: BACVOL.C 1.11 2002/10/15 23:51:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************

   BACVOL.C  created 03/07/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Datalight Backup files can be written to floppy or to
	hard disks.  Each file is self-contained, although
	large files might split multiple floppies.

	Most important routines: StoreFile() and FlushFileStore().
	StoreFile() gets the files from the source disk into memory.
	FlushFileStore() puts the files from memory to the destination
	volume(s).

   DL BACKUP File Format:
   [ ascii identifier for sanity check and so user can use TYPE]
   [ file records]

********************************************************************/
#define __FILENUM__ 4

#define BACKUP 1		// indicate we want the backup messages

// includes
#include <dir.h>
#include <dos.h>
#include <mem.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "backup.h"
#include "bacvol.h"
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"

// prototypes
#include "backup.pt"
#include "bacvol.pt"
#include "bacstore.pt"

// used to store last filename logged/displayed
char szLastFileName[MAXLONGPATH];

/*
	Actually read in and store the given file.  This routine
	is part of a state-machine that knows about disks, destination
	volumes, the memory store and data compression.

	This MAY need to flush data to disk(s) via FlushFileStore() if
	this file exceeds the in-memory store.
*/
void StoreFile(char *pFileName, unsigned char attr,
								unsigned date, unsigned time)
{
	FILE *	fp;
	long		fileSize, curFilePos;
	struct	FileRecord curFileRecord;	// file record for this file
	long		curFileRecordPos;				// position of that record
	ulong		recordSize, curRecordPos;	// size of data in record
	long		memPos;				// temporary memory position
	int		nameSize;			// part of record header
	unsigned	len;

	// validate parameters
	ASSERT(pFileName && *pFileName);	// filename must exist
	ASSERT(!(attr & NOTFILE));	// don't back up directories or volume labels!

	// todo: may be skipping files to get to the beginning
	// of a previously bad volume

	// user said only backup those with attribute set...
	// so don't store if attribute clear
	if(swArchive && !(attr & A_ARCHIVE))
		return;

	// open the file
	fp = fopen(pFileName, "rb");
	SHIPASSERT(fp);	// it HAS to work, we just found the file!

	// determine file size
	fileSize = DetermineFileSize(fp);

	// update .LOG info
	++numFiles;
	fileBytes += fileSize;

	// get copy of the actual file name, not the "root" of
	// the name, just the path part
	nameSize = strlen(pFileName+ROOTLEN) + 1;

	// start with all 0s
	memset(&curFileRecord, 0, sizeof(curFileRecord));
	curFileRecord.sig  = FILE_RECORD_SIG;
	curFileRecord.attr = attr & (~A_ARCHIVE);	// make sure archive is clear
	curFileRecord.date = date;
	curFileRecord.time = time;
	curFileRecord.size = fileSize;
	curFileRecord.nameSize = nameSize;
	curFileRecord.moreVolumes = TRUE;

	// start at beginning, and store whole file
	curFilePos = 0L;
	do
	{
		// remember where this file block starts at
		// (may not be entire file, as the file could be bigger
		//  than the memory store).
		curFileRecord.startOffset = curFilePos;

		// not enough room for header and some data? flush file store
		if(MemStoreLeft() < MIN_RECORD_SIZE)
			FlushFileStore();

		// remember current position in the store so we can update
		// the file record later
		curFileRecordPos = MemStoreTell();

		// write the file record to the memory store
		MemStoreWrite(&curFileRecord, sizeof(curFileRecord)-1);
		MemStoreWrite(pFileName+ROOTLEN, nameSize);

		// determine maximum size of write to this next volume
		// it will be the minimum of:
		//   1) file size
		//   2) bytes left in memory store
		recordSize = fileSize - curFilePos;
		if(MemStoreLeft() < recordSize)
			recordSize = MemStoreLeft();

		// write data to the store
		curRecordPos = 0L;
		while(curRecordPos < recordSize)
		{
			// read in as much as we can from the file
			if(recordSize - curRecordPos < DISKIO_BUFFSIZE)
				len = (unsigned)(recordSize - curRecordPos);
			else
				len = DISKIO_BUFFSIZE;

			// read some bytes from the file
			fread(FileStoreBuff(), 1, len, fp);

			// write them to the memory store
			MemStoreWrite(FileStoreBuff(), len);

			// we wrote that part of the file, on to next
			curRecordPos += len;
		}

		// record the actual length of header+data in the "next" field
		// curFileRecord.next = sizeof(curFileRecord)-1 + nameSize + recordSize;
		memPos = MemStoreTell();
		MemStoreSeek(curFileRecordPos, SEEK_SET);
		MemStoreWrite(&memPos, sizeof(memPos));
		MemStoreSeek(memPos, SEEK_SET);

		// update file position
		curFilePos += recordSize;

	} while(curFilePos < fileSize);

	// done with file
	fclose(fp);

	// done with file, reset archive bit to mark it as backed up
	dlSetFileAttributes(attr & (~A_ARCHIVE), pFileName );
}

/*
	Flush in-memory file store to disk.

	Called during a BACKUP when in-memory store is full. Also called
	at end of BACKUP to flush final in-memory store.

	May require multiple disks if in-memory file store is larger than
	volume size.  Will prompt user for each disk if appropriate.
*/
void FlushFileStore(void)
{
	ulong	memTop;
	int	moreData;

	// nothing to do! no file records in memory
	memTop = MemStoreTell();
	if(!memTop)
		return;

	// not currently in the middle of a volume? get a new one going
	if(!fpVolume)
		CreateNewVolume();

	// keep writing volumes until we have emptied the memory
	// store (may only write a partial volume, depending on
	// memory store size and volume size).
	MemStoreSeek(0L, SEEK_SET);
	while(MemStoreTell() < memTop)
	{
		// flush the next record to disk
		moreData = FlushNextRecord(volumeSize, &volumePos);

		// all done with backup set!
		if(moreData == LAST_VOLUME)
		{
			fclose(fpVolume);
			break;
		}

		// done with this disk? close the volume, and open a new one
		if(volumePos == volumeSize)
		{
			// close the old volume
			fclose(fpVolume);
			fpVolume = 0;

			// open the new
			CreateNewVolume();
		}
	}

	// file store is flushed
	MemStoreSeek(0L, SEEK_SET);
}

/*
	Flush the next record.  Uses records in the MemStore().

	Returns "MORE_DATA" if there is more to write
	Returns "LAST_VOLUME"  if last volume record encountered
*/
int FlushNextRecord(ulong volumeSize, ulong *pVolumePos)
{
#if DEBUG
	static bool done = FALSE;
#endif
	struct FileRecord fileRecord;
	unsigned len;
	ulong 	maxLen;
	long		sizeWrite;
	long		orgMemPos;		// position in Memory Store of file record
	ulong 	splitNext;		// file record might be split across volumes
	long		splitPos = 0L;
	bool		gotName;

#if DEBUG
	// should NEVER be called after we've hit the last record
	ASSERT(!done);
#endif

	// validate parameters
	ASSERT(volumeSize);
	ASSERT(pVolumePos && *pVolumePos < volumeSize);

	// where are we at?
	orgMemPos = MemStoreTell();

	// get this file record
	MemStoreRead(&fileRecord, sizeof(fileRecord)-1);

	// file record must be valid!
	ASSERT(fileRecord.sig == FILE_RECORD_SIG);
	ASSERT(fileRecord.next == 0 || fileRecord.next > orgMemPos);

	// determine maximum length we can write
	maxLen = volumeSize - *pVolumePos;

	// all done? (next of 0 indicates that)
	if(fileRecord.next == 0 && maxLen >= sizeof(fileRecord)-1)
	{
		fwrite(&fileRecord, 1, sizeof(fileRecord)-1, fpVolume);
		return LAST_VOLUME;
	}

	// not enough room for a record? then ask for a new volume
	if(maxLen < MIN_OUT_SIZE)
	{
		// seek back to before this record
		MemStoreSeek(orgMemPos, SEEK_SET);

		// volume is full
		*pVolumePos = volumeSize;

		// but we still have more volumes to go
		return MORE_DATA;
	}

	// does the record need to be split up onto mulitple volumes?
	sizeWrite = fileRecord.next - orgMemPos;
	if((ulong)sizeWrite > maxLen)
	{
		sizeWrite = maxLen;
		splitPos  = orgMemPos + sizeWrite;
		splitNext = fileRecord.next;
	}

	// update file record next field
	fileRecord.next = *pVolumePos + sizeWrite;

	// update starting volume
	if(!fileRecord.startOffset)
		fileRecord.startVol = curVolume;

	// write the updated file record to disk
	fwrite(&fileRecord, 1, sizeof(fileRecord)-1, fpVolume);
	sizeWrite -= sizeof(fileRecord)-1;
	*pVolumePos += sizeof(fileRecord)-1;

	// write out all of this record to disk
	gotName = FALSE;
	while(sizeWrite)
	{
		// get largest size we can write
		if(sizeWrite > DISKIO_BUFFSIZE)
			len = DISKIO_BUFFSIZE;
		else
			len = (unsigned)sizeWrite;

		// get some data to write to disk
		MemStoreRead(FileStoreBuff(), len);

		// display file we are about to add to backup
		if(!gotName && strcmp(szLastFileName, FileStoreBuff()) != 0)
		{
			printf("%s\n", FileStoreBuff());

			// display file to .LOG file
			if(fpLogFile)
				LogThisFile(&fileRecord, FileStoreBuff());

			// store that filename
			strcpy(szLastFileName, FileStoreBuff());

		}

		// the name is at the beginning of the 1st entry in the loop
		gotName = TRUE;

		// write it to disk
		fwrite(FileStoreBuff(), 1, len, fpVolume);

		// wrote some bytes
		sizeWrite -= len;
		*pVolumePos += len;
	}

	// do we have to split up a record?
	if(splitPos)
	{
		// make room for fileRecord and name
		splitPos -= sizeof(fileRecord)-1 + fileRecord.nameSize;
		MemStoreSeek(splitPos, SEEK_SET);

		// update next field for that record
		fileRecord.next = splitNext;
		fileRecord.startOffset += splitPos - orgMemPos;

		// write out file record
		MemStoreWrite(&fileRecord, sizeof(fileRecord)-1);

		// write out file name
		MemStoreSeek(orgMemPos + sizeof(fileRecord)-1, SEEK_SET);
		MemStoreRead(FileStoreBuff(), fileRecord.nameSize);
		MemStoreSeek(splitPos + sizeof(fileRecord)-1, SEEK_SET);
		MemStoreWrite(FileStoreBuff(), fileRecord.nameSize);

		// seek back to split record
		MemStoreSeek(splitPos, SEEK_SET);
	}

	// must be more data to write
	return MORE_DATA;
}

/*
	Displays a single file (used both by Dir and Restore).
*/
void LogThisFile(struct FileRecord *pFileRecord, char szFileName[])
{
	char	szDate[DATESIZE];
	char	szTime[DATESIZE];
	char *p;

	// determine date and itme
	DateToAscii(pFileRecord->date, szDate);
	TimeToAscii(pFileRecord->time, szTime);

	// show the file entry
	p = szLogFile;
	sprintf(p, msgLogLine, szFileName, pFileRecord->size,
				szDate, szTime);
	p += strlen(p);

	// was file continued from previous volume?
	if(pFileRecord->startOffset && pFileRecord->startVol != curVolume)
	{
		sprintf(p, msgContinued, pFileRecord->startVol);
		p += strlen(p);
	}

	// done with line
	sprintf(p, "\n");

	// write it to the file
	fputs(szLogFile, fpLogFile);
}

/*
	Initialize all the volume info
*/
void InitVolume(void)
{
	// set up starting postition
	curVolume = 0;
	volumePos = 0L;

	// determine size of volume on backup drive
	dlIsDriveRemovable( (uchar)(backupDrive+1), &volumeIsRemovable );
	if(!volumeIsRemovable)
	{
		dlGetFreeSpaceOnDrive(backupDrive+1, &volumeSize);
	}

	// for use in debugging, set a small volume size
#if DEBUG
	if(swVolSize)
	{
		volumeSize = 1024L*swVolSize;
		volumeIsRemovable = TRUE;
	}
#endif

	// create an initial volume name
	sprintf(szVolumeName, "%c:%s", backupDrive+'A', szBackupBaseName);
}

/* Is a drive write protected? */
bool IsWriteProtected(int iDrive)
{
	static char szFileName[] = "A:$";
	FILE *fp;

	*szFileName = iDrive+'A';
	fp = fopen(szFileName,"wb");
	if(!fp)
		return TRUE;	// write protected (can't create a file for write)
	fclose(fp);
	remove(szFileName);
	return FALSE;		// not write protected
}

ulong ClusterRoundUp(int iDrive, ulong fileSize)
{
	union REGS r;
	ulong clusterSize;
	ulong clusters;

	// determine cluster size
	r.h.ah = 0x36;
	r.x.dx = iDrive+1;
	intdos(&r,&r);
	if(r.x.ax == 0xFFFF)		/* invalid drive */
		return 0;
	clusterSize = (ulong)r.x.ax * r.x.cx;

	// determine actual # of clusters occupied by file
	clusters = fileSize / clusterSize;
	if(fileSize % clusterSize)
		++clusters;

	// return size of file in cluster increments
	return clusterSize * clusters;
}

/* have we hit the first prompt yet? */
bool bFirstPrompt = TRUE;

/*
	When appending, find the last volume in the set.

	This function will exit if it cannot find the last volume on a
	fixed disk.
*/
void FindLastVolume(void)
{
	FILE * fpVolume = NULL;
	struct ffblk ff;
	char szVolumeID[32];
	struct FileRecord fileRecord;
	int err;

	// loop through, looking for the last volume
	while(1)
	{
		// don't ask again for fixed disks
		if(!bFirstPrompt && !volumeIsRemovable)
			ErrorExit(ERR_NOTFOUND);

		// prompt (but NOT if user hit /Y)
		if(swPrompt || !bFirstPrompt)
		{
			printf(msgInsertLastVol);
			printf(msgReturn);
			WaitForReturn();
		}

		// no longer in the first prompt
		bFirstPrompt = FALSE;

		// look for a volume set
		sprintf(szBackupBaseName, "%c:%s", backupDrive+'A', "DL*.*");
		err = findfirst(szBackupBaseName, &ff, A_NORMAL);
		if(err)
		{
			printf(errMsgNotLastVol);
			continue;
		}

		// does the volume contain a standard volume signature?
		sprintf(szBackupBaseName, "%c:%s", backupDrive+'A', ff.ff_name);
		fpVolume = fopen(szBackupBaseName, "rb");
		SHIPASSERT(fpVolume);
		if(fread(szVolumeID,1,sizeof(szVolumeID),fpVolume) != sizeof(szVolumeID)
			|| memcmp(szVolumeID,szVolumeSignature,sizeof(szVolumeID)) != 0)
		{
			fclose(fpVolume);
			printf(errMsgNotLastVol);
			continue;
		}

		// is it the last volume in the set? (does it end in a NULL record?)
		fseek(fpVolume, -1L*(sizeof(struct FileRecord)-1), SEEK_END);
		if(fread(&fileRecord, 1, sizeof(struct FileRecord)-1, fpVolume) !=
			(sizeof(struct FileRecord)-1) ||
			fileRecord.next != 0 || fileRecord.sig != FILE_RECORD_SIG)
		{
			fclose(fpVolume);
			printf(errMsgNotLastVol);
			continue;
		}

		// no errors! figure out the rest
		break;
	}

	// set volume position
	volumePos = ftell(fpVolume) - (sizeof(struct FileRecord)-1);
	fclose(fpVolume);

	// set volume number
	curVolume = atoi(FindExt(ff.ff_name)+1);

	// set volume name
	sprintf(szVolumeName, "%c:%s", backupDrive+'A', ff.ff_name);
	*FindExt(ff.ff_name) = 0;
	strcpy(szBackupBaseName, ff.ff_name);

#if DEBUG
	// we've already set a volume size for debugging purposes
	if(swVolSize)
		return;
#endif

	// determine volume size
	if(!dlGetFreeSpaceOnDrive(backupDrive+1, &volumeSize))
		ABORT();

	// we'll be replacing the final (NULL) file record, so we have a
	// little more space than is free on disk.
	volumeSize += ClusterRoundUp(backupDrive,
							volumePos + (sizeof(struct FileRecord)-1));
}

/*
	Used for removable media only: formats the disk in the
	floppy drive.
*/
void CreateBlankVolume(void)
{
	char szCheckName[MAXBAKNAME];
	char szLastFound[MAXBAKNAME];
	struct ffblk ff;
	bool overWrite = FALSE;
	int	volFound;

	// we haven't used the check name yet!
	*szLastFound = 0;
	overWrite = FALSE;

	// keep asking for disks until we get one that will work
	// for backup
	while(1)
	{
		// user may have said NO prompting, but we MUST if there
		// is a bad volume or more than one
		if(swPrompt || !bFirstPrompt)
		{
			// give user a chance to insert a new disk
			printf(msgInsertDisk, szVolumeName);
			if(overWrite)
				printf(msgOtherReturn);
			else
				printf(msgReturn);
			WaitForReturn();
		}

		// no longer first time through the loop
		bFirstPrompt = FALSE;

		// is this a disk from our current backup set?
		// don't overwrite it!  allow user to override the disk
		strcpy(szCheckName, szVolumeName);
		SetFileExt(szCheckName, "*");
		if(findfirst(szCheckName, &ff, A_NORMAL) == 0 &&
			strcmp(szLastFound, ff.ff_name) != 0)
		{
			// don't worry about volumes unless they are LESS than
			// the currently expected volume
			volFound = atoi(FindExt(ff.ff_name)+1);
			if(volFound < curVolume)
			{
				// let user know his mistake!
				printf(errMsgFromThisSet, ff.ff_name);

				// remember last name, in case user wants to overwrite the disk
				strcpy(szLastFound, ff.ff_name);
				overWrite = TRUE;

				continue;
			}
		}

		// for removable media, format the volume. Size may change
		// due to bad clusters or different media size. That's OK
		if(!QuickFormat(backupDrive))
		{
			// try again!
			printf(errMsgNoFormat, backupDrive+'A');
			continue;
		}

		// create the volume label on removable media
		// the volume label isn't really used, but makes it obvious
		// this volume is a backup disk.
		SetVolumeLabel(backupDrive, "DL BACKUP");

#if DEBUG
		// we've already set a volume size for debugging purposes
		if(!swVolSize)
#endif
		// get size of new disk (we get this each time because
		// some disks may have bad clusters)
		if(!dlGetFreeSpaceOnDrive(backupDrive+1, &volumeSize))
			ABORT();

		// no errors, so we're done!
		break;

	} // end of loop
}

/*
	Ask user to input a new volume (disk).  May need to format
	the new volume.  Also determine size of new volume.

	This only happens on removable media.
*/
void CreateNewVolume(void)
{
	char	szVolumeExt[6];

	// we are on a new volume #, increment volume #
	++curVolume;

	// reset comparison for names (since we're on a new volume)
	*szLastFileName = 0;

	// create the new volume name
	sprintf(szVolumeExt,".%03d", curVolume);
	SetFileExt(szVolumeName, szVolumeExt);

	do
	{
		// for appending, determine
		if(swAppend)
			FindLastVolume();

		// wait for user to insert a new disk
		else if(volumeIsRemovable)
			CreateBlankVolume();

		// display a return between title and filenames
		printf("\n");

		// validate the volume size
		ASSERT(volumeSize > MIN_RECORD_SIZE);

		// append to volume file
		if(swAppend)
		{
			// last record is a NULL record; we will be overwriting the
			// NULL record.
			fpVolume = NULL;
			if(!IsWriteProtected(backupDrive))
				fpVolume = fopen(szVolumeName, "r+b");
			if(fpVolume)
				fseek(fpVolume, volumePos, SEEK_SET);
		}

		// create the volume file
		else
			fpVolume = fopen(szVolumeName, "wb");

		// floppy or network drive might be write-protected
		if(!fpVolume)
		{
			// show write protect error
			printf(errMsgWriteProtect, *szVolumeName);

			// quit on network drive (nothing else we can do!)
			if(!volumeIsRemovable)
				ErrorExit(ERR_WRITEPROT);
		}

	} while(!fpVolume);

	// the volume had BETTER be able to be opened
	SHIPASSERT(fpVolume);

	// initialize a new volume, complete with signature (but not if
	// we're just appending to a volume)
	if(!swAppend)
	{
		fwrite(szVolumeSignature, 1, sizeof(szVolumeSignature), fpVolume);
		volumePos = ftell(fpVolume);
	}

	// no longer in append mode.
	swAppend = FALSE;

	// display the new volume name in .LOG file
	if(fpLogFile)
	{
		sprintf(szLogFile, msgDirectoryOf, szVolumeName);
		fputs(szLogFile, fpLogFile);
	}
}

/*
	Write the last volume to disk.  There will ALWAYS be room for
	this record, because StoreFile flushes if the memory store is
	too full.
*/
void MarkLastVolume(void)
{
	struct FileRecord fileRecord;

	// must ALWAYS be enough mem in store for this final record
	ASSERT(MemStoreLeft() >= sizeof(fileRecord)-1);

	// set up an "empty" File Record to mark end of Backup
	memset(&fileRecord, 0, sizeof(fileRecord));
	fileRecord.sig  = FILE_RECORD_SIG;

	// write out that "empty" file record
	MemStoreWrite(&fileRecord, sizeof(fileRecord)-1);
}

/*
	Format a drive by just zeroing the FATs.
*/
bool QuickFormat(int drive)
{
	struct BPB bpb;
	unsigned numClus;
	long		startSec;
	int		fat;
	int		numSec, rootDirSectors;

	ASSERT(drive >= 0 && drive < 26);

	// determine size of BPB
	if(!GetBPB(drive, &bpb))
		return FALSE;

	// file storage buffer MUST be big enough for entire FAT,
	// or we'll have to use a different algorithm!
	SHIPASSERT(bpb.fat_size * bpb.sec_size <= DISKIO_BUFFSIZE);

	// calculate # of clusters
	numClus = NumClustersOf(&bpb);

	// read in the whole 1st fat
	startSec = bpb.res_sec;
	if(absread(drive, bpb.fat_size, startSec, FileStoreBuff()) != 0)
		return FALSE;

	// zero the FAT
	ZeroFat(numClus, bpb.media, FileStoreBuff());

	// write out the new, zeroed FAT to ALL fats
	for(fat=0; fat < bpb.num_fats; ++fat)
	{
		if(abswrite(drive, bpb.fat_size, startSec, FileStoreBuff()) != 0)
			return FALSE;

		// on to next FAT
		startSec += bpb.fat_size;
	}

	// zero the root directory in as large of chunks as possible
	rootDirSectors	= bpb.dir_ent / (bpb.sec_size / DIR_ENT_SIZE);
	memset(FileStoreBuff(), 0, DISKIO_BUFFSIZE);
	while(rootDirSectors)
	{
		if(rootDirSectors > DISKIO_BUFFSIZE / bpb.sec_size)
			numSec = (int)(DISKIO_BUFFSIZE / bpb.sec_size);
		else
			numSec = rootDirSectors;

		// write out those zeroed sectors
		if(abswrite(drive, numSec, startSec, FileStoreBuff()) != 0)
			return FALSE;

		// may need to loop if root is big
		rootDirSectors -= numSec;
		startSec += numSec;
	}

	// reset drive so DOS knows about the new FAT
	ResetDrive(drive);

	return TRUE;
}

#define fat0(u) ((u)&0xFFF)
#define fat1(u) ((u)>>4)
#define HICLUS		0xFF6		// highest cluster # for 12-bit fat
#define BADCLUS	0xFF7		// don't zero these, they are marked bad
#define DIRENTSIZE 32

/*
	Get the BPB of the disk currently in the drive.
*/
bool GetBPB(int drive, struct BPB *pBPB)
{
	struct DEVICE_PARMS deviceParms;
	union REGS r;

	ASSERT(drive >= 0 && drive < 26);
	ASSERT(pBPB);

	// get the BPB from DOS
	deviceParms.specFunc = 1;
	r.x.ax = 0x440d;
	r.x.bx = drive+1;
	r.x.cx = 0x860;
	r.x.dx = (unsigned)&deviceParms;
	intdos(&r,&r);

	// error?
	if(r.x.cflag)
		return FALSE;

	// copy the BPB to the passed parameter
	memcpy(pBPB, &deviceParms.bpb, sizeof(struct BPB));

	return TRUE;
}

/*
	Return # of clusters in this BPB
*/
unsigned NumClustersOf(struct BPB *pBPB)
{
	unsigned numClus;
	long totalSectors;

	// validate parameter
	ASSERT(pBPB && (pBPB->media & 0xF0) == 0xF0);

	// determine # of sectors on disk
	if(pBPB->num_sec)
		totalSectors = pBPB->num_sec;
	else
		totalSectors = pBPB->lnumsec;

	// determine sectors in DATA area of disk
	totalSectors -= pBPB->res_sec + (pBPB->num_fats * pBPB->fat_size);
	totalSectors -= pBPB->dir_ent / (pBPB->sec_size / DIRENTSIZE);

	// determine # of clusters
	numClus = 2 + (unsigned)(totalSectors / pBPB->sec_per_clus);

	return numClus;
}

/*
	Zero an in-memory copy of the FAT.  The entire FAT must be
	in memory.
*/
void ZeroFat(unsigned numClus, uchar mediaID, void *pFat)
{
	unsigned clus, clusValue;

	//  set the media ID correctly, regardless of what it was in the
	// old FAT
	SetClusterEntry(0, 0xFF00 | mediaID, pFat);
	SetClusterEntry(1, 0xFFFF, pFat);

	/* mark all non-BAD clusters as free */
	for(clus=2; clus < numClus; ++clus)
	{
		clusValue = GetClusterEntry(clus, pFat);
		if(clusValue != BADCLUS)
			SetClusterEntry(clus, 0, pFat);
	}
}

/*
	Only works for 12-bit fats.
*/
unsigned GetClusterEntry(unsigned n, void *pFat)
{
	char * pEntry = pFat;
	unsigned value;

	// validate parameters
	ASSERT(n <= HICLUS);
	ASSERT(pFat);

	// determine entry value
	pEntry += n + (n>>1);
	value = *(unsigned *)pEntry;

	// return the 12-bit entry
	return (n&1 ? fat1(value) : fat0(value));
}

/*
	Only works for 12-bit fats.
*/
void SetClusterEntry(unsigned n, unsigned newValue, void *pFat)
{
	char * pEntry = pFat;
	unsigned value;

	// validate parameters
	ASSERT(n <= HICLUS);
	ASSERT(pFat);

	// we use FFFF for EOF, convert to 12 bits
	newValue &= 0xFFF;

	// determine entry value
	pEntry += n + (n>>1);
	value = *(unsigned *)pEntry;

	// mark the particular cluster entry
	if(n&1)
		value = (value & 0x000F) | (newValue << 4);
	else
		value = (value & 0xF000) | newValue;

	// update the entry
	*(unsigned *)pEntry = value;
}

/*
	Reset a drive so it knows about updated sectors in FAT.
*/
void ResetDrive(int drive)
{
	union REGS r;

	// reset disk
	r.h.ah = 0x0D;
	intdos(&r,&r);

	// get new DPB for disk - resets internal buffers
	r.h.ah = 0x32;
	r.x.dx = drive+1;
	intdos(&r,&r);
}
