//
//   $Id: BACWILD.C 1.6 2002/10/15 23:51:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   BACWILD.C  created 04/18/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   Handles the matching of wildcards.

********************************************************************/
#define __FILENUM__ 7

#include <ctype.h>
#include <string.h>

#include "backup.h"

// wildcard characters
#define WILDCARD '*'
#define WILDCHAR '?'
#define IsWildCard(c) ((c)==WILDCARD || (c)==WILDCHAR)

/*
	Does the string match the mask string?

	In the mask, a '?' is used as a wildcard character to
	match any character.  The '*' wildcard has already been
	factored out at a higher level.

	Returns TRUE if they match, FALSE if not.
*/
bool WildCompare(char *pMask, int lenMask, char *pStr)
{
	// validate parameters
	ASSERT(pMask);
	ASSERT(lenMask >= 0);
	ASSERT(pStr);

	// check each character.  all must match.
	while(lenMask && *pStr)
	{
		// always matches on a ?
		// ignores case (toupper)
		if(*pMask != WILDCHAR && toupper(*pMask) != toupper(*pStr))
			return FALSE;

		// on to next character
		++pStr;
		++pMask;

		// one more character matched
		--lenMask;
	}

	// if we looked at ALL characters, then string match
	// if not, string was too short
	return (lenMask == 0);
}

/*
	Does the mask string appear some place inside of the
	other string?

	The mask may contain the wildcard '?' which will match
	any one character.  The '*' wildcard has already been
	factored out.

	This function is effectively strstr() with wildcards and
	ignore case.

	Returns TRUE if mask string is found, FALSE if not.
*/
bool MidWildCompare(char *pMask, int lenMask, char *pStr)
{
	char * pMaskThisTry;
	int    lenMaskThisTry;

	// validate parameters
	ASSERT(pMask);
	ASSERT(lenMask >= 0);
	ASSERT(pStr);

	// look until end of string, or found
	while(*pStr)
	{
		// start over in mask for another try
		pMaskThisTry  = pMask;
		lenMaskThisTry = lenMask;

		// try to match characters in this string
		while(*pStr && lenMaskThisTry)
		{
			// ignores case. all chars match "?".
			if(*pMaskThisTry != WILDCHAR &&
					toupper(*pMaskThisTry) != toupper(*pStr))
			{
				// starting searching on following character
				++pStr;
				break;
			}

			// compare OK so far... try next character
			++pStr;
			++pMaskThisTry;

			// one less character to compare
			--lenMaskThisTry;
		}

		// did we find a match? yes, we're done!
		if(lenMaskThisTry == 0)
			break;
	}

	// matches if all characters matched
	return (lenMaskThisTry == 0);
}

/*
	Match a single string string, given a mask.  The mask
	may contain wildcards, and it will search

	Examples:
	*			always matches
	*x*		matches if string contains an "x"
	a* 		matches if string begins with "a"
	*ing		matches if string ends in "ing"
	a*b*c		begins with "a", has a "b" in it, ends with "c"
*/
bool MatchMask(char *pMask, char *pStr)
{
	bool 	 doesMatch = TRUE;
	char * pNextMask;
	int	 lenMask, lenStr;

	// validate parameters
	ASSERT(pMask);
	ASSERT(pStr);

	// empty mask doesn't ever match
	if(*pMask == 0)
		return FALSE;

	// special case: *.* acts just like *, no need to find a
	// dot in the string
	if(strcmp(pMask, "*.*") == 0)
		pMask = "*";

	// find first wildcard
	pNextMask = pMask;
	while(*pNextMask && *pNextMask != WILDCARD)
		++pNextMask;

	// determine length of first wildcard string
	lenMask = (int)(pNextMask-pMask);

	// look for "first" mask match (at start of string)
	doesMatch = WildCompare(pMask, lenMask, pStr);

	// might be NO wildcards, make sure it ends in 0s
	if(doesMatch && lenMask && *pNextMask==0)
	{
		// doesn't match, not at end of string
		if(pStr[lenMask] != 0)
			return FALSE;
	}

	// check for all "mid" ones
	while(doesMatch && *pNextMask)
	{
		// skip wildcard
		pMask = pNextMask+1;

		// find next wildcard
		++pNextMask;
		while(*pNextMask && *pNextMask != WILDCARD)
			++pNextMask;

		// determine length of mask
		lenMask = (int)(pNextMask-pMask);

		// must be a "mid" mask (anywhere in string)
		if(*pNextMask != 0)
			doesMatch = MidWildCompare(pMask, lenMask, pStr);

		// must be a "last" mask (at end of string)
		// only check if there actually IS a last mask string
		else if(lenMask)
		{
			// we need some convenient lengths here
			lenStr  = strlen(pStr);

			// doesn't compare if mask is longer than string!
			if(lenStr < lenMask)
				doesMatch = FALSE;
			else
				doesMatch = WildCompare(pMask, lenMask, pStr+(lenStr-lenMask));
		}
	}

	// return whether it matches or not
	return doesMatch;
}

#if UNIT_TEST
#include <stdio.h>

#define Failed() { printf("Failed! %d", __LINE__); return 1; }

int main(void)
{
	if(!MatchMask("*.*", "x"))
		Failed();
	if(!MatchMask("*", "x"))
		Failed();
	if(!MatchMask(".*", ".x"))
		Failed();
	if(MatchMask(".*", "x"))
		Failed();
	if(!MatchMask("*a*", "a"))
		Failed();
	if(!MatchMask("*a*", "abc"))
		Failed();
	if(!MatchMask("*a*", "variable"))
		Failed();

	if(!MatchMask("*.C", "BACMSG.C"))
		Failed();
	if(MatchMask("*.C", "BACMSG.H"))
		Failed();

	printf("passed");
	return 0;
}
#endif
