//
//   $Id: RESTORE.C 1.11 2002/10/15 23:51:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   RESTORE.C  created 03/07/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   Restores files saved with Datalight Backup.

********************************************************************/
#define __FILENUM__ 1

// includes
#include <ctype.h>
#include <dir.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dosver.h"
#include "backup.h"
#include "bacvol.h"
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"
#include "fatworld.h"
#include "abortlfn.H"

// messages
#define RESTORE 1		// indicate we want the restore messages
#include "dosmsg.c"

// prototypes
#include "restore.pt"
#include "bacstore.pt"
#include "bacwild.pt"

// cmdline switches
bool	swPrompt = TRUE;	// prompt for volume, file overwrites?
bool	swNewerPrompt;		// prompt only on newer or read-only files
bool	swSubdir;			// recurse into subdirecties?
bool	swDir;				// show directory only (don't restore files)
bool	swRaw;				// show raw dump of file records

// Destination path info
char	szDestPath[MAXLONGPATH];		// example C:\DEV\ROMDOS\*.*
static char szDestMask[MAXBAKNAME];	// example *.*
int	rootLength;							// length of root string

// Source drive info
int	backupDrive;						// source drive for restore

// file statistics
ulong	totalFiles;
ulong	totalBytes;

// help for restoring directories
char	szLastPath[MAXLONGPATH];
int	lastPathLen;

// for display of last filename
char	szLastFile[MAXLONGPATH];
bool	ok2Flush = TRUE;

bool fIsFat32World;

void DisplaySignOnMessage( void )
{
	char * szSignOn = "RESTORE %s %s\n%s\n";


	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/*
	Restore program.
*/
int main(int argc, char *argv[])
{
	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	//  Abort if there are LFNs on the command line
	//
	AbortForLFNArgs();

	/* display sign-on */
	DisplaySignOnMessage();

	// process command-line arguments into global switches
	// and restore name.  Will error and show usage if problem
	// in cmdline args.
	ProcessCmdline(argc, argv);

	// initialize country info
	InitCountryInfo();

	// is the backup volume a real drive?
	InitVolume();

	// initialize file storage system.  can only fail if
	// there is not enough memory.
	if(!InitFileStore())
		Error(errMsgNoMem, "", ERR_NOMEM);

	// now, go through each file/directory and back up the stuff
	Restore(szDestPath, szDestMask);

	// done, free the memory
	KillFileStore();

	// everything worked
	return 0;
}

/*
	Scan for switches, and determine source and destination
	names.  Will show help if an errors encountered.
*/
void ProcessCmdline(int argc, char *argv[])
{
	int	i, switchLen;
	char	*pSwitch;
	bool	srcFound, dstFound, moreFound;

	// no arguments, just show help
	if(argc == 1)
		Usage();

	// loop through, looking for switches
	for(i=1; i<argc; ++i)
	{
		// any switches in this cmdline argument?
		pSwitch = argv[i];
		while(pSwitch)
		{
			// look for next switch
			pSwitch = strchr(pSwitch, '/');
			if(!pSwitch)
				break;

			// default switch length (example: /s)
			switchLen = 2;

			// got a switch?
			switch(toupper(pSwitch[1]))
			{
				// recurse into subdirectories
				case 'C':
				case 'Y':
					swPrompt = FALSE;
					break;

				// recurse into subdirectories
				case 'S':
					swSubdir = TRUE;
					break;

				// prompt on overwriting read-only or newer files
				case 'P':
					swNewerPrompt = TRUE;
					break;

				// only Display files, don't restore
				case 'D':
				case 'R':

					// just displaying directory of backup
					swDir = TRUE;

					// display raw file records
					if(toupper(pSwitch[1]) == 'R')
						swRaw = TRUE;
					break;

				// does user want help?
				case '?':
					Usage();
					break;

				default:
					printf(errMsgBadOpt, pSwitch[1]);
					printf(msgUsage);
					exit(ERR_BADOPT);
			}

			// remove switch from the cmdline, remember to copy NUL
			memmove(pSwitch, pSwitch+switchLen, strlen(pSwitch+switchLen)+1);

		}	// end of while (pSwitch)

	}	// end of for(i<argc)

	// find source path and destination drive
	srcFound = dstFound = moreFound = FALSE;
	for(i=1; i<argc; ++i)
	{
		// may skip arg if switches were first
		if(*argv[i] == 0)
			continue;

		// getting source drive?
		if(!srcFound)
		{
			// get drive letter
			backupDrive = toupper(*argv[i]) - 'A';

			// make sure drive is valid
			if(!IsValidDrive(backupDrive))
				Error(errMsgBadSource, (char *)(backupDrive+'A'), ERR_BADPATH);

			// may only want to restore SOME files
			if(*(argv[i]+1) == ':' && *(argv[i]+2))
				CopyNamePart(szDestMask, argv[i]);
			else
				strcpy(szDestMask, "*.*");

			srcFound = TRUE;
		}

		// getting destination path/mask?
		else if(!dstFound)
		{
			// wildcards at the end? remove the name mask
			if(IsWild(argv[i]))
			{
				CopyNamePart(szDestMask, argv[i]);
				RemoveName(argv[i]);
			}

			// try to create the path
			if(!MakeDirTree(argv[i]))
				Error(errMsgBadDest, argv[i], ERR_BADPATH);

			// if above call succeded, path MUST exist
			if(!dlExpandPath(argv[i], szDestPath))
				Error(errMsgBadDest, argv[i], ERR_BADPATH);

			dstFound = TRUE;
		}

		// too many arguments found
		else
			moreFound = TRUE;
	}

	// wrong # of args (too many or too few)
	if(!srcFound || moreFound)
	{
		printf(errMsgWrongNumber);
		printf(msgUsage);
		exit(ERR_BADOPT);
	}

	// no destination? assume current directory
	if(!dstFound)
	{
  		if(!dlExpandPath(".", szDestPath))
			ABORT();
	}

	// set up the root
	rootLength = strlen(szDestPath);
}

/*
	Display simple error message and exit with error code.
*/
void Error(const char *pMsg, char *pSubMsg, int errCode)
{
	// display error message
	printf(errMsg);
	printf(pMsg, pSubMsg);

	// free memory handles, in case we have XMS
	KillFileStore();

	// exit with code
	exit(errCode);
}

/*
	Display usage.
*/
void Usage(void)
{
	printf(msgUsage);
	exit(ERR_USAGE);
}


/*
	Is this file extension of the 3 digit variety? (eg .001)
*/
bool Is3Digit(char *pExt)
{
	int	numDigits = 0;

	// validate parameter
	ASSERT(pExt);

	// skip preceding dot
	if(*pExt == '.')
		++pExt;

	// look for 3 digits
	while(isdigit(*pExt))
	{
		++numDigits;
		++pExt;
	}

	// did we find 3 of them?
	return (numDigits == 3);
}

/*
	Get the next volume for the restore.  Will prompt the user
	for the next volume.  The user may quit by hitting ctrl-break.
*/
void GetNextVolume(void)
{
	// once we've opened the volume, never erase the name part
	static bool openedVolume = FALSE;

	char szExpectedExt[MAXEXT+1];
	char szOtherExt[MAXEXT+1];			// for out of sequence volumes
	char szPromptStr[MAXFILE+1];
	char szStr[32];						// for checking volume signature
	struct ffblk ff;
	int err;

	// validate parameter
#if DEBUG
	if(openedVolume)
		ASSERT(*szVolumeName);
#endif

	// opening a new volume, reset displayed name
	*szLastFile = 0;

	// file corrupt, or error occured opening file
Again:

	// what is the currently expected extension?
	sprintf(szExpectedExt, ".%03d", curVolume);

	// user MAY want to skip a volume or two, and go on to
	// another volume # than the one we expect.  Reset this
	// so we no user hasn't picked another yet
	*szOtherExt = 0;

	// have not yet opened a specific backup volume
	// so look for ANY datalight backup files
	if(!openedVolume)
		sprintf(szVolumeName, "%c:DL*.*", backupDrive+'A');

	// some white space for clarity
	printf("\n");

	while(1)
	{
		// ask user for the backup drive
		if(swPrompt)
		{
			// User HAS already chosen a specific volume set. use it
			if(openedVolume)
			{
				// set up the prompt string
				SetFileExt(szVolumeName, "");
				sprintf(szPromptStr,"%s ", szVolumeName+2);
			}

			// Insert Volume 1 of Backup Set DL970414 in Drive A:
			printf(msgInsertVolume, curVolume,
				(openedVolume ? szPromptStr : ""), backupDrive+'A');

			// and Press Return...
			if(*szOtherExt == 0)
				printf(msgReturn);

			// Or Press Return to Continue From This Point...
			else
				printf(msgOtherReturn);

			WaitForReturn();
		}

		// we've already got the backup set name, just set the ext
		if(openedVolume)
		{
			// try for an exact match
			SetFileExt(szVolumeName, szExpectedExt);
			err = findfirst(szVolumeName, &ff, A_NORMAL);

			// not found, maybe they put in a different Volume #
			if(err)
				SetFileExt(szVolumeName, "*");
		}

		// is there a Volume File on that drive?
		err = findfirst(szVolumeName, &ff, A_NORMAL);

		// something was found, is it the RIGHT thing?
		if(!err)
		{
			// found the right volume? go for it!
			if(strcmp(FindExt(ff.ff_name), szExpectedExt) == 0)
				break;

			// when you can't prompt, give up if we didn't find the
			// exact backup file!
			else if(!swPrompt)
			{
				// this will never return, just error exit out
				Error(msgNotFound, "", ERR_BADPATH);
			}

			// user MAY have selected a different volume.
			if(strcmp(FindExt(ff.ff_name), szOtherExt) == 0)
			{
				// we now have a NEW current volume
				curVolume = atoi(szOtherExt+1);

				break;
			}

			// user put in a different volume.  Should we use that?
			if(Is3Digit(FindExt(ff.ff_name)))
			{
				// copy the extension
				strcpy(szOtherExt, FindExt(ff.ff_name));

				// let user KNOW it's the wrong volume
				printf(msgOtherVolume, szOtherExt+1);
			}

			// doesn't have a numeric extension, so it's not a backup file!
			else
				printf(msgNotFound);
		}

		// didn't find any backup files on THAT disk
		else
			printf(msgNotFound);
	}

	// some white space for clarity
	printf("\n");

	// create the volume name
	sprintf(szVolumeName, "%c:%s", backupDrive+'A', ff.ff_name);

	// OK, now we have a filename, open it
	fpVolume = fopen(szVolumeName, "rb");
	if(!fpVolume)
	{
		printf(errMsgCantOpen, szVolumeName);
		if(swPrompt)
			goto Again;
	}

	// is this our standard backup file?
	if(fread(szStr, 1, sizeof(szStr), fpVolume) != sizeof(szVolumeSignature)
		|| strcmp(szStr, szVolumeSignature) != 0)
	{
		printf(errMsgCorrupted, szVolumeName);
		if(swPrompt)
			goto Again;
	}

	// how big is this first volume?
	volumeSize = DetermineFileSize(fpVolume);

	// volume position starts just after signature
	volumePos = ftell(fpVolume);

	// we have now opened a volume
	openedVolume = TRUE;
}

/*
	Validate the restore drive.  Sets up the initial
	Volume Info for the restore, including the volume name.
*/
void InitVolume(void)
{
	// determine info about the volume
	dlIsDriveRemovable( (uchar)(backupDrive+1), &volumeIsRemovable );

	// no need to prompt on non-removable volumes
	if(!volumeIsRemovable)
		swPrompt = FALSE;

	// start on volume 1 by default
	curVolume = 1;
}

// handle a File Record split across Memory Store reads
bool splitFile;
struct FileRecord splitFileRecord;
char szSplitFileName[MAXLONGPATH];
ulong extraHeaderSize;

/*
	Load File Record(s) from backup volume file into memory.
	Will fill memory, or end of volume, whichever comes first.

	Ignores any files that do not match the file mask.

	Returns TRUE if there is more data from this volume to read.
	Returns FALSE, and sets "moreVolumes" if this is the end
	of the file.
*/
bool LoadStore(bool *pMoreVolumes)
{
	struct FileRecord fileRecord;
	unsigned len, lenRead;
	ulong		memTop, curPos;
	bool		moreData;

	// validate parameters
	ASSERT(fpVolume);			// volume file must be open
	ASSERT(pMoreVolumes);

	// start with memory store empty
	MemStoreSeek(0L, SEEK_SET);

	// was a File Record split from the last LoadStore()?
	if(splitFile)
	{
#if DEBUG
	printf("split %s\n", szSplitFileName);
#endif
		// write adjusted File Record
		MemStoreWrite(&splitFileRecord, sizeof(splitFileRecord)-1);
		MemStoreWrite(szSplitFileName, splitFileRecord.nameSize);

		// we wrote in an extra header, don't record that in
		// volume position update
		extraHeaderSize = MemStoreTell();

		splitFile = FALSE;
	}

	// blast in as much from the disk as possible
	moreData = TRUE;
	while(moreData && MemStoreLeft())
	{
		// read in as much as we can from the file
		if(MemStoreLeft() < DISKIO_BUFFSIZE)
			len = (unsigned)MemStoreLeft();
		else
			len = DISKIO_BUFFSIZE;

		// read some bytes from the file
		lenRead = fread(FileStoreBuff(), 1, len, fpVolume);

		// at end of data stream
		if(lenRead < len)
			moreData = FALSE;

		// write them to the memory store
		// (must ALWAYS work because we set the length above)
		if(MemStoreWrite(FileStoreBuff(), lenRead) != lenRead)
			ABORT();
	}

	// remember top of memory
	memTop = MemStoreTell();

	// adjust all "next" fields in the file records
	fileRecord.next = 0L;
	while(fileRecord.next < memTop)
	{
		// seek to the next position
		curPos = fileRecord.next;
		MemStoreSeek(curPos, SEEK_SET);

		// not enough room for another record, stop here
		if(memTop == MemStoreSize() && memTop - curPos < MIN_OUT_SIZE)
		{
			memTop = MemStoreTell();
			*pMoreVolumes = moreData = TRUE;
			break;
		}

		// get next file record without moving position
		MemStoreRead(&fileRecord, sizeof(fileRecord)-1);
		MemStoreSeek(curPos, SEEK_SET);

		// on last record. we're done. also handle a bogus input file
		if(!fileRecord.next || fileRecord.sig != FILE_RECORD_SIG)
		{
			memTop = MemStoreTell();
			*pMoreVolumes = moreData = FALSE;
			break;
		}

#if DEBUG
	MemStoreSeek(curPos + sizeof(fileRecord)-1, SEEK_SET);
	MemStoreRead(szSplitFileName, fileRecord.nameSize);
	printf("sSplitFileName name %s\n", szSplitFileName);
	MemStoreSeek(curPos, SEEK_SET);
#endif

		// modify next field to be memory relative, not volume relative
		fileRecord.next = fileRecord.next - volumePos + extraHeaderSize;

		// a record is split across the in-memory pool
		if(fileRecord.next > memTop)
		{
			// record ends (splits) at top of memory
			fileRecord.next = memTop;

			// remember the split info, put leave ptr in place
			MemStoreRead(&splitFileRecord, sizeof(fileRecord)-1);
			MemStoreRead(szSplitFileName, splitFileRecord.nameSize);
			MemStoreSeek(curPos, SEEK_SET);

			// we are no longer starting at the same file offset
			splitFileRecord.startOffset += memTop -
				(curPos + sizeof(fileRecord)-1 + splitFileRecord.nameSize);

			// we now have a split file
			splitFile = TRUE;
		}

		// write out updated file record
		MemStoreWrite(&fileRecord, sizeof(fileRecord)-1);
	}

	// communicate memory top
	MemStoreSeek(memTop, SEEK_SET);

	// volume pos ptr has moved ahead by so many bytes
	volumePos += memTop - extraHeaderSize;
	extraHeaderSize = 0L;

	// seek to that position
	fseek(fpVolume, volumePos, SEEK_SET);

	// there MAY be more data
	return moreData;
}

/*
	Get a single key from DOS.  This will be used for a
	Yes/No style question.
*/
char getkey(char *chars)
{
	char c;

	// must expect SOMETHING
	ASSERT(chars);

	while(1)
	{
		c = bdos(8,0,0);
		c = toupper(c);
		if(!chars || strchr(chars, c))
			break;
	}
	printf("%c\n",c);
	return c;
}

/*
	Return length of path part (but not name).
	Returns 0 if no path part of string.

	Example: "C:\TEMP\NAME.EXT" would return 7.
*/
int DeterminePathLen(char szPath[])
{
	char *pNamePart;

	// look for just the name part
	pNamePart = IntlStrRChr(szPath, '\\');
	if(!pNamePart)
		return 0;		// no path!

	// return length of root string.
	return (pNamePart - szPath);
}

/*
	Flush this file record to disk.
	Displays error message if couldn't open or create the file.

	Returns FALSE if disk is full.
*/
bool FlushFileRecord(struct FileRecord *pFileRecord, char szDestPath[])
{
	FILE   *	fp;
	unsigned len;
	unsigned lenWritten;
	bool	worked = TRUE;

	// validate parameters
	ASSERT(pFileRecord);
	ASSERT(szDestPath);

	// create subdirectories as needed for the files
	len = DeterminePathLen(szDestPath);
	ASSERT(len);
	if(len != lastPathLen || strncmp(szLastPath, szDestPath, len) != 0)
	{
		// remember this new tree
		memmove(szLastPath, szDestPath, len);
		szLastPath[len] = 0;
		lastPathLen = len;

		// create the new tree
		MakeDirTree(szLastPath);
	}

	// create if at start of file
	if(!pFileRecord->startOffset)
	{
		// create the file
		fp = fopen(szDestPath, "wb");
	}

	// or append if writing to the middle or end of the file
	else
	{
		fp = fopen(szDestPath, "ab");
		if(!fp)
		{
			fseek(fp, pFileRecord->startOffset, SEEK_SET);
		}
	}

	// did we create/open it?
	if(!fp)
	{
		// display error message!
		Error(errMsgCantCreate, szDestPath, ERR_BADPATH);
		return FALSE;
	}

	// update file totals
	if(!pFileRecord->startOffset)
	{
		// update totals
		++totalFiles;
		totalBytes += pFileRecord->size;
	}

	// write out the entire record
	while(worked && MemStoreTell() < pFileRecord->next)
	{
		// write as much as we can to the file
		if(pFileRecord->next - MemStoreTell() < DISKIO_BUFFSIZE)
			len = (unsigned)(pFileRecord->next - MemStoreTell());
		else
			len = DISKIO_BUFFSIZE;

		// read some bytes from the store
		if(MemStoreRead(FileStoreBuff(), len) != len)
			ABORT();

		// write it to the file
		lenWritten = fwrite(FileStoreBuff(), 1, len, fp);

		// handle case where disk is full
		if(lenWritten != len)
			worked = FALSE;
	}

	// flush the file buffer, so our date sticks
	fflush(fp);

	// update date/time to reflect file date/time
	if(!SetFileDate(fp->fd, pFileRecord->date, pFileRecord->time))
		ABORT();

	// close the file
	fclose(fp);

	// update attributes to match file attributes
	if(!dlSetFileAttributes(pFileRecord->attr, szDestPath))
		ABORT();

	// failed to write all of file! dest drive must be full.
	if(!worked)
		Error(errMsgDestFull, (char *)(*szDestPath), ERR_BADPATH);

	// did the write work?
	return worked;
}

/*
	Flush files from the in-memory store to disk (or screen).
	Will report errors if it can't create the files, or disk is full.

	Returns FALSE if we should quit (disk full, on to subdirs with
	no /S switch).
*/
bool FlushStore(char szRootPath[], char szMask[])
{
	struct FileRecord fileRecord;
	long	curPos, memTop;
	char	*pFileName;
	char	c;
	bool	needCR = FALSE;
	unsigned	fileAttr;

	// validate parameter(s)
	ASSERT(szRootPath);
	ASSERT(szMask);

	// set once on init
	ASSERT(rootLength);

	// determine how much is in the memory pool
	memTop = MemStoreTell();
	ASSERT(memTop);		// must be SOMETHING!

	curPos = 0L;
	while(curPos < memTop)
	{
		MemStoreSeek(curPos, SEEK_SET);

		// read in file record and name
		MemStoreRead(&fileRecord, sizeof(fileRecord)-1);

		// file record had better be valid
		ASSERT(fileRecord.sig == FILE_RECORD_SIG);
		ASSERT(fileRecord.nameSize);

		// set up a buffer for the filename
		pFileName = FileStoreBuff();

		// read in file name (must be NULL terminated)
		MemStoreRead(pFileName, fileRecord.nameSize);
		ASSERT(*pFileName && *(pFileName + fileRecord.nameSize - 1) == 0);

		// is there a PATH part in this name? we're done!
		if( !swSubdir )
		{
			pFileName = LastPart(pFileName);
		}

#if DEBUG
		printf("MatchMask(%s,%s)\n", szMask, pFileName);
#endif
		// only restore files that match
		if(MatchMask(szMask, pFileName))
		{
			// assume it's OK to flush the file records
			if(fileRecord.startOffset == 0L)
				ok2Flush = TRUE;

			// create absolute destination file name
			AddName(szRootPath, pFileName);

			// display the file as it is restored (but only display once,
			// even if file spans multiple disks/records)
			if(strcmp(szLastFile, szRootPath))
			{
				// display the file, only once
				printf("%s",szRootPath);
				strcpy(szLastFile, szRootPath);

				// we will need to terminate line with CR/LF
				needCR = TRUE;
			}

			// Do we need to prompt because we are overwriting?
			// GetFileAttr will both check if file exists, AND get
			// attr for later processing. No need for overwrite
			// prompting if file doesn't exist.
			if(dlGetFileAttributes(szRootPath, &fileAttr))
			{
				// prompt overwrite a file?
				// Only prompt if:
				//   1) At start of file (not a record in the middle)
				//   2) We are prompting at all
				//   3) /P and overwriting rdonly or non-archived
				if(fileRecord.startOffset == 0L &&
					swPrompt &&
					(!swNewerPrompt || (fileAttr & A_OVERWRITE) != 0))
				{
					// ask if user wants to overwrite!
					printf(msgOverWrite);
					c = getkey(overChars);

					// no need to CR/LF, we already did
					needCR = FALSE;

					// OK to overwrite all of 'em
					if(c == 'A')
						swPrompt = FALSE;

					// no, don't overwrite this one!
					else if(c == 'N')
						ok2Flush = FALSE;
				}

				// just in case file was read-only
				if(ok2Flush)
					dlSetFileAttributes(0,szDestPath);
			}

			// no longer need a CR
			if(needCR)
			{
				printf("\n");
				needCR = FALSE;
			}

			// flush this file record to disk
			if(ok2Flush && !FlushFileRecord(&fileRecord, szRootPath))
				return FALSE;

			// truncate the name back off of dest path
			szRootPath[rootLength] = 0;
		}

		// seek to next position
		curPos = fileRecord.next;
	}

	// it all worked
	return TRUE;
}
/*
	Displays a single file record, for use with the /R switch
*/
void ShowRaw(ulong volumePos, struct FileRecord *pFileRecord, char szFileName[])
{
	char	szDate[DATESIZE];
	char	szTime[DATESIZE];

	// determine date and itme
	DateToAscii(pFileRecord->date, szDate);
	TimeToAscii(pFileRecord->time, szTime);

	// show raw header start, size, starting offset
	printf("%08lX %-32s %,7ld %s %s (%,7ld)\n",volumePos,szFileName,
		pFileRecord->size, szDate, szTime, pFileRecord->startOffset);
}

/*
	Displays a single file, for use with /D switch.
*/
void ShowFile(struct FileRecord *pFileRecord, char szFileName[])
{
	char	szDate[DATESIZE];
	char	szTime[DATESIZE];

	// determine date and itme
	DateToAscii(pFileRecord->date, szDate);
	TimeToAscii(pFileRecord->time, szTime);

	// display the name
	printf(msgLogLine, szFileName, pFileRecord->size, szDate, szTime);

	// was file continued from previous volume?
	if(pFileRecord->startOffset && pFileRecord->startVol != curVolume)
		printf(msgContinued, pFileRecord->startVol);

	// done with line
	printf("\n");
}

/*
	Display all the files on this volume.

	Returns TRUE if there are more volumes to process.
*/
bool DisplayFiles(char szMask[])
{
	struct FileRecord fileRecord;
	bool	 moreVolumes;
	char * pFileName;

	// validate parameters
	ASSERT(szMask);
	ASSERT(fpVolume);			// volume file must be open
	ASSERT(volumePos);
	ASSERT(volumeSize);
	ASSERT(FileStoreBuff());

	// get a place we can store a filename
	pFileName = FileStoreBuff();

	// assume more volumes to come
	moreVolumes = TRUE;

	// read in the entire volume
	while(volumePos < volumeSize)
	{
		// read in the file record. MUST be able to read all of it
		if(fread(&fileRecord, 1, sizeof(fileRecord)-1, fpVolume) !=
					sizeof(fileRecord)-1)
			break;

		// sanity check: record signature must be valid
		if(fileRecord.sig != FILE_RECORD_SIG)
			break;

		// end of Backup record, we're done, no more volumes
		if(!fileRecord.next)
		{
			moreVolumes = FALSE;
			break;
		}

		// sanity check: next should always be greater than current position
		if(fileRecord.next <= volumePos)
			break;

		// read in file name. MUST be able to read all of it
		if(fread(pFileName, 1, fileRecord.nameSize, fpVolume) !=
					fileRecord.nameSize)
			break;

		// show the file, if appropriate
		if(MatchMask(szMask, pFileName))
		{
			// update file totals (we may have multiple records
			// for the same file)
			if(fileRecord.startOffset == 0L)
			{
				++totalFiles;
				totalBytes += fileRecord.size;
			}

			// just show raw records
			if(swRaw)
				ShowRaw(volumePos, &fileRecord, pFileName);

			// only show the file once per volume
			else if(strcmp(szLastFile, pFileName))
			{
				ShowFile(&fileRecord, pFileName);
				strcpy(szLastFile, pFileName);
			}
		}

		// on to next entry
		volumePos = fileRecord.next;
		fseek(fpVolume, volumePos, SEEK_SET);
	}

	// are there more volumes to process?
	return moreVolumes;
}

/*
	Restore a single volume.  This might entail many files
	and directories.

	Returns TRUE if there are more volumes to restore.
*/
bool RestoreVolume(char szRootPath[], char szMask[])
{
	bool	moreVolumes;
	bool	moreData;

	// validate parameters
	ASSERT(szRootPath);
	ASSERT(szMask);

	// display files that match the mask
	if(swDir)
		moreVolumes = DisplayFiles(szMask);

	// read volume in as large of chunks as possible
	// restores only filenames that match the mask
	else
	{
		moreData = TRUE;
		while(moreData)
		{
			// load the in-memory store with as many file records as we can
			moreData = LoadStore(&moreVolumes);

			// flush the in-memory store to disk (or screen)
			if(!FlushStore(szRootPath, szMask))
				moreVolumes = moreData = FALSE;
		}
	}

	// are we done yet?
	return moreVolumes;
}

/*
	Restore one or more files from a backup set.  It simply
	reads through the backup disks, searching for matching
	files, then calls "RestoreFile" for each file.
*/
void Restore(char szRootPath[], char szMask[])
{
	static bool firstTime = TRUE;

	ulong	diskFreeSpace;
	bool moreVolumes;

	// we will keep track of file totals
	totalFiles = totalBytes = 0L;

	// for restore path purposes
	lastPathLen = 0;
	*szLastPath = 0;

	// restore each volume
	while(1)
	{
		// get the next volume
		GetNextVolume();

		// show header for the first volume
		if(firstTime)
		{
			if(swDir)
				printf(msgDir, szVolumeName, szMask);
			else
				printf(msgRestoring, szMask, szVolumeName, szRootPath);
			firstTime = FALSE;
		}

		// keep restoring volumes until we are done!
		moreVolumes = RestoreVolume(szRootPath, szMask);
		if(!moreVolumes)
			break;

		// on to next volume
		++curVolume;

		//  1-24-2000 BT  ETR 3218
		//
		//  Close this volume before we go on to the next one or
		//  we will run out of handles.
		//
		fclose(fpVolume);
	}

	// display totals. dest drive has already been validated,
	// so we had BETTER be able to read the free space
	if(!dlGetFreeSpaceOnDrive(*szDestPath-'A' + 1, &diskFreeSpace))
		ABORT();
	printf(msgStatistics, totalFiles, totalBytes, curVolume);
	printf(msgBytesFree, diskFreeSpace);
}
