//
//   $Id: PRINT.C 1.6 2002/10/15 23:53:04Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	PRINT.C

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
	Datalight Confidential and Proprietary

	This file is the PRINT interrupt handler - this part
	stays resident.  Lower level routines pass control to
	these routines upon PRINT interrupts.

**********************************************************/
#include <stdio.h>
#include <dos.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"

#include "print.h"

extern int crit_err;

int column = 0;
struct prdata dat;
struct prdata * pr = &dat;

/* PROTO  Version 1.00 */

/* File PRINT.C */
void spooler(void);
dos_open(char *name, unsigned mode);
dos_read(unsigned handle, char *buffer, unsigned size);
dos_write(unsigned handle, char *buffer, unsigned size);
dos_lseek(unsigned handle, unsigned long offset, unsigned mode);
int is_busy(unsigned handle);
void memcpy(void *dst, void *src, unsigned size);
void form_feed(void);
int dos_break(int get_set, int flag);
int cmp(char far *fp, char *p);

#include "print.pt"

/*
	This handles all interaction between a resident PRINT
	and a PRINT that is adding/deleting files from the
	queue.
*/
#pragma argsused
void interrupt far int2f(BP, DI, SI, DS, ES, DX, CX, BX, AX, IP, CS, FLAGS)
{
	int i, j, total_files=0;
	char far *fp;
	struct sub_packet far *s;

	switch(AX & 0xFF)
	{

		/* Install Check */
		case 0:
			AX = 0x01FF;	/* Return FF = Already installed */
			break;

		/* Submit file "DS:DX" */
		case 1:
			for(i=0; i<pr->max_files; ++i)
			{
				if(pr->files[i][0] == '\0')
					break;
			}
			if(i < pr->max_files)
			{
				s = MK_FP(DS, DX);
				fp = s->file;
				for(j=0; j < NAMESIZE; ++j)
					pr->files[i][j] = fp[j];
			}
			else
				AX = QUEUE_FULL;
			break;

		/* Remove File DS:DX */
		case 2:
			fp = MK_FP(DS, DX);
			for(i=0; i<pr->max_files; ++i)
			{
				if(pr->files[i][0])
					++total_files;
				else
					break;
			}
			for(i=0; i<total_files; ++i)
			{
				/* First file */
				if(cmp(fp, pr->files[i]))
				{
					if(i == 0)
					{
						form_feed();
						pr->total = pr->count = 0;
						memcpy(&pr->files[0], &pr->files[1],
						(total_files)*sizeof(pr->files[1]));
					}

					/* Last file */
					else if(i+1 == pr->max_files)
						pr->files[i][0] = '\0';

					/* Any Other file */
					else
					{
						memcpy(&pr->files[i], &pr->files[i+1],
						(total_files-i)*sizeof(pr->files[1]));
					}
				}
			}
			break;

		/* Cancel all files */
		case 3:
			pr->print_errors = pr->count = pr->total = 0;
			for(i=0; i<pr->max_files; ++i)
				pr->files[i][0] = '\0';
			break;

		/* Hold print Jobs */
		case 4:
			pr->on_hold = TRUE;
			DX = pr->print_errors;
			SI = (unsigned) &pr->files;
			DS = getDS();
			break;

		/* End hold */
		case 5:
			pr->on_hold = FALSE;
			break;

		/* Get "print" structure */
		case 6:
			AX = getDS();
			BX = (unsigned) pr;
			break;
	}
}


/*
	General Spooling function

	This function is called at each int 28H and each
	interrupt 1CH.  Will not be called if on hold.
*/
void spooler(void)
{
	int written;
	int ret;
	int break_flag;
	unsigned char c;
	unsigned char ucModeData = 0;

	extern int errno, overflow;

	void far * old_int24;
	void print_crit_err();

	/* set Critical error handler to our handler */
	old_int24 = get_vector(0x24);
	set_vector(0x24, MK_FP(getCS(), print_crit_err));

	/* Disable DOS break flag */
	break_flag = dos_break(0, 0);		/* remember old break flag */
	dos_break(1, 0);

#if DEBUG
	/* detect stack overflow */
	if(overflow != 0x55AA)
		printf("Overflow = %04X\n", overflow);
#endif

	/* If data in buffer */
	if(pr->count < pr->total)
	{
		pr->outfile = dos_open(pr->device, 1);

      // IR 3364 9-7-01 TD
      //
      // All PRINTing should be done with the printer in RAW mode
      // (otherwise a printing will be halted with a CTRL-Z).
      //
      ucModeData = dos_ioctl( pr->outfile,0,0 );
      if( ucModeData & 0x80 )       // if we're a device
      {
         dos_ioctl( pr->outfile,1,ucModeData|0x20 );
      }

		/* problem opening the printer! */
		if(pr->outfile == 0xFFFF)
			return;

		written = 0;
		while(pr->count < pr->total && !is_busy(pr->outfile))
		{
			c = *(pr->buffer+pr->count);

			/* expand tabs into spaces */
			if(c == '\t')
			{
				c = ' ';
				while((column+1) % 8)
				{
					ret = dos_write(pr->outfile,(char *)(&c),1);
					if(ret != 1)
						break;
					++written;
					++column;
				}
			}
			ret = dos_write(pr->outfile,(char *)(&c),1);

			/* was there an error? */
			if(ret != 1)
			{
				pr->print_errors = TRUE;
				column = 0;
				break;
			}

			/* adjust the column */
			if(c != '\n')
				++column;
			if(c == '\r' || column >=80)
				column = 0;

			/* no errors, no problem */
			pr->print_errors = FALSE;
			++pr->count;
			++written;

			/* don't write too many bytes on each tick */
			if(written > WRITE_SIZE)
				break;
		}

		if(crit_err)
			pr->print_errors = TRUE;

      // IR 3364 9-7-01 TD
      //
      // If we saved off a value before the write, we need to restore it
      //
      if( ucModeData )
      {
         dos_ioctl( pr->outfile,1,ucModeData );
      }
		dlCloseFile(pr->outfile);
	}

	/* If a file to read from */
	else if(pr->files[0][0] != '\0')
	{
		pr->infile = dos_open(pr->files[0], 0);

		/* could not open the file for input, error */
		if(pr->infile == 0xFFFF)
			goto err;

		/* read some bytes from the file */
		dos_lseek(pr->infile, pr->offset, 0);
		pr->total = dos_read(pr->infile, pr->buffer, pr->buffer_size);
		dlCloseFile(pr->infile);

		if(pr->total == 0)
		{
			/* on to the next file */
err:
			memcpy(&pr->files[0], &pr->files[1],
				(pr->max_files-1)*sizeof(pr->files[1]));
			form_feed();
			pr->files[9][0] = '\0';
			pr->offset = 0;
		}

		pr->count = 0;
		pr->offset += pr->total;
	}

	else
		/* Idle */;

	/* Restore DOS break flag and critical error handler */
	set_vector(0x24, old_int24);
	dos_break(1, break_flag);
}

extern errno; 					/* Defined in C0S.OBJ */

/*
	Open this file/device in binary mode.  Returns handle
	or -1 on error.
*/
dos_open(char *name, unsigned mode)
{
	int handle;
	bool fSuccessful;
	unsigned uDevAttributes;


	errno=0;
	fSuccessful = dlOpenFile( name, (uchar)mode, &handle );
	if( !fSuccessful )
	{
		errno = dlCheckDOSError();
		handle = -1;
	}
	else
	{
		/* if this is a device, open it in RAW mode */
		fSuccessful = dlGetDeviceAttributes( handle, &uDevAttributes );
		if( (fSuccessful) && (uDevAttributes | DEVA_QUERY) )
		{
			uDevAttributes |= DEVA_RAWMODE;
//			uDevAttributes &= 0xff;
			dlSetDeviceAttributes( handle, uDevAttributes );
		}
	}

	return (handle);
}

dos_read(unsigned handle, char *buffer, unsigned size)
{
	bool fSuccessful;
	unsigned uByteCount;


	uByteCount = size;
	fSuccessful = dlReadFile( handle, &uByteCount, buffer );
	if( !fSuccessful )
	{
		errno = dlCheckDOSError();
		return -1;
	}
	else
	{
		return uByteCount;
	}
}

dos_write(unsigned handle, char *buffer, unsigned size)
{
	bool fSuccessful;
	unsigned uByteCount;


	uByteCount = size;
	fSuccessful = dlWriteFile( handle, &uByteCount, buffer );
	if( !fSuccessful )
	{
		errno = dlCheckDOSError();
		return -1;
	}
	else
	{
		return uByteCount;
	}
}


dos_lseek(unsigned handle, unsigned long offset, unsigned mode)
{
	bool fSuccessful;
	long lDistance;


	lDistance = offset;
	fSuccessful = dlSeekFile( mode, handle, &lDistance );
	if( !fSuccessful )
	{
		errno = dlCheckDOSError();
		return -1;
	}
	else
	{
		return 0;
	}
}

unsigned busy_ret;

/*
	Determine if the device in question is busy or not.
	Returns FALSE if not busy.
*/
int is_busy(unsigned handle)
{
	unsigned flags;
	extern int crit_err;

	crit_err = 0;				/* Flag no critical error! */

	_BX = handle;
	_AX = 0x4407;
	geninterrupt(0x21);
	busy_ret = _AL;
	flags = _FLAGS;
	if( (flags & CARRY) || crit_err )
	{
		errno = busy_ret;
		return -1;
	}

	busy_ret = !busy_ret;
	return busy_ret;
}

void memcpy(void *dst, void *src, unsigned size)
{
	char *p1 = dst, *p2 = src;
	int i;
	for(i=0; i<size; ++i)
		p1[i] = p2[i];
}

void form_feed(void)
{
	char formfeed = '\f';
	int prn;
	prn = dos_open(pr->device, 1);
	dos_write(prn, &formfeed, 1);
	dlCloseFile(prn);

	column = 0;
}

int dos_break(int get_set, int flag)
{
	_AX = 0x3300 | get_set;
	_DX = flag;
	geninterrupt(0x21);
	return(_DX & 0xFF);
}


int cmp(char far *fp, char *p)
{
	int ret = TRUE;
	do
	{
		if(*fp != *p)
			ret = FALSE;
		++p;
		++fp;
	} while(*p);
	return (ret);
}

// Generic IOCTL routine, block copied from ANSI
//
int dos_ioctl(int handle, int mode, unsigned setvalue)
{
	_AX = 0x4400 + mode;
   _BX = handle;
   _DX = setvalue;
	geninterrupt(0x21);

	return (_DX);
}
