//
//   $Id: RXE_CVT.C 1.8 2002/10/16 15:53:20Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:RXE_CVT.C

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary


	Convert .EXE programs into .RXE programs

	Syntax
		RXE_CVT {Options} <InFile> <DataSegName> <MemorySeg> [OutFile]
	Options
		/C              Continue with errors
		/D              Prepair the MAP file for RDEB debugging
		/I<HexNumber>   Set the RXE interrupt
		/Q              Operate in Quiet mode
		/R              Rename output file to <FileName>.RXE
		/S              Get RXE file size
		/W              Display warning messages
		/L              Lower Case is significant in segment names

	$DOCHISTORY
	10/26/2000 de  Changed command line options to always display warnings
								 and to disallow "quit" mode.
	08/06/1998 dg  Fixed bug
*********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include <dos.h>
#include <fcntl.h>
#include <sys\stat.h>

#include <sys\types.h>
#include <sys\stat.h>

#include "datalght.h"
#include "dostruct.h"
#include "rxe.h"
#include "dosver.h"
#include "exehdr.h"


char USAGE[] =
	"Overview\n"
	"   Convert an .EXE program into a ROMable executable (.RXE) program\n"
	"Syntax\n"
	"   RXE_CVT {Options} <InFile> <DataSegName> <MemorySeg> [OutFile]\n"
	"Options\n"
	"   /C              Continue with errors\n"
	"   /D              Prepair the MAP file for RDEB debugging\n"
	"   /I<HexNumber>   Set the RXE interrupt\n"
	"   /Q              Operate in Quiet mode\n"
	"   /R              Rename output file to <FileName>.RXE\n"
	"   /S              Get RXE file size\n"
	"   /W              Display warning messages\n"
	"   /L              Lower Case is significant in segment names\n"
	"\n"
	"Example:\n"
	"   RXE_CVT /C /I45 args.exe data D014\n"
	"\n"
	"   This example converts ARGS.EXE to ARGS.RXE. The Code block ends\n"
	"   at segment DATA and it will be located in ROM at D014:0. The RXE\n"
	"   interrupt is 45h";

char rxe_name[66];		/* Output File Name */
char exe_name[66];		/* Input file name */

void dump(void *v, unsigned count);

void error(int msg, ...), *malloc(unsigned);

int useLowerCase = FALSE;

int output;

int continue_on_error = FALSE;

int quit = FALSE;

int exit_status;

int for_debugging = FALSE;

int do_rename = FALSE;

int custom_name = FALSE;

int do_getsize = FALSE;

int VersionOnly = FALSE;

// 10/26/2000 - DE
// Warnings are now on by default (well, always now, actually).
// This is required to alert the user to ambiguous fixups that
// are encountered by RXE and which, if wrongly applied, may make
// the user's program crash in any number of mysterious ways.
//
int DisplayWarningMessages = TRUE;

extern char map_name[66];

int rxe_int = 0x18;		/* Default set to BASIC interrupt */

char *rxe_cvt_name;

FILE *RXE_output;			// Output for RXE Error

char *RXERootPath;		// for Root of RXE_CVT.EXE

/*
 * RXE_CVT Main function
 * Exit Codes
 *		0 = Success
 *		Any Other is error Message #
 */
void main(int argc, char *argv[])
{
	int i;
	char *OutFile="CON";
	unsigned d_seg, m_seg, tmp;
	extern unsigned rxe_base_paras;
	extern char found_seg_name[], found_class_name[];

	for(i=1; i<argc; ++i){
		if(argv[i][0] == '/' || argv[i][0] == '-'){
			while(argv[i][0] == '/' || argv[i][0] == '-'){
				switch(argv[i][1]){
					case 'C':
					case 'c':
//  06-20-00  BT
//
//  No longer allow continue on errors because there is no distinction
//  between errors you can continue with and critical error like error
//  allocating memory.  Allow the switch however.
//
//						continue_on_error = TRUE;
						argv[i] += 2;
						continue;

					case 'D':
					case 'd':
						for_debugging = TRUE;
						argv[i] += 2;
						continue;

					case 'E':
					case 'e':
						OutFile = argv[i]+2;
						argv[i] += 2 + strlen(argv[i] + 2);
						continue;

					case 'i':
					case 'I':
						sscanf(argv[i] + 2, "%x", &rxe_int);
						if(rxe_int < 7 || (rxe_int >= 20 &&
								rxe_int <= 0x3F))
							error(16);
						argv[i] += 2;
						continue;

					case 'Q':
					case 'q':
//  10/20/2000  DE
//
//  No longer allow "quit" mode because we need users to see the errors.
//
//						quit = TRUE;
						argv[i] += 2;
						continue;

					case 'R':
					case 'r':
						do_rename = TRUE;
						argv[i] += 2;
						continue;

					case 'S':
					case 's':
						do_getsize = TRUE;
						argv[i] += 2;
						continue;

					case 'V':
					case 'v':
						VersionOnly = TRUE;
						argv[i] += 2;
						continue;

					case 'W':
					case 'w':
						DisplayWarningMessages = TRUE;
						argv[i] += 2;
						continue;

					case 'L':
					case 'l':
						useLowerCase = TRUE;
						argv[i] += 2;
						continue;

					case '?':
					case 'H':
					case 'h':
						goto do_help;

					default:
						continue_on_error = FALSE;
						error(10, argv[i]);
				}
			}
		}
		else
			break;
	}

	RXERootPath = argv[0];			// Save Root path!

	RXE_output = fopen(OutFile, "w");

	if(!quit){
do_help:
		printf("Datalight ROMable EXE Converter\nVersion %s\n", FULLVERSION);
		printf("%s\n\n", COPYRIGHT);

		if(VersionOnly)
			exit(0);

		if(argc < 4){
			printf(USAGE);
			exit(1);
		}
	}

	strupr(argv[i]);

	strcpy(exe_name, argv[i++]);

	tmp = find_segment_value(exe_name, "RXE_BASE");
	if(tmp != 0xFFFF){
		error(15);
		exit(2);			/* In case /c option */
	}

	if(!isdigit(argv[i][0])){		/* Get memory segment */
		if(!useLowerCase)
			strupr(argv[i]);			/* Make the name upper case */
		d_seg = find_segment_value(exe_name, argv[i]);
		if(d_seg == 0xFFFF)
			error(9, argv[i], map_name);
	}
	else
		sscanf(argv[i], "%x", &d_seg);

	++i;

	sscanf(argv[i], "%x", &m_seg);

	++i;
	if(argc > i){					/* If they specified a optional Output file Name */
		custom_name = TRUE;
		strcpy(rxe_name, argv[i]);
		if(no_ext(rxe_name))
			mk_ext(rxe_name, "RXE");
	}
	else {
		strcpy(rxe_name, exe_name);
		mk_ext(rxe_name, "RXE");
	}

	if(no_ext(exe_name))
		mk_ext(exe_name, "EXE");

	_fmode = O_BINARY;
	output = creat(rxe_name, S_IWRITE);

	if(!quit)
		printf("Converting %s to %s with:\n\tData Block at Seg \"%s\" Class \"%s\"  %04X\n\tMemory seg = %04X\n\n",
			exe_name, rxe_name, found_seg_name, found_class_name, d_seg, m_seg);

	convert_exe_to_rxe(exe_name, (unsigned)d_seg, (unsigned)m_seg, rxe_int);
	close(output);
	if(mk_same_date(exe_name, rxe_name) != 0)
		assert(0);

	if(for_debugging)
		update_map(map_name, rxe_base_paras, d_seg);
											/* Add rxe_base to MAP */

	if(!custom_name){
		if(!do_rename){
			unlink(exe_name);
			rename(rxe_name, exe_name);
		}
	}

	fclose( RXE_output );

	exit(0);
}


/*
 * Make the file "f1" and file "f2" have the same date
 *
 */
int mk_same_date(char *f1, char *f2)
{
	int fd1, fd2;
	union REGS regs;
	fd1 = open(f1, O_RDONLY | O_BINARY);
	fd2 = open(f2, O_RDWR   | O_BINARY);
	if(fd1 == -1 || fd2 == -1)
		return -1;
	regs.x.ax = 0x5700;			/* Get File Date */
	regs.x.bx = fd1;
	intdos(&regs, &regs);
	if(regs.x.flags & 1)
		return -1;
	regs.x.ax = 0x5701;			/* Set file Date */
	regs.x.bx = fd2;
	intdos(&regs, &regs);
	if(regs.x.flags & 1)
		return -1;

   close( fd1 );
   close( fd2 );

	return (exit_status);		/* May have been an Error */
}

