//
//   $Id: rxeoptim.C 1.6 2002/10/16 16:02:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  RXEOPTIM.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//  07-19-2000  bt  Created
//


//  Enable debug information
//
#define RXEOPTIM_DEBUG		FALSE


//  Borland libraries
//
#include <dir.h>
#include <stdio.h>
#include <string.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dosver.h"
#include "dosmsg.c"
#include "fatworld.h"
#include "dlapp.h"
#include "dllfn.h"


//  HEADER FILES
//
#include "RXEOPTIM.h"


//  PROTOTYPES
//
#include "RXEOPTIM.pt"
char * LastPart( char * szFileName );
char * FindExt(char szFileName[]);


//
//  [ VARAIABLES ]
//

bool fIsFat32World = FALSE;

//  Flags for the command line switches
//  (accessed through macros in TREE.H)
//
bool fSwitches[MAX_BOOLEAN_SWITCHES];


//  Start of RXEOPTIM
//
uchar main( int argc, char *argv[] )
{
	uchar ucExitCode = ERR_NONE;
	static char szNewDrivePath[LFN_MAX_PATH];
	static char szCurrentDrivePath[LFN_MAX_PATH];
	char * szPath;
	bool fSuccessful;
	ulong ulNewFileSize = 0;
	ulong ulOldFileSize = 0;
	ulong ulPercent;
	unsigned uDelta = 0;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	//  Parse the command line
	//
	ucExitCode = DoParseCommandLine( argc-1, &argv[1] );
	if( ucExitCode == ERR_NONE )
	{
		//  Determine if we should display the signon message and if
		//  so display it
		//
		DisplaySignonMessage();


		//  Get the file size for the file
		//
		szPath = GetRXEPath();
		if( !GetFileSize(szPath, &ulOldFileSize) )
		{
			//  Display an error message stating that the path was invalid
			//  and exit
			//
			DisplayError( ERR_INVALID_PATH );
			return ERR_INVALID_PATH;
		}


		//  Get the current drive and directory
		//
//		SmartExpandPath( 0, ".", szCurrentDrivePath );
		dlGetCurrentDrive( (int *)(&szCurrentDrivePath[0]) );
		szCurrentDrivePath[0] += 'A';
		szCurrentDrivePath[1] = ':';
		szCurrentDrivePath[2] = '\\';
		szCurrentDrivePath[3] = '0';
		SmartGetCurrentDirectory( 0, &szCurrentDrivePath[3] );


		//  Build a string which contains only the drive and directory
		//  for the source file
		//
		strcpy( szNewDrivePath, szPath );
		szPath = LastPart( szNewDrivePath );
		if( *(szPath-1) == '\\' )
		{
			if( *(szPath-2) != ':' )
			{
				szPath--;
			}
		}
		*szPath = 0;


		//  Change to the appropriate drive and path
		//
		fSuccessful = SmartChangeDirectory( szNewDrivePath );
		if( !fSuccessful )
		{
			//  Display an error message stating that the path was invalid
			//  and exit
			//
			DisplayError( ERR_INVALID_PATH );
			return ERR_INVALID_PATH;
		}
		dlSetCurrentDrive( *szNewDrivePath - 'A' );


		//  If this is an RXE file, we will need to replace any fixups
		//  that were not replaced with interrupts.
		//
		ucExitCode = ReplaceImmediateFixups( &uDelta );
		if( ucExitCode != ERR_NONE )
		{
			DisplayError( ucExitCode );
			return ucExitCode;
		}


		//  Create and prepend the summary for the test log
		//
		ucExitCode = OptimizeEXESize( uDelta );


		//  Display a summary of the results if we were successful
		//
		if( ucExitCode == ERR_NONE )
		{
			szPath = GetRXEPath();
			if( GetFileSize( szPath, &ulNewFileSize ) )
			{
				printf( "Previous file size: %10lu\n", ulOldFileSize );
				if( ulNewFileSize == 0 )
				{
					ulPercent = 0;
				}
				else
				{
					ulPercent  = ulNewFileSize * 100;
					ulPercent /= ulOldFileSize;
				}
				printf( "New file size:      %10lu", ulNewFileSize );
				printf( " (%u%% of original size)\n", (unsigned)ulPercent );
			}
		}


		//  Restore the drive and path
		//
		dlSetCurrentDrive( *szCurrentDrivePath - 'A' );
		SmartChangeDirectory( szCurrentDrivePath );
	}


	if( ucExitCode != ERR_NONE )
	{
		DisplayError( ucExitCode );
	}


	//  Set the return code and exit
	//
	return ucExitCode;
}


//  If this is an RXE file, we will need to replace any fixups
//  that were not replaced with interrupts.
//
uchar ReplaceImmediateFixups( unsigned * puDelta )
{
	bool fIsRXE = FALSE;
	uchar ucExitCode = ERR_NONE;


	//  Determine if this is an RXE file and determine the delta of
	//  the reduced header size.
	//
	ucExitCode = DetermineNewEntry( puDelta, &fIsRXE );
	if( fIsRXE )
	{
		ucExitCode = DoFixups( *puDelta );
	}

	return ucExitCode;
}


//
//
uchar DoFixups( unsigned uDelta )
{
	static _XEHDR	EXEHeader;
	static _XEHDR	RXEHeader;
	char * szEXEPath;
	char * szEXEMAPPath;
	char * szRXEPath;
	char * szRXEMAPPath;
	int iAction;
	int iEXEHandle = 0xffff;
	int iRXEHandle = 0xffff;
	uchar ucErrorCode;
	unsigned uEXEFixupListSegment;
	unsigned uEXEFixupListPara;
	LPFIXUP_ENTRY lpEXEFixupList;
	unsigned uRXEFixupListSegment;
	unsigned uRXEFixupListPara;
	LPFIXUP_ENTRY lpRXEFixupList;
	bool fSuccessful;


	//  Open the executable
	//
	szEXEPath = GetEXEPath();
	iAction = ACT_EXIST_OPEN;
	if( !SmartCreateOpenFile(O_RDONLY, A_NORMAL, iAction, szEXEPath,
			&iEXEHandle, &iAction) )
	{
		return ERR_OPEN_EXE;
	}


	//  Open the RXE converted executable
	//
	szRXEPath = GetRXEPath();
	iAction = ACT_EXIST_OPEN;
	if( !SmartCreateOpenFile(O_RDWR, A_NORMAL, iAction, szRXEPath,
			&iRXEHandle, &iAction) )
	{
		dlCloseFile( iEXEHandle );
		return ERR_OPEN_RXE;
	}


	//  Load the EXE header in the EXE file
	//
	ucErrorCode = LoadEXEHeader( iEXEHandle, &EXEHeader );
	if( ucErrorCode != ERR_NONE )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );


		//  Translate the error code if neccessary
		//
		if( ucErrorCode == ERR_SEEK )
		{
			ucErrorCode = ERR_SEEK_EXE;
		}
		else if( ucErrorCode == ERR_READ )
		{
			ucErrorCode = ERR_READ_EXE;
		}

		return ucErrorCode;
	}


	//  Load the EXE header in the RXE converted EXE file
	//
	ucErrorCode = LoadEXEHeader( iRXEHandle, &RXEHeader );
	if( ucErrorCode != ERR_NONE )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );


		//  Translate the error code if neccessary
		//
		if( ucErrorCode == ERR_SEEK )
		{
			ucErrorCode = ERR_SEEK_RXE;
		}
		else if( ucErrorCode == ERR_READ )
		{
			ucErrorCode = ERR_READ_RXE;
		}

		return ucErrorCode;
	}


	//  Allocate enough memory for the fixup list
	//
	uEXEFixupListPara = (unsigned)((((ulong)(EXEHeader.EXE.nreloc) *
			sizeof( FIXUP_ENTRY )) + 0xf) >> 4);
	if( !dlAllocMem(uEXEFixupListPara, &uEXEFixupListSegment) )
	{
		return ERR_MEM;
	}
	lpEXEFixupList = MK_FP( uEXEFixupListSegment, 0 );


	//  Read in the fixup list from the exe header
	//
	ucErrorCode = LoadFixupList( iEXEHandle, &EXEHeader, lpEXEFixupList );
	if( ucErrorCode )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		dlFreeMem(uRXEFixupListSegment);
		dlFreeMem(uEXEFixupListSegment);


		//  Translate the error code if neccessary
		//
		if( ucErrorCode == ERR_SEEK )
		{
			ucErrorCode = ERR_SEEK_EXE;
		}
		else if( ucErrorCode == ERR_READ )
		{
			ucErrorCode = ERR_READ_EXE;
		}

		return ucErrorCode;
	}


	//  Modify the fixups in the RXE file
	//
	ucErrorCode = ModifyFixups( iEXEHandle, &EXEHeader, lpEXEFixupList,
			iRXEHandle, &RXEHeader, uDelta );
	if( ucErrorCode )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		dlFreeMem(uEXEFixupListSegment);
		return ucErrorCode;
	}


	//  Free the memory we allocated for the fixup list
	//
	if( !dlFreeMem(uEXEFixupListSegment) )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		return ERR_FREE;
	}


	//  Close the RXE converted executable
	//
	if( !dlCloseFile( iRXEHandle ) )
	{
		dlCloseFile( iEXEHandle );
		return ERR_CLOSE_RXE;
	}


	//  Close the executable file
	//
	if( !dlCloseFile( iEXEHandle ) )
	{
		return ERR_CLOSE_EXE;
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Modify the fixups
//
uchar ModifyFixups( int iEXEHandle, P_XEHDR pEXEHeader,
		FIXUP_ENTRY huge * hpFixupList, int iRXEHandle, P_XEHDR pRXEHeader,
		unsigned uDelta )
{
	static RXE_BASE RXEBase;
	long lEXELocation;
	long lRXELocation;
	unsigned uFixupCount;
	unsigned uFixupIndex;
	uchar ucEXEInstruction[20];
	uchar ucRXEInstruction[20];
	uchar ucErrorCode;
	bool fConversionValid;
	unsigned uEXEValue;
	unsigned uRXEValue;
	long lEXECmpStart;
	long lRXECmpStart;
	long lEXECmpEnd;
	ulong ulCmpSize;
	uchar ucInstructionBytes;
	unsigned uTemp;
	uchar ucTemp;
	ulong ulImageSize;
	ulong ulTemp;
	uchar ucIndex;
	unsigned uNewSegmentFixup;
	unsigned * puOldFixupValue;
	uchar ucSegOverride;


	//  Load the RXE files RXE_BASE\XIP structure
	//
	ucErrorCode = LoadRXEBASE( iRXEHandle, (P_XEHDR)(pRXEHeader), &RXEBase );
	if( ucErrorCode )
	{
		return ucErrorCode;
	}


	//
	//
	uFixupCount = pEXEHeader->EXE.nreloc;
	for( uFixupIndex = 0; uFixupIndex < uFixupCount; uFixupIndex++ )
	{
		//  Make sure there is a fixup at this location.  Earlier the
		//  fixup list was pruned.  The fixups that will be made during
		//  run-time were removed from the list and replaced with zeros.
		//
		if(( hpFixupList[uFixupIndex].uOffset == 0 ) &&
				( hpFixupList[uFixupIndex].uSegment == 0 ) )
		{
			//
			//  This entry was removed because the fixup will be made
			//  at run-time.  Thus ignore this entry
			//
		}
		else
		{
//  This debug code can be enabled to have a place to set a breakpoint
//  for a specific fixup.  Just change the segment and offset to that
//  of the fixup and set your breakpoint where labeled below
//
#if 0
			if(( hpFixupList[uFixupIndex].uSegment == 0x0 ) &&
					( hpFixupList[uFixupIndex].uOffset == 0x9 ) )
			{
				uchar ucBT;


				ucBT = 1;
			}
#endif

			//  Determine where in the EXE file the fixup is
			//
			lEXELocation  = (long)(pEXEHeader->EXE.hdrlen) << 4;
			lEXELocation += (long)(hpFixupList[uFixupIndex].uSegment) << 4;
			lEXELocation += hpFixupList[uFixupIndex].uOffset;


			//  Determine where in the RXE file the fixup is
			//
			if( hpFixupList[uFixupIndex].uSegment >= RXEBase.cd_len )
			{
				lRXELocation  = (long)(pEXEHeader->EXE.hdrlen) << 4;
				lRXELocation += (long)(hpFixupList[uFixupIndex].uSegment) << 4;
				lRXELocation += (long)(RXEBase.rxe_base_len) << 4;
				lRXELocation += hpFixupList[uFixupIndex].uOffset;
			}
			else
			{
				lRXELocation  = (long)(pEXEHeader->EXE.hdrlen) << 4;
				lRXELocation += (long)(hpFixupList[uFixupIndex].uSegment) << 4;
				lRXELocation += hpFixupList[uFixupIndex].uOffset;
			}


			//  Read the instruction from the EXE file
			//
			ucInstructionBytes = 2;
			ucErrorCode = GetInstructionAtLocation( iEXEHandle,
					lEXELocation - ucInstructionBytes + 2, ucInstructionBytes,
					ucEXEInstruction, &ucSegOverride );
			if( ucErrorCode )
			{
				//  Translate the error code if neccessary
				//
				if( ucErrorCode == ERR_SEEK )
				{
					ucErrorCode = ERR_SEEK_EXE;
				}
				else if( ucErrorCode == ERR_READ )
				{
					ucErrorCode = ERR_READ_EXE;
				}

				return ucErrorCode;
			}


			//  Read the instruction from the RXE file
			//
			ucErrorCode = GetInstructionAtLocation( iRXEHandle,
					lRXELocation - ucInstructionBytes + 2, ucInstructionBytes,
					ucRXEInstruction, &ucSegOverride );
			if( ucErrorCode )
			{
				//  Translate the error code if neccessary
				//
				if( ucErrorCode == ERR_SEEK )
				{
					ucErrorCode = ERR_SEEK_RXE;
				}
				else if( ucErrorCode == ERR_READ )
				{
					ucErrorCode = ERR_READ_RXE;
				}

				return ucErrorCode;
			}


#if RXEOPTIM_DEBUG
			for( ucIndex = 0; ucIndex < ucInstructionBytes; ucIndex++ )
			{
				printf( "%02x ", ucEXEInstruction[ucIndex] );
			}

			printf("-> ");

			for( ucIndex = 0; ucIndex < ucInstructionBytes; ucIndex++ )
			{
				printf( "%02x ", ucRXEInstruction[ucIndex] );
			}
#endif

			puOldFixupValue =
					(unsigned *)(&(ucEXEInstruction[ucInstructionBytes - 2]));

			if( *puOldFixupValue < RXEBase.cd_len )
			{
				uNewSegmentFixup  = ucRXEInstruction[ucInstructionBytes - 1];
				uNewSegmentFixup <<= 8;
				uNewSegmentFixup |= ucRXEInstruction[ucInstructionBytes - 2];
				uNewSegmentFixup -= uDelta;

				ucRXEInstruction[ucInstructionBytes - 2] = (uchar)(uNewSegmentFixup);
				uNewSegmentFixup >>= 8;
				ucRXEInstruction[ucInstructionBytes - 1] = (uchar)(uNewSegmentFixup);

#if RXEOPTIM_DEBUG
				printf("-> ");

				for( ucIndex = 0; ucIndex < ucInstructionBytes; ucIndex++ )
				{
					printf( "%02x ", ucRXEInstruction[ucIndex] );
				}

				printf("\n");
#endif

				//  Write the instruction to the RXE file
				//
				ucErrorCode = SetInstructionAtLocation( iRXEHandle,
						lRXELocation - ucInstructionBytes + 2, ucInstructionBytes,
						ucRXEInstruction );
				if( ucErrorCode )
				{
					//  Translate the error code if neccessary
					//
					if( ucErrorCode == ERR_SEEK )
					{
						ucErrorCode = ERR_SEEK_RXE;
					}
					else if( ucErrorCode == ERR_READ )
					{
						ucErrorCode = ERR_READ_RXE;
					}

					return ucErrorCode;
				}
			}
		}
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Read the instruction which contains the fixup given the handle and
//  the location in the file.
//
uchar SetInstructionAtLocation( int iHandle, long lLocation,
		uchar ucInstructionBytes, uchar * ucInstruction )
{
	unsigned uReadBytes;


	//  Seek to the appropriate place in the file
	//
	if( !dlSeekFile(SEEK_SET, iHandle, &lLocation) )
	{
		return ERR_SEEK;
	}


	//  Write the instruction that contains the fixup
	//
	uReadBytes = ucInstructionBytes;
	if( !dlWriteFile(iHandle, &uReadBytes, (char *)ucInstruction) )
	{
		return ERR_WRITE;
	}
	if( uReadBytes != ucInstructionBytes )
	{
		return ERR_WRITE;
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Read the fixup list in an EXE file into the fixup structures
//
uchar WriteFixupList( int iHandle, P_XEHDR pEXEHeader,
		LPFIXUP_ENTRY lpFixupList )
{
	unsigned uWriteBytes;
	long lLocation;
	unsigned uFixupCount;


	//  Seek to the beginning of the fixup list
	//
	lLocation = (long)(pEXEHeader->EXE.reldsp);
	if( !dlSeekFile(SEEK_SET, iHandle, &lLocation) )
	{
		return ERR_SEEK;
	}


	//  Determine the number of fixups in the list
	//
	uFixupCount = pEXEHeader->EXE.nreloc;


	//  Read in all the fixup locations
	//
	while( uFixupCount )
	{
		uWriteBytes = sizeof( FIXUP_ENTRY );
		if( !dlWriteFile(iHandle, &uWriteBytes, (char far *)lpFixupList) )
		{
			return ERR_WRITE;
		}
		if( uWriteBytes != sizeof(FIXUP_ENTRY) )
		{
			return ERR_WRITE;
		}

		(char huge *)lpFixupList += sizeof( FIXUP_ENTRY );
		uFixupCount--;
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//
//
uchar DetermineNewEntry( unsigned * puDelta, bool * pfIsRXE )
{
	char * szRXEPath;
	bool fSuccessful;
	int iSrcHandle;
	int iAction;
	static RXE_BASE SrcRXEBASE;
	static EXEHDR SrcEXEHDR;
	unsigned uByteCount;
	ulong ulTemp;
	bool fIsRXE;
	uchar ucErrorCode;
	unsigned uOldHeaderSize;


	//  Get the expanded name and path for the name and path specified
	//  on the command line
	//
	szRXEPath = GetRXEPath();


	//  Open the source file (file to be optimized)
	//
	fSuccessful = SmartCreateOpenFile( OPEN_ACCESS_READONLY, A_NORMAL,
			FILE_OPEN, szRXEPath, &iSrcHandle, &iAction );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an invalid path
		//
		DisplayError( ERR_INVALID_PATH );
		return ERR_INVALID_PATH;
	}


	//  Read in the exe header
	//
	uByteCount = sizeof( EXEHDR );
	fSuccessful = dlReadFile( iSrcHandle, &uByteCount,
			(char *)(&SrcEXEHDR) );
	if( (!fSuccessful) || (uByteCount != sizeof(EXEHDR)) )
	{
		//  Display an error message stating that there was an error
		//  reading from the source file
		//
		DisplayError( ERR_READ );
		dlCloseFile( iSrcHandle );
		return ERR_READ;
	}


	//  Make sure that this is an exe file
	//
	if( SrcEXEHDR.sig != 0x5a4d )
	{
		//  Display an error message stating that the source file was not
		//  an EXE/RXE file
		//
		DisplayError( ERR_NOT_EXE );
		dlCloseFile( iSrcHandle );
		return ERR_NOT_EXE;
	}


	//  Determine if this is an RXE program
	//
	ucErrorCode = LoadRXEBASE( iSrcHandle, (P_XEHDR)(&SrcEXEHDR), &SrcRXEBASE );
	fIsRXE = (ucErrorCode == 0);
	*pfIsRXE = fIsRXE;
	if( !fIsRXE )
	{
		dlCloseFile( iSrcHandle );
		return ERR_NONE;
	}


	//  Close the source file
	//
	fSuccessful = dlCloseFile( iSrcHandle );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  closing the source file
		//
		DisplayError( ERR_CLOSE_SOURCE );
		return ERR_CLOSE_SOURCE;
	}


	//  Determine the new header size
	//
	ulTemp  = SrcEXEHDR.nreloc * sizeof(FIXUP_ENTRY);
	ulTemp += sizeof(EXEHDR) + 15;
	ulTemp >>= 4;


	//  Determine the size of the old header
	//
	uOldHeaderSize = SrcEXEHDR.hdrlen;
	if( fIsRXE )
	{
		uOldHeaderSize -= SrcRXEBASE.cd_len;
	}


	//  Save the difference in the size of the old header and the new header
	//
	*puDelta = uOldHeaderSize - (unsigned)(ulTemp);


	return ERR_NONE;
}


//
//
uchar OptimizeEXESize( unsigned uDelta )
{
	uchar ucExitCode = ERR_NONE;
	char * szPath;
	bool fSuccessful;
	int iSourceHandle;
	int iDestinationHandle;
	int iAction;
	static EXEHDR EXEHDR;
	unsigned uByteCount;
	char * szTemp;
	static char szDestName[LFN_MAX_PATH];
	static char szTempName[LFN_MAX_PATH];


	//  Get the expanded name and path for the name and path specified
	//  on the command line
	//
	szPath = GetRXEPath();


	//  Open the source file (file to be optimized)
	//
	fSuccessful = SmartCreateOpenFile( OPEN_ACCESS_READONLY, A_NORMAL,
			FILE_OPEN, szPath, &iSourceHandle, &iAction );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an invalid path
		//
		DisplayError( ERR_INVALID_PATH );
		return ERR_INVALID_PATH;
	}


	//  Read in the exe header
	//
	uByteCount = sizeof( EXEHDR );
	fSuccessful = dlReadFile( iSourceHandle, &uByteCount,
			(char *)(&EXEHDR) );
	if( (!fSuccessful) || (uByteCount != sizeof(EXEHDR)) )
	{
		//  Display an error message stating that there was an error
		//  reading from the source file
		//
		DisplayError( ERR_READ );
		dlCloseFile( iSourceHandle );
		return ERR_READ;
	}


	//  Make sure that this is an exe file
	//
	if( EXEHDR.sig != 0x5a4d )
	{
		//  Display an error message stating that the source file was not
		//  an EXE/RXE file
		//
		DisplayError( ERR_NOT_EXE );
		dlCloseFile( iSourceHandle );
		return ERR_NOT_EXE;
	}


	//  Create a temporary file
	//
	strcpy( szDestName, szPath );
	szTemp = FindExt( szDestName );
	strcpy( szTemp, ".___" );
	fSuccessful = SmartCreateOpenFile( OPEN_ACCESS_READWRITE, A_NORMAL,
			FILE_CREATE, szDestName, &iDestinationHandle, &iAction );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  creating the intermediate file
		//
		DisplayError( ERR_CREATE );
		dlCloseFile( iSourceHandle );
		return ERR_CREATE;
	}


	//  Write the new EXE file while not copying any unused data
	//
	ucExitCode = OptimizedCopy( iSourceHandle, &EXEHDR, iDestinationHandle,
			uDelta );
	if( ucExitCode != ERR_NONE )
	{
		dlCloseFile( iSourceHandle );
		dlCloseFile( iDestinationHandle );
		SmartDelete( szDestName );
		return ucExitCode;
	}


	//  Close the temporary file
	//
	fSuccessful = dlCloseFile( iDestinationHandle );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  closing the intermediate file
		//
		DisplayError( ERR_CLOSE_DESTINATION );
		dlCloseFile( iSourceHandle );
		return ERR_CLOSE_DESTINATION;
	}


	//  Close the source file
	//
	fSuccessful = dlCloseFile( iSourceHandle );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  closing the source file
		//
		DisplayError( ERR_CLOSE_SOURCE );
		return ERR_CLOSE_SOURCE;
	}


	//  Rename the source file
	//
	strcpy( szTempName, szPath );
	szTemp = FindExt( szTempName );
	strcpy( szTemp, ".__1" );
	fSuccessful = SmartRenameFileOrDirectory( LastPart(szPath),
			LastPart(szTempName) );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  renaming the source file
		//
		DisplayError( ERR_RENAME_SOURCE );
		return ERR_RENAME_SOURCE;
	}


	//  Rename the temp file to the source file
	//
	fSuccessful = SmartRenameFileOrDirectory( LastPart(szDestName),
			LastPart(szPath) );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  renaming the source file
		//
		DisplayError( ERR_RENAME_DEST );
		SmartRenameFileOrDirectory( LastPart(szTempName), LastPart(szPath) );
		SmartDelete( szDestName );
		return ERR_RENAME_DEST;
	}


	//  Delete the original source file
	//
	fSuccessful = SmartDelete( szTempName );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  deleting the source file
		//
		DisplayError( ERR_DELETE );
		SmartRenameFileOrDirectory( LastPart(szTempName), LastPart(szPath) );
		SmartDelete( szDestName );
		return ERR_DELETE;
	}


	//  Return the exit code for the program
	//
	return ucExitCode;
}


//
//
uchar OptimizedCopy( int iSrcHandle, PEXEHDR pSrcEXEHDR, int iDstHandle,
		unsigned uDelta )
{
	static RXE_BASE SrcRXEBASE;
	static EXEHDR DstEXEHDR;
	static char cBuffer[BUFFER_SIZE];
	unsigned uByteCount;
	bool fSuccessful;
	ulong ulFixupListByteSize;
	ulong ulNewEXEHeaderSize;
	unsigned uReadWriteSize;
	long lTemp;
	ulong ulImageSize;
	ulong ulNewImageSize;
	bool fIsRXE;
	ulong ulRXEHiddenCodeSize = 0;
	unsigned uIndex;
	unsigned uCheckSumCount;
	unsigned uCheckSumIndex;
	unsigned uCheckSum = 0;
	uchar ucErrorCode;


	//  Determine if this is an RXE program
	//
	ucErrorCode = LoadRXEBASE( iSrcHandle, (P_XEHDR)(pSrcEXEHDR), &SrcRXEBASE );
	fIsRXE = (ucErrorCode == 0);


	//  Print out some of the information in the source EXE header
	//
#if RXEOPTIM_DEBUG
	printf( "pSrcEXEHDR->imgmod = %04xh\n", pSrcEXEHDR->imgmod );
	printf( "pSrcEXEHDR->imglen = %04xh\n", pSrcEXEHDR->imglen );
	printf( "pSrcEXEHDR->nreloc = %04xh (%lu bytes)\n", pSrcEXEHDR->nreloc,
			(ulong)(pSrcEXEHDR->nreloc) * sizeof(FIXUP_ENTRY) );
	printf( "pSrcEXEHDR->hdrlen = %04xh (%lu bytes)\n", pSrcEXEHDR->hdrlen,
			(ulong)(pSrcEXEHDR->hdrlen) * 16 );
	printf( "pSrcEXEHDR->minpar = %04xh\n", pSrcEXEHDR->minpar );
	printf( "pSrcEXEHDR->maxpar = %04xh\n", pSrcEXEHDR->maxpar );
	printf( "pSrcEXEHDR->ssdsp  = %04xh\n", pSrcEXEHDR->ssdsp );
	printf( "pSrcEXEHDR->spofs  = %04xh\n", pSrcEXEHDR->spofs );
	printf( "pSrcEXEHDR->cksum  = %04xh\n", pSrcEXEHDR->cksum );
	printf( "pSrcEXEHDR->ipofs  = %04xh\n", pSrcEXEHDR->ipofs );
	printf( "pSrcEXEHDR->csdsp  = %04xh\n", pSrcEXEHDR->csdsp );
	printf( "pSrcEXEHDR->reldsp = %04xh\n", pSrcEXEHDR->reldsp );
	printf( "pSrcEXEHDR->ovlnum = %04xh\n", pSrcEXEHDR->ovlnum );
	printf("\n");
#endif


	//  Print out some of the information in the source RXEBASE structure
	//
#if RXEOPTIM_DEBUG
	if( fIsRXE )
	{
		printf("SrcRXEBASE->rxe_ver      = %04xh\n", SrcRXEBASE.rxe_ver );
		printf("SrcRXEBASE->cd_seg       = %04xh\n", SrcRXEBASE.cd_seg );
		printf("SrcRXEBASE->ld_seg       = %04xh\n", SrcRXEBASE.ld_seg );
		printf("SrcRXEBASE->int_no       = %04xh\n", SrcRXEBASE.int_no );
		printf("SrcRXEBASE->cd_cs        = %04xh\n", SrcRXEBASE.cd_cs );
		printf("SrcRXEBASE->cd_ip        = %04xh\n", SrcRXEBASE.cd_ip );
		printf("SrcRXEBASE->cd_len       = %04xh\n", SrcRXEBASE.cd_len );
		printf("SrcRXEBASE->cd_sum       = %04xh\n", SrcRXEBASE.cd_sum );
		printf("SrcRXEBASE->rxe_base_len = %04xh\n", SrcRXEBASE.rxe_base_len );
		for( uIndex = 0; uIndex < 16; uIndex++ )
		{
			printf("SrcRXEBASE->rxe_fixup_table = %04xh\n",
					SrcRXEBASE.rxe_fixup_table[uIndex] );
		}
		printf("\n");
	}
#endif


	//  Copy the new exe header
	//
	memcpy( &DstEXEHDR, pSrcEXEHDR, sizeof(EXEHDR) );


	//  Determine the total size of the new exe header including
	//  the fixup list
	//
	ulFixupListByteSize = DstEXEHDR.nreloc * sizeof(FIXUP_ENTRY);
	ulNewEXEHeaderSize  = ulFixupListByteSize;
	ulNewEXEHeaderSize += sizeof(EXEHDR);
	if( fIsRXE )
	{
		ulRXEHiddenCodeSize = ((ulong)(SrcRXEBASE.cd_len)) << 4;
	}


	//  Determine the size of the image in the source exe file
	//
	ulImageSize  = (ulong)(pSrcEXEHDR->imglen) * 512;
	if( pSrcEXEHDR->imgmod )
	{
		ulImageSize -= 512;
		ulImageSize += pSrcEXEHDR->imgmod;
	}
	ulImageSize -= ((ulong)(pSrcEXEHDR->hdrlen)) << 4;


	//  Adjust the start of the fixup list in the destination file,
	//  the size of the EXE header, and the image size.
	//
	DstEXEHDR.reldsp = sizeof(EXEHDR);
	if( fIsRXE )
	{
		DstEXEHDR.hdrlen = (unsigned)(((ulFixupListByteSize + sizeof(EXEHDR) +
				15) >> 4) + SrcRXEBASE.cd_len);
	}
	else
	{
		DstEXEHDR.hdrlen = (unsigned)((ulFixupListByteSize + sizeof(EXEHDR) +
				15) >> 4);
	}
	ulNewImageSize = ulImageSize + ((ulong)(DstEXEHDR.hdrlen) << 4);
	DstEXEHDR.imglen = ulNewImageSize / 512;
	DstEXEHDR.imgmod = ulNewImageSize % 512;
	if( DstEXEHDR.imgmod != 0)
	{
		DstEXEHDR.imglen++;
	}


	//  Print out some of the information in the source EXE header
	//
#if RXEOPTIM_DEBUG
	printf( "DstEXEHDR.imgmod   = %04xh\n", DstEXEHDR.imgmod );
	printf( "DstEXEHDR.imglen   = %04xh\n", DstEXEHDR.imglen );
	printf( "DstEXEHDR.nreloc   = %04xh\n", DstEXEHDR.nreloc );
	printf( "DstEXEHDR.hdrlen   = %04xh\n", DstEXEHDR.hdrlen );
	printf( "DstEXEHDR.minpar   = %04xh\n", DstEXEHDR.minpar );
	printf( "DstEXEHDR.maxpar   = %04xh\n", DstEXEHDR.maxpar );
	printf( "DstEXEHDR.ssdsp    = %04xh\n", DstEXEHDR.ssdsp );
	printf( "DstEXEHDR.spofs    = %04xh\n", DstEXEHDR.spofs );
	printf( "DstEXEHDR.cksum    = %04xh\n", DstEXEHDR.cksum );
	printf( "DstEXEHDR.ipofs    = %04xh\n", DstEXEHDR.ipofs );
	printf( "DstEXEHDR.csdsp    = %04xh\n", DstEXEHDR.csdsp );
	printf( "DstEXEHDR.reldsp   = %04xh\n", DstEXEHDR.reldsp );
	printf( "DstEXEHDR.ovlnum   = %04xh\n", DstEXEHDR.ovlnum );
	printf("\n");
#endif


	//  Write the new exe header
	//
	uByteCount = sizeof( EXEHDR );
#if RXEOPTIM_DEBUG
	printf("EXE Header size:            %04xh (%u)\n", uByteCount, uByteCount );
#endif
	fSuccessful = dlWriteFile( iDstHandle, &uByteCount,
			(char *)(&DstEXEHDR) );
	if( (!fSuccessful) || (uByteCount != sizeof(EXEHDR)) )
	{
		//  Display an error message stating that there was an error
		//  reading from the source file
		//
		DisplayError( ERR_WRITE );
		return ERR_WRITE;
	}


	//  Seek to the fixup list in the source EXE file
	//
	lTemp = (long)(pSrcEXEHDR->reldsp);
	fSuccessful = dlSeekFile( SEEK_SET, iSrcHandle, &lTemp );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  seekin in the source file
		//
		DisplayError( ERR_SEEK );
		return ERR_SEEK;
	}


	//  Copy the fixup list to the destination file
	//
#if RXEOPTIM_DEBUG
	printf("Fixup list size:            %04lxh (%lu)\n", ulFixupListByteSize,
			ulFixupListByteSize );
	if( fIsRXE )
	{
		lTemp = ((long)(DstEXEHDR.hdrlen - SrcRXEBASE.cd_len)) << 4;
	}
	else
	{
		lTemp = ((long)(DstEXEHDR.hdrlen)) << 4;
	}
	lTemp -= sizeof( EXEHDR );
	lTemp -= ulFixupListByteSize;
	printf("Fixup list para round size: %04lxh (%lu)\n", lTemp, lTemp );
#endif
	while( ulFixupListByteSize )
	{
		//  Determine the size of the read
		//
		uReadWriteSize = BUFFER_SIZE;
		if( (ulong)BUFFER_SIZE > ulFixupListByteSize )
		{
			uReadWriteSize = (unsigned)ulFixupListByteSize;
		}


		//  Read the data from the source file
		//
		uByteCount = uReadWriteSize;
		fSuccessful = dlReadFile( iSrcHandle, &uByteCount, cBuffer );
		if( (!fSuccessful) || (uByteCount != uReadWriteSize) )
		{
			//  Display an error message stating that there was an error
			//  reading from the source file
			//
			DisplayError( ERR_READ );
			return ERR_READ;
		}


		//  Write the data to the destination file
		//
		uByteCount = uReadWriteSize;
		fSuccessful = dlWriteFile( iDstHandle, &uByteCount, cBuffer );
		if( (!fSuccessful) || (uByteCount != uReadWriteSize) )
		{
			//  Display an error message stating that there was an error
			//  reading from the source file
			//
			DisplayError( ERR_WRITE );
			return ERR_WRITE;
		}


		//  Adjust the remaining number of bytes to copy
		//
		ulFixupListByteSize -= uReadWriteSize;
	}


	//  Expand the destination file to the at least the size
	//  of the exe header
	//
	if( fIsRXE )
	{
		lTemp = ((long)(DstEXEHDR.hdrlen - SrcRXEBASE.cd_len)) << 4;
	}
	else
	{
		lTemp = ((long)(DstEXEHDR.hdrlen)) << 4;
	}
	fSuccessful = dlSeekFile( SEEK_SET, iDstHandle, &lTemp );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  seekin in the source file
		//
		DisplayError( ERR_SEEK );
		return ERR_SEEK;
	}
	uByteCount = 0;
	fSuccessful = dlWriteFile( iDstHandle, &uByteCount, cBuffer );
	if( (!fSuccessful) || (uByteCount != 0) )
	{
		//  Display an error message stating that there was an error
		//  reading from the source file
		//
		DisplayError( ERR_WRITE );
		return ERR_WRITE;
	}


	//  Seek to the image in the source EXE file
	//
	if( fIsRXE )
	{
		lTemp = ((long)(pSrcEXEHDR->hdrlen - SrcRXEBASE.cd_len)) << 4;
	}
	else
	{
		lTemp = ((long)(pSrcEXEHDR->hdrlen)) << 4;
	}
	fSuccessful = dlSeekFile( SEEK_SET, iSrcHandle, &lTemp );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an error
		//  seekin in the source file
		//
		DisplayError( ERR_SEEK );
		return ERR_SEEK;
	}


	//  Copy the part of the image that does not get loaded from the
	//  RXE file and is hidden by the RXEHDR header size field.
	//
#if RXEOPTIM_DEBUG
	if( ulRXEHiddenCodeSize )
	{
		printf("RXE hidden code size:      %04lxh (%lu)\n", ulRXEHiddenCodeSize,
				ulRXEHiddenCodeSize );
	}
#endif
	while( ulRXEHiddenCodeSize )
	{
		//  Determine the size of the read
		//
		uReadWriteSize = BUFFER_SIZE;
		if( (ulong)BUFFER_SIZE > ulRXEHiddenCodeSize )
		{
			uReadWriteSize = (unsigned)ulRXEHiddenCodeSize;
		}


		//  Read the data from the source file
		//
		uByteCount = uReadWriteSize;
		fSuccessful = dlReadFile( iSrcHandle, &uByteCount, cBuffer );
		if( (!fSuccessful) || (uByteCount != uReadWriteSize) )
		{
			//  Display an error message stating that there was an error
			//  reading from the source file
			//
			printf("uByteCount: %u\n", uByteCount );
			printf("uReadWriteSize: %u\n", uReadWriteSize );
			printf("ulRXEHiddenCodeSize: %lu (%lx)\n", ulRXEHiddenCodeSize, ulRXEHiddenCodeSize );
			DisplayError( ERR_READ );
			return ERR_READ;
		}


		//  Calculate checksum
		//
		uCheckSumCount = uReadWriteSize / 2;
		for( uCheckSumIndex = 0; uCheckSumIndex < uCheckSumCount;
				uCheckSumIndex++ )
		{
			uCheckSum += ((unsigned *)cBuffer)[uCheckSumIndex];
		}


		//  Write the data to the destination file
		//
		uByteCount = uReadWriteSize;
		fSuccessful = dlWriteFile( iDstHandle, &uByteCount, cBuffer );
		if( (!fSuccessful) || (uByteCount != uReadWriteSize) )
		{
			//  Display an error message stating that there was an error
			//  reading from the source file
			//
			DisplayError( ERR_WRITE );
			return ERR_WRITE;
		}


		//  Adjust the remaining number of bytes to copy
		//
		ulRXEHiddenCodeSize -= uReadWriteSize;
	}


	//  If this is an RXE file, we need to modify the RXE_BASE structure
	//  to reflect the new entry point
	//
	if( fIsRXE )
	{
		//  Modify the RXE_BASE structure to reflect the new entry point.
		//
		//  NOTE: cd_ip does not need to be adjusted because we are working
		//        on paragraph boundaries.
		//
		SrcRXEBASE.cd_cs  -= uDelta;
		SrcRXEBASE.cd_seg -= uDelta;


		//  Modify the RXE file checksum
		//
		SrcRXEBASE.cd_sum = uCheckSum;


		//  Print out some of the information in the source RXEBASE structure
		//
#if RXEOPTIM_DEBUG
		printf("\n");
		printf("DstRXEBASE->rxe_ver      = %04xh\n", SrcRXEBASE.rxe_ver );
		printf("DstRXEBASE->cd_seg       = %04xh\n", SrcRXEBASE.cd_seg );
		printf("DstRXEBASE->ld_seg       = %04xh\n", SrcRXEBASE.ld_seg );
		printf("DstRXEBASE->int_no       = %04xh\n", SrcRXEBASE.int_no );
		printf("DstRXEBASE->cd_cs        = %04xh\n", SrcRXEBASE.cd_cs );
		printf("DstRXEBASE->cd_ip        = %04xh\n", SrcRXEBASE.cd_ip );
		printf("DstRXEBASE->cd_len       = %04xh\n", SrcRXEBASE.cd_len );
		printf("DstRXEBASE->cd_sum       = %04xh\n", SrcRXEBASE.cd_sum );
		printf("DstRXEBASE->rxe_base_len = %04xh\n", SrcRXEBASE.rxe_base_len );
		for( uIndex = 0; uIndex < 16; uIndex++ )
		{
			printf("DstRXEBASE->rxe_fixup_table = %04xh\n",
					SrcRXEBASE.rxe_fixup_table[uIndex] );
		}
		printf("\n");
#endif


		//  Write the data to the destination file
		//
		uByteCount = sizeof( RXE_BASE );
		fSuccessful = dlWriteFile( iDstHandle, &uByteCount,
				(char *)(&SrcRXEBASE) );
		if( (!fSuccessful) || (uByteCount != sizeof(RXE_BASE)) )
		{
			//  Display an error message stating that there was an error
			//  reading from the source file
			//
			DisplayError( ERR_WRITE );
			return ERR_WRITE;
		}


		//  Go past the RXE_BASE structure in the source file
		//
		lTemp = sizeof( RXE_BASE );
		fSuccessful = dlSeekFile( SEEK_CUR, iSrcHandle, &lTemp );
		if( !fSuccessful )
		{
			//  Display an error message stating that there was an error
			//  seekin in the source file
			//
			DisplayError( ERR_SEEK );
			return ERR_SEEK;
		}

		//
		//  Subtract the size of the RXE_BASE structure from the image size
		//
		ulImageSize -= sizeof( RXE_BASE );
	}


	//  Copy the image to the destination file
	//
#if RXEOPTIM_DEBUG
	printf("code/data size:             %04lxh (%lu)\n", ulImageSize,
			ulImageSize );
#endif
	while( ulImageSize )
	{
		//  Determine the size of the read
		//
		uReadWriteSize = BUFFER_SIZE;
		if( (ulong)BUFFER_SIZE > ulImageSize )
		{
			uReadWriteSize = (unsigned)ulImageSize;
		}


		//  Read the data from the source file
		//
		uByteCount = uReadWriteSize;
		fSuccessful = dlReadFile( iSrcHandle, &uByteCount, cBuffer );
		if( (!fSuccessful) || (uByteCount != uReadWriteSize) )
		{
			//  Display an error message stating that there was an error
			//  reading from the source file
			//
			printf("uByteCount: %u\n", uByteCount );
			printf("uReadWriteSize: %u\n", uReadWriteSize );
			printf("ulImageSize: %lu (%lx)\n", ulImageSize, ulImageSize );
			DisplayError( ERR_READ );
			return ERR_READ;
		}


		//  Write the data to the destination file
		//
		uByteCount = uReadWriteSize;
		fSuccessful = dlWriteFile( iDstHandle, &uByteCount, cBuffer );
		if( (!fSuccessful) || (uByteCount != uReadWriteSize) )
		{
			//  Display an error message stating that there was an error
			//  reading from the source file
			//
			DisplayError( ERR_WRITE );
			return ERR_WRITE;
		}


		//  Adjust the remaining number of bytes to copy
		//
		ulImageSize -= uReadWriteSize;
	}


#if RXEOPTIM_DEBUG
	printf("\n");
#endif

	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Determine if there are any wildcards in the path
//
bool ContainsWildCards( char * szPath )
{
	while( *szPath )
	{
		if( (*szPath == '*') || (*szPath == '?') )
		{
			return TRUE;
		}
		szPath++;
	}

	return FALSE;
}


//  Process all the command line arguments
//
uchar DoParseCommandLine( int argc, char *argv[] )
{
	char * szPath;
	bool fSuccessful;
	int  iDrive;
	ARGDEF RXEOPTIMOpts[] =
	{
		{ BOOL_SWITCH, &fSwitches[HIDE_SIGNON],      '[', 0 },
		{ 0 }
	};


	//  Process all the boolean switches in the command line
	//
	argc = ParseCmdline(argc, &argv[0], RXEOPTIMOpts);


	//  We should not have more than one path
	//
	if( (argc < 1) || (argc > 2) )
	{
		Usage();
	}


	//  Set the path to the source EXE file
	//
	szPath = GetEXEPath();
	fSuccessful = SmartExpandPath( 0, argv[0], szPath );
	if( !fSuccessful )
	{
		//  Display an error message stating that there was an invalid path
		//
		DisplayError( ERR_INVALID_PATH );
		return ERR_INVALID_PATH;
	}


	//  Make sure that there are no wildcards
	//
	if( ContainsWildCards(szPath) )
	{
		//  Display an error message stating that wildcards are not allowed
		//
		DisplayError( ERR_WILDCARDS );
		return ERR_WILDCARDS;
	}


	if( argc == 2 )
	{
		//  Set the path to the source file (file to be optimized)
		//
		szPath = GetRXEPath();
		fSuccessful = SmartExpandPath( 0, argv[1], szPath );
		if( !fSuccessful )
		{
			//  Display an error message stating that there was an invalid path
			//
			DisplayError( ERR_INVALID_PATH );
			return ERR_INVALID_PATH;
		}
	}
	else
	{
		//  Set the path to the source file (file to be optimized)
		//
		szPath = GetRXEPath();
		fSuccessful = SmartExpandPath( 0, argv[0], szPath );
		if( !fSuccessful )
		{
			//  Display an error message stating that there was an invalid path
			//
			DisplayError( ERR_INVALID_PATH );
			return ERR_INVALID_PATH;
		}
	}


	//  Make sure that there are no wildcards
	//
	if( ContainsWildCards(szPath) )
	{
		//  Display an error message stating that wildcards are not allowed
		//
		DisplayError( ERR_WILDCARDS );
		return ERR_WILDCARDS;
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Display the usage screen and exit
//
void Usage( void )
{
	char szUsage[] =
	"This utility reduces the size of an RXE file be removing the unused\n"
	"space inside the exe header.\n"
	"\n"
	"RXEOPTIM exename rxename\n"
	"\n"
	"exename  name of the original EXE file\n"
	"rxename  name of the RXE file to optimize\n"
	"";

	DisplaySignonMessage();
	printf( szUsage );
	exit(ERR_USAGE);
}


//  Determine if we should display the signon message and if
//  so display it
//
void DisplaySignonMessage( void )
{
	char * szSignOnMessage = "RXEOPTIM %s %s\n%s\n\n";

	if( !DoHideSignon() )
	{
		if( fIsFat32World )
		{
			printf(szSignOnMessage,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(szSignOnMessage,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}


//  Return the address of the variable which contains the
//  starting/root path for the RXE file
//
char * GetRXEPath( void )
{
	static char szPath[LFN_MAX_PATH];
	return szPath;
}


//  Return the address of the variable which contains the
//  starting/root path for the EXE file
//
char * GetEXEPath( void )
{
	static char szPath[LFN_MAX_PATH];
	return szPath;
}


//  Error handling for the command line parser
//
void OptError( char * pMsg )
{
	printf( "Invalid command line option: %s\n", pMsg );
	exit( ERR_USAGE );
}


//  Display an error message
//
void DisplayError( unsigned uErrorIndex )
{
	static char * szErrorMsg[] = {
		"",
		"Invalid name or path!\n",
		"Invalid name or path! Wildcards not allowed.\n",
		"Error reading the source file.\n",
		"Error writing the output file.\n",
		"Error!  Not a valid EXE file.\n",
		"Error creating the intermediate file.\n",
		"Error closing the source file.\n",
		"Error closing the intermediate file.\n",
		"Error renaming the source file.\n",
		"Error renaming the intermediate file.\n",
		"Error deleting the source file.\n",
		"Error seeking in the source file.\n",

		"Error opening EXE file\n",									// 13
		"Error opening RXE converted executable file\n",		// 14
		"Error reading EXE file\n",									// 15
		"Error reading RXE converted executable file\n",		// 16
		"Error closing EXE file\n",									// 17
		"Error closing RXE converted executable file\n",		// 18
		"Error seeking EXE file\n",									// 19
		"Error seeking RXE converted executable file\n",		// 20
		"Error, bad EXE file\n",										// 21
		"Error, bad RXE file\n",										// 22
		"Error, insufficent conventional memory\n",				// 23
		"Error freeing conventional memory\n",						// 24
		"Error, The fixup list in the RXE file contains\n"
		"entries which are not in the EXE file.\n",				// 25
	};


	printf( szErrorMsg[uErrorIndex] );
}


//  Retrieve the file size of a file
//
bool GetFileSize( char * szFileName, ulong * pulFileSize )
{
	struct ffblk FindData;
	bool fSuccessful;


	fSuccessful = SmartFindFirst( A_NORMAL, szFileName, &FindData );
	if(fSuccessful)
	{
		*pulFileSize = (ulong)(FindData.ff_fsize);
		SmartFindClose( &FindData );
	}

	return fSuccessful;
}
