//
//   $Id: RXEVERFY.C 1.4 2002/10/16 16:12:00Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  RXEVERFY.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//  06-22-2000  bt  Created
//


//  Borland libraries
//
#include "dir.h"
//#include "stdio.h"
//#include "stdlib.h"
//#include "ctype.h"
//#include "mem.h"
#include "string.h"


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dosver.h"
#include "fatworld.h"
#include "dlapp.h"
#include "dllfn.h"


//  RXEVERFY header files
//
#include "rxeverfy.h"


//  RXEVERFY Prototypes
//
#include "rxeverfy.pt"


//
//  External variables
//
#define  RXEVERFY	1
#include "dosmsg.c"


//
//  Variables
//

//  Flag for the DOS version being 7.1+
//
bool fIsFat32World = FALSE;

//  Flags for the command line switches
//  (accessed through macros in RXEVERFY.H)
//
bool fSwitches[MAX_BOOLEAN_SWITCHES];


//  Error messages
//
#define ARRAY_ENTRIES( array )  (sizeof(array)/sizeof(array[0]))
char * szErrorMsg[] = {
	"",
	"Error opening EXE file",										// 1
	"Error opening MAP file",										// 2
	"Error opening RXE converted executable file",			// 3
	"Error reading EXE file",										// 4
	"Error reading MAP file",										// 5
	"Error reading RXE converted executable file",			// 6
	"Error closing EXE file",										// 7
	"Error closing MAP file",										// 8
	"Error closing RXE converted executable file",			// 9
	"Error seeking EXE file",										// 10
	"Error seeking MAP file",										// 11
	"Error seeking RXE converted executable file",			// 12
	"Error - bad EXE file",											// 13
	"Error - bad RXE file",											// 14
	"Error - insufficent conventional memory",				// 15
	"Error freeing conventional memory",						// 16
	"Error EXEHDR corrupt",											// 17
	"Error in RXE fixup list",										// 18
	"Error RXEBASE corrupt",										// 19
};


//  Expected file name extensions (these do not change runtime)
//
char * szEXEExtension       = "EX_";
char * szEXEMAPExtension    = "MA_";
char * szAltEXEMAPExtension = "MAP";
char * szRXEExtension       = "EXE";
char * szRXEMAPExtension    = "MAP";


//  Start of the program
//
uchar main( int argc, char *argv[] )
{
	uchar ucExitCode = ERR_NONE;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	//  Parse the command line
	//
	DoParseCommandLine( argc-1, &argv[1] );


	//  Determine if we should display the signon message and if
	//  so display it
	//
	DisplaySignonMessage();


	//  Make sure all of the fixups listed in the EXE have
	//  been handled in the RXE
	//
	ucExitCode = PerformVerify();
	if( ucExitCode != ERR_NONE )
	{
		if( ucExitCode > (ARRAY_ENTRIES(szErrorMsg) - 1) )
		{
			printf("Unexpected error: %u\n", (unsigned)ucExitCode );
		}
		else
		{
			printf("%s\n", szErrorMsg[ucExitCode]);
		}
	}


	//  Set the exit code and exit
	//
	return ucExitCode;
}




//  Make sure all of the fixups listed in the EXE have
//  been handled in the RXE
//
uchar PerformVerify( void )
{
	static _XEHDR	EXEHeader;
	static _XEHDR	RXEHeader;
	char * szEXEPath;
	char * szEXEMAPPath;
	char * szRXEPath;
	char * szRXEMAPPath;
	int iAction;
	int iEXEHandle = 0xffff;
	int iRXEHandle = 0xffff;
	uchar ucErrorCode;
	unsigned uEXEFixupListSegment;
	unsigned uRXEFixupListSegment;
	unsigned uEXEFixupListPara;
	unsigned uRXEFixupListPara;
	LPFIXUP_ENTRY lpEXEFixupList;
	LPFIXUP_ENTRY lpRXEFixupList;


	//  Open the executable
	//
	szEXEPath = GetEXEPath();
	iAction = ACT_EXIST_OPEN;
	if( !SmartCreateOpenFile(O_RDONLY, A_NORMAL, iAction, szEXEPath,
			&iEXEHandle, &iAction) )
	{
		return ERR_OPEN_EXE;
	}


	//  Open the RXE converted executable
	//
	szRXEPath = GetRXEPath();
	iAction = ACT_EXIST_OPEN;
	if( !SmartCreateOpenFile(O_RDONLY, A_NORMAL, iAction, szRXEPath,
			&iRXEHandle, &iAction) )
	{
		dlCloseFile( iEXEHandle );
		return ERR_OPEN_RXE;
	}


	//  Load the EXE header in the EXE file
	//
	ucErrorCode = LoadEXEHeader( iEXEHandle, &EXEHeader );
	if( ucErrorCode != ERR_NONE )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );


		//  Translate the error code if neccessary
		//
		if( ucErrorCode == ERR_SEEK )
		{
			ucErrorCode = ERR_SEEK_EXE;
		}
		else if( ucErrorCode == ERR_READ )
		{
			ucErrorCode = ERR_READ_EXE;
		}

		return ucErrorCode;
	}


	//  Load the EXE header in the RXE converted EXE file
	//
	ucErrorCode = LoadEXEHeader( iRXEHandle, &RXEHeader );
	if( ucErrorCode != ERR_NONE )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );


		//  Translate the error code if neccessary
		//
		if( ucErrorCode == ERR_SEEK )
		{
			ucErrorCode = ERR_SEEK_RXE;
		}
		else if( ucErrorCode == ERR_READ )
		{
			ucErrorCode = ERR_READ_RXE;
		}

		return ucErrorCode;
	}


	//  Allocate enough memory for the fixup list that is in the
	//  EXE file
	//
	uEXEFixupListPara = (unsigned)((((ulong)(EXEHeader.EXE.nreloc) *
			sizeof( FIXUP_ENTRY )) + 0xf) >> 4);
	if( !dlAllocMem(uEXEFixupListPara, &uEXEFixupListSegment) )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		return ERR_MEM;
	}
	lpEXEFixupList = MK_FP( uEXEFixupListSegment, 0 );


	//  Allocate enough memory for the fixup list that is in the
	//  RXE file
	//
	uRXEFixupListPara = (unsigned)((((ulong)(RXEHeader.EXE.nreloc) *
			sizeof( FIXUP_ENTRY )) + 0xf) >> 4);
	if( !dlAllocMem(uRXEFixupListPara, &uRXEFixupListSegment) )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		dlFreeMem(uEXEFixupListSegment);
		return ERR_MEM;
	}
	lpRXEFixupList = MK_FP( uRXEFixupListSegment, 0 );


	//  Read in the fixup list from the exe header
	//
	ucErrorCode = LoadFixupList( iEXEHandle, &EXEHeader, lpEXEFixupList );
	if( ucErrorCode )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		dlFreeMem(uRXEFixupListSegment);
		dlFreeMem(uEXEFixupListSegment);


		//  Translate the error code if neccessary
		//
		if( ucErrorCode == ERR_SEEK )
		{
			ucErrorCode = ERR_SEEK_EXE;
		}
		else if( ucErrorCode == ERR_READ )
		{
			ucErrorCode = ERR_READ_EXE;
		}

		return ucErrorCode;
	}


	//  Read in the fixup list from the exe header
	//
	ucErrorCode = LoadFixupList( iRXEHandle, &RXEHeader, lpRXEFixupList );
	if( ucErrorCode )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		dlFreeMem(uRXEFixupListSegment);
		dlFreeMem(uEXEFixupListSegment);


		//  Translate the error code if neccessary
		//
		if( ucErrorCode == ERR_SEEK )
		{
			ucErrorCode = ERR_SEEK_RXE;
		}
		else if( ucErrorCode == ERR_READ )
		{
			ucErrorCode = ERR_READ_RXE;
		}

		return ucErrorCode;
	}


	//  Sort the fixup list from least to greatest
	//
	SortFixupList( EXEHeader.EXE.nreloc, lpEXEFixupList );


	//  Check the fixups in the EXE and there RXE counter parts
	//
	ucErrorCode = CheckEXERXE( iEXEHandle, &EXEHeader, lpEXEFixupList,
			iRXEHandle, &RXEHeader, lpRXEFixupList );
	if( ucErrorCode )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		dlFreeMem(uRXEFixupListSegment);
		dlFreeMem(uEXEFixupListSegment);
		return ucErrorCode;
	}


	//  Free the memory we allocated for the fixup list
	//
	if( !dlFreeMem(uRXEFixupListSegment) )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		dlFreeMem(uEXEFixupListSegment);
		return ERR_FREE;
	}


	//  Free the memory we allocated for the fixup list
	//
	if( !dlFreeMem(uEXEFixupListSegment) )
	{
		dlCloseFile( iRXEHandle );
		dlCloseFile( iEXEHandle );
		return ERR_FREE;
	}


	//  Close the RXE converted executable
	//
	if( !dlCloseFile( iRXEHandle ) )
	{
		dlCloseFile( iEXEHandle );
		return ERR_CLOSE_RXE;
	}


	//  Close the executable file
	//
	if( !dlCloseFile( iEXEHandle ) )
	{
		return ERR_CLOSE_EXE;
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Check the fixups in the EXE and there RXE counter parts
//
uchar CheckEXERXE( int iEXEHandle, P_XEHDR pEXEHeader,
		FIXUP_ENTRY huge * hpFixupList, int iRXEHandle, P_XEHDR pRXEHeader,
		FIXUP_ENTRY huge * hpRXEFixupList )
{
	static char szErrorMessage[512];
	static FIXUP_ENTRY LastFixup;
	static RXE_BASE RXEBase;
	long lEXELocation;
	long lRXELocation;
	unsigned uFixupCount;
	unsigned uFixupIndex;
	uchar ucEXEInstruction[20];
	uchar ucRXEInstruction[20];
	uchar ucErrorCode;
	bool fConversionValid;
	unsigned uEXEValue;
	unsigned uRXEValue;
	uchar ucInterrupt;
	unsigned uROMSegment;
	ulong ulTotalFixups = 0;
	ulong ulIncorrectFixups = 0;
	long lEXECmpStart;
	long lRXECmpStart;
	long lEXECmpEnd;
	ulong ulCmpSize;
	bool fCmpSame = TRUE;
	ulong ulCmpTests = 0;
	ulong ulCmpFails = 0;
	uchar ucInstructionBytes;
	unsigned uTemp;
	uchar ucTemp;
	ulong ulImageSize;
	ulong ulTemp;
	ulong ulRXEHeaderSize;
   uchar ucSegOverride = 0;
	uchar ucFixupLocation;
	uchar ucMODRM;
	uchar ucFUB;
	ulong ulWarnings = 0;
	bool  fIsOptimizedRXE;
	bool	fTemp;


	//  Load the RXE files RXE_BASE\XIP structure
	//
	ucErrorCode = LoadRXEBASE( iRXEHandle, pRXEHeader, &RXEBase );
	if( ucErrorCode )
	{
		return ucErrorCode;
	}


	//  Get the RXE options for interrupt and segment address
	//
	ucInterrupt = RXEBase.int_no;
	uROMSegment = RXEBase.ld_seg;


	//  Determine the size of the header
	//
	ulRXEHeaderSize  = (ulong)(pRXEHeader->EXE.nreloc) * 4;
	ulRXEHeaderSize += sizeof(EXEHDR);
	ulRXEHeaderSize += 0xf;
	ulRXEHeaderSize &= 0xfffffff0;


	//  Determine if this file has been optimized with RXEOPTIM
	//
	if( ( Force_Optim() ) ||
		(pRXEHeader->EXE.hdrlen ==
		((unsigned)(ulRXEHeaderSize >> 4)) + RXEBase.cd_len) )
	{
		//
		//  Optimized with RXEOPTIM
		//

		//  The calculation for the size of the header has already
		//  been performed above for an optimized file
		//
		fIsOptimizedRXE = TRUE;
	}
	else
	{
		//
		//  Not optimized with RXEOPTIM
		//

		//  Determine the size of the header
		//
		ulRXEHeaderSize = (ulong)(pEXEHeader->EXE.hdrlen) << 4;
		fIsOptimizedRXE = FALSE;
	}


	//  Test the EXEHDR fields
	//
	{
		//  Determine if any of the fields that should not
		//  have changed have been modified.
		//
		fTemp = FALSE;
		if( pEXEHeader->EXE.minpar != pRXEHeader->EXE.minpar )
		{
			fTemp = TRUE;
		}
		if( pEXEHeader->EXE.maxpar != pRXEHeader->EXE.maxpar )
		{
			fTemp = TRUE;
		}
		uTemp  = pEXEHeader->EXE.ssdsp;
		uTemp -= RXEBase.cd_len;
		uTemp += RXEBase.rxe_base_len;
		if( pRXEHeader->EXE.ssdsp != uTemp )
		{
			fTemp = TRUE;
		}
		uTemp  = pEXEHeader->EXE.ssdsp - RXEBase.cd_len;
		uTemp += RXEBase.rxe_base_len;
		if( uTemp != pRXEHeader->EXE.ssdsp )
		{
			fTemp = TRUE;
		}
		if( pEXEHeader->EXE.spofs != pRXEHeader->EXE.spofs )
		{
			fTemp = TRUE;
		}
		if( pEXEHeader->EXE.cksum != pRXEHeader->EXE.cksum )
		{
			fTemp = TRUE;
		}
		uTemp = (unsigned)(&RXEBase.code[0]) - (unsigned)(&RXEBase.rxe_sig);
		if( uTemp != pRXEHeader->EXE.ipofs )
		{
			fTemp = TRUE;
		}

      // New Code Segment Displacement should be zero in all converted RXE
      //
		if( pRXEHeader->EXE.csdsp != 0 )       // by definition
		{
			fTemp = TRUE;
		}
		if( pEXEHeader->EXE.ovlnum != pRXEHeader->EXE.ovlnum )
		{
			fTemp = TRUE;
		}
		if( fIsOptimizedRXE )
		{
			if( pRXEHeader->EXE.reldsp != 0x1c )
			{
				fTemp = TRUE;
			}
			uTemp = (unsigned)(ulRXEHeaderSize >> 4) + RXEBase.cd_len;
			if( uTemp != pRXEHeader->EXE.hdrlen )
			{
				fTemp = TRUE;
			}
		}
		else
		{
			if( pEXEHeader->EXE.reldsp != pRXEHeader->EXE.reldsp )
			{
				fTemp = TRUE;
			}
			uTemp = pEXEHeader->EXE.hdrlen + RXEBase.cd_len;
			if( uTemp != pRXEHeader->EXE.hdrlen )
			{
				fTemp = TRUE;
			}
 		}
		if( fTemp )
		{
			return ERR_EXEHDR_CORRUPT;
		}
	}


	//  Test the RXE BASE structure fields
	//
	{
		//  check cd_cs
		//
		uTemp  = pEXEHeader->EXE.csdsp;
		uTemp += RXEBase.ld_seg;
		uTemp += pRXEHeader->EXE.hdrlen - RXEBase.cd_len;
		if( RXEBase.cd_cs != uTemp )
		{
			return ERR_RXEBASE_CORRUPT;
		}


		//  check cd_seg
		//
		uTemp  = pRXEHeader->EXE.hdrlen - RXEBase.cd_len;
		uTemp += RXEBase.ld_seg;
		if( RXEBase.cd_seg != uTemp )
		{
			return ERR_RXEBASE_CORRUPT;
		}
	}


#if 0
	//  Check to make sure that all the fixups are in the RXE file
	//
	ucErrorCode = CheckRXEFixups( iEXEHandle, pEXEHeader, hpFixupList,
			pRXEHeader, hpRXEFixupList, &RXEBase );
	if( ucErrorCode )
	{
		return ucErrorCode;
	}
#endif


	//  Display the RXE configuration information
	//
	printf("RXE configuration information\n\n");
	printf("RXE File optimized: ");
	if( fIsOptimizedRXE )
	{
		printf("YES\n");
	}
	else
	{
		printf("NO\n");
	}
	printf("RXE Interrupt: %02xh\n", (unsigned)ucInterrupt );
	printf("RXE ROM Segment: %04xh\n", uROMSegment );
	printf("EXE Header length: %xh\n", (unsigned)(ulRXEHeaderSize >> 4) );
	printf("RXE Init/Interrupt handler size: %xh\n", RXEBase.rxe_base_len );
	printf("RXE Data start: %xh\n", RXEBase.cd_len );
	printf("RXE Header length: %xh\n", pRXEHeader->EXE.hdrlen );
	printf("\n");
	for( ucTemp = 0; ucTemp < RXE_FIXUP_TABLE_ENTRIES; ucTemp++ )
	{
		printf("RXE Fixup table index %x: %04xh\n", ucTemp,
					RXEBase.rxe_fixup_table[ucTemp] );
	}
	printf("\n\n");


	//  Display the output header describing the column information
	//
	printf("  Fixup list details\n\n");
	printf("  Image Addr  EXE seq.     RXE seq.  RESULTS\n");
	printf("  ------------------------------------------\n");


	//  Set the address of the last fixup
	//
	LastFixup.uSegment = 0;
	LastFixup.uOffset  = 0;


	//  Validate the memory blocks between each fixup, before the
	//  first fixup, after the last fixup, and verify that each
	//  fixup case was handled correctly
	//
	uFixupCount = pEXEHeader->EXE.nreloc;
	for( uFixupIndex = 0; uFixupIndex < uFixupCount; uFixupIndex++ )
	{


//  This debug code can be enabled to have a place to set a breakpoint
//  for a specific fixup.  Just change the segment and offset to that
//  of the fixup and set your breakpoint where labeled below
//
#if 0
		if( hpFixupList[uFixupIndex].uSegment == 0xe7 )
		{
			if( hpFixupList[uFixupIndex].uOffset == 0xa )
			{
				unsigned uBT;


				//  Set your breakpoint here
				//
				uBT = 1;
			}
		}
#endif


		//  Determine where in the EXE file the fixup is
		//
		lEXELocation  = (long)(pEXEHeader->EXE.hdrlen) << 4;
		lEXELocation += (long)(hpFixupList[uFixupIndex].uSegment) << 4;
		lEXELocation += hpFixupList[uFixupIndex].uOffset;


		//  Determine where in the RXE file the fixup is
		//
		if( hpFixupList[uFixupIndex].uSegment >= RXEBase.cd_len )
		{
			//  Fixup in the data segment
			//
			lRXELocation  = ulRXEHeaderSize;
			lRXELocation += (long)(hpFixupList[uFixupIndex].uSegment) << 4;
			lRXELocation += (long)(RXEBase.rxe_base_len) << 4;
			lRXELocation += hpFixupList[uFixupIndex].uOffset;
		}
		else
		{
			//  Fixup in the code segment
			//
			lRXELocation  = ulRXEHeaderSize;
			lRXELocation += (long)(hpFixupList[uFixupIndex].uSegment) << 4;
			lRXELocation += hpFixupList[uFixupIndex].uOffset;
		}


		//  Determine the size of the instruction
		//
		ucErrorCode = GetInstructionSize(
		       iEXEHandle, lEXELocation,
				 (hpFixupList[uFixupIndex].uSegment >= RXEBase.cd_len),
             iRXEHandle, lRXELocation, ucInterrupt,
				 RXEBase.cd_len,
				 &ucInstructionBytes );
		if( ucErrorCode )
		{
			//  Translate the error code if neccessary
			//
			if( ucErrorCode == ERR_SEEK )
			{
				ucErrorCode = ERR_SEEK_EXE;
			}
			else if( ucErrorCode == ERR_READ )
			{
				ucErrorCode = ERR_READ_EXE;
			}

			return ucErrorCode;
		}


		//  Read the instruction from the EXE file
		//
		ucErrorCode = GetInstructionAtLocation( iEXEHandle,
				lEXELocation - ucInstructionBytes + 2, ucInstructionBytes,
				ucEXEInstruction, &ucSegOverride );
		if( ucErrorCode )
		{
			//  Translate the error code if neccessary
			//
			if( ucErrorCode == ERR_SEEK )
			{
				ucErrorCode = ERR_SEEK_EXE;
			}
			else if( ucErrorCode == ERR_READ )
			{
				ucErrorCode = ERR_READ_EXE;
			}

			return ucErrorCode;
		}


		//  Read the instruction from the RXE file
		//
		ucErrorCode = GetInstructionAtLocation( iRXEHandle,
				lRXELocation - ucInstructionBytes + 2, ucInstructionBytes,
				ucRXEInstruction, &ucSegOverride );
		if( ucErrorCode )
		{
			//  Translate the error code if neccessary
			//
			if( ucErrorCode == ERR_SEEK )
			{
				ucErrorCode = ERR_SEEK_RXE;
			}
			else if( ucErrorCode == ERR_READ )
			{
				ucErrorCode = ERR_READ_RXE;
			}

			return ucErrorCode;
		}


		//  Determine if this fixup is ok
		//
		fConversionValid = TRUE;
		strcpy( szErrorMessage, " *** CONVERSION ERROR ***\n" );
		if( ucInstructionBytes == 2 )
		{
			//
			//  We only need to make sure that the value at the
			//  fixup location was modified correctly
			//

			//  Determine the current value at the fixup location in
			//  the RXE file
			//
			uRXEValue  = ucRXEInstruction[1] << 8;
			uRXEValue += ucRXEInstruction[0];


	 		//  Determine the fixup value in the original EXE file
			//
			uEXEValue  = ucEXEInstruction[1] << 8;
			uEXEValue += ucEXEInstruction[0];


			//  Determine if this fixup is in the class of CODE
			//
			if( hpFixupList[uFixupIndex].uSegment < RXEBase.cd_len )
			{
				//  Determine what the fixup value should be
				//
				uEXEValue += uROMSegment + (unsigned)(ulRXEHeaderSize >> 4);
			}
			else
			{
				if( uEXEValue < RXEBase.cd_len )
				{
					//  Determine what the fixup value should be
					//
					uEXEValue += uROMSegment + (unsigned)(ulRXEHeaderSize >> 4);
				}
				else
				{
					//  Determine what the fixup value should be
					//
					uEXEValue -= RXEBase.cd_len;
					uEXEValue += RXEBase.rxe_base_len;
				}
			}


			//  Make sure that the value at the fixup location in the RXE
			//  file is what we expect it to be
			//
			if( uEXEValue != uRXEValue )
			{
				sprintf( szErrorMessage,"%s  [expected: %04X]\n",
						" *** CONVERSION ERROR ***", uEXEValue );
				fConversionValid = FALSE;
			}
		}
		else if( ucRXEInstruction[0] == 0xCD )
		{
			//
			//  RXE_CVT has replaced this instruction to call
			//  its interrupt handler
			//

			//  Validate the interrupt number
			//
			if( ucRXEInstruction[1] == ucInterrupt )
			{
 				if( ucInstructionBytes == 3 )
				{
					//  Make sure we recognize the instruction
					//
					switch(ucEXEInstruction[0])
					{
						//  mov	ax, imm16
						//
						case 0xb8:
							if( (ucRXEInstruction[2] & 0xe0) != 0x00 )
							{
								fConversionValid = FALSE;
							}
							break;

                  //  mov  cx, imm16
						//
						case 0xb9:
							if( (ucRXEInstruction[2] & 0xe0) != 0x40 )
							{
								fConversionValid = FALSE;
							}
							break;

						//  mov	dx, imm16
						//
						case 0xba:
							if( (ucRXEInstruction[2] & 0xe0) != 0x60 )
							{
								fConversionValid = FALSE;
							}
							break;

						//  mov	bx, imm16
						//
						case 0xbb:
							if( (ucRXEInstruction[2] & 0xe0) != 0x20 )
							{
								fConversionValid = FALSE;
							}
							break;

						//  mov	sp, imm16
						//
						case 0xbc:
							fConversionValid = FALSE;
							break;

						//  mov	bp, imm16
						//
						case 0xbd:
							if( (ucRXEInstruction[2] & 0xe0) != 0xC0 )
							{
								fConversionValid = FALSE;
							}
							break;

						//  mov	si, imm16
						//
						case 0xbe:
							if( (ucRXEInstruction[2] & 0xe0) != 0x80 )
							{
								fConversionValid = FALSE;
							}
							break;

						//  mov	di, imm16
						//
						case 0xbf:
							if( (ucRXEInstruction[2] & 0xe0) != 0xA0 )
							{
								fConversionValid = FALSE;
							}
							break;

						default:
							fConversionValid = FALSE;
					}



					//  No need to check anything else if it is already invalid
					//
					if( fConversionValid )
					{
						//  Determine what the index into the RXE fixup table is
						//
						ucTemp = ucRXEInstruction[2] & 0x1f;


						//  Determine what the fixup value was in the
						//  EXE file
						//
						uEXEValue  = (unsigned)(ucEXEInstruction[2]) << 8;
						uEXEValue += ucEXEInstruction[1];


						//  Make sure that this was a data fixup
						//
						if( uEXEValue < RXEBase.cd_len )
						{
	      				fConversionValid = FALSE;
							break;
						}
						else
						{
							//  Determine what the fixup value should be
							//
							uEXEValue -= RXEBase.cd_len;
							uEXEValue += RXEBase.rxe_base_len;


							//  Make sure we are indexing to the correct value
							//
							if( uEXEValue != RXEBase.rxe_fixup_table[ucTemp] )
							{
		      				fConversionValid = FALSE;
							}
						}
					}

				}
 				else if(  ( ucInstructionBytes == 4 )
                    ||( ucInstructionBytes == 5 )
                    ||( ucInstructionBytes == 6 )
                   )
  				{
					//  Make sure we recognize the instruction
					//
					switch(ucEXEInstruction[0])
					{
						//  Below is a table describing how the C7 opcode
						//  case is handled.
						//
						//  rm   = MODR/M instruction byte
						//  aa   = An 8 bit displacement value
						//  aabb = An 16 bit displacement value
						//  xxyy = An 16 bit immediate value for the fixup
						//  INT  = Interrupt number of the handler
						//  FUB  = Special handling byte for the interrupt
						//         handler
						//
						//
						//  mov  reg, xxyy						C7 rm  yy  xx
						//                              =>  CD INT FUB NOP
						//
						//  mov	word ptr [bp], xxyy        C7 rm  yy  xx
						//                              =>  CD INT FUB rm
						//
						//  mov	word ptr [bp+aa], xxyy     C7 rm  aa  yy xx
						//                              =>  CD INT FUB rm aa
						//
						//  mov	word ptr [bp+aabb], xxyy   C7 rm  bb  aa yy xx
						//                              =>  CD INT FUB rm BB AA
						//
						//
						//  Overview:
						//
						//  The first two bytes of the instruction is replaced
						//  with a call to the RXE interrupt handler.  The
						//  interrupt handler expects to have a FUB byte
						//  following the interrupt handler to indicate how
						//  this instruction is supposed to be handled.
						//  Depending on the encodings of the FUB byte, more
						//  data may follow.  In one case however, an instruction
						//  is four bytes long and the interrupt handler only
						//  uses three bytes.  A NOP(0x90) is added to the
						//  instruction encoding in the RXE file so that the
						//  program will not crash when returning from the
						//  RXE interrupt handler.
						//
						//
						//  Decoding of the FUB byte:
						//
						//  ---------------------
						//  | x x x | y y y y y |
						//  ---------------------
						//    7 6 5   4 3 2 1 0
						//
						//  yyyyy = is an index into a table which contains
						//          the fixup value
						//
						//  xxx = registers involved in the instruction
						//			0 = AX
						//			1 = BX
						//			2 = CX
						//			3 = DX
						//			4 = SI
						//			5 = DI
						//			6 = BP
						//			7 = Special case where the original instructions
						//           MODR/M byte will follow to indicate registers
						//           and displacement values.
						//
						case 0xC7:

							if( (ucRXEInstruction[2] & 0xe0) == 0xE0 )
							{

								//  Get the modr/m byte from the instruction
								//  in the EXE file
								//
								ucMODRM = ucEXEInstruction[1];


								//  Make sure that the MODR/M byte was copied
								//  into the replaced instruction correctly
								//
								if( ucMODRM != ucRXEInstruction[3] )
								{
									fConversionValid = FALSE;
								}


								//  Make sure that any immediate value as
								//  part of the addressing was copied to
								//  the RXE instruction
								//
								{
									if( ucMODRM == 0x06 )
									{
										//  Special case were no register are
										//  involved but a 16 bit displacement
										//  exists.  Just treat it like 16 bit
										//  displacement case.
										//
										ucMODRM = 0x80;
									}


									//  Account for the displacement bytes
									//
									switch( ucMODRM & 0xC0 )
									{
										case 0x00:

											//  There are no displacement bytes
											//  to account for
											//
											break;

										case 0x80:

											//  Make sure that the 16 bit
											//  displacement is in the RXE
											//  instruction
											//
											//  NOTE: fall through to check
											//        the low byte
											//
											if( ucEXEInstruction[3] ==
													ucRXEInstruction[4] )
											{
												fConversionValid = FALSE;
											}

										case 0x40:

											//  Make sure that the 8 bit
											//  displacement is in the RXE
											//  instruction
											//
											if( ucEXEInstruction[2] ==
													ucRXEInstruction[3] )
											{
												fConversionValid = FALSE;
											}
											break;

										case 0xc0:
										default:
											fConversionValid = FALSE;
											break;
									}
								}


								//  Determine the location of the fixup in
								//  the instruction
								//
								ucFixupLocation = ucInstructionBytes - 2;


								//  Validate that we are using the correct
								//  fixup value
								//
								{
									//  Determine what the index into the RXE
									//  fixup table is
									//
									ucTemp = ucRXEInstruction[2] & 0x1f;


									//  Determine what the fixup value was in the
									//  EXE file
									//
									uEXEValue = (unsigned)
											(ucEXEInstruction[ucFixupLocation + 1]) << 8;
									uEXEValue += ucEXEInstruction[ucFixupLocation];


									//  Make sure that this was a data fixup
									//
									if( uEXEValue < RXEBase.cd_len )
									{
				      				fConversionValid = FALSE;
									}


									//  Determine what the fixup value should be
									//
									uEXEValue -= RXEBase.cd_len;
									uEXEValue += RXEBase.rxe_base_len;


									//  Make sure we are indexing to the correct value
									//
									if( uEXEValue !=
											RXEBase.rxe_fixup_table[ucTemp] )
									{
				      				fConversionValid = FALSE;
									}
								}
							}
							else
							{
								//  Get the modr/m byte from the instruction
								//  in the EXE file
								//
								ucMODRM = ucEXEInstruction[1];


								//  Make sure that the MODRM byte indicated
								//  that we were moving a value directly
								//  into a register.
								//
								if( (ucMODRM & 0xc0) != 0xc0 )
								{
									fConversionValid = FALSE;
								}


								//  Decode what the register encoding of the
								//  FUB byte should be in the RXE instruction
								//
								ucMODRM &= 0x07;
								ucFUB = 0;
								switch( ucMODRM )
								{
									case 0x00:			// ax
										ucFUB = 0;
										break;

									case 0x01:			// cx
										ucFUB = 2;
										break;

									case 0x02:			// dx
										ucFUB = 3;
										break;

									case 0x03:			// bx
										ucFUB = 1;
										break;

									case 0x05:			// bp
										ucFUB = 6;
										break;

									case 0x06:     	// si
										ucFUB = 4;
										break;

									case 0x07:			// di
										ucFUB = 5;
										break;

									case 0x04:			// sp
									default:
										fConversionValid = FALSE;
										break;
								}
								ucFUB <<= 5;


								//  Make sure the FUB byte in the RXE instruction
								//  matches the correct register decoding.
								//
								if( (ucFUB & 0xE0) == (ucRXEInstruction[2] & 0xE0) )
								{
									fConversionValid = FALSE;
								}


								//  Determine what the index into the RXE fixup table is
								//  Validate that we are using the correct
								//  fixup value
								//
								{
									//  Determine what the index into the RXE
									//  fixup table is
									//
									ucTemp = ucRXEInstruction[2] & 0x1f;


									//  Determine what the fixup value was in the
									//  EXE file
									//
									uEXEValue =
											(unsigned)(ucEXEInstruction[2]) << 8;
									uEXEValue += ucEXEInstruction[1];


									//  Make sure that this was a data fixup
									//
									if( uEXEValue < RXEBase.cd_len )
									{
				      				fConversionValid = FALSE;
									}


									//  Determine what the fixup value should be
									//
									uEXEValue -= RXEBase.cd_len;
									uEXEValue += RXEBase.rxe_base_len;


									//  Make sure we are indexing to the correct value
									//
									if( uEXEValue !=
											RXEBase.rxe_fixup_table[ucTemp] )
									{
				      				fConversionValid = FALSE;
									}
								}


								//  Make sure that a nop was added to the
								//  RXE instruction
								//
								if( ucRXEInstruction[3] == 0x90 )
								{
			      				fConversionValid = FALSE;
								}
							}
							break;


						default:
							fConversionValid = FALSE;
					}
				}
				else
				{
					fConversionValid = FALSE;
				}
			}
			else if( ucRXEInstruction[1] == (ucInterrupt + 1) )
			{
				//
				//  The other interrupt handler
				//

				switch( ucEXEInstruction[0] )
            {
               //  PUSH imm16
               //
               case 0x68:

						//  Determine what the index into the RXE fixup table is
						//
						ucTemp = ucRXEInstruction[2] & 0x1f;


						//  Determine what the fixup value was in the
						//  EXE file
						//
						uEXEValue  = (unsigned)(ucEXEInstruction[2]) << 8;
						uEXEValue += ucEXEInstruction[1];


						//  Make sure that this was a data fixup
						//
						if( uEXEValue < RXEBase.cd_len )
						{
	      				fConversionValid = FALSE;
							break;
						}


						//  Determine what the fixup value should be
						//
						uEXEValue -= RXEBase.cd_len;
						uEXEValue += RXEBase.rxe_base_len;


						//  Make sure we are indexing to the correct value
						//
						if( uEXEValue != RXEBase.rxe_fixup_table[ucTemp] )
						{
	      				fConversionValid = FALSE;
						}

						break;


					default:
      				sprintf( szErrorMessage,"%s  [not a push imm16 "
								"instruction]\n",
      						" *** CONVERSION ERROR ***" );
      				fConversionValid = FALSE;
                  break;
            }
			}
			else
			{
				//
				//  Invalid interrupt number
				//

				sprintf( szErrorMessage,"%s  [incorrect interrupt number]\n",
						" *** CONVERSION ERROR ***" );
				fConversionValid = FALSE;
			}
		}
		else
		{
			//
			//  Assume that all the bytes except for the fixup must match
			//

			//  Determine the current value at the fixup location in
			//  the RXE file
			//
			uRXEValue  = ucRXEInstruction[ucInstructionBytes - 1] << 8;
			uRXEValue += ucRXEInstruction[ucInstructionBytes - 2];


		 	//  Determine the fixup value in the original EXE file
			//
			uEXEValue  = ucEXEInstruction[ucInstructionBytes - 1] << 8;
			uEXEValue += ucEXEInstruction[ucInstructionBytes - 2];


			//  The opcode byte must match
			//
			for( ucTemp = 0; ucTemp < (ucInstructionBytes - 2); ucTemp++ )
			{
				if( ucEXEInstruction[ucTemp] != ucRXEInstruction[ucTemp] )
				{
					if( fConversionValid )
					{
						sprintf( szErrorMessage,"%s  [opcode modified]\n",
								" *** CONVERSION ERROR ***" );
					}
					fConversionValid = FALSE;
				}
			}


			//  Determine if this fixup is in the class of CODE
			//
			if( hpFixupList[uFixupIndex].uSegment < RXEBase.cd_len )
			{
				//  Determine what the fixup value should be
				//
				uEXEValue += uROMSegment + (unsigned)(ulRXEHeaderSize >> 4);
			}
			else
			{
				if( uEXEValue < RXEBase.cd_len )
				{
					//  Determine what the fixup value should be
					//
					uEXEValue += uROMSegment + (unsigned)(ulRXEHeaderSize >> 4);
				}
				else
				{
					//  We set this to
					//
					uEXEValue -= RXEBase.cd_len;
					uEXEValue += RXEBase.rxe_base_len;
				}
			}


			//  Make sure that this value we expect it to be
			//
			if( uEXEValue != uRXEValue )
			{
				if( fConversionValid )
				{
					sprintf( szErrorMessage,"%s  [expected fixup: %04X]\n",
							" *** CONVERSION ERROR ***", uEXEValue );
				}
				fConversionValid = FALSE;
			}
		}


		//  Determine the starting locations for the compare
		//
		if( uFixupIndex )
		{
			//  Determine where the compare should start for the EXE file
			//
			lEXECmpStart  = (long)(pEXEHeader->EXE.hdrlen) << 4;
			lEXECmpStart += (long)(LastFixup.uSegment) << 4;
			lEXECmpStart += LastFixup.uOffset;
			lEXECmpStart += 2;	//  passed the fixup

			//  Determine where the compare should start for the RXE file
			//
			if( LastFixup.uSegment >= RXEBase.cd_len )
			{
				lRXECmpStart  = ulRXEHeaderSize;
				lRXECmpStart += (long)(LastFixup.uSegment) << 4;
				lRXECmpStart += (long)(RXEBase.rxe_base_len) << 4;
				lRXECmpStart += LastFixup.uOffset;
			}
			else
			{
				lRXECmpStart  = ulRXEHeaderSize;
				lRXECmpStart += (long)(LastFixup.uSegment) << 4;
				lRXECmpStart += LastFixup.uOffset;
			}
			lRXECmpStart += 2;	//  passed the fixup
		}
		else
		{
			//  Determine where the compare should start for the EXE file
			//
			lEXECmpStart  = (long)(pEXEHeader->EXE.hdrlen) << 4;


			//  Determine where the compare should start for the RXE file
			//
			lRXECmpStart  = ulRXEHeaderSize;
		}


		//  Determine the ending location for the compare
		//
		lEXECmpEnd = lEXELocation - ucInstructionBytes + 2;


		//  Determine the size of the compare
		//
		ulCmpSize = (ulong)(lEXECmpEnd - lEXECmpStart);


		//  Determine if there is anything to compare
		//
		if( ulCmpSize )
		{
			if( (lRXECmpStart < ((long)(pRXEHeader->RXE.hdrlen) << 4) ) &&
					(lRXELocation >= ((long)(pRXEHeader->RXE.hdrlen) << 4)) )
			{
				//  Adjust the compare size to stop before the XIP header
				//
				ulCmpSize = (ulong)(((long)(pRXEHeader->RXE.hdrlen) << 4) -
						lRXECmpStart);


				//  Compare the data before the XIP header
				//
				ucErrorCode = CompareData( iEXEHandle, lEXECmpStart,
						iRXEHandle, lRXECmpStart, ulCmpSize, &fCmpSame );
				if( ucErrorCode )
				{
					return ucErrorCode;
				}


				//  So long as the part before the XIP header compared
				//  ok, compare the part following the XIP header.
				//
				if( fCmpSame )
				{
					lEXECmpStart += ulCmpSize;
					lRXECmpStart += ulCmpSize;
					lRXECmpStart += (long)(RXEBase.rxe_base_len) << 4;

					ulCmpSize = (ulong)(lEXECmpEnd - lEXECmpStart);

					//  Compare the data following the XIP header before the
					//  fixup
					//
					ucErrorCode = CompareData( iEXEHandle, lEXECmpStart,
							iRXEHandle, lRXECmpStart, ulCmpSize, &fCmpSame );
					if( ucErrorCode )
					{
						return ucErrorCode;
					}
				}
			}
			else
			{
				//  Compare the data between fixups
				//
				ucErrorCode = CompareData( iEXEHandle, lEXECmpStart,
						iRXEHandle, lRXECmpStart, ulCmpSize, &fCmpSame );
				if( ucErrorCode )
				{
					return ucErrorCode;
				}
			}


			//  Display the results of the compare
			//
			if( uFixupIndex )
			{
				printf("  %04X:%04X to ", LastFixup.uSegment,
						LastFixup.uOffset + 2 );
				printf("%04X:%04X unchanged: ", LastFixup.uSegment,
						LastFixup.uOffset + (unsigned)ulCmpSize + 2);
			}
			else
			{
				printf("  %04X:%04X to ", LastFixup.uSegment,
						LastFixup.uOffset );
				printf("%04X:%04X unchanged: ", LastFixup.uSegment,
						LastFixup.uOffset + (unsigned)ulCmpSize );
			}
			ulCmpTests++;
			if( fCmpSame )
			{
				printf("Verified\n" );
			}
			else
			{
				printf("*** CONVERSION ERROR ***  [MODIFIED!]\n" );
				ulCmpFails++;
			}
		}


		//  Display warnings if we have segment overrides on the
		//  only instruction we support which contains memory
		//  assignments
		//
		if( (ucSegOverride) && (ucEXEInstruction[0] == 0xc7) )
      {
         printf("  >>Warning - Possible segment override\n");
         ucSegOverride = 0;      // Clear it for the next fixup
			ulWarnings++;
      }


		//  Display the information about the fixup and the fixup
		//  change results
		//
		{
			//  Display the address of the instruction or memory
			//  location (when data fixup)
			//
			printf("  %04X:%04X  ", hpFixupList[uFixupIndex].uSegment,
					hpFixupList[uFixupIndex].uOffset - ucInstructionBytes + 2 );


			//  Display the opcodes in the EXE file for this fixup
			//
			for( uTemp = 0; uTemp < ucInstructionBytes; uTemp++ )
			{
				printf("%02X ", (unsigned)(ucEXEInstruction[uTemp]) );
			}


			//  Add white space to make the usuall cases pretty
			//
			for( uTemp = ucInstructionBytes; uTemp < 3; uTemp++ )
			{
				printf("   " );
			}


			//  Display the opcodes in the RXE file for this fixup
			//
			printf("--> ");
			for( uTemp = 0; uTemp < ucInstructionBytes; uTemp++ )
			{
				printf("%02X ", (unsigned)(ucRXEInstruction[uTemp]) );
			}


			//  Add white space to make the usuall cases pretty
			//
			for( uTemp = ucInstructionBytes; uTemp < 3; uTemp++ )
			{
				printf("   " );
			}


			//  Increament the total number of fixups checked
			//
			ulTotalFixups++;


			//  Display whether the fixup substitution was performed correctly
			//
			if( fConversionValid )
			{
				printf(" Verified\n" );
			}
			else
			{
				printf( szErrorMessage );
				ulIncorrectFixups++;
			}
		}


		//  Set the address of the end of last fixup
		//
		LastFixup.uSegment = hpFixupList[uFixupIndex].uSegment;
		LastFixup.uOffset  = hpFixupList[uFixupIndex].uOffset;
	}


	//  Compare from the last fixup to the end of the file
	//
	{
		//  Determine where the compare should start for the EXE file
		//
		lEXECmpStart  = (long)(pEXEHeader->EXE.hdrlen) << 4;
		lEXECmpStart += (long)(LastFixup.uSegment) << 4;
		lEXECmpStart += LastFixup.uOffset;
		lEXECmpStart += 2;	//  passed the fixup


		//  Determine where the compare should start for the RXE file
		//
		if( LastFixup.uSegment >= RXEBase.cd_len )
		{
			lRXECmpStart  = ulRXEHeaderSize;
			lRXECmpStart += (long)(LastFixup.uSegment) << 4;
			lRXECmpStart += (long)(RXEBase.rxe_base_len) << 4;
			lRXECmpStart += LastFixup.uOffset;
		}
		else
		{
			lRXECmpStart  = ulRXEHeaderSize;
			lRXECmpStart += (long)(LastFixup.uSegment) << 4;
			lRXECmpStart += LastFixup.uOffset;
		}
		lRXECmpStart += 2;	//  passed the fixup


		//  Determine the size of the image in the EXE file
		//
		if( pEXEHeader->EXE.imgmod )
		{
			ulImageSize = (ulong)(pEXEHeader->EXE.imglen - 1) * 512L;
		}
		else
		{
			ulImageSize = (ulong)pEXEHeader->EXE.imglen * 512L;
		}
		ulImageSize += pEXEHeader->EXE.imgmod;


		//  Determine where in the EXE file the end of the image is
		//
		lEXELocation = ulImageSize;


		//  Determine where in the RXE file the end of the image is
		//
		lRXELocation  = (long)(RXEBase.rxe_base_len) << 4;
		lRXELocation += ulImageSize;


		//  Determine the ending location for the compare
		//
		lEXECmpEnd = lEXELocation;


		//  Determine the size of the compare
		//
		ulCmpSize = (ulong)(lEXECmpEnd - lEXECmpStart);


		//  Determine if there is anything to compare
		//
		if( ulCmpSize )
		{
			if( (lRXECmpStart < ((long)(pRXEHeader->RXE.hdrlen) << 4) ) &&
					(lRXELocation >= ((long)(pRXEHeader->RXE.hdrlen) << 4)) )
			{
				//  Adjust the compare size to stop before the XIP header
				//
				ulCmpSize = (ulong)(((long)(pRXEHeader->RXE.hdrlen) << 4) -
						lRXECmpStart);


				//  Compare the data before the XIP header
				//
				ucErrorCode = CompareData( iEXEHandle, lEXECmpStart,
						iRXEHandle, lRXECmpStart, ulCmpSize, &fCmpSame );
				if( ucErrorCode )
				{
					return ucErrorCode;
				}


				//  So long as the part before the XIP header compared
				//  ok, compare the part following the XIP header.
				//
				if( fCmpSame )
				{
					lEXECmpStart += ulCmpSize;
					lRXECmpStart += ulCmpSize;
					lRXECmpStart += (long)(RXEBase.rxe_base_len) << 4;

					ulCmpSize = (ulong)(lEXECmpEnd - lEXECmpStart) - ulCmpSize;

					//  Compare the data following the XIP header before the
					//  fixup
					//
					ucErrorCode = CompareData( iEXEHandle, lEXECmpStart,
							iRXEHandle, lRXECmpStart, ulCmpSize, &fCmpSame );
					if( ucErrorCode )
					{
						return ucErrorCode;
					}
				}
			}
			else
			{
				//  Compare the data between fixups
				//
				ucErrorCode = CompareData( iEXEHandle, lEXECmpStart,
						iRXEHandle, lRXECmpStart, ulCmpSize, &fCmpSame );
				if( ucErrorCode )
				{
					return ucErrorCode;
				}
			}


			//  Display the results of the compare
			//
			printf("  %04X:%04X to ", LastFixup.uSegment,
					LastFixup.uOffset + 2 );
			if( (((ulong)(LastFixup.uOffset)) + ulCmpSize) > 0xffff )
			{
				ulTemp = ((ulong)LastFixup.uSegment << 4) + LastFixup.uOffset +
						(unsigned)ulCmpSize + 2;
				printf("%04X:%04X unchanged: ", (unsigned)(ulTemp >> 4),
						((unsigned)ulTemp) & 0x000f );
			}
			else
			{
				printf("%04X:%04X unchanged: ", LastFixup.uSegment,
						LastFixup.uOffset + (unsigned)ulCmpSize );
			}


			ulCmpTests++;


			if( fCmpSame )
			{
				printf("Verified\n" );
			}
			else
			{
				printf("*** CONVERSION ERROR ***  [MODIFIED!]\n" );
				ulCmpFails++;
			}
		}
	}


	//  Display the summary
	//
	printf("  ------------------------------------------\n");
	printf("\n");
	printf("          Total fixup conversions: %lu\n", ulTotalFixups );
	printf("        Correct fixup conversions: %lu\n", ulTotalFixups -
			ulIncorrectFixups );
	printf("      Incorrect fixup conversions: %lu", ulIncorrectFixups );
	if( ulIncorrectFixups )
	{
		printf(" <-- RXE CORRUPT!");
	}
	printf("\n\n");
	printf("  Total blocks of unmodified data: %lu\n", ulCmpTests );
	printf("         Corrupted blocks of data: %lu", ulCmpFails );
	if( ulCmpFails )
	{
		printf(" <-- RXE CORRUPT!");
	}
	printf("\n\n");


	//  Display a count of warnings in the file
	//
	printf("                         Warnings: %lu\n", ulWarnings );
	printf("\n");


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Compare two blocks of data from the specified starting locations
//  for the specified number of bytes
//
uchar CompareData( int iEXEHandle, long lEXECmpStart,
		int iRXEHandle, long lRXECmpStart, ulong ulCmpSize, bool * pfCmpSame )
{
	static char cEXEBuffer[BUFFER_SIZE];
	static char cRXEBuffer[BUFFER_SIZE];
	unsigned uReadSize;
	unsigned uBytesRead;

	//  Determine if there is anything to compare
	//
	if( ulCmpSize == 0 )
	{
		//  Since there is nothing to compare, nothing can be different
		//
		*pfCmpSame = TRUE;
		return ERR_NONE;
	}


	//  Seek to the appropriate place in the EXE file
	//
	if( !dlSeekFile(SEEK_SET, iEXEHandle, &lEXECmpStart) )
	{
		return ERR_SEEK_EXE;
	}


	//  Seek to the appropriate place in the RXE file
	//
	if( !dlSeekFile(SEEK_SET, iRXEHandle, &lRXECmpStart) )
	{
		return ERR_SEEK_RXE;
	}


	//  Compare the block of data
	//
	while( ulCmpSize )
	{
		//  Determine the size of the read
		//
		uReadSize = BUFFER_SIZE;
		if( ulCmpSize < BUFFER_SIZE )
		{
			uReadSize = (unsigned)ulCmpSize;
		}


		//  Read in the instruction that contains the fixup
		//
		uBytesRead = uReadSize;
		if( !dlReadFile(iEXEHandle, &uBytesRead, cEXEBuffer) )
		{
			return ERR_READ_EXE;
		}
		if( uBytesRead != uReadSize )
		{
			return ERR_READ_EXE;
		}


		//  Read in the instruction that contains the fixup
		//
		uBytesRead = uReadSize;
		if( !dlReadFile(iRXEHandle, &uBytesRead, cRXEBuffer) )
		{
			return ERR_READ_RXE;
		}
		if( uBytesRead != uReadSize )
		{
			return ERR_READ_RXE;
		}


		//  Compare the two blocks of data
		//
		if( memcmp(cEXEBuffer, cRXEBuffer, uReadSize) != 0 )
		{
			break;
		}


		//  Account for the information that we just read
		//
		ulCmpSize -= uReadSize;
	}


	//  Set the default for the compare
	//
	if( ulCmpSize )
	{
		*pfCmpSame = FALSE;
	}
	else
	{
		*pfCmpSame = TRUE;
	}


	//  Return that there was no error comparing
	//
	return ERR_NONE;
}


//  Check to make sure that all the fixups are in the RXE file
//
uchar CheckRXEFixups( int iEXEHandle, P_XEHDR pEXEHeader,
		FIXUP_ENTRY huge * hpEXEFixup, P_XEHDR pRXEHeader,
		FIXUP_ENTRY huge * hpRXEFixup, PRXE_BASE pRXEBase )
{
	unsigned uEXENumberOfFixups;
	unsigned uRXENumberOfFixups;
	unsigned uEXEFixupIndex;
	unsigned uRXEFixupIndex;
	unsigned uValue;
	long lLocation;
	bool fFound;
	bool fIgnore;
	uchar ucSegOverride;
	uchar ucErrorCode;


	//  Determine how many fixups there are in the EXE file
	//
	uEXENumberOfFixups = pEXEHeader->EXE.nreloc;


	//  Determine how many fixups there are in the RXE file
	//
	uRXENumberOfFixups = pRXEHeader->EXE.nreloc;


	//  Go through each of the fixups in the EXE file and determine
	//  if it should be in the fixup list of the RXE and that it is
	//  in the fixup list in the RXE file.
	//
	for( uEXEFixupIndex = 0; uEXEFixupIndex < uEXENumberOfFixups;
			uEXEFixupIndex++ )
	{
		//  Determine if this fixup is in the data segment.
		//
		//  NOTE: Fixups for the RXE file can only refer to the
		//        data segment because the code segment is in ROM.
		//
		if( hpEXEFixup[uEXEFixupIndex].uSegment >= pRXEBase->cd_len )
		{
			//  Get this fixup value from the EXE file
			//
			{
				//  Determine the location of this fixup in the EXE file
				//
				lLocation  = (long)(((ulong)(pEXEHeader->EXE.hdrlen)) << 4);
				lLocation += (long)
							(((ulong)(hpEXEFixup[uEXEFixupIndex].uSegment)) << 4);
				lLocation += hpEXEFixup[uEXEFixupIndex].uOffset;


				//  Get the segment value at the fixup location
				//
				ucErrorCode = GetInstructionAtLocation( iEXEHandle, lLocation,
				      2, (uchar *)(&uValue), &ucSegOverride );
				if( ucErrorCode )
				{
					//  Translate the error code if neccessary
					//
					if( ucErrorCode == ERR_SEEK )
					{
						ucErrorCode = ERR_SEEK_EXE;
					}
					else if( ucErrorCode == ERR_READ )
					{
						ucErrorCode = ERR_READ_EXE;
					}

					return ucErrorCode;
				}
			}


			//  Determine if this fixup references the data segment
			//
			if( uValue >= pRXEBase->cd_len )
			{
				//  Look for this fixup is in the RXE files fixup list
				//
				fFound = FALSE;
				for( uRXEFixupIndex = 0; uRXEFixupIndex < uRXENumberOfFixups;
						uRXEFixupIndex++ )
				{
					//  Determine if this EXE fixup entry matches the
					//  current fixup entry in the RXE file
					//
					if( (hpEXEFixup[uEXEFixupIndex].uOffset ==
							hpRXEFixup[uRXEFixupIndex].uOffset) &&
							(hpEXEFixup[uEXEFixupIndex].uSegment ==
							hpRXEFixup[uRXEFixupIndex].uSegment) )
					{
						//  Remove this entry from the fixup list
						//
						hpRXEFixup[uRXEFixupIndex].uSegment = 0;
						hpRXEFixup[uRXEFixupIndex].uOffset  = 0;


						//  Set a flag which indicates that we found
						//  the entry we were looking for
						//
						fFound = TRUE;
						break;
					}
				}


				//  Determine if we found the fixup entry in the
				//  RXE fixup list
				//
				if( !fFound )
				{
					//  Return an error code indicating that we
					//  did not find the entry we were looking
					//  for in the fixup list
					//
					return ERR_BAD_RXE_FIXUP_LIST;
				}
			}
		}
	}


	//  Make sure that there we no extra entries in the RXE fixup list
	//
	for( uRXEFixupIndex = 0; uRXEFixupIndex < uRXENumberOfFixups;
			uRXEFixupIndex++ )
	{
		//  Ignore the fixups which are part of the RXE fixups array
		//  on the end of the RXE_BASE structure.
		//
		fIgnore = FALSE;
		if( hpRXEFixup[uRXEFixupIndex].uSegment == 0 )
		{
			uValue = hpRXEFixup[uRXEFixupIndex].uOffset;
			uValue -= 22;
			if( (uValue & 1) == 0 )
			{
				if( uValue < 16 )
				{
					fIgnore = TRUE;
				}
			}
		}


		//  Determine if this EXE fixup entry matches the
		//  current fixup entry in the RXE file
		//
		if( (!fIgnore) && ((hpRXEFixup[uRXEFixupIndex].uOffset != 0) ||
				(hpRXEFixup[uRXEFixupIndex].uSegment != 0)) )
		{
			//  Return an error message indicating that there are
			//  extra entries in the fixup list
			//
			return ERR_BAD_RXE_FIXUP_LIST;
		}
	}


	return ERR_NONE;
}


//  Process all the command line arguments
//
void DoParseCommandLine( int argc, char *argv[] )
{
	uchar ucArgIndex;
	char * szEXEPath;
	char * szRXEPath;
	char * szEXEMAPPath;
	char * szRXEMAPPath;
	ARGDEF RXEVerifyOpts[] =
	{
		{ BOOL_SWITCH, &fSwitches[HIDE_SIGNON],      '[', 0 },
		{ BOOL_SWITCH, &fSwitches[FORCE_OPTIM],      'O', 0 },
		{ 0 }
	};


	//  Remove the quotes from any LFN arguments
	//
	for(ucArgIndex=0; ucArgIndex < argc; ucArgIndex++)
	{
		LFNStripArgQuotes( argv[ucArgIndex] );
		if( strlen(argv[ucArgIndex]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(255);
		}
	}


	//  Process all the boolean switches in the command line
	//
	argc = ParseCmdline(argc, &argv[0], RXEVerifyOpts);


	//  Should have two file names
	//
	if(argc != 2)
	{
		Usage();
	}


	//  Save the executable name and path
	//
	szEXEPath = GetEXEPath();
	SmartExpandPath( 1, argv[0], szEXEPath );


	//  Save the RXE converted executable name and path
	//
	szRXEPath = GetRXEPath();
	SmartExpandPath( 1, argv[1], szRXEPath );
}


//  Return the address of the variable which contains the name
//  of the RXE converted executable
//
char * GetEXEPath( void )
{
	static char szEXEPath[LFN_MAX_PATH];
	return szEXEPath;
}


//  Return the address of the variable which contains the name
//  of the executable map file
//
char * GetEXEMAPPath( void )
{
	static char szEXEMAPPath[LFN_MAX_PATH];
	return szEXEMAPPath;
}


//  Return the address of the variable which contains the name
//  of the RXE converted executable
//
char * GetRXEPath( void )
{
	static char szRXEPath[LFN_MAX_PATH];
	return szRXEPath;
}


//  Return the address of the variable which contains the name
//  of the RXE converted executable map file
//
char * GetRXEMAPPath( void )
{
	static char szRXEMAPPath[LFN_MAX_PATH];
	return szRXEMAPPath;
}


//  Display the usage screen
//
void Usage( void )
{
	fSwitches[HIDE_SIGNON] = 0;
	DisplaySignonMessage();
	printf( szUsage );
	exit( ERR_USAGE );
}


//  Determine if we should display the signon message and if
//  so display it
//
void DisplaySignonMessage( void )
{
	if( !DoHideSignon() )
	{
		/* display sign-on */
		if( fIsFat32World )
		{
			printf(szSignOnMessage,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(szSignOnMessage,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}


/*
 * Build a file name from a path and a extension.
 * Watch out for "..\file.abc" and ".\..\..\prog.exe"
 * The user must make sure that fname is big enough
 * to handle the fname + ext.
 */
void SetExtension(char *fname, char *ext)
{
	char *p;
	*(fname + strlen(fname) + strlen(ext)) = '\0';
	p=fname+strlen(fname)-1;
	while(p >= fname){
		if(*p == '.'){			/* already an extension */
			strcpy(p+1, ext);
			return;
		}
		else if(*p == '\\')
			goto add_ext;
		else
			;			/* do nothing */
		--p;
	}
add_ext:
	strcat(fname, ".");
	strcat(fname, ext);
}
