//
//   $Id: rxecomon.c 1.4 2002/10/16 16:12:00Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  INSTSIZE.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//  10-26-2000  bt  Created
//


//  Borland libraries
//
#include "dir.h"
//#include "stdio.h"
//#include "stdlib.h"
//#include "ctype.h"
//#include "mem.h"
#include "string.h"


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dosver.h"
#include "fatworld.h"
#include "dlapp.h"
#include "dllfn.h"


#ifdef RXEVERFY

//  RXEVERFY header files
//
#include "rxeverfy.h"


//  RXEVERFY Prototypes
//
#include "rxeverfy.pt"

#endif

#ifdef RXEOPTIM

//  HEADER FILES
//
#include "RXEOPTIM.h"


//  PROTOTYPES
//
#include "RXEOPTIM.pt"

#endif


//  Determine the size of the instruction
//
uchar GetInstructionSize(
      int iEXEHandle, long lEXELocation, bool fIsDataFixup,
      int iRXEHandle, long lRXELocation, uchar ucInterrupt,
		unsigned uDataSegment, uchar * pucInstructionBytes )
{
	unsigned uReadBytes;
	uchar ucEXEInstruction[20];
	uchar ucRXEInstruction[20];
	unsigned * puTemp;
	uchar ucErrorCode;
   uchar ucSegOverride = 0;
	uchar ucHitCount = 0;
	uchar ucHitMask = 0;       // Keeps track of instruction sizes
                              // and a count of the "hits"

	//  Determine if this is a fixup in the data area
	//
	if( fIsDataFixup )
	{
		//  Only the actual fixup needs to be changed in the data area
		//
		*pucInstructionBytes = 2;
	}
	else
	{
		//  Determine if this fixup in the code segment refers to code.
		//  if it does refer to code then we do not need to know the size
		//  of the instruciton.
		//
		{
			//  Go the location of the fixup in the file
			//
			if( !dlSeekFile(SEEK_SET, iEXEHandle, &lEXELocation) )
			{
				return ERR_SEEK;
			}


			//  Read in the fixup
			//
			uReadBytes = 2;
			if( !dlReadFile(iEXEHandle, &uReadBytes, (char *)ucEXEInstruction) )
			{
				return ERR_READ;
			}
			if( uReadBytes != 2 )
			{
				return ERR_READ;
			}


			//  Determine if this fixup refers to code
			//
			puTemp = (unsigned *)(ucEXEInstruction);
			if( *puTemp < uDataSegment )
			{
				//  Set the number of bytes in the instruction to two since
				//  we only need to examine the fixup
				//
				*pucInstructionBytes = 2;

				return ERR_NONE;
			}
		}


		//  Seek to the appropriate place in the file
		//
		lEXELocation -= 4;
		if( !dlSeekFile(SEEK_SET, iEXEHandle, &lEXELocation) )
		{
			return ERR_SEEK;
		}


		//  Read in the instruction that contains the fixup
		//
		uReadBytes = 6;
		if( !dlReadFile(iEXEHandle, &uReadBytes, (char *)ucEXEInstruction) )
		{
			return ERR_READ;
		}
		if( uReadBytes != 6 )
		{
			return ERR_READ;
		}


      // Check the possible cases.
      // If we get more than 1 hit, we have an ambiguous instruction.
      //
      switch( ucEXEInstruction[0] )    // fixup - 6
      {
         case 0x69:        // IMUL
         case 0x81:        // ADD/OR/ADC/SBB/AND/SUB/XOR/CMP
         case 0xC7:        // MOV
         case 0xF7:        // TEST
				*pucInstructionBytes = 6;
            ucHitMask |= 0x80;
            ucHitCount++;
      }
      switch( ucEXEInstruction[1] )    // fixup - 5
      {
         case 0x69:        // IMUL
         case 0x81:        // ADD/OR/ADC/SBB/AND/SUB/XOR/CMP
         case 0xC7:        // MOV
         case 0xF7:        // TEST
				*pucInstructionBytes = 5;
            ucHitMask |= 0x40;
            ucHitCount++;
      }
      switch( ucEXEInstruction[2] )    // fixup - 4
      {
         case 0x69:        // IMUL
         case 0x81:        // ADD/OR/ADC/SBB/AND/SUB/XOR/CMP
         case 0xC7:        // MOV
         case 0xF7:        // TEST
				*pucInstructionBytes = 4;
            ucHitMask |= 0x20;
            ucHitCount++;
      }
      switch( ucEXEInstruction[3] )    // fixup - 3
      {
         case 0x05:        // ADD AX,
         case 0x0D:        // OR AX,
         case 0x15:        // ADC AX,
         case 0x1D:        // SBB AX,
         case 0x25:        // AND AX,
         case 0x2D:        // SUB AX,
         case 0x35:        // XOR AX,
         case 0x3D:        // CMP AX,
         case 0x68:        // PUSH
         case 0xA9:        // TEST AX,
         case 0xB8:        // MOV AX,
         case 0xB9:        // MOV CX,
         case 0xBA:        // MOV DX,
         case 0xBB:        // MOV BP,
         case 0xBC:        // MOV SP,
         case 0xBD:        // MOV BP,
         case 0xBE:        // MOV SI,
         case 0xBF:        // MOV DI,
         case 0xC2:        // RET     (bug)
         case 0xC8:        // ENTER   (bug)
         case 0xCA:        // RETF    (bug)
				*pucInstructionBytes = 3;
            ucHitMask |= 0x10;
            ucHitCount++;
      }

      if( ucHitCount > 1 )
      {
         // We have a potential ambiguity, so go read the bytes from the
         // corresponding location in the RXE file, and find the CD intno

   		//  Read the instruction from the RXE file
   		//
   		ucErrorCode = GetInstructionAtLocation( iRXEHandle,
   				lRXELocation - MAX_INSTRUCTION_LENGTH + 2,
   				MAX_INSTRUCTION_LENGTH, ucRXEInstruction, &ucSegOverride );

   		if( ucErrorCode )
   		{
   			//  Translate the error code if neccessary
   			//
   			if( ucErrorCode == ERR_SEEK )
   			{
   				ucErrorCode = ERR_SEEK_RXE;
   			}
   			else if( ucErrorCode == ERR_READ )
   			{
   				ucErrorCode = ERR_READ_RXE;
   			}

   			return ucErrorCode;
   		}

         // Now we've got it, what are we going to do with it?
         //
         if(  ( ucHitMask & 0x80 )
            &&( ucRXEInstruction[0] == 0xCD)
            &&(  ( ucRXEInstruction[1] == ucInterrupt )
               ||( ucRXEInstruction[1] == ucInterrupt + 1 )
              )
           )
         {
				*pucInstructionBytes = 6;
         }
         else if(  ( ucHitMask & 0x40 )
                 &&( ucRXEInstruction[1] == 0xCD)
                 &&(  ( ucRXEInstruction[2] == ucInterrupt )
                    ||( ucRXEInstruction[2] == ucInterrupt + 1 )
                   )
                )
         {
				*pucInstructionBytes = 5;
         }
         else if(  ( ucHitMask & 0x20 )
                 &&( ucRXEInstruction[2] == 0xCD)
                 &&(  ( ucRXEInstruction[3] == ucInterrupt )
                    ||( ucRXEInstruction[3] == ucInterrupt + 1 )
                   )
                )
         {
				*pucInstructionBytes = 4;
         }
         else if(  ( ucHitMask & 0x10 )
                 &&( ucRXEInstruction[3] == 0xCD)
                 &&(  ( ucRXEInstruction[4] == ucInterrupt )
                    ||( ucRXEInstruction[4] == ucInterrupt + 1 )
                   )
                )
         {
				*pucInstructionBytes = 3;
         }
         else
         {
            // Error case - the "CD" found in the RXE doesn't match any
            // valid instruction length found in the EXE file.
            //
				*pucInstructionBytes = 2;
         }
      }
		else if( ucHitCount != 1 )
		{
			//
			//  We don't know what the instruction is! Default instruction size
			//  will be set to two because no instruction which contains a
			//  fixup can be two.
			//
			*pucInstructionBytes = 2;
		}
		else
		{
			//
			//  We know what the instruction is and the instruction
			//  size is set.
			//
		}
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Read the instruction which contains the fixup given the handle and
//  the location in the file.
//
uchar GetInstructionAtLocation( int iHandle, long lLocation,
		uchar ucInstructionBytes, uchar * ucInstruction,
		uchar * ucSegOverride )
{
	unsigned uReadBytes;


	//  Initialize the variable which may contain a segment override
	//  if there is one
	//
	*ucSegOverride = 0;


	//  Seek to the appropriate place in the file
	//
   if (lLocation > 0)
   {
      // While we're at it, go one more back and pull the possible
      // Segment Override byte
      //

      lLocation--;
   	if( !dlSeekFile(SEEK_SET, iHandle, &lLocation) )
   	{
   		return ERR_SEEK;
   	}
      uReadBytes = 1;
   	if( !dlReadFile(iHandle, &uReadBytes, (char *)ucSegOverride ))
   	{
   		return ERR_READ;
   	}


		// Now test the fixup to see if it's valid - clear it if not.
	   //
	   if(  (*ucSegOverride != 0x26)       // 001 00 110 = ES:
	      &&(*ucSegOverride != 0x2E)       // 001 01 110 = CS:
	      &&(*ucSegOverride != 0x36)       // 001 10 110 = SS:
	      &&(*ucSegOverride != 0x3E)       // 001 11 110 = DS:
//	   	&&(*ucSegOverride != 0x64)       // 011 00 100 = FS: (386 only)
//	   	&&(*ucSegOverride != 0x65)       // 011 00 101 = GS: (386 only)
	     )
	   {
	      *ucSegOverride = 0;        // Not valid, clear it
	   }
	}
   else
   {
   	if( !dlSeekFile(SEEK_SET, iHandle, &lLocation) )
   	{
   		return ERR_SEEK;
   	}
   }


	//  Read in the instruction that contains the fixup
	//
	uReadBytes = ucInstructionBytes;
	if( !dlReadFile(iHandle, &uReadBytes, (char *)ucInstruction) )
	{
		return ERR_READ;
	}
	if( uReadBytes != ucInstructionBytes )
	{
		return ERR_READ;
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Read the fixup list in an EXE file into the fixup structures
//
uchar LoadFixupList( int iHandle, P_XEHDR pEXEHeader,
		LPFIXUP_ENTRY lpFixupList )
{
	unsigned uReadBytes;
	long lLocation;
	unsigned uFixupCount;


	//  Seek to the beginning of the fixup list
	//
	lLocation = (long)(pEXEHeader->EXE.reldsp);
	if( !dlSeekFile(SEEK_SET, iHandle, &lLocation) )
	{
		return ERR_SEEK;
	}


	//  Determine the number of fixups in the list
	//
	uFixupCount = pEXEHeader->EXE.nreloc;


	//  Read in all the fixup locations
	//
	while( uFixupCount )
	{
		uReadBytes = sizeof( FIXUP_ENTRY );
		if( !dlReadFile(iHandle, &uReadBytes, (char far *)lpFixupList) )
		{
			return ERR_READ;
		}
		if( uReadBytes != sizeof(FIXUP_ENTRY) )
		{
			return ERR_READ;
		}

		(char huge *)lpFixupList += sizeof( FIXUP_ENTRY );
		uFixupCount--;
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Determine if the STDOUT is CON
//
bool STDOUTIsCON( void )
{
	unsigned uAttributes;
	bool fSuccessful;


	fSuccessful = dlGetDeviceAttributes( 1, &uAttributes );
	if( fSuccessful )
	{
		if( (uAttributes & 0x11) == 0x11 )
		{
			return TRUE;
		}
		else
		{
			return FALSE;
		}
	}
	else
	{
		return TRUE;
	}
}


//  Sort the fixup list from least to greatest
//
void SortFixupList( unsigned uRelocItems, FIXUP_ENTRY huge * hpFixupList )
{
	FIXUP_ENTRY	Temp;
	unsigned uCurEntry;
	unsigned uCheckEntry;
	ulong ulCurFixup;
	ulong ulCheckFixup;
	static char szPercent[] = "\x0d%3d%% percent complete\x0d";
	unsigned uPercent;
	unsigned uLastPercent = 0xffff;
	bool fSTDOUTIsCON;


	//  Determine if the STDOUT is CON
	//
	fSTDOUTIsCON = STDOUTIsCON();


	//
	//
	printf("Sorting the fixup list ...\n");


	//
	//
	for( uCurEntry = 0; uCurEntry < uRelocItems; uCurEntry++ )
	{
		//  Convert this segmented address into a physical address
		//
		ulCurFixup  = (long)(hpFixupList[uCurEntry].uSegment) << 4;
		ulCurFixup += hpFixupList[uCurEntry].uOffset;


		//  Determine if we are being redirected to a file
		//
		if( fSTDOUTIsCON )
		{
			//  Determine the percentage we have done
			//
			if( uCurEntry != (uRelocItems - 1) )
			{
				uPercent =
						(unsigned)(((ulong)uCurEntry * 100)/ (uRelocItems-1));
			}
			else
			{
				uPercent = 100;
			}
		}
		else
		{
			//  When the output is being redirected, we only need to
			//  display one line with the max percentage.
			//
			uPercent = 100;
		}


		//  Determine if the percentage has changed.  If so update the
		//  percentage on the screen
		//
		if( uPercent != uLastPercent )
		{
			uLastPercent = uPercent;
			printf( szPercent, uPercent );
		}


		//  Sort the fixup list from least to greatest
		//
		for( uCheckEntry = uCurEntry+1; uCheckEntry < uRelocItems;
				uCheckEntry++ )
		{
			//  Convert this segmented address into a physical address
			//
			ulCheckFixup  = (long)(hpFixupList[uCheckEntry].uSegment) << 4;
			ulCheckFixup += hpFixupList[uCheckEntry].uOffset;


			//  Determine if the current entry is at a higher address
			//  than the entry we are checking
			//
			if( ulCurFixup > ulCheckFixup )
			{
				//  Exchange the current entry with the one we are checking
				//
				Temp.uSegment = hpFixupList[uCurEntry].uSegment;
				Temp.uOffset  = hpFixupList[uCurEntry].uOffset;
				hpFixupList[uCurEntry].uSegment =
						hpFixupList[uCheckEntry].uSegment;
				hpFixupList[uCurEntry].uOffset =
						hpFixupList[uCheckEntry].uOffset;
				hpFixupList[uCheckEntry].uSegment = Temp.uSegment;
				hpFixupList[uCheckEntry].uOffset  = Temp.uOffset;


				//  Convert this segmented address into a physical address
				//
				ulCurFixup  = (long)(hpFixupList[uCurEntry].uSegment) << 4;
				ulCurFixup += hpFixupList[uCurEntry].uOffset;
			}
		}
	}

	printf("\n\n");
}



//  Read the EXE header into the provided EXE structure from
//  the specified handle
//
uchar LoadEXEHeader( int iHandle, P_XEHDR pEXEHeader )
{
	unsigned uReadBytes;
	long lLocation;


	//  Seek to the beginning of the file where the EXE header should be
	//
	lLocation = 0;
	if( !dlSeekFile(SEEK_SET, iHandle, &lLocation) )
	{
		return ERR_SEEK;
	}


	//  Read in the EXE header
	//
	uReadBytes = sizeof( _XEHDR );
	if( !dlReadFile(iHandle, &uReadBytes, (char *)pEXEHeader) )
	{
		return ERR_READ;
	}
	if( uReadBytes != sizeof(_XEHDR) )
	{
		return ERR_READ;
	}


	//  Make sure that this appears to be a valid EXE header
	//
	if( pEXEHeader->EXE.sig != 0x5a4d )
	{
		return ERR_BAD_EXE;
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}


//  Read the RXE_BASE/XIP header into the provided structure from
//  the specified handle
//
uchar LoadRXEBASE( int iRXEHandle, P_XEHDR pRXEHeader, PRXE_BASE pRXEBase )
{
	unsigned uReadBytes;
	long lLocation;


	//  Seek to the place in the file where the RXE_BASE/XIP
	//  header should be
	//
	lLocation = (long)(pRXEHeader->RXE.hdrlen) << 4;
	if( !dlSeekFile(SEEK_SET, iRXEHandle, &lLocation) )
	{
		return ERR_SEEK_RXE;
	}


	//  Read in the EXE header
	//
	uReadBytes = sizeof( RXE_BASE );
	if( !dlReadFile(iRXEHandle, &uReadBytes, (char *)pRXEBase) )
	{
		return ERR_READ_RXE;
	}
	if( uReadBytes != sizeof( RXE_BASE) )
	{
		return ERR_READ_RXE;
	}


	//  Make sure that this is the XIP structure
	//
	if( pRXEBase->rxe_sig != XIPSIG )
	{
		return ERR_BAD_RXE;
	}


	//  Return that there was no error
	//
	return ERR_NONE;
}
