//
//   $Id: FTPPARSE.C 1.3 2003/01/27 20:20:38Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Portions Copyright (c) 1993-2002 GP van Niekerk Ondernemings
//   Portions Copyright (c) 1986-1989 Phil Karn, KA9Q
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/* CMDPARSE.C
 * Parse command line, set up command arguments Unix-style, and call function.
 * Note: argument is modified (delimiters are overwritten with nulls)

Updated @ Date 4 November 1995, 23:38 by Gerrit van Niekerk
*/

#include <stdio.h>
#include <string.h>
#include "inetdefs.h"
#include "ftpparse.h"

static void near print_err_line(int argc,char **argv)
{
	while (argc--)
		printf("%s ",*argv++);
	printf("\n");
}

int cmdparse(struct cmds *cmds,char *line,char error)
{
	struct cmds *cmdp;
	char *argv[NARG],*cp;
	int argc,qflag;
	int rslt,len;

	/* Remove cr/lf */
	if((cp = strchr(line,'\r')) != NULL)
		*cp = '\0';
	if((cp = strchr(line,'\n')) != NULL)
		*cp = '\0';

	for (argc = 0;argc < NARG;argc++)
		argv[argc] = NULLCHAR;

	for (argc = 0;argc < NARG;) {
		qflag = 0;
		/* Skip leading white space */
		while (*line == ' ' || *line == '\t')
			line++;
		if (*line == '\0')
			break;
		/* Check for quoted token */
		if (*line == '"') {
			line++;	/* Suppress quote */
			qflag = 1;
		}
		argv[argc++] = line;	/* Beginning of token */
		/* Find terminating delimiter */
		if (qflag) {
			/* Find quote, it must be present */
			if ((line = strchr(line,'"')) == NULLCHAR) {
				return -1;
			}
			*line++ = '\0';
		} else {
			/* Find space or tab. If not present,
			 * then we've already found the last
			 * token.
			 */
			if ((cp = strchr(line,' ')) == NULLCHAR
			 && (cp = strchr(line,'\t')) == NULLCHAR) {
				break;
			}
			*cp++ = '\0';
			line = cp;
		}
	}
//
//	if (argc < 1) {		/* empty command line */
//		argc = 1;
//		argv[0] = "";
//	}
//
	/* Lines beginning with "#" are comments */
	if ( argc < 1 || argv[0] == NULLCHAR || argv[0][0] == '#')
		return 0;

	/* Look up command in table; prefix matches are OK */
	len = strlen(argv[0]);
	for (cmdp = cmds;cmdp->name[0] != 0;cmdp++) {
		if (strnicmp(argv[0],cmdp->name,len) == 0) {
			rslt = (*cmdp->func)(argc,argv);
			if (rslt < 0 && error) {
				printf("Command failed: %.9Fs\n",cmdp->name);
				print_err_line(argc,argv);
			}
			for (len = 0;len < argc - 1;++len)
				argv[len][strlen(argv[len])] = ' ';
			return(rslt);
		}
	}
	if (error) {
		printf("Unknown command: %s (use 'help' for list)\n",argv[0]);
		print_err_line(argc,argv);
	}
	return -1;
}

/* Call a subcommand based on the first token in an already-parsed line */
int subcmd(struct cmds *tab,int argc,char *argv[])
{
	int rslt,len;
	struct cmds *cmdp;

	/* Strip off first token and pass rest of line to subcommand */
	if (argc < 2) {
		printf("Option required: %s [ ",argv[0]);
		goto print_options;
	}
	argc--;
	argv++;
	len = strlen(argv[0]);
	for (cmdp = tab;cmdp->name[0] != 0;cmdp++) {
		if (strnicmp(argv[0],cmdp->name,len) == 0) {
			rslt = (*cmdp->func)(argc,argv);
			if (rslt < 0)
				printf("Option failed: %s\n",argv[0]);
			return(rslt);
		}
	}
	printf("Unknown option: %s [ ",argv[0]);
	print_options:
	for (cmdp = tab;cmdp->name[0] != 0;cmdp++) {
		printf ("%.9Fs ",cmdp->name);
	}
	printf("]\n");
	return -1;
}
