//
//   $Id: READKB.C 1.3 2003/02/15 00:39:42Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/* Keyboard read routine for IBM PC */
/*
Updated @ Date 5 November 1995, 2:47 by Gerrit van Niekerk
*/

#include <dos.h>
#include <stdio.h>
#include <time.h>
#include "inetdefs.h"
#include "compiler.h"
#include "api.h"
#include "ftp.h"
#include "keys.h"

#define	ZF	0x40	/*8086/8 zero flag in flags word */
#define	CF	1	/*8086/8 carry flag */
#define NOCHAR -1
#define FKEY	0	/* IBM function key */
#define EFKEY	0xe0	/* IBM extended function key */

int sec_timer;
char noxkey;
char ttymode;
char dis_stop;			/* display stopped flag */
#define DS_FREEZE	1	/* display frozen */
#define DS_ERROR	2	/* waiting for error acknowledge */

#define COMESC		-2	/* command escape */

#define CTLZ		26

#define TTY_COOKED	0
#define TTY_SECRET	1
#define	TTY_RAW		2

#define	LINESIZE	256
#define ALTF 0			// ALT_F

void (*helpfunc)();
int helpkey = 0;
int helping = 0;

#define KEYBOARD	0x16	/* Keyboard interrupt */
#define	K_READ		0	/* read key */
#define	K_SENSE		1	/* sense key */
#define	K_SHIFT		2	/* read shift status */
#define	K_XREAD		0x10	/* extended read key */
#define	K_XSENSE	0x11	/* extended sense key */
#define	K_XSHIFT	0x12	/* extended read shift status */
#define KSTAT_S	3	/* shift status */
#define	KSTAT_C	4	/* control status */
#define KSTAT_A	8	/* alt status */

char k_read = K_READ;
char k_sense = K_SENSE;

static char linebuf[LINESIZE];
static char *cp = linebuf;

int key_store;		/* single key storage area */
char abort_flag;

/* Initialise keyboard */
void kbinit()
{
	char model = peekb(0xf000,0xfffe);
	char far *idtab;
	char stat;

	if (noxkey)
		return;
	if (model == 0xfc) {	/* AT */
		_AH = 0xc0;
		geninterrupt(0x15);
		stat = _AH;
		idtab = MK_FP(_ES,_BX);
		if (stat != 0 || idtab[2] < 1)
			return;
	}
	else if (model != 0xfb)	/* not recent XT */
		return;
	k_read = K_XREAD;
	k_sense = K_XSENSE;
}

void raw()
{
	ttymode = TTY_RAW;
}

void cooked()
{
	ttymode = TTY_COOKED;
}

void secret()
{
	ttymode = TTY_SECRET;
}

//**************************************************************************
/* sense keyboard and return scancode if any, else check for other work    */
//**************************************************************************
#pragma argsused
int ureadkey(char *pshift,char help_flag)
{
	int c;
	char shift;

	_AH = k_sense;
	geninterrupt(KEYBOARD);
	if (!(_FLAGS & ZF))
	{
		_AH = k_read;
		geninterrupt(KEYBOARD);
		c = _AX;
		_AH = K_SHIFT;
		geninterrupt(KEYBOARD);
		shift = _AL;
	}
	else
		return NOCHAR;
	if (pshift)
		*pshift = shift;
	return c;
}

/****************************************************************************/
/* sense keyboard and return scancode if any 								*/
/****************************************************************************/
static int near readkey(void)
{
	int c;

	if (((c = ureadkey(0,1)) & 0xff) == EFKEY)
		c &= 0xff00;
	return c;
}

int sense_key(void)
{
	int c;

	if ((c = key_store) != 0)
		key_store = 0;
	else if ((c = readkey()) == NOCHAR)
		return NOCHAR;
	if (c & 0xff)
		c &= 0xff;
	return c;
}


/* ------------- get a keyboard character ---------------- */
int get_char()
{
	int c;

	while ((c = sense_key()) == NOCHAR || c == ALTF);
	if (c & 0xff)
		c &= 0xff;
	return c;
}

void abort_read()
{
	abort_flag = 1;
}

getline(char **buf)		/* get a cooked or secret line */
{
	int ch;
	char shown = window_shown;

	if (abort_flag) {
		abort_flag = 0;
		return -3;
	}
	switch (ch = sense_key()) {
	case ALT_X:		/* alt x */
		return COMESC;
	case F3:		/* f3 */
		if (ttymode != TTY_SECRET) {
		hide_window();
		while (*cp && *cp != '\r')
			printf("%c",*cp++);
		}
		break;
	case DEL:		/* Del */
		hide_window();
		while (cp != linebuf) {
			cp--;
			printf("\b \b");
		}
		break;
	case '\b':		/* Backspace */
		if (cp != linebuf) {
			cp--;
			hide_window();
			printf("\b \b");
		}
		break;
	case '\r':		/* CR and LF are equivalent */
	case '\n':
		*cp++ = '\r';
		*cp++ = '\n';
		hide_window();
		printf("\n");
		goto exit;
	default:		/* Other character */
		if (ch & 0xff00)	/* function code */
			return ch;
		*cp++ = ch;
		hide_window();
		printf("%c",ttymode == TTY_SECRET ? 'X' : ch);
		if(cp >= linebuf + LINESIZE){
			exit:
			*buf = linebuf;
			ch = cp - linebuf;
			cp = linebuf;
			if (shown)
				show_window();
			return ch;
		}
		break;
	}
	if (shown)
		show_window();
	return NOCHAR;
}

test_key()			/* test for freeze(alt_f) and alt_x */
{
	int c;

	if ((c = readkey()) == NOCHAR || c == ALT_X) {
		return c;
	}
	key_store = c;
	return NOCHAR;
}

