//
// HTTPPUT.C
// Copyright (c) 2003 Datalight, Inc.
// Portions copyright (c) GPvNO 2003
// All Rights Reserved
// Datalight Confidential and Proprietary
//
// Simple HTTP Client
//
// Revision History
// 08/01/2003 gvn port to Sockets API
// 07/22/2003 gvn make more robust with better error messages
// 05/20/2003 gvn first version
//
//

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <io.h>
#include <string.h>
#include <dir.h>
#include <dos.h>
#include <conio.h>
#include <ctype.h>
#include <signal.h>

#ifdef _Windows
#include <winsock2.h>
#else
#include "socket.h"
#endif

#ifdef DL
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dosver.h"
#include "fatworld.h"
#endif

// prototypes
char *GetSocketError(int iError);
SOCKET HtOpen(char *pszServer,WORD wPort,char *pszRemoteFile,long lLen);
int HtWrite(char *pcBuf,int iBuflen,SOCKET soHtp);
void FixName(char *pszName);
void Fail(char *pszMessage);
void Exit(int iCode);

#define BUFSZ 1000
char arcBuf[BUFSZ];
char arcReadBuf[BUFSZ];
BYTE bServer = 0, bVerbose = 0;
char *pcRead;

// authorization (username:password)
char szAuth[200];	// Base64 encoded

// Base64 encode according to RFC 2045 Paragraph 6.8
char Encode(WORD wIn)
{
	if (wIn < 26)
		return (char)('A' + wIn);
	if (wIn < 52)
		return (char)('a' + wIn - 26);
	if (wIn < 62)
		return (char)('0' + wIn - 52);
	if (wIn == 62)
		return '+';
	return '/';
}

char *Base64Encode(char *pszIn)
{
	char *pszOut = szAuth;
	int iLen = strlen(pszIn),i;
	WORD wIn0,wIn1,wIn2;

	for (i = 0;i < iLen;i += 3)
	{
		wIn0 = *(BYTE *)pszIn++;
		wIn1 = (WORD)(i < iLen - 1 ? *(BYTE *)pszIn++ : 0);
		wIn2 = (WORD)(i < iLen - 2 ? *(BYTE *)pszIn++ : 0);
		*pszOut++ = Encode((WORD)(wIn0 >> 2));
		*pszOut++ = Encode((WORD)(((wIn0 << 4) & 0x30) | (wIn1 >> 4)));
		if (i == iLen - 1)
		{
			*pszOut++ = '=';
			*pszOut++ = '=';
			break;
		}
		*pszOut++ = Encode((WORD)(((wIn1 << 2) & 0x3c) | (wIn2 >> 6)));
		if (i == iLen - 2)
		{
			*pszOut++ = '=';
			break;
		}
		*pszOut++ = Encode((WORD)(wIn2 & 0x3f));
	}
	*pszOut = 0;
	return szAuth;
}

#define LBLEN 200

char cBreak;

void SigBreak(int iSig)
{
	signal(iSig,SigBreak);
	cBreak = 1;
}

int main(int iArgc,char *ppszArgv[])
{
	char *pszArg, *pszArgSwitch, *pszRemoteFile, *psz;
	char *pszURL = 0, *pszLocalFile = 0, *pszServer = 0;
	int i = 0, iCnt, iRet, iLineCount = 0;
	FILE *fp;
	SOCKET soHtp;
	WORD wPort = 0;
	long lLen;
	char linebuf[LBLEN + 1],ch;
	WSADATA 	sWsaData;
	char *pszServiceName = "http";
	struct servent *psService;

#ifdef DL
	printf("Datalight Sockets HTTPPUT ");
	if (dlIsFat32World())
		printf("%s ", SHORTVERSION);
	else
		printf("%s ", SHORTVERSION622);
	printf("%s\n%s\nPortions copyright (c) GPvNO 2003\n%s\n\n",
		BUILDVERSION, COPYRIGHT, _BUILD_NUMBER);
#else
	printf("HTTPPUT version 3.05\n");
#endif
	while (iArgc > 1)
	{
		pszArg = *++ppszArgv;
		--iArgc;
		if (*pszArg == '-' || (*pszArg == '/' && pszArg[1] != '/'))
		{
			pszArgSwitch = pszArg + (pszArg[2] == '=' ? 3 : 2);
			switch (pszArg[1] | ('a' - 'A'))
			{
			case 'a':
				Base64Encode(pszArgSwitch);
				break;
			case 'i':
				printf("Authorization: Basic %s\n",szAuth);
				break;
			case 'p':
				wPort = (WORD)atoi(pszArgSwitch);
				break;
			case 's':
				pszServer = pszArgSwitch;
				bServer = 1;
				break;
			case 'v':
				bVerbose = 1;
				break;
			default:
				goto BadArg;
			}
		}
		else if (i == 0)
		{
			pszURL = pszArg;
			++i;
		}
		else if (i == 1)
		{
			pszLocalFile = pszArg;
			++i;
		}
	}
	if (!i)
	{
		BadArg:
		printf("HTTPPUT [-p=<port>] [-a=<username>:<password>] [-s=<server>] [-v] <URL> [<local file>]\n");
		exit(1);
	}
	if ((i = WSAStartup(0x0101,&sWsaData)) != 0) {
		printf("WSAStartup failed: %s\n",GetSocketError(i));
		Exit(1);
	}


#ifdef __FLAT__
	signal(SIGBREAK,SigBreak);
#endif
	signal(SIGINT,SigBreak);
	if (!bServer)
	{
		// Get server address/local file from URL
		if ((psz = strchr(pszURL,'/')) != 0)
		{
			if (psz[1] == '/')
			{
				// "//" given
				if (psz != pszURL && !pszServer && psz[-1] == ':')
				{
					psz[-1] = 0;
					if (stricmp(pszURL,"http") != 0)
					{
						printf("Unknown protocol: %s\n",pszURL);
						Exit(1);
					}
					pszURL = psz;
				}
				pszServer = psz + 2;
			}
			else if (psz != pszURL)
				pszServer = pszURL;
			else
			{
				printf("Server address cannot be determined\n");
				Exit(1);
			}
			if ((psz = strchr(pszServer,'@')) != 0)
			{
				*psz = 0;
				Base64Encode(pszServer);
				pszServer = psz + 1;
			}
		}
		else
			pszServer = pszURL;
		if ((psz = strchr(pszServer,'/')) != 0)
			*psz++ = 0;
		pszRemoteFile = psz;
	}
	else
		psz = pszURL;
	if (!pszLocalFile)
	{
		if ((pszLocalFile = strrchr(psz,'/')) != 0)
			++pszLocalFile;
		else
			pszLocalFile = psz;
	}
	if (bVerbose)
		printf("Local file: '%s'\n",pszLocalFile);
	if ((fp = fopen(pszLocalFile,"rb")) == 0)
	{
		printf("Unable to open file '%s'\n",pszLocalFile);
		Exit(4);
	}
	if ((lLen = filelength(fp->fd)) < 0)
	{
		printf("Unable to determine length of file '%s'\n",pszLocalFile);
		Exit(4);
	}
	if (!wPort) {
		psService = getservbyname(pszServiceName,"tcp");
		wPort = (WORD)(psService ? htons(psService->s_port) : 80u);
	}
	if (bVerbose)
		printf("HTTP server: %s:%d\nLocal file: %s Len: %ld\n",
								pszServer,wPort,pszLocalFile,lLen);

	if ((soHtp = HtOpen(pszServer,wPort,
						bServer ? pszURL : pszRemoteFile,lLen)) == INVALID_SOCKET) {
		Exit(3);
	}
	while (1)
	{
		if (cBreak || kbhit())
		{
			if (kbhit())
				getch();
			printf("User abort\n");
			break;
		}
		if ((iRet = recv(soHtp,pcRead = arcReadBuf,BUFSZ,0)) > 0)
		{
			while (iRet--) {
				ch = *pcRead++;
				if (ch == '\r')
					continue;
				else if (ch == '\n') {
					linebuf[iLineCount] = 0;
					if (bVerbose)
						printf("HTTP: %s\n",linebuf);
					if (strnicmp(linebuf,"HTTP/",5) == 0)
					{
						char *psz;

						for (psz = linebuf + 5,iLineCount -= 5;
												iLineCount > 0 && *psz;--iLineCount,++psz)
						{
							if (isspace(*psz))
								break;
						}
						for (;iLineCount > 0 && !isdigit(*psz);--iLineCount,++psz);
						if (atoi(psz) != 200)
						{
							printf("Error: %s\n",linebuf);
							if (fp)
								fclose(fp);
							Exit(3);
						}
					}
					iLineCount = 0;
				}
				else if (iLineCount < LBLEN) {
					linebuf[iLineCount++] = ch;
				}
			}
		}
		else if (iRet == SOCKET_ERROR && WSAGetLastError() != WSAEWOULDBLOCK)
		{
			printf("Net read failed: %s\n",GetSocketError(WSAGetLastError()));
			Exit(4);
		}
		else if (iRet == 0) {
			if (bVerbose)
				printf("Net connection closed\n");
			break;
		}
		if (fp)
		{
			if ((iCnt = fread(arcBuf,1,BUFSZ,fp)) > 0)
			{
				if (HtWrite(arcBuf,iCnt,soHtp) <= 0)
				{
					fclose(fp);
					Exit(4);
				}
			}
			else
			{
				if (bVerbose)
					printf("File read completed\n");
				shutdown(soHtp,1);
				fclose(fp);
				fp = 0;
			}
		}
	}
	WSACleanup();
	if (fp)
		fclose(fp);
	return 0;
}

SOCKET HtOpen(char *pszServer,WORD wPort,char *pszRemoteFile,long lLen)
{
	SOCKET so;
	int len;
	unsigned long ulOn = 1l;
	struct sockaddr sSockAddrLocal,sSockAddrRemote;
	struct sockaddr_in *pAddr = (struct sockaddr_in *)&sSockAddrRemote;
	struct hostent *psHost;

	so = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
	if (so == INVALID_SOCKET)
		Fail("socket");
	memset(&sSockAddrLocal,0,sizeof(sSockAddrLocal));
	memset(&sSockAddrRemote,0,sizeof(sSockAddrRemote));
	sSockAddrLocal.sa_family = AF_INET;
	sSockAddrRemote.sa_family = AF_INET;
	if (bind(so,&sSockAddrLocal,sizeof(sSockAddrLocal)) == SOCKET_ERROR)
		Fail("bind");
	if ((psHost = gethostbyname(pszServer)) == 0) {
		if ((pAddr->sin_addr.s_addr = inet_addr(pszServer)) == 0) {
			printf("Unable to resolve: %s\n",pszServer);
			Fail("inet_addr");
		}
	}
	else
		pAddr->sin_addr.s_addr = *(u_long *)(psHost->h_addr_list[0]);
	pAddr->sin_port = htons(wPort);
	if (connect(so,&sSockAddrRemote,sizeof(sSockAddrRemote)) == SOCKET_ERROR)
		Fail("connect");
	if (ioctlsocket(so,FIONBIO,&ulOn) == SOCKET_ERROR)
		Fail("ioctlsocket(FIONBIO)");

	len = sprintf(arcBuf,"POST /%s HTTP/1.0\r\nContent-Length: %ld",
						pszRemoteFile + (*pszRemoteFile == '/' ? 1 : 0),lLen);
	if (*szAuth)
		len += sprintf(arcBuf + len,"\r\nAuthorization: Basic %s",szAuth);
	len += sprintf(arcBuf + len,"\r\n\r\n");
	if (bVerbose)
		printf("%s",arcBuf);
	if (HtWrite(arcBuf,len,so) <= 0)
	{
		printf("Unable to send header\n");
		closesocket(so);
		return INVALID_SOCKET;
	}
	return so;
}

int HtWrite(char *pcBuf,int buflen,SOCKET so)
{
	int iRet,iWritten = 0;

	while (iWritten < buflen) {
		iRet = send(so,pcBuf + iWritten,buflen - iWritten,0);
		if (iRet >= 0)
			iWritten += iRet;
		else if (iRet == SOCKET_ERROR && WSAGetLastError() != WSAEWOULDBLOCK)
		{
			printf("Net write failed: %s\n",GetSocketError(WSAGetLastError()));
			return 0;
		}
		if (cBreak || kbhit())
		{
			if (kbhit())
				getch();
			printf("Net write failed: User abort\n");
			return 0;
		}
	}
	return buflen;
}

void Fail(char *pszMessage)
{
	printf("Fatal error in %s: %s\n",pszMessage,GetSocketError(WSAGetLastError()));
	Exit(3);
}

void Exit(int iCode)
{
	WSACleanup();
	exit(iCode);
}

void fixname(char *pname)
{
	char *oname = pname,ch;
	int previous_is_slash = 0;

	while ((ch = *pname++) != 0) {
		if (ch == '\\' || ch == '/') {
			if (!previous_is_slash)
				*oname++ = ch;
			previous_is_slash = 1;
		}
		else {
			previous_is_slash = 0;
			*oname++ = ch;
		}
	}
	*oname = 0;
}

typedef struct sErr {
	int iErr;
	char *pszErr;
} SERROR;

extern SERROR sError[];

char *GetSocketError(int iError)
{
	int i;

	for (i = 0;sError[i].iErr != 0x7fff;++i) {
		if (iError == sError[i].iErr)
			break;
	}
	return sError[i].pszErr;
}

SERROR sError[] = {
  {WSABASEERR,         	"No Error"},
  {WSAEINTR,        		"Interrupted system call"},
  {WSAEBADF,        		"Bad file number"},
  {WSAEACCES,       		"Permission denied"},
  {WSAEFAULT,       		"Bad address"},
  {WSAEINVAL,       		"Invalid argument"},
  {WSAEMFILE,    	  		"Too many open files"},
  {WSAEWOULDBLOCK,     	"Operation would block"},
  {WSAEINPROGRESS,     	"Operation now in progress"},
  {WSAEALREADY,        	"Operation already in progress"},
  {WSAENOTSOCK,        	"Socket operation on non-socket"},
  {WSAEDESTADDRREQ,    	"Destination address required"},
  {WSAEMSGSIZE,        	"Message too long"},
  {WSAEPROTOTYPE,      	"Protocol wrong type for socket"},
  {WSAENOPROTOOPT,     	"Bad protocol option"},
  {WSAEPROTONOSUPPORT, 	"Protocol not supported"},
  {WSAESOCKTNOSUPPORT, 	"Socket type not supported"},
  {WSAEOPNOTSUPP,      	"Operation not supported on socket"},
  {WSAEPFNOSUPPORT,    	"Protocol family not supported"},
  {WSAEAFNOSUPPORT,    	"Address family not supported by protocol family"},
  {WSAEADDRINUSE,      	"Address already in use"},
  {WSAEADDRNOTAVAIL,   	"Can't assign requested address"},
  {WSAENETDOWN,        	"Network is down"},
  {WSAENETUNREACH,     	"Network is unreachable"},
  {WSAENETRESET,       	"Net dropped connection or reset"},
  {WSAECONNABORTED,    	"Software caused connection abort"},
  {WSAECONNRESET,      	"Connection reset by peer"},
  {WSAENOBUFS,         	"No buffer space available"},
  {WSAEISCONN,         	"Socket is already connected"},
  {WSAENOTCONN,        	"Socket is not connected"},
  {WSAESHUTDOWN,       	"Can't send after socket shutdown"},
  {WSAETOOMANYREFS,    	"Too many references, can't splice"},
  {WSAETIMEDOUT,       	"Connection timed out"},
  {WSAECONNREFUSED,    	"Connection refused"},
  {WSAELOOP,           	"Too many levels of symbolic links"},
  {WSAENAMETOOLONG,    	"File name too long"},
  {WSAEHOSTDOWN,       	"Host is down"},
  {WSAEHOSTUNREACH,    	"No Route to Host"},
  {WSAENOTEMPTY,       	"Directory not empty"},
  {WSAEPROCLIM,        	"Too many processes"},
  {WSAEUSERS,          	"Too many users"},
  {WSAEDQUOT,          	"Disc Quota Exceeded"},
  {WSAESTALE,          	"Stale NFS file handle"},
  {WSAEREMOTE,         	"Too many levels of remote in path"},
  {WSASYSNOTREADY,     	"Network SubSystem is unavailable"},
  {WSAVERNOTSUPPORTED, 	"WINSOCK DLL Version out of range"},
  {WSANOTINITIALISED,  	"Successful WSASTARTUP not yet performed"},
  {WSAHOST_NOT_FOUND,  	"Host not found"},
  {WSATRY_AGAIN,       	"Non-Authoritative Host not found"},
  {WSANO_RECOVERY,     	"Non-Recoverable errors: FORMERR, REFUSED, NOTIMP"},
  {WSANO_DATA,         	"Valid name, no data record of requested type"},
  {0x7fff,				  	"Unkown error"}
};


