//
//   $Id: GETMAIL.C 1.5 2002/12/31 21:52:30Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// GETMAIL.C
// Copyright (c) 2001 Datalight, Inc.
// Portions copyright (c) GPvNO 2001
// All Rights Reserved
// Datalight Confidential and Proprietary
//
// program to retrieve mail from a mail server.
//
// Revision History
// 12/31/2002 jmb corrected ReadLine to handle lines broken between
//                TCP/IP packets correctly as well as to handle
//                lines longer than LINE_SIZE.
// 06/27/2001 jmb added -T command line parameter to change the
//                timeout for the server to respond.
// 04/30/2001 jmb used ROMDOS version number and copyright info.
// 04/30/2001 jmb created header
//
//

#include <stdio.h>
#include <stdlib.h>
#include <mem.h>
#include <string.h>
#include <time.h>
#include "compiler.h"
#include "capi.h"

#ifdef DL
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dosver.h"
#include "fatworld.h"
#endif

#include "getmail.h"
#include "getmail.pt"

char szUsage[] = "GETMAIL <server> <user> <password> [-T<val>]\n\n"
                 "where\n"
                 " -T<val> -- val specifies a timeout in milliseconds\n";
unsigned long ulPopConnectTimeout = 15000l;

// #define USAGE   "GETMAIL <server> <user> <password> [-T<val>]"


char   szLogFile[] = "GETMAIL.LOG";

int main( int argc, char* argv[] )
{
   int   iSock;
   int   iMessages;
   int   i;
   char   chBuf[80];

#ifdef DL
   printf("Datalight Sockets GETMAIL ");
   if (dlIsFat32World())
      printf("%s ", SHORTVERSION);
   else
      printf("%s ", SHORTVERSION622);
   printf("%s\n%s\n%s\n\n",
      BUILDVERSION, COPYRIGHT, _BUILD_NUMBER);
#endif
   if( argc < 4 )
   {
      LogError( ERR_BAD_PARAMS, szUsage );
      return 1;
   }

   if (argc == 5)
   {
      // check cmd line parameter 4 for timeout option
      if (argv[4][0] == '-')
      {
         switch (argv[4][1])
         {
            case 't':
            case 'T':
               ulPopConnectTimeout = (unsigned long)atol(&argv[4][2]);
               break;
               
            default:
               LogError(ERR_BAD_PARAMS, szUsage);
               break;
         }
      }
   }
      
   
   // connect and authorize a user
   iSock = PopConnect( argv[1], argv[2], argv[3], ulPopConnectTimeout );
   if( iSock == -1 )
      return 1;
      
   // get the number of messages from the server
   iMessages = PopGetMessageCount( iSock );
   if( iMessages == -1 )
   {
      PopDisconnect( iSock );
      return 1;
   }
   sprintf( chBuf, "%d messages in maildrop\n", iMessages );
   LogMessage( chBuf, 0 );
   
   // count through the messages
   for( i=1; i <= iMessages; i++ )
   {
      // retrieve this message
      if( PopRetrieve( iSock, i ) == -1 )
      {
         // error case
         PopDisconnect( iSock );
         return 1;
      }
   }

   // disconnect
   PopDisconnect(iSock);
   LogMessage( "000 Retrieved POP3 mail successfully", 1 );
   return 0;
}


/*
   PopConnect

   Gets a socket, connects to the passed server, and
   authorizes with the passed user and password.  If
   anything goes wrong, it logs a message and returns
   -1.
*/
int PopConnect( char* szServer, char* szUser, char* szPassword, 
                unsigned long ulTimeout )
{
   int   iSock;
   int   iRet;
   char   chCommand[80];

   // connect to the server
   iSock = ConnectTo( szServer, 110, ulTimeout );
   if( iSock == -1 )
      return -1;

   // get the response line
   if( PopCheckResponse( iSock, SINGLE_LINE ) == FALSE )
   {
      LogError( ERR_SERVER_NOT_READY, "Server not ready" );
      goto ErrorExit;
   }

   // send user name
   sprintf( chCommand, "user %s", szUser );
   iRet = SendLine( iSock, chCommand );
   if( iRet == -1 )
   {
      LogNetError( "Sending user name" );
      goto ErrorExit;
   }
   if( PopCheckResponse( iSock, SINGLE_LINE ) == FALSE )
   {
      LogError( ERR_BAD_USER, "Invalid user name" );
      goto ErrorExit;
   }
   
   // send secret
   sprintf( chCommand, "pass %s", szPassword );
   iRet = SendLine( iSock, chCommand );
   if( iRet == -1 )
   {
      LogNetError( "Sending password" );
      goto ErrorExit;
   }
   if( PopCheckResponse( iSock, SINGLE_LINE ) == FALSE )
   {
      LogError( ERR_BAD_PASSWORD, "Incorrect password" );
      goto ErrorExit;
   }
   
   // success
   return iSock;

ErrorExit:
   PopDisconnect( iSock );
   return -1;
}


/*
   PopCheckResponse

   Reads from the socket a POP3 response code.  The code
   may be a multi-line response if iLines == MULTI_LINE,
   in which case it reads until it hits "\n\r.\n\r".  

   Returns 1 if the response starts with "+", 0 if "-",
   or -1 on internal error.
*/
int PopCheckResponse( int iSock, int iLines )
{
   char*   pszLine;
   int   fSuccess;

   // get the next line from the server
   pszLine = ReadLine( iSock );
   if( !pszLine )
      return -1;

   // set the return code
   switch( pszLine[0] )
   {
      case '-':
         fSuccess = 0;
         break;
      case '+':
         fSuccess = 1;
         break;
      default:
         LogError( ERR_BAD_DATA, "Unexpected response from server" );
         return -1;
   }

   // done?
   if( iLines == SINGLE_LINE )
      return fSuccess;

   // nope, swallow lines until we see a "." line
   while( strcmp( pszLine, "." ) != 0 )
   {
      pszLine = ReadLine( iSock );
   }

   return fSuccess;   
}


/*
   PopGetMessageCount

   Returns the number of messages in the current maildrop.
   Assumes that the passed socket is already connected to
   a POP3 server and has been authorized.
   
   Returns -1 on error.
*/
int PopGetMessageCount( int iSock )
{
   int   iRet;
   char*   pszLine;
   
   
   // send the status command
   iRet = SendLine( iSock, "stat" );
   if( iRet == -1 )
   {
      LogNetError( "Getting message count" );
      return -1;
   }
   
   // read the response line
   pszLine = ReadLine( iSock );
   if( !pszLine )
   {
      LogNetError( "Reading status response" );
      return -1;
   }

   // check response value
   switch( pszLine[0] )
   {
      // success
      case '+':
         break;

      // server problem
      case '-':
         LogError( ERR_BAD_MAILBOX, "Error retrieving mailbox status" );
         return 0;
         
      // internal error
      default:
         LogError( ERR_BAD_DATA, "Unexpected response from server" );
         return -1;
   }

   // parse the response line
   {
      char*   lpszTemp;
      
      lpszTemp = strchr( pszLine, ' ' );
      if( !lpszTemp )
      {
         LogError( ERR_BAD_MAILBOX, "Error parsing mailbox status" );
         return -1;
      }
      lpszTemp++;
      
      return atoi( lpszTemp );
   }
}


/*
   PopRetrieve
*/
int PopRetrieve( int iSock, int iMessage )
{
   int   iRet;
   char   chCommand[60];
   char*   pszLine;
   FILE*   pfMail;

   // build the command
   sprintf( chCommand, "retr %d", iMessage );

   // send it
   iRet = SendLine( iSock, chCommand );
   if( iRet == -1 )
   {
      LogNetError( chCommand );
      fclose( pfMail );
      return -1;
   }
   if( PopCheckResponse( iSock, SINGLE_LINE ) != TRUE )
   {
      LogError( ERR_BAD_MESSAGE, "Error retrieving message" );
      fclose( pfMail );
      return -1;
   }

   // message waiting at socket, read line-by-line until "." terminator
   pszLine = ReadLine( iSock );
   if (pszLine)
   {
      // get a temporary file name to store this mail
      pfMail = CreateMailFile();
      if( !pfMail )
         return -1;
   }

   while( pszLine )
   {
      // test for terminator
      if( pszLine[0] == '.' )
      {
         // end-of-message?
         if( pszLine[1] == '.' )
            pszLine++;
         else
            break;
      }
      
      // dump the line
      fwrite(pszLine, 1, strlen(pszLine), pfMail);
      fprintf(pfMail, "\n");
//      fprintf( pfMail, "%s\n", pszLine );
      
      // read the next line
      pszLine = ReadLine( iSock );
   }
   
   // close the mail file only if it had been opened...
   if (pfMail)
      fclose( pfMail );

   // check for error -- don't delete mail if we had an error reading
   if (!pszLine)
      return -1;

   sprintf( chCommand, "000 Retrieved message %d.", iMessage );
   LogMessage( chCommand, 1 );

   // build the delete command
   sprintf( chCommand, "dele %d", iMessage );

   // send it
   iRet = SendLine( iSock, chCommand );
   if( iRet == -1 )
   {
      LogNetError( chCommand );
      return -1;
   }
   if( PopCheckResponse( iSock, SINGLE_LINE ) != TRUE )
   {
      LogError( ERR_DELETING, "Error deleting retrieved message" );
      return -1;
   }

   return 1;
}


/*
   PopDisconnect

   Sends QUIT, closes socket, releases socket, returns.
*/
void PopDisconnect( int iSock )
{
   SendLine( iSock, "quit" );
   FlushSocket( iSock );
   EofSocket( iSock );
   ReleaseSocket( iSock );
   return;
}


/*
   SendLine

   Sends passed string+CRLF to socket.  Returns the WriteSocket
   return code.
*/
int SendLine( int iSock, char* lpsz )
{
   char   chBuf[80];
   
   sprintf( chBuf, "%s\r\n", lpsz );
   return WriteSocket( iSock, chBuf, strlen(chBuf), 0 );
}


/*
   ReadLine

   Reads a line of text from the passed socket.  Strips trailing
   CRLF pair, and swallows all characters past the end of the
   buffer to maintain line-alignment.

   Returns a pointer to a static buffer or NULL on socket error.
*/
char* ReadLine( int iSock )
{
   static char   chThisLine[LINE_SIZE+1];
   char         chBuf[LINE_SIZE+1];
   int         iRet;
   char*         pszTemp;

   // read-ahead to see what's available
   iRet = ReadSocket( iSock, chThisLine, LINE_SIZE, 0, NET_FLG_PEEK );
   if( iRet < 1 )
   {
      LogNetError( "Reading from server" );
      return NULL;
   }
   chThisLine[iRet] = '\0';

   // now find EOL
   pszTemp = strchr( chThisLine, '\n' );
   if( pszTemp )
   {
      // only read as much as possible
      if (strlen(chThisLine) > LINE_SIZE)
         iRet = LINE_SIZE;
      else
         iRet = strlen(chThisLine)+2;
         
      // found it!  dequeue the data from TCP and leave
      *(pszTemp-1) = '\0';
      iRet = ReadSocket( iSock, chBuf, iRet, 0, 0 );
      if( iRet < 1 )
         return NULL;
      return chThisLine;
   }


   // dequeue the line to return
   iRet = ReadSocket( iSock, chThisLine, iRet, 0, 0 );

   // line too long.  loop until we find eol
   while( !pszTemp )
   {
      // peek another chunk ahead
      iRet = ReadSocket( iSock, chBuf, LINE_SIZE, 0, NET_FLG_PEEK );
      if( iRet < 1 )
      {
         LogNetError( "Reading from server" );
         return NULL;
      }
      chBuf[iRet] = '\0';

      // look for CRLF
      pszTemp = strchr( chBuf, '\n' );
      if( pszTemp )
      {
         *(pszTemp-1) = '\0';
         iRet = strlen(chBuf);
      }

      //
      // don't attempt to concatenate if the combined string
      // length is greater than LINE_SIZE.  The unfortunate
      // side effect is that really long lines will be broken
      // into multiple pieces.
      //
      if (strlen(chThisLine) + strlen(chBuf) > LINE_SIZE)
         return chThisLine;
      // attach the rest of the line   
      strcat(chThisLine, chBuf);

      // dequeue iRet bytes
      iRet = ReadSocket( iSock, chBuf, iRet+2, 0, 0 );
      
      if( iRet < 1 )
      {
         LogNetError( "Reading from server" );
         return NULL;
      }

   }
   return chThisLine;
}

/*
   ConnectTo

   Open a TCP connection to the passed host and port.
*/
int ConnectTo( char* lpszHostName, int iPort, unsigned long ulTimeout )
{
   NET_ADDR sNetAddr;
   int iSock;

   memset(&sNetAddr, 0, sizeof(NET_ADDR));

   // setup the destination address
   sNetAddr.wRemotePort = iPort;
   sNetAddr.dwRemoteHost = ResolveName( lpszHostName, 0, 0 );
   if( sNetAddr.dwRemoteHost == 0 )
   {
      LogNetError( "Looking up host" );
      return -1;
   }
   
   // get a socket
   iSock = GetSocket();
   if( iSock < 0 )
   {
      LogNetError( "Getting client socket" );
      return -1;
   }

   // set the socket to blocking mode
   if( SetSocketOption( iSock, 0, NET_OPT_NON_BLOCKING, 0, 4 ) == -1 )
   {
      ReleaseSocket( iSock );
      LogNetError( "Setting blocking mode" );
      return -1;
   }

   // set the socket to 15 second timeout
    if( SetSocketOption( iSock, 0, NET_OPT_TIMEOUT, ulTimeout, 4 ) == -1 )
//   if( SetSocketOption( iSock, 0, NET_OPT_TIMEOUT, 15000, 4 ) == -1 )
   {
      ReleaseSocket( iSock );
      LogNetError( "Setting socket timeout" );
      return -1;
   }

   // connect the socket to the server
   if( ConnectSocket(iSock, STREAM, &sNetAddr) < 0 )
   {
      ReleaseSocket( iSock );
      LogNetError( "Connecting to server" );
      return -1;
   }

   return iSock;
}


/*
   CreateMailFile

   Creates and opens a uniquely numbered file following the
   pattern "MAIL####.TXT".

   Returns a file handle to the opened file on success, NULL
   otherwise.
*/
FILE* CreateMailFile( void )
{
   char         chFileName[13];
   FILE*         pfMail = NULL;
   static int   iNumber=0;
   
   while( iNumber < 9999 )
   {
      // make up a file name
      sprintf( chFileName, "MAIL%04d.TXT", iNumber );
      
      // check to see if it exists
      pfMail = fopen( chFileName, "r" );
      if( pfMail )
      {
         // file exists, leave it alone
         fclose( pfMail );
      }
      else
      {
         // no file.  create one
         pfMail = fopen( chFileName, "wt" );
         if( !pfMail )
            LogError( ERR_NO_TEMP_FILE, "Error opening mail storage file" );
         return pfMail;
      }
      
      iNumber++;
   }

   return NULL;
}

/*
   LogNetError

   Transforms an integer error number into a log message and
   writes it to the log file.
*/
void LogNetError( char *pszWhence )
{
   char*   pszError;
   int   iError = iNetErrNo;
   static char rgcUnk[30];
   static char *rgszErrs[] =
   {
      "No error",
      "Connection already in use",
      "DOS error",
      "Out of memory",
      "Connection does not exist",
      "Protocol or Mode not supported",
      "Bad Packet",
      "Invalid or missing host address",
      "Can't open",
      "Network unreachable",
      "Host unreachable",
      "Protocol unreachable on host",
      "Port unreachable",
      "Timeout",
      "Host name lookup failed",
      "No servers",
      "Server error",
      "Bad parameter format",
      "Bad parameter",
      "Connection closed by peer",
      "Connection reset by peer",
      "Operation would block",
      "Socket not bound",
      "Socket not available",
      "Bad parameter in system call",
      "Can't broadcast",
      "Connection has not been established",
      "Kernel Re-entered",
   };

   if( iError == ERR_API_NOT_LOADED )
   {
      pszError = "Sockets API not loaded";
   }
   else
   {
      if( iError > ERR_RE_ENTRY )
      {
         sprintf( rgcUnk, "Unknown error 0x%04X", iError );
         pszError = rgcUnk;
      }
      else
      {
         pszError = rgszErrs[iError];
      }
   }

   // format and log the error
   {
      char   szError[255];
      
      sprintf( szError, "%s (%s)", pszError, pszWhence );
      LogError( 200+iError, szError );
   }
}


/*
   LogError

   Transforms an error number and message into a text line
   for the log, and outputs it.
*/
void LogError( int iError, char *lpszError )
{
   static char   chBuf[257];

   sprintf( chBuf, "%03d %s", iError, lpszError );
   LogMessage( chBuf, TRUE );
}


/*
   LogMessage

   Outputs a message to the screen in DEBUG mode, and   to the 
   log file if the passed flag is TRUE.  Will not log 
   anything in a non-DEBUG case with the flag set to FALSE.
*/
void LogMessage( char* lpszMessage, int fFile )
{
   time_t   lTime;
   char*      pszTime = ctime(&lTime);
   char      szTime[25];
   
   // get time stamp, trim CRLF
   lTime = time(0);
   pszTime = ctime(&lTime);
   strncpy( szTime, pszTime, 24 );
   szTime[24] = '\0';

   if( fFile )
   {
      FILE*   pfLog;

      pfLog = fopen( szLogFile, "a+" );
      if( pfLog )
      {
         fprintf( pfLog, "%s, %s\n", szTime, lpszMessage );
         fclose( pfLog );
      }
      else
      {
         printf( "FAILED TO OPEN LOG!\n" );
         exit( 1 );
      }
   }

   #if DEBUG > 0
      printf( "%s, %s\n", szTime, lpszMessage );
   #endif
}
