//
//   $Id: SNTPCLI.C 1.4 2002/10/16 16:34:40Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//

/******************************************************************************
   SNTPCLI.C

   Copyright (C) 2001 Datalight, Inc.
   All Rights Reserved
   Datalight Confidential and Proprietary

   set dos clock from internet using SNTP

   Revision History
   11/01/2001 jmb corrected time calculation and added TZ check
   07/12/2001 jmb original revision

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <time.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dosver.h"
#include "fatworld.h"

#include "compiler.h"
#include "capi.h"
#include "sntpcli.h"
#include "sntpcli.pt"


//
// prototype for routine to search for sockets interrupt vector
//
int FindSocketsVector(void);

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif





main(int argc, char *argv[])
{
   DWORD host;
   long newtime;
   long addminutes;
   struct date dstruct;
   struct time tstruct;

   printf("Datalight Sockets SNTP Client ");
   if (dlIsFat32World())
      printf("%s ", SHORTVERSION);
   else
      printf("%s ", SHORTVERSION622);
   printf("%s\n%s\nAll Rights Reserved\n%s\n\n",
      BUILDVERSION, COPYRIGHT, _BUILD_NUMBER);

   if (argc < 2) 
   {
      printf("SNTPCLI server [minutes]\n\n");
      printf("server      the server providing the time service.\n");
      printf("[minutes]   an optional parameter which provides an\n");
      printf("            adjustment to the server time in minutes.\n\n");

      exit(3);
   }

   if (FindSocketsVector() == 0)
   {
      printf("Sockets API not loaded\n");
      exit(1);
   }

   if (!getenv("TZ"))
      printf("Warning: TZ is not set -- assuming EST5EDT\n\n");
 
   if (argc == 3)
      addminutes = atol( argv[2] ) * 60L;
   else
      addminutes = 0L;
                               
   host = ResolveName(argv[1], 0, 0);
   if (host == 0)
   {
      printf("Error Looking up host!\n");
      exit(2);
   }

   printf("Getting time from %s (%u.%u.%u.%u)\n", argv[1],
              ((BYTE *)&host)[0],
              ((BYTE *)&host)[1],
              ((BYTE *)&host)[2],
              ((BYTE *)&host)[3]);

   newtime = sntptime( host, addminutes );
   if (newtime)
   {
      printf("Time set to %s", ctime( (time_t far *)&newtime));
      unixtodos( newtime, &dstruct, &tstruct );
      settime( &tstruct );
      setdate( &dstruct );
      exit(0);
   }
   printf("Unable to get the time from that host\n");
   return 1;
}


//
// sntpime() given the host address, returns an Internet based time
//
long sntptime(DWORD addr, long addminutes)
{
   NET_ADDR sNetAddr;
   int s;
   unsigned long temptime = 0l;
   SNTP_REQUEST req;
   long T1, T2, T3, T4, NetLag;

   
   if (addr == 0l)
      return 0;

   temptime = 0L;

   // setup the time server port and address
   sNetAddr.wRemotePort = TIME_PORT;
   sNetAddr.dwRemoteHost = addr;

   // get a socket
   s = GetSocket();
   if (s < 0)
   {   
      printf("unable to allocate client socket!\n");
      return 0;
   }

   // set socket options
   SetSocketOption(s, 0, NET_OPT_TIMEOUT, 10000l, 4);
   
   // connect socket to the server
   if (ConnectSocket(s, DATA_GRAM, &sNetAddr) < 0)
   {
      ReleaseSocket(s);
      printf("Sorry, unable to connect to that machine right now!\n");
      return 0;
   }

   T1 = Now;
   memset(&req, 0,sizeof(req));
   req.LI = 0;
   req.VN = 3;
   req.Mode = 3;
   UnixToSNTP(&req.TransmitTimestamp, Now);
   if (WriteToSocket(s, (char *)&req, sizeof(req), &sNetAddr, NET_FLG_NON_BLOCKING) < 0)
   {
      printf("unable to send request -- aborting!\n");
      goto sock_err;
   }


   if (ReadFromSocket(s, (char *)&req, sizeof(req), &sNetAddr, 0) == sizeof(req))
   {

      T2 = ntohl(req.ReceiveTimestamp.Seconds) - BASE_TIME;
      T3 = ntohl(req.TransmitTimestamp.Seconds) - BASE_TIME;
      T4 = Now;

      NetLag = (T4 - T1) - (T2 - T3);
      ReleaseSocket(s);
      return T3 + NetLag/2 + addminutes;
   }

sock_err:
   switch (iNetErrNo) 
   {
      case NO_ERR:
         return temptime;
      case ERR_EOF : /* foreign host closed */
      case ERR_RESET:
         printf(" Connection closed!\n");
         break;

      case ERR_WOULD_BLOCK:
         printf(" Connection would block!  Aborting!\n");
         break;

      case ERR_TIMEOUT: /* timeout */
         printf(" Connection timed out!\n");
         break;

      default: 
         printf(" Aborting (%d)\n", iNetErrNo);
   }
   ReleaseSocket(s);
   return 0;
}

void Delay(WORD ms)
{
   asm   mov   ax, 1000
   asm   mul   ms
   asm   mov   cx, dx
   asm   mov   dx, ax
   asm   mov   ah, 86h
   asm   int   15h
}


long SNTPToUnix(SNTP_TIMESTAMP tstamp)
{
   return ntohl(tstamp.Seconds);
}

void UnixToSNTP(SNTP_TIMESTAMP *tstamp, long dt)
{
   tstamp->Seconds = htonl(dt);
   tstamp->SubSeconds = 0;
}


