//
//   $Id: ARP.H 1.4 2002/11/26 21:16:12Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/* ARP.H
 * Header file for ARP
 * Revision history:

2002-09-16 gvn DHCP ARP followed by gratuitous ARP

 */

/* Size of ARP hash table */
#define	ARPSIZE	17
/* Maximum ARP queue per target */
#define	MAX_ARP_Q 10
/* Lifetime of a valid ARP entry */
#define	ARPLIFE		900	/* 15 minutes */
/* Lifetime of a pending ARP entry */
#define	PENDTIME	15	/* 15 seconds */

/* ARP definitions (see RFC 826) */

/* Address size definitions */
#define	IPALEN	4		/* Length in bytes of an IP address */
#define	MAXHWALEN	16	/* Maximum length of a hardware address */

/* ARP opcodes */
#define	ARP_REQUEST	1
#define	ARP_REPLY	2

/* Hardware types */
#define	ARP_ETHER	1	/* Assigned to 10 megabit Ethernet */
#define	ARP_EETHER	2	/* Assigned to experimental Ethernet */
#define	ARP_AX25		3	/* Assigned to AX.25 Level 2 */
#define	ARP_PRONET	4	/* Assigned to PROnet token ring */
#define	ARP_CHAOS	5	/* Assigned to Chaosnet */
#define	ARP_IEEE		6	/* RFC 1042 */
#define	ARP_ARCNET	7
#define	ARP_APPLETALK	8
#define	ARP_IPX		9
#define	ARP_SERIAL	20	// used for hardware type, not ARP

//
// ARP types
//
#define ARP_TYPE_DHCP 1    // DHCP ARP
#define ARP_TYPE_GRAT 2    // Gratuitous ARP
#define ARP_TYPE_BOTH (ARP_TYPE_DHCP | ARP_TYPE_GRAT)


extern char *arptypes[];	/* Type fields in ASCII, defined in arpcmd */
#define	NHWTYPES 10

/* Table of hardware types known to ARP */
struct arp_type {
	WORD hwalen;		/* Hardware length */
	WORD iptype;		/* Hardware type field for IP */
	WORD arptype;		/* Hardware type field for ARP */
	BYTE *bdcst;		/* Hardware broadcast address */
	void (*format)(char *,BYTE *);	/* Function that formats addresses */
	void (*scan)(BYTE *,char **,int);		/* Reverse of format */
};
extern struct arp_type arp_type[];
#define	NULLATYPE	(struct arp_type *)0

/* Format of an ARP request or reply packet. From p. 3 */
struct arp {
	WORD hardware;			/* Hardware type */
	WORD protocol;			/* Protocol type */
	BYTE hwalen;			/* Hardware address length, bytes */
	BYTE pralen;			/* Length of protocol address */
	WORD opcode;			/* ARP opcode (request/reply) */
	BYTE shwaddr[MAXHWALEN];	/* Sender hardware address field */
	DWORD sprotaddr;		/* Sender Protocol address field */
	BYTE thwaddr[MAXHWALEN];	/* Target hardware address field */
	DWORD tprotaddr;		/* Target protocol address field */
};

/* Format of ARP table */
struct arp_tab {
	struct arp_tab *next;	/* Doubly-linked list pointers */
	struct arp_tab *prev;
	DWORD ip_addr;		/* IP Address, host order */
	WORD hardware;		/* Hardware type */
	BYTE *hw_addr;		/* Hardware address */
//	char *rt_addr;		/* Router address (used for IPX) */
	BYTE state;		    /* (In)complete */
#define	ARP_PENDING	0
#define	ARP_VALID	1
	struct timer timer;	/* Time until aging this entry */
	struct mbuf *pending;/* Queue of datagrams awaiting resolution */
	BYTE *rif;			// pointer to routing information
};
struct arp_tab *arp_lookup(),*arp_add();
#define	NULLARP	(struct arp_tab *)0
extern struct arp_tab *arp_tab[];

struct arp_stat {
	unsigned recv;		/* Total number of ARP packets received */
	unsigned badtype;	/* Incoming requests for unsupported hardware */
	unsigned badlen;	/* Incoming length field(s) didn't match types */
	unsigned badaddr;	/* Bogus incoming addresses */
	unsigned inreq;		/* Incoming requests for us */
	unsigned replies;	/* Replies sent */
	unsigned outreq;	/* Outoging requests sent */
};

/* prototypes */
arp_init(
struct interface *ifp,	/* interface pointer */
WORD hwtype,	/* ARP Hardware type */
int hwalen,		/* Hardware address length */
int iptype,		/* Subnet's protocol ID for IP */
WORD arptype,		/* Subnet's protocol ID for ARP */
BYTE *bdcst,		/* Subnet's broadcast address (if any) */
void (*format)(),	/* Function to format hardware addresses */
void (*scan)(BYTE *out,char *in[],int cnt));		/* Function to scan addresses in ascii */
void arp_input(struct interface *ifp,struct mbuf *bp,BYTE *rif);
BYTE *res_arp(struct interface *interface,DWORD target,
						struct mbuf *bp,BYTE **rif);
struct arp_tab *arp_add(
DWORD ipaddr,		/* IP address, host order */
WORD hardware,		/* Hardware type */
BYTE *hw_addr,		/* Hardware address, if known; NULLCHAR otherwise */
WORD hw_alen,		/* Length of hardware address */
BYTE *rif);			// rif for tokenring
void arp_drop(struct arp_tab *ap);
struct arp_tab *arp_lookup(WORD hardware,DWORD ipaddr);
int ntoharp(struct arp *arp,struct mbuf **bpp);
void arp_output(struct interface *ifp,WORD hardware,DWORD source,DWORD target);

