//
//   $Id: BOOTP.H 1.5 2003/08/21 21:36:22Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// BOOTP.H
//
// Copyright (c) 2001 Datalight, Inc.
// All Rights Reserved
// Datalight Confidential and Proprietary
// Portions Copyright(c) 2001 GPvNO
//
//
// Revision History
// 08/21/2003 jmb get domain suffix from DHCP/BOOTP server
// 07/23/2001 gvn fixed BOOTP_SIZE
// Updated @ Date 20 April 1995, 19:13 by Gerrit van Niekerk

/*************************************************/
/* Center for Information Technology Integration */
/*	     The University of Michigan 	 */
/*		      Ann Arbor 		 */
/*						 */
/* Dedicated to the public domain.		 */
/* Send questions to info@citi.umich.edu	  */
/*						 */
/* BOOTP is documented in RFC 951 and RFC 1048	 */
/*************************************************/

#ifndef BOOTREQUEST

struct bootp {
	BYTE	op;				/* packet opcode type */
	BYTE	htype;			/* hardware addr type */
	BYTE	hlen;				/* hardware addr length */
	BYTE	hops;				/* gateway hops */
	DWORD	xid;				/* transaction ID */
	unsigned secs;			/* seconds since boot began */
	unsigned flags;
	DWORD	ciaddr;			/* client IP address */
	DWORD	yiaddr;			/* 'your' IP address */
	DWORD	siaddr;			/* server IP address */
	DWORD	giaddr;			/* gateway IP address */
	BYTE	chaddr[16];		/* client hardware address */
	BYTE	sname[64];		/* server host name */
	BYTE	file[128];		/* boot file name */
	BYTE	vend[64];		/* vendor-specific area */
};

//#define BOOTP_SIZE (sizeof(struct bootp))
#define BOOTP_SIZE 548	// 576

/*
 * UDP port numbers, server and client.
 */
#define	IPPORT_BOOTPS	67
#define	IPPORT_BOOTPC	68

#define BOOTREQUEST		1
#define BOOTREPLY		2

#define BOOTP_PAD		0
#define BOOTP_SUBNET	1
#define BOOTP_GATEWAY	3
#define BOOTP_DNS		6
#define BOOTP_DOMAIN_SUFFIX   15
#define BOOTP_HOSTNAME	12
#define DHCP_REQIPA		50	// 9.1. Requested IP Address
#define DHCP_LEASETIME	51	// 9.2. IP Address Lease Time
#define DHCP_OPTOVER	52	// 9.3. Option Overload
#define OVERLOAD_FILE	1	// the 'file' field is used to hold options
#define OVERLOAD_SNAME	2	// the 'sname' field is used to hold options
#define OVERLOAD_BOTH	3	// both fields are used to hold options
#define DHCP_SNAME		66	// 9.4 TFTP server name
#define DHCP_FILE		67	// 9.5 Bootfile name
#define DHCP_MESTYPE	53	// 9.6. DHCP Message Type
#define DHCPDISCOVER	1
#define DHCPOFFER       2
#define DHCPREQUEST     3
#define DHCPDECLINE     4
#define DHCPACK         5
#define DHCPNAK         6
#define DHCPRELEASE     7
#define DHCPINFORM      8
#define DHCP_SERVERID	54	// 9.7. Server Identifier
#define DHCP_PARRLIST	55	// 9.8. Parameter Request List
#define DHCP_MESSAGE	56	// 9.9. Message
#define DHCP_MESSIZE	57	// 9.10. Maximum DHCP Message Size
#define DHCP_RENEWTIME	58	// 9.11. Renewal (T1) Time Value
#define DHCP_REBINDTIME	59	// 9.12. Rebinding (T2) Time Value
#define DHCP_VENDCLSID	60	// 9.13. Vendor class identifier
#define DHCP_CLIENTID	61	// 9.14. Client-identifier
#define DHCP_WWW		72	// 8.17. Default World Wide Web (WWW) Server Option
#define BOOTP_END		0xff

/*
 * "vendor" data permitted for Stanford boot clients.
 */
struct vend {
	unsigned char  v_magic[4];     /* magic number */
	unsigned long  v_flags;        /* flags/opcodes, etc. */
	unsigned char  v_unused[56];   /* currently unused */
};

#define VM_STANFORD	"STAN"	/* v_magic for Stanford */

/* v_flags values */
#define VF_PCBOOT	1	/* an IBMPC or Mac wants environment info */
#define VF_HELP 	2	/* help me, I'm not registered */

extern int WantBootp;
extern char bp_ascii[];

void bootp_print_packet(struct bootp *bp);
int bootp_validPacket(struct ip *ip,struct mbuf **bpp);
int bootp(IFACE *ifp);
void dhcp_decline(IFACE *ifp);
void dhcp_inform(IFACE *ifp);

#endif
