//
//   $Id: ICMP.H 1.2 2002/10/16 16:40:06Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/* ICMP.H
Internet Control Message Protocol
Updated @ Date 21 July 1995, 18:48 by Gerrit van Niekerk
*/

/* Message types */
#define	ECHO_REPLY	0	/* Echo Reply */
#define	DEST_UNREACH	3	/* Destination Unreachable */
#define	QUENCH		4	/* Source Quench */
#define	REDIRECT	5	/* Redirect */
#define	ECHO		8	/* Echo Request */
#define	TIME_EXCEED	11	/* Time-to-live Exceeded */
#define	PARAM_PROB	12	/* Parameter Problem */
#define	TIMESTAMP	13	/* Timestamp */
#define	TIME_REPLY	14	/* Timestamp Reply */
#define	INFO_RQST	15	/* Information Request */
#define	INFO_REPLY	16	/* Information Reply */

/* Internal format of an ICMP header (checksum is missing) */
struct icmp {
	char type;
	char code;
 	union icmp_args {
		int32 unused;
		unsigned char pointer;
		DWORD address;
		struct {
			WORD id;
			WORD seq;
		} echo;
	} args;
};
#define	ICMPLEN		8	/* Length of ICMP header on the net */
#define	NULLICMP	(union icmp_args *)0

/* Destination Unreachable codes */
#define	NET_UNREACH	0	/* Net unreachable */
#define	HOST_UNREACH	1	/* Host unreachable */
#define	PROT_UNREACH	2	/* Protocol unreachable */
#define	PORT_UNREACH	3	/* Port unreachable */
#define	FRAG_NEEDED	4	/* Fragmentation needed and DF set */
#define	ROUTE_FAIL	5	/* Source route failed */

/* Time Exceeded codes */
#define	TTL_EXCEED	0	/* Time-to-live exceeded */
#define	FRAG_EXCEED	1	/* Fragment reassembly time exceeded */

/* Redirect message codes */
#define	REDR_NET	0	/* Redirect for the network */
#define	REDR_HOST	1	/* Redirect for the host */
#define	REDR_TOS	2	/* Redirect for Type of Service, or-ed with prev */

struct icmp_errors {
	unsigned checksum;		/* ICMP Checksum errors */
	unsigned nospace;		/* alloc_mbuf failed someplace */
	unsigned noloop;		/* No ICMP in response to an ICMP */
	unsigned bdcsts;		/* Ignore broadcast ICMPs */
};
#define	ICMP_TYPES	17
struct icmp_stats {
	unsigned input[ICMP_TYPES];	/* ICMP input stats by type */
	unsigned output[ICMP_TYPES];	/* ICMP output stats by type */
};

void icmp_dump();

/* ICMP messages, decoded */
extern char *icmptypes[],*unreach[],*exceed[],*redirect[];

// prototypes

struct mbuf *htonicmp(struct icmp *icmp,struct mbuf *data);
void ntohicmp(struct icmp *icmp,struct mbuf **bpp);
void icmp_output(struct ip *ip,struct mbuf *bp,char type,char code,union icmp_args *args);
void icmp_input(IFACE *ifp,struct mbuf *bp,char protocol,DWORD source,DWORD dest,char tos,WORD length,char rxbroadcast);
void tcp_icmp(DWORD source,DWORD dest,char type,char code,struct mbuf **bpp);
void ping_log(DWORD source,MBUF *bp);
void pingem_solicited(DWORD dest);		// send ping solicited by ID 1234
void echo_proc(IFACE *ifp,DWORD source,DWORD dest,struct icmp *icmp);
