//
//   $Id: IP.H 1.2 2002/10/16 16:40:06Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/* IP.H
Updated @ Date 16 April 1996, 10:35 by Gerrit van Niekerk
*/

#ifndef __IP_H
#define __IP_H
#ifndef __MBUF_H
#include "mbuf.h"
#endif
#ifndef __TIMER_H
#include "timer.h"
#endif
#ifndef __IFACE_H
#include "iface.h"
#endif

//#define IPDEBUG 1	/* put in for code to drop occational dg */

#define	NROUTE	5	/* Number of hash chains in routing table */

// default ip address and netbits
extern DWORD def_ip_addr;	/* Our IP address for ICMP and source routing */
extern BYTE def_netbits;	// number of bits in our net mask

extern BYTE ip_ttl;	    	/* Default time-to-live for IP datagrams */
extern BYTE proxy_arp;  	// some proxy ARP routes declared

#define	IPVERSION	4
/* IP header, INTERNAL representation */
#define IPLEN		20		/* Length of standard IP header */
#define IP_MAXOPT	40		/* Largest option field, bytes */
struct ip {
	BYTE version;			/* IP version number */
	BYTE tos;		    	/* Type of service */
	WORD length;			/* Total length */
	WORD id;			    	/* Identification */
	WORD fl_offs;			/* Flags + fragment offset */

#define	F_OFFSET	0x1fff	/* Offset field */
#define	DF	0x4000		/* Don't fragment flag */
#define	MF	0x2000		/* More Fragments flag */

	WORD checksum;		/* Header checksum */
	BYTE ttl;		    	/* Time to live */
	BYTE protocol;			/* Protocol */
	DWORD source;			/* Source address */
	DWORD dest;		    	/* Destination address */
	BYTE optlen;			/* Length of options field, bytes */
	BYTE options[IP_MAXOPT];/* Options field */
#ifdef TGW
	DWORD time;				// timestamp
	WORD ip_len;			/* IP header length */
	WORD lengthd;			/* Length of data portion */
	DWORD gateway;			/* Gateway IP address */
	IFACE *iface;			/* Output interface, possibly forwarded */
	MBUF *bp;
	struct ip *next;
#endif
};

#define	NULLIP	(struct ip *)0

/* Fields in option type byte */
#define	OPT_COPIED	0x80	/* Copied-on-fragmentation flag */
#define	OPT_CLASS	0x60	/* Option class */
#define	OPT_NUMBER	0x1f	/* Option number */

/* IP option numbers */
#define	IP_EOL		0	/* End of options list */
#define	IP_NOOP		1	/* No Operation */
#define	IP_SECURITY	2	/* Security parameters */
#define	IP_LSROUTE	3	/* Loose Source Routing */
#define	IP_TIMESTAMP 4	/* Internet Timestamp */
#define	IP_RROUTE	7	/* Record Route */
#define	IP_STREAMID	8	/* Stream ID */
#define	IP_SSROUTE	9	/* Strict Source Routing */

/* Timestamp option flags */
#define	TS_ONLY		0	/* Time stamps only */
#define	TS_ADDRESS	1	/* Addresses + Time stamps */
#define	TS_PRESPEC	3	/* Prespecified addresses only */

/* IP routing table entry */
typedef struct route {
	struct route *prev;	/* Linked list pointers */
	struct route *next;
	DWORD target;		/* Target IP address */
	DWORD gateway;		/* IP address of local gateway for this target */
	int metric;		    /* Hop count, whatever */
	struct interface *interface;/* Device interface structure */
//	struct interface *altif;/* Alternate interface */
//	char priority;		/* priority of alt if (0=same,1=less) */
    TIMER timer;        // route timer
	char stat;          // route status
#define RT_PROXY      1 // proxy ARP for this route
#define RT_PRIVATE    2 // route not to be advertised
#define RT_TRIG       4 // trigger pending for route
#define RT_STATIC     8 // route not to be dropped by RIP
    char bits;          // number of significant bits in target
} ROUTE;
#define	NULLROUTE	(struct route *)0
extern ROUTE *routes[33][NROUTE];	/* Routing table */
//extern struct route r_default;	/* Default route entry */

/* Cache for the last-used routing entry, speeds up the common case where
 * we handle a burst of packets to the same destination
 */
struct rt_cache {
	DWORD target;
	struct route *route;
};
extern struct rt_cache rt_cache;

/* Reassembly descriptor */
struct reasm {
	struct reasm *next;	/* Linked list pointers */
	struct reasm *prev;
	DWORD source;			/* These four fields uniquely describe a datagram */
	DWORD dest;
	WORD id;
	char protocol;
	WORD length;			/* Entire datagram length, if known */
	struct timer timer;	/* Reassembly timeout timer */
	struct frag *fraglist;	/* Head of data fragment chain */
};
#define	NULLREASM	(struct reasm *)0

/* Fragment descriptor in a reassembly list */
struct frag {
	struct frag *prev;	/* Previous fragment on list */
	struct frag *next;	/* Next fragment */
	struct mbuf *buf;		/* Actual fragment data */
	WORD offset;			/* Starting offset of fragment */
	WORD last;		    	/* Ending offset of fragment */
};
#define	NULLFRAG	(struct frag *)0

extern struct reasm *reasmq;	/* The list of reassembly descriptors */

/* IP error logging counters */
#define IP_STAT_DEF
struct ip_stats {
	long total;		    	/* Total packets received */
	unsigned runt;			/* Smaller than minimum size */
	unsigned length;		/* IP header length field too small */
	unsigned version;		/* Wrong IP version */
	unsigned checksum;	/* IP header checksum errors */
	unsigned badproto;	/* Unsupported protocol */
	unsigned drop_ratio;	/* Number of datagrams to accept before drop */
	unsigned drop_count;	/* Count since last drop */
	char drop_type;
	ROUTE *((*proutes)[33][NROUTE]);	// pointer to route table
#ifdef	TGW
	unsigned drop_rand;	// random drop number
	unsigned delay;		// packet delay
	unsigned delay_rand;	// random packet delay
	DWORD	thruput;			// maximum throughput
#endif
};
extern struct ip_stats ip_stats;

void ip_route(MBUF *bp,char rxbroadcast,IFACE *ifp);
int rt_add(DWORD target,unsigned int bits,DWORD gateway,
	 int metric,struct interface *interface,DWORD time,char stat,char top);
int rt_drop(DWORD target,unsigned int bits);
ROUTE *rt_lookup(DWORD target,DWORD source,int flags);
// rt_lookup flags
#define USE_DOWN    0   // use interfaces that are down
#define SKIP_DOWN   1   // skip interfaces that are down

void ip_recv(IFACE *ifp,struct ip *ip,MBUF *bp,char rxbroadcast);
ROUTE *rt_blookup(DWORD target,unsigned int bits);
DWORD this_addr(DWORD dest,IFACE **ifp);    // determine source IP address to use
int is_this_addr(DWORD dest);	// check for our host adress(es)
void ip_send(DWORD source,DWORD dest,char protocol,char tos,char ttl,
			struct mbuf *bp,WORD length,WORD id,char df,IFACE *ifp);
MBUF * htonip(struct ip *ip,MBUF *data,int cs_flag);
ntohip(struct ip *ip,MBUF **bpp);
void send_ip(struct ip *pip,IFACE *iface,MBUF *bp,DWORD gateway,
		WORD ip_len,WORD length);

#define HOSTNAMELEN 32
extern BYTE hostname[HOSTNAMELEN];
extern BYTE proxy_arp;
DWORD resolve_nl(char *name,WORD *plen,char **phost);
set_ip_addr(char *name,int netlen);
WORD ip_mtu(DWORD addr);
WORD lcsum(WORD *,int);
void rt_flush_cache(void);

#endif
