//  $Id: compiler.h 1.8 2003/07/21 17:18:42Z johnb Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 2000-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/*
	$DOCFILE:COMPILER.H

	Copyright (C) 2000-2001 Datalight, Inc.
	All Rights Reserved

	Compiler specific header file for the
	Sockets Compatible API (CAPI)

	$DOCHISTORY:
		  07/21/2003 gvn change Borland CLK_TCK floating point definition
		  07/19/2002 jmb added CARRY_FLAG comparison macro
		  10/03/2001 jmb added macros to support DJGPP
	     05/11/2001 jmb updated ISR_ROUTINE
	     03/29/2001 jmb removed MSVC compatibility changaes -- caused too
	                    many problems.
	     03/20/2001 jmb Fixed REALMODE_SEGMENT, REALMODE_OFFSET and
		                 MAKE_REAL POINTERS macros to work under MSVC.
							  I also added a #define for interrupt to be
							  compatible with MSVC.
	1.00 2000-11-29 pkg Extracted from CAPI.H
*/

/*
	This define must be set to produce a 32 bit (seg:off pointer)
*/
#if defined(__DJGPP__) || defined(__WIN32__)
#define D_FAR
#else
#define D_FAR far
#endif

#if defined(__DJGPP__) || defined(__WIN32__)
#define LOADDS
#else
#define LOADDS _loadds
#endif

#ifdef __DJGPP
#define FAR
#endif

/*
	this macro is the if statement for comparing against the carry flag
*/
#ifdef __DJGPP__
#define CARRY_FLAG_CHECK ((r.x.flags & 0x0001) > 0)
#else
#define CARRY_FLAG_CHECK (r.x.cflag)
#endif
/*
	Function declaration prefix, if any, for an ISR routine.
*/
#if defined(__DJGPP__) || defined(__WIN32__)
#define ISR_ROUTINE
#else
#define ISR_ROUTINE __loadds
#endif

/*
	Portable types which must be set:
	BYTE (8bits) WORD (16bits) DWORD (32bits)
*/
#ifndef __WIN_TYPES
 #define __WIN_TYPES
 typedef unsigned char BYTE;
 typedef unsigned short WORD;
 typedef unsigned long DWORD;
#endif

#ifdef __DJGPP__
/*
   The following macro allows a structure to be created without 
   inserting gaps to align for faster access
*/
#define __ATTR__ __attribute__((packed))
#else
#define __ATTR__
#endif

/*
	The following macros are used to work with both native
	and real mode pointers and buffers.  Each is generally
	unique to a given compiler and environment.  For
	example, an application using the 32-bit DJGPP compiler
	will require a specially allocated buffer and a
	means to move data between both native and real mode
	buffers.

	Each macro must be carefully written for a particular
	environment.
*/
#ifndef __DJGPP__
#define REALMODE_SEGMENT(p)     (WORD)((DWORD)(p) >> 16)
#define REALMODE_OFFSET(p)      (WORD)((DWORD)(p) & 0xFFFFL)
#define MAKE_REAL_POINTER(s, o) (DWORD)(((DWORD)(s) << 16) \
                                   + ((DWORD)(o) &0xFFFFL))
#else
#define REALMODE_SEGMENT(p)     (p).rm_segment;
#define REALMODE_OFFSET(p)      (p).rm_offset
#define MAKE_REAL_POINTER(s, o) (DWORD)(((DWORD)(s) << 4l) \
                                   + ((DWORD)(o) & 0xFFFFl))
#endif

#ifndef __DJGPP__
//
// allocate a temporary pointer for use with the REALMODE macros
#define DPMIINFO(i) void * (i)     
//
// no initialization necessary in real mode
#define DPMIINFO_INIT(i) 
//
// no allocation necessary in real mode
#define DPMIINFO_ALLOC(i, s)
//
// no reallocation necessary in real mode
#define DPMIINFO_REALLOC(i, s)
//
// assign the specified pointer to our temporary one allocated with DPMIINFO
#define DPMIINFO_SET(i, p, s) i = (p)
//
// no data to retrieve for real mode
#define DPMIINFO_GET(i, p, s)
//
// no memory to free in real mode
#define DPMIINFO_FREE(i)


// 
// define pointer for callback info
#define DPMI_CB_INFO(i) void D_FAR *(i)
//
// no callback allocation is necessary
#define DPMI_CB_ALLOCRETF(i, p, r) i = (p);
//
// no callback regs are necessary
#define DPMI_CB_REGS(r) void *r

#else

// 
// define structure to hold pm seletor as well as realmode seg and offset
// of transfer buffer
#define DPMIINFO(i)         _go32_dpmi_seginfo (i)
//
// initialize structure to 0's
#define DPMIINFO_INIT(i)  memset(&i, 0, sizeof(_go32_dpmi_seginfo))
//
// allocate pmode transfer buffer to share with real mode, if unable to
// allocate this buffer, set global error variable
#define DPMIINFO_ALLOC(i, s) \
            i.size = (((s) + 15) >> 4) + 1; \
            if (_go32_dpmi_allocate_dos_memory(&i) != 0) \
            { \
               iNetErrNo = ERR_NO_MEM; \
            } \
            else \
               iNetErrNo = 0; \
//
// resize pmode transfer buffer, if unable to resize this buffer, set
// global error variable
#define DPMIINFO_REALLOC(i, s) \
      i.size = (((s) + 15) >> 4) + 1; \
      if (_go32_dpmi_allocate_dos_memory(&i) != 0) \
      { \
         iNetErrNo = ERR_NO_MEM; \
      } \
      else \
         iNetErrNo = 0; \

//
// copy pmode data into transfer buffer
#define DPMIINFO_SET(i, p, s)  dosmemput(p, s, (i.rm_segment<<4)+i.rm_offset)
//
// copy data from transfer buffer to pmode
#define DPMIINFO_GET(i, p, s)  dosmemget((i.rm_segment<<4)+i.rm_offset, s, p)
//
// release transfer buffer
#define DPMIINFO_FREE(i)  _go32_dpmi_free_dos_memory(&i)

//
// same as DPMIINFO
#define DPMI_CB_INFO(i)         _go32_dpmi_seginfo (i)

//
// allocate a callback
#define DPMI_CB_ALLOCRETF(i, p, r) \
   if (i.pm_offset != (long)p) \
   { \
      DPMIINFO_INIT(i); \
      i.pm_offset = (long)p; \
      if (_go32_dpmi_allocate_real_mode_callback_retf(&i, &r)) \
      { \
         iNetErrNo = ERR_NO_MEM; \
      } \
      else \
         iNetErrNo = 0; \
   }


//
// declare registers for callback
#define DPMI_CB_REGS(r) __dpmi_regs r
#endif


/*
	The compatibility API is called called by setting up
	the CPU registers and memory structures and issueing
	an interrupt 0x61 (or appropriate interrupt)  The 
	following macros provide a semi-portable means of
	embedding assembly code in C code.
*/

/* Microsoft C uses _asm and cflag */
#ifdef _MSC_VER
  #ifdef _ASM
        #undef _ASM
  #endif
  #define _ASM _asm
#endif

/* Borland C uses asm and flags*/
#ifdef __TURBOC__
 #ifdef _ASM
  #undef _ASM
 #endif
 #define _ASM asm
 /* Undo the floating point definition of CLK_TCK */
 #ifdef CLK_TCK
  #undef CLK_TCK
 #endif
 #define CLK_TCK 1000
#endif

#ifdef __DJGPP__
#define _ASM __asm__
#endif

/*
	These are required for Borland and Microsoft C, other
	compilers may or may not use them.  The following
	macros must be defined appropriately.
*/
#ifndef _ASM
 #error _ASM not defined.  Required for _CAPI
#endif

/*
	MSDELAY must be defined to sleep or delay execution
	for the given number of milliseconds.  Note this is
	only used for testing purposes.
*/
#ifdef WIN32
 #define MSDELAY(x) sleep((x))
#else
 #define MSDELAY(x) delay((x))
#endif

/*
	The interface to the Sockets driver is performed via
	a software interrupt with appropriate registers
	set to indicate the function, various parameters,
	and to pass pointers for data transfer.
*/
   typedef struct _tagX86Regs
{
	WORD	ax;
	WORD	bx;
	WORD	cx;
	WORD	dx;
	WORD	si;
	WORD	di;
	WORD	ds;
	WORD	es;
} X86Regs;


/* Global API calls used within the Sockets Compatible API */
int CallSocketsDosApi(X86Regs *px86regs);

/* macro to get the current timer ticks from BDA */
#ifdef __DJGPP__
#define GET_TICKS _farpeekw(_dos_ds, 0x46c)
#else
#define GET_TICKS *((WORD D_FAR *)0x46c)
#endif

/* macro to declare the DPMI sections to lock */
#ifdef __DJGPP__
#define DPMI_SECTION_EXTERNS \
   extern char stext _ASM("stext"); \
   extern char etext _ASM("etext"); \
   extern char sdata _ASM("sdata"); \
   extern char edata _ASM("edata"); \
   extern char sbss  _ASM("sbss"); \
   extern char ebss  _ASM("ebss")

#define DPMI_LOCK_SECTIONS \
   _go32_dpmi_lock_code(&stext, (DWORD)&etext - (DWORD)&stext); \
   _go32_dpmi_lock_data(&sdata, (DWORD)&edata - (DWORD)&sdata); \
   _go32_dpmi_lock_data(&sbss,  (DWORD)&ebss  - (DWORD)&sbss)

#else
#define DPMI_SECTION_EXTERNS
#define DPMI_LOCK_SECTIONS
#endif

#ifdef DEBUG
//
// macro to write a value to a debug port
//
#ifdef __DJGPP__
#define DebugWritePort(port, val) ({ \
   _ASM( \
      "pushw   %%ax\n\t" \
      "pushw   %%dx\n\t" \
      "movb    %1, %%al\n\t" \
      "movw    %0, %%dx \n\t" \
      "outb    %%al, %%dx \n\t" \
      "popw    %%dx\n\t" \
      "popw    %%ax" \
      : \
      : "g" (port), "g" (val) \
      : "ax" , "dx", "memory"); \
   })
#else
#define DebugWritePort(port, val) ({ \
   _ASM  push  ax \
   _ASM  push  dx \
   _ASM  mov   al, val \
   _ASM  mov   dx, port \
   _ASM  out   dx, al \
   _ASM  pop   dx \
   _asm  pop   ax \
   })
#endif
#else
#define DebugWritePort(port, val)
#endif

