//
//   $Id: ARP.C 1.6 2002/11/26 21:24:56Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// ARP.C
// Address Resolution Protocol implementation
// Copyright (c) GP van Niekerk Ondernemings 1990-2001
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

/* Revision history:

2002-09-16 gvn DHCP ARP followed by gratuitous ARP

*/

#include <alloc.h>
#include <string.h>
#include "inetdefs.h"
#include "internet.h"
#include "mbuf.h"
#include "timer.h"
#include "iface.h"
#include "enet.h"
#include "arp.h"
#include "ip.h"
#include "bootp.h"

void ip_route(MBUF *bp,char rxbroadcast,IFACE *ifp);
void arp_output(struct interface *ifp,WORD hardware,DWORD source,DWORD target);
struct mbuf *htonarp(struct arp *arp);
static unsigned arp_hash(WORD hardware,DWORD ipaddr);

/* ARP entries for particular subnetwork types. The table values
 * are filled in by calls to arp_init() at device attach time
 */
#define	NTYPES	NHWTYPES
struct arp_type arp_type[NTYPES];

/* Hash table headers */
struct arp_tab *arp_tab[ARPSIZE];

struct arp_stat arp_stat;

//
// ARP type to send
//
int arp_init_type = ARP_TYPE_BOTH;

/* Initialize an entry in the ARP table
 * Called by the device driver when 'interface' is executed
 */
arp_init(
IFACE *ifp,
WORD hwtype,		// ARP Hardware type
int hwalen,			// Hardware address length
int iptype,			// Subnet's protocol ID for IP
WORD arptype,		// Subnet's protocol ID for ARP
BYTE *bdcst,		// Subnet's broadcast address (if any)
void (*format)(),	// Function to format hardware addresses
void (*scan)(BYTE *out,char *in[],int cnt))// Function to scan addresses in ascii
{
	register struct arp_type *at;

	if(hwtype >= NTYPES)
		return -1;	/* Table too small */

	at = &arp_type[hwtype];
	at->hwalen = (WORD)hwalen;
	at->iptype = (WORD)iptype;
	at->arptype = (WORD)arptype;
	at->bdcst = bdcst;
	at->format = format;
	at->scan = scan;
	if ((hwtype == ARP_ETHER || hwtype == ARP_IEEE) &&
					!(ifp->flags & TOKEN_FLAG)) {
		ifp->flags |= ETH_FLAG;
	}
	else if (hwtype == ARP_IPX)
		ifp->flags |= IPX_FLAG;

   // do gratuitous arp?
   if (arp_init_type & ARP_TYPE_DHCP)
   {
      // do DHCP ARP
      arp_output(ifp,hwtype,0,ifp->ip_addr);	// DHCP ARP

      // do Gratuitous ARP if DHCP arp times out?
      if (arp_init_type & ARP_TYPE_GRAT)
      {
      	ifp->flags |= GARP_FLAG;
      	if (run_timer(&ifp->timer))
   	   	ifp->flags |= TIMER_FLAG;
      	set_timer(&ifp->timer,200);
      	start_timer(&ifp->timer);
      }
   }
   else
   {
   	arp_output(ifp,hwtype,ifp->ip_addr,ifp->ip_addr);	// Gratuitous ARP
   }
	return 0;
}

/* Resolve an IP address to a hardware address; if not found,
 * initiate query and return NULLCHAR.	If an address is returned, the
 * interface driver may send the packet; if NULLCHAR is returned,
 * res_arp() will have saved the packet on its pending queue,
 * so no further action (like freeing the packet) is necessary.
 */
BYTE *res_arp(
struct interface *interface,	// Pointer to interface block
DWORD target,						// Target IP address
struct mbuf *bp,					// IP datagram to be queued if unresolved
BYTE **rif)							// pointer to buffer for RIF (token ring only)
{
	register struct arp_tab *arp;
	WORD hardware;					// Hardware type

	if (interface->flags & (IEEE_FLAG | TOKEN_FLAG)) {
		hardware = ARP_IEEE;
	}
	else if (interface->flags & ETH_FLAG)
		hardware = ARP_ETHER;
	else
		hardware = ARP_IPX;
	if((arp = arp_lookup(hardware,target)) != NULLARP && arp->state == ARP_VALID) {
	  if (rif)
		 *rif = arp->rif;
	  return arp->hw_addr;
	}
	/* Create an entry and put the datagram on the
	 * queue pending an answer
	 */
	arp = arp_add(target,hardware,NULLCHAR,0,0);
	if (len_q(arp->pending) < MAX_ARP_Q)
		enqueue(&arp->pending,bp);
	else
		free_p(bp);
	arp_output(interface,hardware,interface->ip_addr,target);
	return NULLCHAR;
}
/* Handle incoming ARP packets. This is almost a direct implementation of
 * the algorithm on page 5 of RFC 826, except for:
 * 1. Outgoing datagrams to unresolved addresses are kept on a queue
 *	  pending a reply to our ARP request.
 * 2. The names of the fields in the ARP packet were made more mnemonic.
 */
void arp_input(struct interface *ifp,struct mbuf *bp,BYTE *rif)
{
	struct arp arp;
	struct arp_tab *ap;
	struct arp_type *at;
	char eth[128];
	int32 temp;
	ROUTE *rp;

	arp_stat.recv++;
	if(ntoharp(&arp,&bp) == -1)	/* Convert into host format */
		return;
	if(arp.hardware >= NTYPES){
		/* Unknown hardware type, ignore */
		arp_stat.badtype++;
		return;
	}
	at = &arp_type[arp.hardware];
	if(arp.protocol != at->iptype){
		/* Unsupported protocol type, ignore */
		arp_stat.badtype++;
		return;
	}
	if (arp.hwalen > MAXHWALEN || uchar(arp.pralen) != sizeof(DWORD)) {
		/* Incorrect protocol addr length (different hw addr lengths
		 * are OK since AX.25 addresses can be of variable length)
		 */
		arp_stat.badlen++;
		return;
	}
	if(memcmp(arp.shwaddr,at->bdcst,at->hwalen) == 0){
		/* This guy is trying to say he's got the broadcast address! */
		arp_stat.badaddr++;
		return;
	}
	if (rif) {
		rif[0] &= 0x1f;	// strip broadcast bits
		rif[1] ^= 0x80;	// change direction
	}
	if_mark_up(ifp,0);
	/* If this guy is already in the table, update its entry
	 * unless it's a manual entry (noted by the lack of a timer)
	 */
	ap = NULLARP;	/* ap plays the role of merge_flag in the spec */
	if((ap = arp_lookup(arp.hardware,arp.sprotaddr)) != NULLARP
	 && ap->timer.start != 0){
		ap = arp_add(arp.sprotaddr,arp.hardware,arp.shwaddr,uchar(arp.hwalen),rif);
	}
	/* See if we're the address they're looking for */
	if (arp.tprotaddr == ifp->ip_addr || is_this_addr(arp.tprotaddr) ||
			(proxy_arp && (rp = rt_lookup(arp.tprotaddr,0,SKIP_DOWN)) != 0 &&
													rp->interface != ifp)) {
		if (ap == NULLARP) {	/* Only if not already in the table */
			arp_add(arp.sprotaddr,arp.hardware,
													arp.shwaddr,uchar(arp.hwalen),rif);
		}
		if (arp.opcode == ARP_REQUEST) {
			/* Swap sender's and target's (us) hardware and protocol
			 * fields, and send the packet back as a reply
			 */
			memcpy(arp.thwaddr,arp.shwaddr,(WORD)uchar(arp.hwalen));
			/* Mark the end of the sender's AX.25 address
			 * in case he didn't
			 */

			memcpy(arp.shwaddr,ifp->hwaddr,at->hwalen);
			if ((temp = arp.sprotaddr) == 0)
				temp = arp.tprotaddr;		// probably a DHCP ARP
			arp.sprotaddr = arp.tprotaddr;
			arp.tprotaddr = temp;
			arp.opcode = ARP_REPLY;
			if ((bp = htonarp(&arp)) != 0)
				(*ifp->output)(ifp,arp.thwaddr,ifp->hwaddr,at->arptype,bp,rif);
			arp_stat.inreq++;
			return;
		}
		else if (arp.opcode == ARP_REPLY)
			arp_stat.replies++;
	}
	if (arp.opcode == ARP_REPLY && is_this_addr(arp.sprotaddr)) { /* duplicate! */
	//{ // force duplicate for debugging
		if(arp_type[arp.hardware].format != 0){
			(*arp_type[arp.hardware].format)(eth,arp.shwaddr);
		} else {
			eth[0] = '\0';
		}
		printf("\nWarning!! Duplicate IP address detected\n"
				 "Hardware address: %s\n",eth);
		dhcp_decline(ifp);
	}
}

/* Add an IP-addr / hardware-addr pair to the ARP table */
struct arp_tab *arp_add(
DWORD ipaddr,		/* IP address, host order */
WORD hardware,		/* Hardware type */
BYTE *hw_addr,		/* Hardware address, if known; NULLCHAR otherwise */
WORD hw_alen,		/* Length of hardware address */
BYTE *rif)			// rif (tokenring)
{
	void arp_drop();
	struct mbuf *bp,*dequeue();
	register struct arp_tab *ap;
	unsigned hashval;

	if((ap = arp_lookup(hardware,ipaddr)) == NULLARP){
		/* New entry */
		if((ap = (struct arp_tab *)calloc(1,sizeof(struct arp_tab),AllocArpArp)) == NULLARP)
			return NULLARP;
		ap->timer.func = (void(*)(char *))arp_drop;
		ap->timer.arg = (char *)ap;
		ap->hardware = hardware;
		ap->ip_addr = ipaddr;

		/* Put on head of hash chain */
		hashval = arp_hash(hardware,ipaddr);
		ap->prev = NULLARP;
		ap->next = arp_tab[hashval];
		arp_tab[hashval] = ap;
		if(ap->next != NULLARP){
			ap->next->prev = ap;
		}
	}
	if(hw_addr == 0){
		/* Await response */
		ap->state = ARP_PENDING;
		ap->timer.start = PENDTIME * (1000 / MSPTICK);
	} else {
		/* Response has come in, update entry and run through queue */
		ap->state = ARP_VALID;
		ap->timer.start = ARPLIFE * (1000 / MSPTICK);
		if(ap->hw_addr != 0)
			free(ap->hw_addr);
		if((ap->hw_addr = (BYTE *)malloc(hw_alen,AllocArpMac)) == 0){
			free((char *)ap);
			return NULLARP;
		}
		memcpy(ap->hw_addr,hw_addr,hw_alen);
		if (rif && (rif[0] &= 0x1f) != 0 && !ap->rif) {
			if ((ap->rif = malloc(rif[0],AllocArpRif)) == 0) {
				free (ap);
				return NULLARP;
			}
			memcpy(ap->rif,rif,rif[0]);
		}
		while((bp = dequeue(&ap->pending)) != NULLBUF)
			ip_route(bp,0,0);
	}
	start_timer(&ap->timer);
	return ap;
}

/* Remove an entry from the ARP table */
void arp_drop(struct arp_tab *ap)
{
	if(ap == NULLARP)
		return;
	stop_timer(&ap->timer);	/* Shouldn't be necessary */
	if(ap->next != NULLARP)
		ap->next->prev = ap->prev;
	if(ap->prev != NULLARP)
		ap->prev->next = ap->next;
	else
		arp_tab[arp_hash(ap->hardware,ap->ip_addr)] = ap->next;
	if(ap->hw_addr != 0)
		free(ap->hw_addr);
	if(ap->rif)
		free(ap->rif);
	free_q(&ap->pending);
	free((char *)ap);
}

/* Look up the given IP address in the ARP table */
struct arp_tab *arp_lookup(WORD hardware,DWORD ipaddr)
{
	register struct arp_tab *ap;

	for(ap = arp_tab[arp_hash(hardware,ipaddr)]; ap != NULLARP; ap = ap->next){
		if(ap->ip_addr == ipaddr && ap->hardware == hardware)
			break;
	}
	return ap;
}

/* Send an ARP request to resolve IP address target_ip */
void arp_output(struct interface *ifp,WORD hardware,DWORD source,DWORD target)
{
	struct arp arp;
	struct mbuf *bp;
	struct arp_type *at;
	static BYTE bc_rif[2] = {0x82,0};

	at = &arp_type[hardware];
	if(ifp->output == NULLFP)
		return;

	arp.hardware = hardware;
	arp.protocol = at->iptype;
	arp.hwalen = at->hwalen;
	arp.pralen = sizeof(DWORD);
	arp.opcode = ARP_REQUEST;
	memcpy(arp.shwaddr,ifp->hwaddr,at->hwalen);
	arp.sprotaddr = source;
	memset(arp.thwaddr,0,at->hwalen);
	arp.tprotaddr = target;
	if ((bp = htonarp(&arp)) != 0)
		(*ifp->output)(ifp,at->bdcst,ifp->hwaddr,at->arptype,bp,bc_rif);
	arp_stat.outreq++;
}

/* Hash a {hardware type, IP address} pair */
static unsigned arp_hash(WORD hardware,DWORD ipaddr)
{
	register unsigned hashval;

	hashval = hardware;
	hashval ^= hiword(ipaddr);
	hashval ^= loword(ipaddr);
	return hashval % (unsigned)ARPSIZE;
}		
/* Copy a host format arp structure into mbuf for transmission */
struct mbuf *htonarp(struct arp *arp)
{
	struct mbuf *bp;
	register BYTE *buf;

	if(arp == (struct arp *)NULL)
		return NULLBUF;
	if((bp = alloc_mbuf(sizeof(struct arp),AllocArpSnd)) == NULLBUF)
		return NULLBUF;

	buf = bp->data;

	buf = put16(buf,arp->hardware);
	buf = put16(buf,arp->protocol);
	*buf++ = arp->hwalen;
	*buf++ = arp->pralen;
	buf = put16(buf,arp->opcode);
	memcpy(buf,arp->shwaddr,(WORD)uchar(arp->hwalen));
	buf += arp->hwalen;
	buf = put32(buf,arp->sprotaddr);
	memcpy(buf,arp->thwaddr,(WORD)uchar(arp->hwalen));
	buf += arp->hwalen;
	buf = put32(buf,arp->tprotaddr);

	bp->cnt = buf - bp->data;
	return bp;
}
/* Convert an incoming ARP packet into a host-format structure */
int ntoharp(struct arp *arp,struct mbuf **bpp)
{
	if(arp == (struct arp *)NULL || bpp == NULLBUFP)
		return -1;

	arp->hardware = pull16(bpp);
	arp->protocol = pull16(bpp);
	arp->hwalen = pullchar(bpp);
	arp->pralen = pullchar(bpp);
	arp->opcode = pull16(bpp);
	pullup(bpp,arp->shwaddr,(WORD)uchar(arp->hwalen));
	arp->sprotaddr = pull32(bpp);
	pullup(bpp,arp->thwaddr,(WORD)uchar(arp->hwalen));

	arp->tprotaddr = pull32(bpp);

	/* Get rid of anything left over */
	free_p(*bpp);
	*bpp = NULLBUF;
	return 0;
}
