//
//   $Id: ASY.C 1.12 2003/08/22 05:48:50Z johnb Exp johnb $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// ASY.C
// Copyright (C) 2001 Datalight, Inc.
// All Rights Reserved
// Datalight Confidential and Proprietary
// Portions Copyright (c) GPvNO 1990-2001
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q
//
// Async driver for IBM-PC comports
//
// Revision History
// 08/20/2003 gvn rename playback command from "trace" to "play"
// 07/20/2002 gvn Better handling of shared interrupt ports - see note below
// 07/15/2002 gvn V25 modifications
// 07/23/2001 gvn Allocation tracing
// 06/19/2001 gvn updated previous fix to handle IOCTL_DISABLE to be able
//						to handle sharing IRQs.
// 06/06/2001 jmb changed continue in the asy interrupt handler to a break
//                the continue was causing hang when the IOCTL_DISABLE
//                command was generated and an interrupt occurred on the
//                serial port.
// 05/29/2001 gvn updates...
//
/*
	Note on interrupt handling of serial ports:

	1. It is possible to have Sockets use one or more ports sharing an
	interrupt with one or more other ports being used by another application.
	To make this work properly the interrupt service routines must be
	written as follows:

	1. If the application is loaded before Sockets, it must always clear the
	PIC/PICs at the end of the service routine, regardless of whether it
	has serviced one of the ports belonging to itself. In this case the /s
	switch should be used in "the interface" command for the port in Sockets.

	2. If the application is loaded after Sockets, it must service its port(s),
	and call or chain to Sockets (using the vector it found when initializing)
	WITHOUT clearing the PIC/PICs.
*/

#if !APL
#pragma option -zCDRIVER_CODE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <ctype.h>
#include <alloc.h>
#include <mem.h>
#include "inetdefs.h"
#include "pc.h"
#include "timer.h"
#include "iface.h"
#include "modem.h"
#include "netfunc.h"
#include "ser.h"
#include "8250.h"
#include "mbuf.h"
#ifdef LAPB
#include "lapb.h"
#endif
#include "slhc.h"
#include "ppp.h"

#ifdef V25
#define VS_TE 0x20
#define VS_RR 0x10
#define VS_PE	4
#define VS_FE  2
#define VS_OE  1
#define VS_ERR (VS_PE | VS_FE | VS_OE)
#endif

int doppp_commands(int argc,char *argv[],void *ifp);
#define SP_CHAR			0x20

#ifdef SER_TEST_R
char far r_buf[0xff00];
char far *pr_buf = r_buf;

void put_rbuf(char ch)
{
	*pr_buf = ch;
	if (++pr_buf >= r_buf + 0xff00)
		pr_buf = r_buf;
}
#endif
#ifdef SER_TEST_X
char far x_buf[0xff00];
char far *px_buf = x_buf;

void put_xbuf(char ch)
{
	*px_buf = ch;
	if (++px_buf >= x_buf + 0xff00)
		px_buf = x_buf;
}
#endif

unsigned nser;
char cAsySched;			// set if frame completed
//char diagstr[120];		// !!! move to malloc area!!!!
extern char window_col[];
extern char badhost[];
extern void interrupt (*asyvecs[])();

void near asyint(SER *serp);
void near asytxint(SER *serp,BYTE bIsInt);
void near asyrxint(SER *serp);
void near asy_start_xmit(SER *serp);

static int near of_stopped(SER *serp);
void asy_disconnect(SER *serp);		// disconnect by dropping DTR for 1 second
int asy_peek(IFACE *ifp,char *buf,WORD cnt);
void modem_command(SER *serp,MODEMPAR *pm,BYTE *pcommand);

#ifdef APL
void chartimeout(SER *serp)		// called on character timeout
{
	serp->deliver |= DEL_TO;
	((IFACE *)serp->iface)->sflags |= SF_DELIVER;
}
#endif

#ifndef V25
void modemtimeout(SER *serp)		/* called on modem timeout */
{
	MODEMPAR *pm = serp->pmpar;
	IFACE *ifp = (IFACE *)serp->iface;
	unsigned int tmstate = pm->tmstate;

	pm->tmstate = TM_IDLE;
	if (serp->flags & FL_DISABLED) {
		start_timer(&pm->timer);	// just keep it going until enabled again
		return;
	}
	if (pm->flags & MF_DISCON) {
		outportb(serp->addr+MCR,serp->mcr |= MCR_DTR);
		pm->flags &= ~(MF_DISCON | MF_DCDLOST);
		asy_modem_init(ifp);
		return;
	}
	switch (tmstate) {
	case TM_WTO:
		modem_command(serp,pm,pm->nextptr);
		break;
	case TM_WDCD:
		if (pm->retry && --pm->retry == 0) {
			if (pm->flags & (MF_DISX | MF_DISR)) {
				printf("%s:retries exhausted\n",ifp->name);
			}
			release_sif(ifp);
			serp->ostate = OS_IDLE;
			asy_disconnect(serp);
		}
		else {
			modem_command(serp,pm,pm->dialcom);
		}
		break;
	case TM_WRCV:
	case TM_WXID:
	case TM_WADD:
		if ((!(inportb(serp->addr + MSR) & MSR_RLSD)) || (pm->retry && --pm->retry == 0)) {
			if (pm->flags & (MF_DISX | MF_DISR)) {
				printf("%s:retries exhausted\n",ifp->name);
			}
			release_sif(ifp);
			serp->ostate = OS_IDLE;
			asy_disconnect(serp);
		}
		else {
			pm->tmstate = TM_IDLE;
			modem_command(serp,pm,pm->state == MODEM_CONNECT ? pm->connect : pm->answer);
		}
		break;
	case TM_PACE:
		if (serp->cnt) {
			// transmit next character
			asy_start_xmit(serp);
		}
		else {
			serp->ostate = OS_IDLE;
			modem_command(serp,pm,pm->nextptr);
		}
		break;
	}
}

void asy_disconnect(SER *serp)		// disconnect by dropping DTR for 10 second
{
	MODEMPAR *pm = serp->pmpar;
	TIMER *ptim;

	if (pm && !(serp->flags & FL_DISABLED)) {
		IFACE *ifp = (IFACE *)serp->iface;
		if (ifp->sflags & SF_UNC)
			release_sif(ifp);
		outportb(serp->addr+MCR,serp->mcr &= ~MCR_DTR);
		pm->flags |= MF_DISCON;
		pm->nextptr = 0;
		pm->tmstate = TM_IDLE;
		ptim = &pm->timer;
		set_timer(ptim,pm->time);
		start_timer(ptim);
		pm->state = MODEM_INIT;
		serp->istate = IS_MODEM;
	}
}

asy_modem_init(IFACE *ifp)
{
	WORD buf8th;
	SER *serp = ifp->dcb;
	MODEMPAR *pm = serp->pmpar;

	serp->flags &= ~FL_CONNECTED;
	serp->flags |= FL_MODEM;
	pm->timer.arg = (char *)serp;
	pm->timer.func = (void (*)(char *))modemtimeout;
	pm->retry = pm->retry_val;
	pm->flags &= ~MF_INIT;
	if (!(ifp->sflags & SF_RAW)) {
		/* Set up receiver FIFO */
		if (!serp->buf && (serp->buf = malloc(serp->bufsize,AllocAsyRbuf)) == 0) {
			return -1;
		}
		buf8th = serp->bufsize >> 3;
		serp->stop_cnt = buf8th * 7;
		serp->start_cnt = buf8th * 6;
		serp->wp = serp->rp = serp->buf;
		serp->icnt = 0;
	}
	pm->state = MODEM_INIT;
	serp->istate = IS_MODEM;
	modem_command(serp,pm,pm->init);
	return 0;
}

void modem_command(SER *serp,MODEMPAR *pm,BYTE *pcommand)
{
	BYTE ch,ch1,*pn;
	TIMER *ptim = &pm->timer;
	IFACE *ifp = serp->iface;

	if (!pcommand) {
		pm->nextptr = 0;
		serp->flags |= FL_EVENT;
		return;
	}
	while ((ch = *pcommand) != 0 && ch == pm->com_char) {
		char set_ptr = 0;
		++pcommand;
		switch(ch = *pcommand++) {
		case 'a':		// wait for IP address
			pm->tmstate = TM_WADD;
			set_ptr = 2;
			goto set_timer;
		case 'd':		/* wait for DCD */
			pm->tmstate = TM_WDCD;
			set_ptr = 1;
			goto set_timer;
		case 'f':		// find XID match
			pm->tmstate = TM_WXID;
			set_ptr = 2;
			goto set_timer;
		case 'm':		// control modem signals DTR/RTS
			if ((ch = *pcommand) == 0 || (ch1 = *++pcommand) == 0)
				break;
			++pcommand;
			if (ch == 'd') {
				if (ch1 == '1')
					serp->mcr |= MCR_DTR;
				else if (ch1 == '0')
					serp->mcr &= ~MCR_DTR;
			}
			else if (ch == 'r') {
				if (ch1 == '1')
					serp->mcr |= MCR_RTS;
				else if (ch1 == '0')
					serp->mcr &= ~MCR_RTS;
			}
			outportb(serp->addr + MCR,serp->mcr);
			break;
		case 'n':		// insert tel #
			pm->nextptr = pcommand;
			if (pm->fdial[0]) {
				if (!pm->dial || *pm->dial == 0)
					pm->dial = pm->fdial;
				serp->data = *pm->dial++;
				serp->cnt = strlen((char *)serp->data);
				goto send_it;
			}
			else
				serp->ostate = OS_IDLE;
			break;
		case 'r':
			pm->tmstate = TM_WRCV;
			set_ptr = 3;
			goto set_timer;
		case 'w':		/* wait */
			pm->tmstate = TM_WTO;
			set_ptr = 1;
			set_timer:
			set_timer(ptim,atol((char *)pcommand));
			while (isdigit(*pcommand)) {
				++pcommand;
			}
			if (set_ptr >= 2) {
				if (*pcommand == pm->com_char && pcommand[1] == '.')
					pcommand += 2;
			}
			if (set_ptr >= 3) {
				serp->data = pcommand;
				for (serp->cnt = 0;*pcommand && *pcommand != pm->com_char;
										++serp->cnt,++pcommand);
				if (*pcommand)
					++pcommand;
			}
			if (set_ptr)
				pm->nextptr = pcommand;
			start_timer(ptim);
			return;
		case 'x':		// insert exchange ID
			pm->nextptr = pcommand;
			serp->data = pm->xchange;
			serp->cnt = strlen((char *)serp->data);
			goto send_it;
		default:
			continue;		// ignore unknown commands
		}
	}
	if (!ch) {
		pm->nextptr = 0;
		serp->flags |= FL_EVENT;
		return;
	}
	serp->data = pcommand;
	serp->cnt = strlen((char *)pcommand);
	if ((pn = (BYTE *)strchr((char *)pcommand,pm->com_char)) != 0)
		serp->cnt = pn - pcommand;
	pm->nextptr = pcommand + serp->cnt;
	send_it:
	if (pm->flags & MF_DISX) {
		printf("%s:%.*s\n",ifp->name,serp->cnt,serp->data);
	}
	serp->ostate = OS_MODEM;
	asy_start_xmit(serp);
}
#else
// ignore for V25
#pragma argsused
void asy_disconnect(SER *serp)	// disconnect by dropping DTR for 10 second
{
}
#endif

// Initialize asynch port
#pragma argsused
int asy_init(IFACE *ifp,unsigned int speed)
{
	extern int nservec;
	WORD intno;
#ifndef V25
	char maskbit,shared = 0;
#endif
	SER *serp = ifp->dcb;
	unsigned base = serp->addr;

	intno = serp->vec
#ifndef V25
							+ (serp->vec >= 8 ? 0x68 : 8)
#endif
							;
	idisable();
	if (serp->oldvec) {
		// port uses shared interrupt
#ifndef V25
		shared = 1;
#endif
		((SER *)serp->oldvec)->next = serp;
		serp->oldvec = 0;
	}
	else {
		// Save original interrupt vector
		serp->oldvec = getvect(intno);

		// Set interrupt vector to handler
		setvect(intno,asyvecs[nservec - 1]);
	}
#ifdef V25
	if (serp->oldvec1) {
		// port uses shared interrupt
		((SER *)serp->oldvec1)->next = serp;
		serp->oldvec1 = 0;
	}
	else {
		// Save original interrupt vector
		serp->oldvec1 = getvect(intno + 1);

		// Set interrupt vector to handler
		setvect(intno + 1,asyvecs[nservec + 1]);
	}
	serp->RxB = MK_FP(0xff00,0xf60 + (base ? 0x10 : 0));
	serp->TxB = serp->RxB + 2;
	serp->SCM = serp->RxB + 8;
	serp->SCE = serp->RxB + 0xb;
#define MB 3
#else
#define MB 1
	/* Purge the receive data buffer and line status register */
	(void)inportb(base+RBR);
	(void)inportb(base+LSR);

	serp->oldlcr = inportb(base+LCR);
	/* Set line control register: 8 bits, no parity */
	outportb(base+LCR, serp->lcr = LCR_8BITS);

	/* determine if 16550A, turn on FIFO mode and clear RX and TX FIFOs */
	outportb(base+FCR,(char) FIFO_ENABLE);

	/* According to National ap note AN-493, the FIFO in the 16550 chip
	 * is broken and must not be used. To determine if this is a 16550A
	 * (which has a good FIFO implementation) check that both bits 7
	 * and 6 of the IIR are 1 after setting the fifo enable bit. If
	 * not, don't try to use the FIFO.
	 */
	if ((inportb(base+IIR) & IIR_FIFO_ENABLED) == IIR_FIFO_ENABLED) {
		serp->flags |= FL_IS16550A;
		outportb(base+FCR,(char) FIFO_SETUP);
	} else {
		/* Chip is not a 16550A. In case it's a 16550 (which has a
		 * broken FIFO), turn off the FIFO bit.
		 */
		serp->flags &= ~FL_IS16550A;
		outportb(base+FCR,(char)0);
	}

	/* Turn on receive interrupt enable in 8250, leave transmit
	 * and modem status interrupts turned off for now
	 */
	outportb(base+IER,IER_DAV | IER_RLS);

	/* Set modem control register: assert DTR, RTS, turn on 8250
	 * master interrupt enable (connected to OUT2)
	 */
	serp->mcr = MCR_OUT2 |
		(serp->iflow == (IF_DTR | IF_INV) ? 0 : MCR_DTR) |
		(serp->iflow == (IF_RTS | IF_INV) ? 0 : MCR_RTS);
	outportb(base + MCR,serp->mcr);
	if (!shared) {
		/* Clear mask (enable interrupt) in 8259 interrupt controller */
		if (serp->vec < 8) {
			serp->oldmask = inportb(INTMASK);
			outportb(INTMASK,
					serp->oldmask & ~(maskbit = (MB << serp->vec)));
		}
		else {
			serp->oldmask = inportb(INTMASK1);
			outportb(INTMASK1,
				serp->oldmask & ~(maskbit = (MB << (serp->vec - 8))));
		}
		serp->oldmask &= maskbit;
	}
	asy_speed(ifp,speed);
#endif
	popf();
#ifndef V25
	if (serp->pmpar) {
		if (asy_modem_init(ifp) < 0)
			return -1;
	}
	else
#endif
			if (!(ifp->sflags & SF_UNC)) {
		serp->flags |= FL_CONNECTED;
	}
	return 0;
}

// Re-initialize asynch port
static near void asy_reinit(IFACE *ifp)
{
	extern int nservec;
	SER *serp = ifp->dcb;
#ifndef V25
	unsigned base = serp->addr;
#endif

	idisable();
#ifndef V25
	// Purge the receive data buffer and line status register
	(void)inportb(base+RBR);
	(void)inportb(base+LSR);

	// Set line control register
	outportb(base+LCR, serp->lcr);

	if (serp->flags & FL_IS16550A)
		outportb(base+FCR,(char) FIFO_SETUP);
	outportb(base+IER,IER_DAV | IER_RLS);
	outportb(base + MCR,serp->mcr);
#endif
	if (serp->vec < 8)
		outportb(INTMASK,inportb(INTMASK) & ~(MB << serp->vec));
	else
		outportb(INTMASK1,inportb(INTMASK1) & ~(MB << (serp->vec - 8)));
	asy_speed(ifp,(unsigned int)serp->speed);
	popf();
#ifndef V25
	if (serp->pmpar) {
		asy_modem_init(ifp);
	}
#endif
}

void asy_stop(IFACE *ifp)
{
	SER *serp = ifp->dcb;
#ifndef V25
	unsigned base;

	if (!(serp->flags & FL_DISABLED)) {
		base = serp->addr;
		/* Ensure that DLAB bit is off */
		outportb(base+LCR,serp->oldlcr);

		/* Turn off all 8250 interrupt enable bits */
		outportb(base+IER,(char)0);

		/* Turn off DTR, RTS, master interrupt enable */
		outportb(base+MCR,serp->mcr = (char)0);

		/* Purge the receive data buffer */
		inportb(base+RBR);

		/* and hardware fifos if available */
		if (serp->flags & FL_IS16550A)
			outportb(base+FCR,(char) 0);
	}
	// Set 8259 interrupt mask (turn off interrupts)
	if (serp->oldmask) {
		if (serp->vec < 8)
			setbit(INTMASK,serp->oldmask);
		else
			setbit(INTMASK1,serp->oldmask);
	}
	// Restore original interrupt vector
	setvect(serp->vec + (serp->vec >= 8 ? 0x68 : 8),serp->oldvec);
#else
	setvect(serp->vec,serp->oldvec);
	setvect(serp->vec + 1,serp->oldvec1);
#endif
}

#if APL
char *ostates[] = {		// Output states
	"IDLE","RAW","FSTART","DATA","FRESC","FREND","CRC1","CRC1ESC",
	"CRC1END","CRC2","CRC2ESC","CRC2END","END","FHDLC","HDLCDATA",
	"HDLCEND","HDLCCRC","HDLCEA","FCS1","FCS1ESC","FCS2","FCS2ESC",
	"MODEM"
};

char *istates[] = {		// Input states
	"RAW","DATA","ESC","IGN","DATAB","ESCB","IGNB","HDLC",
	"PPP","PPPESC","PPPIGN","MODEM"
};
#endif

/* Asynchronous line I/O control */
asy_ioctl(IFACE *ifp,int argc,char *argv[])
{
	SER *serp = ifp->dcb;
	MODEMPAR *pm;
	FSM *psFSM;
	PPP *psPPP;
	char ch,ch1;
	char speedflag;
	WORD i;
	int iParam;
	BYTE bStart;
#if APL
	char clearflag = 0,par,mod;
	WORD errtot;
	char **pcom;
	static char *commands[] = {
		"clear","con","dis","play",0
	};
#endif
	speedflag = 0;
	if (argc == 1 && argv[0][0] == 0) {	/* internal commands */
		if (!serp)
			return -1;
		switch (argv[0][1]) {
		case IOCTL_REL_RSC:	/* release resources */
			free_p(ifp->mbuf);
			break;
		case IOCTL_STP_TMR:	/* stop timer */
			stop_timer(&serp->timer);
			break;
		case IOCTL_GET_CNT:	/* get receive byte count */
			return serp->icnt;
		case IOCTL_PROD_OUTPUT:	/* prod the output incase intr lost */
			if (serp->ostate != OS_IDLE && !(serp->flags & FL_DISABLED)) {
				asy_start_xmit(serp);
			}
			break;
		case IOCTL_ENABLEPORT:		// enable port on ASY
			serp->flags &= ~FL_DISABLED;
			asy_reinit(ifp);
			break;
		case IOCTL_DISABLEPORT:  	// disable port on ASY
			serp->flags |= FL_DISABLED;
			break;
#ifndef V25
		case IOCTL_CONNECT:			// start dial on ASY
		case IOCTL_DISCONNECT:		// drop modem on ASY
		case IOCTL_ENABLEDOD:    	// enable-dial-on-demand
		case IOCTL_DISABLEDOD:   	// disable dial-on-demand
			if ((pm = serp->pmpar) == 0)
				return -1;				// not supported
			switch (argv[0][1]) {
			case IOCTL_CONNECT:			// start dial on ASY
				pm->flags |= MF_DIALNOW;
				break;
			case IOCTL_DISCONNECT:		// drop modem on ASY
				asy_disconnect(serp);
				break;
			case IOCTL_ENABLEDOD:    	// enable-dial-on-demand
				pm->flags &= ~MF_NODOD;
				break;
			case IOCTL_DISABLEDOD:   	// disable dial-on-demand
				pm->flags |= MF_NODOD;
				break;
			}
			break;
#endif
		case IOCTL_GETSTATUS:    	// get dial status
			pm = serp->pmpar;
			i = 0;
			psFSM = ((PPP *)(ifp->prot_ext))->fsm;
			if ((psPPP = ifp->prot_ext) != 0) {
				i = psPPP->phase << 12;
				if ((psFSM = psPPP->fsm) != 0 &&
									psFSM[Lcp].pdv == psFSM[Lcp].pdv_in)
					i |= 0x800;		// STPPP_IN
			}
			return
#ifndef V25
						((inportb(serp->addr + MSR) &
						(MSR_RLSD | MSR_RI | MSR_CTS | MSR_DSR)) >> 2) |
						(serp->mcr & (MCR_DTR | MCR_RTS)) |
#endif
						((serp->flags & FL_CONNECTED) ? 0x40 : 0) |
						(pm ? ((pm->state + 1) & 7) << 8 : 0) | i;
		case IOCTL_PPP_OPEN:		// Open PPP
				bStart = 0;
				goto ppp_open;
		case IOCTL_PPP_START:	// Start PPP
				bStart = 1;
				ppp_open:
				iParam = PARAM_UP;
				goto do_ppp;
		case IOCTL_PPP_CLOSE:	// Close PPP
				iParam = PARAM_DOWN;
				do_ppp:
				if (ifp->prot_ext) {
					ppp_iostatus(ifp,iParam,bStart);
					break;
				}
				// fall tru
		default:
			return -1;
		}
		return 0;
	}
	if (argc > 0) {
		if (isdigit(argv[0][0]) != 0) {
			speedflag = 1;
			goto oldstyle;
		}
#if APL
		for (pcom = commands,i = 0;*pcom;++pcom,++i) {
			if (strncmp(argv[0],*pcom,strlen(argv[0])) == 0)
				break;
		}
		switch (i) {
		case 0:		/* "clear" */
			clearflag = 1;
			goto oldstyle;
#ifdef LAPB
		case 1:		/* "con" */
			if (ifp->sflags & SF_LAPB) {
				if (ifp->indcb)
					lapb_delete(ifp->indcb);
				lapb_init(ifp,0,0,0,0);
			}
			break;
#endif
		case 2:		/* "dis" */
#ifdef LAPB
			if (ifp->sflags & SF_LAPB) {
				if (ifp->indcb) {
					lapb_delete(ifp->indcb);
					ifp->indcb = 0;
				}
			}
			else
#endif
				asy_disconnect(serp);
			break;
#ifdef PT
		case 3:		// trace
		{
			FILE *fp;
			int iCh;
			MBUF *bp;

			if ((fp = fopen(argv[1],"rb")) != 0) {
				if (argc > 2)
					ifp->sflags |= SF_PPP_TRACE;
				while ((iCh = getc(fp)) != EOF) {
					if (iCh == HDLC_FLAG) {
						if ((bp = serp->rbp) != 0 && bp->cnt) {
							enqueue(&serp->rcvq,bp);
							serp->rbp = 0;
							schedule();
						}
					}
					else if ((bp = serp->rbp) == 0 && (bp = serp->rbp =
							alloc_mbuf(serp->bufsize,AllocAsyIbuf0)) == 0) {
						printf("PT:Out of memory\n");
						break;
					}
					else if (bp->cnt < serp->bufsize) {
						bp->data[bp->cnt++] = iCh;
					}
					else {
						printf("PT:Buffer too small\n");
						break;
					}
				}
				free_p(serp->rbp);
				serp->rbp = 0;
				fclose(fp);
				ifp->sflags &= ~SF_PPP_TRACE;
			}
			else
				printf("PT:Open error: %s\n",argv[1]);
		}
		break;
#endif
		default:		// not found, try PPP
#endif
#ifdef PPP_S
			if (ifp->flags & PPP_FLAG)
				doppp_commands(argc,argv,ifp);
			else {
				printf("Unknown parameter for interface '%s': %s\n",
								ifp->name,argv[0]);
				return -1;
			}
#endif
#if APL
			break;
		}
#endif
		return 0;
	}
	oldstyle:
	if (argc > 1) {		/* output flow specified */
		if ((ch = toupper(*argv[1])) == 'I') {
			serp->oflow = OF_INV;
			ch = toupper(argv[1][1]);
			ch1 = toupper(argv[1][2]);
		}
		else {
			serp->oflow = 0;
			ch1 = toupper(argv[1][1]);
		}
		if (ch == 'X')
			serp->oflow = OF_XON;
		else if (ch == 'C')
			serp->oflow |= OF_CTS;
		else if (ch == 'D')
			serp->oflow |= (ch1 == 'C') ?
					OF_DCD : OF_DSR;
	}
	if (argc > 2) {		/* input flow specified */
		if ((ch = toupper(*argv[2])) == 'I') {
			serp->iflow = IF_INV;
			ch = toupper(argv[2][1]);
		}
		else {
			serp->iflow = 0;
		}
		if (ch == 'X')
			serp->iflow = IF_XON;
		else if (ch == 'D')
			serp->iflow |= IF_DTR;
		else if (ch == 'R')
			serp->iflow |= IF_RTS;
#ifndef V25
		serp->mcr = MCR_OUT2 |
			(serp->iflow == (IF_DTR | IF_INV) ? 0 : MCR_DTR) |
			(serp->iflow == (IF_RTS | IF_INV) ? 0 : MCR_RTS);
		outportb(serp->addr + MCR,serp->mcr);
	}
	if (argc > 3) {		/* character format specified */
		char databits = argv[3][0];
		char stopbits = argv[3][1];
		char parity = toupper(argv[3][2]);

		databits = (databits >= '5' && databits <= '8') ?
			databits - '5' : LCR_8BITS;
		stopbits = (stopbits == '2') ? LCR_NSB : 0;
		parity = (parity == 'O') ? LCR_PEN :
			 (parity == 'E') ? (LCR_PEN | LCR_EPS) :
			 (parity == 'M') ? (LCR_PEN | LCR_SP) :
			 (parity == 'S') ? (LCR_PEN | LCR_SP | LCR_EPS) : 0;
		outportb(serp->addr+LCR,
				serp->lcr = databits | stopbits | parity);
#endif
	}
#if APL
	par = serp->lcr & (LCR_EPS | LCR_SP);
	if (argc < 1 || argc > 4 || clearflag) {
		if (ifp->sflags & SF_RAW)
			printf("%s: Raw line",ifp->name);
		else {
			printf("Slip link");
		}
#ifdef SER_TEST_R
		{
			char far *p;
			int i;

			printf("\nReceive buf:\n");
			for (p = r_buf,i = 0;p < pr_buf;++p) {
				printf("%02x ",*p);
				if (++i == 16) {
					printf("\n");
					i = 0;
				}
			}
			printf("\n");
		}
#endif
#ifdef SER_TEST_X
		{
			char far *p;
			int i;

			printf("\nXmit buf:\n");
			for (p = x_buf,i = 0;p < px_buf;++p) {
				printf("%02x ",*p);
				if (++i == 16) {
					printf("\n");
					i = 0;
				}
			}
			printf("\n");
		}
#endif
#ifndef V25
		mod = inportb(serp->addr + MSR);
		printf("%s%s%s%s%s\n",mod & MSR_RLSD ? " DCD" : "",
				  mod & MSR_CTS ? " CTS" : "",
				  mod & MSR_DSR ? " DSR" : "",
				serp->mcr & MCR_DTR ? " DTR" : "",
				serp->mcr & MCR_RTS ? " RTS" : "");
#endif
		printf("Characters received: %lu  sent: %lu\n"
					"ostate '%s' istate '%s' flg: %02x %s%s%s%s\n",
				serp->rcvcount,serp->xmtcount,
				ostates[serp->ostate],
				istates[serp->istate],serp->flags,
				serp->flags & FL_OFSTOPPED ? "OFC " : "",
				serp->flags & FL_IFSTOPPED ? "IFC " : "",
				serp->flags & FL_MODEM	   ? "MODEM " : "",
				serp->flags & FL_CONNECTED ? "CONNECTED" : "");
		if ((errtot = serp->overrun + serp->parerr + serp->frmerr +
					serp->brkerr + serp->buferr) != 0) {
			printf("Line errors: %u - Overrun: %u Parity: %u\n",
				errtot,serp->overrun,serp->parerr);
			printf("Framing: %u Break: %u Buffer: %u\n",
				serp->frmerr,serp->brkerr,serp->buferr);
		}
		if (serp->errors)
			printf("Slip errors: %u\n",serp->errors);
		if (serp->slcomp)
			slhc_status(serp->slcomp,3);
		if (clearflag) {
#ifdef SER_TEST_R
			pr_buf = r_buf;
			_fmemset(r_buf,0,0xff00);
#endif
#ifdef SER_TEST_X
			px_buf = x_buf;
			_fmemset(x_buf,0,0xff00);
#endif
			serp->overrun = serp->parerr = serp->frmerr =
			serp->brkerr = serp->buferr = serp->errors = 0;
			serp->rcvcount = serp->xmtcount = 0;
		}
		ch = serp->oflow & ~OF_INV;
		ch1 = serp->iflow & ~IF_INV;
		printf(	"Speed = %lu, Output flow control = %s%s, "
				"Input flow control = %s%s\n",
			serp->speed,
			(serp->oflow & OF_INV) ? "I" : "",
			ch == OF_XON ? "Xon/Xoff" :
			ch == OF_DCD ? "DCD" :
			ch == OF_DSR ? "DSR" :
			ch == OF_CTS ? "CTS" : "None",
			(serp->iflow & IF_INV) ? "I" : "",
			ch1 == IF_XON ? "Xon/Xoff" :
			ch1 == IF_DTR ? "DTR" :
			ch1 == IF_RTS ? "RTS" : "None");
#ifndef V25
		printf( "Databits = %d Stopbits = %d %s parity\n",
			(serp->lcr & LCR_8BITS) + 5,
			serp->lcr & LCR_NSB ? 2 : 1,
			!(serp->lcr & LCR_PEN) ? "No" :
			par == 0 ? "Odd" :
			par == LCR_EPS ? "Even" :
			par == LCR_SP ? "Mark" : "Space");
#endif
		if (serp->palt)
			printf("Linked to: %s\n",((IFACE *)serp->palt)->name);
#ifndef V25
		if (serp->pmpar)
			dial_status(serp->pmpar);
#endif
	}
#endif
	return speedflag ? asy_speed(ifp,atoi(argv[0])) : 0;
}

/* Set asynch line speed */
#pragma argsused
int asy_speed(IFACE *ifp,unsigned int speed)
{
#ifndef V25
	register unsigned base;
	register unsigned divisor;
	SER *serp = ifp->dcb;
	int dev = ifp->dev;
	unsigned long lspeed = (speed == 49664u) ? 115200l : speed;

	if(speed == 0 || dev >= nser)
		return -1;

	base = serp->addr;

	divisor = (unsigned)(BAUDCLK / lspeed);
	serp->speed = BAUDCLK / (unsigned long)divisor;

	idisable();

	/* Purge the receive data buffer */
	(void)inportb(base+RBR);
	if (serp->flags & FL_IS16550A)
		outportb(base+FCR,(char) FIFO_SETUP);

	/* Turn on divisor latch access bit */
	setbit(base+LCR,LCR_DLAB);

	/* Load the two bytes of the register */
	outportb(base+DLL,(char)(divisor & 0xff));		/* Low byte */
	outportb(base+DLM,(char)((divisor >> 8) & 0xff));	/* Hi byte */

	/* Turn off divisor latch access bit */
	outportb(base+LCR,serp->lcr);

	serp->datamask = (0x1f << (serp->lcr & LCR_8BITS)) | 0xf;
	popf();
#endif
	return 0;
}

// Send a buffer to serial transmitter (cnt != 0)
// Ignore 'buf' for slip(_lapb)
// Returns cnt if started, 0 if unable to send
//				or
// Poll serial transmit (cnt == 0)
// Returns 1 if ready, 0 if not ready to send
asy_xmit(IFACE *ifp,BYTE *buf,WORD cnt)
{
	SER *serp = ifp->dcb;

	if (serp->flags & FL_DISABLED)
		return 0;
#ifndef V25
	if (serp->flags & FL_MODEM) {
		MODEMPAR *pm;
		char rbuf[82];
		char *ptr,*ptr1;
		int rcnt,rcnt1,connecting;
		char return_now = 0;

		pm = serp->pmpar;
//#define SHOW_MODEM
#ifdef SHOW_MODEM
		{
			static char cLastIn,cLastOut;
			char cIn,cOut;

			cIn = inportb(serp->addr + MSR) & (MSR_RLSD|MSR_CTS|MSR_DSR);
			cOut = serp->mcr & (MCR_DTR|MCR_RTS);
			if (cIn != cLastIn || cOut != cLastOut) {
				printf("\n%s%s%s%s%s\n",cIn & MSR_RLSD ? " DCD" : "    ",
						cIn & MSR_CTS ? " CTS" : "    ",
						cIn & MSR_DSR ? " DSR" : "    ",
						cOut & MCR_DTR ? " DTR" : "    ",
						cOut & MCR_RTS ? " RTS" : "");
				cLastIn = cIn;
				cLastOut = cOut;
			}
		}
#endif
		if (pm->flags & MF_DCDLOST) {
			pm->flags &= ~MF_DCDLOST;
			if (pm->flags & MF_DISR)
				printf("\1%s:Carrier lost\2",ifp->name);
#ifdef PPP_S
			if (ifp->flags & FL_PPP) {
				if (((PPP *)ifp->prot_ext)->flags & PPP_CBVCALL) {
					pm->flags |= MF_DIALNOW;
				}
				else
					ppp_iostatus(ifp,PARAM_DOWN,0);
			}
#endif
			release_sif(ifp);
			asy_disconnect(serp);
		}
		if (!(serp->flags & FL_CONNECTED)) {
			rcnt = 80;
			if (serp->icnt && (pm->tmstate == TM_WRCV ||
						pm->tmstate == TM_WXID || pm->tmstate == TM_WADD)) {
				ptr = rbuf;
				rcnt = rcnt1 = asy_peek(ifp,rbuf,80);
				if (pm->tmstate == TM_WXID) {
					if (!(pm->flags & MF_ACKSEEN)) {
						if ((ptr = memchr(rbuf,6,rcnt)) != 0) {
							rcnt = ++ptr - rbuf;
							rcnt1 -= rcnt;
							pm->flags |= MF_ACKSEEN;
						}
					}
					if (pm->flags & MF_ACKSEEN) {
						if ((ptr1 = memchr(ptr,'\r',rcnt1)) != 0) {
							rcnt = ptr1 - rbuf;
							*ptr1 = 0;
							pm->flags &= ~MF_ACKSEEN;
							if (find_mdd(serp->iface,ptr) <= 0) {
								if (pm->retry && --pm->retry) {
									modem_command(serp,pm,pm->answer);
								}
								else
									asy_disconnect(serp);
							}
							else {
								stop_timer(&pm->timer);
								pm->tmstate = TM_IDLE;
								modem_command(serp,pm,pm->nextptr);
							}
							return_now = 1;
							goto flush;
						}
						else rcnt = ptr - rbuf;	// dont use any of XID
					}
				}
				else if (pm->tmstate == TM_WRCV) {	// waiting for match
					ptr1 = ptr;
					while ((ptr = memchr(ptr1,*serp->data,rcnt1)) != 0) {
						rcnt = ptr - rbuf;
						rcnt1 -= (ptr1 - ptr);
						ptr1 = ptr;
						if (rcnt1 >= serp->cnt) {
							if (memcmp(ptr,serp->data,serp->cnt) == 0) {
								stop_timer(&pm->timer);
								pm->tmstate = TM_IDLE;
								modem_command(serp,pm,pm->nextptr);
								rcnt = ptr + serp->cnt - rbuf;
								return_now = 1;
								goto flush;
							}
							else {
								++ptr1;
								--rcnt1;
							}
						}
						else
							break;
					}
				}
				else {		// waiting for IP address
					while (rcnt1 && !isdigit(*ptr)) {
						--rcnt1;
						++ptr;
					}
					if (rcnt1) {
						unsigned int cnt,dotcnt;
						unsigned int ipnum[4];
						ipnum[0] = ipnum[1] = ipnum[2] = ipnum[3] = 0xffff;
						rcnt = ptr - rbuf;
						for (ptr1 = ptr,cnt = rcnt1,dotcnt = 0;cnt;
															--cnt,++ptr1) {
							if (*ptr1 == '.') {
								if (++dotcnt > 3) {
									cnt = 1;
									break;
								}
							}
							else if (isdigit(*ptr1)) {
								if (ipnum[dotcnt] == 0xffff)
									ipnum[dotcnt] = 0;
								ipnum[dotcnt] = ipnum[dotcnt] * 10 + (*ptr1 - '0');
							}
							else
								break;
						}
						if (cnt && !isdigit(*ptr1) && dotcnt >= 3 &&
										ipnum[0] < 256 && ipnum[1] < 256 &&
										ipnum[2] < 256 && ipnum[3] < 256) {
							stop_timer(&pm->timer);
							pm->tmstate = TM_IDLE;
							*ptr1 = 0;
							set_ip_addr(ptr,-1);
							if (def_ip_addr) {
								if (ifp->sflags & SF_UNC)
									ifp = (IFACE *)serp->palt;
								ifp->ip_addr = def_ip_addr;
								ifp->netbits = def_netbits;
							}
							else
								printf(badhost,ptr);
							modem_command(serp,pm,pm->nextptr);
							rcnt = ptr1 - rbuf;
							return_now = 1;
							goto flush;
						}
						else if (cnt)
							rcnt = ptr1 - rbuf;	// flush invalid
					}
				}
			}
			flush:
			if (serp->istate >= IS_MODEM && serp->icnt && rcnt) {
				rcnt = asy_srecv(ifp,rbuf,rcnt);
				rbuf[rcnt] = 0;
				if (pm->flags & MF_DISR) {
					//printf("\1%s\2",rbuf);
					printf("\1%s\2",rbuf);
				}
			}
			if (return_now)
				return 0;
			if (pm->tmstate == TM_WDCD) {
				if (inportb(serp->addr + MSR) & MSR_RLSD) {
					if (pm->flags & MF_DISR)
						printf("\1%s:CD\2\",ifp->name);
					stop_timer(&pm->timer);
					pm->tmstate = TM_IDLE;
					modem_command(serp,pm,pm->nextptr);
				}
				return 0;
			}
			if (serp->flags & FL_EVENT) {
				serp->flags &= ~FL_EVENT;
				if (pm->nextptr && *pm->nextptr) {
					modem_command(serp,pm,pm->nextptr);
					return 0;
				}
				if (pm->state == MODEM_DIAL) {
					pm->state = MODEM_CONNECT;
					modem_command(serp,pm,pm->connect);
					return 0;
				}
				if (pm->state != MODEM_ANSWER && pm->state != MODEM_CONNECT) {
					if (pm->flags & MF_DIAL) {
						goto dial;
					}
					else
						pm->state = MODEM_IDLE;
					return 0;
				}
				if (pm->flags & MF_DISR)
					printf("\1%s:Connected\2\",ifp->name);
				serp->flags |= FL_CONNECTED;
				connecting = (pm->state == MODEM_CONNECT) ||
													(pm->flags & MF_DIALNOW);
				pm->flags &= ~MF_DIALNOW;
				pm->state = MODEM_IDLE;
				if (!(ifp->sflags & SF_RAW)) {
					if (serp->buf) {
						free(serp->buf);
						serp->buf = 0;
					}
					serp->icnt = 0;
				}
				serp->istate =
					(ifp->flags & PPP_FLAG) ? IS_PPP :
					(ifp->sflags & SF_RAW) ? IS_RAW :
					(ifp->sflags & SF_LAPB) ? IS_DATAB : IS_DATA;
#ifdef PPP_S
				if (ifp->flags & FL_PPP)
					ppp_iostatus(ifp,PARAM_UP,connecting);
#endif
				if ((ifp->sflags & SF_RAW) || !serp->tbp) {
					serp->ostate = OS_IDLE;
					return 0;
				}
				serp->ostate = OS_FSTART;
				serp->data = serp->tbp->data;
				serp->cnt = serp->tbp->cnt;
				return 0;
			}
			if (pm->state != MODEM_IDLE)
				return 0;
			else if (inportb(serp->addr + MSR) & MSR_RLSD) {
				if (pm->flags & MF_DISR)
					printf("\1%s:Incoming CD\2\",ifp->name);
				pm->retry = pm->retry_val;
				modem_command(serp,pm,pm->answer);
				pm->state = MODEM_ANSWER;
#ifdef PPP_S
				if (ifp->flags & PPP_FLAG) {
					((PPP *)ifp->prot_ext)->flags &= ~(PPP_CBV | PPP_CBVCALL);
				}
#endif
				return 0;
			}
			else if (pm->fdial[0] &&
							(((cnt || serp->tbp) && !(pm->flags & MF_NODOD)) ||
							(pm->flags & (MF_DIAL | MF_DIALNOW)))) {
				dial:
				pm->flags &= ~MF_DIALNOW;
				pm->state = MODEM_DIAL;
				pm->retry = pm->retry_val;
				modem_command(serp,pm,pm->dialcom);
			}
			return 0;
		}
		else if (!(inportb(serp->addr + MSR) & MSR_RLSD)) {
			serp->flags &= ~FL_CONNECTED;
			pm->flags |= MF_DCDLOST;
			return 0;
		}
	}
#endif
	idisable();
	if (serp->ostate == OS_IDLE) {
		if (cnt == 0) {
			popf();
			return 1;
		}
		if (ifp->sflags & SF_RAW) {
			serp->ostate = OS_RAW;
		}
		else {
			serp->ostate = OS_FSTART;
			buf = serp->tbp->data;
			cnt = serp->tbp->cnt;
		}
		serp->data = buf;
		serp->cnt = cnt;
	}
	else
		cnt = 0;
	popf();
	if ((serp->flags & FL_OFSTOPPED) && of_stopped(serp)) {
		return cnt;
	}
	if (cnt) {
		asy_start_xmit(serp);
	}
	return cnt;
}

/* Receive characters from asynch line
 * Returns count of characters read
 */
int asy_srecv(IFACE *ifp,char *buf,WORD cnt)
{
	unsigned tot,n;
	SER *serp = ifp->dcb;

	tot = 0;
	/* Read from serial I/O input buffer */
	idisable();
	for(;;){
		n = min(cnt,serp->icnt);
		if(n == 0)
			break;
		n = min(n,&serp->buf[serp->bufsize] - serp->rp);
		memcpy(buf,serp->rp,n);
		serp->rp += n;
		if(serp->rp >= &serp->buf[serp->bufsize])
			serp->rp = serp->buf;
		serp->icnt -= n;
		buf += n;
		tot += n;
		cnt -= n;
		if (!(serp->flags & FL_DISABLED) && (serp->flags & FL_IFSTOPPED)
													&& serp->icnt <= serp->start_cnt) {
			serp->flags &= ~FL_IFSTOPPED;
			if (serp->iflow == IF_XON) {
				serp->flags |= FL_SEND_XON;
				asy_start_xmit(serp);
			}
#ifndef V25
			else if (serp->iflow == IF_DTR) {
				outportb(serp->addr+MCR,serp->mcr |= MCR_DTR);
			}
			else if (serp->iflow == IF_RTS) {
				outportb(serp->addr+MCR,serp->mcr |= MCR_RTS);
			}
			else if (serp->iflow == (IF_DTR | IF_INV)) {
				outportb(serp->addr+MCR,serp->mcr &= ~MCR_DTR);
			}
			else if (serp->iflow == (IF_RTS | IF_INV)) {
				outportb(serp->addr+MCR,serp->mcr &= ~MCR_RTS);
			}
#endif
		}
	}
	popf();
	return (int)tot;
}

int asy_peek(IFACE *ifp,char *buf,WORD cnt)
{
	unsigned tot,n,icnt;
	SER *serp = ifp->dcb;
	char *rp;

	tot = 0;
	//Read from serial I/O input buffer
	idisable();
	icnt = serp->icnt;
	rp = serp->rp;
	for(;;){
		n = min(cnt,icnt);
		if(n == 0)
			break;
		n = min(n,&serp->buf[serp->bufsize] - rp);
		memcpy(buf,rp,n);
		rp += n;
		if(rp >= &serp->buf[serp->bufsize])
			rp = serp->buf;
		icnt -= n;
		buf += n;
		tot += n;
		cnt -= n;
	}
	popf();
	return (int)tot;
}

#ifndef V25
/* Interrupt handler for 8250 asynch chip */
void near asyint(SER *serpf)
{
	void near pdr_sched(void);
	register char iir,irseen = 0;
	int base;
	char stat;
	struct mbuf *bp;
	IFACE *ifp;
	SER *serp = serpf;

	while (1) {
		if (!(serp->flags & FL_DISABLED)) {
			base = serp->addr;
			while (((iir = inportb(base + IIR)) & IIR_IP) == 0) {
				irseen = 1;
				switch(iir & IIR_ID){
				case IIR_RDA:	/* Receiver interrupt */
					asyrxint(serp);
					break;
				case IIR_THRE:	/* Transmit interrupt */
					asytxint(serp,1);
					break;
				case IIR_RLS:	/* error interrupt */
					ifp = serp->iface;
					inportb(base + RBR);	/* read data byte */
					if ((stat = inportb(base + LSR)) & LSR_OE)
						++serp->overrun;
					if (stat & LSR_PE)
						++serp->parerr;
					if (stat & LSR_FE)
						++serp->frmerr;
					if (stat & LSR_BI)
						++serp->brkerr;
					if (!(ifp->sflags & (SF_TELNET | SF_FTP))) {
						/* force a slip error */
						if ((bp = serp->rbp) != NULLBUF && serp->istate != IS_PPP)
							bp->cnt = 0;
						/*
						if (serp->istate == IS_PPP || serp->istate == IS_PPPESC)
							serp->istate = IS_PPPIGN;
						*/
						if (serp->istate == IS_DATAB || serp->istate == IS_ESCB)
							serp->istate = IS_IGNB;
						else if (serp->istate == IS_DATA || serp->istate == IS_ESC)
							serp->istate = IS_IGN;
					}
					break;
				}
			}
			if (inportb(base + LSR) & LSR_THRE) {
				asytxint(serp,1);
			}
		}
		if ((serp = serp->next) == 0) {
			if (irseen) {
				serp = serpf;	// try all ports again
				irseen = 0;
			}
			else
				break;
		}
	}
	if (!(serpf->flags & FL_SUB_ISR)) {
		if (serpf->vec >= 8)
			outp(0xa0,0x20);	/* clear slave interrupt controller */
		outp(0x20,0x20);		/* clear interrupt controller */
	}
	else
		(*serpf->oldvec)();
#if !APL
	if (cAsySched) {
		cAsySched = 0;
		pdr_sched();
	}
#endif
}
#endif

/* Process 8250/V25 receiver interrupts */
static void near asyrxint(SER *serp)
{
	struct mbuf *bp;
	BYTE c;
#ifndef V25
	IFACE *ifp;
	unsigned base = serp->addr;
#else
	BYTE bStatus;
#endif

	while (
#ifndef V25
			inportb(base+LSR) & LSR_DR) {
		c = inportb(base+RBR);
#else
			(bStatus= *serp->SCE) & VS_RR) {
		c = *serp->RxB;
		if (bStatus & VS_ERR) {
			// just keep statistics
			if (bStatus & VS_PE)
				++serp->parerr;
			if (bStatus & VS_FE)
				++serp->frmerr;
			if (bStatus & VS_OE)
				++serp->overrun;
		}
#endif
		serp->rcvcount++;
#ifdef SER_TEST_R
		put_rbuf(c);
#endif
		switch (serp->istate) {
#ifndef V25
		case IS_MODEM:
			if (serp->buf == 0)
				break;
		case IS_RAW:
			ifp = serp->iface;
			c &= serp->datamask;
			if (ifp->sflags & SF_FTP) {
				serp->deliver &= ~DEL_TO;
				ifp->sflags &= ~SF_DELIVER;
				start_timer(&serp->timer);
			}
			if (serp->oflow == OF_XON) {
				if (c == XON) {
					serp->flags &= ~FL_OFSTOPPED;
					if (serp->ostate != OS_IDLE) {
						asy_start_xmit(serp);
					}
					if (!(ifp->sflags & SF_PASSFC))
						continue;
				}
				else if (c == XOFF) {
					serp->flags |= FL_OFSTOPPED;
					if (!(ifp->sflags & SF_PASSFC))
						continue;
				}
			}
			/* Process incoming data;
			 * If buffer is full, we have no choice but
			 * to drop the character
			 */
			if (serp->icnt < serp->bufsize) {
				*serp->wp++ = c;
				if(serp->wp >= &serp->buf[serp->bufsize]) {
					/* Wrap around */
					serp->wp = serp->buf;
				}
				serp->icnt++;
			}
			else
				++serp->buferr;
			if (serp->icnt >= serp->stop_cnt && serp->iflow &&
					!(serp->flags & FL_IFSTOPPED)) {
				serp->flags |= FL_IFSTOPPED;
				if (serp->iflow == IF_XON) {
					serp->flags |= FL_SEND_XOFF;
					asy_start_xmit(serp);
				}
				else if (serp->iflow == IF_DTR) {
					outportb(serp->addr + MCR,serp->mcr &= ~MCR_DTR);
				}
				else if (serp->iflow == IF_RTS) {
					outportb(serp->addr + MCR,serp->mcr &= ~MCR_RTS);
				}
				else if (serp->iflow == (IF_DTR | IF_INV)) {
					outportb(serp->addr + MCR,serp->mcr |= MCR_DTR);
				}
				else if (serp->iflow == (IF_RTS | IF_INV)) {
					outportb(serp->addr + MCR,serp->mcr |= MCR_RTS);
				}
			}
			break;
#endif
		case IS_DATA:
			if ((bp = serp->rbp) == 0 && (bp = serp->rbp =
					alloc_mbuf(serp->bufsize,AllocAsyIbuf)) == 0) {
				++serp->buferr;
				break;
			}
			if (c == FR_END) {
				if (bp->cnt) {
					enqueue(&serp->rcvq,bp);
					serp->rbp = alloc_mbuf(serp->bufsize,AllocAsyIbuf1);
					cAsySched = 1;
				}
			}
			else if (c != FR_ESC) {
				goto storc;
			}
			else
				serp->istate = IS_ESC;
			break;
		case IS_ESC:
			if ((bp = serp->rbp) == 0 && (bp = serp->rbp =
					alloc_mbuf(serp->bufsize,AllocAsyIbuf2)) == 0) {
				++serp->buferr;
				break;
			}
			if (c == T_FR_END)
				c = FR_END;
			else if (c == T_FR_ESC)
				c = FR_ESC;
			else {
				goto setign;
			}
			serp->istate = IS_DATA;
			storc:
			if (bp->cnt >= serp->bufsize) {
				setign:
				serp->errors++;
				serp->istate = IS_IGN;
				bp->cnt = 0;
			}
			else {
				bp->data[bp->cnt++] = c;
			}
			break;
		case IS_IGN:
			if (c == FR_END)
				serp->istate = IS_DATA;
			break;
#ifdef PPP_S
		/*
		case IS_PPPIGN:
			if (c == HDLC_FLAG)
				serp->istate = IS_PPP;
			break;
		*/
		case IS_PPP:
			if (c == HDLC_FLAG) {
				if ((bp = serp->rbp) != 0 && bp->cnt) {
					enqueue(&serp->rcvq,bp);
					serp->rbp = alloc_mbuf(serp->bufsize,AllocAsyIbuf3);
					cAsySched = 1;
				}
				break;
			}
			if ((bp = serp->rbp) == 0 && (bp = serp->rbp =
					alloc_mbuf(serp->bufsize,AllocAsyIbuf4)) == 0) {
				++serp->buferr;
				break;
			}
			if (bp->cnt < serp->bufsize) {
				bp->data[bp->cnt++] = c;
			}
			break;
#endif
#ifdef LAPB
		case IS_DATAB:
			if ((bp = serp->rbp) == 0 && (bp = serp->rbp =
					alloc_mbuf(serp->bufsize,AllocAsyDatab)) == 0) {
				++serp->buferr;
				break;
			}
			if (c == FR_END) {
				if (bp->cnt) {
					if (!serp->accum) {
						bp->cnt -= 2;
						enqueue(&serp->rcvq,bp);
						serp->rbp =
							alloc_mbuf(serp->bufsize,AllocAsyDatab1);
						cAsySched = 1;
					}
					else {
						bp->cnt = 0;
						++serp->errors;
					}
					serp->accum = 0xffff;
				}
			}
			else if (c != FR_ESC) {
				goto storcb;
			}
			else
				serp->istate = IS_ESCB;
			break;
		case IS_ESCB:
			if ((bp = serp->rbp) == 0 && (bp = serp->rbp =
					alloc_mbuf(serp->bufsize,AllocAsyDatab2)) == 0) {
				++serp->buferr;
				break;
			}
			if (c == T_FR_END)
				c = FR_END;
			else if (c == T_FR_ESC)
				c = FR_ESC;
			else {
				goto setignb;
			}
			serp->istate = IS_DATAB;
			storcb:
			serp->accum = (serp->accum >> 8) ^
					crctab[(serp->accum ^ c) & 0xff];
			if (bp->cnt >= serp->bufsize) {
				setignb:
				serp->errors++;
				serp->istate = IS_IGNB;
				bp->cnt = 0;
				serp->accum = 0xffff;
			}
			else {
				bp->data[bp->cnt++] = c;
			}
			break;
		case IS_IGNB:
			if (c == FR_END)
				serp->istate = IS_DATAB;
			break;
#endif
		}
	}
}

// start transmit by faking a transmit interrupt
void near asy_start_xmit(SER *serp)
{
	idisable();
#ifndef V25
	setbit(serp->addr + IER,IER_TxE);
#else
	*serp->SCM |= 0x80;
#endif
	asytxint(serp,0);
	popf();
}

/* Handle 8250 transmitter interrupts */
static void near asytxint(SER *serp,BYTE bIsInt)
{
	BYTE ch,bExit = 0;
	struct mbuf *bp;
#ifndef V25
	MODEMPAR *pm;
	TIMER *ptim;
	unsigned base = serp->addr;
#endif

	while(!bExit && (bIsInt ||
#ifndef V25
						(inportb(base+LSR) & LSR_THRE))) {
#else
						(*serp->SCE & VS_TE))) {
		*serp->SCM &= 0x7f;	// disable transmit
#endif


		bIsInt = 0;
		if (serp->flags & FL_SEND_XOFF) {
			serp->flags &= ~FL_SEND_XOFF;
			ch = XOFF;
			goto send_it;
		}
		if (serp->flags & FL_SEND_XON) {
			serp->flags &= ~FL_SEND_XON;
			ch = XON;
			goto send_it;
		}
		if (serp->oflow && of_stopped(serp)) {
			goto clear_int;
		}
#ifndef V25
		if ((serp->flags & (FL_MODEM | FL_CONNECTED)) ==
				(FL_MODEM | FL_CONNECTED) &&
				!(inportb(serp->addr + MSR) & MSR_RLSD)) {
			serp->flags &= ~FL_CONNECTED;
			pm = serp->pmpar;
			pm->flags |= MF_DCDLOST;
			goto clear_int;
		}
#endif
		switch (serp->ostate) {
#ifndef V25
		case OS_RAW:
			ch = *serp->data++;
			if(--serp->cnt == 0){
				// set state to OS_IDLE, send ch and exit
				goto exit;
			}
			break;
		case OS_MODEM:
			pm = serp->pmpar;
			if (pm->tmstate != TM_IDLE || !serp->cnt)
				goto clear_int;
			ch = *serp->data++;
			--serp->cnt;
			if (pm->modem_char_time) {
				ptim = &pm->timer;
				pm->tmstate = TM_PACE;
				set_timer(ptim,pm->modem_char_time);
				start_timer(ptim);
			}
			else if (!serp->cnt) {
				serp->flags |= FL_EVENT;
				// set state to OS_IDLE, send ch and exit
				goto exit;
			}
			break;
#endif
		case OS_FSTART:
			ch = (serp->flags & FL_PPP) ? HDLC_FLAG : FR_END;
			serp->ostate = OS_DATA;
			serp->oaccum = HDLC_FCS_START;	// for PPP
			goto finddata;
		case OS_DATA:
			ch = *serp->data++;
#ifdef LAPB
			if (serp->flags & FL_LAPB) {
				serp->oaccum = (serp->oaccum >> 8) ^
					crctab[(serp->oaccum ^ ch) & 0xff];
			}
#endif
#ifdef PPP_S
			if (serp->flags & FL_PPP) {
				serp->oaccum = (serp->oaccum >> 8) ^
					crctab[(serp->oaccum ^ ch) & 0xff];
				if ( ((ch < SP_CHAR) && (serp->accm_o & (1L << ch))) ||
						(ch == HDLC_ESC_ASYNC) || (ch == HDLC_FLAG)) {
					serp->last_o_ch = ch;
					ch = HDLC_ESC_ASYNC;
					serp->ostate = OS_HDLC_EA;
					break;
				}
			}
			else
#endif
			{
				if (ch == FR_ESC) {
					serp->ostate = OS_FRESC;
					break;
				}
				else if (ch == FR_END) {
					serp->ostate = OS_FREND;
					ch = FR_ESC;
					break;
				}
			}
			goto testend;
#ifdef PPP_S
		case OS_HDLC_EA:
			ch = (serp->last_o_ch ^ HDLC_ESC_COMPL);
			serp->ostate = OS_DATA;
			goto testend;
		case OS_FCS1:
			// Final FCS calculation
			serp->oaccum ^= 0xffff;
			ch = serp->oaccum;	// Least significant byte first
			if ( ((ch < SP_CHAR) && (serp->accm_o & (1L << ch))) ||
					(ch == HDLC_ESC_ASYNC) || (ch == HDLC_FLAG)) {
				serp->last_o_ch = ch;
				ch = HDLC_ESC_ASYNC;
				serp->ostate = OS_FCS1ESC;
			}
			else
				serp->ostate = OS_FCS2;
			break;
		case OS_FCS1ESC:
			ch = (serp->last_o_ch ^ HDLC_ESC_COMPL);
			serp->ostate = OS_FCS2;
			break;
		case OS_FCS2:
			ch = (serp->oaccum >> 8);	// Most significant byte first
			if ( ((ch < SP_CHAR) && (serp->accm_o & (1L << ch))) ||
					(ch == HDLC_ESC_ASYNC) || (ch == HDLC_FLAG)) {
				serp->last_o_ch = ch;
				ch = HDLC_ESC_ASYNC;
				serp->ostate = OS_FCS2ESC;
			}
			else
				serp->ostate = OS_END;
			break;
		case OS_FCS2ESC:
			ch = (serp->last_o_ch ^ HDLC_ESC_COMPL);
			serp->ostate = OS_END;
			break;
#endif
		case OS_FRESC:
			ch = T_FR_ESC;
			serp->ostate = OS_DATA;
			goto testend;
		case OS_FREND:
			ch = T_FR_END;
			serp->ostate = OS_DATA;
			testend:
			if (--serp->cnt == 0) {
				finddata:
				while (!serp->cnt) {
					if ((bp = serp->tbp = free_mbuf(serp->tbp)) == 0) {
						serp->ostate = (serp->flags & FL_PPP) ? OS_FCS1 :
										(serp->flags & FL_LAPB)?	OS_CRC1 :
																OS_END;
						break;
					}
					else {
						serp->cnt = bp->cnt;
						serp->data = bp->data;
					}
				}
			}
			break;
#ifdef LAPB
		case OS_CRC1:
			if ((ch = serp->oaccum) == FR_ESC)
				serp->ostate = OS_CRC1ESC;
			else if (ch == FR_END) {
				serp->ostate = OS_CRC1END;
				ch = FR_ESC;
			}
			else
				serp->ostate = OS_CRC2;
			break;
		case OS_CRC1ESC:
			ch = T_FR_ESC;
			serp->ostate = OS_CRC2;
			break;
		case OS_CRC1END:
			ch = T_FR_END;
			serp->ostate = OS_CRC2;
			break;
		case OS_CRC2:
			if ((ch = serp->oaccum >> 8) == FR_ESC)
				serp->ostate = OS_CRC2ESC;
			else if (ch == FR_END) {
				serp->ostate = OS_CRC2END;
				ch = FR_ESC;
			}
			else
				serp->ostate = OS_END;
			break;
		case OS_CRC2ESC:
			ch = T_FR_ESC;
			serp->ostate = OS_END;
			break;
		case OS_CRC2END:
			ch = T_FR_END;
			serp->ostate = OS_END;
			break;
#endif
		case OS_END:
			ch = serp->flags & FL_PPP ? HDLC_FLAG : FR_END;
			while (!serp->cnt) {
				if (serp->sndq == NULLBUF || (bp = serp->tbp =
						 dequeue(&serp->sndq)) == 0) {
					cAsySched = 1;
#ifndef V25
					exit:
#endif
					serp->ostate = OS_IDLE;
					bExit = 1;
					goto send_it;
				}
				serp->accm_o = *(DWORD *)serp->tbp->data;
				serp->data = (serp->tbp->data += 4);
				serp->cnt = (serp->tbp->cnt -= 4);
			}
			serp->ostate = OS_FSTART;
			break;
		default:
			clear_int:
#ifndef V25
			clrbit(base+IER,IER_TxE);
#endif
			return;
		}
		send_it:
#ifdef SER_TEST_X
		put_xbuf(ch);
#endif
#ifndef V25
		outportb(base + THR,ch);
#else
		*serp->TxB = ch;
		*serp->SCM |= 0x80;
#endif
		++serp->xmtcount;
	}
}

static int near of_stopped(SER *serp)
{
	char flow  = serp->oflow & ~OF_INV;
#ifndef V25
	char inverse = serp->oflow & OF_INV;
	char status;
#endif

	if (serp->flags & FL_OFSTOPPED) {
		switch (flow) {
		case OF_XON:
			break;
#ifndef V25
		case OF_DCD:
			if (((status = inportb(serp->addr + MSR) & MSR_RLSD)
				!= 0 && !inverse) || (!status && inverse))
				goto clearof;
			break;
		case OF_DSR:
			if (((status = inportb(serp->addr + MSR) & MSR_DSR)
				!= 0 && !inverse) || (!status && inverse))
				goto clearof;
			break;
		case OF_CTS:
			if (((status = inportb(serp->addr + MSR) & MSR_CTS)
				!= 0 && !inverse) || (!status && inverse))
				goto clearof;
			break;
#endif
		default:
#ifndef V25
		clearof:
			if (serp->ostate != OS_IDLE)
				setbit(serp->addr+IER,IER_TxE);
#endif
			serp->flags &= ~FL_OFSTOPPED;
			return 0;
		}
		return 1;
	}
#ifndef V25
	switch (flow) {
	case OF_DCD:
		if (((status = inportb(serp->addr + MSR) & MSR_RLSD) == 0
			&& inverse) || (!status && !inverse))
			goto setof;
		break;
	case OF_DSR:
		if (((status = inportb(serp->addr + MSR) & MSR_DSR) == 0
			&& inverse) || (!status && !inverse))
			goto setof;
		break;
	case OF_CTS:
		if (((status = inportb(serp->addr + MSR) & MSR_CTS) == 0
			&& inverse) || (!status && !inverse)) {
			setof:
			serp->flags |= FL_OFSTOPPED;
			return 1;
		}
		break;
	}
#endif
	return 0;
}

/* Set bit(s) in I/O port */

void setbit(unsigned port,char bits)
{
	outportb(port,(char)inportb(port)|bits);
}

/* Clear bit(s) in I/O port */
void clrbit(unsigned port,char bits)
{
	outportb(port,(char)(inportb(port) & ~bits));
}

SER *ser[SER_INT_MAX];

#ifndef V25
void interrupt asy0vec(void)
{
	asyint(ser[0]);
}

void interrupt asy1vec(void)
{
	asyint(ser[1]);
}

void interrupt asy2vec(void)
{
	asyint(ser[2]);
}

void interrupt asy3vec(void)
{
	asyint(ser[3]);
}

void interrupt asy4vec(void)
{
	asyint(ser[4]);
}

void interrupt asy5vec(void)
{
	asyint(ser[5]);
}
#else

#define FINT asm	db 0Fh;asm	db 92h

static void near ClearV25Interrupt(SER *serp)
{
	if (!serp->oldvec) {
		FINT;
	}
	else
		(*serp->oldvec)();
}

void interrupt asy0vec(void)
{
	asyrxint(ser[0]);
	ClearV25Interrupt(ser[0]);
}

void interrupt asy1vec(void)
{
	asytxint(ser[0],1);
	ClearV25Interrupt(ser[0]);
}

void interrupt asy2vec(void)
{
	asyrxint(ser[1]);
	ClearV25Interrupt(ser[1]);
}

void interrupt asy3vec(void)
{
	asytxint(ser[1],1);
	ClearV25Interrupt(ser[1]);
}

void interrupt asy4vec(void)
{
	asyrxint(ser[2]);
	ClearV25Interrupt(ser[2]);
}

void interrupt asy5vec(void)
{
	asytxint(ser[2],1);
	ClearV25Interrupt(ser[2]);
}
#endif
void interrupt (*asyvecs[])() =
		{asy0vec,asy1vec,asy2vec,asy3vec,asy4vec,asy5vec};

