//
//   $Id: BPRINTF.C 1.3 2003/07/17 16:43:36Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*------------------------------------------------------------------------
 * filename - bprintf.c
 *
 * function(s)
 *        __bputn - character output function
 *        biosprintf - sends formatted output to the console
 *-----------------------------------------------------------------------*/

#pragma option -zC_TEXT

#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include "_printf.h"

/*---------------------------------------------------------------------*

Name            __bputn - character output function

Description     prints a string of n chars on the screen

Return value    returns last character printed

Note            template must match putnF typedef from _printf.h

*------------------------------------------------------------------------*/
extern char quiet_flag;
extern char cReverseVideo;

static char normal = 0x07;
static char reverse = 0x70;
static unsigned char col = 0x07;

void near SetColor(void)
{
	if (cReverseVideo) {
		_CX = 1;
		_BX = col;
		_AX = 0x0920;
		geninterrupt(0x10);
	}
}

#pragma argsused
unsigned pascal near __bputn(const void *s, unsigned n, void *dontCare)
{
	unsigned char c = 0;

	while (n--) {
		c = *((char *)s)++;
		if (c == '\n') {
			_BX = 0;
			_AH = 0xe;
			_AL = '\r';
			geninterrupt(0x10);
		}
		if (c == '\t')
			c = ' ';
		if (c == 1)
			col = reverse;
		else if (c == 2) {
			col = normal;
			SetColor();
		}
		else {
			if (col == reverse && c >= 0x20)
				SetColor();
			_BX = 0;
			_AH = 0xe;
			_AL = c;
			geninterrupt(0x10);
		}
	}
	return c;
}

/*---------------------------------------------------------------------*

Name            biosprintf - sends formatted output to the console

Usage           int biosprintf(char *format [,argument, ...]);

Description     member of the printf family.  biosprintf send its output
				directly to the console.  It does not translate linefeed
				characters into CR/LF combinations.

Return value    success : the number of bytes output.
				error : EOF

*------------------------------------------------------------------------*/
int     cdecl biosprintf(const char *fmt, ...)
{
	static char not_first;

	if (!not_first) {
		not_first = 1;
		_BH = 0;
		_AH = 0x8;
		geninterrupt(0x10);
		normal = col = _AH;
		reverse = (col & 0x88) | ((col & 0x70) >> 4) | ((col & 0x7) << 4);
	}
	if (quiet_flag)
		return 0;
	return  __vprinter (__bputn, 0, fmt, (void _ss *) _va_ptr);
}
