//
//   $Id: CMDPARSE.C 1.6 2002/10/16 16:42:10Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// CMDPARSE.C
// Command parsing functions
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1991 Phil Karn, KA9Q

#if APL==0
#pragma option -zAINETINIT
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "inetdefs.h"
#include "cmdparse.h"

static void near print_err_line(int argc,char **argv)
{
	while (argc--)
		printf("%s ",*argv++);
	printf("\n");
}

int cmdparse(struct cmds far *cmds,char *line,char error)
{
	struct cmds far *cmdp;
	char *argv[NARG],*cp;
	int argc,qflag;
	int rslt,len;

	/* Remove cr/lf */
	rip(line);

	for (argc = 0;argc < NARG;argc++)
		argv[argc] = NULLCHAR;

	for (argc = 0;argc < NARG;) {
		qflag = 0;
		/* Skip leading white space */
		while (*line == ' ' || *line == '\t')
			line++;
		if (*line == '\0')
			break;
		/* Check for quoted token */
		if (*line == '"') {
			line++;	/* Suppress quote */
			qflag = 1;
		}
		argv[argc++] = line;	/* Beginning of token */
		/* Find terminating delimiter */
		if (qflag) {
			/* Find quote, it must be present */
			if ((line = strchr(line,'"')) == NULLCHAR) {
				return -1;
			}
			*line++ = '\0';
		} else {
			/* Find space or tab. If not present,
			 * then we've already found the last
			 * token.
			 */
			if ((cp = strchr(line,' ')) == NULLCHAR
			 && (cp = strchr(line,'\t')) == NULLCHAR) {
				break;
			}
			*cp++ = '\0';
			line = cp;
		}
	}
/*
	if (argc < 1) {		/* empty command line */
		argc = 1;
		argv[0] = "";
	}
*/
	/* Lines beginning with "#" are comments */
	if ( argc < 1 || argv[0] == NULLCHAR || argv[0][0] == '#')
		return 0;

	/* Look up command in table; prefix matches are OK */
	len = strlen(argv[0]);
	for (cmdp = cmds;cmdp->name[0] != 0;cmdp++) {
		if (_fstrnicmp(argv[0],cmdp->name,len) == 0) {
			rslt = (*cmdp->func)(argc,argv);
			if (rslt < 0 && error) {
				printf("Command failed: %.9Fs\n",cmdp->name);
				print_err_line(argc,argv);
			}
			for (len = 0;len < argc - 1;++len)
				argv[len][strlen(argv[len])] = ' ';
			return(rslt);
		}
	}
	if (error) {
		printf("Unknown command: %s "
#if APL
				"(use 'help' for list)"
#endif
				"\n",argv[0]);
		print_err_line(argc,argv);
	}
	return -1;
}

/* Call a subcommand based on the first token in an already-parsed line */
int subcmd(struct scmds far *tab,int argc,char *argv[],void *p)
{
	int rslt,len;
	struct scmds far *cmdp;

	/* Strip off first token and pass rest of line to subcommand */
	if (argc < 2) {
		printf("Option required: %s [ ",argv[0]);
		goto print_options;
	}
	argc--;
	argv++;
	len = strlen(argv[0]);
	for (cmdp = tab;cmdp->name[0] != 0;cmdp++) {
		if (_fstrnicmp(argv[0],cmdp->name,len) == 0) {
			rslt = (*cmdp->func)(argc,argv,p);
			if (rslt < 0)
				printf("Option failed: %s\n",argv[0]);
			return(rslt);
		}
	}
	printf("Unknown option: %s [ ",argv[0]);
	print_options:
	for (cmdp = tab;cmdp->name[0] != 0;cmdp++) {
		printf ("%.9Fs ",cmdp->name);
	}
	printf("]\n");
	return -1;
}

struct boolcmd {
	char *str;	/* Token */
	int val;	/* Value */
};

static struct boolcmd Boolcmds[] = {
	{"y",		1},	/* Synonyms for "true" */
	{"yes",	1},
	{"true",	1},
	{"on",	1},
	{"1",		1},
	{"set",	1},
	{"enable",1},

	{"n",		0},	/* Synonyms for "false" */
	{"no",	0},
	{"false",0},
	{"off",	0},
	{"0",		0},
	{"clear",0},
	{"disable",0},
	{NULL,0}
};
/* Subroutine for setting and displaying boolean flags */
int
setbool(var,label,argc,argv)
int *var;
char *label;
int argc;
char *argv[];
{
	struct boolcmd *bc;

	if(argc < 2){
		printf("%s: %s\n",label,*var ? "on":"off");
		return 0;
	}
	for(bc = Boolcmds;bc->str != NULL;bc++){
		if(strcmpi(argv[1],bc->str) == 0){
			*var = bc->val;
			return 0;
		}
	}
	printf("Valid options:");
	for(bc = Boolcmds;bc->str != NULL;bc++)
		printf(" %s",bc->str);

	printf("\n");
	return 1;
}


/* Subroutine for setting and displaying bit values */
int
bit16cmd(bits,mask,label,argc,argv)
WORD *bits;
WORD mask;
char *label;
int argc;
char *argv[];
{
	int doing = (*bits & mask);
	int result = setbool( &doing, label, argc, argv );

	if ( !result ) {
		if ( doing )
			*bits |= mask;
		else
			*bits &= ~mask;
	}
	return result;
}

int setint(int *var,char *label,int argc,char *argv[])
{
	if(argc < 2)
		printf("%s: %i\n",label,*var);
	else
		*var = atoi(argv[1]);
	return 0;
}
