//
//   $Id: ENET.C 1.4 2002/10/16 16:42:10Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// ENET.C
// Ethernet and token ring common functions
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

#include <stdio.h>
#include <string.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "enet.h"

BYTE ether_bdcst[] = {0xff,0xff,0xff,0xff,0xff,0xff};

/* Convert Ethernet header in host form to network mbuf */
struct mbuf *htonether(struct ether *ether)
{
	struct mbuf *bp;
	register BYTE *cp;

	if ((bp = alloc_mbuf(ETHERLEN,AllocEnetEth)) == 0)
		return bp;
	bp->cnt = ETHERLEN;
	cp = bp->data;

	memcpy(cp,ether->dest,EADDR_LEN);
	cp += EADDR_LEN;
	memcpy(cp,ether->source,EADDR_LEN);
	cp += EADDR_LEN;
	put16(cp,ether->type);

	return bp;
}
/* Convert ISO (SNAP) header in host form to network mbuf */
struct mbuf *htoniso(struct ether *ether,WORD len)
{
	struct mbuf *bp;
	register BYTE *cp;

	if ((bp = alloc_mbuf(IEEE_LEN,AllocEnetIeee)) == 0)
		return bp;
	bp->cnt = IEEE_LEN;
	cp = bp->data;

	memcpy(cp,ether->dest,EADDR_LEN);
	cp += EADDR_LEN;
	memcpy(cp,ether->source,EADDR_LEN);
	cp += EADDR_LEN;
	put16(cp,len + 8);
	cp[2] = cp[3] = IEEESAP;	/* dsap = ssap = 170 */
	cp[4] = UI;		/* UI */
	cp[5] = cp[6] = cp[7] = 0;
	put16(cp + 8,ether->type);
	return bp;
}

/* Extract Ethernet (DIX or IEEE header) */
ntohether(struct ether *ether,struct mbuf **bpp)
{
	BYTE dsap,ssap,control;

	pullup(bpp,ether->dest,EADDR_LEN);
	pullup(bpp,ether->source,EADDR_LEN);
	if ((ether->type = pull16(bpp)) <= GIANT) {
		dsap = pullchar(bpp);
		ssap = pullchar(bpp);
		control = pullchar(bpp);
		if (dsap == IEEESAP && ssap == IEEESAP && control == UI) {
			pullup(bpp,0,ID_LEN);
			ether->type = pull16(bpp);
		}
		else if (dsap == IPSAP && ssap == IPSAP && control == UI)
			ether->type = IP_TYPE;
	}
	return ETHERLEN;
}

/* Convert Token ring SNAP header in host form to network mbuf */
struct mbuf *htontoken(struct ether *ether,BYTE *rif)
{
	struct mbuf *bp;
	register BYTE *cp;
	int len = rif ? rif[0] & 0x1f : 0;

	if ((bp = alloc_mbuf(IEEE_LEN + len,AllocEnetToken)) == 0)
		return bp;
	bp->cnt = IEEE_LEN + len;
	cp = bp->data;

	*cp++ = 0x10;			// AC
	*cp++ = 0x40;			// FC
	memcpy(cp,ether->dest,EADDR_LEN);
	cp += EADDR_LEN;
	memcpy(cp,ether->source,EADDR_LEN);
	cp += EADDR_LEN;
	if (len) {
		memcpy(cp,rif,len);
		cp += len;
		bp->data[EADDR_LEN + 2] |= 0x80;
	}
	cp[0] = cp[1] = IEEESAP;	/* dsap = ssap = 170 */
	cp[2] = UI;		/* UI */
	cp[3] = cp[4] = cp[5] = 0;
	put16(cp + 6,ether->type);
	return bp;
}

/* Extract Token Ring (IEEE header) */
ntohtoken(struct ether *ether,struct mbuf **bpp,BYTE *rif)
{
	BYTE dsap,ssap,control;
	BYTE b_lth;

	pull16(bpp);				// ignore AC & FC
	pullup(bpp,ether->dest,EADDR_LEN);
	pullup(bpp,ether->source,EADDR_LEN);
	if (ether->source[0] & 0x80) {		// RIF
		ether->source[0] &= 0x7f;
		rif[0] = pullchar(bpp);
		if ((b_lth = rif[0] & 0x1f) > 30)
			b_lth = 30;	// something wrong, but at least don't crash
		pullup(bpp,rif + 1,b_lth - 1);	// read RIF
	}
	else
		rif[0] = 0;
	dsap = pullchar(bpp);
	ssap = pullchar(bpp);
	control = pullchar(bpp);
	if (dsap == IEEESAP && ssap == IEEESAP && control == UI) {
		pullup(bpp,0,ID_LEN);
		ether->type = pull16(bpp);
	}
	return ETHERLEN;
}

/* Format an Ethernet address into a printable ascii string */
void pether(char *out,BYTE *addr)
{
	sprintf(out,"%02x:%02x:%02x:%02x:%02x:%02x",
		uchar(addr[0]),uchar(addr[1]),
		uchar(addr[2]),uchar(addr[3]),
		uchar(addr[4]),uchar(addr[5]));
}

/* Convert an Ethernet address from Hex/ASCII to binary */
void gether(BYTE *out,char *cp)
{
	register int i;

	for(i=6; i!=0; i--){
		*out++ = htoi(cp);
		if((cp = strchr(cp,':')) == NULLCHAR)	/* Find delimiter */
			break;
		cp++;			/* and skip over it */
	}
}
/* Version of gether for ARP, matches passed args */
#pragma argsused
void gaether(BYTE *out,char *in[],int cnt)
{
	gether(out,in[0]);
}

