//
//   $Id: ICMPCMD.C 1.8 2003/07/01 19:59:04Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// ICMPCMD.C
// ICMP commands
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

#include <stdio.h>
#include <alloc.h>
#include <string.h>
#include <stdlib.h>
#include "inetdefs.h"
#include "iface.h"
#include "mbuf.h"
#include "icmp.h"
#include "netfunc.h"
#include "internet.h"
#include "timer.h"
#include "ping.h"
#include "ip.h"
#include "log.h"

#if APL
static void del_ping(struct ping *pp);
static WORD hash_ping(int32 dest);
static struct ping *add_ping(int32 dest);
#endif

extern char rc_inuse,altconf;

XPING *pxpfirst;	/* first entry in external ping queue */
XPING *pxplast;		/* last entry in external ping queue */

#if APL
int doicmpstat()
{
	extern struct icmp_errors icmp_errors;
	extern struct icmp_stats icmp_stats;
	extern char *icmptypes[];
	register int i;

	printf("ICMP: chksum err %u no space %u icmp %u bdcsts %u\n",
	 icmp_errors.checksum,icmp_errors.nospace,icmp_errors.noloop,
	 icmp_errors.bdcsts);
	printf("type  rcvd  sent\n");
	for(i=0;i<ICMP_TYPES;i++){
		if(icmp_stats.input[i] == 0 && icmp_stats.output[i] == 0)
			continue;
		printf("%-6u%-6u%-6u",i,icmp_stats.input[i],
			icmp_stats.output[i]);
		if(icmptypes[i] != NULLCHAR)
			printf("  %s",icmptypes[i]);
		printf("\n");
	}
	return 0;
}
/* Hash table list heads */
struct ping *ping[PMOD];

/* Counter for generating seq numbers */
static WORD iclk;

/* Increment counter -- called by low level clock tick */
void icmpclk()
{
	iclk++;
}

/* Send ICMP Echo Request packets */
doping(int argc,char *argv[])
{
	int32 dest;
	struct ping *pp1;
	register struct ping *pp;
	void ptimeout();
	WORD hval,val,size = 0xffff;
	int i,id,indx = 0;
	char *inet_ntoa();

	if(argc < 2){
		printf("Host                Sent    Rcvd   %%   Avg RTT  Interval\n");
		for(i=0;i<PMOD;i++){
			for(pp = ping[i];pp != NULLPING;pp = pp->next){
				printf("%-16s",inet_ntoa(pp->remote));
				printf("%8lu%8lu",pp->count,pp->echoes);
				printf("%4lu%10lu%10lu\n",
				 (long)pp->echoes * 100 / pp->count,
				 pp->echoes != 0 ?
				 (long)pp->ttotal * MSPTICK / pp->echoes : 0,
				 ((long)pp->timer.start * MSPTICK + 500) / 1000);
			}
		}
		return 0;
	}
	if(strcmp(argv[1],"clear") == 0){
		for(i=0;i<PMOD;i++){
			for(pp = ping[i];pp != NULLPING;pp = pp1){
				pp1 = pp->next;
				del_ping(pp);
			}
		}
		return 0;
	}
	id =
#ifndef NORC
			altconf ? REMCON :
#endif
			ONESHOT;
	for (i = 1;i < argc;++i) {
		if (argv[i][0] == '/') {
			char *p = argv[i] + 1;
			char option = *p++;
			if (*p == '=')
				++p;
			switch(option) {
			case 'i':
				id = atoi(p);
				break;
			case 's':
				size = atoi(p);
				break;
			}
		}
		else {
			switch (indx++) {
			case 0:						// destination host
				if ((dest = resolve(argv[i])) == 0) {
					printf("Host %s unknown\n",argv[i]);
					return 1;
				}
				/* See if dest is already in table */
				hval = hash_ping(dest);
				for(pp = ping[hval]; pp != NULLPING; pp = pp->next){
					if(pp->remote == (DWORD)dest){
						break;
					}
				}
				break;
			case 1:				// time
				/* Inter-ping time is specified; set up timer structure */
				if(pp == NULLPING)
					pp = add_ping(dest);
				val = atoi(argv[i]);
				strchr(argv[1],2);
				if (strchr(argv[i],'m') != 0)
					pp->timer.start = (val + MSPTICK / 2) / MSPTICK;
				else
					pp->timer.start = val * (1000 / MSPTICK);
				pp->timer.func = ptimeout;
				pp->timer.arg = (char *)pp;
				pp->remote = dest;
				pp->size = size;
				start_timer(&pp->timer);
				pp->count++;
				id = REPEAT;
				goto do_it;
			}
		}
	}
	if (indx) {
		do_it:
		pingem(dest,iclk,id,size);
	}
	return 0;
}

/* Called by ping timeout */
void ptimeout(char *p)
{
	register struct ping *pp;

	/* Send another ping */
	pp = (struct ping *)p;
	pp->count++;
	pingem(pp->remote,iclk,REPEAT,pp->size);
	start_timer(&pp->timer);
}

/* Called with incoming Echo Reply packet */
#pragma argsused
void echo_proc(IFACE *ifp,DWORD source,DWORD dest,struct icmp *icmp)
{
	register struct ping *pp;
	WORD hval;
	WORD rtt;
	char *inet_ntoa();
	char str[80];
	XPING *pxp;
	int id = icmp->args.echo.id;

	if (id == EXTRNL) {
		if ((pxp = malloc(sizeof(XPING),AllocIcmpcXping)) != NULL) {
			pxp->remote = source;
			pxp->seq = icmp->args.echo.seq;
			pxp->next = 0;
			if (pxplast) {
				pxplast->next = pxp;
				pxplast = pxp;
			}
			else
				pxpfirst = pxplast = pxp;
		}
		return;
	}
	if (id == QUERY) {
		if_mark_up(ifp,1);
		return;
	}
	if (id == SILENT) {
		return;
	}
	rtt = iclk - icmp->args.echo.seq;
	hval = hash_ping(source);
	for(pp = ping[hval]; pp != NULLPING; pp = pp->next)
		if(pp->remote == source)
			break;
	if (pp == NULLPING || id != REPEAT) {
		sprintf(str,"%s: echo reply id %u seq %u, %lu ms\n",
		 inet_ntoa(source),
		 icmp->args.echo.id,icmp->args.echo.seq,
		 (long)rtt * MSPTICK);
#ifndef NORC
		if (id == REMCON && rc_inuse && !altconf) {
			altconf = 1;
			printf(str);
			altconf = 0;
		}
		else
#endif
			printf(str);
	} else {
		/* Repeated poll, just keep stats */
		pp->ttotal += rtt;
		pp->echoes++;
	}
}

static WORD hash_ping(int32 dest)
{
	WORD hval;

	hval = (hiword(dest) ^ loword(dest)) % PMOD;
	return hval;
}

/* Add entry to ping table */
static struct ping *add_ping(int32 dest)
{
	struct ping *pp;
	WORD hval,hash_ping();

	pp = (struct ping *)calloc(1,sizeof(struct ping),AllocIcmpcPing);
	if(pp == NULLPING)
		return NULLPING;

	hval = hash_ping(dest);
	pp->prev = NULLPING;
	pp->next = ping[hval];
	if(pp->next != NULLPING)
		pp->next->prev = pp;
	ping[hval] = pp;
	return pp;
}
/* Delete entry from ping table */
static void del_ping(struct ping *pp)
{
	WORD hval,hash_ping();

	stop_timer(&pp->timer);

	if(pp->next != NULLPING)
		pp->next->prev = pp->prev;
	if(pp->prev != NULLPING) {
		pp->prev->next = pp->next;
	} else {
		hval = hash_ping(pp->remote);
		ping[hval] = pp->next;
	}
	free((char *)pp);
}

#else
/* Called with incoming Echo Reply packet */
#pragma argsused
void echo_proc(IFACE *ifp,DWORD source,DWORD dest,struct icmp *icmp)
{
	XPING *pxp;
	int id = icmp->args.echo.id;

	if (id == EXTRNL) {
		if ((pxp = malloc(sizeof(XPING),AllocIcmpPing)) != NULL) {
			pxp->remote = source;
			pxp->seq = icmp->args.echo.seq;
			pxp->next = 0;
			if (pxplast) {
				pxplast->next = pxp;
				pxplast = pxp;
			}
			else
				pxpfirst = pxplast = pxp;
		}
		return;
	}
	if (id == QUERY) {
		if_mark_up(ifp,1);
		return;
	}
}
#endif
extern unsigned long serial_numbers[20];
extern unsigned int  num_ser_num;
extern unsigned int copies;
extern char version[];
extern message[];

void pingem_solicited(DWORD dest)		// send ping solicited by ID 1234
{
	pingem(dest,SILENT,SILENT,0xffff);
}

#if APL
void ping_log(DWORD source,MBUF *bp)
{
	char str[100];

	if (bp && bp->cnt) {
		sprintf(str,"ICMP:%s %.*s",inet_ntoa(source),min(bp->cnt,78),bp->data);
		log(0,str);
	}
}
#endif
/* Send ICMP Echo Request packet */
pingem(DWORD dest,WORD seq,WORD id,WORD size)
{
	struct mbuf *bp,*htonicmp();
	struct icmp icmp;
	extern struct icmp_stats icmp_stats;
	char str[80];
	int len;

	icmp_stats.output[ECHO]++;
	icmp.type = ECHO;
	icmp.code = 0;
	icmp.args.echo.seq = seq;
	icmp.args.echo.id = id;
#ifndef NOLIC
	len = sprintf(str,"INET v%s ser # SL%u %lu %s",
					version, copies, serial_numbers[num_ser_num - 1],message);
#else
	len = sprintf(str,"DL Sockets v%s",version);
#endif
	if (size == 0xffff)
		size = len;
	bp = 0;
	if (size) {
		if ((bp = alloc_mbuf(size,AllocIcmpcSnd)) != 0) {
			bp->cnt = size;
			if (len < size) {
				memcpy(bp->data,str,len);
				memset(bp->data + len,'P',size - len);
			}
			else {
				memcpy(bp->data,str,size);
			}
		}
	}
	if ((bp = htonicmp(&icmp,bp)) != 0)
		ip_send(0,dest,ICMP_PTCL,0,0,bp,len_mbuf(bp),0,0,0);
	return 0;
}
