//
//   $Id: IFACE.C 1.4 2003/08/04 19:46:54Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// IFACE.C
// Interface level functions
// Copyright (c) GP van Niekerk Ondernemings 1990-1998

/* Revision history:

2003-08-04 gvn Fix spurious route to 0.0.0.0 when using DHCP
2002-09-16 gvn DHCP ARP followed by gratuitous ARP

*/

#include <string.h>
#include "inetdefs.h"
#include "iface.h"
#include "ping.h"
#include "arp.h"
#include "ip.h"
#include "rip.h"
#include "log.h"
#ifdef X25_S
#include "x25if.h"
#endif
struct interface *ifaces;

struct interface *if_lookup(char *name)
{
	register struct interface *iface = 0;

	for(iface = ifaces; iface != NULLIF; iface = iface->next)
		if(strcmp(iface->name,name) == 0)
			break;
	return iface;
}

void if_timeout(IFACE *ifp)
{
	char stat;

	if (ifp->flags & GARP_FLAG) {
		set_timer(&ifp->timer,1000L * ifp->up_time);
		arp_output(ifp,ifp->hwtype,ifp->ip_addr,ifp->ip_addr);	// gratuitous ARP
		if (ifp->flags & TIMER_FLAG)
			start_timer(&ifp->timer);
		ifp->flags &= ~(GARP_FLAG | TIMER_FLAG);
		return;
	}
	rt_flush_cache();
	if (ifp->stat & IF_UP) {
		if (ifp->stat & IF_QUERY) {
			set_timer(&ifp->timer,1000L * ifp->down_time);
			ifp->stat &= ~(IF_UP | IF_QUERY | IF_DISC);
#if APL
			log(0,"%s down",ifp->name);
#endif
			rip_trigger(ifp);
		}
		else {
#if APL
			log(0,"%s query",ifp->name);
#endif
			ifp->stat |= IF_QUERY;
		}
	}
#ifdef X25_S
	if ((ifp->flags & X25_FLAG) && !(ifp->stat & IF_UP)) {
		// For X.25 use 2 timeouts per down_time period
		// On first one do query ping and disconnect on second one
		if (ifp->stat & IF_DISC) {
			x25_disconnect(ifp);
			ifp->stat &= ~IF_DISC;
			start_timer(&ifp->timer);
			return;
		}
		set_timer(&ifp->timer,500L * ifp->down_time);
		ifp->stat |= IF_DISC;
	}
#endif
	if (ifp->ip_query) {
		stat = ifp->stat;
		ifp->stat |= IF_UP;
		pingem(ifp->ip_query,0,QUERY,0);
		pingem(ifp->ip_query,0,QUERY,0);	// for osi
		pingem(ifp->ip_query,0,QUERY,0);	// for osi
		ifp->stat = stat;
		rt_flush_cache();
	}
	start_timer(&ifp->timer);
}

void if_mark_up(IFACE *ifp,int flag)
{
	if (ifp && (((ifp->stat & (IF_UP | IF_QUERY)) == IF_UP) ||
									 ((flag & 1) && ifp->ip_query) || !ifp->ip_query)) {
		if (ifp->timer.state == TIMER_RUN) {
			stop_timer(&ifp->timer);
		}
		if (!(ifp->stat	& IF_UP)) {
			ifp->stat |= IF_UP;
			set_timer(&ifp->timer,1000L * ifp->up_time);
			rt_flush_cache();
#if APL
			log(0,"%s up",ifp->name);
#endif
			rip_trigger(ifp);
		}
		ifp->stat &= ~IF_QUERY;
	}
}

void if_mark_down(IFACE *ifp)
{
	ifp->stat &= ~(IF_UP | IF_QUERY);
	stop_timer(&ifp->timer);
	set_timer(&ifp->timer,1000L * ifp->down_time);
	if (ifp->timer.start)
		start_timer(&ifp->timer);
	rt_flush_cache();
#if APL
	log(0,"%s down",ifp->name);
#endif
	rip_trigger(ifp);
}

void if_add(IFACE *ifp)
{
	IFACE *cifp;
	extern int trace_all;

	if ((cifp = ifaces) != NULLIF) {
		for(;cifp->next != NULLIF;cifp = cifp->next);
		cifp->next = ifp;
	}
	else
		ifaces = ifp;
	ifp->stat = IF_UP;
	if (ifp->flags & (MDD_FLAG/* | X25_FLAG*/))
		ifp->stat |= IF_LAST_RIP;
	ifp->timer.func = (void(*)(char *))if_timeout;
	ifp->timer.arg = (char *)ifp;
	set_timer(&ifp->timer,1000L * ifp->up_time);
	ifp->ip_addr = def_ip_addr;
	ifp->netbits = def_netbits;
#if APL
	ifp->trace = trace_all;
#endif
	if (!(ifp->sflags & (SF_RAW | SF_UNC)) && (ifp->ip_addr & ~1)) {
		rt_add(ifp->ip_addr & NETMASK(ifp->netbits),ifp->netbits,0,1,ifp,0,RT_STATIC,0);
	}
}
