//
//   $Id: IP.C 1.6 2003/08/04 19:48:32Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// IP.C
// Copyright (C) 2001 Datalight, Inc.
// All Rights Reserved
// Datalight Confidential and Proprietary
// Portions Copyright (c) GPvNO, 1990-2001
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q
//
// Internet Protocol implementation
//
// Revision History
// 08/04/2003 gvn fix random netbits for unopened PPP interface
// 07/29/2003 gvn hostname set when resolving to 0 address, reset when resolving dotted decimal 0.0.0.1
// 07/23/2001 gvn Allocation tracing
// 05/25/2001 gvn updates...

#define	TLB	30 * (1000/MSPTICK)	/* Reassembly limit time */

#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "timer.h"
#include "netfunc.h"
#include "internet.h"
#include "iface.h"
#include "ip.h"
#include "igmp.h"
#include "icmp.h"
#include "get16.h"
#include "_api.h"

extern char must_schedule;
static struct mbuf *fraghandle(struct ip *ip,struct mbuf *bp);
static void free_reasm(register struct reasm *rp);
static struct reasm *creat_reasm(struct ip *ip);
static void freefrag(struct frag *fp);
static struct reasm *lookup_reasm(struct ip *ip);
static struct frag *newfrag(WORD offset,WORD last,struct mbuf *bp);

BYTE ip_ttl = MAXTTL;	/* Default time-to-live for IP datagrams */
WORD wNoMulticastEcho;	// Don't echo Multicast if set

struct reasm *reasmq;
struct mbuf *loopq;		/* Queue for loopback packets */

#define	INSERT	0
#define	APPEND	1
#define	PREPEND	2

/* Send an IP datagram. Modeled after the example interface on p 32 of
 * RFC 791
 */
void ip_send(
	DWORD source,		/* source address */
	DWORD dest,			/* Destination address */
	char protocol,		/* Protocol */
	char tos,			/* Type of service */
	char ttl,			/* Time-to-live */
	struct mbuf *bp,	/* Data portion of datagram */
	WORD length,		/* Optional length of data portion */
	WORD id,			/* Optional identification */
	char df,				/* Don't-fragment flag */
	IFACE *ifp)			// direct to interface if set, else route
{
	struct ip ip;		/* Pointer to IP header */
	static WORD id_cntr;	/* Datagram serial number */
	static char secstart;

	//printf("ip_send %lx -> %lx\n",source,dest);
	if (!secstart) {
		void termstart(void);

		secstart = 1;
		termstart();
	}
	if(length == 0 && bp != NULLBUF)
		length = len_mbuf(bp);
	if(id == 0)
		id = id_cntr++;
	if(ttl == 0)
		ttl = ip_ttl;

	/* Fill in IP header */
	ip.tos = tos;
	ip.length = IPLEN + length;
	ip.id = id;
	if(df)
		ip.fl_offs = DF;
	else
		ip.fl_offs = 0;
	ip.ttl = ttl;
	ip.protocol = protocol;
	if (source == 0) {
		source = this_addr(dest,&ifp);
	}
	ip.source = source;
	ip.dest = dest;
	ip.optlen = 0;
	if (wNoMulticastEcho == 0 && (dest & 0xf0000000l) == 0xe0000000l) {
		MBUF *tbp = copy_p(bp,len_mbuf(bp),AllocIpMc);
		ip_recv(ifp,&ip,tbp,1);
	}
	if (ifp) {
		send_ip(&ip,ifp,bp,dest,IPLEN,length);
	}
	else {
		bp = htonip(&ip,bp,0);
		ip_route(bp,0,0);		/* Toss it to the router */
	}
}

/* Reassemble incoming IP fragments and dispatch completed datagrams
 * to the proper transport module
 */
void ip_recv(IFACE *ifp,
struct ip *ip,		/* Extracted IP header */
struct mbuf *bp,	/* Data portion */
char rxbroadcast)	/* True if received on subnet broadcast address */
{
#ifdef INET_UTIL
	extern int drop_tcp;
#endif
	void (*recv)(IFACE *,struct mbuf *,char,DWORD,DWORD,char,WORD,char);	/* Function to call with completed datagram */

	// check multicast
	if (rxbroadcast) {
		GROUP_AD *psGroupAd;

		if ((ip->dest & 0xf0000000l) == 0xe0000000l) {	// multicast
			if (ip->dest == 0xe0000001l)
				goto found;
			else for (psGroupAd = psGroupAdFirst;psGroupAd;
												psGroupAd = psGroupAd->psNext) {
				if (psGroupAd->psIface == ifp &&
										psGroupAd->dwGroupAddr == ip->dest)
					goto found;
			}
			free_p(bp);		// not interested in this multicast
			return;
			found:
		}
	}
	// Initial check for protocols we can't handle
	switch(uchar(ip->protocol)){
	case TCP_PTCL:
#ifdef INET_UTIL
		if (drop_tcp) {
			free_p(bp);
			return;
		}
#endif
		recv = tcp_input;
		break;
	case UDP_PTCL:
		recv = udp_input;
		break;
	case ICMP_PTCL:
		recv = icmp_input;
		break;
	case IGMP_PTCL:
		recv = (void *)igmp_input;
      break;
	default:
		/* Send an ICMP Protocol Unknown response... */
		ip_stats.badproto++;
		/* ...unless it's a broadcast */
		if(!rxbroadcast){
			icmp_output(ip,bp,DEST_UNREACH,PROT_UNREACH,(union icmp_args *)NULL);
		}
		free_p(bp);
		return;
	}
	/* If we have a complete packet, call the next layer
	 * to handle the result. Note that fraghandle passes back
	 * a length field that does NOT include the IP header
	 */
	if((bp = fraghandle(ip,bp)) != NULLBUF)
		(*recv)(ifp,bp,ip->protocol,ip->source,ip->dest,ip->tos,
			ip->length - (IPLEN + ip->optlen),rxbroadcast);
}
/* Process IP datagram fragments
 * If datagram is complete, return it with ip->length containing the data
 * length (MINUS header); otherwise return NULLBUF
 */
static struct mbuf *fraghandle(struct ip *ip,struct mbuf *bp)
//struct ip *ip;		/* IP header, host byte order */
//struct mbuf *bp;	/* The fragment itself */
{
	register struct reasm *rp; /* Pointer to reassembly descriptor */
	struct frag *lastfrag,*nextfrag,*tfp;
	struct mbuf *tbp;
	WORD i;
	WORD offset;	/* Index of first byte in fragment */
	WORD last;		/* Index of first byte beyond fragment */
	char mf;		/* 1 if not last fragment, 0 otherwise */

	offset = (ip->fl_offs & F_OFFSET) << 3;	/* Convert to bytes */
	last = offset + ip->length - (IPLEN + ip->optlen);
	mf = (ip->fl_offs & MF) ? 1 : 0;

	rp = lookup_reasm(ip);
	if(offset == 0 && !mf){
		/* Complete datagram received. Discard any earlier fragments */
		if(rp != NULLREASM)
			free_reasm(rp);

		return bp;
	}
	if(rp == NULLREASM){
		/* First fragment; create new reassembly descriptor */
		if((rp = creat_reasm(ip)) == NULLREASM){
			/* No space for descriptor, drop fragment */
			free_p(bp);
			return NULLBUF;
		}
	}
	/* Keep restarting timer as long as we keep getting fragments */
	stop_timer(&rp->timer);
	start_timer(&rp->timer);

	/* If this is the last fragment, we now know how long the
	 * entire datagram is; record it
	 */
	if(!mf)
		rp->length = last;

	/* Set nextfrag to the first fragment which begins after us,
	 * and lastfrag to the last fragment which begins before us
	 */
	lastfrag = NULLFRAG;
	for(nextfrag = rp->fraglist;nextfrag != NULLFRAG;nextfrag = nextfrag->next){
		if(nextfrag->offset > offset)
			break;
		lastfrag = nextfrag;
	}
	/* Check for overlap with preceeding fragment */
	if(lastfrag != NULLFRAG  && offset < lastfrag->last){
		/* Strip overlap from new fragment */
		i = lastfrag->last - offset;
		pullup(&bp,NULLCHAR,i);
		if(bp == NULLBUF)
			return NULLBUF;	/* Nothing left */
		offset += i;
	}
	/* Look for overlap with succeeding segments */
	for(; nextfrag != NULLFRAG; nextfrag = tfp){
		tfp = nextfrag->next;	/* save in case we delete fp */

		if(nextfrag->offset >= last)
			break;	/* Past our end */
		/* Trim the front of this entry; if nothing is
		 * left, remove it.
		 */
		i = last - nextfrag->offset;
		pullup(&nextfrag->buf,NULLCHAR,i);
		if(nextfrag->buf == NULLBUF){
			/* superseded; delete from list */
			if(nextfrag->prev != NULLFRAG)
				nextfrag->prev->next = nextfrag->next;
			else
				rp->fraglist = nextfrag->next;
			if(tfp->next != NULLFRAG)
				nextfrag->next->prev = nextfrag->prev;
			freefrag(nextfrag);
		} else
			nextfrag->offset = last;
	}
	/* Lastfrag now points, as before, to the fragment before us;
	 * nextfrag points at the next fragment. Check to see if we can
	 * join to either or both fragments.
	 */
	i = INSERT;
	if(lastfrag != NULLFRAG && lastfrag->last == offset)
		i |= APPEND;
	if(nextfrag != NULLFRAG && nextfrag->offset == last)
		i |= PREPEND;
	switch(i){
	case INSERT:	/* Insert new desc between lastfrag and nextfrag */
		if ((tfp = newfrag(offset,last,bp)) == 0)
			return NULLBUF;
		tfp->prev = lastfrag;
		tfp->next = nextfrag;
		if(lastfrag != NULLFRAG)
			lastfrag->next = tfp;	/* Middle of list */
		else
			rp->fraglist = tfp;	/* First on list */
		if(nextfrag != NULLFRAG)
			nextfrag->prev = tfp;
		break;
	case APPEND:	/* Append to lastfrag */
		append(&lastfrag->buf,bp);
		lastfrag->last = last;	/* Extend forward */
		break;
	case PREPEND:	/* Prepend to nextfrag */
		tbp = nextfrag->buf;
		nextfrag->buf = bp;
		append(&nextfrag->buf,tbp);
		nextfrag->offset = offset;	/* Extend backward */
		break;
	case (APPEND|PREPEND):
		/* Consolidate by appending this fragment and nextfrag
		 * to lastfrag and removing the nextfrag descriptor
		 */
		append(&lastfrag->buf,bp);
		append(&lastfrag->buf,nextfrag->buf);
		nextfrag->buf = NULLBUF;
		lastfrag->last = nextfrag->last;

		/* Finally unlink and delete the now unneeded nextfrag */
		lastfrag->next = nextfrag->next;
		if(nextfrag->next != NULLFRAG)
			nextfrag->next->prev = lastfrag;
		freefrag(nextfrag);
		break;
	}
	if(rp->fraglist->offset == 0 && rp->fraglist->next == NULLFRAG 
		&& rp->length != 0){
		/* We've gotten a complete datagram, so extract it from the
		 * reassembly buffer and pass it on.
		 */
		bp = rp->fraglist->buf;
		rp->fraglist->buf = NULLBUF;
		/* Tell IP the entire length */
		ip->length = rp->length + (IPLEN + ip->optlen);
		free_reasm(rp);
		return bp;
	} else
		return NULLBUF;
}
static struct reasm *lookup_reasm(struct ip *ip)
{
	register struct reasm *rp;

	for(rp = reasmq;rp != NULLREASM;rp = rp->next){
		if(ip->source == rp->source && ip->dest == rp->dest
		 && ip->protocol == rp->protocol && ip->id == rp->id)
			return rp;
	}
	return NULLREASM;
}
#ifdef	FOO
static
WORD
hash_reasm(source,dest,protocol,id)
int32 source;
int32 dest,
char protocol;
WORD id;
{
	register WORD hval;

	hval = loword(source);
	hval ^= hiword(source);
	hval ^= loword(dest);
	hval ^= hiword(dest);
	hval ^= uchar(protocol);
	hval ^= id;
	return hval % (WORD)RHASH;
}
#endif
/* Create a reassembly descriptor,
 * put at head of reassembly list
 */
static struct reasm *creat_reasm(struct ip *ip)
{
	register struct reasm *rp;
	void ip_timeout();

	if((rp = (struct reasm *)calloc(1,sizeof(struct reasm),AllocIpRa)) == NULLREASM)
		return rp;	/* No space for descriptor */
	rp->source = ip->source;
	rp->dest = ip->dest;
	rp->id = ip->id;
	rp->protocol = ip->protocol;
	rp->timer.start = TLB;
	rp->timer.func = ip_timeout;
	rp->timer.arg = (char *)rp;

	rp->next = reasmq;
	if(rp->next != NULLREASM)
		rp->next->prev = rp;
	reasmq = rp;
	return rp;
}

/* Free all resources associated with a reassembly descriptor */
static void free_reasm(register struct reasm *rp)
{
	register struct frag *fp;

	stop_timer(&rp->timer);
	/* Remove from list of reassembly descriptors */
	if(rp->prev != NULLREASM)
		rp->prev->next = rp->next;
	else
		reasmq = rp->next;
	if(rp->next != NULLREASM)
		rp->next->prev = rp->prev;
	/* Free any fragments on list, starting at beginning */
	while((fp = rp->fraglist) != NULLFRAG){
		rp->fraglist = fp->next;
		free_p(fp->buf);
		free((char *)fp);
	}
	free((char *)rp);
}

/* Handle reassembly timeouts by deleting all reassembly resources */
static void ip_timeout(int *arg)
{
	register struct reasm *rp;

	rp = (struct reasm *)arg;
	free_reasm(rp);
}
/* Create a fragment */
static struct frag *newfrag(WORD offset,WORD last,struct mbuf *bp)
{
	struct frag *fp;

	if((fp = (struct frag *)calloc(1,sizeof(struct frag),AllocIpFrag)) == NULLFRAG){
		/* Drop fragment */
		free_p(bp);
		return NULLFRAG;
	}
	fp->buf = bp;
	fp->offset = offset;
	fp->last = last;
	return fp;
}
/* Delete a fragment, return next one on queue */
static void freefrag(struct frag *fp)
{
	free_p(fp->buf);
	free((char *)fp);
}

set_ip_addr(char *name,int netlen)
{
	char *phost;
	WORD ip_netlen = 0;
	DWORD n;

	n = resolve_nl(name,&ip_netlen,&phost);
	if (netlen != -1)
		ip_netlen = netlen;
	def_netbits = ip_netlen;
	def_ip_addr = n;
	if (!*hostname)
		strncpy((char *)hostname,phost,HOSTNAMELEN);
	return 0;
}

/* resolve name and optional netID length */
DWORD resolve_nl(char *name,WORD *plen,char **phost)
{
	unsigned bits = 32;
	char *bitp;
	int32 addr = 0;
	char separator = 0;

	if (*name == '/') {
		addr = *((long far *)MK_FP(0,4 * htoi(name + 1)));
		*phost = inet_ntoa(addr);
		name++;
	}
	/* If IP address is followed by an optional slash and
	 * a length field, (e.g., 128.96/16) get it;
	 */
	if ((bitp = strchr(name,'/')) != 0 || (bitp = strchr(name,'&')) != 0) {
		separator = *bitp;
		*bitp++ = 0;
		bits = atoi(bitp);
	}
	if (!addr) {
		*phost = name;
		if ((addr = resolve(name)) == 0)
			return 0;
		if (addr == 1L && is_dotted_decimal(name))
			*name = 0;	// No name for 0.0.0.1 (DHCP)
	}
	if (bits == 32) {
		if ((addr & 0x80000000l) == 0)	/* class A network */
			bits = 8;
		else if ((addr & 0xc0000000l) == 0x80000000l)/* B */
			bits = 16;
		else if ((addr & 0xe0000000l) == 0xc0000000l)/* C */
			bits = 24;
	}
	*plen = bits;
	if (separator == '&')
		addr &= (0xffffffffl << (32 - bits));
	return addr;
}

// IP routing functions

#ifdef TGW
#include "task.h"
extern unsigned long master_tick;
char tgw_init;
int poll_ip_q();
struct ip *ip_q,*ip_qe;
#endif

void rt_timeout(void *s);
void rt_flush_cache();
//WORD hash_ip(int32 addr);
#define	hash_ip(addr) ((((WORD *)&addr)[0] ^ ((WORD *)&addr)[1]) % NROUTE)

ROUTE *routes[33][NROUTE];	/* Routing table */

char source_filter;		// set if source to default filter is enabled
int num_filter;			// number of active entries in filter
int max_filter;			// maximum number of allocated entries in filter
DWORD *ip_filter;		// pointer to entries in filter

DWORD def_ip_addr;		// our IP address
BYTE def_netbits;		// number of bits in our net mask
char route_echo;		// set to route over SAME interface

struct ip_stats ip_stats = {0,0,0,0,0,0,0,0,0,&routes};

char dont_time; 		// set if iface timer should not be started
extern char this_is_rip;// this datagram is a RIP

/* send an ip datagram to an interface */
void send_ip(struct ip *pip,IFACE *iface,MBUF *bp,DWORD gateway,
		WORD ip_len,WORD length)
{
	char precedence;		/* Extracted from tos field */
	char delay;
	char throughput;
	char reliability;
	WORD offset;			/* Offset into current fragment */
	WORD fragflag;			/* fragment flag */

	precedence = PREC(pip->tos);
	delay = pip->tos & DELAY;
	throughput = pip->tos & THRUPUT;
	reliability = pip->tos & RELIABILITY;

	if (iface->flags & (MDD_FLAG/* | X25_FLAG*/)) {
		if (this_is_rip) {
			if (iface->stat & IF_LAST_RIP) {
				free_p(bp);	// don't send
				return;
			}
			iface->stat |= IF_LAST_RIP;
		}
		else
			iface->stat &= ~IF_LAST_RIP;            
	}
	if(pip->length <= iface->mtu){
		/* Datagram smaller than interface MTU; put header
		 * back on and send normally
		 */
		bp = htonip(pip,bp,0);
		if (iface->timer.start && iface->timer.state != TIMER_RUN &&
														!dont_time) {
			start_timer(&iface->timer);
		}
		(*iface->send)(bp,iface,gateway,
			precedence,delay,throughput,reliability);
		return;
	}
	/* Fragmentation needed */
	if(pip->fl_offs & DF){
		/* Don't Fragment set; return ICMP message and drop */
		icmp_output(pip,bp,DEST_UNREACH,FRAG_NEEDED,NULLICMP);
		free_p(bp);
		return;
	}
	/* Create fragments */
	offset = (pip->fl_offs & F_OFFSET) << 3;
	fragflag = pip->fl_offs & MF;
	while (length != 0) {		/* As long as there's data left */
		WORD fragsize;		/* Size of this fragment's data */
		struct mbuf *f_data;	/* Data portion of fragment */

		/* After the first fragment, should remove those
		 * options that aren't supposed to be copied on fragmentation
		 */
		pip->fl_offs = (offset >> 3) | fragflag;
		if(length + ip_len <= iface->mtu){
			/* Last fragment; send all that remains */
			fragsize = length;
		} else {
			/* More to come, so send multiple of 8 bytes */
			fragsize = (iface->mtu - ip_len) & 0xfff8;
			pip->fl_offs |= MF;
		}
		pip->length = fragsize + ip_len;

		/* Move the data fragment into a new, separate mbuf */
		if((f_data = alloc_mbuf(fragsize,AllocIpFragSnd)) == NULLBUF){
			free_p(bp);
			break;
		}
		f_data->cnt = pullup(&bp,f_data->data,fragsize);

		/* Put IP header back on */
		if((f_data = htonip(pip,f_data,0)) == NULLBUF){
			free_p(bp);
			break;
		}
		/* and ship it out */
		if (iface->timer.start && iface->timer.state != TIMER_RUN &&
															!dont_time) {
			start_timer(&iface->timer);
		}
		(*iface->send)(f_data,iface,gateway,
			precedence,delay,throughput,reliability);

		offset += fragsize;
		length -= fragsize;
	}
}

IFACE *ifpsecurity;			// set if security dg must be sent

#ifdef NAT
void near FixCheckSum(MBUF *bp,DWORD dwNew,DWORD dwOld,int iProtocol)
{
	WORD wCsAddr;
	DWORD dwCsDiff = 	(WORD)dwNew + (dwNew >> 16) -
							(WORD)dwOld - (dwOld >> 16);
	wCsAddr = 	iProtocol == TCP_PTCL ? 16 :
					iProtocol == UDP_PTCL ? 6 :
					iProtocol == IGMP_PTCL ? 2 : 0;
	if (wCsAddr) {
		int iOverflow;
		dwCsDiff += ~(WORD)get16(bp->data + wCsAddr);
		while ((iOverflow = (int)(dwCsDiff >> 16)) != 0)
			dwCsDiff = (dwCsDiff & 0xffff) + iOverflow;
		put16(bp->data + wCsAddr,~(WORD)dwCsDiff);
	}
}
#endif

/* Route an IP datagram. This is the "hopper" through which all IP datagrams,
 * coming or going, must pass.
 *
 * "rxbroadcast" is set to indicate that the packet came in on a subnet
 * broadcast. The router will kick the dg upstairs regardless of the
 * IP destination address.
 */
void ip_route(MBUF *bp,char rxbroadcast,IFACE *ifp)
{
#ifdef TGW
	struct ip *pip;
#endif
	struct ip ip;			/* IP header being processed */
	WORD ip_len;			/* IP header length */
	WORD length;			/* Length of data portion */
	DWORD gateway;			/* Gateway IP address */
	register ROUTE *rp;	/* Route table entry */
	IFACE *iface;			/* Output interface, possibly forwarded */
	int strict = 0;		/* Strict source routing flag */
	WORD opt_len;			/* Length of current option */
	BYTE *opt;				/* -> beginning of current option */
	BYTE *ptr;				/* -> pointer field in source route fields */
	struct mbuf *tbp;

	ip_stats.total++;
	if(len_mbuf(bp) < IPLEN){
		/* The packet is shorter than a legal IP header */
		ip_stats.runt++;
		free_p(bp);
		return;
	}
	/* Sneak a peek at the IP header's IHL field to find its length */
	ip_len = (bp->data[0] & 0xf) << 2;
	if(ip_len < IPLEN){
		/* The IP header length field is too small */
		ip_stats.length++;
		free_p(bp);
		return;
	}
#ifdef TGW
	if (ip_stats.drop_rand || ip_stats.drop_ratio) {
		if (!ip_stats.drop_count--) {
			ip_stats.drop_count = ip_stats.drop_ratio + random(ip_stats.drop_rand);
			goto bad_check;
		}
	}
#endif
	if(cksum(NULLHEADER,bp,ip_len) != 0){
		/* Bad IP header checksum; discard */
#ifdef TGW
		bad_check:
#endif
		ip_stats.checksum++;
		free_p(bp);
		return;
	}
	if (ifp) {
		if_mark_up(ifp,0);
	}
	/* Extract IP header */
	ntohip(&ip,&bp);

#ifdef NAT
	if (ifp && (ifp->flags & NAT_FLAG) && (ip.dest & ifp->dwNatMask) == ifp->dwIpExternal) {
		DWORD dwNewDest = ifp->dwIpInternal | (ip.dest & ~ifp->dwNatMask);
		FixCheckSum(bp,dwNewDest,ip.dest,ip.protocol);
		ip.dest = dwNewDest;
	}
#endif
	if(ip.version != IPVERSION){
		/* We can't handle this version of IP */
		ip_stats.version++;
		free_p(bp);
		return;
	}
	/* Trim data segment if necessary. */
	length = ip.length - ip_len; /* Length of data portion */
	trim_mbuf(&bp,length);

	/* Process options, if any. Also compute length of secondary IP
	 * header in case fragmentation is needed later
	 */
	strict = 0;
	for(opt = ip.options; opt <	&ip.options[ip.optlen];opt += opt_len){
		DWORD ipaddr = ifp ? ifp->ip_addr : def_ip_addr;

		/* Most options have a length field. If this is a EOL or NOOP,
		 * this (garbage) value won't be used
		 */
		opt_len = uchar(opt[1]);

		switch(opt[0] & OPT_NUMBER){
		case IP_EOL:
			goto no_opt;	/* End of options list, we're done */
		case IP_NOOP:
			opt_len = 1;
			break;		/* No operation, skip to next option */
		case IP_SSROUTE:	/* Strict source route & record route */
			strict = 1;	/* note fall-thru */
		case IP_LSROUTE:	/* Loose source route & record route */
			/* Source routes are ignored unless we're in the
			 * destination field
			 */
			if(ip.dest != ipaddr)
				break;	/* Skip to next option */
			if(uchar(opt[2]) >= opt_len){
				break;	/* Route exhausted; it's for us */
			}
			/* Put address for next hop into destination field,
			 * put our address into the route field, and bump
			 * the pointer
			 */
			ptr = opt + uchar(opt[2]) - 1;
			ip.dest = get32(ptr);
			put32(ptr,ipaddr);
			opt[2] += 4;
			break;
		case IP_RROUTE: /* Record route */
			if(uchar(opt[2]) >= opt_len){
				/* Route area exhausted; kick back an error */
				union icmp_args icmp_args;

				icmp_args.pointer = IPLEN + opt - ip.options;
				icmp_output(&ip,bp,PARAM_PROB,0,&icmp_args);
				free_p(bp);
				return;
			}
			/* Add our address to the route */
			ptr = opt + uchar(opt[2]) - 1;
			ptr = put32(ptr,ipaddr);
			opt[2] += 4;
			break;
		}
	}
no_opt:

	// See if it's a broadcast or addressed to us, or the
	// iface ip address is zero (meaning bootp!) and send it up
	if (ifp && (!ifp->ip_addr || (ip.dest == 0xffffffffl) ||
			(ip.dest == (ifp->ip_addr | HOSTMASK(ifp->netbits))))) {
		rxbroadcast = 1;
	}
	if ((ifp && ip.dest == ifp->ip_addr) || is_this_addr(ip.dest) ||
														rxbroadcast) {
		/* If this is a local loopback packet, place on the loopback
		 * queue for processing in the main loop. This prevents the
		 * infinite stack recursion and other problems that would
		 * otherwise occur when we talk to ourselves, e.g., with ftp
		 */
		if(is_this_addr(ip.source)){
			/* Put IP header back on */
			if((bp = htonip(&ip,bp,0)) == NULLBUF)
				return;

			/* Copy loopback packet into new buffer.
			 * This avoids an obscure problem with TCP which
			 * dups its outgoing data before transmission and
			 * then	frees it when an ack comes, even though the
			 * receiver might not have actually read it yet
			 */
			tbp = copy_p(bp,len_mbuf(bp),AllocIpLb);
			free_p(bp);
			if(tbp != NULLBUF) {
				enqueue(&loopq,tbp);
				must_schedule = 3;
			}
		} else {
			ip_recv(ifp,&ip,bp,rxbroadcast);
		}
		return;
	}

	/* Decrement TTL and discard if zero */
	if(--ip.ttl == 0){
		/* Send	ICMP "Time Exceeded" message */
		icmp_output(&ip,bp,TIME_EXCEED,0,NULLICMP);
		free_p(bp);
		return;
	}
	/* Look up target address in routing table */
	if((rp = rt_lookup(ip.dest,ip.source,SKIP_DOWN)) == NULLROUTE){
		/* No route exists, return unreachable message */
		icmp_output(&ip,bp,DEST_UNREACH,HOST_UNREACH,NULLICMP);
		free_p(bp);
		return;
	}
	/* Check for alternate and send tel_tos on it
	iface =	(tcp_stat.tel_tos && (ip.tos == tcp_stat.tel_tos) &&
					rp->altif) ? rp->altif : rp->interface;
	*/
	iface = rp->interface;
	if (iface == ifp && !route_echo) {
		icmp_output(&ip,bp,DEST_UNREACH,HOST_UNREACH,NULLICMP);
		free_p(bp);
		return;
	}
#ifdef NAT
	if ((iface->flags & NAT_FLAG) && (ip.source & iface->dwNatMask) == iface->dwIpInternal) {
		DWORD dwNewSource = iface->dwIpExternal | (ip.source & ~iface->dwNatMask);
		FixCheckSum(bp,dwNewSource,ip.source,ip.protocol);
		ip.source = dwNewSource;
	}
#endif
	/* Find gateway; zero gateway in routing table means "send direct" */
	if(rp->gateway == (int32)0)
		gateway = ip.dest;
	else
		gateway = rp->gateway;

	if(strict && gateway != ip.dest){
		/* Strict source routing requires a direct entry */
		icmp_output(&ip,bp,DEST_UNREACH,ROUTE_FAIL,NULLICMP);
		free_p(bp);
		return;
	}
	if ((iface->flags & SECUR_FLAG) && !this_is_rip) {
		void termprod(void);
		ifpsecurity = iface;
		iface->flags &= ~SECUR_FLAG;
		termprod();
	}
#ifndef TGW
	send_ip(&ip,iface,bp,gateway,ip_len,length);
}
#else
	if (!ip_stats.delay && !ip_stats.delay_rand && !ip_stats.thruput) {
		send_ip(&ip,iface,bp,gateway,ip_len,length);
		return;
	}
	if (!tgw_init) {
		tgw_init = 1;
		randomize();
		task_create(poll_ip_q,0);
	}
	ip.bp = bp;
	ip.iface = iface;
	ip.gateway = gateway;
	ip.ip_len = ip_len;
	ip.lengthd = length;
	if ((pip = malloc(sizeof(struct ip))) == 0)
		return;
	*pip = ip;
	pip->next = 0;
	pip->time = master_tick + ip_stats.delay + random(ip_stats.delay_rand);
	if (ip_q) {
		ip_qe->next = pip;
		ip_qe = pip;
	}
	else
		ip_q = ip_qe = pip;
}

int poll_ip_q()
{
	struct ip ip,*pip,*pipl;
	extern WORD sec_timer;		/* one second timer */
	static WORD this_sec;
	static DWORD sent;

	if (!ip_q)
		return;
	for (pip = ip_q,pipl = 0;pip;pip = pip->next) {
		if (pip->time > master_tick) {
			if (!ip_stats.delay_rand)
				return;
		}
		else
			break;
		pipl = pip;
	}
	if (!pip)
		return;
	if (this_sec == sec_timer) {
		if (sent + pip->length > ip_stats.thruput)
			return;
		sent += pip->length;
	}
	else {
		sent = pip->length;
		this_sec = sec_timer;
	}
	ip = *pip;
	if (pipl)
		pipl->next = pip->next;
	else
		ip_q = pip->next;
	free(pip);
	send_ip(&ip,ip.iface,ip.bp,ip.gateway,ip.ip_len,ip.lengthd);
}
#endif

struct rt_cache rt_cache;

/* Add an entry to the IP routing table. Returns 0 on success, -1 on failure */
int rt_add(
DWORD target,		/* Target IP address prefix */
unsigned int bits,	/* Size of target address prefix in bits (0-32) */
DWORD gateway,
int metric,
struct interface *interface,
DWORD time,char stat,char top)
{
	ROUTE *rp,*crp,**hp;

	if(interface == NULLIF || (interface->sflags & SF_RAW))
		return -1;

	rt_cache.target = 0;	/* Flush cache */

	if (bits > 32 || (target & ~(0xffffffffl << (32 - bits))))
		bits = 32;
	if((rp = (ROUTE *)calloc(1,sizeof(ROUTE),AllocIpRoute)) == NULLROUTE)
		return -1;	/* No space */
	hp = &routes[bits][hash_ip(target)];
	if ((crp = *hp) != NULLROUTE) {
		if (top) {
			rp->next = crp;
			crp->prev = rp;
			*hp = rp;
		}
		else {
			for (;crp->next != NULLROUTE;crp = crp->next);
			crp->next = rp;
			rp->prev = crp;
		}
	}
	else
		*hp = rp;
	rp->target = target;
	rp->gateway = gateway;
	rp->metric = metric;
	rp->interface = interface;
//	rp->altif = altif;
//	rp->priority = pri;
	if (time) {
		set_timer(&rp->timer,time * 1000L);
		start_timer(&rp->timer);
	}
	rp->timer.func = (void far *)rt_timeout;
	rp->timer.arg = (void *)rp;
	rp->bits = bits;
	rp->stat = stat;
	return 0;
}

/* Remove an entry from the IP routing table. Returns 0 on success, -1
 * if entry was not in table.
 */
int rt_drop(DWORD target,unsigned int bits)
{
	register ROUTE *rp;

	rt_cache.target = 0;	/* Flush the cache */

	if (bits > 32 || (target & ~(0xffffffffl << (32 - bits))))
		bits = 32;
	/* Search appropriate chain for existing entry */
	for(rp = routes[bits][hash_ip(target)];rp != NULLROUTE;rp = rp->next){
		if(rp->target == target)
			break;
	}
	if(rp == NULLROUTE)
		return -1;	/* Not in table */

	stop_timer(&rp->timer);
	if(rp->next != NULLROUTE)
		rp->next->prev = rp->prev;
	if(rp->prev != NULLROUTE)
		rp->prev->next = rp->next;
	else
		routes[bits][hash_ip(target)] = rp->next;

	free((char *)rp);
	return 0;
}

/*
/* Compute hash function on IP address */
static WORD hash_ip(int32 addr)
{
	register WORD ret;

	ret = hiword(addr);
	ret ^= loword(addr);
	return ret % (WORD)NROUTE;
}
*/

/* Given an IP address, return the MTU of the local interface used to
 * reach that destination. This is used by TCP to avoid local fragmentation
 */
WORD ip_mtu(DWORD addr)
{
	register ROUTE *rp;
	struct interface *iface;

	rp = rt_lookup(addr,0,SKIP_DOWN);
	if(rp == NULLROUTE || rp->interface == NULLIF)
		return 0;

	iface =	rp->interface;
	return iface->mtu;
}

DWORD this_addr(DWORD dest,IFACE **pifp) 	 // determine source IP address to use
{
	ROUTE *rp;

	if (!dest)
		return 0l;
	if (is_this_addr(dest))
		return dest;
	if ((dest & 0xf0000000l) == 0xe0000000l) {	// multicast
		if (!pifp)
			return 0;		// nothing we can do
		if (!*pifp)
			*pifp = ifaces;		// first interface is default for multicast
		return (*pifp)->ip_addr;
	}
	if ((rp = rt_lookup(dest,0,SKIP_DOWN)) == 0)
		return 0l;
	return rp->interface->ip_addr;
}

int is_this_addr(DWORD dest)	// check for our host adress(es)
{
	IFACE *ifp;

	if (dest == 0x7f000001L)
		return 1;
	if ((ifp = ifaces) == 0) {
		return (def_ip_addr == dest);
	}
	for (;ifp;ifp = ifp->next) {
		if (ifp->ip_addr == dest && !(ifp->sflags & (SF_RAW | SF_UNC)))
			return 1;
	}
	return 0;
}

/* Look up target in hash table, matching the entry having the largest number
 * of leading bits in common. Return default route if not found;
 * if default route not set, return NULLROUTE
 */
ROUTE *rt_lookup(DWORD target,DWORD source,int flags)
{
	register ROUTE *rp;
	int bits;
	int32 tsave;
	int32 mask;

	if(target == rt_cache.target) {
		if ((rt_cache.route->interface->stat & IF_UP) &&
											rt_cache.route->metric < 16)
			return rt_cache.route;
		rt_cache.target = 0;
	}
	tsave = target;

	mask = ~0;	/* All ones */
	for (bits = 32;bits >= 0; bits--,mask <<= 1) {
		target &= mask;
		for(rp = routes[bits][hash_ip(target)];rp != NULLROUTE;rp = rp->next) {
			if(rp->target == target) {
				if ((rp->interface->stat & IF_UP) && rp->metric < 16) {
					if (source_filter && bits == 0) {
						int i;
						if (!source)
							return rp;
						for (i = 0;i < num_filter;++i) {
							if (ip_filter[i] == source)
								goto ok;
						}
						return 0;
					}
					ok:
					/* Stash in cache and return */
					rt_cache.target = tsave;
					rt_cache.route = rp;
					return rp;
				}
				else if (flags == USE_DOWN) {
					return rp;
				}
			}
		}
	}
	return NULLROUTE;
}

void rt_flush_cache()
{
	rt_cache.target = 0;
}

/* Convert IP header in host format to network mbuf */
struct mbuf * htonip(struct ip *ip,MBUF *data,int cs_flag)
{
	WORD hdr_len;
	struct mbuf *bp;
	BYTE *cp;
	WORD checksum;

	hdr_len = IPLEN + ip->optlen;
	if((bp = alloc_mbuf(hdr_len,AllocIpSnd)) == NULLBUF){
		return NULLBUF;
	}
	bp->cnt = hdr_len;
	cp = bp->data;

	*cp++ = (IPVERSION << 4) | (hdr_len >> 2);
	*cp++ = ip->tos;
	cp = put16(cp,ip->length);
	cp = put16(cp,ip->id);
	cp = put16(cp,ip->fl_offs);
	*cp++ = ip->ttl;
	*cp++ = ip->protocol;
	cp = put16(cp,0);	/* Clear checksum */
	cp = put32(cp,ip->source);
	cp = put32(cp,ip->dest);
	if(ip->optlen != 0)
		memcpy(cp,ip->options,ip->optlen);

	checksum = cs_flag ? ip->checksum : cksum(NULLHEADER,bp,hdr_len);
	put16(bp->data + 10,checksum);

	bp->next = data;
	return bp;
}
/* Extract an IP header from mbuf */
ntohip(struct ip *ip,struct mbuf **bpp)
{
	int ihl;
	BYTE ipbuf[IPLEN];

	if(pullup(bpp,ipbuf,IPLEN) != IPLEN)
		return -1;

	ip->version = (ipbuf[0] >> 4) & 0xf;
	ip->tos = ipbuf[1];
	ip->length = get16(ipbuf + 2);
	ip->id = get16(ipbuf + 4);
	ip->fl_offs = get16(ipbuf + 6);
	ip->ttl = ipbuf[8];
	ip->protocol = ipbuf[9];
	ip->checksum = get16(ipbuf + 10);
	ip->source = get32(ipbuf + 12);
	ip->dest = get32(ipbuf + 16);

	ihl = (ipbuf[0] & 0xf) << 2;
	if(ihl < IPLEN){
		/* Bogus packet; header is too short */
		ip->optlen = 0;
		return -1;
	}
	if ( (ip->optlen = ihl - IPLEN) != 0 ) {
		if ( pullup(bpp,ip->options,ip->optlen) < ip->optlen )
			return -1;
	}
	return ihl;
}
/* old code
{
	char v_ihl;
	WORD ihl;

	v_ihl = pullchar(bpp);
	ip->version = (v_ihl >> 4) & 0xf;
	ip->tos = pullchar(bpp);
	ip->length = pull16(bpp);
	ip->id = pull16(bpp);
	ip->fl_offs = pull16(bpp);
	ip->ttl	= pullchar(bpp);
	ip->protocol = pullchar(bpp);
	(void)pull16(bpp); /* Toss checksum */
	ip->source = pull32(bpp);
	ip->dest = pull32(bpp);

	ihl = (v_ihl & 0xf) << 2;
	if(ihl < IPLEN){
		/* Bogus packet; header is too short */
		return -1;
	}
	ip->optlen = ihl - IPLEN;
	if(ip->optlen != 0)
		pullup(bpp,ip->options,ip->optlen);

	return ip->optlen + IPLEN;
}
*/
/* Perform end-around-carry adjustment */
WORD eac(int32 sum)
//int32 sum;	/* Carries in high order 16 bits */
{
	register WORD csum;

	while((csum = (WORD)(sum >> 16)) != 0)
		sum = csum + (sum & 0xffffL);
	return (WORD) (sum & 0xffffl);	/* Chops to 16 bits */
}
/* Checksum a mbuf chain, with optional pseudo-header */
WORD cksum(struct pseudo_header *ph,struct mbuf *m,WORD len)
{
	register unsigned int cnt, total;
	register int32 sum, csum;
	register char *up;
	WORD csum1;
	int swap = 0;

	sum = 0l;

	/* Sum pseudo-header, if present */
	if(ph != NULLHEADER){
		sum = hiword(ph->source);
		sum += loword(ph->source);
		sum += hiword(ph->dest);
		sum += loword(ph->dest);
		sum += uchar(ph->protocol);
		sum += ph->length;
	}
	/* Now do each mbuf on the chain */
	for(total = 0; m != NULLBUF && total < len; m = m->next) {
		cnt = min(m->cnt, len - total);
		up = (char *)m->data;
		csum = 0;

		if(((long)up) & 1){
			/* Handle odd leading byte */
			if(swap)
				csum = uchar(*up++);
			else
				csum = (WORD)(uchar(*up++) << 8);
			cnt--;
			swap = !swap;
		}
		if(cnt > 1){
			/* Have	the primitive checksumming routine do most of
			 * the work. At this point, up is guaranteed to be on
			 * a short boundary
			 */
			csum1 = lcsum((unsigned short *)up, cnt >> 1);
			if(swap)
				csum1 = (csum1 << 8) | (csum1 >> 8);
			csum += csum1;
		}
		/* Handle odd trailing byte */
		if(cnt & 1){
			if(swap)
				csum += uchar(up[--cnt]);
			else
				csum += (WORD)(uchar(up[--cnt]) << 8);
			swap = !swap;
		}
		sum += csum;
		total += m->cnt;
	}
	/* Do final end-around carry, complement and return */
	return ~eac(sum) & 0xffff;
}
/* Machine-independent, alignment insensitive network-to-host long conversion */
DWORD get32(BYTE *cp)
{
	int32 rval;
/* slow, machine independant
	rval = uchar(*cp++);
	rval <<= 8;
	rval |= uchar(*cp++);
	rval <<= 8;
	rval |= uchar(*cp++);
	rval <<= 8;
	rval |= uchar(*cp);
*/
/* fast, PC specific */
	char *rp = (char *)&rval;
	rp[3] = cp[0];
	rp[2] = cp[1];
	rp[1] = cp[2];
	rp[0] = cp[3];
	return rval;
}

/* Search routing table for entry with specific width */
struct route *rt_blookup(DWORD target,unsigned int bits)
{
	register struct route *rp;

	/* Mask off target according to width */
	target &= ~0L << (32-bits);

	for(rp = routes[bits][hash_ip(target)];rp != NULLROUTE;rp = rp->next){
		if(rp->target == target) {
			break;
		}
	}
	return rp;
}

// source to default filter - another OSI/MJS invention

int source_default_filter(DWORD address,int operation)
{
	int i;

	rt_flush_cache();
	switch(operation) {
	case FILTER_DISABLE:
	case FILTER_ENABLE:
		source_filter = operation;
		return 0;
	case FILTER_FLUSH:
		num_filter = max_filter = 0;
		free(ip_filter);
		ip_filter = 0;
		return 0;
	case FILTER_ADD:
		if (!ip_filter) {
			max_filter = 20;
			if ((ip_filter = malloc(max_filter * sizeof(int32),AllocIpFilter)) == 0)
				return -2;
		}
		for (i = 0;i < num_filter;++i) {
			if (ip_filter[i] == address)
				return -1;
		}
		if (num_filter >= max_filter) {
			if ((ip_filter = realloc(ip_filter,(max_filter + 20) * sizeof(int32),AllocIpFilterRe)) == 0)
				return -2;
			max_filter += 20;
		}
		ip_filter[num_filter++] = address;
		return 0;
	case FILTER_DROP:
		for (i = 0;i < num_filter;++i) {
			if (ip_filter[i] == address) {
				for (;i < num_filter - 1;++i) {
					ip_filter[i] = ip_filter[i + 1];
				}
				--num_filter;
				return 0;
			}
		}
	default:
		return -1;
	}
}

/* code to determine routes by means of pings
   shelved for now

void rt_timeout(ROUTE *rp)
{
	if (rp->stat & RT_QPING_RCVD) {
		rp->stat &= ~RT_QPING_RCVD;
		return;
	}
	start_timer(&rp->timer);
	rp->stat |= RT_UP;			// force usage of route
	pingem(rp->ping_addr,FP_OFF(rp),FP_SEG(rp),0);
	rp->stat &= ~RT_UP;
	rt_cache.target = 0;
}

void rt_rcvping(ROUTE *rpp,int32 source)
{
	ROUTE *rp;
	int bits;
	int32 mask;

	mask = ~0;
	for (bits = 32;bits >= 0; bits--,mask <<= 1) {
		source &= mask;
		for(rp = routes[bits][hash_ip(source)];rp != NULLROUTE;rp = rp->next) {
			if (rp == rpp) {
				rp->stat |= RT_QPING_RCVD;
				if (!(rp->stat & RT_UP)) {
					rp->stat |= RT_UP;
					rt_cache.target = 0;
				}
				return;
			}
		}
	}
}
*/

