//
//   $Id: IPCMD.C 1.4 2003/07/01 20:01:32Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// IPCMD.C
// IP commands
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

#if APL==0
#pragma option -zAINETINIT
#pragma option -zFINETINIT
#endif

#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <stdlib.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "internet.h"
#include "timer.h"
#include "netfunc.h"
#include "iface.h"
#include "arp.h"
#include "ip.h"
#include "cmdparse.h"

void dumproute(void);
void doicmpstat(void);

#define KEYBUFSZ	200
BYTE hostname[HOSTNAMELEN];
BYTE proxy_arp;

#ifdef INI_FILE
#include "ini.h"
void ini_ip(char *ini_file)	// initialise IP
{
	int netlen;

	GetPrivateProfileString("IP","Address","",hostname,
					HOSTNAMELEN,ini_file);
	netlen = GetPrivateProfileInt("IP","NetLen",-1,ini_file);
	ip_ttl = GetPrivateProfileInt("IP","TTL",ip_ttl,ini_file);
	if (*hostname == 0)
		return;
	set_ip_addr(hostname,netlen);
}

#define  ROUTENAMESIZE		20
#define  ROUTEINTERFACESIZE	20

typedef struct {
	char name[ROUTENAMESIZE];
	int netbits;
	int32 dest;
	char interface[ROUTEINTERFACESIZE];
	int32 gateway;
	int metric;
	int stat;
} RouteDesc;

void GetRouteDesc(RouteDesc *pRouteDesc,char *rtname,char *ini_file)
{
	char sz[80],szA[80];
	char r_rtname[40];
	char *pDest;

	if (strlen(rtname) >= ROUTENAMESIZE)
		rtname[ROUTENAMESIZE - 1] = 0;
	strcpy(pRouteDesc->name,rtname);
	sprintf(r_rtname,"Route:%s",rtname);
	GetPrivateProfileString("Route",rtname,"",sz,sizeof sz,ini_file);
	pDest = sz;
	if (GetPrivateProfileString(r_rtname,"Destination","",szA,80,ini_file) != 0)
		pDest = szA;
	pRouteDesc->dest = resolve(pDest);
	GetPrivateProfileString(r_rtname,"Interface","",szA,ROUTEINTERFACESIZE,ini_file);
	if (strlen(szA) >= ROUTEINTERFACESIZE)
		szA[ROUTEINTERFACESIZE - 1] = 0;
	strcpy(pRouteDesc->interface,szA);
	GetPrivateProfileString(r_rtname,"Gateway","",szA,80,ini_file);
	pRouteDesc->gateway = resolve(szA);
	pRouteDesc->netbits = GetPrivateProfileInt(r_rtname,"NetBits",0,ini_file);
	pRouteDesc->metric = GetPrivateProfileInt(r_rtname,"Metric",0,ini_file);
	pRouteDesc->stat =
		(GetPrivateProfileInt(r_rtname,"Private",0,ini_file) ? RT_PRIVATE : 0) |
		(GetPrivateProfileInt(r_rtname,"Static",0,ini_file) ? RT_STATIC : 0) |
		(GetPrivateProfileInt(r_rtname,"Proxy",0,ini_file) ? RT_PROXY : 0);
}

void ini_iproute(char *ini_file)	// initialise IP
{
	IFACE *ifp;
	char szKey[KEYBUFSZ];
	char *pszKey;
	int len;
	RouteDesc RouteDes;

	len = GetPrivateProfileString("Route",NULL,"",szKey,KEYBUFSZ,ini_file);
	if (len && *szKey) for (pszKey = szKey;pszKey < szKey + len;
														pszKey += (strlen(pszKey) + 1)) {
		GetRouteDesc(&RouteDes,pszKey,ini_file);
		if (!RouteDes.dest)
			RouteDes.netbits = 0;
		else if (RouteDes.netbits == 0) {
			RouteDes.netbits = 32;
			if (!(RouteDes.dest & 0x80000000l)) {	// class A
				if (!(RouteDes.dest & 0x00ffffffl))
					RouteDes.netbits = 8;
			}
			else if (!(RouteDes.dest & 0x40000000l)) {
				if (!(RouteDes.dest & 0x0000ffffl))
					RouteDes.netbits = 16;
			}
			else if (!(RouteDes.dest & 0x20000000l)) {
				if (!(RouteDes.dest & 0x000000ff))
					RouteDes.netbits = 24;
			}
		}
		for(ifp=ifaces;ifp != NULLIF;ifp = ifp->next){
			if(strcmp(RouteDes.interface,ifp->name) == 0)
				break;
		}
		if(ifp == NULLIF){
			printf("Interface \"%s\" unknown\n",RouteDes.interface);
			return;
		}
		rt_add(RouteDes.dest,RouteDes.netbits,RouteDes.gateway,
			RouteDes.metric,ifp,0,RouteDes.stat,0);
	}
}
#endif

int doipaddr(),doipstat(),dottl(),doipdrop(),doiptest();
char badhost[] = "Unknown host: %s\n";
struct scmds far ipcmds[] = {
	{"address",	doipaddr},
#ifdef TGW
	{"drop",		doipdrop},
#endif
#if APL
	{"status",	doipstat},
#endif
#ifdef TGW
	{"test",		doiptest},
#endif
	{"ttl",		dottl},
	{"",		NULLFP},
};

doip(int argc,char *argv[])
{
	return subcmd(ipcmds,argc,argv,0);
}

int doipaddr(int argc,char *argv[])
{
	char *inet_ntoa();
	IFACE *ifp;

	if(argc < 2) {
		printf(	"Hostname:\t%s\n",hostname);
		for (ifp = ifaces; ifp != NULLIF; ifp = ifp->next) {
			if (!(ifp->sflags & (SF_RAW | SF_UNC))) {
				printf("%s:\t\t%s\t",ifp->name,inet_ntoa(ifp->ip_addr));
				printf("Net mask:\t%s\n",inet_ntoa(NETMASK(ifp->netbits)));
			}
		}
		return 0;
	}
	else if (argc > 2) {
		printf("Changing interface IP address no longer allowed!\n");
		return -1;
	}
	return set_ip_addr(argv[1],-1);
}

#ifdef TGW
void show_test_val(void)
{
	if (ip_stats.drop_rand)
		printf("Drop 1 in %u to %u datagrams\n",
						ip_stats.drop_ratio ? ip_stats.drop_ratio : 1,
						ip_stats.drop_rand + ip_stats.drop_ratio);
	else if (ip_stats.drop_ratio)
		printf("Drop 1 in %u datagrams\n",ip_stats.drop_ratio);
	if (ip_stats.delay_rand)
		printf("Delay datagrams for %u to %u milli-seconds\n",
						ip_stats.delay * MSPTICK,
						(ip_stats.delay + ip_stats.delay_rand) * MSPTICK);
	else if (ip_stats.delay)
		printf("Delay datagrams for %u milli-seconds\n",
						ip_stats.delay * MSPTICK);
	if (ip_stats.thruput)
		printf("Limit thruput to %lu bytes/second\n",ip_stats.thruput);
}

int doipdrop(int argc,char *argv[])	// set drop ratio
{
	switch (argc) {
	case 3:
		ip_stats.drop_rand = atoi(argv[2]);
	case 2:
		ip_stats.drop_ratio = atoi(argv[1]);
	default:
		show_test_val();
	}
	return 0;
}

int doiptest(int argc,char *argv[])	// set drop ratio
{
	switch (argc) {
	case 4:
		ip_stats.thruput = atol(argv[3]);
	case 3:
		ip_stats.delay_rand = (atoi(argv[2]) + MSPTICK / 2) / MSPTICK;
	case 2:
		ip_stats.delay = (atoi(argv[1]) + MSPTICK / 2) / MSPTICK;
	default:
		show_test_val();
		break;
	}
	return 0;
}
#endif

int dottl(int argc,char *argv[])
{
	if(argc < 2)
		printf("%u\n",uchar(ip_ttl));
	else
		ip_ttl = atoi(argv[1]);
	return 0;
}

/* "route" subcommands */
int doadd(),dodrop(),doroute_echo();
static struct scmds far rtcmds[] = {
	{"add", doadd},
	{"drop", dodrop},
	{"echo", doroute_echo},
	{"", 0},
};

/* Display and/or manipulate routing table */
int doroute(int argc,char *argv[])
{
	if (argc < 2) {
#if APL
		dumproute();
#endif
		return 0;
	}
	return subcmd(rtcmds,argc,argv,0);
}

/* Add an entry to the routing table
 * E.g., "add 1.2.3.4 ec0 [/ec1] 5.6.7.8 3"
 */
int doadd(int argc,char *argv[])
{
	struct interface *ifp;
	int32 dest,gateway;
	WORD bits = 32;
	char *phost;
	char stat = 0;
	int metric,i;

	if(strcmp(argv[1],"default") == 0){
		dest = 0;
		bits = 0;
	}
	else if ((dest = resolve_nl(argv[1],&bits,&phost)) == 0) {
		printf(badhost,argv[1]);
		return 1;
	}
	for(ifp=ifaces;ifp != NULLIF;ifp = ifp->next){
		if(strcmp(argv[2],ifp->name) == 0)
			break;
	}
	if(ifp == NULLIF){
		printf("Interface \"%s\" unknown\n",argv[2]);
		return 1;
	}
	if(argc > 3 && strcmp(argv[3],"none") != 0){
		if((gateway = resolve(argv[3])) == 0){
			printf(badhost,argv[3]);
			return 1;
		}
	} else {
		gateway = 0;
	}
	if(argc > 4)
		metric = atoi(argv[4]);
	else
		metric = 0;
	for (i = 5;argc > i;++i) {
		if (strncmp(argv[i],"proxy",strlen(argv[i])) == 0) {
			stat |= RT_PROXY;
			proxy_arp = 1;
		}
		else if (strncmp(argv[i],"private",strlen(argv[i])) == 0) {
			stat |= RT_PRIVATE;
		}
		else if (strncmp(argv[i],"static",strlen(argv[i])) == 0) {
			stat |= RT_STATIC;
		}
	}
	return rt_add(dest,bits,gateway,metric,ifp,0,stat,0);
}

/* Drop an entry from the routing table
 * E.g., "drop 128.96/16
 */
#pragma argsused
int dodrop(int argc,char *argv[])
{
	WORD bits = 32;
	char *phost;
	int32 n;

	if(strcmp(argv[1],"default") == 0){
		n = 0;
		bits = 0;
	}
	else if ((n = resolve_nl(argv[1],&bits,&phost)) == 0) {
		printf(badhost,argv[1]);
		return 1;
	}
	return rt_drop(n,bits);
}

int doroute_echo(int argc,char *argv[])
{
	extern char route_echo;

	if (argc < 2) {
		printf("Route echo is %s\n",route_echo ? "on" : "off");
		return 0;
	}
	route_echo = !strcmp(argv[1],"on");
	return 0;
}
/* Dump IP routing table
 * Dest 			 Class	 Interface	  Gateway		   Metric
 * 192.001.002.003			 sl0		  192.002.003.004		4
 */
#if APL
void dumproute(void)
{
	register unsigned int i,bits;
	register struct route *rp;
	static char *class[] = {"Net A","Net B","Net C","Host"};

	printf("Destination     Class Interface    Gateway        Metric\n");
	for (bits = 0;bits <= 32;bits++){
		for (i=0;i<NROUTE;i++) {
			for(rp = routes[bits][i];rp != NULLROUTE;rp = rp->next){
				if (bits) {
					printf("%-16s",inet_ntoa(rp->target));
					if (bits % 8)
						printf("Net%02d ",bits);
					else
						printf("%-6s",class[bits / 8 - 1]);
				}
				else
					printf("%-22s","Default");
				printf("%-13s",rp->interface->name);
				/*
				if (rp->altif)
					printf(rp->priority ? "-%-13s" :
						"%-14s",rp->altif->name);
				else
					printf("%-14s","None");
				*/
				printf("%-16s%5u ",rp->gateway ?
					inet_ntoa(rp->gateway) : "",rp->metric);
				if (rp->timer.start) {
					printf("%u ",
						(WORD)((rp->timer.count *
									(long)MSPTICK + 500) / 1000));
				}
				printf("%s%s%s%s\n",rp->stat & RT_PROXY ? "X" : "",
								  rp->stat & RT_PRIVATE ? "P" : "",
								  rp->stat & RT_TRIG ? "T" : "",
								  rp->stat & RT_STATIC ? "S" : "");
			}
		}
	}
}

#pragma argsused
int doipstat(int argc,char *argv[])
{
	extern struct ip_stats ip_stats;
	extern struct reasm *reasmq;
	register struct reasm *rp;
	register struct frag *fp;
	char *inet_ntoa();

	printf("IP: Total %ld Runt %u LenErr %u VersErr %u",
		ip_stats.total,ip_stats.runt,ip_stats.length,ip_stats.version);
	printf(" ChecksumErr %u BadProtocol %u\n",
		ip_stats.checksum,ip_stats.badproto);

	if(reasmq != NULLREASM)
		printf("Reassembly fragments:\n");
	for(rp = reasmq;rp != NULLREASM;rp = rp->next){
		printf("Src %s",inet_ntoa(rp->source));
		printf(" Dest %s",inet_ntoa(rp->dest));
		printf(" ID %u Proto %u Time %u Len %u\n",
			rp->id,uchar(rp->protocol),rp->timer.count,rp->length);
		for(fp = rp->fraglist;fp != NULLFRAG;fp = fp->next){
			printf(" Offset %u Last %u\n",fp->offset,fp->last);
		}
	}
	doicmpstat();
	return 0;
}
#endif
