//
//   $Id: LPDD.C 1.5 2003/01/07 22:56:02Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// LPDD.C
// LPD server loosely following RFC 1179 (like everybody else)
// This one does direct printing without spooling
// Copyright (c) GP van Niekerk Ondernemings 1990-2002
// Revision history
// 2002-12-13 gvn Fix memory leak in LPD
// 2002-04-26 gvn /d=<delay count> to replace global /p=

#ifdef LPDD
#if APL==0
#pragma option -zAPRINTERS
#pragma option -zFPRINTERS
#endif

#include <dos.h>
#include <dir.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys\stat.h>
#include <time.h>
#include <alloc.h>
#include <io.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "timer.h"
#include "iface.h"
#include "icmp.h"
#include "netfunc.h"
#include "tcp.h"
#include "printing.h"
#include "log.h"
#include "task.h"

//===========================================================================
// Definitions
//===========================================================================
#define TRUE					1
#define FALSE					0
#define LPD_SESSIONS			4	// Maximum number of LPD sessions
#define BUFSIZE				200

//===========================================================================
// Global Variables
//===========================================================================
TCB		*ptcbLpdListen;			// Listen TCB

//===========================================================================
// Prototypes
//===========================================================================
void LpdRcvUpcall(TCB *tcb, unsigned int cnt);
void LpdStateUpcall(TCB *tcb, char old, char cNew);
void nprintf(PRINTCON *pPrintCon, char *format, ... );

extern void interrupt (*lptvec)(void);	/* Original interrupt vector [cs:ip] */
//===========================================================================
// Initialize LPDD here. Set scheduler etc.
//===========================================================================
int lpddserv1(int argc, char *argv[])
{
	struct socket 	lsocket;
	char 	*psz, *pszPrinterName = 0;
	char	szName[10];
	int	val;
	WORD	wPrinterNumber = 0;
	char 	cMask = 0xa8;
	char	cStatus = 0x80;
	WORD	wFlags = 0;			// form feed at end of job and cr/lf
	WORD	wDelay = 40;
	PRINTQUE	*pPrintQue;

	for (;argc > 1;argc--,argv++) {
		if (argv[1][0] == '/') {
			psz = argv[1] + (argv[1][2] == '=' ? 3 : 2);
			val = htoi(psz);
			switch (argv[1][1] | ('a' - 'A')) {
			case 'c':
				wFlags |= LPD_CR;
				break;
			case 'd':
				wDelay = atoi(psz);
				break;
			case 'f':
				wFlags |= LPD_FF;
				break;
			case 'm':
				cMask = val;
				break;
			case 'n':
				pszPrinterName = psz;
				break;
			case 's':
				cStatus = val;
				break;
			}
		}
		else
			wPrinterNumber = atoi(argv[1]);
	}
	lptvec = getvect(LPT);
	if (wPrinterNumber > NUMPRINT)
		wPrinterNumber = 0;
	pPrintQue = pque + wPrinterNumber;
	pPrintQue->mask = cMask;
	pPrintQue->stat = cStatus;
	pPrintQue->flags |= wFlags;
	if ((pPrintQue->wDelay = wDelay) == 0)
		++pPrintQue->wDelay;
	if (!pszPrinterName) {
		pszPrinterName = szName;
		strcpy(szName,"printer0");
		szName[7] = '0' + wPrinterNumber;
	}
	if ((pPrintQue->pszName = malloc(strlen(pszPrinterName) + 1,AllocLpdNm)) == 0)
		return -1;		// out of memory
   strcpy(pPrintQue->pszName,pszPrinterName);
	if (ptcbLpdListen)
		return 0;
	lsocket.address = 0;
	lsocket.port 	= 515;				// Default listening port for LPD
	ptcbLpdListen = open_tcp(&lsocket,// Local socket
			NULLSOCK,						// No remote socket
			TCP_SERVER,						// Type TCP_SERVER
			0,									// Receive window size
			LpdRcvUpcall,					// Upcall - receive
			0,									// Upcall - send
			LpdStateUpcall,				// Upcall - state
			0, 0);
	cPrintingUsed = 1;
	return 0;
}

//===========================================================================
// Shut down LPD server
//===========================================================================
#pragma argsused
lpddserv0(int argc,char *argv[])
{
	PRINTQUE	*pPrintQue;

	for(pPrintQue = pque; pPrintQue < pque + NUMPRINT; pPrintQue++) {
		if(pPrintQue->pszName != NULL) {
			free(pPrintQue->pszName); // Done with printer name
			pPrintQue->pszName = 0;
		}
	}
	close_tcp(ptcbLpdListen);
	return 0;
}

//===========================================================================
// Upcall comes here when connection status change in any way.
//===========================================================================
#pragma argsused
void LpdStateUpcall(TCB *tcb, char old, char cNew)
{
	PRINTCON *pPrintCon,*pPrintConQ,*pPrConPrev;
	PRINTQUE *pPrintQue;

	switch (cNew) {
	case ESTABLISHED:
		if((pPrintCon = (PRINTCON *)calloc(LPDCON_LEN, 1, AllocLpdCon)) == 0) {
			close_tcp(tcb);
		}
		else {
#if APL
			log(tcb,"open lpdd");
#endif
			pPrintCon->pTCB = tcb;
			tcb->user = (char *)pPrintCon;
		}
		return;
	case CLOSE_WAIT:
		if (tcb->r_upcall)
			close_tcp(tcb);		// not polling yet, close from our side
		break;
	case TIME_WAIT:
	case CLOSED:
		if ((pPrintCon = (PRINTCON *)tcb->user) != 0) {
			if ((pPrintQue = pPrintCon->pPrintQue) != 0) {
				// remove connection from printer queue
				pPrConPrev = 0;
				for (pPrintConQ = pPrintQue->pPrintCon;pPrintConQ;pPrintConQ = pPrintConQ->psNext) {
					if (pPrintConQ == pPrintCon) {
						goto found;
					}
					pPrConPrev = pPrintConQ;
				}
				goto not_found;
				found:
				if (pPrConPrev)
					pPrConPrev->psNext = pPrintCon->psNext;
				else {
					pPrintQue->pPrintCon = pPrintCon = pPrintCon->psNext;
               pPrintCon->pPrintQue = pPrintQue;
				}
				not_found:
			}
		}
		if (cNew == CLOSED) {
			free(tcb->user);
			del_tcp(tcb);
		}
		break;
	}
}

//===========================================================================
// Upcall comes here when there is data to receive
//===========================================================================
#pragma argsused
void LpdRcvUpcall(TCB *tcb, unsigned int cnt)
{
	MBUF	*bp;
	char cCh;
	PRINTCON *pPrintCon = (PRINTCON *)tcb->user,*pPrintConQ;
	PRINTQUE *pPrintQue;

	while (recv_tcp(tcb, &bp, 1, 0) > 0) {
		cCh = pullchar(&bp);
		switch(pPrintCon->wState) {
		case ST_CONNECTED:
			if (cCh >= LPD_START_PRINTING && cCh <= LPD_REMOVE_JOB) {
				pPrintCon->wState = cCh;
				break;
			}
			SendAck(pPrintCon,BAD_COMMAND);
			return;
		case ST_PRINT_JOBS:
		case ST_RECEIVE_JOB:
		case ST_STATUS_SHORT:
		case ST_STATUS_LONG:
		case ST_REMOVE_JOBS:
			if (pPrintCon->iIndex < PR_BUFLEN) {
				if (cCh != '\n' && cCh != ' ' && cCh != '\t') {
					pPrintCon->pcBuf[pPrintCon->iIndex++] = cCh;
					break;
				}
				pPrintCon->pcBuf[pPrintCon->iIndex] = 0;
				pPrintCon->iIndex = 0;
				for (pPrintQue = pque; pPrintQue < pque + NUMPRINT;++pPrintQue) {
					if (stricmp(pPrintQue->pszName,pPrintCon->pcBuf) == 0)
						goto found;
				}
				pPrintQue = pque;		// not found, use 0 as default
				found:
				switch (pPrintCon->wState) {
				case ST_RECEIVE_JOB:
					if ((pPrintConQ = pPrintQue->pPrintCon) != 0) {
						while (pPrintConQ->psNext)
							pPrintConQ = pPrintConQ->psNext;
						pPrintConQ->psNext = pPrintCon;
					}
					else {
						pPrintQue->pPrintCon = pPrintCon;
						pPrintCon->pPrintQue = pPrintQue;
					}
					tcb->r_upcall = 0;			// from now on we poll
					SendAck(pPrintCon,GOOD);
					print_flag = 1;				// activate print scheduler
					break;
				case ST_STATUS_SHORT:
				case ST_STATUS_LONG:
					if ((pPrintConQ = pPrintQue->pPrintCon) != 0) {
						char *pszStatus = "printing";
						while (pPrintConQ) {
							if (pPrintConQ->wState == ST_SPD) {
								nprintf(pPrintCon,"%s: Socket %s\r\n",
									pPrintQue->pszName,pszStatus);
							}
							else {
								nprintf(pPrintCon,"%s: Job %03u %s\r\n",
									pPrintQue->pszName,pPrintConQ->wJobNumber,pszStatus);
							}
							pszStatus = "queued";
							pPrintConQ = pPrintConQ->psNext;
						}
					}
					else
						nprintf(pPrintCon,"%s: No jobs waiting\r\n",pPrintQue->pszName);
					close_tcp(tcb);
					break;
				case ST_REMOVE_JOBS:
					nprintf(pPrintCon,"%s: LPD remove not yet implemented\r\n",
									pPrintQue->pszName);
					// fall-thru
				case ST_PRINT_JOBS:
					close_tcp(tcb);
					break;
				}
				break;
			}
			SendAck(pPrintCon,BAD_QLEN);
			return;
		}
	}
}

//===========================================================================
// Send a 1-byte acknowledge back. 0: Good, Nonzero = Bad
//===========================================================================
void SendAck(PRINTCON *pPrintCon, int type)
{
	MBUF 	*bp;

	if (pPrintCon->pTCB == 0)
		return;
	if ((bp = alloc_mbuf(1,AllocLpdSndAck)) != NULLBUF)
	{
		bp->cnt = 1;
		bp->data[0] = type;
		send_tcp(pPrintCon->pTCB, bp, 0);
	}
	if (type)
		close_tcp(pPrintCon->pTCB);
}

//===========================================================================
//
//===========================================================================
void nprintf(PRINTCON *pPrintCon, char *format, ... )
{
	va_list ap;
	char 	output_str[120];
	MBUF 	*bp;
	int		count;

	if (pPrintCon->pTCB == 0)
		return;

	va_start(ap, format);
	count = vsprintf(output_str, format, ap);
	va_end(ap);

	if ((bp = alloc_mbuf(count,AllocLpdNprintf)) != NULLBUF) {
		bp->cnt = count;
		memmove(bp->data, output_str, count);
		send_tcp(pPrintCon->pTCB, bp, 0);
	}
}

#endif
