//
//   $Id: MBUF.C 1.4 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// MBUF.C
// Memory buffer functions
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1991 Phil Karn, KA9Q

#define MBTRACE	0		/* set to trace mbufs */

#include <alloc.h>
#include <string.h>
#include "inetdefs.h"
#include "mbuf.h"

#if MBTRACE

typedef int (*PFUNC)();

struct hist {
	PFUNC caller;
	MBUF *mbuf;
	char type;
} mhist[64];

int mhi;

#endif

/* Allocate mbuf with associated buffer of 'size' bytes */
#ifdef ALLOC_TRACE
struct mbuf *Alloc_mbuf(WORD size, unsigned id)
#else
struct mbuf *Alloc_mbuf(WORD size)
#endif
{
	struct mbuf *bp;
	WORD size1;

	if((size1 = size + sizeof(struct mbuf)) < size)
		return NULLBUF;		// overflow!
	if((bp = (struct mbuf *)malloc(size1, id)) == NULLBUF)
		return NULLBUF;
	bp->next = bp->anext = NULLBUF;
	if(size != 0){
		bp->data = (BYTE *)(bp + 1);
	} else {
		bp->data = 0;
	}
	bp->cnt = 0;
#if MBTRACE
	mhist[mhi &= 63].caller = *(PFUNC *)(&size-2);
	mhist[mhi].mbuf = bp;
	mhist[mhi++].type = 'a';
#endif
	//printf("alloc-mbuf(%u %u) %p\n",size,size1,bp);
	return bp;
}

/* Free all resources associated with mbuf
 * Return pointer to next mbuf in packet chain
 */
struct mbuf *free_mbuf(struct mbuf *bp)
{
	register struct mbuf *bp1 = NULLBUF;

#if MBTRACE
	mhist[mhi &= 63].caller = *(PFUNC *)(&bp-2);
	mhist[mhi].mbuf = bp;
	mhist[mhi++].type = 'f';
#endif
	if(bp != NULLBUF){
		bp1 = bp->next;
		free((char *)bp);
	}
	return bp1;
}

/* Free packet (a chain of mbufs). Return pointer to next packet on queue,
 * if any
 */
struct mbuf *free_p(struct mbuf *bp)
{
	struct mbuf *abp;

	if(bp == NULLBUF)
		return NULLBUF;
	abp = bp->anext;
	while(bp != NULLBUF)
		bp = free_mbuf(bp);
	return abp;
}		

/* Free entire queue of packets (of mbufs) */
void free_q(struct mbuf **q)
{
	register struct mbuf *bp;

	while((bp = dequeue(q)) != NULLBUF)
		free_p(bp);
}

/* Count up the total number of bytes in an mbuf */
WORD len_mbuf(struct mbuf *bp)
{
	int cnt;

	cnt = 0;
	while(bp != NULLBUF){
		cnt += bp->cnt;
		bp = bp->next;
	}
	return cnt;
}

/* Count up the number of packets in a queue */
WORD len_q(struct mbuf *bp)
{
	register int cnt;

	for(cnt=0;bp != NULLBUF;cnt++,bp = bp->anext)
		;
	return cnt;
}

/* Trim mbuf to specified length by lopping off end */
void trim_mbuf(struct mbuf **bpp,WORD length)
{
	register WORD tot = 0;
	register struct mbuf *bp;

	if(bpp == NULLBUFP || *bpp == NULLBUF)
		return;	/* Nothing to trim */

	if(length == 0){
		/* Toss the whole thing */
		free_p(*bpp);
		*bpp = NULLBUF;
		return;
	}
	/* Find the point at which to trim. If length is greater than
	 * the packet, we'll just fall through without doing anything
	 */
	for( bp = *bpp; bp != NULLBUF; bp = bp->next){
		if(tot + bp->cnt < length){
			tot += bp->cnt;
		} else {
			/* Cut here */
			bp->cnt = length - tot;
			free_p(bp->next);
			bp->next = NULLBUF;
			break;
		}
	}
}

/* Duplicate/enqueue/dequeue operations based on mbufs */

/* Duplicate first 'cnt' bytes of packet starting at 'offset'.
 * This is done without copying data; only the headers are duplicated,
 * but without data segments of their own. The pointers are set up to
 * share the data segments of the original copy. The return pointer is
 * passed back through the first argument, and the return value is the
 * number of bytes actually duplicated.
 */
#ifdef ALLOC_TRACE
WORD Dup_p(struct mbuf **hp,struct mbuf *bp,WORD offset,WORD cnt, unsigned id)
#else
WORD Dup_p(struct mbuf **hp,struct mbuf *bp,WORD offset,WORD cnt)
#endif
{
	register struct mbuf *cp;
	WORD tot;

	if(cnt == 0 || bp == NULLBUF || hp == NULLBUFP){
		if(hp != NULLBUFP)
			*hp = NULLBUF;
		return 0;
	}
	if((*hp = cp = alloc_mbuf(0,id)) == NULLBUF){
		return 0;
	}
	/* Skip over leading mbufs that are smaller than the offset */
	while(bp != NULLBUF && bp->cnt <= offset){
		offset -= bp->cnt;
		bp = bp->next;
	}
	if(bp == NULLBUF){
		free_mbuf(cp);
		*hp = NULLBUF;
		return 0;	/* Offset was too big */
	}
	tot = 0;
	for(;;){
		cp->data = bp->data + offset;
		cp->cnt = min(cnt,bp->cnt - offset);
		offset = 0;
		cnt -= cp->cnt;
		tot += cp->cnt;
		bp = bp->next;
		if(cnt == 0 || bp == NULLBUF || (cp->next = alloc_mbuf(0,id)) == NULLBUF)
			break;
		cp = cp->next;
	}
	return tot;
}

/* Copy first 'cnt' bytes of packet into a new, single mbuf */
#ifdef ALLOC_TRACE
struct mbuf *Copy_p(struct mbuf *bp,WORD cnt, unsigned id)
#else
struct mbuf *Copy_p(struct mbuf *bp,WORD cnt)
#endif
{
	register struct mbuf *cp;
	BYTE *wp;
	register WORD n;

	if(bp == NULLBUF || cnt == 0 || (cp = alloc_mbuf(cnt,id)) == NULLBUF)
		return NULLBUF;
	wp = cp->data;
	while(cnt != 0 && bp != NULLBUF){
		n = min(cnt,bp->cnt);
		memcpy(wp,bp->data,n);
		wp += n;
		cp->cnt += n;
		cnt -= n;
		bp = bp->next;
	}
	return cp;
}

/* Copy and delete "cnt" bytes from beginning of packet. Return number of
 * bytes actually pulled off
 */
WORD pullup(struct mbuf **bph,BYTE *buf,WORD cnt)
{
	register struct mbuf *bp;
	WORD n,tot;

	tot = 0;
	if(bph == NULLBUFP)
		return 0;
	while(*bph != NULLBUF && cnt != 0){
		bp = *bph;
		n = min(cnt,bp->cnt);
		if(buf != 0 && n != 0){
			memcpy(buf,bp->data,n);
			buf += n;
		}
		tot += n;
		cnt -= n;
		bp->data += n;
		bp->cnt -= n;		
		if(bp->cnt == 0){
			*bph = free_mbuf(bp);
		}
	}
	return tot;
}

/* Append mbuf to end of mbuf chain */
int append(struct mbuf **bph,struct mbuf *bp)
{
	register struct mbuf *p;
	int qcnt = 1;

	if(bph == NULLBUFP || bp == NULLBUF)
		return 0;
	if(*bph == NULLBUF){
		/* First one on chain */
		*bph = bp;
	} else {
		++qcnt;
		for(p = *bph ; p->next != NULLBUF ; p = p->next)
			++qcnt;
		p->next = bp;
	}
	return qcnt;
}

/* Append packet to end of packet queue */
void enqueue(struct mbuf **q,struct mbuf *bp)
{
	register struct mbuf *p;

	if(q == NULLBUFP || bp == NULLBUF)
		return;
	idisable();
	if(*q == NULLBUF){
		/* List is empty, stick at front */
		*q = bp;
	} else {
		for(p = *q ; p->anext != NULLBUF ; p = p->anext)
			;
		p->anext = bp;
	}
	popf();
}

/* Unlink a packet from the head of the queue */
struct mbuf *dequeue(struct mbuf **q)
{
	register struct mbuf *bp;

	if(q == NULLBUFP)
		return NULLBUF;
	idisable();
	if((bp = *q) != NULLBUF){
		*q = bp->anext;
		bp->anext = NULLBUF;
	}
	popf();
	return bp;
}	

/* Copy user data into an mbuf */
#ifdef ALLOC_TRACE
struct mbuf *Qdata(BYTE *data,WORD cnt, unsigned id)
#else
struct mbuf *Qdata(BYTE *data,WORD cnt)
#endif
{
	register struct mbuf *bp;

	if((bp = alloc_mbuf(cnt,id)) == NULLBUF)
		return NULLBUF;
	memcpy(bp->data,data,cnt);
	bp->cnt = cnt;
	return bp;
}

/* Copy mbuf data into user buffer */
WORD dqdata(struct mbuf *bp,BYTE *buf,unsigned cnt)
{
	unsigned n,tot;
	struct mbuf *bp1;

	if(buf == 0)
		return 0;
	
	tot = 0;
	for(bp1 = bp;bp1 != NULLBUF; bp1 = bp1->next){
		n = min(bp1->cnt,cnt);
		memcpy(buf,bp1->data,n);
		cnt -= n;
		buf += n;
		tot += n;
	}
	free_p(bp);
	return tot;
}

/* Pull a 32-bit integer in host order from buffer in network byte order */
DWORD pull32(struct mbuf **bpp)
{
/* machine independant
	int32 rval;
	char buf[4];
	register char *cp;

	if(pullup(bpp,buf,4) != 4){
		/* Return zero if insufficient buffer */
		return 0;
	}
	cp = buf;

	/* Unwound for speed */
	rval = uchar(*cp++);
	rval <<= 8;
	rval |= uchar(*cp++);
	rval <<= 8;
	rval |= uchar(*cp++);
	rval <<= 8;
	rval |= uchar(*cp);
	return rval;
*/
/* fast PC specific */
	int32 rval;
	BYTE buf[4];

	if(pullup(bpp,buf,4) != 4){
		/* Return zero if insufficient buffer */
		return 0;
	}
	((char *)&rval)[3] = buf[0];
	((char *)&rval)[2] = buf[1];
	((char *)&rval)[1] = buf[2];
	((char *)&rval)[0] = buf[3];
	return rval;
}

/* Pull a 16-bit integer in host order from buffer in network byte order */
WORD pull16(struct mbuf **bpp)
{
/* machine independant
	WORD rval;
	char buf[2];
	register char *cp;

	if(pullup(bpp,buf,2) != 2){
		/* Return zero if insufficient buffer */
		return 0;
	}
	cp = buf;

	rval = uchar(*cp++);
	rval <<= 8;
	rval |= uchar(*cp);
*/
/* pc specific */
	WORD rval;
	BYTE buf[2];

	if(pullup(bpp,buf,2) != 2){
		/* Return zero if insufficient buffer */
		return 0;
	}
	((char *)&rval)[1] = buf[0];
	((char *)&rval)[0] = buf[1];

	return rval;
}

/* Pull single character from mbuf */
int pullchar(struct mbuf **bpp)
{
	int c;

	if(pullup(bpp,(BYTE *)&c,1) != 1)
		/* Return -1 if nothing left */
		return -1;
	return c & 0xff;
}

/* Put a long in host order into a char array in network order */
BYTE *put32(BYTE *cp,DWORD x)
{
/* nice machine independant code
	*cp++ = x >> 24;
	*cp++ = x >> 16;
	*cp++ = x >> 8;
	*cp++ = x;
	return cp;
*/
/* nasty fast PC specific code */
	char *xp = (char *)&x;

	cp[0] = xp[3];
	cp[1] = xp[2];
	cp[2] = xp[1];
	cp[3] = xp[0];

	return cp + 4;
}

/* Put a short in host order into a char array in network order */
BYTE *put16(BYTE *cp,WORD x)
{
/* nice machine independant code
	*cp++ = x >> 8;
	*cp++ = x;
	return cp;
*/
/* nasty fast PC specific code */
	char *xp = (char *)&x;

	cp[0] = xp[1];
	cp[1] = xp[0];

	return cp + 2;
}
