//
//   $Id: NETFUNC.C 1.4 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// NETFUNC.C
// Network functions
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

#include <ctype.h>
#include <stdio.h>
#include <alloc.h>
#include <stdlib.h>
#include <string.h>
#include "inetdefs.h"
#include "netfunc.h"
#if APL
#include "twindow.h"
#include "keys.h"
#endif

int findhost(char *hostent,char *host);

int net_error;

#define LINELEN 256

/* Convert Internet address in ascii dotted-decimal format (44.0.0.1) to
 * binary IP address
 */
DWORD aton(register char *s)
{
	int32 n;
	register int i;

	n = 0;
	for(i=24;i>=0;i -= 8){
		n |= (int32)atoi(s) << i;
		if((s = strchr(s,'.')) == NULLCHAR)
		break;
		s++;
	}
	return n;
}

static struct {
	char *name;
	int32 address;
} cache;

/* Resolve a host name into an IP address. IP addresses in dotted-decimal
 * notation are distinguished from domain names if it starts with a decimal
 * and have at least one '.' in it eg 192.1.1.6
 * host_term is set to name of terminal emulator (/name), if found
 * host_port is set to port of terminal emulator (/p=port)
 */
#define HTSIZE 16		/* size of terminal emulator name */
char host_term[HTSIZE];		/* terminal emulator name */
int host_port;

add_cache(char *name,int len,int32 addr)
{
	if(cache.name != NULLCHAR)
		free(cache.name);
	if ((cache.name = malloc(len + 1,AllocNetfIpA)) != 0) {
		strncpy(cache.name,name,len);
		cache.name[len] = 0;
	}
	cache.address = addr;
	return;
}

int is_dotted_decimal(char *host)
{
	char *ph,digitseen = 0,num_dot = 0;

	for (ph = host,num_dot = 0;;++ph) {
		if (*ph == 0) {
			if (digitseen && num_dot == 3) {
				return 1;
			}
			break;
		}
		else if (*ph == '.') {
			if (!digitseen)
				break;
			digitseen = 0;
			++num_dot;
		}
		else if (isdigit(*ph)) {
			digitseen = 1;
		}
		else
			break;
	}
	return 0;
}

DWORD resolve(char *host)
{
	int32 hostnum = 0;
	char hostent[LINELEN];
	FILE *sfile;

	if (!*host)
		return 0;
	if (*host == '[') {
		/* Brackets indicate IP address in dotted-decimal form */
		*host_term = 0;
		return aton(host + 1);
	}
	if (is_dotted_decimal(host) != 0) {
		*host_term = 0;
		return aton(host);
	}
	if(cache.name != NULLCHAR && strcmp(cache.name,host) == 0)
		return cache.address;

	/* Not a numerical IP address, search the host table */
	if((sfile = fopen(getfilename(hostent,hosts,0),"r")) == NULL){
		goto trydns;
	}
	while (fgets(hostent,LINELEN,sfile) != 0) {
		if (findhost(hostent,host)) {
			hostnum = cache.address;
			break;
		}
	}
	/* No address found */
	fclose(sfile);
	if (!hostnum) {
		trydns:
		if ((hostnum = resolve_dns(host)) == 0)
			return 0;
		*host_term = 0;
		add_cache(host,strlen(host),hostnum);
	}
	return hostnum;
}

int findhost(char *hostent,char *host)
{
	register char *cp,*cp1,*cp2;
	int i;
	char *names;

	host_port = 0;
	if (*hostent == '*')
		++hostent;
	rip(hostent);
	if ((cp = strchr(hostent,'#')) != 0)
		*cp = 0;
	cp = hostent;
	if(!(*cp))
		return 0;		/* Comment or invalid line */
	while(!isspace(*cp))
		++cp;
	*cp++ = 0;
	names = cp;
	while(cp != NULLCHAR){
		/* Skip white space */
		while(*cp == ' ' || *cp == '\t')
			cp++;
		if(*cp == '\0')
			break;
		/* Look for next token, find length of this one */
		cp2 = strchr(cp,' ');
		if ((cp1 = strchr(cp,'\t')) == 0 && cp2 == 0) {
			i = strlen(cp);
		}
		else {
			if (cp1 == 0 || (cp2 && cp1 && cp2 < cp1))
				cp1 = cp2;
			i = cp1 - cp;
		}
		if (i > 29)
			i = 29;
		if(!host || (strlen(host) == i && strnicmp(host,cp,i) == 0)) {
			long ip;
			/* Found it, put in cache */
			if (is_dotted_decimal(hostent) != 0)
				ip = aton(hostent);
			else if ((ip = resolve_dns(hostent)) == 0)
				return 0;
			add_cache(cp,i,ip);
			*host_term = 0;
			cp = names;
			while ((cp = strchr(cp,'/')) != NULLCHAR) {
				if ((cp[1] | ('a' - 'A')) == 'p' && cp[2] == '=') {
					host_port = atoi(cp += 3);
				}
				else {
					for (++cp,cp1 = host_term;
						*cp && *cp != '\t' && *cp != ' ' &&
						cp1 < host_term + HTSIZE - 1;
								++cp,++cp1) {
						*cp1 = *cp;
					}
					*cp1 = 0;
				}
			}
			return 1;
		}
		/* That one didn't match, try the next one */
		cp = cp1;
	}
	return 0;
}
#if APL
#define NUMNAMES 15

extern char enter_col[];
FILE *sfile;

static int last_num = -1;

char *get_host_line(char *hostent,int num)
{
	char *line = hostent + 7;

	if (num <= last_num) {
		rewind(sfile);
		last_num = -1;
	}
	while (last_num < num) {
		if (fgets(line,72,sfile) == 0)
			return 0;
		if (*line != '#' && *line != '*')
				++last_num;
	}
	line[strlen(line) - 1] = 0;
	sprintf(hostent,"%-2u ",num);
	strcat(hostent,line);
	return hostent;
}

char *host_select(int x,int y)
{
	static int csel;
	char str[80];
	char *retv = 0,*p;
	int i;

	if((sfile = fopen(getfilename(str,hosts,0),"r")) == NULL){
		return 0;
	}
	for (i = 0;get_host_line(str,i);++i);
	if (window_select(x,y," Hosts ",&csel,i,70,NUMNAMES,0,get_host_line) != ESC) {
		get_host_line(str,csel);
		if ((p = strchr(str,' ')) != 0) {
			if (*++p == ' ')
				++p;
			findhost(p,0);
			retv = cache.name;
		}
	}
	fclose(sfile);
	return retv;
}

/* window based host selection - old implementation
char *host_select(int x,int y)
{
	WINDOW *wnd;
	char hostent[LINELEN];
	FILE *sfile;
	int i,j;
	static char selstr[] = "0123456789abcdefghijklmnop";
	long fptrs[NUMNAMES];
	static int chost = 1;
	int ckey;
	char *retv = 0;
	char eof = 0;

	if((sfile = fopen(getfilename(hostent,hosts),"r")) == NULL){
		return 0;
	}
	wnd = establish_window(x,y,NUMNAMES + 2,70);
	set_title(wnd," Hosts ");
	set_wcol(wnd, enter_col);
	set_border(wnd, 1);
	display_window(wnd);
	while (1) {
		clear_window(wnd);
		WCURS =
		SCROLL =
		i = 0;
		while (i < NUMNAMES) {
			if (feof(sfile)) {
				eof = 1;
				break;
			}
			fptrs[i] = ftell(sfile);
			*hostent = 0;
			fgets(hostent,LINELEN,sfile);
			if (*hostent != '#' && *hostent != '*' &&
							isdigit(*hostent))
				++i;		// looks valid
		}
		for (j = 0;j < i;++j) {
			fseek(sfile,fptrs[j] , 0);
			fgets(hostent,LINELEN,sfile);
			hostent[strlen(hostent) - 1] = 0;
			wprintf(wnd,"%c %.66s%s",selstr[j],hostent,
				j < i - 1 ? "\n" : "");
		}
		ignore:
		ckey = get_selection(wnd,&chost,0,selstr);
		if (ckey == ESC) {
			break;
		}
		if (ckey == PGUP) {
			chost = 1;
			rewind(sfile);
			eof = 0;
			continue;
		}
		else if (ckey == PGDN) {
			if (eof) {
				chost = i;
				goto ignore;
			}
			chost = 1;
			continue;
		}
		else if (chost <= NUMNAMES) {
			fseek(sfile,fptrs[chost - 1] , 0);
			fgets(hostent,LINELEN,sfile);
			findhost(hostent,0);
			retv = cache.name;
			break;
		}
	}
	fclose(sfile);
	delete_window(wnd);
	return retv;
}
*/

#endif
/* Convert an internet address (in host byte order) to a dotted decimal ascii
 * string, e.g., 255.255.255.255\0
 */
char *inet_ntoa(DWORD a)
{
	static char buf[16];

	sprintf(buf,"%u.%u.%u.%u",
		hibyte(hiword(a)),
		lobyte(hiword(a)),
		hibyte(loword(a)),
		lobyte(loword(a)) );
	return buf;
}
/* Convert a socket (address + port) to an ascii string of the form
 * aaa.aaa.aaa.aaa:ppppp
 */
char *psocket(struct socket *s)
{
	static char buf[30];

	sprintf(buf,"%s:%u",inet_ntoa(s->address),s->port);
	return buf;
}
/* Convert hex-ascii string to long integer */
long htol(char *s)
{
	long ret;
	char c;

	ret = 0;
	while ((c = *s++) != '\0') {
		c &= 0x7f;
		if(c >= '0' && c <= '9')
			ret = ret*16 + (c - '0');
		else if(c >= 'a' && c <= 'f')
			ret = ret*16 + (10 + c - 'a');
		else if (c >= 'A' && c <= 'F')
			ret = ret*16 + (10 + c - 'A');
		else
			break;
	}
	return ret;
}
