//
//   $Id: PKTDRVI.C 1.7 2003/08/16 19:12:08Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// PKTDRVI.C
// Copyright (C) 2001 Datalight, Inc.
// All Rights Reserved
// Datalight Confidential and Proprietary
// Portions Copyright (c) GPvNO 1990-2001
//
// Packet driver init functions
//
// Revision History
// 08/16/2003 gvn improve interrupt vector searching
// 07/23/2001 gvn Allocation tracing
// 05/29/2001 gvn Search for PDR int if not given or 0.
//						Search for h/w int if not given or 0.

#if APL==0
#pragma option -zAINETINIT
#pragma option -zFINETINIT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <alloc.h>
#include <dos.h>
#include <string.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "iface.h"
#include "timer.h"
#include "arp.h"
#include "pktdrv.h"
#include "enet.h"
#include "bootp.h"
#include "ip.h"
typedef char far * fptr;

#define NPDRMAX 4			/* max number of packet drivers */
extern unsigned npdr;		/* number of PACKET DRIVER adaptors */

#ifndef __LARGE__
extern unsigned pstack_start;
#endif
extern char *obuf;			/* output buffer for send */
extern unsigned obufsize;	// size of output buffer
extern void far (*receiver[])();
#if !APL
extern void interrupt (*pdr_schedtab[])();
#endif
extern PDCB pktcb[4];
void pdr_stop(IFACE *ifp);
pdr_ioctl(struct interface *ifp,int argc,char *argv[]);
pdr_raw(struct interface *ifp,struct mbuf *bp);
void slp_send();
void pdr_recv(IFACE *ifp);
void slp_recv(IFACE *ifp);

enum {
	A_HWTYPE,
	A_IFNAME,
	A_MODE,
	A_MTU,
	A_BUFLIM,
	A_IOAD,
	A_VEC,
};

/* Attach a PACKET driver to the system
 * argv[0]: must be 'pdr'
 * argv[1]: interface label, e.g., "ec0"
 * argv[2]: mode (dix,token,slip,ieee,pdr)
 * argv[3]: maximum transmission unit, bytes, e.g., "1500"
 * argv[4]: maximum number of packets allowed on receive queue, e.g., "5"
 * argv[5]: interrupt in hex
 * argv[6]: hardware vector (optional)
 */
struct pdr_s {
	char *name;
	char class;
};
struct pdr_s pdr_class[] = {
	{"dix",CLASS_DIX},
	{"token",CLASS_8025},
	{"slip",CLASS_SLIP},
	{"ieee",CLASS_8023},
	{"",0},
};

extern BYTE bMultiCast;

#define DOSINT 0x21

typedef struct mcb_s {
	char id;
	unsigned psp;
	unsigned len;
	char junk[3];
	char name[8];
	char env[16];	// next segment if environment
} MCB;


pdr_setup(char *ifname,char *ifclass,int mtu,int buflim,int pdrint,int iHwIntNum)
{
	register struct interface *ifp;
	extern struct interface *ifaces;
	struct pdr_s *ppc;
	void ec_send();
	int ec_output();
	PDCB *ppdcb;
#ifndef __LARGE__
	void *pstack;
#endif
	int i;
	static int intno = 0x20;
	char far *id;
	static char pid[] = "PKT DRVR";
	struct REGPACK regs;
	char if_class,if_number;
	int if_type;
	int dixoff = 6;
	int arp_hwtype = ARP_ETHER;
	static char arp_type[8] = {170,170,3,0,0,0,ARP_TYPE >> 8,
							ARP_TYPE & 0xff};
	static char ip_type[8] = {170,170,3,0,0,0,IP_TYPE >> 8,
						  IP_TYPE & 0xff};

	if (npdr >= NPDRMAX) {
		printf("Too many Packet drivers\n");
		return -1;
	}
	if (pdrint)
		intno = pdrint;
	for (;pdrint || intno < 0x100;++intno) {
		id = *(fptr far *)MK_FP(0,4 * intno);
		for (i = 0;i < 9;++i) {
			if (id[i + 3] != pid[i]) {
				if (!pdrint)
					break;
				else {
					printf("No packet driver at 0x%x\n",intno);
					return -1;
				}
			}
		}
		if (i == 9)
			goto found;
	}
	printf("Packet driver not found\n");
	return -1;
	found:
	if((ifp = (IFACE *)calloc(1,sizeof(struct interface),AllocPktiIf)) == 0 ||
		(ifp->name = malloc((unsigned)strlen(ifname) + 1,AllocPktiIfNm)) == 0 ||
		(ifp->hwaddr = malloc(EADDR_LEN,AllocPktiMac)) == 0) {
		goto nomem;
	}
	ppdcb = pktcb + npdr;
	ppdcb->ifp = ifp;
	ifp->dcb = ppdcb;
	ifp->dev = intno;
	for (ppc = pdr_class;ppc->class && strcmp(ifclass,ppc->name);
									++ppc);
	if_class = ppc->class;
	regs.r_ax = (DRIVER_INFO << 8) + 0xff;
	intr(intno,&regs);
	if (!if_class)
		if_class = regs.r_cx >> 8;
	if_type = regs.r_dx;
	if_number = regs.r_cx;
	ifp->mtu = mtu;
	if ((if_class == CLASS_DIX) && ifp->mtu > 1500)
		ifp->mtu = 1500;
	if ((if_class == CLASS_8023) && ifp->mtu > 1492)
		ifp->mtu = 1492;
#ifndef __LARGE__
	if (!npdr) {
		if ((pstack = malloc(100,AllocPktiStack)) == 0) {
			goto nomem;
		}
		pstack_start = (unsigned)pstack + 98;
	}
#endif
	if (obufsize < ifp->mtu + IEEE_LEN) {
		if (obuf)
			free(obuf);
		if ((obuf = malloc(ifp->mtu + IEEE_LEN,AllocPktiSbuf)) == 0) {
			nomem:
			printf("Not enough memory!\n");
			return -1;
		}
		obufsize = ifp->mtu + IEEE_LEN;
	}
	if (if_class == CLASS_DIX ||
		(arp_hwtype = ARP_IEEE,dixoff = 0,if_class == CLASS_8023) ||
						(if_class == CLASS_8025)) {
		regs.r_ax = (ACCESS_TYPE << 8) + if_class;
		regs.r_bx = if_type;
		regs.r_dx = if_number;
		regs.r_cx = 8 - dixoff;			/* typelen */
		regs.r_si = FP_OFF(arp_type + dixoff);
		regs.r_ds = FP_SEG(arp_type);
		regs.r_es = FP_SEG(receiver[npdr]);
		regs.r_di = FP_OFF(receiver[npdr]);
		intr(intno,&regs);
		if (regs.r_flags & CF) {
			goto acc_err;
		}
		ppdcb->arphandle = regs.r_bx = regs.r_ax;
		if (bMultiCast) {
			// set receive mode to include multicast
			regs.r_ax = SET_RCV_MODE << 8;
			regs.r_cx = 5;	// receive multicast
			intr(ifp->dev,&regs);
			if (regs.r_flags & CF) {
				printf("Warning! - Packet driver does not support Multicast (%d)\n",
							regs.r_dx >> 8);
			}
		}
		regs.r_cx = 8 - dixoff;			/* typelen */
	}
	else {
		arp_hwtype = 0;
		regs.r_cx = 0;
	}
	regs.r_ax = (ACCESS_TYPE << 8) + if_class;
	regs.r_bx = if_type;
	regs.r_dx = if_number;
	regs.r_si = FP_OFF(ip_type + dixoff);
	regs.r_ds = FP_SEG(ip_type);
	regs.r_es = FP_SEG(receiver[npdr]);
	regs.r_di = FP_OFF(receiver[npdr]);
	intr(intno,&regs);
	if (regs.r_flags & CF) {
		acc_err:
		printf("'access_type' error %u\n",regs.r_dx >> 8);
		return -1;
	}
	ppdcb->iphandle = regs.r_bx = regs.r_ax;
	// get MAC address
	regs.r_bx = ppdcb->iphandle;
	regs.r_ax = GET_ADDRESS << 8;
	regs.r_cx = EADDR_LEN;
	regs.r_es = FP_SEG(ifp->hwaddr);
	regs.r_di = FP_OFF(ifp->hwaddr);
	intr(intno,&regs);
	strcpy(ifp->name,ifname);
	ifp->stop = pdr_stop;
	ifp->ioctl = pdr_ioctl;
	if (!iHwIntNum) {	// try to find h/w vec
	  MCB far * pmcb;
	  unsigned mcbseg;
	  int iIntNum;
	  DWORD dwStart,dwEnd,dwVect,dwHwVect;
	  void interrupt (*v)();

		_AH = 0x52;
		geninterrupt(DOSINT);
		mcbseg = _ES;
		mcbseg = peek(mcbseg, _BX - 2);
		pmcb = MK_FP(mcbseg,0);
		v = getvect(intno);
		dwVect = ((DWORD)FP_SEG(v) << 4) + FP_OFF(v);
		while (1) {
			dwStart = (DWORD)FP_SEG(pmcb) << 4;
			dwEnd = dwStart + ((DWORD)pmcb->len << 4);
			if (dwStart <= dwVect && dwVect < dwEnd) {
				for (iIntNum = 10;iIntNum < 0x80;++iIntNum) {
					if (iIntNum == 0x10)
						iIntNum = 0x70;
					v = getvect(iIntNum);
					dwHwVect = ((DWORD)FP_SEG(v) << 4) + FP_OFF(v);
					if (dwStart <= dwHwVect && dwHwVect < dwEnd) {
						iHwIntNum = iIntNum;
						goto hw_found;
					}
				}
			}
			if (pmcb->id == 'Z')
				break;
			else if (pmcb->id != 'M') {
				printf("DOS memory corruption\n");
				return -1;
			}
			FP_SEG(pmcb) += pmcb->len + 1;
		}
	}
	else
		iHwIntNum += (iHwIntNum >= 8 ? 0x68 : 8);
	if (iHwIntNum) {
		hw_found:
		/* Save original interrupt vector */
		ppdcb->oldvec = getvect(iHwIntNum);
		ppdcb->vec = iHwIntNum;
		/* Set interrupt vector */
#if !APL
		setvect(iHwIntNum,pdr_schedtab[npdr]);
#endif
		iHwIntNum -= (iHwIntNum > 0x10 ? 0x68 : 8);
	}
	else
		iHwIntNum = -1;
	printf("%s: Packet driver at 0x%x using h/w int %d\n",ifp->name,intno,iHwIntNum);
	/* Initialize device */
	ppdcb->rcvmax = buflim;
	if_add(ifp);
	++npdr;
	if (arp_hwtype) {
		ifp->send = ec_send;
		ifp->output = ec_output;
		ifp->raw = pdr_raw;
		ifp->recv = pdr_recv;
		if (if_class == CLASS_8023)
			ifp->flags |= IEEE_FLAG;
		else if (if_class == CLASS_8025)
			ifp->flags |= TOKEN_FLAG;
		if ((arp_hwtype == ARP_ETHER || arp_hwtype == ARP_IEEE) &&
						!(ifp->flags & TOKEN_FLAG)) {
			ifp->flags |= ETH_FLAG;
		}
		ifp->hwtype = arp_hwtype;
		ifp->hwalen = EADDR_LEN;
		if ((ifp->ip_addr & 0xfffffffeL) == 0L) {
			ifp->dhcp_state = ifp->ip_addr ? D_INIT : D_BOOTP;
			ifp->ip_addr = 0;
			bootp(ifp);
		}
		else {
			arp_init(ifp,arp_hwtype,EADDR_LEN,IP_TYPE,ARP_TYPE,ether_bdcst,
							pether,gaether);
		}
	}
	else {
		ifp->send = slp_send;
		ifp->recv = slp_recv;
		ifp->hwtype = ARP_SERIAL;
	}

	return 0;
}

pdr_iface(int argc,char *argv[])
{
	return pdr_setup(argv[A_IFNAME],argv[A_MODE],atoi(argv[A_MTU]),
		atoi(argv[A_BUFLIM]),argc > A_IOAD ? htoi(argv[A_IOAD]) : 0,
		argc > A_VEC ? atoi(argv[A_VEC]) : 0);
}

#ifdef INI_FILE
void pdr_ini(InterfaceDesc *pif)
{
	pdr_setup(pif->name,pif->Class,pif->mtu,pif->numbuf,pif->intvec,0);
}
#endif
