//
//   $Id: RIP.C 1.4 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// RIP.C

/* Implements RIP v 1 according to RFC 1058
 * RIP v2 according to RFC 1723

  The first four octets of a RIP message contain the RIP header.  The
   remainder of the message is composed of 1 - 25 route entries (20
   octets each).  The new RIP message format is:

	0                   1                   2                   3 3
	0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   | Command (1)   | Version (1)   |           unused              |
   +---------------+---------------+-------------------------------+
   | Address Family Identifier (2) |        Route Tag (2)          |
   +-------------------------------+-------------------------------+
   |                         IP Address (4)                        |
   +---------------------------------------------------------------+
   |                         Subnet Mask (4)                       |
   +---------------------------------------------------------------+
   |                         Next Hop (4)                          |
   +---------------------------------------------------------------+
   |                         Metric (4)                            |
   +---------------------------------------------------------------+
 */

/* This file contains code to implement the Routing Information Protocol (RIP)
 * and is derived from 4.2BSD code. Mike Karels of Berkeley has stated on
 * TCP-IP that the code may be freely used as long as UC Berkeley is
 * credited. (Well, here's some credit :-). AGB 4-28-88

 * Further documentation on the RIP protocol is now available in Charles
 * Hedrick's draft RFC, as yet unnumbered. AGB 5-6-88
 *
 * The RIP RFC has now been issued as RFC1058. AGB 7-23-88
 *
 * Code gutted and substantially rewritten. KA9Q 9/89
 * Adapted for INET by GvN 1991
 */
#if APL==0
#pragma option -zFINETINIT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "netfunc.h"
#include "udp.h"
#include "timer.h"
#include "iface.h"
#include "ip.h"
#include "internet.h"
#include "rip.h"
#include "arp.h"
#include "cmdparse.h"
#include "log.h"

extern void *logfp;
char rip_advertise;
char rip_use;
int rip_open;
int adv_self;
char this_is_rip;		// flag passed to send_ip in iproute.c

BYTE rip_version = 1;

WORD route_timeout;

TIMER rip_timer;

static void far rip_rx (SOCKET *psock,int cnt,IFACE *iface);
static void proc_rip (IFACE *iface,DWORD gateway,
	struct rip_route *ep,int32 ttl,int bits);
static BYTE *putheader (BYTE *cp,char command,char version);
static BYTE *putentry(register BYTE *cp,WORD fam,WORD route_tag,
int32 target,int32 subnet_mask,int32 metric);
static void rip_timeout (void);
static void send_routes (IFACE *iface,int32 dest,WORD port,int split,int trig,
	int us,char version);

/* Send RIP CMD_RESPONSE packet(s) to all directly connected interfaces */
static void rip_timeout(void)
{
	IFACE *ifp;

	stop_timer(&rip_timer);
	for (ifp = ifaces;ifp;ifp = ifp->next) {
		if ((ifp->sflags & (SF_RAW | SF_UNC)) || !(ifp->stat & IF_UP) ||
													(ifp->stat & IF_NORIPADV))
			continue;
		send_routes(ifp,-1l,RIP_PORT,(RIP_SPLIT),0,adv_self,rip_version);
	}
	start_timer(&rip_timer);
}

/* Send the routing table. */
static void send_routes(
	IFACE *iface,		/* interface to send to */
	int32 dest,		// destination IP address, -1l for broadcast
	WORD port,
	int split,		/* Do split horizon? */
	int trig,		/* Send only triggered updates? */
	int us,			/* Include our address in update */
	char version)
{
	BYTE *cp;
	int i,bits,numroutes,maxroutes;
	WORD pktsize;
	struct mbuf *bp;
	struct route *rp,*rp1;
	struct socket lsock,fsock;
	IFACE *ifr;

	if ((iface->stat & IF_NORIPADV) ||
		((iface->flags & (MDD_FLAG/* | X25_FLAG*/)) &&
					(iface->stat & IF_LAST_RIP)))
		return; 	// no advertising allowed
	/* Compute maximum packet size and number of routes we can send */
	pktsize = iface->mtu - IPLEN;
	pktsize = min(pktsize,MAXRIPPACKET);
	maxroutes = (pktsize - RIPHEADER) / RIPROUTE;

	lsock.address = iface->ip_addr;
	lsock.port = RIP_PORT;
	fsock.address = dest == -1l ? ((iface->ip_addr & NETMASK(iface->netbits)) |
					HOSTMASK(iface->netbits)) : dest;
	fsock.port = port;

	/* Allocate space for a full size RIP packet and generate header */
	if((bp = alloc_mbuf(pktsize,AllocRipSnd)) == NULLBUF)
		return;
	numroutes = 0;
	cp = putheader(bp->data,RIPCMD_RESPONSE,version);

	/* Emit route to ourselves, if requested */
	if(us){
		cp = putentry(cp,RIP_IPFAM,0,iface->ip_addr,
			version == 2 ? NETMASK(iface->netbits) : 0,1);
		numroutes++;
	}
	/* Emit default route, if appropriate */
	if (routes[0][0] && (ifr = routes[0][0]->interface) != NULLIF &&
				!(routes[0][0]->stat & RT_PRIVATE) &&
				(!trig || (routes[0][0]->stat & RT_TRIG))) {
		if ((!split || iface != ifr) && (ifr->stat & IF_UP)) {
			cp = putentry(cp,RIP_IPFAM,0,0,0,routes[0][0]->metric);
			numroutes++;
		} else if (trig) {
			cp = putentry(cp,RIP_IPFAM,0,0,0,RIP_INFINITY);
			numroutes++;
		}
	}
	for (bits = 1;bits < 33;bits++) {
		for (i = 0;i < NROUTE;i++) {
			for (rp = routes[bits][i];rp;rp = rp->next) {
				if((rp->stat & RT_PRIVATE) ||
										(trig && !(rp->stat & RT_TRIG))) {
					skip_this_one:
					continue;
				}
				for (rp1 = rp->next;rp1;rp1 = rp1->next) {
					if (rp->target == rp1->target &&
						rp->metric >= rp1->metric &&
						(rp1->interface->stat & IF_UP)) {
						goto skip_this_one;
					}
				}
				for (rp1 = rp->prev;rp1;rp1 = rp1->prev) {
					if (rp->target == rp1->target &&
						rp->metric > rp1->metric &&
						(rp1->interface->stat & IF_UP)) {
						goto skip_this_one;
					}
				}
				if(numroutes >= maxroutes){
					/* Packet full, flush and make another */
					bp->cnt = RIPHEADER + numroutes * RIPROUTE;
					this_is_rip = 1;
					send_udp(&lsock,&fsock,0,0,bp,bp->cnt,0,0,iface);
					this_is_rip = 0;
					//Rip_stat.output++;
					if((bp = alloc_mbuf(pktsize,AllocRipSnd1)) == NULLBUF)
						return;
					numroutes = 0;
					cp = putheader(bp->data,RIPCMD_RESPONSE,version);
				}
				if((!split || iface != rp->interface) &&
											(rp->interface->stat & IF_UP)){
					cp = putentry(cp,RIP_IPFAM,0,rp->target,
						version == 2 ? NETMASK(rp->bits) : 0,rp->metric);
					numroutes++;
				} else if(trig){
					cp = putentry(cp,RIP_IPFAM,0,rp->target,
						version == 2 ? NETMASK(rp->bits) : 0,RIP_INFINITY);
					numroutes++;
				}
			}
		}
	}
	if(numroutes != 0){
		bp->cnt = RIPHEADER + numroutes * RIPROUTE;
		this_is_rip = 1;
		send_udp(&lsock,&fsock,0,0,bp,bp->cnt,0,0,iface);
		this_is_rip = 0;
		//Rip_stat.output++;
	} else {
		free_p(bp);
	}
}

/* function to output a RIP CMD_RESPONSE packet for the rip_trigger list */
void rip_trigger(IFACE *ifp)
{
//	register struct rip_list *rl;
	int bits,i;
	struct route *rp;

	if (!rip_advertise)
		return;
	if (ifp) {
		for (bits = 0;bits < 33;bits++) {
			for (i = 0;i < NROUTE;i++) {
				for(rp = routes[bits][i];rp != NULLROUTE;rp = rp->next) {
					if (rp->interface == ifp)
						rp->stat |= RT_TRIG;
				}
			}
		}
	}
	for (ifp = ifaces;ifp;ifp = ifp->next) {
		if ((ifp->sflags & (SF_RAW | SF_UNC)) || !(ifp->stat & IF_UP) ||
													(ifp->stat & IF_NORIPADV))
			continue;
		send_routes(ifp,-1l,RIP_PORT,(RIP_SPLIT),1,adv_self,rip_version);
	}
	/* Clear the trigger list */
	for (bits=0;bits < 33;bits++) {
		for (i = 0;i < NROUTE;i++) {
			for (rp = routes[bits][i];rp != NULLROUTE;rp = rp->next) {
				rp->stat &= ~RT_TRIG;
			}
		}
	}
}

void rip_drop_iface(IFACE *ifp)
{
	int bits,i;
	struct route *rp;

	if (ifp) {
		for (bits = 0;bits < 33;bits++) {
			for (i = 0;i < NROUTE;i++) {
				for(rp = routes[bits][i];rp != NULLROUTE;rp = rp->next) {
					if (rp->interface == ifp && !(rp->stat | RT_STATIC))
						rp->metric = RIP_INFINITY;
				}
			}
		}
	}
	rip_trigger(ifp);
}

/* Start RIP agent listening at local RIP UDP port */
int rip_init()
{
	struct socket lsock;

	lsock.address = 0;
	lsock.port = RIP_PORT;

	if (rip_open <= 0) {
		if ((rip_open = open_udp(&lsock,rip_rx,0)) <= 0)
			return -1;
	}
	return 0;
}

/* Process RIP input received from 'interface'. */
#pragma argsused
static void far rip_rx(SOCKET *psock,int cnt,IFACE *iface)
{
	struct mbuf *bp;
	struct socket fsock;
	int cmd;
	char version;
	struct rip_route entry;
	struct route *rp;
	int32 ttl;

	/* receive the RIP packet */
	recv_udp(psock,&fsock,&bp);

	cmd = pullchar(&bp);
	/* Check the version of the frame */
	if((version = pullchar(&bp)) < 1){
		free_p(bp);
		return;
	}
	switch(cmd){
	case RIPCMD_RESPONSE:
		//Rip_stat.response++;
		/* See if this interface is on our broadcast list; if so,
		 * use its interval to calculate entry lifetimes. Otherwise,
		 * use default
		 */
		if (!rip_use)
			break;
		ttl = route_timeout;
		(void)pull16(&bp);	/* remove one word of padding */
		while (len_mbuf(bp) >= RIPROUTE) {
			int bits = -1;
			pullentry(&entry,&bp);
			if (version > 1) {
				int32 mask = entry.subnet_mask;
				for (bits = 0;mask;mask <<= 1,bits++);
			}
			proc_rip(iface,fsock.address,&entry,ttl,bits);
		}
		/* If we can't reach the sender of this update, or if
		 * our existing route is not through the interface we
		 * got this update on, add him as a host specific entry
		 */
		if ((rp = rt_blookup(fsock.address,32)) != NULLROUTE) {
			/* Host-specific route already exists, refresh it */
			start_timer(&rp->timer);
		} else if((rp = rt_lookup(fsock.address,0,USE_DOWN)) == NULLROUTE ||
												rp->interface != iface) {
			entry.addr_fam = RIP_IPFAM;
			entry.target = fsock.address;
			entry.metric = 0; /* will get incremented to 1 */
			proc_rip(iface,fsock.address,&entry,ttl,32);
		}
		rip_trigger(0);
		break;
	case RIPCMD_REQUEST:

		if((rp = rt_lookup(fsock.address,0,SKIP_DOWN)) != NULLROUTE)
			send_routes(rp->interface,fsock.address,fsock.port,
							fsock.port == RIP_PORT,0,adv_self,version);
		break;
	default:
		break;
	} /* switch */
	free_p(bp);
}

/* Apply a set of heuristics for determining the number of significant bits
 * (i.e., the address mask) in the target address. Needed since RIP doesn't
 * include the address mask for each entry.
 */
int nbits(DWORD target)
{
	int bits,minbits;
	DWORD mask;

	if(target == 0)
		return 0;	/* Special case: 0.0.0.0 is the default route */

	/* Since we don't have the netmask, take a guess at it:
	 * The class gives a lower limit of 8,16 or 24 for classes A, B or C.
	 * Otherwise we look at the position of the least significant 1 and
	 * assume that is last bit of the netmask. The effect is that we might
	 * think that we have routes to more destinations than we actually have
	 * which is better than thinking that we have less. If we get one which
	 * is more specific, we will use it rather than the wider one.
	 */
	if ((minbits = (hibyte(hiword(target)) >> 6) * 8) == 0)
		minbits = 8;
	for (bits = 32,mask = 1;bits > minbits && !(target & mask);
														bits--,mask <<= 1);
	return bits;
}

/* Remove and process a RIP response entry from a packet */
static void proc_rip(IFACE *iface,DWORD gateway,struct rip_route *ep,
								int32 ttl,int ibits)
{
#if APL
	char str[60];
#endif
	unsigned int bits;
	register struct route *rp;
	int add = 0;	/* action flags */
	int drop = 0;
	int trigger = 0;

	if (iface->stat & IF_NORIPUSE)
		return;
	if(ep->addr_fam != RIP_IPFAM) {
		/* Skip non-IP addresses */
		//Rip_stat.addr_family++;
		return;
	}
	/* Guess at the mask unless given, since it's not explicit in ver 1 */
	bits = ibits < 0 ? nbits(ep->target) : ibits;

	/* Don't ever add a route to myself through somebody! */
	if (is_this_addr(ep->target)) {
		return;
	}
	if (ep->target == 0 && ep->metric == RIP_INFINITY)
		return;		// ignore poisened default routes emitted by Unix
	/* Update metric to reflect link cost */
	ep->metric++;
	ep->metric = min(ep->metric,RIP_INFINITY);

	/* Find existing entry, if any */
	rp = rt_blookup(ep->target,bits);

	/* Don't touch private routes */
	if(rp != NULLROUTE && (rp->stat & RT_PRIVATE))
		return;

	if(rp == NULLROUTE){
		if(ep->metric < RIP_INFINITY){
			/* New route; add it and trigger an update */
			add++;
			trigger++;
		}
	} else if (rp->metric == RIP_INFINITY && ep->metric == RIP_INFINITY) {
		/* Route is in hold-down; ignore this guy */
	} else if (rp->gateway == gateway && rp->interface == iface) {
		if (rp->stat & RT_STATIC) {
			add++;
		}
		else {
			// This is the gateway for the entry we already have;
			// restart the timer
			set_timer(&rp->timer,ttl*1000L);
			start_timer(&rp->timer);
			if(rp->metric != ep->metric) {
				/* Metric has changed. Update it and trigger an
				 * update. If route has become unavailable, start
				 * the hold-down timeout.
				*/
				if (ep->metric == RIP_INFINITY)
					rt_timeout(rp);	/* Enter hold-down timeout */
				else
					rp->metric = (unsigned)ep->metric;
				trigger++;
			}
		}
	} else {
		/* Entry is from a different gateway than the current route */
		ROUTE *rp1 = rp;
		while (rp && rp->metric == RIP_INFINITY)
			rp = rp->next;
		if (!rp)
			rp = rp1;
		if (ep->metric < rp->metric || !(rp->interface->stat & IF_UP)) {
			/* Switch to a new gateway */
			drop++;
			add++;
			trigger++;
		}
	}
	if (drop && !(rp->stat & RT_STATIC)) {
		/* Switching to a better gateway; delete old entry */
#if APL
		if (logfp) {
			sprintf(str,"RtDrop %u.%u.%u.%u/%u",
				((unsigned char *)&ep->target)[3],
				((unsigned char *)&ep->target)[2],
				((unsigned char *)&ep->target)[1],
				((unsigned char *)&ep->target)[0],
				bits);
			log(0,str);
		}
#endif
		rt_drop(ep->target,bits);
	}
	if (add) {
		/* Add a new entry */
#if APL
		if (logfp) {
			sprintf(str,"RtAdd %u.%u.%u.%u/%u %s %u.%u.%u.%u %u",
				((unsigned char *)&ep->target)[3],
				((unsigned char *)&ep->target)[2],
				((unsigned char *)&ep->target)[1],
				((unsigned char *)&ep->target)[0],
				bits,iface->name,
				((unsigned char *)&gateway)[3],
				((unsigned char *)&gateway)[2],
				((unsigned char *)&gateway)[1],
				((unsigned char *)&gateway)[0],ep->metric);
			log(0,str);
		}
#endif
		rt_add(ep->target,(unsigned) bits,gateway,
		 (int) ep->metric,iface,ttl,trigger && rip_advertise ? RT_TRIG : 0,1);
	}
}

/* Send a RIP request packet to the specified destination */
int ripreq(DWORD dest,WORD replyport)
{
	struct mbuf *bp;
	struct socket lsock,fsock;
	BYTE *cp;

	lsock.address = 0;
	lsock.port = replyport;

	/* if we were given a valid dest addr, ask it (the routers on that net)
	 * for a default gateway
	 */
	if(dest == 0)
		return 0;

	fsock.address = dest;
	fsock.port = RIP_PORT;

	/* Send out one RIP Request packet as a broadcast to 'dest'  */
	if((bp = alloc_mbuf(RIPHEADER + RIPROUTE,AllocRipSnd2)) == NULLBUF)
		return -1;

	cp = putheader(bp->data,RIPCMD_REQUEST,rip_version);
	cp = putentry(cp,RIP_IPFAM,0,0L,0L,RIP_INFINITY);
	bp->cnt = RIPHEADER + RIPROUTE;
	this_is_rip = 1;
	send_udp(&lsock, &fsock,0,0,bp,bp->cnt,0,0,0);
	this_is_rip = 0;
	//Rip_stat.output++;
	return 0;
}

void pullentry(register struct rip_route *ep,struct mbuf **bpp)
{
	ep->addr_fam = pull16(bpp);
	ep->route_tag = pull16(bpp);
	if (ep->addr_fam != AUTHENT) {
		ep->target = pull32(bpp);
		ep->subnet_mask = pull32(bpp);
		ep->next_hop = pull32(bpp);
		ep->metric = pull32(bpp);
	}
	else
		pullup(bpp,authent(ep),16);
}

/* Write the header of a RIP packet */
static BYTE *putheader(BYTE *cp,char command,char version)
{
	*cp++ = command;
	*cp++ = version;
	return put16(cp,0);
}

/* Write a single entry into a rip packet */
static BYTE *putentry(BYTE *cp,WORD fam,WORD route_tag,
int32 target,int32 subnet_mask,int32 metric)
{
	cp = put16(cp,fam);				// address family, 0xffff for authenticaton
	cp = put16(cp,route_tag);		// Route tag
	cp = put32(cp,target);			// target address
	cp = put32(cp,subnet_mask);		// subnet mask
	cp = put32(cp,0L);				// next hop
	return put32(cp,metric);
}

/* Route timeout handler. If route has already been marked for deletion
 * then delete it. Otherwise mark for deletion and restart timer.
 */
void rt_timeout(void *s)
{
	register struct route *rp = (struct route *)s;
#if APL
	char str[60];
#endif

	stop_timer(&rp->timer);
	if(rp->metric < RIP_INFINITY){
		rp->metric = RIP_INFINITY;
		/* wait 2/3 of timeout before garbage collect */
		set_timer(&rp->timer,route_timeout * 2000L / 3);
		start_timer(&rp->timer);
		/* Route changed; mark it for triggered update */
		if (rip_advertise) {
			rp->stat |= RT_TRIG;
			rip_trigger(0);
		}
	} else if (!(rp->stat & RT_STATIC)) {
#if APL
		if (logfp) {
			sprintf(str,"RtDrop %u.%u.%u.%u/%u",
				((unsigned char *)&rp->target)[3],
				((unsigned char *)&rp->target)[2],
				((unsigned char *)&rp->target)[1],
				((unsigned char *)&rp->target)[0],
				rp->bits);
			log(0,str);
		}
#endif
		rt_drop(rp->target,rp->bits);
	}
}

#pragma argsused
int doripadvert(int argc,char *argv[],void *p)
{
	if (rip_init() == -1)
		return -1;
	rip_advertise = 1;
	rip_timer.func = (void(*)(char *))rip_timeout;
	set_timer(&rip_timer,(argc > 1 ? atoi(argv[1]) : 30) * 1000L);
	rip_timeout();
	while (argc > 2) {
		char *parg = argv[argc - 1];
		char digit = parg[0];
		if (parg[1] == 0 && (digit == '1' || digit == '2'))
			rip_version = digit - '0';
		else
			adv_self = 1;
		--argc;
	}
	return 0;
}

#pragma argsused
int doripuse(int argc,char *argv[],void *p)
{
	if (rip_init() == -1)
		return -1;
	rip_use = 1;
	route_timeout = argc > 1 ? atoi(argv[1]) : 240;
	ripreq(0xffffffffL,RIP_PORT);
	return 0;
}

struct scmds far ripcmds[] = {
	{"advertise",	doripadvert},
	{"use",			doripuse},
	{"",				NULLFP},
};

dorip(int argc,char *argv[])
{
	return subcmd(ripcmds,argc,argv,0);
}
