//
//   $Id: SCHEDULE.C 1.6 2003/08/11 03:28:28Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// SCHEDULE.C
// Scheduling functions
// Copyright (c) GP van Niekerk Ondernemings 1990-1998

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <stdarg.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <dos.h>
#include <string.h>
#include <time.h>
#include <alloc.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "netfunc.h"
#include "timer.h"
#include "iface.h"
#include "icmp.h"
#include "ip.h"
#include "tcp.h"
#include "pc.h"
#include "task.h"
#include "bootp.h"
#if APL
#include "ini.h"
#include "trace.h"
#define TRACE 1
#endif

void iostop(void);

extern struct mbuf *loopq;
extern int in_tcp;

c_break()			/* control break routine */
{
	return 1;		/* ignore error */
}

/* fatal error routine */

#if APL
WORD ax,bx,cx,dx,sp,bp,si,di,ds,es,ss,flags;
#endif

void fatal(char *ermes)
{
#if APL
	int fho,i;
	char *buf;

	ax = _AX;bx = _BX;cx = _CX;dx = _DX;
	sp = _SP;bp = _BP;si = _SI;di = _DI;
	ds = _DS;es = _ES;ss = _SS;flags = _FLAGS;
	printf("FATAL SYSTEM ERROR: %s\n",ermes);
	if ((fho = open("inet.dmp",O_WRONLY|O_BINARY|O_CREAT|O_TRUNC,S_IWRITE)) == -1) {
		perror("Unable to open INET.DMP");
		goto err2;
	}
	for (i = 0,buf = 0;i < 4;i++,buf += 0x4000) {
		if (write(fho,buf,0x4000) != 0x4000) {
			perror("Write error on INET.DMP");
			goto err1;
		}
	}
	printf("Dump to INET.DMP successful\n");
	err1:
	close(fho);
	err2:
	iostop();
	restore_palette();
	exit(1);
#else
	printf("FATAL SYSTEM ERROR: %s\n",ermes);
#endif
}

extern char did_schedule;

#if APL
TASK *tasks;			/* tasks header */

TASK *task_create(void (*poll)(),void *param)
{
	TASK *ptask;

	if ((ptask = calloc(sizeof(TASK),1,AllocSchedTasks)) != 0) {
		ptask->poll = poll;
		ptask->param = param;
		ptask->next = tasks;
		tasks = ptask;
	}
	return ptask;
}

int task_delete(TASK *ptaskd)
{
	TASK *ptask,*ptaskp = 0;

	for (ptask = tasks;ptask;ptask = ptask->next) {
		if (ptask == ptaskd) {
			if (ptaskp)
				ptaskp->next = ptask->next;
			else
				tasks = ptask->next;
			free(ptask);
			return 0;
		}
		ptaskp = ptask;
	}
	return -1;
}

/* Schedule an event */
void schedule(void)
{
	void check_time();
	struct interface *ifp;
	TASK *ptask;
	struct mbuf *bp;
	static char insched;
	extern int async;
	extern struct interface loopback;

#ifdef DEB1C
extern int aiTimer[3];						// debug int 1c
	if (in_tcp == 0x41)
		++aiTimer[1];
#endif
	if (insched || !async)
		return;
	insched = 1;
#ifdef DEB1C
	if (in_tcp == 0x41)
		++aiTimer[2];
#endif
	/* Service the loopback queue */
	if ((bp = dequeue(&loopq)) != NULLBUF) {
		struct ip ip;
#ifdef TRACE
		dump(&loopback,IF_TRACE_IN,TRACE_IP,bp);
#endif
		/* Extract IP header */
		ntohip(&ip,&bp);
		ip_recv(0,&ip,bp,0);
	}
	/* Service the interfaces */
	for (ifp = ifaces; ifp != NULLIF; ifp = ifp->next) {
		if(ifp->recv != 0) {
			(*ifp->recv)(ifp);
		}
	}
	/* service the clock */
	check_time();
	insched = 0;
	/* service tasks */
	for (ptask = tasks;ptask;) {
		TASK *ntask = ptask->next;
		(*ptask->poll)(ptask->param);
		ptask = ntask;
	}

#else
void schedule(void)
{
	void sched_printer(void);
	void check_time();
	struct interface *ifp;
	struct mbuf *bp;
	static char insched;
	extern char print_flag;

#ifdef DEB1C
extern int aiTimer[3];						// debug int 1c
	if (in_tcp == 0x41)
		++aiTimer[1];
#endif
	if (insched)
		return;
	insched = 1;
#ifdef DEB1C
	if (in_tcp == 0x41)
		++aiTimer[2];
#endif
	/* Service the loopback queue */
	if ((bp = dequeue(&loopq)) != NULLBUF) {
		struct ip ip;
		/* Extract IP header */
		ntohip(&ip,&bp);
		ip_recv(0,&ip,bp,0);
	}
	/* Service the interfaces */
	for (ifp = ifaces; ifp != NULLIF; ifp = ifp->next) {
		if(ifp->recv != 0) {
			(*ifp->recv)(ifp);
		}
	}
	/* service the clock */
	check_time();
	/* service printers */
	if (print_flag)
		sched_printer();
	insched = 0;
	did_schedule = 1;
#endif
}
