//
//   $Id: SLIP.C 1.4 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// SLIP.C
// Serial Link Internet Protocol implementation
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

#include <stdio.h>
#include <string.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "iface.h"
#include "pc.h"
#include "timer.h"
#include "modem.h"
#include "ser.h"
#include "netfunc.h"
#include "slhc.h"
#include "rip.h"

#if APL
#define	TRACE	1
#include "trace.h"
#endif

static char ioctl_po[] = {0,IOCTL_PROD_OUTPUT};
char *pioctl_po[1] = {ioctl_po};

void ip_route(MBUF *bp,char rxbroadcast,IFACE *ifp);

// Send routine for point-to-point slip
#pragma argsused
void slip_send(
	struct mbuf *bp,	/* Buffer to send */
	IFACE *ifp,			/* Pointer to interface control block */
	int32 gateway,		/* Ignored (SLIP is point-to-point) */
	char precedence,
	char delay,
	char throughput,
	char reliability)
{
	struct ser *serp;
	int type;

	if (ifp == NULLIF) {
		free_p(bp);
		return;
	}
	serp = ifp->dcb;
#ifdef TRACE
	dump(ifp,IF_TRACE_OUT,TRACE_IP,bp);
#endif
	if (serp->slcomp) {
		/* Attempt IP/ICP header compression */
		type = slhc_compress(serp->slcomp,&bp,1);
		bp->data[0] |= type;
#ifdef TRACE
		if ((ifp->trace & (IF_TRACE_OUT | IF_TRACE_COMP)) ==
					(IF_TRACE_OUT | IF_TRACE_COMP)) {
			dump(ifp,IF_TRACE_OUT,0,bp);
		}
#endif
	}
	(*ifp->raw)(ifp,bp);
}

/* Send a raw slip frame
 * Put on link output queue, and kick transmitter
 */
void slip_raw(IFACE *ifp,MBUF *bp)
{
	register SER *serp;
	MBUF *nbp;
	WORD size;

	serp = ifp->dcb;
	if ((size = len_q(serp->sndq)) > 0 /*ifp->gpvar / 2*/) {
		/* try and restart */
		if (ifp->ioctl)
			(*ifp->ioctl)(ifp,1,pioctl_po);
		if (size >= ifp->gpvar) {
			/* too many entries on queue, just drop it */
			free_p(bp);
			return;
		}
	}
	/* Copy packet to a new mbuf to solve problem where
	 * an ack of a segment while a retransmit is taking place, frees
	 * the mbuf data causing garbage to be sent. Could also be used to
	 * re-use used portions of the mbuf after VJ compression since we are
	 * guarenteed to not have a duplicated header with no real data, BUT
	 * the code will have to be moved to before compression.
	 */
	nbp = copy_p(bp,len_mbuf(bp),AllocSlipDup);
	free_p(bp);
	if (nbp)
		bp = nbp;
	else
		return;
	enqueue(&serp->sndq,bp);
	if(serp->tbp == NULLBUF && serp->ostate == OS_IDLE) {
		if (serp->sndq == NULLBUF || (serp->tbp = dequeue(&serp->sndq)) == 0) {
			return;	/* No work */
		}
		(*ifp->xmit)(ifp,0,1);
	}
}

/* Process SLIP line I/O */
void doslip(IFACE *ifp)
{
	struct mbuf *bp;
	struct ser *serp;
	 char c;

	serp = ifp->dcb;
	/* Process any pending input */
	if ((bp = dequeue(&serp->rcvq)) != NULLBUF) {
		if (serp->slcomp) {
#ifdef TRACE
			if ((ifp->trace & (IF_TRACE_IN | IF_TRACE_COMP)) ==
					(IF_TRACE_IN | IF_TRACE_COMP)) {
				dump(ifp,IF_TRACE_IN,0,bp);
			}
#endif
			if ((c = bp->data[0]) & SL_TYPE_COMPRESSED_TCP) {
				if (slhc_uncompress(serp->slcomp,&bp) <= 0 ) {
					free_p(bp);
					serp->errors++;
					return;
				}
			} else if (c >= SL_TYPE_UNCOMPRESSED_TCP) {
				bp->data[0] &= 0x4f;
				if (slhc_remember(serp->slcomp,&bp) <= 0 ) {
					free_p(bp);
					serp->errors++;
					return;
				}
			}
		}
#ifdef TRACE
		dump(ifp,IF_TRACE_IN,TRACE_IP,bp);
#endif
		ip_route(bp,0,ifp);
	}
	(*ifp->ioctl)(ifp,1,pioctl_po);
	/* Kick the transmitter if it's idle */
	(*ifp->xmit)(ifp,0,0);
}

void dummy()
{
}

void find_link(
	struct mbuf *bp,	/* Buffer to send */
	IFACE *ifp,			/* Pointer to interface control block */
	DWORD gateway,		/* Ignored (SLIP is point-to-point) */
	char precedence,
	char delay,
	char throughput,
	char reliability)
{
	IFACE *sifp;
	SER *serp;
	MODEMPAR *pm;

	pm = ifp->indcb;
	if (pm->fdial[0]) {
		for (sifp = ifaces;sifp;sifp = sifp->next) {
			if ((sifp->sflags & (SF_UNC | SF_INUSE)) == SF_UNC)
				goto found;
		}
	}
	free_p(bp);
	return;
	found:
	serp = ifp->dcb = sifp->dcb;
	if ((ifp->sflags & SF_COMPRESS) && 
				(serp->slcomp = slhc_init(16,16)) == 0) {
		printf("No memory for compression\n");
		return;
	}
	sifp->sflags |= SF_INUSE;
	sifp->recv = doslip;
	ifp->send = (void *)slip_send;
	serp->bufsize = ifp->mtu;
	serp->istate = IS_MODEM;
	serp->palt = (SER *)ifp;
	if (serp->rbp)
		free_p(serp->rbp);
	serp->rbp = alloc_mbuf(serp->bufsize,AllocSlipBuf);
	if (ifp->indcb && serp->pmpar) {
		MODEMPAR *pm_mdd = ifp->indcb;
		MODEMPAR *pm_asl = serp->pmpar;
		pm_asl->dial =
		pm_asl->fdial = pm_mdd->fdial;
		pm_asl->xchange = pm_mdd->xchange;
		pm_asl->connect = pm_mdd->connect;
	}
	slip_send(bp,ifp,gateway,precedence,delay,throughput,reliability);
}

int find_mdd(IFACE *sifp,char *xid)
{
	SER *serp = sifp->dcb;
	IFACE *mifp;
	MODEMPAR *pm;

	for (mifp = ifaces;mifp;mifp = mifp->next) {
		if ((mifp->flags & MDD_FLAG) && (pm = mifp->indcb) != 0 &&
				mifp->dcb == 0 && strcmp(xid,(char *)pm->xchange) == 0)
			goto found;
	}
	return 0;
	found:
	if (sifp->sflags & SF_INUSE)
		release_sif(sifp);			// terminate outgoing call
	if ((mifp->sflags & SF_COMPRESS) &&
				(serp->slcomp = slhc_init(16,16)) == 0) {
		printf("No memory for compression\n");
		return 0;
	}
	mifp->dcb = serp;
	sifp->sflags |= SF_INUSE;
	sifp->recv = doslip;
	mifp->send = (void *)slip_send;
	serp->bufsize = mifp->mtu;
	serp->palt = (SER *)mifp;
	return 1;
}

void release_sif(IFACE *ifp)
{
	SER *serp;
	IFACE *mifp;
	MODEMPAR *pm_asl;

	if (!ifp)
		return;
	serp = ifp->dcb;
	if (serp->tbp) {
		free_mbuf(serp->tbp);
		serp->tbp = 0;
	}
	if (serp->sndq)
		free_q(&serp->sndq);
	serp->istate = IS_MODEM;
	slhc_free(serp->slcomp);
	serp->slcomp = 0;
	if (ifp->sflags & SF_UNC) {
		ifp->sflags &= ~SF_INUSE;
		if ((mifp = (IFACE *)serp->palt) == 0)
			return;
		mifp->send = find_link;
		mifp->dcb = 0;
		ifp = mifp;
		serp->palt = 0;
		if ((pm_asl = serp->pmpar) != 0) {
			pm_asl->dial =
			pm_asl->fdial = 0;
			pm_asl->xchange = 
			pm_asl->connect =
			pm_asl->nextptr = 0;
			pm_asl->tmstate = TM_IDLE;
		}
	}
	rip_drop_iface(ifp);
}

mdd_ioctl(IFACE *ifp,int argc,char *argv[])
{
	SER * serp;
	IFACE *ifpl;

	if (argc == 1 && argv[0][0] == 0) {	/* internal commands */
		if ((serp = ifp->dcb) != 0) {
			ifpl = (IFACE *)serp->iface;
			return (*ifpl->ioctl)(ifpl,argc,argv);
		}
		return 0;
	}
#if APL
	if (argc > 1)
		return 0;
	printf("mdd interface: %s\n",ifp->name);
	if ((serp = ifp->dcb) != 0) {
		ifpl = (IFACE *)serp->iface;
		printf("Linked to: %s\n",ifpl->name);
		dial_status(serp->pmpar);
		if (serp->sndq) {
			printf("Send Q %u, Limit %u\n",len_q(serp->sndq),ifp->gpvar);
		}
	}
	else {
		printf("Not linked\n");
		if (ifp->indcb)
			dial_status(ifp->indcb);
	}
#endif
	return 0;
}
