//
//   $Id: SLIPB.C 1.4 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// SLIPB.C
// Send and receive IP datagrams on serial lines with LAPB
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

#if APL==0
#pragma option -zAINETINIT
#pragma option -zFINETINIT
#endif

#include <stdio.h>
#include <stdlib.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "iface.h"
#ifdef LAPB
#include "x25id.h"
#include "lapb.h"
#include "trace.h"
#endif
#include "pc.h"
#include "timer.h"
#include "modem.h"
#include "ser.h"
#include "slhc.h"

unsigned *crctab;
extern char *pioctl_po[];

#ifdef LAPB
/* Send routine for point-to-point slip
 */
int
slipb_send(bp,interface,gateway,precedence,delay,throughput,reliability)
struct mbuf *bp;		/* Buffer to send */
struct interface *interface;	/* Pointer to interface control block */
int32 gateway;			/* Ignored (SLIP is point-to-point) */
char precedence;
char delay;
char throughput;
char reliability;
{
	struct x25_cb *axp,*find_ax25(),*open_ax25();
	struct mbuf *tbp;
	extern WORD axwindow;
	void ax_incom();
	WORD size,bsize,seq;

	if(interface == NULLIF){
		free_p(bp);
		return;
	}
	axp = interface->indcb;
	if ((size = len_q(axp->txq)) > interface->gpvar / 2) {
		/* try and restart */
		(*interface->ioctl)(interface,1,pioctl_po);
		if (size >= interface->gpvar) {
			/* too many entries on queue, just drop it */
			free_p(bp);
			return;
		}
	}
	/* If datagram is too big for one frame, send all but the last with
	 * the extension PID. Note: the copy to a new buf is necessary because
	 * LAPB may continue retransmitting the frame after a local TCB has
	 * gone away, and using the buf directly would cause heap garbage to be
	 * transmitted.
	 */
	bsize = len_mbuf(bp);
	seq = 0;
	while(bsize != 0){
		size = min(bsize,axp->paclen);
		/* Allocate buffer, allowing space for PID */
		if((tbp = alloc_mbuf(size + 1,AllocSlipbBuf)) == NULLBUF)
			break;		/* out of memory! */
		*tbp->data = PID_IP;
		if(seq++ == 0)
			*tbp->data |= PID_FIRST;	 /* First in sequence */
		if(size == bsize)
			*tbp->data |= PID_LAST;		/* That's all of it */
		/* else more to follow */

		tbp->cnt = 1;
		tbp->cnt += pullup(&bp,tbp->data + 1,size);
		enqueue(&axp->txq,tbp);
		lapb_output(axp);
		bsize -= size;
	}
	free_p(bp);	/* Shouldn't be necessary */
}

/* Process SLIP line I/O */
int doslipb(struct interface *interface)
{
	struct mbuf *bp;
	struct ser *serp;

	serp = interface->dcb;
	/* Process any pending input */
	if ((bp = dequeue(&serp->rcvq)) != NULLBUF) {
		dump(interface,IF_TRACE_IN,TRACE_X25,bp);
		lapb_input(interface->indcb,bp,interface);
	}
	/* Kick the transmitter if it's idle */

	(*interface->xmit)(interface,0,0);
}
#endif

static
unsigned near crchw(unsigned data,unsigned genpoly,unsigned accum)
{
	int i;

	data <<= 1;
	for (i = 8;i > 0;i--) {
		data >>= 1;
		if ((data ^ accum) & 1)
			accum = (accum >> 1) ^ genpoly;
		else
			accum >>= 1;
	}
	return accum;
}

bcrctab()
{
	unsigned i;

	if (crctab)
		return 1;		/* already built */
	if ((crctab = (unsigned *)malloc(512,AllocSlipbCrc)) == 0)
		return -1;
	for (i = 0;i < 256;++i) {
		crctab[i] = crchw(i,0x8408,0);	/* ccitt */
	}
	return 1;
}
