//
//   $Id: S_IFACE.C 1.4 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// S_IFACE.C
// Serial interface functions
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

// 2002-07-23 gvn <iovec>[/s]

#if APL==0
#pragma option -zAINETINIT
#pragma option -zFINETINIT
#endif

#include <stdlib.h>
#include <stdio.h>
#include <alloc.h>
#include <string.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "iface.h"
#include "pc.h"
#include "timer.h"
#ifdef LAPB
#include "lapb.h"
#endif
#include "modem.h"
#include "ser.h"
//#include "trace.h"
#include "netfunc.h"
#include "slhc.h"
#include "arp.h"
#include "ppp.h"

int bcrctab(void);
int nservec;

enum {
	SLIP_MODE,		/* standard slip */
	CSLIP_MODE,		/* VJ compressed slip */
	PPP_MODE,		// Point to Point Protocol
	RAW_MODE,		/* no protocol */
	SLIPB_MODE,		/* slip with lapb */
	CLOCK_MODE,		/* clock generator */
	HLAPB_MODE,		/* lapb over hdlc */
	HDLC_MODE,		/* lapb over hdlc */
	INV_MODE		/* invalid mode */
};

char *mode_tab[] = {
	"slip","cslip","ppp","raw","slip_lapb","clock","lapb","hdlc"
};

void doslip(),doslipb(), dummy();
void find_link();
int mdd_ioctl();
void slip_send(),slipb_send();
int slip_raw();
int slip_recv(),slipb_recv();
void raw_recv();
void chartimeout();

enum {
	A_HWTYPE,
	A_IFNAME,
	A_MODE,
	A_MTU,
	A_BUFLIM,
	A_IOAD,
	A_IOVEC,
	A_SPEED,
	A_MODEM,
};

/* Define a serial interface
 * argv[0]: hardware type, must be "asy","aslink","mdd" or "scc"
 * argv[1]: interface label, e.g., "sl0"
 * argv[2]: mode, may be:
 *		"slip" (point-to-point SLIP)
 *		"slip_lapb" for lapb over slip
 *		"raw" (no frame protocol)
 * argv[3]: maximum transmission unit, bytes
 * argv[3]: receive timeout (raw)
 * argv[4]: receiver ring buffer size in bytes (raw)
 * argv[4]: slip and lapb max send queue size
 * argv[5]: I/O address, e.g., "0x3f8"
 * argv[6]: vector, e.g., "4"
 * argv[7]: interface speed, e.g, "9600"
 * argv[8]: modem spec file
 */

#define ASY_SER	1		// async with link
#define ASY_MDD	2		// async interface
#define SCC_SER 4		// scc

#pragma argsused
ser_setup(char *hwtype,char *ifname,char *ifclass,int mtu,int timeout,
	int buflim,int bufsize,int ioad,int iovec,unsigned speed,
	char *modem,IFACE **pifp,WORD wFlags)
{
	register struct interface *ifp;
	SER *serp,*pser;
	extern struct interface *ifaces;
	extern WORD paclen;		/* I field packet length */
	WORD dev;
	int mode,i;
	char type = ASY_SER;
	char compress = 0;
#if APL
#ifdef SERIAL
	WORD buf8th;
	struct timer *ptim;
#endif
#endif
	if (strcmp(hwtype,"asy") == 0)
		type = ASY_SER | ASY_MDD;
	else if (strcmp(hwtype,"mdd") == 0)
		type = ASY_MDD;
	else if (strcmp(hwtype,"scc") == 0)
		type = SCC_SER;
	for (mode = SLIP_MODE;mode < INV_MODE;mode++) {
		if(strcmp(ifclass,mode_tab[mode]) == 0) {
			if (mode == HDLC_MODE)
				mode = HLAPB_MODE;
			goto mode_ok;
		}
	}
	printf("Class %s unknown for interface %s\n",ifclass,ifname);
	return -1;
	mode_ok:
	if (mode == CSLIP_MODE) {
		mode = SLIP_MODE;
		compress = 1;
	}
//	if ((type == ASY_SER || type == ASY_MDD) && mode != SLIP_MODE)
//		return -1;
	if (type & (SCC_SER | ASY_SER)) {
		dev = nser++;
		if ((serp = calloc(1,sizeof(SER),AllocSifSer)) == 0)
			return -1;
		serp->flags = wFlags;
		/* Initialize hardware-level control structure */
		serp->addr = ioad;
		// search for vec shared with previous port(s)
		for (i = 0;i < dev && (pser = ser[i]) != 0;++i) {
			while (pser->vec == iovec && pser->next != 0)
				pser = pser->next;
			if (pser->vec == iovec) {
				// save pointer to previous ser for use in asy_init()
				(SER *)(serp->oldvec) = pser;
				break;
			}
		}
		if (i == nservec) {
			// new interrupt vector
			if (nservec >= SER_INT_MAX) {
				printf("Too many serial interrupt vectors\n");
				return -1;
			}
			ser[nservec++] = serp;
		}
		serp->vec = iovec;
	}
	else
		serp = 0;
	/* Create interface structure and fill in details */
	if ((ifp = (struct interface *)calloc(1,sizeof(struct interface),AllocSifIf)) == 0)
		return -1;
	if ((ifp->name = malloc(strlen(ifname) + 1,AllocSifNm)) == 0)
		return -1;
	strcpy(ifp->name,ifname);
	ifp->mtu = mtu;
	ifp->dev = dev;
	ifp->dcb = serp;
   ifp->hwtype = ARP_SERIAL;
	if (serp) {
		serp->iface = ifp;
		ifp->recv = doslip;
	}
	else {
		if (modem) {
			if ((ifp->indcb = modem_par(modem)) == 0)
				return -1;
		}
		ifp->send = find_link;
		ifp->raw = slip_raw;
		ifp->ioctl = mdd_ioctl;
		ifp->gpvar = buflim;
		ifp->stop = (void *)dummy;
		ifp->xmit = asy_xmit;
		ifp->srecv = asy_srecv;
		ifp->flags |= MDD_FLAG;
		if (compress)
			ifp->sflags |= SF_COMPRESS;
		goto add_if;
	}
	switch(mode){
#ifdef PPP_S
	case PPP_MODE:
		ppp_init(ifp);
		ifp->flags = PPP_FLAG;
		serp->flags |= FL_PPP;
		if (bcrctab() == -1)
			return -1;
		serp->bufsize = 2 * ifp->mtu + 20;
		serp->istate = IS_PPP;
		goto slip_common;
#endif
	case SLIP_MODE:
		ifp->send = slip_send;
		ifp->raw = slip_raw;
		ifp->flags = 0;
		serp->bufsize = ifp->mtu;
		serp->istate = IS_DATA;
		slip_common:
		if (type == ASY_SER)
			ifp->sflags = SF_UNC;
		if (compress && (type & ASY_MDD)) {
			ifp->sflags |= SF_COMPRESS;
		}
		goto setqlim;
#if APL
#ifdef SERIAL
	case RAW_MODE:
		ifp->mtu = timeout;
		ifp->recv = raw_recv;
		serp->iflow = IF_XON;		/* default flow control */
		ptim = &serp->timer;
		set_timer(ptim,ifp->mtu);
		ptim->func = chartimeout;
		ptim->arg = (char *)serp;
		serp->bufsize = bufsize;
		ifp->sflags = SF_RAW;
		serp->istate = IS_RAW;
		/* Set up receiver FIFO */
		if ((serp->buf = malloc(serp->bufsize,AllocSifBuf)) == NULLCHAR) {
			return -1;
		}
		buf8th = serp->bufsize >> 3;
		serp->stop_cnt = buf8th * 7;
		serp->start_cnt = buf8th * 6;
		serp->wp = serp->rp = serp->buf;
		serp->icnt = 0;
		break;
#endif
#endif
#ifdef	LAPB
	case SLIPB_MODE:
		serp->flags |= FL_LAPB;
		serp->accum = 0xffff;
		serp->bufsize = paclen + 5;
		serp->istate = IS_DATAB;
		goto lapb_mode;
	case HLAPB_MODE:
		serp->flags |= (FL_HDLC | FL_LAPB);
		serp->istate = IS_HDLC;
		serp->bufsize = paclen + 4;
		lapb_mode:
		ifp->sflags = SF_LAPB;
		ifp->send = slipb_send;
		ifp->raw = slip_raw;
		ifp->recv = doslipb;
#endif
		setqlim:
		ifp->gpvar = buflim;
		break;
	case CLOCK_MODE:
		serp->flags |= FL_HDLC;
		ifp->recv = dummy;
	}
	if (modem) {
		if ((serp->pmpar = modem_par(modem)) == 0)
			return -1;
	}
	if (type & ASY_SER) {
#ifdef ASYNC
		ifp->stop = asy_stop;
		ifp->ioctl = asy_ioctl;
		ifp->xmit = asy_xmit;
		ifp->srecv = asy_srecv;
		if (asy_init(ifp,speed) < 0)
			return -1;
		if ((ifp->sflags & SF_COMPRESS) &&
				(serp->slcomp = slhc_init(16,16)) == 0) {
			return -1;
		}
#endif
	}
#ifdef SCC
	else {
		int i = strlen(ifname);

		serp->flags |= (ifname[i - 1] & 1 ? 0 : FL_CHANB);
		ifp->stop = scc_stop;
		ifp->ioctl = scc_ioctl;
		ifp->xmit = scc_xmit;
		ifp->srecv = scc_srecv;
		scc_init(dev,mode == HLAPB_MODE ? SCC_HDLC :
				mode == CLOCK_MODE ? 0xff : SCC_ASYNC);
		scc_speed(dev,speed);
	}
#endif
#ifdef LAPB
	if (mode == SLIPB_MODE || mode == HLAPB_MODE) {
		if (mode == SLIPB_MODE && bcrctab() == -1)/* build CRC table */
			return -1;
		/* initialize link */
		if (lapb_init(ifp,0,0,0,0) == 0)
			return -1;
	}
#endif
	add_if:
	if_add(ifp);
	if (pifp)
		*pifp = ifp;
	return 0;
}

void ser_iface(int argc,char *argv[])
{
	int mtu = atoi(argv[A_MTU]);
	int buflim = atoi(argv[A_BUFLIM]);

	ser_setup(argv[A_HWTYPE],argv[A_IFNAME],argv[A_MODE],mtu,mtu,
		buflim,buflim,htoi(argv[A_IOAD]),atoi(argv[A_IOVEC]),
		atoi(argv[A_SPEED]),argc > A_MODEM ? argv[A_MODEM] : 0,0,
		strchr(argv[A_IOVEC],'//') ? FL_SUB_ISR : 0);
}

#ifdef INI_FILE
ser_ini(InterfaceDesc *pif)
{
	int retv;
	IFACE *ifp;
	char *argv[4];
	char dsp[4];
	char baud[10];

	if ((retv = ser_setup(pif->type,pif->name,pif->Class,pif->mtu,pif->timeout,
			pif->buflim,pif->bufsize,pif->ioad,pif->intvec,pif->speed,
			*pif->modem_file ? pif->modem_file : 0,&ifp,0)) < 0) {
		return retv;
	}
	sprintf(baud,"%u",pif->speed);
	argv[0] = baud;
	argv[1] = pif->oflow;
	argv[2] = pif->iflow;
	sprintf(dsp,"%u%u%c",pif->databits % 10,pif->stopbits & 3,
						pif->parity[0]);
	argv[3] = dsp;
	ifp->ioctl(ifp,3,argv);
	return retv;
}
#endif
