//
//   $Id: TCPCMD.C 1.6 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// TCPCMD.C
// TCP commands
// Copyright (c) GP van Niekerk Ondernemings 1990-2002
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

/*
02/13/2002 gvn IR#3936:
               Show/set AckOnPush to allow immediate ACKnowledgement of
               a TCP packet.
*/

#if APL==0
#pragma option -zAINETINIT
#pragma option -zFINETINIT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <alloc.h>
#include <string.h>
#include "inetdefs.h"
#include "timer.h"
#include "mbuf.h"
#include "netfunc.h"
#include "internet.h"
#include "tcp.h"
#include "cmdparse.h"

int tstat(void);

typedef TCB * PTCB;
PTCB far *tcbindex;
int tcbcount;

/* TCP connection states */
#if APL
char *tcpstates[] = {
	"Closed",
	"Listen",
	"SYN sent",
	"SYN received",
	"Established",
	"FIN wait 1",
	"FIN wait 2",
	"Close wait",
	"Closing",
	"Last ACK",
	"Time wait"
};

/* TCP closing reasons */
char *reasons[] = {
	"Normal",
	"Reset",
	"Timeout",
	"ICMP"
};
#endif

#ifdef INI_FILE
#include "ini.h"
void ini_tcp(char *ini_file)	// initialise TCP
{
	tcp_stat.tcp_irtt = GetPrivateProfileInt("TCP","irtt",tcp_stat.tcp_irtt,ini_file);
	tcp_stat.tcp_time_max = GetPrivateProfileInt("TCP","timemax",tcp_stat.tcp_irtt,ini_file);
	tcp_stat.tcp_mss = GetPrivateProfileInt("TCP","mss",tcp_stat.tcp_mss,ini_file);
	tcp_stat.tcp_smss = GetPrivateProfileInt("TCP","smss",tcp_stat.tcp_mss,ini_file);
	tcp_stat.tcp_retry = GetPrivateProfileInt("TCP","retry",tcp_stat.tcp_retry,ini_file);
	tcp_stat.tcp_window = GetPrivateProfileInt("TCP","window",tcp_stat.tcp_window,ini_file);
}
#endif

/* TCP subcommand table */
int domss(),doirtt(),dortt(),doretry(),dotcpstat(),dowindow(),dotcpkick(),
	dotcpreset(),dolport(),dotcptos(),dosendmss(),dotimemax(),AckOnPush();
struct scmds far tcpcmds[] = {
	{"ackonpush",AckOnPush},
	{"irtt",		doirtt},
	{"lport",	dolport},
	{"mss",		domss},
	{"prod",		dotcpkick},
	{"reset",	dotcpreset},
	{"retry",	doretry},
	{"rtt",		dortt},
#if APL
	{"status",	dotcpstat},
#endif
	{"smss",		dosendmss},
	{"timemax",	dotimemax},
	{"tos",		dotcptos},
	{"window",	dowindow},
	{"",		NULLFP},
};
int
dotcp(argc,argv)
int argc;
char *argv[];
{
	return subcmd(tcpcmds,argc,argv,0);
}

/* Eliminate a TCP connection */
#pragma argsused
static int dotcpreset(int argc,char *argv[])
{
	register struct tcb *tcb;
	extern char notval[];
	int i;

	if ((i = atoi(argv[1])) < tcbcount && tcpval(tcb = tcbindex[i])) {
		reset_tcp(tcb);
		return 0;
	}
	printf(notval);
	return 1;
}

/* print tcp default values */
static void near tcp_defaults(void)
{
	printf("TCP: IRTT %u ms Retry %u MSS %u SMSS %u Window %u TimeMax %u ms%s\n",
		tcp_stat.tcp_irtt,tcp_stat.tcp_retry,tcp_stat.tcp_mss,
		tcp_stat.tcp_smss,tcp_stat.tcp_window,tcp_stat.tcp_time_max,
		tcp_stat.cAckOnPush ? "" : " Delayed Ack");
}

/* Set initial round trip time for new connections */
static int AckOnPush(int argc,char *argv[])
{
	if(argc < 2)
		tcp_defaults();
	else
		tcp_stat.cAckOnPush = atoi(argv[1]);
	return 0;
}

/* Set initial round trip time for new connections */
static int
doirtt(argc,argv)
int argc;
char *argv[];
{
	if(argc < 2)
		tcp_defaults();
	else tcp_stat.tcp_irtt = atoi(argv[1]) * (strchr(argv[1],'m') ?
								1 : 1000);
	return 0;
}

/* Set maximum retry time */
static int
dotimemax(argc,argv)
int argc;
char *argv[];
{
	if(argc < 2)
		tcp_defaults();
	else tcp_stat.tcp_time_max = atoi(argv[1]) * (strchr(argv[1],'m') ?
								1 : 1000);
	return 0;
}

/* Set tcp retry count */
static int
doretry(argc,argv)
int argc;
char *argv[];
{
	if(argc < 2)
		tcp_defaults();
	else
		tcp_stat.tcp_retry = atoi(argv[1]);
	return 0;
}

/* Set smoothed round trip time for specified TCB */
#pragma argsused
static int dortt(int argc,char *argv[])
{
	register struct tcb *tcb;
	extern char notval[];
	long atol();
	int i;

	if ((i = atoi(argv[1])) < tcbcount && tcpval(tcb = tcbindex[i])) {
		tcb->srtt = atol(argv[2]);
		return 0;
	}
	printf(notval);
	return 1;
}

/* Set/examine local port */
static int
dolport(argc,argv)
int argc;
char *argv[];
{
	extern WORD tcp_lport,tcp_lport_i;

	if(argc < 2)
		printf("Lport = %u\n",tcp_lport);
	else
		tcp_lport = tcp_lport_i = atoi(argv[1]);
	return 0;
}

/* Set/examine TOS used by Telnet */
static int
dotcptos(argc,argv)
int argc;
char *argv[];
{
	if(argc < 2)
		printf("Telnet TOS = %u\n",tcp_stat.tel_tos);
	else
		tcp_stat.tel_tos = atoi(argv[1]);
	return 0;
}

/* Force a retransmission */
#pragma argsused
static int dotcpkick(int argc,char *argv[])
{
	register struct tcb *tcb;
	extern char notval[];
	int i;

	if ((i = atoi(argv[1])) < tcbcount && tcpval(tcb = tcbindex[i]) &&
						(kick_tcp(tcb) != -1)) {
		return 0;
	}
	printf(notval);
	return 1;
}

/* Set default maximum segment size for receive */
static int domss(int argc,char *argv[])
{
	if(argc < 2)
		tcp_defaults();
	else
		tcp_stat.tcp_mss = atoi(argv[1]);
	return 0;
}

/* Set default maximum segment size for send */
static int dosendmss(int argc,char *argv[])
{
	if(argc < 2)
		tcp_defaults();
	else
		tcp_stat.tcp_smss = atoi(argv[1]);
	return 0;
}

/* Set default window size */
static int
dowindow(argc,argv)
int argc;
char *argv[];
{
	if(argc < 2)
		tcp_defaults();
	else
		tcp_stat.tcp_window = atoi(argv[1]);
	return 0;
}

/* Display status of TCBs */
#if APL
static int
dotcpstat(argc,argv)
int argc;
char *argv[];
{
	register struct tcb *tcb;
	int i;
	extern char notval[];

	if(argc < 2){
		tstat();
	} else {
		if ((i = atoi(argv[1])) < tcbcount &&
						tcpval(tcb = tcbindex[i]))
			state_tcp(tcb,i);
		else
			printf(notval);
	}
	return 0;
}

/* Dump TCP stats and summary of all TCBs, build TCB index
 * Con# Rcv-Q Snd-Q  Local socket           Remote socket          State
 *   1      0     0  xxx.xxx.xxx.xxx:xxxxx  xxx.xxx.xxx.xxx:xxxxx  Established
 */

int tstat(void)
{
	register int i;
	register struct tcb *tcb;
	int num;
	char *psocket();
	PTCB far *ptcb;

	printf("TCP status:\n"
	   "ConOut %u ConIn %u ResetOut %u Runt %u ChksumErr %u Bdcsts %u\n",
		tcp_stat.conout,tcp_stat.conin,tcp_stat.resets,tcp_stat.runt,
		tcp_stat.checksum,tcp_stat.bdcsts);
	printf("Con# Rcv-Q  Snd-Q  Local socket          Remote socket         State\n");
	for(i=0,num = 0;i<NTCB;i++){
		for(tcb=tcp_stat.tcbs[i];tcb != NULLTCB;tcb = tcb->next){
			printf("%3u %6u %6u  ",num++,tcb->rcvcnt,tcb->sndcnt);
			printf("%-22s",psocket(&tcb->conn.local));
			printf("%-22s",psocket(&tcb->conn.remote));
			printf("%-s",tcpstates[tcb->state]);
			if(tcb->state == LISTEN && (tcb->flags & CLONE))
				printf(" (S)");
			printf("\n");
		}
	}
	if (tcbindex)
		farfree(tcbindex);
	if ((tcbindex = farmalloc(num * sizeof(TCB *))) != 0) {
		tcbcount = num;
		for (i = 0,ptcb = tcbindex;i<NTCB;i++) {
			for (tcb=tcp_stat.tcbs[i];tcb != NULLTCB;tcb = tcb->next) {
				*ptcb++ = tcb;
			}
		}
	}
	else
		tcbcount = 0;
	return 0;
}
/* Dump a TCP control block in detail */
static void state_tcp(struct tcb *tcb,int num)
{
	int32 sent,recvd;

	if(tcb == NULLTCB)
		return;
	/* Compute total data sent and received; take out SYN and FIN */
	sent = tcb->snd.una - tcb->iss;	/* Acknowledged data only */
	recvd = tcb->rcv.nxt - tcb->irs;
	switch(tcb->state){
	case LISTEN:
	case SYN_SENT:		/* Nothing received or acked yet */
		sent = recvd = 0;	
		break;
	case SYN_RECEIVED:
		recvd--;	/* Got SYN, no data acked yet */
		sent = 0;
		break;
	case ESTABLISHED:	/* Got and sent SYN */
	case FINWAIT1:		/* FIN not acked yet */
		sent--;
		recvd--;
		break;
	case FINWAIT2:		/* Our SYN and FIN both acked */
		sent -= 2;
		recvd--;
		break;
	case CLOSE_WAIT:	/* Got SYN and FIN, our FIN not yet acked */
	case CLOSING:
	case LAST_ACK:
		sent--;
		recvd -= 2;
		break;
	case TIME_WAIT:		/* Sent and received SYN/FIN, all acked */
		sent -= 2;
		recvd -= 2;
		break;
	}
	if (sent < 0)
		sent = 0;
	printf("TCP con# %u:\nLocal: %s",num,psocket(&tcb->conn.local));
	printf(" Remote: %s",psocket(&tcb->conn.remote));
	printf(" State: %s\n",tcpstates[tcb->state]);
	printf("  Init seq      Unack       Next Resent CWind Thrsh  Wind  MSS Queue     Total\n");
	printf("%10lu",tcb->iss);
	printf("%11lu",tcb->snd.una);
	printf("%11lu",tcb->snd.nxt);
	printf("%7lu",tcb->resent);
	printf("%6u",tcb->cwind);
	printf("%6u",tcb->ssthresh);
	printf("%6u",tcb->snd.wnd);
	printf("%5u",tcb->mss);
	printf("%6u",tcb->sndcnt);
	printf("%10lu\n",sent);

	printf("%10lu",tcb->irs);
	printf("           ");
	printf("%11lu",tcb->rcv.nxt);
	printf("%7lu",tcb->rerecv);
	printf("      ");
	printf("      ");
	printf("%6u",tcb->rcv.wnd);
	printf("     ");
	printf("%6u",tcb->rcvcnt);
	printf("%10lu\n",recvd);

	if(tcb->reseq != (struct reseq *)NULL){
		register struct reseq *rp;

		printf("Reassembly queue:\n");
		for(rp = tcb->reseq;rp != (struct reseq *)NULL; rp = rp->next){
			printf("  seq %lu %u bytes\n",rp->seg.seq,rp->length);
		}
	}
	if(tcb->backoff > 0 || tcb->retry > 0)
		printf("Backoff %u %u ",tcb->backoff,tcb->retry);
	if(tcb->flags & RETRAN)
		printf("Retrying ");
	switch(tcb->timer.state){
//	case TIMER_STOP:
//		printf("Timer stopped ");
//		break;
	case TIMER_RUN:
		printf("%simer: %ld/%ld ms\n",
		 (tcb->flags1 & F1_KATIMING) ? "Keep-alive t" : "T",
		 (long)MSPTICK * (tcb->timer.start - tcb->timer.count),
		 (long)MSPTICK * tcb->timer.start);
		break;
//	case TIMER_EXPIRE:
//		printf("Timer expired ");
	}
	printf("SRTT %ld ms Mean dev %ld ms\n",tcb->srtt,tcb->mdev);
}
#endif
