//
//   $Id: DataBase.java 1.2 2002/10/16 17:01:48Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
package inet;

import java.io.*;
import java.net.*;
import java.awt.*;
import java.util.*;
import java.awt.event.*;


/*
Any setting like key mapping, dialog name, screen size, etc.

  Contains 3 possibilities: the value, an override value and a default value.
*/

class KeyStroke
{
	int key;
	int modFlags;

	KeyStroke(int key, int modFlags)
	{
		this.key = key;
		this.modFlags = modFlags;
	}

	public int hashCode()
	{
		return key ^ modFlags;
	}

	public boolean equals(Object o)
	{
		if (!(o instanceof KeyStroke))
			return false;
		KeyStroke k = (KeyStroke)o;
		return k.key == key && k.modFlags == modFlags;
	}
}

class Int
{
	//thoe only thing this class contain
	int i;

	Int (int i) {
		this.i = i;
	}
	public int hashCode()
	{
		return i;
	}

	public boolean equals(Object o)
	{
		if (!(o instanceof Int))
			return false;
		return ((Int)o).i == i;
	}
	Int set(int n)
	{
		i = n;
		return this;
	}
	public String toString()
	{
		return ""+i;
	}
}

class Setting
{
	String key;
	Object val;
	int flags = 0;
	static final int FINAL = 1;//cannot be overriden
	//static final int FINAL = 1;//cannot be overriden
	//static final int FINAL = 1;//cannot be overriden

	Setting(DataBase db, String key, Object val) 
	{
		db.put(key, this);
		this.key = key;
		this.val = val;
	}

	Setting(DataBase db, String key, int val)
	{
		this(db, key, new Int(val));
	}

	String getS()
	{
		return val instanceof String ? (String)val : null;
	}

	/*
	called when setting already exists and see if it can be edited, etc.
	Look at things like unsettable, already set, etc.
	*/
	void modify(Setting s)
	{
	}

	int getI()
	{
		return val instanceof Int ? ((Int)val).i : -1;
	}

	Color getColor()
	{
		return val instanceof Color ? (Color)val : null;
	}

	Entry getEntry()
	{
		return val instanceof Entry ? (Entry)val : null;
	}
	
	Object getVal()
	{
		return val;
	}
	

	static int stringToInt(String val)
	{

		try {
			return Integer.parseInt(val);
		}
		catch(Exception e) {
			return -1;
		}
	}
	static int hexStringToInt(String val)
	{
		try {
			return Integer.parseInt(val, 16);
		}
		catch(Exception e) {
			return -1;
		}
	}
	static Color stringToColor(String val)
	{
		try {
			return new Color(hexStringToInt(val));
		}
		catch(Exception e) {
			return null;
		}
	}
}

class Entry implements Cloneable
{
	public static final int KEYMAP = 0;
	public static final int SETTING = 1;
	//public static final int 
	int id;//position in file, needed for writing file in correct order
	int[] firstmap;
	int[] secondmap;
	int key, mods;
	String inetRep;
	String description;
	String comment;
	String setting;
	String value; 

//		return inetRep + " key:"+ key + " mods:" + mods+" id:"+id;
	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		if (inetRep != null)
			sb.append("i{").append(inetRep).append("} ");
		if (firstmap != null)
			sb.append("m{").append(DataBase.getHexStringFromInts(firstmap)).append("} ");
		if (secondmap != null)
			sb.append("n{").append(DataBase.getHexStringFromInts(secondmap)).append("} ");
		if (description != null)
			sb.append("d{").append(description).append("} ");
		if (setting != null)
			sb.append("s{").append(setting).append("} ");
		if (value != null)
			sb.append("v{").append(value).append("} ");
		if (comment != null)
			sb.append(comment);
		return sb.toString();
	}
}



class DataBase
{

	Hashtable keyTable = new Hashtable();
	private static Hashtable table = new Hashtable();//A database containing key codes to defined inet string variables
	static Hashtable defsets = new Hashtable();
	Hashtable settings = new Hashtable();//settings (customization + internationilation)
	Hashtable keys = new Hashtable();	//mapped keys are contained here
	Hashtable indexer = new Hashtable();

	//little ugly, must use as synchronized
	Int itemp = new Int(0);
	//utility function
	Int setI(int i) 
	{
		return itemp.set(i);
	}


	private boolean modified = false;
	private String pathname;//pathname coud be anything like c:\xyz, deva/etc/xyz, http://at/... or ftp://196.10....
	/*
	All the specific settings classes may use.
	Setting's constructor sets the default value, thus all these settings are guarenteed to
	provide a meaningful value. Other values can be set with setupfiles or dialogs (customizing)

    All are added to database and indexed with [number / string?] so they can be reached for
    customization
	*/
	public static final String
		Connect_title = addToProps("Enter connection parameters"),
		Connect_remport = addToProps("remport", "23"),
		Connect_remhost = addToProps("remhost", ""),
		KeyMap_title    = addToProps("Setup for keybord mappings"),
		CurSpeed_title  = addToProps("Set cursor speed"),
		Setup_title     = addToProps("General setup options"),

		Label_ENVal  = addToProps("Enter new values"),
		Label_presKey = addToProps("Press any key"),
		Label_curAss = addToProps("Current assignment"),
		Label_mapTo  = addToProps("map to"),
		Label_mapToH = addToProps("in hex"),
		Label_func   = addToProps("function"),
		Label_host      = addToProps("Host to log on to"),
		Label_port      = addToProps("Remote port number"),
		Mes_HostNotFound = addToProps("Inet cannot establish a connection to the specified host and port."),
		Mes_ConFailed = addToProps("Connection failed"),

		ConNotS_title = addToProps("ConNotS.title", "Settings not saved - "),
		ConNotS_Mes = addToProps("ConNotS.Mes", "Do you want to save your settings?"),
		Screen_rows     = addToProps("Screen.rows", "24"),
		Screen_cols     = addToProps("Screen.cols", "80"),
		Screen_pages    = addToProps("Screen.pages", "6"),
		F_FACE = addToProps("Screen.fontface", "Courier"),
		F_SIZE = addToProps("Screen.fontsize", "13"),

		SET_LOAD = addToProps("Load settings"),
		SET_SAVE = addToProps("Save settings"),
		SET_TITLE = addToProps("Settings for Inet java terminal emulators"),

		//all button labels
		ok = addToProps("Ok"),
		cancel = addToProps("Cancel"),
		yes = addToProps("Yes"),
		no = addToProps("No"),
		retry = addToProps("Retry"),
		done = addToProps("Done"),
		clear = addToProps("Clear"),
		assign = addToProps("Assign")
	;
/*
	static Entry getEntry(String key, Object o)
	{
		Entry e = new Entry(o);
		defsets.put (key, e);
		return e;
	}
*/
	static String addToProps (String name, String value)
	{
		defsets.put(name, value);
		return name;
	}

	static String addToProps (String name)
	{
		return addToProps(name, name);
	}

	boolean isModified()
	{
		return modified;
	}
	/**
	Gets the value of the specified key. Return null if key doesnt exist
	*/
	String getS(String key)
	{
		Object o = settings.get(key);
		if (o == null)
			return (String)defsets.get(key);
		return ((Entry)o).value;
	}

	/**
	Gets the value of the specified key. Return the defualt supplied if key doesnt exist
	*/
	String getS(String key, String def)
	{
		String s = getS(key);
		if (null == s) return def;
		return s;
	}

	int getI(String key)
	{
		try {
			return Integer.parseInt(getS(key));
		}
		catch(Exception e) {
			return -1;
		}
	}
	
	int getHI(String key)
	{
		try {
			return Integer.parseInt(getS(key), 16);
		}
		catch(Exception e) {
			return -1;
		}
	}

	Color getColor(String key)
	{
		try {
			return new Color(getHI(key));
		}
		catch(Exception e) {
			return null;
		}
	}
	
	int getI(String key, int def) {
		int i = getI(key);
		if (-1 == i) return def;
		return i;
	}


	/**
	insert a new entry into the table
	*/
	void insertKey(Entry e)
	{
		if (e.inetRep != null)
			keyTable.put(new Int(e.mods + (e.key << 4)), e);
		else if (e.setting != null)
			settings.put(e.setting, e);
		indexer.put(new Int(e.id), e);
	}

	void addSetting(String setting, String value)
	{
		Entry e = new Entry();
		e.setting = setting;
		e.value = value;
		insertKey(e);
		modified = true;
	}

	//adds a new keymapping to the lot
	void addKey(int key, int mods, String hexString)
	{
		Entry e = new Entry();
		e.key = key;
		e.mods = mods;
		e.inetRep = getInetRep(key, mods);
		e.id = id++;
		e.firstmap = getIntsFromHexString(hexString);
		insertKey(e);
		modified = true;
	}

	//Only settings can be put in database
	void put(String s, Setting o)
	{
		Object b = settings.get(s);
		if (b == null)//not yet contained
			settings.put(s, o);
		else
			((Setting)b).modify(o);
	}

	static String getInetRep(int key, int mods)
	{
		Int itemp = new Int(0);
		String s = "";
		if ((mods & InputEvent.CTRL_MASK) != 0)
			s += table.get(itemp.set(KeyEvent.VK_CONTROL)) + " ";
		if ((mods & InputEvent.ALT_MASK) != 0)
			s += table.get(itemp.set(KeyEvent.VK_ALT)) + " ";
		if ((mods & InputEvent.SHIFT_MASK) != 0)
			s += table.get(itemp.set(KeyEvent.VK_SHIFT)) + " ";
		if ((mods & InputEvent.META_MASK) != 0)
			s += table.get(itemp.set(KeyEvent.VK_META)) + " ";
		s += table.get(itemp.set(key));
		return s;
	}

	void remove(int key, int mods)
	{
		Entry e = (Entry)keyTable.remove(itemp.set(mods + (key << 4)));
		if (e == null)
			return;
		indexer.remove(itemp.set(e.id));
		modified = true;
	}

	Entry getKeyMap(int key, int mods)
	{
		return (Entry)keyTable.get(itemp.set(mods + (key << 4)));
	}
	/**
	query functions to database

  May grow to long list according to need
	*/
	public int[] getMappingByKey(int key, int mods)
	{
		Entry e = getKeyMap(key, mods);
		if (e != null)
			return e.firstmap;
		return null;
	}

	static int[] getIntsFromHexString(String s)
	{
		try {
			StringTokenizer st = new StringTokenizer(s," ,;.\t");
			int[] ret = new int[st.countTokens()];
			for (char i = 0; i < ret.length; i++)
				ret[i] = Integer.parseInt(st.nextToken(), 16);
			return ret;
		}
		catch(Exception e) {
			if (Debug.dv)
				Debug.out(e+"  "+s);
			return null;
		}
	}

	public static String getHexStringFromInts(int[] ints)
	{
		StringBuffer hex = new StringBuffer();
		for (char i = 0; i < ints.length; i++)
			hex.append(Integer.toHexString(ints[i])).append(' ');
		return new String(hex);
	}



	/**
	Creates a database from parsing a file. The file should be at the same place as
	from where the classes are loaded, as this class' classloader is used to load it.
	*/
		/**
	Creates a database by parsing a stream
	*/
	private int id = 0;
	public DataBase(String file)
	{
		InputStream in;
		try {
			//Debug.out(getClass()+" "+file);
			Class c = getClass();
/*			URL u = c.getResource(file);
			String s1 = u.getFile();
			String s2 = u.getProtocol();
			String s3 = u.getRef();
			Debug.out("url = "+u+" file "+s1+" prot "+s2+" ref "+s3);//+""++""
*/
			//String f = c.getResource(file).getFile();
			//pathname = f.substring(f.indexOf('/'));
			pathname = file;
			in = c.getResourceAsStream(file);
			//Debug.out(in);

			int l = 0;
			byte b;
			final int SIZE = 1024;
			byte[] data = new byte[SIZE];
			while (true) {
				int i = in.read();
				data[l++] = b = (byte)i;
				if (b == '\r' || b == '\n' || i == -1 || l == SIZE) {
					if (l > 1) {
						doLine(data, l - 1);
					}
					l = 0;
					if (i == -1)
						break;
				}
			}
		}
		catch (IOException e)
		{
			if (Debug.dv)
				Debug.out(e);
		}
	}

	/*
	Format of line: eg Shift+F1=a0 12 or B=42
	*/
	private void doLine(byte[] line, int length)
	{
		
	}
	
	private void doLineBracketed(byte[] line, int length)
	{
		int i = 0, start, end;
		char type;
		String s;
		String comment;
		Entry k = new Entry();
		while(true) {
			while ("imndsv#".indexOf(line[i]) < 0 && i < length)
				i++;
			if (i >= length) 
				break;
			type = (char)line[i];
			if (type == '#') {
				k.comment = new String (line, i, length - i);
				break;
			}
			while (line[i++] != '{' && i < length);
			start = i;//inclusive
			while (line[++i] != '}' && i < length);
			end = i;//exclusive
			s = new String(line, start, end - start);
			switch (type) {
			case 'i'://key mapping
				k.inetRep = s;
				StringTokenizer st = new StringTokenizer(s, " \t+");
				while (st.hasMoreTokens()) {
					int key = ((Int)table.get(st.nextToken())).i;
					switch (key) {
					case 0://undef
						Debug.out("Undefined key:" + k+"  id"+id);
						break; 
					case KeyEvent.VK_ALT:
						k.mods |= InputEvent.ALT_MASK;
						break;
					case KeyEvent.VK_CONTROL:
						k.mods |= InputEvent.CTRL_MASK;
						break;
					case KeyEvent.VK_META:
						k.mods |= InputEvent.META_MASK;
						break;
					case KeyEvent.VK_SHIFT:
						k.mods |= InputEvent.SHIFT_MASK;
						break;
					default:
						k.key = key;
					}
				}
				break;
			case 'm'://first mapping
				k.firstmap = getIntsFromHexString(s);
				break;
			case 'n'://second mapping
				k.secondmap = getIntsFromHexString(s);
				break;
			case 'd'://description
				k.description = s;
				break;
			case 's'://setting
				k.setting = s;
				break;
			case 'v'://value
				k.value = s;
				break;
			}
		}
		k.id = id++;
		insertKey(k);
	}

	private static void add(String inetRep, int key)
	{
		Int i = new Int(key);
		table.put(inetRep, i);
		table.put(i, inetRep);//other way around as well for saving
	}

	private static void add(int c)
	{
		add(""+(char)c, c);
	}

	static
	{
		table = new Hashtable(60);
		add("F1", KeyEvent.VK_F1);
		add("F2", KeyEvent.VK_F2);
		add("F3", KeyEvent.VK_F3);
		add("F4", KeyEvent.VK_F4);
		add("F5", KeyEvent.VK_F5);
		add("F6", KeyEvent.VK_F6);
		add("F7", KeyEvent.VK_F7);
		add("F8", KeyEvent.VK_F8);
		add("F9", KeyEvent.VK_F9);
		add("F10", KeyEvent.VK_F10);
		add("F11", KeyEvent.VK_F11);
		add("F12", KeyEvent.VK_F12);

		add("Alt", KeyEvent.VK_ALT);
		add("Ctrl", KeyEvent.VK_CONTROL);
		add("Meta", KeyEvent.VK_META);
		add("Shift", KeyEvent.VK_SHIFT);

		add("upArrow", KeyEvent.VK_UP);
		add("downArrow", KeyEvent.VK_DOWN);
		add("leftArrow", KeyEvent.VK_LEFT);
		add("rightArrow", KeyEvent.VK_RIGHT);

		add("Home", KeyEvent.VK_HOME);
		add("End", KeyEvent.VK_END);
		add("Ins", KeyEvent.VK_INSERT);
		add("Del", KeyEvent.VK_DELETE);

		add("PgUp", KeyEvent.VK_PAGE_UP);
		add("PgDn", KeyEvent.VK_PAGE_DOWN);

		add("Accept", KeyEvent.VK_ACCEPT);
		add("+", KeyEvent.VK_ADD);
		add("'", KeyEvent.VK_BACK_QUOTE);
		add("\\", KeyEvent.VK_BACK_SLASH);
		add("Backspace", KeyEvent.VK_BACK_SPACE);
		add("Cancel", KeyEvent.VK_CANCEL);
		add("Clear", KeyEvent.VK_CLEAR);
		add("Convert", KeyEvent.VK_CONVERT);
		add("Decimal", KeyEvent.VK_DECIMAL);
		add("Divide", KeyEvent.VK_DIVIDE);
		add("Enter", KeyEvent.VK_ENTER);
		add("esc", KeyEvent.VK_ESCAPE);
		add("Final", KeyEvent.VK_FINAL);
		add("Help", KeyEvent.VK_HELP);
		add("Kana", KeyEvent.VK_KANA);
		add("Kanji", KeyEvent.VK_KANJI);
		add("Modechange", KeyEvent.VK_MODECHANGE);
		add("*", KeyEvent.VK_MULTIPLY);
		add("Nonconvert", KeyEvent.VK_NONCONVERT);
		add("Num_lock", KeyEvent.VK_NUM_LOCK);
		add("Caps_lock", KeyEvent.VK_CAPS_LOCK);
		add("Scroll_Lock", KeyEvent.VK_SCROLL_LOCK);
		add("Numpad_0", KeyEvent.VK_NUMPAD0);
		add("Numpad_1", KeyEvent.VK_NUMPAD1);
		add("Numpad_2", KeyEvent.VK_NUMPAD2);
		add("Numpad_3", KeyEvent.VK_NUMPAD3);
		add("Numpad_4", KeyEvent.VK_NUMPAD4);
		add("Numpad_5", KeyEvent.VK_NUMPAD5);
		add("Numpad_6", KeyEvent.VK_NUMPAD6);
		add("Numpad_7", KeyEvent.VK_NUMPAD7);
		add("Numpad_8", KeyEvent.VK_NUMPAD8);
		add("Numpad_9", KeyEvent.VK_NUMPAD9);
		add("[", KeyEvent.VK_OPEN_BRACKET);
		add("]", KeyEvent.VK_CLOSE_BRACKET);
		add("Pause", KeyEvent.VK_PAUSE);
		add(".", KeyEvent.VK_PERIOD);
		add(",", KeyEvent.VK_COMMA);
		add("PrnScr", KeyEvent.VK_PRINTSCREEN);
		add("quote", KeyEvent.VK_QUOTE);
		add(";", KeyEvent.VK_SEMICOLON);
		add("Separater", KeyEvent.VK_SEPARATER);
		add("=", KeyEvent.VK_EQUALS);
		add("/", KeyEvent.VK_SLASH);
		add("space", KeyEvent.VK_SPACE);
		add("-", KeyEvent.VK_SUBTRACT);
		add("Tab", KeyEvent.VK_TAB);
		int c;
		for (c = 'A'; c <= 'Z'; c++)
			add(c);
		for (c = '0'; c <= '9'; c++)
			add(c);
	}

	void save(OutputStream o)
		throws IOException
	{
		Object e = null;
		int id = 0;
		byte[] nl = {13, 10};
		while (true) {
			if (indexer.isEmpty())
				break;
			e = indexer.remove(itemp.set(id++));
			if (e == null)
				continue;
			//writes all data to outputstream
			o.write(e.toString().getBytes());
			o.write(nl);
		}
		o.flush();
	}

	void saveAs(Frame cFrame)
		throws IOException
	{
		try {
			OutputStream os = new FileOutputStream(pathname);
			save(os);
			os.close();
			modified = false;
			return;
		}
		catch (Exception e) {
		}
		FileDialog d = new FileDialog(cFrame, "Save settings", FileDialog.SAVE);
		d.show();
		if (d.getDirectory() == null)
			return;
		pathname = d.getDirectory() + d.getFile();
		saveAs(cFrame);
	}
}
