//
//   $Id: DIRUTIL.C 1.7 2002/11/21 17:57:00Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//	dirutil.c
//
// Copyright (C) 2001 Datalight, Inc.
// All Rights Reserved
// Datalight Confidential and Proprietary
// Portions Copyright (c) GP van Niekerk Ondernemings 1990-2000
//
// Directory handling functions
//
// Revision History
// 07/27/2001 gvn added SetSchedule prototype


#include <stdio.h>
#include <sys\stat.h>
#include <dos.h>
#include <dir.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <alloc.h>
#include <time.h>
#include "dostruct.h"	//  DOS structures
#include "dosdef.h"
#include "datalght.h"
#include "dldosio.h" 	//  some int 21h (DOS)
#include "dllfn.h"  		//  LFN and LFN "smart" functions
#include "dirutil.h"

char upper_case;		// flag for UPPER case display

extern int errno;
extern char *sys_errlist[];

#ifndef	FALSE
#define	FALSE	(0)
#endif

#ifndef	TRUE
#define	TRUE	!(FALSE)
#endif

#define	SET_DTA		0x1a
#define	FIND_FIRST	0x4e
#define	FIND_NEXT	0x4f

#define	NULLENT	(PFIND)0





long dmytodays(int day,int month,int year);
void list1(PFIND sbuf,int full,FILE *file,char *path);

// Create a directory listing in a temp file and return the resulting file
// descriptor. If full == 1, give a full listing; else return just a list
// of names.

FILE *dir(char *path,char full)
{
	FILE *fp,*tmpfile();

	fp = tmpfile();
	getdir(path,full,fp);
	fseek(fp,0L,0);
	return fp;
}

// do a directory list to the stream
// full = 0 -> short form, 1 is long
/* FINDFIRST attributes */
#define A_READONLY	0x01
#define A_HIDDEN		0x02
#define A_SYSTEM		0x04
#define A_VOLUME		0x08
#define A_SUBDIR		0x10
#define A_ARCHIVE		0x20
#define A_DEVICE		0x40
#define A_ALLFILES	0x06	/* hidden and system */
#define ALL_ATTR		0x16

#pragma argsused
getdir(char *path,char full,FILE *file)
{
	FIND sbuf;
	struct stat statbuf;
	register char *cp;
	char cflag = 0;
	bool	bFound;

	// Root directory is a special case
	if(path == 0 || *path == '\0' || strcmp(path,"\\") == 0 || 
               strcmp(path,"/") == 0)
		path = "/*.*";
	else if(strchr(path,'*') == 0
	 && stat(path,&statbuf) != -1
	 && (statbuf.st_mode & S_IFDIR)) {
		if((cp = malloc(strlen(path) + 10)) == 0)
			return -1;
		sprintf(cp,"%s/%s",path,"*.*");
		path = cp;
		cflag = 1;
	}
	// do directories first
	bFound = SmartFindFirst(A_SUBDIR, path, &sbuf);
	while (bFound) {
		if ((sbuf.attr & FA_DIREC) && sbuf.name[0] != '.') {	// exclude "." and ".."
			list1(&sbuf,full,file,path);
		} // if (".")
		bFound = SmartFindNext(&sbuf);
	}
	SmartFindClose(&sbuf);
	bFound = SmartFindFirst(0, path, &sbuf);
	while (bFound) {
		list1(&sbuf,full,file,path);
		bFound = SmartFindNext(&sbuf);
	}
	SmartFindClose(&sbuf);
	if(cflag)
		free(path);
	return 0;
}

void list1(PFIND sbuf,int full,FILE *file, char *path)
{
	char cbuf[30];				// for making line_buf
	char szFileName[260],*pszFileName,szPath[260];
	struct tm tm;
	long lCurDateDays, lFileDateDays;

//ff_fdate 15..........98.....54.......0
// ͹
//		     ͼ
//		   Years since   Month     Day
//			  1980

struct tm {
  int tm_sec;      // Seconds
  int tm_min;      // Minutes
  int tm_hour;     // Hour (0--23)
  int tm_mday;     // Day of month (1--31)
  int tm_mon;      // Month (0--11)
  int tm_year;     // Year (calendar year minus 1900)
  int tm_wday;     // Weekday (0--6; Sunday = 0)
  int tm_yday;     // Day of year (0--365)
  int tm_isdst;    // 0 if daylight savings time is not in effect)
};

struct date tCurrentDOSDate;

	strcpy(szPath,path);
	if ((pszFileName = strrchr(szPath,'/')) == 0)
		pszFileName = szPath;
	else
		++pszFileName;
	strcpy(pszFileName,sbuf->name);
	if (SmartExpandPath(FULLPATH_LONGNAME, szPath, szFileName))
	{
		if ((pszFileName = strrchr(szFileName,'\\')) == 0)
			pszFileName = szFileName;
		else
			++pszFileName;
	}
	else
		pszFileName = sbuf->name;
//	if (!upper_case)
//		strlwr(pszFileName);
	if (full) {
		tm.tm_year = (sbuf->date >> 9) + 80;
		tm.tm_mon  = ((sbuf->date >> 5) & 0xf) - 1;
		tm.tm_mday = (sbuf->date & 0x1f);

      tm.tm_hour = (sbuf->time >> 11);
      tm.tm_min  = (sbuf->time >> 5) & 0x3f;
      tm.tm_sec  = (sbuf->time & 0x1f);
//drwxrwxrwx   1 owner   group         512 Sep  6  1994 bin
//-rwxrwxrwx   1 owner   group     5062418 Apr 10 04:46 dls-lR
      getdate(&tCurrentDOSDate);
      lCurDateDays = dmytodays(tCurrentDOSDate.da_day,
                              tCurrentDOSDate.da_mon, 
                              tCurrentDOSDate.da_year);
      lFileDateDays = dmytodays(tm.tm_mday, tm.tm_mon+1, tm.tm_year);
      
      if (lFileDateDays >= lCurDateDays - 180)
         strftime(cbuf, sizeof(cbuf), "%b %d %H:%M",&tm);
      else
   		strftime(cbuf,sizeof(cbuf),"%b %d  %Y",&tm);
		fprintf(file,"%crwxrwxrwx   1 owner   group %12lu %s %s\r\n",
				(sbuf->attr & FA_DIREC) ? 'd' : '-',sbuf->size,
				cbuf,pszFileName);
	}
	else {
		fprintf(file,"%s%s\r\n",pszFileName,(sbuf->attr & FA_DIREC) ? "/" : "");
	}
}

//
// Day-Month-Year ->  days since 1980
//
long dmytodays(int day,int month,int year)
{
   int dpm[]=     {31,28,31,30,31,30,31,31,30,31,30,31};
   int dpmleap[]= {31,29,31,30,31,30,31,31,30,31,30,31};

   long days=0l;
   int x;

   //
   // Range check
   //
   if (year < 80) 
   {
      year += 2000;
   }
   
   if (year < 1980) 
   {
      year += 1900;
   }
   
   if (!month || month > 12) 
   {
      month = 1;
   }

      
   if (!day)
   {
      day = 1;
   }
   
   //
   // Adjust for year
   //

   for (x=1980; x!=year;x++)
   {
      if (x&3) 
         days += 365;
      else 
         days += 366;
   }

   //
   // Adjust for month
   //

   for (x=0; x<month; x++)
   {
      if (year & 3)
         days += dpm[x];
      else
         days += dpmleap[x];
   }


   //
   // Adjust for day
   //

   days += day-1;


   return days;
}
