//
//   $Id: FTPACOM.C 1.3 2003/02/15 02:52:26Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   Portions Copyright (c) 1993-2002 GP van Niekerk Ondernemings
//   Portions Copyright (c) 1986-1989 Phil Karn, KA9Q
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
ftpcom.c
Routines common to both the FTP server and client
Updated @ Date 16 May 1996, 10:53 by Gerrit van Niekerk
*/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <alloc.h>
#include <string.h>
#include <errno.h>
#include "inetdefs.h"
#include "comiler.h"
#include "api.h"
#include "rftp.h"
#include "ftps.h"

extern char rxok[],txok[],rxerr[],txerr[],noconn[];

struct ftp *ftps;		// start of ftp chain
struct ftp *ftpe;		// end of ftp chain

/* FTP Data channel Receive upcall handler */
void FtpDataReceiveHandler(struct ftp *ftp)
{
	TCB far *tcb;
	int cnt;

	if (ftp->data == 0)
		return;	// ???
	tcb = (TCB far *)MK_FP(tcbs,ftp->data);
	if (ftp->state != RECEIVING_STATE && ftp->state != LIST_STATE &&
		ftp->state != RECV_RDY_STATE && ftp->state != LIST_RDY_STATE) {
//		if (tcb->state == FINWAIT2)
//			reset_tcp(FP_OFF(tcb));	/* shut him up for good */
		return;
	}
	while ((cnt = RecvTcp(FP_OFF(tcb),ftp->dbuf,DBUFSZ)) > 0) {
		if (ftp->fp && fwrite(ftp->dbuf,1,cnt,ftp->fp) != cnt) {
			ftp->error = errno;
			goto close_it;
		}
	}
	if (cnt != - WOULDBLK && cnt != -RE_ENTER) {
		close_it:
		CloseTcp(ftp->data);
		ftp->data = 0;
		if (ftp->state == RECEIVING_STATE ||
						ftp->state == RECV_RDY_STATE) {
			// End of file received on incoming file
			fclose(ftp->fp);
			ftp->fp = 0;
			if (!(ftp->flags & FF_CLIENT)) {
				// server done
				if (tcb->reason != R_NORMAL)
					ftp->error = -1;
				if (ftp->error) {
					if (ftp->error != -1) {
						Tprintf(ftp,rxerr);
					}
					else {
						// Data connection was reset, complain about it
						Tprintf(ftp,noconn);
					}
					ftp->status = FSTAT_RX_ERR;
					ftp->error = 0;
				}
				else {
					Tprintf(ftp,rxok);
					ftp->status = FSTAT_RX_OK;
				}
				strncpy(ftp->status_line,ftp->filename,99);
				ftp->state = COMMAND_STATE;
			}
			else if (ftp->state == RECV_RDY_STATE) {
				goto rcvok;
			}
			else {
				ftp->state = RECEIVED_STATE;
			}
		}
		else if (ftp->state == LIST_STATE || ftp->state == LIST_RDY_STATE) {
			rewind(ftp->fplist = ftp->fp);
			ftp->fp = 0;
			if (ftp->state == LIST_RDY_STATE) {
				if (ftp->type == IMAGE_TYPE)
					Tprintf(ftp,"TYPE I\r\n");
				rcvok:
				ftp->state = COMMAND_STATE;
				if (ftp->fplist) {
					startget(ftp,0,0);
				}
			}
			else
				ftp->state = LISTED_STATE;
		}
	}
}

/* FTP Data channel Transmit upcall handler */
void FtpDataTransmitHandler(struct ftp *ftp)
{
	TCB far *tcb;
	int cnt;

	if (ftp->data == 0)
		return;	// ???
	tcb = (TCB far *)MK_FP(tcbs,ftp->data);

	if ((ftp->state != SENDING_STATE && ftp->state != SEND_RDY_STATE) || ftp->fp == 0) {
		return;
	}
	if ((cnt = RecvTcp(ftp->data,0,0)) != 0) {
		goto tcp_error;
	}
	if (tcb->state >= LISTEN && tcb->state < ESTABLISHED)
		return;
	if (!ftp->dcnt && !(ftp->flags & FF_SNDEOF)) {
		if ((cnt = min(DBUFSZ,tcb->window - tcb->sndcnt)) != 0) {
			ftp->dcnt = fread(ftp->dbuf,1,cnt,ftp->fp);
			ftp->dbufc = ftp->dbuf;
			if (ftp->dcnt != cnt)
				ftp->flags |= FF_SNDEOF;
			if (ftp->dcnt > 0) {
				if (ftp->type == ASCII_TYPE) {
					char *p;
					if ((p = memchr(ftp->dbuf,CTLZ,ftp->dcnt)) != 0) {
						ftp->dcnt = p - ftp->dbuf;
						ftp->flags |= FF_SNDEOF;
					}
				}
			}
		}
	}
	while (ftp->dcnt) {
		if ((cnt = SendTcp(FP_OFF(tcb),ftp->dbufc,ftp->dcnt)) < 0) {
			if (cnt == -NO_SPACE || cnt == -WOULDBLK || cnt == -RE_ENTER)
				return;		// try again later
			tcp_error:
			ftp->error = -1;
			goto close;
		}
		ftp->dcnt -= cnt;
		ftp->dbufc += cnt;
	}
	if (ftp->flags & FF_SNDEOF) {		// EOF seen
		if (ferror(ftp->fp) != 0)
			ftp->error = errno;
		close:
		fclose(ftp->fp);
		ftp->fp = NULL;
		CloseTcp(FP_OFF(tcb));
		ftp->data = 0;
		if (!(ftp->flags & FF_CLIENT)) {
			// server done
			if (ftp->error) {
				if (ftp->error != -1) {
					Tprintf(ftp,txerr);
				}
				else if (tcb->reason != R_NORMAL) {
					// Data connection was reset, complain about it
					Tprintf(ftp,noconn);
				}
				ftp->status = FSTAT_TX_ERR;
				ftp->error = 0;
			}
			else {
				Tprintf(ftp,txok);
				ftp->status = FSTAT_TX_OK;
			}
			strncpy(ftp->status_line,ftp->filename,99);
			ftp->state = COMMAND_STATE;
		}
		else {
			ftp->state = SENT_STATE;
		}
	}
}

/* Allocate an FTP control block */
struct ftp *ftp_create()
{
	register struct ftp *ftp;
	static int id;

	if((ftp = (struct ftp *)calloc(1,sizeof (struct ftp))) == NULLFTP)
		return NULLFTP;
	if ((ftp->buf = malloc(CBUFSZ)) == 0) {
		free(ftp);
		return NULLFTP;
	}
	if ((ftp->dbuf = malloc(DBUFSZ)) == 0) {
		free(ftp->buf);
		free(ftp);
		return NULLFTP;
	}
	if (ftpe) {
		ftp->prev = ftpe;
		ftpe->next = ftp;
		ftpe = ftp;
	}
	else
		ftpe = ftps = ftp;
	ftp->state = COMMAND_STATE;
	ftp->id = ++id;
	ftp->type = ASCII_TYPE;	/* Default transfer type */
	return ftp;
}

// Get ftp control block, checking for validity
struct ftp *get_ftp(FTPPar far *pf)
{
	struct ftp *ftp;
	int par5 = pf->par5;
	int par6 = pf->par6;

	for (ftp = ftps;ftp;ftp = ftp->next) {
		if (ftp == (struct ftp *)par5 && ftp->id == par6)
			break;
	}
	return ftp;
}

/* Free resources, delete control block */
void ftp_delete(register struct ftp *ftp)
{
	struct ftp *ftp1;

	for (ftp1 = ftps;ftp1;ftp1 = ftp1->next) {
		if (ftp == ftp1 && ftp->id == ftp1->id)
			goto found;
	}
	return;		// does not exist!
	found:
	CloseTcp(ftp->control);
	DelTcp(ftp->control);
	if (ftp->fp != NULL)
		fclose(ftp->fp);
	if (ftp->data != 0) {
		CloseTcp(ftp->data);
		DelTcp(ftp->data);
	}
	if (ftp->username != NULL)
		free(ftp->username);
	if (ftp->path != NULL)
		free(ftp->path);
	if (ftp->buf != NULL)
		free(ftp->buf);
	if (ftp->dbuf != NULL)
		free(ftp->dbuf);
	if (ftp->linebuf != NULL)
		free(ftp->linebuf);
	if (ftp->cd != NULL)
		free(ftp->cd);
	if (ftp->oldname != NULL)
		free(ftp->oldname);
	if (ftp->remotename != NULL)
		free(ftp->remotename);
	if (ftp->filename)
		free(ftp->filename);
	if (ftp->next)
		ftp->next->prev = ftp->prev;
	else
		ftpe = ftp->prev;
	if (ftp->prev)
		ftp->prev->next = ftp->next;
	else
		ftps = ftp->next;
	free((char *)ftp);
}
