//
//   $Id: arpstat.c 1.2 2002/10/16 17:06:36Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
	$DOCFILE:ARPSTAT.C

	Copyright (C) 2002 Datalight, Inc.
	All Rights Reserved

	ARP API demonstration program

	WARNING: The listing of ARP entries depends on internal
	Sockets structures which may change in new revisions.

	$DOCHISTORY:
	10/01/2002  gvn first version
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dos.h>
#include "compiler.h"
#include "capi.h"
#include "api.h"				// For TIMER structure
#include "arpapi.h"

#define	MSPTICK		55		// Milliseconds per tick

char *arptypes[] = {
	"",
	"10 Mb Ethernet",
	"3 Mb Ethernet",
	"AX.25",
	"Pronet",
	"Chaos",
	"IEEE",
	"Arcnet",
	"AppleTalk",
	"IPX",
};

char *IPAddressString(DWORD dwIPAddress)
{
	static char sz[20];

	sprintf(sz, "%u.%u.%u.%u",
			((BYTE *)&dwIPAddress)[0],((BYTE *)&dwIPAddress)[1],
			((BYTE *)&dwIPAddress)[2],((BYTE *)&dwIPAddress)[3]);
	return sz;
}

/* Convert an internet address (in host byte order) to a dotted decimal ascii
 * string, e.g., 255.255.255.255\0
 */
char *inet_ntoa(DWORD a)
{
	static char buf[16];

	sprintf(buf,"%u.%u.%u.%u",
		((BYTE *)&a)[3],
		((BYTE *)&a)[2],
		((BYTE *)&a)[1],
		((BYTE *)&a)[0]);
	return buf;
}

int main(int argc, char *argv[])
{
	WORD wLen,wSeg;
	char szNameBuf[80], szVersion[20];
	int i,iVersion,iNumber = 0;
	DWORD dwIPAddress;
	struct arp_tab near * far *pArpTab;
	struct arp_tab far *pfArp;

	printf("Datalight Sockets ARP demonstration program\n");
	wLen = sizeof(szNameBuf);
	szNameBuf[0] = 0;
	GetKernelInformation(0, K_INF_HOSTNAME, 0, szNameBuf, &wLen);
	if ((iVersion = GetSocketsVersion()) < 0)
	{
		if (iNetErrNo == ERR_API_NOT_LOADED)
			sprintf(szVersion, "API not loaded");
		else
			sprintf(szVersion, "build < 205");
	}
	else
		sprintf(szVersion, "build %04u", iVersion);
	if (iNetErrNo == ERR_API_NOT_LOADED)
   {
      printf("%s\n", szVersion);
		exit(1);
   }
	printf("Running on '%s', Sockets %s\n", szNameBuf, szVersion);

	while (argc-- > 1) {
		// Drop ARP entries for all IP addresses on command line
		// 0.0.0.0 will drop all ARP entries
		dwIPAddress = ResolveName(argv++[1], szNameBuf, sizeof(szNameBuf));
		printf("Drop %s %s\n",IPAddressString(dwIPAddress),
			ArpApi(ARP_API_DROP,ARP_ETHER, dwIPAddress,0) < 0 ? "failed" : "succeeded");
	}
	// list all ARP entries
	// Get a pointer to the Sockets ARP table
	if (ArpApi(ARP_API_GETTABLE,ARP_ETHER, 0,(void far*)&pArpTab) < 0) {
		printf("Unable to access ARP table, please use more recent build of Sockets\n");
		exit(1);
	}
	wSeg = FP_SEG(pArpTab);
	// Run through the table using segmented memory gymnastics
	for (i = 0;i < ARPSIZE;i++) {
		for(pfArp = MK_FP(wSeg,(WORD)pArpTab[i]);FP_OFF(pfArp);FP_OFF(pfArp) = (WORD)pfArp->next){
			if (iNumber++ > 100) {
				printf("loop!\n");
				break;
			}
			//printf("%Fp: ",pfArp);
			printf("%-16s",inet_ntoa(pfArp->ip_addr));
			printf("%-15s",arptypes[pfArp->hardware]);
			printf("%-5ld",pfArp->timer.count*(long)MSPTICK/1000);
			if(pfArp->state == ARP_PENDING)
				printf("D ");
			else
				printf("  ");
			if(pfArp->state == ARP_VALID) {
				int iIndex;
				for (iIndex = 0;iIndex < 6;++iIndex)
					printf("%02x%c",((BYTE far *)MK_FP(wSeg,(WORD)pfArp->hw_addr))[iIndex],iIndex == 5 ? ' ' : ':');
			} else {
				printf("[unknown]");
			}
			if (pfArp->rif) {
				int i;
				for (i = 0;i < (((BYTE far *)MK_FP(wSeg,(WORD)pfArp->rif))[0] & 0x1f);++i)
					printf(" %02x",((BYTE far *)MK_FP(wSeg,(WORD)pfArp->rif))[i]);
			}
			printf("\n");
		}
	}
	return 0;
}
