// IPTEST.C
                                        
#ifdef __DJGPP__
#include <dpmi.h>
#include <go32.h>
#endif


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>
#include <process.h>
#include <stdarg.h>
#include <time.h>
#include <dos.h>
#include <ctype.h>
#include <signal.h>

#include "compiler.h"
#include "capi.h"


#ifndef __DJGPP__
#ifndef enable
	#define enable _enable
#endif
#ifndef disable
	#define disable _disable
#endif
#endif
char *GetErrorString(unsigned uErrCode);
char *IPAddressString(DWORD dwIPAddress);

DPMI_SECTION_EXTERNS;
volatile BYTE bControlBreak;

void ControlBreak(int iSignal)
{
	bControlBreak = iSignal;
}

int Fail (char *szFormat, ...)
{
	va_list ap;
	char szLine[400];

	if (szFormat)
	{
		va_start(ap, szFormat);
		vsprintf(szLine, szFormat, ap);
		va_end(ap);
		printf("Fail: %s - %s\n", GetErrorString(iNetErrNo),szLine);
	}
	else
		printf("Fail: %s\n", GetErrorString(iNetErrNo));
	return 0;
}

int main(int argc, char *argv[])
{
	WORD wLen;
	char szNameBuf[80], szVersion[20];
	BYTE pbBuf[16];
	int i, iVersion, iDev;
	DWORD dwMyIPAddress,dwHisIPAddress;


	printf("IP test version 1.00 (%s)\n",__DATE__);
	printf("Copyright (C) 2003 Datalight, Inc.\nAll Rights Reserved\n"
			 "Portions copyright (c) GPvNO 2003\n\n");
	while (argc > 1 && argv[1][0] == '/')
	{
		//char *pszArg = argv[1] + ((argv[1][2] == '=') ? 3 : 2);
		switch (tolower(argv[1][1]))
		{
			default:
				exit(0);
		}
		--argc;
		++argv;
	}

	wLen = sizeof(szNameBuf);
	szNameBuf[0] = 0;
	i = GetKernelInformation(0, K_INF_HOSTNAME, 0, szNameBuf, &wLen);
	if ((iVersion = GetSocketsVersion()) < 0)
	{
		if (iNetErrNo == ERR_API_NOT_LOADED)
			sprintf(szVersion, "API not loaded");
		else
			sprintf(szVersion, "build < 205");
	}
	else
		sprintf(szVersion, "build %04u", iVersion);
	printf("Sockets %s\nHostname: %s\n", szVersion, szNameBuf);
	if (iNetErrNo == ERR_API_NOT_LOADED)
		exit(1);
	signal(SIGINT, ControlBreak);
	if (i < 0)
		Fail("GetKernelInformation(K_INF_HOSTNAME)");

	wLen = 16;
	if (GetKernelInformation(0, K_INF_DNS_SERVERS, 0, pbBuf, &wLen) < 0)
		Fail("GetKernelInformation(K_INF_DNS_SERVERS)");
	else
	{
		printf("DNS(%d): ",wLen);
		for (i = 0;i < (int)wLen;i += 4)
			printf("%s ",IPAddressString(*(DWORD *)(pbBuf + i)));
		printf("\n");
	}

	for (iDev = 0;iDev < 10;++iDev) {
		wLen = 6;
		if (GetKernelInformation(0, K_INF_MAC_ADDR, iDev, pbBuf, &wLen) < 0)
		{
			if (iDev)
				break;
			Fail("GetKernelInformation(K_INF_MAC_ADDR)");
		}
		else
		{
			printf("MAC address %d (%d): ",iDev,wLen);
			for (i = 0;i < (int)wLen;i++)
				printf("%02X ",pbBuf[i]);
			printf("\n");
		}
	}
	for (iDev = 0;iDev < 10;++iDev) {
		wLen = 4;
		if (GetKernelInformation(0, K_INF_IP_ADDR, iDev, (char *)&dwMyIPAddress, &wLen) < 0)
		{
			if (iDev)
				break;
			Fail("GetKernelInformation(K_INF_IP_ADDR)");
		}
		else
			printf("Local IP address %d:  %s\n",iDev,IPAddressString(dwMyIPAddress));
	}
	wLen = 4;
	for (iDev = 0;iDev < 10;++iDev) {
		if (GetKernelInformation(0, K_INF_REMOTE_IP, iDev, (char *)&dwHisIPAddress, &wLen) < 0)
		{
			if (iDev)
				break;
			Fail("GetKernelInformation(K_INF_REMOTE_IP)");
		}
		else
			printf("Remote IP address %d: %s\n",iDev,IPAddressString(dwHisIPAddress));
	}
	return 0;
}

WORD Getch(void)
{
	WORD wKey;
	char *pszArgvl[2];	//used in spawning COMSPEC

	wKey = getch();
	if (wKey == 0)
		/* function key */
		wKey = 1000 + getch();
	switch (wKey)
	{
		case 1045:		/* alt X */
			printf("User exit\n");
			exit(2);
			break;
		case 1035:	// Alt H
			printf( "\n"
					"Escape        Abort test\n"
					"Alt-E         Execute command.com\n"
					"Alt-H         Help\n"
					"Alt-X         eXit\n"
					"Any other key Display current test\n");
			break;
		case 1018:	// alt e
			if ((pszArgvl[0] = getenv("COMSPEC")) == 0)
				pszArgvl[0] = "\\COMMAND.COM";
			pszArgvl[1] = 0;
			spawnvp(P_WAIT, pszArgvl[0], pszArgvl);
			break;
	}
	return wKey;
}


/*
Create a human understandable string from a Sockets error code.

Argument:
	uErrCode - The sockets error code.

Returns:
	A pointer to the (static) string representation of the error.
*/
char *GetErrorString(unsigned uErrCode)
{
	static char rgcUnk[30];
	static char *rgszErrs[] =
	{
		"NO_ERR",					// 0	// No error
		"ERR_IN_USE",				// 1	// A connection already exists
		"ERR_DOS",					// 2	// A DOS error occured
		"ERR_NO_MEM",				// 3	// No memory to perform function
		"ERR_NOT_NET_CONN",		// 4	// Connection does not exist
		"ERR_ILLEGAL_OP",			// 5	// Protocol or mode not supported
		"ERR_BAD_PACKET",
		"ERR_NO_HOST",				// 7	// No host address specified
		"ERR_CANT_OPEN",
		"ERR_NET_UNREACHABLE",
		"ERR_HOST_UNREACHABLE",
		"ERR_PROTOCOL_UNREACHABLE",
		"ERR_PORT_UNREACHABLE",
		"ERR_TIMEOUT",				// 13	// The function timed out
		"ERR_HOST_UNKNOWN",		// 14	// Unknown host has been specified
		"ERR_NO_SERVERS",
		"ERR_SERVER_ERR",
		"ERR_BAD_FORMAT",
		"ERR_BAD_ARG",				// 18	// Bad arguments
		"ERR_EOF",					// 19	// The connection has been closed by peer
		"ERR_RESET",				// 20	// The connection has been reset by peer
		"ERR_WOULD_BLOCK",		// 21	// Operation would block
		"ERR_UNBOUND",				// 22	// The descriptor has not been assigned yet
		"ERR_NO_SOCKET",			// 23	// No descriptor is available
		"ERR_BAD_SYS_CALL",		// 24	// Bad parameter in call
		"ERR_CANT_BROADCAST",
		"ERR_NOT_ESTAB",			// 26	// The connection has not been established
		"ERR_RE_ENTRY",			// 27	// The kernel is in use, try again later
		"ERR_NETWORK",          // 28	 /* Network error eg ICMP message */
		"ERR_TERMINATING",      // 29  /* Kernel unloading */
		"ERR_INF_LOCKED",			//	30	 /* Kernel info locked */
		"ERR_BAD_INTERFACE"		//	31	 /* Bad intreface specified */
	};

	if (uErrCode == ERR_API_NOT_LOADED)
		return "Sockets API not loaded";
	if ((uErrCode & 0xff) > ERR_BAD_INTERFACE)
	{
		sprintf(rgcUnk, "Unknown error 0x%04X", uErrCode);
		return rgcUnk;
	}
	return rgszErrs[uErrCode & 0xff];
}

char *IPAddressString(DWORD dwIPAddress)
{
	static char sz[20];

	sprintf(sz, "%u.%u.%u.%u",
			((BYTE *)&dwIPAddress)[0],((BYTE *)&dwIPAddress)[1],
			((BYTE *)&dwIPAddress)[2],((BYTE *)&dwIPAddress)[3]);
	return sz;
}


