//
//   $Id: CACHE.C 1.22 2003/03/14 21:03:20Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*

	CACHE.C
	Created 06-28-2001  by Keith Garvin

	Copyright (c) 2001-2002 Datalight, Inc.
	All rights reserved.
	Confidential and Proprietary

	TODO: Does not yet implement uDataOffset.

	Provides public disk oriented APIs to the XMS cache.

	InitializeDiskCache()
	DeinitializeDiskCache()
	RegisterDiskInformation()
	InvalidateDiskCache()
	FlushDiskCache()
	FlushOneCacheElement()
	CacheReadWrite()

*/

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dlxms.h"
#include "stdc.h"

#include "lrucache.h"
#include "xmscache.h"
#include "cache.h"
#include "cachinit.h"


#include "cache.pt"


#if USE_ELEMENT_INFO

//  These arrays are used to quickly retrieve a status mask  for
//  a particular sector within an element
//
extern unsigned urElementSectorToMask[];

//  This pointer to an array is used to track the valid and dirty
//  sectors within an element
//
extern LPELEMENT_INFO lprElementInfo;

#endif // #if USE_ELEMENT_INFO


/* FIXME: Where do these belong? */
#define READ TRUE
#define WRITE FALSE


/*
	The following two are maintained in the XMSCACHE module,
	usefull here for parsing the xms cache list.
*/
extern unsigned uXmsCacheElementSize;
extern unsigned uXmsNumberOfElements;


#if (CLUSTER_ALIGNMENT || USE_ELEMENT_INFO)
//  Disk information arrays used by DiskReadWrite
//
extern INT13_DRIVE_INFO Int13DriveInfo[26];
#endif


/*
	Local near buffer in real mode address space.  Used for
	double buffering, read ahead, and write behind caching.
	Size is defined by uXmsBufferSize.
*/
static uchar far * lprucLocalBuffer = NULL;
static unsigned uElementsInLocalBuffer = 0;


static unsigned uLocalBufferSize = 0;

/*
	The local buffer is used as a temporary place to
	store data between the XMS and application.
	The fLocalBufferValid flag indicates wether or not
	there is a valid data contents and if the
	ulLocalBufferSector acurrately describes it.
*/
bool fLocalBufferValid = FALSE;
unsigned uLocalBufferXmsMapIndex = 0;
bool fLocalBufferCommitted;


/*
	XMS handle to manage blocks of memory.
	Note that the code does not support more than one handle.
*/
static unsigned uXmsHandle = 0;
static ulong ulXmsPhysicalAddress = 0;
static XMS_MOVE_STRUCT xmsMove;

/*
	The Disk info structure holds information regarding
	what disks are cached, their sector size, ...
*/
#define MAX_VALID_DISKS 26
DISK_INFO diskInfo[MAX_VALID_DISKS];


/* Private proto types */
static bool PASCAL ReadCacheElementFromXms(unsigned uXmsMapIndex,
						void far * vpDataBuffer);
static bool PASCAL WriteCacheElementToXms(void far * vpDataBuffer,
						unsigned uXmsMapIndex);
static unsigned PASCAL FlushCacheElement(unsigned uXmsMapIndex);
static void PASCAL MakeCacheElementClean(unsigned uXmsMapIndex);
#if !USE_ELEMENT_INFO
 static void PASCAL MakeCacheElementDirty(unsigned uXmsMapIndex);
#else
 static void PASCAL MakeCacheElementDirty(unsigned uXmsMapIndex,
		unsigned uSectorOffsetInElement, unsigned uCount );
 static void PASCAL MakeCacheElementValid(unsigned uXmsMapIndex,
		unsigned uSectorOffsetInElement, unsigned uCount );
#endif

#if !USE_ELEMENT_INFO
 static bool DoReadAhead( void );
#else // #if !USE_ELEMENT_INFO
#if CONDITIONAL_READAHEAD
 static bool DoReadAhead( unsigned uSectorShiftValue,
 		unsigned * puMaxSectorsToRead );
#endif // #if CONDITIONAL_READAHEAD
#endif // #if !USE_ELEMENT_INFO


#if TESTING
static unsigned PASCAL DiskReadWrite(bool fRead, unsigned uDisk,
			ulong ulStartSector, unsigned uNumberOfSectors,
			void far * pvData, unsigned * puSecsTransfered);
#endif

/*
	Given the size of our cache, allocate our necessary
	buffers, the cache, and a few global bits of information
*/
bool PASCAL InitializeDiskCache(unsigned uNumberOfElements,
				unsigned uElementSize, unsigned uBufferSize)
{
	unsigned uIndex;
	unsigned uSegment;
	unsigned uSizeInParagraphs;
	unsigned uElementIndex;
	unsigned uSizeInKiloBytes;


	/*
		XMS is required for the cache, also, the install
		check below initializes the XMS subsystem.
	*/
	if(CheckXMMInstalled() != XMS_SUCCESS)
	{
		return FALSE;
	}

	/* Init the XMS and LRU caches */
	if(!InitializeLruCache(uNumberOfElements))
	{
		DeinitializeDiskCache();
		return FALSE;
	}
	if(!InitializeXmsCache(uNumberOfElements, uElementSize))
	{
		DeinitializeDiskCache();
		return FALSE;
	}

	/* Allocate the local maps XMS map and disk map */
	uLocalBufferSize = uBufferSize;
	uSizeInParagraphs = (unsigned)(((ulong)uLocalBufferSize + 0x0F) / 0X10);
	if(!dlAllocMem(uSizeInParagraphs, &uSegment))
	{
		DeinitializeDiskCache();
		return FALSE;
	}
	if(uSegment == 0)
	{
		DeinitializeDiskCache();
		return FALSE;
	}
	lprucLocalBuffer = (uchar far *) MK_FP(uSegment, 0);
	_fmemset(lprucLocalBuffer, 0xcc, uLocalBufferSize);
	uElementsInLocalBuffer = uLocalBufferSize / uElementSize;

	/* Invalidate the local buffer */
	fLocalBufferValid = FALSE;
	uLocalBufferXmsMapIndex = 0;

	/* Init the LRU with the indices of our cache */
	for(uElementIndex = 0;
			uElementIndex < uNumberOfElements; ++uElementIndex)
	{
		UpdateCacheEntryToMru(uElementIndex);
	}

	/* Initialize the disk information structure */
	for(uIndex = 0; uIndex < MAX_VALID_DISKS; ++uIndex)
	{
		diskInfo[uIndex].fValid = FALSE;
	}

	/* Allocate and lock the needed XMS memory */
	uSizeInKiloBytes = uNumberOfElements * (uElementSize / 1024);
	ASSERT(uSizeInKiloBytes);

	if(AllocateXMSBlock(uSizeInKiloBytes, &uXmsHandle) != XMS_SUCCESS)
	{
		DeinitializeDiskCache();
		return FALSE;
	}

	/* Sucessful initialization */
	return TRUE;
}


/*
	Given the size of our cache, allocate our necessary
	buffers, the cache, and a few global bits of information
*/
void PASCAL DeinitializeDiskCache(void)
{
	unsigned uIndex;

	/* Deinit the XMS and LRU caches */
	DeinitializeLruCache();
	DeinitializeXmsCache();

	/* Invalidate the local buffer */
	fLocalBufferValid = FALSE;
	uLocalBufferXmsMapIndex = 0;

	/* Free the local buffers */
	if(lprucLocalBuffer)
	{
		_fmemset(lprucLocalBuffer, 0, uLocalBufferSize);
		dlFreeMem(FP_SEG(lprucLocalBuffer));
		lprucLocalBuffer = NULL;
	}

	/* Initialize the disk information structure */
	for(uIndex = 0; uIndex < MAX_VALID_DISKS; ++uIndex)
	{
		diskInfo[uIndex].fValid = FALSE;
	}

	/* Invalidate the XMS cache */
	uLocalBufferSize = 0;
	uXmsHandle = 0;
	ulXmsPhysicalAddress = 0;

	/* Free the XMS cache */
	FreeXMSBlock(uXmsHandle);
}


/*
	Record disk information needed by the cache.  The cache
	will only cache the disks registered with valid information
	through this API.
*/
bool PASCAL RegisterDiskInformation(unsigned uDisk, unsigned uSectorSize,
				unsigned uDataOffset)
{
	unsigned uTemp;


	/* A little sanity checking */
	if(uDisk >= 26)
	{
		return FALSE;
	}

	if(uSectorSize < 512)
	{
		return FALSE;
	}

	if(uDataOffset % 512)
	{
		return FALSE;
	}

	/* Record the disk information */
	for( uTemp = 0; uSectorSize > 1; uTemp++ )
	{
		uSectorSize >>= 1;
	}
	diskInfo[uDisk].uSectorShiftValue = uTemp;
	diskInfo[uDisk].uDataOffset = uDataOffset;
	diskInfo[uDisk].fValid = TRUE;

	return TRUE;
}

/*
	Invalidate the all cache elements for the given disk.
	Note this does *not* flush, just nuke the data contents!
*/
void PASCAL InvalidateDiskCache(unsigned uDisk)
{
//unused	unsigned uDiskMapIndex;
	unsigned uXmsMapIndex;
	DISK_ENTRY deCacheEntry;
	unsigned uOldDiskMapIndex;
	extern unsigned uNumberOfDiskEntries;
	extern DISK_ENTRY far * lprdeXmsCacheMap;
	extern unsigned far * lpruDiskSectorMap;


	/* Verify its a valid disk */
	ASSERT((diskInfo[uDisk].fValid) && (uDisk <= MAX_VALID_DISKS));


	/*
		Parse through the XMS map looking for elements
		assigned to the requested disk drive
	*/
	for(uXmsMapIndex = 0; uXmsMapIndex < uXmsNumberOfElements;
			uXmsMapIndex++ )
	{
		/*
			If it is the given disk , delete it.  Ignore
			error since there is nothing we can do about it.
		*/
		GetXmsDiskEntry(uXmsMapIndex, &deCacheEntry);

		if( DE_VALID(deCacheEntry) &&	(GET_DE_DISK(deCacheEntry) == uDisk) )
		{
			/* Remove the element from the cache */
			RemoveDiskEntryFromXmsCache(uXmsMapIndex, &uOldDiskMapIndex);
		}
	}
}


/*
	Flush data contents belonging to the given disk and which
	is marked as dirty, to disk.  Returns success after all
	dirty elements belonging to the given disk have been written
	back to the disk.  Returns FALSE only on a write error.
*/
bool PASCAL FlushDiskCache(unsigned uDisk)
{
	unsigned uDiskMapIndex;
	unsigned uXmsMapIndex;
	DISK_ENTRY deCacheEntry;
//unused	unsigned uOldDiskMapIndex;
	extern unsigned uNumberOfDiskEntries;
	extern DISK_ENTRY far * lprdeXmsCacheMap;
	extern unsigned far * lpruDiskSectorMap;
	bool fSuccessful = TRUE;


	/* Verify its a valid disk */
	ASSERT((diskInfo[uDisk].fValid) && (uDisk <= MAX_VALID_DISKS));


	/*
		Parse through the XMS map looking for dirty
		elements to flush from the cache.
	*/
	for(uDiskMapIndex = 0;
			uDiskMapIndex < uNumberOfDiskEntries; uDiskMapIndex++ )
	{
		/* If it is the given disk and it is dirty */
		uXmsMapIndex = lpruDiskSectorMap[uDiskMapIndex];
		if(!GetXmsDiskEntry(uXmsMapIndex, &deCacheEntry))
		{
			return fSuccessful;
		}

		if( DE_VALID_AND_DIRTY(deCacheEntry) )
		{
			if( GET_DE_DISK(deCacheEntry) == uDisk )
			{
				/* Write out the cache element */
				if( IO_SUCCESS != FlushCacheElement(uXmsMapIndex) )
				{
					/* On failure, remove this element from the cache */
					_RemoveDiskEntryFromXmsCache(uDiskMapIndex);
					fSuccessful = FALSE;
				}
			}
		}
	}

	return fSuccessful;
}


/*
	Flush one dirty cache element back to the disk.  Cache will
	be searched from the least recently used element to the most
	recent.  If there is no dirty elements to flush return with
	success.  Returns FALSE only on a write error.
*/
bool PASCAL FlushOneCacheElement(void)
{
	unsigned uLruIndex;
	unsigned uXmsMapIndex;
	DISK_ENTRY deCacheEntry;
	unsigned uOldDiskMapIndex;
	bool fSuccessful = TRUE;


	/* Parse the LRU cache */
	for(uLruIndex = 0; uLruIndex < uXmsNumberOfElements; ++uLruIndex)
	{
		/* If the element is dirty */
		if(!GetLruEntry(uLruIndex, &uXmsMapIndex))
		{
			return fSuccessful;
		}
		if(!GetXmsDiskEntry(uXmsMapIndex, &deCacheEntry))
		{
			return fSuccessful;
		}

		/* Check valid and dirty status */
		if( DE_VALID_AND_DIRTY(deCacheEntry) )
		{
			/*  Write out the cache element */
			//
			//  This will also mark the element as clean
			//
			if(IO_SUCCESS != FlushCacheElement(uXmsMapIndex))
			{
				/* On failure, remove this element from the cache */
				RemoveDiskEntryFromXmsCache(uXmsMapIndex, &uOldDiskMapIndex );
				fSuccessful = FALSE;
			}

			/* Tell the caller, we flushed the data! */
			return fSuccessful;
		}
	}

	/* No data to flush */
	return fSuccessful;
}


bool PASCAL FlushLastLRUElements( void )
{
	unsigned uLruIndex;
	unsigned uXmsMapIndex;
	unsigned uNumberOfElements;
	unsigned uOldDiskMapIndex;
	bool fSuccessful = TRUE;


	if( uElementsInLocalBuffer < uXmsNumberOfElements )
	{
		uNumberOfElements = uElementsInLocalBuffer;
	}
	else
	{
		uNumberOfElements = uXmsNumberOfElements;
	}


	/* Parse the LRU cache */
	for(uLruIndex = 0; uLruIndex < uNumberOfElements; uLruIndex++)
	{
		/* If the element is dirty */
		if(!GetLruEntry(uLruIndex, &uXmsMapIndex))
		{
			return fSuccessful;
		}

		/*  Write out the cache element */
		//  This will only commit dirty elements and will mark them
		//  clean as well
		//
		if(IO_SUCCESS != FlushCacheElement(uXmsMapIndex))
		{
			/* On failure, remove this element from the cache */
			RemoveDiskEntryFromXmsCache(uXmsMapIndex, &uOldDiskMapIndex );
			fSuccessful = FALSE;
		}
	}

	/* No data to flush */
	return fSuccessful;
}


#if CONDITIONAL_READAHEAD

//  Determine if we should perform the read ahead for this call
//
#if !USE_ELEMENT_INFO
 static bool DoReadAhead( void )
#else // #if !USE_ELEMENT_INFO
 static bool DoReadAhead( unsigned uSectorShiftValue,
 		unsigned * puMaxSectorsToRead )
#endif // #if !USE_ELEMENT_INFO
{
	extern LPLOL lpLOL;
	extern unsigned uSFTIndex;
	extern bool fIsFileRead;
	unsigned uIndex;
	LPDCB lpDCB;
	LPSFT lpSFT;
#if USE_ELEMENT_INFO
	unsigned uSectorSize;
	ulong ulSectorsLeftInFile;
#endif


	//
	//
	if( !fIsFileRead )
	{
#if USE_ELEMENT_INFO
		//  Indicate that as many sectors can be requested as desired
		//
		*puMaxSectorsToRead = 0xffff;
#endif // #if USE_ELEMENT_INFO

		return FALSE;
	}


	//  Get the SFT for this file
	//
	uIndex = uSFTIndex;
	lpDCB = lpLOL->first_dcb;
	while( TRUE )
	{
		if( uIndex < lpDCB->numsft )
		{
			lpSFT = &(lpDCB->sft[uIndex]);
			break;
		}
		else
		{
			uIndex -= lpDCB->numsft;
			lpDCB = lpDCB->next;
		}
	}


#if USE_ELEMENT_INFO

	//  Determine the number of sectors that remain in the file
	//
	uSectorSize = 1 << uSectorShiftValue;
	ulSectorsLeftInFile = lpSFT->size - lpSFT->offset + (uSectorSize - 1);
	ulSectorsLeftInFile >>= uSectorShiftValue;


	//  Indicate that we should not read past the end of the file
	//
	*puMaxSectorsToRead = (unsigned)ulSectorsLeftInFile;

#endif // #if USE_ELEMENT_INFO


	//
	//
	return ((lpSFT->size - lpSFT->offset) > uXmsCacheElementSize );
}

#endif // #if CONDITIONAL_READAHEAD


/*
	The meat of the cache, all disk access travels through
	this function allowing us to track reads and writes
	into our cache.
*/
unsigned PASCAL CacheReadWrite(bool fRead, unsigned uDisk, ulong ulStartSector,
			unsigned uNumberOfSectors, void far * pvData, bool fCacheWrite,
			unsigned * puSectorsReadWritten)
{
	unsigned uSkipSectorsInFirstElement;
	unsigned uCopySectorsInFirstElement;
	unsigned uMoveLength;
	unsigned uXmsMapIndex;
	unsigned uSectorShiftValue;
	unsigned uSectorsPerElement;
	unsigned uAccessReturn;
	unsigned uReturnCode;
	ulong ulTempStartSector;
#if CLUSTER_ALIGNMENT
	ulong ulNonAlignedStartSector;
#endif
	DISK_ENTRY deTemp;
	DISK_ENTRY deTarget;
	unsigned uTemp;
	bool fReadAheadBufferValid;
	bool fFoundInCache;
	unsigned uDiskMapIndex;
	unsigned uOldDiskMapIndex;
	extern unsigned far * lpruDiskSectorMap;
#if USE_ELEMENT_INFO
	bool fDoReadAhead;
	unsigned uMaxSectorsToRead;
	unsigned uSectorsToRead;
	unsigned uSectorsReadInElement;
	unsigned uSectorOffsetInElement;
	unsigned uTempIndex;
	bool fIsFirstTime;
	unsigned uSectorsToReadInFirstElement;
	extern bool fIsFileRead;
	static uLastAccessedDisk = 0xff;
#endif



#if USE_ELEMENT_INFO
	if( uLastAccessedDisk != uDisk )
	{
		if( uLastAccessedDisk != 0xff )
		{
			FlushDiskCache( uDisk );
		}
	}
#endif


#if CLUSTER_ALIGNMENT
	//  Cluster alignement
	//
	ulStartSector += Int13DriveInfo[uDisk].ulSkipSectors;
#endif


	/* Start with zero sectors transfered */
	*puSectorsReadWritten = 0;
	uReturnCode = IO_SUCCESS;


	/* Verify its a valid disk */
	ASSERT((diskInfo[uDisk].fValid) && (uDisk > MAX_VALID_DISKS));


	/*
		All of our requests are completed on an element boundary
		which is a multiple of the sector size.  There may be an
		odd number of sectors at the start or end of the request
	*/
	uSectorShiftValue = diskInfo[uDisk].uSectorShiftValue;
	uSectorsPerElement = uXmsCacheElementSize >> uSectorShiftValue;


	uSkipSectorsInFirstElement = (unsigned)(ulStartSector %
			uSectorsPerElement);
	uTemp = uSectorsPerElement - uSkipSectorsInFirstElement;
	if(uNumberOfSectors > uTemp )
	{
		//  this line was wrong.
		//
		//	uCopySectorsInFirstElement = uSkipSectorsInFirstElement %
		//			uSectorsPerElement;
		//
		//  corrected below:
		//
		uCopySectorsInFirstElement = uTemp;
	}
	else
	{
		uCopySectorsInFirstElement = uNumberOfSectors;
	}


	/* Continue for as long as we have sectors to get to/from the cache */
	while(uNumberOfSectors)
	{
		deTarget  = MAKE_DE_SECTOR(ulStartSector, uSectorsPerElement);
		deTarget |= MAKE_DE_DISK(uDisk);
		fFoundInCache = FindFirstXmsCacheElementMatch( deTarget, &uDiskMapIndex);
		if(!fFoundInCache)
		{
			//  Flush the last couple of LRU buffers so that we can
			//  read ahead
			//
			if( !FlushLastLRUElements() )
			{
				// This failed ???
				//
            asm mov bx,0xFFFF;
				asm int 3
				asm hlt
			}

			/* Make space in our cache */
			GetLeastRecentLruCacheEntry(&uXmsMapIndex);
			GetXmsDiskEntry(uXmsMapIndex, &deTemp);
			if( DE_VALID(deTemp) )
			{
				if(!RemoveDiskEntryFromXmsCache(uXmsMapIndex, &uOldDiskMapIndex))
				{
					return IO_FAILURE;
				}

				//  Adjust the disk map index for the one that was just removed
				//  from the list
				//
				if( uOldDiskMapIndex < uDiskMapIndex )
				{
					uDiskMapIndex--;
				}
			}


			//  Get the starting sector for the element
			//
			ulTempStartSector = ulStartSector - (ulStartSector %
					uSectorsPerElement);


			//  Make sure the local buffer is committed before we
			//  destroy it
			//
			CommitLocalBuffer();


#if !USE_ELEMENT_INFO
			//  determine if we really need to hit the disk
			//
			fReadAheadBufferValid = FALSE;
			if( (fRead) || ((uCopySectorsInFirstElement) ||
					(uNumberOfSectors < uSectorsPerElement)) )
			{
				unsigned uSectorsToRead;


#if CHECK_COMMIT_OF_LOCALBUFFER
				if( (fLocalBufferValid) && !fLocalBufferCommitted )
				{
               asm mov bx,0xFFFE;
					asm int 3
				}
#endif // #if CHECK_COMMIT_OF_LOCALBUFFER


				//  We are about to change what is in the local buffer
				//  so mark it invalid
				//
				fLocalBufferValid = FALSE;
				if( (!fRead) ||
					(uNumberOfSectors == uCopySectorsInFirstElement) )
				{
					uSectorsToRead = uSectorsPerElement;
				}
				else
				{
					if( DoReadAhead() )
					{
						uSectorsToRead = uSectorsPerElement * uElementsInLocalBuffer;
						fReadAheadBufferValid = TRUE;
					}
					else
					{
						uSectorsToRead = uSectorsPerElement;
					}
				}


#if !CLUSTER_ALIGNMENT
				uReturnCode = DiskReadWrite(READ, uDisk, ulTempStartSector,
						uSectorsToRead, lprucLocalBuffer, &uAccessReturn);
#else // #if !CLUSTER_ALIGNMENT
				if( ulTempStartSector )
				{
					//  Don't have to worry about an invalid sector number here
					//
					ulNonAlignedStartSector = ulTempStartSector -
							Int13DriveInfo[uDisk].ulSkipSectors;

					uReturnCode = DiskReadWrite(READ, uDisk,
							ulNonAlignedStartSector, uSectorsToRead,
							lprucLocalBuffer,	&uAccessReturn);
				}
				else
				{
					//  Do have to worry about an invalid sector number here
					//
					uReturnCode = DiskReadWrite( READ, uDisk,
							ulTempStartSector, uSectorsToRead -
							Int13DriveInfo[uDisk].ulSkipSectors,
							lprucLocalBuffer +
							(Int13DriveInfo[uDisk].ulSkipSectors * 512),
							&uAccessReturn );
					uAccessReturn += Int13DriveInfo[uDisk].ulSkipSectors;
				}
#endif // #if !CLUSTER_ALIGNMENT
				if(uReturnCode != IO_SUCCESS)
				{
					fReadAheadBufferValid = FALSE;
					return uReturnCode;
				}
			}


			//  If this is a write, we are about to change what is in
			//  the local buffer anyway and WriteCacheElementToXms() will
			//  be called again, so we don't need to call it now
			//
			if( fRead )
			{
				if(!WriteCacheElementToXms(lprucLocalBuffer, uXmsMapIndex))
				{
					return IO_FAILURE;
				}
			}
			else
			{
#if CHECK_COMMIT_OF_LOCALBUFFER
				if( (fLocalBufferValid) || (fLocalBufferCommitted) )
				{
               asm mov bx,0xFFFD;
					asm int 3
				}
#endif // #if CHECK_COMMIT_OF_LOCALBUFFER

				fLocalBufferValid = TRUE;
				uLocalBufferXmsMapIndex = uXmsMapIndex;
				fLocalBufferCommitted = FALSE;
			}


#else // #if !USE_ELEMENT_INFO


			//  Indicate that the read ahead buffer is no longer valid
			//
			fReadAheadBufferValid = FALSE;


			//  determine if we really need to hit the disk
			//
			if( fRead )
			{
#if CHECK_COMMIT_OF_LOCALBUFFER
				if( (fLocalBufferValid) && !fLocalBufferCommitted )
				{
               asm mov bx,0xFFFC;
					asm int 3
				}
#endif // #if CHECK_COMMIT_OF_LOCALBUFFER


				//  We are about to change what is in the local buffer
				//  so mark it invalid.  It was already committed.
				//
				fLocalBufferValid = FALSE;


#if CONDITIONAL_READAHEAD
				//  Determine how many sectors to read
				//
				fDoReadAhead = DoReadAhead( uSectorShiftValue,
						&uMaxSectorsToRead );
#else // #if CONDITIONAL_READAHEAD

				fDoReadAhead = TRUE;
				uMaxSectorsToRead = 0xffff;

#endif // #if CONDITIONAL_READAHEAD


				if( fDoReadAhead )
				{
					uSectorsToRead = uSectorsPerElement * uElementsInLocalBuffer;
				}
				else
				{
					uSectorsToRead = uSectorsPerElement;
				}
				uSectorsToRead -= uSkipSectorsInFirstElement;


				//  Determine if the number of sectors to read exceeds
				//  the maximum number of sectors to read ahead
				//
				if( uMaxSectorsToRead < uSectorsToRead )
				{
					uSectorsToRead = uMaxSectorsToRead;
				}
				if( (uSectorsToRead + uSkipSectorsInFirstElement) >
						uSectorsPerElement )
				{
					fReadAheadBufferValid = TRUE;
				}


#if CHECK_COMMIT_OF_LOCALBUFFER
				if( (fLocalBufferValid) && !fLocalBufferCommitted )
				{
               asm mov bx,0xFFFB;
					asm int 3
				}
#endif // #if CHECK_COMMIT_OF_LOCALBUFFER


#if !CLUSTER_ALIGNMENT
				//  Read in the sectors into the local buffer
				//
				uReturnCode = DiskReadWrite(READ, uDisk,
						ulTempStartSector + uSkipSectorsInFirstElement,
						uSectorsToRead, lprucLocalBuffer +
						(uSkipSectorsInFirstElement << uSectorShiftValue ),
						&uAccessReturn);
#else // #if !CLUSTER_ALIGNMENT
				lalalala
#endif // #if !CLUSTER_ALIGNMENT
				if(uReturnCode != IO_SUCCESS)
				{
					fReadAheadBufferValid = FALSE;
					return uReturnCode;
				}


				//  Update the local buffer structures with the XMS map index
				//
				if(!WriteCacheElementToXms(lprucLocalBuffer, uXmsMapIndex))
				{
					return IO_FAILURE;
				}
			}


#endif // #if !USE_ELEMENT_INFO


			/* Add the new cache element to our cache */
			if( !AddDiskEntryToXmsCache(uXmsMapIndex, uDiskMapIndex,
					deTarget) )
			{
				return IO_FAILURE;
			}


#if USE_ELEMENT_INFO

			if( fRead )
			{
				//  Determine the number of sectors that were read in for
				//  the first element
				//
				if( fReadAheadBufferValid )
				{
					uSectorsReadInElement = uSectorsPerElement -
							uSkipSectorsInFirstElement;
				}
				else
				{
					uSectorsReadInElement = uSectorsToRead;
				}


				//  Update the sectors within the element that are valid
				//
				MakeCacheElementValid( uXmsMapIndex, uSkipSectorsInFirstElement,
						uSectorsReadInElement );
			}

#endif // #if USE_ELEMENT_INFO

			/* Update use of the LRU cache and our progress */
			UpdateCacheEntryToMru(uXmsMapIndex);


#if !USE_ELEMENT_INFO

			if( fReadAheadBufferValid )
			{
				ulong ulNextElementStartSector;
				uchar far * lpBuffer;
				unsigned uElementCount;
				unsigned uTempXmsMapIndex;


				lpBuffer = lprucLocalBuffer;
				ulNextElementStartSector = ulTempStartSector;
				for( uElementCount = 1; uElementCount < uElementsInLocalBuffer;
						uElementCount++ )
				{
					ulNextElementStartSector += uSectorsPerElement;
					lpBuffer += uXmsCacheElementSize;
					deTarget  = MAKE_DE_SECTOR(ulNextElementStartSector,
							uSectorsPerElement);
					deTarget |= MAKE_DE_DISK(uDisk);
					if( !FindFirstXmsCacheElementMatch( deTarget, &uDiskMapIndex) )
					{
						uTempXmsMapIndex = lpruDiskSectorMap[uDiskMapIndex];

						/* Make space in our cache */
						GetLeastRecentLruCacheEntry(&uTempXmsMapIndex);
						GetXmsDiskEntry(uTempXmsMapIndex, &deTemp);
						if( DE_VALID(deTemp) )
						{
							if(!RemoveDiskEntryFromXmsCache(uTempXmsMapIndex,
									&uOldDiskMapIndex))
							{
								return IO_FAILURE;
							}

							//  Adjust the disk map index for the one that
							//  was just removed from the list
							//
							if( uOldDiskMapIndex < uDiskMapIndex )
							{
								uDiskMapIndex--;
							}
						}


						if(!WriteCacheElementToXms(lpBuffer, uTempXmsMapIndex))
						{
							return IO_FAILURE;
						}


						/* Add the new cache element to our cache */
						if( !AddDiskEntryToXmsCache(uTempXmsMapIndex,
								uDiskMapIndex,	deTarget) )
						{
							return IO_FAILURE;
						}


						/* Update use of the LRU cache and our progress */
						UpdateCacheEntryToMru(uTempXmsMapIndex);
					}
				}
			}

#else // #if !USE_ELEMENT_INFO

			if( (fRead) && (fReadAheadBufferValid) )
			{
				ulong ulNextElementStartSector;
				uchar far * lpBuffer;
//unused				unsigned uElementCount;
				unsigned uTempXmsMapIndex;


				lpBuffer = lprucLocalBuffer;
				ulNextElementStartSector = ulTempStartSector;
				uSectorsToRead -=	uSectorsReadInElement;
				while( uSectorsToRead )
				{
					ulNextElementStartSector += uSectorsPerElement;
					lpBuffer += uXmsCacheElementSize;
					deTarget  = MAKE_DE_SECTOR(ulNextElementStartSector,
							uSectorsPerElement);
					deTarget |= MAKE_DE_DISK(uDisk);
					if( !FindFirstXmsCacheElementMatch( deTarget, &uDiskMapIndex) )
					{
						/* Make space in our cache */
						GetLeastRecentLruCacheEntry(&uTempXmsMapIndex);
						GetXmsDiskEntry(uTempXmsMapIndex, &deTemp);
						if( DE_VALID(deTemp) )
						{
							if(!RemoveDiskEntryFromXmsCache(uTempXmsMapIndex,
									&uOldDiskMapIndex))
							{
								return IO_FAILURE;
							}

							//  Adjust the disk map index for the one that
							//  was just removed from the list
							//
							if( uOldDiskMapIndex < uDiskMapIndex )
							{
								uDiskMapIndex--;
							}
						}


						//  Write the element data to XMS
						//
						if( uSectorsToRead < uSectorsPerElement )
						{
							if( !WritePartCacheElementToXms(lpBuffer,
									uTempXmsMapIndex, 0,
									uSectorsToRead << uSectorShiftValue ) )
							{
								return IO_FAILURE;
							}
						}
						else
						{
							if(!WriteCacheElementToXms(lpBuffer, uTempXmsMapIndex))
							{
								return IO_FAILURE;
							}
						}


						/* Add the new cache element to our cache */
						if( !AddDiskEntryToXmsCache(uTempXmsMapIndex,
								uDiskMapIndex,	deTarget) )
						{
							return IO_FAILURE;
						}


						//  Update the sectors within the element that are valid
						//
						if( uSectorsToRead < uSectorsPerElement )
						{
							MakeCacheElementValid( uTempXmsMapIndex, 0,
									uSectorsToRead );
						}
						else
						{
							MakeCacheElementValid( uTempXmsMapIndex, 0,
									uSectorsPerElement );
						}


						/* Update use of the LRU cache and our progress */
						UpdateCacheEntryToMru(uTempXmsMapIndex);
					}
					else
					{
						//
						//  Don't handle this case yet.  We will need to
						//  make sure that each of the sector we add does
						//  not overwrite a valid and dirty sector when
						//  we do add this case.
						//
					}


					//  Update the number of sectors left to be added to
					//  the cache which still reside in the local buffer
					//
					if( uSectorsToRead < uSectorsPerElement )
					{
						uSectorsToRead = 0;
					}
					else
					{
						uSectorsToRead -= uSectorsPerElement;
					}
				}
			}
#endif // #if !USE_ELEMENT_INFO
		}
		else
		{
			//
			//  An element was found that contains the sector
			//  range that is being requested
			//
			uXmsMapIndex = lpruDiskSectorMap[uDiskMapIndex];


			/* Update use of the LRU cache and our progress */
			UpdateCacheEntryToMru(uXmsMapIndex);


#if USE_ELEMENT_INFO
			//  An element was found which may contain all the
			//  sectors we are looking for but we must make sure
			//  that all of the sectors being requested are in
			//  the element if this is a read request
			//
			if( fRead )
			{
				fIsFirstTime = TRUE;


				//  Get the starting sector for the element
				//
				ulTempStartSector = ulStartSector - (ulStartSector %
						uSectorsPerElement);


				if( uCopySectorsInFirstElement )
				{
					uSectorsToReadInFirstElement = uCopySectorsInFirstElement;
				}
				else if(uNumberOfSectors >= uSectorsPerElement)
				{
					uSectorsToReadInFirstElement = uSectorsPerElement;
				}
				else
				{
					uSectorsToReadInFirstElement = uNumberOfSectors;
				}


				for( uSectorOffsetInElement = uSkipSectorsInFirstElement;
						uSectorOffsetInElement <
						(uSkipSectorsInFirstElement + uSectorsToReadInFirstElement);
						uSectorOffsetInElement++ )
				{
					//  Make sure that this requested sector is valid
					//
					if( !EI_VALID(uXmsMapIndex,uSectorOffsetInElement) )
					{
						if( fIsFirstTime )
						{
							//  Make sure that the local buffer is committed
							//  before it is overwritten
							//
							CommitLocalBuffer();


							//  Copy this element down into the local buffer
							//
							if(!ReadCacheElementFromXms(uXmsMapIndex,
									lprucLocalBuffer))
							{
								return IO_FAILURE;
							}


							//  Set the flag indicating that we don't want
							//  to invalidate or overwrite the local buffer
							//  with the XMS data twice
							//
							fIsFirstTime = FALSE;
						}


						//  Initialize the number of sectors to read
						//
						uSectorsToRead = 1;


						//  Determine how many sectors that follow this one
						//  need to be read too
						//
						for( uTempIndex = uSectorOffsetInElement + 1;
								uTempIndex <
								(uSkipSectorsInFirstElement +
								uSectorsToReadInFirstElement); uTempIndex++ )
						{
							//  Determine if this sector is already in
							//  the cache
							//
							if( EI_VALID(uXmsMapIndex,uTempIndex) )
							{
								break;
							}


							//  Increment the number of sectors to read
							//
							uSectorsToRead++;
						}


#if CHECK_COMMIT_OF_LOCALBUFFER
						if( (fLocalBufferValid) && !fLocalBufferCommitted )
						{
                     asm mov bx,0xFFFA;
							asm int 3
						}
#endif


						//  Read this many sectors into the local buffer
						//
#if !CLUSTER_ALIGNMENT
						fReadAheadBufferValid = FALSE;
						uReturnCode = DiskReadWrite(READ, uDisk,
								ulTempStartSector + uSectorOffsetInElement,
								uSectorsToRead, lprucLocalBuffer +
								(uSectorOffsetInElement << uSectorShiftValue ),
								&uAccessReturn);
#else // #if !CLUSTER_ALIGNMENT
						lalalala
#endif // #if !CLUSTER_ALIGNMENT
						if(uReturnCode != IO_SUCCESS)
						{
							return uReturnCode;
						}
						fLocalBufferCommitted = FALSE;


						//  Mark these sectors as valid
						//
						MakeCacheElementValid( uXmsMapIndex, uSectorOffsetInElement,
								uSectorsToRead );


						//  Adjust the place where we should continue checking
						//  for valid sectors
						//
						uSectorOffsetInElement += uSectorsToRead;
					}
				}
			}
#endif
		}


		/* The first access may not be aligned, special case */
		if(uCopySectorsInFirstElement)
		{
			uMoveLength = uCopySectorsInFirstElement << uSectorShiftValue;
			ASSERT(uMoveLength);

			/* Move only the needed amount of data */
			if(fRead)
			{
				/* Odd number of reads, get into our buffer first */
				if(!ReadPartCacheElementFromXms(uXmsMapIndex, pvData,
						(uSkipSectorsInFirstElement << uSectorShiftValue),
						uMoveLength ))
				{
					return IO_FAILURE;
				}
			}
			else /* Its a write */
			{
#if !USE_ELEMENT_INFO
				if( !fFoundInCache )
				{
					/* Odd number of reads, get into our buffer first */
					if(!ReadCacheElementFromXms(uXmsMapIndex, lprucLocalBuffer))
					{
						return IO_FAILURE;
					}
				}

				if( fFoundInCache )
				{
					if(!WritePartCacheElementToXms(pvData, uXmsMapIndex,
							(uSkipSectorsInFirstElement << uSectorShiftValue),
							uMoveLength ))
					{
						return IO_FAILURE;
					}
				}
				else
				{
#if CHECK_COMMIT_OF_LOCALBUFFER
					if( (fLocalBufferValid) && !fLocalBufferCommitted )
					{
                  asm mov bx,0xFFF9;
						asm int 3
					}
#endif


					/* Copy data into our cache */
					_fmemcpy(lprucLocalBuffer +
							(uSkipSectorsInFirstElement << uSectorShiftValue),
							(uchar far *)pvData, uMoveLength);

					if(!WriteCacheElementToXms(lprucLocalBuffer, uXmsMapIndex))
					{
						return IO_FAILURE;
					}

				}

#else // #if !USE_ELEMENT_INFO

				if(!WritePartCacheElementToXms(pvData, uXmsMapIndex,
						(uSkipSectorsInFirstElement << uSectorShiftValue),
						uMoveLength ))
				{
					return IO_FAILURE;
				}

#endif // #if !USE_ELEMENT_INFO

#if !USE_ELEMENT_INFO
				MakeCacheElementDirty(uXmsMapIndex);
#else // #if !USE_ELEMENT_INFO
				MakeCacheElementValid(uXmsMapIndex, uSkipSectorsInFirstElement,
						uCopySectorsInFirstElement );
				MakeCacheElementDirty(uXmsMapIndex, uSkipSectorsInFirstElement,
						uCopySectorsInFirstElement );
#endif // #if !USE_ELEMENT_INFO
				if(!fCacheWrite)
				{
					/* Commit the write now! */
					uReturnCode = FlushCacheElement(uXmsMapIndex);
					if( uReturnCode != IO_SUCCESS )
					{
   					/* On failure, remove this element from the cache */
   					RemoveDiskEntryFromXmsCache(uXmsMapIndex, &uOldDiskMapIndex );
						return uReturnCode;
					}
				}
			}

			/* Keep the user's sector count updated */
			*puSectorsReadWritten += uCopySectorsInFirstElement;

			(uchar far *)pvData += uMoveLength;
			uNumberOfSectors -= uCopySectorsInFirstElement;
			ulStartSector += uCopySectorsInFirstElement;
			uCopySectorsInFirstElement = 0;
			uSkipSectorsInFirstElement = 0;
		}
		else if(uNumberOfSectors >= uSectorsPerElement)
		{
			/* Full cache element length - read directly into user data */
			if(fRead)
			{
				/* FIXME: Picks up the one in XMS rather than local */
				ReadCacheElementFromXms(uXmsMapIndex, pvData);
			}
			else /* Its a write */
			{
				if(!WriteCacheElementToXms(pvData, uXmsMapIndex))
				{
					return IO_FAILURE;
				}

#if !USE_ELEMENT_INFO
				MakeCacheElementDirty(uXmsMapIndex);
#else // #if !USE_ELEMENT_INFO
				MakeCacheElementValid(uXmsMapIndex, 0, uSectorsPerElement);
				MakeCacheElementDirty(uXmsMapIndex, 0, uSectorsPerElement);
#endif // #if !USE_ELEMENT_INFO
				if(!fCacheWrite)
				{
					/* Commit the write now! */
					uReturnCode = FlushCacheElement(uXmsMapIndex);
					if(uReturnCode != IO_SUCCESS)
					{
   					/* On failure, remove this element from the cache */
   					RemoveDiskEntryFromXmsCache(uXmsMapIndex, &uOldDiskMapIndex );
						return uReturnCode;
					}
				}
			}

			/* Keep the user's sector count updated */
			*puSectorsReadWritten += uSectorsPerElement;

			(uchar far *) pvData += uXmsCacheElementSize;
			uNumberOfSectors -= uSectorsPerElement;
			ulStartSector += uSectorsPerElement;
		}
		else /* Last odd length */
		{
			uMoveLength = uNumberOfSectors << uSectorShiftValue;

#if !USE_ELEMENT_INFO
			/* Read the remaining sectors from the cache */
			ReadCacheElementFromXms(uXmsMapIndex, lprucLocalBuffer);
#endif // #if !USE_ELEMENT_INFO

			if(fRead)
			{
				/* Read the remaining sectors from the cache */
				ReadPartCacheElementFromXms(uXmsMapIndex, pvData, 0,
						uMoveLength );
			}
			else /* Its a write */
			{
#if !USE_ELEMENT_INFO
				if( fFoundInCache )
				{
					if(!WritePartCacheElementToXms( pvData, uXmsMapIndex, 0,
							uMoveLength ))
					{
						return IO_FAILURE;
					}
				}
				else
				{
#if CHECK_COMMIT_OF_LOCALBUFFER
					if( (fLocalBufferValid) && !fLocalBufferCommitted )
					{
                  asm mov bx,0xFFF8;
						asm int 3
					}
#endif


					/* Move only the needed amount of data */
					_fmemcpy(lprucLocalBuffer, pvData, uMoveLength);
					if(!WriteCacheElementToXms(lprucLocalBuffer, uXmsMapIndex))
					{
						return IO_FAILURE;
					}
				}

#else // #if !USE_ELEMENT_INFO

				if(!WritePartCacheElementToXms( pvData, uXmsMapIndex, 0,
						uMoveLength ))
				{
					return IO_FAILURE;
				}

#endif // #if !USE_ELEMENT_INFO


#if !USE_ELEMENT_INFO
				MakeCacheElementDirty(uXmsMapIndex);
#else // #if !USE_ELEMENT_INFO
				MakeCacheElementValid(uXmsMapIndex, 0, uNumberOfSectors);
				MakeCacheElementDirty(uXmsMapIndex, 0, uNumberOfSectors);
#endif // #if !USE_ELEMENT_INFO
				if(!fCacheWrite)
				{
					/* Commit the write now! */
					uReturnCode = FlushCacheElement(uXmsMapIndex);
					if(uReturnCode != IO_SUCCESS)
					{
   					/* On failure, remove this element from the cache */
   					RemoveDiskEntryFromXmsCache(uXmsMapIndex, &uOldDiskMapIndex );
						return uReturnCode;
					}
				}
			}

			/* Keep the user's sector count updated */
			*puSectorsReadWritten += uNumberOfSectors;

			//
			//  These are not needed if we are done servicing the call
			//
			//			(uchar far *) pvData += uMoveLength;
			//			uNumberOfSectors -= uNumberOfSectors;
			//			ulStartSector += uNumberOfSectors;
			//
			uNumberOfSectors = 0;
		}
	}


	return uReturnCode;
}



#if !USE_ELEMENT_INFO

/*
	Mark a given Xms cache element as dirty
*/
 static void PASCAL MakeCacheElementDirty(unsigned uXmsMapIndex)
{
	extern DISK_ENTRY far * lprdeXmsCacheMap;

	lprdeXmsCacheMap[uXmsMapIndex] |= DISK_STATUS_DIRTY;
}

#else

/*
	Mark a given Xms cache element as dirty
*/
 static void PASCAL MakeCacheElementDirty(unsigned uXmsMapIndex,
		unsigned uSectorOffsetInElement, unsigned uCount )
{
	extern DISK_ENTRY far * lprdeXmsCacheMap;
	unsigned uIndex;


	lprdeXmsCacheMap[uXmsMapIndex] |= DISK_STATUS_DIRTY;


	for( uIndex = 0; uIndex < uCount; uIndex++ )
	{
		//  Mark this sector within the element as dirty
		//
		lprElementInfo[uXmsMapIndex].uDirtySectorBits |=
				urElementSectorToMask[uIndex + uSectorOffsetInElement];
	}
}

#endif


/*
	Mark a given Xms cache element as clean
*/
static void PASCAL MakeCacheElementClean(unsigned uXmsMapIndex)
{
	extern DISK_ENTRY far * lprdeXmsCacheMap;

	lprdeXmsCacheMap[uXmsMapIndex] &= ~DISK_STATUS_DIRTY;

#if USE_ELEMENT_INFO

	//  Mark all the sectors as clean
	//
	lprElementInfo[uXmsMapIndex].uDirtySectorBits = 0;

#endif

}


#if USE_ELEMENT_INFO

/*
	Mark a sector within a cache element as valid
*/
static void PASCAL MakeCacheElementValid(unsigned uXmsMapIndex,
		unsigned uSectorOffsetInElement, unsigned uCount )
{
	extern DISK_ENTRY far * lprdeXmsCacheMap;
	unsigned uIndex;


	lprdeXmsCacheMap[uXmsMapIndex] |= DISK_STATUS_VALID;


	for( uIndex = 0; uIndex < uCount; uIndex++ )
	{
		//  Mark this sector within the element as valid
		//
		lprElementInfo[uXmsMapIndex].uValidSectorBits |=
				urElementSectorToMask[uIndex + uSectorOffsetInElement];
	}
}


#endif


#if !USE_ELEMENT_INFO

/*
	Given an index into the XMS cache, write out it's
	contents to the appropriate disk and sector.
*/
static unsigned PASCAL FlushCacheElement(unsigned uXmsMapIndex)
{
	unsigned uDisk;
	ulong ulStartSector;
	unsigned uNumberOfSectors;
	unsigned uSectorShiftValue;
	DISK_ENTRY deCache;
	unsigned uSectorsPerElement;
	ulong ulNextElementStartSector;
	uchar far * lpBuffer;
	unsigned uElementCount;
	unsigned uOldDiskMapIndex;
#if CLUSTER_ALIGNMENT
	unsigned uReturnCode;
	ulong ulNonAlignedStartSector;
#endif

	/* Get the disk information from the cache */
	if(!GetXmsDiskEntry(uXmsMapIndex, &deCache))
	{
		return IO_FAILURE;
	}

	/* Write out the cache element */
	if( DE_VALID_AND_DIRTY(deCache) )
	{
		/*
			Calculate the original disk and sector
		*/
		uDisk = (unsigned)GET_DE_DISK(deCache);
		uSectorShiftValue = diskInfo[uDisk].uSectorShiftValue;
		uSectorsPerElement = uXmsCacheElementSize >> uSectorShiftValue;
		uNumberOfSectors = uSectorsPerElement;
		ulStartSector = GET_DE_SECTOR( deCache, uSectorsPerElement );

		/* Mark the cache element as clean regardless of success */
		MakeCacheElementClean(uXmsMapIndex);

		/* Get the xms cache element into our local buffer */
		if(ReadCacheElementFromXms(uXmsMapIndex, lprucLocalBuffer))
		{
			//  Determine if we have the sectors following this element
			//
			ulNextElementStartSector = ulStartSector;
			lpBuffer = lprucLocalBuffer;

//
//  Local buffer is not being changed in the following code so there
//  is no need to invalidate it.  The part of the local buffer
//  that is being used in the following code is in the local
//  buffer but is past the first element.  The fLocalBufferValid
//  is only used for the first element in the local buffer.
//
//			fLocalBufferValid = FALSE;
//

			for( uElementCount = 1; uElementCount < uElementsInLocalBuffer;
					uElementCount++ )
			{
				//  Determine if we have the sectors following this element
				//
				ulNextElementStartSector += uSectorsPerElement;
				if( FindElementInXmsCache( uDisk, ulNextElementStartSector,
						uSectorShiftValue, &uXmsMapIndex) )
				{
					/* Get the disk information from the cache */
					if( GetXmsDiskEntry(uXmsMapIndex, &deCache) )
					{
						//  Determine if this element has been modified
						//
						if( DE_VALID_AND_DIRTY(deCache) )
						{
							//  Update the buffer address
							//
							lpBuffer += uXmsCacheElementSize;


							//  Get the buffer from XMS
							//
							if( !ReadCacheElementFromXms(uXmsMapIndex,
									lpBuffer) )
							{
								break;
							}


 							//  Mark the cache element as clean regardless
							//  of success
							//
							MakeCacheElementClean(uXmsMapIndex);


							//  Update the number of sectors to write
							//
							uNumberOfSectors += uSectorsPerElement;
							continue;
						}
					}
				}
				break;
			}


			/*
				Write out the actual data!
				Note that no error handling ocurrs here, that is done
				by the caller which has the ability to handle errors
				more appropriately.
			*/
#if !CLUSTER_ALIGNMENT
			return DiskReadWrite(WRITE, uDisk, ulStartSector,
					uNumberOfSectors, lprucLocalBuffer, &uNumberOfSectors);
#else
			if( ulStartSector )
			{
				//  Don't have to worry about an invalid sector number here
				//
				ulNonAlignedStartSector = ulStartSector -
						Int13DriveInfo[uDisk].ulSkipSectors;

				uReturnCode = DiskReadWrite(WRITE, uDisk,
						ulNonAlignedStartSector, uNumberOfSectors,
						lprucLocalBuffer,	&uNumberOfSectors );
			}
			else
			{
				//  Do have to worry about an invalid sector number here
				//
				uReturnCode = DiskReadWrite( WRITE, uDisk,
						ulStartSector, uNumberOfSectors -
						Int13DriveInfo[uDisk].ulSkipSectors,
						lprucLocalBuffer +
						(Int13DriveInfo[uDisk].ulSkipSectors * 512),
						&uNumberOfSectors );
				uNumberOfSectors += Int13DriveInfo[uDisk].ulSkipSectors;
			}
			return uReturnCode;
#endif
		}
		else
		{
			//  Should you be able to call flush for with an invalid index ???
			//
         asm mov bx,0xFFF7;
			asm	int 3
			asm	hlt

			/* Remove the element from the cache */
			RemoveDiskEntryFromXmsCache(uXmsMapIndex, &uOldDiskMapIndex );
		}
	}

	return IO_SUCCESS;
}

#else // #if !USE_ELEMENT_INFO

/*
	Given an index into the XMS cache, write out it's
	contents to the appropriate disk and sector.
*/
static unsigned PASCAL FlushCacheElement(unsigned uXmsMapIndex)
{
	unsigned uDisk;
	ulong ulStartSector;
	unsigned uNumberOfSectors;
	unsigned uSectorShiftValue;
	DISK_ENTRY deCache;
	unsigned uSectorsPerElement;
//	ulong ulNextElementStartSector;
//	uchar far * lpBuffer;
//	unsigned uElementCount;
	unsigned uOldDiskMapIndex;
#if CLUSTER_ALIGNMENT
	unsigned uReturnCode;
	ulong ulNonAlignedStartSector;
#endif // #if CLUSTER_ALIGNMENT
	unsigned uIndex;
	unsigned uFirstDirtySectorOffset;
	ulong ulDirtyStartSector;
	bool fFoundValidAndDirtySector;
	unsigned uByteOffsetInElement;
	unsigned uReturnCode;


	/* Get the disk information from the cache */
	if(!GetXmsDiskEntry(uXmsMapIndex, &deCache))
	{
		return IO_FAILURE;
	}

	/* Write out the cache element */
	if( DE_VALID_AND_DIRTY(deCache) )
	{
		/*
			Calculate the original disk and sector
		*/
		uDisk = (unsigned)GET_DE_DISK(deCache);
		uSectorShiftValue = diskInfo[uDisk].uSectorShiftValue;
		uSectorsPerElement = uXmsCacheElementSize >> uSectorShiftValue;
		ulStartSector = GET_DE_SECTOR( deCache, uSectorsPerElement );

#if CLUSTER_ALIGNMENT

		//  todo
		lalalala

#else // #if CLUSTER_ALIGNMENT

		//  Before we overwrite the contents of the local buffer, commit it
		//
		CommitLocalBuffer();


		//  Set the starting point within the element to start looking
		//  for a valid and dirty sector
		//
		uFirstDirtySectorOffset = 0;


		//  Set an initial number of sectors to commit
		//
		uNumberOfSectors = 0;


		//  Commit each block of contiguios sectors within an element
		//
		while( TRUE )
		{
			fFoundValidAndDirtySector = FALSE;

			//  Determine the first valid and dirty sector within this element
			//
			for( uIndex = uFirstDirtySectorOffset + uNumberOfSectors;
					uIndex < uSectorsPerElement; uIndex++ )
			{
				//  Until we find one that is dirty and valid ...
				//
				fFoundValidAndDirtySector = EI_VALID_DIRTY(uXmsMapIndex,uIndex);
				if( fFoundValidAndDirtySector )
				{
					//  Record the sector offset within the element of the
					//  first dirty sector
					//
					uFirstDirtySectorOffset = uIndex;


					//  Record the starting sector number of the first
					//  valid and dirty sector
					//
					ulDirtyStartSector = ulStartSector + uFirstDirtySectorOffset;


					//  The first dirty sector has been found
					//
					break;
				}
			}


			//  Determine if a valid and dirty sector was found
			//
			if( !fFoundValidAndDirtySector )
			{
				//  Nothing more to do in this element
				//
				break;
			}


			//  Adjust the index to be the next sector following the
			//  one that was found
			//
			uIndex++;


			//  Initialize a count of the number of sectors that we need to
			//  write.  At this point we only know of one dirty and valid
			//  sector.
			//
			uNumberOfSectors = 1;


			//  Determine just how many sectors are valid and dirty past
			//  the first one that we found.
			//
			for( ; uIndex < uSectorsPerElement; uIndex++ )
			{
				//  If one is found that is not valid and dirty ...
				//
				if( !EI_VALID_DIRTY(uXmsMapIndex,uIndex) )
				{
					//  Stop when a sector that is not dirty and valid is found
					//
					break;
				}


				//  Increment the number of sectors that will be written
				//  consecutively.
				//
				uNumberOfSectors++;
			}


#if CHECK_COMMIT_OF_LOCALBUFFER
			if( (fLocalBufferValid) && !fLocalBufferCommitted )
			{
            asm mov bx,0xFFF6;
				asm int 3
			}
#endif // #if CHECK_COMMIT_OF_LOCALBUFFER


			//  Manually mark the local buffer invalid since its contents
			//  is not a full element.  Its contents were already committed
			//  when we first decided to use it.
			//
			fLocalBufferValid = FALSE;


			//  Read this portion of the element into the local buffer
			//
			uByteOffsetInElement = uFirstDirtySectorOffset << uSectorShiftValue;
			if( !ReadPartCacheElementFromXms(uXmsMapIndex,
							lprucLocalBuffer + uByteOffsetInElement,
							uByteOffsetInElement,
							uNumberOfSectors << uSectorShiftValue ))
			{
				//  Should you be able to call flush for with an invalid index ???
				//
            asm mov bx,0xFFF5;
				asm	int 3
				asm	hlt

				/* Remove the element from the cache */
				RemoveDiskEntryFromXmsCache(uXmsMapIndex, &uOldDiskMapIndex );
			}


			//  Write this portion of the element
			//
			//  Note that no error handling ocurrs here, that is done
			//  by the caller which has the ability to handle errors
			//  more appropriately.
			//
			uReturnCode = DiskReadWrite(WRITE, uDisk, ulDirtyStartSector,
			      uNumberOfSectors, lprucLocalBuffer + uByteOffsetInElement,
					&uNumberOfSectors);
			if(uReturnCode != IO_SUCCESS)
			{
				return uReturnCode;
			}
		}


		/* Mark the cache element as clean regardless of success */
		MakeCacheElementClean(uXmsMapIndex);


		//  Manually mark the local buffer invalid since its contents
		//  is not a full element.  Its contents were already committed
		//  when we first decided to use it.
		//
		fLocalBufferValid = FALSE;

#if CHECK_COMMIT_OF_LOCALBUFFER
		if( (fLocalBufferValid) && !fLocalBufferCommitted )
		{
         asm mov bx,0xFFF4;
			asm int 3
		}
#endif // #if CHECK_COMMIT_OF_LOCALBUFFER

#endif // #if CLUSTER_ALIGNMENT
	}

	return IO_SUCCESS;
}

#endif // #if !USE_ELEMENT_INFO


//  Commit what is in the local buffer to XMS
//
static void CommitLocalBuffer( void )
{
	if( fLocalBufferValid )
	{
		if( !fLocalBufferCommitted )
		{
			/* Move the XMS data into the local buffer */
			xmsMove.length = uXmsCacheElementSize;
			xmsMove.destHandle = uXmsHandle;
			xmsMove.destOffset.offset32 =
						(ulong)uLocalBufferXmsMapIndex * uXmsCacheElementSize;
			xmsMove.sourceHandle = 0;
			xmsMove.sourceOffset.p.lpOff = FP_OFF(lprucLocalBuffer);
			xmsMove.sourceOffset.p.lpSeg = FP_SEG(lprucLocalBuffer);

			/* Request XMS to perform the move */
			if(MoveXMSBlock(&xmsMove) != XMS_SUCCESS)
			{
				fLocalBufferValid = FALSE;
				fLocalBufferCommitted = FALSE;
#if CHECK_COMMIT_OF_LOCALBUFFER
				if( (fLocalBufferValid) && !fLocalBufferCommitted )
				{
               asm mov bx,0xFFF3;
					asm int 3
				}
#endif // #if CHECK_COMMIT_OF_LOCALBUFFER
			}
			else
			{
				fLocalBufferCommitted = TRUE;
			}
		}
	}

#if CHECK_COMMIT_OF_LOCALBUFFER
	if( (fLocalBufferValid) && !fLocalBufferCommitted )
	{
      asm mov bx,0xFFF2;
		asm int 3
	}
#endif // #if CHECK_COMMIT_OF_LOCALBUFFER
}


/*
	Write a given element to the XMS cache from the
	speficied buffer.  This function also keeps local
	buffer cache information updated.
*/
static bool PASCAL WriteCacheElementToXms(void far * vpDataBuffer,
						unsigned uXmsMapIndex)
{
	ASSERT(vpDataBuffer);


	if(vpDataBuffer == lprucLocalBuffer)
	{
		fLocalBufferValid = TRUE;
		uLocalBufferXmsMapIndex = uXmsMapIndex;
		fLocalBufferCommitted = FALSE;
	}
	else
	{
		//  If this information is in the local buffer, then update the local
		//  buffer
		//
		if( (fLocalBufferValid) && (uLocalBufferXmsMapIndex == uXmsMapIndex) )
		{
			_fmemcpy( lprucLocalBuffer, vpDataBuffer, uXmsCacheElementSize );
			fLocalBufferCommitted = FALSE;
		}
		else
		{
			/* Move the XMS data into the local buffer */
			xmsMove.length = uXmsCacheElementSize;
			xmsMove.destHandle = uXmsHandle;
			xmsMove.destOffset.offset32 =
						(ulong) uXmsMapIndex * uXmsCacheElementSize;
			xmsMove.sourceHandle = 0;
			xmsMove.sourceOffset.p.lpOff = FP_OFF(vpDataBuffer);
			xmsMove.sourceOffset.p.lpSeg = FP_SEG(vpDataBuffer);

			/* Request XMS to perform the move */
			if(MoveXMSBlock(&xmsMove) != XMS_SUCCESS)
			{
				if(vpDataBuffer == lprucLocalBuffer)
				{
					fLocalBufferValid = FALSE;
				}
				return FALSE;
			}
		}
	}

	/* Update information about our local cache */
	if(vpDataBuffer == lprucLocalBuffer)
	{
		fLocalBufferValid = TRUE;
		uLocalBufferXmsMapIndex = uXmsMapIndex;
	}

	return TRUE;
}


/*
	Write a given element to the XMS cache from the
	speficied buffer.  This function also keeps local
	buffer cache information updated.
*/
static bool PASCAL WritePartCacheElementToXms(void far * vpDataBuffer,
						unsigned uXmsMapIndex, unsigned uOffsetInElement,
						unsigned uByteCount )
{
	ASSERT(vpDataBuffer);


	//  If this information is in the local buffer, then update the local
	//  buffer
	//
	if( (fLocalBufferValid) && (uLocalBufferXmsMapIndex == uXmsMapIndex) )
	{
		_fmemcpy( &lprucLocalBuffer[uOffsetInElement], vpDataBuffer,
				uByteCount );
		fLocalBufferCommitted = FALSE;
	}
	else
	{
		/* Move the XMS data into the local buffer */
		xmsMove.length = uByteCount;
		xmsMove.destHandle = uXmsHandle;
		xmsMove.destOffset.offset32 = (((ulong)uXmsMapIndex) *
				uXmsCacheElementSize) + uOffsetInElement;
		xmsMove.sourceHandle = 0;
		xmsMove.sourceOffset.p.lpOff = FP_OFF(vpDataBuffer);
		xmsMove.sourceOffset.p.lpSeg = FP_SEG(vpDataBuffer);

		/* Request XMS to perform the move */
		if(MoveXMSBlock(&xmsMove) != XMS_SUCCESS)
		{
			if(vpDataBuffer == lprucLocalBuffer)
			{
				fLocalBufferValid = FALSE;
			}
			return FALSE;
		}
	}

	return TRUE;
}


/*
	Read a given element from the XMS cache into the
	speficied buffer.  This function also keeps the local
	buffer cache information updated.
*/
static bool PASCAL ReadCacheElementFromXms(unsigned uXmsMapIndex,
						void far * vpDataBuffer)
{
	ASSERT(vpDataBuffer);

	/* Check in the local buffer */
	if( (fLocalBufferValid) && (uLocalBufferXmsMapIndex == uXmsMapIndex) )
	{
		if( vpDataBuffer == lprucLocalBuffer )
		{
			/* Nothing to do! */
		}
		else
		{
			_fmemcpy( vpDataBuffer, lprucLocalBuffer, uXmsCacheElementSize );
		}
		return TRUE;
	}

	if( vpDataBuffer == lprucLocalBuffer )
	{
		CommitLocalBuffer();
	}

#if CHECK_COMMIT_OF_LOCALBUFFER
	if( IS_LOCALBUFFER(vpDataBuffer) && (fLocalBufferValid) && !fLocalBufferCommitted )
	{
      asm mov bx,0xFFF1;
		asm int 3
	}
#endif

	/* Move the XMS data into the local buffer */
	xmsMove.length = uXmsCacheElementSize;
	xmsMove.sourceHandle = uXmsHandle;
	xmsMove.sourceOffset.offset32 =
				(ulong) uXmsMapIndex * uXmsCacheElementSize;
	xmsMove.destHandle = 0;
	xmsMove.destOffset.p.lpOff = FP_OFF(vpDataBuffer);
	xmsMove.destOffset.p.lpSeg = FP_SEG(vpDataBuffer);

	/* Request XMS to perform the move */
	if(MoveXMSBlock(&xmsMove) != XMS_SUCCESS)
	{
		if(vpDataBuffer == lprucLocalBuffer)
		{
 			fLocalBufferValid = FALSE;
#if CHECK_COMMIT_OF_LOCALBUFFER
			if( fLocalBufferCommitted )
			{
            asm mov bx,0xFFF0;
				asm int 3
			}
#endif
		}
		return FALSE;
	}

	/* If the data is going into our local cache, record it */
 	if(vpDataBuffer == lprucLocalBuffer)
	{
		/* Track what is in the local buffer */
		fLocalBufferValid = TRUE;
		uLocalBufferXmsMapIndex = uXmsMapIndex;
		fLocalBufferCommitted = TRUE;
	}

	return TRUE;
}


/*
	Read part a given element from the XMS cache into the
	speficied buffer.  This function also keeps the local
	buffer cache information updated.
*/
//
//  NOTE:  This is not intended to be use for the local buffer
//
static bool PASCAL ReadPartCacheElementFromXms(unsigned uXmsMapIndex,
						void far * vpDataBuffer, unsigned uOffsetInElement,
						unsigned uByteCount )
{
	ASSERT(vpDataBuffer);

	/* Check in the local buffer */
	if( (fLocalBufferValid) && (uLocalBufferXmsMapIndex == uXmsMapIndex) )
	{
		_fmemcpy( vpDataBuffer, lprucLocalBuffer + uOffsetInElement,
					uByteCount );
		return TRUE;
	}

	if( vpDataBuffer == lprucLocalBuffer )
	{
		CommitLocalBuffer();
	}

#if CHECK_COMMIT_OF_LOCALBUFFER
	if( IS_LOCALBUFFER(vpDataBuffer) && (fLocalBufferValid) && !fLocalBufferCommitted )
	{
      asm mov bx,0xFFEF;
		asm int 3
	}
#endif

	/* Move the XMS data into the local buffer */
	xmsMove.length = uByteCount;
	xmsMove.sourceHandle = uXmsHandle;
	xmsMove.sourceOffset.offset32 = (((ulong)uXmsMapIndex) *
			uXmsCacheElementSize) + uOffsetInElement;
	xmsMove.destHandle = 0;
	xmsMove.destOffset.p.lpOff = FP_OFF(vpDataBuffer);
	xmsMove.destOffset.p.lpSeg = FP_SEG(vpDataBuffer);

	/* Request XMS to perform the move */
	if(MoveXMSBlock(&xmsMove) != XMS_SUCCESS)
	{
		if(vpDataBuffer == lprucLocalBuffer)
		{
 			fLocalBufferValid = FALSE;
#if CHECK_COMMIT_OF_LOCALBUFFER
			if( fLocalBufferCommitted )
			{
            asm mov bx,0xFFEE;
				asm int 3
			}
#endif
		}
		return FALSE;
	}

	/* If the data is going into our local cache, record it */
 	if(vpDataBuffer == lprucLocalBuffer)
	{
		/* Track what is in the local buffer */
		fLocalBufferValid = TRUE;
		uLocalBufferXmsMapIndex = uXmsMapIndex;
		fLocalBufferCommitted = TRUE;
	}

	return TRUE;
}


#if TESTING


#define TESTX(x) if(!(x)) { \
			printf("Failed file: %s line: %u", \
			__FILE__, __LINE__); return FALSE; }

#define TEST_VALUE 0x57
#define TEST_BUFF_SIZE 4096


 static unsigned PASCAL DiskReadWrite(bool fRead, unsigned uDisk,
			ulong ulStartSector, unsigned uNumberOfSectors,
			void far * pvData, unsigned * puSecsTransfered)
{
	uchar ucByte;

	ASSERT(vpDataBuffer);

	/* A repeatable patern based on the disk and sector number */
	ucByte = (uchar) ((uDisk << 5) + (ulStartSector & 0x1F));
	if(ucByte == TEST_VALUE)
	{
		++ucByte;
	}

	/* Only supports read */
	if(fRead)
	{
		_fmemset(pvData, ucByte, uNumberOfSectors * 512);
	}

	/* Fake it */
	*puSecsTransfered = uNumberOfSectors;

	/* No error code */
	return IO_SUCCESS;
}


bool PASCAL TestCache(void)
{
	static uchar rucTempBuff[4096]; /* Two elements */
	unsigned uNumSecsTransfered;
	unsigned uDisk;

	/*
		The following tests are configured for an 8 element cache
	*/

	TESTX(InitializeDiskCache(6, 2048, 2048));
	TESTX(RegisterDiskInformation(3, 512, 0));
	TESTX(RegisterDiskInformation(2, 512, 0));


	/* Try an invalid disk drive or two */
	TESTX(IO_SUCCESS != CacheReadWrite(READ, 13, 9, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(IO_SUCCESS != CacheReadWrite(READ, 39, 9, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(!FlushDiskCache(5));
	TESTX(!FlushDiskCache(29));

	uDisk = 3;

	/* Alignment tests different elements (all misses) */
	{
		/* Read one sector top aligned */
		_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
		TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 9, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
		TESTX(rucTempBuff[0] != TEST_VALUE);
		TESTX(rucTempBuff[512] == TEST_VALUE);

		/* Read one sector midle of element */
		_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
		TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 15, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
		TESTX(rucTempBuff[0] != TEST_VALUE);
		TESTX(rucTempBuff[512] == TEST_VALUE);

		/* Read one sector midle of element */
		_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
		TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 16, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
		TESTX(rucTempBuff[0] != TEST_VALUE);
		TESTX(rucTempBuff[512] == TEST_VALUE);
	}
	/* Elements cached (LRU -> MRU): 2, 3, 4 */

	/* Alignment tests same element (all hits) */
	{
		/* Read one sector bottom aligned */
		_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
		TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 16, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
		TESTX(rucTempBuff[0] != TEST_VALUE);
		TESTX(rucTempBuff[512] == TEST_VALUE);

		/* Read one sector top aligned */
		_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
		TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 17, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
		TESTX(rucTempBuff[0] != TEST_VALUE);
		TESTX(rucTempBuff[512] == TEST_VALUE);

		/* Read one sector top aligned */
		_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
		TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 18, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
		TESTX(rucTempBuff[0] != TEST_VALUE);
		TESTX(rucTempBuff[512] == TEST_VALUE);

		/* Read one sector midle of element */
		_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
		TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 19, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
		TESTX(rucTempBuff[0] != TEST_VALUE);
		TESTX(rucTempBuff[512] == TEST_VALUE);
	}
	/* Elements cached (LRU -> MRU): 2, 3, 4, 5 */


	/* Full element read (miss) */
	_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 20, 4, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[0] != TEST_VALUE);

	/* Full element read (hit) */
	_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 20, 4, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[0] != TEST_VALUE);

	/* Elements cached (LRU -> MRU): 2, 3, 4, 5, 6 */

	/* New Disk */
	uDisk = 2;

	/* Element sized read, crossing element boundary (miss) */
	_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 26, 4, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[0] != TEST_VALUE);
	TESTX(rucTempBuff[1023] != rucTempBuff[1024]);
	TESTX(rucTempBuff[1024] != TEST_VALUE);

	/* Element sized read, crossing element boundary (hit) */
	_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 26, 4, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[0] != TEST_VALUE);
	TESTX(rucTempBuff[1023] != rucTempBuff[1024]);
	TESTX(rucTempBuff[1024] != TEST_VALUE);

	/* Cache (LRU -> MRU): 3/2, 3/3, 3/4, 3/5, 3/6, 2/1, 2/2 */

	/* Write Tests */

	/* Two sector sized (NON cached) write, (hit) */
	_fmemset(rucTempBuff, ~TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(WRITE, uDisk, 24, 2, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[0] == (uchar) ~TEST_VALUE);

	/* Two sector sized (CACHED) write, (hit) */
	_fmemset(rucTempBuff, ~TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(WRITE, uDisk, 28, 2, rucTempBuff, TRUE, &uNumSecsTransfered));
	TESTX(rucTempBuff[0] == (uchar) ~TEST_VALUE);

	/* Read back the data in the cache (cached data still there) */

	/*
		Two sector sized read, (hit)
		First 1K is ~TEST data and the second is original read
	*/
	_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 24, 4, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[1023] == (uchar) ~TEST_VALUE);
	TESTX(rucTempBuff[1024] != TEST_VALUE);

	/* Two sector sized read, (hit) */
	_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 28, 4, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[1023] == (uchar) ~TEST_VALUE);
	TESTX(rucTempBuff[1024] != TEST_VALUE);

	/* Cache (LRU -> MRU): 3/2, 3/3, 3/4, 3/5, 3/6, 2/1, 2/2 */

	/* Cache is now full, go back to fist disk */
	uDisk = 3;

	/* Read the original disk sector (no longer LRU) */
	_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 8, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[0] != TEST_VALUE);
	TESTX(rucTempBuff[511] != TEST_VALUE);
	TESTX(rucTempBuff[512] == TEST_VALUE);

	/* Cache (LRU -> MRU): 3/3, 3/4, 3/5, 3/6, 2/1, 2/2, 3/2 */

	/* Test a flush of one element */
	TESTX(FlushOneCacheElement());

	/* Cache (LRU -> MRU): 3/3, 3/4, 3/5, 3/6, 2/1, 2/2, 3/2 */

	/*
		Now the second element is the LRU, write cache it
	*/
	_fmemset(rucTempBuff, ~TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(WRITE, uDisk, 12, 1, rucTempBuff, TRUE, &uNumSecsTransfered));
	TESTX(rucTempBuff[0] == (uchar) ~TEST_VALUE);

	_fmemset(rucTempBuff, TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 12, 1, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[0] == (uchar) ~TEST_VALUE);
	TESTX(rucTempBuff[512] == TEST_VALUE);

	/* Cache (LRU -> MRU): 3/3, 3/4, 3/5, 3/6, 2/1, 2/2, 3/2 */

	/* And all of the second disk */
	TESTX(FlushDiskCache(3));

	/* Read the original disk sector (no longer LRU) */
	_fmemset(rucTempBuff, ~TEST_VALUE, TEST_BUFF_SIZE);
	TESTX(IO_SUCCESS != CacheReadWrite(READ, uDisk, 12, 4, rucTempBuff, FALSE, &uNumSecsTransfered));
	TESTX(rucTempBuff[512] != TEST_VALUE);

	/* And all of the second disk */
	InvalidateDiskCache(3);
	TESTX(FlushDiskCache(2));
	InvalidateDiskCache(2);


	DeinitializeDiskCache();

	return TRUE;
}
#endif
