//
//   $Id: CACHRES.C 1.18 2003/02/18 15:36:00Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  CACHRES.C
//
//  Copyright (c) 2001-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//  06-28-2001  bt  Created
//


//  Borland libraries
//
//#include <stdio.h>
#include <mem.h>
//#include <string.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"


//  Program definitions
//
#include "cachinit.h"
#include "cache.h"


//  Prototypes
//
#include "int2f.pt"
#include "int1302.pt"
#include "int1342.pt"
#include "cachres.pt"
#include "devcall.pt"

void DoPreviousService( void );
void PASCAL RebootTheMachine(void);


//  Externs
//
extern CALLREG Int2fRegisters;
extern bool fHandledCall;


//  Globals
//
ulong ulElementSize = 8192;
ulong ulBufferSize  = 16384;
ulong ulInitCacheSizeKB = 0;
ulong ulWinCacheSizeKB  = 0;
INT13_DRIVE_INFO Int13DriveInfo[26];
uchar ucDriveInfoEntries;
uchar ucFirstCDROMEntry;
bool fDontWriteBeforeCommand;
ulong ulCacheHits;
ulong ulCacheMisses;


#if DEBUG_BAD_ERROR_CODES
unsigned uGlobalErrorCode;
#endif

#if DEBUG_BAD_DATA
uchar ucGlobalBuffer[2048];
#endif


#if  COUNT_READS_AND_WRITES
ulong ulHDReadRequests = 0;
ulong ulHDWriteRequests = 0;
ulong ulCDReadRequests = 0;
ulong ulCDWriteRequests = 0;
#endif


//  Flush all the unwritten buffers in the cache for all drives
//
void PASCAL FlushAllBuffers( void )
{
	uchar ucIndex;


	//  Go through each of the drives in the drive info list
	//
	for( ucIndex = 0; ucIndex < ucDriveInfoEntries; ucIndex++ )
	{
		//  Determine if the drive is cached
		//
		if( Int13DriveInfo[ucIndex].fIsCached )
		{
			//  Determine if the drive is write cached
			//
			if( Int13DriveInfo[ucIndex].fIsWriteCached )
			{
				//  Flush any uncommitted sectors for this drive
				//
				FlushDiskCache( ucIndex );
			}
		}
	}
}


//  Invalidate all the buffers in the cache for all drives
//
void PASCAL InvalidateAllBuffers( void )
{
	uchar ucIndex;


	//  Go through each of the drives in the drive info list
	//
	for( ucIndex = 0; ucIndex < ucDriveInfoEntries; ucIndex++ )
	{
		//  Determine if the drive is cached
		//
		if( Int13DriveInfo[ucIndex].fIsCached )
		{
			//  Determine if the drive is write cached
			//
			if( Int13DriveInfo[ucIndex].fIsWriteCached )
			{
				//  Flush any uncommitted sectors for this drive
				//
				FlushDiskCache( ucIndex );
			}


			//  Invalidate all sectors for this drive
			//
			InvalidateDiskCache( ucIndex );
		}
	}
}


//  All device calls get dispatched through here
//
void PASCAL _DeviceDispatch( void )
{
	extern LPDEVREQHDR lpRequest;
	extern bool fDeviceCallHandled;
	extern bool fInFDHD;
	bool fIsRead  = FALSE;
	bool fIsWrite = FALSE;
	ulong ulTempCacheMisses;
	ulong ulPhysicalSector;
	unsigned uNumberOfSectors;
	uchar ucDriveIdentifier;
	bool fIsWriteCached;
	unsigned uErrorCode;
	uchar far * lpucData;
	ulong ulSize;
	LPBPB lpBPB;


	if( fInFDHD )
	{
		//  Determine the drive info index
		//
		ucDriveIdentifier = lpRequest->unit;


		//  Determine if this drive is cached
		//
		if( !Int13DriveInfo[ucDriveIdentifier].fIsCached )
		{
			//  We are not caching this drive
			//
			DoPreviousService();
			return;
		}


		//  Determine if this drive is write cached
		//
		fIsWriteCached = Int13DriveInfo[ucDriveIdentifier].fIsWriteCached;


		//  Due to the complexity of the whole media change concept,
		//  removable media will not be write cached
		//
		if( Int13DriveInfo[ucDriveIdentifier].fIsRemovable )
		{
			fIsWriteCached = FALSE;
		}


		//  Determine how to handle DOS device commands
		//
		switch( lpRequest->command )
		{
			//  Media change check
			//
			case 0x01:
				{
					//
					//
					DoPreviousService();
					if( *((uchar far *)(&(lpRequest->trans))) != 1 )
					{
						//  Disk has changed, invalidate all our buffers
						//
						//  NOTE:  This code assumes that we are not write
						//         caching removable media.  There was code
						//         added before the call to cachereadwrite
						//         to assure this assumption was true
						//
						InvalidateDiskCache( ucDriveIdentifier );
					}
				}
				break;


			//  Rebuild BPB
			//
			case 0x02:
				{
					//  This is called whenever the media is changed.  Perform
					//  the requested operation, then update the internal
					//  structures with the number of sectors on the disk
					//
					DoPreviousService();
					lpBPB = *((LPBPB far * )(&lpRequest->count));
					ulSize = lpBPB->num_sec;
					if( ulSize == 0 )
					{
						ulSize = lpBPB->lnumsec;
					}
					Int13DriveInfo[ucDriveIdentifier].ulEnd = ulSize;
					return;
				}
//unreachable				break;


			//  Device read
			//
			case 0x04:
				{
					//  Determine the starting sector number
					//
					if( lpRequest->start == 0xffff )
					{
						ulPhysicalSector = lpRequest->start32;
					}
					else
					{
						ulPhysicalSector = lpRequest->start;
					}


#if COMM_PORT_LOGGING
					CommLog( "HR", ucDriveIdentifier, ulPhysicalSector, lpRequest->count );
#endif

#if  COUNT_READS_AND_WRITES
					ulHDReadRequests++;
#endif

					fIsRead = TRUE;
				}
				break;


			//  Device write and write with verify
			//
			case 0x08:  // Write
			case 0x09:  // Write with verify
				{
					//  Determine the starting sector number
					//
					if( lpRequest->start == 0xffff )
					{
						ulPhysicalSector = lpRequest->start32;
					}
					else
					{
						ulPhysicalSector = lpRequest->start;
					}


#if COMM_PORT_LOGGING
					CommLog( "HW", ucDriveIdentifier, ulPhysicalSector, lpRequest->count );
#endif

#if  COUNT_READS_AND_WRITES
					ulHDWriteRequests++;
#endif

					fIsWrite = TRUE;
				}
				break;


			//  These functions that affect the information that we
			//  have cached
			//
			case 0x13:
				{
					switch( *((uchar far *)(&lpRequest->trans)) )
					{
						case 0x40:
						case 0x41:
						case 0x42:
						case 0x46:
						case 0x61:
						case 0x62:
							{
								if( fIsWriteCached )
								{
									FlushDiskCache( ucDriveIdentifier );
								}
								InvalidateDiskCache( ucDriveIdentifier );
							}
					}
				}
				break;


			//  Do nothing
			//
			default:
				{
					//  Do nothing
				}
				break;
		}
	}
	else
	{
		//  Determine the drive info index
		//
		ucDriveIdentifier = lpRequest->unit + ucFirstCDROMEntry;


		//  Determine if this drive is cached
		//
		if( !Int13DriveInfo[ucDriveIdentifier].fIsCached )
		{
			//  We are not caching this drive
			//
			DoPreviousService();
			return;
		}


		//  Determine if this drive is write cached (Don't write cache CDROMs)
		//
		fIsWriteCached = FALSE;


		//  Determine how to handle CDEX/ATAPI device commands
		//
		switch( lpRequest->command )
		{
			//  Read long
			//
			case 0x80:
			{
				//  Determine the starting sector number
				//
		 		ulPhysicalSector = *(ulong far *)(&lpRequest->start);


#if COMM_PORT_LOGGING
					CommLog( "CR", ucDriveIdentifier, ulPhysicalSector, lpRequest->count );
#endif

#if  COUNT_READS_AND_WRITES
					ulCDReadRequests++;
#endif

				fIsRead = TRUE;
			}
			break;


			//  Write long and Write long with verify
			//
			case 0x86:
			case 0x87:
			{
				//  Determine the starting sector number
				//
		 		ulPhysicalSector = *(ulong far *)(&lpRequest->start);


#if COMM_PORT_LOGGING
					CommLog( "CW", ucDriveIdentifier, ulPhysicalSector, lpRequest->count );
#endif

#if  COUNT_READS_AND_WRITES
					ulCDWriteRequests++;
#endif

				fIsWrite = TRUE;
			}
			break;


			//  Input flush (7h)
			//  Output flush (11h)
			//
			//  These functions that affect the information that we
			//  have cached
			//
			case 0x7:
			case 0x11:
				{
					if( fIsWriteCached )
					{
						FlushDiskCache( ucDriveIdentifier );
					}
					InvalidateDiskCache( ucDriveIdentifier );
				}
				break;


			//  IOCTL Output
			//
			//  These functions that affect the information that we
			//  have cached
			//
			case 0x12:
				{
					//  Get the address of the data packet
					//
					lpucData = lpRequest->trans;


					//  Determine which command
					//
					switch( lpucData[0] )
					{
						//  Eject disk
						//
						case 0x00:
							{
								if( fIsWriteCached )
								{
									FlushDiskCache( ucDriveIdentifier );
								}
								InvalidateDiskCache( ucDriveIdentifier );
							}
					}
				}
				break;


			//  IOCTL Read
			//
			case 0x03:
				{
					//  Get the address of the data packet
					//
					lpucData = lpRequest->trans;


					//  Perform the media change request
					//
					DoPreviousService();


					//  Check for media change request
					//
					if( lpucData[0] == 0x09 )
					{
						//  Check the result of the request
						//
						if( lpucData[1] != 1 )
						{
							InvalidateDiskCache( ucDriveIdentifier );
						}
					}
					else if( lpucData[0] == 0x08 )
					{
						//  This is called whenever the media is changed.  Perform
						//  the requested operation, then update the internal
						//  structures with the number of sectors on the disk
						//
						ulSize = *((ulong far *)(&lpucData[1]));
						Int13DriveInfo[ucDriveIdentifier].ulEnd = ulSize;
					}


					//  Request has been serviced
					//
					return;
				}


			//  Do nothing
			//
			default:
				{
					//  Do nothing
				}
				break;
		}
	}


	//  Determine if this was a read or write
	//
	if( fIsRead || fIsWrite )
	{
		//  Set the flag indicating that this call was handled
		//
		fDeviceCallHandled = TRUE;


		//  Save the current cache misses in a temporary variable
		//
		ulTempCacheMisses = ulCacheMisses;


#if DEBUG_BAD_DRIVES
		if( ucDriveIdentifier > ucDriveInfoEntries )
		{
			asm int 3
		}
#endif


#if DEBUG_BAD_ERROR_CODES
		uGlobalErrorCode = 0;
#endif


		//  Determine if the requested sectors are past the end of the disk
		//
		if( (ulPhysicalSector + lpRequest->count) >
				Int13DriveInfo[ucDriveIdentifier].ulEnd )
		{
			//  Return sector not found for requests past the end of disk
			//
			uErrorCode = 0x8108;
#if DEBUG_BAD_ERROR_CODES
			uGlobalErrorCode = uErrorCode;
#endif
		}
		else
		{
			//  Perform the read/write operation
			//
			uErrorCode = CacheReadWrite( fIsRead, ucDriveIdentifier,
					ulPhysicalSector, lpRequest->count, lpRequest->trans,
					fIsWriteCached, &uNumberOfSectors );
		}


#if DEBUG_BAD_ERROR_CODES
		if( uGlobalErrorCode != uErrorCode )
		{
			asm int 3
			uErrorCode = uGlobalErrorCode;
		}
#endif


		if( uErrorCode )
		{
			lpRequest->status = uErrorCode;
		}
		else
		{
#if DEBUG_BAD_DATA
			if( fIsRead )
			{
				unsigned uTemp;
				unsigned uTempSectorsRead;
				unsigned uTempIndex;
				unsigned uTempSectorSize;


				uTempSectorSize = Int13DriveInfo[ucDriveIdentifier].uSectorSize;

				for( uTempIndex = 0; uTempIndex < lpRequest->count;
						uTempIndex++ )
				{
					if( fInFDHD )
					{
						uTemp = DebugRead( TRUE, ucDriveIdentifier,
								ulPhysicalSector + uTempIndex,
								1, (uchar far *)(&ucGlobalBuffer[0]),
								&uTempSectorsRead );
						uTempSectorsRead = 1;
					}
					else
					{
						uTemp = DeviceReadWrite( TRUE, ucDriveIdentifier,
								ulPhysicalSector + uTempIndex,
								1, (uchar far *)(&ucGlobalBuffer[0]),
								&uTempSectorsRead );
					}
					if( (uTemp) || (uTempSectorsRead != 1) )
					{
						//  can't read the sector(s) we just got?
						//
						asm int 3
						break;
					}
					else
					{
						if( _fmemcmp( (uchar far *)(lpRequest->trans) +
								(uTempIndex * uTempSectorSize), ucGlobalBuffer,
								uTempSectorSize) != 0 )
						{
							//  User got bad data (overwriting bad data with good)
							//
							asm int 3
							_fmemcpy( (uchar far *)(lpRequest->trans) +
									(uTempIndex * uTempSectorSize), ucGlobalBuffer,
									uTempSectorSize );
						}
					}
				}
			}
#endif // #if DEBUG_BAD_DATA


			lpRequest->status = 0x0100;
		}


		//  Update the device request with the results of
		//  the read or write
		//
		lpRequest->count = uNumberOfSectors;


		//  Determine if we had to hit the disk
		//
		if( ulCacheMisses == ulTempCacheMisses )
		{
			ulCacheHits++;
		}
		else
		{
			ulCacheMisses = ulTempCacheMisses + 1;
		}
	}
	else
	{
		DoPreviousService();
	}
}


//  All int 13h functions get dispatched through here
//
void PASCAL _Int13Dispatch( void )
{
	uchar ucIndex;
	uchar ucBIOSDriveNumber;
	bool fIsRead = FALSE;


	//  Determine how to handle this function
	//
	ucBIOSDriveNumber = LOWBYTE(Int2fRegisters.dx);
	switch( HIGHBYTE(Int2fRegisters.ax) )
	{
		case 0x02:	//  CHS read
		case 0x42:	//  LBA extended read
			fIsRead = TRUE;

		case 0x03:	//  CHS write
		case 0x43:	//  LBA extended write
		case 0x05:	//  CHS format track
		case 0x06:	//  CHS format cylinder
		case 0x07:  //  CHS format drive from specified cylinder
		case 0x0a:	//  CHS read long sector
		case 0x0b:	//  CHS write long sector
			{
				//  Determine if this is a request for a drive that
				//  is being cached
				//
				for( ucIndex = 0; ucIndex < ucFirstCDROMEntry; ucIndex++ )
				{
					//  Determine if the drive is cached
					//
					if( Int13DriveInfo[ucIndex].fIsCached )
					{
						//  Determine if this is the drive number for the
						//  int 13h request
						//
						if( Int13DriveInfo[ucIndex].ucBIOSDriveNumber ==
								ucBIOSDriveNumber)
						{
							//  NOTE: Don't break out because the BIOS drive
							//        number can be used for more than one
							//        logical drive
							//
							FlushDiskCache(ucIndex);
							if( !fIsRead )
							{
								InvalidateDiskCache(ucIndex);
							}
						}
					}
				}
			}
			break;
	}
}


//  All int 25h and int 26h functions get dispatched through here
//
void PASCAL _AbsRDWRDispatch( void )
{
	extern bool fABSRDWRIsRead;
	extern bool fLeaveFlagsOnStack;
	unsigned uError;
	unsigned uNumberOfSectors;
	unsigned uSectorsRead;
	ulong ulStartingSector;
	uchar ucDOSDriveNumber;
	uchar far * lpucData;
	LPABS_RDWR lpABS;


	//  Get the DOS drive number
	//
	ucDOSDriveNumber = LOWBYTE(Int2fRegisters.ax);


	//  Determine if this drive is cached
	//
	if( !Int13DriveInfo[ucDOSDriveNumber].fIsCached )
	{
		//  Let the previous handler service the call
		//
		return;
	}


	//  Get the values for the ABS read/write
	//
	uNumberOfSectors = Int2fRegisters.cx;
	if( uNumberOfSectors == 0xffff )
	{
		//  Request values is in the request packet
		//
		lpABS = MK_FP( Int2fRegisters.ds, Int2fRegisters.bx );
		uNumberOfSectors = lpABS->uCount;
		ulStartingSector = lpABS->ulStartSector;
		lpucData = lpABS->lpBuffer;
	}
	else
	{
		//  Request values are in the registers.
		//
		ulStartingSector = Int2fRegisters.dx;
		lpucData = MK_FP( Int2fRegisters.ds, Int2fRegisters.bx );
	}


	//  Perform the read/write
	//
	uError = CacheReadWrite( fABSRDWRIsRead, ucDOSDriveNumber,
			ulStartingSector, uNumberOfSectors, lpucData,
			Int13DriveInfo[ucDOSDriveNumber].fIsWriteCached, &uSectorsRead );


	//  Determine if there was an error
	//
	if( uError )
	{
		Int2fRegisters.ax = uError;
		Int2fRegisters.flags |= 1;	// set carry
	}
	else
	{
		Int2fRegisters.flags &= ~1;	// clear carry
	}


	//  Indicate that the call was serviced
	//
	fHandledCall = TRUE;


	//  Set the flag indicating to leave the flags on the stack for
	//  emulating Int 25h and Int 26h
	//
	fLeaveFlagsOnStack = TRUE;
}


//
//
#pragma argsused
unsigned PASCAL DiskReadWrite( bool fIsRead, unsigned ucDrive, ulong ulStartingSector,
		unsigned uNumberOfSectors, void far * lpucData,
		unsigned * puSectorsRead )
{
#if 0
	uchar ucBIOSDriveNumber;
	unsigned uCylinder;
	uchar ucHead;
	uchar ucSector;
	bool fIsCHS;
	static LBAPACKET LBAPacket;
#endif
	unsigned uErrorCode;
	ulong ulTemp;


	//  Only reads are counted as cache misses
	//
	if( fIsRead )
	{
		ulCacheMisses++;
	}


//  This is not used.  It is better to call the device to perform
//  read and writes.  This code is left here in case we find a reason
//  to do otherwise.
//
#if 0
	//  Determine if this is an Int 13 request call showing up here
	//
	if( ucDrive < ucFirstCDROMEntry )
	{
		//  Get the BIOS drive number for the drive
		//
		ucBIOSDriveNumber = Int13DriveInfo[ucDrive].ucBIOSDriveNumber;
		ulStartingSector += Int13DriveInfo[ucDrive].ulStart;


		//  Attempt to convert the sector number to CHS.  If it fails
		//  use LBA
		//
		fIsCHS = LBAToCHS( ucBIOSDriveNumber, ulStartingSector, &uCylinder,
				&ucHead, &ucSector );
		if( !fIsCHS )
		{
			//  Build an LBA request packet
			//
			memset( &LBAPacket, 0, sizeof(LBAPacket) );
			LBAPacket.ucPacketSize = 0x10;
			LBAPacket.uNumberOfSectors = uNumberOfSectors;
			LBAPacket.ulSectorLowDWORD = ulStartingSector;
			LBAPacket.lpucData = lpucData;


			//  Call int 13h behind our cache
			//
			uErrorCode = LBAReadWrite( fIsRead, ucBIOSDriveNumber, &LBAPacket,
					puSectorsRead );
		}
		else
		{
			//  Call int 13h behind our cache
			//
			uErrorCode = CHSReadWrite( fIsRead, ucBIOSDriveNumber, uCylinder,
					ucHead, ucSector, lpucData, uNumberOfSectors, puSectorsRead );
		}
	}
	else
#endif
	{
		//  Determine if the read past the end of disk needs to be clipped
		//
		ulTemp = Int13DriveInfo[ucDrive].ulEnd - ulStartingSector;
		if( ulTemp < uNumberOfSectors )
		{
			uNumberOfSectors = (unsigned)ulTemp;
		}
		else
		{
			ulTemp = uNumberOfSectors;
		}


		//  Call the Floppy/Hard disk driver or the CDEX/ATAPI driver
		//
		uErrorCode = DeviceReadWrite( fIsRead, ucDrive, ulStartingSector,
				uNumberOfSectors, lpucData, puSectorsRead );


		//  Adjust the number of sectors read to match the number of
		//  sectors requested so long as the it was equal to the adjusted
		//  requested amount
		//
		if( *puSectorsRead == (unsigned)ulTemp )
		{
			*puSectorsRead = uNumberOfSectors;
		}
	}


	return uErrorCode;
}



//  DEBUG only
//
#if 0

//
//
#pragma argsused
unsigned PASCAL DebugRead( bool fIsRead, unsigned ucDrive, ulong ulStartingSector,
		unsigned uNumberOfSectors, void far * lpucData,
		unsigned * puSectorsRead )
{
	uchar ucBIOSDriveNumber;
	unsigned uCylinder;
	uchar ucHead;
	uchar ucSector;
	bool fIsCHS;
	static LBAPACKET LBAPacket;
	unsigned uErrorCode;


	//  Determine if this is an Int 13 request call showing up here
	//
	if( ucDrive < ucFirstCDROMEntry )
	{
		//  Get the BIOS drive number for the drive
		//
		ucBIOSDriveNumber = Int13DriveInfo[ucDrive].ucBIOSDriveNumber;
		ulStartingSector += Int13DriveInfo[ucDrive].ulStart;


		//  Attempt to convert the sector number to CHS.  If it fails
		//  use LBA
		//
		fIsCHS = LBAToCHS( ucBIOSDriveNumber, ulStartingSector, &uCylinder,
				&ucHead, &ucSector );
		if( !fIsCHS )
		{
			//  Build an LBA request packet
			//
			memset( &LBAPacket, 0, sizeof(LBAPacket) );
			LBAPacket.ucPacketSize = 0x10;
			LBAPacket.uNumberOfSectors = uNumberOfSectors;
			LBAPacket.ulSectorLowDWORD = ulStartingSector;
			LBAPacket.lpucData = lpucData;


			//  Call int 13h behind our cache
			//
			uErrorCode = LBAReadWrite( fIsRead, ucBIOSDriveNumber, &LBAPacket,
					puSectorsRead );
		}
		else
		{
			//  Call int 13h behind our cache
			//
			uErrorCode = CHSReadWrite( fIsRead, ucBIOSDriveNumber, uCylinder,
					ucHead, ucSector, lpucData, uNumberOfSectors, puSectorsRead );
		}
	}
	else
	{
		//  Call the CDEX/ATAPI driver
		//
		uErrorCode = DeviceReadWrite( fIsRead, ucDrive, ulStartingSector,
				uNumberOfSectors, lpucData, puSectorsRead );
	}


	return uErrorCode;
}
#endif


//  Call the device driver
//
unsigned PASCAL DeviceReadWrite( bool fIsRead, uchar ucDrive, ulong ulStartingSector,
		unsigned uNumberOfSectors, uchar far * lpucData,
		unsigned * puSectorsRead )
{
	static DEVREQHDR Request;
	extern LPDEVREQHDR lpRequest;
	LPDEVREQHDR lpRequest2;
	unsigned uReturnStatus;
//unused	uchar ucIndex;


#if DEBUG_BAD_DRIVES
	if( ucDrive > ucDriveInfoEntries )
	{
		asm int 3
	}
#endif


	//  Initialize the default values in the request structure
	//
	memset( &Request, 0, sizeof(Request) );
	if( fIsRead )
	{
		_fmemcpy( &Request, lpRequest, sizeof(Request) );
	}
	lpRequest2 = &Request;


	//  Build the request structure
	//
	lpRequest2->unit = Int13DriveInfo[ucDrive].ucUnitNumber;
	if( ucDrive < ucFirstCDROMEntry )
	{
		lpRequest2->length = 0x1e;
#if 0
		lpRequest2->media = 0;
#else
		lpRequest2->media = Int13DriveInfo[ucDrive].lpMaxBPB->media;
#endif
		if( fIsRead )
		{
			lpRequest2->command = 4;
		}
		else
		{
			lpRequest2->command = 8;
		}
	}
	else
	{
		lpRequest2->length = 0x1d;
		if( fIsRead )
		{
			lpRequest2->command = 0x80;
		}
		else
		{
			lpRequest2->command = 0x86;
		}
		lpRequest2->media = 0;
	}
	lpRequest2->status = 0;
	lpRequest2->trans = lpucData;
	lpRequest2->count = uNumberOfSectors;
	if( ucDrive < ucFirstCDROMEntry )
	{
		lpRequest2->start = 0xffff;
		lpRequest2->start32 = ulStartingSector;
//		lpRequest2->drive = Int13DriveInfo[ucDrive].ucDOSDriveNumber;
	}
	else
	{
		*(ulong far *)(&lpRequest2->start) = ulStartingSector;
	}


#if DEBUG_CHECK_SECTOR_NUMBERS
	if( fIsRead )
	{
		ulong ulChkSectorNumber;
		ulong ulChkSectorNumber2;
		unsigned uSectorsInElement;
		unsigned uChkNumberOfSectors;


		uSectorsInElement = ulElementSize /
				Int13DriveInfo[ucDrive].uSectorSize;


 		uChkNumberOfSectors = lpRequest->count;
		if( uChkNumberOfSectors % uSectorsInElement )
		{
			uChkNumberOfSectors += uSectorsInElement -
					(uChkNumberOfSectors % uSectorsInElement);
		}

		if( ucDrive < ucFirstCDROMEntry )
		{
			ulChkSectorNumber  = lpRequest2->start32;
			ulChkSectorNumber2 = lpRequest->start32;
		}
		else
		{
			ulChkSectorNumber  = *(ulong far *)(&lpRequest2->start);
			ulChkSectorNumber2 = *(ulong far *)(&lpRequest->start);
		}
		ulChkSectorNumber2 -= ulChkSectorNumber2 % uSectorsInElement;


		while( uChkNumberOfSectors )
		{
			if( ulChkSectorNumber2 == ulChkSectorNumber )
			{
				break;
			}
 			ulChkSectorNumber2  += uSectorsInElement;
			uChkNumberOfSectors -= uSectorsInElement;
		}
		if( ulChkSectorNumber2 != ulChkSectorNumber )
		{
			//  Sector number doesn't match original request
			//
			asm int 3
		}
	}
#endif


#if DEBUG_WRITE_PROTECT
	//  For DEBUGGING we essentially write protected the disk
	//
	if( fIsRead )
	{
		//  Call the device
		//
		DevCall( lpRequest2, Int13DriveInfo[ucDrive].lpOriginalDevice );
		if( (lpRequest2->status & 0xff00) == 0x0100 )
		{
			lpRequest2->status = 0;
		}


		*puSectorsRead = lpRequest2->count;
	}
	else
	{
		lpRequest2->status = 0x8100;
		*puSectorsRead = 0;
	}
#else

	//  Call the device
	//
	DevCall( lpRequest2, Int13DriveInfo[ucDrive].lpOriginalDevice );
 	*puSectorsRead = lpRequest2->count;
	uReturnStatus = lpRequest2->status;
	if( (uReturnStatus & 0xff00) == 0x0100 )
	{
		uReturnStatus = 0;
	}

#endif


#if DEBUG_BAD_ERROR_CODES
		uGlobalErrorCode = uReturnStatus;
#endif

	return uReturnStatus;
}


//  Get the physical sector number for this CHS value
//
bool PASCAL CHSToLBA( uchar ucBIOSDriveNumber, unsigned uCylinder, uchar ucHead,
		uchar ucSector, ulong * pulPhysicalSector )
{
	uchar ucIndex;
	LPBPB lpBPB;


	//  Determine if this is a request for a drive that
	//  is being cached
	//
	for( ucIndex = 0; ucIndex < ucFirstCDROMEntry; ucIndex++ )
	{
		//  Determine if this is the drive number for the
		//  int 13h request
		//
		if( Int13DriveInfo[ucIndex].ucBIOSDriveNumber ==
				ucBIOSDriveNumber)
		{
			lpBPB = Int13DriveInfo[ucIndex].lpMaxBPB;
			break;
		}
	}
	if( ucIndex == ucFirstCDROMEntry )
	{
		return FALSE;
	}


	*pulPhysicalSector  = ((ulong)uCylinder * lpBPB->sides) + ucHead;
	*pulPhysicalSector *= lpBPB->sec_per_trak;
	*pulPhysicalSector += ucSector - 1;


	return TRUE;
}


#if 0
//  Get the CHS values for this physical sector number
//
bool PASCAL LBAToCHS( uchar ucBIOSDriveNumber, ulong ulPhysicalSector,
		unsigned * puCylinder, uchar * pucHead, uchar * pucSector )
{
	uchar ucIndex;
	LPBPB lpBPB;


	//  Determine if this is a request for a drive that
	//  is being cached
	//
	for( ucIndex = 0; ucIndex < ucFirstCDROMEntry; ucIndex++ )
	{
		//  Determine if this is the drive number for the
		//  int 13h request
		//
		if( Int13DriveInfo[ucIndex].ucBIOSDriveNumber ==
				ucBIOSDriveNumber)
		{
			lpBPB = Int13DriveInfo[ucIndex].lpMaxBPB;
			break;
		}
	}
	if( ucIndex == ucFirstCDROMEntry )
	{
		return FALSE;
	}


	*pucSector = (uchar)(ulPhysicalSector % lpBPB->sec_per_trak) + 1;
	ulPhysicalSector /= lpBPB->sec_per_trak;
	*pucHead = (uchar)(ulPhysicalSector % lpBPB->sides);
	ulPhysicalSector /= lpBPB->sides;
	if( ulPhysicalSector > 1023 )
	{
		return FALSE;
	}
	*puCylinder = (unsigned)ulPhysicalSector;


	return TRUE;
}
#endif


//  All int 28h functions get dispatched through here
//
void PASCAL _Int28Dispatch( void )
{
	FlushOneCacheElement();
}


//  All int 2fh functions get dispatched through here
//
void PASCAL _Int2fDispatch( void )
{
	uchar ucIndex;


	if( Int2fRegisters.ax == 0x1122 )
	{
		//  This is the process termination hook.  It is broadcast by
		//  DOS when a process terminates.  We will hook this vector
		//  so that when the "flush before returning to command" option
		//  is enabled, we can flush our buffers
		//
		//  Note:  This call must be passed down the int 2fh chain when
		//         we are done flushing our buffers
		//
		if( !fDontWriteBeforeCommand )
		{
			FlushAllBuffers();
		}
	}
	else if( Int2fRegisters.ax == 0x4a10 )
	{
		//  Go ahead and assume that the call has been handled
		//
		fHandledCall = TRUE;


		//  Determine the subfunction number
		//
		switch( Int2fRegisters.bx )
		{
			//  Int 2fh ax=4a10 bx=0000
			//
			//  Smartdrv installation check and general query
			//
			case 0x0000:
				{
					//  Serviced/valid flag
					//
					Int2fRegisters.ax = 0xbabe;


					//  Number of cache hits in DX:BX
					//
					Int2fRegisters.dx = HIGHWORD(ulCacheHits);
					Int2fRegisters.bx = LOWWORD(ulCacheHits);


					//  Number of cache misses in DI:SI
					//
					Int2fRegisters.di = HIGHWORD(ulCacheMisses);
					Int2fRegisters.si = LOWWORD(ulCacheMisses);


					//  Number of dirty elements in CX
					//
					//  NOTE: Returning nothing because we don't
					//        know how compatible our element
					//        implementation is
					//
					Int2fRegisters.cx = 0;


					//  Version number in BP
					//
					//  NOTE: MS compatible
					//
					Int2fRegisters.bp = 0x0501;
				}
				break;


#if COUNT_READS_AND_WRITES


			//  Int 2fh ax=4a10 bx=ffff (RD ONLY)
			//
			//  Smartdrv extended debug info.  Returns HD read/write
			//  request counts
			//
			case 0xffff:
				{
					//  Serviced/valid flag
					//
					Int2fRegisters.ax = 0xbabe;


					//  Number of cache hits in DX:BX
					//
					Int2fRegisters.dx = HIGHWORD(ulHDReadRequests);
					Int2fRegisters.bx = LOWWORD(ulHDReadRequests);


					//  Number of cache misses in DI:SI
					//
					Int2fRegisters.di = HIGHWORD(ulHDWriteRequests);
					Int2fRegisters.si = LOWWORD(ulHDWriteRequests);
				}
				break;


			//  Int 2fh ax=4a10 bx=fffe (RD ONLY)
			//
			//  Smartdrv extended debug info.  Returns CD read/write
			//  request counts
			//
			case 0xfffe:
				{
					//  Serviced/valid flag
					//
					Int2fRegisters.ax = 0xbabe;


					//  Number of cache hits in DX:BX
					//
					Int2fRegisters.dx = HIGHWORD(ulCDReadRequests);
					Int2fRegisters.bx = LOWWORD(ulCDReadRequests);


					//  Number of cache misses in DI:SI
					//
					Int2fRegisters.di = HIGHWORD(ulCDWriteRequests);
					Int2fRegisters.si = LOWWORD(ulCDWriteRequests);
				}
				break;


#endif // #if COUNT_READS_AND_WRITES


			//  Int 2fh ax=4a10 bx=0001
			//
			//  Smartdrv commit cache
			//
			case 0x0001:
				{
					FlushAllBuffers();
				}
				break;


			//  Int 2fh ax=4a10 bx=0002
			//
			//  Smartdrv restart cache
			//
			case 0x0002:
				{
					InvalidateAllBuffers();
				}
				break;


			//  Int 2fh ax=4a10 bx=0003
			//
			//  Smartdrv get/set cache status
			//
			case 0x0003:
				{
					//  Determine if this is a minor function number
					//  we recognize
					//
					if( LOWBYTE(Int2fRegisters.dx) <= 4  )
					{
						//  Determine the index into the drive info list
						//  for this DOS drive number
						//
						for( ucIndex = 0; ucIndex < ucDriveInfoEntries;
								ucIndex++ )
						{
							if( Int13DriveInfo[ucIndex].ucDOSDriveNumber ==
								  LOWBYTE(Int2fRegisters.bp) )
							{
								break;
							}
						}
						if( ucIndex == ucDriveInfoEntries )
						{
							//  This is not a drive number we recognize
							//
							Int2fRegisters.ax = 0xbabe;
							Int2fRegisters.dx |= 0xff;
							break;
						}


						//  Handle the requested action
						//
						Int2fRegisters.ax = 0xbabe;
						switch( LOWBYTE(Int2fRegisters.dx) )
						{
							//  Enable read cache
							//
							case 0x01:
								{
									Int13DriveInfo[ucIndex].fIsCached = TRUE;
								}
								break;


							//  Disable read cache
							//
							case 0x02:
								{
									FlushDiskCache(ucIndex);
									InvalidateDiskCache(ucIndex);
									Int13DriveInfo[ucIndex].fIsCached = FALSE;
								}
								break;


							//  Enable write cache
							//
							case 0x03:
								{
									Int13DriveInfo[ucIndex].fIsCached = TRUE;
									Int13DriveInfo[ucIndex].fIsWriteCached = TRUE;
								}
								break;


							//  Disable write cache
							//
							case 0x04:
								{
									FlushDiskCache(ucIndex);
									Int13DriveInfo[ucIndex].fIsWriteCached = FALSE;
								}
								break;


							//  Do nothing for 0x00
							//
							default:
								break;
						}


						//  Return the status for drive in dx
						//
						Int2fRegisters.dx &= 0xff00;
						Int2fRegisters.dx |=
								Int13DriveInfo[ucIndex].ucUnitNumber & 0x3f;
						if( !Int13DriveInfo[ucIndex].fIsCached )
						{
							Int2fRegisters.dx |= 0x80;
						}
						if( !Int13DriveInfo[ucIndex].fIsWriteCached )
						{
							Int2fRegisters.dx |= 0x40;
						}
					}
					else
					{
						//  This is not a function number we recognize
						//
						fHandledCall = FALSE;
					}
				}
				break;


			//  Int 2fh ax=4a10 bx=0004
			//
			//  Smartdrv get cache size
			//
			case 0x0004:
				{
					//  Number of elements in full cache in AX
					//
					Int2fRegisters.ax =
							(unsigned)((ulInitCacheSizeKB * 1024)/ulElementSize);


					//  Number of elements currently used in cache in BX
					//
					Int2fRegisters.bx = Int2fRegisters.ax;


					//  Size of element in bytes in CX
					//
					Int2fRegisters.cx = (unsigned)ulElementSize;


					//  Number of elements in cache when Windows
					//  is running in DX
					//
					Int2fRegisters.dx = 0;
				}
				break;


			//  Int 2fh ax=4a10 bx=0005
			//
			//  Smartdrv get double buffering status
			//
			case 0x0005:
				{
					//
					//  Do nothing when there is no double buffering
					//
				}
				break;


			//  Int 2fh ax=4a10 bx=0007
			//
			//  Smartdrv get original driver address
			//
			case 0x0007:
				{
					//  Determine the index into the drive info list
					//  for this DOS drive number
					//
					for( ucIndex = 0; ucIndex < ucDriveInfoEntries;
							ucIndex++ )
					{
						if( Int13DriveInfo[ucIndex].ucDOSDriveNumber ==
								(LOWBYTE(Int2fRegisters.bp) - 1) )
						{
							break;
						}
					}
					if( ucIndex == ucDriveInfoEntries )
					{
						//  This is not a drive number we recognize
						//
						Int2fRegisters.dx &= 0xff00;
						Int2fRegisters.dx |= 0x3f;
						break;
					}


					//
					//
					Int2fRegisters.dx &= 0xff00;
					Int2fRegisters.dx |= Int13DriveInfo[ucIndex].ucUnitNumber;
					Int2fRegisters.es =
							HIGHWORD(Int13DriveInfo[ucIndex].lpOriginalDevice);
					Int2fRegisters.di =
							LOWWORD(Int13DriveInfo[ucIndex].lpOriginalDevice);
				}
				break;


			//  Int 2fh ax=4a10 bx=0008
			//
			//  Smartdrv get/set command flushing mode
			//
			case 0x0008:
				{
					if( LOWBYTE(Int2fRegisters.dx) == 0x01  )
					{
						//  Get the command flushing mode
						//
						if( fDontWriteBeforeCommand )
						{
							Int2fRegisters.dx = 0x0001;
						}
						else
						{
							Int2fRegisters.dx = 0x0101;
						}
					}
					else if( Int2fRegisters.dx == 0x0002  )
					{
						fDontWriteBeforeCommand = TRUE;
					}
					else if( Int2fRegisters.dx == 0x0102  )
					{
						fDontWriteBeforeCommand = FALSE;
					}
					else
					{
						//  This is not a function number we recognize
						//
						fHandledCall = FALSE;
					}
				}
				break;


			//  Int 2fh ax=4a10 bx=0006
			//
			//  Smartdrv check drive cachability
			//
			case 0x0006:

				//
				//  Do nothing, this is mainly an API for other
				//  programs to use
				//
				//  Fall through to default to pass on call
				//


			//  Not our call
			//
			default:
				{
					//  Indicate the call has not been serviced so that
					//  it will be passed down the int 2fh chain
					//
					fHandledCall = FALSE;
				}
				break;
		}
	}
}


#if COMM_PORT_LOGGING

//  Send the device request information to the comm port
//  for logging purposes
//
void CommLog( char * szDeviceString, uchar ucDrive, ulong ulSector, unsigned uCount )
{
	static char szTemp[256];
	unsigned uIndex;


	sprintf( szTemp, "%s Drv: %xh Sec: %lxh Cnt: %xh\x0d\x0a", szDeviceString,
			(unsigned)ucDrive, ulSector, uCount );
	for( uIndex = 0; szTemp[uIndex]; uIndex++ )
	{
		comm_putc( szTemp[uIndex] );
	}
}

#endif
