//
//   $Id: MISERY.C 1.13 2003/02/18 15:57:34Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//  MISERY.C
//
//  Copyright (c) 2001-2002 Datalight, Inc.
//  All rights reserved.
//  Confidential and Proprietary
//
//  History:
//
//  ??-??-01  bt  created
//


//  Borland libraries
//
#include <stdio.h>
#include <mem.h>
#include <string.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "crc32.h"
#include "dlxms.h"


//  Cache includes
//
#include "lrucache.h"
#include "xmscache.h"


//  REMINDER
//
//  Need to change the data ptr size in this function
//
bool PASCAL dlGetCRC32( ulong * pulCRC, char far * lpucData,
		unsigned uSize );


void TestCommOutput( char * szInputFileName );
void exit(int);


//  Program definitions
//
#include "misery.h"
#include "cache.h"


//  Prototypes
//
#include "cache.pt"
#include "misery.pt"


//  Globals
//
DISKINFO MiseryDiskInfo[MAX_DRIVES];
//uchar ucBuffer[MAX_RW_SEC_COUNT * MAX_SECTOR_SIZE];
uchar far * lpucBuffer;
bool fValidationTestMode = FALSE;


ulong ulDiskReadHitCount = 0;
ulong ulDiskWriteHitCount = 0;
ulong ulElementSize = 8192;
ulong ulBufferSize  = 16384;
ulong ulInitCacheSizeKB = 0;


//  Globals which are referenced by the cache
//
LOL FakeLOL;
LPLOL lpLOL;
unsigned uSFTIndex = 0;
bool fIsFileRead = FALSE;


//  Globals used in building the FAKE LOL, DCBs, and SFTs
//
DCB FakeDCB;


//  Misery test program globals
//
unsigned uWriteCountEntries = 0;
unsigned uMaxWriteCountEntries;
LPWRITE_COUNTS lpWriteCounts;


//
//
uchar main( void )
{
	uchar ucDrive;
	bool fSuccessful;
	ulong ulPara;
	unsigned uSegment;
	extern unsigned uXmsNumberOfElements;


	printf("\"Misery\" cache tests running...\n");


	//  Allocate buffer space
	//
	ulPara = ((ulong)(MAX_RW_SEC_COUNT)) * MAX_SECTOR_SIZE;
	ulPara >>= 4;
	fSuccessful = dlAllocMem( (unsigned)ulPara, &uSegment );
	if( !fSuccessful )
	{
		printf("Insufficient memory for temp buffer\n");
		exit(1);
	}
	lpucBuffer = MK_FP( uSegment, 0 );


	//
	//
	DetermineCacheSize();
	fSuccessful = InitializeDiskCache(
			((ulInitCacheSizeKB * 1024)/ulElementSize),
			ulElementSize,	ulBufferSize );
	if( !fSuccessful )
	{
		printf("Not enough XMS\n");
		ErrorExit(1);
	}


	//  Allocate space for the write count array
	//
	ulPara  = (uXmsNumberOfElements * (ulElementSize / MIN_SECTOR_SIZE)) +
			MAX_RW_SEC_COUNT;
	uMaxWriteCountEntries = (unsigned)ulPara;
	ulPara *= sizeof(WRITE_COUNTS);
	ulPara >>= 4;
	fSuccessful = dlAllocMem( (unsigned)ulPara, &uSegment );
	if( !fSuccessful )
	{
		printf("Insufficient memory for write count array\n");
		exit(1);
	}
	lpWriteCounts = MK_FP( uSegment, 0 );


	//  Initialize the disk info
	//
	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Determine if this is the last drive in the disk info array
		//
		if( (ucDrive == MAX_DRIVES - 1) && (ucDrive != 0) )
		{
			//  Set this drive sector size to that of a CD
			//
			MiseryDiskInfo[ucDrive].uSectorSize = CD_SECTOR_SIZE;
		}
		else
		{
			//  Set this drive sector size to that of a hard drive
			//
			MiseryDiskInfo[ucDrive].uSectorSize = HD_SECTOR_SIZE;
		}


		//  Mark the disks as write cached
		//
		MiseryDiskInfo[ucDrive].fIsWriteCached = TRUE;


		//  Register each drive
		//
		RegisterDiskInformation( ucDrive, MiseryDiskInfo[ucDrive].uSectorSize, 0 );
	}


	//  Test this programs validation code
	//
	TestProgramValidation();


	//  Create a FAKE LOL structure and initialize the pointer to
	//  the FAKE LOL
	//
	FakeLOL.first_dcb = &FakeDCB;
	FakeDCB.numsft = 1;
	FakeDCB.sft[0].size = 1000000L;
	FakeDCB.sft[0].offset = 0;
	fIsFileRead = TRUE;
	lpLOL = &FakeLOL;


#if 1
	//  Test invalidating the cache
	//
	TestSet1();
	TestSet5();


	//  Test reading sectors sequentially on the disk
	//
	TestSet2();


	//  Test reading recently cached sectors
	//
	TestSet3();


	//  Test reading sectors out of order and multiple sector
	//  requests crossing multiple elements
	//
	TestSet4();


	//  Test part of a real world case
	//
	TestSet6();
#endif


#if 1
	//  Test a real world case as it was recorded by the COMM
	//  output feature of our SMARTDRV
	//
	TestCommOutput( "COMMOUT1.DAT" );
#endif


#if 1
	//  Exercise the cache code paths
	//
	ExersiseCodePaths();
#endif


	//  Return that the test was successful
	//
	printf("Misery tests  PASSED\n");
	return 0;
}


//  Exercise the cache code paths
//
void ExersiseCodePaths( void )
{
	uchar ucDrive = 1;
	bool fIsRead;
	unsigned uSectorsInEachElement;
	unsigned uElementsInLocalBuffer;
	unsigned uSectorsInLocalBuffer;
	unsigned uSectorCount;
	unsigned uSectorIndexIntoElement;
	unsigned uSectorBitIndex;
	ulong ulSector;
	ulong ulStartingSector;
	ulong ulPossibleElementWriteStates;
	ulong ulElementWriteState;


	uSectorsInEachElement = (unsigned)(ulElementSize /
			MiseryDiskInfo[ucDrive].uSectorSize);
	uElementsInLocalBuffer = (unsigned)(ulBufferSize / ulElementSize);
	uSectorsInLocalBuffer = uElementsInLocalBuffer * uSectorsInEachElement;


	//  Test all the conditions of looking for an element in the cache
	//
	{
		//  Read condition where the sectors are not in the cache
		//
		ulStartingSector = 0;
		fIsRead = TRUE;
		uSectorCount = 1;
		for( uSectorIndexIntoElement = 0;
				uSectorIndexIntoElement < uSectorsInEachElement;
				uSectorIndexIntoElement++ )
		{
			ulSector = ulStartingSector + uSectorIndexIntoElement;
			TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, TRUE );
			ulStartingSector += uSectorsInLocalBuffer;
		}


		//  Read condition where the sectors are in the cache from
		//  previous reads
		//
		ulStartingSector = 0;
		fIsRead = TRUE;
		uSectorCount = uSectorsInLocalBuffer;
		for( uSectorIndexIntoElement = 0;
				uSectorIndexIntoElement < uSectorsInEachElement;
				uSectorIndexIntoElement++ )
		{
			ulSector = ulStartingSector;
			TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, TRUE );
			ulStartingSector += uSectorsInLocalBuffer;
		}


		//  Write condition where the sectors are in the cache from
		//  prior reads
		//
		ulStartingSector = 0;
		fIsRead = FALSE;
		uSectorCount = 1;
		for( uSectorIndexIntoElement = 0;
				uSectorIndexIntoElement < uSectorsInEachElement;
				uSectorIndexIntoElement++ )
		{
			ulSector = ulStartingSector + uSectorIndexIntoElement;
			TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, TRUE );
			ulStartingSector += uSectorsInLocalBuffer;
		}


		//  Make sure all dirty sectors get committed
		//
		FlushDiskCache( ucDrive );
		ValidateUncommittedSectors();


		//  Invalidate all the information in the cache
		//
		InvalidateDiskCache( ucDrive );
		InvalidateWriteCountArray();


		//  Write condition where the sectors are not in the cache
		//
		ulStartingSector = 0;
		fIsRead = FALSE;
		uSectorCount = 1;
		for( uSectorIndexIntoElement = 0;
				uSectorIndexIntoElement < uSectorsInEachElement;
				uSectorIndexIntoElement++ )
		{
			ulSector = ulStartingSector + uSectorIndexIntoElement;
			TestSectors( WRITE, ucDrive, ulSector, uSectorCount, TRUE );
			ulStartingSector += uSectorsInLocalBuffer;
		}


		//  Read condition where the sectors are in the cache
		//  from a prior write
		//
		ulStartingSector = 0;
		fIsRead = TRUE;
		uSectorCount = 1;
		for( uSectorIndexIntoElement = 0;
				uSectorIndexIntoElement < uSectorsInEachElement;
				uSectorIndexIntoElement++ )
		{
			ulSector = ulStartingSector + uSectorIndexIntoElement;
			TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, TRUE );
			ulStartingSector += uSectorsInLocalBuffer;
		}


		//  Write condition where the sectors are in the cache from
		//  prior writes
		//
		ulStartingSector = 0;
		fIsRead = FALSE;
		uSectorCount = 1;
		for( uSectorIndexIntoElement = 0;
				uSectorIndexIntoElement < uSectorsInEachElement;
				uSectorIndexIntoElement++ )
		{
			ulSector = ulStartingSector + uSectorIndexIntoElement;
			TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, TRUE );
			ulStartingSector += uSectorsInLocalBuffer;
		}


		//  Make sure all dirty sectors get committed
		//
		FlushDiskCache( ucDrive );
		ValidateUncommittedSectors();


		//  Invalidate all the information in the cache
		//
		InvalidateDiskCache( ucDrive );
		InvalidateWriteCountArray();


		//  Read condition where some of the sectors are in the cache
		//  from prior writes
		//
		//  This will test every possibility for having a write
		//  to one or more sectors in an element and a read following it
		//
		ulStartingSector = 0;
		fIsRead = TRUE;
		uSectorCount = uSectorsInEachElement;
		ulPossibleElementWriteStates = 2L << uSectorsInEachElement;
		for( ulElementWriteState = 1;
				ulElementWriteState < ulPossibleElementWriteStates;
				ulElementWriteState++ )
		{
//			printf("Element Sector = %lx\n", ulStartingSector );
//			printf("Element State  = %02lx\n", ulElementWriteState );
			uSectorBitIndex = 1;
			for( uSectorIndexIntoElement = 0;
					uSectorIndexIntoElement < uSectorsInEachElement;
					uSectorIndexIntoElement++ )
			{
				if( (unsigned)ulElementWriteState & uSectorBitIndex )
				{
 					ulSector = ulStartingSector + uSectorIndexIntoElement;
					TestSectors( WRITE, ucDrive, ulSector, 1, TRUE );
				}
				uSectorBitIndex <<= 1;
			}

			ulSector = ulStartingSector;
			TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, TRUE );
			ulStartingSector += uSectorsInLocalBuffer;
		}
	}


	printf("Exercise code path testing  PASSED\n");
}


//  Check the internal cache index tables
//
void CheckInternalCacheArrays( bool fIsRead, uchar ucDrive, ulong ulSector,
		unsigned uSectorCount, bool fWriteCached )
{
	extern LRU_ENTRY far * lpuCacheArray;
	extern unsigned far * lpruDiskSectorMap;
	extern DISK_ENTRY far * lprdeXmsCacheMap;
	extern unsigned uNumberOfDiskEntries;
	extern unsigned uXmsNumberOfElements;
	unsigned uIndex;
	unsigned uCheckIndex;
	unsigned uMissingIndexes;
	bool fEntryFound;
	bool fEntryValid;
	DISK_ENTRY DiskEntry;
	DISK_ENTRY LastDiskEntry;


	//  Check limits
	//
	if( uNumberOfDiskEntries > uXmsNumberOfElements )
	{
		//  Error, the number of entries in the array exceed the
		//  number of elements possible in the arrays
		//
		printf("Error uNumberOfDiskEntries > uXmsNumberOfElements\n");
		exit(4);
	}


	//  Check the LRU array for duplicate or missing entries
	//
	for( uIndex = 0; uIndex < uXmsNumberOfElements; uIndex++ )
	{
		//  Find this index in the LRU array
		//
		fEntryFound = FALSE;
		for( uCheckIndex = 0; uCheckIndex < uXmsNumberOfElements;
				uCheckIndex++ )
		{
			if( lpuCacheArray[uCheckIndex] == uIndex )
			{
				if( fEntryFound )
				{
					//  Duplicate indexes in array
					//
					printf("Duplicate indexes in LRU array\n");
					exit(4);
				}
				fEntryFound = TRUE;
			}
		}
		if( !fEntryFound )
		{
			//  Missing index
			//
			printf("Missing index in LRU array\n");
			exit(4);
		}
	}


	//  Check the sorted array for duplicate or missing entries
	//
	uMissingIndexes = 0;
	for( uIndex = 0; uIndex < uXmsNumberOfElements; uIndex++ )
	{
		//  Determine whether this entry is valid
		//
		DiskEntry = lprdeXmsCacheMap[uIndex];
		fEntryValid = ((DiskEntry & DISK_STATUS_VALID) == DISK_STATUS_VALID);


		//  Find this index in the sorted array
		//
		fEntryFound = FALSE;
		for( uCheckIndex = 0; uCheckIndex < uNumberOfDiskEntries;
				uCheckIndex++ )
		{
			if( lpruDiskSectorMap[uCheckIndex] == uIndex )
			{
				if( fEntryFound )
				{
					//  Duplicate indexes in array
					//
					printf("Duplicate indexes in sorted array\n");
					exit(4);
				}
				fEntryFound = TRUE;
				if( !fEntryValid )
				{
					printf("Used disk entry not marked as valid\n");
					exit(4);
				}
			}
		}
		if( !fEntryFound )
		{
			if( fEntryValid )
			{
				printf("Warning: Unused disk entry marked valid\n");
			}
			uMissingIndexes++;
		}
	}


	//  Determine if the missing entry count is what we expected
	//
	if( uMissingIndexes != (uXmsNumberOfElements - uNumberOfDiskEntries) )
	{
		printf("uMissingIndexes != (uXmsNumberOfElements - uNumberOfDiskEntries)\n");
		exit(4);
	}


	//  Check and make sure that the sorted array is sorted
	//
	LastDiskEntry = lprdeXmsCacheMap[lpruDiskSectorMap[0]] &
			DISK_NUM_ELEMENT_MASK;
	for( uIndex = 1; uIndex < uNumberOfDiskEntries; uIndex++ )
	{
		//  Get the value of this entry
		//
		DiskEntry = lprdeXmsCacheMap[lpruDiskSectorMap[uIndex]] &
				DISK_NUM_ELEMENT_MASK;
		if( LastDiskEntry >= DiskEntry )
		{
 			LastDiskEntry = lprdeXmsCacheMap[lpruDiskSectorMap[uIndex-1]];
 			DiskEntry = lprdeXmsCacheMap[lpruDiskSectorMap[uIndex]];
			printf("Sorted array not sorted!\n\n");
			printf("Previous entry: %08lxh  lpruDiskSectorMap index: %xh  DE index: %xh\n",
					LastDiskEntry,	uIndex - 1, lpruDiskSectorMap[uIndex-1] );
			printf("Current entry:  %08lxh  lpruDiskSectorMap index: %xh  DE index: %xh\n\n",
					DiskEntry, uIndex, lpruDiskSectorMap[uIndex] );

			printf("uXmsNumberOfElements: %xh\n", uXmsNumberOfElements );
			printf("uNumberOfDiskEntries: %xh\n", uNumberOfDiskEntries );
			printf("\n");
			printf("Cache request info\n\n");
			if( fIsRead )
			{
				printf("fIsRead: TRUE\n");
			}
			else
			{
				printf("fIsRead: FALSE\n");
			}
			printf("ucDrive: %xh\n", (unsigned)ucDrive );
			printf("ulSector: %lxh\n", ulSector );
			printf("uSectorCount: %x\n", uSectorCount );
			if( fWriteCached )
			{
				printf("fWriteCached: TRUE\n");
			}
			else
			{
				printf("fWriteCached: FALSE\n");
			}
			exit(4);
		}
		LastDiskEntry = DiskEntry;
	}
}


//  Test the code that validates sector data
//
void PASCAL TestProgramValidation( void )
{
	unsigned uIndex;
	bool fTestPassed = TRUE;
	unsigned uTestFailureNumber = 0;


	//  Set a flag so that other validation tests to
	//  not terminate this test function
	//
	fValidationTestMode = TRUE;


	//  Build a sector, make sure it is valid, and change every
	//  byte in the sector checking each time that it was not
	//  valid
	//
	if( fTestPassed )
	{
		SerializeSector( READ, 0, 0, &lpucBuffer[0]);
 		if( !ValidateSerializeSector(0, 0, &lpucBuffer[0]) )
		{
			fTestPassed = FALSE;
			if( !uTestFailureNumber )
			{
				uTestFailureNumber = 1;
			}
		}


		for( uIndex = 0; uIndex < MiseryDiskInfo[0].uSectorSize; uIndex++ )
		{
			lpucBuffer[uIndex] = lpucBuffer[uIndex] + 1;
	 		if( ValidateSerializeSector(0, 0, &lpucBuffer[0]) )
			{
				fTestPassed = FALSE;
				if( !uTestFailureNumber )
				{
					uTestFailureNumber = 2;
				}
			}
			lpucBuffer[uIndex] = lpucBuffer[uIndex] - 1;
	 		if( !ValidateSerializeSector(0, 0, &lpucBuffer[0]) )
			{
				fTestPassed = FALSE;
				if( !uTestFailureNumber )
				{
					uTestFailureNumber = 3;
				}
			}
		}
	}


	//  Determine if the tests passed
	//
	if( !fTestPassed )
	{
		printf("\nFAILURE!  TestProgramValidation\n");
		printf("  Test %u Failed\n", uTestFailureNumber );
		ErrorExit(1);
	}


	//  Clear a flag so that other validation tests to
	//  not terminate this test function
	//
	//  NOTE: Removing this line from this function would
	//        not be in your best intrest.  See Brandon
	//
	fValidationTestMode = FALSE;
}


//  Read or write a sector
//
unsigned PASCAL DiskReadWrite( bool fIsRead, unsigned uDrive,
		ulong ulStartingSector, unsigned uNumberOfSectors,
		void far * lpucData, unsigned * puSectorsRead )
{
	unsigned uErrorCode;


	//  Determine if this is a read or write
	//
	if( fIsRead )
	{
		//  Read the sectors
		//
		uErrorCode = FakeReadSector( (uchar)uDrive, uNumberOfSectors,
				ulStartingSector, lpucData, puSectorsRead );
	}
	else
	{
		//  Write the sectors
		//
		uErrorCode = FakeWriteSector( (uchar)uDrive, uNumberOfSectors,
				ulStartingSector, lpucData, puSectorsRead );
	}


	//  Return the error code
	//
	return uErrorCode;
}


//
//
unsigned PASCAL FakeReadSector( uchar ucDrive,
		unsigned uNumberOfSectors,	ulong ulStartingSector,
		uchar far * lpucData, unsigned * puSectorsRead )
{
	unsigned uSectorCount;
	ulong ulSector;


	//  Increment the read hit counter
	//
	ulDiskReadHitCount++;


	//  Get the sector information for each of the requested
	//  sectors
	//
	*puSectorsRead = 0;
	ulSector = ulStartingSector;
	for( uSectorCount = 0; uSectorCount < uNumberOfSectors;
			uSectorCount++ )
	{
		//  Build a sector which is serialized with the drive number,
		//  sector number, a write number, and 32-bit CRC of the data.
		//
		SerializeSector( READ, ucDrive, ulSector, lpucData );


		//  Set the data address for the next sector
		//
		lpucData += MiseryDiskInfo[ucDrive].uSectorSize;


		//  Increment the sector number
		//
		ulSector++;


		//  Increment the number of sectors read
		//
		(*puSectorsRead)++;
	}


	//
	//
	return 0;
}


//  Build a sector which is serialized with the drive number,
//  sector number, a write number, and 32-bit CRC of the data.
//
void PASCAL SerializeSectors( bool fIsRead, uchar ucDrive,
		ulong ulSectorNumber, unsigned uCount, uchar far * lpucData )
{
	unsigned uIndex;


	for( uIndex = 0; uIndex < uCount; uIndex++ )
	{
		SerializeSector( fIsRead, ucDrive, ulSectorNumber + uIndex,
				&lpucData[uIndex * MiseryDiskInfo[ucDrive].uSectorSize] );
	}
}


//  Build a sector which is serialized with the drive number,
//  sector number, a write number, and 32-bit CRC of the data.
//
void PASCAL SerializeSector( bool fIsRead, uchar ucDrive,
		ulong ulSectorNumber, uchar far * lpucData )
{
	unsigned uByteIndex;
	LPSECTORINFO lpSectorInfo;
	unsigned uBytesToWriteInSector;
	static uchar ucByteData;
	ulong ul32BitCRC = 0;


	//  Serialize each byte in this sector
	//
	uBytesToWriteInSector = MiseryDiskInfo[ucDrive].uSectorSize -
			sizeof(SECTORINFO);
	for( uByteIndex = 0; uByteIndex < uBytesToWriteInSector;
			uByteIndex++ )
	{
		ucByteData += PRIME_VALUE;
		lpucData[uByteIndex] = ucByteData;
	}


	//  Get the address of the sector info structure in the sector
	//
	lpSectorInfo = (LPSECTORINFO)(lpucData + MiseryDiskInfo[ucDrive].uSectorSize -
			sizeof(SECTORINFO));


	//  Update all the sector info except for the 32-bit bit sector CRC
	//
	lpSectorInfo->ulSectorNumber = ulSectorNumber;
	if( fIsRead )
	{
		lpSectorInfo->uWriteCount = 0;	// Just zero for now
	}
	else
	{
		lpSectorInfo->uWriteCount = IncrementWriteCount(ucDrive,
				ulSectorNumber);
	}
	lpSectorInfo->ucDrive = ucDrive;


	//  Calculate the 32-bit for every byte in the sector except for
	//  the last couble bytes where the 32-bit CRC is stored
	//
	dlGetCRC32( &ul32BitCRC, lpucData, MiseryDiskInfo[ucDrive].uSectorSize -
			sizeof(lpSectorInfo->ul32BitCRC) );


	//  Store the 32-bit CRC in the sector info structure
	//
	lpSectorInfo->ul32BitCRC = ul32BitCRC;
}


//  Validate a mumber of sectors which are serialized with the drive number,
//  sector number, a write number, and 32-bit CRC of the data.
//
bool PASCAL ValidateSerializeSectors( uchar ucDrive, ulong ulSectorNumber,
		unsigned uSectorCount, uchar far * lpucData )
{
	unsigned uCount;
	bool fPassed;


	//  Validate each sector
	//
	for( uCount = 0; uCount < uSectorCount; uCount++ )
	{
		//  Validate this sector
		//
		fPassed = ValidateSerializeSector( ucDrive, ulSectorNumber +
				uCount, lpucData );
		if( !fPassed )
		{
			return fPassed;
		}


		//  Set the data address for the next sector
		//
		lpucData += MiseryDiskInfo[ucDrive].uSectorSize;
	}


	//  Return the results
	//
	return fPassed;
}


//  Validate a sector which is serialized with the drive number,
//  sector number, a write number, and 32-bit CRC of the data.
//
bool PASCAL ValidateSerializeSector( uchar ucDrive, ulong ulSectorNumber,
		uchar far * lpucData )
{
	LPSECTORINFO lpSectorInfo;
	ulong ul32BitCRC = 0;
	bool  fIsValid = TRUE;
	unsigned uWriteCount;


	//  Get the address of the sector info structure in the sector
	//
	lpSectorInfo = (LPSECTORINFO)(lpucData + MiseryDiskInfo[ucDrive].uSectorSize -
			sizeof(SECTORINFO));


	//
	//
	dlGetCRC32( &ul32BitCRC, lpucData, MiseryDiskInfo[ucDrive].uSectorSize -
			sizeof(lpSectorInfo->ul32BitCRC) );
	if( ul32BitCRC != lpSectorInfo->ul32BitCRC )
	{
		fIsValid = FALSE;
	}

	if( lpSectorInfo->ucDrive != ucDrive )
	{
		fIsValid = FALSE;
	}


	if( lpSectorInfo->ulSectorNumber != ulSectorNumber )
	{
		fIsValid = FALSE;
	}


	//  Determine the "Write Count" for this sector
	//
	uWriteCount = GetSectorWriteCount( ucDrive, ulSectorNumber );
	if( lpSectorInfo->uWriteCount != uWriteCount )
	{
		fIsValid = FALSE;
	}


	//  Display a message if the sector was not valid
	//
	if( (!fIsValid) && (!fValidationTestMode) )
	{
		printf("\nFAILURE!  ValidateSerializeSector\n");
		printf("\n");
		printf("             Expected   Recieved\n");
		printf("             --------   --------\n");
		printf("     Drive:  %02xh        %02xh\n", (unsigned)(ucDrive),
				(unsigned)(lpSectorInfo->ucDrive) );
		printf("    Sector:  %08lxh  %08lxh\n", ulSectorNumber,
				lpSectorInfo->ulSectorNumber );
		printf("       CRC:  %08lxh  %08lxh\n", ul32BitCRC,
				lpSectorInfo->ul32BitCRC );
		printf("  WriteNum:  %04xh      %04xh\n", uWriteCount,
				lpSectorInfo->uWriteCount );
		ErrorExit(1);
	}


	return fIsValid;
}


//  Determine if this sector has been written multiple times in order
//  to determine the "write count" for this sector.  Unwritten sectors
//  should return a "write count" of zero.
//
unsigned GetSectorWriteCount( uchar ucDrive, ulong ulSectorNumber )
{
	unsigned uWriteCount = 0;
	unsigned uIndex;


	//  Determine if this sector exists in the write count array
	//
	for( uIndex = 0; uIndex < uWriteCountEntries; uIndex++ )
	{
		if( lpWriteCounts[uIndex].ulSectorNumber == ulSectorNumber )
		{
			if( lpWriteCounts[uIndex].ucDrive == ucDrive )
			{
				uWriteCount = lpWriteCounts[uIndex].uWriteCount;
				break;
			}
		}
	}


	return uWriteCount;
}


//  Increment the "write count" for a specific sector in the write
//  count array (adding it if necessary) and return what the write
//  count is.
//
unsigned IncrementWriteCount( uchar ucDrive, ulong ulSectorNumber )
{
	unsigned uWriteCount = 0;
	unsigned uIndex;
	bool fFound = FALSE;


	//  Determine if this sector exists in the write count array
	//
	for( uIndex = 0; uIndex < uWriteCountEntries; uIndex++ )
	{
		if( lpWriteCounts[uIndex].ulSectorNumber == ulSectorNumber )
		{
			if( lpWriteCounts[uIndex].ucDrive == ucDrive )
			{
				fFound = TRUE;
				break;
			}
		}
	}


	//  if the entry was not found ...
	//
	if( !fFound )
	{
		//  Determine if there is room to add it to the array
		//
		if( (uWriteCountEntries + 1) > uMaxWriteCountEntries )
		{
			//  Error, the write count array is not large enough
			//  to hold all the sector write information
			//
			printf("Insufficient room in the write count array\n");
			ErrorExit(1);
		}


		//  Increment the number of entries in the write count array
		//
		uWriteCountEntries++;


		//  Add this entry to the array
		//
		uIndex = uWriteCountEntries - 1;
		lpWriteCounts[uIndex].ucDrive = ucDrive;
		lpWriteCounts[uIndex].ulSectorNumber = ulSectorNumber;
		lpWriteCounts[uIndex].uWriteCount = uWriteCount;
	}


	//  Increment the write count and indicate that the sector
	//  has not been committed
	//
	lpWriteCounts[uIndex].fCommitted = FALSE;
	lpWriteCounts[uIndex].uWriteCount++;
	uWriteCount = lpWriteCounts[uIndex].uWriteCount;


	return uWriteCount;
}


//  Mark this dirty sector as committed
//
void MarkWriteCountEntryCommitted( uchar ucDrive, ulong ulSectorNumber )
{
	unsigned uIndex;
	bool fFound = FALSE;


	//  Determine if this sector exists in the write count array
	//
	for( uIndex = 0; uIndex < uWriteCountEntries; uIndex++ )
	{
		if( lpWriteCounts[uIndex].ulSectorNumber == ulSectorNumber )
		{
			if( lpWriteCounts[uIndex].ucDrive == ucDrive )
			{
				fFound = TRUE;
				break;
			}
		}
	}


	//  if the entry was not found ...
	//
	if( !fFound )
	{
		//  Error, the entry was not found in the write count array
		//
		printf("Error marking write count entry committed.  Entry not found.\n");
		ErrorExit(1);
	}


	//  Make sure the sector was dirty since it is being committed to disk
	//
	if( lpWriteCounts[uIndex].fCommitted != FALSE )
	{
		printf("\nFAILURE!  Clean sector committed to disk\n");
		printf("\n");
		printf("     Drive:  %02xh\n",
				(unsigned)(lpWriteCounts[uIndex].ucDrive) );
		printf("    Sector:  %08lxh\n", lpWriteCounts[uIndex].ulSectorNumber );
		printf("  WriteNum:  %04xh\n", lpWriteCounts[uIndex].uWriteCount );
		ErrorExit(1);
	}


	//  Indicate that this entry has been committed
	//
	lpWriteCounts[uIndex].fCommitted = TRUE;
}


//  Reinitialize the write count array
//
void InvalidateWriteCountArray( void )
{
	uWriteCountEntries = 0;
}


//  Remove an entry from the write count array.  This should only
//  occure once an entry has been flushed from the cache and is
//  being removed from the cache
//
void RemoveWriteCountEntry( uchar ucDrive, ulong ulSectorNumber )
{
	unsigned uIndex;
	bool fFound = FALSE;


	//  Determine if this sector exists in the write count array
	//
	for( uIndex = 0; uIndex < uWriteCountEntries; uIndex++ )
	{
		if( lpWriteCounts[uIndex].ulSectorNumber == ulSectorNumber )
		{
			if( lpWriteCounts[uIndex].ucDrive == ucDrive )
			{
				fFound = TRUE;
				break;
			}
		}
	}


	//  if the entry was not found ...
	//
	if( !fFound )
	{
//  Since this will be called from the cache directly, it will be
//  called for valid but potentially unwritten sectors so we can
//  not just error here.
//
#if 0
		//  Error, the entry was not found in the write count array
		//
		printf("Error removing write count entry.  Entry not found.\n");
		ErrorExit(1);
#else
		return;
#endif
	}


	//  Make sure the dirty sector was committed to disk
	//
	if( lpWriteCounts[uIndex].fCommitted != TRUE )
	{
		printf("\nFAILURE!  Dirty sector removed from cache\n");
		printf("\n");
		printf("     Drive:  %02xh\n",
				(unsigned)(lpWriteCounts[uIndex].ucDrive) );
		printf("    Sector:  %08lxh\n", lpWriteCounts[uIndex].ulSectorNumber );
		printf("  WriteNum:  %04xh\n", lpWriteCounts[uIndex].uWriteCount );
		ErrorExit(1);
	}


	//  Remove this entry by shifting all the other entries down and
	//  reducing the number of entries in the array
	//
	_fmemmove( &(lpWriteCounts[uIndex]), &(lpWriteCounts[uIndex+1]),
			((uWriteCountEntries - 1) - uIndex) * sizeof(WRITE_COUNTS) );
	uWriteCountEntries--;
}


//  Make sure that all dirty sectors were committed to disk
//
void ValidateUncommittedSectors( void )
{
	unsigned uIndex;


	//  Determine if this sector exists in the write count array
	//
	for( uIndex = 0; uIndex < uWriteCountEntries; uIndex++ )
	{
		if( !lpWriteCounts[uIndex].fCommitted )
		{
			printf("\nFAILURE!  Dirty sector not committed to disk\n");
			printf("\n");
			printf("     Drive:  %02xh\n",
					(unsigned)(lpWriteCounts[uIndex].ucDrive) );
			printf("    Sector:  %08lxh\n",
					lpWriteCounts[uIndex].ulSectorNumber );
			printf("  WriteNum:  %04xh\n", lpWriteCounts[uIndex].uWriteCount );
			ErrorExit(1);
		}
	}
}


//
//
unsigned PASCAL FakeWriteSector( uchar ucDrive,
		unsigned uNumberOfSectors,	ulong ulStartingSector,
		uchar far * lpucData, unsigned * puSectorsRead )
{
#if 0
	LPSECTORINFO lpSectorInfo;
#endif
	ulong ulSector;
	unsigned uSectorCount;


	//  Increment the write hit counter
	//
	ulDiskWriteHitCount++;


	//  Get the sector information for each of the requested
	//  sectors
	//
	*puSectorsRead = 0;
	ulSector = ulStartingSector;
	for( uSectorCount = 0; uSectorCount < uNumberOfSectors;
			uSectorCount++ )
	{
		//  Build a sector which is serialized with the drive number,
		//  sector number, a write number, and 32-bit CRC of the data.
		//
		ValidateSerializeSector( ucDrive, ulSector, lpucData );


		//  Update that this write count entry has been committed
		//
		MarkWriteCountEntryCommitted( ucDrive, ulSector );


#if 0
		//  Get the address of the sector info structure in the sector
		//
		lpSectorInfo = (LPSECTORINFO)(lpucData +
				MiseryDiskInfo[ucDrive].uSectorSize - sizeof(SECTORINFO));


		//  Determine if this sectors write count was changed
		//
		if( lpSectorInfo->uWriteCount == 0 )
		{
			printf("\nFAILURE!  FakeWriteSector\n");
			printf("\n");
			printf("     Drive:  %02xh\n", (unsigned)(ucDrive) );
			printf("    Sector:  %08lxh\n", ulSectorNumber );
			printf("  WriteNum:  %04xh\n", lpSectorInfo->uWriteCount );
			ErrorExit(1);
		}
#endif

		//  Set the data address for the next sector
		//
		lpucData += MiseryDiskInfo[ucDrive].uSectorSize;


		//  Increment the sector number
		//
		ulSector++;


		//  Increment the number of sectors read
		//
		(*puSectorsRead)++;
	}


	return 0;
}


void TestSectors( bool fIsRead, uchar ucDrive, ulong ulSector,
		unsigned uSectorCount, bool fWriteCached )
{
	unsigned uErrorCode;
	unsigned uSectorsRead;


	if( !fIsRead )
	{
		SerializeSectors( WRITE, ucDrive, ulSector, uSectorCount,
				&lpucBuffer[0] );
		ValidateSerializeSectors( ucDrive, ulSector, uSectorCount,
				&lpucBuffer[0] );
	}
	else
	{
		_fmemset( &lpucBuffer[0], 0,
				uSectorCount * MiseryDiskInfo[ucDrive].uSectorSize );
	}


	uErrorCode = CacheReadWrite( fIsRead, ucDrive, ulSector, uSectorCount,
			&lpucBuffer[0], fWriteCached, &uSectorsRead );
	if( uErrorCode )
	{
		printf( "Unexpected error code %xh\n", uErrorCode );
		if( fIsRead )
		{
			printf("fIsRead: TRUE\n");
		}
		else
		{
			printf("fIsRead: FALSE\n");
		}
		printf("ucDrive: %xh\n", (unsigned)ucDrive );
		printf("ulSector: %lxh\n", ulSector );
		printf("uSectorCount: %x\n", uSectorCount );
		if( fWriteCached )
		{
			printf("fWriteCached: TRUE\n");
		}
		else
		{
			printf("fWriteCached: FALSE\n");
		}
		printf("uSectorsRead: %x\n", uSectorsRead );
		exit(4);
	}
	if( uSectorsRead != uSectorCount )
	{
		printf( "Didn't read/write expected sectors\n" );
		exit(4);
	}
	ValidateSerializeSectors(ucDrive, ulSector, uSectorCount, &lpucBuffer[0]);


	//  Check the internal cache index tables
	//
	CheckInternalCacheArrays( fIsRead, ucDrive, ulSector, uSectorCount,
			fWriteCached );
}


//  Determine the total size of the cache, the buffer size, and the
//  element size
//
void PASCAL DetermineCacheSize( void )
{
	int iErrorCode;
	unsigned uIndex;
	unsigned uNumberOfElements;
	unsigned uXMSLargestFreeK;
	unsigned uXMSTotalFreeK;
	static unsigned uValidElementSizes[] = {
		1024, 2048, 4096, 8192,
	};


	//  Determine if the element size is a valid value.  If not, round
	//  up to the nearest valid value
	//
	for( uIndex = 0; uIndex < (ARRAY_ENTRIES(uValidElementSizes) - 1); uIndex++ )
	{
		if( ulElementSize <= (ulong)(uValidElementSizes[uIndex]) )
		{
			break;
		}
	}
	ulElementSize = uValidElementSizes[uIndex];


	//  Determine if the buffer size is a valid value.  If it is not
	//  a multiple of the element size, then round down.
	//
	ulBufferSize  = (ulBufferSize / ulElementSize);
	if( ulBufferSize == 0 )
	{
		ulBufferSize++;
	}
	ulBufferSize *= ulElementSize;


	//  Determine the amount of XMS in KB
	//
	iErrorCode = CheckXMMInstalled();
	if( iErrorCode == XMS_SUCCESS )
	{
		iErrorCode = GetXMSFree( &uXMSLargestFreeK, &uXMSTotalFreeK );
		if( uXMSLargestFreeK < MIN_XMS_KB )
		{
			printf("Not enough XMS\n");
			ErrorExit(1);
		}
	}
	if( iErrorCode != XMS_SUCCESS )
	{
		printf( "No XMS\n" );
		ErrorExit(1);
	}


	//  Compute the number of elements that will be in the cache
	//
	//  Note:  We are using the same type of logic as MS to figure out
	//         what size the cache should be.  It seems a bit convoluted
	//         until you realize that they don't change the element
	//         size from 8K to what the user specified on the command line
	//         until after the number of elements is calculated.
	//
	if( ulInitCacheSizeKB == 0 )
	{
		if( uXMSLargestFreeK > (6 * 1024) )
		{
			ulInitCacheSizeKB = 2 * 1024;
		}
		else if( uXMSLargestFreeK > (4 * 1024) )
		{
			ulInitCacheSizeKB = 2 * 1024;
		}
		else if( uXMSLargestFreeK > (2 * 1024) )
		{
			ulInitCacheSizeKB = 1024;
		}
		else if( uXMSLargestFreeK > 1024 )
		{
			ulInitCacheSizeKB = 1024;
		}
		else
		{
			ulInitCacheSizeKB = 2 * 1024;
		}
	}
	else
	{
		if( ulInitCacheSizeKB < MIN_XMS_KB )
		{
			ulInitCacheSizeKB = MIN_XMS_KB;
		}
	}
	uNumberOfElements = ulInitCacheSizeKB / ELEMENT_SIZE_DEFAULT;


	//  Compute the size of the cache
	//
	ulInitCacheSizeKB  = uNumberOfElements;
	ulInitCacheSizeKB *= (ulElementSize / 1024);
}


void PASCAL ErrorExit( uchar ucExitCode )
{
	DeinitializeDiskCache();
	exit( ucExitCode );
}


//  Test invalidating the cache
//
void PASCAL TestSet1( void )
{
	uchar ucDrive = 0;
	ulong ulSector;
	unsigned uSectorsRead;
	unsigned uErrorCode;
	ulong ulNumberOfSectors;
	unsigned uSectorsInEachElement;
	ulong ulReadHits;
	ulong ulWriteHits;
	unsigned uElementsInLocalBuffer;
	unsigned uSectorsInLocalBuffer;



	//  Determine the number of sectors that have to be read to fill
	//  up the cache
	//
	ulNumberOfSectors  = ulInitCacheSizeKB / (ulElementSize / 1024);
	uSectorsInEachElement = (unsigned)(ulElementSize /
			MiseryDiskInfo[ucDrive].uSectorSize);
	ulNumberOfSectors *= uSectorsInEachElement;


	uElementsInLocalBuffer = (unsigned)(ulBufferSize / ulElementSize);
	uSectorsInLocalBuffer = uElementsInLocalBuffer * uSectorsInEachElement;


	//  Sequentially read each sector on the disk until the cach is full
	//
	//  NOTE: Cache is reading a full element worth of sectors with
	//        each read request so there is no need to read less than
	//        element size
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++ )
	{
		uSectorsRead = 0xffff;
		uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
				1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
				&uSectorsRead );
		if( uErrorCode )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected error on read sector %lxh line %u\n", ulSector,
					__LINE__ );
			printf("  expected: 0000h  %04xh\n", uErrorCode );
			printf("  uSectorsInEachElement: %u\n", uSectorsInEachElement );
			printf("  uElementsInLocalBuffer: %u\n", uElementsInLocalBuffer );
			printf("  uSectorsInLocalBuffer: %u\n", uSectorsInLocalBuffer );
			ErrorExit(1);
		}
		if( uSectorsRead != 1 )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Wrong sectors read/written value on sector request "
					"%lxh\n", ulSector );
			ErrorExit(1);
		}
		ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
		if( (ulSector % uSectorsInLocalBuffer) == 0)
		{
			ulReadHits++;
		}
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Extra or missing read requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
	}


	//  Make sure each of the sectors are cached
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++  )
	{
		uSectorsRead = 0xffff;
		uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
				1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
				&uSectorsRead );
		if( uErrorCode )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected error on read sector %lxh line %u\n", ulSector,
					__LINE__ );
			printf("  expected: 0000h  %04xh\n", uErrorCode );
			ErrorExit(1);
		}
		if( uSectorsRead != 1 )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Wrong sectors read/written value on sector request "
					"%lxh\n", ulSector );
			ErrorExit(1);
		}
		ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected read requests line %u\n", __LINE__ );
			printf("  ulSector: %lu\n", ulSector );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
	}


	//  Invalidate all the cached sectors for this drive
	//
	InvalidateDiskCache( ucDrive );
	if( ulReadHits != ulDiskReadHitCount )
	{
		printf("\nFAILURE!  TestSet1\n");
		printf("  Unexpected read requests line %u\n", __LINE__ );
		printf("  ulSector: %lu\n", ulSector );
		ErrorExit(1);
	}
	if( ulWriteHits != ulDiskWriteHitCount )
	{
		printf("\nFAILURE!  TestSet1\n");
		printf("  Unexpected write requests line %u\n", __LINE__ );
		ErrorExit(1);
	}
	InvalidateWriteCountArray();


	//  Read each sector that was in the cache making sure that
	//  we hit the disk
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++  )
	{
		uSectorsRead = 0xffff;
		uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
				1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
				&uSectorsRead );
		if( uErrorCode )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected error on read sector %lxh line %u\n", ulSector,
					__LINE__ );
			printf("  expected: 0000h  %04xh\n", uErrorCode );
			ErrorExit(1);
		}
		if( uSectorsRead != 1 )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Wrong sectors read/written value on sector request "
					"%lxh\n", ulSector );
			ErrorExit(1);
		}
		ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
		if( (ulSector % uSectorsInLocalBuffer) == 0)
		{
			ulReadHits++;
		}
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Extra or missing read requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
	}


	//  Invalidate all the cached sectors for this drive
	//
	InvalidateDiskCache( ucDrive );
	if( ulReadHits != ulDiskReadHitCount )
	{
		printf("\nFAILURE!  TestSet1\n");
		printf("  Unexpected read requests line %u\n", __LINE__ );
		printf("  ulSector: %lu\n", ulSector );
		ErrorExit(1);
	}
	if( ulWriteHits != ulDiskWriteHitCount )
	{
		printf("\nFAILURE!  TestSet1\n");
		printf("  Unexpected write requests line %u\n", __LINE__ );
		ErrorExit(1);
	}
	InvalidateWriteCountArray();


#if 0
	//  Sequentially read each sector on the disk until the cache is full
	//
	//  NOTE: Cache is reading a full element worth of sectors with
	//        each read request so there is no need to read less than
	//        element size
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	ulNumberOfSectors /= MAX_DRIVES;
	ulNumberOfSectors /= (MiseryDiskInfo[MAX_DRIVES-1].uSectorSize / 512);
	for( ulSector = 0; ulSector < ulNumberOfSectors;
			ulSector++ )
	{
		for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
		{
			uSectorsRead = 0xffff;
			uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
					1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
					&uSectorsRead );
			if( uErrorCode )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Unexpected error on read sector %lxh line %u\n",
						ulSector, __LINE__ );
				printf("  expected: 0000h  %04xh\n", uErrorCode );
				ErrorExit(1);
			}
			if( uSectorsRead != 1 )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Wrong sectors read/written value on sector request "
						"%lxh\n", ulSector );
				ErrorExit(1);
			}
			ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
			uSectorsInEachElement = (unsigned)(ulElementSize /
					MiseryDiskInfo[ucDrive].uSectorSize);
			uSectorsInLocalBuffer = uElementsInLocalBuffer *
					uSectorsInEachElement;
			if( (ulSector % uSectorsInLocalBuffer) == 0)
			{
				ulReadHits++;
			}
			if( ulReadHits != ulDiskReadHitCount )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Extra or missing read requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
			if( ulWriteHits != ulDiskWriteHitCount )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Unexpected write requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
		}
	}


	//  Make sure each of the sectors are cached
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++  )
	{
		for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
		{
			uSectorsRead = 0xffff;
			uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
					1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
					&uSectorsRead );
			if( uErrorCode )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Unexpected error on read sector %lxh line %u\n",
						ulSector, __LINE__ );
				printf("  expected: 0000h  %04xh\n", uErrorCode );
				ErrorExit(1);
			}
			if( uSectorsRead != 1 )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Wrong sectors read/written value on sector request "
						"%lxh\n", ulSector );
				ErrorExit(1);
			}
			ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
			if( ulReadHits != ulDiskReadHitCount )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Unexpected read requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
			if( ulWriteHits != ulDiskWriteHitCount )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Unexpected write requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
		}
	}


	//  Invalidate all the cached sectors for this drive
	//
	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Invalidate all the cached sectors for this drive
		//
		InvalidateDiskCache( ucDrive );
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected read requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
	}
	InvalidateWriteCountArray();


	//  Read each sector that was in the cache making sure that
	//  we hit the disk
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++  )
	{
		for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
		{
			uSectorsRead = 0xffff;
			uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
					1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
					&uSectorsRead );
			if( uErrorCode )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Unexpected error on read sector %lxh line %u\n",
						ulSector, __LINE__ );
				printf("  expected: 0000h  %04xh\n", uErrorCode );
				ErrorExit(1);
			}
			if( uSectorsRead != 1 )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Wrong sectors read/written value on sector request "
						"%lxh\n", ulSector );
				ErrorExit(1);
			}
			ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
			uSectorsInEachElement = (unsigned)(ulElementSize /
					MiseryDiskInfo[ucDrive].uSectorSize);
 			uSectorsInLocalBuffer = uElementsInLocalBuffer *
					uSectorsInEachElement;
			if( ((ulSector % uSectorsInLocalBuffer) == 0) &&
					(ucDrive == (MAX_DRIVES - 1)) )
			{
				ulReadHits++;
			}
			if( ulReadHits != ulDiskReadHitCount )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Extra or missing read requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
			if( ulWriteHits != ulDiskWriteHitCount )
			{
				printf("\nFAILURE!  TestSet1\n");
				printf("  Unexpected write requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
		}
	}


	//  Invalidate all the cached sectors for this drive
	//
	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Invalidate all the cached sectors for this drive
		//
		InvalidateDiskCache( ucDrive );
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected read requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet1\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
	}
	InvalidateWriteCountArray();
#endif


	printf("TestSet1  PASSED\n");
}


//  Test invalidating the cache
//
void PASCAL TestSet5( void )
{
	uchar ucDrive;
	ulong ulSector;
	unsigned uSectorsRead;
	unsigned uErrorCode;
	ulong ulNumberOfSectors;
	unsigned uSectorsInEachElement;
	ulong ulReadHits;
	ulong ulWriteHits;
	unsigned uElementsInLocalBuffer;
	unsigned uSectorsInLocalBuffer;


	//  Determine the number of sectors that have to be read to fill
	//  up the cache
	//
	ulNumberOfSectors  = ulInitCacheSizeKB / (ulElementSize / 1024);
	uSectorsInEachElement = (unsigned)(ulElementSize /
			MiseryDiskInfo[ucDrive].uSectorSize);
	ulNumberOfSectors *= uSectorsInEachElement;


	uElementsInLocalBuffer = (unsigned)(ulBufferSize / ulElementSize);
	uSectorsInLocalBuffer = uElementsInLocalBuffer * uSectorsInEachElement;


	//  Invalidate all the sectors in the cache
	//
	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Invalidate all the cached sectors for this drive
		//
		InvalidateDiskCache( ucDrive );
	}
	InvalidateWriteCountArray();


	//  Sequentially read each sector on the disk until the cache is full
	//
	//  NOTE: Cache is reading a full element worth of sectors with
	//        each read request so there is no need to read less than
	//        element size
	//
	ucDrive = 1;
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++ )
	{
		SerializeSector( READ, ucDrive, ulSector, &lpucBuffer[0]);
		uSectorsRead = 0xffff;
		uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
				1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
				&uSectorsRead );
		if( uErrorCode )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected error on read sector %lxh line %u\n", ulSector,
					__LINE__ );
			printf("  expected: 0000h  %04xh\n", uErrorCode );
			printf("  uSectorsInEachElement: %u\n", uSectorsInEachElement );
			printf("  uElementsInLocalBuffer: %u\n", uElementsInLocalBuffer );
			printf("  uSectorsInLocalBuffer: %u\n", uSectorsInLocalBuffer );
			ErrorExit(1);
		}
		if( uSectorsRead != 1 )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Wrong sectors read/written value on sector request "
					"%lxh\n", ulSector );
			ErrorExit(1);
		}
		if( (ulSector % uSectorsInLocalBuffer) == 0)
		{
			ulReadHits++;
		}
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Extra or missing read requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
	}


	//  Make sure each of the sectors are cached
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++  )
	{
		uSectorsRead = 0xffff;
		uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
				1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
				&uSectorsRead );
		if( uErrorCode )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected error on read sector %lxh line %u\n", ulSector,
					__LINE__ );
			printf("  expected: 0000h  %04xh\n", uErrorCode );
			ErrorExit(1);
		}
		if( uSectorsRead != 1 )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Wrong sectors read/written value on sector request "
					"%lxh\n", ulSector );
			ErrorExit(1);
		}
		ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected read requests line %u\n", __LINE__ );
			printf("  ulSector: %lu\n", ulSector );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
	}


	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Flush the cache
		//
		FlushDiskCache( ucDrive );
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected read requests line %u\n", __LINE__ );
			printf("  ulSector: %lu\n", ulSector );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
		ValidateUncommittedSectors();
	}


	//  Make sure each of the sectors are still cached
	//
	ucDrive = 1;
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++  )
	{
		uSectorsRead = 0xffff;
		uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
				1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
				&uSectorsRead );
		if( uErrorCode )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected error on read sector %lxh line %u\n", ulSector,
					__LINE__ );
			printf("  expected: 0000h  %04xh\n", uErrorCode );
			ErrorExit(1);
		}
		if( uSectorsRead != 1 )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Wrong sectors read/written value on sector request "
					"%lxh\n", ulSector );
			ErrorExit(1);
		}
		ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected read requests line %u\n", __LINE__ );
			printf("  ulSector: %lu\n", ulSector );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
	}


	//  Invalidate all the cached sectors for this drive
	//
	InvalidateDiskCache( ucDrive );
	InvalidateWriteCountArray();


	//  Read each sector that was in the cache making sure that
	//  we hit the disk
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++  )
	{
		uSectorsRead = 0xffff;
		uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
				1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
				&uSectorsRead );
		if( uErrorCode )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected error on read sector %lxh line %u\n", ulSector,
					__LINE__ );
			printf("  expected: 0000h  %04xh\n", uErrorCode );
			ErrorExit(1);
		}
		if( uSectorsRead != 1 )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Wrong sectors read/written value on sector request "
					"%lxh\n", ulSector );
			ErrorExit(1);
		}
		ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
		if( (ulSector % uSectorsInLocalBuffer) == 0)
		{
			ulReadHits++;
		}
		if( ulReadHits != ulDiskReadHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Extra or missing read requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
		if( ulWriteHits != ulDiskWriteHitCount )
		{
			printf("\nFAILURE!  TestSet5\n");
			printf("  Unexpected write requests line %u\n", __LINE__ );
			ErrorExit(1);
		}
	}


	//  Invalidate all the cached sectors for this drive
	//
	InvalidateDiskCache( ucDrive );
	if( ulReadHits != ulDiskReadHitCount )
	{
		printf("\nFAILURE!  TestSet5\n");
		printf("  Extra or missing read requests line %u\n", __LINE__ );
		ErrorExit(1);
	}
	if( ulWriteHits != ulDiskWriteHitCount )
	{
		printf("\nFAILURE!  TestSet5\n");
		printf("  Unexpected write requests line %u\n", __LINE__ );
		ErrorExit(1);
	}
	InvalidateWriteCountArray();


#if 0
	//  Sequentially read each sector on the disk until the cache is full
	//
	//  NOTE: Cache is reading a full element worth of sectors with
	//        each read request so there is no need to read less than
	//        element size
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	ulNumberOfSectors /= MAX_DRIVES;
	ulNumberOfSectors /= (MiseryDiskInfo[MAX_DRIVES-1].uSectorSize / 512);
	for( ulSector = 0; ulSector < ulNumberOfSectors;
			ulSector++ )
	{
		for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
		{
			uSectorsRead = 0xffff;
			uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
					1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
					&uSectorsRead );
			if( uErrorCode )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Unexpected error on read sector %lxh line %u\n",
						ulSector, __LINE__ );
				printf("  expected: 0000h  %04xh\n", uErrorCode );
				ErrorExit(1);
			}
			if( uSectorsRead != 1 )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Wrong sectors read/written value on sector request "
						"%lxh\n", ulSector );
				ErrorExit(1);
			}
			ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
			uSectorsInEachElement = (unsigned)(ulElementSize /
					MiseryDiskInfo[ucDrive].uSectorSize);
			uSectorsInLocalBuffer = uElementsInLocalBuffer *
					uSectorsInEachElement;
			if( (ulSector % uSectorsInLocalBuffer) == 0)
			{
				ulReadHits++;
			}
			if( ulReadHits != ulDiskReadHitCount )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Extra or missing read requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
			if( ulWriteHits != ulDiskWriteHitCount )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Unexpected write requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
		}
	}


	//  Make sure each of the sectors are cached
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++  )
	{
		for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
		{
			uSectorsRead = 0xffff;
			uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
					1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
					&uSectorsRead );
			if( uErrorCode )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Unexpected error on read sector %lxh line %u\n",
						ulSector, __LINE__ );
				printf("  expected: 0000h  %04xh\n", uErrorCode );
				ErrorExit(1);
			}
			if( uSectorsRead != 1 )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Wrong sectors read/written value on sector request "
						"%lxh\n", ulSector );
				ErrorExit(1);
			}
			ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
			if( ulReadHits != ulDiskReadHitCount )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Unexpected read requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
			if( ulWriteHits != ulDiskWriteHitCount )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Unexpected write requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
		}
	}


	//  Invalidate all the cached sectors for this drive
	//
	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		InvalidateDiskCache( ucDrive );
	}
	InvalidateWriteCountArray();


	//  Read each sector that was in the cache making sure that
	//  we hit the disk
	//
	ulReadHits = ulDiskReadHitCount;
	ulWriteHits = ulDiskWriteHitCount;
	for( ulSector = 0; ulSector < ulNumberOfSectors; ulSector++  )
	{
		for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
		{
			uSectorsRead = 0xffff;
			uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
					1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
					&uSectorsRead );
			if( uErrorCode )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Unexpected error on read sector %lxh line %u\n",
						ulSector, __LINE__ );
				printf("  expected: 0000h  %04xh\n", uErrorCode );
				ErrorExit(1);
			}
			if( uSectorsRead != 1 )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Wrong sectors read/written value on sector request "
						"%lxh\n", ulSector );
				ErrorExit(1);
			}
			ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
			uSectorsInEachElement = (unsigned)(ulElementSize /
					MiseryDiskInfo[ucDrive].uSectorSize);
 			uSectorsInLocalBuffer = uElementsInLocalBuffer *
					uSectorsInEachElement;
			if( ((ulSector % uSectorsInLocalBuffer) == 0) &&
					(ucDrive == (MAX_DRIVES - 1)) )
			{
				ulReadHits++;
			}
			if( ulReadHits != ulDiskReadHitCount )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Extra or missing read requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
			if( ulWriteHits != ulDiskWriteHitCount )
			{
				printf("\nFAILURE!  TestSet5\n");
				printf("  Unexpected write requests line %u\n", __LINE__ );
				ErrorExit(1);
			}
		}
	}
#endif


	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Invalidate all the cached sectors for this drive
		//
		InvalidateDiskCache( ucDrive );
	}
	InvalidateWriteCountArray();


	printf("TestSet5  PASSED\n");
}


//  Test reading sectors sequentially on the disk
//
void PASCAL TestSet2( void )
{
	uchar ucDrive = 0;
	ulong ulSector;
	unsigned uSectorsRead;
	unsigned uErrorCode;


	//  Go through each of the drives
	//
	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Go through each sector skipping around
		//
		for( ulSector = 0; ulSector < MAX_SECTORS; ulSector++ )
		{
			uSectorsRead = 0xffff;
			uErrorCode = CacheReadWrite( READ, ucDrive, ulSector,
					1, &lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached,
					&uSectorsRead );
			if( uErrorCode )
			{
				printf("\nFAILURE!  TestSet2\n");
				printf("  Unexpected error on read sector %lxh\n", ulSector );
				printf("  expected: 0000h  %04xh\n", uErrorCode );
				ErrorExit(1);
			}
			if( uSectorsRead != 1 )
			{
				printf("\nFAILURE!  TestSet2\n");
				printf("  Wrong sectors read/written value on sector request "
						"%lxh\n", ulSector );
				ErrorExit(1);
			}
			ValidateSerializeSector(ucDrive, ulSector, &lpucBuffer[0]);
		}
	}

	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Invalidate all the cached sectors for this drive
		//
		InvalidateDiskCache( ucDrive );
	}
	InvalidateWriteCountArray();


	printf("TestSet2  PASSED\n");
}


//  Test reading recently cached sectors
//
void PASCAL TestSet3( void )
{
	uchar ucDrive = 0;
	ulong ulSector;
	unsigned uSectorAhead = 0;
	unsigned uSectorsRead;
	unsigned uErrorCode;


	//  Go through each sector skipping around
	//
	for( ulSector = 0; ulSector < MAX_SECTORS; ulSector++ )
	{
		//  Go through each of the drives
		//
		for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
		{
			for( uSectorAhead = 0; uSectorAhead < 100; uSectorAhead += 10 )
			{
				uSectorsRead = 0xffff;
				uErrorCode = CacheReadWrite( READ, ucDrive,
						ulSector + uSectorAhead, 1, &lpucBuffer[0],
						MiseryDiskInfo[ucDrive].fIsWriteCached,
						&uSectorsRead );
				if( uErrorCode )
				{
					printf("\nFAILURE!  TestSet3\n");
					printf("  Unexpected error on read sector %lxh + %xh\n",
							ulSector, uSectorAhead );
					printf("  expected: 0000h  %04xh\n", uErrorCode );
					ErrorExit(1);
				}
				if( uSectorsRead != 1 )
				{
					printf("\nFAILURE!  TestSet3\n");
					printf("  Wrong sectors read/written value on sector request "
							"%lxh + %xh\n", ulSector, uSectorAhead );
					ErrorExit(1);
				}
				ValidateSerializeSector(ucDrive, ulSector + uSectorAhead,
						&lpucBuffer[0]);
			}
		}
	}


	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Invalidate all the cached sectors for this drive
		//
		InvalidateDiskCache( ucDrive );
	}
	InvalidateWriteCountArray();


	printf("TestSet3  PASSED\n");
}


//
//
void PASCAL TestSet4( void )
{
	uchar ucDrive = 0;
	ulong ulSector;
	unsigned uSectorCount;
	unsigned uSectorsRead;
	unsigned uErrorCode;


	ulSector = 0x201;
	uSectorCount = 0x1;
	uErrorCode = CacheReadWrite( READ, ucDrive, ulSector, uSectorCount,
			&lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached, &uSectorsRead );
	ValidateSerializeSectors(ucDrive, ulSector, uSectorCount, &lpucBuffer[0]);

	ulSector = 0x95254;
	uSectorCount = 0x1;
	uErrorCode = CacheReadWrite( READ, ucDrive, ulSector, uSectorCount,
			&lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached, &uSectorsRead );
	ValidateSerializeSectors(ucDrive, ulSector, uSectorCount, &lpucBuffer[0]);

	ulSector = 0x26;
	uSectorCount = 0x1;
	uErrorCode = CacheReadWrite( READ, ucDrive, ulSector, uSectorCount,
			&lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached, &uSectorsRead );
	ValidateSerializeSectors(ucDrive, ulSector, uSectorCount, &lpucBuffer[0]);

	ulSector = 0x95255;
	uSectorCount = 0x2c;
	uErrorCode = CacheReadWrite( READ, ucDrive, ulSector, uSectorCount,
			&lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached, &uSectorsRead );
	ValidateSerializeSectors(ucDrive, ulSector, uSectorCount, &lpucBuffer[0]);

	ulSector = 0x95361;
	uSectorCount = 0x2a;
	uErrorCode = CacheReadWrite( READ, ucDrive, ulSector, uSectorCount,
			&lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached, &uSectorsRead );
	ValidateSerializeSectors(ucDrive, ulSector, uSectorCount, &lpucBuffer[0]);

	ulSector = 0x9538b;
	uSectorCount = 0x1;
	uErrorCode = CacheReadWrite( READ, ucDrive, ulSector, uSectorCount,
			&lpucBuffer[0], MiseryDiskInfo[ucDrive].fIsWriteCached, &uSectorsRead );
	ValidateSerializeSectors(ucDrive, ulSector, uSectorCount, &lpucBuffer[0]);


	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Invalidate all the cached sectors for this drive
		//
		InvalidateDiskCache( ucDrive );
	}
	InvalidateWriteCountArray();


	printf("TestSet4  PASSED\n");
}


//
//
void PASCAL TestSet6( void )
{
	bool fIsRead = TRUE;
	uchar ucDrive = 0;
	ulong ulSector = 0;
	unsigned uSectorCount = 0;


	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x201;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x221;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x261;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x26a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x3;
	ulSector = 0x1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x3;
	ulSector = 0x201;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x201;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2a1;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2aa;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x221;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2e1;
	uSectorCount = 0xa;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2eb;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x223;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x222;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x321;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x32a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x361;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x36a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x225;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x224;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x3a1;
	uSectorCount = 0xa;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x3ab;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x226;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x3e1;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x3ea;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x421;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x42a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x227;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x461;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x46a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x228;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x4a1;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x4aa;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x229;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x4e1;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x4ea;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x521;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x52a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x22a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x561;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x56a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x5a1;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x5aa;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x22b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x222;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x5e1;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x5ea;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x621;
	uSectorCount = 0xa;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x62b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x22d;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x22c;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x661;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x66a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x22f;
	uSectorCount = 0x2;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x22e;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x6a1;
	uSectorCount = 0xa;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x6ab;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x6e1;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x6ea;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x231;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x721;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x72a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x232;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x761;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x76a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x233;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x7a1;
	uSectorCount = 0x9;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x7aa;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x234;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x7e1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x7e2;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x235;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x821;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x822;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x236;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x861;
	uSectorCount = 0x2;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x863;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x8a1;
	uSectorCount = 0x2a;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x8cb;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x8e1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x237;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x921;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x238;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x961;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x9a1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x239;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x223;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x101;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x9e1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xa21;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xa61;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x23a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xaa1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xae1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x23b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xb21;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xb61;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xba1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xbe1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x23c;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xc21;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xc61;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x23d;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xca1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xce1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x23e;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xd21;
	uSectorCount = 0x4;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xd25;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x23f;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xd61;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xd66;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x240;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x3;
	ulSector = 0x1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x241;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xda1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xda6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x242;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x224;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xde1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xde6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xe21;
	uSectorCount = 0x4;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xe25;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x243;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xe61;
	uSectorCount = 0x6;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xe67;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xea1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xea6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x244;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xee1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xee6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xf21;
	uSectorCount = 0x6;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xf27;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xf61;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xf66;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x245;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xfa1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xfa6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xfe1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0xfe6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1021;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1026;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1061;
	uSectorCount = 0x6;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1067;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x10a1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x10a6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x10e1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x10e6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1121;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1126;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1161;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1166;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x11a1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x11a6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x246;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x225;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x11e1;
	uSectorCount = 0x18;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x11f9;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1221;
	uSectorCount = 0x1a;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x123b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x247;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1261;
	uSectorCount = 0x3f;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x12a0;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x12a1;
	uSectorCount = 0x25;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x12c6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x248;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x12e1;
	uSectorCount = 0x11;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x12f2;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x249;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x3;
	ulSector = 0x1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x24a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1321;
	uSectorCount = 0x23;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1344;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1361;
	uSectorCount = 0xa;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x136b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x24b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x13a1;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x13a6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x13e1;
	uSectorCount = 0x15;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x13f6;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1421;
	uSectorCount = 0x40;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1461;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x24c;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x14a1;
	uSectorCount = 0x22;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x14c3;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x24d;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x14e1;
	uSectorCount = 0x4a;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x152b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x24e;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1561;
	uSectorCount = 0x58;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x15b9;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x15e1;
	uSectorCount = 0x34;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1615;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x250;
	uSectorCount = 0x3;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x24f;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1621;
	uSectorCount = 0x46;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1667;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x16a1;
	uSectorCount = 0x8;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x16a9;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x253;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x226;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x16e1;
	uSectorCount = 0x2;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x16e3;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x254;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1721;
	uSectorCount = 0x7f;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x255;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x17e1;
	uSectorCount = 0xc;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x17ed;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1821;
	uSectorCount = 0x5;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1826;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x256;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1861;
	uSectorCount = 0x2;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1863;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x18a1;
	uSectorCount = 0x7;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x18a8;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x257;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x18e1;
	uSectorCount = 0x7f;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x19a1;
	uSectorCount = 0x36;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x19d7;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x19e1;
	uSectorCount = 0x11;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x19f2;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x258;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1a21;
	uSectorCount = 0x42;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1a63;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x25a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x259;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1aa1;
	uSectorCount = 0x2d;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1ace;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x25b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1ae1;
	uSectorCount = 0x7a;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1b5b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x25d;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x25c;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1b61;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1ba1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x25e;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1be1;
	uSectorCount = 0x19;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1bfa;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1c21;
	uSectorCount = 0xe;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1c2f;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x227;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1c61;
	uSectorCount = 0x1a;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1c7b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x260;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x25f;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1ca1;
	uSectorCount = 0x23;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1cc4;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x261;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1ce1;
	uSectorCount = 0x1c;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1cfd;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x262;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1d21;
	uSectorCount = 0x4;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1d25;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1d61;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1d62;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x263;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1da1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1da2;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1de1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1e21;
	uSectorCount = 0x2;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1e23;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x265;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x264;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1e61;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1ea1;
	uSectorCount = 0x7;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1ea8;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x267;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x266;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1ee1;
	uSectorCount = 0x7;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1ee8;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1f21;
	uSectorCount = 0x37;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1f58;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x268;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1f61;
	uSectorCount = 0x19;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1f7a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1fa1;
	uSectorCount = 0x43;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x1fe4;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x269;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2021;
	uSectorCount = 0x2c;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x204d;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2061;
	uSectorCount = 0x71;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x20d2;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x26a;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x228;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x20e1;
	uSectorCount = 0x60;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2141;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x26b;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2161;
	uSectorCount = 0x5d;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x21be;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x21e1;
	uSectorCount = 0x28;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x101;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2209;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x26c;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2221;
	uSectorCount = 0x6f;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2290;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x22a1;
	uSectorCount = 0x7f;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x26d;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2361;
	uSectorCount = 0x50;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x23b1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x23e1;
	uSectorCount = 0x7f;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x26e;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2521;
	uSectorCount = 0x51;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2572;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x26f;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x25a1;
	uSectorCount = 0x6b;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x260c;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x270;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2621;
	uSectorCount = 0x15;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2636;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2661;
	uSectorCount = 0x5d;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x26be;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x271;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x26e1;
	uSectorCount = 0x23;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2704;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2721;
	uSectorCount = 0x2d;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x274e;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x272;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x3;
	ulSector = 0x1;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x273;
	uSectorCount = 0x2;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2761;
	uSectorCount = 0x23;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2784;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x275;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x27a1;
	uSectorCount = 0x14;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x27b5;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x27e1;
	uSectorCount = 0x26;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x2807;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = FALSE;
	ucDrive = 0x3;
	ulSector = 0x276;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x4;
	ulSector = 0x229;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );

	fIsRead = TRUE;
	ucDrive = 0x3;
	ulSector = 0x201;
	uSectorCount = 0x1;
	TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, FALSE );


	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		//  Invalidate all the cached sectors for this drive
		//
		InvalidateDiskCache( ucDrive );
	}
	InvalidateWriteCountArray();


	printf("TestSet6  PASSED\n");
}
