//
//   $Id: XMSCACHE.C 1.15 2003/02/18 18:24:18Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*

	XMSCACHE.C
	Created 06-28-2001  by Keith Garvin

	Copyright (c) 2001-2002 Datalight, Inc.
	All rights reserved.
	Confidential and Proprietary

	The XMS cache maintains two arrays: one that maps an XMS cache
	between disk drives and associated sectors and the second which
	maintains a sort list of the sectors saved within the cache.

	Given that the XMS cache is first broke into a number of
	elements (uXmsNumberOfElements) each of a certain size
	(uXmsCacheElementSize), the index into the XmsCacheMap
	cooresponds to the physical offset (in elements) from the start
	of the XMS cache.

	The XmsCacheMap holds DISK_ENTRY's each which describe the
	state of the data within that element, the disk it belongs to,
	and the element within the disk it represents.

	The disk also, is broke into a number of elements of which have
	the same size as the cache elements.  Elements on the disk start
	at some offset into the disk taht represents where data is
	stored within the disk.

	Note that the XMS cache only maintains track of elements.
	Requests for sectors on disks which have a smaller sector size
	will return the element (within the disk or cache) which
	contains the same sector.  Higher level software is required
	to maintain disk specific information.

	The drive numbers tracked in the cache are not interpreted
	by the XMS cache allowing any kind of disk numbers to be used.

	The second structure maintained contains an array of indices
	into the XmsCacheMap and is called the uDiskSectorMap.  Data
	contents describe the index into the XmsCacheMap.  Contents
	are sorted based upon the XmsCacheMap's indexed content's
	disk element number.  Allowing fast search of the DiskSectorMap
	to find a sector in the cache and then it's cooresponding
	data in the cache.

	Provides general public APIs to the XMS cache.

	InitializeXmsCache()
	DeinitializeXmsCache()
	SetXmsDiskEntry()
	GetXmsDiskEntry()
	FindElementInXmsCache()
	RemoveDiskEntryFromXmsCache()
	AddDiskEntryToXmsCache()

*/


#define 		__FILENUM__		3


#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "stdc.h"

#include "cache.h"
#include "xmscache.h"

#include "xmscache.pt"

#ifndef MISERY_TESTING
#define MISERY_TESTING		FALSE
#endif

#if MISERY_TESTING
void RemoveWriteCountEntry( uchar ucDrive, ulong ulSectorNumber );
#endif


bool fLastFoundValid = FALSE;


/*
	All access to and from both the disk and the XMS cache is
	completed in blocks of uXmsCacheElementSize (bytes).  The
	cache is composed of uXmsNumberOfElements of uXmsCacheElementSize.
	For example a 1MB cache might be composed of 256 elements
	each of 4096 bytes in size.
*/
unsigned uXmsCacheElementSize = 0;
unsigned uXmsNumberOfElements = 0;


/*
	The XmsCacheMap provides a mechanism to track what disk
	elements are located within the XMS cache.  the XMS is first
	organized into a number of elements of the recorded size
	above.  Second each index into the XMS cache map corresponds
	to the same element within the XMS cache.

	The DISK_ELEMENT identifies the status of the XMS cache
	element (free, valid, dirty), the disk drive number, and
	the disk element number.

	Given a known physical location within the cache, the
	XmsCacheMap identifies what is there.
*/
DISK_ENTRY far * lprdeXmsCacheMap = NULL;


/*
	The DiskCacheMap provides a mechanism to track where disk
	elements are located within the XmsCache.  This map is
	simply a sorted list of the disk elements located within
	the XMS cache.  Each data element within this array is
	an index into the XmsCacheMap which then has particular
	information regarding the data there.

	Note that since the data within the cache may be assigned
	to different disks and same elements are possible, the
	a binary search only lands within +/- total number of
	disks cached.  A brute force search is then needed possibly,
	before and after the first element match.
*/
unsigned far * lpruDiskSectorMap = NULL;
unsigned uNumberOfDiskEntries = 0;


#if USE_ELEMENT_INFO


//  These arrays are used to quickly retrieve a status mask  for
//  a particular sector within an element
//
unsigned urElementSectorToMask[] = {
	0x0001,	0x0002,	0x0004,	0x0008,
	0x0010,	0x0020,	0x0040,	0x0080,
	0x0100,	0x0200,	0x0400,	0x0800,
	0x1000,	0x2000,	0x4000,	0x8000,
};


//  This pointer to an array is used to track the valid and dirty
//  sectors within an element
//
LPELEMENT_INFO lprElementInfo;


#endif // #if USE_ELEMENT_INFO


/*
	Given the size of our cache, allocate our necessary
	buffers, the cache, and a few global bits of information
*/
bool PASCAL InitializeXmsCache(unsigned uNumberOfElements,
				unsigned uElementSize)
{
	unsigned uSegment;
	unsigned uNumOfParagraphs;
	ulong ulNumOfBytes;

	ASSERT(uNumberOfElements);

	/* Record the organization of the cache */
	uXmsCacheElementSize = uElementSize;
	uXmsNumberOfElements = uNumberOfElements;

	/* At this point there are no sectors in our buffers */
	uNumberOfDiskEntries = 0;


	//  Allocate the DISK_ENTRY array
	//
	ulNumOfBytes = (ulong)uXmsNumberOfElements * sizeof(DISK_ENTRY);
	uNumOfParagraphs = (unsigned)((ulNumOfBytes + 0x0F) / 0X10);
	if(!dlAllocMem(uNumOfParagraphs, &uSegment))
	{
		return FALSE;
	}
	if(uSegment == 0)
	{
		return FALSE;
	}
	lprdeXmsCacheMap = (DISK_ENTRY far *) MK_FP(uSegment, 0);
   _fmemset(lprdeXmsCacheMap, 0,
            (unsigned)(uXmsNumberOfElements * sizeof(DISK_ENTRY)) );

#if USE_ELEMENT_INFO

	//  Allocate the ELEMENT_INFO array
	//
	ulNumOfBytes = (ulong)uXmsNumberOfElements * sizeof(ELEMENT_INFO);
	uNumOfParagraphs = (unsigned)((ulNumOfBytes + 0x0F) / 0X10);
	if(!dlAllocMem(uNumOfParagraphs, &uSegment))
	{
		return FALSE;
	}
	if(uSegment == 0)
	{
		return FALSE;
	}
	lprElementInfo = (ELEMENT_INFO far *) MK_FP(uSegment, 0);
   _fmemset(lprdeXmsCacheMap, 0,
            (unsigned)(uXmsNumberOfElements * sizeof(ELEMENT_INFO)) );

#endif // #if USE_ELEMENT_INFO


	//  Allocate ???
	//
	ulNumOfBytes = (ulong)uXmsNumberOfElements * sizeof(unsigned);
	uNumOfParagraphs = (unsigned)((ulNumOfBytes + 0x0F) / 0X10);
	if(!dlAllocMem(uNumOfParagraphs, &uSegment))
	{
		DeinitializeXmsCache();
		return FALSE;
	}
	if(uSegment == 0)
	{
		DeinitializeXmsCache();
		return FALSE;
	}
	lpruDiskSectorMap = (unsigned far *) MK_FP(uSegment, 0);
   _fmemset(lpruDiskSectorMap, 0,
            (unsigned)(uXmsNumberOfElements * sizeof(unsigned)) );

	return TRUE;
}


/*
	Free all the memory we had been using pior.
*/
void PASCAL DeinitializeXmsCache(void)
{
	/* Boom!  No more sectors in this cache and no going back. */
	uNumberOfDiskEntries = 0;


	//  Free the DISK_ENTRY array
	//
	if(lprdeXmsCacheMap)
	{
		_fmemset(lprdeXmsCacheMap, 0,
				uXmsNumberOfElements * sizeof(DISK_ENTRY));
		dlFreeMem(FP_SEG(lprdeXmsCacheMap));
		lprdeXmsCacheMap = NULL;
	}


#if USE_ELEMENT_INFO

	//  Free the ELEMENT_INFO array
	//
	if(lprElementInfo)
	{
		_fmemset(lprElementInfo, 0,
				uXmsNumberOfElements * sizeof(ELEMENT_INFO));
		dlFreeMem(FP_SEG(lprElementInfo));
		lprElementInfo = NULL;
	}

#endif // #if USE_ELEMENT_INFO

	//  Free the ???
	//
	if(lpruDiskSectorMap)
	{
		_fmemset(lpruDiskSectorMap, 0,
				uXmsNumberOfElements * sizeof(unsigned));
		dlFreeMem(FP_SEG(lpruDiskSectorMap));
		lpruDiskSectorMap = NULL;
	}


	/* Invalidate the description of our cache */
	uXmsCacheElementSize = 0;
	uXmsNumberOfElements = 0;

}


/*
	Return the actual disk entry from the XMS cache map.
	Used by the middle cache code to get access directly
	to the XMS cache data.
*/
bool PASCAL GetXmsDiskEntry(unsigned uXmsMapIndex, DISK_ENTRY * pDiskEntry)
{
	ASSERT(uXmsMapIndex < uXmsNumberOfElements);

	/* Just return the raw data! */
	*pDiskEntry = lprdeXmsCacheMap[uXmsMapIndex];
	return TRUE;
}


/*
	Return the actual disk entry from the XMS cache map.
	Used by the middle cache code to get access directly
	to the XMS cache data.
*/
void PASCAL SetXmsDiskEntry(unsigned uXmsMapIndex, DISK_ENTRY deCacheEntry)
{
	ASSERT(uXmsMapIndex < uXmsNumberOfElements);

	lprdeXmsCacheMap[uXmsMapIndex] = deCacheEntry;
}


/*
	Given a typical disk number and sector number, parse
	the XMS cache and determine if it is already in the
	cache and if so, record the XMSCacheMap index that
	describes it and return TRUE.
*/
bool PASCAL FindElementInXmsCache(unsigned uDisk, ulong ulSectorNumber,
					unsigned uDiskSectorSizeShift, unsigned * puXmsMapIndex)
{
	unsigned uDiskMapIndex;
	DISK_ENTRY deTarget;
	bool fFound;
	unsigned uSectorsPerElement;


	ASSERT(uDiskSectorSize);
	ASSERT(puXmsMapIndex);
	ASSERT(uDisk <= MAKE_DE_DISK(0xff) );

	/* Quick case - nothing to do */
	if(uNumberOfDiskEntries == 0)
		return FALSE;


	/*
		Determine what disk entry to look for in the cache
		For the moment, we are unconcerned with the disk number
		and can assume the status bits are valid.
	*/
	uSectorsPerElement = uXmsCacheElementSize >> uDiskSectorSizeShift;
	deTarget = MAKE_DE_SECTOR(ulSectorNumber, uSectorsPerElement);


	/* Add the disk number into the entry we are looking for */
	deTarget |= MAKE_DE_DISK(uDisk);

	fFound = FindFirstXmsCacheElementMatch(deTarget, &uDiskMapIndex);

	/* If the entry is found!  Record it's index for the caller */
	if(fFound)
	{
		*puXmsMapIndex = lpruDiskSectorMap[uDiskMapIndex];
	}


	return fFound;
}


/*
	Removes an entry from the cache tables given an index to the
	XMS cache map.  The cooresponding disk cach entry will also
	be removed.  This function assumes the disk entry is indeed
	valid and within the cache.
*/
bool PASCAL RemoveDiskEntryFromXmsCache(unsigned uXmsMapIndex,
		unsigned * puOldDiskMapIndex )
{
	DISK_ENTRY deTarget;
	unsigned uDiskMapIndex;
	extern bool fLocalBufferValid;
	extern unsigned uLocalBufferXmsMapIndex;


	CHECK_SORT();


	fLastFoundValid = FALSE;

	/* Validate the parameters */
	if(uXmsMapIndex >= uXmsNumberOfElements)
	{
		return FALSE;
	}

	if( uLocalBufferXmsMapIndex == uXmsMapIndex )
	{
		fLocalBufferValid = FALSE;
	}


#if MISERY_TESTING
	{
		uchar ucDrive;
		ulong ulStartingSector;
		unsigned uSectorIndex;
		unsigned uSectorBit = 1;
 		DISK_ENTRY deCacheEntry;
		unsigned uSectorShiftValue;
		unsigned uSectorsPerElement;
		unsigned uDisk;
		extern DISK_INFO diskInfo[];


		GetXmsDiskEntry(uXmsMapIndex, &deCacheEntry);
		if( DE_VALID(deCacheEntry) )
		{
			uDisk = (unsigned)(GET_DE_DISK(deCacheEntry));
			ucDrive = (uchar)uDisk;
			uSectorShiftValue = diskInfo[uDisk].uSectorShiftValue;
			uSectorsPerElement = uXmsCacheElementSize >> uSectorShiftValue;
			ulStartingSector = GET_DE_SECTOR(deCacheEntry,  uSectorsPerElement );
			for( uSectorIndex = 0; uSectorIndex < uSectorsPerElement;
					uSectorIndex++ )
			{
#if USE_ELEMENT_INFO
				if( lprElementInfo[uXmsMapIndex].uValidSectorBits & uSectorBit )
				{
					RemoveWriteCountEntry( ucDrive, ulStartingSector + uSectorIndex );
				}
				uSectorBit <<= 1;
#else // #if USE_ELEMENT_INFO
	 			RemoveWriteCountEntry( ucDrive, ulStartingSector + uSectorIndex );
#endif // #if USE_ELEMENT_INFO
			}
		}
	}
#endif // #if MISERY_TESTING


#if USE_ELEMENT_INFO
	lprElementInfo[uXmsMapIndex].uValidSectorBits = 0;
	lprElementInfo[uXmsMapIndex].uDirtySectorBits = 0;
#endif // #if USE_ELEMENT_INFO


	/* Find within the disk cache */
	deTarget = lprdeXmsCacheMap[uXmsMapIndex] & DISK_NUM_ELEMENT_MASK;
	if(!FindFirstXmsCacheElementMatch(deTarget, &uDiskMapIndex))
	{
		return FALSE;
	}


	*puOldDiskMapIndex = uDiskMapIndex;


	/*
		Bye-bye data! - Hope the caller meant to do this.
		Invalidate the XmsMap's entry, and slide the disk map's
		data down to delete the requested entry.
	*/
	lprdeXmsCacheMap[uXmsMapIndex] = DISK_STATUS_FREE;
	if((uNumberOfDiskEntries - 1) - uDiskMapIndex)
	{
		_fmemmove(&lpruDiskSectorMap[uDiskMapIndex],
			&lpruDiskSectorMap[uDiskMapIndex + 1],
			((uNumberOfDiskEntries - 1) - uDiskMapIndex) * sizeof(unsigned));
	}


#if USE_ELEMENT_INFO

	//
	//  There is no need to do anything with the ELEMENT_INFO entry
	//  since it will be initialized when the element becomes valid
	//  again.
	//

#endif


	/* Delete the very last entry in the list */
	lpruDiskSectorMap[uNumberOfDiskEntries-1] = 0;
	--uNumberOfDiskEntries;


	return TRUE;
}


/*
	Removes an entry from the cache tables given an index to the
	XMS disk sector map.  The cooresponding disk cach entry will also
	be removed.  This function assumes the disk entry is indeed
	valid and within the cache.
*/
bool PASCAL _RemoveDiskEntryFromXmsCache(unsigned uDiskMapIndex)
{
	unsigned uXmsMapIndex;
	extern bool fLocalBufferValid;
	extern unsigned uLocalBufferXmsMapIndex;


	CHECK_SORT();


	fLastFoundValid = FALSE;

	uXmsMapIndex = lpruDiskSectorMap[uDiskMapIndex];
	if( uLocalBufferXmsMapIndex == uXmsMapIndex )
	{
		fLocalBufferValid = FALSE;
	}


#if MISERY_TESTING
	{
		uchar ucDrive;
		ulong ulStartingSector;
		unsigned uSectorIndex;
		unsigned uSectorBit = 1;
 		DISK_ENTRY deCacheEntry;
		unsigned uSectorShiftValue;
		unsigned uSectorsPerElement;
		unsigned uDisk;
		extern DISK_INFO diskInfo[];


		GetXmsDiskEntry(uXmsMapIndex, &deCacheEntry);
		if( DE_VALID(deCacheEntry) )
		{
			uDisk = (unsigned)(GET_DE_DISK(deCacheEntry));
			ucDrive = (uchar)uDisk;
			uSectorShiftValue = diskInfo[uDisk].uSectorShiftValue;
			uSectorsPerElement = uXmsCacheElementSize >> uSectorShiftValue;
			ulStartingSector = GET_DE_SECTOR(deCacheEntry,  uSectorsPerElement );
			for( uSectorIndex = 0; uSectorIndex < uSectorsPerElement;
					uSectorIndex++ )
			{
#if USE_ELEMENT_INFO
				if( lprElementInfo[uXmsMapIndex].uValidSectorBits & uSectorBit )
				{
					RemoveWriteCountEntry( ucDrive, ulStartingSector + uSectorIndex );
				}
				uSectorBit <<= 1;
#else // #if USE_ELEMENT_INFO
	 			RemoveWriteCountEntry( ucDrive, ulStartingSector + uSectorIndex );
#endif // #if USE_ELEMENT_INFO
			}
		}
	}
#endif // #if MISERY_TESTING


#if USE_ELEMENT_INFO
	lprElementInfo[uXmsMapIndex].uValidSectorBits = 0;
	lprElementInfo[uXmsMapIndex].uDirtySectorBits = 0;
#endif // #if USE_ELEMENT_INFO


	/*
		Bye-bye data! - Hope the caller meant to do this.
		Invalidate the XmsMap's entry, and slide the disk map's
		data down to delete the requested entry.
	*/
	lprdeXmsCacheMap[uXmsMapIndex] = DISK_STATUS_FREE;
	if((uNumberOfDiskEntries - 1) - uDiskMapIndex)
	{
		_fmemmove(&lpruDiskSectorMap[uDiskMapIndex],
			&lpruDiskSectorMap[uDiskMapIndex + 1],
			((uNumberOfDiskEntries - 1) - uDiskMapIndex) * sizeof(unsigned));
	}


#if USE_ELEMENT_INFO

	//
	//  There is no need to do anything with the ELEMENT_INFO entry
	//  since it will be initialized when the element becomes valid
	//  again.
	//

#endif


	/* Delete the very last entry in the list */
	lpruDiskSectorMap[uNumberOfDiskEntries-1] = 0;
	uNumberOfDiskEntries--;


	return TRUE;
}


/*
	Adds a disk entry into the cache maps.  After this call, both
	the disk cache map and the xms cache map are updated to reflect
	the new data located within the cache.  This function assumes
	the entry is not already in the list.
*/
 bool PASCAL AddDiskEntryToXmsCache(unsigned uXmsMapIndex,
		unsigned uDiskMapIndex, DISK_ENTRY deTarget )
{
	fLastFoundValid = FALSE;

	/* There must already be room in the disk cache */
	if(uNumberOfDiskEntries >= uXmsNumberOfElements)
	{
		//  Should this error be a design problem?
		//
		asm int 3
		asm hlt

		return FALSE;
	}

	/* Slide the cache entries up to insert the new entry */
	if(uNumberOfDiskEntries - uDiskMapIndex)
	{
		ASSERT((DISK_ELEMENT_MASK(lprdeXmsCacheMap[lpruDiskSectorMap[uDiskMapIndex]]))
				>= (DISK_ELEMENT_MASK(deTarget)) );
		_fmemmove(&lpruDiskSectorMap[uDiskMapIndex + 1],
				&lpruDiskSectorMap[uDiskMapIndex],
				(uNumberOfDiskEntries - uDiskMapIndex) * sizeof(unsigned));
	}

	/* Add the entry into the disk cache */
	lpruDiskSectorMap[uDiskMapIndex] = uXmsMapIndex;
	uNumberOfDiskEntries++;

	/* Plunk the disk information into the given index. */
	deTarget |= DISK_STATUS_VALID;
	lprdeXmsCacheMap[uXmsMapIndex] = deTarget;


	CHECK_SORT();


#if USE_ELEMENT_INFO
	lprElementInfo[uXmsMapIndex].uValidSectorBits = 0;
	lprElementInfo[uXmsMapIndex].uDirtySectorBits = 0;
#endif


	return TRUE;
}


/*
	Find the first match of the given disk entry within
	the XMS cache ignoring the disk number, and only
	searching on the given element number.  Additional
	elements on other disks are located above the index
	returned to the caller.  Return TRUE if it is found,
	FALSE otherwise.  The puDiskMapIndex is always
	updated to point to where the entrie is on a find,
	or where the entry belongs, if not already in the
	list.
*/
bool PASCAL FindFirstXmsCacheElementMatch(DISK_ENTRY deTarget,
					unsigned * puDiskMapIndex)
{
	int iLow, iMid, iHigh;
	DISK_ENTRY deCurrentCache;
	bool fFound;
	static bool fLastFound;
	static DISK_ENTRY deLastCurrentCache;
	static unsigned uLastDiskMapIndex;


	CHECK_SORT();


	ASSERT(puDiskMapIndex);

	/* Ensure there is a cache datum to search */
	if(uNumberOfDiskEntries == 0)
	{
		/* An empty cache can take the slot zero */
		*puDiskMapIndex = 0;
		return FALSE;
	}


	deTarget &= DISK_NUM_ELEMENT_MASK;


	if( fLastFoundValid )
	{
		if( deTarget == deLastCurrentCache )
		{
			//  Retrieve search optimization results
			//
			fFound = fLastFound;
			*puDiskMapIndex = uLastDiskMapIndex;
			deCurrentCache = deLastCurrentCache;
		}
		else
		{
			fLastFoundValid = FALSE;
		}
	}


	if( !fLastFoundValid )
	{
		/* Start the binary search */
		fFound = FALSE;
		iLow = 0;
		iHigh = uNumberOfDiskEntries - 1;
		while(iLow <= iHigh)
		{
			/* Check the middle entry first */
			iMid = (iLow + iHigh) >> 1;


			/*
				Read the actual disk entry from the cache map
			*/
			deCurrentCache = lprdeXmsCacheMap[lpruDiskSectorMap[iMid]]
						& DISK_NUM_ELEMENT_MASK;

			/* If the client block is less than this alloc, its too small */
			if(deCurrentCache > deTarget)
			{
				/* Reduce the upper (iHigh) boundary */
				iHigh = iMid - 1;
				continue;
			}

			/* If the Allocation is above this... */
			if(deCurrentCache < deTarget)
			{
				/* Go down by half */
				iLow = iMid + 1;
				continue;
			}

			/*
				Now we are close, if not exactly, pointing to
				the corrent entry in the cache.
			*/
			fFound = TRUE;

			break;
		}


		/* Always return where the entry should go if it were there */
		if( fFound )
		{
			*puDiskMapIndex = iMid;


			//  Store search optimization results
			//
			fLastFoundValid = TRUE;
			fLastFound = fFound;
			uLastDiskMapIndex = *puDiskMapIndex;
			deLastCurrentCache = deCurrentCache;
		}
		else
		{
			*puDiskMapIndex = iLow;
		}

		ASSERT(iMid <= uXmsNumberOfElements);
		ASSERT(iMid >= 0);
	}

	return fFound;
}


#if ENABLE_CHECK_SORT

void CheckSort( unsigned uFile, unsigned uLine )
{
	DISK_ENTRY deLastCacheEntry;
	DISK_ENTRY deCurrentCacheEntry;
	unsigned uIndex;
	static char cBuffer[256];


	if(uNumberOfDiskEntries < 2)
	{
		//  Must have at least 2 entries to validate sort
		//
		return;
	}


	deLastCacheEntry = lprdeXmsCacheMap[lpruDiskSectorMap[0]]
				& DISK_NUM_ELEMENT_MASK;
	for( uIndex = 1; uIndex < uNumberOfDiskEntries; uIndex++ )
	{
		deCurrentCacheEntry = lprdeXmsCacheMap[lpruDiskSectorMap[uIndex]]
					& DISK_NUM_ELEMENT_MASK;
		if( deLastCacheEntry >= deCurrentCacheEntry )
		{
			//  Cache sorted lookup is hosed
			//
			sprintf( cBuffer, "Cache not sorted file: %u  line: %u\n",
					uFile, uLine );
			asm int 3
		}
		deLastCacheEntry = deCurrentCacheEntry;
	}
}

#endif


#if TESTING

#define TESTX(x) if(!(x)) { \
			printf("Failed file: %s line: %u\n", \
			__FILE__, __LINE__); return FALSE; }

bool PASCAL TestXmsCache(void)
{
	unsigned uXmsMapIndex = 0;
	DISK_ENTRY deTemp;

	/* Test a few pass cases */
	TESTX(InitializeXmsCache(128, 8192));
	DeinitializeXmsCache();

	TESTX(InitializeXmsCache(128, 1024));
	DeinitializeXmsCache();

	TESTX(InitializeXmsCache(512, 1024));
	DeinitializeXmsCache();

	/*
		The Real testing begins....
		Start with initialization of the caches
	*/
	TESTX(InitializeXmsCache(4, 1024));

	/* Empty list, the following calls should fail */
	TESTX(!FindElementInXmsCache(0x18, 200, 512, &uXmsMapIndex));
	uXmsMapIndex = 0;
	TESTX(!RemoveDiskEntryFromXmsCache(uXmsMapIndex));

	/*
		Add an entry, 0x13 will be our cached drive for the test
	*/
	uXmsMapIndex = 0;
	TESTX(AddDiskEntryToXmsCache(uXmsMapIndex, 0x13, 100, 9));

	/*
		Make certain each of these work appropriately:
			1) Wrong drive AND Wrong sector
			2) Wrong drive AND Correct sector
			3) Correct drive AND Wrong sector
			4) Correct drive AND Correct sector
	*/
	TESTX(!FindElementInXmsCache(0x18, 200, 512, &uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x17, 100, 512, &uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x13, 102, 512, &uXmsMapIndex));

	/* Note this search is successful since the element size is 1KB */
	TESTX( FindElementInXmsCache(0x13, 101, 512, &uXmsMapIndex));
	TESTX( FindElementInXmsCache(0x13, 100, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex < uXmsNumberOfElements);

	/*
		Now remove the entry and we should be back to an empty list
	*/
	uXmsMapIndex = 0;
	TESTX(RemoveDiskEntryFromXmsCache(uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x13, 100, 512, &uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x18, 200, 512, &uXmsMapIndex));
	TESTX(!RemoveDiskEntryFromXmsCache(uXmsMapIndex));

	/*
		Now add three entries where two have the same sector,
		but different drives and the other two have different
		sectors, but the same drive.
	*/
	uXmsMapIndex = 0;
	TESTX(!FindElementInXmsCache(0x13, 500, 512, &uXmsMapIndex));
	TESTX(AddDiskEntryToXmsCache(uXmsMapIndex, 0x13, 500, 9));
	GetXmsDiskEntry(uXmsMapIndex, &deTemp);
	TESTX(deTemp & DISK_STATUS_VALID);

	uXmsMapIndex = 1;
	TESTX(!FindElementInXmsCache(0x13, 100, 512, &uXmsMapIndex));
	TESTX(AddDiskEntryToXmsCache(uXmsMapIndex, 0x13, 100, 9));
	GetXmsDiskEntry(uXmsMapIndex, &deTemp);
	TESTX(deTemp & DISK_STATUS_VALID);

	uXmsMapIndex = 2;
	TESTX(!FindElementInXmsCache(0x17, 100, 512, &uXmsMapIndex));
	TESTX(AddDiskEntryToXmsCache(uXmsMapIndex, 0x17, 100, 9));
	GetXmsDiskEntry(uXmsMapIndex, &deTemp);
	TESTX(deTemp & DISK_STATUS_VALID);

	/* Make certain both the set and get entries work */
	uXmsMapIndex = 1;
	GetXmsDiskEntry(uXmsMapIndex, &deTemp);
	TESTX(deTemp & DISK_STATUS_VALID);

	/* Invalidate the entry */
	deTemp &= ~DISK_STATUS_VALID;
	SetXmsDiskEntry(uXmsMapIndex, deTemp);
	GetXmsDiskEntry(uXmsMapIndex, &deTemp);
	TESTX(!(deTemp & DISK_STATUS_VALID));

	/* Mark it as valid again */
	deTemp = 0x12345678UL;
	GetXmsDiskEntry(uXmsMapIndex, &deTemp);
	deTemp |= DISK_STATUS_VALID;
	SetXmsDiskEntry(uXmsMapIndex, deTemp);


	/* Make certain we can find each one individually */
	TESTX(FindElementInXmsCache(0x17, 100, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 2);
	TESTX(FindElementInXmsCache(0x13, 100, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 1);
	TESTX(FindElementInXmsCache(0x13, 500, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 0);

	/* And that we don't find similar, but different ones */
	TESTX(!FindElementInXmsCache(0x17, 500, 512, &uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x13, 200, 512, &uXmsMapIndex));

	/* Fill the cache up */
	uXmsMapIndex = 3;
	TESTX(!FindElementInXmsCache(0x17, 499, 512, &uXmsMapIndex));
	TESTX(AddDiskEntryToXmsCache(uXmsMapIndex, 0x17, 499, 9));
	TESTX( FindElementInXmsCache(0x17, 499, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 3);

	/* Make certain we can't do much with a full cache */
	TESTX(!FindElementInXmsCache(0x18, 999, 512, &uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x10, 29, 512, &uXmsMapIndex));
	TESTX(!AddDiskEntryToXmsCache(uXmsMapIndex, 0x10, 29, 9));
	TESTX(!AddDiskEntryToXmsCache(uXmsMapIndex, 0x10, 999, 9));

	/* Remove a middle one (1) */
	uXmsMapIndex = 1;
	TESTX(RemoveDiskEntryFromXmsCache(uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x13, 100, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 1);

	/* Fill the cache back up with another unique sector */
	uXmsMapIndex = 1;
	TESTX(!FindElementInXmsCache(0x13, 300, 512, &uXmsMapIndex));
	TESTX(AddDiskEntryToXmsCache(uXmsMapIndex, 0x13, 300, 9));
	TESTX( FindElementInXmsCache(0x13, 300, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 1);

	/* Remove a middle one (2) */
	uXmsMapIndex = 2;
	TESTX(RemoveDiskEntryFromXmsCache(uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x17, 100, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 2);

	/* Remove a the first physical one (0) */
	uXmsMapIndex = 0;
	TESTX(RemoveDiskEntryFromXmsCache(uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x13, 500, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 0);

	/* Remove a middle one again (1) */
	uXmsMapIndex = 1;
	TESTX(RemoveDiskEntryFromXmsCache(uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x13, 100, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 1);

	/* And finally the last physical one (3) */
	uXmsMapIndex = 3;
	TESTX(RemoveDiskEntryFromXmsCache(uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x17, 499, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 3);

	/* Empty list, the following calls should fail now */
	uXmsMapIndex = 7;
	TESTX(!FindElementInXmsCache(0x13, 100, 512, &uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x17, 499, 512, &uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x13, 500, 512, &uXmsMapIndex));
	TESTX(!FindElementInXmsCache(0x17, 100, 512, &uXmsMapIndex));
	TESTX(uXmsMapIndex == 7);

	/* Attempt to remove entries that no longer exist */
	TESTX(!RemoveDiskEntryFromXmsCache(0));
	TESTX(!RemoveDiskEntryFromXmsCache(2));
	TESTX(!RemoveDiskEntryFromXmsCache(4));
	TESTX(!RemoveDiskEntryFromXmsCache(7));

	DeinitializeXmsCache();

	return TRUE;
}
#endif
