//
//   $Id: XMSCACHE.H 1.10 2002/10/16 21:38:52Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*

	XMSCACHE.H
	Created 06-29-2001  by Keith Garvin

	Provides public APIs and types to the XMS cache.

	Copyright (c) 2001-2002 Datalight, Inc.
	All rights reserved.
	Confidential and Proprietary

*/


#define USE_ELEMENT_INFO			TRUE


/*
	bit  31....: If set, indicates valid data
	bit  30....: If set, indicates dirty cache element
	bit  29....: Reserved and set to zero
	bits 28..24: Disk drive number
	bits 23...0: Disk element number

	The disk element number is an offset within the disk
	that is zero modulo of the element size defined
	at initialization of the cache.

	Disk drive number is the drive number used by the client
	of the cache.

	Status field indicates availability of the element
	within the cache.
*/
typedef unsigned long DISK_ENTRY;


#if USE_ELEMENT_INFO

//  This is structure used to manage the status of each of
//  the sectors within an element
//
typedef struct ELEMENT_INFO ELEMENT_INFO;
typedef ELEMENT_INFO * PELEMENT_INFO;
typedef ELEMENT_INFO far * LPELEMENT_INFO;
struct ELEMENT_INFO
{
	unsigned uValidSectorBits;
	unsigned uDirtySectorBits;
};

#endif



/* AND masks to isolate portions of each disk entry */
#define DISK_STATUS_MASK(de)  ((de) & 0XE0000000UL)
#define DISK_NUM_MASK(de)     ((de) & 0X1F000000UL)
#define DISK_ELEMENT_MASK(de)	((de) & 0X00FFFFFFUL)

/* The following mask isolates both the disk and element number */
#define DISK_NUM_ELEMENT_MASK 0X1FFFFFFFUL

//  Macro to get the element/sector number from a disk entry
//
#define GET_DE_SECTOR(element,uSectorsPerElement) (DISK_ELEMENT_MASK(element) * (uSectorsPerElement))

//  Macro to make the element/sector number for a disk entry
//
#define MAKE_DE_SECTOR(ulSector,uSectorsPerElement) (DISK_ELEMENT_MASK(((ulSector) / (uSectorsPerElement))))

//  Macro to get the drive number from a disk entry
//
#define GET_DE_DISK(element)	(DISK_NUM_MASK(element) >> 24)

//  Macro to make the drive number for a disk entry
//
#define MAKE_DE_DISK(disk)		(DISK_NUM_MASK(((ulong)(disk)) << 24))

/* OR masks used to check against the status field */
#define DISK_STATUS_FREE      0X00000000UL
#define DISK_STATUS_VALID     0X80000000UL
#define DISK_STATUS_DIRTY     0X40000000UL


//  This macro determines if an element is valid and dirty
//
#define DE_VALID_AND_DIRTY(de)	(((de) & (DISK_STATUS_DIRTY | DISK_STATUS_VALID)) ==	(DISK_STATUS_DIRTY | DISK_STATUS_VALID) )


//  This macro determines if an element is valid
//
#define DE_VALID(de)  (((de) & DISK_STATUS_VALID) ==	DISK_STATUS_VALID )

#if USE_ELEMENT_INFO

#define EI_VALID(uXmsMapIndex,uSectorOffset)        \
 (lprElementInfo[(uXmsMapIndex)].uValidSectorBits & \
  urElementSectorToMask[(uSectorOffset)])

#define EI_VALID_DIRTY(uXmsMapIndex,uSectorOffset)    \
 ((lprElementInfo[(uXmsMapIndex)].uValidSectorBits  & \
   lprElementInfo[(uXmsMapIndex)].uDirtySectorBits) & \
   urElementSectorToMask[(uSectorOffset)])

#endif


/* Prototypes */
bool PASCAL InitializeXmsCache(unsigned uNumberOfElements,
				unsigned uElementSize);
void PASCAL DeinitializeXmsCache(void);
bool PASCAL GetXmsDiskEntry(unsigned uXmsMapIndex, DISK_ENTRY * pDiskEntry);
void PASCAL SetXmsDiskEntry(unsigned uXmsMapIndex, DISK_ENTRY deCacheEntry);
void PASCAL GetXmsCacheConfiguration(unsigned * puNumberOfElements,
				unsigned * puElementSize);
bool PASCAL FindElementInXmsCache(unsigned uDisk, ulong ulSectorNumber,
					unsigned uDiskSectorSize, unsigned * puXmsMapIndex);

bool PASCAL AddDiskEntryToXmsCache(unsigned uXmsMapIndex,
		unsigned uDiskMapIndex, DISK_ENTRY deTarget );
bool PASCAL RemoveDiskEntryFromXmsCache(unsigned uXmsMapIndex,
		unsigned * puOldDiskMapIndex );

bool PASCAL FindFirstXmsCacheElementMatch(DISK_ENTRY deTarget,
					unsigned * puDiskMapIndex);
bool PASCAL _RemoveDiskEntryFromXmsCache(unsigned uDiskMapIndex);

#if TESTING
bool PASCAL TestXmsCache(void);
#endif
