//
//   $Id: CHKBUF.C 1.13 2003/03/04 22:05:42Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  CHKBUF.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  This module contains the buffering scheme functions for chkdsk
//
//  HISTORY:
//
//  03-15-2000  bt  Created (Started over from scratch)
//  03-23-1989  Drew Gislason (created old chkdsk)
//


#define __FILENUMBER__		2


//  Borland libraries
//
#include <dir.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <mem.h>
#include <string.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dlxms.h"
#include "dlmem.h"
#include "crc32.h"
#include "fatlib.h"


//  Chkdsk header files
//
#include "chkdsk.h"
#include "chkasert.h"


//  Prototypes
//
#include "chkdsk.pt"


//  Variables
//
extern char szInsufficientMemory[];
extern char szNotReady[];
extern bool fIsFat32World;

//  Segment of the only allocated conventional memory block.
//
unsigned uConvMemSegment;

//  These are the MRU buffers
//
LPBUF_MGT lpMRUConvMemBuffer;
LPSTORE_BUF_MGT lpMRUXMSMemBuffer;


#if DEBUG_BUFFERS
unsigned uAllocatedConvMemBuffers;
unsigned uAllocatedStoreMemBuffers;
#endif // #if DEBUG_BUFFERS


//  Used by the cache hit counting code
//
#if DEBUG_VIEW_CACHE_HITS
ulong ulSectorRequests       = 0;
ulong ulConvMemBuffCacheHit  = 0;
ulong ulMemStoreBuffCacheHit = 0;
#endif //  #if DEBUG_VIEW_CACHE_HITS


//  Initialize the buffering scheme
//
void InitializeBuffers( void )
{
	LPDPB lpDPB;
	ulong ulVisitedListByteSize;
	ulong ulMemStoreSize;
	unsigned uConvMemPara;
	unsigned uConvMemBuffers;
	unsigned uXMSMemBuffers;
	ulong ulTemp;
	bool fSuccessful;
	uchar * pucVisitedListSwap;
	ulong ulCount;
	ulong ulPhysicalAddress;
	unsigned uCount;
	LPBUF_MGT lpConvBuffMgt;
	LPBUF_MGT lpPrevConvBuffMgt;
	LPSTORE_BUF_MGT lpXMSBuffMgt;
	LPSTORE_BUF_MGT lpPrevXMSBuffMgt;
	ulong ulTotalConventionalMemory;
	ulong ulConvBufByteSize;
	ulong ulXMSBufByteSize;
	ulong ulOffsetInStore;


	DEBUG_STEP_MESSAGE( "GetDPB( &lpDPB )" );


	//  Determine how much memory is required for the FAT
	//  "visited" list by determining how many bytes it
	//  would take to represent all the fat entries with
	//  four bits per entry.
	//
	//  ulVisitedListByteSize = MaxCluster * 4 / 8 = MaxCluster / 2
	//
	if( !GetDPB( &lpDPB ) )
	{
		//  Display an error message and exit
		//
		printf(szNotReady);
		exit(255);
	}
	ASSERT_LPDPB( lpDPB );
	if( IsFat32Drive(lpDPB) )
	{
		ulVisitedListByteSize = lpDPB->ulMaxCluster;
	}
	else
	{
		ulVisitedListByteSize = lpDPB->hiclus;
	}
#if 0
	ulVisitedListByteSize++;
	ulVisitedListByteSize >>= 1;	// ulVisitedListByteSize /= 2;
	ulVisitedListByteSize += ulVisitedListByteSize & 1;  // round up
#else
   ulVisitedListByteSize >>= 9;
   ulVisitedListByteSize++;
   ulVisitedListByteSize <<= 8;
#endif // #if 0

	DEBUG_STEP_MESSAGE( "InitMemStore()" );


	//  Determine the total amount of memory we have to work with
	//
	if( !InitMemStore() )
	{
		CHKDSK_INTERNAL_ERROR();
	}
	ulMemStoreSize = MemStoreSize();
	KillMemStore();


	//  Validate that the buffer structures are word sized
	//
	ASSERT_EVEN( sizeof(BUF_MGT) );
	ASSERT_EVEN( sizeof(STORE_BUF_MGT) );


	//  Determine if there is sufficient memory to run chkdsk
	//
	ulTemp = lpDPB->bytes_per_sec + sizeof(BUF_MGT);	// size of 1 conv buffer
	ulTemp *= MIN_CONV_BUFFERS;
	ulTemp += lpDPB->bytes_per_sec + sizeof(STORE_BUF_MGT);  //  1 store buffer
	ulTemp += 0x10;		//  MCB for conventional memory allocation
	ulTemp += ulVisitedListByteSize;  //  visited list size
	if( ulTemp > ulMemStoreSize )
	{

//  Debug code
//
#if 0
		printf("ulMemStoreSize: %lx\n", ulMemStoreSize );
		printf("ulVisitedListByteSize: %lx\n", ulVisitedListByteSize );
		printf("conv buffer size: %x\n", lpDPB->bytes_per_sec + sizeof(BUF_MGT) );
		printf("store buffer size: %x\n", lpDPB->bytes_per_sec + sizeof(STORE_BUF_MGT) );
		printf("MIN_CONV_BUFFERS: %x\n", MIN_CONV_BUFFERS );
#endif

		//  Display an isufficient memory error message and exit
		//
		printf(szInsufficientMemory);
		exit(255);
	}


	DEBUG_STEP_MESSAGE( "dlAllocMem( MAX_PARAGRAPH_SIZE, &uConvMemPara )" );


	//  Determine the amount of available conventional memory
	//
	dlAllocMem( MAX_PARAGRAPH_SIZE, &uConvMemPara );


	//  Reduce the amount of conventional mememory by 1 MCB header
	//
	uConvMemPara--;
//	printf("uConvMemPara: %xh\n", uConvMemPara );


	//  Determine the byte size of the conventional memory
	//  we have to work with.
	//
	ulTemp = ulMemStoreSize - ulVisitedListByteSize - lpDPB->bytes_per_sec;
	if( ulTemp > ((ulong)uConvMemPara << 4) )
	{
		ulTemp = ((ulong)uConvMemPara << 4);
	}
	ulTotalConventionalMemory = ulTemp;


	//  Automatically, there must be 1 XMS buffer
	//
	ulTotalConventionalMemory -= sizeof(STORE_BUF_MGT);


	//  Determine the number of conventional memory buffers
	//
	ulTemp = ulTotalConventionalMemory;
	ulTemp /= (lpDPB->bytes_per_sec + sizeof(BUF_MGT));
	if( ulTemp > MAX_CONV_MEM_BUFFERS )
	{
		ulTemp = MAX_CONV_MEM_BUFFERS;
	}
	uConvMemBuffers = (unsigned)ulTemp;
	ulConvBufByteSize = (ulong)(uConvMemBuffers) *
			(lpDPB->bytes_per_sec + sizeof(BUF_MGT));


	//  Adjust the total amount of conventional memory remaining
	//
	ulTotalConventionalMemory -= ulConvBufByteSize;


	//  Determine the number of XMS buffers
	//
	ulTemp = ulTotalConventionalMemory;
	ulTemp /= sizeof(STORE_BUF_MGT);
	if( ulTemp > MAX_XMS_MEM_BUFFERS )
	{
		ulTemp = MAX_XMS_MEM_BUFFERS;
	}
	uXMSMemBuffers = (unsigned)ulTemp;
	uXMSMemBuffers++;  			//  Add in the automatic one
	ulXMSBufByteSize = (ulong)(uXMSMemBuffers) * sizeof(STORE_BUF_MGT);



	//  Make sure that there is enough room in the memory store for
	//  the number of XMS buffers we can have.  Adjust the number
	//  as necessary.
	//
	ulTemp = ulMemStoreSize - ulVisitedListByteSize - lpDPB->bytes_per_sec;
	ulTemp -= ulConvBufByteSize;
	ulTemp -= ulXMSBufByteSize;
	ulTemp /= lpDPB->bytes_per_sec;
	if( ulTemp < (ulong)uXMSMemBuffers )
	{
		uXMSMemBuffers = (unsigned)ulTemp;
	}



#if DEBUG_BUFFERS
	uAllocatedConvMemBuffers = uConvMemBuffers;
	printf("uAllocatedConvMemBuffers: %xh\n", uAllocatedConvMemBuffers );
#endif  // #if DEBUG_BUFFERS


	DEBUG_STEP_MESSAGE( "dlAllocMem( uConvMemPara, &uConvMemSegment )" );


	//  Allocate the memory for the conventional memory buffers
	//
	uConvMemSegment = 0;
	ulTemp = (ulong)(lpDPB->bytes_per_sec + sizeof(BUF_MGT)) *
			uConvMemBuffers;
	ulTemp += (ulong)uXMSMemBuffers * sizeof(STORE_BUF_MGT);
	uConvMemPara = (unsigned)((ulTemp + 0xf) >> 4);
#if DEBUG_ASSERTS
	uAllocatedPara = uConvMemPara;
#endif  // #if DEBUG_ASSERTS
	fSuccessful = dlAllocMem( uConvMemPara, &uConvMemSegment );
	if( !fSuccessful )
	{
		//  The calculations above should assure that we are not
		//  allocating more memory than is available.
		//
		CHKDSK_INTERNAL_ERROR();
	}
//	printf("uConvMemPara: %xh\n", uConvMemPara );
	ASSERT_NONZERO( uConvMemSegment );
#if DEBUG_ASSERTS
	uAllocatedSegment = uConvMemSegment;
#endif  // #if DEBUG_ASSERTS


	DEBUG_STEP_MESSAGE( "InitMemStore()" );


	//  Initialize the file store
	//
	if( !InitMemStore() )
	{
		CHKDSK_INTERNAL_ERROR();
	}


	DEBUG_STEP_MESSAGE( "MemStoreSize()" );


	//  Determine the size of the file store
	//
	ulMemStoreSize = MemStoreSize();


	//  Determine how much memory is left after the visited list is
	//  placed in the memory store
	//
	ulMemStoreSize -= ulVisitedListByteSize;


#if DEBUG_BUFFERS
	uAllocatedStoreMemBuffers = uXMSMemBuffers;
	printf("uAllocatedStoreMemBuffers: %xh\n", uAllocatedStoreMemBuffers );
#endif  // #if DEBUG_BUFFERS


	//  Initialize the visited list
	//
	{
		//  Determine the location of the visited list swap area
		//
		pucVisitedListSwap = GetVisitedListSwapAddress();


		//  Zero the entire visited list swap area
		//
		memset( pucVisitedListSwap, 0, VISITED_LIST_SWAP_SIZE );


		//  Determine the offset of the unitialized part of the
		//  visited list
		//
		ulPhysicalAddress = 0;


		//  Initialize each entry in the visited list
		//
		ulCount = ulVisitedListByteSize;
		while( ulCount )
		{
			//  Set the next location for a write to the memory store
			//
			MemStoreSeek( ulPhysicalAddress, SEEK_SET );


			//  if the size of the uninitialize part of the visited
			//  list is greater than or equal to the size of the
			//  visited list swap area...
			//
			if( ulCount > VISITED_LIST_SWAP_SIZE )
			{
				//  Write the vistited list swap area into the memory store
				//
				MemStoreWrite( pucVisitedListSwap, VISITED_LIST_SWAP_SIZE );


				//  Determine the offset of the unitialized part of the
				//  visited list
				//
				ulPhysicalAddress += VISITED_LIST_SWAP_SIZE;


				//  Adjust the amount left to write
				//
				ulCount -= VISITED_LIST_SWAP_SIZE;
			}
			else
			{
				//  Write the remaining part of the visited list swap area
				//  into the memory store
				//
				MemStoreWrite( pucVisitedListSwap, (unsigned)ulCount );


				//  Determine the offset of the unitialized part of the
				//  visited list
				//
				ulPhysicalAddress += ulCount;


				//  Adjust the amount left to write
				//
				ulCount -= ulCount;
			}
		}
	}


	//  Initialize the buffer management structures for each of
	//  the sector buffers in conventional memory
	//
	{
		//  Determine the size of the conventional buffer
		//  management structure and the size of the buffer
		//  together.
		//
		ulTemp = sizeof(BUF_MGT) + lpDPB->bytes_per_sec;


		//  Determine the physical memory address of the first
		//  conventional memory buffer
		//
		ulPhysicalAddress = (ulong)uConvMemSegment << 4;


		//  Set the initial MRU conventional memory buffer
		//
		lpMRUConvMemBuffer = PhysicalToSegmentedAddress( ulPhysicalAddress );


		//  Set up the double linked list for the conventional
		//  memory buffers
		//
		uCount = uConvMemBuffers;
 		lpPrevConvBuffMgt = NULL;
		while( uCount )
		{
			//  Determine the address of the conventional memory buffer
			//
			lpConvBuffMgt = PhysicalToSegmentedAddress( ulPhysicalAddress );


			//  Set status flag to invalid (buffer) in the conventional
			//  memory buffer managment status field
			//
			lpConvBuffMgt->ucStatusFlags = 0;


			//  if this is the first conventional memory buffer...
			//
			if( uCount == uConvMemBuffers )
			{
				//  Set the address of the next conventional memory buffer
				//  in the next field of the the conventional memory buffer
				//  managment structure
				//
				lpConvBuffMgt->lpNext =
						PhysicalToSegmentedAddress( ulPhysicalAddress + ulTemp );


				//  Set the address of the last conventional memory buffer
				//  in the previous field of the conventional memory buffer
				//  management structure
				//
				lpConvBuffMgt->lpPrev =
						PhysicalToSegmentedAddress( ulPhysicalAddress +
						(ulTemp * (uConvMemBuffers - 1)) );
			}
			else if( uCount == 1 )
			{
				//
				//  if this is the last conventional memory buffer
				//

				//  Set the address of the first conventional memory
				//  buffer in the next field of the the conventional
				//  memory buffer managment structure
				//
				lpConvBuffMgt->lpNext =
						PhysicalToSegmentedAddress( (ulong)uConvMemSegment << 4 );


				//  Determine the address of the previous conventional
				//  memory buffer and set the address of the previous
				//  conventional memory buffer in the previous field
				//  of the conventional memory buffer management structure
				//
				lpConvBuffMgt->lpPrev = lpPrevConvBuffMgt;
			}
			else
			{
				//  Determine the address of the next conventional memory
				//  buffer and set the address of the next conventional
				//  memory buffer in the next field of the the conventional
				//  memory buffer managment structure
				//
				lpConvBuffMgt->lpNext =
						PhysicalToSegmentedAddress( ulPhysicalAddress + ulTemp );


				//  Determine the address of the previous conventional
				//  memory buffer and set the address of the previous
				//  conventional memory buffer in the previous field
				//  of the conventional memory buffer management structure
				//
				lpConvBuffMgt->lpPrev = lpPrevConvBuffMgt;
			}


			//  Determine the address of the last conventional memory
			//  buffer
			//
			lpPrevConvBuffMgt = lpConvBuffMgt;


			//  Determine the location of the next memory store
			//  buffer management structure
			//
			ulPhysicalAddress += ulTemp;


			//  Determine the number of buffers we have left
			//
			uCount--;
		}


		//  Make sure that the buffering scheme is OK
		//
		ASSERT_BUFFERS();
	}


	//  Initialize the buffer management structures for each of
	//  the sector buffers in XMS memory.
	//
	//  NOTE: buffer headers are in conventional memory
	//
	{
		//  Determine the size of the XMS buffer
		//  management structure.
		//
		ulTemp = sizeof(STORE_BUF_MGT);


		//  Set the initial MRU XMS memory buffer
		//
		lpMRUXMSMemBuffer = PhysicalToSegmentedAddress( ulPhysicalAddress );


		//  Set the initial memory store offset
		//
		ulOffsetInStore = ulVisitedListByteSize;


		//  Set up the double linked list for the conventional
		//  memory buffers
		//
		uCount = uXMSMemBuffers;
		lpPrevXMSBuffMgt = NULL;
		while( uCount )
		{
			//  Determine the address of the conventional memory buffer
			//
			lpXMSBuffMgt = PhysicalToSegmentedAddress( ulPhysicalAddress );


			//  Set status flag to invalid (buffer) in the conventional
			//  memory buffer managment status field
			//
			lpXMSBuffMgt->ucStatusFlags = 0;


			//  Set the offset of the buffer data into the memory store
			//
			lpXMSBuffMgt->ulOffsetInStore = ulOffsetInStore;


			//  if this is the first XMS memory buffer...
			//
			if( uCount == uXMSMemBuffers )
			{
				//  Set the address of the next XMS memory buffer
				//  in the next field of the the XMS memory buffer
				//  managment structure
				//
				lpXMSBuffMgt->lpNext =
						PhysicalToSegmentedAddress( ulPhysicalAddress + ulTemp );


				//  Set the address of the last XMS memory buffer
				//  in the previous field of the XMS memory buffer
				//  management structure
				//
				lpXMSBuffMgt->lpPrev =
						PhysicalToSegmentedAddress( ulPhysicalAddress +
						(ulTemp * (uXMSMemBuffers - 1)) );
			}
			else if( uCount == 1 )
			{
				//
				//  if this is the last XMS memory buffer
				//

				//  Set the address of the first XMS memory
				//  buffer in the next field of the the XMS
				//  memory buffer managment structure
				//
				lpXMSBuffMgt->lpNext = lpMRUXMSMemBuffer;


				//  Determine the address of the previous XMS
				//  memory buffer and set the address of the previous
				//  XMS memory buffer in the previous field
				//  of the XMS memory buffer management structure
				//
				lpXMSBuffMgt->lpPrev = lpPrevXMSBuffMgt;
			}
			else
			{
				//  Determine the address of the next XMS memory
				//  buffer and set the address of the next XMS
				//  memory buffer in the next field of the the XMS
				//  memory buffer managment structure
				//
				lpXMSBuffMgt->lpNext =
						PhysicalToSegmentedAddress( ulPhysicalAddress + ulTemp );


				//  Determine the address of the previous XMS
				//  memory buffer and set the address of the previous
				//  XMS memory buffer in the previous field
				//  of the XMS memory buffer management structure
				//
				lpXMSBuffMgt->lpPrev = lpPrevXMSBuffMgt;
			}


			//  Determine the address of the last conventional memory
			//  buffer
			//
			lpPrevXMSBuffMgt = lpXMSBuffMgt;


			//  Determine the location of the next memory store
			//  buffer management structure
			//
			ulPhysicalAddress += ulTemp;


			//  Determine the location of the next memory store
			//  buffer in the memory store
			//
			ulOffsetInStore += lpDPB->bytes_per_sec;


			//  Determine the number of buffers we have left
			//
			uCount--;
		}


		//  Make sure that the buffering scheme is OK
		//
		ASSERT_STOREBUFFERS();
	}


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();
}


//  Free all the memory used by our sector buffers
//
void ReleaseAllMemory( void )
{
	KillMemStore();
	dlFreeMem( uConvMemSegment );
}


//  Get the address of a sector in conventional memory
//
void far * GetSectorAddress( ulong ulSectorNumber )
{
	void far * lpBuffer;
	unsigned * puBufferHandle;
	uchar ucDrive;


	//  Get the drive number
	//
	ucDrive = GetWorkingDriveDriveNumber();


	//  DEBUG
	//
#if 0
	if( ulSectorNumber == 0x201 )
	{
		unsigned bt;

		bt = 0;
	}
#endif  // #if 0


	//  Increament the counter of requested sectors
	//
	SECTOR_REQUEST_HIT()


	//  Get the sector
	//
	ReadSectorBuffer( ucDrive,	ulSectorNumber, (void **)&puBufferHandle );


	//  Get the address
	//
	GetBufferDataAddress( puBufferHandle, &lpBuffer );
	ASSERT_RANGE_CONV_BUFF( lpBuffer );


	//  DEBUG
	//
	//  Compare the sector read off of disk.  This will only work
	//  initially because the sector buffers may be modified
	//
#if 0
	{
		static char cTempBuffer[MAX_SECTOR_SIZE];
		LPDPB lpDPB;


		//  Get the DPB for the drive so that we can have the
		//  sector size
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Read the sector in for comparison
		//
		if( fIsFat32World )
		{
			dlAbsRead32( ucDrive+1, 1, ulSectorNumber,
					cTempBuffer );
		}
		else
		{
			dlAbsDiskRead( ucDrive, 1, ulSectorNumber,
					cTempBuffer );
		}

		//  Compare the sector
		//
		if( _fmemcmp( lpBuffer, cTempBuffer, lpDPB->bytes_per_sec ) != 0 )
		{
			printf("Sector buffering failed!\n");
			exit(255);
		}
	}
#endif  // #if 0


	//  Return the address
	//
	return lpBuffer;
}


//  Return the buffer information in the conventional memory buffers
//  for the given sector.  If this sector is not in the conventional
//  memory buffers, this function returns FALSE.
//
bool GetConvMemBuffInfo( ulong ulSectorNumber, LPBUF_MGT * plpConvBuffMgt )
{
	LPBUF_MGT lpConvBuffMgt;


	//  Determine if CHKDSK is syncronizing the FAT copies
	//
	if( GetSFCFlag() )
	{
		return FALSE;
	}


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();


	//  Get the MRU conventional memory buffer
	//
	lpConvBuffMgt = lpMRUConvMemBuffer;


	//  Go through all of the conventional memory buffers looking
	//  for a valid copy of this sector
	//
	do
	{
		//  Determine if this buffer is valid.  If this buffer is
		//  not valid ...
		//
		if( (lpConvBuffMgt->ucStatusFlags & STATUS_VALID) == 0 )
		{
			//  Make sure that the buffering scheme is OK
			//
			ASSERT_BUFFERS();


			//  This is an optimization for searching for buffers.  We
			//  will never have an invalid buffer except when we first
			//  start filling the buffers.  We will spend a lot of time
			//  looking through the buffers when nothing is in them.
			//  So we are done looking in this buffer list.
			//
			return FALSE;
		}


		//  Determine if this buffer is for the specified sector.  If this
		//  buffer is for the specifed sector ...
		//
		if( lpConvBuffMgt->ulSectorNumber == ulSectorNumber )
		{
			//  This is the sector (buffer) we were looking for
			//
			*plpConvBuffMgt = lpConvBuffMgt;


			//  Update the cache hit statistics
			//
			CONV_MEM_BUFF_CACHE_HIT();


			//  Make sure that the buffering scheme is OK
			//
			ASSERT_BUFFERS();


			return TRUE;
		}


		//  Get the next buffer
		//
		lpConvBuffMgt = lpConvBuffMgt->lpNext;

	}while(lpConvBuffMgt != lpMRUConvMemBuffer );


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();


	//  We did not find the specified sector
	//
	return FALSE;
}


//  Set this sector as the most recently used
//
void SetConvMemBufferAsMRU( LPBUF_MGT lpNewMRUConvBuff )
{
	LPBUF_MGT lpNextConvMemBuffer;
	LPBUF_MGT lpPrevConvMemBuffer;


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();


	//  Determine what the MRU buffer is.  If the new buffer
	//  is not already the MRU buffer, make it the MRU buffer
	//
	if( lpMRUConvMemBuffer != lpNewMRUConvBuff )
	{
		//  Remove the buffer from the linked list
		//
		lpPrevConvMemBuffer = lpNewMRUConvBuff->lpPrev;
		lpNextConvMemBuffer = lpNewMRUConvBuff->lpNext;
		lpPrevConvMemBuffer->lpNext = lpNextConvMemBuffer;
		lpNextConvMemBuffer->lpPrev = lpPrevConvMemBuffer;


		//  Insert the buffer into the linked list at the MRU
		//  conventional memory	buffer
		//
		lpPrevConvMemBuffer = lpMRUConvMemBuffer->lpPrev;
		lpNewMRUConvBuff->lpNext = lpMRUConvMemBuffer;
		lpNewMRUConvBuff->lpPrev = lpPrevConvMemBuffer;
		lpPrevConvMemBuffer->lpNext = lpNewMRUConvBuff;
		lpMRUConvMemBuffer->lpPrev = lpNewMRUConvBuff;


		//  Set the new buffer as the MRU buffer
		//
		lpMRUConvMemBuffer = lpNewMRUConvBuff;
	}


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();
}


//  Get the LRU conventional memory buffer
//
LPBUF_MGT GetLRUConvMemBuffer( void )
{
	return lpMRUConvMemBuffer->lpPrev;
}


//  Get the MRU conventional memory buffer
//
LPBUF_MGT GetMRUConvMemBuffer( void )
{
	return lpMRUConvMemBuffer;
}


//  Free the MRU buffer and make it the LRU buffer if it is not
//  dirty.
//
//  NOTE: This function is an optimization for Syncronizing the
//        FAT copies.  This is used so that the buffers do not
//        cache all the duplicate FAT sectors.
//
void FreeMRUConvBufferAndMakeItLRU( void )
{
	LPBUF_MGT lpBuffer;


	lpBuffer = GetMRUConvMemBuffer();
	if( (lpBuffer->ucStatusFlags & STATUS_DIRTY) == 0 )
	{
		lpBuffer->ucStatusFlags = STATUS_AVAILABLE;
		lpMRUConvMemBuffer = lpBuffer->lpNext;
	}
}


//  Get the LRU memory store buffer
//
LPSTORE_BUF_MGT GetLRUMemStoreBuffer( void )
{
	//  Return the address/offset of the LRU buffer
	//
	return lpMRUXMSMemBuffer->lpPrev;
}


//  Get the LRU memory store buffer
//
LPSTORE_BUF_MGT GetMRUMemStoreBuffer( void )
{
	//  Return the address/offset of the MRU buffer
	//
	return lpMRUXMSMemBuffer;
}


//  Return the buffer information in the XMS memory buffers
//  for the given sector.  If this sector is not in the XMS
//  memory buffers, this function returns FALSE.
//
bool GetMemStoreBuffInfo( ulong ulSectorNumber,
		LPSTORE_BUF_MGT * plpXMSBuffMgt )
{
	LPSTORE_BUF_MGT lpXMSBuffMgt;


	//  Determine if CHKDSK is syncronizing the FAT copies
	//
	if( GetSFCFlag() )
	{
		return FALSE;
	}


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_STOREBUFFERS();


	//  Get the MRU XMS memory buffer
	//
	lpXMSBuffMgt = lpMRUXMSMemBuffer;


	//  Go through all of the XMS memory buffers looking
	//  for a valid copy of this sector
	//
	do
	{
		//  Determine if this buffer is valid.  If this buffer is
		//  not valid ...
		//
		if( (lpXMSBuffMgt->ucStatusFlags & STATUS_VALID) == 0 )
		{
			//  Make sure that the buffering scheme is OK
			//
			ASSERT_STOREBUFFERS();


			//  This is an optimization for searching for buffers.  We
			//  will never have an invalid buffer except when we first
			//  start filling the buffers.  We will spend a lot of time
			//  looking through the buffers when nothing is in them.
			//  So we are done looking in this buffer list.
			//
			return FALSE;
		}


		//  Determine if this buffer is for the specified sector.  If this
		//  buffer is for the specifed sector ...
		//
		if( lpXMSBuffMgt->ulSectorNumber == ulSectorNumber )
		{
			//  This is the sector (buffer) we were looking for
			//
			*plpXMSBuffMgt = lpXMSBuffMgt;


			//  Update the cache hit statistics
			//
			MEM_STORE_BUFF_CACHE_HIT();


			//  Make sure that the buffering scheme is OK
			//
			ASSERT_STOREBUFFERS();


			return TRUE;
		}


		//  Get the next buffer
		//
		lpXMSBuffMgt = lpXMSBuffMgt->lpNext;

	}while(lpXMSBuffMgt != lpMRUXMSMemBuffer );


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_STOREBUFFERS();


	//  We did not find the specified sector
	//
	return FALSE;
}


//  Set this sector as the most recently used
//
void SetMemStoreBufferAsMRU( LPSTORE_BUF_MGT lpNewMRUXMSBuff )
{
	LPSTORE_BUF_MGT lpNextXMSMemBuffer;
	LPSTORE_BUF_MGT lpPrevXMSMemBuffer;


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_STOREBUFFERS();


	//  Determine what the MRU buffer is.  If the new buffer
	//  is not already the MRU buffer, make it the MRU buffer
	//
	if( lpMRUXMSMemBuffer != lpNewMRUXMSBuff )
	{
		//  Remove the buffer from the linked list
		//
		lpPrevXMSMemBuffer = lpNewMRUXMSBuff->lpPrev;
		lpNextXMSMemBuffer = lpNewMRUXMSBuff->lpNext;
		lpPrevXMSMemBuffer->lpNext = lpNextXMSMemBuffer;
		lpNextXMSMemBuffer->lpPrev = lpPrevXMSMemBuffer;


		//  Insert the buffer into the linked list at the MRU
		//  XMS memory	buffer
		//
		lpPrevXMSMemBuffer = lpMRUXMSMemBuffer->lpPrev;
		lpNewMRUXMSBuff->lpNext = lpMRUXMSMemBuffer;
		lpNewMRUXMSBuff->lpPrev = lpPrevXMSMemBuffer;
		lpPrevXMSMemBuffer->lpNext = lpNewMRUXMSBuff;
		lpMRUXMSMemBuffer->lpPrev = lpNewMRUXMSBuff;


		//  Set the new buffer as the MRU buffer
		//
		lpMRUXMSMemBuffer = lpNewMRUXMSBuff;
	}


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_STOREBUFFERS();


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();
}


//  Set the memory store buffer as the MRU conventional memory buffer
//
void MakeMemStoreBufferMRUConvMemBuffer( LPSTORE_BUF_MGT lpXMSBuffMgt )
{
	ulong ulTempSectorNumber;
	uchar ucTempStatusFlags;
	char * pTempSectorBuffer;
	LPDPB lpDPB;
	LPBUF_MGT lpLRUConvMemBuffer;
#if DEBUG_CRC_BUFFERS
	ulong ulCRCStoreBuffer;
#endif //  #if DEBUG_CRC_BUFFERS


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_STOREBUFFERS();


	//  Set this memory store buffer as the MRU memory store buffer
	//
	SetMemStoreBufferAsMRU( lpXMSBuffMgt );


	//  Get the address of the temp buffer
	//
	pTempSectorBuffer = GetTempSectorBuffer();


	//  Get the DPB for the drive so that we can have the
	//  sector size
	//
	GetDPB( &lpDPB );
	ASSERT_LPDPB( lpDPB );


	//  Save the sector number and status flags for the buffer
	//  we just copied into the temp buffer
	//
	ulTempSectorNumber = lpXMSBuffMgt->ulSectorNumber;
	ucTempStatusFlags  = lpXMSBuffMgt->ucStatusFlags;
#if DEBUG_CRC_BUFFERS
	ulCRCStoreBuffer   = lpXMSBuffMgt->ulCRC32;
#endif //  #if DEBUG_CRC_BUFFERS


	//  DEBUG
	//
#if 0
	if( lpXMSBuffMgt->ulSectorNumber == 0x201 )
	{
		unsigned uBT;

		uBT = 0;
	}
#endif  // #if 0


	//  Copy the memory store buffer into the temp buffer
	//
	MemStoreSeek( lpXMSBuffMgt->ulOffsetInStore, SEEK_SET );
	MemStoreRead( pTempSectorBuffer,	lpDPB->bytes_per_sec );


	//  Make sure the CRC is correct on the memory store buffer
	//
	CHECK_STORE_BUFFER_CRC( lpXMSBuffMgt, pTempSectorBuffer );


	//  Get the LRU conventional memory buffer
	//
	lpLRUConvMemBuffer = GetLRUConvMemBuffer();


	//  Make sure the CRC is correct on the conventional memory buffer
	//
	CHECK_CONV_BUFFER_CRC( lpLRUConvMemBuffer );


	//  Set the sector number and status flags in the memory
	//  store buffer structure
	//
	lpXMSBuffMgt->ulSectorNumber = lpLRUConvMemBuffer->ulSectorNumber;
	lpXMSBuffMgt->ucStatusFlags  = lpLRUConvMemBuffer->ucStatusFlags;
#if DEBUG_CRC_BUFFERS
	lpXMSBuffMgt->ulCRC32        = lpLRUConvMemBuffer->ulCRC32;
#endif //  #if DEBUG_CRC_BUFFERS


	//  DEBUG
	//
#if 0
	if( lpLRUConvMemBuffer->ulSectorNumber == 0x201 )
	{
		char far * lpBT;

		lpBT = BUFFER_ADDR(lpLRUConvMemBuffer);
	}
#endif  // #if 0


	//  Copy the LRU conventional memory buffer into the MRU
	//  memory store buffer
	//
	MemStoreSeek( lpXMSBuffMgt->ulOffsetInStore, SEEK_SET );
	MemStoreWrite( BUFFER_ADDR(lpLRUConvMemBuffer), lpDPB->bytes_per_sec );


	//  Copy the buffer data in the temp buffer into LRU conventional
	//  memory buffer
	//
	_fmemcpy( BUFFER_ADDR(lpLRUConvMemBuffer), pTempSectorBuffer,
			lpDPB->bytes_per_sec );


	//  Set the sector number and status flags in the LRU conventional
	//  memory buffer
	//
	lpLRUConvMemBuffer->ulSectorNumber = ulTempSectorNumber;
	lpLRUConvMemBuffer->ucStatusFlags  = ucTempStatusFlags;
#if DEBUG_CRC_BUFFERS
	lpLRUConvMemBuffer->ulCRC32        = ulCRCStoreBuffer;
#endif //  #if DEBUG_CRC_BUFFERS


	//  Set the LRU conventional memory buffer as the MRU conventional
	//  memory buffer
	//
	lpMRUConvMemBuffer = lpLRUConvMemBuffer;


	//  Make sure the CRC is correct on the conventional memory buffer
	//
	CHECK_CONV_BUFFER_CRC( lpMRUConvMemBuffer );


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_STOREBUFFERS();


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();
}


//  Make a conventional memory buffer available, flushing a
//  sector buffer as neccessary, and set this buffer as the
//  MRU conventional memory buffer
//
void MakeAvailableBufferMRU( void )
{
	LPBUF_MGT lpLRUConvMemBuffer;
	LPSTORE_BUF_MGT lpXMSBuffMgt;
	static LPDPB lpDPB = 0;


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_STOREBUFFERS();


	//  Determine if we already have the DPB address.  If we do not
	//  already have the DPB address ...
	//
	if( !lpDPB )
	{
		//  Get the DPB for the drive
		//
		GetDPB( &lpDPB );
	}
	ASSERT_LPDPB( lpDPB );


	//  Get the LRU memory store buffer
	//
	lpLRUConvMemBuffer = GetLRUConvMemBuffer();


	//  Determine if this buffer is available.  If this buffer is
	//  available ...
	//
	if( (lpLRUConvMemBuffer->ucStatusFlags & STATUS_VALID) != 0 )
	{
		//  Make sure the CRC is correct on the conventional memory buffer
		//
		CHECK_CONV_BUFFER_CRC( lpLRUConvMemBuffer );


		//  Get the buffer information
		//
		lpXMSBuffMgt = GetLRUMemStoreBuffer();


		//  Determine if this buffer is available.  If this buffer
		//  is not available ...
		//
		if( (lpXMSBuffMgt->ucStatusFlags & STATUS_VALID) != 0 )
		{
			//  Flush the LRU memory store buffer
			//
			FlushLRUMemStoreBuffer( lpXMSBuffMgt );
		}


		//  DEBUG
		//
#if 0
		if( lpLRUConvMemBuffer->ulSectorNumber == 0x201 )
		{
			unsigned bt;

			bt = 0;
		}
#endif  // #if 0


		//  Copy the LRU conventional memory buffer contents
		//  to the LRU memory store buffer
		//
		MemStoreSeek( lpXMSBuffMgt->ulOffsetInStore,	SEEK_SET );
		MemStoreWrite( BUFFER_ADDR(lpLRUConvMemBuffer), lpDPB->bytes_per_sec );
		lpXMSBuffMgt->ulSectorNumber = lpLRUConvMemBuffer->ulSectorNumber;
		lpXMSBuffMgt->ucStatusFlags  = lpLRUConvMemBuffer->ucStatusFlags;
#if DEBUG_CRC_BUFFERS
		lpXMSBuffMgt->ulCRC32        = lpLRUConvMemBuffer->ulCRC32;
#endif //  #if DEBUG_CRC_BUFFERS


		//  Set the LRU memory store buffer as the MRU memory store
		//  buffer
		//
		lpMRUXMSMemBuffer = lpXMSBuffMgt;


		//  Since the LRU conventional memory has been copied, the
		//  buffer is now invalid
		//
		lpLRUConvMemBuffer->ucStatusFlags = 0;
	}


	//  Set the available LRU conventional memory buffer to the
	//  MRU conventional memory buffer
	//
	lpMRUConvMemBuffer = lpLRUConvMemBuffer;


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_STOREBUFFERS();


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();
}



//  Flush the LRU memory store buffer
//
void FlushLRUMemStoreBuffer( LPSTORE_BUF_MGT lpXMSBuffMgt )
{
	LPDPB lpDPB;
	char * pTempSectorBuffer;
	uchar ucCopies;
	ulong ulSectorNumber;
	ulong ulFatLength;
	bool fSuccessful;
	ulong ulFatEnd;
#if !DEBUG_NO_WRITES
	uchar ucDrive;
#endif  // #if !DEBUG_NO_WRITES


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_STOREBUFFERS();


	//  Determine if this buffer is valid.  If this buffer
	//  is valid ...
	//
	if( (lpXMSBuffMgt->ucStatusFlags & STATUS_VALID) != 0 )
	{
		//  Get the DPB for the drive so that we can have the
		//  sector size
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Determine the distance between copies
		//
		if( IsFat32Drive(lpDPB) )
		{
			ulFatLength = lpDPB->ulSectorsPerFAT;
		}
		else
		{
			ulFatLength = lpDPB->fat_len;
		}


		//  Determine the sector number where the FAT ends
		//
		ulFatEnd = lpDPB->fat_start + ulFatLength;


#if DEBUG_CRC_BUFFERS

		//  Get the address of the temp sector buffer
		//
		pTempSectorBuffer = GetTempSectorBuffer();


		//  Copy the sector contents to conventional memory
		//
		MemStoreSeek( lpXMSBuffMgt->ulOffsetInStore, SEEK_SET );
		MemStoreRead( pTempSectorBuffer,	lpDPB->bytes_per_sec );

#endif //  #if DEBUG_CRC_BUFFERS


		//  Make sure the CRC is correct on the memory store buffer
		//
		CHECK_STORE_BUFFER_CRC( lpXMSBuffMgt, pTempSectorBuffer );


		//  Determine if this buffer is dirty.  If this buffer
		//  is dirty ...
		//
		if( (lpXMSBuffMgt->ucStatusFlags & STATUS_DIRTY) != 0 )
		{
			//  As soon as we commit this sector it will not be
			//  dirty anymore.  Go ahead and set this buffer
			//  not dirty
			//
			lpXMSBuffMgt->ucStatusFlags &= ~STATUS_DIRTY;


			//  We only need to write the sector if we are
			//  committing the disk changes
			//
			if( DoFixErrors() )
			{
				//  Get the address of the temp sector buffer
				//
				pTempSectorBuffer = GetTempSectorBuffer();


#if !DEBUG_NO_WRITES
				//  Get the drive number we are checking
				//
				ucDrive = GetWorkingDriveDriveNumber();
#endif  // #if !DEBUG_NO_WRITES

				//  Get the DPB for the drive so that we can have the
				//  sector size
				//
				GetDPB( &lpDPB );
				ASSERT_LPDPB( lpDPB );


				//  Get the sector number
				//
				ulSectorNumber = lpXMSBuffMgt->ulSectorNumber;


				//  We are going to commit this sector to disk but
				//  we have to copy it into conventional memory
				//  first.
				//
				MemStoreSeek( lpXMSBuffMgt->ulOffsetInStore,	SEEK_SET );
				MemStoreRead( pTempSectorBuffer,	lpDPB->bytes_per_sec );


				//  Determine the number of copies of this
				//  sector
				//
				{
					//  Determine if this is a FAT sector.  If this
					//  is a FAT sector ...
					//
					if( (ulSectorNumber >= lpDPB->fat_start) &&
							(ulSectorNumber < ulFatEnd) )
					{
						ucCopies = GetFATCopies();
					}
					else
					{
						ucCopies = 1;
					}
				}


				//  Write each copy of this sector
				//
				while( ucCopies )
				{
					//  Determine which abswrite call to use and perform
					//  the write
					//
#if DEBUG_NO_WRITES
					printf("writing sector: %lu\n", ulSectorNumber );
					fSuccessful = TRUE;
#else  // #if DEBUG_NO_WRITES
					if( fIsFat32World )
					{
						//  Write the sectors for this copy
						//  (drive 0=default,1=A,...)
						//
						fSuccessful = dlAbsWrite32( ucDrive+1, 1, ulSectorNumber,
								pTempSectorBuffer, ABS32WRITE_UNKNOWN );
					}
					else
					{
						//  Write the sectors for this copy (drive 0=A,1=B,...)
						//
						fSuccessful = dlAbsDiskWrite( ucDrive, 1,
								ulSectorNumber, pTempSectorBuffer );

					}
#endif  // #if DEBUG_NO_WRITES

					//  Determine if there was an error
					//
					if(!fSuccessful)
					{
						//  ???
						//
						printf("Failed to write changes!\n");
						exit(255);
					}


					//  Mark that we wrote this copy
					//
					ucCopies--;


					//  Determine the starting sector for the next copy
					//
					ulSectorNumber += ulFatLength;
				}
			}
		}
	}


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();
}


//  Flush all the conventional memory buffers
//
void FlushAllConventionalMemoryBuffers( void )
{
	LPBUF_MGT lpLRUStart;
	LPBUF_MGT lpLRUBuffer;
	LPDPB lpDPB;
	ulong ulSectorNumber;
	uchar ucCopies;
	ulong ulFatLength;
	ulong ulFatEnd;
	bool fSuccessful;
#if !(DEBUG_NO_WRITES)
	char far * lpSectorBuffer;
#endif  // #if !(DEBUG_NO_WRITES)


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();


	//  Get the DPB for the drive so that we can have the
	//  sector size
	//
	GetDPB( &lpDPB );
	ASSERT_LPDPB( lpDPB );


	//  Determine the distance between copies
	//
	if( IsFat32Drive(lpDPB) )
	{
		ulFatLength = lpDPB->ulSectorsPerFAT;
	}
	else
	{
		ulFatLength = lpDPB->fat_len;
	}


	//  Determine the sector number where the FAT ends
	//
	ulFatEnd = lpDPB->fat_start + ulFatLength;


	//  Determine the current LRU buffer
	//
	lpLRUStart = GetLRUConvMemBuffer();


	//  Determine the first buffer we will flush
	//
	lpLRUBuffer = lpLRUStart;


	//  Go through all the buffers flushing them as we go
	//
	do
	{
		//  Flush the current LRU buffer
		//
		{
			//  Determine if this buffer is valid.  If this buffer
			//  is valid ...
			//
			if( (lpLRUBuffer->ucStatusFlags & STATUS_VALID) != 0 )
			{
				//  Make sure the CRC is correct on the conventional
				//  memory buffer
				//
				CHECK_CONV_BUFFER_CRC( lpLRUBuffer );


				//  Determine if this buffer is dirty.  If this buffer
				//  is dirty ...
				//
				if( (lpLRUBuffer->ucStatusFlags & STATUS_DIRTY) != 0 )
				{
					//  As soon as we commit this sector it will not be
					//  dirty anymore.  Go ahead and set this buffer
					//  not dirty
					//
					lpLRUBuffer->ucStatusFlags &= ~STATUS_DIRTY;


					//  We only need to write the sector if we are
					//  committing the disk changes
					//
					if( DoFixErrors() )
					{
#if !(DEBUG_NO_WRITES)
						//  Get the address of the buffer
						//
						lpSectorBuffer = ((LPBUFFER)lpLRUBuffer)->cStartOfBuffer;
#endif  // #if !(DEBUG_NO_WRITES)

						//  Get the sector number
						//
						ulSectorNumber = lpLRUBuffer->ulSectorNumber;


						//  Determine the number of copies of this
						//  sector
						//
						{
							//  Determine if this is a FAT sector.  If this
							//  is a FAT sector ...
							//
							if( (ulSectorNumber >= lpDPB->fat_start) &&
									(ulSectorNumber < ulFatEnd) )
							{
								ucCopies = GetFATCopies();
							}
							else
							{
								ucCopies = 1;
							}
						}


						//  Write each copy of this sector
						//
						while( ucCopies )
						{
							//  Determine which abswrite call to use and perform
							//  the write
							//
#if DEBUG_NO_WRITES
  							printf("writing sector: %lu\n", ulSectorNumber );
							fSuccessful = TRUE;
#else   // #if DEBUG_NO_WRITES
							if( IsFat32Drive(lpDPB) )
							{
								//  Write the sectors for this copy
								//  (drive 0=default,1=A,...)
								//
								fSuccessful = dlAbsWrite32( lpDPB->drive+1, 1,
										ulSectorNumber, lpSectorBuffer,
										ABS32WRITE_UNKNOWN );
							}
							else
							{
								//  Write the sectors for this copy (drive 0=A,1=B,...)
								//
								fSuccessful = dlAbsDiskWrite( lpDPB->drive, 1,
										ulSectorNumber, lpSectorBuffer );
							}
#endif  // #if DEBUG_NO_WRITES

							//  Determine if there was an error
							//
							if(!fSuccessful)
							{
								//  ???
								//
								printf("Failed to write changes!\n");
								exit(255);
							}


							//  Mark that we wrote this copy
							//
							ucCopies--;


							//  Determine the starting sector for the next copy
							//
							ulSectorNumber += ulFatLength;
						}
					}
				}
			}
		}


		//  Determine the next buffer
		//
		lpLRUBuffer = lpLRUBuffer->lpPrev;

	}while( lpLRUBuffer != lpLRUStart );


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();
}


//  Validate that the circularly linked list is valid for the
//  conventional memory buffers
//
#if DEBUG_BUFFERS
bool CheckIfConvMemBuffOK( void )
{
	LPBUF_MGT lpLRUStart;
	LPBUF_MGT lpLRUBuffer;
	unsigned uCount = 0;

	lpLRUStart  = GetLRUConvMemBuffer();
	lpLRUBuffer = lpLRUStart;

	do{
		uCount++;
		if( uCount > uAllocatedConvMemBuffers )
		{
			printf("Conventional memory buffers corrupt!\n");
			return FALSE;
		}

		lpLRUBuffer = lpLRUBuffer->lpPrev;

	}while( lpLRUBuffer != lpLRUStart );

	if( uCount != uAllocatedConvMemBuffers )
	{
		printf("Conventional memory buffers corrupt!\n");
		return FALSE;
	}

	return TRUE;
}
#endif //#if DEBUG_BUFFERS


//  Make sure all the memory store buffers are ok
//
#if DEBUG_BUFFERS
bool CheckIfStoreMemBuffOK( void )
{
	LPSTORE_BUF_MGT lpXMSBuffMgt;
	LPSTORE_BUF_MGT lpStartBuffMgt;
	unsigned uCount = 0;


	//  Get the MRU memory store buffer
	//
	lpStartBuffMgt = GetMRUMemStoreBuffer();
	lpXMSBuffMgt   = lpStartBuffMgt;


	//  Go through all of the memory store buffers looking
	//  for a valid copy of this sector
	//
	do
	{
		uCount++;
		if( uCount > uAllocatedStoreMemBuffers )
		{
			printf("Memory store buffers corrupt!\n");
			return FALSE;
		}


		//  Get the next buffer
		//
		lpXMSBuffMgt = lpXMSBuffMgt->lpNext;

	}while( lpXMSBuffMgt != lpStartBuffMgt );


	if( uCount != uAllocatedStoreMemBuffers )
	{
		printf("Memory store buffers corrupt!\n");
		return FALSE;
	}

	return TRUE;
}
#endif //#if DEBUG_BUFFERS


//  Flush all the memory store buffers
//
void FlushAllMemoryStoreBuffers( void )
{
	LPSTORE_BUF_MGT lpXMSBuffMgt;
	LPSTORE_BUF_MGT lpStartBuffMgt;


	//  Determine the current LRU buffer
	//
	lpStartBuffMgt = GetLRUMemStoreBuffer();
	lpXMSBuffMgt   = lpStartBuffMgt;


	//  Go through all the buffers flushing them as we go
	//
	do
	{
		//  Flush the current LRU buffer
		//
		FlushLRUMemStoreBuffer( lpXMSBuffMgt );


		//  Set the LRU buffer as the MRU buffer
		//
		lpMRUXMSMemBuffer = lpXMSBuffMgt;


		//  Determine the current LRU buffer
		//
		lpXMSBuffMgt = GetLRUMemStoreBuffer();

	}while( lpXMSBuffMgt != lpStartBuffMgt );
}


//  Flush all the conventional memory buffers and all the memory
//  store buffers.
//
void FlushAllBuffers( void )
{
	//  Determine if we need to flush the buffers.  If we need
	//  to flush the buffers ...
	//
	if( DoFixErrors() )
	{
		//  Flush all the conventional memory buffers
		//
		FlushAllConventionalMemoryBuffers();


		//  Flush all the memory store buffers
		//
		FlushAllMemoryStoreBuffers();
	}
}


//  Return a handle for the specified sector
//
#pragma argsused
bool PASCAL ReadSectorBuffer( uchar ucDOSDriveNumber,
		ulong ulAbsSector, void ** ppBufferHandle )
{
	LPBUF_MGT lpConvBuffMgt;
	LPSTORE_BUF_MGT lpXMSBuffMgt;
	uchar ucDrive;
	bool fSuccessful;
	char * pTempSectorBuffer;
	LPDPB lpDPB;


	//  DEBUG
	//
#if DEBUG_SECTOR_REQUEST
	{
		static ulong ulLastSector = 0;
		if( ulLastSector != ulAbsSector )
		{
			printf("Get sector: %lxh\n", ulAbsSector );
			ulLastSector = ulAbsSector;
		}
	}
#endif  // #if DEBUG_SECTOR_REQUEST


	//  Determine if this sector is in the	conventional memory buffers.
	//  If this sector is in the conventional memory buffers, ...
	//
	if( GetConvMemBuffInfo(ulAbsSector, &lpConvBuffMgt) )
	{
		//  Set this sector as the most recently used
		//
		SetConvMemBufferAsMRU( lpConvBuffMgt );


		//  The MRU conventional memory buffer contains the
		//  requested sector
		//
		return TRUE;
	}
	else if( GetMemStoreBuffInfo(ulAbsSector, &lpXMSBuffMgt) )
	{
		//
		//  Determine if this sector is in the memory store buffers.  If
		//  this sector is in the memory store buffers, ...
		//

		//  Set the memory store buffer as the MRU conventional
		//  memory buffer
		//
		MakeMemStoreBufferMRUConvMemBuffer( lpXMSBuffMgt );
	}
	else
	{
		//  Make sure that the buffering scheme is OK
		//
		ASSERT_BUFFERS();


		//  Make a conventional memory buffer available, flushing a
		//  sector buffer as neccessary, and set this buffer as the
		//  MRU conventional memory buffer
		//
		MakeAvailableBufferMRU();


		//  Get the drive number
		//
		ucDrive = GetWorkingDriveDriveNumber();


		//  Get the address of the temporary sector buffer
		//
		pTempSectorBuffer = GetTempSectorBuffer();


		//  Get the address of the DPB so that we can have the
		//  sector size
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Read the sector from disk into the LRU conventional
		//  memory buffer
		//
		if( fIsFat32World )
		{
			fSuccessful = dlAbsRead32( ucDrive+1, 1, ulAbsSector,
					pTempSectorBuffer );
		}
		else
		{
			fSuccessful = dlAbsDiskRead( ucDrive, 1, ulAbsSector,
					pTempSectorBuffer );
		}
		if(!fSuccessful)
		{
			//  ???
			//
			printf("Error reading sector!\n");
//			printf( "ulAbsSector: %lxh\n", ulAbsSector );
			exit(255);
		}


		//  Copy the sector data into the buffer
		//
		_fmemcpy( BUFFER_ADDR(lpMRUConvMemBuffer), pTempSectorBuffer,
				lpDPB->bytes_per_sec );


		//  Set the sector number in the buffer management structure
		//
		lpMRUConvMemBuffer->ulSectorNumber = ulAbsSector;


		//  Mark the MRU buffer as valid and not dirty
		//
		lpMRUConvMemBuffer->ucStatusFlags = STATUS_VALID;


		//  Calculate the CRC for the conventional memory buffer
		//
		COMPUTE_CONV_BUFFER_CRC( lpMRUConvMemBuffer );


		//  Make sure the CRC is correct on the conventional memory buffer
		//
		CHECK_CONV_BUFFER_CRC( lpMRUConvMemBuffer );


		//  Make sure that the buffering scheme is OK
		//
		ASSERT_BUFFERS();
	}


	//  The MRU conventional memory buffer contains the
	//  requested sector
	//
	return TRUE;
}


//  Return the address of the buffer for the specified handle
//
#pragma argsused
bool PASCAL GetBufferDataAddress( void * pBufferHandle,
		void far * * plpBufferAddress )
{
	//
	//  Presently we are not using handles.  How do you know if
	//  handle is stale?  So instead it is assumed that the
	//  address you are seeking is for what was just set as
	//  the MRU sector when you first requested the sector.
	//

	//  Set the address of the request sector in the return variable
	//
	*plpBufferAddress = BUFFER_ADDR(lpMRUConvMemBuffer);

	return TRUE;
}


//  Required by the FAT libraries but not used in chkdsk
//
#pragma argsused
bool PASCAL SetBufferCopiesAndDistance( void * pBufferHandle,
		unsigned uCopiesOfThisSector, ulong ulDistanceToNextSector )
{
	//
	//  This feature of the FAT libraries is not being used
	//  in this implementation of chkdsk because we might
	//  not be writing multiple FAT copies.
	//

	return TRUE;
}


//  Mark the MRU buffer as modified
//
#pragma argsused
bool PASCAL MarkBufferDirty( void * pBufferHandle )
{
	//
	//  Presently we are not using handles.  How do you know if
	//  handle is stale?  So instead it is assumed that the
	//  address you are seeking is for what was just set as
	//  the MRU sector when you first requested the sector.
	//


#if DEBUG_MODIFIED
	if( (lpMRUConvMemBuffer->ucStatusFlags & STATUS_DIRTY) == 0 )
	{
		printf("Sector buffer %lu is dirty\n",
				lpMRUConvMemBuffer->ulSectorNumber );
	}
#endif  // #if DEBUG_MODIFIED


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();


	//  Set the MRU buffer as modified
	//
	lpMRUConvMemBuffer->ucStatusFlags |= STATUS_DIRTY;


	//  Calculate the CRC for the conventional memory buffer
	//
	COMPUTE_CONV_BUFFER_CRC( lpMRUConvMemBuffer );


	//  Make sure the CRC is correct on the conventional memory buffer
	//
	CHECK_CONV_BUFFER_CRC( lpMRUConvMemBuffer );


	//  Make sure that the buffering scheme is OK
	//
	ASSERT_BUFFERS();


	return TRUE;
}


//  Return the address of a temporary sector buffer
//
char * GetTempSectorBuffer( void )
{
	static char cTempSectorBuffer[MAX_SECTOR_SIZE];
	return cTempSectorBuffer;
}



#if DEBUG_CRC_BUFFERS

//  Compute the 32-bit CRC value for a given buffer and set it
//  in the buffer header structure
//
void ComputeConvBufferCRC( LPBUF_MGT lpConvBuffMgt )
{
	static unsigned uCRCSize = 0;
	LPDPB lpDPB;
	ulong ulTempCRC32 = 0;


	//  Determine if this is the first time this functiion has been called
	//
	if( !uCRCSize )
	{
		//  Get the DBP for the drive we are checking
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Get the sector size of the drive we are checking
		//
		uCRCSize = lpDPB->bytes_per_sec;
	}


	//  Determine the CRC value for the buffer
	//
	dlGetCRC32( &ulTempCRC32, (char far *)(BUFFER_ADDR(lpConvBuffMgt)),
			uCRCSize );


	//  DEBUG
	//
#if 0
	printf("ComputeConvBufferCRC: %08lx\n", ulTempCRC32 );
#endif


	//  Set the CRC value in the buffer header structure
	//
	lpConvBuffMgt->ulCRC32 = ulTempCRC32;
}


//  Compute the 32-bit CRC value for a given buffer and make sure
//  it matches what is set in the buffer header structure
//
void CheckConvBufferCRC( LPBUF_MGT lpConvBuffMgt, unsigned uLineNumber )
{
	static unsigned uCRCSize = 0;
	LPDPB lpDPB;
	ulong ulTempCRC32 = 0;


	//  Determine if this is the first time this functiion has been called
	//
	if( !uCRCSize )
	{
		//  Get the DBP for the drive we are checking
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Get the sector size of the drive we are checking
		//
		uCRCSize = lpDPB->bytes_per_sec;
	}


	//  Determine the CRC value for the buffer
	//
	dlGetCRC32( &ulTempCRC32, (char far *)(BUFFER_ADDR(lpConvBuffMgt)),
			uCRCSize );


	//  Set the CRC value in the buffer header structure
	//
	if( lpConvBuffMgt->ulCRC32 != ulTempCRC32 )
	{
		//  DEBUG
		//
#if 0
		printf("CheckConvBufferCRC   result: %08lx\n", ulTempCRC32 );
		printf("CheckConvBufferCRC expected: %08lx\n", lpConvBuffMgt->ulCRC32 );
#endif


		CHKDSKInternalError( __FILENUMBER__, uLineNumber );
	}
}

#endif //  #if DEBUG_CRC_BUFFERS



#if DEBUG_CRC_BUFFERS

//  Compute the 32-bit CRC value for a given buffer and set it
//  in the buffer header structure
//
void ComputeStoreBufferCRC( LPSTORE_BUF_MGT lpXMSBuffMgt, char * pSectorBuffer )
{
	static unsigned uCRCSize = 0;
	LPDPB lpDPB;
	ulong ulTempCRC32 = 0;


	//  Determine if this is the first time this functiion has been called
	//
	if( !uCRCSize )
	{
		//  Get the DBP for the drive we are checking
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Get the sector size of the drive we are checking
		//
		uCRCSize = lpDPB->bytes_per_sec;
	}


	//  Determine the CRC value for the buffer
	//
	dlGetCRC32( &ulTempCRC32, pSectorBuffer, uCRCSize );


	//  DEBUG
	//
#if 0
	printf("ComputeStoreBufferCRC: %08lx\n", ulTempCRC32 );
#endif


	//  Set the CRC value in the buffer header structure
	//
	lpXMSBuffMgt->ulCRC32 = ulTempCRC32;
}


//  Compute the 32-bit CRC value for a given buffer and make sure
//  it matches what is set in the buffer header structure
//
void CheckStoreBufferCRC( LPSTORE_BUF_MGT lpXMSBuffMgt, char * pSectorBuffer,
		unsigned uLineNumber )
{
	static unsigned uCRCSize = 0;
	LPDPB lpDPB;
	ulong ulTempCRC32 = 0;


	//  Determine if this is the first time this functiion has been called
	//
	if( !uCRCSize )
	{
		//  Get the DBP for the drive we are checking
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Get the sector size of the drive we are checking
		//
		uCRCSize = lpDPB->bytes_per_sec;
	}


	//  Determine the CRC value for the buffer
	//
	dlGetCRC32( &ulTempCRC32, pSectorBuffer, uCRCSize );


	//  Set the CRC value in the buffer header structure
	//
	if( lpXMSBuffMgt->ulCRC32 != ulTempCRC32 )
	{
		//  DEBUG
		//
#if DEBUG_CRC_BUFFERS
		printf("CheckStoreBufferCRC   result: %08lx\n", ulTempCRC32 );
		printf("CheckStoreBufferCRC expected: %08lx\n", lpXMSBuffMgt->ulCRC32 );
#endif

		CHKDSKInternalError( __FILENUMBER__, uLineNumber );
	}
}

#endif //  #if DEBUG_CRC_BUFFERS

