//
//   $Id: CHKDSK.H 1.8 2002/10/16 21:40:24Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  CHKDSK.H
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//  03-15-2000  bt  Created (Started over from scratch)
//  03-23-1989  Drew Gislason (created old chkdsk)
//


//  Debugging defines
//
#define DEBUG_BUFFERS					FALSE
#define DEBUG_STOREBUFFERS				FALSE
#define SHOW_SCANDISK_MSG				FALSE
#define DEBUG_NO_WRITES					FALSE
#define DEBUG_MODIFIED					FALSE
#define DEBUG_ASSERTS					FALSE
#define DEBUG_VIEW_CACHE_HITS			FALSE
#define DEBUG_SECTOR_REQUEST			FALSE
#define DEBUG_CRC_BUFFERS				FALSE
#define DEBUG_STEP_MESSAGES			FALSE


//  Optimization variables
//
#define MAX_CONV_MEM_BUFFERS		500
#define MAX_XMS_MEM_BUFFERS		100



//  Debug macros
//


#if DEBUG_STEP_MESSAGES

#define DEBUG_STEP_MESSAGE( szMessage )  printf( szMessage ); printf("\n");

#else  // #if DEBUG_STEP_MESSAGES

#define DEBUG_STEP_MESSAGE( szMessage )

#endif  // #if DEBUG_STEP_MESSAGES


//  Make sure that the buffering scheme is OK
//
#if DEBUG_BUFFERS

#define ASSERT_BUFFERS()                     \
{                                            \
	if( !CheckIfConvMemBuffOK() )             \
	{                                         \
		printf("Convention Memory buffer ");	\
		printf("Error line: %u", __LINE__ );   \
		exit(255);                             \
	}                                         \
}

#else  // #if DEBUG_BUFFERS
#define ASSERT_BUFFERS() {}
#endif  // #if DEBUG_BUFFERS


//  Make sure that the buffering scheme is OK
//
#if DEBUG_STOREBUFFERS

#if DEBUG_BUFFERS
#define ASSERT_STOREBUFFERS()                \
{                                            \
	if( !CheckIfStoreMemBuffOK() )            \
	{                                         \
		printf("Memory store buffer ");			\
		printf("Error line: %u", __LINE__ );   \
		exit(255);                             \
	}                                         \
}
#else  // #if DEBUG_BUFFERS
#define ASSERT_STOREBUFFERS() {}
#endif  // #if DEBUG_BUFFERS

#else  // #if DEBUG_STOREBUFFERS
#define ASSERT_STOREBUFFERS() {}
#endif  // #if DEBUG_STOREBUFFERS


//  Macros
//
#define	CHKDSK_INTERNAL_ERROR()					\
			CHKDSKInternalError( __FILENUMBER__, __LINE__ );


//
//
typedef struct BUF_MGT BUF_MGT;
typedef BUF_MGT * PBUF_MGT;
typedef BUF_MGT far * LPBUF_MGT;
struct BUF_MGT {
	LPBUF_MGT lpNext;				// 00
	LPBUF_MGT lpPrev;				// 04
	ulong ulSectorNumber;		// 08
	uchar ucStatusFlags;			// 0c
	uchar ucReserved;				// 0d
#if DEBUG_CRC_BUFFERS
	ulong ulCRC32;					// 0e
#endif //  #if DEBUG_CRC_BUFFERS
};


//
//
typedef struct BUFFER BUFFER;
typedef BUFFER * PBUFFER;
typedef BUFFER far * LPBUFFER;
struct BUFFER {
	BUF_MGT	BuffMgt;
	char		cStartOfBuffer[];
};


//
//
typedef struct STORE_BUF_MGT STORE_BUF_MGT;
typedef STORE_BUF_MGT * PSTORE_BUF_MGT;
typedef STORE_BUF_MGT far * LPSTORE_BUF_MGT;
struct STORE_BUF_MGT {
	LPSTORE_BUF_MGT lpNext;		// 00
	LPSTORE_BUF_MGT lpPrev;		// 04
	ulong ulSectorNumber;		// 08
	uchar ucStatusFlags;			// 0c
	uchar ucReserved;				// 0d
	ulong ulOffsetInStore;		// 0e
#if DEBUG_CRC_BUFFERS
	ulong ulCRC32;					// 12
#endif //  #if DEBUG_CRC_BUFFERS
};


//  This is a macro to get at the conventional memory buffer address
//
#define BUFFER_ADDR(lpBuffMgt) (((LPBUFFER)lpBuffMgt)->cStartOfBuffer)


//  Status flags
//
#define STATUS_AVAILABLE	0x00
#define STATUS_VALID			0x01
#define STATUS_DIRTY			0x02
//#define 0x04
//#define 0x08
//#define 0x10
//#define 0x20
//#define 0x40
//#define 0x80


//  These constants are for the buffering requirements
//
#define MIN_CONV_BUFFERS			3
#define VISITED_LIST_SWAP_SIZE   256


//  This is used to determine the amount of free
//  conventional memory
//
#define MAX_PARAGRAPH_SIZE		0xffff


//  This is the size of a paragraph in bytes
//
#define PARAGRAPH_SIZE			0x10


//  Typedefs for packed date and time
//
typedef struct DOSFileDate * pPACKEDDATE;
typedef struct DOSFileTime * pPACKEDTIME;


//
//
#define FIRST_PASS		TRUE
#define SECOND_PASS		FALSE



//  Error codes
//
#define ERR_NONE			0


//  Indexes into the command line switches array
//
#define DONT_CONFIRM		0
#define FIX_ERRORS		1
#define SHOW_VERBOSE		2
#define HIDE_SIGNON		3
#define SWITCH_COUNT		4


//  Macros for using boolean switches passed in on the command line
//
extern bool fSwitches[SWITCH_COUNT];
#define DoConfirm()				(fSwitches[DONT_CONFIRM]     == 0)
#define DoFixErrors()			(fSwitches[FIX_ERRORS]       != 0)
#define DoShowVerbose()			(fSwitches[SHOW_VERBOSE]     != 0)
#define DoHideSignon()			(fSwitches[HIDE_SIGNON]      != 0)


//  Defualt answers for when no confirm option is used
//
#define DEFAULT_YES		'Y'
#define DEFAULT_NO		'N'


//  Error codes for name validation
//
#define NAME_ERROR_NONE					0
#define NAME_ERROR_CASE					1
#define NAME_ERROR_INVALID				2


/* Statistics information */
struct stat
{
	ulong 	bytes;
	unsigned number;
	bool		fInKBytes;
};


//  Indexes into the statistics array of strings
//
#define DTOTAL				0
#define DHIDDEN			1
#define DDIR				2
#define DUSER				3
#define DLOST				4
#define DBAD				5
#define DAVAIL				6
#define DALLOCSIZE		7
#define DALLOCTTOTAL		8
#define DALLOCAVAIL		9
#define MTOTAL				10
#define MAVAIL				11
#define MAXSTAT			12


//  This structure is passed to DOS to retrieve the
//  serial number and other disk information
//
typedef struct DiskInfo DISK_INFO;
typedef DISK_INFO * pDISK_INFO;
typedef DISK_INFO far * lpDISK_INFO;
struct DiskInfo{
	unsigned uInfoLevel;
	ulong    ulSerialNumber;
	char		cVolumeLabel[11];
	char		cFatTypeString[8];
};


//  Parameter block for setting drive free space
//
typedef struct SETFREEPARM SETFREEPARM;
struct SETFREEPARM{
	unsigned	uSize;
	unsigned	uVersion;
	ulong		ulFunction;
	ulong		ulFreeCount;
	ulong		ulNextFree;
	ulong		ulReserved[2];
};


//  This is the maximum sector size
//
#define MAX_SECTOR_SIZE			1024


//  This is used to determine the amount of free
//  conventional memory
//
#define MAX_PARAGRAPH_SIZE		0xffff


//  This is the size of a paragraph in bytes
//
#define PARAGRAPH_SIZE			0x10


//  These are constants for evaluating the visited list entries
//
#define VL_USED					0x01
#define VL_CROSS_LINKED			0x02
#define VL_DO_COPY				0x04
#define VL_IN_PROCESS			0x08


//  These are constants returned after following a chain
//
#define CHAIN_NO_ERROR			0x00
#define CHAIN_INVALID			0x01
#define CHAIN_CROSS_LINKED		0x02
#define CHAIN_FIRST_INVALID	0x04
#define CHAIN_ADJUST_SIZE		0x08

#define GENERAL_ERROR				0x01
#define INVALID_DIRECTORY			0xff
#define EXIT_CODE_FOUND_ERRORS	0xff

//  Macros for examining cache and disk hits
//
#if DEBUG_VIEW_CACHE_HITS

#define CONV_MEM_BUFF_CACHE_HIT()	ulConvMemBuffCacheHit++
#define MEM_STORE_BUFF_CACHE_HIT()	ulMemStoreBuffCacheHit++
#define SECTOR_REQUEST_HIT()			ulSectorRequests++

#else

#define SECTOR_REQUEST_HIT()
#define CONV_MEM_BUFF_CACHE_HIT()
#define MEM_STORE_BUFF_CACHE_HIT()

#endif


//  Macros for computing 32-bit CRCs on conventional memory and
//  memory store buffers
//
#if DEBUG_CRC_BUFFERS

#define COMPUTE_CONV_BUFFER_CRC( lpBuffer ) ComputeConvBufferCRC( lpBuffer )
#define CHECK_CONV_BUFFER_CRC( lpBuffer )   CheckConvBufferCRC( lpBuffer, __LINE__ )
#define COMPUTE_STORE_BUFFER_CRC( pBuffer, pSectorData ) ComputeStoreBufferCRC( pBuffer, pSectorData )
#define CHECK_STORE_BUFFER_CRC( pBuffer, pSectorData )   CheckStoreBufferCRC( pBuffer, pSectorData, __LINE__ )

#else // #if DEBUG_CRC_BUFFERS

#define COMPUTE_CONV_BUFFER_CRC( lpBuffer )
#define CHECK_CONV_BUFFER_CRC( lpBuffer )
#define COMPUTE_STORE_BUFFER_CRC( pBuffer, pSectorData )
#define CHECK_STORE_BUFFER_CRC( pBuffer, pSectorData )

#endif // #if DEBUG_CRC_BUFFERS


extern bool fIsFat32World;
extern bool fLFNsSupported;


//  Global functions
//
//void FreeMRUConvBufferAndMakeItLRU( void );
