//
//   $Id: chkdsk.C 1.23 2003/03/14 17:12:28Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  CHKDSK.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//  03-15-2000  bt  Created (Started over from scratch)
//  03-23-1989  Drew Gislason (created old chkdsk)
//


#define __FILENUMBER__		1


//  Borland libraries
//
#include <dir.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <mem.h>
#include <string.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "fatworld.h"
#include "dosver.h"
#include "dlapp.h"
#include "fatlib.h"
#include "dlpath.h"



//  Chkdsk header files
//
#include "chkdsk.h"
#include "chkasert.h"


//  Prototypes
//
#include "chkdsk.pt"



//
//  Debugging defines (in chkdsk.h)
//



//  External Message strings
//
#define CHKDSK		1
#include "dosmsg.c"


//  Variables
//

bool fIsFat32World  = FALSE;
bool fLFNsSupported = FALSE;
char * szCommandLinePath = 0;

//  Array of boolean command line switches.  Accessed through macros.
//
bool fSwitches[SWITCH_COUNT];

/* Statistics information */
struct stat stat[MAXSTAT];


//  CHKDSK uses a lot of stack space.  Increase the stack size.
//
unsigned _stklen = 0xD000;



//  Start of chkdsk
//
uchar main( int argc, char *argv[] )
{
	uchar ucExitCode = ERR_NONE;
	uchar ucReturnCode;


	dlResetDisk();
	DEBUG_STEP_MESSAGE( "main()" );


	//  Determine if this is DOS 7.1
	//
	fIsFat32World = dlIsFat32World();


	//  Determine if LFNs are support
	//
	fLFNsSupported = LFNPresent();


	DEBUG_STEP_MESSAGE( "SetInterruptHandlers()" );


	//  Set interrupt vectors
	//
	SetInterruptHandlers();


	DEBUG_STEP_MESSAGE( "DoParseCommandLine( argc-1, &argv[1] )" );


	//  Parse the command line
	//
	DoParseCommandLine( argc-1, &argv[1] );


	DEBUG_STEP_MESSAGE( "DisplaySignOnMessage()" );


	//  Display the signon message
	//
	DisplaySignOnMessage();


	DEBUG_STEP_MESSAGE( "IsAValidDrive()" );


	//  Determine if the drive is valid.  If the drive is not
	//  valid, display an error message, set the appropriate
	//  error code and exit
	//
	if( !IsAValidDrive() )
	{
		printf( szErrInvalidDrive );

		//  MSDOS 7.1 chkdsk does not set an error code
		//
		exit(0);
	}


	DEBUG_STEP_MESSAGE( "IsASubstDrive()" );


	//  Determine if this is a subst drive.  If the drive is a subst
	//  drive, display an error message, set the appropriate error
	//  code and exit.
	//
	if( IsASubstDrive() )
	{
		printf( szErrSubstDrive );

		//  MSDOS 7.1 chkdsk does not set an error code
		//
		exit(0);
	}


	DEBUG_STEP_MESSAGE( "IsARemoteDrive()" );


	//  Determine if the drive is local or remote.  If the drive is
	//  remote, display an error message, set the appropriate error
	//  code and exit
	//
	if( IsARemoteDrive() )
	{
		printf( szErrRemoteDrive );

		//  MSDOS 7.1 chkdsk does not set an error code
		//
		exit(0);
	}


	DEBUG_STEP_MESSAGE( "SaveCurrentDrive()" );


	//  Save the current drive
	//
	SaveCurrentDrive();


	DEBUG_STEP_MESSAGE( "SaveCurrentDirectory()" );


	//  Save the current directory
	//
	SaveCurrentDirectory();


	dlResetDisk();


	DEBUG_STEP_MESSAGE( "InitializeBuffers()" );


	//  Initialize the buffers
	//
	InitializeBuffers();


	DEBUG_STEP_MESSAGE( "LockDrive()" );


	//  Lock the logical disk
	//
	LockDrive();


#if 0
	{
		char far * lpSector;
		uchar ucCount;
		ulong ulSectorNumber = 0;


		for( ucCount = 0; ucCount < 6; ucCount++ )
		{
			lpSector = GetSectorAddress( ulSectorNumber + ucCount );
		}

		lpSector = GetSectorAddress( ulSectorNumber );
		lpSector = GetSectorAddress( ulSectorNumber );
	}
#endif


	DEBUG_STEP_MESSAGE( "CheckUserAbort()" );


	//  Check if the user wants to abort CTRL-C
	//
	CheckUserAbort();


	DEBUG_STEP_MESSAGE( "DisplayVolumeInfo()" );


	//  Display information about the disk like the disk label
	//  and serial number.
	//
	DisplayVolumeInfo();


	DEBUG_STEP_MESSAGE( "SyncronizeFATCopies()" );


	//  Fix FAT copies
	//
	if( SyncronizeFATCopies() )
	{
		ucExitCode = EXIT_CODE_FOUND_ERRORS;
	}


	DEBUG_STEP_MESSAGE( "CheckUserAbort()" );


	//  Check if the user wants to abort CTRL-C
	//
	CheckUserAbort();


	DEBUG_STEP_MESSAGE( "CountRootEntries()" );


	//  Count the number of root entries that are available
	//
	CountRootEntries();


	DEBUG_STEP_MESSAGE( "CheckUserAbort()" );


	//  Check if the user wants to abort CTRL-C
	//
	CheckUserAbort();


	DEBUG_STEP_MESSAGE( "CheckTheEntireDrive(FIRST_PASS)" );


	//  Check the drive
	//
	ucReturnCode = CheckTheEntireDrive(FIRST_PASS);
	if( ucReturnCode )
	{
		printf("\n");
		ucExitCode = EXIT_CODE_FOUND_ERRORS;
	}


	DEBUG_STEP_MESSAGE( "CheckUserAbort()" );


	//  Check if the user wants to abort CTRL-C
	//
	CheckUserAbort();


	DEBUG_STEP_MESSAGE( "FixLostClusters()" );


	//  Fix lost clusters
	//
	if( FixLostClusters() )
	{
		ucExitCode = EXIT_CODE_FOUND_ERRORS;
	}


	DEBUG_STEP_MESSAGE( "CheckUserAbort()" );


	//  Check if the user wants to abort CTRL-C
	//
	CheckUserAbort();


	DEBUG_STEP_MESSAGE( "CheckTheEntireDrive(SECOND_PASS)" );


	//  Fix cross-linked files
	//
	ucReturnCode = CheckTheEntireDrive(SECOND_PASS);
	if( ucReturnCode )
	{
		printf("\n");
		ucExitCode = EXIT_CODE_FOUND_ERRORS;
	}


	DEBUG_STEP_MESSAGE( "CheckUserAbort()" );


	//  Check if the user wants to abort CTRL-C
	//
	CheckUserAbort();


	DEBUG_STEP_MESSAGE( "FlushAllBuffers()" );


	//  Commit all our sector buffers
	//
	FlushAllBuffers();


	//  Free all the memory used by our sector buffers
	//
	if( !DoFixErrors() )
	{
		DEBUG_STEP_MESSAGE( "ReleaseAllMemory()" );


		ReleaseAllMemory();
	}


	//  Initialize the buffers.  This will reset all previously buffered
	//  information so that we will read unmodified sectors off the disk.
	//  In fix mode however, what is in the buffers should be what is on
	//  the disk.
	//
	if( !DoFixErrors() )
	{
		DEBUG_STEP_MESSAGE( "InitializeBuffers()" );


		InitializeBuffers();
	}


	DEBUG_STEP_MESSAGE( "FixFAT32DriveFreespace()" );


	//  Check FAT32 drive freespace
	//
	if( FixFAT32DriveFreespace() )
	{
		ucExitCode = EXIT_CODE_FOUND_ERRORS;
	}


	DEBUG_STEP_MESSAGE( "CheckUserAbort()" );


	//  Check if the user wants to abort CTRL-C
	//
	CheckUserAbort();


	DEBUG_STEP_MESSAGE( "UnlockDrive()" );


	//  Unlock the logical disk
	//
	UnlockDrive();


	DEBUG_STEP_MESSAGE( "DisplaySummary()" );


	//  Display the chkdsk drive usage summary
	//
	if( !ucExitCode )
	{
		printf("\n");
	}
	DisplaySummary();


	DEBUG_STEP_MESSAGE( "CheckUserAbort()" );


	//  Check if the user wants to abort CTRL-C
	//
	CheckUserAbort();


	DEBUG_STEP_MESSAGE( "ShowContigiousFileInfo()" );


	//  Display contigious file information
	//
	ShowContigiousFileInfo();


	DEBUG_STEP_MESSAGE( "ReleaseAllMemory()" );


	//  Free all the memory used by our sector buffers
	//
	ReleaseAllMemory();


	//  Display a message recomending the use of scandisk
	//
	ShowScandiskMsg();


	DEBUG_STEP_MESSAGE( "ReleaseAllMemory()" );


	//  Restore the users current drive
	//
	RestoreCurrentDrive();


	DEBUG_STEP_MESSAGE( "RestoreCurrentWorkingDirectory()" );


	//  Restore the users current working directory
	//
	RestoreCurrentWorkingDirectory();


	DEBUG_STEP_MESSAGE( "RestoreInterruptVectors()" );


	//  Restore interrupt vectors
	//
	RestoreInterruptVectors();


	DEBUG_STEP_MESSAGE( "exit( ucExitCode )" );


	dlResetDisk();


	//  Set the error code and exit
	//
	return ucExitCode;
}


//  Flush buffers, free memory, unlock the drive, restore current
//  drive and directory, set exit code and exit
//
void ErrorExit( uchar ucErrorCode )
{
	//  Commit all our sector buffers
	//
	FlushAllBuffers();


	//  Free all the memory used by our sector buffers
	//
	ReleaseAllMemory();


	//  Unlock the logical disk
	//
	UnlockDrive();


	//  Restore the users current drive
	//
	RestoreCurrentDrive();


	//  Restore the users current working directory
	//
	RestoreCurrentWorkingDirectory();


	//  Restore interrupt vectors
	//
	RestoreInterruptVectors();


	dlResetDisk();


	//  Set the error code and exit
	//
	exit( ucErrorCode );
}


//  Get the address of our local copy of the DPB
//
bool PASCAL GetDPB( LPDPB * plpDPB )
{
	static bool fDPBRetrieved = FALSE;
	static bool fSuccessful = FALSE;
	static char pBuffer[sizeof(DPB)+2];
	static PDPB pDPB;
	uchar ucDrive;


	//  Get the drive number
	//
	ucDrive = GetWorkingDriveDriveNumber();


	//  Determine if we have retrieved the DPB already.  If we have
	//  not retrieved the DPB ...
	//
	if( !fDPBRetrieved )
	{
		//  Set the address of the DPB;
		//
		pDPB = (DPB*)(pBuffer+2);


		//  Determine if we are in DOS 7.1.  If we are in DOS 7.1 ...
		//
		if( fIsFat32World )
		{
			//  Get the DPB for this drive
			//
			fSuccessful = dlGetDriveDPB( ucDrive+1, plpDPB );
			fSuccessful = dlGetExtDPB( ucDrive+1, sizeof(DPB)+2, pBuffer );
		}


		//  If we were not successful at getting the DPB in DOS 7.1
		//  (an occasional weird windows fluke) or if we are not in
		//  a 7.1 world ...
		//
		if(!fSuccessful)
		{
			//  Get the DPB for this drive
			//
			fSuccessful = dlGetDriveDPB( ucDrive+1, plpDPB );


			//  If we were successful at retrieving the DPB ...
			//
			if( fSuccessful )
			{
				//  Copy the DPB into our local copy
				//
				_fmemcpy( pDPB, *plpDPB, sizeof(DPB) );
			}
		}
	}


	//  Set the address of our local copy of the DPB in the return variable
	//
	if( fSuccessful )
	{
		*plpDPB = pDPB;
	}


	//  Set a flag indicating that we already got (or attempted to get)
	//  the DPB
	//
	fDPBRetrieved = TRUE;


	//  Return whether the function was successful
	//
	return fSuccessful;
}


//  Check if the FAT32 drive freespace is correct.  Fix it if it
//  is not
//
bool FixFAT32DriveFreespace( void )
{
	GDF32 DiskSpace;
	SETFREEPARM pParmBlock;
	uchar ucDrive;
	LPDPB lpDPB;
	bool fFreespaceIncorrect = FALSE;
	ulong ulCluster;
	ulong ulNextCluster;
	ulong ulMaxClusterNumber;
	ulong ulAvailableClusters = 0;


	//  Get the drive number
	//
	ucDrive = GetWorkingDriveDriveNumber();


	//  Get the DPB for the drive
	//
	GetDPB( &lpDPB );
	ASSERT_LPDPB( lpDPB );


	//  We only need to do this if we are checking a FAT32 drive
	//
	if( !IsFat32Drive(lpDPB) )
	{
		return fFreespaceIncorrect;
	}


	//  Determine the maximum cluster number on this drive
	//
	ulMaxClusterNumber = lpDPB->ulMaxCluster;


	//  Display a message stating that chkdsk is checking the freespace
	//  and that this could take a while
	//
	printf( szChkFreeSpace );


	//  Count all the free clusters on the drive
	//
	for( ulCluster = 2; ulCluster <= ulMaxClusterNumber; ulCluster++ )
	{
		//  Determine if the FAT entry for this cluster is available.
		//  If the FAT entry for this cluster is not available ...
		//
		GetFatEntry( lpDPB, ulCluster, &ulNextCluster );


		//  Determine if this cluster is available.  If this cluster
		//  is available ...
		//
		if( !ulNextCluster )
		{
			ulAvailableClusters++;
		}


		//  Check if the user wants to abort CTRL-C
		//
		CheckUserAbort();
	}


	//  Determine what DOS thinks the freespace is
	//
	SmartGetFreeSpace( ucDrive+1, &DiskSpace );


	//  Determine if the freespace is the same as what we calculated
	//
	if( ulAvailableClusters != DiskSpace.ulAvailableClusters )
	{
		//  Show the message stating that we found errors on
		//  this disk, if we are not in fix mode and we
		//  have not already shown this message.
		//
		ShowFoundErrorsMsg();


		//  Display a message stating that the freespace was incorrect
		//  but has been corrected.
		//
		printf(szIncorrectFreeSpace);


		//  Set a flag indicating that the freespace was incorrect
		//
		fFreespaceIncorrect = TRUE;


		//  Correct the drive freespace (function is misnamed).  This
		//  is Int 21h fn 7304 subfn 00h
		//
		if( DoFixErrors() )
		{
#if DEBUG_NO_WRITES
         printf( "... was %lu, fix to %lu\n",
                 DiskSpace.ulAvailableClusters, ulAvailableClusters );
#else
			pParmBlock.uSize = sizeof( SETFREEPARM );
			pParmBlock.uVersion = 0;
			pParmBlock.ulFunction = 0;
			pParmBlock.ulFreeCount = ulAvailableClusters;
			pParmBlock.ulNextFree = 2;
			dlSetDBPForFormat( ucDrive+1, &pParmBlock );
#endif // #if DEBUG_NO_WRITES
		}
	}


	//  Display a blank line
	//
	printf("\n");


	//  Return whether the free space was incorrect
	//
	return fFreespaceIncorrect;
}


//  Display contigious file information
//
void ShowContigiousFileInfo( void )
{
	static char szExpandedPath[255];
	LPDPB lpDPB;
	FIND Find;
	bool fSuccessful;
	ulong ulNonContigiousCount;
	ulong ulCluster;
	ulong ulNextCluster;
	ulong ulMaxClusterNumber;
	LPDIR lpDir;
	unsigned uEntryIndex;
	ulong ulSectorNumber;
	unsigned uEntriesPerSector;


	//  Determine if we need to show in contigious file information
	//
	if((szCommandLinePath) && (szCommandLinePath[2] != 0))
	{
		//  Get the DPB for the drive
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Determine the maximum cluster number on this drive
		//
		if( IsFat32Drive(lpDPB) )
		{
			ulMaxClusterNumber = lpDPB->ulMaxCluster;
		}
		else
		{
			ulMaxClusterNumber = lpDPB->hiclus;
		}


		//  Display a blank line for clarity
		//
		printf("\n");


		//  Get the expanded path for the one given on the command line
		//
//		printf("szCommandLinePath: %s\n", szCommandLinePath );
		SmartExpandPath( 1, szCommandLinePath, szExpandedPath );
//		printf(" szExpandedPath: %s\n", szExpandedPath );


		//  Set the DTA to our FIND block for the find first
		//
		dlSetDTA( &Find );


		//  Determine if any files match this mask
		//
		fSuccessful = dlFindFirstFile( A_ALL & (~(A_VOLUME | A_SUBDIR)),
				szExpandedPath );
		if( !fSuccessful )
		{
			//  Display an error message stating that no files matched
			//  this search mask
			//
			printf( szFileNotFound );
		}
		else
		{
			//  Check if the user wants to abort CTRL-C
			//
			CheckUserAbort();


			//  Remove the name part of the expanded path
			//
			RemoveName( szExpandedPath );


			//  Determine the number of direcory entries in one sector
			//
			uEntriesPerSector = lpDPB->bytes_per_sec / sizeof(DIR);


			//  Display any contigious file information
			//
			while( fSuccessful )
			{
				//  Files which do not exceed the size of a cluster will
				//  not have multiple clusters allocated to them.  Determine
				//  if this file has multiple clusters allocated to it.
				//  If this file has multiple clusters allocated to it ...
				//
				//  TD 1/24/01 However, for MS compatibility we print the
				//  "All file(s) contiguous" message even in this case.
				//  (Appropriate "if" statement removed.)

				//  Get the cluster number of the root if this is a FAT32
				//  drive and the cluster number is zero
				//
				ulCluster = Find.ulStartingCluster;
				if( (ulCluster == 0) && IsFat32Drive(lpDPB) )
				{
					ulCluster = lpDPB->ulRootDirStartCluster;
				}


				//  Get the starting sector number for this cluster
				//
				if( ulCluster )
				{
          		ulSectorNumber = ulCluster - 2;
           		ulSectorNumber <<= lpDPB->shift;

					if( IsFat32Drive(lpDPB) )
					{
						ulSectorNumber += lpDPB->ulFirstDataSector;
					}
					else
					{
						ulSectorNumber += lpDPB->data_start;
					}
				}
				else
				{
					ulSectorNumber = lpDPB->dir_start;
				}


				//  Adjust the entry index and sector number so that
				//  it is an index into a sector
				//
				ulSectorNumber += Find.dirnum / uEntriesPerSector;
				uEntryIndex = Find.dirnum % uEntriesPerSector;


				//  Get the starting cluster number for this directory
				//  entry
				//
				lpDir = GetDirEntryAtSector( uEntryIndex, ulSectorNumber );
				if( IsFat32Drive(lpDPB) )
				{
					ulCluster = Fat32StartLP(lpDir);
				}
				else
				{
					ulCluster = lpDir->uStartingClusterLo;
				}


				//  Determine the number of non-contigious clusters
				//
				ulNonContigiousCount = 0;
				while( TRUE )
				{
					//  Check if the user wants to abort CTRL-C
					//
					CheckUserAbort();


					//  Determine if this cluster is pointing to
					//  another cluster.  If this cluster is not
					//  pointing to another cluster then we are
					//  through counting contigious clusters.
					//
					if( (ulCluster < 2) ||
							(ulCluster > ulMaxClusterNumber) )
					{
						break;
					}


					//  Get the FAT entry for this cluster
					//
					GetFatEntry( lpDPB, ulCluster, &ulNextCluster );


					//  Determine if this cluster is pointing to
					//  another cluster.  If this cluster is not
					//  pointing to another cluster then we are
					//  through counting contigious clusters.
					//
					if( (ulNextCluster < 2) ||
							(ulNextCluster > ulMaxClusterNumber) )
					{
						break;
					}


					//  Determine if the next cluster is non-contigious
					//
					if( ulCluster+1 != ulNextCluster )
					{
						ulNonContigiousCount++;
					}


					//  Set the next cluster number
					//
					ulCluster = ulNextCluster;
				}


				//  Adjust the non-contigious count so that it
				//  matches MS chkdsk
				//
				if( ulNonContigiousCount )
				{
					ulNonContigiousCount++;
				}


				//  Display the path and name for any file which
				//  has non-contigious clusters as well as the
				//  number of non-contigious clusters
				//
				if( ulNonContigiousCount )
				{
					//  Add the name of the file that was found to
					//  the expanded path
					//
					AddName( szExpandedPath, Find.name );


					//  Display the name, path, and non-contigious cluster
					//  count.
					//
					printf( szNonContigious, szExpandedPath,
							ulNonContigiousCount );


					//  Remove the name part of the expanded path
					//
					RemoveName( szExpandedPath );
				}
				else
				{
					printf( szContigious );
				}

				//  Determine if there are more files which match the mask
				//
				fSuccessful = dlFindNextFile();
			}
		}
	}
}


//  Make sure that the FAT copies are in sync.  Fix it if they
//  are not
//
uchar SyncronizeFATCopies( void )
{
	static uchar pFirstFATCopy[MAX_SECTOR_SIZE];
	LPDPB lpDPB;
	uchar ucFATCopies;
	ulong ulFATSectors;
	ulong ulSectorNumber;
	ulong ulLastFATSector;
	unsigned uFATIndex;
	unsigned uSectorSize;
	bool fHaveShownMsg = FALSE;
	char far * lpFATSector;
	bool fFoundErrors = FALSE;


	//  Get the DPB for the drive
	//
	GetDPB( &lpDPB );
	ASSERT_LPDPB( lpDPB );


	//  Determine the number of FAT copies
	//
	ucFATCopies = lpDPB->num_fats;


	//  Save the number of FAT copies
	//
	SetFATCopies( ucFATCopies );


	//  Determine the sector size
	//
	uSectorSize = lpDPB->bytes_per_sec;


	//  Determine if there is more than one copy of the FAT.  If there
	//  is not more than one copy of the FAT.
	//
	if( ucFATCopies < 2 )
	{
		//  There is no FAT copies to compare
		//
		return fFoundErrors;
	}


	//  Determine the first sector number of the first FAT
	//
	ulSectorNumber = lpDPB->fat_start;


	//  Determine if this is a FAT32 drive.  If this is
	//  a FAT32 drive.
	//
	if( IsFat32Drive(lpDPB) )
	{
		//  Determine the number of FAT sectors
		//
		ulFATSectors = lpDPB->ulSectorsPerFAT;
	}
	else
	{
		//  Determine the number of FAT sectors
		//
		ulFATSectors = lpDPB->fat_len;
	}


	//  Determine the last FAT sector of the first FAT
	//
	ulLastFATSector = ulSectorNumber + ulFATSectors - 1;


	//  Display a message stating that we are comparing the FATs
	//  and that this could take a while.
	//
	if( IsFat32Drive(lpDPB) )
	{
		printf( szCmpFATCopies );
	}


	//  Set SFC (synconizing FAT copies) flag
	//
	SetSFCFlag();


	//  Compare each copy of the FAT with the first copy
	//
	for( ; ulSectorNumber <= ulLastFATSector; ulSectorNumber++ )
	{
		//  Get the FAT sector from the first copy of the FAT
		//
		lpFATSector = GetSectorAddress( ulSectorNumber );


		//  Copy this sector into a local buffer
		//
		_fmemcpy( pFirstFATCopy, lpFATSector, uSectorSize );


		//  Compare this FAT sector with the other copies of
		//  the FAT.
		//
		for( uFATIndex = 1; uFATIndex < ucFATCopies; uFATIndex++ )
		{
			//  Check if the user wants to abort CTRL-C
			//
			CheckUserAbort();


			//  Get the address in memory for this FAT sector
			//
			lpFATSector = GetSectorAddress( ulSectorNumber +
					(ulFATSectors * uFATIndex) );


			//  Compare this copy with the first copy.  If this FAT copy
			//  does not match the first FAT copy ...
			//
			if( _fmemcmp(pFirstFATCopy, lpFATSector, uSectorSize) != 0 )
			{
				//  Set a flag indicating that we found errors
				//
				fFoundErrors = TRUE;


				//  Show the message stating that we found errors on
				//  this disk, if we are not in fix mode and we
				//  have not already shown this message.
				//
				ShowFoundErrorsMsg();


				//  Determine if we have already shown a message stating
				//  that the FAT copies were out of sync.  If we have
				//  not shown the message ...
				//
				if( !fHaveShownMsg )
				{
					//  Display a message indicating that the FAT copies
					//  were out of sync
					//
					printf( szSyncFATCopies );

					//  Set a flag indicating that we have shown this
					//  message already.  We only want to show it once.
					//
					fHaveShownMsg = TRUE;
				}


				//  If we are not fixing errors, there is no need to
				//  perform the copy.  These sectors will not be committed
				//  to disk anyway.  Determine if there is actually a need
				//  to perform the copy
				//
				if( DoFixErrors() )
				{
					//  Update the FAT sector with the information in
					//  the first FAT sector
					//
					_fmemcpy( lpFATSector, pFirstFATCopy, uSectorSize );


					//  Record the fact that this sector has
					//  been modified.
					//
					MarkBufferDirty( &ulSectorNumber );
				}
			}


			//  Free the MRU buffer and make it the LRU buffer if it is not
			//  dirty.
			//
			//  NOTE: This function is an optimization for Syncronizing the
			//        FAT copies.  This is used so that the buffers do not
			//        cache all the duplicate FAT sectors.
			//
			FreeMRUConvBufferAndMakeItLRU();
		}
	}


	//  Display a blank line
	//
	if( fFoundErrors || IsFat32Drive(lpDPB) )
	{
		printf("\n");
	}


	//  Clear SFC (synconizing FAT copies) flag
	//
	ClearSFCFlag();


	return fFoundErrors;
}


//  Check for, fix, and count lost clusters
//
bool FixLostClusters( void )
{
	LPDPB lpDPB;
	ulong ulMaxClusterNumber;
	ulong ulCluster;
	ulong ulNextCluster;
	ulong ulLostClusterCount;
	uchar ucVisitedListEntry;
	ulong ulCountOfLostClusters = 0;
	ulong ulBytesInLostClusters = 0;
	uchar ucBPSShift;
	uchar ucBPCShift;
	unsigned uBPS;
	unsigned uCHKFileNumber;
	bool fErrorsFound = FALSE;


	//  Get the DPB for the drive
	//
	GetDPB( &lpDPB );
	ASSERT_LPDPB( lpDPB );


	//  Determine the maximum cluster number on this drive
	//
	if( IsFat32Drive(lpDPB) )
	{
		ulMaxClusterNumber = lpDPB->ulMaxCluster;
	}
	else
	{
		ulMaxClusterNumber = lpDPB->hiclus;
	}


	//  Go through each entry looking for lost clusters
	//
	for( ulCluster = 2; ulCluster <= ulMaxClusterNumber; ulCluster++ )
	{
		//  Check if the user wants to abort CTRL-C
		//
		CheckUserAbort();


		//  Get the visited list entry for this cluster
		//
		ucVisitedListEntry = GetVisitedListClusterEntry( ulCluster );


		//  Determine if this cluster is marked used.  If this cluster
		//  is not marked used in the visited list ...
		//
		if( (ucVisitedListEntry & VL_USED) == 0 )
		{
			//  Determine if the FAT entry for this cluster is available.
			//  If the FAT entry for this cluster is not available ...
			//
			GetFatEntry( lpDPB, ulCluster, &ulNextCluster );
			if( ulNextCluster )
			{
				//  Determine if this FAT entry is marked bad.  If this
				//  cluster entry is not marked bad ...
				if( ulNextCluster != BAD_CLUSTER )
				{
					//  Determine if this is part of another lost cluster
					//  chain.  If this is not part of another lost cluster
					//  chain ...
					//
					if( (ucVisitedListEntry & VL_IN_PROCESS) == 0 )
					{
						//  Set a flag indicating that lost clusters were
						//  found.
						//
						fErrorsFound = TRUE;


						//  Show the message stating that we found errors on
						//  this disk, if we are not in fix mode and we
						//  have not already shown this message.
						//
						ShowFoundErrorsMsg();


						//  Increament the count of lost cluster chains
						//
						ulCountOfLostClusters++;


						//  Count the lost chains
						//
						_FollowLostClusterChain( ulCluster, &ulLostClusterCount,
								TRUE, TRUE, FALSE );

//						printf("Cluster: %lu  Count: %lu\n", ulCluster,
//								ulLostClusterCount );

						//  Update the lost cluster count
						//
						ulBytesInLostClusters += ulLostClusterCount;
					}
				}
				else
				{
					//  Increament the count of bad clusters
					//
					stat[DBAD].number++;
					stat[DBAD].bytes++;
				}
			}
		}
	}


	//  Calculate the bytes per cluster conversion shift factor
	//
	{
		//  Determine the shift factor for the bytes per sector
		//
		ucBPSShift = 0;
		uBPS = lpDPB->bytes_per_sec;
		while( uBPS )
		{
			uBPS >>= 1;
			ucBPSShift++;
		}
		ucBPSShift--;


		//  Determine the shift factor for bytes per cluster
		//
		ucBPCShift = ucBPSShift + lpDPB->shift;
	}


	//  Determine if there are any lost clusters.  If there are any lost
	//  clusters...
	//
	if( ulCountOfLostClusters )
	{
		//  Display the lost clusters statistics
		//
		printf( szLostMsg, ulBytesInLostClusters, ulCountOfLostClusters );


		//  Determine if we are fixing problems.  If we are fixing problems ...
		//
		if( DoFixErrors() )
		{
			//  Determine if the user wants to save these as files.  If
			//  the user wants to save these as files ...
			//
			if( getyesno(szConvLostClusMsg, DEFAULT_YES ) )
			{
				//  Save all the lost clusters as files in the root
				//
				{
					//  Go through each entry fixing lost clusters
					//
					for( ulCluster = 2; ulCluster <= ulMaxClusterNumber;
							ulCluster++ )
					{
						//  Check if the user wants to abort CTRL-C
						//
						CheckUserAbort();


						//  Get the visited list entry for this cluster
						//
						ucVisitedListEntry = GetVisitedListClusterEntry(
								ulCluster );


						//  Determine if this cluster is marked used.  If this
						//  cluster is not marked used in the visited list ...
						//
						if( (ucVisitedListEntry & VL_USED) == 0 )
						{
							//  Determine if the FAT entry for this cluster is
							//  available.  If the FAT entry for this cluster
							//  is not available ...
							//
							GetFatEntry( lpDPB, ulCluster, &ulNextCluster );
							if( ulNextCluster )
							{
								//  Determine if this FAT entry is marked bad.  If
								//  this cluster entry is not marked bad ...
								//
								if( ulNextCluster != BAD_CLUSTER )
								{
									//  Determine if this is part of another lost
									//  cluster chain.  If this is not part of
									//  another lost cluster chain ...
									//
									if( (ucVisitedListEntry & VL_IN_PROCESS) != 0 )
									{
										//  Allocate an entry in the root
										//
										if( !CanAllocateRootEntry(&uCHKFileNumber) )
										{
											return fErrorsFound;
										}
										else
										{
											//  Save the lost chains
											//
											_FollowLostClusterChain( ulCluster,
													&ulLostClusterCount, FALSE, FALSE,
													FALSE );


											//  Save this chain to a CHK file in the
											//  root directory
											//
											CreateRootEntry( uCHKFileNumber, ulCluster,
													ulLostClusterCount );

											//  Increament the LOST cluster recovery
											//  statistics
											//
											stat[DLOST].number++;
											stat[DLOST].bytes += ulLostClusterCount;
										}
									}
								}
							}
						}
					}


					//  Don't need to say anything when saving lost clusters
					//
					printf("\n");
				}
			}
			else
			{
				//  Free all the lost clusters
				//
				{
					//  Go through each entry fixing lost clusters
					//
					for( ulCluster = 2; ulCluster <= ulMaxClusterNumber;
							ulCluster++ )
					{
						//  Check if the user wants to abort CTRL-C
						//
						CheckUserAbort();


						//  Get the visited list entry for this cluster
						//
						ucVisitedListEntry = GetVisitedListClusterEntry(
								ulCluster );


						//  Determine if this cluster is marked used.  If this
						//  cluster is not marked used in the visited list ...
						//
						if( (ucVisitedListEntry & VL_USED) == 0 )
						{
							//  Determine if the FAT entry for this cluster is
							//  available.  If the FAT entry for this cluster
							//  is not available ...
							//
							GetFatEntry( lpDPB, ulCluster, &ulNextCluster );
							if( ulNextCluster )
							{
								//  Determine if this FAT entry is marked bad.  If
								//  this cluster entry is not marked bad ...
								//
								if( ulNextCluster != BAD_CLUSTER )
								{
									//  Determine if this is part of another lost
									//  cluster chain.  If this is not part of
									//  another lost cluster chain ...
									//
									if( (ucVisitedListEntry & VL_IN_PROCESS) != 0 )
									{
										//  Delete the lost chains
										//
										_FollowLostClusterChain( ulCluster,
												&ulLostClusterCount, FALSE, FALSE,
												TRUE );
									}
								}
							}
						}
					}
				}


				//  Display the amount that was freed
				//
				printf( szLostMsg3, (ulBytesInLostClusters << ucBPCShift) );
			}
		}
		else
		{
			//  Display the amount that would be freed
			//
			printf( szLostMsg2, (ulBytesInLostClusters << ucBPCShift) );
		}
	}


	//  Return whether any errors were found
	//
	return fErrorsFound;
}


//  Get the address of the variable which contains the available
//  root entry count
//
ulong * GetVarAvailableRootEntries( void )
{
	static ulong ulAvailableRootEntries = 0;
	return &ulAvailableRootEntries;
}


//  Get the available root entry count
//
ulong GetAvailableRootEntries( void )
{
	ulong * pulAvailableRootEntries;
	pulAvailableRootEntries = GetVarAvailableRootEntries();
	return *pulAvailableRootEntries;
}


//  Increament the available root entry count
//
void IncAvailableRootEntries( void )
{
	ulong * pulAvailableRootEntries;
	pulAvailableRootEntries = GetVarAvailableRootEntries();
	(*pulAvailableRootEntries)++;
}


//  Increament the available root entry count
//
void DecAvailableRootEntries( void )
{
	ulong * pulAvailableRootEntries;
	pulAvailableRootEntries = GetVarAvailableRootEntries();
	(*pulAvailableRootEntries)--;
}


//  Determine if there are empty entries in the root
//
bool IsAvailableRootEntry( void )
{
	LPDPB lpDPB;
	ulong ulAvailableRootEntries;
	ulong ulAllocatedClusterNumber;


	//  Get the available number of root entries
	//
	ulAvailableRootEntries = GetAvailableRootEntries();


	//  Determine if there are any root entries available
	//
	if( ulAvailableRootEntries )
	{
		//  Return that there are available root entries
		//
		return TRUE;
	}
	else
	{
		//  Get the DPB
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Determine if this is a FAT32 drive
		//
		if( IsFat32Drive(lpDPB) )
		{
			//  Determine if there is an available cluster
			//
			if( AllocateCluster( &ulAllocatedClusterNumber, TRUE ) )
			{
				//  Return that there are available root entries
				//
				return TRUE;
			}
		}
	}


	//  Return that there are no available root entries
	//
	return FALSE;
}


//  Determine if we can allocate an entry in the root directory
//
bool CanAllocateRootEntry( unsigned * puCHKFileNumber )
{
	//  Determine if there are empty entries in the root
	//
	if( !IsAvailableRootEntry() )
	{
		//  Display a message stating that there are no more
		//  available entries in the root.
		//
		printf( szInsufRoot );


		//  Return that there are no more available entries in the
		//  root.
		//
		return FALSE;
	}


	//  Determine the next "chk" file name number part
	//
	if( !GetNextChkFileNumber(puCHKFileNumber) )
	{
		//  Display a message stating that all the FILE*.CHK file names
		//  have been used.
		//
		printf( szTooManyCHKFiles );


		//  Return that all the CHK file name possibilities were
		//  used.
		//
		return FALSE;
	}


	//  We were successful in creating the root entry
	//
	return TRUE;
}


//  Determine the next "chk" file name
//
bool GetNextChkFileNumber( unsigned * puCHKFileNumber )
{
	static char szENAME[ENAMESIZE+1];
	static unsigned uLastCHKFileNumberFound = 0xffff;


	//  Start seaching for a chk file number with the number following
	//  the last one found.
	//
	uLastCHKFileNumberFound++;


	//  Search for the next valid chk file name to use
	//
	while( uLastCHKFileNumberFound <= 9999 )
	{
		//  Create the name of the file we are looking for
		//
		sprintf( szENAME, szCHKName, uLastCHKFileNumberFound );


		//  Determine if this name is already used.  If this
		//  name is not already used ...
		//
		if( !ExistInRoot(szENAME) )
		{
			//  Set the number for the chk file in the return
			//  variable
			//
			*puCHKFileNumber = uLastCHKFileNumberFound;


			//  Return that we found a chk file name to use
			//
			return TRUE;
		}


		//  Determine the next chk file number to look for
		//
		uLastCHKFileNumberFound++;
	}


	//  Return that we were unable to create a chk file.
	//  This would occure because all the file names
	//  between FILE0000.CHK and FILE9999.CHK were used.
	//
	return FALSE;
}


//  Determine if the specified name exists in the root directory.
//  Wildcards are not handled.
//
bool ExistInRoot( char * szENAME )
{
	static LPDPB lpDPB = 0;
	uchar ucSPC;
	ulong ulSectorNumber;
	ulong ulCurCluster;
	ulong ulNextCluster;
	unsigned uCount;
	unsigned uEntryIndex;
	LPDIR lpDir;
	unsigned uEntriesPerSector;
	unsigned uRootEntries;


	//  Determine if we have already calculated these values.  If
	//  we have not already calculated these values ...
	if( !lpDPB )
	{
		//  Get the DPB for the drive
		//
		GetDPB( &lpDPB );
	}
	ASSERT_LPDPB( lpDPB );


	//  Determine the number of entries in a sector
	//
	uEntriesPerSector = lpDPB->bytes_per_sec / sizeof(DIR);


	//  Determine if this is a FAT32 drive.  If this is a
	//  FAT32 drive ...
	//
	if( IsFat32Drive(lpDPB) )
	{
		//  Determine the starting cluster of the root directory
		//
		ulCurCluster = lpDPB->ulRootDirStartCluster;


		//  Determine the number of entries in a cluster
		//
		ucSPC = lpDPB->ms_spc + 1;


		//  Go through the entries in the root looking for an
		//  available entry.
		//
		while( TRUE )
		{
			//  Determine the starting sector number for this
			//  cluster
			//
			ulSectorNumber = (ulCurCluster - 2) << lpDPB->shift;
			ulSectorNumber += lpDPB->ulFirstDataSector;


			//  Initialize the variable which contains the sector
			//  index into this cluster we are working on to zero.
			//
			uCount = 0;


			//  Go through the entries in this cluster looking for an
			//  available entry.
			//
			while( uCount < ucSPC )
			{
				//  Check if the user wants to abort CTRL-C
				//
				CheckUserAbort();


				//  Initialize the variable which contains the current
				//  directory entry index into the current sector to
				//  zero.
				//
				uEntryIndex = 0;


				//  Go through the entries in this sector looking for
				//  an available entry.
				//
				while( uEntryIndex < uEntriesPerSector )
				{
					//  Get the directory entry being indexed in this
					//  sector.
					//
					lpDir = GetDirEntryAtSector( uEntryIndex,
							ulSectorNumber );


					//  Determine if this entry is the one we are looking
					//  for.  If this is the entry we are looking for ...
					//
					if( _fmemicmp(lpDir->name, szENAME, ENAMESIZE) == 0 )
					{
						//  Return that we found this entry.
						//
						return TRUE;
					}


					//  determine if this is the last entry in the root
					//  directory
					//
					if( lpDir->name[0] == 0 )
					{
						return FALSE;
					}


					//  Increament the entry index
					//
					uEntryIndex++;
				}


				//  Increament the sector number
				//
				ulSectorNumber++;


				//  Increament the count of sectors we have looked
				//  through in this cluster.
				//
				uCount++;
			}


			//  Get the next cluster in the chain
			//
			GetFatEntry( lpDPB, ulCurCluster, &ulNextCluster );


			//  Deterine if this is the end of the FAT chain.  If this
			//  is the end of the FAT chain ...
			//
			if( ulNextCluster >= END_OF_FAT_CHECK )
			{
				//  We are done searching the root
				//
				break;
			}


			//  Update the current cluster number
			//
			ulCurCluster = ulNextCluster;
		}
	}
	else
	{
		//  Determine the number of entries in the root
		//
		uRootEntries = lpDPB->dir_ent;


		//  Determine the starting sector number of the root
		//  directory
		//
		ulSectorNumber = lpDPB->dir_start;


		//  Initialize the variable which contains the current
		//  directory entry index into the current sector to
		//  zero.
		//
		uEntryIndex = 0;


		//  Go through the entries in the root looking for an
		//  available entry.
		//
		while( uEntryIndex < uRootEntries )
		{
			//  Check if the user wants to abort CTRL-C
			//
			CheckUserAbort();


			//  Get the directory entry being indexed in this
			//  sector.
			//
			lpDir = GetDirEntryAtSector( uEntryIndex,
					ulSectorNumber );


			//  Determine if this entry is the one we are looking
			//  for.  If this is the entry we are looking for ...
			//
			if( _fmemicmp(lpDir->name, szENAME, ENAMESIZE) == 0 )
			{
				//  Return that we found this entry.
				//
				return TRUE;
			}


			//  determine if this is the last entry in the root
			//  directory
			//
			if( lpDir->name[0] == 0 )
			{
				return FALSE;
			}


			//  Determine the next entry index
			//
			uEntryIndex++;


			//  Determine if we have checked all the entries in
			//  this sector.  If we have checked all the entries
			//  in this sector
			//
			if( uEntryIndex >= uEntriesPerSector )
			{
				//  Increament the sector number
				//
				ulSectorNumber++;


				//  Reset the directory entry index
				//
				uEntryIndex = 0;


				//  Adjust the number of entries in the root
				//
				uRootEntries -= uEntriesPerSector;
			}
		}
	}


	//  Return that this entry does not exist in the root
	//
	return FALSE;
}


//  Count the number of root entries that are available
//
void CountRootEntries( void )
{
	LPDPB lpDPB;
	uchar ucSPC;
	ulong ulSectorNumber;
	ulong ulCurCluster;
	ulong ulNextCluster;
	unsigned uCount;
	unsigned uEntryIndex;
	LPDIR lpDir;
	unsigned uEntriesPerSector;
	unsigned uRootEntries;
	bool fLastEntryFound = FALSE;


	//  Get the DPB for the drive
	//
	GetDPB( &lpDPB );
	ASSERT_LPDPB( lpDPB );


	//  Determine the number of entries in a sector
	//
	uEntriesPerSector = lpDPB->bytes_per_sec / sizeof(DIR);


	//  Determine if this is a FAT32 drive.  If this is a
	//  FAT32 drive ...
	//
	if( IsFat32Drive(lpDPB) )
	{
		//  Determine the starting cluster of the root directory
		//
		ulCurCluster = lpDPB->ulRootDirStartCluster;


		//  Determine the number of entries in a cluster
		//
		ucSPC = lpDPB->ms_spc + 1;


		//  Go through the entries in the root looking for an
		//  available entry.
		//
		while( TRUE )
		{
			//  Determine the starting sector number for this
			//  cluster
			//
			ulSectorNumber = (ulCurCluster - 2) << lpDPB->shift;
			ulSectorNumber += lpDPB->ulFirstDataSector;


			//  Initialize the variable which contains the sector
			//  index into this cluster we are working on to zero.
			//
			uCount = 0;


			//  Go through the entries in this cluster looking for an
			//  available entry.
			//
			while( uCount < ucSPC )
			{
				//  Initialize the variable which contains the current
				//  directory entry index into the current sector to
				//  zero.
				//
				uEntryIndex = 0;


				//  Go through the entries in this sector looking for
				//  an available entry.
				//
				while( uEntryIndex < uEntriesPerSector )
				{
					//  Check if the user wants to abort CTRL-C
					//
					CheckUserAbort();


					if( !fLastEntryFound )
					{
						//  Get the directory entry being indexed in this
						//  sector.
						//
						lpDir = GetDirEntryAtSector( uEntryIndex,
								ulSectorNumber );


						//  Determine if this is a deleted entry
						//
						if( lpDir->name[0] == (char)0xe5 )
						{
							IncAvailableRootEntries();
						}


						//  Determine if this is the last root entry
						//
						if( lpDir->name[0] == 0 )
						{
							IncAvailableRootEntries();
							fLastEntryFound = TRUE;
						}
					}
					else
					{
						IncAvailableRootEntries();
					}


					//  Increament the entry index
					//
					uEntryIndex++;
				}


				//  Increament the sector number
				//
				ulSectorNumber++;


				//  Increament the count of sectors we have looked
				//  through in this cluster.
				//
				uCount++;
			}


			//  Get the next cluster in the chain
			//
			GetFatEntry( lpDPB, ulCurCluster, &ulNextCluster );


			//  Deterine if this is the end of the FAT chain.  If this
			//  is the end of the FAT chain ...
			//
			if( ulNextCluster >= END_OF_FAT_CHECK )
			{
				//  We are done searching the root
				//
				break;
			}


			//  Update the current cluster number
			//
			ulCurCluster = ulNextCluster;
		}
	}
	else
	{
		//  Determine the number of entries in the root
		//
		uRootEntries = lpDPB->dir_ent;


		//  Determine the starting sector number of the root
		//  directory
		//
		ulSectorNumber = lpDPB->dir_start;


		//  Initialize the variable which contains the current
		//  directory entry index into the current sector to
		//  zero.
		//
		uEntryIndex = 0;


		//  Go through the entries in the root looking for an
		//  available entry.
		//
		while( uEntryIndex < uRootEntries )
		{
			//  Check if the user wants to abort CTRL-C
			//
			CheckUserAbort();


			if( !fLastEntryFound )
			{
				//  Get the directory entry being indexed in this
				//  sector.
				//
				lpDir = GetDirEntryAtSector( uEntryIndex,
						ulSectorNumber );


				//  Determine if this is a deleted entry
				//
				if( lpDir->name[0] == (char)0xe5 )
				{
					IncAvailableRootEntries();
				}


				//  Determine if this is the last root entry
				//
				if( lpDir->name[0] == 0 )
				{
					IncAvailableRootEntries();
					fLastEntryFound = TRUE;
				}
			}
			else
			{
				IncAvailableRootEntries();
			}


			//  Determine the next entry index
			//
			uEntryIndex++;


			//  Determine if we have checked all the entries in
			//  this sector.  If we have checked all the entries
			//  in this sector
			//
			if( uEntryIndex >= uEntriesPerSector )
			{
				//  Increament the sector number
				//
				ulSectorNumber++;


				//  Reset the directory entry index
				//
				uEntryIndex = 0;


				//  Adjust the number of entries in the root
				//
				uRootEntries -= uEntriesPerSector;
			}
		}
	}
}


//  Create an entry in the root directory
//
void CreateRootEntry( unsigned uCHKFileNumber, ulong ulStartingCluster,
		ulong ulClusters )
{
	static DIR DIR;
	static char szName[ENAMESIZE+1];
	unsigned uFileTime;
	unsigned uFileDate;
	LPDPB lpDPB;
	unsigned uBPS;
	uchar ucSPC;
	uchar ucBPSShift;
	uchar ucBPCShift;
	ulong ulSectorNumber;
	ulong ulCurCluster;
	ulong ulNextCluster;
	ulong ulTemp;
	unsigned uCount;
	unsigned uEntryIndex;
	LPDIR lpDir;
	unsigned uEntriesPerSector;
	unsigned uRootEntries;
	char far * lpSectorData;
	bool fMarkNextEntryAsLast = FALSE;
	ulong ulNewAvailableEntries = 0;


	//  Determine if we are actually fixing the problems.  If we are
	//  not fixing the problems, hence we are simulating fixing the
	//  problems, there is not a need to perform all these operations.
	//  Besides, this would not work in simulation mode for a number
	//  of reasons.
	//
	if( DoFixErrors() )
	{
		//  Get the DPB for the drive
		//
		GetDPB( &lpDPB );
		ASSERT_LPDPB( lpDPB );


		//  Calculate the bytes per cluster conversion shift factor
		//
		{
			//  Determine the shift factor for the bytes per sector
			//
			ucBPSShift = 0;
			uBPS = lpDPB->bytes_per_sec;
			while( uBPS )
			{
				uBPS >>= 1;
				ucBPSShift++;
			}
			ucBPSShift--;


			//  Determine the shift factor for bytes per cluster
			//
			ucBPCShift = ucBPSShift + lpDPB->shift;
		}


		//  Fill in all the information in this local copy of a
		//  directory entry
		//
		{
			//  Set the file name
			//
			sprintf( szName, szCHKName, uCHKFileNumber );
//			printf( "szName: %s\n", szName );
			memcpy( &DIR, szName, 11 );


			//  Set the attributes
			//
			DIR.attr = A_ARCHIVE;


			//  Determine the current date and time and set the
			//  date and time in the directory entry
			//
			GetDateAndTime( &uFileDate, &uFileTime );
			DIR.date = uFileDate;
			DIR.time = uFileTime;


			//  Set the starting cluster number
			//
			if( IsFat32Drive(lpDPB) )
			{
				//  FAT32 drives use big clusters
				//
				DIR.uStartingClusterLo = LOWORD(ulStartingCluster);
				DIR.uStartingClusterHi = HIWORD(ulStartingCluster);
			}
			else
			{
				//  FAT12/FAT16 drives use small clusters
				//
				DIR.uStartingClusterLo = LOWORD(ulStartingCluster);
				DIR.uStartingClusterHi = 0;
			}


			//  Set the file size
			//
			DIR.size = ulClusters << ucBPCShift;
		}


		//  Place this entry in the root directory
		//
		{
			//  Determine the number of entries in a sector
			//
			uEntriesPerSector = lpDPB->bytes_per_sec / sizeof(DIR);


			//  Determine if this is a FAT32 drive.  If this is a
			//  FAT32 drive ...
			//
			if( IsFat32Drive(lpDPB) )
			{
				//  Determine the starting cluster of the root directory
				//
				ulCurCluster = lpDPB->ulRootDirStartCluster;


				//  Determine the number of entries in a cluster
				//
				ucSPC = lpDPB->ms_spc + 1;


				//  Go through the entries in the root looking for an
				//  available entry.
				//
				while( TRUE )
				{
					//  Determine the starting sector number for this
					//  cluster
					//
					ulSectorNumber = (ulCurCluster - 2) << lpDPB->shift;
					ulSectorNumber += lpDPB->ulFirstDataSector;
//					printf("lpDPB->ulFirstDataSector: %lu\n",
//							lpDPB->ulFirstDataSector );
//					printf("ulSectorNumber: %lu\n", ulSectorNumber );


					//  Initialize the variable which contains the sector
					//  index into this cluster we are working on to zero.
					//
					uCount = 0;


					//  Go through the entries in this cluster looking for an
					//  available entry.
					//
					while( uCount < ucSPC )
					{
						//  Initialize the variable which contains the current
						//  directory entry index into the current sector to
						//  zero.
						//
						uEntryIndex = 0;


						//  Go through the entries in this sector looking for
						//  an available entry.
						//
						while( uEntryIndex < uEntriesPerSector )
						{
							//  Check if the user wants to abort CTRL-C
							//
							CheckUserAbort();


							//  Get the directory entry being indexed in this
							//  sector.
							//
							lpDir = GetDirEntryAtSector( uEntryIndex,
									ulSectorNumber );


							//  Determine if we are marking this entry as the
							//  last entry
							//
							if( fMarkNextEntryAsLast )
							{
								//  mark this entry as the last entry
								//
								lpDir->name[0] = 0;


								//  Record the fact that this sector has
								//  been modified.
								//
								MarkBufferDirty( &ulSectorNumber );


								//  we are done
								//
								return;
							}


							//  Determine if this is the last of the directory
							//  entries
							//
							if( lpDir->name[0] == 0 )
							{
								//  Set a flag indicating that we will need to
								//  mark the next entry as the last entry
								//
								fMarkNextEntryAsLast = TRUE;
							}


							//  Determine if this entry is available.  If this
							//  entry is available ...
							//
							if( ((uchar)(lpDir->name[0]) == 0xe5) ||
									(lpDir->name[0] == 0) )
							{
								//  Record that there is one less available root
								//  entry
								//
								DecAvailableRootEntries();


//								printf( "uEntryIndex: %u\n", uEntryIndex );
								//  Set the directory entry information
								//
								_fmemcpy( lpDir, &DIR, sizeof(DIR) );


								//  Record the fact that this sector has
								//  been modified.
								//
								MarkBufferDirty( &ulSectorNumber );


								//  We are done if we don't have to mark the
								//  next entry as the last entry
								//
								if( !fMarkNextEntryAsLast )
								{
									return;
								}
							}


							//  Increament the entry index
							//
							uEntryIndex++;
						}


						//  Increament the sector number
						//
						ulSectorNumber++;


						//  Increament the count of sectors we have looked
						//  through in this cluster.
						//
						uCount++;
					}


					//  Get the next cluster in the chain
					//
					GetFatEntry( lpDPB, ulCurCluster, &ulNextCluster );


					//  Deterine if this is the end of the FAT chain.  If this
					//  is the end of the FAT chain ...
					//
					if( ulNextCluster >= END_OF_FAT_CHECK )
					{
						//  We do not need to set the next entry as the last
						//  entry if we need to allocate a new cluster to do
						//  so
						//
						if( fMarkNextEntryAsLast )
						{
							//  We are done
							//
							return;
						}


						//  Allocate another cluster for the root directory.
						//  (Can't fail, a previous check made sure that
						//   another cluster was available.)
						//
						if( !AllocateCluster(&ulNextCluster,FALSE) )
						{
							CHKDSK_INTERNAL_ERROR();
						}


						//  Write zeros to all the bytes in this cluster.
						//
						{
							//  Determine the starting sector number for this
							//  cluster.
							//
							ulSectorNumber = (ulCurCluster - 2) << lpDPB->shift;
							ulSectorNumber += lpDPB->ulFirstDataSector;


							//  Initialize an index for this sector in a cluster
							//
							uCount = 0;


							//  Zero each of the sectors in this cluster
							//
							while( uCount < ucSPC )
							{
								//  Check if the user wants to abort CTRL-C
								//
								CheckUserAbort();


								//  Get the address in memory for the data
								//  for this sector
								//
								lpSectorData = GetSectorAddress( ulSectorNumber );


								//  Zero this sector
								//
								_fmemset( lpSectorData, 0, lpDPB->bytes_per_sec );


								//  Record the fact that this sector has
								//  been modified.
								//
								MarkBufferDirty( &ulSectorNumber );


								//  Increament the sector number
								//
								ulSectorNumber++;


								//  Increament the count
								//
								uCount++;
							}
						}


						//  Place an EOF at in this cluster entry
						//
						PutFatEntry( lpDPB, ulNextCluster, END_OF_FAT, &ulTemp );


						//  Point the end of the root cluster chain to the
						//  newly allocated cluster
						//
						PutFatEntry( lpDPB, ulCurCluster, ulNextCluster, &ulTemp );


						//  Account for the new root entries
						//
						ulNewAvailableEntries = (ulong)uEntriesPerSector * ucSPC;
						ulNewAvailableEntries /= sizeof(DIR);
						while( ulNewAvailableEntries )
						{
							IncAvailableRootEntries();
							ulNewAvailableEntries--;
						}
					}


					//  Update the current cluster number
					//
					ulCurCluster = ulNextCluster;
				}
			}
			else
			{
				//  Determine the number of entries in the root
				//
				uRootEntries = lpDPB->dir_ent;


				//  Determine the starting sector number of the root
				//  directory
				//
				ulSectorNumber = lpDPB->dir_start;


				//  Initialize the variable which contains the current
				//  directory entry index into the current sector to
				//  zero.
				//
				uEntryIndex = 0;


				//  Go through the entries in the root looking for an
				//  available entry.
				//
				while( uEntryIndex < uRootEntries )
				{
					//  Check if the user wants to abort CTRL-C
					//
					CheckUserAbort();


					//  Get the directory entry being indexed in this
					//  sector.
					//
					lpDir = GetDirEntryAtSector( uEntryIndex,
							ulSectorNumber );


					//  Determine if we are marking this entry as the
					//  last entry
					//
					if( fMarkNextEntryAsLast )
					{
						//  mark this entry as the last entry
						//
						lpDir->name[0] = 0;


						//  Record the fact that this sector has
						//  been modified.
						//
						MarkBufferDirty( &ulSectorNumber );


						//  we are done
						//
						return;
					}


					//  Determine if this is the last of the directory
					//  entries
					//
					if( lpDir->name[0] == 0 )
					{
						//  Set a flag indicating that we will need to
						//  mark the next entry as the last entry
						//
						fMarkNextEntryAsLast = TRUE;
					}


					//  Determine if this entry is available.  If this
					//  entry is available ...
					//
					if( ((uchar)(lpDir->name[0]) == 0xe5) ||
							(lpDir->name[0] == 0) )
					{
						//  Record that there is one less available root entry
						//
						DecAvailableRootEntries();


						//  Set the directory entry information
						//
						_fmemcpy( lpDir, &DIR, sizeof(DIR) );


						//  Record the fact that this sector has
						//  been modified.
						//
						MarkBufferDirty( &ulSectorNumber );


						//  We are done
						//
						if( !fMarkNextEntryAsLast )
						{
							return;
						}
					}


					//  Increament the entry index
					//
					uEntryIndex++;


					//  Determine if we have checked all the entries in
					//  this sector.  If we have checked all the entries
					//  in this sector
					//
					if( uEntryIndex >= uEntriesPerSector )
					{
						//  Increament the sector number
						//
						ulSectorNumber++;


						//  Reset the directory entry index
						//
						uEntryIndex = 0;


						//  Adjust the number of entries in the root
						//
						uRootEntries -= uEntriesPerSector;
					}
				}


				//  We don't need to set the last entry marker
				//
				if( fMarkNextEntryAsLast )
				{
					return;
				}
			}


			//  We should never end up here.  Our calculations for determining
			//  that there was an available directory entry were flawed
			//
			CHKDSK_INTERNAL_ERROR();
		}
	}
}


//  Display a message stating that there was an internal error
//  at the specified line number
//
void CHKDSKInternalError( unsigned uFileNumber, unsigned uLineNumber )
{
	printf( "\nCHKDKS Internal Error! (%u:%u)", uFileNumber, uLineNumber );
	exit(255);
}


//  Return the cluster number for the next allocatable cluster.  This
//  updates the visited lists cluster usage as well.
//
bool AllocateCluster( ulong * pulAllocatedClusterNumber,
		bool fAvailableCheck )
{
	static ulong ulLastAllocatedCluster = 1;
	static ulong ulMaxClusterNumber = 0;
	static LPDPB lpDPB;
	uchar ucVisitedListEntry;
	ulong ulNextCluster;


	//  Determine if we already have these values.  If we do not
	//  already have the maximum cluster number ...
	//
	if( !ulMaxClusterNumber )
	{
		//  Get the DPB for the drive
		//
		GetDPB( &lpDPB );


		//  Determine the maximum cluster number
		//
		if( IsFat32Drive(lpDPB) )
		{
			ulMaxClusterNumber = lpDPB->ulMaxCluster;
		}
		else
		{
			ulMaxClusterNumber = lpDPB->hiclus;
		}
	}
	ASSERT_LPDPB( lpDPB );


	//  Go through the entire FAT looking for an available entry
	//
	while( ulLastAllocatedCluster < ulMaxClusterNumber )
	{
		//  Check if the user wants to abort CTRL-C
		//
		CheckUserAbort();


		//  Determine what the next cluster number we will check to
		//  see if it is available will be.
		//
		ulLastAllocatedCluster++;


		//  Determine if this cluster is marked available in the
		//  visited list.  If this cluster is marked available in
		//  the visited list ...
		//
		ucVisitedListEntry = GetVisitedListClusterEntry(
				ulLastAllocatedCluster );
		if( ucVisitedListEntry == 0 )
		{
			//  Determine if the FAT entry for this cluster is available.
			//  If the FAT entry for this cluster is available ...
			//
			GetFatEntry( lpDPB, ulLastAllocatedCluster, &ulNextCluster );
			if( !ulNextCluster )
			{
				//  Determine if we are only checking for a free cluster
				//
				if( fAvailableCheck )
				{
					//  We were only checking to see if there was an
					//  available cluster.  Go back one cluster so
					//  that this cluster will be the next found
					//
					ulLastAllocatedCluster--;
				}
				else
				{
					//  Mark this cluster as used in the visited list
					//
					SetVisitedListClusterEntry( ulLastAllocatedCluster,
							ucVisitedListEntry );


					//  Set the newly allocated cluster number in the return
					//  variable
					//
					*pulAllocatedClusterNumber = ulLastAllocatedCluster;
				}


				//  Return that we were successful at allocating a cluster.
				//
				return TRUE;
			}
		}
	}

	//  Return that we were unable to allocate a cluster
	//
	return FALSE;
}


//  Make sure the the '.' or '..' entry information is correct.  Fix
//  it if it is incorrect.
//
bool ModifiedSpecialEntry( LPDIR lpDir, ulong ulStartingCluster )
{
	bool fModified = FALSE;
	static LPDPB lpDPB = 0;

	if( !lpDPB )
	{
		//  Get the DPB for the drive
		//
		GetDPB( &lpDPB );
	}
	ASSERT_LPDPB( lpDPB );


	//  Make sure that the size is zero.  If the
	//  size is not zero ...
	//
	if( lpDir->size != 0 )
	{
		lpDir->size = 0;
		fModified = TRUE;
	}


	//  Make sure that the attribute is a normal
	//  directory attribute.  If the attribute
	//  is not just a directory ...
	//
	if( lpDir->attr != A_SUBDIR )
	{
		lpDir->attr = A_SUBDIR;
		fModified = TRUE;
	}


//  Don't know if MS checks or fixes this
//
#if 0
	//  Make sure that the time is that of the
	//  parent.  If the time is not that of
	//  the parent.
	//
	if( lpDir->time != uParentTime )
	{
		lpDir->time = uParentTime;
		fModified = TRUE;
	}


	//  Make sure that the date is that of the
	//  parent.  If the date is not that of
	//  the parent.
	//
	if( lpDir->date != uParentDate )
	{
		lpDir->date = uParentDate;
		fModified = TRUE;
	}
#endif


	//  The '..' entries on a FAT32 drive do not contain the
	//  root starting cluster number.  Like a FAT16 drive, they
	//  contain 0.  Compensate for this case.
	//
	if( IsFat32Drive(lpDPB) )
	{
		if( lpDPB->ulRootDirStartCluster == ulStartingCluster )
		{
			ulStartingCluster = 0;
		}
	}


	//  Make sure that the starting cluster is
	//  correct.  If the starting cluster is
	//  not correct
	//
	if( Fat32StartLP(lpDir) != ulStartingCluster )
	{
		lpDir->uStartingClusterLo = LOWORD(ulStartingCluster);
		lpDir->uStartingClusterHi = HIWORD(ulStartingCluster);
		fModified = TRUE;
	}


	//  Return whether we fixed (modified) this entry
	//
	return fModified;
}


//
//
uchar CheckTheEntireDrive( bool fFirstPass )
{
	char * szWorkingPath;
	LPDPB lpDPB;
	uchar ucDrive;
	static char szFileName[13];
	static uchar ucError;
	static ulong ulClusterCount;
	static unsigned uBPS;
	static uchar ucBPSShift;
	static uchar ucBPCShift;
	static ulong ulClusterSize;
	static ulong ulClustersInFile;
	static ulong ulFirstCrossLink;
	static uchar ucCheckCode;
	uchar ucReturnCode = 0;
	unsigned uEntryIndex;
	unsigned uNumberOfEntries;
	LPDIR lpDir;
	ulong ulSectorNumber;
	unsigned uEntriesPerSector;
	ulong ulDirStartingCluster;
	ulong ulAllocatedClusterCount;
	static bool fDeleteLFNSequence;
	static bool fInLFNSequence;
	static uchar ucLFNSequenceNumber;
	static uchar ucLFNStartSequenceNumber;
	static unsigned uLFNEntryIndex;
	static ulong ulLFNSectorNumber;
	static uchar ucLFNChecksum;
	static uchar ucTemp;
	static uchar ucNameError;
	bool fShowedInvalidEntriesMsg = FALSE;
	bool fRestOfDirIsAvailable = FALSE;


	//  Get the drive number
	//
	ucDrive = GetWorkingDriveDriveNumber();


	//  Set the variable which contains the current path to
	//  the drive ':' and the root.
	//
	szWorkingPath = GetWorkingPath();
	szWorkingPath[0] = 'A' + ucDrive;
	szWorkingPath[1] = ':';
	szWorkingPath[2] = '\\';
	szWorkingPath[3] = 0;


	//  Get the DPB for the drive
	//
	GetDPB( &lpDPB );
	ASSERT_LPDPB( lpDPB );


	//  Calculate the bytes per cluster conversion shift factor
	//
	{
		//  Determine the shift factor for the bytes per sector
		//
		ucBPSShift = 0;
		uBPS = lpDPB->bytes_per_sec;
		while( uBPS )
		{
			uBPS >>= 1;
			ucBPSShift++;
		}
		ucBPSShift--;


		//  Determine the shift factor for bytes per cluster
		//
		ucBPCShift = ucBPSShift + lpDPB->shift;


		//  Determine the cluster size
		//
		ulClusterSize = 1L << ucBPCShift;
	}


	//  Determine if we are checking a FAT32 drive.  If we are
	//  checking a FAT32 drive...
	//
	if( IsFat32Drive(lpDPB) )
	{
		if( fFirstPass )
		{
			ucError = FollowFATChain( lpDPB->ulRootDirStartCluster,
					&ulClusterCount, &ulFirstCrossLink,
					&ulAllocatedClusterCount, fFirstPass );
			if( ucError )
			{
				if( ucError & CHAIN_INVALID )
				{
					//  Determine if we are fixing errors.  If we
					//  are not fixing errors, display a message
					//  stating that errors were found but will
					//  not be fixed.
					//
					ShowFoundErrorsMsg();


					//
					//  Determine if the chain contains invalid
					//  cluster values.  If the chain contains
					//  invalid cluster values ...
					//

					//  Display the current path
					//
					printf("%s\n", szWorkingPath );


					//  Display a message stating that the
					//  root size has been adjusted
					//
					printf( szInvalidRootMsg );


					//  Update the return code
					//
					ucReturnCode |= ucError;
				}
			}
		}

		//  Determine the cluster number of the root directory and
		//  check the directory at the specified cluster
		//
		ucCheckCode |= CheckDirectoryAtCluster( lpDPB->ulRootDirStartCluster,
				fFirstPass, 0 );
		ucReturnCode |= ucCheckCode;
		if( ucCheckCode == INVALID_DIRECTORY )
		{
			//  Display an error message
			//
			if( !DoShowVerbose() && fFirstPass )
			{
				printf( "%s\n", szWorkingPath );
				printf( szInvalidDirMsg );
			}
		}
	}
	else
	{
		//
		//  Check the root directory
		//


		//
		//
		fInLFNSequence = FALSE;
		fDeleteLFNSequence = FALSE;


		//  Display the directory where we are if the user requested so
		//
		if( fFirstPass && DoShowVerbose() )
		{
			printf( showdir_str, szWorkingPath );
		}


		//  Initialize the variable which contains the
		//  current entry index to zero
		//
		uEntryIndex = 0;


		//  Determine the number of directory entries that are
		//  in the root
		//
		uNumberOfEntries = lpDPB->dir_ent;


		//  Determine the number of direcory entries in one sector
		//
		uEntriesPerSector = lpDPB->bytes_per_sec / sizeof(DIR);


		//  Determine the starting sector of the root directory
		//
		ulSectorNumber = lpDPB->dir_start;


		//  Determine if we have examined all the directory entries
		//  the root. if we have examined all the directory entries
		//  we are done checking the root
		//
		while( uEntryIndex < uNumberOfEntries )
		{
			//  Check if the user wants to abort CTRL-C
			//
			CheckUserAbort();


			//  Get the directory entry at the specified sector
			//  and index
			//
			lpDir = GetDirEntryAtSector( uEntryIndex, ulSectorNumber );


			//  Detemine if this is a blank entry. if this is
			//  a blank entry ...
			//
			if( lpDir->name[0] == 0 )
			{
				//  Set a flag indicating that all of the following
				//  entries should be blank
				//
				fRestOfDirIsAvailable = TRUE;
			}
			else if( fRestOfDirIsAvailable )
			{
				//
				//  This entry is invalid
				//

				//  Determine if this is the first pass through the
				//  drive
				//
				if( fFirstPass )
				{
					//  Determine if we are fixing errors.  If we
					//  are not fixing errors, display a message
					//  stating that errors were found but will
					//  not be fixed.
					//
					ShowFoundErrorsMsg();


					if( !fShowedInvalidEntriesMsg )
					{
						//  Display a message stating that an orphaned
						//  LFN entry was found and deleted.
						//
						if( !DoShowVerbose() )
						{
							printf( "Directory %s\n", szWorkingPath );
						}
						else
						{
							printf( "%s\n", szWorkingPath );
						}
						printf( szInvalidEntries );
						fShowedInvalidEntriesMsg = TRUE;
					}


					//  Mark this entry as deleted SCANDISK fashion.
					//
					lpDir->name[0] = 0;


					//  Record that there is one more available
					//  root entry if this is the root directory
					//
					if( szWorkingPath[3] == 0 )
					{
						IncAvailableRootEntries();
					}


					//  Record the fact that this sector has
					//  been modified.
					//
					MarkBufferDirty( &ulSectorNumber );
				}
			}


			//  Determine if this is a deleted entry. if this
			//  is not a deleted entry ...
			//
			else if( (uchar)(lpDir->name[0]) != 0xE5 )
			{
				//  Determine if this is the volume label
				//
				if( (lpDir->attr & (~A_ARCHIVE)) == A_VOLUME )
				{
					if( fInLFNSequence )
					{
						//  Determine if we are fixing errors.  If we
						//  are not fixing errors, display a message
						//  stating that errors were found but will
						//  not be fixed.
						//
						ShowFoundErrorsMsg();


						//  Display a message stating that an orphaned
						//  LFN entry was found and deleted.
						//
						printf( szOrphanedLFNEntry );


						//  Set a flag indicating to delete the LFN
						//  sequence.
						//
						fInLFNSequence = FALSE;
						fDeleteLFNSequence = TRUE;


						//  Set the sequence number to start deleting
						//  the LFN sequence
						//
						ucLFNSequenceNumber = ucLFNStartSequenceNumber;


						//
						//
						uEntryIndex = uLFNEntryIndex;
						ulSectorNumber = ulLFNSectorNumber;
						continue;
					}


					//  We are no longer deleting an LFN sequence
					//
					fDeleteLFNSequence = FALSE;


					//  Validate the volume label
					//

				}
				else if( lpDir->attr == A_LONGNAME )
				{
					//
					//  if this is a long file name entry
					//
					if( fLFNsSupported )
					{
						//  Validate the long directory entry
						//
						{
							//  Determine if we are deleting an LFN sequence
							//
							if( fDeleteLFNSequence )
							{
								//  Make sure that this sequence matches what
								//  it is supposed to
								//
								if( (lpDir->name[0] == ucLFNSequenceNumber) &&
										(lpDir->bCentiSeconds == ucLFNChecksum) )
								{
									//  Determine what the next sequence
									//  number should be
									//
									ucLFNSequenceNumber = lpDir->name[0];
									if( ucLFNSequenceNumber > 0x40 )
									{
										ucLFNSequenceNumber -= 0x40;
									}
									ucLFNSequenceNumber--;


									//  Mark this entry as deleted.
									//
									lpDir->name[0] = (char)0xe5;


									//  Record that there is one more available
									//  root entry if this is the root directory
									//
									if( szWorkingPath[3] == 0 )
									{
										IncAvailableRootEntries();
									}

									//  Record the fact that this sector has
									//  been modified.
									//
									MarkBufferDirty( &ulSectorNumber );
								}
								else
								{
									//  We are no longer deleting an LFN sequence
									//
									fDeleteLFNSequence = FALSE;


									//
									//
									continue;
								}
							}
							else
							{
								//  Determine if we are are in the middle
								//  of processing an LFN sequence
								//
								if( fInLFNSequence )
								{
									//
									//
									if( (lpDir->name[0] != ucLFNSequenceNumber) ||
											(lpDir->bCentiSeconds != ucLFNChecksum) )
									{
										//  Determine if we are fixing errors.  If we
										//  are not fixing errors, display a message
										//  stating that errors were found but will
										//  not be fixed.
										//
										ShowFoundErrorsMsg();


										//  Display a message stating that an orphaned
										//  LFN entry was found and deleted.
										//
										printf( szOrphanedLFNEntry );


										//  Set a flag indicating to delete the LFN
										//  sequence.
										//
										fInLFNSequence = FALSE;
										fDeleteLFNSequence = TRUE;


										//  Set the sequence number to start deleting
										//  the LFN sequence
										//
										ucLFNSequenceNumber = ucLFNStartSequenceNumber;


										//
										//
										uEntryIndex = uLFNEntryIndex;
										ulSectorNumber = ulLFNSectorNumber;
										continue;
									}
									else
									{
										//  Determine the next LFN sequence number
										//
										ucLFNSequenceNumber = lpDir->name[0];
										if( ucLFNSequenceNumber > 0x40 )
										{
											ucLFNSequenceNumber -= 0x40;
										}
										ucLFNSequenceNumber--;
									}
								}
								else
								{
									//  Record the entry index and sector for the
									//  first entry in the sequence.
									//
									uLFNEntryIndex = uEntryIndex;
									ulLFNSectorNumber = ulSectorNumber;
									ucLFNStartSequenceNumber = lpDir->name[0];


									//  Record the chksum
									//
									ucLFNChecksum = lpDir->bCentiSeconds;


									//  Determine if this is the first LFN entry
									//  in the sequence.  If this is the first
									//  in the sequence ...
									//
									if( lpDir->name[0] > 0x40 )
									{
										//  Set a flag to indicate that we are
										//  processing an LFN sequence
										//
										fInLFNSequence = TRUE;


										//  Determine the next LFN sequence number
										//
										ucLFNSequenceNumber = lpDir->name[0] - 0x40 - 1;
									}
									else
									{
										//  Determine if we are fixing errors.  If we
										//  are not fixing errors, display a message
										//  stating that errors were found but will
										//  not be fixed.
										//
										ShowFoundErrorsMsg();


										//  Display a message stating that an orphaned
										//  LFN entry was found and deleted.
										//
										printf( szOrphanedLFNEntry );


										//  Set a flag indicating to delete the LFN
										//  sequence.
										//
										fDeleteLFNSequence = TRUE;


										//  Set the sequence number to start deleting
										//  the LFN sequence
										//
										ucLFNSequenceNumber = lpDir->name[0];


										//
										//
										continue;
									}
								}
							}
						}
					}
				}
				else
				{
					//
					//  Validate the 8.3 directory entry
					//
					{
						//  Determine if we are at the short name part
						//  of an LFN entry.  If we are at the short
						//  name part of an LFN entry ...
						//
						if( fInLFNSequence )
						{
							//  Compute the checksum for the short name entry
							//
							ComputeENAMEChecksum( lpDir->name, &ucTemp );

							//  Determine if the previous LFN sequences checksum
							//  matches the short name.  If the checksums do not
							//  match ...
							//
							if( (ucTemp != ucLFNChecksum) ||
									(ucLFNSequenceNumber != 0) )
							{
								//  Determine if we are fixing errors.  If we
								//  are not fixing errors, display a message
								//  stating that errors were found but will
								//  not be fixed.
								//
								ShowFoundErrorsMsg();


								//  Display a message stating that an orphaned
								//  LFN entry was found and deleted.
								//
								printf( szOrphanedLFNEntry );


								//  Set a flag indicating to delete the LFN
								//  sequence.
								//
								fInLFNSequence = FALSE;
								fDeleteLFNSequence = TRUE;


								//  Set the sequence number to start deleting
								//  the LFN sequence
								//
								ucLFNSequenceNumber = ucLFNStartSequenceNumber;


								//
								//
								uEntryIndex = uLFNEntryIndex;
								ulSectorNumber = ulLFNSectorNumber;
								continue;
							}
						}


						//  We are no longer deleting an LFN sequence
						//
						fDeleteLFNSequence = FALSE;
						fInLFNSequence = FALSE;


						//  Convert the file name in ENAME form into a
						//  ASCIIZ string and append it to our current
						//  path.
						//
						EnameToDot( lpDir->name, szFileName );
						AddName( szWorkingPath, szFileName );


						//  Determine the number of clusters which would
						//  be allocated for the files size
						//
						ulClustersInFile = lpDir->size >> ucBPCShift;
						if( lpDir->size & (ulClusterSize-1) )
						{
							ulClustersInFile++;
						}


						//  Determine the starting cluster for this
						//  directory
						//
						ulDirStartingCluster = lpDir->uStartingClusterLo;
						if( IsFat32Drive(lpDPB) )
						{
							ulDirStartingCluster |=
									(ulong)(lpDir->uStartingClusterHi) << 16;
						}


						//  Determine if this is our first time going through
						//  the directory tree of the drive.  If this is our
						//  first time going through the directory tree of
						//  of the drive ...
						//
	  					if( fFirstPass )
						{
							//  Display the file we are checking if the user
							//  requested so
							//
							if( DoShowVerbose() )
							{
								//  Display the full path for the user
								//
								if( lpDir->attr & A_SUBDIR )
								{
									printf( showdir_str, szWorkingPath );
								}
								else
								{
									printf("%s\n", szWorkingPath );
								}
							}


							//  Examine and/or fix the FAT chain
							//
							ucError = FollowFATChain( ulDirStartingCluster,
									&ulClusterCount, &ulFirstCrossLink,
									&ulAllocatedClusterCount, fFirstPass );
							if( ucError )
							{
								//  Determine if we are fixing errors.  If we
								//  are not fixing errors, display a message
								//  stating that errors were found but will
								//  not be fixed.
								//
								ShowFoundErrorsMsg();


								//  Determine if the first cluster of this
								//  chain was invalid
								//
								if( ucError & CHAIN_FIRST_INVALID )
								{
									//  Update the return code
									//
									ucReturnCode |= ucError;


									//  Display the current path
									//
									printf("%s\n", szWorkingPath );


									//  Display a message stating that the
									//  file size has been adjusted
									//
									printf( szFirstInvalidAllocationMsg );


									//  Get the directory entry at the specified
									//  sector and index
									//
									lpDir = GetDirEntryAtSector( uEntryIndex,
											ulSectorNumber );


									//  The starting cluster is invalid.  Update
									//  the directory entry
									//
									lpDir->uStartingClusterHi = 0;
									lpDir->uStartingClusterLo = 0;


									//  Record the fact that this sector has
									//  been modified.
									//
									MarkBufferDirty( &ulSectorNumber );
								}
								else if( ucError & CHAIN_INVALID )
								{
									//
									//  Determine if the chain contains invalid
									//  cluster values.  If the chain contains
									//  invalid cluster values ...
									//

									//  Display the current path
									//
									printf("%s\n", szWorkingPath );


									//  Display a message stating that the
									//  file size has been adjusted
									//

									printf( szInvalidAllocationMsg );


									//  Update the return code
									//
									ucReturnCode |= ucError;
								}


								//  Only adjust the file size for invalid
								//  chains.
								//
								if( ucError & ( CHAIN_FIRST_INVALID |
										CHAIN_INVALID | CHAIN_ADJUST_SIZE) )
								{

									//  Only adjust the file size if the number
									//  of clusters required for file size is
									//  different than the number of clusters
									//  in the chain
									//
									if( ulClustersInFile != ulClusterCount )
									{
		  								//  Get the directory entry at the specified
										//  sector and index
										//
										lpDir = GetDirEntryAtSector( uEntryIndex,
												ulSectorNumber );


										//  Update the file size
										//
										lpDir->size = ulClusterCount << ucBPCShift;


										//  Record the fact that this sector has
										//  been modified.
										//
										MarkBufferDirty( &ulSectorNumber );
									}
								}
							}
						}
						else
						{
							//  Examine and/or fix the FAT chain
							//
							ucError = _FollowFATChain( ulDirStartingCluster,
									&ulClusterCount, &ulFirstCrossLink,
									&ulAllocatedClusterCount, fFirstPass,
									TRUE );
							if( ucError )
							{
								//  Determine if the file contains cross-linked
								//  clusters.  If the file contains cross-linked
								//  clusters ...
								//
								if( ucError & CHAIN_CROSS_LINKED )
								{
									//  Update the return code
									//
									ucReturnCode |= ucError;


									//  Display the current path
									//
									printf("%s\n", szWorkingPath );


									//  Display a message stating that cross-linked
									//  files were found
									//
									if( DoFixErrors() )
									{
										printf( szFixCrossLinkedMsg, szWas,
												ulFirstCrossLink );
									}
									else
									{
										printf( szFixCrossLinkedMsg, szIs,
												ulFirstCrossLink );
									}


									//  Fix the cross-linked files
									//
	 								_FollowFATChain( ulDirStartingCluster,
											&ulClusterCount, &ulFirstCrossLink,
											&ulAllocatedClusterCount, fFirstPass,
											FALSE );
								}
							}
						}


						//  Get the directory entry at the specified sector
						//  and index
						//
						lpDir = GetDirEntryAtSector( uEntryIndex,
								ulSectorNumber );


						if( fFirstPass )
						{
							//  Determine if the directory entry name is valid
							//
							ucNameError = IsValidName( lpDir->name, szFileName );
							if( ucNameError )
							{
								//  Determine if we are fixing errors.  If we
								//  are not fixing errors, display a message
								//  stating that errors were found but will
								//  not be fixed.
								//
								ShowFoundErrorsMsg();


								//  Determine if the only error was that
								//  there was lower case letters
								//
								if( ucNameError == NAME_ERROR_CASE )
								{
									//  Display a message stating that there
									//  was an invalid directory entry name
									//  and that it was fixed
									//
									printf( szFixed, szInvalidNameMsg );
								}
								else
								{
									//  This is an invalid file name.  We will
									//  need to move this entry to the root
									//
									printf( szRenamedAndRelocated,
											szInvalidNameMsg );
								}
							}


							//  Determine if the directory attributes are valid
							//
							{
							}
						}


						//  Determine if the file size is correct
						//
						if( (lpDir->attr & (A_SUBDIR | A_VOLUME)) == 0 )
						{
							//  Determine the number of clusters which would
							//  be allocated for the files size
							//
							ulClustersInFile = lpDir->size >> ucBPCShift;
							if( lpDir->size & (ulClusterSize-1) )
							{
								ulClustersInFile++;
							}


							//  Display an error message if the file size
							//  and the number of clusters in the chain
							//  do not coincide.
							//
	  						if( ulClusterCount != ulClustersInFile )
							{
								//  Determine if this is the first pass through
								//  the drive.  If this is the first pass through
								//  the drive ...
								//
								if( fFirstPass )
								{
									//  Determine if we are fixing errors.  If we
									//  are not fixing errors, display a message
									//  stating that errors were found but will
									//  not be fixed (committed).
									//
									ShowFoundErrorsMsg();


									//  Update the return code
									//
									ucReturnCode |= GENERAL_ERROR;


									//  Display the current path
									//
									printf("%s\n", szWorkingPath );


									//  Display a message stating that the
									//  file size has been adjusted
									//
									printf( szFileSizeErrorMsg );
								}
							}


							//  Determine if there are the correct number
							//  of clusters allocated to this file.  If the
							//  correct number of clusters is not allocated ...
							//
							if( ulClusterCount != ulClustersInFile )
							{
								//  Adjust the file size
								//
								lpDir->size = ulClusterCount << ucBPCShift;


								//  Record the fact that this sector has
								//  been modified.
								//
								MarkBufferDirty( &ulSectorNumber );
							}
						}


						//  On the first pass we will build all our statistics
						//
						if( fFirstPass )
						{
							//
							//
							if( lpDir->attr & A_SUBDIR )
							{
								//  Update the directory count statistics
								//
								stat[DDIR].number++;


								//  Update the number of bytes used by this
								//  directory.  For now, this number will be
								//  the number of clusters allocated to
								//  the directory
								//
								stat[DDIR].bytes += ulAllocatedClusterCount;
							}
							else
							{
								//  Update the hidden file count and size
								//  statistics
								//
								if( lpDir->attr & A_HIDDEN )
								{
									stat[DHIDDEN].number++;
									stat[DHIDDEN].bytes += ulClusterCount;
								}
								else
								{
									//  Update the file count statistics
									//
									stat[DUSER].number++;
									stat[DUSER].bytes += ulClusterCount;
								}
							}
						}


						//  If this is a directory then we need to recurse
						//  into it and examine the rest of the files
						//
						if( lpDir->attr & A_SUBDIR )
						{
							//
							//  Examine this directory
							//
#if 1
							ucCheckCode = CheckDirectoryAtCluster(
									ulDirStartingCluster, fFirstPass, 0 );
#else
	ucCheckCode = 0;
#endif
							ucReturnCode |= ucCheckCode;
							if( ucCheckCode == INVALID_DIRECTORY )
							{
								//  Display an error message
								//
								if( !DoShowVerbose() && fFirstPass )
								{
									printf( "%s\n", szWorkingPath );
									printf( szInvalidDirMsg );
								}


								//  Determine if we are fixing errors.  If we
								//  are fixing errors ...
								//
								if( DoFixErrors() )
								{
									//  Determine if the user wants to save this
									//  invalid directory entry as a file.  If
									//  the user wants to save this invalid
									//  directory entry as a file ...
									//
									if( getyesno(szConvertDirToFileMsg, DEFAULT_YES) )
									{
										//  Get the directory entry at the
										//  specified sector and index
										//
										lpDir = GetDirEntryAtSector( uEntryIndex,
												ulSectorNumber );


										//  Change this directory entry attribute
										//  to that of a file
										//
										lpDir->attr = A_NORMAL;


										//  Update the size of the file
										//
										lpDir->size = ulClusterCount << ucBPCShift;


										//  Record the fact that this sector has
										//  been modified.
										//
										MarkBufferDirty( &ulSectorNumber );
									}
								}
							}


							//  Get the directory entry at the specified sector
							//  and index
							//
							lpDir = GetDirEntryAtSector( uEntryIndex,
									ulSectorNumber );
						}


						//  Set the working path back to the current directory
						//
	 					RemoveName( szWorkingPath );
					}
				}
			}
			else
			{
				//
				//
				if( fInLFNSequence )
				{
					//  Determine if we are fixing errors.  If we
					//  are not fixing errors, display a message
					//  stating that errors were found but will
					//  not be fixed.
					//
					ShowFoundErrorsMsg();


					//  Display a message stating that an orphaned
					//  LFN entry was found and deleted.
					//
					printf( szOrphanedLFNEntry );


					//  Set a flag indicating to delete the LFN
					//  sequence.
					//
					fInLFNSequence = FALSE;
					fDeleteLFNSequence = TRUE;


					//  Set the sequence number to start deleting
					//  the LFN sequence
					//
					ucLFNSequenceNumber = ucLFNStartSequenceNumber;


					//
					//
					uEntryIndex = uLFNEntryIndex;
					ulSectorNumber = ulLFNSectorNumber;
					continue;
				}


				//
				//
				fDeleteLFNSequence = FALSE;
			}


			//  Increment the directory entry index
			//
			uEntryIndex++;


			//  Determine if we have exceeded the number of
			//  entries in this sector.  If we have exceeded
			//  the number of entries in this sector...
			//
			if( uEntryIndex >= uEntriesPerSector )
			{
				//  Determine the next sector number in the root
				//
				ulSectorNumber++;


				//  Adjust the number of entries remaining and the entry index
				//
				uEntryIndex = 0;
				uNumberOfEntries -= uEntriesPerSector;
			}
		}
	}

	return ucReturnCode;
}


//  Determine if the directory entry name is valid
//
uchar IsValidName( char far * lpENAME, char * szName )
{
	static char * szInvalidNames[] = {
		"CON",
		"COM1",
		"COM2",
		"COM3",
		"COM4",
		"LPT1",
		"LPT2",
		"LPT3",
		"PRN",
		"STDOUT",
		"STDIN",
		"STDERR",
		""
	};
	uchar ucNameError = NAME_ERROR_NONE;
	uchar ucIndex;
	char cCharacter;


	//  Go through each of the characters in the file name and
	//  make sure that they are valid
	//
	for( ucIndex=0; ucIndex < ENAMESIZE; ucIndex++ )
	{
		//  Get this character into a local variable
		//
		cCharacter = lpENAME[ucIndex];


		//  Determine if this is a valid character.  If this is not
		//  a valid character ...
		//
		if( !IsValidNameCharacter(cCharacter) )
		{
			//  Make sure that this isn't just a case problem.  If
			//  this is just a case problem ...
			//
			if( cCharacter != toupper(cCharacter) )
			{
				//  We can fix a case problem
				//
				lpENAME[ucIndex] = toupper(cCharacter);
				ucNameError = NAME_ERROR_CASE;
			}
			else
			{
				//  Return that this is an invalid name containing invalid
				//  characters
				//
				ucNameError = NAME_ERROR_INVALID;
				break;
			}
		}
	}


	//  If there is no invalid characters thus far ...
	//
	if( ucNameError != NAME_ERROR_INVALID )
	{
		//  Make sure this file name is not reserved
		//
		ucIndex = 0;
		while( szInvalidNames[ucIndex][0] )
		{
			//  Determine if this name matches one of the names in the
			//  device list
			//
			if( strcmpi( szInvalidNames[ucIndex], szName) == 0 )
			{
				//  You cannot access/create a file/dir with a device name
				//
				ucNameError = NAME_ERROR_INVALID;
				break;
			}

			//  Go to the next name
			//
			ucIndex++;
		}
	}


	//  Return the error code
	//
	return ucNameError;
}


//  Determine if this is a valid character.
//
bool IsValidNameCharacter( char cCharacter )
{
	static char * szInvalidList = ".|\"\\/[]?><=:;,+*()";
	uchar ucIndex;


	//  For optimization, determine if this character is A-Z or 0-9 or 0x20
	//
	if( (cCharacter >= 'A') && (cCharacter <= 'Z') )
	{
		return TRUE;
	}
	if( (cCharacter >= '0') && (cCharacter <= '9') )
	{
		return TRUE;
	}
	if( cCharacter == 0x20 )
	{
		return TRUE;
	}


	//  Perform some range check validation
	//
	if( cCharacter <= 0x19 )
	{
		return FALSE;
	}


	//  Perform some range check validation
	//
	if( cCharacter >= 0x7f )
	{
		return FALSE;
	}


	//  If this is a lower case letter a-z then it is not valid
	//
	if( toupper(cCharacter) != cCharacter )
	{
		return FALSE;
	}


	//  Determine if this character is in the invalid list
	//
	ucIndex = 0;
	while( szInvalidList[ucIndex] )
	{
		//  Determine if this character is in the invalid list
		//
		if( szInvalidList[ucIndex] == cCharacter )
		{
			return FALSE;
		}


		//  Go on to the next invalid character
		//
		ucIndex++;
	}


	//  Return that the character is valid
	//
	return TRUE;
}


//  check the directory at the specified cluster
//
uchar CheckDirectoryAtCluster( ulong ulStartingCluster, bool fFirstPass,
		ulong ulParentCluster )
{
	static LPDPB lpDPB;
	static ulong ulMaxClusterNumber;
	static unsigned uEntriesPerSector;
	static unsigned uEntriesPerCluster;
	static char szFileName[13];
	static ulong ulNextCluster;
	static uchar ucError;
	static ulong ulFirstCrossLink;
	static ulong ulClustersInFile;
	static uchar ucBPCShift;
	static uchar ucBPSShift;
	static unsigned uBPS;
	static ulong ulClusterSize;
	static ulong ulClusterCount;
	static char * szWorkingPath;
	static unsigned uLength;
	static uchar ucCheckCode;
	static bool fFound;
	static unsigned uFileTime;
	static unsigned uFileDate;
	uchar ucReturnCode = 0;
	ulong ulCurDirCluster;
	bool fDoneFirstEntry;
	bool fDoneSecondEntry;
	static bool fDeleteLFNSequence;
	static bool fInLFNSequence;
	static uchar ucLFNSequenceNumber;
	static uchar ucLFNStartSequenceNumber;
	static unsigned uLFNEntryIndex;
	static ulong ulLFNSectorNumber;
	static uchar ucLFNChecksum;
	static uchar ucTemp;
	unsigned uCountSPC;
	bool fDoneProcessingDir = FALSE;
	unsigned uEntryIndex;
	LPDIR lpDir;
	ulong ulSectorNumber;
	ulong ulDirStartingCluster;
	ulong ulAllocatedClusterCount;
	bool fRestOfDirIsAvailable = FALSE;
	bool fShowedInvalidEntriesMsg = FALSE;


#if 0
	static ulong ulCount = 1;
	static ulong ulDepth = 0;
#endif


	//  Determine the current working directory
	//
	szWorkingPath = GetWorkingPath();


	//  Determine if this cluster number is zero.  If this cluster
	//  number is zero...
	if( !ulStartingCluster )
	{
		//  This is not an error.  No clusters have been
		//  allocated to this directory
		//
		return ucReturnCode;
	}


	//  If we don't already have the maximum cluster number, then
	//  get the maximum cluster number and some of the other
	//  values we need
	//
	if( !ulMaxClusterNumber )
	{
		//  Get the DPB for the drive
		//
		GetDPB( &lpDPB );


		//  Determine the maximum cluster number
		//
		if( IsFat32Drive(lpDPB) )
		{
			ulMaxClusterNumber = lpDPB->ulMaxCluster;
		}
		else
		{
			ulMaxClusterNumber = lpDPB->hiclus;
		}

		//  Determine the number of directory entries in a cluster
		//
		uEntriesPerSector = lpDPB->bytes_per_sec / sizeof(DIR);
		uEntriesPerCluster = uEntriesPerSector << lpDPB->shift;
//		printf("uEntriesPerCluster: %u\n", uEntriesPerCluster );


		//  Calculate the bytes per cluster conversion shift factor
		//
		{
			//  Determine the shift factor for the bytes per sector
			//
			ucBPSShift = 0;
			uBPS = lpDPB->bytes_per_sec;
			while( uBPS )
			{
				uBPS >>= 1;
				ucBPSShift++;
			}
			ucBPSShift--;


			//  Determine the shift factor for bytes per cluster
			//
			ucBPCShift = ucBPSShift + lpDPB->shift;


			//  Determine the cluster size
			//
			ulClusterSize = 1L << ucBPCShift;
		}
	}
	ASSERT_LPDPB( lpDPB );


	//  Make sure that this this is a valid cluster entry
	//
	{
		//  If this cluster value is less than two or if this cluster
		//  value exceeds the maximum cluster number ...
		//
		if( (ulStartingCluster < 2) ||
				(ulStartingCluster > ulMaxClusterNumber) )
		{
			//  this is not a valid cluster number
			//
		}
	}


	//
	//
	ulCurDirCluster = ulStartingCluster;


	//  Determine if this is the root.  If this is the root
	//  directory ...
	//
	if( (szWorkingPath[3] == 0) && (szWorkingPath[2] == '\\') )
	{
		//  Clear the flags which will cause us to check for the '.' and '..'
		//  entries in the subdirectory
		//
		fDoneFirstEntry  = TRUE;
		fDoneSecondEntry = TRUE;
	}
	else
	{
		//  Set the flags which will cause us to check for the '.' and '..'
		//  entries in the subdirectory
		//
		fDoneFirstEntry  = FALSE;
		fDoneSecondEntry = FALSE;
	}


	//  Check all the entries in this directory
	//
	while(TRUE)
	{
		//
		//
		uCountSPC = 0;


		//  Determine the starting sector number for this cluster
		//
		{
			ulSectorNumber = ulCurDirCluster - 2;
			ulSectorNumber *= (lpDPB->ms_spc + 1);
			if( IsFat32Drive(lpDPB) )
			{
				ulSectorNumber += lpDPB->ulFirstDataSector;
			}
			else
			{
				ulSectorNumber += lpDPB->data_start;
			}
		}


		//
		//
		while( (uCountSPC < (lpDPB->ms_spc + 1)) && (!fDoneProcessingDir) )
		{
			//  Initialize the variable which contains the
			//  current entry index to zero
			//
			uEntryIndex = 0;


			//
			//
			while( uEntryIndex < uEntriesPerSector )
			{
				//  Check if the user wants to abort CTRL-C
				//
				CheckUserAbort();


				//  Get the directory entry at the specified sector
				//  and index
				//
				lpDir = GetDirEntryAtSector( uEntryIndex, ulSectorNumber );


				//  Check if we are looking for the '.' entry
				//
				if(!fDoneFirstEntry)
				{
					//  A special condition where microsoft displays a
					//  different message
					//
					if( lpDir->name[0] == 0 )
					{
						//  Display the error information
						//
						if( fFirstPass )
						{
							if( DoShowVerbose() )
							{
								printf( "%s\n", szWorkingPath );
								printf( szDirEmptyMsg );
								printf( szUnrecoverableMsg );
							}


							//  Set a special return code
							//
							ucReturnCode = INVALID_DIRECTORY;
						}


						//  We cannot process this directory
						//
						fDoneProcessingDir = TRUE;
						break;
					}


					//  Make sure that this entry is the '.' entry.  If
					//  this is the '.' entry
					//
					EnameToDot( lpDir->name, szFileName );
					if( strcmp( szFileName, "." ) == 0 )
					{
						//  Validate the '.' entry.  If the '.' entry was
						//  not correct ...
						//
						if( ModifiedSpecialEntry( lpDir, ulStartingCluster ) )
						{
							//  Record the fact that this sector has
							//  been modified.
							//
							MarkBufferDirty( &ulSectorNumber );
						}


						//  Proceed onto the next entry
						//
						uEntryIndex++;
						lpDir = GetDirEntryAtSector( uEntryIndex, ulSectorNumber );


						//  Indicate that we are done looking for the '.' entry
						//
						fDoneFirstEntry = TRUE;
					}
					else
					{
						//  Determine if we are fixing errors.  If we
						//  are not fixing errors, display a message
						//  stating that errors were found but will
						//  not be fixed (committed).
						//
						ShowFoundErrorsMsg();


						//  Display this error information
						//
						if( fFirstPass && DoShowVerbose() )
						{
							uLength = strlen( szWorkingPath );
		 					AddName( szWorkingPath, "." );
							printf( "%s\n", szWorkingPath );
							printf( szNoExist );
							szWorkingPath[uLength] = 0;
						}


						//  Determine if we can fix this problem.  If
						//  we cannot fix this problem ...
						//
						if( lpDir->name[0] != (char)0xe5 )
						{
							if( fFirstPass )
							{
								if( DoShowVerbose() )
								{
									uLength = strlen( szWorkingPath );
			 						AddName( szWorkingPath, "." );
									printf( "%s\n", szWorkingPath );
									printf( szNoRecover1, ". ", szNoRecover2 );
									szWorkingPath[uLength] = 0;
								}


								//  Set a special return code
								//
								ucReturnCode = INVALID_DIRECTORY;
							}
						}
						else
						{
							//  Display an error message
							//
							if( !DoShowVerbose() && fFirstPass )
							{
								printf( "%s\n", szWorkingPath );
								printf( szInvalidDirMsg );
							}


							//  fix the problem
							//
							{
								//  Set the name of the entry
								//
								_fmemset( lpDir->name, 0x20, 11 );
								lpDir->name[0] = '.';


								//  Set the attributes
								//
								lpDir->attr = A_SUBDIR;


								//  Set the size of the entry
								//
								lpDir->size = 0;


								//  Set the starting cluster number
								//
								lpDir->uStartingClusterLo =
										(unsigned)ulStartingCluster;
								lpDir->uStartingClusterHi =
										(unsigned)(ulStartingCluster >> 16);


								//  Determine the current date and time and set the
								//  date and time in the directory entry
								//
								GetDateAndTime( &uFileDate, &uFileTime );
								lpDir->date = uFileDate;
								lpDir->time = uFileTime;


								//  Record the fact that this sector has
								//  been modified.
								//
								MarkBufferDirty( &ulSectorNumber );
							}


							//  Proceed onto the next entry
							//
							uEntryIndex++;
							lpDir = GetDirEntryAtSector( uEntryIndex, ulSectorNumber );
						}
					}
				}


				//  Check if we are looking for a '..' entry
				//
				if(!fDoneSecondEntry)
				{
					//  Get the name in the directory entry
					//
					EnameToDot( lpDir->name, szFileName );


					//  Look for the '..' entry
					//
					fFound = (strcmp( szFileName, ".." ) == 0 );


//  This makes sense, but MS does not do this
//
#if 0
					//  If we did not find the '.' or '..' entries
					//
					if( !fFound && !fDoneFirstEntry )
					{

						//  Proceed onto the next entry
						//
						uEntryIndex++;
						lpDir = GetDirEntryAtSector( uEntryIndex, ulSectorNumber );


						//  Get the name in the directory entry
						//
						EnameToDot( lpDir->name, szFileName );


						//  Look for the '..' entry
						//
						fFound = (strcmp( szFileName, ".." ) == 0 );
					}
#endif

					//  Determine if we found the '..' entry.  If we found
					//  the '..' entry ...
					//
					if( fFound )
					{
						//  The '..' entry can not be the first entry.
						//
						if( uEntryIndex == 0 )
						{
							//  Display a message stating that this
							//  suituation is unrecoverable only if
							//  we are on the first pass and in
							//  verbose mode
							//
							if( fFirstPass )
							{
								if( DoShowVerbose() )
								{
									printf( szUnrecoverableMsg );
								}


								//  Set a special return code
								//
								ucReturnCode = INVALID_DIRECTORY;
							}


							//  We cannot process this directory
							//
							fDoneProcessingDir = TRUE;
							break;
						}

						//  Validate the '..' entry.  If the '..' entry
						//  was not correct ...
						//
						if( ModifiedSpecialEntry( lpDir, ulParentCluster ) )
						{
							//  Record the fact that this sector has
							//  been modified.
							//
							MarkBufferDirty( &ulSectorNumber );
						}
					}
					else
					{
						//  Determine if we are fixing errors.  If we
						//  are not fixing errors, display a message
						//  stating that errors were found but will
						//  not be fixed (committed).
						//
						ShowFoundErrorsMsg();


						//  Display this error information only if we
						//  are in verbose mode
						//
						if( fFirstPass && DoShowVerbose() )
						{
							uLength = strlen( szWorkingPath );
	 						AddName( szWorkingPath, ".." );
							printf( "%s\n", szWorkingPath );
							printf( szNoExist );
							szWorkingPath[uLength] = 0;
						}


						//  Determine if we can fix this problem.  If
						//  we cannot fix this problem ...
						//
						if( lpDir->name[0] != (char)0xe5 )
						{
							if( fFirstPass )
							{
								if( DoShowVerbose() )
								{
									uLength = strlen( szWorkingPath );
		 							AddName( szWorkingPath, ".." );
									printf( "%s\n", szWorkingPath );
									printf( szNoRecover1, "..", szNoRecover3 );
									szWorkingPath[uLength] = 0;
									printf( szUnrecoverableMsg );
								}


								//  Set a special return code
								//
								ucReturnCode = INVALID_DIRECTORY;
							}


							//  We cannot process this directory
							//
							fDoneProcessingDir = TRUE;
							break;
						}
						else
						{
							//  Display an error message
							//
							if( !DoShowVerbose() && fFirstPass )
							{
								printf( "%s\n", szWorkingPath );
								printf( szInvalidDirMsg );
							}


							//  fix the problem
							//
							{
								//  Set the name of the entry
								//
								_fmemset( lpDir->name, 0x20, 11 );
								lpDir->name[0] = '.';
								lpDir->name[1] = '.';


								//  Set the attributes
								//
								lpDir->attr = A_SUBDIR;


								//  Set the size of the entry
								//
								lpDir->size = 0;


								//  Set the starting cluster number
								//
								lpDir->uStartingClusterLo =
										(unsigned)ulParentCluster;
								lpDir->uStartingClusterHi =
										(unsigned)(ulParentCluster >> 16);


								//  Determine the current date and time and set the
								//  date and time in the directory entry
								//
								GetDateAndTime( &uFileDate, &uFileTime );
								lpDir->date = uFileDate;
								lpDir->time = uFileTime;


								//  Record the fact that this sector has
								//  been modified.
								//
								MarkBufferDirty( &ulSectorNumber );
							}
						}
					}


					//  Indicate that we are done looking for the '.' entry
					//
					fDoneFirstEntry = TRUE;


					//  Indicate that the '..' entry has been checked
					//
					fDoneSecondEntry = TRUE;


					//  Proceed onto the next entry
					//
					uEntryIndex++;
					lpDir = GetDirEntryAtSector( uEntryIndex, ulSectorNumber );
				}


				//  Detemine if this is a blank entry. if this is
				//  a blank entry ...
				//
				if( lpDir->name[0] == 0 )
				{
					//  Set a flag indicating that all of the following
					//  entries should be blank
					//
					fRestOfDirIsAvailable = TRUE;
					fDoneProcessingDir = TRUE;
				}
				else if( fRestOfDirIsAvailable )
				{
					//
					//  This entry is invalid
					//

					//  Determine if this is the first pass through the
					//  drive
					//
					if( fFirstPass )
					{
						//  Determine if we are fixing errors.  If we
						//  are not fixing errors, display a message
						//  stating that errors were found but will
						//  not be fixed.
						//
						ShowFoundErrorsMsg();


						if( !fShowedInvalidEntriesMsg )
						{
							//  Display a message stating that an orphaned
							//  LFN entry was found and deleted.
							//
							if( !DoShowVerbose() )
							{
								printf( "Directory %s\n", szWorkingPath );
							}
							else
							{
								printf( "%s\n", szWorkingPath );
							}
							printf( szInvalidEntries );
							fShowedInvalidEntriesMsg = TRUE;
						}


						//  Mark this entry as deleted SCANDISK fashion.
						//
						lpDir->name[0] = 0;


						//  Record that there is one more available
						//  root entry if this is the root directory
						//
						if( szWorkingPath[3] == 0 )
						{
							IncAvailableRootEntries();
						}


						//  Record the fact that this sector has
						//  been modified.
						//
						MarkBufferDirty( &ulSectorNumber );
					}
				}


				//  Determine if this is a deleted entry. if this
				//  is not a deleted entry ...
				//
				else if( (uchar)(lpDir->name[0]) != 0xE5 )
				{
					//  Determine if this is the volume label
					//
					if( (lpDir->attr & (~A_ARCHIVE)) == A_VOLUME )
					{
						if( fInLFNSequence )
						{
							//  Determine if we are fixing errors.  If we
							//  are not fixing errors, display a message
							//  stating that errors were found but will
							//  not be fixed.
							//
							ShowFoundErrorsMsg();


							//  Display a message stating that an orphaned
							//  LFN entry was found and deleted.
							//
							printf( szOrphanedLFNEntry );


							//  Set a flag indicating to delete the LFN
							//  sequence.
							//
							fInLFNSequence = FALSE;
							fDeleteLFNSequence = TRUE;


							//  Set the sequence number to start deleting
							//  the LFN sequence
							//
							ucLFNSequenceNumber = ucLFNStartSequenceNumber;


							//
							//
							uEntryIndex = uLFNEntryIndex;
							ulSectorNumber = ulLFNSectorNumber;
							continue;
						}


						//  We are no longer deleting an LFN sequence
						//
						fDeleteLFNSequence = FALSE;


						//  Validate the volume label
						//

					}
					else if( lpDir->attr == A_LONGNAME )
					{
						//
						//  if this is a long file name entry
						//


						//  Validate the long directory entry
						//
						if( fLFNsSupported )
						{
							//  Validate the long directory entry
							//
							{
								//  Determine if we are deleting an LFN sequence
								//
								if( fDeleteLFNSequence )
								{
									//  Make sure that this sequence matches what
									//  it is supposed to
									//
									if( (lpDir->name[0] == ucLFNSequenceNumber) &&
											(lpDir->bCentiSeconds == ucLFNChecksum) )
									{
										//  Determine what the next sequence
										//  number should be
										//
										ucLFNSequenceNumber = lpDir->name[0];
										if( ucLFNSequenceNumber > 0x40 )
										{
											ucLFNSequenceNumber -= 0x40;
										}
										ucLFNSequenceNumber--;


										//  Mark this entry as deleted.
										//
										lpDir->name[0] = (char)0xe5;


										//  Record that there is one more available
										//  root entry if this is the root directory
										//
										if( szWorkingPath[3] == 0 )
										{
											IncAvailableRootEntries();
										}


										//  Record the fact that this sector has
										//  been modified.
										//
										MarkBufferDirty( &ulSectorNumber );
									}
									else
									{
										//  We are no longer deleting an LFN sequence
										//
										fDeleteLFNSequence = FALSE;


										//
										//
										continue;
									}
								}
								else
								{
									//  Determine if we are are in the middle
									//  of processing an LFN sequence
									//
									if( fInLFNSequence )
									{
										//
										//
										if( (lpDir->name[0] != ucLFNSequenceNumber) ||
												(lpDir->bCentiSeconds != ucLFNChecksum) )
										{
											//  Determine if we are fixing errors.  If we
											//  are not fixing errors, display a message
											//  stating that errors were found but will
											//  not be fixed.
											//
											ShowFoundErrorsMsg();


											//  Display a message stating that an orphaned
											//  LFN entry was found and deleted.
											//
											printf( szOrphanedLFNEntry );


											//  Set a flag indicating to delete the LFN
											//  sequence.
											//
											fInLFNSequence = FALSE;
											fDeleteLFNSequence = TRUE;


											//  Set the sequence number to start deleting
											//  the LFN sequence
											//
											ucLFNSequenceNumber = ucLFNStartSequenceNumber;


											//
											//
											uEntryIndex = uLFNEntryIndex;
											ulSectorNumber = ulLFNSectorNumber;
											continue;
										}
										else
										{
											//  Determine the next LFN sequence number
											//
											ucLFNSequenceNumber = lpDir->name[0];
											if( ucLFNSequenceNumber > 0x40 )
											{
												ucLFNSequenceNumber -= 0x40;
											}
											ucLFNSequenceNumber--;
										}
									}
									else
									{
										//  Record the entry index and sector for the
										//  first entry in the sequence.
										//
										uLFNEntryIndex = uEntryIndex;
										ulLFNSectorNumber = ulSectorNumber;
										ucLFNStartSequenceNumber = lpDir->name[0];


										//  Record the chksum
										//
										ucLFNChecksum = lpDir->bCentiSeconds;


										//  Determine if this is the first LFN entry
										//  in the sequence.  If this is the first
										//  in the sequence ...
										//
										if( lpDir->name[0] > 0x40 )
										{
											//  Set a flag to indicate that we are
											//  processing an LFN sequence
											//
											fInLFNSequence = TRUE;


											//  Determine the next LFN sequence number
											//
											ucLFNSequenceNumber = lpDir->name[0] - 0x40 - 1;
										}
										else
										{
											//  Determine if we are fixing errors.  If we
											//  are not fixing errors, display a message
											//  stating that errors were found but will
											//  not be fixed.
											//
											ShowFoundErrorsMsg();


											//  Display a message stating that an orphaned
											//  LFN entry was found and deleted.
											//
											printf( szOrphanedLFNEntry );


											//  Set a flag indicating to delete the LFN
											//  sequence.
											//
											fDeleteLFNSequence = TRUE;


											//  Set the sequence number to start deleting
											//  the LFN sequence
											//
											ucLFNSequenceNumber = lpDir->name[0];


											//
											//
											continue;
										}
									}
								}
							}
						}
					}
					else
					{
						//
						//  Validate the 8.3 directory entry
						//
						{
							//  Determine if we are at the short name part
							//  of an LFN entry.  If we are at the short
							//  name part of an LFN entry ...
							//
							if( fInLFNSequence )
							{
								//  Compute the checksum for the short name entry
								//
								ComputeENAMEChecksum( lpDir->name, &ucTemp );


								//  Determine if the previous LFN sequences checksum
								//  matches the short name.  If the checksums do not
								//  match ...
								//
								if( (ucTemp != ucLFNChecksum) ||
										(ucLFNSequenceNumber != 0) )
								{
									//  Determine if we are fixing errors.  If we
									//  are not fixing errors, display a message
									//  stating that errors were found but will
									//  not be fixed.
									//
									ShowFoundErrorsMsg();


									//  Display a message stating that an orphaned
									//  LFN entry was found and deleted.
									//
									printf( szOrphanedLFNEntry );


									//  Set a flag indicating to delete the LFN
									//  sequence.
									//
									fInLFNSequence = FALSE;
									fDeleteLFNSequence = TRUE;


									//  Set the sequence number to start deleting
									//  the LFN sequence
									//
									ucLFNSequenceNumber = ucLFNStartSequenceNumber;


									//
									//
									uEntryIndex = uLFNEntryIndex;
									ulSectorNumber = ulLFNSectorNumber;
									continue;
								}
							}


							//  We are no longer deleting an LFN sequence
							//
							fDeleteLFNSequence = FALSE;
							fInLFNSequence = FALSE;


							//  Convert the file name in ENAME form into a
							//  ASCIIZ string and append it to our current
							//  path.
							//
							EnameToDot( lpDir->name, szFileName );
							AddName( szWorkingPath, szFileName );


							//  Determine the number of clusters which would
							//  be allocated for the files size
							//
							ulClustersInFile = lpDir->size >> ucBPCShift;
							if( lpDir->size & (ulClusterSize-1) )
							{
								ulClustersInFile++;
							}


							//  Determine the starting cluster for this
							//  directory
							//
							ulDirStartingCluster = lpDir->uStartingClusterLo;
							if( IsFat32Drive(lpDPB) )
							{
								ulDirStartingCluster |=
										(ulong)(lpDir->uStartingClusterHi) << 16;
							}


							//  Determine if this is our first time going through
							//  the directory tree of the drive.  If this is our
							//  first time going through the directory tree of
							//  of the drive ...
							//
		  					if( fFirstPass )
							{
								//  Display the file we are checking if the user
								//  requested so
								//
								if( DoShowVerbose() )
								{
									//  Display the full path for the user
									//
									if( lpDir->attr & A_SUBDIR )
									{
										printf( showdir_str, szWorkingPath );
									}
									else
									{
										printf("%s\n", szWorkingPath );
									}
								}


								//  Examine and/or fix the FAT chain
								//
								ucError = FollowFATChain( ulDirStartingCluster,
										&ulClusterCount, &ulFirstCrossLink,
										&ulAllocatedClusterCount, fFirstPass );
								if( ucError )
								{
									//  Determine if we are fixing errors.  If we
									//  are not fixing errors, display a message
									//  stating that errors were found but will
									//  not be fixed.
									//
									ShowFoundErrorsMsg();


									//  Determine if the first cluster of this
									//  chain was invalid
									//
									if( ucError & CHAIN_FIRST_INVALID )
									{
										//  Update the return code
										//
										ucReturnCode |= ucError;


										//  Display the current path
										//
										printf("%s\n", szWorkingPath );


										//  Display a message stating that the
										//  file size has been adjusted
										//
										printf( szFirstInvalidAllocationMsg );


										//  Get the directory entry at the specified
										//  sector and index
										//
										lpDir = GetDirEntryAtSector( uEntryIndex,
												ulSectorNumber );


										//  The starting cluster is invalid.  Update
										//  the directory entry
										//
										lpDir->uStartingClusterHi = 0;
										lpDir->uStartingClusterLo = 0;


										//  Record the fact that this sector has
										//  been modified.
										//
										MarkBufferDirty( &ulSectorNumber );
									}
									else if( ucError & CHAIN_INVALID )
									{
										//
										//  Determine if the chain contains invalid
										//  cluster values.  If the chain contains
										//  invalid cluster values ...
										//

										//  Update the return code
										//
										ucReturnCode |= ucError;


										//  Display the current path
										//
										printf("%s\n", szWorkingPath );


										//  Display a message stating that the
										//  file size has been adjusted
										//
										printf( szInvalidAllocationMsg );
									}


									//  Only adjust the file size for invalid
									//  chains.
									//
									if( ucError & ( CHAIN_FIRST_INVALID |
											CHAIN_INVALID | CHAIN_ADJUST_SIZE) )
									{

										//  Only adjust the file size if the number
										//  of clusters required for file size is
										//  different than the number of clusters
										//  in the chain
										//
										if( ulClustersInFile != ulClusterCount )
										{
			  								//  Get the directory entry at the specified
											//  sector and index
											//
											lpDir = GetDirEntryAtSector( uEntryIndex,
													ulSectorNumber );


											//  Update the file size
											//
											lpDir->size = ulClusterCount << ucBPCShift;


											//  Record the fact that this sector has
											//  been modified.
											//
											MarkBufferDirty( &ulSectorNumber );
										}
									}
								}
							}
							else
							{
								//  Examine and/or fix the FAT chain
								//
								ucError = _FollowFATChain( ulDirStartingCluster,
										&ulClusterCount, &ulFirstCrossLink,
										&ulAllocatedClusterCount, fFirstPass,
										TRUE );
								if( ucError )
								{
									//  Determine if the file contains cross-linked
									//  clusters.  If the file contains cross-linked
									//  clusters ...
									//
									if( ucError & CHAIN_CROSS_LINKED )
									{
										//  Update the return code
										//
										ucReturnCode |= ucError;


										//  Display the current path
										//
										printf("%s\n", szWorkingPath );


										//  Display a message stating that cross-linked
										//  files were found
										//
										if( DoFixErrors() )
										{
											printf( szFixCrossLinkedMsg, szWas,
													ulFirstCrossLink );
										}
										else
										{
											printf( szFixCrossLinkedMsg, szIs,
													ulFirstCrossLink );
										}


										//  Fix the cross-linked files
										//
		 								_FollowFATChain( ulDirStartingCluster,
												&ulClusterCount, &ulFirstCrossLink,
												&ulAllocatedClusterCount, fFirstPass,
												FALSE );
									}
								}
							}


							//  Get the directory entry at the specified sector
							//  and index
							//
							lpDir = GetDirEntryAtSector( uEntryIndex,
									ulSectorNumber );


							if( fFirstPass )
							{
								//  Determine if the directory attributes are valid
								//
								{
								}
							}


							//  Determine if the file size is correct
							//
							if( (lpDir->attr & (A_SUBDIR | A_VOLUME)) == 0 )
							{
								//  Determine the number of clusters which would
								//  be allocated for the files size
								//
								ulClustersInFile = lpDir->size >> ucBPCShift;
								if( lpDir->size & (ulClusterSize-1) )
								{
									ulClustersInFile++;
								}


								//  Display an error message if the file size
								//  and the number of clusters in the chain
								//  do not coincide.
								//
		  						if( ulClusterCount != ulClustersInFile )
								{
									//  Determine if this is the first pass through
									//  the drive.  If this is the first pass through
									//  the drive ...
									//
									if( fFirstPass )
									{
										//  Determine if we are fixing errors.  If we
										//  are not fixing errors, display a message
										//  stating that errors were found but will
										//  not be fixed (committed).
										//
										ShowFoundErrorsMsg();


										//  Update the return code
										//
										ucReturnCode |= GENERAL_ERROR;


										//  Display the current path
										//
										printf("%s\n", szWorkingPath );


										//  Display a message stating that the
										//  file size has been adjusted
										//
										printf( szFileSizeErrorMsg );
									}
								}


								//  Determine if there are the correct number
								//  of clusters allocated to this file.  If the
								//  correct number of clusters is not allocated ...
								//
								if( ulClusterCount != ulClustersInFile )
								{
									//  Adjust the file size
									//
									lpDir->size = ulClusterCount << ucBPCShift;


									//  Record the fact that this sector has
									//  been modified.
									//
									MarkBufferDirty( &ulSectorNumber );
								}
							}


							//  On the first pass we will build all our statistics
							//
							if( fFirstPass )
							{
								//
								//
								if( lpDir->attr & A_SUBDIR )
								{
									//  Update the directory count statistics
									//
									stat[DDIR].number++;


									//  Update the number of bytes used by this
									//  directory.  For now, this number will be
									//  the number of clusters allocated to
									//  the directory
									//
									stat[DDIR].bytes += ulAllocatedClusterCount;
								}
								else
								{
									//  Update the hidden file count and size
									//  statistics
									//
									if( lpDir->attr & A_HIDDEN )
									{
										stat[DHIDDEN].number++;
										stat[DHIDDEN].bytes += ulClusterCount;
									}
									else
									{
										//  Update the file count statistics
										//
										stat[DUSER].number++;
										stat[DUSER].bytes += ulClusterCount;
									}
								}
							}


							//  If this is a directory then we need to recurse
							//  into it and examine the rest of the files
							//
							if( lpDir->attr & A_SUBDIR )
							{
								//
								//  Examine this directory
								//
#if 0
	{
//		if( ulCount <= 2 )
		{
			printf("ulCount: %lu  ulDepth: %lu\n", ulCount, ulDepth );
			if( ulCount > ulDepth )
			{
				ulDepth = ulCount;
			}
			ulCount++;
								ucCheckCode = CheckDirectoryAtCluster(
										ulDirStartingCluster, fFirstPass,
										ulStartingCluster );
			ulCount--;
		}
	}
#else
#if 1
								ucCheckCode = CheckDirectoryAtCluster(
										ulDirStartingCluster, fFirstPass,
										ulStartingCluster );
#else
	ucCheckCode = 0;
#endif
#endif
								ucReturnCode |= ucCheckCode;
								if( ucCheckCode == INVALID_DIRECTORY )
								{
									//  Display an error message
									//
									if( !DoShowVerbose() && fFirstPass )
									{
										printf( "%s\n", szWorkingPath );
										printf( szInvalidDirMsg );
									}


									//  Determine if we are fixing errors.  If we
									//  are fixing errors ...
									//
									if( DoFixErrors() )
									{
										//  Determine if the user wants to save this
										//  invalid directory entry as a file.  If
										//  the user wants to save this invalid
										//  directory entry as a file ...
										//
										if( getyesno(szConvertDirToFileMsg,
												DEFAULT_YES) )
										{
											//  Get the directory entry at the
											//  specified sector and index
											//
											lpDir = GetDirEntryAtSector( uEntryIndex,
													ulSectorNumber );


											//  Change this directory entry attribute
											//  to that of a file
											//
											lpDir->attr = A_NORMAL;


											//  Update the size of the file
											//
											lpDir->size = ulClusterCount << ucBPCShift;


											//  Record the fact that this sector has
											//  been modified.
											//
											MarkBufferDirty( &ulSectorNumber );
										}
									}
								}


								//  Get the directory entry at the specified sector
								//  and index
								//
								lpDir = GetDirEntryAtSector( uEntryIndex,
										ulSectorNumber );
							}


							//  Set the working path back to the current directory
							//
		 					RemoveName( szWorkingPath );
						}
					}
				}
				else
				{
					//
					//
					if( fInLFNSequence )
					{
						//  Determine if we are fixing errors.  If we
						//  are not fixing errors, display a message
						//  stating that errors were found but will
						//  not be fixed.
						//
						ShowFoundErrorsMsg();


						//  Display a message stating that an orphaned
						//  LFN entry was found and deleted.
						//
						printf( szOrphanedLFNEntry );


						//  Set a flag indicating to delete the LFN
						//  sequence.
						//
						fInLFNSequence = FALSE;
						fDeleteLFNSequence = TRUE;


						//  Set the sequence number to start deleting
						//  the LFN sequence
						//
						ucLFNSequenceNumber = ucLFNStartSequenceNumber;


						//
						//
						uEntryIndex = uLFNEntryIndex;
						ulSectorNumber = ulLFNSectorNumber;
						continue;
					}


					//
					//
					fDeleteLFNSequence = FALSE;
				}


				//  Increment the directory entry index
				//
				uEntryIndex++;
			}


			//  Check if we are done processing this directory
			//
			if( fDoneProcessingDir )
			{
				//  We are done processing this directory
				//
				break;
			}


			//
			//
			uCountSPC++;


			//
			//
			ulSectorNumber++;
		}


		//  Check if we are done processing this directory
		//
		if( fDoneProcessingDir )
		{
			//  We are done processing this directory
			//
			break;
		}


		//  Get the next cluster number in the chain
		//
		GetFatEntry( lpDPB, ulCurDirCluster, &ulNextCluster );
		if( ulNextCluster >= END_OF_FAT_CHECK )
		{
			//  We are done processing this directory
			//
			break;
		}
		ulCurDirCluster = ulNextCluster;
	}

	return ucReturnCode;
}


//  Copy the contents of one cluster to another cluster.  This does
//  not update the FAT.
//
void CopyCluster( ulong ulSourceCluster, ulong ulDestinationCluster )
{
	LPDPB lpDPB;
	LPBUF_MGT lpMRUConvBuffer;
	static unsigned uSPC;
	static ulong ulDataStart;
	static ulong ulSourceSector;
	static ulong ulDestinationSector;


	//  Get the DPB for the drive
	//
	GetDPB( &lpDPB );
	ASSERT_LPDPB( lpDPB );


	//  Determine the number of sectors in a cluster
	//
	uSPC = lpDPB->ms_spc + 1;


	//  Determine the start of the data area on the disk.
	//
	if( IsFat32Drive(lpDPB) )
	{
		ulDataStart = lpDPB->ulFirstDataSector;
	}
	else
	{
		ulDataStart = lpDPB->data_start;
	}


	//  Determine the starting sector number for the source cluster
	//
	ulSourceSector = ulDataStart +
			((ulSourceCluster - 2) << lpDPB->shift);


	//  Determine the starting sector number for the destination cluster
	//
	ulDestinationSector = ulDataStart +
			((ulDestinationCluster - 2) << lpDPB->shift);


	//  Copy each sector in the source cluster to the destination cluster
	//
	while( uSPC )
	{
		//  Read this sector into memory.  This will also set this buffer
		//  to the MRU conventional memory buffer.
		//
		GetSectorAddress( ulSourceSector );


		//  Change the sector number for this sector to the destination
		//  sector number
		//
		lpMRUConvBuffer = GetMRUConvMemBuffer();
		lpMRUConvBuffer->ulSectorNumber = ulDestinationSector;


		//  Mark the MRU conventional memory buffer as modified.
		//
		MarkBufferDirty( &ulDestinationSector );


		//  Increament the sector numbers for both the source and
		//  destination clusters
		//
		ulSourceSector++;
		ulDestinationSector++;


		//  Decrement the count of sectors remaining
		//
		uSPC--;
	}
}


//  Get the address of the directory for this entry index and sector
//
LPDIR GetDirEntryAtSector( unsigned uEntryIndex, ulong ulSectorNumber )
{
	char far * lpBuffer;


	//  Get the address of this sector in conventional memory
	//
	lpBuffer = GetSectorAddress( ulSectorNumber );


	//  Return the address of the request directory entry
	//
	return (LPDIR)(lpBuffer + ( uEntryIndex * sizeof(DIR) ));
}


//  Return the address of the string which contains the working path
//
char * GetWorkingPath( void )
{
	static char szWorkingPath[LFN_MAX_PATH];
	return szWorkingPath;
}


//  Process all the command line arguments
//
void DoParseCommandLine( int argc, char *argv[] )
{
	uchar ucArgIndex;
	uchar ucDriveLetter;
	int	iDrive;
	static ARGDEF CHKDSKOpts[] =
	{
		{ BOOL_SWITCH, &fSwitches[DONT_CONFIRM],     'c', 0 },
		{ BOOL_SWITCH, &fSwitches[FIX_ERRORS],       'f', 0 },
		{ BOOL_SWITCH, &fSwitches[SHOW_VERBOSE],     'v', 0 },
		{ BOOL_SWITCH, &fSwitches[HIDE_SIGNON],      '[', 0 },
		{ 0 }
	};


	//  Remove the quotes from any LFN arguments
	//
	for(ucArgIndex=0; ucArgIndex < argc; ucArgIndex++)
	{
		LFNStripArgQuotes( argv[ucArgIndex] );
		if( strlen(argv[ucArgIndex]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(255);
		}
	}


	//  This is for the command line parsing library function.  It
	//  indicates that no command line options are ok
	//
	SetParseFlags( PARSEF_NOARGSOK );


	//  Process all the boolean switches in the command line
	//
	argc = ParseCmdline(argc, &argv[0], CHKDSKOpts);


	//  If there is anything, it should only be a drive or
	//  path/name combination.
	//
	if(argc > 1)
		Usage();


	//  Determine which drive we are checking or which drive the file
	//  we are checking is on
	//
	if( argc )
	{
		//  Save the address of the path in the command line arguments
		//
		szCommandLinePath = argv[0];


		if(argv[0][1] == ':')
		{
			ucDriveLetter = toupper(argv[0][0]);
			if( (ucDriveLetter < 'A') || (ucDriveLetter > 'Z') )
			{
				Usage();
			}

			iDrive = ucDriveLetter - 'A';
		}
		else
		{
			dlGetCurrentDrive( &iDrive );
		}
	}
	else
	{
		dlGetCurrentDrive( &iDrive );
	}
	SetWorkingDriveNumber( (char)iDrive );
}


//  Display the usage screen and exit.  Called by the command line
//  parsing routines if an error is encountered
//
void Usage( void )
{
	DisplaySignOnMessage();
	printf( szUsage );

	//  MSDOS 7.1 chkdsk does not set error codes for invalid
	//  command line options.
	//
	exit( ERR_NONE );
}


//  Display information about the disk like the disk label
//  and serial number.
//
void DisplayVolumeInfo( void )
{
	static DISK_INFO DiskInfo;
	FIND FIND;
	bool fSuccessful;
	char * szSearchMask = "?:*.*";
	uchar ucDrive;
	pPACKEDDATE pPackedDate;
	pPACKEDTIME pPackedTime;
	unsigned uHour;
	char cMidnightChar;
	unsigned uIndex  = 0;
	char szENAME[12];
	unsigned uENAMEIndex;


	//  Get the drive number we are checking
	//
	ucDrive = GetWorkingDriveDriveNumber();


	//  Display the volume label if there is one
	//
	szSearchMask[0] = ucDrive + 'A';
	dlSetDTA( &FIND );
	fSuccessful = dlFindFirstFile( A_VOLUME, szSearchMask );
	if( fSuccessful )
	{
		for( uIndex=0,uENAMEIndex=0; uIndex < 8; uIndex++ )
		{
			if( FIND.name[uIndex] == '.' )
			{
				break;
			}
			if( FIND.name[uIndex] == 0 )
			{
				break;
			}
			szENAME[uENAMEIndex] = FIND.name[uIndex];
			uENAMEIndex++;
		}
		if( FIND.name[uIndex] == '.' )
		{
			uIndex++;
		}

		while( uENAMEIndex < 8 )
		{
			szENAME[uENAMEIndex] = ' ';
			uENAMEIndex++;
		}
 		szENAME[uENAMEIndex] = 0;

		for( ; uIndex < 12; uIndex++ )
		{
			if( FIND.name[uIndex] == 0 )
			{
				break;
			}
			szENAME[uENAMEIndex] = FIND.name[uIndex];
			uENAMEIndex++;
		}

		while( uENAMEIndex < 11 )
		{
			szENAME[uENAMEIndex] = ' ';
			uENAMEIndex++;
		}
 		szENAME[uENAMEIndex] = 0;

		pPackedDate = (pPACKEDDATE)(&FIND.date);
		pPackedTime = (pPACKEDTIME)(&FIND.time);
		uHour = pPackedTime->hour;
		if( uHour < 12 )
		{
			cMidnightChar = 'a';
		}
		else
		{
			uHour -= 12;
			cMidnightChar = 'p';
		}

		if( uHour == 0 )
		{
			uHour = 12;
		}

		printf(vol_str1, szENAME, "",
				pPackedDate->month,
				pPackedDate->day,
				pPackedDate->year + 1980,
				uHour,
				pPackedTime->minute,
				cMidnightChar );
		printf("\n");
	}


	//  Get the serial number
	//
	if( fIsFat32World )
	{
		dlGenericIOCTLForBlockDevices( ucDrive+1,
				0x48, 0x66,	&DiskInfo );
	}
	else
	{
		dlGenericIOCTLForBlockDevices( ucDrive+1,
				0x08, 0x66,	&DiskInfo );
	}



	if( fSuccessful || fIsFat32World )
	{
		//  Display the volume serial number
		//
		printf( vol_str2,
				(unsigned)((DiskInfo.ulSerialNumber & 0xFFFF0000L) >> 16),
				(unsigned)((DiskInfo.ulSerialNumber & 0x0000FFFFL)) );
	}
}


//  Display the chkdsk drive usage summary
//
void DisplaySummary( void )
{
	GDF32 DiskSpace;
	extern unsigned _psp;
	uchar ucIndex;
	uchar ucDrive;
	LPDPB lpDPB;
	unsigned uBPS;
	uchar ucBPSShift;
	uchar ucBPCShift;
	uchar ucTemp;


	//  Get the drive number we are checking
	//
	ucDrive = GetWorkingDriveDriveNumber();


	//  Get the drive specs and free space
	//
	SmartGetFreeSpace( ucDrive+1, &DiskSpace );
	GetDPB( &lpDPB );
	ASSERT_LPDPB( lpDPB );


	//  Determine the shift factor for the bytes per sector
	//
	ucBPSShift = 0;
	uBPS = lpDPB->bytes_per_sec;
	while( uBPS )
	{
		uBPS >>= 1;
		ucBPSShift++;
	}
	ucBPSShift--;


	//  Determine the shift factor for bytes per cluster
	//
	ucBPCShift = ucBPSShift + lpDPB->shift;


	//  Determine the total disk space
	//
	if( CalculationOverflowed( DiskSpace.ulTotalClusters, ucBPCShift,
			&stat[DTOTAL].bytes) )
	{
		stat[DTOTAL].fInKBytes = TRUE;
		if( ucBPCShift >= 10 )
		{
			ucTemp = ucBPCShift - 10;
			stat[DTOTAL].bytes = DiskSpace.ulTotalClusters << ucTemp;
		}
		else
		{
			ucTemp = 10 - ucBPCShift;
			stat[DTOTAL].bytes = DiskSpace.ulTotalClusters >> ucTemp;
		}
	}
	stat[DTOTAL].number++;


	//  Determine the number of hidden files
	//
	if( stat[DHIDDEN].bytes )
	{
//		stat[DHIDDEN].bytes <<= ucBPCShift;
		if( CalculationOverflowed( stat[DHIDDEN].bytes, ucBPCShift,
					&stat[DHIDDEN].bytes) )
		{
			stat[DHIDDEN].fInKBytes = TRUE;
		}
	}
	else
	{
		stat[DHIDDEN].number = 0;
	}


	//  Determine the number of bytes used by directories
	//
	if( CalculationOverflowed( stat[DDIR].bytes, ucBPCShift,
				&stat[DDIR].bytes) )
	{
		stat[DDIR].fInKBytes = TRUE;
	}


	//  Determine the number of bytes used by file
	//
	if( stat[DUSER].bytes )
	{
		if( CalculationOverflowed( stat[DUSER].bytes, ucBPCShift,
					&stat[DUSER].bytes) )
		{
			stat[DUSER].fInKBytes = TRUE;
		}
	}
	else
	{
		stat[DUSER].number = 0;
	}


	//  Determine the number of bytes which were recovered in lost clusters
	//
	if( CalculationOverflowed( stat[DLOST].bytes, ucBPCShift,
				&stat[DLOST].bytes) )
	{
		stat[DLOST].fInKBytes = TRUE;
	}


	//  Determine the number of bytes which are bad
	//
	if( CalculationOverflowed( stat[DBAD].bytes, ucBPCShift,
				&stat[DBAD].bytes) )
	{
		stat[DBAD].fInKBytes = TRUE;
	}


	//  Determine the amount of available bytes on disk
	//
	if( CalculationOverflowed( DiskSpace.ulAvailableClusters, ucBPCShift,
			&stat[DAVAIL].bytes) )
	{
		if( ucBPCShift >= 10 )
		{
			ucTemp = ucBPCShift - 10;
			stat[DAVAIL].bytes = DiskSpace.ulAvailableClusters << ucTemp;
		}
		else
		{
			ucTemp = 10 - ucBPCShift;
			stat[DAVAIL].bytes = DiskSpace.ulAvailableClusters >> ucTemp;
		}
		stat[DAVAIL].fInKBytes = TRUE;
	}
	stat[DAVAIL].number++;


	//  Determine the amount of bytes in each allocation unit
	//
	stat[DALLOCSIZE].bytes = 1L << ucBPCShift;
	stat[DALLOCSIZE].number++;


	//  Determine the total amount of allocation units
	//
	stat[DALLOCTTOTAL].bytes = DiskSpace.ulTotalAllocationUnits;
	stat[DALLOCTTOTAL].number++;


	//  Determine the total amount of availble allocation units
	//
	stat[DALLOCAVAIL].bytes = DiskSpace.ulAvailableAllocationUnits;
	stat[DALLOCAVAIL].number++;


	//  Determine the total amount of memory
	//
	stat[MTOTAL].bytes= psp_memory();
	stat[MTOTAL].number++;


	//  Determine the amount of available memroy
	//
	stat[MAVAIL].bytes = stat[MTOTAL].bytes - DetermineChkdskMemory();
	stat[MAVAIL].number++;


	//  Check if the user wants to abort CTRL-C
	//
	CheckUserAbort();


	//  Display all the calculations
	//
	for(ucIndex=0;ucIndex<MAXSTAT;++ucIndex)
	{
		if(stat[ucIndex].number)
		{
			printf("%,13lu ",stat[ucIndex].bytes);
			if( stat[ucIndex].fInKBytes )
			{
				printf( szKiloStr );
			}
			printf(stat_str[ucIndex],stat[ucIndex].number);
#if 0
			if( (ucIndex==DTOTAL) && (fTotalSpaceOverFlow) )
			{
				printf( szScaledTotalSpace );
			}
			else if( (ucIndex==DAVAIL) && (fFreeSpaceOverFlow) )
			{
				printf( szScaledFreeSpace );
			}
			else
			{
				printf(stat_str[ucIndex],stat[ucIndex].number);
			}
#endif
			printf("\n");
		}
		if((ucIndex==DAVAIL) || (ucIndex==DALLOCAVAIL) )
		{
			printf("\n");
		}
	}
}


//  Determine the amount of memory used by chkdsk
//
ulong DetermineChkdskMemory( void )
{
	extern unsigned _psp;
	struct PSP far * lpPSP;

	lpPSP = MK_FP( _psp, 0 );
	if( lpPSP->env_seg < _psp )
	{
		return (lpPSP->env_seg) * 0x10L;
	}
	else
	{
		return _psp * 0x10L;
	}
}


//  Determine if this calculation will overflow an unsigned long
//
bool CalculationOverflowed( ulong ulValue, uchar ucShiftFactor,
		ulong * pulResult )
{
	while( ucShiftFactor )
	{
		if(ulValue & 0xC0000000L)
		{
			return TRUE;
		}

		ucShiftFactor--;
		ulValue <<= 1;
	}

	*pulResult = ulValue;

	return FALSE;
}


//  Calculate free memory
//
ulong psp_memory(void)
{
	extern unsigned _psp;
	unsigned far * memtop = MK_FP(_psp,2);
//	return *memtop / 0x40;
	return *memtop * 0x10L;
}


//  Determine if this is a remote drive.
//
bool IsARemoteDrive( void )
{
	bool fIsRemoteDrive = FALSE;
	uchar ucDrive;


	ucDrive = GetWorkingDriveDriveNumber();
	if( IsAValidDrive() )
	{
		dlIsDriveRemote( ucDrive+1, &fIsRemoteDrive );
	}

	return fIsRemoteDrive;
}


//  Determine if this is a subst drive.
//
bool IsASubstDrive( void )
{
	bool fIsSubstDrive = FALSE;
	uchar ucDrive;


	ucDrive = GetWorkingDriveDriveNumber();
	if( IsAValidDrive() )
	{
		dlIsDriveSubst( ucDrive+1, &fIsSubstDrive );
	}

	return fIsSubstDrive;
}


//  Determine if this is a subst drive.
//
bool IsAValidDrive( void )
{
	LPCDS lpCDS;
	uchar ucDrive;


	ucDrive = GetWorkingDriveDriveNumber();
	if( dlGetCDSForDrive( ucDrive, &lpCDS ) )
	{
		return _IsValidDrive(lpCDS);
	}

	return FALSE;
}


//  Lock the logical disk
//
void LockDrive( void )
{
	uchar ucDrive;
	bool fSuccessful;

	ucDrive = GetWorkingDriveDriveNumber();
	fSuccessful = dlLockLogicalVolume( ucDrive+1, 0, 0 );
	if( !fSuccessful )
	{
	}
}


//  Unlock the logical disk
//
void UnlockDrive( void )
{
	uchar ucDrive;

	ucDrive = GetWorkingDriveDriveNumber();
	dlUnlockLogicalVolume( ucDrive+1 );
	dlUnlockLogicalVolume( ucDrive+1 );
}


//  Save the drive number which we will be checking
//
void SetWorkingDriveNumber( uchar ucDOSDriveNumber )
{
	uchar * pucWorkingDrive;

	pucWorkingDrive = GetAddressWorkingDriveVariable();
	*pucWorkingDrive = ucDOSDriveNumber;
}


//  Restore the users current drive
//
uchar GetWorkingDriveDriveNumber( void )
{
	uchar * pucWorkingDrive;

	pucWorkingDrive = GetAddressWorkingDriveVariable();
	return *pucWorkingDrive;
}


//  Get the address of the variable which contains the working drive
//
uchar * GetAddressWorkingDriveVariable( void )
{
	static uchar ucWorkingDrive;
	return &ucWorkingDrive;
}


//  Save the current directory
//
void SaveCurrentDirectory( void )
{
	char * szCurrentDirectory;

	szCurrentDirectory = GetAddressCurrentDirectoryVariable();
	dlGetCurrentDirectory( 0, szCurrentDirectory );
}


//  Restore the users current working directory
//
void RestoreCurrentWorkingDirectory( void )
{
	char * szCurrentDirectory;

	szCurrentDirectory = GetAddressCurrentDirectoryVariable();
	dlSetCurrentDirectory( szCurrentDirectory );
}


//  Get the address of the variable which contains the current drive
//
char * GetAddressCurrentDirectoryVariable( void )
{
	static char szCurrentDirectory[MAXNAMELEN];
	return szCurrentDirectory;
}


//  Save the current drive
//
void SaveCurrentDrive( void )
{
	uchar * pucCurrentDrive;
	int iDrive;

	pucCurrentDrive = GetAddressCurrentDriveVariable();
	dlGetCurrentDrive( &iDrive );
	*pucCurrentDrive = (uchar)iDrive;
}


//  Restore the users current drive
//
void RestoreCurrentDrive( void )
{
	uchar * pucCurrentDrive;

	pucCurrentDrive = GetAddressCurrentDriveVariable();
	dlSetCurrentDrive( (int)*pucCurrentDrive );
}


//  Get the address of the variable which contains the current drive
//
uchar * GetAddressCurrentDriveVariable( void )
{
	static uchar ucCurrentDrive;
	return &ucCurrentDrive;
}


//  Display the program name, version, and copyright information
//
void DisplaySignOnMessage( void )
{
	if( !DoHideSignon() )
	{
		if( fIsFat32World )
		{
			printf(szSignon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(szSignon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}


//  Show the usage for a bad command line parameter
//
void OptError(char * szMsg)
{
	unsigned uIndex = 0;

	while( (szMsg[uIndex] != ' ') && (szMsg[uIndex] != 0) )
	{
		if( szMsg[uIndex] == '/' )
		{
			uIndex++;
			if( szMsg[uIndex] != 0 )
			{
				uIndex++;
			}
			break;
		}
		uIndex++;
	}
	szMsg[uIndex] = 0;


	printf("Invalid switch - %s\n", szMsg );
	exit(255);
}


//  Determine if we are fixing errors.  If we are not fixing
//  errors, display a message stating that errors were found
//  but will not be fixed.
//
void ShowFoundErrorsMsg( void )
{
	static bool fHaveShownMsg = FALSE;


	if( !DoFixErrors() )
	{
		if( !fHaveShownMsg )
		{
			printf(chk_nofix);
			fHaveShownMsg = TRUE;
		}
	}
}


//  Display a message stating that there was not
//  enough free space to fix the problem and
//  instruct the user to free some space and
//  run chkdsk again.
//
void DisplayInsufficientSpaceErrorMsg( void )
{
	static bool fHaveShownMsg = FALSE;


	if( !fHaveShownMsg )
	{
		printf(szInsufficientSpace);
		fHaveShownMsg = TRUE;
	}
}


//  Display a message recomending the use of scandisk
//
void ShowScandiskMsg( void )
{
#if SHOW_SCANDISK_MSG
	if( !DoFixErrors() )
	{
		printf(szUseScandiskMsg);
	}
#endif
}


//  Return the number of copies of the FAT.  This is set to something
//  other than what is in the DPB if the FAT copies are not identicle
//  and the user does not want to fix it.
//
unsigned GetFATCopies( void )
{
	unsigned * puFATCopies;

	puFATCopies = GetFATCopiesVar();
	return *puFATCopies;
}


//  Set the number of copies of the FAT.  This is set to something
//  other than what is in the DPB if the FAT copies are not identicle
//  and the user does not want to fix it.
//
void SetFATCopies( unsigned uFATCopies )
{
	unsigned * puFATCopies;

	puFATCopies = GetFATCopiesVar();
	*puFATCopies = uFATCopies;
}


//  Return the address of the variable which contains the number of FATs
//
unsigned * GetFATCopiesVar( void )
{
	static unsigned uFATCopies = 2;
	return &uFATCopies;
}


//  Return the address of the variable that contains the SFC flag
//
bool * GetSFCFlagPointer( void )
{
	static bool fSFCFlag = FALSE;
	return &fSFCFlag;
}


//  Return the SFC flag
//
bool GetSFCFlag( void )
{
	bool * pfSFCFlag;


	pfSFCFlag = GetSFCFlagPointer();
	return *pfSFCFlag;
}


//  Set the SFC flag
//
void SetSFCFlag( void )
{
	bool * pfSFCFlag;


	pfSFCFlag = GetSFCFlagPointer();
	*pfSFCFlag = TRUE;
}


//  Set the SFC flag
//
void ClearSFCFlag( void )
{
	bool * pfSFCFlag;


	pfSFCFlag = GetSFCFlagPointer();
	*pfSFCFlag = FALSE;
}

