//
//   $Id: CHOICE.C 1.9 2003/03/20 15:05:38Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
	CHOICE [/C[:]keys] [/N] [/S] [/T[:]c,nn] [Prompt Text]

   Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary
*/

#include "dosver.h"
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "fatworld.h"
#include "stdc.h"

#include "choice.pt"


#define  NUL   '\0'


char syntax[]=
"\nWaits for the user to choose one of a set of choices.\n"
"\n"
"CHOICE [/C[:]choices] [/N] [/S] [/T[:]c,nn] [text]\n"
"\n"
"/C[:]choices Specifies allowable keys. Default is YN\n"
"/N           Do not display choices and ? at end of prompt string.\n"
"/S           Treat choice keys as case sensitive.\n"
"/T[:]c,nn    Default choice to c after nn seconds\n"
"text         Prompt string to display\n"
"\n"
"ERRORLEVEL is set to offset of key user presses in choices.\n"
"\n";




void putc(char c)
{
	if( c == '\n' )
	{
		putc('\r');
	}
	asm {
		mov	dl, c
		mov	ah, 2
		int 	21h
	}
}

void puts(char *s)
{
	while(*s)
	{
		putc(*s++);
	}
}


long CentiSeconds(void)
{
	long t;

	asm {
		push	bx
		mov 	ah, 2Ch
		int	21h
		mov	al, 60				// minutes * 60 = seconds in full minutes
		mul	cl
		xor	bx, bx
		xchg	bl, dl				// bx = centiseconds as a word (dl == 0)
		xchg	dl, dh				// dx = seconds into current minute
		add	ax, dx				// ax = full seconds into this hour
		mov	dx, 100				// dx:ax = centiseconds in full seconds
		mul	dx
		add	ax, bx
		adc	dx, 0	   			// dx:ax = centiseconds into current hour.
		mov	word ptr [t],   ax
		mov	word ptr [t][2], dx
		pop	bx
	}
	return t;
}

long ElapsedSince(long then)
{
	long now, since;

	now = CentiSeconds();

	// if the minutes "wrapped" then add the number of centiseconds
	// in a minute to the current time
	//
	if(now < then)
//		now += 6000;
      now += 360000;
	since = now - then;
	return since;				// return difference in centiseconds
}


int main( void )
{
	int i;
	int c;
	int numChoices=2;
	int usersChoice;
	char * szPSPCommandLine;
	char * opt;
	char * promptStarts;
	int  timeSpec = 0;
	long waitingTime = 0;
	unsigned uIndex;
	long startTime;
	long elapsedTime;
	int  showingChoices = TRUE;
	int  caseSensitive = FALSE;
	char defChar = FALSE;
	char * choices = "YN";
	bool fIsFat32World;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	szPSPCommandLine = (char *)(0x80);
	opt = &szPSPCommandLine[1];
	opt[*szPSPCommandLine] = 0;

	promptStarts = opt;
	while((*opt != NUL) && (*opt != '/'))
		++opt;

	while(*opt == '/')
	{
		++opt;
	  	switch(toupper(*opt++))
		{
		case 'C':                  		// allowable Chars
			numChoices = 0;					// changing default list
	  		if(*opt == ':')  					// optional ':'
				++opt;
			choices = opt;                // point to start of list
			while(*opt != NUL)				// count chars in list
			{
				if((*opt <= ' ') || (*opt == '/'))
					break;
				++opt;
				++numChoices;
			}
			break;

		case 'N':
			showingChoices = FALSE;			// display No choices
			break;

		case 'S':
			caseSensitive = TRUE;			// input is case Sensitive
			break;

		case 'T':								// max Time to wait for keys
			// get current time
			startTime = CentiSeconds();
	  		if(*opt == ':')					// optional ':'
				++opt;
			defChar = *opt++;	 				// default char

			// RD specific enhancement: allow default char to be specified
			// without a time delay spec causing CHOICE to wait forever and
			// return the default char when ENTER is pressed.
			//
	  		if(*opt == ',')					// skip ','
				++opt;
			else
				break;
			while(*opt != NUL)				// get delay in seconds
			{
				if((*opt < '0') || (*opt > '9'))
					break;
				timeSpec = (timeSpec * 10) + (*opt++ - '0');
			}
			if(timeSpec > 99)
				timeSpec = 99;					// limit spec to 99 seconds

			// RD specific enhancement: convert seconds to centiseconds
			// for increased timing resolution (waitingTime = timeSpec * 100).
			//
			asm {
				mov	ax, timeSpec
				mov	dx, 100
				mul	dx
				mov	word ptr [waitingTime], ax
				mov	word ptr [waitingTime+2], dx
			}
			break;

		case '?':       						// help
			puts("CHOICE ");
			if( fIsFat32World )
			{
				puts(SHORTVERSION);
			}
			else
			{
				puts(SHORTVERSION622);
			}
			puts(" " BUILDVERSION "\n" COPYRIGHT "\n" );
			puts(syntax);
			return(255);

		default:
			puts("\a\nInvalid Switch.\n");	// bad input
			puts(syntax);
			return(255);
		}

		// If not at next arg inc opt: We could be at the end of
		// the choices string and would otherwise make prompt NULL
		// when we NUL terminate the choices string.
		//
		if((*opt != NUL) && (*opt != '/'))
			++opt;
		promptStarts = opt;							// prompt is always last option
		while((*opt != NUL) && (*opt != '/'))	// find next option spec
			++opt;
	}
	choices[numChoices] = NUL;				// NUL terminate choices string
	if(!caseSensitive)
	{
		// do upper case comparisons
		uIndex = 0;
		while( choices[uIndex] )
		{
			choices[uIndex] = toupper(choices[uIndex]);
			uIndex++;
		}
		defChar = toupper(defChar);
	}
	puts(promptStarts);						// display prompt, if any

	if(showingChoices)						// if no /N argument
	{
		putc('[');								// list choices, (ie [1,(2),3]?)
		i = 0;
		while( choices[i] )
		{
			c = choices[i++];
			if(defChar && (c == defChar))	// RD: enclose default choice in parens
			{
				putc('(');
				putc(c);
				putc(')');
			}
			else
				putc(c);

			if( choices[i] )
			{
				putc(',');
			}
		}
		puts("]? ");
	}
	do
	{
		if(waitingTime)							// see if delay expired
		{
			elapsedTime = ElapsedSince(startTime);
			if(elapsedTime >= waitingTime)	// return default choice if so
			{
				c = defChar;
				usersChoice = Lookup(c,choices,caseSensitive);
				break;
			}
		}

		if( (c = GetKey()) == 0 )
		{
			continue;
		}

		// RD specific enhancement: if default char specified and user
		// hits ENTER, supply default responce.
		//
		if(defChar && (c == '\r'))
		{
			c = defChar;
		}

		usersChoice = Lookup(c,choices,caseSensitive);
		if(usersChoice > 0)					// check if c is in set of OK chars
			break;

	} while(TRUE);

	if(!caseSensitive)
		c = toupper(c);
	putc(c);										// display choice
	putc('\n');
	return usersChoice; 						// return ordinal of choice in list
}

// Returns ordinal of c if c is in {s} else 0.
// If caseSensitive != 0 then {s} must be all upper case.
//
int Lookup(char c, char *s, int caseSensitive)
{ int i=0;

	if(!caseSensitive)
		c = toupper(c);

	while(s[i])								// for every c in s
		if(c == s[i++])  					// if we have a match
			return i;						// return its ordinal

	putc('\a');								// else ring the BEL
	return FALSE;							// and return FALSE
}

// returns next char input from STDIN, no echo
//
int GetKey(void)
{
	char c;

	//  if it's an extended scan code, skip scan code, get char code
	//
	asm	mov	ah, 0Bh						// if STDIN pending
	asm	int	21h
	asm	and	ax, 1
	asm	jz		NoKey
	asm	mov	ah, 8							// read a key
	asm	int	21h
NoKey:
	asm	mov	c, al

	return c;								// return char code
}
