//
//   $Id: COMMAPI.C 1.8 2002/10/16 21:42:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  COMMAPI.C
//
//  This file contains all the API functions that a comm transport
//  layer interface is expected to perform.
//
//  History:
//
//  04-01-2001  bt  Created
//


//  Borland libraries
//
#include <mem.h>
#include <stdio.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dl8250.h"
#include "crc32.h"
#include "dlapp.h"


//  Local definitions
//
#include "commapi.h"
#include "commint.h"


//  Prototypes
//
#include "commapi.pt"

void DisplaySignonMessage( void );


//  Globals
//
unsigned uSecondsToTimeout = 10;


//  ========================================================================
//
//  The following functions are required APIs
//
//  ========================================================================




//  TXRXCommInit
//
//
//  Initialize the COMM interface
//
//
//  On entry:
//
//    argc = The number of options in the argv array
//    argv = Array of pointers to command line options
//
//  On exit:
//
//  on success,
//
//  	function returns TRUE
//
//  on error,
//
//  	function returns FALSE
//
//
bool TXRXCommInit( int argc, char *argv[] )
{
	static uchar RXBuffer[200];
	bool fSuccessful;
	unsigned uIndex;
	static ulong ulBaudRate = 115200L;
	static unsigned uComPortNumber = 1;
	static unsigned uIRQNumber = 4;
	static unsigned uUserIRQNumber = 0;
	static ARGDEF CommOpts[] =
	{
		{ INT_SWITCH,  &uComPortNumber,    'C', 0 },
		{ LONG_SWITCH, &ulBaudRate,        'B', 0 },
		{ INT_SWITCH,  &uUserIRQNumber,    'I', 0 },
		{ INT_SWITCH,  &uSecondsToTimeout, 'W', 0 },
		{ 0 }
	};
	static BAUDXREF BaudXREF[] = {
		{ 115200L,   115 },
		{  57600L, (unsigned)(57600L) },
		{  38400L, (unsigned)(38400L) },
		{  19200L, 19200 },
		{   9600L,  9600 },
		{   4800L,  4800 },
		{   2400L,  2400 },
		{   1200L,  1200 },
		{    300L,   300 },
	};


	SetParseFlags( PARSEF_NOARGSOK );
	argc = ParseCmdline(argc, &argv[0], CommOpts);
	if( argc )
	{
		OptError( argv[0] );
	}


	if( (uComPortNumber < 1) || (uComPortNumber > 4) )
	{
		printf("Invalid COM port number\n");
		return FALSE;
	}


	if( uUserIRQNumber )
	{
		uIRQNumber = uUserIRQNumber;
		if( (uIRQNumber < 1) || (uIRQNumber > 15) )
		{
			printf("Invalid IRQ number\n");
			return FALSE;
		}
	}
	else
	{
		//  Determine if the COM port is COM1/COM3 or COM2/COM4
		//
		if( uComPortNumber & 0x01 )
		{
			uIRQNumber = 4;
		}
		else
		{
			uIRQNumber = 3;
		}
	}


	if( uSecondsToTimeout > 60 )
	{
		printf("Invalid network timeout value.  Max timeout is 60 seconds\n");
		return FALSE;
	}


	for( uIndex = 0; uIndex < ARRAY_ENTRIES(BaudXREF); uIndex++ )
	{
		if( BaudXREF[uIndex].ulBaudRate == ulBaudRate )
		{
			break;
		}
	}
	if( BaudXREF[uIndex].ulBaudRate != ulBaudRate )
	{
		printf("Invalid baud rate\n");
		return FALSE;
	}


	//  Determine if we should display the signon message and if
	//  so display it
	//
	DisplaySignonMessage();


	printf("Initializing ");
	printf("COM%u  ", uComPortNumber );
	printf("IRQ:%u  ", uIRQNumber );
	printf("Baud:%lu  ", ulBaudRate );
	printf("Network timeout set to %u seconds\n", uSecondsToTimeout );


	fSuccessful = comm_open( BaudXREF[uIndex].uCommOpenValue, DEFPARMS,
			uComPortNumber-1, uIRQNumber, 200, RXBuffer);
	if( !fSuccessful )
	{
		printf("Unable to initialize the COM port\n");
	}

	return fSuccessful;
}


//  TXRXCommDeInit
//
//
//  Uninitialize the COMM interface
//
//
//  On entry:
//
//    nothing
//
//  On exit:
//
//  on success,
//
//  	function returns TRUE
//
//  on error,
//
//  	function returns FALSE
//
//
bool TXRXCommDeInit( void )
{
	comm_close();
	return TRUE;
}


//  TXCommPacket
//
//
//  Transmit a packet over the a comm port
//
//
//  On entry:
//
//    pPacket     = The address of the packet (data)
//    uPacketSize = The size of the packet to transmit
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns an error code defined in COMMAPI.H
//
//
unsigned TXCommPacket( void * pPacket, unsigned uPacketSize )
{
	PTXRXPKT pTXRXPacket;
	uchar * pucTXRXPacket;
	uchar * pPacketData;
	unsigned uTotalTXRXPackets;
	unsigned uPacketNumber;
	ulong ulPacketCRC;
	int iValue;
	unsigned uIndex;


	//  Get the data space where the comm packets will be build
	//  and transmitted
	//
	pTXRXPacket = GetGlobalTansmitPacketAddress();
	(PTXRXPKT)pucTXRXPacket = GetGlobalTansmitPacketAddress();


	//  Get the address of the packet data that will be transmitted
	//
	pPacketData = pPacket;


	//  Determine the number of packets that will need to be sent
	//
	uTotalTXRXPackets = uPacketSize / sizeof(pTXRXPacket->ucPacketData);
	if( uPacketSize % sizeof(pTXRXPacket->ucPacketData) )
	{
		uTotalTXRXPackets++;
	}


	//  Build and transmit the packets
	//
	uPacketNumber = 1;
	while( uPacketNumber <= uTotalTXRXPackets )
	{
		//  Create the packet
		//
		{
			//  Fill in the packet header
			//
			pTXRXPacket->uNumberOfPackets = uTotalTXRXPackets;
			pTXRXPacket->uPacketNumber = uPacketNumber;


			//  Copy in the callers data
			//
			memcpy( pTXRXPacket->ucPacketData, pPacketData,
					sizeof(pTXRXPacket->ucPacketData) );


			//  Perform the CRC on the packet
			//
			ulPacketCRC = 0;
			dlGetCRC32( &ulPacketCRC, (char *)(pTXRXPacket), sizeof(TXRXPKT) -
					sizeof(pTXRXPacket->ulPacketCRC) );
			pTXRXPacket->ulPacketCRC = ulPacketCRC;
		}


		//  Send this packet
		//
		while( TRUE )
		{
			//  Determine if the receiving end is ready
			//
Restart:
			comm_putc( COMM_TXRX_READY );
			asm pushf
			asm sti
			iValue = COMM_TGETC( COMM_TIMEOUT_TICKS );
			asm popf
			if( iValue == ERR )
			{
				return COMM_TIMEOUT;
			}
			if( iValue != COMM_TXRX_ACK )
			{
				goto Restart;
//				return COMM_READY_FAILED;
			}


			//  Transmit the packet
			//
			for( uIndex = 0; uIndex < sizeof(TXRXPKT); uIndex++ )
			{
				comm_putc( (char)(pucTXRXPacket[uIndex]) );
			}


			//  Determine if the packet was received
			//
			asm pushf
			asm sti
			iValue = COMM_TGETC( COMM_TIMEOUT_TICKS * 2 );
			asm popf
			if( iValue == ERR )
			{
				return COMM_TIMEOUT;
			}
			if( iValue == COMM_TXRX_RESEND )
			{
				continue;
			}
			else if( iValue != COMM_TXRX_ACK )
			{
				return COMM_SEND_FAILED;
			}


			//  Packet received successfully.  Now on to the next packet.
			//
			break;
		}


		//  Set the start of the next part of the packet data
		//
		pPacketData += sizeof(pTXRXPacket->ucPacketData);


		//  Increment the packet number
		//
		uPacketNumber++;
	}


	//  Return that the packet was sent
	//
	return COMM_SERVICED;
}


//  RXCommPacket
//
//
//  Receive a packet from the comm port
//
//
//  Note:  This function must be able to receive the maximum
//         possible packet size
//
//
//  On entry:
//
//    pPacket = The address of the packet (data)
//
//  On exit:
//
//  on success,
//
//    *pPacket = Updated with the packet (data)
//
//  	function returns 0
//
//  on error,
//
//		function returns an error code defined in COMMAPI.H
//
//
unsigned RXCommPacket( void * pPacket )
{
	PTXRXPKT pTXRXPacket;
	uchar * pucTXRXPacket;
	uchar * pPacketData;
	ulong ulPacketCRC;
	int iValue;
	unsigned uIndex;


	//  Get the data space where the comm packets will be build
	//  and transmitted
	//
	pTXRXPacket = GetGlobalTansmitPacketAddress();
	(PTXRXPKT)pucTXRXPacket = GetGlobalTansmitPacketAddress();


	//  Get the address of the packet data that will be transmitted
	//
	pPacketData = pPacket;


	//  Get the service packet back from the server
	//
	while( TRUE )
	{
Restart:
		//  Determine if the server is ready
		//
		asm pushf
		asm sti
		iValue = COMM_TGETC( COMM_TIMEOUT_TICKS );
		asm popf
		if( iValue == ERR )
		{
			return COMM_TIMESHARE_TIMEOUT;
		}
		if( iValue != COMM_TXRX_READY )
		{
			return COMM_READY_FAILED;
		}
		comm_putc( COMM_TXRX_ACK );


		//  Get a packet back from the server
		//
		for( uIndex = 0; uIndex < sizeof(TXRXPKT); uIndex++ )
		{
			asm pushf
			asm sti
			iValue = COMM_TGETC( COMM_TIMEOUT_TICKS );
			asm popf
			if( iValue == ERR )
			{
				comm_putc( COMM_TXRX_RESEND );
				goto Restart;
//				return COMM_TIMEOUT;
			}
			pucTXRXPacket[uIndex] = (uchar)iValue;
		}


		//  Determine if the packet is valid
		//
		ulPacketCRC = 0;
		dlGetCRC32( &ulPacketCRC, (char *)(pTXRXPacket), sizeof(TXRXPKT) -
				sizeof(pTXRXPacket->ulPacketCRC) );
		if( ulPacketCRC != pTXRXPacket->ulPacketCRC )
		{
			comm_putc( COMM_TXRX_RESEND );
			continue;
		}
		else
		{
			comm_putc( COMM_TXRX_ACK );
		}


		//  Copy the packet data into the request packet
		//
		memcpy( pPacketData, pTXRXPacket->ucPacketData,
				sizeof(pTXRXPacket->ucPacketData) );
		pPacketData += sizeof(pTXRXPacket->ucPacketData);


		//  Determine if this was the last packet
		//
		if( pTXRXPacket->uNumberOfPackets == pTXRXPacket->uPacketNumber )
		{
			break;
		}
	}


	//  Return that the packet was sent
	//
	return COMM_SERVICED;
}




//  ========================================================================
//
//  The remainder of this file is specific to the implementation.
//
//  ========================================================================




//  GetGlobalTansmitPacketAddress
//
//
//  Return the address of the data space used to form packets
//
//
//  Note:  This data space must be able to hold the maximum
//         possible packet size
//
//
//  On entry:
//
//    Nothing
//
//  On exit:
//
//  	function returns = The address of the data space for a packet
//
//
PTXRXPKT GetGlobalTansmitPacketAddress( void )
{
	static TXRXPKT TransmitPacket;
	return &TransmitPacket;
}
