//
//   $Id: COMMAPI.H 1.7 2002/10/16 21:42:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  COMMAPI.H
//
//  Copyright (c) 2001-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  This file contains all the API function prototypes and constants
//  that a comm transport layer interface uses
//
//  History:
//
//  04-01-2001  bt  Created
//




//  ========================================================================
//
//  The following functions are required APIs
//
//  ========================================================================




//  TXRXCommInit
//
//
//  Initialize the COMM interface
//
//
//  On entry:
//
//    argc = The number of options in the argv array
//    argv = Array of pointers to command line options
//
//  On exit:
//
//  on success,
//
//  	function returns TRUE
//
//  on error,
//
//  	function returns FALSE
//
//
bool TXRXCommInit( int argc, char *argv[] );




//  TXRXCommDeInit
//
//
//  Uninitialize the COMM interface
//
//
//  On entry:
//
//    nothing
//
//  On exit:
//
//  on success,
//
//  	function returns TRUE
//
//  on error,
//
//  	function returns FALSE
//
//
bool TXRXCommDeInit( void );




//  TXCommPacket
//
//
//  Transmit a packet over the a comm port
//
//
//  On entry:
//
//    pPacket     = The address of the packet (data)
//    uPacketSize = The size of the packet to transmit
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns an error code defined in COMMAPI.H
//
//
unsigned TXCommPacket( void * pPacket, unsigned uPacketSize );




//  RXCommPacket
//
//
//  Receive a packet from the comm port
//
//
//  Note:  This function must be able to receive the maximum
//         possible packet size
//
//
//  On entry:
//
//    pPacket = The address of the packet (data)
//
//  On exit:
//
//  on success,
//
//    *pPacket = Updated with the packet (data)
//
//  	function returns 0
//
//  on error,
//
//		function returns an error code defined in COMMAPI.H
//
//
unsigned RXCommPacket( void * pPacket );




//  ========================================================================
//
//  The remainder of this file is specific to the implementation.
//
//  ========================================================================








//  This is the number of timer ticks (18.2 per second) that
//  the comm APIs will wait before returning a timeout error
//
#define ENABLE_TIMEOUT			TRUE
//#define COMM_TIMEOUT_TICKS		(18.2 * 10)
//#define COMM_TIMESHARE_TICKS	COMM_TIMEOUT_TICKS
#define COMM_TIMESHARE_TICKS		2
#define COMM_TIMEOUT_TICKS			(2)


//  These are the values returned by the API
//
#define COMM_TIMESHARE_TIMEOUT			0x51
#define COMM_TIMEOUT							0x52
#define COMM_SEND_FAILED					0x53
#define COMM_READY_FAILED					0x54
#define COMM_SERVICED						0x55


//  This is the maximum amount of caller data contained in one
//  comm network packet.  Multiple packets are sent to send
//  larger caller data
//
//#define COMM_MAX_CALLER_DATA		0x100
#define COMM_MAX_CALLER_DATA			0x80



//  This is the packets sent between the comm interface
//
typedef struct TransmitPacket TXRXPKT;
typedef TXRXPKT * PTXRXPKT;
typedef TXRXPKT far * LPTXRXPKT;
struct TransmitPacket
{
	unsigned uNumberOfPackets;
	unsigned uPacketNumber;
	uchar ucPacketData[COMM_MAX_CALLER_DATA];
	ulong ulPacketCRC;
};




//  GetGlobalTansmitPacketAddress
//
//
//  Return the address of the data space used to form packets
//
//
//  Note:  This data space must be able to hold the maximum
//         possible packet size
//
//
//  On entry:
//
//    Nothing
//
//  On exit:
//
//  	function returns = The address of the data space for a packet
//
//
PTXRXPKT GetGlobalTansmitPacketAddress( void );


#define ARRAY_ENTRIES(array) (sizeof(array)/sizeof(array[0]))

typedef struct BaudXREF BAUDXREF;
struct BaudXREF
{
	ulong ulBaudRate;
	unsigned uCommOpenValue;
};
