//
//   $Id: IFSDISP.C 1.11 2002/10/16 21:42:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  IFSDISP.C
//
//  Copyright (c) 2001-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  This file contains the int 2fh dispatch functions for the IFS
//
//  History:
//
//  04-01-2001  bt  Created
//


//  Borland libraries
//
#include <mem.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"


//  Local includes
//
#include "ifsdisp.h"
#include "cliapi.h"
#include "client.h"
#include "netapi.h"


//  Prototypes
//
#include "ifsdisp.pt"


void INT2fReturn( void );


//  This variable countains the number of used entries in
//  the CDS map structure
//
unsigned uNumberOfMappedDrives = 0;


//  This variable contains the segment address for all the CDSs
//
unsigned uCDSSegment;


//  This array contains the CDS offset, local drive letter, and
//  remote drive letter, for all the mapped drives
//
CDSMAP DriveMap[26];
unsigned uDriveMapSize = sizeof(DriveMap);


#if 0
char szDebugString[512];
#endif



//  IsOurCDS
//
//
//  Determine if this CDS is mapped to a remote drive and if so,
//  return the local drive letter and remote drive letter for
//  the supplied CDS.
//
//
//  On entry:
//
//    lpCDS                = The address of the CDS in question
//    pucLocalDriveLetter  = The place to store the local drive letter
//    pucRemoteDriveLetter = The place to store the remote drive letter
//
//  On exit:
//
//  on success,
//
//    pucLocalDriveLetter  = Updated with the local drive letter for the CDS
//    pucRemoteDriveLetter = Updated with the remote drive letter for the CDS
//
//    function returns TRUE
//
//  on failure,
//
//    function returns FALSE
//
//
bool IsOurCDS( LPCDS lpCDS, uchar * pucLocalDriveLetter,
		uchar * pucRemoteDriveLetter )
{
	bool fSuccessful = FALSE;
	uchar ucIndex;


	//  Determine if this is the segment for the array of CDS.
	//
	//  NOTE: All CDSs should have the same segment
	//
	if( uCDSSegment == FP_SEG(lpCDS) )
	{
		//  Search through our list of mapped CDSs until we
		//  find a mapped CDS
		//
		for( ucIndex = 0; ucIndex < uNumberOfMappedDrives; ucIndex++ )
		{
			//  Determine if this CDS offset matches the current
			//  entry in the CDS map
			//
			if( DriveMap[ucIndex].uCDSOffset == FP_OFF(lpCDS) )
			{
				//  Save the local drive letter that this CDS references
				//
				*pucLocalDriveLetter =
						DriveMap[ucIndex].ucLocalDriveLetter;


				//  Save the remote drive letter that this CDS references
				//
				*pucRemoteDriveLetter =
						DriveMap[ucIndex].ucRemoteDriveLetter;


				//  Set the return variable to indicate that a mapped drive
				//  was found
				//
				fSuccessful = TRUE;


				//  No need to search any further, because a match has been
				//  found
				//
				break;
			}
		}
	}



	//  Return whether the CDS was mapped by our IFS
	//
	return fSuccessful;
}


//  IsOurDrive
//
//
//  Determine if this drive letter is mapped to a remote drive and
//  if so, return the remote drive letter for the supplied drive.
//
//
//  On entry:
//
//    ucLocalDriveLetter   = The drive letter in question
//    pucRemoteDriveLetter = The place to store the remote drive letter
//    plpCWD               = Address of a variable to contain the CDS CWD
//
//  On exit:
//
//  on success,
//
//    pucRemoteDriveLetter = Updated with the remote drive letter
//                           for the supplied local drive
//    plpCWD               = Updated withe address of the CDS CWD
//
//    function returns TRUE
//
//  on failure,
//
//    function returns FALSE
//
//
bool IsOurDrive( uchar ucLocalDriveLetter, uchar * pucRemoteDriveLetter,
	LPCDS * plpCDS )
{
	bool fSuccessful = FALSE;
	uchar ucIndex;


	//  Search through our list of mapped CDSs until we
	//  find a mapped CDS
	//
	for( ucIndex = 0; ucIndex < uNumberOfMappedDrives; ucIndex++ )
	{
		//  Determine if the local drive letter matches one of
		//  the entries in mapped CDSs
		//

		if( DriveMap[ucIndex].ucLocalDriveLetter == ucLocalDriveLetter )
		{
			//  Save the remote drive letter that this CDS references
			//
			*pucRemoteDriveLetter =
					DriveMap[ucIndex].ucRemoteDriveLetter;


			//  Save the address of the CDS in a return variable
			//
			*plpCDS = (LPCDS)(MK_FP(uCDSSegment ,DriveMap[ucIndex].uCDSOffset));


			//  Set the return variable to indicate that a mapped drive
			//  was found
			//
			fSuccessful = TRUE;


			//  No need to search any further, because a match has been
			//  found
			//
			break;
		}
	}



	//  Return whether the CDS was mapped by our IFS
	//
	return fSuccessful;
}


//  IFSDispatch
//
//
//  This function is called by the int 2fh interrupt handler
//  to dispatch the request to the apropriate function.
//
//
void IFSDispatch( void )
{
	extern CALLREG CallersRegisters;
	unsigned uAX;
	unsigned uDI;
	unsigned uES;
	bool fIsOurCall = FALSE;
	LPCDS lpCDS;
	LPSFT lpSFT;
	LPFIND lpFind;
	uchar far * lpucDTA;
	LPDIR lpDir;
	unsigned uErrorNumber = 0;
	char far * lpszSDAfname1;
	char far * lpszSDAfname2;
	uchar far * lpucSDASearchAttribute;
	unsigned far * lpuSDAExtendedOpenAction;
	unsigned far * lpuSDAExtendedOpenAttr;
	unsigned far * lpuSDAExtendedOpenMode;
	unsigned far * lpuStackWord;
	unsigned uIndex;
	uchar ucLocalDriveLetter;
	uchar ucRemoteDriveLetter;
	bool fChangedSDAfname1Letter = FALSE;
	bool fChangedSDAfname2Letter = FALSE;
	unsigned uRemoteHandle;
	extern unsigned uOnExitPtrSegment;
	extern unsigned uOnExitPtrOffset;
	extern unsigned _psp;
	uchar ucSPC;
	uchar ucMediaID;
	unsigned uTotalClusters;
	unsigned uBPS;
	unsigned uFreeClusters;
	unsigned uAttributes;
	ulong ulSize;
	ulong ulPosition;
	unsigned uAction;
	unsigned uBytesRead;
	unsigned uBytesWritten;
	extern bool bIFSHandledCall;
	bool fUpdateDateTime;
	unsigned uFileTime;
	unsigned uFileDate;
	unsigned uOpenMode;
	bool fDuplicateHandleAndClose;
	unsigned far * lpuInt21CallersRegisters;
	char far * lpTemp;
	extern bool fIsFat32World;
	extern bool fInWindows;
	LPDEV_HDR lpDev;
	extern LPLOL lpLOL;
	uchar ucLength;
	uchar ucLength2;
	bool fCreateNew;
	unsigned uCurrentPSP;


	//  Set some local variable with some of the commonly used
	//  variables in the following case statement
	//
	uAX = CallersRegisters.ax;
	uDI = CallersRegisters.di;
	uES = CallersRegisters.es;
	lpCDS  = *(LPCDS far *)(MK_FP(CallersRegisters.ss, 0x5a2) );
	lpSFT  = MK_FP( uES, uDI );
	lpFind  = *(LPFIND far *)(MK_FP(CallersRegisters.ss, 0x32c) );
	uCurrentPSP = *(unsigned far *)(MK_FP(CallersRegisters.ss, 0x330) );
	lpucDTA = *(uchar far * far *)(MK_FP(CallersRegisters.ss, 0x32c) );
	lpszSDAfname1 = MK_FP(CallersRegisters.ss, 0x3be);
	lpuStackWord = MK_FP(CallersRegisters.ss,	(CallersRegisters.sp) + 6 );


	//  Go to the apropriate handler for this function request
	//
	switch( (uchar)uAX )
	{
		//  Installation check
		//
		//  Int 2fh fn 1100h
		//  Int 2fh fn 1131h  (THIS IFS ONLY)
		//
		//  This function is used to determine if an IFS is installed.
		//  It also returns internal information so that an unload
		//  can be performed if the cookies are received.
		//
		case 0x00:
		case 0x31:
			{
				//  Set a flag indicating that an IFS is installed
				//
				CallersRegisters.ax  = (unsigned)('D') << 8;
				CallersRegisters.ax |= 0xff;


				//  Determine if the caller provided the extra
				//  info cookies
				//
				if( CallersRegisters.bx == 0xbbbb )
				{
					if( CallersRegisters.cx == 0xcccc )
					{
						if( CallersRegisters.dx == 0xdddd )
						{
							//
							//  Caller has provided the extra info
							//  cookies
							//


							//  Set a flag indicating that this is the
							//  client software and the version
							//
							CallersRegisters.cx  =  CLI_TOKEN << 8;
							CallersRegisters.cx |= (uchar)(CLI_VERSION);


							//  Set a flag indicating the network connection
							//  method (serial, modem, network, ... ) and
							//  any other information the network type wants
							//  to return (like comm port IRQ) so that a
							//  subsiquent instance won't conflict
							//
							CallersRegisters.si = GetNetworkTypeWord();
							CallersRegisters.di =
									GetNetworkTypeDetailsWord();


							//  Return the address of a far pointer in the
							//  data segment to the on-exit routines used to
							//  uninstall interrupt handlers and such.
							//
							CallersRegisters.bx = uOnExitPtrOffset;
							CallersRegisters.es = uOnExitPtrSegment;


							//  Return the segment of the PSP
							//
							CallersRegisters.dx = _psp;
						}
					}
				}
				else if( CallersRegisters.dx == 0xbbbb )
				{
					if( CallersRegisters.cx == 0xcccc )
					{
						if( CallersRegisters.bx == 0xdddd )
						{
							//  Get drive map address request
							//
							CallersRegisters.si = (unsigned)(&DriveMap);
							CallersRegisters.es = FP_SEG((void far *)(&DriveMap));
						}
					}
				}
				else if( CallersRegisters.bx == 0xdddd )
				{
					if( CallersRegisters.cx == 0xdddd )
					{
						if( CallersRegisters.dx == 0xdddd )
						{
							//  Caller wants us to unload
							//
							UnloadClient();
						}
					}
				}
			}
			break;


		//  Get remote disk info
		//
		//  Int 2fh fn 110ch
		//
		case 0x0c:
			{
				fIsOurCall = IsOurCDS( MK_FP(uES,uDI), &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  Issue the get drive info request
					//
					uErrorNumber = GetRemoteDriveInfo( ucRemoteDriveLetter,
							&ucSPC, &ucMediaID, &uTotalClusters, &uBPS,
							&uFreeClusters );


					//  Save the results of the drive info call in the
					//  registers
					//
					if( !uErrorNumber )
					{
						CallersRegisters.ax = (((unsigned)ucMediaID) << 8) | ucSPC;
						CallersRegisters.bx = uTotalClusters;
						CallersRegisters.cx = uBPS;
						CallersRegisters.dx = uFreeClusters;
					}
				}
			}
			break;


		//  Remove remote directory
		//
		//  Int 2fh fn 1101h
		//
		//  SS = DOS data segment
		//
		//
		case 0x01:
			{
				fIsOurCall = IsOurCDS( lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Issue the remove directory request
					//
					uErrorNumber = RemoveRemoteDirectory( ucRemoteDriveLetter,
							&(lpCDS->cwd[2]), lpszSDAfname1 );


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Make remote directory
		//
		//  Int 2fh fn 1103h
		//
		case 0x03:
			{
				fIsOurCall = IsOurCDS( lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Issue the make directory request
					//
					uErrorNumber = CreateRemoteDirectory( ucRemoteDriveLetter,
							&(lpCDS->cwd[2]), lpszSDAfname1 );


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Change remote directory
		//
		//  Int 2fh fn 1105h
		//
		case 0x05:
			{
				fIsOurCall = IsOurCDS( lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Issue the change directory request
					//
					uErrorNumber = ChangeRemoteDirectory( ucRemoteDriveLetter,
							&(lpCDS->cwd[2]), lpszSDAfname1 );


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Remote file find first
		//
		//  Int 2fh fn 111bh
		//
		case 0x1b:
			{
				fIsOurCall = IsOurCDS(lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter);
				if( fIsOurCall )
				{
					lpucSDASearchAttribute = MK_FP(CallersRegisters.ss, 0x56d);


					//  Get the address in the SDA for the DIR entry
					//
					lpDir = (LPDIR)(((char far *)lpFind) + 0x15);


					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Issue the find first request
					//
					uErrorNumber = RemoteFindFirst( ucRemoteDriveLetter,
							&(lpCDS->cwd[2]), lpszSDAfname1,
							*lpucSDASearchAttribute, lpFind, lpDir );
					if( uErrorNumber == 3 )
					{
						if( !fInWindows )
						{
							lpTemp = _fstrrchr( lpszSDAfname1, '\\' );
							if( lpTemp )
							{
								lpTemp++;
								lpDev = &(lpLOL->first_dev);
								while( FP_OFF(lpDev) != 0xffff )
								{
									ucLength = 0;
									for( uIndex = 0;
											(uIndex < 8) && (lpDev->name[uIndex] != ' ');
											uIndex++ )
									{
										ucLength++;
									}
									ucLength2 = _fstrlen( lpTemp );
									if( ucLength2 == ucLength )
									{
										if( _fstrnicmp(lpTemp, lpDev->name, ucLength) == 0 )
										{
											uErrorNumber = ERR_LASTFILE;
											break;
										}
									}
									lpDev = lpDev->next;
								}
							}
						}
					}


					//  Set the local drive letter in the find first
					//  structure
					//
					lpFind->drive = toupper(ucLocalDriveLetter) - 'A' + 1;


					//  Set the high bit of the drive letter in the find
					//  first structure.
					//
					//  NOTE:  This is a required operation for an IFS
					//
					lpFind->drive |= 0x80;


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Remote file find next
		//
		//  Int 2fh fn 111ch
		//
		case 0x1c:
			{
				fIsOurCall = IsOurDrive( (lpFind->drive & 0x7f) - 1 + 'A',
						&ucRemoteDriveLetter, &lpCDS );
				if( fIsOurCall )
				{
					//  Get the address in the SDA for the DIR entry
					//
					lpDir = (LPDIR)(((char far *)lpFind) + 0x15);


					//  Set the remote drive letter in the find first
					//  structure
					//
					lpFind->drive = toupper(ucRemoteDriveLetter) - 'A' + 1;
					lpFind->drive |= 0x80;


					//  Issue the find next request
					//
					uErrorNumber = RemoteFindNext( ucRemoteDriveLetter,
							&(lpCDS->cwd[2]), lpFind, lpDir );


					//  Set the local drive letter in the find first
					//  structure
					//
					lpFind->drive = toupper(ucLocalDriveLetter) - 'A' + 1;


					//  Set the high bit of the drive letter in the find
					//  first structure.
					//
					//  NOTE:  This is a required operation for an IFS
					//
					lpFind->drive |= 0x80;
				}
			}
			break;


		//  Qualify remote file name
		//
		//  Int 2fh fn 1123h
		//
		//  NOTE: DOS will do this if the function request fails
		//
		//  On Entry:
		//
		//  DS:SI = ASCIIZ file name to canonicalize
		//  ES:DI = 128 byte buffer for qualified name
		//
		//  On Exit:
		//
		//  ES:DI = Buffer updated
		//
		case 0x23:
			{
				//  Do nothing
			}
			break;


		//  Get remote file attributes and size
		//
		//  Int 2fh fn 110fh
		//
		//  On Entry:
		//
		//
		//  On Exit:
		//
		//  AX    = File attributes
		//  BX:DI = Size of file
		//
		case 0x0f:
			{
				fIsOurCall = IsOurCDS(lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter);
				if( fIsOurCall )
				{
					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Issue the get file attributes and size request
					//
					uErrorNumber = GetRemoteFileAttributesAndSize(
							ucRemoteDriveLetter, &(lpCDS->cwd[2]), lpszSDAfname1,
							&uAttributes, &ulSize );


					//  Save the results of the drive info call in the
					//  registers
					//
					if( !uErrorNumber )
					{
						CallersRegisters.ax = uAttributes;
						CallersRegisters.bx = (unsigned)(ulSize >> 16);
						CallersRegisters.di = (unsigned)ulSize;
					}


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Set remote file attributes
		//
		//  Int 2fh fn 110eh
		//
		//  On Entry:
		//
		//
		//  On Exit:
		//
		//
		case 0x0e:
			{
				fIsOurCall = IsOurCDS(lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter);
				if( fIsOurCall )
				{
					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Issue the get file attributes and size request
					//
					uErrorNumber = SetRemoteFileAttributes(
							ucRemoteDriveLetter, &(lpCDS->cwd[2]), lpszSDAfname1,
							*lpuStackWord );


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Open existing remote file
		//
		//  Int 2fh fn 1116h
		//
		//  On Entry:
		//
		//
		//  On Exit:
		//
		case 0x16:
			{
				ucLocalDriveLetter = toupper(lpszSDAfname1[0]);
				fIsOurCall = IsOurDrive( ucLocalDriveLetter,
						&ucRemoteDriveLetter, &lpCDS );
				if( fIsOurCall )
				{
					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Issue the extended open create request
					//
					uOpenMode = (uchar)(*lpuStackWord);
					uErrorNumber = RemoteOpenFile( ucRemoteDriveLetter,
							&(lpCDS->cwd[2]), lpszSDAfname1, uOpenMode,
							uCurrentPSP, &uRemoteHandle, &uFileTime, &uFileDate,
							&ulSize, &uAttributes );


					//  Save the results of the drive info call in the
					//  registers
					//
					if( !uErrorNumber )
					{
						//  Fill in the SFT
						//
						lpSFT->owner  = uCurrentPSP;
						lpSFT->rwmode = uOpenMode;
						lpSFT->attr   = uAttributes;
						lpSFT->ioctl  = 0x8000;
						lpSFT->time   = uFileTime;
						lpSFT->date   = uFileDate;
						lpSFT->ioctl |= toupper(ucLocalDriveLetter) - 'A';
						lpSFT->size   = ulSize;
						lpSFT->device = 0;   //  No DPB for network drives
						lpSFT->offset = 0;   //  Don't know if this needs to be set
						lpSFT->uRelClusterHigh = 0;
						lpSFT->ulStartingCluster = 0;
						lpSFT->shareid = 0;
						lpSFT->ulAbsCluster = 0;
#if USING_FAT32 && LSEEK_ENHANCEMENT
						lpSFT->uRDlseekHigh = 0;
						lpSFT->uRDlseekLow = 0;
#else
						lpSFT->shareoff = 0;
						lpSFT->unknown = 0;
#endif

						//  Put the ENAME in the SFT
						//
						DOTToEname( lpszSDAfname1, lpSFT->name );


						//  Update these field so that we know
						//  that this SFT belongs to us
						//
						lpSFT->uRelClusterLow = uRemoteHandle;
						*(LPCDS far *)(&(lpSFT->dirsec)) = lpCDS;


//  BT 5-17-01
//
//  MS-DOS seems to do this for us and yet we still have to
//  decrement the handle count on the close ???
//
#if 0
						//  Mark this SFT as used
						//
						lpSFT->handles++;  //  DOS does it
#endif
					}


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Create or truncate remote file
		//
		//  Int 2fh fn 1117h
		//
		//  On Entry:
		//
		//
		//  On Exit:
		//
		case 0x17:
			{
				ucLocalDriveLetter = toupper(lpszSDAfname1[0]);
				fIsOurCall = IsOurDrive( ucLocalDriveLetter,
						&ucRemoteDriveLetter, &lpCDS );
				if( fIsOurCall )
				{
					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					uAttributes = (uchar)(*lpuStackWord);
					fCreateNew = (*lpuStackWord) >> 8;
					uOpenMode = O_RDWR;
					uErrorNumber = RemoteCreateFile( ucRemoteDriveLetter,
							&(lpCDS->cwd[2]), lpszSDAfname1, uAttributes,
							fCreateNew, uCurrentPSP, &uRemoteHandle, &uFileTime,
							&uFileDate, &ulSize, &uAttributes );


					//  Save the results of the drive info call in the
					//  registers
					//
					if( !uErrorNumber )
					{
						//  Fill in the SFT
						//
						lpSFT->owner  = uCurrentPSP;
						lpSFT->rwmode = uOpenMode;
						lpSFT->attr   = uAttributes;
						lpSFT->ioctl  = 0x8000;
						lpSFT->time   = uFileTime;
						lpSFT->date   = uFileDate;
						lpSFT->ioctl |= toupper(ucLocalDriveLetter) - 'A';
						lpSFT->size   = ulSize;
						lpSFT->device = 0;   //  No DPB for network drives
						lpSFT->offset = 0;   //  Don't know if this needs to be set
						lpSFT->uRelClusterHigh = 0;
						lpSFT->ulStartingCluster = 0;
						lpSFT->shareid = 0;
						lpSFT->ulAbsCluster = 0;
#if USING_FAT32 && LSEEK_ENHANCEMENT
						lpSFT->uRDlseekHigh = 0;
						lpSFT->uRDlseekLow = 0;
#else
						lpSFT->shareoff = 0;
						lpSFT->unknown = 0;
#endif


						//  Put the ENAME in the SFT
						//
						DOTToEname( lpszSDAfname1, lpSFT->name );


						//  Update these field so that we know
						//  that this SFT belongs to us
						//
						lpSFT->uRelClusterLow = uRemoteHandle;
						*(LPCDS far *)(&(lpSFT->dirsec)) = lpCDS;


//  BT 5-17-01
//
//  MS-DOS seems to do this for us and yet we still have to
//  decrement the handle count on the close ???
//
#if 0
						//  Mark this SFT as used
						//
						lpSFT->handles++;  //  DOS does it
#endif
					}


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Extended open, create, or truncate remote file
		//
		//  Int 2fh fn 112eh
		//
		//  On Entry:
		//
		//
		//  On Exit:
		//
		case 0x2e:
			{
				ucLocalDriveLetter = toupper(lpszSDAfname1[0]);
				fIsOurCall = IsOurDrive( ucLocalDriveLetter,
						&ucRemoteDriveLetter, &lpCDS );
				if( fIsOurCall )
				{
					lpuSDAExtendedOpenAction = MK_FP(CallersRegisters.ss, 0x5fd);
					lpuSDAExtendedOpenAttr   = MK_FP(CallersRegisters.ss, 0x5ff);
					lpuSDAExtendedOpenMode   = MK_FP(CallersRegisters.ss, 0x601);
					lpuInt21CallersRegisters =
							*((unsigned far * far *)(MK_FP( CallersRegisters.ss,
							0x584 )));


					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Issue the extended open create request
					//
					uAction = *lpuSDAExtendedOpenAction;
					uErrorNumber = RemoteExtendedOpenCreateFile(
							ucRemoteDriveLetter, &(lpCDS->cwd[2]), lpszSDAfname1,
							*lpuSDAExtendedOpenMode, *lpuSDAExtendedOpenAttr,
							uCurrentPSP, &uAction, &uRemoteHandle, &uFileTime,
							&uFileDate, &ulSize, &uAttributes );


					//  Save the results of the drive info call in the
					//  registers
					//
					if( !uErrorNumber )
					{
						//  Update the action
						//
						CallersRegisters.cx = uAction;
						(lpuInt21CallersRegisters)[2] = uAction;


						//  Fill in the SFT
						//
						lpSFT->owner  = uCurrentPSP;
						lpSFT->rwmode = *lpuSDAExtendedOpenMode;
						lpSFT->attr   = *lpuSDAExtendedOpenAttr;
						lpSFT->ioctl  = 0x8000;
						lpSFT->time   = uFileTime;
						lpSFT->date   = uFileDate;
						lpSFT->ioctl |= toupper(ucLocalDriveLetter) - 'A';
						lpSFT->size   = ulSize;
						lpSFT->device = 0;   //  No DPB for network drives
						lpSFT->offset = 0;   //  Don't know if this needs to be set
						lpSFT->uRelClusterHigh = 0;
						lpSFT->ulStartingCluster = 0;
						lpSFT->shareid = 0;
						lpSFT->ulAbsCluster = 0;
#if USING_FAT32 && LSEEK_ENHANCEMENT
						lpSFT->uRDlseekHigh = 0;
						lpSFT->uRDlseekLow = 0;
#else
						lpSFT->shareoff = 0;
						lpSFT->unknown = 0;
#endif


						//  Put the ENAME in the SFT
						//
						DOTToEname( lpszSDAfname1, lpSFT->name );


						//  Update these field so that we know
						//  that this SFT belongs to us
						//
						lpSFT->uRelClusterLow = uRemoteHandle;
						*(LPCDS far *)(&(lpSFT->dirsec)) = lpCDS;


//  BT 5-17-01
//
//  MS-DOS seems to do this for us and yet we still have to
//  decrement the handle count on the close ???
//
#if 0
						//  Mark this SFT as used
						//
						lpSFT->handles++;  //  DOS does it
#endif
					}


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Set the position of a remote file
		//
		//  Int 2fh fn 1121h
		//
		//  On Entry:
		//
		//  ES:DI = SFT
		//  CX:DX = Bytes from end of file
		//
		//  On Exit:
		//
		//  DX:AX = Number of bytes actually read
		//  SFT updated
		//
		case 0x21:
			{
				lpCDS = *(LPCDS far *)(&(lpSFT->dirsec));
				fIsOurCall = IsOurCDS( lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  Perform the read operation
					//
					uRemoteHandle = lpSFT->uRelClusterLow;
					ulPosition   = CallersRegisters.cx;
					ulPosition <<= 16;
					ulPosition  |= CallersRegisters.dx;
					uErrorNumber = SetRemoteFilePosition( uRemoteHandle,
							ulPosition, &ulPosition );


					//  Determine if there was an error
					//
					if( !uErrorNumber )
					{
						//  Update the position in the return registers
						//
						CallersRegisters.dx = (unsigned)(ulPosition >> 16);
						CallersRegisters.ax = (unsigned)(ulPosition);


						//  Update the position in the SFT
						//
						lpSFT->offset = ulPosition;
					}
				}
			}
			break;


		//  Read from remote file
		//
		//  Int 2fh fn 1108h
		//
		//  On Entry:
		//
		//  ES:DI           = SFT
		//  CX              = Number of bytes to read
		//  SDA DTA pointer = Address of users buffer
		//
		//  On Exit:
		//
		//  CX = Number of bytes actually read
		//  SFT updated
		//
		case 0x08:
			{
				lpCDS = *(LPCDS far *)(&(lpSFT->dirsec));
				fIsOurCall = IsOurCDS( lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  Perform the read operation
					//
					uRemoteHandle  = lpSFT->uRelClusterLow;
					ulPosition = lpSFT->offset;
					uErrorNumber = ReadRemoteFile( uRemoteHandle, ulPosition,
							CallersRegisters.cx, lpucDTA,
							&uBytesRead );


					//  Determine if there was an error
					//
					if( !uErrorNumber )
					{
						//  Update the number of bytes read in the return
						//  registers
						//
						CallersRegisters.cx = uBytesRead;


						//  Update the position in the SFT
						//
						lpSFT->offset += uBytesRead;
					}
				}
			}
			break;


		//  Write to remote file
		//
		//  Int 2fh fn 1109h
		//
		//  On Entry:
		//
		//  ES:DI = SFT
		//
		//  On Exit:
		//
		//  SFT updated
		//
		case 0x09:
			{
				lpCDS = *(LPCDS far *)(&(lpSFT->dirsec));
				fIsOurCall = IsOurCDS( lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  Perform the write operation
					//
					uRemoteHandle = lpSFT->uRelClusterLow;
					ulPosition = lpSFT->offset;
					uErrorNumber = WriteRemoteFile( uRemoteHandle, ulPosition,
							CallersRegisters.cx, lpucDTA,
							&uBytesWritten );


					//  Determine if there was an error
					//
					if( !uErrorNumber )
					{
						//  Update the number of bytes written in the return
						//  registers
						//
						CallersRegisters.cx = uBytesWritten;


						//  Update the position in the SFT
						//
						lpSFT->offset += uBytesWritten;
						if( lpSFT->offset > lpSFT->size )
						{
							lpSFT->size = lpSFT->offset;
						}
					}
				}
			}
			break;


		//  Close remote file
		//
		//  Int 2fh fn 1106h
		//
		//  On Entry:
		//
		//  ES:DI = SFT
		//
		//  On Exit:
		//
		//  SFT updated
		//
		case 0x06:
			{
				lpCDS = *(LPCDS far *)(&(lpSFT->dirsec));
				fIsOurCall = IsOurCDS( lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  DOS will never open the same SFT multiple times using
					//  the DOS open functions.  However, DOS will allow the
					//  handle to be duplicated but won't bother telling us,
					//  the IFS, about it.  So, if the handle count is greater
					//  than one, someone has duplicated the handle and are
					//  calling for a close of one of the handles.  The result
					//  of the expression below is passed to the server and
					//  in the case of it being TRUE, the server end will
					//  duplicate the handle and close the duplicated handle.
					//
					fDuplicateHandleAndClose = (lpSFT->handles > 1);


					//  Perform the close operation
					//
					uRemoteHandle = lpSFT->uRelClusterLow;
					if( (uchar)(lpSFT->rwmode) )
					{
						fUpdateDateTime = TRUE;
						uFileTime = lpSFT->time;
						uFileDate = lpSFT->date;
					}
					else
					{
						fUpdateDateTime = FALSE;
					}
					uErrorNumber = CloseRemoteFile( uRemoteHandle, fUpdateDateTime,
							uFileTime, uFileDate, fDuplicateHandleAndClose );


					//  Determine if there was an error
					//
					if( !uErrorNumber )
					{
						lpSFT->handles--;
					}
				}
			}
			break;


		//  Close all files for remote file termination
		//
		//  Int 2fh fn 111dh
		//
		//  On Entry:
		//
		//
		//  On Exit:
		//
		//
		case 0x1d:
			{
				CloseAllRemoteFiles( uCurrentPSP );
				fIsOurCall = FALSE;		//  Allow other IFS to recieve the call
			}
			break;


		//  Commit remote file
		//
		//  Int 2fh fn 1107h
		//
		//  On Entry:
		//
		//  ES:DI = SFT
		//
		//  On Exit:
		//
		//  SFT updated
		//
		case 0x07:
			{
				lpCDS = *(LPCDS far *)(&(lpSFT->dirsec));
				fIsOurCall = IsOurCDS( lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  Perform the close operation
					//
					uRemoteHandle = lpSFT->uRelClusterLow;
					if( (uchar)(lpSFT->rwmode) )
					{
						fUpdateDateTime = TRUE;
						uFileTime = lpSFT->time;
						uFileDate = lpSFT->date;
					}
					else
					{
						fUpdateDateTime = FALSE;
					}
					uErrorNumber = CommitRemoteFile( uRemoteHandle,
							fUpdateDateTime, uFileTime, uFileDate );
				}
			}
			break;


		//  Rename remote file
		//
		//  Int 2fh fn 1111h
		//
		case 0x11:
			{
				fIsOurCall = IsOurCDS(lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					lpszSDAfname2 = MK_FP(CallersRegisters.ss, 0x43e);


					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname2[1] == ':' )
					{
						if( toupper(lpszSDAfname2[0]) != ucLocalDriveLetter )
						{
							//  Don't rename accross drives
							//
							uErrorNumber = 0x0005;
							break;
						}
						lpszSDAfname2[0] = ucRemoteDriveLetter;
						fChangedSDAfname2Letter = TRUE;
					}

					//  Issue the rename file request
					//
					uErrorNumber = RenameRemoteFile( ucRemoteDriveLetter,
							&(lpCDS->cwd[2]), lpszSDAfname1, lpszSDAfname2 );


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname2Letter )
					{
						lpszSDAfname2[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Delete remote file
		//
		//  Int 2fh fn 1113h
		//
		case 0x13:
			{
				fIsOurCall = IsOurCDS(lpCDS, &ucLocalDriveLetter,
						&ucRemoteDriveLetter );
				if( fIsOurCall )
				{
					//  Change the local drive letter to the
					//  remote drive letter, if there was a drive
					//  letter
					//
					if( lpszSDAfname1[1] == ':' )
					{
						lpszSDAfname1[0] = ucRemoteDriveLetter;
						fChangedSDAfname1Letter = TRUE;
					}


					//  Issue the delete file request
					//
					uErrorNumber = DeleteRemoteFile( ucRemoteDriveLetter,
							&(lpCDS->cwd[2]), lpszSDAfname1 );


					//  Restore the local drive letter if it
					//  was changed to the remote drive letter
					//
					if( fChangedSDAfname1Letter )
					{
						lpszSDAfname1[0] = ucLocalDriveLetter;
					}
				}
			}
			break;


		//  Don't know what this function is.  Must not be our call.
		//
		default:
			break;
	}


	//  Determine if the call was handled
	//
	bIFSHandledCall = fIsOurCall;
	if( fIsOurCall )
	{
		//  Set the registers and return to the caller
		//
		if( uErrorNumber )
		{
			CallersRegisters.ax = uErrorNumber;
			CallersRegisters.flags |= FLAG_CARRY;
		}
		else
		{
			CallersRegisters.flags &= ~FLAG_CARRY;
		}
	}
	INT2fReturn();
}



//  DOTToEname
//
//
//  Create the ENAME form of the file name rom the DOT form
//
//
//  On Entry:
//
//    lpszDotName = File name in DOT form (with path OK)
//
//
//  On Exit:
//
//    lpszENAME = updated with ENAME
//
//
void DOTToEname( char far * lpszDotName, char far * lpszENAME )
{
	char far * lpszTemp;
	unsigned uIndex;


	lpszTemp = _fstrrchr( lpszDotName, '\\' );
	if( !lpszTemp )
	{
		lpszTemp = _fstrchr( lpszDotName, ':' );
		if( !lpszTemp )
		{
			lpszTemp = lpszDotName;
		}
		else
		{
			lpszTemp++;
		}
	}
	else
	{
		lpszTemp++;
	}
	for( uIndex = 0; uIndex < 8; uIndex++ )
	{
		if( (*lpszTemp == '.') || (*lpszTemp == 0) )
		{
			lpszENAME[uIndex] = ' ';
		}
		else
		{
			lpszENAME[uIndex] = *lpszTemp;
			lpszTemp++;
		}
	}
	if( *lpszTemp == '.' )
	{
		lpszTemp++;
	}
	for( uIndex = 8; uIndex < 11; uIndex++ )
	{
		if( *lpszTemp == 0 )
		{
			lpszENAME[uIndex] = ' ';
		}
		else
		{
			lpszENAME[uIndex] = *lpszTemp;
			lpszTemp++;
		}
	}
}


//  GetSFTForIndex
//
//
//  Find the SFT for the given index into the handle table pointed
//  to by the PSP.
//
//
//  On Entry:
//
//    ucSFTIndex = Value in the handle table
//    plpSFT     = Address of variable to return the SFT address
//
//
//  On Exit:
//
//  on success,
//
//    plpSFT = Updated
//
//    function returns TRUE
//
//  on failure,
//
//    function returns FALSE
//
//
bool GetSFTForIndex( uchar ucSFTIndex, LPSFT * plpSFT )
{
	extern LPLOL lpLOL;
	LPDCB lpDCB;


	if( ucSFTIndex != 0xff )
	{
		lpDCB = lpLOL->first_fcb;
		while( ucSFTIndex > lpDCB->numsft )
		{
			ucSFTIndex -= lpDCB->numsft;
			lpDCB = lpDCB->next;
		}
		*plpSFT = &(lpDCB->sft[ucSFTIndex]);
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}


//  MapRemoteDrive
//
//
//  Add this drive to the list of remote drives that are serviced.
//  This function also sets up the CDS for the local drive mapping.
//
//
//  On Entry:
//
//    ucRemoteDriveLetter = The drive letter of the remote drive
//    fIsRemovable        = Whether the remote drive is removable
//    fIsRemote           = Whether the remote drive is remote
//
//
//  On Exit:
//
//    On Success,
//
//      Function returns TRUE
//
//    On Failure,
//
//      Function returns FALSE
//
//
bool MapRemoteDrive( uchar ucRemoteDriveLetter,	bool fIsRemovable,
		bool fIsRemote )
{
	static uchar ucIndex = 0;
	LPCDS lpCDS;
	static bool fShownOnce = FALSE;
	bool fSuccessful;
	uchar ucLocalDriveLetter;
	unsigned uLookupIndex;


	//  Only 26 drive letters can be mapped
	//
	if( ucIndex >= 26 )
	{
		return FALSE;
	}


	//  Determine if there is a list of drives that we are to only map
	//
	if( szOnlyMapList[0] )
	{
		//  Find this drive in the "only map these drives list"
		//
		uLookupIndex = 0;
		while( szOnlyMapList[uLookupIndex] )
		{
			if( szOnlyMapList[uLookupIndex] == ucRemoteDriveLetter )
			{
 				//  Drive is in the "only map these drives list"
				//
				break;
			}
			uLookupIndex++;
		}
		if( szOnlyMapList[uLookupIndex] == 0 )
		{
			//  Drive is not in the "only map these drives list"
			//
			return FALSE;
		}
	}


	//  Determine if there is a list of drives that the user
	//  does not want mapped
	//
	if( szDontMapList[0] )
	{
		//  Find this drive in the "dont map these drives list"
		//
		uLookupIndex = 0;
		while( szDontMapList[uLookupIndex] )
		{
			if( szDontMapList[uLookupIndex] == ucRemoteDriveLetter )
			{
 				//  Drive is in the "dont map these drives list"
				//
				return FALSE;
			}
			uLookupIndex++;
		}
	}


	//  Determine if the remote drive is remote
	//
	if( fIsRemote )
	{
		//  Determine if mapping remote drives will be allowed
		//
		if( DontMapNetworkDrives() )
		{
			return FALSE;
		}
	}


	//  Determine if the remote drive is removable
	//
	if( fIsRemovable )
	{
		//  Determine if mapping removable drives will be allowed
		//
		if( DontMapRemovableDrives() )
		{
			return FALSE;
		}
	}


	//  Find an available CDS
	//
	fSuccessful = FindAvailableCDS( &lpCDS, &ucLocalDriveLetter );
	if( !fSuccessful )
	{
		if( !fShownOnce )
		{
			fShownOnce = TRUE;
			printf("Not enough drives letters to map all remote drives.\n");
		}
		return FALSE;
	}


	//  Add this entry to the remote drive list
	//
	DriveMap[ucIndex].uCDSOffset = FP_OFF( lpCDS );
	DriveMap[ucIndex].ucRemoteDriveLetter = toupper(ucRemoteDriveLetter);
	DriveMap[ucIndex].ucLocalDriveLetter  = toupper(ucLocalDriveLetter);
	uCDSSegment = FP_SEG(lpCDS);


	//  Increment the count of mapped drives2
	//
	uNumberOfMappedDrives++;


	//  Set up the CDS for DOS to use
	//
	_fmemset( lpCDS, 0, sizeof(CDS) );
	lpCDS->cwd[0] = ucLocalDriveLetter;
	lpCDS->cwd[1] = ':';
	lpCDS->cwd[2] = '\\';
	lpCDS->cwd[3] = 0;
	lpCDS->skip = 2;
	lpCDS->ulSubDirStartingCluster = 0xffffffffL;
//	lpCDS->unknown2 = 0xffff;
	lpCDS->status = C_NETWORK | C_VALID;


	//  Increament the index to the next available DriveMap
	//
	ucIndex++;


	//  Return that the drive mapping was succcessful
	//
	return TRUE;
}


//  FindAvailableCDS
//
//
//  Return the address of an unused CDS and its drive letter
//
//
//  On Entry:
//
//    plpCDS         = Variable for the address of the available CDS
//    pucDriveLetter = Variable for the drive letter of the CDS
//
//
//  On Exit:
//
//    On Success,
//
//      plpCDS         = Updated with the available CDS
//      pucDriveLetter = Updated with the drive letter of the CDS
//
//      Function returns TRUE
//
//    On Failure,
//
//      Function returns FALSE
//
//
bool FindAvailableCDS( LPCDS * plpCDS, uchar * pucDriveLetter )
{
	extern LPLOL lpLOL;
	static unsigned uCDSCount;
	static LPCDS lpCDS;


	//  Determine the number of entries in the CDS array
	//
	uCDSCount = lpLOL->lastdrive;


	//  Get the first CDS
	//
	lpCDS = lpLOL->first_cds;


	//  Find an unused CDS
	//
	*pucDriveLetter = 'A';
	*pucDriveLetter += lpLOL->lastdrive - uCDSCount;
	while( uCDSCount )
	{
		//  Determine if this CDS is available
		//
		if( lpCDS->status == 0 )
		{
			*plpCDS = lpCDS;
			return TRUE;
		}


		//  Set the address of the next CDS
		//
		lpCDS++;


		//  Increment the drive letter
		//
		(*pucDriveLetter)++;


		//  Decrement the remaining number of CDSs left to search
		//
		uCDSCount--;
	}


	return FALSE;
}


//  DisplayDriveMapping
//
//
//  Display a list of all the mapped drives
//
//
//  On Entry:
//
//    Nothing
//
//
//  On Exit:
//
//    Nothing
//
//
void DisplayDriveMapping( void )
{
	unsigned uIndex;


	for( uIndex = 0; uIndex < 26; uIndex++ )
	{
		if( DriveMap[uIndex].ucRemoteDriveLetter )
		{
			printf("%c:\\ assigned to remote drive => %c:\\\n",
					DriveMap[uIndex].ucLocalDriveLetter,
					DriveMap[uIndex].ucRemoteDriveLetter );
		}
		else
		{
			break;
		}
	}
}


//  UnmapDrives
//
//
//  Invalidate all the drives that were mapped to the IFS both in
//  the CDS settings and in the internal table
//
//
//  On entry:
//
//    nothing
//
//
//  On exit:
//
//    nothing
//
//
void UnmapDrives( void )
{
	unsigned uIndex;
	uchar ucRemoteDriveLetter;
	LPCDS lpCDS;
	bool fSuccessful;


	for( uIndex = 0; uIndex < 26; uIndex++ )
	{
		if( DriveMap[uIndex].ucLocalDriveLetter )
		{
			fSuccessful = IsOurDrive( DriveMap[uIndex].ucLocalDriveLetter,
					&ucRemoteDriveLetter, &lpCDS );
			if( fSuccessful )
			{
				lpCDS->status = 0;
				DriveMap[uIndex].ucLocalDriveLetter = 0;
			}
		}
		else
		{
			break;
		}
	}
}


#if 0
//  DebugPrintf
//
//
//  Output a character to the display going directly through INT 10h
//
//
//  On Entry:
//
//    szString = the string to output
//
//
//  On Exit:
//
//    Nothing
//
//
void DebugPrintf( char * szString )
{
	while( *szString )
	{
		if( *szString == '\\' )
		{
			szString++;
			switch( toupper(*szString) )
			{
				case 0x00:
					break;

				case 'R':
					DebugPutc( 0x0a );
					break;

				case 'B':
					DebugPutc( 0x08 );
					break;

				case 'N':
					DebugPutc( 0x0d );
					DebugPutc( 0x0a );
					break;

				default:
					DebugPutc( *szString );
					break;

			}
			szString++;
		}
		else
		{
			if( *szString == 0x0d )
			{
				DebugPutc( *szString );
//				DebugPutc( 0x0a );
			}
			else
			{
				DebugPutc( *szString );
			}
			szString++;
		}
	}
}


//  DebugPutc
//
//
//  Output a character to the display going directly through INT 10h
//
//
//  On Entry:
//
//    cCharacter = the character to output
//
//
//  On Exit:
//
//    Nothing
//
//
void DebugPutc( char cCharacter )
{
	asm	mov	ah, 0eh
	asm	mov	al, cCharacter
	asm	mov	bx, 0101h
	asm	int   10h
}
#endif
