//
//   $Id: NETAPI.C 1.8 2002/10/16 21:42:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  NETAPI.C
//
//  Copyright (c) 2001-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  This file contains all the API functions for a network transport
//  layer interface is expected to perform.
//
//  History:
//
//  04-01-2001  bt  Created
//


//  Local libraries
//
#include "datalght.h"
#include "dlbiosio.h"


//  Local definitions
//
#include "netapi.h"
#include "commapi.h"


//  Prototypes
//
#include "netapi.pt"



//  ========================================================================
//
//  The following functions are required APIs
//
//  ========================================================================



//  NetworkTXRXInit
//
//
//  Initialize the "network" interface
//
//
//  On entry:
//
//    argc = The number of options in the argv array
//    argv = Array of pointers to command line options
//
//  On exit:
//
//  on success,
//
//  	function returns TRUE
//
//  on error,
//
//  	function returns FALSE
//
//
bool NetworkTXRXInit( int argc, char *argv[] )
{
	bool fSuccessful;


	fSuccessful = TXRXCommInit( argc, &argv[0] );
	return fSuccessful;
}



//  NetworkTXRXDeInit
//
//
//  Uninitialize the "network" interface
//
//
//  On entry:
//
//    nothing
//
//  On exit:
//
//  on success,
//
//  	function returns TRUE
//
//  on error,
//
//  	function returns FALSE
//
//
bool NetworkTXRXDeInit( void )
{
	bool fSuccessful;


	fSuccessful = TXRXCommDeInit();
	return fSuccessful;
}



//  NetworkTX
//
//
//  Transmit a packet over the network
//
//
//  On entry:
//
//    pPacket     = The address of the packet (data)
//    uPacketSize = The size of the packet to transmit
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns an error code defined in NETAPI.H
//
//
unsigned NetworkTX( void * pPacket, unsigned uPacketSize )
{
	return TXCommPacket( pPacket, uPacketSize );
}


//  NetworkRX
//
//
//  Receive a packet from the network
//
//
//  Note:  This function must be able to receive the maximum
//         possible packet size
//
//
//  On entry:
//
//    pPacket = The address of the packet (data)
//
//  On exit:
//
//  on success,
//
//    *pPacket = Updated with the packet (data)
//
//  	function returns 0
//
//  on error,
//
//		function returns an error code defined in NETAPI.H
//
//
unsigned NetworkRX( void * pPacket )
{
	return RXCommPacket( pPacket );
}




//  ========================================================================
//
//  The remainder of this file is specific to the implementation.
//
//  ========================================================================




//  ServiceRequest
//
//
//  Transmit a packet and wait to receive a packet
//
//
//  Note:  This function must be able to receive the maximum
//         possible packet size
//
//
//  On entry:
//
//    pPacket = The address of the packet (data)
//
//  On exit:
//
//  on success,
//
//    *pPacket = Updated with the packet (data)
//
//  	function returns 0
//
//  on error,
//
//		function returns an error code defined in NETAPI.H
//
//
unsigned ServiceRequest( void * pRequestPacket, unsigned uPacketSize )
{
	unsigned uError;
	ulong ulWaitTicks;
	ulong ulCurrentTicks;
	ulong ulEndTicks;
	extern unsigned uSecondsToTimeout;


	//  Transmit the request packet
	//
	ulWaitTicks  = 182 * uSecondsToTimeout;
	ulWaitTicks /= 10;
	dlGetTimerTicks( &ulCurrentTicks );
	ulEndTicks = ulCurrentTicks + ulWaitTicks;
	while( ulCurrentTicks < ulEndTicks )
	{
		uError = TXCommPacket( pRequestPacket, uPacketSize );
		if( uError == NET_SERVICED )
		{
			dlGetTimerTicks( &ulCurrentTicks );
			ulEndTicks = ulCurrentTicks + ulWaitTicks;
 			while( ulCurrentTicks < ulEndTicks )
			{
				//  Receive the service packet
				//
				uError = RXCommPacket( pRequestPacket );
				if( uError == NET_SERVICED )
				{
					break;
				}
				dlGetTimerTicks( &ulCurrentTicks );
			}
			break;
		}
		dlGetTimerTicks( &ulCurrentTicks );
	}

	return uError;
}


//  GetNetworkTypeDetailsWord
//
//
unsigned GetNetworkTypeDetailsWord( void )
{
	return 0;
}

//  GetNetworkTypeWord
//
//
unsigned GetNetworkTypeWord( void )
{
	return 0;
}
