//
//   $Id: COMM.C 1.5 2002/10/16 21:42:38Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//

/*******************************************************************
  COMM.C  created 04/20/88  by Drew Gislason
  Modified  01/08/2002  pb  Changed the SIGNED int line number to an
                            UNSIGNED long since line numbers can't be
									 negative... EVER!!!!
  Modified  05/15/88  DG  added in dial/parms
  Modified  06/02/88  DG  added support for env and cmd line options

   Copyright (c) 1987-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

  Environment:
    COMM= {command line option(s)}

  Command Line Options:
    /B#    Set Baud 300,1200,2400...115200
    /COM#  Set Port one of {1,2,3 or 4}
	 /IRQ#  Set IRQ one of {3,4,5,6,7,8,9,10,11,12,13,14,15}
    /N81   Set Data Bits
    /E71
	 /H#	  Set handshaking mode: 0=none, 1=XON/XOFF, 2=RTS/CTS, 3=DTR/DSR

  Keys:
    Alt-B    Baud Rate
    Alt-C    Clear Screen
    Alt-D    Dial
    Alt-E    Echo on/off
	 Alt-H    Hang Up
	 Alt-P    Parameters
    Alt-T    Toggle CR/LF
    Alt-X    Exit
    PgUp     Upload   (send a file)
    PgDn     Download (receive a file)

  Terminal Emulation (ANSI only):
    Esc[#A    Up    (default is 1 for all #s)
    Esc[#B    Down
    Esc[#C    Right
    Esc[#D    Left
    Esc[#;#H  Position cursor to (row);(col) (1,1 == 0,0)
    Esc[#J    Erase to EndOfPage (if #==2 move to 0,0 then eeop)
    Esc[#K    Erase to EndOfLine
    Esc[#m    Set fore color (30-37), background color (40-47), or
              intensity (0 or 1)

 Download which file:
 Download ASCII, Xmodem (no CRC), or ZMODEM -- (a,x or z):
 Autodial:

 Uploading FILE.C
 27945 bytes, 128 block(s)
 transfer time: 10 min 34 sec

********************************************************************/
#include <dos.h>
#include <ctype.h>
#include <fcntl.h>
#include <io.h>
#include <mem.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* standard stuff */
typedef int boolean;

#include "datalght.h"
#include "dlbiosio.h"
#include "dl8250.h"

#include "comm.pt"
#include "disp2.pt"
#include "bgetch.pt"
#include "checksum.pt"

#ifndef FALSE
#define FALSE	(1==0)
#define TRUE	!FALSE
#endif

#if !USE_ZMODEM
#include "break.pt"
#endif

#define TICKS_PER_SEC		  18

#define NO_HANDSHAKING 		  0
#define HARDWARE_HANDSHAKING 1
#define SOFTWARE_HANDSHAKING 2

#if !USE_ZMODEM
#define RXD_BUFF_SIZE	1048
#define TXD_BUFF_SIZE   0
#else
#define RXD_BUFF_SIZE	8196
#define TXD_BUFF_SIZE	8196
#endif
unsigned char rxdBuffer[RXD_BUFF_SIZE+TXD_BUFF_SIZE];

// messages
#define COMM 1
#include "dosver.h"
#include "dosmsg.c"


// The MCOMMIFC.C file contains a bunch of translation code from the
// Datalight 8250 module interface to the MCOMM routines. The bulk
// of the existing code can be used without changes.  It also includes
// MCOMM libary specific routines to perform ZMODEM file transfers.
//
#if USE_ZMODEM
#include "mcommifc.c"
#undef MONO
#else
#include "clock.h"
#endif
#include "bios.h"


#define ERR   -1


/* available keys (commands) */
#define ALTB Alt_Z+4
#define ALTC Alt_Z+2
#define ALTD Alt_A+2
#define ALTE Alt_Q+2
#define ALTH Alt_A+5
#define ALTK Alt_A+7
#define ALTP Alt_Q+9
#define ALTR Alt_Q+3
#define ALTT Alt_Q+4
#define ALTX Alt_Z+1
#define PAGEUP PgUp
#define PAGEDN PgDn
#define BELL 7

/* some other defines */
#define COMROW1 0
#define COMROW2 23      /* eventually we shall have chat mode */
#define beep() bdos(6,0,7)

/* the comm options */
int port;					/* 0,1 */
int iIrqNumber=0;			/* 3,4,5,6,7,8,9,10,11,12,13,14,15 */
int baud=4;					/* 300,1200,2400,4800,9600,19200,38400,115200 */
unsigned ubaud = 9600;
int parms;					/* 0-4 */
int rdisk;
boolean echo;
boolean crtrans;
boolean opt_dest;			/* destructive backspace -- not implemented */

#define NUMPARMS 10
int parmlist[] = { N81, N81, E71, E71, N82, N82, O81, O81, E81, E81 };
char *parmname[] = { "N81", "8N1", "E71", "7E1", "N82", "8N2", "O81", "8O1", "E81", "8E1" };

/* ANSI.SYS support variables and screen stuff */
#define MAX_ANSI_STATE   5
#define FIRST_ANSI_PARM  2
int  ansirow=0, ansicol=0;
int  ansiForeColor = 7;		// initial color is white on black
int  ansiBackColor = 0;
int  ansiIntensity = 0;
unsigned char ansiAttr = 7;
int  ansistate=0;
int  ansiparm[MAX_ANSI_STATE - FIRST_ANSI_PARM];
char ansis[4];
int  ansiIntDigit;

/*
	Display handles ANSI translations.  We currently
	understand the codes:

	 Esc[#A    Up     (default is 1 for all #s)
	 Esc[#B    Down
	 Esc[#C    Right
	 Esc[#D    Left
	 Esc[#;#H  Position cursor to (row);(col) (1,1 == 0,0)
	 Esc[#J    Erase to EndOfPage (if #==2 move to 0,0 then eeop)
	 Esc[#K    Erase to EndOfLine
	 Esc[#m    Set fore color (30-37), background color (40-47), or
				  intensity (0 or 1)
*/
void disp(unsigned char c)
{
	// we support the following ansi screen codes
	static char supportedCodes[] = "ABCDHJKm";
	unsigned char oldAttr;

	// ansi codes start with excape
	if(c==ESC)
		ansistate=1, ansiIntDigit=0;

	// already in an ansi state, see how far
	else if(ansistate)
	{
		// must have a open bracket next
		if(ansistate==1)
		{
			if(c=='[')
				++ansistate;

			// 2nd character wasn't a bracket
			// no longer in esc sequence
			else
				ansistate=0;
		}

		// we allow up to 3 parms (states 2-4)
		else if(ansistate >= FIRST_ANSI_PARM && ansistate < MAX_ANSI_STATE)
		{
			// got the ansi code
			if(strchr(supportedCodes, c))
			{
				// we have some digits, that means this PARM
				if(ansiIntDigit)
				{
					ansis[ansiIntDigit] = 0;
					ansiparm[ansistate - FIRST_ANSI_PARM] = atoi(ansis);
					++ansistate;
				}

				// OK, we're done
				ansicode(c);
				return;
			}

			// a digit, continue
			else if(c>='0' && c<='9')
			{
				// ignore too many digits
				if(ansiIntDigit < sizeof(ansis)-1)
				{
					ansis[ansiIntDigit] = c;
					++ansiIntDigit;
				}
			}

			// separator between numbers
			else if(c==';')
			{
				// done with this parameter
				ansis[ansiIntDigit] = NULL;
				ansiparm[ansistate-FIRST_ANSI_PARM] = atoi(ansis);

				// on to the next
				++ansistate;
				ansiIntDigit = 0;
			}

			// not a recognized code
			else
				ansistate=0;
		}
		else
			ansistate=0;
	}

	// display digits outside of ansistate
	if(!ansistate)
	{
		// display all printable characters
		if(isprint(c) || c>0x80)
		{
			// go to ansi position
			disp_move(ansirow, ansicol);

			// display the character, in the right color
			oldAttr = disp_setattr(ansiAttr);
			disp_putc(c);
			disp_setattr(oldAttr);

			++ansicol;
			if(ansicol>=80)
			{
				ansicol=0;
				if(ansirow==COMROW2)
					scrollup();
				else
					++ansirow;
			}
		}

		/* non-destructive backspace */
		else if(c==BS)
		{
			if(ansicol>0)
			{
				--ansicol;
				if(opt_dest)
				{
					disp_move(ansirow,ansicol);
					disp_putc(' ');
				}
			}
		}

		else if(c==CR)
		{
			ansicol=0;
			if(crtrans)
				disp(LF);
		}

		else if(c==LF)
		{
			if(ansirow==COMROW2)
				scrollup();
			else
				++ansirow;
		}

		else if(c==BELL)
			bdos(6, 0,BELL);

		disp_move(ansirow,ansicol);
		disp_flush();
	}	/* end of if !state */
}

void reset_cursor( void )
{
	disp_move(ansirow,ansicol);
	disp_flush();
}

/*
	Actually perform the ANSI code
*/
void ansicode(int c)
{
	static unsigned char colors[8] = { 0, 4, 2, 6, 1, 5, 3, 7 };
	boolean worked=TRUE;
	int i, n;

	/* set all undefined parameters to 1 */
	for(i=ansistate; i<MAX_ANSI_STATE; ++i)
		ansiparm[i-FIRST_ANSI_PARM] = 1;

	switch(c)
	{
		case 'A':	/* up */
			ansirow-=ansiparm[0];	break;
		case 'B':	/* down */
			ansirow+=ansiparm[0];	break;
		case 'C':	/* right */
			ansicol+=ansiparm[0];	break;
		case 'D':	/* left */
			ansicol-=ansiparm[0];	break;
		case 'H':	/* cursepos */
			ansirow=ansiparm[0]-1;
			ansicol=ansiparm[1]-1;
			break;
		case 'J':	/* erase to end of page */
			if(ansiparm[0]==2)
				ansirow=ansicol=0;
			disp_move(ansirow,ansicol);
			disp_eeop();
			status();
			break;
		case 'K':	/* erase to end of line */
			disp_move(ansirow,ansicol);
			disp_eeol();
			break;

		case 'm':	/* set colors */

			ansiForeColor = 7;		// establish dim white on black defaults
			ansiBackColor = 0;
			ansiIntensity = 0;

			// get all parameters
			for(i=0; i<ansistate-FIRST_ANSI_PARM; ++i)
			{
				n = ansiparm[i];

				// forground color
				if(n >= 30 && n <= 37)
					ansiForeColor = colors[n - 30];

				// background color
				else if(n >= 40 && n <= 47)
					ansiBackColor = (colors[n - 40] << 4);

				// intensity of foreground color
				else if(n == 0)
					ansiIntensity = 0;

				else if(n == 1)
					ansiIntensity = 8;
			}

			/* set new color */
			ansiAttr = (ansiBackColor | ansiForeColor | ansiIntensity);
			break;

		default: worked=FALSE;	break;
	}

	if(worked)
	{
		if(ansirow<0) ansirow=0;
		if(ansirow>COMROW2) ansirow=COMROW2;
		if(ansicol>79) ansicol=79;
		if(ansicol<0) ansicol=0;

		disp_move(ansirow,ansicol);
		disp_flush();
	}

	/* done */
	ansistate = 0;
}

/***** serial port routines ******/

void comm_timed_flush(void)
{
	cl_twait(TICKS_PER_SEC);
	while(comm_poll()!=ERR)
		/* do nothing */;
}

void ecomputc(int c)
{
	if(echo)
		disp(c);
	comm_putc(c);
}

void ecomputs(char *s)
{
	while(*s) {
		ecomputc(*s);
		++s;
		}
}

void comm_puts(char *s)
{
	while(*s) {
		comm_putc(*s);
		++s;
		}
}

/********************************************************
* Area: ASCII file transfers                            *
*********************************************************/
int ascii_txd(int fd)
{
	char s[16];
	int c;
	int i,len;
	unsigned long line=0;

	center("Press Esc to Abort");
	while(TRUE) {
		if((c=b_poll())==ESC) {
			comm_timed_flush();
			return ERR;
			}
		while((c=comm_poll())!=ERR)
			disp(c);
		if((len=read(fd,s,16))<=0)
			break;
		for(i=0;i<len;++i) {
			ecomputc(s[i]);
			if(s[i]==CR) {
				while((c=comm_tgetc(1))!=ERR)	/*( wait a bit at eol )*/
					disp(c);
				disp_move(24,70);
				disp_printf("Line%5lu",++line);
				}
			}
		}
		while((c=comm_tgetc(1))!=ERR)	 /*( wait a bit at eol )*/
			disp(c);
	return TRUE;
}

int ascii_rxd(int fd)
{
	int c;
	int i=0;
	unsigned long line=0;
	char s[16];

	center("Press Esc when Done");
	while(TRUE) {
		if((c=b_poll())!=ERR) {
			if(c==ESC)
				break;
			comm_putc(c);
			}
		if((c=comm_poll())!=ERR) {
			disp(c);
			if(c==CR) {
				disp_move(24,70);
				disp_printf("Line%5lu",++line);
				}
			s[i++]=c;
			if(i==16)
				write(fd,s,16),i=0;
			}
		}
	if(i)
		write(fd,s,i);
	comm_timed_flush();
	return TRUE;
}

/********************************************************
* Area: Xmodem file transfers
* Description:
*********************************************************/
int xmodem_txd(int fd)
{
	unsigned char buffer[132];
	int block=0,errors=0;
	int c;
	int i,bytes;
	int worked=TRUE;

	center("Press Esc to abort");
	comm_timed_flush();

	/* wait to begin the transmit */
	while(TRUE) {
		c=comm_tgetc(10*TICKS_PER_SEC);
		if(escout() || (c==CAN && (c=comm_tgetc(10*TICKS_PER_SEC))==CAN)) {
			worked= ERR;
			goto exitout;
			}
		if(c==NAK)
			break;
#ifndef _DEBUG_
		disp('.');
#else
		if(c=='C')
			disp('C');
		else {
			sprintf(s,"[%d]",c);
			disps(s);
			}
#endif
		if(++errors >= 10) {
			worked=FALSE;
			goto exitout;
			}
		}

	/* transmit the blocks */
	c=ACK;
	while(TRUE) {
		if(escout())
		{
			worked=ERR;
			goto doneout;
		}

		// Build a new packet IFF prev packet was ACKed.
		// Pred(1st packet) ACKed by definition.
		//
		if(c==ACK)
		{
			errors   =  0;			// new packet has no errors
			++block;					// get next packet sequence number
			block   %=  256;

			buffer[0]=  SOH;		// write the synch byte then
			buffer[1]=  block;	// the packet number and its 1s complement
			buffer[2]= ~block;

			// Read next bytes from file, padding packet data field
			// to length with ASCII SUB (CP/M EOF) chars.
			//
			if((bytes=read(fd,&buffer[3],128))==0)
				break;
			if(bytes<128)			/* set extra bytes to ^Zs */
				memset(&buffer[3+bytes],26,128-bytes); // should be last block!

			// 8 bit chksum catches 95% of single bit errors. Corrects none.
			//
			buffer[131]=checksum(&buffer[3],128);
		}

		// (re)send the current block
		//
		for(i=0;i<132;++i)
			comm_putc(buffer[i]);

		/* make sure the other end received it OK */
		c=comm_tgetc(10*TICKS_PER_SEC);
		if(c==ACK)
			disp('T');
		else if(c==CAN && (c=comm_tgetc(TICKS_PER_SEC))==CAN) {
			worked=ERR;
			goto exitout;
			}
		else {
#ifndef _DEBUG_
			disp('.');
#else
			sprintf(s,"[%d]",c);
			disps(s);
#endif
			if(++errors>=10) {
				worked=FALSE;
				goto exitout;
				}
			}
	}

doneout:
	/* end the transmission */
	while(TRUE) {
		comm_putc(EOT);
		c=comm_tgetc(10*TICKS_PER_SEC);
		if(c==ACK)
			break;
		if(errors>=10) {
			worked=FALSE;
			goto exitout;
			}
		++errors;
		}

exitout:
	return worked;
}

/* abort_str is ^Xs and ^Cs with backspaces */
char *abort_str = "\030\030\030\030\030\010\010\010\010\010\
\003\003\003\003\003\010\010\010\010\010";

boolean xmodem_rxd(int fd)
{
	unsigned char *buffer;
	unsigned char block=1;
	int timeout;
	int c;
	int i,bytes;
	boolean worked=TRUE;

	/* begin the transmission */
	center("Press Esc to abort");
	buffer= malloc(1032);
	comm_timed_flush();
	comm_putc(NAK);

	/* get the blocks */
	while(TRUE) {
		while(TRUE) {
			c=comm_tgetc(10*TICKS_PER_SEC);

			if(c==SOH || c==STX) {
				*buffer= c;
				break;
				}
			if(c==EOT) {
				comm_putc(ACK);
				goto exitout;
				}
			if(c==CAN && (c=comm_tgetc(TICKS_PER_SEC))==CAN) {
				worked=ERR;
				goto exitout;
				}
			if(escout()) {
				worked=ERR;
				goto exitout;
				}

			disp('.');
			comm_timed_flush();
			comm_putc(NAK);
			}

		/* get the block (normal or 1K protocol) */
		bytes= (c==SOH ? 132 : 1028);
		for(i=1,timeout=FALSE;i<bytes && !timeout;++i) {
			c=comm_tgetc(TICKS_PER_SEC);
			if(c==ERR)
				timeout=TRUE;
			buffer[i]=c;
			}

		/* check if a valid block */
		if(!timeout && validblock(buffer,block,bytes)) {
			write(fd,&buffer[3],bytes-4);
			disp('R');
			++block;
			block %= 256;
			comm_putc(ACK);
			}
		else {
			disp('E');
			comm_timed_flush();
			comm_putc(NAK);
			}

		if(escout()) {
			worked=ERR;
			goto exitout;
			}

		}	/* end of the while loop */

exitout:
	if(worked==ERR) {
		comm_puts(abort_str);
		comm_timed_flush();
		}
	free(buffer);
	return worked;
}

boolean validblock(unsigned char *buffer,unsigned char block,int bytes)
{

	if(block != buffer[1] || buffer[1] != (unsigned char)~buffer[2])
		return FALSE;

	if(checksum(buffer+3,bytes-4)!=buffer[bytes-1])
		return FALSE;

	return TRUE;
}

void disps(char *s)
{
	while(*s)
		disp(*s++);
}

void scrollup( void )
{
	disp_scroll(1,COMROW1,0,COMROW2,79,7);
}

extern int disp_cursorrow,disp_cursorcol;

boolean getstring(char s[],int maxlen)
{
	int c;
	int n=0;
	int begrow,begcol;

	begrow= disp_cursorrow;
	begcol= disp_cursorcol;

	s[0]=NULL;
	while(TRUE) {
		disp_move(begrow,begcol);
		disp_printf("%s",s);
		disp_flush();
		disp_eeol();
		c=b_getch();
		if(c==CR || c==ESC)
			break;
		else if(c==BS && n>0)
			s[--n]=NULL;
		else if(n<maxlen && isprint(c)) {
			s[n++]=c;
			s[n]=NULL;
			}
		}
	return (c==CR);
}

void prompt(char *s)
{
	disp_move(24,0);
	disp_printf(s);			/* display the prompt */
	disp_flush();				/* set the cursor */
	disp_eeol();				/* clear the line */
}

void center(char *s)
{
	disp_move(24,0);
	disp_eeol();
	disp_move(24,(80-strlen(s))/2);
	disp_printf(s);
	disp_flush();
}

#define NUMBAUD 9
long baud_rate[]= { 300L,1200L,2400L,4800L,9600L,19200L,38400L,57600L,115200L };
int baud_divisor[]  = { B300,B1200,B2400,B4800,B9600,B19200,B38400,B57600,B115200 };

unsigned nextbaud(int baud)
{
	++baud;
	if(baud>=NUMBAUD)
		baud=0;
	return baud;
}

int baudnum(long lbaud)
{
	int i;

	if(lbaud == 115)
	  lbaud = 115200L;

	for(i=0;i<NUMBAUD;++i)
		if(lbaud == baud_rate[i])
			return i;
	return 1; // default 1200
}

int makebaud(int baud)
{
	return baud_divisor[baud];
}

#define NUMOPTS 4
char *cmd_opt[] = { "B", "COM", "IRQ", "H" };

void setopt(char *opt)
{
	int i,len;
	long atol();

	for(i=0;i<NUMOPTS;++i)
	{
		len=strlen(cmd_opt[i]);
		if(!strnicmp(opt,cmd_opt[i],len))
		{
			switch(i)
			{
			case 0:
				baud  = baudnum(atol(opt+len));
				ubaud = (unsigned)(baud_rate[baud]);
				break;
			case 1:
				port = atoi(opt+len);
				if((port <=0) || (port > 4))
					port = 0;
				else
					port -= 1;
				break;
			case 2:
				iIrqNumber = atoi(opt+len);
				if( ( iIrqNumber < 3 ) || ( iIrqNumber > 15 ) )
				{
					// irq number out of range
					//
					exit(0);
				}
				break;
#if USE_ZMODEM
			case 3:
				handshaking = atoi(opt+len);
				break;
#endif
			}
			break;
		}
	}
	for(i=0;i<NUMPARMS;++i)
	{
		len=strlen(parmname[i]);
		if(!strnicmp(opt,parmname[i],len))
		{
			parms = i;
			break;
		}
	}
}

void process_env( void )
{
	char *p;

	/* adjust for local time zone for ZModem transfers */
	tzset();

	if((p=getenv("COMM"))!=NULL) {
		if(*p!='/')
			p=strchr(p,'/');
		while(p) {
			setopt(p+1);
			p=strchr(p+1,'/');
			}
		}
}

void process_cmdline(int argc,char *argv[])
{
	int i;
	for(i=1;i<argc;++i)
	{
		if(*argv[i]=='/')
			setopt(argv[i]+1);
	}
}

boolean escout( void )
{
	int c;
	int test=FALSE;

	while((c=b_poll())!=ERR)
		if(c==ESC)
			test=TRUE;
	return test;
}

void filetransfer(int type)		/* 0-Upload, 1-Download */
{
	char typestr[10], s[80], file[70];
	int c;
	int  fd;
	long bytes;
	int  worked;
	unsigned blocks;
	unsigned long time;


	// Get the filename
	//
	strcpy(typestr, type ? "Download" : "Upload");
	*file = '\0';

	// find out how to transfer the file
	//
#if USE_ZMODEM
	sprintf(s,"%s %s ASCII, Xmodem or Zmodem (A, X or Z): ",typestr,file);
#else
	sprintf(s,"%s %s ASCII or Xmodem (A or X): ",typestr,file);
#endif
	prompt(s);
	do {
		c=b_getch();
		if(c==ESC)
		{
			status();
			return;
		}
		c=toupper(c);
#if USE_ZMODEM
	} while(c!='A' && c!='X' && c!='Z');
#else
	} while(c!='A' && c!='X' && c!='Z');
#endif
	disp_putc(c);

#if USE_ZMODEM
	if((c != 'Z') || (type == 0))
	{
		sprintf(s,"%s which file(s): ",typestr);
#else
	if(type == 0)
	{
		sprintf(s,"%s which file: ",typestr);
#endif
		prompt(s);
		if(!getstring(file,58))
		{
			status();
			return;
		}
		strupr(file);
	}

#if USE_ZMODEM
	if(c == 'Z')
	{ int c1;
		sprintf(s,"Enable crash recovery? [y/n]: ");
		prompt(s);
		do {
			c1=b_getch();
			if(c1==ESC)
			{
				status();
				return;
			}
			c1=toupper(c1);
		} while(c1!='Y' && c1!='N');
		disp_putc(c1);
		doZRecover = (c1 == 'Y');
		center("Press Esc to abort");
		if(type)
			ReceiveZMODEM();
		else
			SendZMODEM(file);
		status();
		return;
	}
#endif

	// open the file
	//
	if(type)
	{
		fd=_creat(file,0);
	}
	else
	{
		fd=_open(file,O_RDONLY|O_BINARY);
	}
	if(fd<=0)
	{
		sprintf(s," -- Failed to open %s!\r\n",file);
		disps(s);
		status();
		return;
	}


	// do the file transfer
	//
	sprintf(s,"\r\n%sing %s\r\n",typestr,file);
	disps(s);

	if(c=='A')
	{											/*** ASCII ***/
		if(type)
			worked=ascii_rxd(fd);
		else
			worked=ascii_txd(fd);
	}
	else if(c=='X')
	{											/*** XMODEM ***/
		if(type)
			worked=xmodem_rxd(fd);
		else
		{
			/* find number of blocks, etc.. */
			bytes=lseek(fd,0L,2);			/* seek to end */
			lseek(fd,0L,0);						/* seek back to beginning */
			blocks= (unsigned)((bytes-1L)/128L +1L);

			/* display some info on file */
			sprintf(s,"%ld bytes, %u block(s)\r\n",bytes,blocks);
			disps(s);

			/* time of transfer */
			time= (blocks*134L)/(baud_rate[baud]/10);
			sprintf(s,"transfer time: %lu min %2lu sec\r\n",time/60,time%60);
			disps(s);
			worked=xmodem_txd(fd);
		}
	}


	// check if things have worked
	//
	if(c!='A')
	{
		if(!worked)
			disps(" -- Failed\r\n");			/* Failed */
		else if(worked<0)
			disps(" -- Aborted\r\n");			/* Aborted */
		else
		{
				disps(" -- Done\r\n");			/* Worked */
		}
		beep();
		cl_twait(1);
		beep();
	}

	// close and exit
	//
	close(fd);
	status();
}

void hangup( void )
{
	cl_twait(2*TICKS_PER_SEC);
	ecomputs("+++");
	cl_twait(2*TICKS_PER_SEC);
	ecomputs("ATH0\r");
}

void dialup( void )
{
	static char number[32]= "1-360-487-9089";
	char s[50];

	sprintf(s,"Autodial [%s]: ",number);
	prompt(s);
	if(getstring(s,30)) {
		if(strlen(s))
			strcpy(number,s);
		strcpy(s,"ATDT");
		strcat(s,number);
		strcat(s,"\r");
		ecomputs(s);
		}
	status();
}

void status( void )
{
	// no more remote disk support
#if 0
	extern int packetmsg;
#endif

	sprintf(statusMsg,"%s v%s %s",szProgramName, REVISION, szHelpPrompt);
	disp_move(24,0);
	disp_printf(statusMsg);
	disp_eeol();

	if(rdisk) {
		disp_move(24,35);
		disp_printf("Rdisk: %c",rdisk-1+'A');
		}

	disp_move(24,47);
	disp_printf("Com%d:%6ld",port+1,baud_rate[baud]);
	disp_printf(" %s  ",parmname[parms]);
	disp_printf("Echo %s",echo ? "On " : "Off");
	disp_printf("  CR%s",crtrans ? "/LF" : "");
	reset_cursor();
}

#define SCRSIZE (80*24*2)
void help_screen( void )
{
	void *oldscr;
	int i;
	extern disp_base;

	oldscr = malloc(SCRSIZE);
	if( !oldscr )
	{
		disps( "Can't display help screen, memory too low.");
		return;
	}

	// save original screen
	_fmemcpy( MK_FP(_DS,oldscr), MK_FP(disp_base,0), SCRSIZE);

	/* display the whole help screen */
	for(i=0; i<24; ++i)
	{
		disp_move(i,0);
		disp_printf(helpstr[i]);
		disp_eeol();
	}
	b_getch();

	// restore original screen
	_fmemcpy( MK_FP(disp_base,0), MK_FP(_DS,oldscr), SCRSIZE );
	free(oldscr);
}

/*****************************************************************
  Check environment, then command line for comm_open() defaults

  Environment:
    COMM= {command line option(s)}

  Command Line Options:
	 /B#    Set Baud 300,1200,2400...19200
    /COM#  Set Port 1 or 2
	 /IRQ#  Set IRQ (3-15)
	 /N81   Set Data Bits
    /E71

******************************************************************/

#if USE_ZMODEM
int WantsAutoDownload(int c)
{ 	static int sawChar = 0;
	switch(sawChar)
	{
	case 0:
	case 1:
		sawChar = (c == '*') ? ++sawChar : 0;
		break;
	case 2:
		sawChar = (c == CAN) ? ++sawChar : 0;
		break;
	case 3:
		sawChar = (c == 'B') ? ++sawChar : 0;
		break;
	case 4:
		sawChar = (c == '0') ? ++sawChar : 0;
		break;
	case 5:
		sawChar = 0;
		if(c == '0')
			return TRUE;
	}
	return FALSE;
}
#endif

/* Is this an Alt-Key? */
boolean IsAltKey(int c)
{
	return (c >= 128 && c < 255);
}

int main(int argc,char *argv[])
{
	int c, c2;

	/* process the environments */
	process_env();
	process_cmdline(argc,argv);

	//  if no irq number supplied at the command line
	//  set the default irq for the specified port
	//
	if( iIrqNumber == 0 )
	{
		if( ( port == 0 ) || ( port == 2 ) )
		{
			iIrqNumber = 4;
		}
		else
		{
			iIrqNumber = 3;
		}
	}



	/* initialize the screen, etc... */
	disp_open();
	disp_move(0,0);
	disp_eeop();

	/* open the comm port */
#if USE_ZMODEM
	comm_open(baud_rate[baud], parmname[parms], port, iIrqNumber );
#else
	comm_open(ubaud, parmlist[parms], port, iIrqNumber,
			RXD_BUFF_SIZE, rxdBuffer);
#endif
	status();

	while(TRUE)
	{
		if((c=comm_poll())!=ERR)
		{
#if USE_ZMODEM
			if(WantsAutoDownload(c))
			{
  				center("Press Esc to abort auto-download");
 				ReceiveZMODEM();
				status();
				continue;
			}
#endif
			disp(c);
		}
		if((c2=b_poll())!=ERR) {
			switch(c2) {
				case F1:						/* help */
					help_screen();
					break;

// Remote disk operation no longer supported, just not ripped out.
//
#if 0
				case ALTR:					/* remote disk operation */
					remote_disk();
					status();
					break;
#endif

				case ALTB:					/* baud */
					baud=nextbaud(baud);
					status();
#if USE_ZMODEM
					comm_setup((long)baud_rate[baud], parmname[parms]);
#else
					comm_setup(makebaud(baud),parmlist[parms]);
#endif
					break;
				case ALTC:					/* clear screen */
					disps("\033[2J"); break;
				case ALTD:					/* dial */
					dialup();
					break;
				case ALTE:					/* echo on/off */
					echo= !echo;
					status();
					break;
				case ALTH:					/* hang up */
					hangup();
					break;
				case ALTK:
					send_brk();
					break;
				case ALTP:					/* parameters */
					parms += 2;
					if(parms>=NUMPARMS)
						parms=0;
#if USE_ZMODEM
					comm_setup((long)baud_rate[baud], parmname[parms]);
#else
					comm_setup(makebaud(baud),parmlist[parms]);
#endif
					status();
					break;
				case ALTT:					/* translate cr/lf */
					crtrans = !crtrans;
					status();
					break;
				case ALTX:					/* exit */
					disp_move(24,0); disp_eeol();
					disp_move(23,0); disp_flush();
					comm_close();
					exit(0);
				case PAGEUP:
					filetransfer(0);
					break;
				case PAGEDN:
					filetransfer(1);
					break;
				default:
					/*
						the getkey routines convert scan codes by adding 112.
						convert key back to DOS 2-key style for things like F5.
					*/
					if(IsAltKey(c2))
					{
						ecomputc(0);		/* DOS getkey says 0 followed by... */
						c2 -= 112;			/* actual scan code */
					}
					ecomputc(c2);
					break;
				}	/* end of switch */
			}

	// no more remote disk support
#if 0
		/* check for packet message */
		if(packetmsg && ((clock() - MsgTicks) > 2*TICKS_PER_SEC)) {
			packetmsg = 0;
			status();
			}
#endif

	}	/* end of while() */
}
