;
;   $Id: BITSETS.ASM 1.4 2002/10/16 21:45:08Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;
; BITSETS.ASM
;
; Copyright (c) 1989-2002 Datalight, Inc.
; All Rights Reserved.
; Datalight Confidential and Proprietary
;
; DEFRAG bit sets for FAT sector dirty flags
;
; NOTE:	This module uses 386 specific instructions.
; 	80186 equivalent operations would either take
;       8 times as much data or run much slower.
;
include ..\include\datalght.inc

InitModel compact, pascal, use16


; The worst case is (rounding up)...
;
; (65536 entries * 2 bytes/enty) / (512 bytes/sector)  = 256 sectors
; 256 sectors / 16 sectors/flag word = 16 flag words
;
FAT_SECTOR_FLAG_WORDS		= 	16	; (65536 / (512 * 16))

; There are 256 16-bit cluster entries per FAT sector. 256 = 2^8
;
INDEX_16_TO_SECTOR_SHIFT	=	8

; Sector = (3n/2)/512 = (n + (n shr 1)) / 512 = (n + (n shr 1)) shr 9
;
INDEX_12_TO_SECTOR_SHIFT	=	9

; ALWAYS
;
BYTES_PER_SECTOR		=	512

; 16  = 2^4:  (2^4 SHR 4) == 1
;
SECTOR_TO_FLAG_WORD_SHIFT	=	4

; 16 bits in a word, numbered 0...15d
;
SECTOR_TO_FLAG_MASK		=	000Fh

; This converts a 16-bit cluster number to a flag address.
;
; ->	ax	cluster number
; <-	ax	flag number (0..15d)
; 	bx	even aligned word index into the flag word array
;
; regs	bx, ax
;
CLUSTER_NUMBER_16_TO_FLAG	macro

	; Get the FAT sector containing the cluster, dup it in BX
	;
	shr	ax, INDEX_16_TO_SECTOR_SHIFT
	mov	bx, ax

	; Mask off the low nybble of the sector, this is the flag
	; number
	;
	and	ax, SECTOR_TO_FLAG_MASK

	; Convert the sector number into an index into the even
	; aligned flag word array
	;
	shr	bx, SECTOR_TO_FLAG_WORD_SHIFT
	shl	bx, 1
endm

; This converts a 12-bit cluster number to a flag address.
;
; ->	ax	cluster number
; <-	ax	flag number (0..15d)
; 	bx	even aligned word index into the flag word array
;       cx	(byte address of word holding 12 bit FAT element) mod BPS
;		if this value = (BPS -1) then element crosses a sector
;		boundary and when this element is made dirty, so are
;		TWO FAT SECTORS.
;
; regs	bx, ax
;
CLUSTER_NUMBER_12_TO_FLAG	macro

	mov	bx, ax
	shr	ax, 1
	add	ax, bx		; ax = byte offset of word containing entry
	mov	cx, ax
	and	cx, (BYTES_PER_SECTOR - 1)	; cx = bytes MOD bytes/sector

	shr	ax, INDEX_12_TO_SECTOR_SHIFT
	mov	bx, ax

	and	ax, SECTOR_TO_FLAG_MASK
	shr	bx, SECTOR_TO_FLAG_WORD_SHIFT
	shl	bx, 1
endm


.Data

EVENDATA

; The dirty flags.  There is one flag for each sector in the
; largest of all possible FATs. Flags are accessed as words.
;
; The dirty flag for a sector is set whenever a write occures to
; any of its allocated clusters.
;
; There are a maximum of 8 flag words in a 16 bit FAT.
;
	public dirtyFATSectorFlags
	dirtyFATSectorFlags	dw FAT_SECTOR_FLAG_WORDS dup(0)


.Code
	assume	cs:@curseg, ds:@data

.386


; This clears all the dirty bits for the FAT.
;
; ->	void
; <-	void
;
; regs  cx, bx, ax
;
COMMENT ~ ;prototype;
void PASCAL AllFATSectorsAreClean(void)
{}
~
ProcBegin	AllFATSectorsAreClean	public
ProcUses	ds
ProcSetup

	mov	ax, SEG dirtyFATSectorFlags
	mov	ds, ax

	xor	ax, ax
	mov	bx, ax
	mov	cx, FAT_SECTOR_FLAG_WORDS

    FSAC_loop:
	mov	dirtyFATSectorFlags[bx], ax
	inc	bx
	inc	bx
    	loop	FSAC_loop
ProcEnd


; This sets the dirty bit for a given FAT sector and returns its
; original value as a C boolean value in AX. The dirty flag for a
; sector is set whenever a write occures to any of its allocated
; clusters. In the case of a 12-bit fat entry that crosses a
; sector boundary, we return TRUE if either sector was dirty before.
;
; ->	clusterNumber: number of the cluster to mark dirty
;       useFAT16: TRUE if FAT has 16-bit entries, FALSE if 12-bit entries
; <-	0	if flag was set by this call
;	1	if flag was already set
;
; regs	dx, cx, bx, ax
;
COMMENT ~ ;prototype;
bool PASCAL FATEntryIsDirty(ushort clusterNumber, bool useFAT16)
{}
~
ProcBegin	FATEntryIsDirty, public
ProcUses	ds
ProcArgs	clusterNumber:word, useFAT16:word
ProcSetup

	mov	ax, SEG dirtyFATSectorFlags
	mov	ds, ax

	mov	ax, useFAT16
	or	ax, ax

	; get the FAT cluster number argument into AX
	;
	mov	ax, clusterNumber
	jz	FSIS_FAT12

	; compute the cluster's flag address
	;
	CLUSTER_NUMBER_16_TO_FLAG

	; Set the flag and get its previous value into the C flag
	;
	bts	dirtyFATSectorFlags[bx], ax

	; Put only the CF in AX, we will return 1=TRUE if the flag
	; was set before (meaning the sector was already dirty)
	;
	mov	ax, 0
	rcl	ax, 1
	jmp	short FSIS_done


	; 12-bit FATs will have two sectors become dirty when
	; the FAT entry is at offset 511d in the sector.
	;
	; In such cases we return TRUE if either sector was dirty
	; before.
	;
    FSIS_FAT12:
	CLUSTER_NUMBER_12_TO_FLAG		; get flag address

	bts	dirtyFATSectorFlags[bx], ax	; test & set first flag
	mov	dx, 0
	rcl	dx, 1

	cmp	cx, (BYTES_PER_SECTOR - 1)	; if not at offset 511d
	jnz	FSIS_FAT12_done			; then done

	mov	cx, 4				; compute next flag address
	inc	ax
	bt	ax, cx				; may be in next word
	adc	bx, 0
	and	ax, SECTOR_TO_FLAG_MASK

	bts	dirtyFATSectorFlags[bx], ax	; test & set next flag
	mov	ax, 0
	rcl	ax, 1
	or	dx, ax				; combine results

    FSIS_FAT12_done:
	mov	ax, dx				; return old flags in ax

    FSIS_done:
ProcEnd


; This tests the dirty bit of a given FAT sector and returns its
; value as a C boolean AX.
;
; The dirty flag for a sector is set whenever a write occures to
; any of its allocated clusters.
;
;
; ->	sectorInFAT: the index of the FAT sector to mark clean
; <-	0	if the sector was not marked dirty
;	1	if the sector is dirty
;
; regs	bx, ax
;
COMMENT ~ ;prototype;
bool PASCAL IsThisFATSectorDirty(ushort sectorInFAT)
{}
~
ProcBegin	IsThisFATSectorDirty, public
ProcUses	ds
ProcArgs	sectorInFAT:word
ProcSetup

	mov	ax, SEG dirtyFATSectorFlags
	mov	ds, ax

	; get the FAT sector argument into AX
	;
	mov	ax, sectorInFAT

	; compute the cluster's flag address
	;
	mov	bx, ax
	and	ax, SECTOR_TO_FLAG_MASK
	shr	bx, SECTOR_TO_FLAG_WORD_SHIFT
	shl	bx, 1

	; Test the flag, getting its previous value into the C flag
	;
	bt	dirtyFATSectorFlags[bx], ax

	; Put only the CF in AX, we will return 1=TRUE if the flag
	; was set before (meaning the sector was already dirty)
	;
	mov	ax, 0
	rcl	ax, 1
ProcEnd


; This clears the dirty bit of a given FAT sector and returns its
; value as a C boolean AX.
;
; The dirty flag for a sector is set whenever a write occures to
; any of its allocated clusters.
;
;
; ->	sectorInFAT: the index of the FAT sector to mark clean
; <-	1	if the sector was dirty
;	0	if the sector was not marked dirty before
;
; regs	bx, ax
;
COMMENT ~ ;prototype;
bool PASCAL ThisFATSectorIsClean(ushort sectorInFAT)
{}
~
ProcBegin	ThisFATSectorIsClean, public
ProcUses	ds
ProcArgs	sectorInFAT:word
ProcSetup

	mov	ax, SEG dirtyFATSectorFlags
	mov	ds, ax

	; get the FAT sector argument into AX
	;
	mov	ax, sectorInFAT

	; compute the cluster's flag address
	;
	mov	bx, ax
	and	ax, SECTOR_TO_FLAG_MASK
	shr	bx, SECTOR_TO_FLAG_WORD_SHIFT
	shl	bx, 1

	; Test the flag, getting its previous value into the C flag
	;
	btc	dirtyFATSectorFlags[bx], ax

	; Put only the CF in AX, we will return 1=TRUE if the flag
	; was set before (meaning the sector was already dirty)
	;
	mov	ax, 0
	rcl	ax, 1
ProcEnd


@curseg	ends
	end
