//
//   $Id: DEFCHKDK.C 1.4 2002/10/16 21:45:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// DEFCHKDK.C
// Disk checking module for Datalight disk DEFRAGmentation utility.
// Created 03-13-97 Dennis Edwards
//
// Copyright (c) 1989-2002 Datalight, Inc.
//	All Rights Reserved.
//	Datalight Confidential and Proprietary
//
//	This code is mostly just copied, verbose comments and all, from CHKDSK.C.
// We've added a little code to detect the number of discontinuous
// FAT chain entries and to compute a value for percent fragmentation
// (and to fit it into DEFRAG).
//
#define	__FILENUM__	6

#include <dos.h>
#include <string.h>
#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>

#include "dostruct.h"

#include "defrag.h"
#include "defmsg.h"		// contents may eventually be added to DOSMSG.C

#include "defmain.pt"
#include "defchkdk.pt"
#include "deferr.pt"
#include "defutila.pt"


// defmain.c
//
extern	uchar driveLetter;
extern	ushort huge *FAT, freeClusters, bytesToClusters,maxClusterIndex;
extern   ushort totalClusters, rootDIREntries, DIREntriesPerSector, DOSDrive;
extern 	ushort dataStart, rootDIRStart, sectorsPerCluster, bytesPerSector;

// cluster table values
//
#define FAT_ERR	0xF7		/* lowest value for an error flag */
#define FAT_BAD	0xF7		/* when FATs go bad (bad spot on disk) */
#define FAT_FREE	0xFA		/* unused clusters (100% FAT free) */
#define FAT_LOST	0xFC		/* a lost cluster (where's its mommy?) */
#define FAT_HEAD	0xFD		/* the head of a FAT chain */

// the statistics areas
//
#define DTOTAL  0
#define DHIDDEN 1
#define DDIR    2
#define DUSER   3
#define DBAD    4
#define DLOST   5
#define DAVAIL  6
#define MTOTAL  7
#define MAVAIL  8
#define MAXSTAT 9

#define NAMESIZE	66

char dirName[NAMESIZE];					// directory name if showing the path
unsigned char far *lpClusterTable;	// cluster table
void huge *lpRootDir;					// the whole root directory
int fat_changed=FALSE;	// where there ANY errors in the FAT?
int root_changed=FALSE;	// where any ROOT directory entries modified?
int show_path=FALSE;		// CONSTANT don't show paths
FCB fcb;			// for findfirst, etc

// Statistics information
//
struct stat
{
	ulong bytes;
	unsigned number;
};
struct stat stat[MAXSTAT];

char *stat_str[] =
{
	chk_totaldisk_str,
	chk_hidden_str,
	chk_dir_str,
	chk_user_str,
	chk_bad_str,
	chk_lost_str,
	chk_freedisk_str,
	chk_totalmem_str,
	chk_freemem_str
};

char *error_msg[] =
{
	chk_invalid_msg,
	chk_dir2file_msg,
	chk_baddir_msg,
	chk_alloc_msg
};

/* kludge: affects routines when showing crosslinked clusters */
int display_cross_linked=FALSE, errorsInDIRs;

// We use a formula that represents the percentage of discontinuities
// in an average length FAT allocation chain.
//
short percentFragmented;

// discontinuitiesAtStart holds the number of dicontinuities in
// all visible FAT allocation chains when the program started up.
// We use discontinuities to track the number of discontinuities left
// to fix. If (because of no adequate free block) we cannot defragment
// a given directory chain, discontinuities is still reduced by the
// number of discontinuities in that chain, if we reach zero discontinuities
// then we have done all we can do. dicontinuitiesFixed is the actual
// number of discontinuities that have been removed. The program
// will have really defragged the disk only if, when discontinuities is
// zero, discontinuitiesAtStart == discontinuitiesFixed
//
ulong discontinuities, discontinuitiesAtStart, discontinuitiesFixed;

// This routine (de)allocates memory for and then orchestrates running
// the included CHKDSK excerpts. This also computes the %fragmented value
// if no disk errors were discoverd. When we do discover errors we print
// the standard CHKDSK status message.
//
// <- TRUE if disk errors were detected and reported
//		FALSE if disk appears OK
//
bool PASCAL ReportedDiskErrors(void)
{ int lostClusterChains, crossLinkedChains,i;
  ushort totalChains, rootDIRSectors, averageChain;

	lpClusterTable		= DefragAlloc(maxClusterIndex+1,  1);
	lpRootDir 			= DefragAlloc(rootDIREntries, sizeof(DIR));
	if(!FAT || !lpClusterTable || !lpRootDir)
		FatalError(EL_ENOMEM);

	rootDIRSectors = dataStart - rootDIRStart;
	DefragAbsRead(DOSDrive, rootDIRSectors, rootDIRStart, (void *)lpRootDir);
	yield(0);

	ShowVolume(DOSDrive);
	yield(0);

	// Flush all buffers and reset the drive.
	// Recurse through the directory tree validating file size
	// and the FAT chain
	//
	CheckDIREntries(0,0);
	yield(0);

	// look for lost clusters
	//
	lostClusterChains = CheckFat();
	yield(0);

	// look for cross linked clusters
	//
	crossLinkedChains = FoundCrossLinks();
	yield(0);

	// reset the drive again and free our CHKDSK tables
	//
	ResetDrive(DOSDrive);
	DefragFarFree(lpClusterTable);
	DefragFarFree((void far*)lpRootDir);
	yield(0);

	// calculate available bytes on disk
	//
	printf("\n");
	++stat[DTOTAL].number;
	stat[DTOTAL].bytes = (ulong)bytesPerSector * (totalClusters) *
								sectorsPerCluster;
	++stat[DAVAIL].number;
	stat[DAVAIL].bytes = stat[DTOTAL].bytes;
	for(i=1; i<DAVAIL; ++i)
		stat[DAVAIL].bytes-= stat[i].bytes;

	// calculate free memory
	//
	stat[MTOTAL].bytes= (long)1024 * psp_memory();
	stat[MTOTAL].number = 1;
	stat[MAVAIL].bytes= stat[MTOTAL].bytes - (ulong)0x10 * _psp;
	stat[MAVAIL].number = 1;

	// display standard CHKDSK statistics
	//
	display_stats();
	yield(0);

	// If we found any errors the we will have described them already.
	// Tell user to run CHKDSK.
	//
	if(errorsInDIRs || lostClusterChains || crossLinkedChains )
	{
		// say we're helpless
		//
		printf(chk_nofix);
		return TRUE;
	}
	else
	{
		discontinuitiesFixed = 0;
		discontinuitiesAtStart = discontinuities;

		totalChains  = stat[DDIR].number+stat[DUSER].number;
		averageChain = (totalClusters - freeClusters) / totalChains;

		// measure the disk fragmentation, limit {0<=x<=100} and round to
		// nearest whole point. Return FALSE (did NOT report errors).
		//
		percentFragmented = MAX(0, (short)(((ulong)discontinuities * 1000L) /
			                                (ulong)(totalChains * averageChain)));

		percentFragmented = (percentFragmented % 10 >= 5)?
									percentFragmented / 10 + 1 :
									percentFragmented / 10;
		percentFragmented = MIN( percentFragmented, 100 );

		printf("%5u discontinuities in %5u visible chains (averaging %2u clusters each).\n\n",
				 (ushort)discontinuities,totalChains, averageChain);
		printf(percentfragmented_msg,percentFragmented);

		return FALSE;

	}
}

unsigned psp_memory(void)
{
	extern unsigned _psp;
	unsigned far * memtop = MK_FP(_psp,2);
	return *memtop / 0x40;
}

// This has been modified to display "allocation unit" info ala MS-DOS
//
void display_stats(void)
{
	int i;

	if(show_path)
		printf("\n");

	for(i=0;i<MAXSTAT;++i)
	{
		if(stat[i].number)
		{
			printf("%,13lu ",stat[i].bytes);
			printf(stat_str[i],stat[i].number);
			printf("\n");
		}
		if(i==DAVAIL)
		{
			printf("%,13u %s\n",1 << bytesToClusters,chk_bpc_str);
			printf("%,13u %s\n",totalClusters,chk_totalclusters_str);
			printf("%,13u %s\n\n",freeClusters,chk_freeclusters_str);
		}
	}
}

// error message numbers modified to fit in DEFRAG scheme
//
void PASCAL prerror(int msgNumber)
{
	printf("%s\n   %s\n",dirName,error_msg[msgNumber-100]);
}

// Show the date and time in an FCB
//
void show_time(unsigned time)
{
	unsigned hour, minute;
	char c;

	if(!time)
		return;

	hour = time >> 11;
	minute = (time >> 5) & 0x1F;
	if(hour >= 12)
		hour -= 12, c = 'p';
	else
		c = 'a';
	if(hour==0)
		hour=12;

	printf("%2d:%02d%c", hour, minute, c);
}
void show_date(unsigned date)
{
	unsigned year, month, day;

	if(!date)
		return;

	year = (date >> 9) + 1980;
	month = (date >> 5) & 0xF;
	day  = date & 0x1F;
	printf("%2d-%02d-%02d", month, day, year%100);
}


//	Simply used to get the volume label.  Returns TRUE
//	if worked, FALSE if failed.
//
int dos1_findfirst(int drive, char *name, int attr)
{
	extern FCB fcb;

	/* these MUST be static because SS!=DS in Compact model */
	static char blk[7+sizeof(FCB)];
	static char dta[7+sizeof(FCB)];

	FCB * fcbptr;
	uchar al;

	blk[0] = 0xFF;
	blk[6] = attr;
	fcbptr = (void *)(blk+7);

	strcpy((char*)(fcbptr->name), name);
	fcbptr->drive = drive+1;

	/* set the DTA to our dta area */
	SetDTA(dta);

	/* use DOS1 find first to get volume label */
	al = bdos(0x11,(unsigned)blk,0);
	if(al)
		return FALSE;

	/* copy the results into the global FCB */
	memmove(&fcb, dta+7, 12);				/* drive and name */
	fcb.date = *(unsigned *)(dta+32);
	fcb.time = *(unsigned *)(dta+30);

	return TRUE;
}


//	Get and display the volume label
//
void ShowVolume(int drive)
{
	// determine/display the volume label
	//
	if(dos1_findfirst(drive, "???????????", FA_LABEL))
	{
		printf(chk_vol_str,fcb.name);
		show_date(fcb.date);
		printf("  ");
		show_time(fcb.time);
		printf("\n\n");
	}
}

//	Follow the FAT allocation chain for this file or directory
// and sets the clusters variable to the number of clusters in in
//	the FAT chain
//
//	Returns a CHKDSK error code (0=OK)
//
#pragma argsused
int PASCAL follow_chain(unsigned start, unsigned * clusters, bool inSubDir)
{
	unsigned clus;

	// no clusters in chain to begin with
	//
	*clusters = 0;
	if(!start)
		return 0;

	// the start of this file should not be out of bounds
	//
	if(start < 2 || start > maxClusterIndex)
		return ERR_INVALID;

	// walk the entire FAT chain
	//
	while(!IS_FAT_EOF(start))
	{
		// count this cluster
		//
		++(*clusters);

		// Mark this cluster as used, check for cross-links.
		//
		if(display_cross_linked)
		{
			// A cross-linked cluster will have been referenced more than once.
			//
			if((lpClusterTable[start] > 1) && (lpClusterTable[start] < FAT_ERR))
			{
				printf(cross_linked,dirName,start);
				break;
			}
		}
		else if(lpClusterTable[start] < FAT_ERR)
		{
			if(++lpClusterTable[start] > 1)
			{
				--(*clusters);
				break;
			}
		}

		// Read our copy of the FAT to get the link to the next cluster
		// in this allocation chain.
		//
		clus = ReadFATInRAM(start);

		// Validate the cluster link
		//
		if( !IS_FAT_EOF(clus) && !display_cross_linked)
		{
			if((clus < 2) || (clus > maxClusterIndex))
			{
				return ERR_INVALID;
			}

			// determine how badly the disk is fragmented
			//
			if(!IS_FAT_EOF(clus) && (clus != start+1))
			{
				++discontinuities;
			}

			if(lpClusterTable[clus] > 1)
			{
				++lpClusterTable[clus];
				break;
			}

			// Move to the next cluster in the chain
			//
		}
		start = clus;
	}

	return 0;
}


//	Determine the size (in sectors) of this subdirectory.
//	This function will follow the FAT chain, fixing it if
//	there are BAD clusters in the chain.
//
//	Returns 0 on success, ERR code on failure.  Also sets
//	*sectors on success.
//
int PASCAL sizeof_dir(unsigned start, unsigned *sectors)
{
	int err = FALSE;	/* was there an error? */

	*sectors = 0;

	// the subdir must have a valid starting cluster
	//
	if(start < 2 || start > maxClusterIndex)
		return ERR_INVALID;

	// follow the cluster chain and count the number of clusters
	//
	err = follow_chain(start, sectors, TRUE);
	if(!err)
		*sectors *= sectorsPerCluster;

	return err;
}

//	Assumes the FAT chain is terminated correctly.
//
void PASCAL read_subdir(unsigned start, void * dir)
{
	char huge * buff = dir;
	while(start && !IS_FAT_EOF(start))
	{
		DefragAbsRead(DOSDrive, sectorsPerCluster,
				  			dataStart+(long)(start-2)*sectorsPerCluster,
				  			(void far*)(buff));
		buff += (long)sectorsPerCluster * bytesPerSector;
		start = ReadFATInRAM(start);
	}
}

// converts an ename to a dotname and prepares it for appending
// to a path string
//
void PASCAL make_name(char *name, char *ename)
{
	uchar len;
	uchar end = 8;
	char * dotname = name;

	/* place the name part into the dotname string */
again:
	for(len = end; len > 0 && ename[len-1] == ' '; --len)
		/* just count */;
	if(len)
	{
		memmove(dotname,ename,len);
		dotname += len;
	}
	if(end == 8)
	{
		ename += 8;
		if(*ename != ' ')
			*dotname++ = '.';
		end = 3;
		goto again;
	}

	*dotname = 0;

	/* do NOT allow backslashes in the name */
	dotname = name;
	while(*dotname)
	{
		if(*dotname == '\\')
			*dotname = '/';
		++dotname;
	}
}

// add/remove the last 8.3 name from a path string
//
void PASCAL add_name(char *ename)
{
	char name[14];
	make_name(name, ename);
	if(dirName[strlen(dirName)-1] != '\\')
		strcat(dirName,"\\");
	strcat(dirName,name);
}
void PASCAL remove_name(void)
{
	char * p;
	p = strrchr(dirName,'\\');
	if(p)
	{
		*p = 0;
		if(!strchr(dirName,'\\'))
			strcat(dirName,"\\");
	}
}

//	Traverse the directory tree, checking all directory
//	entries.  This will make sure the filesize agrees
//	with the FAT chain, and that there are not invalid
//	clusters in the chain.
//
//	This function is recursive.
//
//	"subdir" is the cluster # of this subdirectory "." (0=root).
//	"parent" is the cluster # of the parent directory ".."
//
//	Returns CHKDSK error code if there was a problem, otherwise 0.
//
int PASCAL CheckDIREntries(unsigned subdir, unsigned parent)
{
	DIR huge * lpDirEntry;
	void huge * lpDir;
	int i;
	unsigned sectors;		/* how many sectors in this directory? */
	int bin;
	int err = FALSE;
	unsigned clus_in_file;		/* clusters according to FAT chain */
	ulong clus_in_entry;			/* clusters according to dir entry */


	// the root directory has already been read in
	//
	yield(0);
	if(subdir == 0)
	{
		strcpy(dirName,"A:\\");
		*dirName  = driveLetter;
		sectors   = dataStart - rootDIRStart;
		lpDir 	 = lpRootDir;
	}

	// processing a subdirectory, read it in
	//
	else
	{
		// allocate the memory for the subdirectory
		//
		err = sizeof_dir(subdir, &sectors);
		if(err)
	  		errorsInDIRs = err;
		if(display_cross_linked)
			err = 0;
		lpDir = (void huge *)DefragAlloc(sectors, bytesPerSector);
		if(!lpDir)
			FatalError(EL_ENOMEM);

		// read in the contents of the directory
		//
		read_subdir(subdir, (void far *)lpDir);
	}
	yield(0);

	// make sure the subdirectory looks valid
	//
	lpDirEntry = (DIR huge *)lpDir;
	if(subdir)
	{
		/* was the FAT chain for this subdir truncated? */
		if(err)
		{
	  		errorsInDIRs = err;
			prerror(ERR_INVALID);
			err = 0;
		}

		// the "." and ".." directories must be valid
		//
		if((strncmp(lpDirEntry->name,    ".       ",8) != 0) ||
		   (strncmp((lpDirEntry+1)->name,"..      ",8) != 0))
		{
			if(display_cross_linked)
			{
				err = 0;
				goto errout;
			}
	  		err = (errorsInDIRs = ERR_BADDIR);		// not repairable
			prerror(ERR_BADDIR);
			goto errout;
		}
		if(lpDirEntry->start != subdir)
		{
			if(display_cross_linked)
			{
				err = 0;
				goto errout;
			}
	  		err = (errorsInDIRs = ERR_BADDIR);		// repairable
			prerror(ERR_BADDIR);
			err = 0;
		}
		if((lpDirEntry+1)->start != parent)
		{
			if(display_cross_linked)
			{
				err = 0;
				goto errout;
			}
	  		err = (errorsInDIRs = ERR_BADDIR);		// repairable
			prerror(ERR_BADDIR);
			err = 0;
		}

		// skip the first 2 entries
		//
		lpDirEntry += 2;
	}
	yield(0);

	// stop at the first 0 entry
	//
	for(i = subdir ? 2 : 0; *lpDirEntry->name &&
								   i<sectors*DIREntriesPerSector;  ++i, ++lpDirEntry)
	{
		yield(0);
		// ignore deleted entries
		//
		if(*(uchar *)lpDirEntry->name == DELETED)
			continue;

		// long filename entries are fatal problems
		//
		if(lpDirEntry->attr == A_LONGNAME)
		{
// !!!
//
// must change this to support LFNs
//
			FatalError(EL_LONG_NAMES);
		}
		// make a normal name from the expanded one
		//
		add_name(lpDirEntry->name);

		// directory entry
		//
		if(lpDirEntry->attr & A_SUBDIR)
		{
			if(display_cross_linked)
				CheckDIREntries(lpDirEntry->start, subdir);

			else
			{
				// if invalid first
				//
				if(lpDirEntry->start < 2 ||
					lpDirEntry->start > maxClusterIndex ||
					!ReadFATInRAM(lpDirEntry->start))
				{
					err = ERR_DIR2FILE;
	  				errorsInDIRs = err;
					prerror(ERR_DIR2FILE);

				}

				// recurse into subdirectory, if error, convert dir into a file
				//
				else
				{
					err = CheckDIREntries(lpDirEntry->start, subdir);
					if(err)
	  					errorsInDIRs = err;
				}
				err = 0;
			}
			yield(0);
		}


		// user or hidden files
		//
		else
		{
			if(display_cross_linked)
				follow_chain(lpDirEntry->start, &clus_in_file, FALSE);

			else
			{
				// display the file if requested
				//
				if(show_path)
					printf("      %s\n",dirName);

				bin = (lpDirEntry->attr & A_HIDDEN) ? DHIDDEN : DUSER;
				err = follow_chain(lpDirEntry->start, &clus_in_file, FALSE);
				if(err)
				{
	  				errorsInDIRs = err;
					prerror(ERR_INVALID);
					err = 0;
				}

				// check if file size disagrees with FAT chain length
				//
				if(lpDirEntry->size)
					clus_in_entry= (lpDirEntry->size-1)/
										((ulong)sectorsPerCluster * bytesPerSector)+1;
				else
					clus_in_entry= 0L;
				if(clus_in_entry != clus_in_file)
				{
					err = ERR_ALLOC;
	  				errorsInDIRs = err;
					prerror(ERR_ALLOC);
					lpDirEntry->size = (ulong)clus_in_file * sectorsPerCluster *
											  bytesPerSector;
				}

				// update the statistics
				//
				if((lpDirEntry->attr & 0x0F) != A_VOLUME)
				{
					stat[bin].bytes += (ulong)clus_in_file * sectorsPerCluster *
 											  bytesPerSector;
					++stat[bin].number;
				}
			}
		}
		yield(0);

		// remove the name
		//
		remove_name();
	}

	// clean up and error out
	//
errout:

	// if the directory was changed, save it
	//
	yield(0);
	if(subdir)
	{
		DefragFarFree((void far *)lpDir);

		if(sectors && !display_cross_linked)
		{
			stat[DDIR].bytes += (ulong)sectors * bytesPerSector;
			++stat[DDIR].number;
		}
	}
	return err;
}

/*
	This will account for all clusters - the only ones in
	the lpClusterTable[] that wont be marked will be lost clusters.
*/
void PASCAL lock_out_bad_entries(void)
{
	unsigned i;
	unsigned val;

	// check through the whole table
	//
	for(i=2; i<=maxClusterIndex; ++i)
	{
		if(!lpClusterTable[i])
		{
			val = ReadFATInRAM(i);

			// if the cluster is free, mark it as such
			//
			if(val == 0)
				lpClusterTable[i] = FAT_FREE;

			// this marks an unusable portion
			//
			else if(val == BAD_CLUSTER)
			{
				lpClusterTable[i] = FAT_BAD;
				stat[DBAD].bytes += sectorsPerCluster * bytesPerSector;
				++stat[DBAD].number;
			}
		}
	}
}

//	Loop through the FAT looking for lost chains.
// A lost chain is a linked set of FAT entries that is
// not referenced by a directory entry (possibly because
// a FAT link value was destroyed or possibly because a
// DIR entry was wrong).
//
unsigned PASCAL count_lost_chains(unsigned * lost)
{
	unsigned chains = 0;
	unsigned clus, oldclus;
	unsigned i;

	*lost = 0;
	for(i=2; i<=maxClusterIndex; ++i)
	{
		/* find the next unclaimed entry */
		if(lpClusterTable[i] == 0)
		{
			/* call this a head */
			lpClusterTable[i] = FAT_HEAD;
			++chains;

			oldclus = i;
			clus = ReadFATInRAM(i);

			/* follow that chain! */
			while(clus && !IS_FAT_EOF(clus))
			{
				/* chain points into a head, attach and we are done */
				if(lpClusterTable[clus] == FAT_HEAD)
				{
					lpClusterTable[clus] = FAT_LOST;
					--chains;
					break;
				}

				/* chain points into the middle of another chain, truncate */
				else if(lpClusterTable[clus])
				{
					WriteFATInRAM(oldclus, FAT_EOF);
					break;
				}

				/* add this cluster to the chain */
				else
				{
					lpClusterTable[clus] = FAT_LOST;
					++(*lost);
				}

				oldclus = clus;
				clus = ReadFATInRAM(clus);
			}

			/* record that it's lost */
			++(*lost);
		}
		yield(0);
	}
	return chains;
}

//	Check the in-memory FAT for lost clusters.
//
int PASCAL CheckFat(void)
{
	unsigned chains, lost;

	// lock out bad clusters
	//
	lock_out_bad_entries();

	// caclulate the # of chains
	//
	chains = count_lost_chains(&lost);
	if(chains)
		printf(lost_clusters,lost,chains);

	return chains;
}


//	Are there any cross linked clusters?  If yes, display
//	them on the screen.
//
int PASCAL FoundCrossLinks(void)
{
  int is_cross_linked = FALSE;
  unsigned i;

	// are there ANY cross linked clusters?
	//
	for(i=2; i<=maxClusterIndex && !cross_linked; ++i)
		if(lpClusterTable[i] > 1 && lpClusterTable[i] < FAT_ERR)
			is_cross_linked = TRUE;

	if(is_cross_linked)
	{
		// display all cross linked clusters
		//
		display_cross_linked = TRUE;
		CheckDIREntries(0, 0);
		display_cross_linked = FALSE;

		printf("\n");
	}
	return is_cross_linked;
}

void PASCAL ResetDrive(int drive)
{
	union REGS r;

#if 0

	//
	// DOS v7 support
	//

   /* LOCK logical drive */
	r.x.ax = 0x440D;
	r.x.cx = 0x084A;
	r.x.dx = 0x0004;
	r.x.bx = drive+1;
	intdos(&r,&r);

   /* RESET logical drive */
	r.x.ax = 0x710D;
	r.x.cx = 0x0001;					/* flush buffers and cache and reset drive */
	r.x.dx = drive;
	intdos(&r,&r);

   /* UNLOCK logical drive */
	r.x.ax = 0x440D;
	r.x.cx = 0x086A;
	r.x.bx = drive+1;
	intdos(&r,&r);
#endif

	/* reset disk */
	r.h.ah = 0x0D;
	intdos(&r,&r);

	/* get new DPB for disk - resets internal buffers */
	r.h.ah = 0x32;
	r.x.dx = drive+1;
	intdos(&r,&r);

}
