//
//   $Id: DEFDIRS.C 1.4 2002/10/16 21:45:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// DEFDIRS.C
// Directory optimizing module for Datalight disk DEFRAGmentation utility.
// Created 03-13-97 Dennis Edwards
//
// Copyright (c) 1989-2002 Datalight, Inc.
//	All Rights Reserved.
//	Datalight Confidential and Proprietary
//

// !!!
//
// Note:	we use the absread/write file I/O buffer for storing DIR entries.
//			The max number of std. sectors in that buffer is 65536u / 512 = 128
//			This means that the max dir entries we can defrag is
// 		128*512/32 = 2048. This is the same limit as in ROM-DOS' COMMAND.COM.
//
#define	__FILENUM__	5

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <setjmp.h>
#include <alloc.h>
#include <errno.h>
#include <dos.h>

#include "dostruct.h"

#include "defrag.h"
#include "defmsg.h"

#include "defmain.pt"
#include "defdirs.pt"
#include "deferr.pt"
#include "defutil.pt"
#include "defutila.pt"
#include "defpaths.pt"

// defmain.c -- globals reflecting properties of the disk to defrag
// and the user specified sort options
//
extern ushort huge *FAT;         	// our in memory copy of the FAT
extern ushort huge *BackwardFAT;	  	// out free last and FAT chain back-links
extern ushort firstFree;				// the head of the BackwardFAT free list
extern ushort rootDIRStart;			// starting sector of root dir
extern ushort sectorsToClusters;		// shift count sectors <-> clusters
extern ushort sectorsPerCluster;
extern ushort bytesToSectors;			// shift count bytes <-> sectors
extern ushort partialSectorMask;		// AND mask to detect partial sectors
extern ushort partialClusterSectorsMask; // AND mask to detect partial clusters
extern ushort dataStart;					// starting sector of cluster FAT[2]
extern ushort DIREntriesToSectors;  // shift count sectors <-> DIR entries
extern ushort DIREntriesPerSector;
extern LPDPB lpDPB;						// far ptr to DPB for drive

// defopts.c
extern char  sortOpts[];				// user specified sort options string
extern ushort DOSDrive;					// the drive to work on and
extern char driveLetter;				// its name

// defragf.c -- the file I/O buffer and its size in sectors, dir entries
// and bytes
//
extern uchar far *lpSectorBuffer;
extern ushort bufferSectors, bufferDirEntries;
extern ulong bufferSize;

// defchkdk.c
extern ulong discontinuities;		  		// total remaining in all FAT chains
extern ulong discontinuitiesAtStart;	// total in all FAT chains at start
extern ulong discontinuitiesFixed;		// total resolved

// deferr.c
//
extern ushort abortingBecause;		// ERRORLEVEL (see DEFERR.C/DEFMSG.H)


// The jmp_buf used to, as they say in Prolog, FAIL a recursive
// subroutine call back to a safe error handler. If yield is called
// with a non-null jmp_buf argument then yield() will longjmp() to that
// error handler if a fatal error has occured since the last yield() call.
// If a critical error has occurred and yield(0) is called then yield()
// will not call longjmp() but will call FatalExit() instead.
//
//	ONLY CALL yield(0) WHEN THE PROGRAM CAN SAFELY BE ABORTED!
//
// yield() could be expanded to save the program state and thus facilitate
// the time-slicing neccessary for a background DEFRAG to work.
//
jmp_buf dirJmpBuf;


// The place holder file name (see OptimizeDirectories())
//
char throwAwayName[] = "DEFRAG00";
char pathName[MAXPATH];

// The number of dir trees we could not optimize
//
ulong dirsSkipped;

// We need to allocate space on the drive for a defragged parent sub-dir
// before we can recursively defrag the children of the dir. We push each
// of the original starting clusters of any subdir that is relocated so
// that we can free the original dir clusters when the entire dir tree
// has been defragged.
//
ushort dirStartingClusterStack[MAX_NODES_IN_SUBDIR], dirStackSP = 0;

// saves the starting cluster of a sub-dir FAT chain to free
//
bool PushDirFATChainToFree(ushort originalStartingClusterOfSubDir)
{
	if(dirStackSP < MAX_NODES_IN_SUBDIR)
	{
		dirStartingClusterStack[dirStackSP++] = originalStartingClusterOfSubDir;
		return TRUE;
	}
	abortingBecause = EL_TOO_MANY_NODES;
	return FALSE;
}

// retrieves the starting cluster of a sub-dir FAT chain to free.
//
// <- 0 	if all sub-dir FAT chains have been returned
//		>0 the starting cluster of the next sub-dir FAT chain to free
//
ushort PopDirFATChainToFree(void)
{ ushort value;

	if(dirStackSP)
		value = dirStartingClusterStack[--dirStackSP];
	else
		value = 0;

	return value;
}

// Rewinds the dir specified by the lpDir struct
//
void RewindThisDir(struct ReadDirStruct far *lpDir)
{
	lpDir->currentCluster = lpDir->startingCluster;
	lpDir->sectorsRead    = 0;
	if(lpDir->startingCluster == 0)
	{
		lpDir->sectorsToRead = (ulong)(dataStart) - (ulong)(rootDIRStart);
		lpDir->currentSector = (ulong)(rootDIRStart);
		lpDir->clustersToRead = (ushort)(lpDir->sectorsToRead
											>> sectorsToClusters);
		if((lpDir->sectorsToRead & partialClusterSectorsMask) != 0)
			++lpDir->clustersToRead;
	}
	else
	{
		lpDir->currentSector  = (ulong)(lpDir->currentCluster -2)
											<< sectorsToClusters;
		lpDir->currentSector += dataStart;
		lpDir->sectorsToRead  = (ulong)(lpDir->clustersToRead)
											<< sectorsToClusters;
	}
}


// Sets things up to read a (sub)dir as though it were a file (sort of).
// Does not actually read anything
//
// ->	startingCluster: number of the subdir or 0 if dir is the root dir
//		lpDir: a far ptr to a ReadDirStruct to initialize (see defrag.h)
//    dirFileCluster: if opening the DIR file grew the subdir size, then this
//						will be cluster that does not have valid allocation
//						information in our copy of the FAT and BackwardFAT
//				      (it won't be in the sub-dir's FAT chain).
// <-	TRUE if dir is ready to be read
// 	FALSE if there is an error in the FAT
//    or the DIR is too big to be held entirely in RAM
//
bool OpenThisDir(ushort startingCluster, 				// of original directory
					  struct ReadDirStruct far *lpDir,	// a ptr to the dir data
					  ushort dirFileCluster)
{ ushort clusterID, clusterNumber, newClusterNumber;

	ASSERT(FAT != 0L);

	// if not the root dir, verify valid FAT chain and count clusters in chain
	//
	if(startingCluster != 0)
	{
		// initialize some vars
		//
		lpDir->clustersToRead = 1;
		lpDir->discontinuities = 0;
		clusterID = ReadFATInRAM(startingCluster);
		clusterNumber = startingCluster;
		if(clusterNumber == dirFileCluster)
			dirFileCluster = 0;

		// while more entries in the FAT chain...
		//
		while(!IS_FAT_EOF(clusterID))
		{
			SHIPASSERT((clusterID != FREE_CLUSTER) && (clusterID != BAD_CLUSTER));

			// if dirFileCluster is in the FAT chain then we didn't grow
			// the sub-dir with our new DIR file entry. Get the next
			// member of the FAT chain
			//
			clusterNumber = clusterID;
			if(clusterNumber == dirFileCluster)
				dirFileCluster = 0;
			clusterID = ReadFATInRAM(clusterNumber);

			// count clusters and discontinuities
			//
			++lpDir->clustersToRead;
			if(!IS_FAT_EOF(clusterID) && (clusterID != (clusterNumber + 1)))
				++lpDir->discontinuities;
		}
		if(dirFileCluster != 0)
		{
			// We grew the sub-dir.
			// Make sure that we know what slot it grew into...
			//
			newClusterNumber = GetLeastFree();
			if(!((newClusterNumber >= 2) && (newClusterNumber <= 0xFFEF)))
			{
			}

#if DEBUG
			if(newClusterNumber != dirFileCluster)
			{
				printf("Allocated wrong dir cluster: wanted %u\t got %u\n",
						 dirFileCluster, newClusterNumber);
			}
#endif
			SHIPASSERT(newClusterNumber == dirFileCluster);

			// update our data structures and count the new cluster
			//
			WriteFATInRAM(clusterNumber, newClusterNumber);
			WriteFATInRAM(newClusterNumber, FAT_EOF);
			BackwardFAT[dirFileCluster] = clusterNumber;
			++lpDir->clustersToRead;
		}
	}
	lpDir->startingCluster = startingCluster;
	RewindThisDir(lpDir);
	return TRUE;
}


// We want to read as many sectors worth of directory information as we
// possibly can at any one time. There are several cases we must consider
// 1) The dir being read is the root dir in which we just read up to
//		sectorsLeftInBuffer per call until we have read the whole thing.
// 2) The dir is a subdir
//    a) SPC <= sectorsLeftInBuffer when we read as many clusters as will
//			fit in the buffer. If we have adjacent FAT sectors we will try to
//			read all of them in the same call.
//    b) SPC > sectorsLeftInBuffer
//
// ->	lpDir:	a far * to the ReadDirStruct previously opened for this
//					directory
// 	lpDirBuffer:	the buffer to hold the directory contents
//	   sectorsLeftInBuffer: the number of sectors available in the buffer
//
// <-	>0: the number of sectors read this call.
//     0: the entire dir was already read successfully (harmlessly redundant).
//		<0: then a disk I/O error occured.
//
short ReadDirSectors(struct ReadDirStruct far *lpDir,
							LPDIR lpDirBuffer,
							ushort sectorsLeftInBuffer)
{ ushort sectorsReadThisCall, nextClusterToRead, clusterToReadNow;
  ushort  sectorsToReadNow, clustersToReadNow;

	ASSERT(lpDir != 0L);
	ASSERT(lpDirBuffer != 0L);
	ASSERT(sectorsLeftInBuffer <= bufferSectors);
	ASSERT(sectorsLeftInBuffer != 0);

	sectorsReadThisCall = 0;
	if(lpDir->startingCluster == 0)		// reading the ROOT DIR
	{	if(lpDir->sectorsRead < lpDir->sectorsToRead)
		{	 sectorsToReadNow = (ushort)(lpDir->sectorsToRead -
												  lpDir->sectorsRead);
			 if(sectorsToReadNow > sectorsLeftInBuffer)
				 sectorsToReadNow = sectorsLeftInBuffer;
			 if(!DefragAbsRead(DOSDrive,sectorsToReadNow,
			 						lpDir->currentSector,
									(void far*)(lpDirBuffer)))
				 return -1;

			 lpDir->sectorsRead   += sectorsToReadNow;
			 lpDir->currentSector += sectorsToReadNow;
			 sectorsReadThisCall  += sectorsToReadNow;
		}
	}
	else		// reading a subdir
	{
		// if more dir entries to read
		//
		if(lpDir->sectorsRead < lpDir->sectorsToRead)
		{
			// if at least 1 cluster/buffer
			//
			if(sectorsPerCluster <= sectorsLeftInBuffer)
			{	do
				{
					clustersToReadNow = 0;
					sectorsToReadNow  = 0;
					nextClusterToRead = lpDir->currentCluster;
					lpDir->currentSector  = (ulong)(lpDir->currentCluster -2)
														<< sectorsToClusters;
					lpDir->currentSector += dataStart;

					// while more clusters in chain, room in buffer for another
					// cluster and the next cluster is subsequent to the current
					// one
					//
					do
					{	clustersToReadNow++;
						sectorsToReadNow 		+= sectorsPerCluster;
						clusterToReadNow      = nextClusterToRead;
						nextClusterToRead 	 = ReadFATInRAM(clusterToReadNow);
					}
					while((!IS_FAT_EOF(nextClusterToRead)) &&
					   ((lpDir->sectorsRead + sectorsToReadNow + sectorsPerCluster)
									< sectorsLeftInBuffer) &&
					      ((clusterToReadNow + 1) == nextClusterToRead));

					if(!DefragAbsRead(DOSDrive,sectorsToReadNow,
									  lpDir->currentSector,(void far*)(lpDirBuffer)))
						return -1;

					lpDir->currentCluster = nextClusterToRead;
					lpDir->currentSector  = ((ulong)(nextClusterToRead-2)
													<< sectorsToClusters)
													+ dataStart;
					lpDir->sectorsRead   += sectorsToReadNow;
					sectorsReadThisCall  += sectorsToReadNow;
					lpDirBuffer          += sectorsToReadNow << DIREntriesToSectors;
				}
				while((lpDir->sectorsRead < lpDir->sectorsToRead) &&
						(sectorsReadThisCall < sectorsLeftInBuffer));

				return sectorsReadThisCall;
			}
			else	// clusters are more than 64K, have multiple buffers/cluster
			{
				sectorsToReadNow = 0;

				// if in the middle of a cluster
				//
				if((lpDir->sectorsRead & partialClusterSectorsMask) > 0)
				{
					// get sectors left to read in cluster
					//
					sectorsToReadNow = sectorsPerCluster -
							(ushort)(lpDir->sectorsRead & partialClusterSectorsMask);

					// limit read to size of buffer
					//
					if(sectorsToReadNow > sectorsLeftInBuffer)
						sectorsToReadNow = sectorsLeftInBuffer;

					// does this finish off the cluster? get next if so
					//
					if(((sectorsToReadNow+lpDir->sectorsRead) &
							partialClusterSectorsMask) == 0)
					{
						lpDir->currentCluster = ReadFATInRAM(lpDir->currentCluster);
					}
				}
				else	// starting a new cluster
				{
					sectorsToReadNow = sectorsLeftInBuffer;
				}
				if(!DefragAbsRead(DOSDrive,sectorsToReadNow,
									lpDir->currentSector,(void far*)(lpDirBuffer)))
					return -1;

				if(((sectorsToReadNow+lpDir->sectorsRead) &
						partialClusterSectorsMask) == 0)
				{
					lpDir->currentSector = ((ulong)(lpDir->currentCluster-2)
													<< sectorsToClusters)
													+ dataStart;
				}
				else
				{
					lpDir->currentSector += sectorsToReadNow;
				}
				lpDir->sectorsRead   += sectorsToReadNow;
				sectorsReadThisCall   = sectorsToReadNow;
			}
		} // have unread sectors
	} // is sub-dir
	return sectorsReadThisCall;
}


// This compacts a directory by moving all deleted entries to the end of
// the DIR list. It sets nextEntryInBuffer to the first deleted DIR entry
// in the compacted list of DIR entries currently in the buffer.
//
// ->	nextEntryInBuffer: a ptr to a word index of the next deleted entry
// 	in the buffer. On the initial call to this routine for a given dir
//		the value must be 0 and not otherwise altered. If this routine
//		returns TRUE this variable will hold the count of the valid DIR
//		entries in the directory.
//
// <-	TRUE:	if all valid directory entries have been processed.
//		FALSE: the end of the list of DIR entries was not encountered.
//			In this case nextEntryInBuffer has the index of the next
//			available slot in the buffer into which subsequent dir entries
//			can be read. The number of entries read must not extend past the
//			end of the buffer. Directories larger than 64K in size with
//			no more than 2048 valid entries can be processed in this way.
//
bool CompactDirEntries(ushort far *nextEntryInBuffer)
{ LPDIR lpNextEntryToUse, lpNextEntryToConsider, lpNextEntryToMove;
  ushort entriesToMove;

	ASSERT(nextEntryInBuffer != 0L);

	lpNextEntryToUse = (LPDIR)(lpSectorBuffer)+*nextEntryInBuffer;

FIND_DELETED_ENTRIES:
	entriesToMove = 0;
	while(!IS_DELETED((uchar)(*(lpNextEntryToUse->name))))
	{
		// if we found a free dir entry then we have processed all valid
		// entries in the dir chain
		//
		ASSERT((lpNextEntryToUse->name[0]==0)? (lpNextEntryToUse->attr==0): 1);

		if(IS_END_OF_DIR_LIST(lpNextEntryToUse))
			return TRUE;

		// otherwise count this valid entry
		//
		++lpNextEntryToUse;
		++*nextEntryInBuffer;

		// if too many valid entries to fit in buffer then we have troubles...
		if(*nextEntryInBuffer >= bufferDirEntries)
			return FALSE;
	}
	// next valid entry can only come after the next deleted entry
	// if free entry encountered there then we're done
	// (nextEntryInBuffer is index is first of n>=1 deleted entries
	// that make up the last n non-empty dir entries)
	//
	lpNextEntryToConsider = lpNextEntryToUse+1;

	ASSERT((lpNextEntryToConsider->name[0]==0)?
			 (lpNextEntryToConsider->attr==0): 1);

	if(IS_END_OF_DIR_LIST(lpNextEntryToConsider))
		return TRUE;

	// Look for the first entry that has not been deleted or quit when
	// we find the first free dir entry. If our pointer wraps then
	// we have no more valid entries.
	//
	while((IS_DELETED((uchar)(*lpNextEntryToConsider->name))) &&
			(lpNextEntryToConsider > lpNextEntryToUse))
	{
		// else keep looking for a valid entry
		++lpNextEntryToConsider;

		// if free entry encountered then we're done
		//
		ASSERT((lpNextEntryToConsider->name[0]==0)?
			 	 (lpNextEntryToConsider->attr==0): 1);

		if(IS_END_OF_DIR_LIST(lpNextEntryToConsider))
			return TRUE;
	}

	// we have a valid entry to replace the deleted entry or the buffer is
	// padded with deleted entries. If the latter FALSE to say we haven't
	// reached the end of the dir; nextEntryInBuffer will have the index
	// of the deleted element.
	//
	if((!IS_DELETED((uchar)(*lpNextEntryToConsider->name))) &&
		(lpNextEntryToConsider > lpNextEntryToUse))
	{
		lpNextEntryToMove = lpNextEntryToConsider;

		while((!IS_DELETED((uchar)(*lpNextEntryToConsider->name))) &&
			   (lpNextEntryToConsider > lpNextEntryToUse) &&
				!(IS_END_OF_DIR_LIST(lpNextEntryToConsider)))
		{
			++entriesToMove;
			++lpNextEntryToConsider;

			ASSERT((lpNextEntryToConsider->name[0]==0)?
			 	 	 (lpNextEntryToConsider->attr==0): 1);
		}
		memmove(lpNextEntryToUse,lpNextEntryToMove,
				  sizeof(DIR)*entriesToMove);

		lpNextEntryToUse   += entriesToMove;
		*nextEntryInBuffer += entriesToMove;
		lpNextEntryToMove  += entriesToMove;

		if(IS_END_OF_DIR_LIST(lpNextEntryToConsider))
			return TRUE;	// we just moved the last valid dir entry down
		else
			lpNextEntryToConsider = lpNextEntryToUse; // want to delete old dups

		// if the buffer was full of valid entries
		//
		if((lpNextEntryToConsider > lpNextEntryToMove))
			return FALSE;

		// delete all of the old entries duplicated by the copy op
		//
		while(lpNextEntryToConsider < lpNextEntryToMove)
		{
			(uchar)(*lpNextEntryToConsider->name) = DELETED;
			++lpNextEntryToConsider;
		}

		// repeat until no deleted entries remain in the dir
		//
		goto FIND_DELETED_ENTRIES;
	}

	// we can only get here if the end of the buffer is filled with deleted
	// entries
	//
	return FALSE;
}


// This tries to read in a complete directory into our buffer. If
// the directory size exceeeds the size of the buffer but the space
// for all valid entries does not, this will still work since it
// compacts the directory as it goes along.
//
//	->	a pointer to the open directory to be read, must have opened by
//	   a previous call to OpenThisDir().
// <-	>0: number of valid DIR entries in the directory and in our buffer else
//		 0: the entire directory could not fit into our buffer or
//		<0: a disk I/O error occured when trying to read the directory.
//
short ReadInAllDirEntries(struct ReadDirStruct far *lpDir)
{ ushort nextEntryInBuffer, entriesInBuffer, sectorsLeftInBuffer;
  LPDIR lpDirBuffer;
  short result;

	ASSERT(lpDir != 0L);

	memset((void far*)(lpSectorBuffer),0,0xFFFF);

	lpDirBuffer				= (LPDIR)(lpSectorBuffer);
	nextEntryInBuffer 	= 0;
	sectorsLeftInBuffer 	= bufferSectors;
	entriesInBuffer 		= (ushort)(bufferSize >> DIREntriesToBytes);
	do
	{
		if(!sectorsLeftInBuffer)
			return 0;

		// read as many sectors as we can from the current dir
		//
		result = ReadDirSectors(lpDir,(void far*)(lpDirBuffer+nextEntryInBuffer),
							sectorsLeftInBuffer);
		if(result < 0)
			return result;

		// compress the dir by replacing deleted entries by any valid
		// dir entries that may come after them in the array of dir structures
		// on disk.
		if(CompactDirEntries(&nextEntryInBuffer))
		{
			// pad the buffer with zeros to indicate that these are free entries
			//
			lpDirBuffer += nextEntryInBuffer;
			memset((void far *)(lpDirBuffer), 0,
					 (entriesInBuffer - nextEntryInBuffer) << DIREntriesToBytes);

			// return the index of the first free entry in the buffer
			// which is the same as the number of valid entries in the buffer
			//
			return nextEntryInBuffer;
		}
		// unless all the dir sectors have been read (meaning the dir is exactly
		// full), then this dir is too big for us to defrag
		//
		else if((nextEntryInBuffer >= bufferDirEntries) &&
				  (lpDir->sectorsRead < lpDir->sectorsToRead))
			return 0;

		// else more dir entries to read, fill up the remainder of the buffer.
		sectorsLeftInBuffer = (entriesInBuffer - nextEntryInBuffer)
										>> DIREntriesToSectors;
	}
	while(lpDir->sectorsRead < lpDir->sectorsToRead); // sectors left to read

	// we'll only get here if the dir is a full 64K in size.
	return nextEntryInBuffer;
}

// This is the comparison function used in sorting the dir entries.
// The rules are:
//		1) Sub-dir entries have the lowest collating order of any attribute.
//	    	a) Among sub-dirs the current collating order applies.
//		2) Free and deleted entries have the highest collating order.
//			a) Among free and deleted entries all are equal.
// 	3) For all other attributes the current user specified collating
//			order applies. If no collating order has been specified by the
//			user, we return entry1 = entry2 for stability.
// Note:
//		We need our special temp file at the head of the normal file list
//		no matter what the sorder order. We need it there to avoid creating
//		inversions when defragging files and to serve as a place holder
//		so that we don't accidently grow sub-dirs when we defrag files.
//
//	->	entry1, entry2: LPDIRs to entries in the DIR being sorted.
// <-	<0:	entry1 < entry2
//		 0:   entry1 = entry2
//    >0:	entry1 > entry2
//
int _USERENTRY CompareDirEntries(const void * a, const void * b)
{ ushort attr1, attr2;
  LPDIR entry1, entry2;
  char *curSortOpt;
  short result = 5;
  long  lresult;

	ASSERT(a != 0L);
	ASSERT(b != 0L);

	entry1 = (LPDIR)(a);
	entry2 = (LPDIR)(b);
	attr1 = entry1->attr;
	attr2 = entry2->attr;


	if((attr1 & A_SUBDIR) != 0)		// if 1st is a subdir and 2nd is not
	{
		if((attr2 & A_SUBDIR) == 0)
			return -1;						// 1st < 2nd
	}
	else if((attr2 & A_SUBDIR) != 0)	// if 2nd is a subdir and 1st is not
	{
			return +1;						// 1st > 2nd
	}
#if 0
	if((attr1 & (A_HIDDEN | A_SYSTEM)) != 0)
	{
		// if 1st is SYSTEM and 2nd is not
		if((attr2 & (A_HIDDEN | A_SYSTEM)) == 0)
			return -1;						// 1st < 2nd
		else
			return 0;
	}
	else if((attr2 & (A_HIDDEN | A_SYSTEM)) != 0)
	{
			// if 2nd is SYSTEM and 1st is not
			return +1;						// 1st > 2nd
	}
#endif
	if((attr1 & A_VOLUME) != 0)		// if 1st is a label and 2nd is not
	{
		if((attr2 & A_VOLUME) == 0)
			return -1;						// 1st < 2nd
	}
	else if((attr2 & A_VOLUME) != 0)	// if 2nd is a label and 1st is not
	{
			return +1;						// 1st > 2nd
	}
	else if(IS_DELETED((uchar)(*entry1->name)))
	{
		if(IS_DELETED((uchar)(*entry2->name)))
			return  0;						// if both deleted then 1st = 2nd
		else
			return +1;						// if 1st deleted but not 2nd, 1st > 2nd
	}
	else if(IS_DELETED((uchar)(*entry2->name)))
	{
			return -1;						// if 2nd deleted bun not 1st, 2nd > 1st
	}
	else if(IS_END_OF_DIR_LIST(entry1))
	{
		if(IS_END_OF_DIR_LIST(entry2))
			return  0;						// if both entries free, 1st = 2nd
		else
			return +1;						// if 1st is end of list, 1st > 2nd
	}
	else if(IS_END_OF_DIR_LIST(entry2))
	{
			return -1;						// if 2nd is end of list, 2nd > 1st
	}
	else if(strncmp(throwAwayName,entry2->name,8) == 0)
	{
			return +1;
	}
	else if(strncmp(throwAwayName,entry1->name,8) == 0)
	{
			return -1;
	}

	// By now we know that neither entry has been deleted nor is it free.
	// Also, if either of the entries is a sub-dir then both are.
	// We also know that our special tempory file has been designated
	// the initial normal file (regardless of the specified order).
	//
	// Perform the normal comparison functions, as specified by the user.
	//
	curSortOpt = sortOpts;
	do
	{	switch(*curSortOpt)
		{
		case 'N':	// alphabetic by name
			result = strncmp(entry1->name, entry2->name, 8);
			break;

		case 'E': 	// alphabetic order by extension
			result = strncmp(entry1->ext, entry2->ext, 3);
			break;

		case 'D':	// date order
			lresult = ((long)(entry1->time) - (long)(entry2->time));
			result  = (lresult == 0) ? 0 : ((lresult < 0) ? -1 : 1);
			break;

		case 'S':	// size order
			lresult = (entry1->size - entry2->size);
			result  = (lresult == 0) ? 0 : ((lresult < 0) ? -1 : 1);
			break;

		case '\0':	// end of list
		default:
			if(curSortOpt == sortOpts)
				// if no valid options: entry1 = entry2 to promote stability
				// This will only sorta work since QuickSort is inherently
				// unstable.
				//
				return 0;
			else
				return result;
		}
		curSortOpt++;
		if(*curSortOpt == '-')
		{
			result = -result;
			++curSortOpt;
		}
	} while((result == 0) && (*curSortOpt != '\0'));

	return result;
}

// writes entries to a dir file
// returns TRUE on success
//
bool WriteDirFileEntries(short dirFileHandle, // DOS file handle
								 ushort firstEntry,	 // first entry in dir to write
								 ushort entriesToWrite) // entry count to write
{ ushort bytesWritten, bytesToWrite;
  long filePos;
  bool oneExtra;

	ASSERT(dirFileHandle >= 0);
	ASSERT(firstEntry < bufferDirEntries);
	ASSERT(entriesToWrite <= bufferDirEntries);
	ASSERT(entriesToWrite > 0);


	filePos = (long)(firstEntry) * sizeof(DIR);
	SeekFile(dirFileHandle,&filePos,SEEK_SET);

	// we can't actually write 64K at a time so if that's what the user
	// wants then we have to make a special trip to pick up the
	// one extra one that we couldn't carry the first time
	//
	if(entriesToWrite == bufferDirEntries)
	{
		oneExtra = TRUE;
		bytesToWrite = (bufferDirEntries -1) << DIREntriesToBytes;
	}
	else
	{
		oneExtra = FALSE;
		bytesToWrite = entriesToWrite << DIREntriesToBytes;
	}

	if(!DefragWrite(dirFileHandle,lpSectorBuffer,&bytesToWrite))
		return FALSE;
	if(oneExtra)
	{
		bytesWritten = bytesToWrite;
		bytesToWrite = sizeof(DIR);
		if(!DefragWrite(dirFileHandle,lpSectorBuffer+bytesWritten,&bytesToWrite))
			return FALSE;
	}
	return TRUE;
}


// reads entries from a DIR file.
// returns TRUE on success
//
short ReadDirFileEntries(short dirFileHandle, // DOS file handle
								ushort firstEntry,	// first entry to read
								ushort entriesToRead)	// count to read
{ ushort bytesRead, bytesToRead;
  long filePos;
  bool oneExtra;

	ASSERT(dirFileHandle >= 0);
	ASSERT(firstEntry < bufferDirEntries);
	ASSERT(entriesToRead <= bufferDirEntries);
	ASSERT(entriesToRead > 0);

	filePos = (long)(firstEntry) * sizeof(DIR);
	SeekFile(dirFileHandle,&filePos,SEEK_SET);

	if(entriesToRead == bufferDirEntries)
	{
		oneExtra = TRUE;
		bytesToRead = (bufferDirEntries -1) << DIREntriesToBytes;
	}
	else
	{
		oneExtra = FALSE;
		bytesToRead = entriesToRead << DIREntriesToBytes;
	}

	if(!DefragRead(dirFileHandle,lpSectorBuffer,&bytesToRead))
		return FALSE;
	if(oneExtra)
	{
		bytesRead   = bytesToRead;
		bytesToRead = sizeof(DIR);
		if(!DefragRead(dirFileHandle,lpSectorBuffer+bytesRead,&bytesToRead))
			return FALSE;
	}
	return TRUE;
}


// A directory file might not fill up an entire cluster with
// valid entries (especially if the clusters are big). In such
// a case we want to pad the cluster with 0's so DOS treats
// them as end of list markers.
// returns TRUE on success
//
bool PadDirFile(short dirFileHandle, 		// DOS file handle
					ushort validDirEntries,		// valid DIR entry count
	     			ushort newDirSectorCount)	// sectors of padding needed
{ ulong padEntriesToWrite, paddedEntries, filePos;
  ushort padEntriesToWriteNow, bytesToWriteNow;

	ASSERT(newDirSectorCount > 0);
	ASSERT(validDirEntries > 0);
	ASSERT(validDirEntries <= bufferDirEntries);
	ASSERT(dirFileHandle > 0);

	padEntriesToWrite  = (ulong)newDirSectorCount << DIREntriesToSectors;
	padEntriesToWrite -= validDirEntries;
	if(padEntriesToWrite == 0)
		return 0;

	paddedEntries 	    = 0;
	filePos            = 0;
	SeekFile(dirFileHandle,&(long)filePos,SEEK_END);
	memset(lpSectorBuffer,0,(ushort)(bufferSize-32));
	while(paddedEntries < padEntriesToWrite)
	{
		padEntriesToWriteNow =
					(ushort)(MIN(padEntriesToWrite,(ulong)(bufferDirEntries-1)));
		bytesToWriteNow = padEntriesToWriteNow << DIREntriesToBytes;
		if(!DefragWrite(dirFileHandle,lpSectorBuffer,&bytesToWriteNow))
			return FALSE;

		paddedEntries += padEntriesToWriteNow;
	}
	return TRUE;
}



// Deletes all temporary dir files in a directory tree. This is
// called when we longjmp() back from a recursive call to
// OptimizeSubDirectories() as a result of a Fatal Error.
//
// -> dirName: 8.3 name of the level 1 root of the sub-dir tree
//					(the root dir entry.)

void DeleteAllDirFiles(char *dirName)
{ int wasFound;
  FIND found;

	ChangeDir(dirName);
	SetDTA((void far*)&found);

  	wasFound = FindFirst(throwAwayName, A_NORMAL);
  	if(wasFound)
  	{
		DeleteFile(throwAwayName);

		wasFound = FindFirst("*.*", A_SUBDIR);
		while(wasFound)
		{
			// skip '.' and '..'
			//
			if((found.attr & A_SUBDIR) && (found.name[0] != '.'))
			{
				DeleteAllDirFiles(found.name);
				SetDTA((void far*)&found);
			}
			wasFound = FindNext();
		}
  	}
  	ChangeDir("..");
}

// It may happen that a sub-dir is too big to defrag but its parent
// may have been moved. In such a case we need to update the back
// link in the "too-big-dir". This routine does that. It assumes
// that the back-link is correctly stored in the second dir entry.
//
bool UpdateSubdirBacklink(ushort firstSubdirCluster, ushort parentDirCluster)
{ LPDIR lpSubdir;
  ulong dirSector;

		dirSector  = ((ulong)(firstSubdirCluster) -2L) << sectorsToClusters;
		dirSector += dataStart;
		lpSubdir   = (LPDIR)(lpSectorBuffer);

		if(DefragAbsRead(DOSDrive, 1, dirSector, (void far *)lpSubdir))
		{
			lpSubdir[1].start = parentDirCluster;
			return DefragAbsWrite(DOSDrive, 1, dirSector, (void far *)lpSubdir);
		}
		return FALSE;
}

// This routine is called by OptimizeDirectories() that handles the
// root and links in the results of this recursive routine.
// When we ultimately return from this routine we will return a
// sector number that is the start of a new sub-dir file which
// is linked to child subdir files. To optimize the dir, we update
// the root dir to point to this link and delete the old sub-dir
// clusters.
//
ushort OptimizeSubDirectories(ushort originalStartingCluster, // were we start
										ushort parentDirCluster,		// were we return
										char far *lpThisDirName) // for user's info
{ LPDIR lpCurrentDirEntry, qSortPtr;
  ushort newDirSectors, newDirClusters, action,
			originalChildDirCluster, newChildDirCluster, newStartingCluster;
  long   newDirBytes;
  char dotName[14];
  short i, validDirEntries, dirFileHandle;
  struct ReadDirStruct thisDir;
  LPSFT lpSFT;

	ASSERT(lpThisDirName != 0L);
	ASSERT(BackwardFAT[originalStartingCluster] == FAT_EOF);
	ASSERT((parentDirCluster == 0) ||
		    (BackwardFAT[parentDirCluster] == FAT_EOF));


	// We need to add a file to the sub-dir right away for two reasons.
	// The 1st is to see if the dir grows when opening another file.
	// If so we have to update our copy of the FAT (we can tell if the
	// sub-dir grows if the cluster number in the SFT is not in our
	// copy of the FAT chain that we read in at startup). The second
	// reason is the same as for the root dir -- to maintain the sort
	// order once we start defragging files (see OptimizeDiretories()).
	//
	lpDPB->newclus = firstFree-1;
	action = ACT_EXIST_REPLACE | ACT_NEXIST_CREATE;
	if(!OpenOrCreateFile(throwAwayName, (O_WRITENOW | O_RDWR), A_NORMAL,
		     					&action, &dirFileHandle))
	{
		abortingBecause = EL_DISK_WRITE_ERR;
		longjmp(dirJmpBuf,1);
	}
	lpSFT              = HandleToSFT(dirFileHandle);
	CloseFile(dirFileHandle);
	newStartingCluster = 2+(ushort)((lpSFT->dirsec - dataStart)
								 				>> sectorsToClusters);
	OpenThisDir(originalStartingCluster, &thisDir, newStartingCluster);

	// assume no movement of the start of the dir
	//
	newStartingCluster = originalStartingCluster;

	// Get the dir entries into our buffer
	//
	validDirEntries = ReadInAllDirEntries(&thisDir);
	if(validDirEntries == 0)			// sub-dir >64K
	{
		++dirsSkipped;
		WarnUserAboutDir(lpThisDirName);
		if(parentDirCluster)
			UpdateSubdirBacklink(originalStartingCluster, parentDirCluster);
		DeleteFile(throwAwayName);
		goto END_OF_IT;
	}
	else if(validDirEntries < 0)		// I/O error
	{
		DeleteFile(throwAwayName);
		longjmp(dirJmpBuf,1);
	}
	yield(dirJmpBuf);

	// ignore "." and "..", sort the DIR (our collating order will put
	// all the sub-dir entries at the top of the list and all deleted
	// entries at the bottom)
	//
	i = 2;
	lpCurrentDirEntry = (LPDIR)(lpSectorBuffer);
	if(validDirEntries > 2)
	{
		qSortPtr = (LPDIR)(MK_FP((FP_SEG(lpCurrentDirEntry)+4),0));
		qsort((void far*)(qSortPtr), validDirEntries-2,
				sizeof(DIR), CompareDirEntries);
	}
	yield(dirJmpBuf);

	// We now have our place holder file at the top of the normal
	// file list. Delete it (we can't call DeleteFile() for this since
	// none of the subdirs exist until the whole optimized chain has
	// be linked into the dir tree by returning to the root dir handler).
	//
	for(i=2; i < validDirEntries; i++)
	{	if(strncmp((char*)(&lpCurrentDirEntry[i].name),throwAwayName,8)==0)
		{
			lpCurrentDirEntry[i].name[0] = DELETED;
			break;
		}
	}
	i = 2;

	// Figure out how many clusters we really need for
	// this sub-dir now that it has been compacted.
	//
	newDirBytes    = (long)(validDirEntries) << DIREntriesToBytes;
	newDirSectors  = (ushort)(newDirBytes >> bytesToSectors) +
						  ((newDirBytes & partialSectorMask) != 0);
	if(newDirSectors < sectorsPerCluster)
		newDirSectors = sectorsPerCluster;
	newDirClusters = (newDirSectors >> sectorsToClusters) +
						  ((newDirSectors & partialClusterSectorsMask) != 0);
	newDirSectors  = newDirClusters << sectorsToClusters;

	// allocate space for the dir using our copy of the FAT in RAM
	//
	newStartingCluster = GetFirstFit(&newDirClusters);
	if(newStartingCluster == 0)
	{
		// out of disk space
		//
		abortingBecause = EL_ENODISK;
		longjmp(dirJmpBuf,1);

		// can't allocate space for the optimized dir.
		//
		++dirsSkipped;
		if(parentDirCluster)
			UpdateSubdirBacklink(originalStartingCluster, parentDirCluster);
		DeleteFile(throwAwayName);
		WarnUserAboutDir(pathName);
	   newStartingCluster = originalStartingCluster;
	   goto END_OF_IT;
	}
	yield(dirJmpBuf);

	// Update the locations of "." and ".."
	//
	lpCurrentDirEntry++->start = newStartingCluster;
	lpCurrentDirEntry++->start = parentDirCluster;

	// Tell DOS to write the new file to the newly allocated
	// region of the FAT
	//
	lpDPB->newclus = newStartingCluster-1;
	action = ACT_EXIST_REPLACE | ACT_NEXIST_CREATE;
	if(!OpenOrCreateFile(throwAwayName, (O_WRITENOW | O_RDWR), A_NORMAL,
							&action, &dirFileHandle))
	{
#if DEBUG
		printf("Could not create dir file.\n");
#endif
		abortingBecause = EL_DISK_WRITE_ERR;
		longjmp(dirJmpBuf,1);
	}

	// This write actually allocates all the space we will need on
	// the disk for the current dir and writes out all the sorted
	// entries into a compacted file. If there is room left over in
	// the final cluster, pad it with 0's.
	//
	yield(dirJmpBuf);
	WriteDirFileEntries(dirFileHandle, 0, validDirEntries);
	yield(dirJmpBuf);
	PadDirFile(dirFileHandle,validDirEntries,newDirSectors);
	yield(dirJmpBuf);
	ReadDirFileEntries(dirFileHandle, 2, 1);
	yield(dirJmpBuf);

#if DEBUG
	lpSFT = HandleToSFT(dirFileHandle);
	if(lpSFT->start != newStartingCluster)
	{
#if DEBUG
		printf("Allocated wrong dir file cluster: wanted %u\t got %u\n",
						 lpSFT->start, newStartingCluster);
#endif
		asm	int	3
		ASSERT(lpSFT->start == newStartingCluster);
	}
	else
	{
		printf("Created temp DIR file %s\n\tParent: %4X Org: %4X New: %4X\n",
					pathName,parentDirCluster,originalStartingCluster,
					newStartingCluster);
	}
#endif
	CloseFile(dirFileHandle);

	// put the original starting cluster on our stack of FAT chains
	// to free.
	//
	PushDirFATChainToFree(originalStartingCluster);
	yield(dirJmpBuf);

	// Optimize each of the children of the current dir. All of
	// the sub-dir entries will be at the front of the dir file.
	//
	i = 2;
	lpCurrentDirEntry = (LPDIR)(lpSectorBuffer);
	while((lpCurrentDirEntry->attr & A_SUBDIR) != 0)
	{
		yield(dirJmpBuf);

		// make the child dir the current dir
		//
		ename_to_dot(dotName, lpCurrentDirEntry->name);
		AddName(pathName,dotName);
		ChangeDir(dotName);

		// optimize the child's directory tree
		//
		originalChildDirCluster = lpCurrentDirEntry->start;
		newChildDirCluster 		= OptimizeSubDirectories(originalChildDirCluster,
																		 newStartingCluster,
																		 dotName);
		yield(dirJmpBuf);

		// we are back in our own current dir.
		// open our directory file and get back the dir entry for
		// the child sub-dir we just optimized.
		//
		action = ACT_EXIST_OPEN;
		if(!OpenOrCreateFile(throwAwayName, (O_WRITENOW | O_RDWR), A_NORMAL,
								&action, &dirFileHandle))
		{
			abortingBecause = EL_DISK_READ_ERR;
			longjmp(dirJmpBuf,1);
		}

		yield(dirJmpBuf);
		ReadDirFileEntries(dirFileHandle, i, 1);

		// update the starting cluster of the child
		//
		yield(dirJmpBuf);
		lpCurrentDirEntry        = ((LPDIR)(lpSectorBuffer));
		lpCurrentDirEntry->start = newChildDirCluster;

		// update the disk file
		//
		yield(dirJmpBuf);
		WriteDirFileEntries(dirFileHandle, i++, 1);

#if DEBUG
		printf("Linked in %s\n",pathName);
#endif
		RemoveName(pathName);

		// get the next DIR entry
		//
		ReadDirFileEntries(dirFileHandle, i, 1);
		CloseFile(dirFileHandle);

	}
	// we actually defragged this sub-dir
	//
	discontinuitiesFixed += thisDir.discontinuities;

	// change back to the parent dir and return the (presumably) new location
	// for this (hopefully optimized) sub-dir
	//
END_OF_IT:

	// don't look for more disconinuities than there are, even if we can't
	// fix this one particular set
	//
	discontinuities -= thisDir.discontinuities;
	ChangeDir("..");
	return newStartingCluster;
}


// This optimizes the root dir and links in all the subdir files
// created by OptimizeSubDirectories().
// returns TRUE on success
//
bool OptimizeDirectories(void)
{ LPDIR lpCurrentDirEntry;
  ushort newDirSectors, padDirSectors, action,
  			originalChildDirCluster, newChildDirCluster;
  char dotName[14];
  short i, validDirEntries, dirFileHandle;
  struct ReadDirStruct thisDir;
  short result, systemFiles;

	ASSERT(FAT != 0L);
	ASSERT(BackwardFAT != 0L);

	pathName[0] = driveLetter;
	pathName[1] = ':';
	pathName[2] = '\\';
	pathName[3] = '\0';

	// We need to create an extra file in the root dir that will
	// be sorted to the top of the list of normal files regardless
	// of the sorted order. Once we have a file in this position
	// we have to delete it. The defrag routines process the dir
   // entries in order, copying the contents of fragmented files and
	// then deleting the original file and renaming the copy to the
	// original file's name. Without this extra deleted dir entry,
	// the first file, once renamed would end up at the bottom of
	// the list. With the placeholder file the deleted entry gets
	// "bubbled up" to the bottom of the normal file list as each
	// file gets deleted and renamed -- maintaining the sort order
   // established in this module.
	//
	action = ACT_EXIST_REPLACE | ACT_NEXIST_CREATE;
	if(!OpenOrCreateFile(throwAwayName, (O_WRITENOW | O_RDWR), A_NORMAL,
							&action, &dirFileHandle))
		FatalError(EL_DISK_WRITE_ERR);
	CloseFile(dirFileHandle);

	// Get the root dir entries into our buffer
	//
	OpenThisDir(0, &thisDir, 0);
	yield(0);
	validDirEntries = ReadInAllDirEntries(&thisDir);
	if(validDirEntries == 0)
	{
		// root dir too big
		//
		yield(0);
		++dirsSkipped;
		WarnUserAboutDir(pathName);
		return FALSE;
	}
	else if(validDirEntries < 0)
	{
		// disk I/O error
		//
		yield(0);
		return FALSE;
	}
	yield(0);

	// sort the root dir
	//
	systemFiles = 0;
	lpCurrentDirEntry = (LPDIR)(lpSectorBuffer);
	while((lpCurrentDirEntry->attr & (A_HIDDEN | A_SYSTEM)) != 0)
	{
		++systemFiles;
		++lpCurrentDirEntry;
	}
	qsort((void far*)(lpCurrentDirEntry), validDirEntries-systemFiles,
				sizeof(DIR), CompareDirEntries);
	newDirSectors  = (validDirEntries >> DIREntriesToSectors) +
							((validDirEntries & 0x0F) != 0);
	yield(0);

	// Write out all the sorted entries into the root directory.
	// If there is room left over in the final sectors, pad them
	// with 0's.
	//
	DefragAbsWrite(DOSDrive, newDirSectors, (long)(rootDIRStart),
							(void far*)(lpSectorBuffer));
	yield(0);
	padDirSectors = (ushort)(dataStart - rootDIRStart) - newDirSectors;
	if(padDirSectors > 0)
	{
		memset(lpSectorBuffer,0,(ushort)(newDirSectors << bytesToSectors));
		DefragAbsWrite(DOSDrive, padDirSectors,
							(long)(rootDIRStart+newDirSectors),
							(void far*)(lpSectorBuffer));
	}
	yield(0);

	// Delete our place holder file.
	//
	DeleteFile(throwAwayName);

	// Optimize each of the sub-dirs. All the the sub-dir entries will
	// be at the front of the dir file, right after any boot files,
	// because of our collating order.
	//
	i = systemFiles;
	DefragAbsRead(DOSDrive, 1, (long)(rootDIRStart),
							(void far*)(lpSectorBuffer));
	lpCurrentDirEntry  = ((LPDIR)(lpSectorBuffer));
	lpCurrentDirEntry += systemFiles;

	while((lpCurrentDirEntry->attr & A_SUBDIR) != 0)
	{
		yield(0);

		// make the sub-dir the current dir and optimize the directory tree
		//
		ename_to_dot(dotName, lpCurrentDirEntry->name);
		originalChildDirCluster  = lpCurrentDirEntry->start;
		AddName(pathName,dotName);
		ChangeDir(dotName);

		// if we run into an unrecoverable problem while processing
		// any of the sub-dirs we will longjmp() back to here. In
		// such a case we have to delete all of our temporary files
		// and then terminate by calling yield() (which will call
		// FatalError() which will call exit() which will call all
		// the atexit() routines to clean up after ourselves.)
		//
		result = setjmp(dirJmpBuf);
		if(result)
		{
			ChangeDir("\\");

FILE_IO_ERROR:
			// !!!
			//
			// We're not doing a good job of cleaning up on a ^C.
			// For one thing we don't deallocate the DIR files in the BackwardFAT
			// when we delete them. This makes us assert(). We SHOULD
			// handle this better but for now we'll just wait for the
			// user to press <Enter> and exit.
			//
			DeleteAllDirFiles(dotName);
			CannotContinue();
			yield(0);
		}
		newChildDirCluster = OptimizeSubDirectories(originalChildDirCluster, 0,
																  dotName);

		// we are back in the root dir.
		// read in the root dir sector containing the sub-dir entry for
		// the sub-dir tree we just optimized.
		//
		if(!DefragAbsRead(DOSDrive,1,
								(long)(rootDIRStart + (i >> DIREntriesToSectors)),
								(void *)(lpSectorBuffer)))
			goto FILE_IO_ERROR;

		// update the starting cluster of the child in the DIR entry
		// in our buffer then write out the root dir sector containing
		// the sub-dir entry
		//
		(((LPDIR)(lpSectorBuffer))[i & (DIREntriesPerSector-1)]).start = newChildDirCluster;
		if(!DefragAbsWrite(DOSDrive,1,
								(long)(rootDIRStart + (i >> DIREntriesToSectors)),
								(void far*)(lpSectorBuffer)))
			goto FILE_IO_ERROR;

#if DEBUG
		printf("Linked in %s\n",pathName);
#endif
		RemoveName(pathName);

		// The sub-dir files are now sub-dirs linked into the root dir.
		// The old sub-dirs are now just lost clusters. Free them in
		// our copy of the FAT and in the two FAT copies on disk.
		//
		while((originalChildDirCluster = PopDirFATChainToFree()) != 0)
			FreeFATChainInRAM(originalChildDirCluster);

		WriteDirtyFATSectors(0);
		WriteDirtyFATSectors(1);
		yield(0);


		// get the next sub-dir entry to process
		//
		++i;
		DefragAbsRead(DOSDrive,1,
						  (long)(rootDIRStart + (i >> DIREntriesToSectors)),
						  (void *)(lpSectorBuffer));
		yield(0);
		lpCurrentDirEntry = &(((LPDIR)(lpSectorBuffer))[i & (DIREntriesPerSector-1)]);
	}

	// We're done. Delete our place holder file.
	//
	DeleteFile(throwAwayName);
	return TRUE;
}
