//
//   $Id: DEFERR.C 1.4 2002/10/16 21:45:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// DEFERR.C
// Common error handling routines for Datalight disk DEFRAGmentation utility.
// Created 03-13-97 Dennis Edwards
//
// Copyright (c) 1989-2002 Datalight, Inc.
//	All Rights Reserved.
//	Datalight Confidential and Proprietary
//
#define	__FILENUM__	4

#include <process.h>
#include <signal.h>
#include <float.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <alloc.h>
#include <ctype.h>
#include <dos.h>


#include "dostruct.h"

#include "defrag.h"
#include "defmsg.h"

#include "defutila.pt"
#include "deferr.pt"
#include "defintr.pt"
#include "absrdwr.pt"


#if	BC2

#define  _harderr(x) 		harderr(x)
#define  _hardresume(x) 	hardresume(x)
#define  _HARDERR_IGNORE	0
#define  _HARDERR_RETRY    1
#define  _HARDERR_ABORT    2
#define  _HARDERR_FAIL     3

#else

#define	Int24Translator	Int24Handler

#endif


// defmain.c
extern ushort DOSDrive;
extern ushort bytesPerSector;
extern bool isBig;

ushort abortingBecause = 0, canContinue=TRUE;

void far *allocStack[MAX_DEFRAG_ALLOCS];
short allocStackPtr;


// Installs all of the error handlers. Each function called
// is defined by the BCC runtime library (see BCC DOS Reference)
// except the <Ctrl><Alt><Del> handler. SignalHandler() just
// sets abortingBecause to the appropriate exit code (EL_USER_ABORT
// or EL_INTERNAL_ERR).
//
// The BCC RTL defines the following signal values:
//
//  Constant     Meaning
//
//  SIGABRT      Abnormal termination. The default action terminates
//	  the calling program with exit code 3. (Real and protected mode.)
//
//  SIGBREAK     CTRL+BREAK signal. The default action terminates the
//	  calling program. (Protected mode only.)
//
//  SIGFPE       Floating-point error, such as overflow, division by
//	  zero, or invalid operation. The default action
//	  terminates the calling program. (Real and protected
//	  mode.)
//
//  SIGILL       Illegal instruction. This signal is not generated by
//	  DOS or OS/2, but is supported for ANSI compatibility.
//	  The default action terminates the calling program.
//	  (Real and protected mode.)
//
//  SIGINT       CTRL+C interrupt. The default action issues INT 23H.
//	  (Real and protected mode.)
//
//  SIGSEGV      Illegal storage access. This signal is not generated
//	  by DOS or OS/2, but is supported for ANSI
//	  compatibility. The default action terminates the
//	  calling program. (Real and protected mode.)
//
//  SIGTERM      Termination request sent to the program. This signal
//	  is not generated by DOS or OS/2, but is supported for
//	  ANSI compatibility. The default action terminates the
//	  calling program. (Real and protected mode.)
//
//  FPE_EXPLICITGEN what-ever a floating point lib might decide...
//
void HandleErrors(void)
{
	InstallCtrlAltDelHandler();
	atexit(RemoveCtrlAltDelHandler);

	signal(SIGINT  ,SignalHandler);	// user abort...
	signal(SIGTERM ,SignalHandler);
	signal(SIGABRT ,SignalHandler);

	signal(SIGFPE  ,SignalHandler);	// internal error...
	signal(SIGSEGV ,SignalHandler);
	signal(SIGILL  ,SignalHandler);
	signal(FPE_EXPLICITGEN,SignalHandler);

	_harderr(Int24Translator);
	ctrlbrk(CtrlBrkHandler);

}


// This is the interface to the error handling code from the rest of
// the program. All the Disk I/O and memory allocation routines in
// in this module set the abortingBecause variable when they encounter
// an error. The outside routines periodically call yield() which can
// cause the program to abort. The other routines must only call yield()
// when the program/machine is in a safe state.
//
// ->	jumping: a jmp_buf as defined in SETJMP.H. if this is non-zero
//					longjmp() will be called under the assumption that this
//					will invoke error handling code to clean up the machine
//					and put it in a safe state.
//
void yield(jmp_buf jumping)
{

	if(abortingBecause)
		if(jumping)
			longjmp(jumping,1);
		else
			FatalError(abortingBecause);
	else
		// save the program state here to allow time-sliced operations
		{;}
}



// Response ID's returned by GetResponse()
//
#define  CRIT_ABORT	0
#define  CRIT_RETRY	1
#define  CRIT_FAIL	2
#define  CRIT_IGNORE 3

// Qualifies a user response to a critical error condition.
// If the user types in an invalid response it will beep otherwise
// it will return one of the CRIT_xxx response IDs listed above.
//
// -> void
// <- returns one of the CRIT_xxx response IDs listed above.
//
ushort GetResponse(void)
{ ushort response, i;
  char *validResponses=criterr_responses;

	while(1)
	{
		response = toupper(ReadKbd() & 0xFF);
		for(i = 0; i < strlen(validResponses); i++)
			if(validResponses[i] == response)
				return i;
		TTYOut('\a');
	}
}


// Displays a message using the BIOS video TTY output
//
// -> msg	A ptr to an ASCIIZ string.
// <- void
//
void Display(char *msg)
{
	while(*msg)
	{
		TTYOut(*msg++);
	}
}


// DEFRAG's very own assert!
//
// ->	file: __FILENUM__ of file that failed
//    line: line # of file that failed
//
void InternalError(int fileNum, int lineNum)
{
	char assertNumber[20];

	Display("DEFRAG internal error: ");
	itoa(fileNum,assertNumber,10);
	Display(assertNumber);
	Display("-");
	itoa(lineNum,assertNumber,10);
	Display(assertNumber);
	Display("\n\r");

	exit(EL_INTERNAL_ERR);
}



// Displays a warning message that a file is too big to defrag.
//
// -> lpName: The 8.3 name of the file
//
#pragma argsused
void WarnUserAboutFile(char *lpName)
{
#if	DEBUG
	Display("\n");
	Display(warning);
	Display(file_msg);
	Display(lpName);
	Display(toobig_msg);
#endif
}

// Displays a warning message that a directory is too big to defrag.
//
// -> lpName: The 8.3 name of the file
//
#pragma argsused
void WarnUserAboutDir(char *lpName)
{
#if	DEBUG
	Display("\n");
	Display(warning);
	Display(lpName);
	Display(dir_msg);
#endif
}


// If we get a user abort signal and can't clean up enough to continue
// we have to clean up enought to exit and then do so. Call this
// ONLY when an unrecoverable error has been detected.
//
void CannotContinue(void)
{
	canContinue = FALSE;
}

// Helper routine for Fatal Error. Used to display user options and
// qualify input in response to a <Ctrl><Alt><Del> or a <Ctrl><C/Break>
// keyboard input.
//
// <- TRUE if we should continue DEFRAG else FALSE if user is aborting.
//
bool ShouldWeContinue(void)
{ int response;

	if(canContinue)
		Display(tocontinue_msg);
	Display(toexitnow_msg );
	while(1)
	{
		response = toupper(ReadKbd() & 0xFF);
		if(response == '\r')
		{
			Display("\n\r");
			return FALSE;
		}
		else if(canContinue && (response == ' '))
		{
			Display("\n\r");
			return TRUE;
		}
		else
			Display("\a");
	}
}


// Displays a question and waits for a yes or no answer.
// Returns TRUE if the answer was affirmative
//
bool YesOrNo(char *question)
{ int response;

	printf(ask_yesno_msg,question,yesno_str);
	while(1)
	{
		response = toupper(ReadKbd() & 0xFF);
		if(response == yesno_responses[0])
		{
			Display("\n\r");
			return TRUE;
		}
		else if(response == yesno_responses[1])
		{
			Display("\n\r");
			return FALSE;
		}
		else
			Display("\a");
	}
}


// Prints the specified fatal error message and exit()s with the
// specified error code.
//
void FatalError(short exitCode)
{

// !!!
//
// can we elaborate on errror message?
//

	if((exitCode != EL_USER_ABORT) && (exitCode != EL_CTRL_ALT_DEL))
		Display(fatal_msg);

	switch(exitCode)
	{
	//
	//	The following error codes are the same as MS-DOS v6.22 DEFRAG
	//

	// 4   A general error occurred.
	//
	default:
	case  EL_GENERAL:
		//
		// fall through...

	// 1   An internal error occurred.
	//
	case EL_INTERNAL_ERR:
		SHIPABORT();
		break;


	// 2   The disk contained no free clusters. To operate, DEFRAG needs 1 free
	//     cluster.
	//
	case  EL_ENODISK:
		Display(fenodisk_msg);
		break;


	case	EL_CTRL_ALT_DEL:
		PassOnCtrlAltDel();
		Display(ctrlaltdel_msg);
		if(ShouldWeContinue())
		{
			abortingBecause = 0;
			TrapCtrlAltDel();
			return;
		}
		Display(feaborted_msg);
		break;

	// 3   The user pressed CTRL+C to stop the process.
	//
	case  EL_USER_ABORT:
		Display(ctrlbrk_msg);
		if(ShouldWeContinue())
		{
			abortingBecause = 0;
			return;
		}
		Display(feaborted_msg);
		break;

	// 5   DEFRAG encountered an error while reading a cluster.
	//
	case  EL_DISK_READ_ERR:
		Display(fediskrd_msg);
		break;

	// 6   DEFRAG encountered an error while writing a cluster.
	//
	case  EL_DISK_WRITE_ERR:
		Display(fediskwr_msg);
		break;

	// 7   An allocation error occurred. To correct the error, use the <SCANDISK>
	//     command.
	//
	case  EL_BAD_DISK:
		Display(feinvalid_msg);
		break;

	// 8   A memory error occurred.
	//
	case  EL_MEM_ERR:
		//
		// fall through...

	// 9   There was insufficient memory to defragment the disk.
	//
	case  EL_ENOMEM:
		Display(fenomem_msg);
		break;

	//
	//	The following error codes are specific to Datalight's DEFRAG
	//

	// 11  Long file names encountered
	//
	case 	EL_LONG_NAMES:
		Display(felongfilenames_msg);
		break;

	// 12 Sub-dir tree has more than MAX_NODES_IN_SUBDIR nodes
	//
	case 	EL_TOO_MANY_NODES:
		Display(fetoomanynodes_msg);
		break;

	// 13  Need at least a '386 CPU for bit set ops
	//
	case  EL_BAD_CPU:
		Display(febadcpu_msg);
		break;

	// 14 Windows was running
	//
	case  EL_WINDOWS_RUNNING:
		Display(fewindozed_msg);
		break;

	// 15 Open files detected
	//
	case EL_OPEN_FILES:
		Display(feopenfiles_msg);
		break;
	}
	exit(exitCode);
}


// Our critical error handler. This displays a prompt to the user and
// returns with their response. It fails all errors not associated with
// the disk being defragged.
//
// The shape of this error handler is specified by the BCC runtime
// library int 24h handler.  See BCC DOS Reference.
//
#pragma argsused;
#if  BC2
int Int24Translator(int __errval, int __ax, int __bp, int __si)
{ ushort di;

	di = _DI;
	Int24Handler((di & 0xFF), __errval, MK_FP((ushort)__bp,__si));
	return(_HARDERR_ABORT);
}
#endif

#pragma argsused
void far Int24Handler(ushort deverr, ushort errval, ushort far *devhdr)
{ ushort response;
  char newline[]= " \n\r";

	abortingBecause = 0;
	if(!(deverr & 0x8000))					// disk error?
	{
		if((deverr & 0xFF) == DOSDrive) 	// our drive?
		{
			// display the standard critical error message
			//
			Display(criterr_msg);
			if(errval & 0x100)
			{
				Display(critwr_msg);
			}
			else
			{
				Display(critrd_msg);
			}
			Display(critdrv_msg);
			newline[0] = ((char)(deverr) + 'A');
			Display(newline);

			Display(criterr_prompt);
			response = GetResponse();
			newline[0] = (char)(response);
			Display(newline);

			// GetResponse() will get the user to select one of
			// the cases below.
			//
			switch(response)
			{
			case CRIT_ABORT :
				if(errval & 0x100)		// if a disk error
				{
					abortingBecause = EL_DISK_WRITE_ERR;
				}
				else
				{
					abortingBecause = EL_DISK_READ_ERR;
				}
				_hardresume( _HARDERR_ABORT  );

			case CRIT_RETRY :
				_hardresume( _HARDERR_RETRY  );

			case CRIT_FAIL	 :
				_hardresume( _HARDERR_FAIL   );

			case CRIT_IGNORE:
				_hardresume( _HARDERR_IGNORE );
			}
		}
	}
	_hardresume( _HARDERR_FAIL );
}


// Handles ^C and ^Brk keystrokes. This routine is called by the
// BCC runtime libraries int 23h handler.
//
int CtrlBrkHandler(void)
{
	if(KbdInputPending())
		ReadKbd();							// throw away the ^C
	abortingBecause = EL_USER_ABORT;
	return(1);
}


// Handles various C signals by exiting the program in a defined way.
// This signal handler is defined by the BCC runtime library (See
// the BCC DOS Reference).
//
void _USERENTRY SignalHandler(int sig)
{

  	signal(sig, SIG_IGN);           // Tell the RTL to ignore the signal
  	signal(sig, SignalHandler);     // and then reset the signal handler

	switch(sig)
	{
	case SIGINT  :
	case SIGTERM :
	case SIGABRT :
		abortingBecause = EL_USER_ABORT;
		break;

	case SIGFPE  :
	case SIGSEGV :
	case SIGILL  :
	case FPE_EXPLICITGEN:
		abortingBecause = EL_INTERNAL_ERR;
	}
}


// free all dynamic memory we allocated
//
void DefragFarFreeAll(void)
{
	while(allocStackPtr > 0)
	{
		farfree(allocStack[--allocStackPtr]);
	}
}


// removes a ptr from the allocStack and frees the associated block
//
void DefragFarFree(void far * block)
{ushort i,j;

	ASSERT(block);
	ASSERT(allocStackPtr > 0);

	for(i = 0; i < allocStackPtr; i++)
	{
		if(allocStack[i] != block)
			continue;

		// free the block
		farfree(block);

		// delete the block from the stack and return
		for(j = i+1; j < allocStackPtr; j++,i++)
		{
			allocStack[i] = allocStack[j];
		}
		--allocStackPtr;
		return;
	}
}


// calls farcalloc() and saves all non-NULL return ptrs to free atexit().
//
// -> same arguments as farcalloc()
// <- same as farcalloc()
//
void far * PASCAL	DefragAlloc(unsigned long nunits, unsigned long unitsz)
{ void far *lpVoid;
  static bool firstTime=TRUE;

	ASSERT(nunits);
	ASSERT(unitsz);

	if(firstTime)
	{
		atexit(DefragFarFreeAll);
		firstTime = FALSE;
	}
	ASSERT(allocStackPtr < MAX_DEFRAG_ALLOCS);

	lpVoid = farcalloc(nunits, unitsz);
	if(lpVoid != 0L)
		allocStack[allocStackPtr++] = lpVoid;
	else
		abortingBecause = EL_ENOMEM;
	return lpVoid;
}


// encapsolates the RTL abswrite() to allow orderly exits on the next yield().
// Returns TRUE on success
//
bool DefragAbsWrite(int drive, int nsects, long lsect, void far *buffer)
{ int result;
  int secs;
  int maxsecs;
  ulong normalizedNumber;

	ASSERT(nsects);
	ASSERT(buffer);


	result  =  0;
	maxsecs = (bytesPerSector >= 0x100 ? 0x80/(bytesPerSector/0x100) : 0x80);

	while(nsects>0)
	{
		secs = (nsects > maxsecs) ? maxsecs : nsects;

		result = _abswrite(isBig,drive,secs,lsect,buffer);
		if(result != 0)
		{
			errno = result;
			abortingBecause = EL_DISK_WRITE_ERR;
			break;
		}
		nsects -= secs;
		lsect  += secs;

		normalizedNumber  = ((ulong)FP_SEG(buffer) << 4) + FP_OFF(buffer);
		normalizedNumber += (ulong)secs * bytesPerSector;

		buffer  = MK_FP((ushort)(normalizedNumber >> 4),
					   	 (ushort)(normalizedNumber & 0xF));
	}
	return (result == 0);
}


// encapsolates the RTL absread() to allow orderly exits on the next yield().
// unlike DefragAbsWrite(), this will take nsects such that nsects*BPS > 64K.
// Returns TRUE on success
//
bool DefragAbsRead(int drive, int nsects, long lsect, void far *buffer)
{ int result;
  int secs;
  int maxsecs;
  ulong normalizedNumber;

	ASSERT(nsects);
	SHIPASSERT(buffer);

	result  =  0;
	maxsecs = (bytesPerSector >= 0x100 ? 0x80/(bytesPerSector/0x100) : 0x80);

	while(nsects>0)
	{
		secs = (nsects > maxsecs) ? maxsecs : nsects;

		result = _absread(isBig,drive,secs,lsect,buffer);
		if(result != 0)
		{
			errno = result;
			abortingBecause = EL_DISK_READ_ERR;
			break;
		}
		nsects -= secs;
		lsect  += secs;

		normalizedNumber  = ((ulong)FP_SEG(buffer) << 4)+(ushort)FP_OFF(buffer);
		normalizedNumber += (ulong)secs * bytesPerSector;

		buffer  = MK_FP((ushort)(normalizedNumber >> 4),
							 (ushort)(normalizedNumber & 0xF));
	}
	return (result == 0);
}


// encapsolates WriteFile() to allow orderly exits on the next yield().
// Returns TRUE on success and updates len to the bytes actually written
//
bool DefragWrite(int handle, void *buf, ushort far * len)
{bool successful;

	ASSERT(len);
	ASSERT(buf);
	ASSERT(handle > 0);

	successful = WriteFile(handle, buf, len);
	if(!successful)
		abortingBecause = EL_DISK_WRITE_ERR;
	return successful;
}


// encapsolates ReadFile() to allow orderly exits on the next yield().
// Returns TRUE on success and updates len to the bytes actually read
//
bool DefragRead(int handle, void *buf, ushort far * len)
{bool successful;

	ASSERT(len);
	SHIPASSERT(buf);
	ASSERT(handle > 0);

	successful = ReadFile(handle, buf, len);
	if(!successful)
		abortingBecause = EL_DISK_READ_ERR;
	return successful;
}
