//
//   $Id: DEFPATHS.C 1.4 2002/10/16 21:45:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  DEFPATHS.C  Created 04-08-97 Dennis Edwards
//
//  Copyright (c) 1989-2002 Datalight, Inc.
//	 All Rights Reserved.
//	 Datalight Confidential and Proprietary
//
//  Routines to handle path names for Datalight Disk DEFRAGmentation utility.
//

#define __FILENUM__	9


#include <ctype.h>
#include <string.h>

#include "dostruct.h"

#include "defrag.h"

#include "defpaths.pt"


// international stuff not yet coded
//
#define IntlStrRChr(str, c) strrchr(str, c)
#define IntlStrChr(str, c) strchr(str, c)
#define IntlStrUpr(str) strupr(str)


// note: not implemented yet!
// Is this character one of the double-byte characters?
#pragma argsused
bool Is2ByteChar(unsigned char c)
{
#if 0
	static unsigned char far * c_dbcs_ptr;
	int i;

	// check the list of double byte characters
	for(i=0; c_dbcs_ptr[i]; i+=2)
	{
		if(c >= c_dbcs_ptr[i] && c <= c_dbcs_ptr[i+1])
			return TRUE;
	}
#endif
	return FALSE;
}

// Return the last character in the string.
char LastChar(char *pStr)
{
	char lastChar;

	// string must be valid
	ASSERT(pStr);

	// nothing in string, so return empty
	if(!pStr || *pStr == 0)
		return 0;

	// find last char
	while(*pStr)
	{
		// remember last character
		lastChar = *pStr;

		// go on to next char
		if(Is2ByteChar(lastChar))
			++pStr;
		++pStr;
	}

	// return last character
	return lastChar;
}


//	Add the name to the end of a path.
//	Return FALSE if strings together are too long.
//
bool AddName(char *pPath, const char *pName)
{
	// validate parameters
	ASSERT(pPath);
	ASSERT(pName);

	// don't allow strings too long
	if(strlen(pPath) + strlen(pName) + 2 >= MAXPATH)
		return FALSE;

	// tack on the backslash
	if(LastChar(pPath) != '\\')
		strcat(pPath, "\\");

	// add on the name
	strcat(pPath, pName);

	return TRUE;
}


//	Remove a name that was added with AddName.
//	Returns FALSE if no path items to remove.
//
bool RemoveName(char *pPath)
{
	char *pLastSlash;

	// validate parameter
	ASSERT(pPath);

	// find the final slash (there MUST be one because we previously
	// used AddName).
	pLastSlash = IntlStrRChr(pPath, '\\');
	if(!pLastSlash)
		return FALSE;

	// remove the last name from the string
	if(*(pLastSlash -1) != ':')
		*pLastSlash = 0;
	else
		*(pLastSlash+1) = 0;
	return TRUE;
}


//	Convert an expanded name form to a "FILE.EXT" form.
//	That is "FILE    C  " to "FILE.C".  Does no case
//	conversion.
//
void	PASCAL ename_to_dot(char *dotname, char *ename)
{
	uchar len;
	uchar end = 8;

	// validate parameters
	ASSERT(dotname);
	ASSERT(ename);

	/* place the name part into the dotname string */
again:
	for(len = end; len > 0 && ename[len-1] == ' '; --len)
		/* just count */;
	if(len)
	{
		if(end == 3)
			*dotname++ = '.';
		memmove(dotname,ename,len);
		dotname += len;
	}
	if(end == 8)
	{
		ename += 8;
		end = 3;
		goto again;
	}

	*dotname = 0;
}
