//
//   $Id: DEFRAGF.C 1.4 2002/10/16 21:45:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// DEFRAGF.C  Created 03-13-97 Dennis Edwards
//
// Copyright (c) 1989-2002 Datalight, Inc.
// All Rights Reserved.
// Datalight Confidential and Proprietary
//
// File oriented defraging module for Datalight disk DEFRAGmentation utility.
//
#define  __FILENUM__ 3

#include <setjmp.h>
#include <alloc.h>
#include <errno.h>
#include <ctype.h>
#include <mem.h>
#include <dos.h>
#include <stdio.h>

#include "dostruct.h"

#include "defrag.h"
#include "defmsg.h"

#include "defmain.pt"
#include "defutil.pt"
#include "defutila.pt"
#include "defragf.pt"
#include "deferr.pt"
#include "defpaths.pt"


// defmain.c
extern ushort bytesPerSector;    // Number of bytes/sector on target disk
extern ushort bytesToSectors;    // shift count to convert bytes->sectors
extern ushort bytesToClusters;   // shift count to convert bytes->clusters
extern ushort partialClusterBytesMask;  // AND mask for partial clusters
extern LPDPB lpDPB;    // ptr to DOS DPB in the LOL
extern ushort huge *BackwardFAT; // our backward links and free list

// defchkdk.c
extern ulong discontinuities;		  		// total remaining in all FAT chains
extern ulong discontinuitiesAtStart;	// total in all FAT chains at start
extern ulong discontinuitiesFixed;		// total resolved

// defdirs.c
extern char throwAwayName[];
extern char pathName[];

// The file I/O buffer ptr and size info
//
uchar far *lpSectorBuffer;
ushort bufferSectors, bufferDirEntries;
ulong  bufferSize;

// The number of files we couldn't defragment
//
ulong filesSkipped;

// The jmp_buf used to, as they say in Prolog, FAIL a recursive
// subroutine call back to a safe error handler. If yield is called
// with a non-null jmp_buf argument then yield() will longjmp() to that
// error handler if a fatal error has occured since the last yield() call.
// If a critical error has occurred and yield(0) is called then yield()
// will not call longjmp() but will call FatalExit() instead.
//
// ONLY CALL yield(0) WHEN THE PROGRAM CAN SAFELY BE ABORTED!
//
//  yield() could be expanded to save the program state and thus facilitate
// the time-slicing neccessary for a background DEFRAG to work.
//
jmp_buf fileJmpBuf;

// Allocates a 64K I/O buffer for absread/write. When using RTL read/write
// routines the extra bytes will go unused since they impose a limit of
// 65535 bytes of I/O max (just like DOS).
//
void InitFileIO(void)
{
   // the RTL memory allocators use the first 4 bytes of a block for their
   // own purposes. We use farcalloc() which can allocate blocks > 64K but,
   // because the minimum offset is 4, we lose access to a whole sector unless
   // we rely on huge pointers to make up for the 4 bytes of offset overflow.
   // By allocating 12 extra bytes we can use all of the last paragraph in
   // the block, get a full 64K buffer and allow it to be paragraph aligned.
   // Note that (as a practical matter) we don't really need to ask for the
   // extra 12 bytes because the RTL has DOS paragraph granularity. Doing so,
   // however, makes our pointer adjustment "correct".
   //
   bufferSize    = 65536L;
   bufferSectors = (ushort)(bufferSize >> bytesToSectors);
   bufferDirEntries = (ushort)(bufferSize >> DIREntriesToBytes);
   lpSectorBuffer= (uchar far*)DefragAlloc((ulong)(bufferSize+12L), 1L);
   lpSectorBuffer= (uchar far*)(MK_FP(FP_SEG(lpSectorBuffer)+1,0));

}


// The file oriented disk defrag routine. This does not compact disks
// or defrag directories but will recurse dirs looking for files.
// This method needs at least as much continuous free space as the
// largest file on disk in order to work.
//
void PASCAL DoDefrag(void)
{ ushort j, k, clustersNeeded, clustersStart, newClustersStart;
  ushort bytes;
  short  srceHandle, destHandle, thisFilesDiscontinuities;
  bool   wasFound;
  FIND found;
  LPSFT lpSFT;
  long	filePos;
  ulong  fileLeft;

   ASSERT(lpSectorBuffer != 0L);

   // Always set the DTA on entry so we can recursively process
   // the dir tree
   //
   SetDTA((void far*)&found);

   // We exclude devices, volume labels and (hidden | system | read-only) files
   // While anything matches...
   //
   wasFound = FindFirst("*.*", ~(A_DEVICE | A_VOLUME | A_HIDDEN | A_SYSTEM));

   while(wasFound && (discontinuities > 0))
   {
      yield(0);

		if(found.attr & A_READONLY)
			goto NEXT_ONE;

		AddName(pathName,found.name);

      // Process non-zero length files and subdirs
      //
      if(found.attr & A_SUBDIR)
      {	if(found.name[0] != '.')
			{
	         ChangeDir(found.name);
   	      DoDefrag();
      	   SetDTA((char*)&found);
			}
      }
      else
      {

#if DEBUG
			printf("%s: ",pathName);
#endif

			thisFilesDiscontinuities = 0;
         clustersNeeded  = (ushort)(found.size >> bytesToClusters);
         clustersNeeded += (found.size & partialClusterBytesMask) ? 1 : 0;
         if(clustersNeeded > 1)
         {

#if DEBUG
			printf("source, ");
#endif
            // change file access writes to allow write/delete
            // open the matching file for reading
            //
            if(found.attr != A_NORMAL)
               SetFileAttributes(found.name, A_NORMAL);
				k = ACT_EXIST_OPEN;
            if(!OpenOrCreateFile(found.name,
				   				  O_RDONLY,
									  A_NORMAL,
									  &k, &srceHandle))
					FatalError(EL_DISK_READ_ERR);

				// find the SFT from the handle to find out the
				// starting cluster of the file
				//
            lpSFT          = HandleToSFT(srceHandle);

            SHIPASSERT(lpSFT != 0);

            clustersStart  = lpSFT->start;
            j              = clustersStart;
            ASSERT(clustersStart >= 2);

            // If the file is already defragmented we can skip it
            // else count its discontinuities
				//
            k = ReadFATInRAM(j);
            while(!IS_FAT_EOF(k))
            {  if(k != j+1)
						++thisFilesDiscontinuities;
               j = k;
               k = ReadFATInRAM(j);
            }

				// if fragmented...
				//
            if(thisFilesDiscontinuities > 0)
            {
               // allocate a new region of the FAT
               //
               k = clustersNeeded;
               newClustersStart = GetFirstFit(&k);

               // If the file is too big to defrag... what can we do?
               //
               if((newClustersStart == 0) || (k != clustersNeeded))
               {
						++filesSkipped;
                  WarnUserAboutFile(found.name);
                  goto AVOID_FILE;
               }
					fileLeft = found.size;

#if DEBUG
			printf("dest, ");
#endif
               // Tell DOS to write the new file to the newly allocated
               // region of the FAT then open the defrag file
               //
               lpDPB->newclus = newClustersStart-1;
					k = ACT_EXIST_REPLACE | ACT_NEXIST_CREATE;
            	if(!OpenOrCreateFile(throwAwayName,
				   				  	(O_WRITENOW | O_RDWR),
									  	A_NORMAL,
									  	&k, &destHandle))
					{
               	FreeFATChainInRAM(newClustersStart);
						++filesSkipped;
                  WarnUserAboutFile(found.name);
                  goto AVOID_FILE;
					}

               // rewind the file and copy it using as big a buffer as we can.
               //
					filePos = 0;
               SeekFile(srceHandle,&filePos,SEEK_SET);
               do
               {
                  bytes = (fileLeft > 0xFFFFL) ? 0xFFFF : (ushort)(fileLeft);
						if(DefragRead(srceHandle,lpSectorBuffer,&bytes))
                   	if(DefragWrite(destHandle,lpSectorBuffer,&bytes))
								fileLeft -= bytes;
               }
               while(bytes == 0xFFFF);

					// Set the copied file to the original file's date/time
					//
					SetFileDateAndTime(destHandle, found.date, found.time);

               // Close the new and old file. Rename new file to old file.
               //
               CloseFile(srceHandle);
               CloseFile(destHandle);
               DeleteFile(found.name);
               Rename(throwAwayName, found.name);
#if DEBUG
			printf("done\n");
#endif

               // add the old file's clusters to our internal free list
               //
               FreeFATChainInRAM(clustersStart);

               // build our internal FAT chain and back links for the new file
               //
               for(k =0, j = clustersNeeded-1; j > 0; j--, k++)
               {
                  WriteFATInRAM(newClustersStart+k, newClustersStart+k+1);
                  BackwardFAT[newClustersStart+j] = newClustersStart+j-1;
               }
               WriteFATInRAM(newClustersStart+clustersNeeded-1, FAT_EOF);
               BackwardFAT[newClustersStart] = FAT_EOF;
// !!!
//
// so what if we do this, how do we get from a dir entry to an
// SFT to update the dir entry of an open file when defragging
// the dir...?
//
//             lpSFT->start = newClustersStart;

					// we actually defragged this file
					//
					discontinuitiesFixed += thisFilesDiscontinuities;

            }
            else
				{
#if DEBUG
			printf("contiguous\n");
#endif
AVOID_FILE:
               CloseFile(srceHandle);
				}

            // restore file access mode
            //
            SetFileAttributes(found.name,found.attr);
         }
			// don't look for more disconinuities than there are,
			// even if we can't fix this one particular set
			//
			discontinuities -= thisFilesDiscontinuities;
      }

		RemoveName(pathName);

NEXT_ONE:
      wasFound = FindNext();
   }
   ChangeDir("..");
}
