//
//   $Id: DLDELTRE.C 1.7 2002/10/16 21:45:44Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// DLDELTRE.C  Created 05-01-97 Dennis Edwards
//
// Copyright (c) 1989-2002 Datalight, Inc.
//	All Rights Reserved.
//	Datalight Confidential and Proprietary
//
// This modules provides a DeleteTree() function to delete
// entire directory trees in one call. It relies on Recurse()
// and WildUnlink() from the standard DLx.LIB.
//
//  History:
//
//  01-30-1999 bt  Added dostruct.h include
//


//  Datalight libraries
//
#include "stdc.h"
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlpath.h"
#include "dlrecurs.h"
#include "dlerrcod.h"


//  Contains a pointer to the Global LFN search name in the datalight
//  libraries
//
extern char * szLFN;


//  Globals for knowing when we are in a FAT32 or LFN environment
//
extern bool fIsFat32World;
extern bool fLFNsSupported;


//  Delete a directory and all of its contents
//
bool DeleteTree(char * szTargetSpec)
{
	bool fSuccessful = TRUE;
	FIND Find;
	bool fWasFound;


	AddName( szTargetSpec, "*.*" );


	//  Search for the first occurance
	//
	fWasFound = SmartFindFirst( (A_ANYTHING & ~A_VOLUME),
			szTargetSpec, &Find );
	RemoveName( szTargetSpec );
	while(fWasFound)
	{
		//  if so make sure we don't process "." and ".." in subdirs
		//
		if(*(Find.name) == '.')
		{
			//  Try to find the next wild card match...
			//  (if none specified we'll just drop out of this loop)
			//
			fWasFound = SmartFindNext( &Find );
			continue;
		}


		//  Add the found name to the path
		//
		AddName( szTargetSpec, Find.name );


		if((Find.attr & (A_HIDDEN|A_SYSTEM|A_READONLY)) != A_NORMAL)
		{
			SmartSetFileAttributes( szTargetSpec, A_NORMAL );
		}


		//  use different routines for trees and single files
		//
		if((Find.attr & A_SUBDIR) == A_SUBDIR)
		{
			if( !DeleteTree(szTargetSpec) )
			{
				fSuccessful = FALSE;
			}
		}
		else
		{
			if( !SmartDelete(szTargetSpec) )
			{
				fSuccessful = FALSE;
			}
		}


		//  Remove the found name from the path
		//
		RemoveName( szTargetSpec );


		// Try to find the next wild card match...
		// (if none specified we'll just drop out of this loop)
		//
		fWasFound = SmartFindNext( &Find );
	}


	//  Close the find handle
	//
	SmartFindClose( &Find );


	if( !SmartRemoveDirectory(szTargetSpec) )
	{
		fSuccessful = FALSE;
	}


	return fSuccessful;
}
