//
//   $Id: DISK2IMG.C 1.5 2002/10/16 21:46:16Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************

	DISK2IMG.C  created 10/06/94  by Drew Gislason

	Copyright (c) 1994-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary

	Converts a full disk into an image.  Disk can be a
	floppy or RAM or ROM disk.

	Usage = DISK2IMG <d:> <filespec>
	Examples
		C>DISK2IMG a: flop.img
		C>DISK2IMG e: C:\TEMP\compress.bin

	History

	06-14-00  td  Added LFN support

********************************************************/
#include <ctype.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <dir.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dosver.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlapp.h"

#include "fatworld.h"
#include "abortlfn.h"
#include "disk2img.pt"

#define DISK2IMG 1
#include "dosmsg.c"


char signon[] = "DISK2IMG %s %s\n%s\n\n";

bool fIsFat32World;
bool fHideSignon = FALSE;

int iFileHandle;


void DisplaySignOnMessage(void)
{
	if( !fHideSignon )
	{
		if( fIsFat32World )
		{
			printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}

void Usage(void)
{
	DisplaySignOnMessage();
	printf( szUsage );
	exit(1);
}

int currentdrv(void)
{
	unsigned char drive;
	drive = bdos(0x19,0,0);
	return drive;
}

unsigned long freespace(int drive)
{
	unsigned long free = 0L;
	union REGS r;

	r.h.ah = 0x36;
	r.x.dx = drive + 1;
	intdos(&r,&r);
	if(r.x.ax != 0xFFFF)
		free = (unsigned long)r.x.bx * r.x.ax * r.x.cx;

	return free;
}

unsigned char buff[(unsigned)32*1024];		// large chunk at a whack

int main(int argc, char *argv[])
{
	int err, iActionTaken;
	int drive, destdrv;
	char *fname;
	LPBPB lpbpb;
	unsigned sec_size, count;
	unsigned long sectors, startsec;
	int secs_at_a_pop, n;
	bool fSuccessful;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	//  Parse the command line
	//
	argc = DoParseCommandLine( argc-1, &argv[1] );


	//  Display the program signon message
	//
	DisplaySignOnMessage();


	// determine drive
	drive = toupper(*argv[1]) - 'A';
	fname = argv[2];

	// determine size of disk image
	err = absread(drive, 1, 0L, buff);
	if(err)
	{
		printf(read_fail_msg, drive+'A');
		return 2;
	}
	lpbpb = (LPBPB)(buff+11);
	sec_size = lpbpb->sec_size;
	sectors = (lpbpb->num_sec ? lpbpb->num_sec : lpbpb->lnumsec);
	if(!sectors || !sec_size)
	{
		printf(bpb_err_msg, drive+'A');
		return 3;
	}

	// make sure there's enough free space on destination drive
	printf(create_msg,fname,sectors * sec_size,drive+'A');
	if(fname[1] == ':')
		destdrv = toupper(*fname) - 'A';
	else
		destdrv = currentdrv();
	if(freespace(destdrv) < sectors * sec_size)
	{
		printf(not_enough_space_msg, destdrv + 'A');
		return 4;
	}

	//  Create or truncate the destination (output) file
	//
	fSuccessful = SmartCreateOpenFile(OPEN_ACCESS_READWRITE, A_NORMAL,
			FILE_TRUNCATE | FILE_CREATE, fname,	&iFileHandle, &iActionTaken);
	if(!fSuccessful)
	{
		printf(cannot_create_msg, fname);
		exit(5);
	}

	// write out the whole image in big chunks
	secs_at_a_pop = sizeof(buff) / sec_size;
	printf("  ");
	for(startsec = 0L; startsec < sectors; startsec += n)
	{
		n = secs_at_a_pop;
		if(sectors-startsec < secs_at_a_pop)
			n = (unsigned)(sectors-startsec);
		err = absread(drive, n, startsec, buff);
		if(err)
		{
			printf(read_fail_msg, drive);
			return 10;
		}

		//  write out the sectors
		//
		count = sec_size * n;
		fSuccessful = dlWriteFile(iFileHandle, &count, (char *)(buff) );
		if((!fSuccessful) || (count != (sec_size * n)) )
		{
			//  We should close our open files before erroring out
			//
			dlCloseFile(iFileHandle);

			printf(write_fail_msg, fname);
			return 11;
		}
		printf(".");
	}


	//  Close the destination (output) file
	//
	dlCloseFile(iFileHandle);

	return 0;
}


//  Process all the command line arguments
//
int DoParseCommandLine( int argc, char *argv[] )
{
	uchar ucArgIndex;
	static ARGDEF DISK2IMGOpts[] =
	{
		{ BOOL_SWITCH, &fHideSignon,      '[', 0 },
		{ 0 }
	};


	//  Remove the quotes from any LFN arguments
	//
	for(ucArgIndex=0; ucArgIndex < argc; ucArgIndex++)
	{
		LFNStripArgQuotes( argv[ucArgIndex] );
		if( strlen(argv[ucArgIndex]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(255);
		}
	}


	//  This is for the command line parsing library function.  It
	//  indicates that no command line options are ok
	//
	SetParseFlags( PARSEF_NOARGSOK );


	//  This is for the command line parsing library function.  It
	//  indicates that no command line options are ok
	//
	SetParseFlags( PARSEF_NOARGSOK );


	//  Process all the boolean switches in the command line
	//
	argc = ParseCmdline(argc, &argv[0], DISK2IMGOpts);


	//  If there is anything, it should only be a drive or
	//  path/name combination.
	//
	if( argc != 2)
		Usage();


	return argc;
}
