//
//   $Id: EXE2BIN.C 1.4 2002/10/16 21:48:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/**********************************************************

	Exe2bin Utility

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
	Datalight Confidential and Proprietary

***********************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <dir.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "exehdr.h"
#include "dosver.h"
#include "dldosio.h"
#include "dllfn.h"
#include "fatworld.h"

char USAGE[] = "EXE2BIN %s %s\n%s\n\n\tUsage = EXE2BIN infile [outfile]\n";

#define EXE2BIN 1

struct exehdr exehdr;
char buffer[512];
bool fIsFat32World;

int input, output, read(int, void *, unsigned);
unsigned count;
long l, lseek(int, long, int);

extern int _doserrno;

#if 0
#include "error.c"

#else
/*
 * Error message and display for all ROM-DOS utilities.
 */

#include <stdio.h>
#include <stdarg.h>
#include <assert.h>

void error(int, ...);

char *msgs[] =
{
			"Msg 0",
#if COMP					/* Compare utility */
			"Cannot open '%s'",
			"Files are different sizes",
			"Files different at %05lX '%02X' and '%02X'",
			"Files are Completely different",
#endif


#if EXE2BIN					/* Exe2bin utility */
		"Cannot open '%s'\n",
		"Read error on input file\n",
		"'%s' not an EXE file\n",
		"'%s' cannot have any relocatable items\n",
		"'%s' starting address not 0 or 0x100\n",
		"Cannot creat '%s'\n",
#endif
};

void error(int msg, ...)
{
	va_list arg_ptr;
	va_start(arg_ptr, msg);
	if(msg < 0 || msg > (sizeof(msgs)/sizeof(msgs[0])))
	{
		printf("error msg = %d\n", msg);
		assert(0);
	}
	vfprintf(stderr, msgs[msg], arg_ptr);
	fprintf(stderr, "\n");
}
#endif

#if 0
#include "mk_ext.c"
#else
mk_ext(char *name, char *ext)
{
	char *p = name + strlen(name);
	int i;
	for(i=4; i >= 0; --i){
		if(*p == '.'){
			++p;
			strcpy(p, ext);
			return 0;
		}
		if(*p == '\\')
			break;
		--p;
	}
	strcat(name, ".");
	strcat(name, ext);
	return 0;
}
#endif

void ShowUsage( void )
{
	if( fIsFat32World )
	{
		printf(USAGE,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(USAGE,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
	exit(1);
}

main(int argc, char *argv[])
{
	int iActionTaken, i;
	long lOffset;
	char fname[LFN_MAX_PATH];
	char infile[LFN_MAX_PATH];

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	for(i=1; i <= argc; i++)
	{
		LFNStripArgQuotes( argv[i] );
		if( strlen(argv[i]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(255);
		}
	}

	if(argc < 2)
	{
		ShowUsage();
	}

	if( strchr( argv[1], '?' ) != NULL )
	{
		ShowUsage();
	}

	strcpy(infile, argv[1]);


	SmartCreateOpenFile(OPEN_ACCESS_READONLY, 0, FILE_OPEN, infile, &input,
		&iActionTaken);
	if(!input)
	{
		mk_ext(infile, "exe");
 		SmartCreateOpenFile(OPEN_ACCESS_READONLY, 0, FILE_OPEN, infile, &input,
			&iActionTaken);
		if(!input)
		{
			error(1, infile);
			exit(1);
		}
	}

	count = sizeof(exehdr);
	dlReadFile(input, &count, (char *)&exehdr);
	if(count != sizeof(exehdr))
	{
		error(2);
		exit(1);
	}

	if(exehdr.sig != 0x5a4d){		/* If not an EXE */
		error(3, argv[1]);
		exit(1);
	}
	if(exehdr.nreloc != 0){			/* If any Relocatable items */
		error(4, argv[1]);
		exit(1);
	}

	if(!(exehdr.ipofs == 0 || exehdr.ipofs == 0x100)){
		error(5, argv[1]);
		exit(1);
	}

	lOffset = (exehdr.hdrlen*16L) + (long) exehdr.ipofs;
	dlSeekFile(SEEK_SET, input, &lOffset);

	if(argc == 2)
	{
		strcpy(fname, argv[1]);
		mk_ext(fname, "COM");
	}
	else
		strcpy(fname, argv[2]);

	SmartCreateOpenFile(OPEN_ACCESS_READWRITE, 0, FILE_TRUNCATE | FILE_CREATE, fname, &output,
		&iActionTaken);
	if(output == -1)
	{
		error(6, fname);
		exit(1);
	}

	while(count)
	{
		count = sizeof(buffer);
		dlReadFile(input, &count, buffer);
		dlWriteFile(output, &count, buffer);
	}

	dlCloseFile(input);
	dlCloseFile(output);
	return 0;
}
