//
//   $Id: FDISK.C 1.12 2003/09/02 17:45:48Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:FDISK.C

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	FDISK creates, deletes, displays and otherwise manages partitions
	on a harddisk (both DOS and non-DOS). All access to the hard disk
	is accomplished through Int 13h.

	New Proposal:
	-------------
	Usage = FDISK [BiosDrive] [/3] [/A|/C|/MBR|/P[#]|/R|/V] [/N] [/S#] [/?]
	BiosDrive is 80h for the first hard disk, 81h for the 2nd hard disk, etc...
	Options
		/3    Enable support to create FAT32 partitions
		/A    Delete all partitions
	   /C    Create partition
	   /MBR  Don't make partition. Refresh Master Boot Record code only
	   /P[#] Create SuperBoot partition (default ID 98h)
		/R    View partitions(s), raw format
	   /V    View partition(s)
		/N    Don't write MBR code for Super-Boot partitions
	   /S#   Maximum partition size in megabytes. The default (no /S) is to
	   		create a single FAT32 partition, or as many 2 Gig DOS partitions
				as required to fill the disk if FAT32 support not enabled.
	   /?    This help screen

		Examples:
		FDISK 80 /R      Refresh MBR code on drive 80h
		FDISK /P /S5     Create a 5 Megabyte SuperBoot partition on drive 80h
		FDISK 81 /3 /S   Create a FAT 32 partition, as large as possible

	Existing Usage:
	---------------
	Usage = FDISK [drive] [/3] [/C] [/I# [/N]] [/B|/R|/V] [/S#] [/?]
	Options
	   /3   Enable Fat32 mode
	   /B   Make partition. Force update of Master Boot Record code
	   /C   Don't confirm (no operator input)
	   /I#  Make Super-Boot partition with this ID (default 98h)
	   /N   Don't write MBR code for Super-Boot partitions
	   /R   Refresh Master Boot Record code only
	   /S#  Partition size in megabytes (default: large as possible)
	   /V   View partitions and exit
	   /?   This help screen

	Examples:

	FDISK 80 /R       Refresh MBR code on drive 80h
	FDISK /I98 /S5    Create a 5 Megabyte SuperBoot partition
	FDISK 80 /3 /S0\t\tCreate a FAT 32 partition, as large as possible\


	Assumptions:
	1. sectors are always 512 bytes.
	2. only 1 extended DOS partition per hard disk (the one extended
     partition may have many extended partitions inside of it).
   3. Supports up to 24 hard disks (C:-Z:)

	FDISK MBR Partition Rules:
	1. The Master Boot Record (MBR) and extended partitions have different
		rules.
	2. MBR allows up to 4 partitions
	3. DOS partitions (including SuperBoot) begin on head 1. Extended
		partitions begin on head 0. If there is no primary DOS extension,
		but there is an extended partition, it will begin on cylinder 1,
		not cylinder 0, in order to begin on head 0.
	4. DOS drive letters are calculated in the following way:
		a. All bootable DOS partitions, all drives are first
		b. Next all non-bootable primary DOS partitions, all drives
		c. All extended partitions, all drives.
	5. The CHS fields are limited to 8 bits for the head, 6 bits for the
		SPT, and 10 bits for the cylinder. Heads and cylinders are 0 based,
		while SPT is 1 based. That limits heads to 256 (0-FFh), cylinders
		to 1024 (0-3FFh) and SPT to 63 (1-3F). The top 2 bits of the cylinder
		are in the top 2 bits of the SPT byte. For example, a CHS of C=2FEh,
		H=99h, S=3Fh would be represented with the bytes 99h,BFh,FEh.
	6. The terms cylinder and track are used synonimously in this program.
	7. The LBA boundary is just below 8 Gigabytes (1024 * 256 * 63), but
		it may be lower due to drive geometry.
	8. The CHS bytes fields in a partition record are limit checked in the
		MBR (they wrap in extended partitions if they extended beyond the
		LBA boundary. For example, if the ending cylinder is 1029 (405h), it
		will be limit checked to 1023 (3FFh).
	9. If any partition in the MBR crosses the LBA boundary, the LBA ID
		is used. See DOSTRUCT.H for a list of possible partition IDs.
	10. The luHidden field in a partition record is relative to sector 0.
	11. extended partitions can't begin on cylinder 0.

	FDISK Extended Partition Rules:
	1. The Master Boot Record (MBR) and extended partitions have different
		rules.
	2. The extended partition in the MBR is called the Master Extended
		Partition.
	3. The CHS bytes fields in a partition record wrap in extended
		partitions (they are limit checked in the MBR). For example, if the
		ending cylinder is 1029 (405h), it is wrapped to just 5 (0Fh). A
		CHS for C=405h, H=99h, S=3Fh would be 99h, 3Fh, 05h.
	4. Only the Master Extended Partition may use the LBA ID. All other IDs
		in extended partitions use the non-LBA IDs, both for DOS partitions
		and subsequent extended partitions.
	5. The luHidden field inside extended partitions is relative to the
		Master Extended Partition for extended partitions, but relative to
		the current partition for DOS partitions.
	6. The end of each extended partition goes is the end of the Master
		Extended Partition.

	Other FDISK Rules:
	1. Leave partition records alone, unless deleting or creating.
	2. Update MBR code only if code not valid (or unless creating SuperBoot
		partitions). Test is the 55AAH at end of MBR sector.
	3. When creating a DOS partition, zero FAT, root DIR and boot sector.

	$DOCHISTORY
	04/30/1999 dg  Fixed numerous bugs:
						Drive limit check was off by 1 for cmdline mode.
						Super-Boot partitions now limited to 2 Gig
	04/23/1999 dg  Updated cmdline options to match interactive mode
	04/05/1999 dg  Added FAT32 support
	03/31/1999 dg  Added LBA support
	09/24/1998 dg  Fixed "ask" bug with /C on writer over partition
	09/24/1998 dg  /V from cmdline now views all drives
   09/24/1998 dg  Fixed create so it shows disk AFTER created fm cmdline
	09/21/1998 dg  Super-Boot ID now recorded in Master Boot Sector
	08/27/1998 dg  Added Super-Boot capability
	08/10/1990 dg  Created by Drew Gislason
*********************************************************************/
#define __FILENUM__ 1

#include <ctype.h>
#include <dos.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dlapp.h"
#include "dlint13.h"
#include "dosver.h"
#include "dlmath.h"
#include "fatworld.h"

#include "fdisk.h"
#include "hboot.h"
#include "fdisk.pt"

//  Conditional compile switch for supporting FAT32 drives
//
#ifdef ENABLEF32
#define SUPPORT_FAT32_DRIVES		TRUE
#else
#define SUPPORT_FAT32_DRIVES		FALSE
#endif

#define FDISK 1
#include "dosmsg.c"

/* FDISK menu options */
struct menu menu[] =
{
	{ VIEW_CHAR,   szViewPartitions, ViewPartitions },
	{ RAW_CHAR,    szDisplayRawPartitions, DisplayRawPartitions },
	{ CREATE_CHAR, szCreatePartition, CreatePartition },
	{ SUPER_CHAR,  szCreateSuperPart, CreateSuperPartition },
	{ DELETE_CHAR, szDeletePartition, DeletePartition },
	{ DELALL_CHAR, szDeleteAllPartitions, DeleteAllPartitions },
	{ BOOT_CHAR,   szToggleBoot, ToggleBoot },
	{ CODE_CHAR,   szRewriteMBRCode, RewriteMBRCode },
	{ QUIT_CHAR,   szQuitButDontSave, QuitButDontSave },
	{ SAVE_CHAR,   szSaveAndExit, SaveAndExit },
	{ 0 }
};

char *pBufferDMA;				/* working INT 13H disk buffer */
uchar berr;						/* last disk error */
char zz[100];					/* a VERY temporary string */

/* a far function pointer to FFFF:0 (to reboot) */
typedef void (far * FUNCPTR)(void);
int (far *reboot)() = MK_FP(0xFFFF,0);

/* general things about all the fixed disks */
char	 startletter = 'C';			/* starting letter for 1st partition */
int    iNumHardDisks, iNumDosParts;	/* maximum fixed disk */
int    iMaxDosParts;
bool   fChanged;						/* has ANYTHING been changed */
struct HardDisk hdisk[MAXFIXED];	/* describe each  fixed disk */
int	 iPrevSuperBootID = DEFAULT_SUPERBOOT_ID;		/* previous ID used */
int	 iCmdlineDrive;
int    iLineCount;
#define MAX_LINES 25

/* cmdline switches */
bool swFat32 = -1;			/* /3    */
bool swBootCode;				/* /B    */
bool fDontConfirm;			/* /C    */
int  iSuperBootID = -1;		/* /I[#] */
bool swNoSuperMBR;			/* /N    */
bool swRefreshCode;			/* /R    */
long lPartMegs = -1L;		/* /S[#] size in megabytes */
bool swViewOnly;				/* /V    */
bool swDeleteAll = FALSE;

ARGDEF argDefs[] =
{
	{ BOOL_SWITCH, &swFat32,			'3' },
	{ BOOL_SWITCH, &swBootCode,      'B' },
	{ BOOL_SWITCH, &fDontConfirm,		'C' },
	{ HEXINT_SWITCH,  &iSuperBootID, 'I'},
	{ BOOL_SWITCH,  &swNoSuperMBR,   'N'},
	{ BOOL_SWITCH, &swRefreshCode,   'R' },
	{ LONG_SWITCH, &lPartMegs,       'S' },
	{ BOOL_SWITCH, &swViewOnly,      'V' },
	{ 0 }
};

bool  fIsFat32World;

/*
	$DOCPROC:Usage

	Display how to use FDISK.
*/
void Usage(void)
{
	if( fIsFat32World )
	{
		printf(USAGE_P1);
		printf(USAGE_P2);
		printf(USAGE_P3);
		printf(USAGE_P4);
	}
	else
	{
		printf(USAGE_P1);
		printf(USAGE_P3);
	}
	exit(EXIT_USAGE);
}

void DisplaySignOnMessage( void )
{
	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/*
	FDISK, here is the main event loop
*/
void main(int argc,char *argv[])
{
	int i, nParms;
//	char c;
	FUNC_PTR func;
	int	iBiosDrive;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
#if SUPPORT_FAT32_DRIVES
	fIsFat32World = dlIsFat32World();
#else // #if SUPPORT_FAT32_DRIVES
	fIsFat32World = FALSE;
#endif // #if SUPPORT_FAT32_DRIVES

	/* display sign-on */
	DisplaySignOnMessage();

	/* parse the cmdline */
	SetParseFlags(PARSEF_NOARGSOK);
	nParms = ParseCmdline(argc-1, argv+1, argDefs);

	if(nParms > 1)
	{
		if(stricmp( argv[2], "[deleteall]") == 0 )
		{
			nParms--;
			argv[2] = argv[3];
			swDeleteAll = TRUE;
		}
	}

	if(nParms > 1)
		Usage();

	/* Make sure that mutually incompatible options aren't selected. */
	i = 0;
	if(swBootCode)			// /B
		++i;
	if(swRefreshCode)		// /R
		++i;
	if(swViewOnly)			// /V
		++i;
	if(swDeleteAll)		// [deleteall]
		++i;

	if( !fIsFat32World && swFat32 && (swFat32 != -1))
	{
		printf(szErrNotFAT32World);
		exit(255);
	}

	/* at most, one of these options allowed */
	if(i > 1)
	{
		printf(szErrIncompatibleOptions);
		exit(EXIT_USAGE);
	}

	/* determine the # of physical hard disks */
	iNumHardDisks = NumFixedDisks();
	if(!iNumHardDisks)
	{
		printf(szErrNoFixedDisks);
		exit(EXIT_NODISKS);
	}

	/* make sure buffer does not cross DMA boundary. BIOS requires this */
	pBufferDMA = AllocDMACompatible(SECSIZE);
	if(!pBufferDMA)
	{
		printf(szErrNoMem);
		exit(EXIT_NODISKS);
	}

	/* view partitions on all drives */
	if(swViewOnly)
	{
		//
		//  Note: Always viewed with large disk support
		//

		/* read in all partitions on all drives */
		iNumDosParts = ReadPartitions();
		iMaxDosParts = 1 + ('Z' - startletter);

		fDontConfirm = TRUE;
		ViewPartitions();
		exit(0);
	}

	/* create partitions, and other automated stuff */
	if(nParms == 1)
	{
		/* user didn't specify /3 switch, so no FAT32 */
		if(swFat32 == -1)
			swFat32 = FALSE;


		//  Determine whether the smart functions can only see the
		//  CHS part of the drive
		//
		if( !swFat32 )
		{
			DlSmartSetCHSLimitFlag( TRUE );
		}


		/* read in all partitions on all drives */
		iNumDosParts = ReadPartitions();
		iMaxDosParts = 1 + ('Z' - startletter);

		/* make sure drive is valid */
		iBiosDrive = atoh(argv[1]);
		iCmdlineDrive = (iBiosDrive - HD1);
		if(iCmdlineDrive < 0 || iCmdlineDrive >= iNumHardDisks)
		{
			printf(szErrInvalidDrive);
			exit(EXIT_BADDISK);
		}

		/*
			Doing one of 3 operations:
			1) Refresh boot code only /R
			2) Create a partition (either DOS or Super-Boot) /S#
			3) Create a partition, with forced updated of boot code /B
		*/

		if(swDeleteAll)
		{
			DeleteAllPartitions();

			/* display the partitions for this drive */
			DisplayPartitions(iCmdlineDrive);

			/* now write out the partition */
			SaveAndExit();
		}

		// update boot code only
		else if(swRefreshCode)
		{
			ReadSector(iCmdlineDrive, 0, pBufferDMA);
			memmove(pBufferDMA, array, PARTOFF);
			WriteSector(iCmdlineDrive, 0, pBufferDMA);
		}

		// create a partition
		else
		{
			/* create the partition */
			if(iSuperBootID != -1)
				CreateSuperPartition();
			else
				CreatePartition();

			/* always write boot code? */
			if(swBootCode)
				hdisk[iCmdlineDrive].fCodeOK = FALSE;

			/* display the partitions for this drive */
			DisplayPartitions(iCmdlineDrive);

			/* now write out the partition */
			SaveAndExit();
		}

		/* done, everything worked */
		exit(0);
	}

	if( fIsFat32World )
	{
		/* does user want FAT32 mode enabled? */
		if(swFat32 == -1)
			swFat32 = (GetYesNoWithDefault(szEnableFat32, YES) == YES);
	}
	else
	{
		swFat32 = FALSE;
	}


	//  Determine whether the smart functions can only see the
	//  CHS part of the drive
	//
	if( !swFat32 )
	{
		DlSmartSetCHSLimitFlag( TRUE );
	}


	/* read in all partitions on all drives */
	iNumDosParts = ReadPartitions();
	iMaxDosParts = 1 + ('Z' - startletter);


	/* disable ctrl-break (restored on program exit) */
	//  Install a critical error handler so that we can abort cleanly
	//
	InstallInt23Handler();

	/* display the main menu */
	while(1)
	{
		/* display the menu and get response */
		func = GetMenuCommand(menu);
		(*func)();
	}
}

/*
	$DOCPROC:ViewPartitions

	Menu Command.

	Display all the partitions for this drive.
*/
void ViewPartitions(void)
{
	int iDrive;

	for(iDrive=0; iDrive<iNumHardDisks; ++iDrive)
		DisplayPartitions(iDrive);
	GetEsc();
}

/*
	$DOCPROC:DisplayRawPartitions

	Menu Command.

	Display all partitions, all drives, in a simple raw format.
*/
void DisplayRawPartitions(void)
{
	int	iDrive;

	/* determine for which drive */
	if((iDrive = GetDriveNumber(szViewWhichDrive)) == CANCEL)
		return;

   iLineCount = 0;

	/* actually do the display */
	DumpRawPartitions(iDrive);

	/* done, wait for a key */
	GetEsc();
}

/*
	$DOCPROC:DumpRawPartitions

	This displays the raw partition chain.
*/
void DumpRawPartitions(int iDrive)
{
	struct ExtPart *ep;
	int	i;
	char	szStr[10];
	int	iPart;
	uchar ucTemp;


	/* display the size of the disk */
	DisplayDiskSize(iDrive);
   iLineCount++;

	/* display each partition of the disk (up to 4) */
	for(i=0; i<NUMPART; ++i)
	{
		GetPartitionName(szStr, iDrive, &hdisk[iDrive].pPartition[i]);
		printf("%s ",szStr);

		Dump(hdisk[iDrive].pPartition+i, sizeof(struct PARTITION));
	}
	printf("\n");
	iLineCount++;

	/* walk the extended partition chain */
	ep = hdisk[iDrive].pepExt;
	iPart = 0;
	while(ep)
	{
		printf(szPartSect, TrackToSector(iDrive, ep->ulStartTrack));
		iLineCount++;
		for(i=0; i<2; ++i)
		{
			if(IsDosPartitionId(ep->pPartition[i].id))
				++iPart;
			GetPartitionName(szStr, iDrive, &ep->pPartition[i]);
			printf("%s ",szStr);
			Dump(ep->pPartition+i, sizeof(struct PARTITION));
		}
		iLineCount++;
      if( iLineCount + 4 > MAX_LINES )
      {
      	printf(szStrikeKey);
   		dlGetKey( (char *)(&ucTemp) );
         printf("\n");

         iLineCount = 0;
      }
		printf("\n");
		ep = ep->pepNext;
	}
}

/*
	$DOCPROC:CreatePartition

	Menu Command.

	Create one or more DOS partitions. Understands all forms of a DOS
	partition, including FAT32 and LBA partitions. Fills the system sectors
	(FAT and root DIR) with 00s, fills the boot sector with F6s.
*/
void CreatePartition(void)
{
	int	iDrive;
	ulong ulStartTrack, ulNumTracks;
	bool	fFirstTime;
	struct PARTITION * pPartition;
	uchar ucIndex;

	/* get the drive to create partitions on */
	if(fDontConfirm)
		iDrive = iCmdlineDrive;
	else if((iDrive = GetDriveNumber(szCreateWhich)) == CANCEL)
		return;

	/* create 1 or more partitions */
	fFirstTime = TRUE;
	while(1)
	{
		/* show user what has been created */
		if(!fDontConfirm)
			DisplayPartitions(iDrive);

		/* don't make too many partitions */
		if(iNumDosParts >= iMaxDosParts)
		{
			printf(szNoDrivesLeft);
			GetEsc();
			break;
		}

		/* any available space? */
		if(!GetAvailableTracks(iDrive, TRUE, &ulStartTrack, &ulNumTracks))
		{
	NoSpace:
			printf(szDiskFilled, iDrive+1);
			GetEsc();
			break;
		}

		// special case: if creating an extended partition, don't allow
		// start track of 0
		if(hdisk[iDrive].iDosParts && ulStartTrack==0)
		{
			++ulStartTrack;
			--ulNumTracks;
			if(!ulNumTracks)
				goto NoSpace;
		}

		/* 2nd time around, ask the user if s/he'd like to continue */
		if(!fFirstTime && GetYesNo(szCreateAnother) != YES)
			break;
		fFirstTime = FALSE;

		/* get start/size of partition */
		if(fDontConfirm)
			GetCmdlineSize(iDrive, lPartMegs, &ulNumTracks);
		else if(!GetPartitionSizeFromUser(iDrive, &ulNumTracks))
			break;

		/* if no Primary Partition, create that first */
		if(hdisk[iDrive].iDosParts == 0)
		{
			pPartition = AddPartitionEntry(iDrive, hdisk[iDrive].pPartition,
				0, ulStartTrack, ulNumTracks,
				DeterminePartitionId(iDrive, ulStartTrack, ulNumTracks, NEW_STYLE));

			/* if error, no free entry */
			if(!pPartition)
				break;

			//  Clear any other bootable partition flag
			//
			for( ucIndex = 0; ucIndex < 4; ucIndex++ )
			{
				if(hdisk[iDrive].pPartition[ucIndex].boot)
				{
					hdisk[iDrive].pPartition[ucIndex].boot = 0;
					hdisk[iDrive].fModified[ucIndex] = TRUE;
				}
			}

			/* DOS partition always starts out as bootable */
			pPartition->boot = BOOTABLE_PART;

			/* mark as created */
			SetCreatedFlag(hdisk[iDrive].fModified,
					(unsigned)(pPartition-hdisk[iDrive].pPartition));
			++hdisk[iDrive].iDosParts;
		}

		/* we must get space from the extended partition */
		else
		{
			if(!CreateExtendedPartition(iDrive, ulStartTrack, ulNumTracks))
				break;
		}
		++iNumDosParts;

		/* don't loop if cmdline driven */
		if(fDontConfirm)
			break;
	}
}

/*
	$DOCPROC:CreateSuperPartition

	Menu Command.

	Create a Super-Boot partition. The ID can be anything, so this
	can be used to create non-DOS partitions as well.
*/
void CreateSuperPartition(void)
{
	int	iId;
	int	iDrive;
	ulong ulStartTrack;
	ulong	ulNumTracks;
	ulong	ulMax2Gig;
	struct PARTITION * pPartition;
	#define MIN_SUPER_ID		0x10
	#define MAX_SUPER_ID		0xFF

	/* get the drive to create nonDOS partitions on */
	if(fDontConfirm)
		iDrive = iCmdlineDrive;

	/* interactive mode, get drive from user */
	else
	{
		/* get drive from user */
		if((iDrive = GetDriveNumber(szCreateWhich)) == CANCEL)
			return;

		/* show user what has previously been created */
		DisplayPartitions(iDrive);
	}

	/* check for free space */
	if(!GetAvailableTracks(iDrive, FALSE, &ulStartTrack, &ulNumTracks))
	{
		printf(szDiskFilled, iDrive+1);
		GetEsc();
		return;
	}

	/* limit Super-Boot partitions to 2 Gigabytes */
	ulMax2Gig = GetMaxTracksFor2Gigabytes(iDrive);
	if(ulNumTracks > ulMax2Gig)
		ulNumTracks = ulMax2Gig;

	/* get the ID value (but don't allow DOS values) */
	if(fDontConfirm)
	{
		/* get ID from cmdline parameters */
		if(iSuperBootID == -1)
			iId = DEFAULT_SUPERBOOT_ID;
		else
			iId = iSuperBootID;

		/* make sure ID is valid (don't allow standard DOS IDs for a Super-Boot
			partition */
		if(iId < MIN_SUPER_ID || iId > MAX_SUPER_ID)
		{
			printf(szInvalidID);
			exit(EXIT_CREATE);
		}
	}

	/* get ID from user */
	else
	{
		iId = GetHexValue(szSuperID, iPrevSuperBootID, MIN_SUPER_ID, MAX_SUPER_ID);
		if(iId == CANCEL)
			return;
		if(   (iId != DEFAULT_SUPERBOOT_ID)
         && (iId != RELIANCE) )
			printf(szWarningID);
	}

	/* get start/size of partition */
	if(fDontConfirm)
		GetCmdlineSize(iDrive, lPartMegs, &ulNumTracks);
	else if(!GetPartitionSizeFromUser(iDrive,  &ulNumTracks))
		return;

	/* record previous ID */
	iPrevSuperBootID = iId;

	/* record ID in master boot sector */
	array[SUPERBOOTID_OFF] = iId;

	/* add the partition to the table */
	pPartition = AddPartitionEntry(iDrive, hdisk[iDrive].pPartition, 0,
		ulStartTrack, ulNumTracks, iId);
	if(!pPartition)
		return;

	/* mark as created */
	SetCreatedFlag(hdisk[iDrive].fModified,
			(unsigned)(pPartition-hdisk[iDrive].pPartition));
	++hdisk[iDrive].iSuperBootParts;

	/* show user what has previously been created now */
	DisplayPartitions(iDrive);

	/*
		by default, when adding a Super-Boot partition, always update the
		code in the MBR (user can override with the /N switch)
	*/
	if(!swNoSuperMBR)
		hdisk[iDrive].fCodeOK = FALSE;
}

/*
	$DOCPROC:DeletePartition

	Menu Command.

	Delete one or more partitions.  This can delete extended, DOS, and
	non-DOS partitions.
*/
void DeletePartition(void)
{
	int  iDrive;
	int  iPartition;
	char cLetter;
	struct PARTITION * pPartition;

	/* get the drive to delete partitions from */
	if((iDrive = GetDriveNumber(szDelPartOnDrive))==CANCEL)
		return;

	/* check for any partitions */
	if(!(hdisk[iDrive].iDosParts+hdisk[iDrive].iExtDosParts+hdisk[iDrive].iSuperBootParts))
	{
		printf(szDiskEmpty,iDrive+1);
		GetEsc();
		return;
	}

	/* display paritions on this drive */
	DisplayPartitions(iDrive);

	/* possibly delete a nonDOS partition */
	if(hdisk[iDrive].iSuperBootParts)
	{
		if(GetYesNo(szDelSuper) == YES)
		{
			DeleteSuperBootPart(iDrive);
			return;
		}
	}

	/* get the drive letter to delete */
	iPartition = GetPartitionLetter(iDrive, &cLetter);
	if(iPartition == CANCEL)
		return;

	/* give user a change to abort */
	sprintf(zz, szWarnDrive, cLetter);
	printf(zz);
	if(GetYesNo(szAreYouSure) == YES)
	{
		/* delete this extended partition */
		if(iPartition >= hdisk[iDrive].iDosParts)
			DeleteExtendedPartition(iDrive, iPartition - hdisk[iDrive].iDosParts);

		/* delete the primary DOS partition */
		else
		{
			/* delete the partition */
			--hdisk[iDrive].iDosParts;
			pPartition = FindPartitionType(hdisk[iDrive].pPartition, DOS16);
			if(pPartition)
			{
				memset(pPartition, 0, sizeof(struct PARTITION));
				SetChangedFlag(hdisk[iDrive].fModified,
						(unsigned)(pPartition-hdisk[iDrive].pPartition));
			}
		}

		/* give user a chance to digest this delete */
		--iNumDosParts;
		printf(szDeleted);
		GetEsc();
	}
}

/*
	$DOCPROC:DeleteAllPartitions

	Menu Command.

	Delete all the partitions on a given drive. Can also delete
	all partitions on all drives.
*/
void DeleteAllPartitions(void)
{
	int iDrive;
	int i;

	/* get drive to delete partitions from */
	if(fDontConfirm)
		iDrive = iCmdlineDrive;
	else if((iDrive = GetDriveNumber(szDeleteAll)) == CANCEL)
		return;

	if(!fDontConfirm)
	{
		/* check for any partitions */
		if(!(hdisk[iDrive].iDosParts+hdisk[iDrive].iExtDosParts+hdisk[iDrive].iSuperBootParts))
		{
			printf(szDiskEmpty,iDrive+1);
			GetEsc();
			return;
		}

		/* display paritions on this drive */
		DisplayPartitions(iDrive);

		/* if the user is sure, delete all partitions on this drive */
		sprintf(zz,szWarnDisk,iDrive+1);
		printf(zz);
		if(GetYesNo(szAreYouSure) == YES)
		{
			iNumDosParts -= hdisk[iDrive].iDosParts + hdisk[iDrive].iExtDosParts;

			/* delete any extended parts */
			ClearAllExtended(iDrive);
			ClearAllParts(iDrive);
			for(i=0; i<NUMPART; ++i)
				SetChangedFlag(hdisk[iDrive].fModified, i);

			/* give user a chance to digest this delete */
			printf(szDeleted);
			GetEsc();
		}
	}
	else
	{
		ClearAllExtended(iDrive);
		ClearAllParts(iDrive);
		for(i=0; i<NUMPART; ++i)
			SetChangedFlag(hdisk[iDrive].fModified, i);

		/* give user a chance to digest this delete */
		printf(szDeleted);
	}
}

/*
	$DOCPROC:RewriteMBRCode

	Menu Command.

	Instruct FDISK to rewrite the boot code when it saves and exits.
*/
void RewriteMBRCode(void)
{
	int iDrive;

	/* get the drive to write master boot code */
	if((iDrive = GetDriveNumber(szWriteMBRForDrive))==CANCEL)
		return;

	printf(szWriteMBR);
	hdisk[iDrive].fCodeOK = FALSE;
	SetChangedFlag(hdisk[iDrive].fModified, 0);
}

/*
	$DOCPROC:ToggleBoot

	Menu Command.

	Toggle whether there is a bootable drive or not.
*/
void ToggleBoot(void)
{
	int iDrive;
	struct PARTITION * pPartition;

	/* get the drive to delete partitions from */
	if((iDrive = GetDriveNumber(szToggleDrive))==CANCEL)
		return;

	pPartition = FindPartitionType(hdisk[iDrive].pPartition, DOS16);
	if(!pPartition)
		printf(szToggleNoDos);
	else
	{
		pPartition->boot = pPartition->boot ? 0 : BOOTABLE_PART;
		sprintf(zz,szToggleDos,iDrive+1,pPartition->boot ? "" : szUn);
		printf(zz);
		SetChangedFlag(hdisk[iDrive].fModified,
				(unsigned)(pPartition-hdisk[iDrive].pPartition));
	}
	GetEsc();
}

/*
	$DOCPROC:QuitButDontSave

	Menu Command.

	Quit without saving anything.
*/
void QuitButDontSave(void)
{
	if(fChanged)
	{
		if(GetYesNo(szQuit) != YES)
			return;
	}
	printf(szNoChange);
	exit(0);
}

/*
	$DOCPROC:SaveAndExit

	Menu Command.

	Save the changes to hard disk and exit (reboots) if there are
	changes.
*/
void SaveAndExit(void)
{
	char cTemp;


	if(fChanged)
	{
		/* the current disk has been modified, write it out */
		WritePartitions();

		if(fDontConfirm)
			printf(szSavedNoBoot);

		else
		{
			printf(szSaved);
   		dlGetKey( &cTemp );

			/* disable interrupts and reboot */
			disable();
			((FUNCPTR)(*reboot))();
		}
	}

	/* not changed, exit */
	else
		printf("No changes\n");

	exit(0);
}

/*
	$DOCPROC:DeleteExtendedPartition

	On Entry
		iDrive		Drive to delete this on
		iExtPart		Extended partition # to delete

	Delete this extended partition. Used by DeletePartition.
*/
void PASCAL DeleteExtendedPartition(int iDrive, int iExtPart)
{
	struct PARTITION * pPartition;
	struct ExtPart * ep;
	struct ExtPart * ep_prev;

	/* validate parameter */
	SHIPASSERT(iExtPart < hdisk[iDrive].iExtDosParts);

	/* find the extended DOS partition */
	ep = ep_prev = hdisk[iDrive].pepExt;
	while(ep)
	{
		pPartition = FindPartitionType(ep->pPartition, DOS16);
		if(pPartition)
		{
			/* we've found the one we want, delete it */
			if(iExtPart == 0)
			{
				/* if not the 1st in the list, unlink the sector */
				if(ep_prev != ep)
				{
					/* copy the extended partition */
					memmove(ep_prev->pPartition+1, ep->pPartition+1, sizeof(struct PARTITION));
					SetChangedFlag(ep_prev->fModified, 1);
					ep_prev->pepNext = ep->pepNext;
					dlFreeMem( FP_SEG(ep) );
				}
				else
				{
					memset(pPartition, 0, sizeof(struct PARTITION));
					SetChangedFlag(ep->fModified,
							(unsigned)(pPartition - ep->pPartition));
				}
				--hdisk[iDrive].iExtDosParts;
				break;
			}

			/* not the one we want, keep looking */
			--iExtPart;
		}

		/* go on to next extended partition */
		ep_prev = ep;
		ep = ep->pepNext;
	}

	/*
		If that was the last extended DOS partition, delete the
		extended record in the Master Boot Sector as well.
	*/
	if(!hdisk[iDrive].iExtDosParts)
	{
		pPartition = FindPartitionType(hdisk[iDrive].pPartition, DOSEXT);
		if(pPartition)
		{
			memset(pPartition, 0, sizeof(struct PARTITION));
			SetChangedFlag(hdisk[iDrive].fModified,
					(unsigned)(pPartition-hdisk[iDrive].pPartition));
		}

		/* free up entire list of eps */
		ClearAllExtended(iDrive);
	}
}


/*
	$DOCPROC:DeleteSuperBootPart

	Delete a Super-Boot partition. Used by DeletePartition.
*/
void PASCAL DeleteSuperBootPart(int iDrive)
{
	int	iId;
	unsigned	uIndex;
	struct PARTITION * pPartition;

	/* get ID value from user */
	do
	{
#if 0
      //TODO code in development to allow partition deletion by list offset
      //
      char szVal[6];
      char *psz;
      int oldval = 1;
      int value;

      printf("%s [%u]: ",szDelSuperID,oldval);

      /* get the string, check for abort */
      if(GetString(szVal,5,TRUE,FALSE) < 0)
      {
         value = CANCEL;
         break;
      }

      /* just use default value */
      if(*szVal == 0)
      {
         value = oldval;
      }
      /* user typed in a value */
      else
      {
         // skip whitespace then get the value
         psz = szVal;
         while(DL_IS_SPACE(*psz))
            ++psz;
         value = atoi(psz);
      }

      /* find this ID */
//      pPartition = FindPartitionEntry(hdisk[iDrive].pPartition, iId);

#else
      /* get ID, but don't allow */
      iId = GetHexValue(szDelSuperID,iPrevSuperBootID,7,0xFF);
      if(iId == CANCEL)
         return;

      /* find this ID */
      pPartition = FindPartitionEntry(hdisk[iDrive].pPartition, iId);
#endif

	} while(!pPartition);

	/* warn the user */
	sprintf(zz, szWarnSuper, iId);
	printf(zz);
	if(GetYesNo(szAreYouSure) == YES)
	{
		/* delete the partition */
		memset(pPartition, 0, sizeof(struct PARTITION));
		--hdisk[iDrive].iSuperBootParts;

		/* mark entry as modified */
		uIndex = (unsigned)(pPartition - hdisk[iDrive].pPartition);
		SetChangedFlag(hdisk[iDrive].fModified, uIndex );

		/* give user a chance to digest this delete */
		printf(szDeleted);
		GetEsc();
	}
}

/*
	$DOCPROC:CreateExtendedPartition

	Returns TRUE if worked, FALSE if failed.
*/
bool PASCAL CreateExtendedPartition(int iDrive, ulong ulStartTrack, ulong ulNumTracks)
{
	struct PARTITION * pPartition;
	struct ExtPart * ep;
	struct ExtPart * epNew;
//	ulong		ulSector;
	ulong		ulMaxTracks;
	ulong		ulMasterTrack;
	unsigned uIndex;
	bool fSuccessful;
	unsigned uTempSeg;


	/* validate parameters */
	ASSERT( iDrive >= 0 && iDrive < iNumHardDisks );
	ASSERT( ulNumTracks );

	/* if no extended partition, create one */
	if(!hdisk[iDrive].pepExt)
	{
		// the extended partition will be the MAXIMUM size
		if(!GetAvailableDiskSpace(iDrive, &ulStartTrack, &ulMaxTracks))
			return FALSE;

		// special case: extended partitions can't begin on track 0
		if(ulStartTrack == 0)
		{
			++ulStartTrack;
			--ulMaxTracks;
			if(!ulMaxTracks)
				return FALSE;
		}

		// assume initially that ALL space on the disk will be
		// used for the extended partition(s)
		pPartition = AddPartitionEntry(iDrive, hdisk[iDrive].pPartition, 0,
				ulStartTrack, ulMaxTracks,
				DetermineExtentionId(iDrive, ulStartTrack, ulMaxTracks));
		if(!pPartition)
			return FALSE;

		/* mark as created */
		SetChangedFlag(hdisk[iDrive].fModified,
				(unsigned)(pPartition - hdisk[iDrive].pPartition));
	}

	// if the extended partition doesn't contain a DOS partition
	// this might happen if MS-DOS' FDISK made the extended partion
	// but didn't put any DOS partitions in it.
	if(hdisk[iDrive].pepExt && hdisk[iDrive].pepExt->pPartition->id == 0)
		epNew = hdisk[iDrive].pepExt;

	// allocate a new structure for this extended partition
	else
	{
		/* get the memory for the structure */
		fSuccessful = dlAllocMem( CVT_PARA(sizeof(struct ExtPart)), &uTempSeg );
		SHIPASSERT( fSuccessful );
		epNew = MK_FP( uTempSeg, 0 );
		memset(epNew, 0, sizeof(struct ExtPart));

		/* remember start of this extended partition */
		epNew->ulStartTrack = ulStartTrack;

		/* this structure may the first in the chain */
		if(!hdisk[iDrive].pepExt)
			hdisk[iDrive].pepExt = epNew;

		/* insert new extended partition at end of chain */
		else
		{
			// get head of chain
			ep = hdisk[iDrive].pepExt;

			// find the end of the chain, then add it
			while(ep)
			{
				if(!ep->pepNext || ep->ulStartTrack > ulStartTrack)
				{
					epNew->pepNext = ep->pepNext;
					ep->pepNext = epNew;

					// added to the middle (as opposed to the end)
					if(ep->pepNext)
					{
						memmove(epNew->pPartition+1,ep->pPartition+1,
							sizeof(struct PARTITION));
						SetChangedFlag(epNew->fModified, 1);
					}

					// add the new extended partition. always use ID of DOSEXT,
					// not DOSEXT_LBA for subsequent extended partitions.
					ep->pPartition[1].id = 0;
					ulMasterTrack = hdisk[iDrive].pepExt->ulStartTrack;
					pPartition = AddPartitionEntry(iDrive, ep->pPartition,
						ulMasterTrack, ulStartTrack, ulNumTracks, DOSEXT);
					uIndex = (unsigned)(pPartition - ep->pPartition);
					SetChangedFlag(ep->fModified, uIndex);
					break;
				}
				ep = ep->pepNext;
			}
		}
	}

	// allocate the DOS part of the partition, always relative to track of
	// master extended partition
	pPartition = AddPartitionEntry(iDrive, epNew->pPartition,
		epNew->ulStartTrack, ulStartTrack, ulNumTracks,
		DeterminePartitionId(iDrive, ulStartTrack, ulNumTracks, OLD_STYLE));
	SHIPASSERT(pPartition);

	/* mark as created */
	SetCreatedFlag(epNew->fModified, 0);
	++hdisk[iDrive].iExtDosParts;

	// everything worked
	return TRUE;
}

bool IsDosDrivePartitionID( uchar ucPartitionID )
{
	switch( ucPartitionID )
	{
		case 0x01:
		case 0x04:
		case 0x06:
			return TRUE;

		default:
			if( fIsFat32World )
			{
				switch( ucPartitionID )
				{
					case 0x0b:
					case 0x0c:
					case 0x0e:
						return TRUE;

					default:
						return FALSE;
				}
			}
			return FALSE;
	}
}


/*
	$DOCPROC:PartitionLetter

	Determine a partition's drive letter. Rules on drive order:
	1) all bootable, primary partitions, all drives
	2) all other primary partitions, all drives
	3) all extended partitions, all drives
	4) all super-boot partitions, all drives
*/
char PASCAL PartitionLetter(int iDrive, struct PARTITION *pPartition)
{
	int i;
	int iPartIndex;
	int iExtPartIndex;
	int iDosPartsFound;
	char letter;
	struct ExtPart * pExtPart;
	unsigned uPartIndex;


	/* count primary partitions first */
	letter = startletter - 1;


	//  1) all bootable, primary partitions, all drives
	//
	for(i=0; i<iNumHardDisks; ++i)
	{
		iDosPartsFound = 0;
		for( iPartIndex = 0; iDosPartsFound < hdisk[i].iDosParts;
				iPartIndex++)
		{
			if( IsDosDrivePartitionID(
					hdisk[i].pPartition[iPartIndex].id ) )
			{
				iDosPartsFound++;
			}

			if(hdisk[i].pPartition[iPartIndex].boot != 0)
			{
				if( IsDosDrivePartitionID(
						hdisk[i].pPartition[iPartIndex].id ) )
				{
					letter++;
				}

				//  Is this the drive partition and letter we are
				//  looking for?
				//
				if( (i == iDrive) &&
						(&(hdisk[i].pPartition[iPartIndex]) == pPartition) )
				{
					return letter;
				}
			}
		}
	}

	//  2) all other primary partitions, all drives
	//
	for(i=0; i<iNumHardDisks; ++i)
	{
		iDosPartsFound = 0;
		for( iPartIndex = 0; iDosPartsFound < hdisk[i].iDosParts;
				iPartIndex++)
		{
			if( IsDosDrivePartitionID(
					hdisk[i].pPartition[iPartIndex].id ) )
			{
				iDosPartsFound++;
			}

			if(hdisk[i].pPartition[iPartIndex].boot == 0)
			{
				if( IsDosDrivePartitionID(
						hdisk[i].pPartition[iPartIndex].id ) )
				{
					letter++;
				}

				//  Is this the drive partition and letter we are
				//  looking for?
				//
				if( (i == iDrive) &&
						(&(hdisk[i].pPartition[iPartIndex]) == pPartition) )
				{
					return letter;
				}
			}
		}
	}

	//  3) all extended partitions, all drives
	//
	for(i=0; i<iNumHardDisks; ++i)
	{
		pExtPart = hdisk[i].pepExt;
		for( iExtPartIndex = 0; iExtPartIndex < hdisk[i].iExtDosParts; )
		{
			for( uPartIndex = 0; (uPartIndex < NUMPART) &&
					(iExtPartIndex < hdisk[i].iExtDosParts); uPartIndex++ )
			{
				if( IsDosDrivePartitionID(pExtPart->pPartition[uPartIndex].id))
				{
					iExtPartIndex++;
					letter++;

					if(i==iDrive)
					{
						if( (&pExtPart->pPartition[uPartIndex]) == pPartition )
						{
							return letter;
						}
					}
				}
			}
			pExtPart = pExtPart->pepNext;
		}
	}

	return letter;
}

/*
	$DOCPROC:TrackToSector

	Convert a track number into a linear sector number
*/
ulong PASCAL TrackToSector(int iDrive, ulong ulTrack)
{
	return ulTrack * GetHeads(iDrive) * GetSPT(iDrive);
}

/*
	$DOCPROC:SectorToTrack

	Convert a linear sector number into a track number.
*/
ulong PASCAL SectorToTrack(int iDrive, ulong ulSector)
{
	return ulSector / (GetHeads(iDrive) * GetSPT(iDrive));
}

/*
	$DOCPROC:SectorToTrack

	Convert a linear sector number into a track number.
*/
ulong PASCAL NumTracks(int iDrive, ulong ulSectors)
{
	ulong		ulSectorsEachTrack = GetHeads(iDrive) * GetSPT(iDrive);
	ulong		ulTracks;
	ulTracks = ulSectors / ulSectorsEachTrack;
	if(ulSectors % ulSectorsEachTrack)
		++ulTracks;
	return ulTracks;
}

/*
	$DOCPROC:SectorToBigCHS

	Convert from a linear sector number to the big CHS values (may be
	larger than the old BIOS CHS values.

	On Entry
		iDrive		drive
		ulSector		linear sector #
	On Exit
		*pulTrack	track # (may be > 1024)
		*pulHead		head
		*pulSector	sector

	Returns FALSE (failed) only if # heads or SPT is invalid.
*/
bool PASCAL SectorToBigCHS(int iDrive, ulong ulSector,
					ulong * pulTrack, ulong * pulHead, ulong * pulSector)
{
	ulong	ulHeads = GetHeads(iDrive);
	ulong	ulSPT   = GetSPT(iDrive);

	// no dividing by 0
	if(!ulSPT || !ulHeads)
		return FALSE;

	// sector is 1 based, head and track are 0 based
	*pulSector = (ulSector % ulSPT) + 1;
	ulSector  /= ulSPT;
	*pulHead   = ulSector % ulHeads;
	ulSector  /= ulHeads;
	*pulTrack  = ulSector;

	return TRUE;
}

/*
	$DOCPROC:SectorToWrapCHS

	Cannot fail. Always wraps to CHS.
*/
void PASCAL SectorToWrapCHS(int iDrive, ulong ulSector, uchar *pucTrack,
			uchar *pucHead, uchar *pucSector)
{
	ulong		ulTrack, ulHead, ulSPT;

	// drive geometry had better be valid
	if(!SectorToBigCHS(iDrive, ulSector, &ulTrack, &ulHead, &ulSPT))
		ErrorExit(szErrDriveParms);

	// set up the original BIOS CHS values
	*pucHead    = (uchar)ulHead;
	*pucTrack   = (uchar)ulTrack;
	*pucSector  = (((int)ulTrack>>2) & 0xC0);
	*pucSector |= (uchar)ulSPT;
}

//  Display a critical error message and exit
//
void ErrorExit( char * szErrorMessage )
{
	printf( szErrorMessage );
	exit(255);
}


/*
	$DOCPROC:SectorToSmallCHS

	Returns TRUE if the CHS fits into the original BIOS CHS fields.
	Returns FALSE if not.
*/
bool PASCAL SectorToSmallCHS(int iDrive, ulong ulSector, uchar *pucTrack,
			uchar *pucHead, uchar *pucSector)
{
	ulong		ulTrack, ulHead, ulSPT;
	bool		fIsSmall = TRUE;

	// drive geometry had better be valid
	if(!SectorToBigCHS(iDrive, ulSector, &ulTrack, &ulHead, &ulSPT))
		ErrorExit(szErrDriveParms);

	// convert longs to original bios CHS form
	if(ulTrack > 1023)
	{
		ulTrack = 1023;
		fIsSmall = FALSE;
	}
	if(ulHead > 255)
	{
		ulHead = 255;
		fIsSmall = FALSE;
	}
	if(ulSPT > 63)
	{
		ulSPT = 63;
		fIsSmall = FALSE;
	}

	// set up the original BIOS CHS values
	*pucHead    = (uchar)ulHead;
	*pucTrack   = (uchar)ulTrack;
	*pucSector  = (((int)ulTrack>>2) & 0xC0);
	*pucSector |= (uchar)ulSPT;

	return	fIsSmall;
}

/*
	$DOCPROC:CalcSystemLen

	Determine size of FATs + root directory for a partition with this
	many sectors. Returns size in sectors.
*/
ulong PASCAL CalcSystemLen(int iDrive, ulong ulSectors)
{
	ulong ulSystemLen;
	unsigned uSectorsPerCluster;

	// Calculate size of FATs
	uSectorsPerCluster = CalculateSectorsPerCluster(ulSectors);

	// determine length of FAT
	if( ulSectors > SPC12BIT*MAXFAT12 )
	{
		if(swFat32 && ulSectors >= _512MEG)
			ulSystemLen = ((ulSectors/uSectorsPerCluster) * 4) / SECSIZE + 3;
		else
			ulSystemLen = ((ulSectors/uSectorsPerCluster) * 2) / SECSIZE + 2;
	}
	else
		ulSystemLen = (((ulSectors/uSectorsPerCluster) * 3) / 2) / SECSIZE + 2;

	// assume two FATs
	ulSystemLen *= 2;

	// add the typical hidden track
	ulSystemLen += GetSPT(iDrive);

	// assume typical root dir (240 entries) at 16 entries per sector
	ulSystemLen += 240/0x10;

	// return length in sectors
	return ulSystemLen;
}

/*
	$DOCPROC:FindValidSystemTrack

	Given a starting track and a total size in tracks of a partition,
	this function calculates the size of the system area and finds the
	first contiguous group of valid sectors large enough.

	Adjusts startTrack and totalTracks if the system area of the
	specified partition contains bad Tracks.

	TotalTracks may return 0 if a valid system area cannot be found.

	Example:
		hardDrive = 1;						// Second hard drive
		startPart = 100;					// Partition starts on Track 100
		lengthPart = 200;					// and is 200 Tracks long.
		FindValidSystemTrack( hardDrive, &startPart, &lengthPart );

	Example Return
		Nothing changed if the first couple of tracks are good.

		Assuming that the example partition needed 4 tracks for
		the system area, and Track 103 was bad, these values would
		return: startPart == 104  and  lengthPart == 196
*/
void PASCAL FindValidSystemTrack(int iDrive, ulong * pulStartTrack,
		ulong * pulNumTracks)
{
	ulong ulSectors;
	ulong ulSectorsEachTrack;
	ulong ulSystemSectors;
	ulong	ulSystemTracks;
	ulong ulBadTracks;

	// Partition size, in sectors
	ulSectorsEachTrack = GetHeads(iDrive) * GetSPT(iDrive);
	ulSectors = *pulNumTracks * ulSectorsEachTrack;

	// determine length (in sectors) of system areas
	ulSystemSectors = CalcSystemLen(iDrive, ulSectors);

	// convert system sectors to cylinders
	ulSystemTracks = ulSystemSectors / ulSectorsEachTrack;
	if(ulSystemSectors % ulSectorsEachTrack)
		++ulSystemTracks;

	// so long as the partition is 1 or more Tracks long...
	while( *pulNumTracks )
	{
		// look for a bad Track
		ulBadTracks = VerifyContiguousTracks( iDrive, *pulStartTrack, ulSystemTracks);

		// no bad Track found?  Good!  QUIT LOOKING!
		if( !ulBadTracks )
			break;

		// move the partition start forward
		*pulStartTrack += ulBadTracks;

		// shrink the partition the same amount, but never
		// fewer than 0 Tracks.
		*pulNumTracks = max( *pulNumTracks - ulBadTracks, 0 );
	}
}

/*
	$DOCPROC:VerifyContiguousTracks

	Verifies every trak from iStartTrak to (iStartTrak+iNumTraks-1)
	Returns 0 if all traks were good, or the trak offset of the
	first bad trak found.

	Example:
		VerifyContiguousTracks( 0, 10, 4 );		// Verify tracks 10-13
															// on BIOS drive 80
		Returns:
			0 if traks 10-13 are good
			1 if trak 10 is bad
			2 if trak 11 is bad, etc
*/
ulong PASCAL VerifyContiguousTracks(int iDrive, ulong ulStartTrack, ulong ulNumTracks)
{
	ulong	ulTrack;
	ulong	ulHead, ulNumHeads;
	ulong	ulSPT;
	ulong	ulSector;
	LSTP	lstpTransfer;

	// Check for invalid drive #s passed (0 based)
	SHIPASSERT(iDrive >= 0 && iDrive < iNumHardDisks);

	// We should never have to check for 0 good traks
	SHIPASSERT(ulNumTracks);

	// determine heads and SPT
	ulNumHeads = GetHeads(iDrive);
	ulSPT      = GetSPT(iDrive);

	// set up the transfer buffer and size just once
	lstpTransfer.lpBuffer      = pBufferDMA;
	lstpTransfer.iNumSectors   = (int)ulSPT;

	// verify each track
	ulSector = TrackToSector(iDrive, ulStartTrack);
	for(ulTrack = 0; ulTrack < ulNumTracks; ulTrack++ )
	{
		for( ulHead = 0; ulHead < ulNumHeads; ulHead++ )
		{
			lstpTransfer.ulStartSector = ulSector;
			if(!DlSmartLbaVerifySectors(&hdisk[iDrive].lsipInfo, &lstpTransfer))
				return ulTrack+1;
			ulSector += ulSPT;
		}
	}

	// everything worked, return no bad tracks
	return 0;
}

/*
	$DOCPROC:CalculateSectorsPerCluster

	Calculate sectors per cluster based on size of hard disk
	partition. Returns sectors per cluster.
*/
unsigned PASCAL CalculateSectorsPerCluster(ulong len)
{
	unsigned val = (unsigned)((len-1)/0xFFF0);
	unsigned char shift;

	/* small partition, either 8 or 4 spc */
	if(val<=1)
	{
		if(len <= SPC12BIT*MAXFAT12)
			return 8;
      return 4;
   }

	// 8 SPC (4KB) for FAT32 partitions
	if(swFat32 && len >= _512MEG)
		return 8;

	shift = 8;

	while((val&0x80) == 0)
	{
		val = val<<1;
		--shift;

	}
	val = 1;
	while(shift--)
		val = val<<1;
	return val;
}

/*
	$DOCPROC:GetAvailableTracks

	Get available cylinders on this drive. Finds the first available
	space, not the largest. Returns TRUE if worked, along with the
	starting cylinder and length in cylinders.
*/
bool PASCAL GetAvailableTracks(int iDrive, bool fDos,
					ulong *pulStartTrack, ulong *pulNumTracks)
{
	/* validate parameters */
	ASSERT(iDrive >= 0 && iDrive < iNumHardDisks);
	ASSERT(pulStartTrack && pulNumTracks);

	/* get space from extended partition if primary and ext partitions
		are already created */
	if(fDos && hdisk[iDrive].pepExt && hdisk[iDrive].iDosParts)
	{
		if(!GetAvailableExtendedDiskSpace(iDrive, pulStartTrack, pulNumTracks))
			return FALSE;
	}

	/* is disk filled (or all free space is bad? */
	else
	{
		if(!GetAvailableDiskSpace(iDrive, pulStartTrack, pulNumTracks))
			return FALSE;
	}

	/* everything worked */
	return TRUE;
}

/*
	$DOCPROC:GetAvaliableExtendedDiskSpace

	Look for a free block in the extended partition.

	Returns TRUE and the start and size of this block in tracks if worked.
	Returns FALSE if failed.
*/
bool PASCAL GetAvailableExtendedDiskSpace(int iDrive,
					ulong *pulStartTrack, ulong *pulNumTracks)
{
	struct ExtPart * ep;
	struct PARTITION * pPartition;
	ulong		ulExtStartTrack;
	ulong		ulExtNumTracks;
	ulong		ulNext, ulEnd;

	/* if no extended partition, give up */
	pPartition = FindPartitionType(hdisk[iDrive].pPartition, DOSEXT);
	if(!pPartition)
		hdisk[iDrive].pepExt = 0;
	ep = hdisk[iDrive].pepExt;
	if(!ep)
		return FALSE;

	/* get the start and size of the extended partition */
	ulExtStartTrack = SectorToTrack(iDrive, pPartition->hidden);
	ulExtNumTracks  = SectorToTrack(iDrive, pPartition->size);

	/* the extended partition MAY have no DOS partition in it yet */
	if(ep->pPartition->id == 0)
	{
		*pulStartTrack = ep->ulStartTrack;
		if(ep->pepNext)
			*pulNumTracks = ep->pepNext->ulStartTrack - *pulStartTrack;
		else
			*pulNumTracks = ulExtStartTrack + ulExtNumTracks;

		FindValidSystemTrack(iDrive, pulStartTrack, pulNumTracks);

		if( *pulNumTracks >= MINTRAK )
			return TRUE;
	}

	/* look through ExtPart list for gaps */
	while(ep)
	{
		ulEnd = ep->ulStartTrack + SectorToTrack(iDrive, ep->pPartition->size);
		if(ulEnd + 1 >= GetTracks(iDrive))
			break;

		if(ep->pepNext)
			ulNext = ep->pepNext->ulStartTrack;
		else
			ulNext = ulExtStartTrack + ulExtNumTracks;

		if(ulEnd+1+MINTRAK < ulNext)
		{
			*pulStartTrack = ulEnd+1;
			*pulNumTracks  = ulNext - *pulStartTrack;
			FindValidSystemTrack( iDrive, pulStartTrack, pulNumTracks );
			if( *pulNumTracks >= MINTRAK )
				return TRUE;
		}

		ep = ep->pepNext;
	}

	/* no free space found */
	return FALSE;
}

/*
	$DOCPROC:GetAvailableDiskSpace

	Search through list for any available space on the disk.
	Returns the starting cylinder or -1 if no free space is found.
	Sets "*pulNumTracks" to the maximum # of cylinders in this
	block.
*/
bool PASCAL GetAvailableDiskSpace(int iDrive,
					ulong *pulStartTrack, ulong *pulNumTracks)
{
	int	i,j;
	struct PARTITION * pPartition;
	ulong	ulStartTrack[NUMPART];			/* array of elements to sort */
	ulong	ulEndTrack[NUMPART];
	ulong	ulTmp;
	int	iNumElements;						/* # of elements to sort */
	bool	fFound = FALSE;

	/* get all current partitions (start and end) */
	iNumElements = 0;
	for(i=0, pPartition = hdisk[iDrive].pPartition; i<NUMPART; ++i, ++pPartition)
	{
		if(pPartition->id)
		{
			ulStartTrack[iNumElements] =
				SectorToTrack(iDrive, pPartition->hidden);
			ulEndTrack[iNumElements]   =
				SectorToTrack(iDrive, pPartition->hidden + pPartition->size - 1);
			SHIPASSERT(ulEndTrack[iNumElements] < GetTracks(iDrive));
			++iNumElements;
		}
	}

	/* sort the partitions into ascending order */
	for(i=0; i+1<iNumElements; ++i)
	{
		for(j=i+1; j<iNumElements; ++j)
		{
			// out of order? swap 'em
			if(ulStartTrack[j] < ulStartTrack[i])
			{
				// swap ulStartTrack
				ulTmp = ulStartTrack[j];
				ulStartTrack[j] = ulStartTrack[i];
				ulStartTrack[i] = ulTmp;

				// swap ulEndTrack
				ulTmp = ulEndTrack[j];
				ulEndTrack[j] = ulEndTrack[i];
				ulEndTrack[i] = ulTmp;
			}
		}
	}

	/* no records yet defined, we have the whole disk */
	if(!iNumElements)
	{
		*pulStartTrack = 0;
		*pulNumTracks  = GetTracks(iDrive);
		fFound = TRUE;
	}

	/* look for free space */
	else
	{
		/* look for space at beginning */
		if(ulStartTrack[0] >= MINTRAK)
		{
			*pulStartTrack = 0;
			*pulNumTracks = ulStartTrack[0];
			fFound = TRUE;
		}

		/* look for hole */
		else
		{
			for(i=0; i+1 < iNumElements; ++i)
			{
				if(ulEndTrack[i]+MINTRAK < ulStartTrack[i+1])
				{
					*pulStartTrack = ulEndTrack[i] + 1;
					*pulNumTracks  = ulStartTrack[i+1] - *pulStartTrack;
					fFound = TRUE;
					break;
				}
			}

			/* not found yet? look for space at the end */
			if(!fFound && ulEndTrack[i]+MINTRAK < GetTracks(iDrive))
			{
				*pulStartTrack = ulEndTrack[i] + 1;
				*pulNumTracks  = GetTracks(iDrive) - *pulStartTrack;
				fFound = TRUE;
			}
		}
	}

	// if found, verify system tracks
	if(fFound)
	{
		/* make sure the system areas are writable */
		FindValidSystemTrack(iDrive, pulStartTrack, pulNumTracks);

		/* enough cylinders? */
		if( *pulNumTracks < MINTRAK )
			return FALSE;
	}

	// return if we found room or not
	return fFound;
}

/*
	$DOCPROC:AddPartitionEntry

	This adds a partition into the given partition table.
	Adjusts the first partition entry to begin on head 1.

	Returns the partition found, or NULL if no partition entry available.
*/
struct PARTITION * PASCAL AddPartitionEntry(int iDrive,
	struct PARTITION pPartitionTable[], ulong ulParentTrack,
	ulong ulStartTrack, ulong ulNumTracks, uchar ucId)
{
	struct PARTITION * pPartition;
	ulong		ulStartSector;
	ulong		ulNumSectors;
	ulong		ulSPT = GetSPT(iDrive);

	// all but DOS partitions start on sector 0
	ulStartSector  = TrackToSector(iDrive, ulStartTrack);
	ulNumSectors   = TrackToSector(iDrive, ulNumTracks);

#if 0
	// DOS partitions start at head 1, including SuperBoot partitions.
	// extended partitions start on head 0
	if(!IsExtPartitionId(ucId))
	{
		ulStartSector += ulSPT;
		ulNumSectors  -= ulSPT;
	}
#endif

	// find a blank entry
	pPartition = FindPartitionEntry(pPartitionTable, 0);
	if(!pPartition)
	{
		printf(szNoPartsLeft);
		GetEsc();
		return NULL;
	}
   // IR 3839 TD 10/29/01
   //
   // Research has shown that the first partition entry must start on
   // head one, but further entries in the table start on head zero.
   //
   // NOTE: the competition will not do this when the first partition
   // is an Extended Partition!  However, we do not allow the user to
   // create anything like that, so no problem (so far...)
   //
   else if( pPartition==pPartitionTable )
   {
		ulStartSector += ulSPT;
		ulNumSectors  -= ulSPT;
   }

	// fill in partition entry
	pPartition->boot   = 0;
	pPartition->id     = ucId;
	pPartition->hidden = ulStartSector - TrackToSector(iDrive, ulParentTrack);
	pPartition->size   = ulNumSectors;

	// MBR partitions are limited to maximum CHS values
	if(ulParentTrack == 0)
	{
		SectorToSmallCHS(iDrive, ulStartSector,
			&pPartition->strak, &pPartition->shead, &pPartition->ssect);
		SectorToSmallCHS(iDrive, ulStartSector+ulNumSectors-1,
			&pPartition->etrak, &pPartition->ehead, &pPartition->esect);
	}

	// partitions inside extended partitions wrap
	else
	{
		SectorToWrapCHS(iDrive, ulStartSector,
			&pPartition->strak, &pPartition->shead, &pPartition->ssect);
		SectorToWrapCHS(iDrive, ulStartSector+ulNumSectors-1,
			&pPartition->etrak, &pPartition->ehead, &pPartition->esect);
	}

	return pPartition;
}

// helper function to get the heads for the given drive
ulong PASCAL GetHeads(int iDrive)
{
	return DlSmartLbaGetHeads(&hdisk[iDrive].lsipInfo);
}

// helper function to get the heads for the given drive
ulong PASCAL GetSPT(int iDrive)
{
	return DlSmartLbaGetSPT(&hdisk[iDrive].lsipInfo);
}

// helper function to get the tracks for the given drive
ulong PASCAL GetTracks(int iDrive)
{
	ulong ulTracks;


	ulTracks = DlSmartLbaGetCylinders(&hdisk[iDrive].lsipInfo);

	// DOS 6.22 dos not have LBA support.  If we are in DOS 6.22
	// mode then we need to clip the tracks at the limit
	//
	if(!swFat32)
	{
		if( ulTracks > 1024 )
		{
			ulTracks = 1024;
		}
	}

	return ulTracks;
}

/*
	Return the maximum cylinders in a 2Gig (DOS 6.22 style) partition.
*/
ulong PASCAL GetMaxTracksFor2Gigabytes(int iDrive)
{
	ulong ulMax2GigTracks;

	ulMax2GigTracks = ((ulong)0x80000000L / SECSIZE) /
			(GetHeads(iDrive) * GetSPT(iDrive));
	return ulMax2GigTracks;
}

/*
	$DOCPROC:DetermineExtensionId
*/
int PASCAL DetermineExtentionId(int iDrive, ulong ulStartTrack,
					ulong ulNumTracks)
{
	uchar	ucTrak, ucHead, ucSect;
	ulong	ulEndSector;

	// return LBA if it's big.
	ulEndSector = TrackToSector(iDrive, ulStartTrack + ulNumTracks) - 1;
	if(!SectorToSmallCHS(iDrive, ulEndSector, &ucTrak, &ucHead, &ucSect))
		return DOSEXT_LBA;

	// a normal extension works
	return DOSEXT;
}

/*
	$DOCPROC:DeterminePartitionId

	For this # of tracks, return what type of partition this should
	be? Can be one of: DOS12, DOS16, DOSBIG, DOS32, DOS32_LBA,
	DOSBIG_LBA. Does not determine the ID for the extended partition,
	just the DOS partitions.

	On Entry
	iDrive			0= 1st HD, 1= 2nd HD, etc..
	ulStartTrack	starting track (cylinder)
	ulNumTracks		ending track (cylinder)
	fNewStyle		allow LBA extensions
*/
int PASCAL DeterminePartitionId(int iDrive, ulong ulStartTrack,
					ulong ulNumTracks, bool fNewStyle)
{
	ulong ulNumSectors;
	ulong	ulEndSector;
	bool	fLba;
	uchar	ucTrak, ucHead, ucSect;

	// determine # of sectos in the partition
	ulNumSectors = TrackToSector(iDrive, ulNumTracks);
	ulEndSector  = TrackToSector(iDrive, ulStartTrack + ulNumTracks) - 1;

	// if we can't fit trak/head/sector into old CHS values, then it's an
	// LBA partition. Only allow LBA codes if newstyle.
	if(fNewStyle &&
		!SectorToSmallCHS(iDrive, ulEndSector, &ucTrak, &ucHead, &ucSect))
		fLba = TRUE;
	else
		fLba = FALSE;

	// make a FAT32 partition if possible
	if(swFat32 && ulNumSectors >= _512MEG)
	{
		// never return DOS32_LBA. Confuses MS-DOS.
		if(fLba)
			return DOS32_LBA;
		return DOS32;
	}

	// for LBA partitions, always DOSBIG, even if a small Fat-12 partition
	if(fLba)
		return DOSBIG_LBA;

	// make a big partition if possible
	if(ulNumSectors > MAX32MSEC)
		return DOSBIG;
	if((ulNumSectors / SPC12BIT) > MAXFAT12)
		return DOS16;
	return DOS12;
}

/*
	Like FindPartitionEntry but looks for a category, rather than a specific
	ID.

	Returns a pointer to the first type found.
*/
struct PARTITION * PASCAL FindPartitionType(struct PARTITION pPartitionTable[],
									int iIdType)
{
	struct PARTITION * pPartition = pPartitionTable;
	int i;
	int PASCAL (*pfnIdType)(uchar);

	// what ID type are we looking for?
	if(iIdType == DOS16)
		pfnIdType = IsDosPartitionId;
	else if(iIdType == DOSEXT)
		pfnIdType = IsExtPartitionId;
	else
		pfnIdType = IsSuperBootPartitionId;

	for(i=0; i<NUMPART; ++i)
	{
		if((*pfnIdType)(pPartition->id))
			return pPartition;
		++pPartition;
	}
	return NULL;
}

/*
	$DOCPROC:FindPartitionEntry
	Given a partition table, find the entry that has the requested id.
*/
struct PARTITION * PASCAL FindPartitionEntry(struct PARTITION pPartitionTable[],
									int iId)
{
	struct PARTITION * pPartition = pPartitionTable;
	int i;

	for(i=0; i<NUMPART; ++i)
	{
		if(pPartition->id == iId)
			return pPartition;
		++pPartition;
	}
	return NULL;
}

/*
	$DOCPROC:Dump
	Dump a near peice up memory, up to 64K.
*/
void PASCAL Dump(void *buff, unsigned length)
{
	unsigned start=0, len, i;
	uchar *buffer = (uchar *)buff;

	while(length>0)
	{
		len=min(length,16);
		for(i=0;i<16;++i)
		{
			printf(i<len ? "%02X " : "   ",buffer[start+i]);
			if(i==7)
				printf(i<len ? "| " : "  ");
		}
		printf(" ");
		for(i=0;i<16;++i)
		{
			if(i<len)
				printf(isprint(buffer[start+i]) ? "%c" : ".",buffer[start+i]);
			else
				printf(" ");
		}
		printf("\n");
		iLineCount++;

		length -= len;
		start  += len;
	}
}

/*
	^break points here, so that break is effectively disabled.
*/
//  Int 23 handler
//
#pragma argsused
 void interrupt far Int23Handler(BP, DI, SI, DS, ES, DX, CX, BX, AX,
		IP, CS, FLAGS)
{
	//  Do nothing
}


//  Install a control break error handler so that we can abort cleanly
//
void InstallInt23Handler( void )
{
	void far * far * lplpOldCtrlBrkHandler;


	lplpOldCtrlBrkHandler = MK_FP( 0, 0x23*4 );
	*lplpOldCtrlBrkHandler = (void far *)Int23Handler;
}


/*
	$DOCPROC:AllocDMACompatible

	Allocate some memory, making sure the memory does not cross a DMA
	boundary (ie 1000:0000, 2000:0000, etc)
*/
void * PASCAL AllocDMACompatible(int size)
{
	char *p;
//	unsigned boundary;
//	unsigned ds = _DS;		/* Turbo C specific (_DS) */
	unsigned uTempSeg;


	/* get some memory */
	uTempSeg = 0;
	dlAllocMem( CVT_PARA(size), &uTempSeg );
	p = MK_FP( uTempSeg, 0 );

	return p;
}

/*
	$DOCPROC:NumFixedDisks

	Ask the BIOS how many fixed disks there are:
	how many fixed disks are in the system?
*/
int PASCAL NumFixedDisks(void)
{
	BDP	bdpParameters;

	// determine # of hard disks
	if(!DlBiosGetDriveParameters(HD1, &bdpParameters))
		return 0;

	// limit check the # of fixed disks
	if(bdpParameters.ucNumDrives > MAXFIXED)
		return MAXFIXED;

	// return the # of disks
	return bdpParameters.ucNumDrives;
}

/*
	Write a single sector to the harddisk.
*/
void PASCAL WriteSector(int iDrive, ulong ulSector, char *pBuffer)
{
	LSTP	lstpTransfer;

	// write the sector
	lstpTransfer.ulStartSector = ulSector;
	lstpTransfer.lpBuffer      = pBuffer;
	lstpTransfer.iNumSectors   = 1;
	if(!DlSmartLbaWriteSectors(&hdisk[iDrive].lsipInfo, &lstpTransfer))
		ErrorExit(szErrWriting);
}

/*
	$DOCPROC:ReadSector
	Read a single sector from the harddisk. Failures are ignored.
*/
void PASCAL ReadSector(int iDrive, ulong ulSector, char *pBuffer)
{
	LSTP	lstpTransfer;

	// read the sector
	lstpTransfer.ulStartSector = ulSector;
	lstpTransfer.lpBuffer      = pBuffer;
	lstpTransfer.iNumSectors   = 1;
	if(!DlSmartLbaReadSectors(&hdisk[iDrive].lsipInfo, &lstpTransfer))
		ErrorExit(szErrReading);
}

/*
	$DOCPROC:WritePartitions

	This is the work-horse routine that writes to the disk. Up to now,
	only the in-memory image has been updated. Write the new
	partitions for all drives. Only the changes will be written out,
	the other partitions will be left as they were.
*/
void PASCAL WritePartitions(void)
{
	int	iDrive;
//	int	partent;
	int	i;
	struct PARTITION * pPartition;

	/* go through all the disks */
	for(iDrive=0; iDrive<iNumHardDisks; ++iDrive)
	{
		/* save changes to the Master Partition Sector */
		if(IsChanged(hdisk[iDrive].fModified))
		{
			/* if Master Partition Sector is already valid don't update CODE */
			if(hdisk[iDrive].fCodeOK)
				ReadSector(iDrive, 0, pBufferDMA);
			else
				memmove(pBufferDMA, array, sizeof(array));

			/* write the new partition information */
			memmove(pBufferDMA+PARTOFF, hdisk[iDrive].pPartition, NUMPART*sizeof(struct PARTITION));
			WriteSector(iDrive, 0, pBufferDMA);
		}

		/*
			One or more DOS and/or super-boot partitions may have been
			created. Make sure to zero the system areas.
		*/
		pPartition = hdisk[iDrive].pPartition;
		for(i=0; i<NUMPART; ++i)
		{
			/* if this partition was freshly created, zero system areas */
			if((pPartition->id==iPrevSuperBootID ||
				IsDosPartitionId(pPartition->id)) &&
				IsCreated(hdisk[iDrive].fModified,
						(unsigned)(pPartition - hdisk[iDrive].pPartition)))
			{
				EraseSystemAreas(iDrive, 0, pPartition);
			}
			++pPartition;
		}

		/* save any extended partitions */
		WriteExtendedPartitions(iDrive);
	}
}

/*
	$DOCPROC:SetCreatedFlag

	Inform global hard disk structure that this partition information
	has been created.
*/
void PASCAL SetCreatedFlag(bool *marray, unsigned n)
{
	// validate parameters
	SHIPASSERT(n < NUMPART);
	SHIPASSERT(marray);

	// inform FDISK something has changed
	fChanged = TRUE;

	// specifically, a partition has been created
	marray[n] = CREATED;
}

/*
	$DOCPROC:SetChangedFlag

	Set a partition entry as modified in this array.
*/
void PASCAL SetChangedFlag(bool *marray, int n)
{
	// validate parameters
	SHIPASSERT(n >= 0 && n < NUMPART);
	SHIPASSERT(marray);

	// let FDISK know something has changed
	fChanged = TRUE;

	// specifically, a partition has changed
	// only update if not already set, as there may be a created code
	if(!marray[n])
		marray[n] = TRUE;
}

/*
	$DOCPROC:IsCreated
	Has this partition entry been created?
*/
bool PASCAL IsCreated(bool * marray, int n)
{
	// validate parameters
	SHIPASSERT(n >= 0 && n < NUMPART);
	SHIPASSERT(marray);

	// is this parition created this session?
	return (marray[n]==CREATED);
}

/*
	$DOCPROC:IsChanged
	Has a partition entry been modified in this array?
*/
bool PASCAL IsChanged(bool *marray)
{
	int modified = FALSE;
	int i;

	// validate parameters
	SHIPASSERT(marray);

	/* has the partition sector been modified? */
	for(i=0; i<NUMPART; ++i)
		if(marray[i])
			modified = TRUE;
	return modified;
}

/*
	$DOCPROC:WriteExtendedPartitions
	write the extended partitions for this drive
*/
void PASCAL WriteExtendedPartitions(int iDrive)
{
	struct PARTITION * pPartition;
	struct ExtPart * ep;
	unsigned uIndex;

	ep = hdisk[iDrive].pepExt;
	while(ep)
	{
		/* if part sec is modified, write it out */
		if(IsChanged(ep->fModified))
		{
			// we only want the ID word, and the partition table.
			// all other bytes should be 0 in the extended partition sector
			memset(pBufferDMA, 0, sizeof(array));
			memmove(pBufferDMA+IDOFF, array+IDOFF, sizeof(unsigned));
			memmove(pBufferDMA+PARTOFF, ep->pPartition, NUMPART * sizeof(struct PARTITION));

			// write the extended partition sector
			WriteSector(iDrive, TrackToSector(iDrive, ep->ulStartTrack), pBufferDMA);
		}

		/* if a DOS partition has been created, initialize its boot sector */
		pPartition = FindPartitionType(ep->pPartition, DOS16);
		uIndex = (unsigned)(pPartition - ep->pPartition);
		if(pPartition && IsCreated(ep->fModified, uIndex))
			EraseSystemAreas(iDrive, ep->ulStartTrack, pPartition);

		// any more extended partitions?
		ep = ep->pepNext;
	}
}

/*
	Is the BPB valid?
*/
bool PASCAL IsValidBPB(BPB *pBPB)
{
	/* does BPB look valid? */
	if(pBPB->sec_size==SECSIZE && pBPB->media==HDID && pBPB->res_sec==1)
		return TRUE;
	return FALSE;
}

/*
	Returns size of drive in sectors.
*/
ulong PASCAL SizeDrive(BPB *pBPB)
{
	ASSERT(IsValidBPB(pBPB));
	if(pBPB->num_sec != 0)
		return pBPB->num_sec;
	return pBPB->lnumsec;
}

/*
	$DOCPROC:EraseSystemAreas

	Invalidate the boot sector, FAT and root directory. Ask user if
	the partition looks like it's the same size.
*/
void PASCAL EraseSystemAreas(int iDrive, ulong ulStartTrack,
					struct PARTITION *pPartition)
{
	ulong	ulFatRootSize;		/* size of FATs+root dir */
	ulong	ulSector;

	ulSector = TrackToSector(iDrive, ulStartTrack) + pPartition->hidden;

	/* does BPB look valid? And is it the same size? */
	if(!fDontConfirm)
	{
		/* if BPB looks valid and is the same size, ask if they REALLY
			want to wipe it out */
		ReadSector(iDrive, ulSector, pBufferDMA);
		if( (*(unsigned *)(pBufferDMA+IDOFF) == IDVAL) &&
			IsValidBPB((BPB *)(pBufferDMA+BPBOFFSET)) &&
			SizeDrive((BPB *)(pBufferDMA+BPBOFFSET)) == pPartition->size )
		{
			/* then ask if they REALLY want to wipe it out */
			printf(szOverwrite, iDrive+HD1, SectorToTrack(iDrive, ulSector));

			if(GetYesNo(szEraseAnyway) != YES)
				return;
		}
	}

	/* let user know what's going on */
	printf("Erasing system areas...\n");

	/* fill "boot sector" with F6s so DOS knows we need to FORMAT */
	memset(pBufferDMA, 0xF6, SECSIZE);
	WriteSector(iDrive, ulSector, pBufferDMA);

	/* calculate size of FAT and root directory based on partition size */
	ulFatRootSize = CalcSystemLen(iDrive, pPartition->size) - 1;

	/* fill FAT and root with 0s */
	memset(pBufferDMA, 0, SECSIZE);
	++ulSector;
	while(ulFatRootSize--)
	{
		WriteSector(iDrive, ulSector, pBufferDMA);
		++ulSector;
	}
}

/*
	Is this a DOS partition according to ID byte?
*/
int PASCAL IsDosPartitionId(uchar ucId)
{
	return (ucId==DOS12 || ucId==DOS16 || ucId==DOSBIG || ucId==DOS32 ||
		ucId==DOS32_LBA || ucId==DOSBIG_LBA);
}

/*
	Is this a nonDOS partition according to ID byte?
*/
int PASCAL IsSuperBootPartitionId(uchar ucId)
{
	return (ucId && !IsDosPartitionId(ucId) && !IsExtPartitionId(ucId));
}

/*
	Is this an extended DOS partition?
*/
int PASCAL IsExtPartitionId(uchar ucId)
{
	return (ucId == DOSEXT || ucId==DOSEXT_LBA);
}

/*
	$DOCPROC:ClearAllExtended
	Delete all extended partitions for this drive.
*/
void PASCAL ClearAllExtended(int iDrive)
{
	struct ExtPart * ep = 0;
	struct ExtPart * ep2;

	if(hdisk[iDrive].pepExt)
		ep = hdisk[iDrive].pepExt;

	/* free up the ExtPart structures */
	while(ep)
	{
		ep2 = ep->pepNext;
		dlFreeMem( FP_SEG(ep) );
		ep = ep2;
	}

	/* all extended partitions are deleted */
	hdisk[iDrive].pepExt = 0;
	hdisk[iDrive].iExtDosParts = 0;
}

/*
	$DOCPROC:ClearAllParts
	Clear all partition entries for this drive.
*/
void PASCAL ClearAllParts(int iDrive)
{
	int i;

	hdisk[iDrive].iDosParts =
	hdisk[iDrive].iExtDosParts =
	hdisk[iDrive].iSuperBootParts = 0;
	for(i=0; i<NUMPART; ++i)
		hdisk[iDrive].fModified[i] = FALSE;
	memset(hdisk[iDrive].pPartition, 0, NUMPART*sizeof(struct PARTITION));
}

/*
	$DOCPROC:ReadPartitions

	Read in all the partitions for all hard disks. This function assumes
	a maximum of one extended partition in the master boot record.

	Valid master boot record (sector 0)
	1) AA55 present
	2) max 1 active part
	3) max 1 extended part
	4) possible non-DOS parts

	Returns the count of all DOS partitions on all hard disks
*/
int PASCAL ReadPartitions(void)
{
	int iDrive, i;
	int iNumDosParts;
//	ulong ulBigPhysicalSectors;

	/* for each harddrive, record # of cylinders, heads, sectors */
	for(iDrive = 0; iDrive < iNumHardDisks; ++iDrive)
	{
		// get info about the drive
		if(!DlSmartLbaGetDriveParameters(iDrive+HD1, &hdisk[iDrive].lsipInfo))
			ErrorExit(szErrDriveParms);

		/* read in the Master Boot Sector (sector 0) */
		ReadSector(iDrive, 0, pBufferDMA);

		/* reset all parts of the disk structure */
		ClearAllParts(iDrive);

		/* if Master Boot Sector looks OK, see about the partitions */
		hdisk[iDrive].fCodeOK = (*(unsigned *)(pBufferDMA+IDOFF) == IDVAL);
		if(hdisk[iDrive].fCodeOK)
		{
			/* there are 4 partition table entries on a disk */
			memmove(hdisk[iDrive].pPartition, pBufferDMA+PARTOFF, IDOFF-PARTOFF);
			for(i=0; i<NUMPART; ++i)
			{
				/* if id field indicates used, validate ulpartition fields */
				if( hdisk[iDrive].pPartition[i].id &&
					(!hdisk[iDrive].pPartition[i].size ||
					!hdisk[iDrive].pPartition[i].hidden) )
				{
					/* invalid, show user raw dump */
					printf(szErrInvalidPartition, i);
					DumpRawPartitions(iDrive);
					if(fDontConfirm || GetYesNo(szDeleteInvalidPartition) != YES)
						exit(EXIT_BADDISK);

					/* delete that partition and continue */
					memset(&hdisk[iDrive].pPartition[i],0,sizeof(struct PARTITION));
				}

				/* DOS partition */
				if(IsDosPartitionId(hdisk[iDrive].pPartition[i].id))
					++hdisk[iDrive].iDosParts;

				/* nonDOS partition */
				else if(IsSuperBootPartitionId(hdisk[iDrive].pPartition[i].id))
					++hdisk[iDrive].iSuperBootParts;
			}

			/* read in all the partitions inside the extended partition */
			ReadExtendedParts(iDrive);

		} /* end of fCodeOK */

	} /* end of for(iDrive) loop */

	iNumDosParts = 0;
	for(iDrive = 0; iDrive < iNumHardDisks; ++iDrive)
		iNumDosParts += hdisk[iDrive].iDosParts + hdisk[iDrive].iExtDosParts;

	return iNumDosParts;
}

/*
	$DOCPROC:ReadExtendedParts
	Read in any extended DOS partitions for this drive.
*/
void PASCAL ReadExtendedParts(int iDrive)
{
	struct PARTITION * pPartition;
	struct ExtPart * ep;
	struct ExtPart * epPrev;
	ulong		ulStartTrack;
	ulong		ulSector;
	int		i;
	bool fSuccessful;
	unsigned uTempSeg;


	//  Set an initial value which can be asserted
	//
	epPrev = 0;


	/* check for the extended partition */
	pPartition = FindPartitionType(hdisk[iDrive].pPartition, DOSEXT);
	if(!pPartition)
	{
		hdisk[iDrive].pepExt = 0;
		return;
	}

	/* read in all the extended boot records */
	ulStartTrack = 0;
	while(pPartition)
	{
		/* read in the extended partition sector */
		ulSector = TrackToSector(iDrive, ulStartTrack) + pPartition->hidden;
		ReadSector(iDrive, ulSector, pBufferDMA);

		/* make sure the sector is valid */
		if(*(unsigned *)(pBufferDMA+IDOFF) != IDVAL)
			break;
		pPartition = (struct PARTITION *)(pBufferDMA+PARTOFF);

		// must have a valid DOS partition and extended partition to be valid
		if(!(IsDosPartitionId(pPartition[0].id) ||
				IsExtPartitionId(pPartition[1].id)))
			break;

		// must NOT have any other partitions in the sector
		if(pPartition[2].id || pPartition[3].id)
			break;

		/* allocate an extended structure to this */
		fSuccessful = dlAllocMem( CVT_PARA(sizeof(struct ExtPart)), &uTempSeg );
		SHIPASSERT( fSuccessful );
		ep = MK_FP( uTempSeg, 0 );
		if(!hdisk[iDrive].pepExt)
		{
			hdisk[iDrive].pepExt = ep;
		}
		else
		{
			SHIPASSERT(epPrev);
			epPrev->pepNext = ep;
		}
		epPrev = ep;

		/* record the information */
		ep->pepNext = 0;
		ep->ulStartTrack = SectorToTrack(iDrive, ulSector);

		for(i=0; i<NUMPART; ++i)
			ep->fModified[i] = FALSE;
		memmove(ep->pPartition, pBufferDMA+PARTOFF, NUMPART * sizeof(struct PARTITION));
		pPartition = FindPartitionType(ep->pPartition, DOS16);
		if(pPartition)
			++hdisk[iDrive].iExtDosParts;

		/* go on to next partition */
		pPartition = FindPartitionType(ep->pPartition, DOSEXT);

		/* all other "hidden" offsets are relative to the master partition */
		if(!ulStartTrack)
			ulStartTrack = ep->ulStartTrack;

	} /* end of while(pPartition) */
}

/*
	$DOCPROC:DisplayDiskSize
	Display the disk size (and cyls,heads,sectors) for this hard disk.
*/
void PASCAL DisplayDiskSize(int iDrive)
{
	ulong	ulTracks, ulHeads, ulSPT;
	char szSize[MAX_ASCII_MEGS];

	// determine disk size
	ulTracks = GetTracks(iDrive);
	ulHeads  = GetHeads(iDrive);
	ulSPT    = GetSPT(iDrive);
	TracksToAscii(iDrive, ulTracks, szSize);

	// report info about the disk
	printf(szFixedInfo, iDrive+1, iNumHardDisks,
		ulTracks, ulHeads, ulSPT, szSize);
}

/*
	Display all partitions on this drive.  This routine
	understands drive letters

	Fixed Disk 1 of 2:  Cylinders 307, Heads 8, Sectors 35, Size 33,123,456
	Drive  Bootable  Type       StartCyl    EndCyl   Size   Bytes
	  C:      Y      DOSBIG          1        200     199   33,123,456
	  E:      N      DOS16         308       1023    1234    9,123,456
	  F:      N      FAT32         xxx        yyy     zzz
	  Su      N      Super(98h)    201        307       5    9,123,456

	Fixed Disk 1 of 2:  Cylinders 307, Heads 8, Sectors 35, Size 33,123,456
	        No partitions
*/
void PASCAL DisplayPartitions(int iDrive)
{
	struct PARTITION * pPartition;
	struct PARTITION * pTempPartition;
	struct ExtPart * ep;
	int iPart;
	int i;

	/* show the size of the harddisk */
	DisplayDiskSize(iDrive);

	/* no partitions on this drive */
	if(!(hdisk[iDrive].iDosParts + hdisk[iDrive].iSuperBootParts + hdisk[iDrive].iExtDosParts))
	{
		printf(szNoPartitions);
		printf("\n");
		return;
	}

	/* print the header for the partitions */
	printf(szPartHeader);

	/* display each PRIMARY partition */
	iPart = 0;
	for(i=0, pPartition = hdisk[iDrive].pPartition; i<NUMPART; ++i, ++pPartition)
	{
		/* 0 means unused */
		if(!pPartition->id)
			continue;

		/* DOS or Super Boot partition (not an extension) */
		if(!IsExtPartitionId(pPartition->id))
		{
			DisplayOnePartition(iDrive, 0, pPartition);
			if(IsDosPartitionId(pPartition->id))
				++iPart;
		}
	} /* end of for loop */

	/* display each EXTENDED partition */
	pPartition = hdisk[iDrive].pPartition;
	for(i=0; i<NUMPART; ++i)
	{
		/* extended partition */
		if(IsExtPartitionId(pPartition->id))
		{
			ep = hdisk[iDrive].pepExt;
			while(ep)
			{
				pTempPartition = FindPartitionType(ep->pPartition,DOS16);
				if(pTempPartition)
				{
					DisplayOnePartition(iDrive,
						TrackToSector(iDrive, ep->ulStartTrack), pTempPartition );
					++iPart;
				}
				ep = ep->pepNext;
			}
		}

		++pPartition;
	}

	printf("\n");
}

/*
	$DOCPROC:DisplayOnePartition

	Display a single partition in human readable form.
*/
void PASCAL DisplayOnePartition(int iDrive, ulong ulStartSector,
		struct PARTITION *pPartition)
{
	char	szLetter[4];		/* "A:", "--" */
	char	cBoot;				/* 'Y' or 'N' */
	char	szTypeStr[16];		/* DOS16 */
	ulong	ulStartTrack;
	ulong	ulEndTrack;
	ulong	ulNumTracks;
	char	szSize[MAX_ASCII_MEGS];

	/* don't allow invalid partition types */
	if(pPartition->id == 0)
		return;

	/* calculate drive letter */
	GetPartitionName(szLetter, iDrive, pPartition);

	/* is this partition bootable? */
	cBoot = pPartition->boot ? 'Y' : 'N';

   /* what type of partition is this? */
   switch(pPartition->id)
   {
      case DOS12:      strcpy(szTypeStr,szId12); break;
      case DOS16:      strcpy(szTypeStr,szId16); break;
      case DOSEXT:     strcpy(szTypeStr,szIdExt); break;
      case DOSBIG:     strcpy(szTypeStr,szIdBig); break;
      case DOS32:      strcpy(szTypeStr,szId32); break;
      case DOS32_LBA:  strcpy(szTypeStr,szId32Lba); break;
      case DOSBIG_LBA: strcpy(szTypeStr,szIdBigLba); break;
      case DOSEXT_LBA: strcpy(szTypeStr,szIdExtLba); break;
      case RELIANCE:   strcpy(szTypeStr,szIdDLRel); break;
      case DEFAULT_SUPERBOOT_ID:
         sprintf(szTypeStr,szIdSuper,pPartition->id); break;
      default:
         sprintf(szTypeStr,szIdNonDos,pPartition->id); break;
   }

	/* determine starting and ending cylinder */
	ulStartTrack = SectorToTrack(iDrive, ulStartSector + pPartition->hidden);
	ulNumTracks  = NumTracks(iDrive, pPartition->size);
	ulEndTrack   = ulStartTrack + ulNumTracks - 1;

	/* determine size of partition in bytes */
	TracksToAscii(iDrive, ulNumTracks, szSize);

	/* display it */
	printf(szPartInfo,szLetter,cBoot,szTypeStr,ulStartTrack,
			ulEndTrack,ulNumTracks,szSize);
}

/*
	$DOCPROC:GetPartitionName

	Returns the partition in a "name" form. For use with display
	partition.
*/
void PASCAL GetPartitionName(char *szStr, int iDrive,
		struct PARTITION *pPartition )
{
	int iId;


	iId = pPartition->id;

	if(IsDosPartitionId(iId))
	{
		strcpy(szStr, "A:");
		*szStr = PartitionLetter(iDrive,pPartition);
	}
	else if(IsExtPartitionId(iId))
		strcpy(szStr, "Ex");		// extended partition
	else if(iId == DEFAULT_SUPERBOOT_ID)
		strcpy(szStr, "SB");		// Super Boot partition
	else if(iId)
		strcpy(szStr, "--");		// Unknown
	else
		strcpy(szStr,"  ");		// unused
}

/*
	$DOCPROC:MegsToTracksWithLimitCheck

	Returns 0 if failed. Returns # of tracks otherwise.
*/
ulong PASCAL MegsToTracksWithLimitCheck(int iDrive, ulong ulMegs,
			ulong ulMaxTracks)
{
	ulong	ulNumTracks;
	ulong	ulMaxMegs;

	ASSERT(ulMaxTracks);
	ASSERT(iDrive >= 0 && iDrive < iNumHardDisks);

	/* switch says partitions can't be bigger than 2 Gig */
	if(!swFat32 && ulMegs > 2048)
		return 0;

	/* does it fit the maximum? */
	ulMaxMegs = TracksToMegs(iDrive, ulMaxTracks);
	if(ulMegs > ulMaxMegs)
		return 0;

	/* return in track form */
	ulNumTracks = MegsToTracks(iDrive, ulMegs);
	return ulNumTracks;
}

/*
	$DOCPROC:MegsToTracks

	Convert megabytes to tracks (aka cylinders) for this drive.
*/
ulong PASCAL MegsToTracks(int iDrive, ulong ulMegs)
{
	ulong ulSectors;
	ulong	ulSectorsEachTrack;
	ulong ulRemainder;
	ulong ulTracks;

	/* validate parameters */
	ASSERT(ulMegs < TWO_TERABYTES);
	ASSERT(iDrive >= 0 && iDrive < iNumHardDisks);

	/* determine # of sectors in 1 cylinder */
	ulSectorsEachTrack = GetHeads(iDrive) * GetSPT(iDrive);

	/* convert Megs to # of sectors */
	ulSectors = ulMegs * SECTORS_PER_MEG;
	ulTracks  = ulSectors / ulSectorsEachTrack;

	/* round up if .5 or greater */
	ulRemainder = ulSectors % ulSectorsEachTrack;
	if(ulRemainder)
	{
		++ulTracks;
	}

	/* return the # of cylinders */
	return ulTracks;
}

/*
	$DOCPROC:TracksToMegs

	Convert tracks (aka cylinders) to megabytes for this drive.
*/
ulong PASCAL TracksToMegs(int iDrive, ulong ulNumTracks)
{
	ulong	ulSectors;
	ulong	ulSectorsEachTrack;
	ulong ulRemainder;
	ulong ulMegabytes;

	/* validate parameters */
	ASSERT(iDrive >= 0 && iDrive < iNumHardDisks);

	/* determine total # of sectors */
	ulSectorsEachTrack = GetHeads(iDrive) * GetSPT(iDrive);
	ulSectors = ulNumTracks * ulSectorsEachTrack;

	/* convert to megabytes */
	ulMegabytes = ulSectors / SECTORS_PER_MEG;

	/* remainder? round up */
	ulRemainder = ulSectors % SECTORS_PER_MEG;
	if(ulRemainder >= SECTORS_PER_MEG/2)
		++ulMegabytes;

	/* return megabytes */
	return ulMegabytes;
}

/*
	$DOCPROC:TracksToAscii

	Convert a number of tracks for this drive into an ascii string.
	Always displays in Megabytes if greater than 1 Meg.
*/
void PASCAL TracksToAscii(int iDrive, ulong ulTracks, char szBytes[])
{
	static char szWithCommas[] = "%,lu MB";
	ulong		ulMegabytes;

	ASSERT( iDrive >= 0 && iDrive < iNumHardDisks );

	// determine size in megabytes
	ulMegabytes = TracksToMegs(iDrive, ulTracks);

	// convert to a comma separated string
	sprintf(szBytes, szWithCommas, ulMegabytes);
}

/*
	$DOCPROC:GetPartitionSizeFromUser

	Does no disk access, but simply gets a number of cylinders from
	the user. Accepts cylinders, Kbytes, Megabytes, and Gigabytes.

	On Entry
		*pulNumTracks		max size of partition on disk
	On Exit
		*pulNumTracks		size user has selected
	Returns TRUE if worked, FALSE if user aborted.
*/
bool PASCAL GetPartitionSizeFromUser(int iDrive, ulong *pulNumTracks)
{
	char	szBytes[20];
	ulong	ulNumTracks2Gig;
	ulong	ulMaxTracks, ulNewSize;

	/* don't allow partitions larger than 2 Gigabytes unless FAT32 enabled */
	ulMaxTracks  = *pulNumTracks;
	if(!swFat32)
	{
		ulNumTracks2Gig = GetMaxTracksFor2Gigabytes(iDrive);
		if(ulMaxTracks > ulNumTracks2Gig)
			ulMaxTracks = ulNumTracks2Gig;
	}

	/* get the size of the partition (may be in cyls, or megs) */
	while(1)
	{
		/* display maximum partition size (in both cyls and bytes) */
		TracksToAscii(iDrive, ulMaxTracks, szBytes);
		printf(szMaxPart, ulMaxTracks, szBytes);

		/* get the size of the partition, in cylinders or megs */
		printf(szGetPartSize, ulMaxTracks);
		if(GetString(szBytes, 7, FALSE, TRUE ) < 0)
			return FALSE;						/* user aborted */

		/* user accepted default */
		if(*szBytes == 0)
		{
			ulNewSize = ulMaxTracks;
			break;
		}

		/* get the size the user requested */
		ulNewSize = atol(szBytes);
		strupr(szBytes);
		if(strchr(szBytes,'M'))
			ulNewSize = MegsToTracksWithLimitCheck(iDrive, ulNewSize, ulMaxTracks);

		/* got a valid size in tracks */
		if(ulNewSize > 0 && ulNewSize <= ulMaxTracks)
			break;

		/* invalid size, ask again */
		printf(szInvalidSize);
	}

	/* double space it */
	printf("\n");

	/* everything worked */
	*pulNumTracks = ulNewSize;
	return TRUE;
}

/*
	$DOCPROC:GetCmdlineSize

	Get the partition size based on cmdline input. The input can be
	in any form (bytes, Kbytes Mbytes), but the result is converted
	to tracks.

	On Entry
		*pulNumTracks		maximum # tracks available
	On Exit
		*pulNumTracks		# tracks requested by user

*/
void PASCAL GetCmdlineSize(int iDrive, long lMegabytes, ulong *pulNumTracks)
{
	ulong	ulNumTracks;
	ulong	ulMaxTracks;

	/* validate parameters */
	ASSERT(iDrive >= 0 && iDrive < iNumHardDisks);
	ASSERT(pulNumTracks && *pulNumTracks);

	/* make sure partition size is in range */
	ulMaxTracks = *pulNumTracks;
	if(!swFat32)
	{
		ulMaxTracks = GetMaxTracksFor2Gigabytes(iDrive);
		if(ulMaxTracks > *pulNumTracks)
			ulMaxTracks = *pulNumTracks;
	}

	/* user said to use maximum */
	if(lMegabytes == 0 || lMegabytes == -1L)
		ulNumTracks = ulMaxTracks;

	/* convert user input megabytes to tracks */
	else
	{
		ulNumTracks = MegsToTracksWithLimitCheck(iDrive, lMegabytes, ulMaxTracks);
		if(!ulNumTracks)
			ulNumTracks = ulMaxTracks;
	}

	/* return the number of tracks */
	*pulNumTracks = ulNumTracks;
}

/*
	$DOCPROC:GetPartitionLetter

	Get a drive letter from the user.
*/
int PASCAL GetPartitionLetter(int iDrive, char *pcLetter)
{
	int	iChoice;
	int	i;
	char	szList[28];		/* a list of drive letters */
	struct ExtPart * pExtPart;
	unsigned uPartIndex;

	/* if no DOS partitions, no drive letter is possible */
	if(!(hdisk[iDrive].iDosParts + hdisk[iDrive].iExtDosParts))
		return CANCEL;

	/* build a list of possible drive letters */
	iChoice = 0;
	for(i=0; i<hdisk[iDrive].iDosParts; ++i)
	{
		if(IsDosPartitionId(hdisk[iDrive].pPartition[i].id))
		{
			szList[iChoice++] = PartitionLetter(iDrive,
					&hdisk[iDrive].pPartition[i] );
		}
	}
	pExtPart = hdisk[iDrive].pepExt;
	for(i=0; i<hdisk[iDrive].iExtDosParts; )
	{
		for( uPartIndex = 0; (uPartIndex < NUMPART) &&
				(i < hdisk[iDrive].iExtDosParts) ; uPartIndex++ )
		{
			if(IsDosPartitionId(pExtPart->pPartition[uPartIndex].id))
			{
				i++;
				szList[iChoice++] = PartitionLetter(iDrive,
						&pExtPart->pPartition[uPartIndex] );
			}
		}
		pExtPart = pExtPart->pepNext;
	}
	szList[iChoice] = 0;

	/* get the drive letter from the user */
	iChoice = GetChoice(szDelByLetter,szList,szList[iChoice-1]);
	if(iChoice >= 0)
		*pcLetter = szList[iChoice];

	return iChoice;
}

/*
	$DOCPROC:GetMenuCommand

	Display the menu and get a choice. Returns a pointer to the
	function to execute.
*/
FUNC_PTR PASCAL GetMenuCommand(struct menu *menu)
{
	#define MAXCHOICE	20
	int i;
	char choices[MAXCHOICE+1];

	/* display the menu */
	for(i=0; menu[i].cmd && i<MAXCHOICE; ++i)
	{
		printf("%c) %s\n", menu[i].cmd, menu[i].str);
		choices[i] = menu[i].cmd;
	}
	choices[i] = 0;
	printf("\n");

	/* Get the command from the user, abort is not allowed here */
	do
	{
		i = GetChoice(szGetCmd, choices, *choices);
	} while(i == CANCEL);
	printf("\n");

	return menu[i].func;
}

/*
	$DOCPROC:GetYesNoWithDefault

	Gets a yes or no question. The default may be yes or no.

	Waits until the user presses Yes, No, Esc (cancel) or Return
	(default).
*/
int PASCAL GetYesNoWithDefault(char szPrompt[], bool fYes)
{
	char c;
	char cDefault;

	/* set up a default character */
	if(fYes == YES)
		cDefault = YES_CHAR;
	else
		cDefault = NO_CHAR;

	printf(szAskYesNo, szPrompt, cDefault);
	while(1)
	{
   	dlGetKey( &c );
		c = toupper(c);
		if(c == '\n')	/* ignore linefeeds */
			continue;

		/* use default*/
		if(c == '\r')
			c = cDefault;

		/* display the yes or no */
		if(c==YES_CHAR || c==NO_CHAR)
		{
			printf("%c\n",c);
			fYes = (c==YES_CHAR ? YES : NO);
			break;
		}

		/* display the cancel */
		if(c==ESC)
		{
			printf(szEsc);
			fYes = CANCEL;
			break;
		}
	}

	printf("\n");
	return fYes;
}

/*
	$DOCPROC:GetYesNo

	Get a yes or no answer. Yes returns "YES", No returns "NO", ESC
	returns "CANCEL". No other characters are accepted. Defaults to NO.
*/
int PASCAL GetYesNo(char szPrompt[])
{
	return GetYesNoWithDefault(szPrompt, NO);
}

/*
	$DOCPROC:GetEsc

	Wait for the escape key before returning to main menu. An error
	has occurred.
*/
void PASCAL GetEsc(void)
{
	char c;

	/* cmdline version should just exit */
	if(fDontConfirm)
		exit(EXIT_CREATE);

	printf(szPressEsc);
	do
	{
		dlGetKey(&c);
	} while(c != ESC);
	printf("\n\n");
}

/*
	$DOCPROC:GetChoice
	Get a value, displaying the default. Returns -1 on abort.
*/
int PASCAL GetChoice(const char *prompt, char *choices, char default_choice)
{
	char c;
	char *p;
	int i;

	/* choices MUST be uppercase, and there must be at least 1 choice */
	strupr(choices);
	if(*choices == 0)
		return CANCEL;

	/* set a default choice */
	if(!strchr(choices,default_choice))
		default_choice = *choices;

	while(1)
	{
		printf("%s [%c]: ",prompt,default_choice);
		dlGetKey(&c);
		c = toupper(c);
		if(c == '\n')	/* ignore linefeeds */
			continue;

		/* user aborted */
		if(c == ESC)
		{
			printf(szEsc);
			return CANCEL;
		}

		/* user chose default */
		if(c == '\r')
			c = default_choice;
		printf("%c\n",c);

		/* check for bad choice */
		p=strchr(choices,c);
		if(!c || !p)
		{
			printf("Please enter one of (");
			for(i=0; choices[i]; ++i)
			{
				printf("%c",choices[i]);
				if(choices[i+1])
					printf(",");
			}
			printf(")\n");
		}
		else
			break;
	}

	return (int)(p - choices);
}

/*
	$DOCPROC:GetHexValue

	Get a hex value (number from min-max) and return it, or -1
	if user aborted.  The number must be a positive integer.
*/
int PASCAL GetHexValue(const char *szPrompt,int oldval,int minval,int maxval)
{
	char szVal[6];
	char *psz;
	int  value;

	while(1)
	{
		printf("%s [%Xh]: ",szPrompt,oldval);

		/* get the string, check for abort */
		if(GetString(szVal,5,TRUE,FALSE) < 0)
		{
			value = CANCEL;
			break;
		}

		/* just use default value */
		if(*szVal == 0)
			value = oldval;

		/* user typed in a value */
		else
		{
			// skip whitespace then get the value
			psz = szVal;
			while(DL_IS_SPACE(*psz))
				++psz;
			value = atoh(psz);
		}

		/* make sure value is inbounds */
		if(value<minval || value>maxval)
			printf(szErrBadValue, minval, maxval);

		else
			break;
	}

	/* return the number */
	return value;
}

/*
	$DOCPROC:GetDriveNumber

	Return the drive number for the given operation. If only 1 drive,
	then return that drive without prompting user.
*/
int PASCAL GetDriveNumber(char *szDrive)
{
	static int iLastDrive = 0;
	char	acDriveList[MAXFIXED+1];
	int	iDrive;

	/* only one drive, no need to ask */
	if(iNumHardDisks <= 1)
		return 0;

	/* create the drive list */
	for(iDrive=0; iDrive<iNumHardDisks; ++iDrive)
		acDriveList[iDrive] = iDrive+'1';
	acDriveList[iDrive] = 0;

	/* ask user for drive, defaults to last drive picked */
	sprintf(zz, szDrive, iNumHardDisks);
	iDrive = GetChoice(zz,acDriveList,acDriveList[iLastDrive]);
	if(iDrive != CANCEL)
		iLastDrive = iDrive;
	return iDrive;
}

/*
	$DOCPROC:GetString

	Returns the length of the string, or "CANCEL" if none.
*/
int PASCAL GetString(char *s, int maxlen, bool fIsHex, bool fCanHaveM )
{
	int len=0;
	char c;

	while(1)
	{
		dlGetKey(&c);
		c = toupper(c);

		if(c == ESC)
		{
			printf(szEsc);
			len = CANCEL;
			break;
		}
		if(c == '\n' || c == '\r')
		{
			s[len] = 0;
			break;
		}
		else if(c == '\b' && len)
		{
			printf("\b \b");
			--len;
		}
		else
		{
			if( ((fIsHex) && (c >= 'A') && (c <= 'F')) ||
					((c >= '0') && (c <= '9')) ||
					((fCanHaveM) && (c == 'M')) )
			{
				if(isprint(c))
				{
					if(len+1 >= maxlen)
						printf("\a");
					else
					{
						printf("%c",c);
						s[len++] = c;
					}
				}
			}
		}
	}

	printf("\n");
	return len;
}

/*
	A simple ASSERT function.
*/
void InternalError(int iFileNum, int iLineNum)
{
	printf(szErrInternal, iFileNum, iLineNum);
	exit(EXIT_ASSERT);
}
