//
//   $Id: FDISK.H 1.6 2002/10/16 21:49:00Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:FDISK.H

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
   Datalight Confidential and Proprietary.

	$DOCHISTORY
	09/09/1998 dg  Added Super-Boot capability.
	08/13/1990 dg  Created by Drew Gislason.
*********************************************************************/

#define CVT_PARA( bytes ) ((bytes+0xf) / 0x10)


/* default Super-Boot identifier */
#define DEFAULT_SUPERBOOT_ID	0x98

/* first hard drive from a BIOS perspective */
#define HD1		0x80

/* partition table information */
#define SUPERBOOTID_OFF	0x1BB		/* offset of Super-Boot ID */
#define SUPERHOTKEY_OFF	0x1BC		/* offset of hotkey */

#define NUMPART	4			/* 4 partitions per boot sector */
#define IDOFF		0x1FE
#define IDVAL		0xAA55	/* value of boot record id */

/* partition IDs in extended partitions are always old-style IDs */
#define OLD_STYLE	0
#define NEW_STYLE 1

/* offset of BPB in boot sector */
#define BPBOFFSET 11
#define HDID		0xF8

#define MAXFIXED	24			/* max # of fixed disks */
#define SECSIZE	512
#define MINTRAK	1			/* minimum traks in a partition */

/* how many bytes are required to hold ascii megabytes, 2,097,152 MB */
#define MAX_ASCII_MEGS	16
#define SECTORS_PER_MEG	2048		// sectors in 2 megs, with 512 byte sectors
#define TWO_TERABYTES (2L*1024*1024)	// 2 Terabytes expressed in megs

/* value for determining partition sysid */
#define MAXBIOSCYLINDERS 1024 // max tracks (cylinders) BIOS int 13h supports

#define _512MEG	0x100000L	/* sectors in a 512 Meg partition */
#define MAX32MSEC	0xFFF0		/* maximum sectors in a 32Meg partition */
#define SPC12BIT	8				/* spc for a 12 bit FAT partition */
#define MAXFAT12	0xFF7			/* maximum clusters in a 12 bit partition */

#define EXIT_USAGE   1
#define EXIT_NODISKS 2
#define EXIT_BADDISK 3
#define EXIT_CREATE	4		/* error creating partition */
#define EXIT_ASSERT  10

// for GetYesNo()
#define YES     1
#define NO      0
#define CANCEL -1

// changed flag
#define CREATED	5

/*
	The extended partition is broken up into 1 or more subsections.
	That is, the 1 extended partition in the Master Boot Record contains
	ALL these extended partitions.

	Example: [ext1 [ext2 [ext3]]]
*/
struct ExtPart
{
	/* pointer to the next subsection of this extended partition */
	struct ExtPart * pepNext;

	/* where is this particular sector? */
	ulong	ulStartTrack;

	/* extended partition info */
	struct PARTITION pPartition[NUMPART];
	bool fModified[NUMPART];
};

/*
	This structure defines each physical hard disk and its
	logical partitions.
*/
struct HardDisk
{
	LSIP	lsipInfo;				// info packet about disk
	int	iDosParts;				// # of primary DOS partitions on this disk
	int	iExtDosParts;			// # of extended DOS partitions
	int	iSuperBootParts;		// # of super boot partitions
	bool	fCodeOK;					// was the master boot sector valid?
	struct ExtPart *pepExt;		// if there are any extended partitions...

	/* there are 4 partition table entries on a disk */
	struct PARTITION pPartition[NUMPART];	// the partition table
	bool fModified[NUMPART];					// is each partition modified?
};

// function pointer
typedef void (*FUNC_PTR)(void);

/*
	The FDISK menu.
*/
struct menu
{
	char cmd;			/* command character */
	char *str;			/* string describing command */
	FUNC_PTR func;		/* function pointer */
};
