//
//   $Id: FIND.C 1.9 2002/10/16 21:49:50Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	FIND.C  created by Roy Sherrill

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

**********************************************************/


#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dosver.h"
#include "fatworld.h"
#include "dldosio.h"
#include "dllfn.h"
#include "stdc.h"

#define FIND_COM 1
#include "dosmsg.c"

char signon[] = "FIND %s %s\n%s\n\n";

#define LINESIZE 250

int input = STDIN;
char first_c, *str, line[LINESIZE];

int Count_Only, Line_Number,  Not_Match;
int line_no, Match_in_line;
unsigned count;
bool fNonCaseSensitive = FALSE;

bool fIsFat32World = FALSE;
bool fHideSignon = FALSE;
bool fLFNsSupported = FALSE;

char * cFgets(int iFileHandle, char * szReadBuf, int iMaxRead);

void DisplaySignOnMessage( void )
{
	if( !fHideSignon )
	{
		if( fIsFat32World )
		{
			printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}

void do_help(void)
{
	DisplaySignOnMessage();
	printf(USAGE);
}

void main(int argc, char *argv[])
{
	int i, iActionTaken;
	char *p;
	bool fResult;
	bool fFirstArgFound = FALSE;
	bool fBadFormat = FALSE;
	unsigned uDeviceAttributes;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	//  Determine if LFNs are support
	//
	fLFNsSupported = LFNPresent();


	for(i=1; i <= argc; i++)
	{
		if( !fFirstArgFound )
		{
			if( argv[i][0] != '"' )
			{
				if( argv[i][0] != '/' )
				{
					fBadFormat = TRUE;
				}
			}
			else
			{
				fFirstArgFound = TRUE;
			}
		}

		LFNStripArgQuotes( argv[i] );
		if( strlen(argv[i]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(255);
		}
	}

	if(argc < 2)
	{
		printf( szMissingParm );
		exit(2);
	}

	for(i=1; i<argc; ++i)
	{
		if(argv[i][0] == '/' || argv[i][0] == '-')
		{
			switch(argv[i][1])
			{
				case 'c':
				case 'C':
					Count_Only = TRUE;
					break;
				case 'h':
				case 'H':
				case '?':
					do_help();
					exit(0);

				case '[':
					fHideSignon = TRUE;
					break;

				case 'n':
				case 'N':
					Line_Number = TRUE;
					break;

				case 'v':
				case 'V':
					Not_Match = TRUE;
					break;

				case 'i':
				case 'I':
					fNonCaseSensitive = TRUE;
					break;

				default:
					printf(bad_opt, argv[i]);
					exit(2);
			}
		}
		else
			break;
	}

	if( fBadFormat )
	{
		printf( szBadFormat );
		exit(255);
	}

	str = argv[i];
	++i;				/* Move to next for filename */
	first_c = *str;

// BT 5-20-01
//
//  Replaced the isatty function with a datalight library function
//
//	for(i=i; i<argc || (i == argc && !isatty(fileno(stdin))); ++i)
	dlGetDeviceAttributes( STDIN, &uDeviceAttributes );
	for(i=i; (i<argc) || ((i == argc) && !(uDeviceAttributes & 0x80)); ++i)
	{
		if( (fIsFat32World) || (fLFNsSupported) )
		{
			//
			//  Don't change the case of the file name if in DOS 7.1
			//  or an LFN enviornment.
			//
		}
		else
		{
			strupr( argv[i] );
		}

		dlGetDeviceAttributes( STDIN, &uDeviceAttributes );
		if( uDeviceAttributes & 0x80 )
		{
			SmartCreateOpenFile(OPEN_ACCESS_READONLY, 0, FILE_OPEN, argv[i],
				&input, &iActionTaken);
			if(input == NULL)
			{
				fprintf( STDERR, cant_open_str, argv[i]);
				if( (!fIsFat32World) && (!fLFNsSupported) )
				{
					fprintf(STDERR, "\n" );
				}
				exit(2);
			}
		}

		printf("\n---------- %s", argv[i]);

		if(!Count_Only)
		{
			printf("\n");
		}

		line_no = 0;
		while(cFgets(input, line, LINESIZE-3))
		{
			if( *line )
			{
				p = line + strlen(line) - 1;
				if( *p == 0x0a )
				{
					*p = 0;
				}
			}

			if( *line )
			{
				p = line + strlen(line) - 1;
				if( *p == 0x0d )
				{
					*p = 0;
				}
			}

			++line_no;
			p=line;
			Match_in_line = FALSE;

			while( *p )
			{
				/* look at all occurances of first_c */
				if( fNonCaseSensitive  )
				{
					fResult = (strncmpi(p, str, strlen(str)) == 0);
				}
				else
				{
					fResult = (strncmp(p, str, strlen(str)) == 0);
				}
				if( fResult )
				{
					Match_in_line = TRUE;
					if(!Not_Match)
					{
						if(!Count_Only)
						{
							if(Line_Number)
								printf("[%u]",line_no);
							printf("%s\n", line);
						}
						++count;
					}
					break;		/* stop if found one */
				}
				else
					++p;		/* move past current first_c */
			}

			if(Match_in_line == FALSE && Not_Match)
			{
				++count;
				if(!Count_Only)
				{
					if(Line_Number)
						printf("[%u]",line_no);
					printf("%s\n", line);
				}
			}
		}

		dlCloseFile(input);
	}

	if(Count_Only)
	{
		printf(match_str, count);
	}

	if( count )
	{
		exit(0);
	}
	else
	{
		exit(1);
	}
}




/*
	$DOCPROC: cFgets

	char * cFgets(int iFileHandle, char * szReadBuf, int iMaxRead);

	Reads one line (CR/LF terminated) from the file.  Meant to be a
	substitute for fgets().

	On Entry:
	iFileHandle = DOS file handle of file to read from
	szReadBuff = Buffer into which text is read
	iMaxRead = Maximum number of bytes that should be read

*/
char * cFgets(int iFileHandle, char * szReadBuf, int iMaxRead)
{
	int iReadCount=0;
	unsigned uBytes;
	bool fSuccessful;

	while(iReadCount < iMaxRead)
	{
		uBytes=1;
		fSuccessful = dlReadFile(iFileHandle, &uBytes, szReadBuf);
		if(!fSuccessful)
			return(NULL);

		if(uBytes == 0 && iReadCount == 0)
			return(NULL);

		// If the end of the file was reached, terminate string and return.
		if(uBytes == 0)
		{
			*szReadBuf = NULL;
			return(szReadBuf-iReadCount);
		}

		// If last 2 bytes read where a newline code(CRLF), terminate string
		// and return.
		if(*szReadBuf == '\n')
		{
			*(szReadBuf+1) = NULL;
			return(szReadBuf-iReadCount);
		}

		iReadCount++;
		szReadBuf++;
	}
	return(0);
}
