;
;   $Id: FIXPAC.ASM 1.4 2002/10/16 21:49:50Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;---------------------------------------------------------------------------
; $DOCMODULE:FIXPAC.ASM
;
; Copyright (c) 1999-2002 Datalight, Inc.
; All Rights Reserved.
; Datalight Proprietary and Confidential.
;
; This program is used to "fix" older programs that were linked with the
; Microsoft "pack" feature. These older programs have a segment math
; error if DOS is loaded high (less than 64K memory is used in lower RAM).
;
; Usage = FIXPAC [/U]
;
; $DOCHISTORY
; 06/02/1999 dg  Created by Drew Gislason
;---------------------------------------------------------------------------
.model small

;---------------------------------------------------------------------------
; INCLUDE FILES
;---------------------------------------------------------------------------
include dosver.inc


;---------------------------------------------------------------------------
; EQUATES AND DEFINES
;---------------------------------------------------------------------------
THRESHOLD	equ	1000h	; segment 1000h = 64k

PSP_CMDLINE	equ	81h
PSP_ENVSEG	equ	2Ch
PSP_PARAS	equ	10h	; 100h bytes in PSP = 10h paras
MCB_ID		equ	0
MCB_OWNER	equ	1
MCB_PARAS	equ	3
OPTION_CHAR	equ	'/'
SPACE_CHAR	equ	' '
HELP_CHAR	equ	'?'
UNLOAD_CHAR	equ	'U'
CR		equ	0Dh
LF		equ	0Ah
TAB		equ	09h

; error exit code
ERR_NONE	equ	0
ERR_USAGE	equ	1
ERR_INSTALL	equ	2


;---------------------------------------------------------------------------
; PROGRAM CODE
;---------------------------------------------------------------------------
.code

	; start of program
	org	100h
_astart:

	; unique 16 byte signature to identify FixPac in memory
	jmp	short EndSignature
	db	55h, "FixPac", 0AAh, 'D', 'A', 'G', 'A', 'J', 'G'
    EndSignature:

	; display signon
	mov	dx, cs
	mov	ds, dx
	mov	dx, offset _TEXT:szSignOn
	mov	ah, 9
	int	21h

	; find option (if there is one) ignoring spaces
	mov	bx, PSP_CMDLINE
     AgainChar:
	cmp	byte ptr es:[bx], ' '
	jz	IgnoreChar
	cmp	byte ptr es:[bx], TAB
	jnz	OnOption
    IgnoreChar:
	inc	bx
	jmp	AgainChar
    OnOption:
	cmp	byte ptr es:[bx], CR
	jz	InstallFixPac

	; is it help?
	cmp	byte ptr es:[bx], OPTION_CHAR
	jnz	BadOption
	inc	bx
	mov	al, byte ptr es:[bx]
	cmp	al, HELP_CHAR
	jnz	NotHelp
	mov	al, ERR_USAGE
	mov	dx, offset _TEXT:szUsage
	jmp	ErrorExit
    NotHelp:

	; is it unload?
	and	al, not 20h	; toupper
	cmp	al, UNLOAD_CHAR
	jnz	BadOption

	;------------------------------------
	; user said to unload FixPac
	;------------------------------------
	call	IsFixPacInstalled
	or	ax, ax
	jz	NotInstalled

	; found it, now free up former fixpac memory
	push	ax
	mov	es, ax
	mov	es, es:[PSP_ENVSEG]
	mov	ah, 49h
	int	21h		; free the environment
	pop	ax
	mov	es, ax
	mov	ah, 49h
	int	21h		; free former FixPac itself

	; display message and exit
	mov	al, ERR_NONE
	mov	dx, offset _TEXT:szRemoved
	jmp	MessageExit

	; trying to uninstall when fixpac is not installed
    NotInstalled:
	mov	al, ERR_INSTALL
	mov	dx, offset _TEXT:szErrNotInstalled
	jmp	ErrorExit

	; bad option
    BadOption:
	mov	al, ERR_USAGE
	mov	dx, offset _TEXT:szErrOption
	jmp	ErrorExit

	;------------------------------------
	; user said to install FixPac
	;------------------------------------
    InstallFixPac:

	; is fixpac already installed? don't install twice
	call	IsFixPacInstalled
	or	ax, ax
	jz	Ok2Install
	mov	al, ERR_INSTALL
	mov	dx, offset _TEXT:szErrAlreadyInstalled
	jmp	ErrorExit
    Ok2Install:

	; determine end of FixPac code
	mov	dx, offset EndSignature
	mov	cl, 4
	shr	dx, cl
	add	dx, PSP_PARAS	; dx now has paragraphs
	mov	ax, cs
	add	ax, dx

	; don't install if alread over threshold
	cmp	ax, THRESHOLD
	jae	NoNeedToInstall

	; let user know we installed
	push	dx
	mov	dx, offset _TEXT:szInstalled
	mov	ah, 9
	int	21h
	pop	dx

	; determine amount of memory to eat up, then...
	; Terminate and Stay Resident
	mov	ax, cs
	mov	dx, THRESHOLD
	sub	dx, ax			; dx = paras to keep
	mov	ah, 31h
	int	21h

	; should never get here! (but just in case)
	jmp	Exit

    NoNeedToInstall:
	mov	al, ERR_INSTALL
	mov	dx, offset _TEXT:szErrNoNeedToInstall
	jmp	ErrorExit

	; display the error message, then exit
    MessageExit:
    ErrorExit:
	push	ax	; preserve error code (may be 0) in al
	mov	ah, 9
	int	21h
	pop	ax

	; all done, exit
    Exit:
	mov	ah, 4Ch
	int	21h


;---------------------------------------------------------------------------
; $DOCPROC:IsFixPacInstalled
;
; OnEntry	nothing
; OnExit	ax = 0 if fixpac not found, ax = seg psp if found
;---------------------------------------------------------------------------
IsFixPacInstalled proc near
	push	ds
	push	es
	push	si
	push	di

	; get first PSP from DOS
	mov	ah, 52h
	int	21h
	mov	es, es:[bx-2]

	; get our own code segment
	mov	ax, cs
	mov	ds, ax

    LookAgainForPSP:

	; stop when we've reached the FixPac that's running or
	; when at end of chain
	mov	ax, ds
	mov	dx, es
	cmp	dx, ax
	jae	FixPacNotInstalled
	cmp	byte ptr es:[MCB_ID], 'Z'
	jz	FixPacNotInstalled

	; don't look at free blocks, or non-PSP blocks
	cmp	word ptr es:[MCB_OWNER], 0
	jz	NextMCB

	; is this the FixPac program?
	push	es
	mov	ax, es
	inc	ax

	; prepare for compare. PSP is 1 paragraph past MCB
	mov	es, ax
	mov	si, 100h
	mov	di, si
	mov	cx, (offset _TEXT:EndSignature - offset _TEXT:_astart) + 1

	; not a PSP, ignore
	cmp	word ptr es:[0], 20CDh		; int 20h
	jnz	NotAPSP		; cx will not be 0, so not found

	; are the signatures identical?
	cld
	repz	cmpsb		; compare signatures, cx=0 if identical
	mov	ax, es		; put PSP in ax
    NotAPSP:
	pop	es		; restore MCB
	jcxz	FixPacFoundInMemory

	; not found, try next MCB in the chain
    NextMCB:
	mov	ax, es:[MCB_PARAS]
	mov	dx, es
	add	ax, dx
	inc	ax
	mov	es, ax
	jmp	LookAgainForPSP

    FixPacNotInstalled:
	xor	ax, ax
    FixPacFoundInMemory:
	pop	di
	pop	si
	pop	es
	pop	ds
	ret
IsFixPacInstalled endp

;---------------------------------------------------------------------------
; Text Messages
;---------------------------------------------------------------------------
szSignOn db "FixPac ",FULLVERSION,CR,LF
	 db COPYRIGHT,CR,LF
	 db CR,LF,'$'

szRemoved	label byte
	db "FixPac removed from memory.",CR,LF,'$'
szInstalled	label byte
	db "FixPac installed.",CR,LF,'$'

szErrOption	label byte
	db "Invalid option!",CR,LF
	db CR,LF	; fall through to szUsage
szUsage		label byte
	db "Usage = FixPac [/U]",CR,LF
	db 'Resolves the "Packed File Corrupt" message. /U unloads FixPac.',CR,LF,'$'
szErrAlreadyInstalled	label byte
	db "FixPac already installed!",CR,LF,'$'
szErrNotInstalled	label byte
	db "Can't unload. FixPac not installed!",CR,LF,'$'
szErrNoNeedToInstall	label byte
	db "Memory already occupied. No need to install.",CR,LF,'$'

	end	_astart
