//
//   $Id: ZMDOS.C 1.2 2002/10/16 21:54:04Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*/////////////////////////////////////////////////////////////////////
//                                                                   //
//  ZMDOS.C -- zmodem protocol for MSC, QC, TC, ZTC                  //
//                                                                   //
//   (c) 1991-94, Mike Dumdei, 6 Holly Lane, Texarkana TX, 75503     //
//                                                                   //
//////////////////////////////////////////////////////////////////// */

/*

	Portions Copyright (c) 1997-1998 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary

	01-10-98 (DE)
		1. Defined a "file not found" error code for ZMODEM transmission.

*/

#pragma optimize("t", off)
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <dos.h>
#include <bios.h>
#include <io.h>
#include "comm.h"
#include "extra.h"
#include "zmdos.h"

#ifdef __TURBOC__
  #include <dir.h>
	#if(__BORLANDC__ < 0x300)
 	 int _bios_keybrd(int);
	#endif
#endif

/*--------------------------------------------------------------------+
|                                                                     |
|                             Data Section                            |
|                                                                     |
+--------------------------------------------------------------------*/
#define GLOBAL
/*--------------------------------------------+
|  initialized globals used by zmodem only    |
+--------------------------------------------*/
GLOBAL int XonTimeOut = 72;     /* max ticks before sending XON for RxChar */
GLOBAL int XoffTimeOut = 144;   /* max ticks before override for rx'd XOFF */
GLOBAL int TxWindow = 0;        /* max chars in transit, 0 if unlimited */

 /*++ zrinit capability flags ++*/
GLOBAL BYTE ZFR0 = CANFDX|CANOVIO|CANBRK|CANFC32;
GLOBAL BYTE ZFR1 = 0;

 /*++ zsinit flags ++*/
GLOBAL BYTE ZFS0 = 0;           /* don't need control characters esc'd */

 /*++ zfile management flags ++*/
GLOBAL ZHEADER ZFMgt = { 0L };      /* not requesting any file management */
GLOBAL ZHEADER ZFMgtMask = { 0L };  /* mask for incoming file mgt requests */

/*--------------------------------------------+
|  uninitialized globals used by zmodem only  |
+--------------------------------------------*/
GLOBAL int BinHdr;              /* ZBIN32 if using CRC32, ZBIN if CRC16 */
GLOBAL char rData32 = 0;        /* use 32 bit CRC for rx'd data packets */
GLOBAL char tData32;            /* use 32 bit CRC for tx'd data packets */
GLOBAL ZHEADER ZFOpts;          /* option flags received in last ZFILE hdr */
GLOBAL ZHEADER ZHdr;            /* I/O buf for zmodem header data */
GLOBAL BYTE ZSMask = 0;         /* send char table mask, default = 0x08 */
GLOBAL BYTE LastChTxd = 0;      /* last char transmitted by ZTxChar */
GLOBAL WORD ErrCnt;             /* number of errors on current file */

/*--------------------------------------------+
|  static variables                           |
+--------------------------------------------*/
static BYTE ZRMask;             /* receive char table mask, default = 0x80 */
static long LastRxdPos;         /* last file offset received from remote */
static long LastSyncPos;        /* last offset where FilePos == RxdFilePos */
static int RemoteBufLen;        /* remote's buffer size, 0 if unlimited */
static int ReSyncs;             /* # times ZRPOS'd to same file offset */
static char *AttnStr;           /* remote's attention string */
static WORD PreHdrMax;          /* max bytes to scan when waiting for hdr */
static int OrgBlkLen;           /* max negotiated size of packets */

/*------------------------------------------------------------+
| Bit masks for program flow control based on current header. |
|  Bit 7 =                                                    |
|  Bit 6 =                                                    |
|  Bit 5 = ZRSendZFIN spec    ZFIN                            |
|  Bit 4 = ZSyncFilePos spec  ZRINIT,ZSKIP                    |
|  Bit 3 = Send ZAbortSess    LABORT,SYNCERR,FERR,MEMERR      |
|  Bit 2 = Timeout mask       TIMEOUT                         |
|  Bit 1 = Non-fatal mask     GARBCNT,GARBLED,BADCRC,LONGPKT  |
|  Bit 0 = Fatal error mask   ABORT,LCARR,SYNCERR,FERR,MEMERR |
+------------------------------------------------------------*/
#define TestZHdr(ch,x)  ((*(&ZHdrTable[12] + (ch))) & (x))
#define isBadData(ch)   ((*(&ZHdrTable[12] + (ch))) & '\x02')
#define isFatalErr(ch)  ((*(&ZHdrTable[12] + (ch))) & '\x01')

static BYTE ZHdrTable[] =
{
    /* GARB_CNT          LONG_PKT          GARBLED_PKT       BAD_CRC      */
       '\x02',           '\x02',          '\x02',           '\x02',
    /* TIMED_OUT         spare slot        SYNC_ERR          MEM_ERR      */
       '\x04',           '\x00',          '\x09',           '\x09',
    /* FILE_ERR          LOST_CARRIER      REM_ABORT         LCL_ABORT    */
       '\x09',           '\x01',          '\x01',           '\x09',
    /* ZRQINIT           ZRINIT            ZSINIT            ZACK         */
       '\x00',           '\x10',          '\x00',           '\x00',
    /* ZFILE             ZSKIP             ZNAK              ZABORT       */
       '\x00',           '\x10',          '\x00',           '\x01',
    /* ZFIN              ZRPOS             ZDATA             ZEOF         */
       '\x20',           '\x00',          '\x00',           '\x00',
    /* ZFERR             ZCRC              ZCHALLENGE        ZCOMPL       */
       '\x00',           '\x00',          '\x00',           '\x00',
    /* ZCAN              ZFREECNT          ZCOMMAND          ZSTDERR      */
       '\x01',           '\x00',          '\x00',           '\x00',
};

/*------------------------------------------------------------+
| Bit masks for program processing based on current character.|
|  Bit 7 = ZRx7Bit returns if ZRMask == 0x80                  |
|  Bit 6 = ZRx7Bit returns if ZRMask == 0x40                  |
|  Bit 5 = '0' - '9'                                          |
|  Bit 4 = ZDLE, ZCRCE, ZCRCG, ZCRCQ, or ZCRCW                |
|  Bit 3 = ZTxBlock send raw if ZSMask == 0x08                |
|  Bit 2 = ZTxBlock send raw if ZSMask == 0x04                |
|  Bit 1 = !(ZDLE, XON, XOFF, XON+parity, or XOFF+parity)     |
|  Bit 0 = !(DEL or DEL+parity)                               |
+------------------------------------------------------------*/
BYTE ZTable[256] =
{ /*  0/8    1/9    2/A    3/B    4/C    5/D    6/E    7/F   */
    '\x8B','\x8B','\x8B','\x8B','\x8B','\x8B','\x8B','\x8B',  /* 00 */
    '\x8B','\x8B','\xCB','\x8B','\x8B','\xC3','\x8B','\x8B',
    '\x83','\x01','\x8B','\x01','\x8B','\x8B','\x8B','\x8B',  /* 10 */
    '\xD1','\x8B','\x8B','\x8B','\x8B','\x8B','\x8B','\x8B',
    '\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF',  /* 20 */
    '\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF',
    '\xEF','\xEF','\xEF','\xEF','\xEF','\xEF','\xEF','\xEF',  /* 30 */
    '\xEF','\xEF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF',
    '\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF',  /* 40 */
    '\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF',
    '\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF',  /* 50 */
    '\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF',
    '\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF',  /* 60 */
    '\xDF','\xDF','\xDF','\xDF','\xCF','\xCF','\xCF','\xCF',
    '\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF',  /* 70 */
    '\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\xCF','\x8E',
    '\x0B','\x0B','\x0B','\x0B','\x0B','\x0B','\x0B','\x0B',  /* 80 */
    '\x0B','\x0B','\x0B','\x0B','\x0B','\x03','\x0B','\x0B',
    '\x03','\x01','\x0B','\x01','\x0B','\x0B','\x0B','\x0B',  /* 90 */
    '\x0B','\x0B','\x0B','\x0B','\x0B','\x0B','\x0B','\x0B',
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',  /* A0 */
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',  /* B0 */
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',  /* C0 */
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',  /* D0 */
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',  /* E0 */
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F',  /* F0 */
    '\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0F','\x0E'
};

/*--------------------------------------------+
|  32 bit CRC table                           |
+--------------------------------------------*/
DWORD Crc32Tbl[256] =
{
    0x00000000LU, 0x77073096LU, 0xEE0E612CLU, 0x990951BALU, 0x076DC419LU, 0x706AF48FLU,
    0xE963A535LU, 0x9E6495A3LU, 0x0EDB8832LU, 0x79DCB8A4LU, 0xE0D5E91ELU, 0x97D2D988LU,
    0x09B64C2BLU, 0x7EB17CBDLU, 0xE7B82D07LU, 0x90BF1D91LU, 0x1DB71064LU, 0x6AB020F2LU,
    0xF3B97148LU, 0x84BE41DELU, 0x1ADAD47DLU, 0x6DDDE4EBLU, 0xF4D4B551LU, 0x83D385C7LU,
    0x136C9856LU, 0x646BA8C0LU, 0xFD62F97ALU, 0x8A65C9ECLU, 0x14015C4FLU, 0x63066CD9LU,
    0xFA0F3D63LU, 0x8D080DF5LU, 0x3B6E20C8LU, 0x4C69105ELU, 0xD56041E4LU, 0xA2677172LU,
    0x3C03E4D1LU, 0x4B04D447LU, 0xD20D85FDLU, 0xA50AB56BLU, 0x35B5A8FALU, 0x42B2986CLU,
    0xDBBBC9D6LU, 0xACBCF940LU, 0x32D86CE3LU, 0x45DF5C75LU, 0xDCD60DCFLU, 0xABD13D59LU,
    0x26D930ACLU, 0x51DE003ALU, 0xC8D75180LU, 0xBFD06116LU, 0x21B4F4B5LU, 0x56B3C423LU,
    0xCFBA9599LU, 0xB8BDA50FLU, 0x2802B89ELU, 0x5F058808LU, 0xC60CD9B2LU, 0xB10BE924LU,
    0x2F6F7C87LU, 0x58684C11LU, 0xC1611DABLU, 0xB6662D3DLU, 0x76DC4190LU, 0x01DB7106LU,
    0x98D220BCLU, 0xEFD5102ALU, 0x71B18589LU, 0x06B6B51FLU, 0x9FBFE4A5LU, 0xE8B8D433LU,
    0x7807C9A2LU, 0x0F00F934LU, 0x9609A88ELU, 0xE10E9818LU, 0x7F6A0DBBLU, 0x086D3D2DLU,
    0x91646C97LU, 0xE6635C01LU, 0x6B6B51F4LU, 0x1C6C6162LU, 0x856530D8LU, 0xF262004ELU,
    0x6C0695EDLU, 0x1B01A57BLU, 0x8208F4C1LU, 0xF50FC457LU, 0x65B0D9C6LU, 0x12B7E950LU,
    0x8BBEB8EALU, 0xFCB9887CLU, 0x62DD1DDFLU, 0x15DA2D49LU, 0x8CD37CF3LU, 0xFBD44C65LU,
    0x4DB26158LU, 0x3AB551CELU, 0xA3BC0074LU, 0xD4BB30E2LU, 0x4ADFA541LU, 0x3DD895D7LU,
    0xA4D1C46DLU, 0xD3D6F4FBLU, 0x4369E96ALU, 0x346ED9FCLU, 0xAD678846LU, 0xDA60B8D0LU,
    0x44042D73LU, 0x33031DE5LU, 0xAA0A4C5FLU, 0xDD0D7CC9LU, 0x5005713CLU, 0x270241AALU,
    0xBE0B1010LU, 0xC90C2086LU, 0x5768B525LU, 0x206F85B3LU, 0xB966D409LU, 0xCE61E49FLU,
    0x5EDEF90ELU, 0x29D9C998LU, 0xB0D09822LU, 0xC7D7A8B4LU, 0x59B33D17LU, 0x2EB40D81LU,
    0xB7BD5C3BLU, 0xC0BA6CADLU, 0xEDB88320LU, 0x9ABFB3B6LU, 0x03B6E20CLU, 0x74B1D29ALU,
    0xEAD54739LU, 0x9DD277AFLU, 0x04DB2615LU, 0x73DC1683LU, 0xE3630B12LU, 0x94643B84LU,
    0x0D6D6A3ELU, 0x7A6A5AA8LU, 0xE40ECF0BLU, 0x9309FF9DLU, 0x0A00AE27LU, 0x7D079EB1LU,
    0xF00F9344LU, 0x8708A3D2LU, 0x1E01F268LU, 0x6906C2FELU, 0xF762575DLU, 0x806567CBLU,
    0x196C3671LU, 0x6E6B06E7LU, 0xFED41B76LU, 0x89D32BE0LU, 0x10DA7A5ALU, 0x67DD4ACCLU,
    0xF9B9DF6FLU, 0x8EBEEFF9LU, 0x17B7BE43LU, 0x60B08ED5LU, 0xD6D6A3E8LU, 0xA1D1937ELU,
    0x38D8C2C4LU, 0x4FDFF252LU, 0xD1BB67F1LU, 0xA6BC5767LU, 0x3FB506DDLU, 0x48B2364BLU,
    0xD80D2BDALU, 0xAF0A1B4CLU, 0x36034AF6LU, 0x41047A60LU, 0xDF60EFC3LU, 0xA867DF55LU,
    0x316E8EEFLU, 0x4669BE79LU, 0xCB61B38CLU, 0xBC66831ALU, 0x256FD2A0LU, 0x5268E236LU,
    0xCC0C7795LU, 0xBB0B4703LU, 0x220216B9LU, 0x5505262FLU, 0xC5BA3BBELU, 0xB2BD0B28LU,
    0x2BB45A92LU, 0x5CB36A04LU, 0xC2D7FFA7LU, 0xB5D0CF31LU, 0x2CD99E8BLU, 0x5BDEAE1DLU,
    0x9B64C2B0LU, 0xEC63F226LU, 0x756AA39CLU, 0x026D930ALU, 0x9C0906A9LU, 0xEB0E363FLU,
    0x72076785LU, 0x05005713LU, 0x95BF4A82LU, 0xE2B87A14LU, 0x7BB12BAELU, 0x0CB61B38LU,
    0x92D28E9BLU, 0xE5D5BE0DLU, 0x7CDCEFB7LU, 0x0BDBDF21LU, 0x86D3D2D4LU, 0xF1D4E242LU,
    0x68DDB3F8LU, 0x1FDA836ELU, 0x81BE16CDLU, 0xF6B9265BLU, 0x6FB077E1LU, 0x18B74777LU,
    0x88085AE6LU, 0xFF0F6A70LU, 0x66063BCALU, 0x11010B5CLU, 0x8F659EFFLU, 0xF862AE69LU,
    0x616BFFD3LU, 0x166CCF45LU, 0xA00AE278LU, 0xD70DD2EELU, 0x4E048354LU, 0x3903B3C2LU,
    0xA7672661LU, 0xD06016F7LU, 0x4969474DLU, 0x3E6E77DBLU, 0xAED16A4ALU, 0xD9D65ADCLU,
    0x40DF0B66LU, 0x37D83BF0LU, 0xA9BCAE53LU, 0xDEBB9EC5LU, 0x47B2CF7FLU, 0x30B5FFE9LU,
    0xBDBDF21CLU, 0xCABAC28ALU, 0x53B39330LU, 0x24B4A3A6LU, 0xBAD03605LU, 0xCDD70693LU,
    0x54DE5729LU, 0x23D967BFLU, 0xB3667A2ELU, 0xC4614AB8LU, 0x5D681B02LU, 0x2A6F2B94LU,
    0xB40BBE37LU, 0xC30C8EA1LU, 0x5A05DF1BLU, 0x2D02EF8DLU
};

/*--------------------------------------------------------------------+
|                                                                     |
|                             Code Section                            |
|                                                                     |
+--------------------------------------------------------------------*/
#ifdef __MSC__
  #if (_MSC_VER >= 600)
    #define _FT_  _N_ _fastcall
  #endif
#endif
#ifndef _FT_
  #define _FT_  _N_ pascal
#endif
/*--------------------------------------------+
|  prototypes for local functions             |
+--------------------------------------------*/
static void _FT_ ZAbortSession(void);
static int  _FT_ ZClose(int ch);
static void _FT_ ZDelay(int ticks);
static int  _FT_ ZInitVars(ASYNC *comport, int receiving);
static int  _FT_ ZRecvFileData(void);
static int  _FT_ ZRecvHdr(void);
static int  _FT_ ZReqRecv(int hdrtype);
static int  _FT_ ZRSendFin(void);
static int  _FT_ ZRx7Bit(void);
static int  _FT_ ZRxHex(void);
static int  _FT_ ZSAckFin(void);
static int  _FT_ ZSendAttn(void);
static int  _FT_ ZSendFile(int lenblk0);
static int  _FT_ ZSendFileData(void);
static int  _FT_ ZSendHdr(WORD style, WORD type);
static void _FT_ ZSendPacket(BYTE *data, int bytes, int zcrc);
static int  _FT_ ZSyncFilePos(int FilePosNEOk);

#ifdef NO_ASM
  #define Upd32Crc(crc,c)  ((Crc32Tbl[((BYTE)(crc) ^ (c))]) ^ (((crc) >> 8)))
  static DWORD _FT_ Blk32Crc(register BYTE *block, int bytes);
  static int _FT_ ZRecvPacket(BYTE *buf, int bytes);
  static int _FT_ ZRxChar(void);
  static void _FT_ ZTxBlock(BYTE *block, int bytes);
#else
  extern DWORD _N_ Upd32Crc(DWORD crc, BYTE ch);
  extern DWORD _N_ Blk32Crc(BYTE *block, int bytes);
  extern int _N_ ZRecvPacket(BYTE *buf, int bytes);
  extern int _N_ ZRxChar(void);
  extern void _N_ ZTxBlock(BYTE *block, int bytes);
#endif

#ifdef NO_ASM
#pragma optimize("t", on)
#pragma check_stack(off)
/*/////////////////////////////////////////////////////////
//  Blk32Crc - calculates 32 bit crc for a block of bytes//
//   returns: 32 bit crc                                 //
//:blk////////////////////////////////////////////////// */
static DWORD _FT_ Blk32Crc(register BYTE *block, int bytes)
{
    register DWORD crc = -1L;

    while (bytes--)
        crc = Upd32Crc(crc, *block++);
    return (crc);
}
#pragma check_stack()
#pragma optimize("t", off)
#endif

#pragma check_stack(off)
/*/////////////////////////////////////////////////////////
//  ZAbortSession - Abort transfer (send CAN string)     //
//:zab////////////////////////////////////////////////// */
static void _FT_ ZAbortSession(void)
{
    static char buf[] = { 24,24,24,24,24,24,24,24,8,8,8,8,8,8,8,8 };

    FlushTxBuf();
    TxBlock(buf, 16);
    ZMsg(M_SHDR, LCL_ABORT, 3);
    Delay(9);
    FlushRxBuf();
}

/*/////////////////////////////////////////////////////////
//  ZClose - cleanup zmodem session                      //
//   returns: zmodem call exit code                      //
//:zcl////////////////////////////////////////////////// */
static int _FT_ ZClose(int ch)
{
    static int rvals[7] = { 0, LCL_ABORT, REM_ABORT, LOST_CARRIER,
        FILE_ERR, MEM_ERR, SYNC_ERR };
    int i;

    if (ch == ZCOMPL || ch == ZFIN)
        ch = 0;
    if (ch == ZCAN || ch == ZABORT)
        ch = REM_ABORT;
    for (i = 0; ch != rvals[i] && i < 6; ++i)
        ;
    ch = rvals[i];                  /* unknowns get changed to SYNC_ERR */
    if (TestZHdr(ch, '\x08'))
        ZAbortSession(); /* send CANs if LCL_ABORT, SYNC, MEM, or FILE_ERR */
    CloseFile();
    if (IOBuf)
        free(IOBuf);
    AsyncMode(0);
    while (KBHIT)
        KBREAD;
    return (ch);
}

/*/////////////////////////////////////////////////////////
//  ZDelay -- pause with M_IDLE calls to ZMsg            //
//////////////////////////////////////////////////////// */
static void _FT_ ZDelay(int ticks)
{
    long timeout;

    set_timeout(&timeout, ticks);
    do {
        ZMsg(M_IDLE);
    } while (!timed_out(&timeout));
}

/*/////////////////////////////////////////////////////////
//  ZInitVars - inititialize zmodem variables            //
//   returns: 0 if successful, else MEM_ERR              //
//   globals/statics: initializes all                    //
//:zin////////////////////////////////////////////////// */
static int _FT_ ZInitVars(ASYNC *comport, int receiving)
{
    tPort = comport;
    if ((IOBuf = calloc(MAXPACKET + MAX_PATH + 50, 1)) == NULL)
        return MEM_ERR;
    AsyncMode(1);
    TFlag.F.Receiving = receiving;
    if (TFlag.F.EscCtl)
        ZRMask = '\x40', ZSMask = '\x04', ZFR0 |= ESCCTL, ZFS0 |= TESCCTL;
    else
        ZRMask = '\x80', ZSMask = '\x08', ZFR0 &= ~ESCCTL, ZFS0 &= ~TESCCTL;
    AttnStr = (char *)&IOBuf[MAXPACKET + 8];
    PathName = &AttnStr[36];
    TotalBytes = 0L;
    TotalFiles = ReSyncs = SerNbr = LastChTxd = 0;
    BinHdr = ZBIN;
    OrgBlkLen = BlkLen = MAXPACKET;
    if (TxWindow)
    {
        TxWindow = (TxWindow < 128) ? 128 : TxWindow & ~0x7f;
        if ((TxWindow / 4) < BlkLen)
            OrgBlkLen = BlkLen = TxWindow / 4;
    }
    if (ConnectBaud == 0L)
        ConnectBaud = atol(tPort->BPDSstr);
    PreHdrMax = (ConnectBaud > 38400L) ? 0 : (WORD)(BlkLen + ConnectBaud);
    return 0;
}

/*/////////////////////////////////////////////////////////
//  ZRecvFileData - receive file data                    //
//   returns: ZEOF, ZSKIP, fatal error                   //
//:zre////////////////////////////////////////////////// */
static int _FT_ ZRecvFileData(void)
{
    register int ch = 0, stat;
    BYTE sendzrpos;
    int errlimit = 10;

    while (1)
    {
        ZHdr.Pos = FilePos;
        if((stat = ZSendHdr(ZHEX, ZRPOS)) < 0)
		  	   return stat;
        for (sendzrpos = 0; !sendzrpos; )
        {
            if ((ch = ZRecvHdr()) == ZDATA)
            {
                if (ZHdr.Pos != FilePos)
                {
                    if (--errlimit < 0)
                        return SYNC_ERR;
                    stat = ZSendAttn();
						  if(stat < 0)
								return stat;
                    break;
                }
                while (1)
                {
                    ch = ZRecvPacket(IOBuf, MAXPACKET);
                    if (ch >= GOTCRCE && ch <= GOTCRCW)
                    {
                        errlimit = 10;
                        if (WriteDisk(BytesRxd) == FILE_ERR)
                            return FILE_ERR;
                        FilePos += (long)BytesRxd;
                        ZMsg(M_DATA, ch & ~0x100);
                        if (ch == GOTCRCG)
                            continue;
                        if (ch == GOTCRCE)
                            break;
                        if (ch == GOTCRCW)
                            FlushDisk();
                        ZHdr.Pos = FilePos;
                        stat = ZSendHdr(ZHEX, ZACK);
								if(stat < 0)
									return stat;
                        if (ch == GOTCRCQ)
                            continue;
                        TxChar(XON);
                        break;
                    }
                    ZMsg(M_RHDR, ch);
                    ++ErrCnt;
                    if (isFatalErr(ch))
                        return (ch);
                    if (ch != TIMED_OUT)
						  {
                    		stat = ZSendAttn();
						  		if(stat < 0)
										return stat;
						  }
                    sendzrpos = 1;
                    break;
                }
                continue;
            }
            switch (ch)                         /* ch = ZRecvHdr result */
            {                                   /*  except for ZDATA    */
              case ZNAK:
                ++ErrCnt;
              case TIMED_OUT:
                if (--errlimit < 0)
                    return SYNC_ERR;
                sendzrpos = 1;
                continue;
              case ZFILE:
                ZRecvPacket(IOBuf, MAXPACKET);
                sendzrpos = 1;
                continue;
              case ZEOF:
                if (ZHdr.Pos != FilePos)
                    continue;
              case ZSKIP:
                return (ch);
              default:
                ++ErrCnt;
                if (isBadData(ch) && --errlimit >= 0)
                {                                   /* test hdr for: */
                    stat = ZSendAttn();             /*  GARBLED_PKT, */
						  	if(stat < 0)
									return stat;
                    sendzrpos = 1;                  /*  GARB_COUNT,  */
                    continue;                       /*  BAD_CRC      */
                }                                   /*  LONG_PKT     */
                return (isFatalErr(ch)) ? ch : SYNC_ERR;
            }
        }
    }
}
#pragma check_stack()

/*/////////////////////////////////////////////////////////
//  ZmodemRecv - zmodem receive  (entry point)           //
//                                                       //
//   returns:      0 - normal completion                 //
//   -1    LCL_ABORT - user pressed ESC                  //
//   -2    REM_ABORT - remote aborted transfer           //
//   -3 LOST_CARRIER - lost carrier                      //
//   -4     FILE_ERR - file error on local system        //
//   -5      MEM_ERR - couldn't malloc I/O buffers       //
//   -6     SYNC_ERR - couldn't sync, lost sync, or ??   //
//                                                       //
// If using a locked baud rate, the global 'ConnectBaud' //
// must be set prior to calling ZmodemRecv.  Also other  //
// global variables may be set to control handling of    //
// name clashes with existing files, escaping controls,  //
// crash recovery, etc.  See global variables.           //
//:zmo////////////////////////////////////////////////// */
int ZmodemRecv(ASYNC *comport)
{
    register int ch = ZRINIT;
    char openmode[4];
    FILELIST f;
    int fmode, atrib;

    if (ZInitVars(comport, 1))
        return MEM_ERR;

    do {
        ch = ZReqRecv(ch);

        if (ch == ZFILE)
        {
            atrib = DosAtrib;
            TFlag.F.EndOfFile = 0;
            FilePos = 0L;
            strcpy(openmode, "wb");
            if ((fmode = ParseFileList(&f, F_FIRST, PathName)) != 0)
            {
                switch (TFlag.F.ExistOpts)
                {
                  case 0:               /* skip */
                    ch = ZSKIP;
                    break;
                  case 1:               /* resume at end of file */
                    if (fmode == F_RDWRITE && FileSize < RxdFileSize)
                        FilePos = FileSize, strcpy(openmode, "r+b");
                    else
                        ch = ZSKIP;
                    break;
                  case 2:               /* create unique name */
                    if (ParseFileList(&f, F_UNIQUE, PathName) == 0)
                        ch = ZSKIP;
                    break;
                  case 3:               /* replace existing file */
                    if (fmode != F_RDWRITE)
                        ch = ZSKIP;
                    break;
                }
                ZMsg(M_CLASH);
            }
            if (ch == ZSKIP)
                ZMsg(M_EOF, ZSKIP);
            else if (!OpenFile(PathName, openmode) || Seek(FilePos) != 0)
                ZMsg(M_EOF, FILE_ERR), ch = ZSKIP;
            else
            {
                ZMsg(M_FILE);
                ErrCnt = 0;
                ch = ZRecvFileData();
                CloseFile();
                if (TFlag.F.KeepAtrib)
                    DosSetFileAttr(PathName, atrib);
					 if(ch == FILE_ERR)
                	ZMsg(M_EOF, FILE_ERR);
					 else
                	ZMsg(M_EOF, ch);
            }
        }
    } while (ch == ZEOF || ch == ZSKIP);
    return (ZClose(ch));
}

#pragma optimize("t", on)
#pragma check_stack(off)
/*/////////////////////////////////////////////////////////
//  ZRecvHdr - receive zmodem header                     //
//   returns: header type, TIMED_OUT, fatal error,       //
//            non-fatal packet error                     //
//:zre////////////////////////////////////////////////// */
static int _FT_ ZRecvHdr(void)
{
    register int ch = 0;
    WORD zdles, crc, i = PreHdrMax;
    DWORD crcl;
    BYTE lbuf[10], *p1, *p2, style;

    while (1)
    {
        if (--i == 0)
            ch = GARB_COUNT;
        if (ch & 0xff00)
        {
            if (ch == GOTCAN)
                ch = REM_ABORT;
            ZMsg(M_RHDR, ch, 3);
            return (ch);
        }
        if (ch == ZDLE || (ch = RxChar()) == ZDLE)
        {
            for (zdles = 5; --zdles && (ch = RxChar()) == ZDLE; )
                ;
            if (ch == ZDLE || ch == ZCRCW)
                ch = (ch == ZDLE) ? GOTCAN : GARBLED_PKT;
        }
        if ((ch & 0xff7f) == ZPAD)
        {
            while ((ch = ZRx7Bit()) == ZPAD)
                ;
            if (ch == ZDLE)
            {
                p1 = lbuf;
                if ((ch = ZRx7Bit()) < ZBIN || ch > ZBIN32)
                    continue;
                if ((style = (BYTE)ch) == ZBIN32)
                    p2 = p1 + 9, crcl = -1L;
                else
                    p2 = p1 + 7, crc = 0;
                do {
                    if ((ch = (style == ZHEX) ? ZRxHex() : ZRxChar()) & 0xff00)
                        break;
                    *p1++ = (BYTE)ch;
                    if (style == ZBIN32)
                        crcl = Upd32Crc(crcl, (BYTE)ch);
                    else
                        crc = Upd16Crc(crc, ch);
                } while (p1 < p2);
                if (ch & 0xff00)
                    continue;
                if (style != ZBIN32)
                {
                    crc = Upd16Crc(crc, 0);
                    crcl = Upd16Crc(crc, 0) + 0xdebb20e3L;
                }
                if (crcl != 0xdebb20e3L)
                    ch = BAD_CRC;
                else
                {
                    ZHdr.Data = *(long *)&lbuf[1];
                    if (style == ZBIN32)
                        rData32 = BinHdr = ZBIN32;
                    else
                    {
                        rData32 = 0;
                        if (style == ZHEX)
                        {
                            RxTimeOut = 2;
                            if ((RxChar() & 0xff7f) == '\r')
                                ch = RxChar();
                            RxTimeOut = DefRxTimeOut;
                        }
                    }
                    ZMsg(M_RHDR, *lbuf, style);
                    return (*lbuf);
                }
            }
        }
    }
}
#pragma optimize("t", off)

#ifdef NO_ASM
#pragma optimize("t", on)
/*/////////////////////////////////////////////////////////
//  ZRecvPacket - receive zmodem data packet             //
//   returns: crctype (GOTCRCx), TIMED_OUT, fatal error, //
//            non-fatal packet error                     //
//   statics: BytesRxd                                   //
//:zre////////////////////////////////////////////////// */
static int _FT_ ZRecvPacket(BYTE *buf, int bytes)
{
    register int ch;
    WORD crc, crctype;
    DWORD crcl;
    BYTE lbuf[4], *p1 = buf;

    crctype = crc = 0;
    crcl = -1L;
    ++bytes;
    while (1)
    {
        if (rData32)
        {
            while (--bytes >= 0 && !((ch = ZRxChar()) & ~0xff))
                crcl = Upd32Crc(crcl, *p1++ = (BYTE)ch);
        }
        else
        {
            while (--bytes >= 0 && !((ch = ZRxChar()) & ~0xff))
                crc = Upd16Crc(crc, *p1++ = (BYTE)ch);
        }
        if (ch >= GOTCRCE && ch <= GOTCRCW)
        {
            if (crctype)
                crctype = 0, ch = GARBLED_PKT;
            else
            {
                crctype = ch;
                ch &= 0xff;
                if (rData32)
                    crcl = Upd32Crc(crcl, (BYTE)ch);
                else
                    crc = Upd16Crc(crc, (BYTE)ch);
                BytesRxd = p1 - buf, p1 = lbuf;
                bytes = (rData32) ? 4 : 2;
                continue;
            }
        }
        break;
    }
    if (bytes < 0 && ch >= 0)
    {
        if (crctype == 0)
            ch = LONG_PKT;
        if (!rData32)
        {
            crc = Upd16Crc(crc, 0);
            crcl = Upd16Crc(crc, 0) + 0xdebb20e3L;
        }
        return (crcl == 0xdebb20e3L) ? crctype : BAD_CRC;
    }
    BytesRxd = 0;
    return (ch == GOTCAN) ? REM_ABORT : ch;
}
#endif

#pragma optimize("t", off)
/*/////////////////////////////////////////////////////////
//  ZReqRecv - send zmodem receive request               //
//   returns: ZFILE, ZCOMPL, fatal err                   //
//:zre////////////////////////////////////////////////// */
static int _FT_ ZReqRecv(int hdrtype)
{
    register int ch, stat;
    int errlimit = 10;

    if (hdrtype == ZEOF)
        hdrtype = ZRINIT;
    while (1)
    {
        ZHdr.BufSz = BufSize;
        ZHdr.Flg[F0] = ZFR0;
        ZHdr.Flg[F1] = ZFR1;
        stat = ZSendHdr(ZHEX, hdrtype);
		  if(stat < 0)
				return stat;
        do
        {
            ch = ZRecvHdr();
            if (hdrtype == ZSKIP && ch != ZNAK)
                hdrtype = ZRINIT;
            switch (ch)
            {
              case ZFILE:
                ZFOpts.Data = ZHdr.Data;
                if (ZRecvPacket(IOBuf, MAXPACKET) == GOTCRCW)
                {
                    ParseYZBlk0((char *)IOBuf);
                    return ZFILE;
                }
                ch = ZNAK;
                break;
              case ZSINIT:
                if (ZHdr.Flg[F0] & TESCCTL)
                     ZSMask = '\x04', ZRMask = '\x40';
                if (ZRecvPacket((BYTE *)AttnStr, 34) == GOTCRCW)
                    ZHdr.Data = 1L, ch = ZACK;
                else
                    ch = ZNAK;
                break;
              case ZFREECNT:
                ZHdr.Data = DosGetDiskFree(0);
                ch = ZACK;
                break;
              case ZCOMMAND:
                ZHdr.Data = 0L;
                if (ZRecvPacket(IOBuf, MAXPACKET) != GOTCRCW)
                    ch = ZNAK;
                else
                {
                    errlimit = 5;
                    do {
                        stat = ZSendHdr(ZHEX, ZCOMPL);
								if(stat < 0)
									 return stat;
                        if (isFatalErr(ch = ZRecvHdr()))
                            return (ch);
                    } while (--errlimit >= 0 && ch != ZFIN);
                    if (ch != ZFIN)
                        return SYNC_ERR;
                }
                break;
              case ZFIN:
                ZSAckFin();
                return ZCOMPL;
              case ZCOMPL:
                continue;
              default:
                if (isFatalErr(ch))
                    return (ch);
                if (--errlimit < 0)
                    return SYNC_ERR;
                ch = 0;
                continue;
            }
            stat = ZSendHdr(ZHEX, ch);
				if (stat < 0)
					return stat;
        } while (ch);
    }
}

/*/////////////////////////////////////////////////////////
//  ZRSendFin - ZFIN transmit by sender                  //
//   returns: ZFIN, TIMED_OUT, fatal error               //
//:zrs////////////////////////////////////////////////// */
static int _FT_ ZRSendFin(void)
{
    register int ch, stat;

    while (1)
    {
        ZHdr.Data = 0L;
        stat = ZSendHdr(ZHEX, ZFIN);
		  if(stat < 0)
			  return stat;
        if (TestZHdr(ch = ZRecvHdr(), '\x25'))  /* ZFIN,TIMEOUT,fatal */
        {
            if (ch == ZFIN)
            {
                TxBlock("OO", 2);
                Delay(4);
            }
            return (ch);
        }
    }
}

/*/////////////////////////////////////////////////////////
//  ZRx7Bit - read char, strip parity & xflow chars      //
//   returns: received char less high bit, TIMED_OUT,    //
//            fatal error                                //
//:zrx////////////////////////////////////////////////// */
static int _FT_ ZRx7Bit(void)
{
    register int ch;

    while (1)
    {
        if (((ch = RxChar()) < 0) || (ZTable[ch &= 0x7f] & ZRMask))
            return (ch);
    }
}

#ifdef NO_ASM
/*/////////////////////////////////////////////////////////
//  ZRxChar - receive char with zmodem processing        //
//   returns: received char, TIMED_OUT, fatal error,     //
//            non-fatal packet error, GOTCAN, GOTCRC#    //
//:zrx////////////////////////////////////////////////// */
static int _FT_ ZRxChar(void)
{
    register int ch;
    int zdles;

    while (1)
    {
        if (((ch = RxChar()) < 0) || (ZTable[ch] & '\x02'))
            return (ch);
        if (ch != ZDLE)
            continue;
        for (zdles = 5; --zdles && (ch = RxChar()) == ZDLE; )
            ;
        if (ch > 0)
        {
            if (ZTable[ch] & '\x10')
                ch |= 0x100;
            else if ((ch & 0x60) == 0x40)
                ch ^= 0x40;
            else if (ch == ZRUB0)
                ch = 0x7f;
            else if (ch == ZRUB1)
                ch = 0xff;
            else if ((ZTable[ch] & '\x02') == 0)
                continue;
            else
                return GARBLED_PKT;
        }
        return (ch);
    }
}
#endif

/*/////////////////////////////////////////////////////////
//  ZRxHex - read 2 ASCII hex digits                     //
//   returns: 2 ascii hex digits as hex byte, TIMED_OUT, //
//            fatal error, non-fatal packet error        //
//:zrx////////////////////////////////////////////////// */
static int _FT_ ZRxHex(void)
{
    register int ch;
    int i, rval;

    for (i = rval = 0; i < 2; ++i)
    {
        if ((ch = ZRx7Bit()) < 0)
            return (ch);
        if ((ch -= (ZTable[ch] & '\x20') ? '0' : ('a' - '\x0a')) & ~0xf)
            return GARBLED_PKT;
        rval = (rval << 4) | ch;
    }
    return (rval);
}

/*/////////////////////////////////////////////////////////
//  ZSAckFin - Ack ZFIN packet                           //
//  return 0 on success or LCL_ABORT if aborted				//
//:zsa////////////////////////////////////////////////// */
static int _FT_ ZSAckFin(void)
{
    register int ch = 0, stat;
    int i = 3;

    ZHdr.Data = 0L;
    while (ch != 'O' && (ch >= 0 || ch == TIMED_OUT) && --i >= 0)
    {
        stat = ZSendHdr(ZHEX, ZFIN);
		  if(stat < 0)
				break;
        RxTimeOut = 100;
        if ((ch = RxChar()) == 'O')
            RxTimeOut = 2, ch = RxChar();
    }
    RxTimeOut = DefRxTimeOut;
	 return stat;
}

/*/////////////////////////////////////////////////////////
//  ZSendAttn - Send zmodem attention string             //
//  return 0 on success or LCL_ABORT if aborted				//
//:zse////////////////////////////////////////////////// */
static int _FT_ ZSendAttn(void)
{
	 int status;
    register char *p1;

    FlushTxBuf();
    for (p1 = AttnStr; *p1; ++p1)
    {
        if (*p1 == '\xde')
            ZDelay(18);
        else if (*p1 == '\xdd')
            SendBreak();
        else
        {
            TxChar(*p1);
            ZDelay(1);
        }
    }
    while ((status = WaitTxLevel(0)) == 0)
        ;
	 return status;
}
#pragma check_stack()

/*/////////////////////////////////////////////////////////
//  ZmodemSend - send files using zmodem                 //
//:zmo////////////////////////////////////////////////// */
int ZmodemSend(ASYNC *comport, char *files)
{
    register int ch, stat;
    int errlimit = 10;
    FILELIST f;

    if (ZInitVars(comport, 0))
        return MEM_ERR;
    if (GetTotalFiles(files, F_NOSYS) == 0)
    {
        ZRSendFin();
        return ERR_NO_FILES;
    }
    TxBlock("\rrz\r", 4);
    ZHdr.Data = 0L;
    stat = ZSendHdr(ZHEX, ZRQINIT);
	 if(stat < 0)
	     return stat;
    while (1)
    {
        switch (ch = ZRecvHdr())
        {
          case ZRINIT:
            BinHdr = (ZFR0 & ZHdr.Flg[F0] & CANFC32) ? ZBIN32 : ZBIN;
            if (ZHdr.Flg[F0] & ESCCTL)
                ZSMask = '\x04', ZRMask = '\x40';
            if (!(ZHdr.Flg[F0] & CANFDX))
                TxWindow = 0;
            if (((RemoteBufLen = ZHdr.BufSz) != 0) && BlkLen > RemoteBufLen)
                OrgBlkLen = BlkLen = RemoteBufLen;
            if ((ZFS0 & TESCCTL) && !(ZHdr.Flg[F0] & ESCCTL))
            {
                while (1)
                {
                    ZHdr.Data = 0L, ZHdr.Flg[F0] = ZFS0;
                    stat = ZSendHdr(ZHEX, ZSINIT);
	 					  if(stat < 0)
	     				     return stat;
                    ZSendPacket((BYTE *)"", 1, ZCRCW);
                    while ((stat = WaitTxLevel(0)) == 0)
                        ;
	 					  if(stat < 0)
	     				     return stat;
                    if ((ch = ZRecvHdr()) == ZACK)
                        break;
                    if (isFatalErr(ch))
                        return (ZClose(ch));
                    if (--errlimit < 0)
                        return (ZClose(SYNC_ERR));
                }
            }
            ParseFileList(&f, F_NOSYS | F_LOWER | F_SLASH, files);
            do {
                if (OpenFile(PathName, "rb"))
                {
                    ch = ZSendFile(BuildYZBlk0());
                    CloseFile();
                    ZMsg(M_EOF, ch);
                }
                if (--TotalFiles < 0)
                    ++TotalFiles;
                if ((TotalBytes -= FileSize) < 0L)
                    TotalBytes = 0L;
            } while (ch >= 0 && ParseFileList(&f, 0));
            if (ch >= 0)
                ch = ZRSendFin();
            return (ZClose(ch));
          case ZCHALLENGE:
            stat = ZSendHdr(ZHEX, ZACK);
				if(stat < 0)
					return stat;
            continue;
          case ZRQINIT:
            if (ZHdr.Flg[F0] == ZCOMMAND)
                continue;
          case ZCOMMAND:
            ch = TIMED_OUT;
          default:
            if (isFatalErr(ch))
                return (ZClose(ch));
            if (--errlimit < 0)
                return (ZClose(SYNC_ERR));
            ZHdr.Data = 0L;
            stat = ZSendHdr(ZHEX, (ch == TIMED_OUT) ? ZRQINIT : ZNAK);
	 			if(stat < 0)
	     			return stat;
            continue;
        }
    }
}

#pragma check_stack(off)
/*/////////////////////////////////////////////////////////
//  ZSendFile - send one file using zmodem               //
//   returns: ZRINIT, ZSKIP, fatal error                 //
//:zse////////////////////////////////////////////////// */
static int _FT_ ZSendFile(int lenblk0)
{
    register int ch = 0, stat;
    int i = 5;

    while (1)
    {
        if (ch != ZPAD)
        {
            ZHdr.Data = ZFMgt.Data;
            stat = ZSendHdr(BinHdr, ZFILE);
				if(stat < 0)
					return stat;
            ZSendPacket(IOBuf, lenblk0, ZCRCW);
            while ((stat = WaitTxLevel(0)) == 0)
                ;
	 			if(stat < 0)
	     			return stat;
        }
        if ((ch = ZRecvHdr()) == ZRPOS)
        {
            if (Seek(ZHdr.Pos) == FILE_ERR)
                return FILE_ERR;
            ErrCnt = 0;
            TFlag.F.EndOfFile = 0;
            LastSyncPos = (FilePos = LastRxdPos = ZHdr.Pos) - 1;
            ZMsg(M_FILE);
            return (ZSendFileData());
        }
        if (isFatalErr(ch) || ch == ZSKIP)
            return (ch);
        if (--i < 0)
            return SYNC_ERR;

        if (ch == ZRINIT)
        {
            RxTimeOut = 54;
            while ((ch = RxChar()) > 0 && ch != ZPAD)
                ;
            RxTimeOut = DefRxTimeOut;
            if (ch < 0 && ch != TIMED_OUT)
                return (ch);
        }
    }
}

/*/////////////////////////////////////////////////////////
//  ZSendFileData - send file data                       //
//   returns: ZRINIT, ZSKIP, fatal error                 //
//:zse////////////////////////////////////////////////// */
static int _FT_ ZSendFileData(void)
{
    register int ch, zpkt, stat;
    int bytes, junkbytes, reqcrcw, reqcrcq, reqcrcqlen, level;

    level = (int)uldiv(ConnectBaud, 100L);
    if (TxWindow)
        reqcrcqlen = TxWindow / 4;
    while (1)
    {
        ZHdr.Pos = FilePos;
        stat = ZSendHdr(BinHdr, ZDATA);
		  if(stat < 0)
			  return stat;
        junkbytes = 0;
        reqcrcq = reqcrcqlen;
        reqcrcw = RemoteBufLen;
        do {
            if ((bytes = ReadDisk(BlkLen)) == FILE_ERR)
                return FILE_ERR;
            if (bytes != BlkLen)
                zpkt = ZCRCE, TFlag.F.EndOfFile = 1;
            else if (junkbytes > 20 || FilePos == LastSyncPos
             || (RemoteBufLen && (reqcrcw -= bytes) <= 0))
                zpkt = ZCRCW;
            else if (TxWindow && (reqcrcq -= bytes) <= 0)
                zpkt = ZCRCQ, reqcrcq = reqcrcqlen;
            else
                zpkt = ZCRCG;
            FilePos += (long)bytes;
            ZSendPacket(IOBuf, bytes, zpkt);
            if (BlkLen < OrgBlkLen && FilePos > (LastSyncPos + (BlkLen << 2)))
            {
                if ((BlkLen *= 2) > OrgBlkLen)
                    BlkLen = OrgBlkLen;
            }
            if (zpkt == ZCRCW)
                break;
            while (zpkt && ((ch = WaitTxLevel(level)) <= 0 || !RxBufEmpty()))
            {
                if (ch < 0)
                    return (ch);
                while (!RxBufEmpty() && zpkt)
                {
                    if ((ch = RxChar()) != ZPAD && ch != ZDLE)
                        ++junkbytes;
                    else if ((ch = ZSyncFilePos(1)) != ZACK)
                        zpkt = 0;
                }
            }
            if (zpkt)
                ZMsg(M_DATA, zpkt);
            while (TxWindow && (FilePos - LastRxdPos) >= TxWindow && zpkt)
            {
                if (zpkt != ZCRCQ)
                {
                    ZSendPacket(IOBuf, 0, ZCRCQ);
                    while ((stat = WaitTxLevel(0)) == 0)
                        ;
	 					  if(stat < 0)
	     						return stat;
                }
                if ((ch = ZSyncFilePos(1)) != ZACK)
                    zpkt = 0;
            }
            if (TFlag.F.EndOfFile && zpkt)
            {
                XonTimeOut = 180;
                ZHdr.Pos = FilePos;
                stat = ZSendHdr(BinHdr, ZEOF);
		  			 if(stat < 0)
			  			 return stat;
                break;
            }
        } while (zpkt);
        if (!zpkt)
        {
            ++ErrCnt;
            ZSendPacket(IOBuf, 0, ZCRCE);
        }
        else
        {
            ch = ZSyncFilePos(0);
            XonTimeOut = 72;
            if (zpkt == ZCRCW)
                ZMsg(M_DATA, zpkt);
        }
        if (ch != ZACK && ch != ZRPOS)
        {
            if (ch == ZRINIT)
            {
                do {
                    FlushRxBuf();
                    ZDelay(2);
                } while (!RxBufEmpty());
            }
            FlushTxBuf();
            return (ch);
        }
    }
}

/*/////////////////////////////////////////////////////////
//  ZSendHdr - send zmodem header                        //
//  return 0 on success or LCL_ABORT if aborted				//
//:zse////////////////////////////////////////////////// */
static int _FT_ ZSendHdr(WORD style, WORD type)
{
    static char table[] =
     { '0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f' };
    static char hdr[] = { '*', '*', ZDLE, 0 };
    BYTE lbuf[10], *p1 = lbuf, *p2 = lbuf;
    WORD crc;
	 int stat;

    ((hdr[3] = (char)style) == ZHEX) ? TxBlock(hdr, 4) : TxBlock(&hdr[1], 3);
    *p1++ = (BYTE)type;
    *(long *)p1 = ZHdr.Data;
    p1 += 4;
    if (style == ZBIN32)
    {
        tData32 = ZBIN32;
        *(long *)p1 = ~Blk32Crc(lbuf, 5);
        p1 += 4;
    }
    else
    {
        tData32 = 0;
        p1[0] = p1[1] = '\0';
        crc = Blk16Crc(lbuf, 7);
        *p1++ = (BYTE)(crc >> 8);
        *p1++ = (BYTE)crc;
    }
    if (style != ZHEX)
        ZTxBlock(p2, p1 - p2);
    else
    {
        while (p2 < p1)
        {
            TxChar(table[(*p2 >> 4) & 0x0f]);
            TxChar(table[*p2++ & 0x0f]);
        }
        TxBlock("\r\n", 2);
        if (type != ZFIN && type != ZACK)
            TxChar(XON);
    }
    ZMsg(M_SHDR, type, style);
    while ((stat = WaitTxLevel(0)) == 0)
        ;
	 return stat;
}

#pragma optimize("t", on)
/*/////////////////////////////////////////////////////////
//  ZSendPacket - send zmodem data packet                //
//:zse////////////////////////////////////////////////// */
static void _FT_ ZSendPacket(BYTE *data, int bytes, int zcrc)
{
    DWORD crcl;
    WORD crc;

    ZTxBlock(data, bytes);
    TxChar(ZDLE);
    TxChar((char)zcrc);
    if (tData32)
    {
        crcl = Blk32Crc(IOBuf, bytes);
        crcl = ~Upd32Crc(crcl, (BYTE)zcrc);
        ZTxBlock((BYTE *)&crcl, 4);
    }
    else
    {
        crc = Blk16Crc(IOBuf, bytes);
        crc = Upd16Crc(crc, zcrc);
        crc = Upd16Crc(crc, 0), crc = Upd16Crc(crc, 0);
        crc = (crc >> 8) | (crc << 8);
        ZTxBlock((BYTE *)&crc, 2);
    }
    if (zcrc == ZCRCW)
        TxChar(XON);
}
#pragma optimize("t", off)

/*/////////////////////////////////////////////////////////
//  ZSyncFilePos - synchronize file position offsets     //
//   returns: ZACK, ZRPOS, ZRINIT, ZSKIP, fatal error    //
//:zsy////////////////////////////////////////////////// */
static int _FT_ ZSyncFilePos(int FilePosNEOk)
{
    register int ch, stat;
    int i = 5;

    while (1)
    {
        switch (ch = ZRecvHdr())
        {
          case ZACK:
            if (FilePos == (LastRxdPos = ZHdr.Pos) || FilePosNEOk)
                return (ch);
            continue;
          case ZRPOS:
            FlushTxBuf();
            if (Seek(FilePos = LastRxdPos = ZHdr.Pos) != 0)
                return FILE_ERR;
            TFlag.F.EndOfFile = 0;
            if (LastSyncPos == ZHdr.Pos && ++ReSyncs == 2 && BlkLen > 32)
                ReSyncs = 0, BlkLen /= 2;
            LastSyncPos = ZHdr.Pos;
            return (ch);
          default:
            if (TestZHdr(ch, '\x11'))  /* ZRINIT, ZSKIP, fatal err */
                return(ch);
            if (ch == ZFIN || --i < 0)
                return SYNC_ERR;
            if ((i > 1) && TFlag.F.EndOfFile && !TFlag.F.Receiving)
            {
                ZHdr.Pos = FilePos;
                stat = ZSendHdr(BinHdr, ZEOF);
		  			 if(stat < 0)
			  			 return stat;
            }
            else
            {
                ZHdr.Pos = 0L;
                stat = ZSendHdr(BinHdr, ZNAK);
		  			 if(stat < 0)
			  			 return stat;
                break;
            }
        }
    }
}

#ifdef NO_ASM
#pragma optimize("t", on)
/*/////////////////////////////////////////////////////////
//  ZTxBlock - transmit a block with zmodem processing   //
//:ztx////////////////////////////////////////////////// */
static void _FT_ ZTxBlock(BYTE *block, int bytes)
{
    register BYTE _N_ *p1;
    static BYTE _N_ tbuf[24];
    WORD ch;

    p1 = &tbuf[1], *tbuf = LastChTxd;
    while (1)
    {
        if (--bytes < 0 || p1 > &tbuf[21])
        {
            TxBlock(&tbuf[1], p1 - &tbuf[1]);
            if (bytes < 0)
            {
                LastChTxd = *(p1 - 1);
                return;
            }
            *tbuf = *(p1 - 1);
            p1 = &tbuf[1];
        }
        ch = *block++;
        if ((ZTable[ch] & ZSMask) || (ZSMask == '\x08' &&
         (ch & 0x7f) == '\r' && (*(p1 - 1) & 0x7f) != '@'))
            *p1++ = ch;
        else
        {
            *p1++ = ZDLE;
            *p1++ = (ZTable[ch] & '\x01') ? (ch ^ 0x40) :
             ((ch == 0x7f) ? ZRUB0 : ZRUB1);
        }
    }
}
#endif
#pragma check_stack()

