//
//   $Id: MEM.C 1.5 2002/10/16 21:54:04Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   MEM.C  created 01/15/97  by Roy Sherrill

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	history
	3/3/97 - updated to "Solid Code" standards.
	         added to ROM-DOS project - dg

	09-26-97 DE corrected largest program size and
				total conventional free.

**********************************************************/
#define __FILENUM__ 1

#include <ctype.h>
#include <dos.h>
#include <mem.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>

#include "dosver.h"
#include "datalght.h"
#include "fatworld.h"
#include "dostruct.h"
#include "dldosio.h"

#define MEM 1
#include "mem.h"
#include "dosmsg.c"

// prototypes
#include "mem.pt"

// Program Options
int Classify = FALSE;
int BiosExtent = FALSE;
int Quiet = FALSE;
int Raw = FALSE;

bool fIsFat32World;

// Address of the XMS Driver if it exists
void (far * XMS_Control)(void);

void DisplaySignOnMessage( void )
{
	char * signon = "ROM-DOS MEM %s %s\n%s\n\n";

	if( fIsFat32World )
	{
		printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}


/*
 * Main entry for the MEM command!
 */
int main(int argc, char *argv[])
{
	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	// don't display signon if quiet switch is used
	// used for comparison with that "other" mem.
	if(!Quiet)
	{
		DisplaySignOnMessage();
	}

	// what options does user want?
	ProcessCmdline(argc, argv);

	// show all bios extensions
	if(BiosExtent)
	{
		ShowBiosExtensions();
		return 0;		// no errors
	}

	// show the programs by category, but don't exit here,
	// instead, fall through to ShowMemory so user sees both.
	if(Classify)
	{
		ShowByProgram();
	}

	// just show raw MCB chain.  Does display it in K and paras.
	if(Raw)
	{
		ShowRawDump();
		return 0;		// no errors
	}

	// show the simple memory used/free space screen
	ShowMemory();

	// no errors
	return 0;
}

/*
	Determine what optios the user picked.  Display error and
	exit if bad option detected.
*/
void ProcessCmdline(int argc, char *argv[])
{
	int i;

	for(i=1; i<argc; ++i)
	{
		if(argv[i][0] != '/')
			goto BadOption;

		switch(toupper(argv[i][1]))
		{
			// Bios Search option.
			case 'B':
				BiosExtent = TRUE;
				break;

			// Classify the memory in MCB's
			case 'C':
				Classify = TRUE;
				break;

			// Raw Dump of the MCB's
			case 'R':
				Raw = TRUE;
				break;

			// used for comparing output to MS-MEM, no signon
			case '[':
				Quiet = TRUE;
				break;

			case '?':
			case 'H':
				printf(Usage);
				exit(ERR_USAGE);
				break;

			default:
BadOption:
				printf(errMsgBadOpt,  argv[i]);
				printf(Usage);
				exit(ERR_BADOPT);
		}
	}
}

/*
	Return the # of bytes expressed as K bytes, rounded up or
	down depending on where it is closest to.
*/
unsigned long K(unsigned long bytes)
{
	unsigned remainder;
	unsigned long kBytes;

	// K bytes had better fit in an unsigned
	ASSERT(bytes / 1024 < 64L*1024);

	// convert bytes to K, round up or down accordingly
	kBytes = (bytes / 1024);
	remainder = (unsigned)(bytes % 1024);
	if(remainder >= 512)
		++kBytes;

	// return # of K bytes
	return kBytes;
}

/*
	Returns the number of paragraphs taken up by the program
	and its environment block.
*/
unsigned long BytesWeOwn(void)
{ unsigned ourParas;
#if 0
	asm {
		mov	ax, _psp
		mov 	es, ax
		mov	bx, es:[44]
		dec	ax
		mov	es, ax
		mov	ax, es:[3]
		mov	ourParas, ax
		dec	bx
		mov	es, bx
		mov	ax, es:[3]
		add	ourParas, ax
	}
#endif
	asm {
		mov	ax, _psp
		dec	ax
		mov	es, ax
		mov	ax, es:[3]
		mov	ourParas, ax
	}
	return ourParas * 16L;
}

/*
	Recursively exhaust the free DOS MCBs, counting the total
	number of bytes we could allocate (includes holes left by
	original environment, etc).
*/
unsigned long GetFreeConventionalMem(void)
{ unsigned long freeBytes = 0;
  unsigned freeParas = 0, allocSeg;

   asm {
		push	bx
		mov	ah, 48h
		mov	bx, 0FFFFh
		int	21h
		or		bx, bx
		jz		NO_MORE_MEM

		mov	freeParas, bx
		mov	ah, 48h
		int	21h
		mov	allocSeg, ax
  	}

	freeBytes = GetFreeConventionalMem();
	asm {
		mov	ah, 49h
		mov	es, allocSeg
		int	21h
	}
NO_MORE_MEM:
	asm {
		pop	bx
	}

	return freeBytes+(freeParas*16L);
}

/*
	Determine the size, in bytes, of the single largest free MCB.
*/
unsigned long GetContiguousFreeMem(void)
{ unsigned freeParas, maxProgSize;

   asm {
		push	bx
		mov	ah, 48h
		mov	bx, 0FFFFh
		int	21h
		mov	freeParas, bx

		mov	ah, 4Ah
		mov	es, _psp
		mov	bx, 0FFFFh
		int	21h
		mov	maxProgSize, bx
		pop	bx
  	}
	if(freeParas > maxProgSize)
		return freeParas*16L;
	else
		return maxProgSize*16L;
}

/*
	Returns the segment value of the last paragraph of conventional memory.
*/
unsigned TopMemSeg(void)
{
	return GetConvMem() << 6;
}

/*
	Returns the LIM 3.2 page frame segment address or -1 if no EMM
*/
unsigned GetLIM32PageFrame(void)
{
	bool fSuccessful;
	unsigned frameSeg = 0xFFFF;
	unsigned handle;
	unsigned uAction = ACT_EXIST_OPEN;


	fSuccessful = dlOpenOrCreateFile( O_RDONLY, A_NORMAL, &uAction,
			"EMMXXXX0", (int *)(&handle));
	if(!fSuccessful)
	{
		return frameSeg;
	}
	dlCloseFile(handle);

	asm {
		push	bx
		mov	ah, 40h
		int	67h
		or		ah, ah
		jnz	DONE
		mov	ah, 46h
		int 	67h
		mov	bx, -1
		cmp	al, 40h
		jne	DONE
		mov	ah, 41h
		int   67h
		mov	frameSeg, bx
	}
DONE:
	asm{
		pop	bx
	}

	return frameSeg;
}

/*
	Sets arguments to the count total and free 16K LIM 3.2 pages.
	Only call this if GetLIM32PageFrame() != -1.
*/
void GetStdEMSPageCounts(unsigned long *pTotal, unsigned long *pFree)
{ unsigned total, free;
	asm {
		push	bx
		xor	bx, bx
		mov	dx, bx
		mov	ah, 42h
		int	67h
		mov	total, dx
		mov	free, bx
		pop	bx
	}
	*pTotal = total * 16L * 1024;
	*pFree  = free  * 16L * 1024;

}

/*
	Show the normal memory used/free chart
*/
void ShowMemory(void)
{
	unsigned long total, free, biggestFree;
	unsigned long totalUMB, freeUMB, biggestFreeUMB;
	unsigned long totalROM;
	unsigned long totalXMS, freeXMS;
	unsigned long totalEMS=0, freeEMS=0;
	char szSizeStr[20];

	// display chart header
	printf(BasicStrings[0]);
	printf(BasicStrings[1]);


	if(GetLIM32PageFrame() != 0xFFFF)
	{
//		unsigned total, free;

		GetStdEMSPageCounts(&totalEMS, &freeEMS);
	}

	// display conventional memory usage
	total = 1024L * GetConvMem();
	free = GetFreeConventionalMem();
	if(_psp < TopMemSeg())
		free += BytesWeOwn();
	printf(BasicStrings[2], K(total), K(total)-K(free), K(free));

	// display upper memory usage
	GetUMBSize(&totalUMB, &freeUMB, &biggestFreeUMB);
	printf(BasicStrings[3], K(totalUMB), K(totalUMB) - K(freeUMB), K(freeUMB));

	// Show free space between 640K and 1 Meg
	// the top of memory might NOT be 640K, so use formula
	totalROM = 1024L*1024 - total;
	printf(BasicStrings[4], K(totalROM), K(totalROM), 0L);

	// display XMS statistics
	GetExtendedMemoryStats(&totalXMS, &freeXMS);
	printf(BasicStrings[5], K(totalXMS), K(totalXMS) - K(freeXMS), K(freeXMS));

	// display totals for whole system
	printf(BasicStrings[6]);
	printf(BasicStrings[7],
		K(total+totalROM+totalXMS),
		K(total+totalROM+totalXMS) - K(free+freeUMB+freeXMS),
		K(free+freeUMB+freeXMS));

	// display totals for under 1 Meg
	printf(BasicStrings[8],
		K(total + totalUMB),
		K(total + totalUMB) - K(free + freeUMB),
		K(free + freeUMB));

	// show EMS, if there
	if(totalEMS)
	{
		sprintf(szSizeStr, msgBytes, totalEMS);
		printf(BasicStrings[12], K(totalEMS), szSizeStr);
		sprintf(szSizeStr, msgBytes, freeEMS);
		printf(BasicStrings[13], K(freeEMS), szSizeStr);
	}

	// show largest free in Conventional Memory
	biggestFree = GetContiguousFreeMem();
	sprintf(szSizeStr, msgBytes, biggestFree);
	printf(BasicStrings[9], K(biggestFree), szSizeStr);

	// show largest free in Upper Memory
	sprintf(szSizeStr, msgBytes, biggestFreeUMB);
	printf(BasicStrings[10], K(biggestFreeUMB), szSizeStr);

	// show where DOS is
	printf(BasicStrings[11], GetDOSName(), GetDOSLocation());
}

/*
	Get Conventional memory size using the BIOS interrupt 12.
	Returns the # of K bytes.
*/
unsigned GetConvMem(void)
{
	unsigned memK;

	asm{
		int	12H
		mov	memK, ax
	};


	return(memK);
}

/*
	Display all BIOS extensions in the system.
*/
void ShowBiosExtensions(void)
{
	int found = 0;
	struct
	{
		unsigned Sig;
		unsigned char Size;		// size in pages (each page is 512 bytes)
	} huge * Extents;
	unsigned Size;

	// start looking for BIOS extension at C000:0
	Extents = MK_FP(BIOSEXT_STARTSEG, 0);

	// only look up to F800:0
	while(FP_SEG(Extents) < BIOSEXT_ENDSEG)
	{
		// must be a BIOS extension
		if(Extents->Sig == BIOSEXT_SIG)
		{
			// display BIOS extension and its size
			Size = Extents->Size * BIOSEXT_PAGESIZE;
			printf(msgBiosExt, FP_SEG(Extents), FP_OFF(Extents), Size/1024, Size);

			// on to next bios extension
			(char huge *)Extents += Size;

			// found another bios extension
			++found;
		}

		// check at 2K increments
		else
			(char huge *)Extents += BIOSEXT_INCR;
	}

	// no bios extensions.
	if(!found)
		printf(msgNoBiosExt);
}

/*
	Get pointer to the list of lists.
*/
void far *GetLOL(void)
{
	unsigned uSegment, uOffset;

	asm{
		push	bx
		mov	AX,0x5200
		int	21H
		mov	uSegment, ES
		mov	uOffset, BX
		pop	bx
	};

	return(MK_FP(uSegment, uOffset));
}


/*
	Return the segment of the first MCB.  The word is found
	in DOS' data just prior to the list of lists.  Cannot fail.
*/
unsigned Get1stMCB(void)
{
	unsigned far *p = GetLOL();
	--p;
	return (*p);
}


/*
 * Get the Next MCB in the Chain including the UMB chain.  This routine
 * takes into account an MCB following the end MCB (MCB with a 'Z' signiture)
 * is the start of the UMB Chain.
 */
struct MCB far *GetNextMCB(struct MCB far *nextMCB)
{
	struct MCB far *tmp;

	if(nextMCB->id == 'Z')
	{
		tmp = MK_FP(FP_SEG(nextMCB) + nextMCB->para_cnt + 1, 0);
		if(tmp->id == 'M')
			return tmp;
		else
			return (NULL);
	}

	return MK_FP(FP_SEG(nextMCB) + nextMCB->para_cnt + 1, 0);
}

/*
	Copy only the valid characters of the name
*/
void CopyValidName(char *pDstName, char far * lpSrcName)
{
	int i;

	for(i=0; i<8; ++i)
	{
		if(!isprint(*lpSrcName))
			break;
		*pDstName++ = *lpSrcName++;
	}

	*pDstName = 0;
}

/*
	Get the name from an MCB.  Is smart, and determines
	if this MCB should have a name, and who owns it.
*/
void GetMCBName(char mcbName[], struct MCB far *lpMCB)
{
	// this MCB might have a name
	if(lpMCB->owner == 0)
		strcpy(mcbName, "Free");

	else if(lpMCB->owner == SYSTEM_MCB)
	{
		CopyValidName(mcbName, lpMCB->name);
		mcbName[2] = 0;		// by luck might be extra chars, chop em
	}

	// does MCB have a name?
	else if(FP_SEG(lpMCB)+1 == lpMCB->owner)
		CopyValidName(mcbName, lpMCB->name);

	// MCB does NOT have name, but it's owner might
	else
	{
		lpMCB = MK_FP(lpMCB->owner-1, 0);
		CopyValidName(mcbName, lpMCB->name);
	}
}

/*
 * Display the passed MCB.
 */
void DisplayMCB(struct MCB far *lpMCB)
{
	char mcbName[10];
	char szBytes[24];

	// get name from the MCB
	GetMCBName(mcbName, lpMCB);

	if(lpMCB->owner == SYSTEM_MCB && strcmp(mcbName, "SD") == 0)
		ShowSystemTable(FP_SEG(lpMCB)+1);

	else if(strcmp(mcbName, "SC") == 0)
		; // DO Nothing;

	else if(lpMCB->owner == 0 || lpMCB->para_cnt == 0)
		;

	else
	{
		sprintf(szBytes, "(%,ld bytes)", 16L * lpMCB->para_cnt);
		printf("%-8s  %04X:0   %3dK %15s\n",
			mcbName, FP_SEG(lpMCB), lpMCB->para_cnt/(1024/16), szBytes);
	}
}

/*
	DOS builds an MCB chain within the first MCB chain.  This
	contains structures like FILES, BUFFERS, CDS, etc.  Display
	each item in that chain.
*/
void ShowSystemTable(unsigned mcbSeg)
{
	struct MCB far *MCB;

	MCB = MK_FP(mcbSeg, 0);
	while(MCB != NULL)
	{
		if(MCB->id == 'D')
			DisplayMCB(MCB);

		MCB = GetNextMCB(MCB);
	}
}


/*
 * Raw Dump ofthe MCB CHain!
 * Datalight MEM specific function.
 */
void ShowRawDump(void)
{
	struct MCB far * lpMCB;
	char mcbName[10];

	lpMCB = MK_FP(Get1stMCB(), 0);

	printf(msgRawHeader);

	while(lpMCB != NULL)
	{
		GetMCBName(mcbName, lpMCB);

		printf("   %04X:0000   %c   %04X    %04X     %s\n",
				FP_SEG(lpMCB), lpMCB->id, lpMCB->owner, lpMCB->para_cnt,  mcbName);

		lpMCB = GetNextMCB(lpMCB);
	}
}


/*
 * Show the programs loaded into memory.
 */
void ShowByProgram(void)
{
	struct MCB far *lpMCB;

	// get the first MCB
	lpMCB = MK_FP(Get1stMCB(), 0);

	// display header to module list
	printf(msgByProgramHeader);

	while(lpMCB != NULL)
	{
		// determine size in bytes
		if(lpMCB->para_cnt)
			DisplayMCB(lpMCB);

		lpMCB = GetNextMCB(lpMCB);
	}

	printf("\n");
}


/*
	Get the first MCB of the UMB Chain.
*/
unsigned Get1stUMB(void)
{
	struct MCB far *lpMCB;
	struct MCB far *lpUMB;

	unsigned para = Get1stMCB(), oldUMBLink;

	asm {
		push	bx
		mov	ax, 5802h
		int 	21h
		xor	ah, ah
		mov	oldUMBLink, ax
		mov 	ax, 5803h
		xor 	bx, bx
		int	21h
		pop	bx
	}

	// walk the standard MCB list
	do
	{
		lpMCB = MK_FP(para, 0);

		para += (lpMCB->para_cnt + 1);

		// sanity check
		if(lpMCB->id != 'M' && lpMCB->id != 'Z')
		{
			printf(errMsgMcbChain);
			exit(ERR_MCBCHAIN);
		}

	} while(lpMCB->id != 'Z');

	asm {
		push	bx
		mov 	ax, 5803h
		mov 	bx, oldUMBLink
		int	21h
		pop	bx
	}


	// are there UMBs past the 'Z' entry?
	lpUMB = MK_FP(FP_SEG(lpMCB) + lpMCB->para_cnt + 1, 0);
	if(lpUMB->id == 'M')
		return (FP_SEG(lpUMB) + lpUMB->para_cnt + 1);

	// no UMBs
	return (0);
}


/*
	Get the size of the UMB area, total, free space, and largest free.

	Returns sizes in bytes.
	Returns 0s in everything if no UMBs.
 */
void GetUMBSize(unsigned long *pTotal, unsigned long *pFree,
						unsigned long *pBiggestFree)
{
	struct MCB far * lpMCB;
	unsigned total, free, biggest;	// in paragraphs
	unsigned segUMB;

	// get first entry in UMB list
	segUMB = Get1stUMB();

	// no UMBs?
	if(segUMB == 0)
	{
		*pTotal = *pFree = *pBiggestFree = 0;
		return;
	}

	// count the sizes
	total = free = biggest = 0;
	do
	{
		// get pointer to the MCB
		lpMCB = MK_FP(segUMB, 0);

		// sanity check
		if(lpMCB->id != 'M' && lpMCB->id != 'Z')
		{
			printf(errMsgMcbChain);
			exit(ERR_MCBCHAIN);
		}

		// don't count system MCBs with 'SC' - system code as their
		// names.  They are just filler to get around gaps in UMBs.
		if(lpMCB->owner != SYSTEM_MCB || *(unsigned far *)lpMCB->name != 'SC')
			total += lpMCB->para_cnt;

		// if the UMB is free, we need to update free sizes
		if(lpMCB->owner == 0)
		{
			free += lpMCB->para_cnt;
			if(lpMCB->para_cnt > biggest)
				biggest = lpMCB->para_cnt;
		}

		// go on to next MCB in the UMB chain
		segUMB += (lpMCB->para_cnt + 1);

	} while(lpMCB->id != 'Z');

	// convert paragraphs to bytes
	*pTotal			= 16L * total;
	*pFree			= 16L * free;
	*pBiggestFree	= 16L * biggest;

	if(_psp > TopMemSeg())
	{ unsigned long ourBytes = BytesWeOwn();

		*pFree += ourBytes;
		asm {
			push	bx
			mov	ah, 4Ah
			mov	es, _psp
			mov	bx, 0FFFFh
			int	21h
			mov	ax, bx
			xor	dx, dx
			shl	ax, 1
			rcl	dx, 1
			shl	ax, 1
			rcl	dx, 1
			shl	ax, 1
			rcl	dx, 1
			shl	ax, 1
			rcl	dx, 1
			mov	word ptr ourBytes, ax
			mov	word ptr ourBytes[2], dx
			pop	bx
		}
		if(*pBiggestFree < ourBytes)
			*pBiggestFree = ourBytes;
	}
}


/*
	Calls the XMS driver and gets the address of XMS handler.
	Returns 0 if XMS (HIMEM) not present.
*/
void far * GetXMSAddress(void)
{
	unsigned xmsSeg=0, xmsOff=0;

	asm{
		push	bx
		mov	ax, 0x4300
		int	0x2f
		cmp	al, 80H
		jne	Done				// himem not loaded

		mov	ax, 0x4310
		int	0x2f
		mov	xmsSeg, es
		mov	xmsOff, bx
	}
Done:
	asm{
		pop	bx
	}
	return MK_FP(xmsSeg, xmsOff);
}


/*
	Get XMS memory Status calls the XMS driver and gets avail
	and used memory.
	Returns TRUE if HIMEM is loaded, FALSE if not.
*/
void GetExtendedMemoryStats(unsigned long *pXMSTotal, unsigned long *pXMSFree)
{
	unsigned freeK, totalK;
	unsigned dosSeg;
	unsigned far * lpTotalXMS;

	// get total size of Extended Memory from DOS
	asm{
		push	bx
		mov	ax, 5200h
		int	21h
		mov	dosSeg, es
		pop	bx
	}
	lpTotalXMS = MK_FP(dosSeg, 0x6B);
	totalK = *lpTotalXMS;

	// determine address of XMS handler
	if(XMS_Control == NULL)
		XMS_Control = GetXMSAddress();

	// no XMS, so just report what DOS says, but in bytes
	if(XMS_Control == NULL)
	{
		*pXMSTotal = *pXMSFree = 1024L * totalK;
		return;
	}

	// ask XMS for free space of largest block
	asm	push	bx
	asm   mov	dx, 0
	asm	push	ds
	asm	mov	ah, 8
	(*XMS_Control)();
	asm	pop	ds
	asm	mov	freeK, dx
	asm	pop	bx

	// return the sizes back to caller in bytes
	*pXMSTotal = 1024L * totalK;
	*pXMSFree  = 1024L * freeK;
}

/*
 * Get the name of the DOS that is running.
 * Use special ROM-DOS DB Int 21 function to see if we are running
 * ROM-DOS else it must be MS-DOS!
 */
char * GetDOSName(void)
{
	asm	push	bx
	asm	mov	ax, 30DBH
	asm	mov	cx, 0
	asm	mov	di, 3
	asm	mov	si, 0B2D2H
	asm	int 	21H
	asm	pop	bx
	asm 	cmp 	cx, 0B2D2H
	asm 	jz		its_ROM_DOS

	return ("MS-DOS");

its_ROM_DOS:
	return ("ROM-DOS");
}


/*
	Get the Location in ROM, Conventional, or HMA of the DOS
	that is running.
*/
char * GetDOSLocation(void)
{
	unsigned char DosPosition;
	char *pWhereIsDos;

	// default to conventional
	pWhereIsDos = msgInConventional;

	// ask DOS where it is
	asm	push	bx						// BC 5.2 uses BX register for variables
	asm	mov	dh, 0
	asm	mov	ax, 3306H
	asm	int	21H
	asm	mov	DosPosition, dh
	asm	pop	bx						// BC 5.2 uses BX register for variables

	// DOS in ROM
	if(DosPosition & 8)
		pWhereIsDos = msgInROM;

	// DOS in the High Memory Area
	if(DosPosition & 16)
		pWhereIsDos = msgInHMA;

	return pWhereIsDos;
}
