//
//   $Id: MORE.C 1.13 2003/08/26 16:46:54Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
	$DOCMODULE:MORE.C

	Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
	Datalight Confidential and Proprietary.

	Display paged output to the console.

	$DOCHISTORY
	02/15/2000 js	Added LFN compatabiliy
	08/20/1999 dg  Added knowlege of rows/cols. Uses dirsize.
****************************************************************************/

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dllfn.h"
#include "dosver.h"
#include "fatworld.h"
#include "dldosio.h"
#include "abortlfn.h"
#include "stdc.h"

int eprintf(const char *fmt, ...);

#define isspace( c )		(c == ' ')

// Various More Exit (ME) Conditions
// for errorlevel compatibility with MS
//
#define ME_PATH_TOO_LONG 255
#define ME_BAD_OPTION    1

#define ME_MAIN_EXIT     0
#define ME_CANT_OPEN     0
#define ME_READ_0_BYTES  0
#define ME_HELP_REQUEST  0


#define SCREEN_SIZE (iScreenLines-1) * iScreenWidth

#define MORE 1
#include "dosmsg.c"

char signon[] = "MORE %s %s\n%s\n\n";

long current_position, current_bottom_line, length, ftell();

int  iScreenLines = 25;
int  iScreenWidth = 80;
int  iCount;
bool fDoneWithFile = FALSE;

int iFileHandle = 0;          //Initialize

char *FileName;

bool fIsFat32World;
bool fHideSignon  = FALSE;

// prototypes
void show(int count);
void goto_line(long position);
int xgetc(void);
void do_args(int argc, char *argv[]);
void do_help(void);
void GetScreenSize(int * piRows, int * piColumns);
char * NextAtoi(char *psz);


void main(int argc, char *argv[])
{
	int c, iActionTaken;
	long ret;
	char *cstr = "";
	bool fSuccessful;
	unsigned uDeviceAttributes;
   int iCurrentArgCount;
   int iRetVal = ME_MAIN_EXIT;
   bool bNoPipeInput;
   bool not_yet_done = TRUE;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	//  Get initial screen rows (lines) and columns (width)
	GetScreenSize(&iScreenLines, &iScreenWidth);


//  BT 8-9-00
//
//  More does not need any arguments. it is mainly used as a pipe!
//
//	//  Do command line parsing
//	if(argc < 2)
//		do_help();
//

   for(iCurrentArgCount=1; iCurrentArgCount <= argc; iCurrentArgCount++)
   {
      LFNStripArgQuotes( argv[iCurrentArgCount] );
      if( strlen(argv[iCurrentArgCount]) >= LFN_MAX_PATH )
      {
         printf("Error: Path too long.\n");
         exit(ME_PATH_TOO_LONG);
      }
   }

   // Get all the switches --- filenames are handled in the loops below
   //
   do_args(argc, argv);

   printf("\n");

   iCurrentArgCount = 1;
   dlGetDeviceAttributes( STDIN, &uDeviceAttributes );
   bNoPipeInput = uDeviceAttributes & 0x80;

   // Process all the files supplied.  First, handle the pipes;
   // then continue through the command line.
   //
   while(not_yet_done)
   {
		fDoneWithFile = FALSE;

      if( bNoPipeInput )
      {
         while(   ( argv[iCurrentArgCount][0] == '/' )
               && ( iCurrentArgCount < argc ) )
         {
            iCurrentArgCount++;
         }
         if( iCurrentArgCount >= argc )
         {
            not_yet_done = FALSE;
         }
         else
         {
            FileName = argv[iCurrentArgCount];

            fSuccessful = SmartCreateOpenFile(OPEN_ACCESS_READONLY, 0,
               FILE_OPEN, FileName,	&iFileHandle, &iActionTaken);
            if(!fSuccessful)
            {
               eprintf(cant_open_str, FileName);

               iRetVal = ME_CANT_OPEN;
               not_yet_done = FALSE;
            }
         }
         // Done with this argument, move to the next one
         //
         iCurrentArgCount++;
      }
      else
      {
         // Deal with the pipe input, then next time hit the args
         //
         bNoPipeInput = TRUE;
      }

      // If there are no errors up to this point, continue with the display
      //
      if( not_yet_done )
      {
			ret = -1L;
		 	dlSeekFile( 2, iFileHandle, &ret); /* Move to end of file */
		 	length = ret+1;								 // ret = EOF - 1
			ret = 0;
			dlSeekFile( 0, iFileHandle, &ret); /* Move to beginning of the file */

			// Screen lines are numbered  0..iScreenLines
		   // Blank line at the top of the screen and --More-- space below
		   //
			iCount = iScreenLines-1;
		   current_bottom_line = 0;

			while(TRUE)
			{
				/* display the line(s) */
				show(iCount);

				//  Check to see if we are done with the current file.
				//
				if( fDoneWithFile )
				{
					break;
				}

		      current_bottom_line += iCount;
		      iCount = iScreenLines-1;         // reset to default

			  	ret = 0;
				dlSeekFile( 1, iFileHandle, &ret);

				/* determine current position and display it */
				current_position = ret;
				printf(more_str, (current_position*100L)/length);

				/* wait for some user input */
				c = xgetc();
				c = toupper(c);
				printf(clear_more);	/* Clear -- More %u line */

				/* b = back one screenful */
				if( c == 'B')
				{
		         if ((current_bottom_line-2*iCount) <= 0)
		         {
		            // past the beginning of the file, same as a "t" or "T"
		            c = 'T';
		         }
		         else
		         {
		            // we will display the next screen from this new spotf
		            goto_line( current_bottom_line-(long)2*iCount );
		            current_bottom_line -= 2*iCount;
		         }
				}

				/* t = top of file */
				if(c == 'T')
				{
					ret=0L;
					dlSeekFile(SEEK_SET, iFileHandle, &ret); /* Move to Start of file */
		         current_bottom_line = 0;
		         printf ("\n");
				}

				/* q = quit */
				else if(c == 'Q')
					break;

				/* ctrl-C = quit */
				else if(c == ctrl('C'))
				{
					cstr = "--More--^C\n";
					break;
				}

				/* CR = one line at a time */
				else if(c == '\r')
		      {
					iCount = 1;
		   		printf(clear_more);	/* Clear -- More %u line */
		      }
			}

			/* blank out the final prompt */
		   printf("%s",cstr);

         // Clean up the mess...
         //
         if( iFileHandle )
         {
            dlCloseFile( iFileHandle );
         }
      }
   }

	exit(iRetVal);
}


/*
	Show one or more lines. Knows about screen height and width.
*/
void show(int iCount)
{
	static char szLine[256];
	unsigned uBytesInBuffer = 0;
	unsigned uBytesRead;
	unsigned uLineCount;
	unsigned uBufferIndex;
	bool fSuccessful;
	long lSeekDist;
	char cSavedChar;
	unsigned uWidth;
	unsigned uMaxIndex;


	uWidth  = iScreenWidth;
	uWidth++;


	for( uLineCount = 0; uLineCount < iCount; uLineCount++ )
	{
		//  Fill the buffer only as needed
		//
		if( uBytesInBuffer < (uWidth + 1) )
		{
			//  Fill up the file buffer
			//
			uBytesRead  = sizeof(szLine);
			uBytesRead -= uBytesInBuffer;
			fSuccessful = dlReadFile(iFileHandle, &uBytesRead,
					&(szLine[uBytesInBuffer]) );
			if( !fSuccessful )
	      {
				exit(ME_READ_0_BYTES);
	      }

			if( uBytesRead == 0 )
			{
				if( uBytesInBuffer == 0 )
				{
					fDoneWithFile = TRUE;
					return;
				}
			}

			uBytesInBuffer += uBytesRead;
		}


		//  Find the end of this line which could be the screenwidth or
		//  a 0x0d
		//
		uMaxIndex = uWidth;
		if( uMaxIndex > uBytesInBuffer )
		{
			uMaxIndex = uBytesInBuffer;
		}
		for( uBufferIndex = 0; uBufferIndex < uMaxIndex; uBufferIndex++ )
		{
			if( szLine[uBufferIndex] == 0x0d )
			{
				break;
			}
		}


		//  Display the current line
		//
		cSavedChar = szLine[uBufferIndex];
		szLine[uBufferIndex] = 0;
		printf("%s", szLine);
		if( uBufferIndex < uWidth )
		{
			printf("\n", szLine);
		}

      // End Of File handling -- the "end in nothing" case
      //
      if(uBufferIndex < uBytesInBuffer)
      {
         // End Of File handling -- the "end in 0x0D" case
         //
         szLine[uBufferIndex] = cSavedChar;
         if( szLine[uBufferIndex] == 0x0d )
         {
            // End Of File handling -- the "end in 0x0D0A" case
            //
            uBufferIndex++;
            if(uBufferIndex < uBytesInBuffer)
            {
               if( szLine[uBufferIndex] == 0x0a )
               {
                  uBufferIndex++;
               }
            }
         }
      }

		//  Back all the bytes in the buffer to the front of the buffer
		//
		uBytesInBuffer -= uBufferIndex;
		if( uBytesInBuffer )
		{
			memmove( szLine, &(szLine[uBufferIndex]), uBytesInBuffer );
		}
	}


	// If the string displayed was shorter than the string read,
	// seek back to begining of next string.
	//
	lSeekDist = 0L - uBytesInBuffer;
	dlSeekFile(SEEK_CUR, iFileHandle, &lSeekDist);
}


/*
	From the top, go to a particular line, showing none along the way.
   Note this is a starting line, the page will display from here.
*/
void goto_line(long target_line)
{
	long position;
	unsigned uBytesRead;
	bool fSuccessful;
	char szLine[80], *pEndOfLine;
	long lSeekDist;

   lSeekDist=0L;
   dlSeekFile(SEEK_SET, iFileHandle, &lSeekDist);  // top of file

	for(position=0; position<target_line; position++)
	{
		// Read max number of characters that will be in a line
		uBytesRead = iScreenWidth;
		fSuccessful = dlReadFile(iFileHandle, &uBytesRead, szLine);
		if((!fSuccessful) || (uBytesRead == 0) )
      {
			exit(ME_READ_0_BYTES);
      }

		// Truncate string read from file before performing string search
		*(szLine + uBytesRead) = 0;

		// Truncate string after newline ('\n')
		pEndOfLine = strchr(szLine, (char)0x0d );
		if(pEndOfLine == NULL) 						// Newline not in string
		{
			pEndOfLine = szLine + uBytesRead;
			*(pEndOfLine) = 0;
		}
		else
			*(pEndOfLine++) = 0;

		// If the string displayed was shorter than the string read,
		// seek back to begining of next string.
		lSeekDist = (long)(pEndOfLine - szLine) - uBytesRead;
		dlSeekFile(SEEK_CUR, iFileHandle, &lSeekDist);
	}
}


/*
	Determine screen rows and columns. Uses the BIOS described
	rows/columns by default. They can be overridden by the DIRSIZE
	environment variable.
*/
void GetScreenSize(int * piRows, int * piColumns)
{
	#define MAX_ROWS	256
	#define MAX_COLS	256
	unsigned char far * lpByte;
	char * pszEnv;
	int	i;
	static char szEnvTemp[10];
	char far * lpszTemp;


	// get rows and columns from BIOS data area
	lpByte = MK_FP(0x40, 0x84);		// highest row
	*piRows = *lpByte + 1;
	lpByte = MK_FP(0x40, 0x4A);		// # columns
	*piColumns = *lpByte;

	// allow DIRSIZE to override default
	lpszTemp = GetEnvEntry("DIRSIZE");
   if( lpszTemp )
   {
      pszEnv = szEnvTemp;
      _fmemcpy( pszEnv, lpszTemp, sizeof(szEnvTemp) );
      // get rows, validate range
      i = atoi(pszEnv);
      if(i > 1 && i <= MAX_ROWS)
      {
         *piRows = i;
      }

      // get cols, validate range
      pszEnv = NextAtoi(pszEnv);
      i = atoi(pszEnv);
      if(i > 1 && i <= MAX_COLS)
      {
         *piColumns = i;
      }
   }
   (*piColumns)--;
   (*piRows)--;
}

/*
	Skips past an atoi() argument. Example "   12   9" would point to
	the first whitespace just past the "12".
*/
char * NextAtoi(char *psz)
{
	// first, skip preceding whitespace
	while(*psz && isspace(*psz))
		++psz;

	// next, skip non-whitespace
	while(*psz && !isspace(*psz))
		++psz;

	return psz;
}

int xgetc(void)
{
//	union REGS regs;
	char c;

	//  wait for a keypress (so cmdlog can be used in testing)
	//
CheckAgain:
	asm	mov	ah, 01h
	asm	int	16h
	asm	jz		CheckAgain


	//  Get the key
	//
	asm	mov	ah, 00h
	asm	int	16h
	asm	mov	c, al

	return (c);
}


void do_args(int argc, char *argv[])
{
	int i;

	for(i=1; i<argc; ++i)
	{
		if(argv[i][0] == '/')
		{
			switch(argv[i][1])
			{
				case 'h':
				case 'H':
				case '?':
					{
						do_help();
						break;
					}

				case '[':
					{
						fHideSignon = TRUE;
						break;
					}

				default:
               {
					   printf(bad_opt, argv[i]);
						exit( ME_BAD_OPTION );
               }
			}
		}
	}
}

void DisplaySignOnMessage( void )
{
	if( !fHideSignon )
	{
		if( fIsFat32World )
		{
			printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}

void do_help(void)
{
	DisplaySignOnMessage();
	printf(USAGE);
	exit(ME_HELP_REQUEST);
}
