//
//   $Id: PRINT.C 1.4 2002/10/16 21:58:28Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   PRINT.C  created 01/17/90  by Drew Gislason

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	This file contains printf.

**********************************************************/

#include <ctype.h>
#include <dos.h>
#include <string.h>
#include <stdlib.h>
#include <mem.h>

// do not print anything over 80 columns in length
int silent;
char prbuff[100];

void printf_on(void)
{
	silent = 0;
}

void printf_off(void)
{
	silent = 1;
}

/*
	Print the string to CONsole.  This function does NOT go
	though handles - it uses INT 29H directly.
*/
void	internal_puts(char *s)
{
	if(silent)
		return;
	while(*s)
	{
		_BX = 7;
		_AL = *s;
		_AH = 0x0E;
		geninterrupt(0x10);
		++s;
	}
}

void (*print_func)(char *) = internal_puts;
static char fillchar;

/*
	Change a number from a word to a decimal or hexidecimal
	number.  Radix can be anything from 2 to 16.  A radix
	of 2 or 16 will fill with 0s.
*/
void xitoa(char *buffer, unsigned long num, int radix, int fill)
{
	static char hex[] = "0123456789ABCDEF";
	char str[18];
	int size = 0;

	do
	{
		str[size] = hex[(int)(num % radix)];
		num = num/radix;
		++size;
	} while(num);

	/* user may have asked for a fill w/ 0s */
	fill -= size;
	while(fill > 0)
	{
		*buffer++ = fillchar;
		--fill;
	}

	/* reverse the string */
	while(size-- > 0)
		*buffer++ = str[size];
	*buffer = 0;
}


/*
	Don't accept cheap imitations!  This is the real printf
	specifically for ROM-DOS.  What a treat!  However,
	there are limitations: No # formatting at all (that is
	%-5.3s will fail, you must use just %s).

	%c			Print a character
	%d			Print an integer
	%s			Print a string
	%p			Print a near pointer (padded with 0s)
	%x			Print an integer in hex
	%ld		Print a long
	%ls		Print a far string
	%lp		Print a far pointer (padded with 0s)
	%lx		Print a long in hex
*/
void vprintf(char *fmt, unsigned *args)
{
	/*
		[ externals ]
	*/
	extern char prbuff[];


	/*
		[ locals ]
	*/
	unsigned segment, offset;
	char *out;			/* the output */
	long l;
	int fill;

	/*
		[ code ]
	*/

	/* fill the string with our formatted text */
	out = prbuff;
	while(*fmt)
	{
		/* if we are filling up the buffer print it */
		if(out-prbuff > sizeof(prbuff)-10)
		{
			*out = 0;
			(*print_func)(prbuff);
			out = prbuff;
		}

		/* general output */
		if(*fmt != '%')
		{
			if(*fmt == '\n')
				*out++ = '\r';			/* CR before LF */
			*out++ = *fmt;
		}

		/* this is a formatted output */
		else
		{
			++fmt;

			// check for a fill value
			fill = 0;
			fillchar = '0';
			if(isdigit(*fmt))
			{
				fill = atoi(fmt);
				fillchar = (*fmt == '0' ? '0' : ' ');
				while(isdigit(*fmt))
					++fmt;
			}

			switch(*fmt)
			{
				/* unknown format, do not translate */
				default:
badform:
					*out++ = '%';
					*out++ = *fmt;
					break;

				/* character format */
				case 'c':
					*out++ = *args;
					break;

				/* integer format */
				case 'd':
				case 'u':
					xitoa(out, (long)(*args), 10, fill);
					out += strlen(out);
					break;

				/* near pointer format */
				case 'p':
					/* same as below */

				/* hexadecimal output */
				case 'x':
				case 'X':
					xitoa(out, (long)(*args), 16, fill);
					out += strlen(out);
					break;

				/* near string format */
				case 's':
					memmove(out,(char *)*args,strlen((char *)*args)+1);
					out += strlen(out);
					break;

				/* the long formats */
				case 'l':
					++fmt;
					l = *(long *)args;
					offset  = *args++;
					segment = *args;

					switch(*fmt)
					{
						default:
							--fmt;
							goto badform;

						case 'd':
						case 'x':
						case 'X':
							xitoa(out, l, *fmt == 'd' ? 10 : 16, fill);
							out += strlen(out);
							break;

						case 'p':
							xitoa(out, segment, 16, 4);
							out += strlen(out);
							*out++ = ':';
							xitoa(out, offset, 16, 4);
							out += strlen(out);
							break;
					}
					break;
			}

			++args;

		}		/* end of formatted output */

		++fmt;
	}

	/* actually display the output */
	*out = 0;
	(*print_func)(prbuff);
}


/*
	This is the entry for a printf.  It calls the real
	workhorse vprintf.  The function declaration is tabbed
	over, and a prototype is included in this comment for
	the use of my prototyping program.  That is how
	DOSPROTO.H is generated.

void printf(char *fmt, ...)
*/
	void	printf(char *fmt, unsigned arglist)
{
	vprintf(fmt,&arglist);
}
