//
//   $Id: REM8250.H 1.4 2002/10/16 21:58:28Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	REM8250.H  created 02/05/90  by Drew Gislason

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	This file contains 8250 UART Defines and Structures for.

**********************************************************/

/* Normal Defines */
#ifndef TRUE
#define TRUE  1
#endif
#define FALSE 0
#define ERR  -1
#define OK    0

/* do this in-line for speed (at 115K speed is important) */
#define comm_data_rdy() while(!(inportb(LINESTATUS) & LS_DRDY)) /* wait */

/* The Defaults */
#define BUFSIZE   200		/* a little over 2K, 2080 */
#define PARMS     N81
#define BAUD    B9600
#define PORT        0


/* 8259 Programable Interrupt Controler */
#define PIC_EOIVAL		0x20
#define PIC_EOIPORT		0x20
#define PIC_MASKPORT		0x21	/* each bit: 0=interrupt enabled, 1=masked off */

/* dual PIC support */
#define PIC_EOIPORT2		0xA0
#define PIC_MASKPORT2	0xA1	/* each bit: 0=interrupt enabled, 1=masked off */

/* 8250 UART Port Addresses */
#define DATAREG      (cp.base)
#define BAUDLOW      (cp.base)
#define BAUDHIGH     (cp.base+1)
#define INTENABLE    (cp.base+1)
#define INTID        (cp.base+2)		// read only interrupt ID
#define FIFOCONTROL  (cp.base+2)		// write only FIFO control
#define LINECONTROL  (cp.base+3)
#define MODEMCONTROL (cp.base+4)
#define LINESTATUS   (cp.base+5)
#define MODEMSTATUS  (cp.base+6)

/*
	Stop/Data/Parity Bits for 8250 UART

	Bits 43210
	     ---xx		Data Bits (00=5, 01=6, 10=7, 11=8)
	     --x--		Stop Bits (0=1, 1=2)
	     xx---		Parity (00=none, 01=odd, 11=even)
*/
#define N81     3
#define E71    26


/* 8250 UART Baud Rate Divisors (based on 115K clock) */
#define B300  384
#define B1200  96
#define B2400  48
#define B4800  24
#define B9600  12
#define B19200  6
#define B38400  3
#define B57600  2
#define B115200 1			/* highest PC clock rate */


/*
	(Line Status Register)          (Modem Status Register)
	1....... Timeout Error           1....... Carrier Detect
	.1...... Trans Sh-Reg Empty      .1...... Ring Indicator
	..1..... Trans Hold-Reg Empty    ..1..... Data-Set-Ready
	...1.... Break-Detect Error      ...1.... Clear-To-Send
	....1... Framing Error           ....1... (change in above)
	.....1.. Parity Error            .....1..        ""
	......1. Overrun Error           ......1.        ""
	.......1 Data Ready              .......1        ""
*/
#define LS_DRDY 		0x01
#define LS_OVERRUN	0x02
#define LS_TRDY 		0x20
#define MS_CARRIER	0x80
#define MS_RING 		0x40
#define MS_DSR  		0x20
#define MS_CTS  		0x10
#define MC_DTR			0x01
#define MC_RTS			0x02
#define MC_INT			0x08


/* this structure is used by the I/O routines */
typedef struct {
	int   base;					// I/O port base address eg 3F8H, 2F8H
	int   head, tail;			// head and tail of ring buffer
	int   bufsize;				// size of ring buffer
	unsigned char modemcontrol;
	unsigned char port;
	unsigned uiIrqNumber;
	unsigned char buffer[BUFSIZE];	// ring buffer
	} comport;

/* Externals */
extern comport cp;
