//
//   $Id: REMDISK.C 1.5 2002/10/16 21:58:28Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   REMDISK.C  created 10/05/90  by Drew Gislason

	Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	History:

	01-30-1999 bt  Renamed rdsk_bpb to std_bpb for memdisk support

**********************************************************/
#include <dos.h>

#include "mem.h"
#include "string.h"

#include "dosver.h"
#include "datalght.h"
#include "fatworld.h"
#include "dlbiosio.h"

#include "remote.h"
#include "remdisk.h"

#include "remdisk.pt"
#include "remdisk2.pt"
#include "remdisk3.pt"
#include "memdisk.pt"
#include "print.pt"

#define IOCTL 1		// allow Quit command IOCTL

extern int memerr;
extern uchar rdsk_debug;


extern char USAGE[];
extern char signon_str[];
extern char uninstall_str[];
extern char irq_range_str[];
extern char uninstall_fail_str[];
extern char install_err_str[];
extern char complete_str[];
extern char baud_str[];
extern char com_str[];
extern char irq_str[];
extern char drive_str[];
extern char time_str[];
extern char not_FAT16_str[];

extern int BAUD_POS;
extern int COM_POS;
extern int IRQ_POS;
extern int DRIVE_POS;
extern int TIME_POS;


extern int opt_unload;
extern int opt_showhelp;

extern struct BPB std_bpb;
extern uchar rdsk_err;

void rem_packet(char cmd, unsigned long sector, unsigned dataoff,
        unsigned dataseg, int num);

unsigned REM_TIMEOUT = REMDISK_TIMEOUT;

bool fIsFat32World;

/*
        Print the options message (via BIOS) and open the
        communications.  Used only upon initialization.
        Example:

        Installed as Drive A: /COM1 /B9600+
*/
void signon(void)
{
	int len;
	extern uchar rdsk_drive;
	extern uchar rdsk_mode;
	extern int rdsk_port;
	extern unsigned rdsk_baud;
	extern unsigned rdsk_uiIrqNumber;

	/* display signon and drive */
	drive_str[DRIVE_POS] = rdsk_drive+'A';
	bios_puts(drive_str);

	/* show the COM port */
	com_str[COM_POS] = rdsk_port + '1';
	bios_puts(com_str);

	/* show the IRQ number */
	if( rdsk_uiIrqNumber < 10 )
	{
		irq_str[IRQ_POS] = rdsk_uiIrqNumber + '0';
	}
	else
	{
		irq_str[IRQ_POS] = '1';
		irq_str[IRQ_POS+1] = (rdsk_uiIrqNumber - 10) + '0';
	}
	bios_puts(irq_str);

	/* show the baud rate */
	len = utoa(rdsk_baud,baud_str+BAUD_POS);
	if(rdsk_mode == M_PACKET)
         	baud_str[BAUD_POS+len] = '+';
	else
         	baud_str[BAUD_POS+len] = '-';
	baud_str[BAUD_POS+len+1] = ' ';
	bios_puts(baud_str);

	/* show the timeout */
	utoa(REM_TIMEOUT/18,time_str+TIME_POS);
	bios_puts(time_str);
	bios_puts("\r\n\n");

	/* initialize serial port */
	enter_rdsk();
	leave_rdsk();
}

/*
   Initialize the device.  This is called only once.  It
   fills in the Bios Parameter Block structure, and returns
   the amount of memory used by the device.

   bpb       The BPB must be filled in for any device.
   cmdline   This is the command line passed to a device
             that is installed via CONFIG.SYS.  This
             parameter is meaningless for built-in devices.
   *endseg   This unsigned is the paragraph just beyond
             the code and data of the device.  Update this
             only if the device requires more RAM from DOS
             (ala a RAM disk in conventional memory).

   Returns FALSE(zero) if failed, # of drives if worked.
*/
#pragma argsused
int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, char drive)
{
	extern char dl_copyright[];

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	// display copyright notice
	bios_puts(signon_str);
	if( fIsFat32World )
	{
		bios_puts( SHORTVERSION );
		bios_puts(" ");
		bios_puts( BUILDVERSION );
	}
	else
	{
		bios_puts( SHORTVERSION622 );
		bios_puts(" ");
		bios_puts( BUILDVERSION );
	}
	bios_puts("\n\r");
	bios_puts(COPYRIGHT);
	bios_puts("\n\r");

	parse_cmdline(drive, cmdline);

	//  set the default irq for the specified port if no irq was specified
	//
	if( rdsk_uiIrqNumber == 0 )
	{
		if( ( rdsk_port == 0 ) || ( rdsk_port == 2 ) )
		{
			rdsk_uiIrqNumber = 4;
		}
		else
		{
			rdsk_uiIrqNumber = 3;
		}
	}

	//  range check the irq
	//
	if( ( rdsk_uiIrqNumber < 3 ) || ( rdsk_uiIrqNumber > 15 ) )
	{
		bios_puts(irq_range_str);
      return 2;
	}


	if(opt_showhelp)
	{
		bios_puts(USAGE);
      return 0;
	}

	if(opt_unload)
	{
      bios_puts(uninstall_str);
      if(!tsr_uninstall())
      	bios_puts(uninstall_fail_str);
      return 0;
	}

	// don't install twice
	if(memdev_already_loaded())
	{
      bios_puts(install_err_str);
      return 0;
	}

	signon();

#if IBM
	// IBM wants a delay of 1 second so user can see drive
	{
		ulong ulCurrentTimerTick;
		ulong ulEndTimerTick;


		dlGetTimerTicks( &ulEndTimerTick );
		ulEndTimerTick += 18;
		do
		{
			dlGetTimerTicks( &ulCurrentTimerTick );
		}
		while( ulEndTimerTick > ulCurrentTimerTick );
	}
#endif

	memcpy( &(*bpb), &std_bpb, sizeof(struct BPB) );

	return 1;		/* 1 drive */
}

void memuninit(void)
{
   bios_puts(complete_str);
}


/*
   This routine determines if the disk has changed.  Update
   the BPB structure if the device HAS changed or MAY have
   changed.

   Returns -1 if Device HAS changed
   Returns  0 if Device MAY have changed
   Returns  1 if Device has NOT changed
*/


int  memchanged(struct BPB *bpb)
{
	extern char rdsk_chg;
	static char s[2];
	int retry_count = 0;

	if(rdsk_debug)
      bios_puts("C");

RETRY1:
	rem_packet('C',0L,0,0,0);

	if(rdsk_err)
	{
		if(++retry_count < MAX_RETRIES)
		{
      	if(rdsk_debug)
          	bios_puts("\007chgerr retry");
			goto RETRY1;
		}
      rdsk_chg = CHG_MAYBE;
	}
	else if(rdsk_chg != 1)
	{
RETRY2:
      // get the BPB
		//
      rem_packet('B',0L,0,0,0);
		if(rdsk_err)
		{
			if(++retry_count < MAX_RETRIES)
			{
      		if(rdsk_debug)
          		bios_puts("\007chgerr retry");
				goto RETRY2;
			}
         rdsk_chg = CHG_MAYBE;
		}
		memcpy( &(*bpb), &std_bpb, sizeof(struct BPB) );
	}

	if(rdsk_debug)
	{
      *s = rdsk_chg+'0';
      bios_puts(s);
	}

	return rdsk_chg;
}

bool memrdwr(int wr, long offset, unsigned len, char far * buffer)
{
	char cmd = wr ? 'W' : 'R';
	long sector = offset / std_bpb.sec_size;
	static char s[2];

	if(rdsk_debug)
	{
   	*s=cmd;
   	bios_puts(s);
   	printf("|%ld| ",sector);
	}

	// length of read
	len /= std_bpb.sec_size;

	// do the actual rd/wr
	rem_packet(cmd, sector, FP_OFF(buffer), FP_SEG(buffer), len);

	if(rdsk_err)
	{
   	if(rdsk_debug)
      	bios_puts("err");

   	return 0;
	}
   return 1;
}

/*
   This function reads data (usually 1 or more sectors)
   from the ROM/RAM/Flash disk into the buffer specified
   by DOS.

   offset   The offset (in bytes) into the disk where the
            read should start.
   len      The length of the read in bytes.  The length
            will not exceed 62K (62*1024 bytes).
   buffer   The buffer, in segment:offset form.

   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/

bool memread(long offset, unsigned len, char far * buffer)
{
	bool rc;
	int retry_count = 0;

RETRY:
	rc = memrdwr(0,offset,len,buffer);
	if (!rc)
	{
		if(++retry_count < MAX_RETRIES)
		{
      	if(rdsk_debug)
          	bios_puts("\007read retry");
			goto RETRY;
		}
		memerr = DEVERR_READ;
	}
	return rc;
}


/*
   This function writes data (usually 1 or more sectors)
   to the ROM/RAM/Flash disk from the buffer specified
   by DOS.

   offset   The offset (in bytes) into the disk where the
            write should start.
   len      The length of the write in bytes.  The length
            will not exceed 62K (62*1024 bytes).
   buffer   The buffer, in segment:offset form.

   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/

bool memwrite(long offset, unsigned len, char far * buffer)
{
	bool rc;
	int retry_count = 0;

RETRY:
	rc = memrdwr(1,offset,len,buffer);
	if (!rc)
	{
		if(++retry_count < MAX_RETRIES)
		{
      	if(rdsk_debug)
          	bios_puts("\007read retry");
			goto RETRY;
		}
		memerr = DEVERR_WRITE;
	}
	return rc;
}

#if IOCTL
/*
	Allow an external program to quit REMSERV
*/
bool memioctl(unsigned category, char far * buffer)
{
	// 86 the REMSERV application (quit)?
	if(category==0x886 && *(unsigned char far *)buffer==0x86)
	{
		rem_packet('X',0L,0,0,0);
		return 1;
	}
	return 0;
}
#endif
