//
//   $Id: REMHOST.C 1.5 2002/10/16 21:58:28Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	REMHOST.C  created 02/05/90  by Drew Gislason

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	An application (on the SERVER side) may call this
	routine to emulate the remote disk.

	History:

	01-30-1999 bt  Renamed all the bpb_def to BPB so that it
						would be using the BPB structure in dostruct.h

**********************************************************/
#include <ctype.h>
#include <mem.h>
#include <dos.h>

#include "datalght.h"

#include "remote.h"
#include "remdisk.h"

#include "remhost.pt"
#include "absread.pt"
#include "remtext.pt"
#include "remserv.pt"
#include "remote.pt"
#include "print.pt"

bool PASCAL dlResetDisk( void );
bool PASCAL dlGetDriveDPB( uchar ucDriveNumber, LPDPB * pfpDPB );
bool PASCAL dlGetFreeSpaceOnDrive(int drive, ulong *pulFreeSpace);

uchar PacChanged = 1;                                   /* has the bpb changed? */
struct BPB PacBPB = { 512,2,1,2,112, 720,0xFD,2, 9,2 };

int     rdsk_debug;              /* 1=print debugging, 0=don't */
ifuncptr rdsk_charfunc;          /* routine called by remdisk_prepacket() */
int     rdisk;                   /* 1=A, 2=B, etc... */
int     rdsk_len;                /* length of the received packet */
uchar   rdsk_buff[1028];         /* just over 1 sector size */
uchar   diskbuff[512];

int _absread(int drive, int num, long start, void * buff);
int _abswrite(int drive, int num, long start, void * buff);
int bios_poll(void);

int maybe_changed;

/*
        remdisk_emulate

        Open the Remote Disk (host end) on the specified
        drive.  This routine returns ONLY if the target
        end tells it to terminate Remote Disk Emulation.

        mode = M_PACKET, M_CHAR

        Returns
                -1              invalid drive or cannot open port/mode or other error
                0               worked
*/
int remdisk_emulate(uchar drive,int mode,unsigned port,unsigned uiIrqNumber,
		unsigned baud)
{
        int retcode;
        int terminate=0;
        void reset_disk(uchar drive);
        extern char packet_err_str[];
        int rc;

        /* open the remote link */
        if(remote_open(mode,port,uiIrqNumber,baud,0))
                return -1;

        /* initialize the globals */
        PacChanged = 1;
        rdisk = drive;

        /* we must be in character mode for PREPACKET sequence */
        if(rem_mode == M_PACKET)
                comm_to_char();

#if IBM
        mouse_button_pressed();
#endif

        /* wait to receive commands */
        while(1)
        {
                while ((rc = remdisk_prepacket()) == -1) ;  // while disk err

                if(rc == -2)            /* user terminated */
                {
                        retcode = -1;
                        break;
                }

                /* now we are processing packets: do them until done */
                while(1)
                {
//              printf("pack ");

                        /* get a packet from the TARGET machine */
                        rdsk_len = remote_getpacket(rdsk_buff);

                        /* if we failed to get the packet, there is something wrong */
                        if(rdsk_len < 0)
                        {
                                if(rdsk_debug)
                                        printf("*");

//                              printf("*");
                                break;
                        }

                        /* process the packet */
                        retcode = remdisk_dopacket();

                        /* we are done with disk operations */
                        if(retcode)
                        {
                                /* the packet failed */
                                if(retcode<0)
                                {
                                        if(rdsk_debug)
                                                printf(packet_err_str,retcode);
                                }

                                /* we want to terminate, wait for the Done packet */
                                else if(retcode == ERR_TERM)
                                {
                                        terminate = 1;
                                        continue;
                                }

                                /* we got the done packet, acknowlege to target */
                                else
                                {
                                        if(rem_mode == M_PACKET)
                                        {
                                                comm_to_char();
                                                send_nul_packet();
                                        }

                                        else if(rem_mode == M_CHAR)
                                        {
                                                // send 2 byte length, nothing more
                                                remote_putc(0);
                                                remote_putc(0);
                                        }
                                }
                                break;
                        }

                } /* process packet loop */

                /* we are done doing Remote Disk I/O, give back control */
                if(terminate)
                {
                        retcode = 0;
                        break;
                }
        }

        /* close the remote link */
        remote_close();

        /* reset host drive */
        reset_disk(drive);

        return retcode;
}


void reset_disk(uchar drive)
{
        DPB far *dpb;         /* Drive Parameter Block */
        union REGS r;
        struct SREGS sr;
		  ulong ulFreeSpace;


        /* flush all buffers */
		  dlResetDisk();

        // get *DPB for the host's drive

		  dlGetDriveDPB( drive, &dpb );
		  dpb->dpbFirstAccess = 0xFF;


		  // access the drive
		  //
		  dlGetFreeSpaceOnDrive( drive, &ulFreeSpace );
}

#if IBM
int mouse_button_pressed(void)
{
        extern int mouse_startx;
        extern int mouse_stopx;
        extern int mouse_starty;
        extern int mouse_stopy;
        #define MOUSEINT 0x33
        int x,y;
        int count;

        // return # of mouse button presses
        _AX = 5;                                        // get # of mouse presses
        _BX = 0;                                        // left button
        geninterrupt(MOUSEINT);
        count = _BX;                    // how many button presses since last call?

        // maybe we need a range check here?
        x = _CX;
        y = _DX;

        if(count)
        {
                if(rdsk_debug)
                        printf("cnt %d, x %d, y %d\n",count,x,y);

                // x range
                if(mouse_stopx && (x < mouse_startx || x > mouse_stopx))
                        count = 0;

                // y range
                if(mouse_stopy && (y < mouse_starty || y > mouse_stopy))
                        count = 0;
        }

        return count;
}
#endif

/*
        remote_prepacket

        Wait for a signal from other end to enter into
        packet processing mode.  This routine will only be
        used on the HOST end of Remote Disk communications.
        This routine acknowleges to the TARGET that we are
        ready to receive packets.

        Returns
                -1              i/o failure
                >=0     worked, no problem
*/
int remdisk_prepacket(void)
{
        extern int rem_mode;
        vfuncptr charin;
        int c,len;
        ulong ticks = cl_start();

//      printf("pre ");

        comm_to_char();

        /* get the approriate character input routine */
        switch(rem_mode)
        {
                case M_PACKET:
                case M_CHAR  : charin = comm_poll; break;
        }

        /* wait forever for the prepacket character */
        while(1)
        {
                c = remote_tgetc(charin);

                /* other end signals packet i/o */
                if(c == PREPACKET)
                        break;

                if (cl_timer(ticks) >= (18*2) )
                        maybe_changed = 1;

                /* some other char came across, display it */
                if(c >= 0)
                {
                        if(rdsk_charfunc)
                                (*rdsk_charfunc)(c);
                }

                /* timeout, simply wait */
                else
                {
#if IBM
                        if(mouse_button_pressed())
                                return -2;
                        if(unusual_exit())
                                return -2;
#endif

                        if(bios_poll() > 0)
                                return -2;
                        if(rdsk_debug)
                                printf(".");
                }
        }

        /* if packet mode, point back to packet ISR */
        if(rem_mode == M_PACKET)
                comm_to_packet();

        /* tell other end we are ready */
        c = ~PREPACKET;
        len = remote_sendpacket((void *)(&c),1);

//              printf("pre 7 len= %d ",len);

        return len;
}


/*
        remdisk_dopacket

        A packet has been received.  Do whatever needs to be
        done with it.  Valid commands are:
                B               Return the BPB
                C               Change of disk
                D               Done with commands for now
                R               Read a sector
                W               Write a sector
                X               Stop emulating a remote disk

        Returns
                -1              failed
                0               target end told us we are Done with commands for now
                1               processed packet like a champ
                2               this was the Terminate packet.  Quit out.
*/
int remdisk_dopacket(void)
{
        uchar cmd;
        unsigned count = 0;
        unsigned char chg;
        int diskerr = 0;
        int send = 1;
        int exitcode = ERR_NONE;
        int len;
        extern char invalid_cmd_str[];

        /* process the command */
        cmd = PAC_CMD;

        /* user may want to know what's going on */
        if(rdsk_debug)
                printf("%c",cmd);

        switch(cmd)
        {
                /*** Read a sector from disk ***/
                case 'R':
                        diskerr = pac_read_sect();
                        send = 0;
                        break;


                /*** Write a sector to disk ***/
                case 'W':
                        diskerr = pac_write_sect();
                        send = 0;
                        break;

                /*** Detect if Disk has Changed ***/
                case 'C':
                        if(PacChanged)
                        {
                                pac_get_bpb();                          /* get the BPB off the disk */
                                chg = CHG_YES;                          /* definitely changed */
                                PacChanged = 0;
                                maybe_changed = 0;
                        }

                        else if (maybe_changed)
                        {
                                pac_get_bpb();
                                chg = CHG_MAYBE;
                                maybe_changed = 0;

                        }



/* NOTE: NO FLOPPY TIME TEST FOR NOW */
#if 0
                        /* 3 second timeout for floppies only */
                        else if(PacBPB.media >= 0xF9 &&  cl_timer(PacTicks) > 3*SECOND)
                        {
                                pac_get_bpb();
                                chg = CHG_MAYBE;
                        }
#endif
                        else
                                chg = CHG_NO;


/* NOTE: NO FLOPPY TIME TEST FOR NOW */
#if 0
                        PacTicks = cl_start();
#endif

                        PAC_CHGCODE = chg;
                        count = 2;
                        break;


                /*** Get the BPB (BIOS Paramter Block) ***/
                case 'B':
                        if(PacChanged)
                        {
                                pac_get_bpb();
                                PacChanged = 0;
                        }
                        memmove(PAC_BPB,&PacBPB,sizeof(struct BPB));


/* NOTE: NO FLOPPY TIME TEST FOR NOW */
#if 0
                        PacTicks = cl_start();
#endif
                        count = 1 + sizeof(struct BPB);
                        break;

                /*** Done Packet ***/
                case 'D':
                        count = 1;
                        exitcode = ERR_DONE;
                        break;

                /*** Exit Packet Mode ***/
                case 'X':
                        count = 1;
                        exitcode = ERR_TERM;
                        break;

                default:
                        printf(invalid_cmd_str,cmd,cmd);
                        return -1;
        }


        /* check for diskette read/write error */
        if(diskerr)
        {
                /* debugging */
                if(rdsk_debug)
                        printf("\ndiskerr %d\n",diskerr);

                count       = 3;
                PAC_SECTOR      = diskerr;
                PAC_CMD     = 'e';
        }


        /* send the response to this packet */
        if(send)
        {

                /* set up the packet command field */
                PAC_CMD   = tolower(PAC_CMD);

                if(exitcode != ERR_DONE)
                {
                        len = remote_sendpacket(rdsk_buff,count);
                        if(len < 0)
                                exitcode = ERR;
                }
        }

        /* exitcode may be: */
        return exitcode;
}



/*
        pac_read_sect

        Read a sector from the HOST disk.  Sector is sent in one big
        whack to the target.

        Returns 0 on success, error on failure.
*/
int pac_read_sect(void)
{
        int diskerr;
        int len;
        unsigned long sector;

        /* read the sector */
        sector = PAC_SECTOR;
        if(rdsk_debug>1)
                printf("sector %08lX ",sector);

        diskerr = _absread(rdisk-1,1,sector,rdsk_buff+1);

        if(rdsk_debug>1)
                printf("err %d, %02X %02X %02X\n",diskerr,rdsk_buff[1],rdsk_buff[2],rdsk_buff[3]);

        /* check for disk errors */
        if(diskerr)
        {
                if(rdsk_debug>1)
                        printf("\ndisk read error: sector %08lX, error %04X\n",sector,diskerr);

                if((diskerr & 0xFF00) == 0x8000)
                        diskerr = 2;                                    /* not ready */
                else
                        diskerr = 0xC;                                  /* general failure */
        }

        /* REMDISK.SYS wants to receive the sector */
        else
        {
                PAC_CMD   = 'r';
                len = remote_sendpacket(rdsk_buff,1 + PacBPB.sec_size);
                if(len < 0)
                        return -1;
        }

        return diskerr;
}


/*
        pac_write_sect

        Write a single sector to the HOST disk.  Sector
        is read in one big whack from the target.

        Returns 0 on success, error on failure.
*/
int pac_write_sect(void)
{
        unsigned long sector;
        int diskerr=0;
        int len;

        /* read and display the sector */
        sector = PAC_SECTOR;

        /* write it and check for errors */
        diskerr = _abswrite(rdisk-1,1,sector,rdsk_buff+5);
        if(diskerr)
        {
                if(rdsk_debug>1)
                        printf("\ndisk write error: sector %04X, error %04X\n",sector,diskerr);

                switch(diskerr & 0xFF00)
                {
                        case 0x8000: diskerr = 2; break;                /* not ready */
                        case 0x0300: diskerr = 0; break;                /* write protected */
                        default: diskerr = 0xC; break;          /* general failure */
                }
        }

        /* tell REMDISK.SYS that we are done writing sector */
        else
        {
                PAC_CMD   = 'W';
                len = remote_sendpacket(rdsk_buff,1);
                if(len < 0)
                        return -1;
        }

        return diskerr;
}


/*
        Get the BPB from disk and store it in our global
        BPB definition.
*/
int pac_get_bpb(void)
{
        /* a list of hard-coded BPBs */
        static struct BPB hard[] =
        {
                { 512,1,1,2,112, 360,0xFC,2, 9,1,0 },   /* FC  180K disk */
                { 512,2,1,2,112, 720,0xFD,2, 9,2,0 },   /* FD  360K disk */
                { 512,1,1,2, 64, 320,0xFE,1, 8,1,0 },   /* FE  160K disk */
                { 512,2,1,2,112, 640,0xFF,1, 8,2,0 }    /* FF  320K disk */
        };

        /*
                WARNING: MS-DOS gets confused if absread() is the 1st access
                ever... so do a findfirst to "initialize" disk access.
        */
        diskinit(rdisk-1);

        /* Read the first FAT sector off the disk */
        if(_absread(rdisk-1,1,1,diskbuff))
                return 0;

        /* check if we use hardcoded bpb or read bpb from the disk */
        if(diskbuff[0] >= 0xFE)
                memmove(&PacBPB,hard+diskbuff[0]-0xFC,sizeof(struct BPB));
        else
        {
                if(_absread(rdisk-1,1,0,diskbuff))
                        return 0;
                memmove(&PacBPB,diskbuff+11,sizeof(struct BPB));
        }
        return 0;
}



/*
        bios_poll

        Poll the BIOS keyboard interrupt for a key.

        Returns
                -1              no key waiting
                0-n     the key (no scan code)
*/
int bios_poll(void)
{
        #define ZF 0x40

        _AH = 1;
        geninterrupt(0x16);
        if(_FLAGS & ZF)
                return -1;
        _AH = 0;
        geninterrupt(0x16);
        _AH = 0;                        // ignore scan code
        return _AX;             // return the character
}

#if 0

#define isprint(c) ((c)>=' ' && (c)<='~')
#define min(a,b) ((a)<=(b) ? (a) : (b))

/*
        Dump a near peice up memory, up to 32K

        For debugging only.
*/
void dump(char *buff, int length)
{
        int start=0, len, i;
        unsigned char *buffer = (unsigned char *)buff;

        while(length>0)
        {
                len=min(length,16);
                for(i=0;i<16;++i)
                {
                        printf(i<len ? "%02X " : "   ",buffer[start+i]);
                        if(i==7)
                                printf(i<len ? "| " : "  ");
                }
                printf(" ");
                for(i=0;i<16;++i)
                {
                        if(i<len)
                                printf(isprint(buffer[start+i]) ? "%c" : ".",buffer[start+i]);
                        else
                                printf(" ");
                }
                printf("\n");
                length -= len;
                start  += len;
        }
}

#endif
