//
//   $Id: SUBST.C 1.9 2003/04/07 17:23:28Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	SUBST.C  created 07/17/95  by Drew Gislason

   Copyright (c) 1995-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Associates a path with a drive letter.

	SUBST [drive1: [drive2:]path]
	SUBST drive1: /D

	  drive1:        Specifies a virtual drive to which you want to assign a path.
	  [drive:2]path  Specifies a physical drive and path you want to assign to
	                 a virtual drive.
	  /D             Deletes a substituted (virtual) drive.

	Type SUBST with no parameters to display a list of current virtual drives.

	08-26-98 (BT)  Modified subst so that the path to be substituted
						could be reletive path.

	01-13-98 (DE)	Added ability to SUBST valid FAT drives.
						Removed string from DOSMSG.C that complains about
						SUBST of a valid (FAT) drive.

*********************************************************/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <dir.h>
#include <process.h>

#define SUBST 1
#include "dosver.h"
#include "dosmsg.c"
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "fatworld.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlapp.h"
#include "windows.h"

int ExpandPathNameString( char * pszRelativePath, char * pszExpandedPath );

bool fIsFat32World = FALSE;
bool fInWindows = FALSE;
bool fHideSignon = FALSE;
bool fDeleteSubstedDrive = FALSE;

/* status bits of the CDS */
#define C_NETWORK	0x8000
#define C_VALID	0x4000
#define C_JOIN		0x2000
#define C_SUBST	0x1000


void DisplaySignOnMessage( void )
{
	if( !fHideSignon )
	{
		if( fIsFat32World )
		{
			printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}

void RemoveWildCardName( char * szPath )
{
	unsigned uLength;

	uLength = strlen( szPath );
	if( szPath[uLength] == '*' )
	{
		szPath[uLength] = 0;
		uLength--;
		if( szPath[uLength] == '.' )
		{
			szPath[uLength] = 0;
			uLength--;
		}
		if( szPath[uLength] == '*' )
		{
			szPath[uLength] = 0;
		}
		if( szPath[uLength] == '\\' )
		{
			szPath[uLength] = 0;
		}
	}
}

int get_num_drives(void)
{
	LPLOL lol;
	unsigned dos_seg;

	// get pointer to DOS data area
	asm	mov	AH, 52H
	asm	int	21H
	asm	mov	dos_seg, ES
	lol = MK_FP(dos_seg, 0);

	return lol->lastdrive;
}

/*
	Get CDS for current drive
*/
LPCDS get_cds_ptr(int drive)
{
	unsigned dos_seg;
	LPLOL lol;
	LPCDS cds_ptr;

	// get pointer to DOS data area
	asm	mov	AH, 52H
	asm	int	21H
	asm	mov	dos_seg, ES
	lol = MK_FP(dos_seg, 0);

	// is the drive valid?
	if(drive < 0 || drive >= lol->lastdrive)
		return 0L;

	// determine pointer to the CDS
	cds_ptr = lol->first_cds;
	cds_ptr += drive;

	// return it
	return cds_ptr;
}

/*
	Get far ptr to DPB for specified drive (0=A:) (does NOT use the CDS)
	Returns 0L if no valid DPB exists for the drive.
*/
LPDPB lpDpbFromDrive(int drive)
{
	unsigned dos_seg;
	LPLOL lpLol;
	LPDPB lpDpb;

	// get pointer to DOS data area
	asm	mov	AH, 52H
	asm	int	21H
	asm	mov	dos_seg, ES
	lpLol = MK_FP(dos_seg, 0);

	// is the drive valid?
	if(drive < 0 || drive >= lpLol->lastdrive)
		return 0L;

	// determine pointer to the CDS
	lpDpb = lpLol->first_dpb;

	while(drive--)
	{
		lpDpb = lpDpb->next;
		if(((long)lpDpb == -1L) || ((long)lpDpb == 0L))
		{
			lpDpb = 0L;
			break;
		}
	}

	// return it
	return lpDpb;
}

/*
	We have a drive and a path.  SUBST them.
	Example:
	drive = 4 (drive E:)
	path = "B:\TEMP"
*/
int subst_drive(int drive, char path[])
{
	int org_drive;
	LPCDS cds_ptr;
	LPCDS org_cds_ptr;
	char org_cwd[LFN_MAX_PATH];

	// is drive valid?
	cds_ptr = get_cds_ptr(drive);
	if(!cds_ptr)
		return FALSE;

	// get original drive.  Is it valid?
	org_drive = toupper(*path) - 'A';
	org_cds_ptr = get_cds_ptr(org_drive);
	if(!org_cds_ptr)
		return FALSE;

	// don't point the same thing to itself
	if(cds_ptr == org_cds_ptr)
		return FALSE;

	// is dest drive OK?
	strcpy(org_cwd, "A:\\");
	*org_cwd = org_drive + 'A';
	if(!SmartGetCurrentDirectory(org_drive+1, org_cwd+3))
		return FALSE;

	// is dest path OK?
	if(!SmartChangeDirectory(path))
	{
		SmartChangeDirectory(org_cwd);
		return FALSE;
	}

   if( fInWindows )
   {
      if( !LFNSubstFunction( LFN_CREATE_SUBST, drive+1, path ) )
      {
         //
         // error handling done elsewhere?
         //
   		return FALSE;
      }
   }
   else
   {
   	// fill in cwd, status, dpb and skip fields
   	SmartExpandPath( 1, path, path );
   	RemoveWildCardName( path );

   	_fmemcpy(cds_ptr->cwd, path, _fstrlen(path)+1);
   	cds_ptr->status = org_cds_ptr->status | C_SUBST;
   	cds_ptr->lpDPB = org_cds_ptr->lpDPB;
   }

	if( fIsFat32World )
	{
		cds_ptr->ulSubDirStartingCluster =
				org_cds_ptr->ulSubDirStartingCluster;
	}
	else
	{
		cds_ptr->ulSubDirStartingCluster = 0xffff0000;
	}

	cds_ptr->skip = _fstrlen(cds_ptr->cwd);
	if(cds_ptr->skip == 3)
		--cds_ptr->skip;

	// restore original path
	SmartChangeDirectory(org_cwd);

	return TRUE;
}

/*
	Delete a drive that has been substituted.
*/
int delete_drive(int drive)
{
	unsigned status;
	LPDPB lpDpb;
	LPCDS cds_ptr;
	int iCurrentDrive;
	char szDirectory[LFN_MAX_PATH];


	// make sure drive exists
	cds_ptr = get_cds_ptr(drive);
	if(!cds_ptr)
		return FALSE;

   // don't delete unless it is already SUBST
   if( fInWindows )
   {
      if( LFNSubstFunction( LFN_QUERY_SUBST, drive+1, szDirectory ) )
      {
         if( !LFNSubstFunction( LFN_TERMINATE_SUBST, drive+1, szDirectory ) )
         {
            //
            // error handling done elsewhere?
            //
            return FALSE;
         }
      }
      else
      {
         return FALSE;
      }
   }
   else
   {
      if((cds_ptr->status & C_SUBST) == 0)
         return FALSE;

   	// If the drive being deleted has a DPB then make the status C_VALID.
   	// The CWD will revert to the root of the actual (vs. SUBST) drive.
   	//
   	status = 0;
   	lpDpb  = lpDpbFromDrive(drive);
   	if((long)lpDpb != 0L)
   	{
   		status = C_VALID;
   	}

   	//  Change the current drive and directory back to the drive
   	//  and directory that was substituted if the current drive is
   	//  the substed drive
   	//
   	_fstrcpy( szDirectory, cds_ptr->cwd );
   	dlGetCurrentDrive( &iCurrentDrive );
   	if( iCurrentDrive == drive )
   	{
   		dlSetCurrentDrive( (toupper(szDirectory[0]) - 'A') );
   		SmartChangeDirectory( szDirectory );
   	}

   	// reset CDS to normal
   	cds_ptr->status = status;
   	cds_ptr->lpDPB	 = lpDpb;
   	cds_ptr->skip 	 = 2;

   	if( fIsFat32World )
   	{
   		cds_ptr->ulSubDirStartingCluster = 0xFFFFFFFF;
   	}
   	else
   	{
   		cds_ptr->ulSubDirStartingCluster |= 0xFFFF;
   	}

      _fmemcpy(cds_ptr->cwd, "A:\\", 4);
      *cds_ptr->cwd = drive + 'A';
   }

	return TRUE;
}

/*
	Show all drives that have been substituted.
*/
void display_subst_drives(void)
{
	int i;
	int drives;
	int num_subst = 0;
	LPCDS cds_ptr;
	char szExpanded[LFN_MAX_PATH];

	drives = get_num_drives();
	for(i=0; i<drives; ++i)
	{
      if( fInWindows )
      {
         if( LFNSubstFunction( LFN_QUERY_SUBST, i+1, szExpanded ) )
         {
   			printf("%c: => %s\n",i+'A',szExpanded);
   			++num_subst;
         }
      }
      else
      {
   		cds_ptr = get_cds_ptr(i);
   		if( cds_ptr && (cds_ptr->status & C_SUBST) )
   		{
   			_fstrcpy( szExpanded, cds_ptr->cwd );

            // zero terminate the path at the SUBST point
            //
            szExpanded[cds_ptr->skip] = (char)0x00;

   			SmartExpandPath( 2, szExpanded, szExpanded );
   			RemoveWildCardName( szExpanded );
   			printf("%c: => %s\n",i+'A',szExpanded);
   			++num_subst;
   		}
      }
	}

	if(!num_subst)
		printf("No SUBST drives!\n");
}


//  Display the usage screen and exit
//
void Usage( void )
{
	DisplaySignOnMessage();
	printf(szUSAGE);
	exit(1);
}


int main(int argc, char *argv[])
{
	int drive;
	int org_drive;
	LPCDS cds_ptr;
	char szExpandedPath[LFN_MAX_PATH];
	int bSuccessful;
	int i;
	char cDriveLetter;
	unsigned uTemp;
	static ARGDEF SubstOpts[] =
	{
		{ BOOL_SWITCH, &fHideSignon, '[', 0 },
		{ BOOL_SWITCH, &fDeleteSubstedDrive, 'd', 0 },
		{ 0 }
	};


	//  Determine if we are in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	//  Determine if we are in windows
	//
	fInWindows = dlInWindows();

	//  Remove the quotes from the arguments
	//
	for(i=1; i <= argc; i++)
	{
		LFNStripArgQuotes( argv[i] );
		if( strlen(argv[i]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			return 255;
		}
	}

	//  Process all the boolean switches in the command line
	//
	SetParseFlags(PARSEF_NOARGSOK);
	argc = ParseCmdline(argc, &argv[0], SubstOpts);


	DisplaySignOnMessage();

	// just display drives?
	if(argc == 1)
	{
      if( fDeleteSubstedDrive )
      {
			printf( szErrInvalidParm,"/d" );
			return 1;
      }
      else
      {
   		display_subst_drives();
   		return 0;
      }
	}

	// make sure first arg is a drive
	if( argv[1][1] != ':' )
	{
		printf( szErrInvalidParm, argv[1] );
		return 1;
	}
	cDriveLetter = toupper(argv[1][0]);
	if( (cDriveLetter < 'A') || (cDriveLetter > 'Z') )
	{
		printf( szErrInvalidParm, argv[1] );
		return 1;
	}
	drive = cDriveLetter - 'A';

	cds_ptr = get_cds_ptr(drive);
	if(!argv[1][0] || argv[1][1] != ':' || !cds_ptr)
	{
		printf(err_invalid_drive);
		return 2;
	}

	// delete the drive?
	if( fDeleteSubstedDrive )
	{
		if(!delete_drive(drive))
		{
			printf(err_not_subst);
			return 3;
		}
	}

	// SUBSTitue the drive
	else
	{
		// The drive to be SUBSTituted can be either an invalid or a FAT drive
		//
		if(!((cds_ptr->status == 0) || (cds_ptr->status == C_VALID)))
		{
			printf(szAlreadySubsted);
			return 5;
		}

		// make sure 2nd arg starts with a drive
		//
//		bSuccessful = ExpandPathNameString( argv[2], szExpandedPath );
		bSuccessful = SmartExpandPath( 0, argv[2], szExpandedPath );
		if( !bSuccessful )
		{
	   	printf( szErrInvalidParm, argv[1] );
			return 1;
		}

		RemoveWildCardName( szExpandedPath );

		org_drive = toupper(szExpandedPath[0])-'A';
		cds_ptr   = get_cds_ptr(org_drive);
		if(!szExpandedPath[0] || szExpandedPath[1] != ':')
		{
			printf( err_bad_path,szExpandedPath );
			return 1;
		}

		// The drive whose path we are using for the substitution must be
		// a valid FAT drive.
		//
		if(((long)cds_ptr == 0L) || (cds_ptr->status != C_VALID))
		{
			printf( err_bad_path, szExpandedPath );
			return 1;
		}

		// do the substitute
		if(!subst_drive(drive, szExpandedPath))
		{
			printf( err_bad_path, szExpandedPath );
			return 1;
		}
	}

	return 0;
}
