;
;   $Id: BOOT.ASM 1.4 2002/10/16 22:01:36Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;
;
;   BOOT.ASM
;
;   Copyright (c) 1989-2002 Datalight, Inc.
;   All Rights Reserved.
;   Datalight Confidential and Proprietary
;
;   History
;
;   01/27/1999 bt  Original
;
;   The BS(Boot Sector) which is used for FAT12 and FAT16
;   formatted floppies and hard drives.  This boot sector
;   does support partitions which exist above the 8Gb boundary
;
;


;
;
; [ CONSTANTS ]
;

;
; All constants are in this file
;

include BOOT.INC


;
;
; [ MACROS ]
;

;
; All macros are in this file
;

include BOOT.MAC


;
;
;
;

_DATA	SEGMENT BYTE PUBLIC 'DATA'
		ASSUME	cs:_DATA, ds:_DATA, es:_DATA, ss:_DATA

		;
		; The boot sector program will take up 1024 bytes.  The
		; the first 512 bytes are written to the boot sector and
		; the remainder gets prepended to the IBMBIO.COM file.
		; We want to have any extra space filled with zeros.
		;

		org	0000h
		db	1024 dup(0)
		org	0000h

;
;
; Program entry point - skip over BPB info
;


	;
	;  This label is used by SYS.COM and FORMAT.COM to locate
	;  the start of the boot sector for FAT12 and FAT16 drives.
	;

	public	_rgucBootSector

	_rgucBootSector:
	BOOT_SECTOR_START:

		jmp	short REAL_START		; offset 00h
		nop					; offset 02h


;
;
; System name and BPB
;

	SYSTEM_NAME:
IFDEF ISRD7
		db	'DLDOS710'			; offset 03h
ELSE
		db	'DLDOS622'			; offset 03h
ENDIF
		;
		; BPB for a 1.44Mb floppy
		;

	BPB:
		bps	dw	0200h			; offset 0bh
		spc	db	01h			; offset 0dh
		res	dw	0001h			; offset 0eh
		nof	db	02h			; offset 10h
		rde	dw	00e0h			; offset 11h
		ts1	dw	0b40h			; offset 13h
		mid	db	0f0h			; offset 15h
		spf	dw	0009h			; offset 16h
		spt	dw	0012h			; offset 18h
		head	dw	0002h			; offset 1ah
		hid	dd	00000000h		; offset 1ch
		ts2	dd	00000000h		; offset 20h
		drv	db	00h			; offset 24h
		u1	db	00h			; offset 25h
		sid	db	29h			; offset 26h
		ser	dd	235607dbh		; offset 27h
		vl	db	'NO NAME    '		; offset 2bh
		fid	db	'FAT12   '		; offset 36h


;
;
; Set up the stack
;

		;
		; Actual program start point
		;

	REAL_START:

		;
		; Setup the stack
		;

		sti
		xor	cx, cx
		mov	ss, cx
		mov	sp, PROGRAM_OFFSET

		;
		; Set the base pointer for all the variables
		;

		mov	bp, sp


;
;
; Fixup the BPB heads and sectors per track
;

		; MSDOS fixes up the BPB at device initialization.
		; They do not update the heads and sectors per
		; track in the boot sector.  Because of this there
		; are some situations where MSDOS will not boot.
		;
		; We want to be able to boot all the time so we
		; fixup the heads and sectors per track using the
		; BIOS function 08h call but not for floppies.
		;

		PatchHeadsAndSectorsPerTrack


;
;
; Set up the DBT (Disk Base Table)
;

		;
		; Set up the segment registers
		;

		mov	es, cx

		;
		; Save the address of the old DBT.  Copy the DBT to
		; a new location.  Change the new DBTs sectors per
		; track to the value for this disk.  Change the new
		; DBTs idle time to a default value.
		;
		; ON ENTRY
		;
		; CX	= 0
		; SP	= Offset of the boot sector
		; SP	= Segment of the boot sector
		; ES    = 0
		;
		; ON EXIT
		;
		; AX   	= Preserved
		; BX    = DESTROYED
		; CX    = Preserved
		; DX    = Preserved
		; SI	= DESTROYED
		; DI	= DESTROYED
		; BP	= Offset of the boot sector
		; SP	= Preserved
	 	; SS	= Preserved
		; DS	= 0
		; ES	= Preserved
		;

		SetNewDBT


;
;
; Save a modified version of the partition ID
;

		;
		; Load the partition table (MBR), locate the partition which
		; this boot sector was loaded from, 'AND' off some bits from
		; the partition ID, then save the modified partition ID in
		; a variable.
		;
		; ON ENTRY
		;
		; CX    = 0
		; DS    = Segment of boot sector
		; BP	= Offset of boot sector
		;
		; ON EXIT
		;
		; SI	= High word of the number of hidden sectors
		; DI	= Low word of the number of hidden sectors
		; AX	= DESTROYED
		; BX    = DESTROYED
		; CX    = Preserved
		; DX	= Preserved
		; BP	= Preserved
		; DS	= Preserved
		; ES	= Preserved
		;

		SaveModifiedPartitionID


;
;
; Load the root directory
;

		;
		; Calculate the number of sectors before the FAT
		;
		; DX is still zero from above code
		;
		; variables already loaded:
		;
		; SI = word ptr [bp + (OFFSET_HID + 0)]
		; DI = word ptr [bp + (OFFSET_HID + 2)]
		;

	LOAD_ROOT_DIRECTORY:

		mov	ax, [bp + (OFFSET_RES + 0)]
		add	ax, si
		adc	dx, di

		;
		; Save the number of sectors before the FAT
		;

		mov	word ptr [bp + FAT_START + 0], ax
		mov	word ptr [bp + FAT_START + 2], dx

		;
		; Calculate the number of sectors occupied by FATs.
		;

		CalcFATSectors

		;
		; Add the number of reserved and hidden sectors
		;

		add	ax, word ptr [bp + FAT_START + 0]
		adc	dx, word ptr [bp + FAT_START + 2]

	       	;
		; Get the number of root directory entries to search
		;

		mov	cx, word ptr [bp + OFFSET_RDE]

	       	;
		; Save the number of root directory entries to search
		;

		push	cx

		;
		; Set the number of sectors used by the root directory
		;

 		shr	cx, 4

		;
		; Save the start of the root directory
		;

		push	ax
		push	dx

		;
		; Add the sector size of both fats.  This value is
		; the physical sector value of the first sector in
		; the data area.
		;

		add	ax, cx
		adc	dx, 0

		;
		; Save the physical sector number of the start of the
		; data area on the disk
		;

		mov	[bp + DATA_START + 0], ax
		mov	[bp + DATA_START + 2], dx

		;
		; Restore the start of the root directory
		;

		pop	dx
		pop	ax

		;
		; Set the buffer address
		;

		mov	bx, ROOT_DIR_BUFFER_OFFSET

		;
		; Preserve the buffer address
		;

		push	bx

		;
		; Read in the root directory
		;

		call	ReadSector

		;
		; Restore the buffer address into the source search
		; register
		;

		pop	si


;
;
; Search for the directory entry which contains the name of our
; second stage loader.
;

		;
		; Set the number of root directory entries to search
		;

		pop	bx

		;
		; Set the size of the search
		;

	REPEAT_SEARCH:

		add	si, cx
		mov	cl, 20h

		;
		; Set the buffer address as the other search string
		;

		mov	di, PROGRAM_OFFSET + (offset ROMDOS_FILENAME - offset BOOT_SECTOR_START )

		;
		; Perform the search
		;

		rep	cmpsb

		;
		; Check if we found the string
		;

		cmp	cl, 14h

		;
		; Jump if we found the string
		;

		jbe	LOAD_STAGE2_LOADER

		;
		; Decrement the search counter
		;

		dec	bx

		;
		; Jump if we are not done searching
		;

		jnz	REPEAT_SEARCH

		;
		; Display an error message saying that there were
		; no system files
		;
		; No jump necessary
		;

;
;
; Display no system on disk error message and reboot
;


	DISPLAY_NO_SYS_ON_DISK:

		;
		; Set the pointer to the 'No system on disk' string
		;

		mov	si, PROGRAM_OFFSET + ( offset MSG_NO_SYS_ON_DISK - offset BOOT_SECTOR_START )

IF 0
		;
		; Display the string
		;

		call	DisplayString
ELSE  ; IF 0
		; DisplayAsciizString
		;
		; Display a zero terminated string being pointed to by
		; DS:SI by successivly calling BIOS display character
		; functions.
		;
		; ON ENTRY
		;
		; DS:SI = Pointer to an ASCIIZ string
		; DS    = Segment of boot sector
		; BP	= Offset of boot sector
		;
		; ON EXIT
		;
		; AX	= DESTROYED
		; BX    = DESTROYED
		; CX    = Preserved
		; DX	= Preserved
		; SI	= DESTROYED
		; DI	= DESTROYED
		; BP	= Preserved
		; DS	= Preserved
		; ES	= Preserved
		;

		DisplayAsciizString
ENDIF ; IF 0
		;
		; Get a keypress, restore the DBT vector and
		; warmboot the computer.
		;
		; ON ENTRY
		;
		; BP    = Offset of the boot sector
		; SS    = Segment of the boot sector
		;
		; ON EXIT
		;
		; does not return
		;

		Restart


IF 0
;
;
; Display string function
;
; Display a zero terminated string one character at a
; time using BIOS display character functions.
;
; ON ENTRY
;
; DS:SI = Pointer to a zero terminated string
; DS    = Segment of boot sector
; BP	= Offset of boot sector
;
; ON EXIT
;
; AX	= DESTROYED
; BX    = DESTROYED
; CX    = Preserved
; DX	= Preserved
; SI	= DESTROYED
; DI	= DESTROYED
; BP	= Preserved
; DS	= Preserved
; ES	= Preserved
;

DisplayString	proc	near

		; Display a zero terminated string being pointed to by
		; DS:SI by successivly calling BIOS display character
		; functions.
		;
		; ON ENTRY
		;
		; DS:SI = Pointer to an ASCIIZ string
		; DS    = Segment of boot sector
		; BP	= Offset of boot sector
		;
		; ON EXIT
		;
		; AX	= DESTROYED
		; BX    = DESTROYED
		; CX    = Preserved
		; DX	= Preserved
		; SI	= DESTROYED
		; DI	= DESTROYED
		; BP	= Preserved
		; DS	= Preserved
		; ES	= Preserved
		;

		DisplayAsciizString

		;
		; We are done
		;

		ret

DisplayString	ENDP

ENDIF  ; IF 0

;
;
; Load the second stage loader
;

		;
		; Load the starting cluster number of the file
		;

	LOAD_STAGE2_LOADER:

		mov	di, word ptr [si + 0eh]

		;
		; Zero the sector counter
		;

		xor	cx, cx

		;
		; Set the destination address
		;

		mov	bx, OTHER_HALF_OF_LOADER_OFFSET

		;
		; Save the cluster number
		;

		push	di

		;
		; Determine the physical sector number of the cluster
		;

		ClusterToPhysicalSector

		;
		; Set the number of sectors to read
		;

		inc	cx

		;
		; Load the sector
		;

		call	ReadSector

		;
		; Restore the cluster number
		;

		pop	di


;
;
; Transfer control to the second stage loader
;

		;
		; Go load up romdos
		;

		jmp	LOAD_ROMDOS


;
;
; Read a number of sectors
;
; Read a number of sectors, one sector at a time, into
; a buffer.  This function does properly increment a
; segmented buffer address so that more than 64K worth
; of data can be loaded.
;
; ON ENTRY
;
; DX:AX = Starting physical sector number
; CX	= Number of sectors to read
; ES:BX = Buffer address
; DS    = Segment of boot sector
; BP	= Offset of boot sector
;
; ON EXIT
;
; DX:AX	= Starting physical sector number + CX
; ES:BX = Buffer address + (CX + 512)
; CX    = 0
; SI	= Preserved
; DI	= Preserved
; BP	= Preserved
; DS	= Preserved
;

ReadSector	PROC	NEAR

		; Read a number of sectors, one sector at a time, into
		; a buffer.  This function does properly increment a
		; segmented buffer address so that more than 64K worth
		; of data can be loaded.
		;
		; ON ENTRY
		;
		; DX:AX = Starting physical sector number
		; CX	= Number of sectors to read
		; ES:BX = Buffer address
		; DS    = Segment of boot sector
		; BP	= Offset of boot sector
		;
		; ON EXIT
		;
		; DX:AX	= Starting physical sector number + CX
		; ES:BX = Buffer address + (CX + 512)
		; CX    = 0
		; SI	= Preserved
		; DI	= Preserved
		; BP	= Preserved
		; DS	= Preserved
		;

		ReadXNumberOfSectors

ReadSector	ENDP


;
;
; System file names
;

		;
		; Second stage loader name (first sector is second
		; stage loader
		;

		;
		; Romdos startup code filename
		;
		; NOTE: the CR in the MSG_NO_SYS_ON_DISK string is
		;       used as a no match carracter when searching
		;       for the directory entry
		;

	ROMDOS_FILENAME:

		db	'IBMBIO  COM'
;		db	0ffh

;
;
; Error strings
;

	MSG_NO_SYS_ON_DISK:

		db	0Dh, 0Ah
		db	'Non-system disk or disk error'
		db	0Dh, 0Ah
		db	'Press a key to reboot'
		db	0Dh, 0Ah
		db	0

;
;
; Boot sector marker
;

		org	01feh

		; Boot sector marker
		;

		dw	0aa55h


;
;
; Start of the second stage loader
;
; [ VARIABLES ]
;
; These variables are not needed in the first part of the
; boot sector.  They are put here so that they don't waste
; space in the part of the boot sector which has a size
; constraint.
;

		org	0200h

		;
		; This is the address which romdos will be loaded
		;

	LOAD_ADDR label dword

		;
		; Offset to load ROM-DOS
		;

		LOAD_OFF  		dw	0		; offset 3eh

		;
		; Segment to load ROM-DOS
		;

		LOAD_SEG  		dw	DEF_LOADADDR	; offset 40h


		;
		; This is a variable used to determine the end of
		; a cluster chain.  This is set by the boot sector
		; at run-time for a 12-bit FAT
		;
		; 12-bit FAT = 0ff8h
		; 16-bit FAT = 0fff8h
		;

		EndOfClusterChainMarker		dw	FAT_16_BIT_END

		;
		; This is the number of cluster entries in a 12-bit
		; or a 16-bit FAT assuming a 512 byte sector.  This
		; is set by the boot sector at run-time for a 12-bit
		; FAT
		;
		; 12-bit_entries	equ	0400h (index into 3 sectors)
		; 16-bit_entries	equ	0300h (index into 3 sector)
		;

		FATClusterEntriesPerSector	dw 	FAT_16_BIT_ENTRIES

		;
		; This is the physical sector number of the the fat
		; which is currently loaded in the fat buffer
		;

		CurrentFatSector		dd	0


;
;
; Load ROMDOS into memory
;
; This part of the code will actually be contained in the
; file IBMBIO.COM.  The boot sector will search for the file
; IBMBIO.COM in the root directory.  It will then load the
; remainder of the boot sector program at 0000:7e00.  Then
; the boot program will end up here to load romdos.
;


		;
		; Load the starting cluster number of the file
		;
		; DI = Cluster number

	LOAD_ROMDOS:

		;
		; Set the destination address
		;

;		les	bx, dword ptr [bp + OFFSET_LOAD_ADDR]
		les	bx, dword ptr [bp + 200h]

		;
		; Save the cluster number
		;

	LOAD_NEXT_CLUSTER:

		push	di

		;
		; Determine the physical sector number for the cluster
		;

		ClusterToPhysicalSector

		;
		; Set the size of a cluster
		;

		mov	di, [bp + OFFSET_SPC]
		and	di, 0ffh

		;
		; Skip the first sector of romdos.  The first
		; sector is the second stage loader and has already
		; been loaded.
		;

		cmp	byte ptr [bp + OFFSET_BS_PART_LOADED], 00h
		jz	DONT_SKIP_SECTOR
		mov	byte ptr [bp + OFFSET_BS_PART_LOADED], 00h
		inc	ax
		adc	dx, 0
		jmp	short SKIP_FIRST_SECTOR

	DONT_SKIP_SECTOR:

		;
		; Set the number of sectors to read
		;

	READ_NEXT_SECTOR:

		inc	cx

		;
		; Load the sector
		;

		call	ReadSector

		;
		; Check for an overflow in the buffer offset
		;

		or	bx, bx

		;
		; Jump if there is not an overflow in the buffer offset
		;

		jnz	NO_BUFFER_OFFSET_OVERFLOW

		;
		; Adjust the segment register to account for the new
		; offset.
		;

		mov	si, es
		add	si, 1000h
		mov	es, si

	NO_BUFFER_OFFSET_OVERFLOW:

	SKIP_FIRST_SECTOR:

		;
		; Decrement the sector we just read from the sectors
		; per cluster counter.
		;

		dec	di

		;
		; Check if we read in the whole cluster
		;

		or	di, di

		;
		; Jump if we did not read in the whole cluster
		;

		jnz	READ_NEXT_SECTOR

		;
		; Restore the sector number
		;

		pop	di

		;
		; Get next cluster value
		;

		call	LookupNextCluster

		;
		; Check if we are at the end of a cluster chain.
		;

		cmp	di, word ptr [bp + (offset EndOfClusterChainMarker - offset BOOT_SECTOR_START)]

		;
		; Jump if we are not at the end of the file
		;

		jb	short LOAD_NEXT_CLUSTER

		;
		; We are done loading the file
		;


;
;
; Transfer control to ROMDOS
;

		;
		; Save the boot drive letter for romdos
		;
		; When ROM-DOS sees an FFFFH in the "segment" part
		; of INT 21H, it knows the boot drive is in the
		; "offset" part.
		;
		; Alternatly, romdos startup code could look directly
		; at the boot sector which is still at 0000:7c00 for
		; the bpb drive letter.
		;

		mov	dl, byte ptr [bp + OFFSET_DRV]
		mov	bx, INT21_VECTOR
		mov	word ptr [bx + 0], dx
		mov	word ptr [bx + 2], 0ffffh

		;
		; Load the old DBT pointer
		;

		mov	di, word ptr [bp + OLD_DBT_ADDRESS + 0]
		mov	es, word ptr [bp + OLD_DBT_ADDRESS + 2]

		;
		; Set the DBT vector with the old DBT pointer
		;

		mov	bx, DBT_VECTOR
		mov	word ptr [bx], di
		mov	word ptr [bx+2], es

		;
		; This is a far jump to the location ROMDOS was loaded
		;

	TRANSFER_CONTROL:

;		jmp	dword ptr [bp + OFFSET_LOAD_ADDR]
		jmp	dword ptr [bp + 200h]

;
;
; Lookup next cluster number
;

LookupNextCluster	PROC	NEAR

		;
		; Save the current buffer address
		;

		push	es
		push	bx

		;
		; Set the FAT buffer address
		;

		mov	bx, FAT_BUFFER_OFFSET
		mov	es, cx

		;
		; Check if this is a 16-bit FAT
		;
		cmp	word ptr [bp + OFFSET_TS1], cx
		jz	LOOKUP_16_BIT_FAT

		push	ax
		push	dx
		push	cx
		xor	dx, dx
		mov	ax, word ptr [bp + OFFSET_TS1]
		mov	cl, byte ptr [bp + OFFSET_SPC]
		div	cx
		cmp	ax, 0FF0h
		pop	cx
		pop	dx
		pop	ax
		ja	LOOKUP_16_BIT_FAT

	LOOKUP_12_BIT_FAT:

		;
		; Set the 12-bit FAT variables
		;

		mov	dx, FAT_12_BIT_END
		mov	[bp + (offset EndOfClusterChainMarker - offset BOOT_SECTOR_START)], dx
		mov	dx, FAT_12_BIT_ENTRIES
		mov	[bp + (offset FATClusterEntriesPerSector - offset BOOT_SECTOR_START)], dx

		;
		; Read in the FAT sector which contains the
		; cluster value we are looking for and set
		; that cluster value in SI
		;

		call	FatRead

		;
		; Get the word wich contains the next cluster number
		;

		mov	di, si
		shr	di, 1
		add	di, si
		mov	di, word ptr [bx + di]

		;
		; Check if this is an odd entry
		;

		test	si, 1

		;
		; Jump if this is an odd entry
		;

		jnz	ODD_FAT_ENTRY

		;
		; Set the next cluster value
		;

	EVEN_FAT_ENTRY:

		and	di, 0fffh

		;
		; We are done
		;

		jmp	short DONE_LOOKUP_FAT_12

	ODD_FAT_ENTRY:

		;
		; Set the next cluster value
		;

		shr	di, 4

		;
		; Set the next cluster number in DI
		;

	DONE_LOOKUP_FAT_12:

		;
		; Jump if we are not done loading
		;

		jmp	short DONE_LOOKUP

		;
		; Read in the FAT sector which contains the
		; cluster value we are looking for and set
		; that cluster value in SI
		;

	LOOKUP_16_BIT_FAT:

		call	FatRead

		;
		; Set the next cluster value
		;

		shl	si, 1
		mov	di, word ptr [bx+si]

		;
		; Restore the previous buffer address
		;

	DONE_LOOKUP:

		pop	bx
		pop	es

		;
		; We are done
		;

		ret

LookupNextCluster	ENDP


;
;
; Read in a specific FAT sector
;
; Read in the FAT sector which contains the cluster value we
; are looking for and set that cluster value in SI
;

FatRead		PROC	NEAR

		;
		; Calculate the FAT sector which contains the this
		; cluster number.
		;
		; We read in three fat sectors at a time so that we
		; can easily index a FAT entry in a 12-bit FAT.
		;

		xor	dx, dx
		mov	ax, di
		div	word ptr [bp + (offset FATClusterEntriesPerSector - offset BOOT_SECTOR_START)]
		push	dx
		mov	dx, 3
		mul	dx

		;
		; Add the number of sectors before the FAT
		;

		add	ax, word ptr [bp + (FAT_START + 0)]
		adc	dx, word ptr [bp + (FAT_START + 2)]

		;
		; Check if we have this sector loaded already
		;

		cmp	word ptr [bp + ((offset CurrentFatSector + 0) - offset BOOT_SECTOR_START)], ax
		jne	DO_READ
		cmp	word ptr [bp + ((offset CurrentFatSector + 2) - offset BOOT_SECTOR_START)], dx
		je	SKIP_READ

		;
		; Set the number of sectors to read
		;

	DO_READ:

		mov	cl, 3

		;
		; Update a variable with the fat sector number
		; that is currently loaded.
		;

		mov	word ptr [bp + ((offset CurrentFatSector + 0) - offset BOOT_SECTOR_START)], ax
		mov	word ptr [bp + ((offset CurrentFatSector + 2) - offset BOOT_SECTOR_START)], dx

		;
		; Load the FAT sector which contains the this
		; cluster number.
		;

		push	bx
		call	ReadSector
		pop	bx

		;
		; Get the cluster number index
		;

	SKIP_READ:

		pop	si

		;
		; Done
		;

		ret

FatRead		ENDP


;
;
; Boot sector marker
;

		org	03feh

		; Boot sector marker
		;

		dw	0aa55h


;
;
; End of the boot program
;

	;
	;  This label is used by SYS.COM and FORMAT.COM to locate
	;  the end of the boot sector.
	;

	public	_rgucBootSectorEnd
	_rgucBootSectorEnd:


_DATA	ENDS

		END
