;
;   $Id: BOOT32.INC 1.4 2002/10/16 22:01:36Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;
;
;   BOOT32.INC
;
;   Copyright (c) 1989-2002 Datalight, Inc.
;   All Rights Reserved.
;   Datalight Confidential and Proprietary
;
;   History
;
;   01/27/99 bt  Original
;
;   These are constants for the BOOT32.ASM file which is the boot sector
;   code for a FAT32 drive.
;
;


;
;
; [ CONSTANTS ]
;


;
; This is the address which ROMDOS will be loaded
;

DEF_LOADADDR	EQU	1000H	; 1000:0 (used to be 70:0)

;
; This is the offset that the boot sector gets loaded.  It can not be
; changed.  It is only here for clearity.
;

PROGRAM_OFFSET			EQU	7c00h

;
; This is the offset that the root directory will be loaded
;

ROOT_DIR_BUFFER_OFFSET		EQU	(PROGRAM_OFFSET - 0600h)

;
; This is the offset that the other half of the loader gets
; loaded at.  Do not change this offset without changing all
; the relative offsets in the code.
;

REMAINING_BOOT_SECTORS_BUFFER	EQU	(PROGRAM_OFFSET + 0200h)

;
; This is the offset that the FAT gets loaded
;

FAT_BUFFER_OFFSET		EQU	(PROGRAM_OFFSET - 0400h)

;
; This is a variable which contains the partition ID if the
; partition is a FAT32 partition which uses the extended BIOS
; function call interface for reads and writes.  Thus, it
; will be either 90h or 0ch.
;

OFFSET_PARTITION_ID		EQU	2

;
; This is the offset that the MBR gets loaded
;

PARTITION_BUFFER_OFFSET		EQU	(PROGRAM_OFFSET - 0400h)

;
; This is the size of a partition table entry
;

PARTITION_ENTRY_SIZE		EQU	10h

;
; This is the number of partition table entries that exist
; in the MBR.
;

NUMBER_OF_PARTITION_ENTRIES	EQU	4

;
; This is the offset of the first partition entry inside
; of the MBR.
;

OFFSET_PART_ENTRY		EQU	1beh

;
; This is the offset of the INT-1Eh vector.  The segment is zero.  This
; is the location of the DBT (Disk Base Table)
;

DBT_VECTOR			EQU	(1eh * 4)

;
; This is the size of the DBT (Disk Base Table)
;

DBT_SIZE			EQU	0bh

;
; This is the offset for the new location of the DBT (Disk Base Table).
; The segment is 0000h.
;

NEW_DBT_OFFSET			EQU	1000h

;
; This is the offset of the INT-1Eh vector.  The segment is zero.  This
; is the location of the DBT (Disk Base Table)
;

INT21_VECTOR			EQU	(21h * 4)

;
; We are using these constants because using the word
; offset caused the compiler to interprete the offsets
; as a word an not a byte.  This caused some indexed
; variable instructions to be longer than they needed
; to be.
;

;
; These are offsets into the BPB (BIOS Parameter Block) at the begining
; of the boot sector.
;

OFFSET_SPC			EQU	0dh
OFFSET_RES			EQU	0eh
OFFSET_NOF			EQU	10h
OFFSET_RDE			EQU	11h
OFFSET_TS1			EQU	13h
OFFSET_SPF			EQU	16h
OFFSET_SPT			EQU	18h
OFFSET_HEAD			EQU	1ah
OFFSET_HID			EQU	1ch
OFFSET_SPF32			EQU	24h
OFFSET_ROOTSTARTCLUSTER		EQU	2ch
OFFSET_DRV			EQU	40h
OFFSET_FID			EQU	52h

;
; These are offsets into a partition table entry
;

PARTITION_ID			EQU	4
PARTITION_HID			EQU	8

;
; This is the offset to the end of the cluster chain marker.
;

OFFSET_END_CLUSTER_CHAIN	EQU	5ah

;
; These are offsets to
;

FAT_ENTRIES_PER_SECTOR		EQU	OFFSET_END_CLUSTER_CHAIN + 4

;
; This is the offset for the ROMDOS load address variable
;

OFFSET_LOAD_ADDR		EQU	FAT_ENTRIES_PER_SECTOR + 4

;
; This is the absolute sector number of the the fat which is
; currently loaded.
;

OFFSET_CURRENT_FAT_SECTOR	EQU	OFFSET_LOAD_ADDR + 4

;
; We a using some code space for variable locations.  The
; code space was already used and will not be used again
; by the time we start writing variables there.
;

;
; This is the variable location to save the DBT address
;

OLD_DBT_ADDRESS			EQU	6eh

;
; This is the absolute sector value for the start of the data
; area on the disk.
;

DATA_START			EQU	72h

;
; This is the number of sectors we have left to read in from
; a cluster for the ReadSectorOfClusterChain function
;

SECTORS_LEFT_IN_CLUSTER		EQU	7ah

;
; This is the cluster number we are currently reading from
; in the ReadSectorOfClusterChain function
;

READING_CLUSTER			EQU	82h

;
; This is the absolute sector value for the start of the FAT.
;

FAT_START			EQU	86h
