//
//   $Id: NEWSYS.C 1.10 2003/07/03 14:11:10Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  NEWSYS.C  created 01/30/1999 by Brandon Thomas
//
//
//  Copyright (c) 1989-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//
//  USAGE
//
//
//  SYS <drive> {options}
//
//  Options
//  /C      Do not confirm before transfering system
//  /H      Do not hide system files
//  /[      Silent - don't display signon
//
//	 1) Determine if drive is valid.
//  2) Get BPB info from drive (verify we can read disk).
//  3) Check for enough free clusters for both ROM-DOS and
//     COMMAND.  ROM-DOS.SYS space must be contiguous.
//  4) Give user a chance to abort if an old DOS already exists
//  5) Write out the new ROM-DOS.SYS under a temporary name.
//     (RDTMPDOS.SYS)  This verifies that we can write to disk.
//  6) Write out the new command processor.
//  7) Delete the old DOS
//  8) Switch with entries at start of disk so IBMBIO.COM and
//     IBMDOS.COM always appear at start of disk
//  9) Update Boot Sector
//
//
//  HISTORY
//
//  06-15-2000 td
//  - Added LFN support
//  02-08-1999 BT
//  - Now identifies a FAT32 drive by the number of sectors per FAT
//    field being zero
//  02-08-1999 BT
//  - No longer Pass the pointer to the structure for sCLSwitch, now
//    pass the boolean switches individually.
//  02-08-1999 BT
//  - Made WriteBootSector determine the address of the FAT12/FAT16 or
//    FAT32 buffer.  Also made it determine the number sectors to write
//    for the boot code.
//  02-08-1999 BT
//  - Changed all the (*x).y type dereferencing to x->y dereferencing
//  02-08-1999 BT
//  - Changed all the hungarian type notation of "ui" to "u"
//  02-05-1999 BT
//  - Added a check to make sure that the boot sector code for
//    FAT12/FAT16 and FAT32 all had a boot sector marker in the
//    last two bytes of each sector.  Also checked to make sure
//    that the boot code was a multiple of 512 bytes.  This is
//    only done in debug mode.
//
//  01/27/99  BT
//  -  Rewrote sys.  Created a boot sector which only needs the
//     BPB updated and does not require that the sytem files be
//     contigious or that the system files be the first entries
//     in the root directory.  This sys program just reads the
//     BPB, and validates it, puts in in the new boot sector code.
//     Then copies the system files and writes a new boot sector.
//     This code supports SYSing a FAT32 drive however we are still
//     developing FAT32.  FAT32 support has been turned off.
//


//
//  [ CONSTANTS ]
//

//  Conditional compile switch for supporting FAT32 drives
//
#ifdef ENABLEF32
#define SUPPORT_FAT32_DRIVES		TRUE
#else
#define SUPPORT_FAT32_DRIVES		FALSE
#endif

//  This constant is used so that other source modules will
//  no that SYS is being compiled.
//
#define SYS									1

//
//  [ LIBRARIES ]
//

#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <mem.h>
#include <ctype.h>
#include <stdlib.h>
#include <dir.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "windows.h"

#include "dosver.h"
#include "dosmsg.c"
#include "newsys.h"
#include "fatworld.h"

//
//  [ PROTOTYPES ]
//

//  All function prototypes are created before compilation
//
#include "newsys.pt"
#include "syscomn.pt"

//
//  [ VARIABLES ]
//

bool fIsFat32World  = FALSE;
bool fLFNsSupported = FALSE;
bool fInWindows = FALSE;

//
//  [ EXTERNALS ]
//

//  These are the labels at the start of the code for a FAT12\FAT16
//  and a FAT32 boot sector.  This code was linked in as data.
//
extern uchar rgucBootSector[];
extern uchar rgucBootSector32[];

//  These are the labels at the end of the code for a FAT12\FAT16
//  and a FAT32 boot sector.  This code was linked in as data.
//
extern uchar rgucBootSectorEnd[];
extern uchar rgucBootSector32End[];

extern char * err_msg[];

extern bool fIsFat32Binary;

CLSwitchType sCLSwitch = { DISPLAY_SIGNON_MESSAGE, HIDE_SYSTEM_FILES,
		CONFIRM_OLD_SYSTEM_DELETE };

//  Program start
//
uchar main( int argc, char *argv[] )
{
	//  [ LOCALS ]
	//
	bool  bSysFAT32Drive = DONT_SYS_FAT32;
	uchar ucDestinationDriveLetter = INVALID_DRIVE;
	uchar ucBootDriveLetter = INVALID_DRIVE;
   int i;


	//
	//  [ CODE ]
	//


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
#if SUPPORT_FAT32_DRIVES
	fIsFat32World = dlIsFat32World();
#else // #if SUPPORT_FAT32_DRIVES
	fIsFat32World = FALSE;
#endif // #if SUPPORT_FAT32_DRIVES


	//  Determine if LFNs are support
	//
	fLFNsSupported = LFNPresent();


	//  Determine if Windows is running
	//
	fInWindows = dlInWindows();


	//  Cleanup the quotes for each argv[]
	//
	for(i=1; i <= argc; i++)
	{
		LFNStripArgQuotes( argv[i] );
		if( strlen(argv[i]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(255);
		}
	}

	//  Check command line options
	//
	ParseCommandLineOptions( argc, argv, &ucDestinationDriveLetter,
			&sCLSwitch );

	//  Display signon message
	//
	DisplaySignOn( sCLSwitch.bShowSignOn );


	//  To keep from sys'ng a Windows development machine unintentionally ...
	//
	if( fInWindows )
	{
		//  Get the boot drive
		//
		dlGetBootDriveNumber( &ucBootDriveLetter );
		ucBootDriveLetter += 'A' - 1;


		//  Determine if the boot drive is the same as the destination
		//
		if( ucBootDriveLetter == ucDestinationDriveLetter )
		{
			printf( err_msg[ERR_SYSWIN] );
			exit(255);
		}
	}


	//  BT 2-5-1999
	//
	//  Added a check to make sure that the boot sector code for
	//  FAT12/FAT16 and FAT32 all had a boot sector marker in the
	//  last two bytes of each sector.  Also checked to make sure
	//  that the boot code was a multiple of 512 bytes.  This is
	//  only done in debug mode.
	//
#if DEBUG
	//  Make sure the boot sector code is valid
	//
	ValidateBootSectorCode();
#endif

	//  Set the BPB in the boot sector in memory
	//
	GetBPB( ucDestinationDriveLetter, &bSysFAT32Drive );

	//  Read in the COMMAND.COM and ROM-DOS.SYS files.  We are also
	//  passing in the name and path for SYS.COM so that we can
	//  look for the system files at the same location.
	//
	GetSystemFiles( argv[0] );

	//  Make sure we are sysing a FAT32 drive with a FAT32 kernel
	//
	if( bSysFAT32Drive && (!fIsFat32Binary) )
	{
		printf(szErrorNotFat32Kernel);
		exit(255);
	}

	//  Copy the system files to the destination drive
	//
	CopySystemFiles( ucDestinationDriveLetter, sCLSwitch.bConfirmDelete,
			sCLSwitch.bHideSystemFiles, bSysFAT32Drive );

   //  Free the memory allocated by GetSystemFiles()
   //
   FreeAllMemory();

	//  Write the boot sector to the destination drive
	//
	WriteBootSector( ucDestinationDriveLetter, bSysFAT32Drive );

	//  Exit with a NO ERROR error code
	//
	printf( err_msg[NO_ERROR] );
	return NO_ERROR;
}


//  BT 2-5-1999
//
//  Added a check to make sure that the boot sector code for
//  FAT12/FAT16 and FAT32 all had a boot sector marker in the
//  last two bytes of each sector.  Also checked to make sure
//  that the boot code was a multiple of 512 bytes.  This is
//  only done in debug mode.
//
#if DEBUG
//  Make sure the boot sector code is valid
//
void ValidateBootSectorCode( void )
{
	//  [LOCALS]
	//
	unsigned uBootSectorSize;
	unsigned uBootSectorMarkerIndex;


	//
	//  [CODE]
	//

	//  Check the FAT12/FAT16 boot sector size
	//
	uBootSectorSize = (unsigned)rgucBootSectorEnd - (unsigned)rgucBootSector;
	if( uBootSectorSize % SECTOR_SIZE )
	{
		printf("Boot sector code is not a multiple of 512 bytes.");
		exit(255);
	}

	//  Check the FAT12/FAT16 boot sector code for boot sector markers
	//  in each sector
	//
	uBootSectorMarkerIndex = SECTOR_SIZE - 2;
	while( uBootSectorMarkerIndex < uBootSectorSize )
	{
		if( (*(unsigned *)(&rgucBootSector[uBootSectorMarkerIndex])) !=
				BOOT_SECTOR_MARKER )
		{
 			printf("Boot sector code does not contain a boot sector marker\n");
			printf("in every sector of the boot code.");
			exit(255);
		}
		uBootSectorMarkerIndex += SECTOR_SIZE;
	}

	//  Check the FAT32 boot sector size
	//
	uBootSectorSize = (unsigned)rgucBootSector32End - (unsigned)rgucBootSector32;
	if( uBootSectorSize % SECTOR_SIZE )
	{
		printf("Boot sector code is not a multiple of 512 bytes.");
		exit(255);
	}

	//  Check the FAT32 boot sector code for boot sector markers
	//  in each sector
	//
	uBootSectorMarkerIndex = SECTOR_SIZE - 2;
	while( uBootSectorMarkerIndex < uBootSectorSize )
	{
		if( (*(unsigned *)(&rgucBootSector32[uBootSectorMarkerIndex])) !=
				BOOT_SECTOR_MARKER )
		{
 			printf("Boot sector code does not contain a boot sector marker\n");
			printf("in every sector of the boot code.");
			exit(255);
		}
		uBootSectorMarkerIndex += SECTOR_SIZE;
	}
}
#endif  //#if DEBUG

//  Set the BPB in the boot sector in memory
//
void GetBPB( uchar ucDestinationDriveLetter, bool * pbSysFAT32Drive )
{
	//  [ LOCALS ]
	//
	void far * fpBuffer;
	bool  bFunctionSuccessful;
//	unsigned uDriveNumber;
	unsigned uMemorySegment;
	bool  bIsValid;


	//
	//  [ CODE ]
	//

	//  Allocate some memory so that we can read in the boot
	//  sector.
	//
	bFunctionSuccessful = dlAllocMem(
			(unsigned)(MAX_BOOT_SECTOR_SIZE/PARAGRAPH_SIZE), &uMemorySegment );
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		ErrorExit( ERR_NOMEMORY, NULL, NULL );
	}

	//  Create a pointer to our buffer
	//
	fpBuffer = MK_FP( uMemorySegment, INITIAL_BUFFER_OFFSET );

	//  Read in the boot sector (BPB) of the destination drive
	//
	ReadBootSector( ucDestinationDriveLetter, fpBuffer,
			MAX_NUMBER_OF_SECTORS_IN_BOOT_SECTOR );

#if USING_FAT32

	//  In case this is a FAT12 or FAT16 drive, save the BPB
	//  in the FAT12/FAT16 boot sector code.
	//
	_fmemcpy( &rgucBootSector[BPB_START_IN_BOOT_SECTOR],
			MK_FP( uMemorySegment, BPB_START_IN_BOOT_SECTOR),
			(sizeof(struct BootSectorBPB16) + 26));

#else

	//  In case this is a FAT12 or FAT16 drive, save the BPB
	//  in the FAT12/FAT16 boot sector code.
	//
	_fmemcpy( &rgucBootSector[BPB_START_IN_BOOT_SECTOR],
			MK_FP( uMemorySegment, BPB_START_IN_BOOT_SECTOR),
			(sizeof(BPB) + 26));

#endif

	//  In case this is a FAT32drive, save the BPB in the
	//  FAT32 boot sector code as well as the BIG FAT INFO
	//  sector in the FAT32 code
	//
	_fmemcpy( &rgucBootSector32[BPB_START_IN_BOOT_SECTOR],
			MK_FP( uMemorySegment, BPB_START_IN_BOOT_SECTOR),
			0x4e);

	_fmemcpy( &rgucBootSector32[512],
			MK_FP( uMemorySegment, 512),
			512);

	//  Release the memory we allocated to read in the boot sector
	//
	bFunctionSuccessful = dlFreeMem( uMemorySegment );
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		dlFreeMem( uMemorySegment );
		ErrorExit( ERR_RELEASE, NULL, NULL );
	}

	//  Validate the BPB we read in
	//
	bIsValid = ValidateBPB( pbSysFAT32Drive );
	if( !bIsValid )
	{
		//  Invalid BPB
		//
		NoErrorFreeAllMemory();
		ErrorExit( ERR_BADBPB, NULL, ucDestinationDriveLetter );
	}
}


//  Make sure we have a valid BPB
//
//  02-08-1999 BT
//  - Now identifies a FAT32 drive by the number of sectors per FAT
//    field being zero
//
#pragma argsused
bool ValidateBPB( bool * pbSysFAT32Drive )
{
	//  [ LOCALS ]
	//
	BPB * psBPB;
#if USING_FAT32
	BPB * psBPB32;
#endif
	uchar ucMedia;


	//
	//  [ CODE ]
	//

	//  Set the pointers to the BPBs
	//
	psBPB   = (void *)&rgucBootSector[0x0b];
#if USING_FAT32
	psBPB32 = (void *)&rgucBootSector32[0x0b];
#endif

	//  Make sure the media ID is valid
	//
	ucMedia = psBPB->media & 0xF0;
	if( ucMedia != 0xF0)
	{
		//  Invalid BPB
		//
		return FALSE;
	}

	//  Make sure the sector size is valid
	//
	if(( psBPB->sec_size % 128 ) || (!psBPB->sec_size))
	{
		//  Invalid BPB
		//
		return FALSE;
	}

	//  We wrote this to sys a disk with a 512 byte sectors. It might
	//  work with 1024 byte sectors but it definitly won't work with
	//  128 byte sectors.
	//
	if( psBPB->sec_size == 0x128 )
	{
		ErrorExit( ERR_SECSIZE, NULL, NULL );
	}

	//  Check the total sectors
	//
	if(( psBPB->num_sec == 0 ) && ( psBPB->lnumsec == 0 ) )
	{
		//  Invalid BPB
		//
		return FALSE;
	}

#if USING_FAT32
	//  Check if this is a FAT32 BPB
	//
	if( psBPB->uFatSectors == 0 )
	{
		//
		//  BPB is a FAT32 BPB.  Check to make sure that the
		//  BPB contains some value for the starting cluster
		//  of the root directory.
		//

		if( psBPB32->ulRootDirStartCluster < 2 )
		{
			return FALSE;
		}

		//  There is no longer a limit to the number of root directory
		//  entries you can have.  In a FAT32 BPB this field is unused
		//  and should be zero.
		//
		if( psBPB32->uDirEntries != 0 )
		{
			return FALSE;
		}

		//  Set the flag to indicate we are sysing a FAT32 partition
		//
		*pbSysFAT32Drive = TRUE;
	}
#else // #if USING_FAT32
	//  Check if this is a FAT32 BPB
	//
	if( psBPB->fat_size == 0 )
	{
		//  If we are not implementing FAT32 support then we
		//  should display an error message saying we don't
		//  support FAT32 drives.
		//
		ErrorExit( ERR_FAT_TYPE, NULL, NULL );
	}
#endif // #if USING_FAT32

	//  Seems to be a valid BPB
	//
	return TRUE;
}

//  Return TRUE if this is a FAT32 drive
//
bool IsThisAFAT32Drive( uchar ucDriveNumber )
{
	LPDPB lpDPB;
	bool bSuccessful;


	bSuccessful = dlGetDPBForDrive( ucDriveNumber, &lpDPB );
	if( bSuccessful )
	{
#if SUPPORT_FAT32_DRIVES
		return ((lpDPB->fat_len) == 0);
#else // #if SUPPORT_FAT32_DRIVES
		if((lpDPB->fat_len) == 0)
		{
			bSuccessful = FALSE;
		}
		else
		{
			return FALSE;
		}
#endif // #if SUPPORT_FAT32_DRIVES
	}
	if( bSuccessful == FALSE )
	{
		//  User gave an invalid drive letter
		//
		DisplayUsage();
		ErrorExit( ERR_BADDRIVE, NULL, (ucDriveNumber+'A') );
	}

	return FALSE;
}

//  Read in the boot sector using the INT-25 when there is no FAT32
//  function support.
//
void ReadBootSector( uchar ucDriveLetter, void far * fpBuffer,
		unsigned uNumberOfSectors )
{
	//  [ LOCALS ]
	//


	//
	//  [ CODE ]
	//

	//  Determine if there is MSDOS 7.1 function support
	//
	if( IsFAT32FunctionSupport( ucDriveLetter ) )
	{
#if USING_FAT32
		//  Use the FAT32 supported function
		//
		ReadBootSector32( ucDriveLetter, fpBuffer, uNumberOfSectors );
#else
		//  Use the INT-25 a function which does not support
		//  FAT32 drives.
		//
		ReadBootSector16( ucDriveLetter, fpBuffer, uNumberOfSectors );
#endif
	}
	else
	{
		//  Use the INT-25 a function which does not support
		//  FAT32 drives
		//
		ReadBootSector16( ucDriveLetter, fpBuffer, uNumberOfSectors );
	}
}

//  We read in the boot sector using the INT-25 abs read.
//
void ReadBootSector16( uchar ucDriveLetter, void far * fpBuffer,
		unsigned uNumberOfSectors )
{
	//  [ LOCALS ]
	//
	int  iDriveNumber;
	unsigned uErrorCode;
	bool  bFunctionSuccessful;
	char * szDriveString = "A";


	//
	//  [ CODE ]
	//

	//  Set the drive letter in a string form in case we have to
	//  display an error for this drive.
	//
	szDriveString[0] = (char)ucDriveLetter;

	//  Determine the drive number based on a drive letter.  The drive
	//  number for the abs read function starts with 0.
	//
	iDriveNumber = ucDriveLetter - 'A';

	//  Read in the boot sector
	//
	bFunctionSuccessful = dlAbsRead( TRUE, iDriveNumber, uNumberOfSectors,
			STARTING_SECTOR_OF_BOOT_SECTOR, fpBuffer );
	if( !bFunctionSuccessful )
	{
		//  Get the error code
		//
		uErrorCode = __dlCheckAbsRdWrError();

		//  Determine how to handle this error
		//
		switch( uErrorCode )
		{
			case ABS_ERROR_UNKNOWN_UNIT:
 				ErrorExit( ERR_BADDRIVE, szDriveString, NULL );

			case ABS_ERROR_DRIVE_NOT_READY:
				ErrorExit( ERR_DRVNOTREADY, szDriveString, NULL );

			case ABS_ERROR_INVALID_CHANGE:
				ErrorExit( ERR_CHANGE, NULL, NULL );

			default:
				ErrorExit( ERR_DRVSPECS, NULL, NULL );
		}
	}
}

#if USING_FAT32

//  Read in the boot sector using the INT-21 function 7305h which
//  is an MSDOS 7.2 function.
//
void ReadBootSector32( uchar ucDriveLetter, void far * fpBuffer,
		unsigned uNumberOfSectors )
{
	//  [ LOCALS ]
	//
	int  iDriveNumber;
	unsigned uErrorCode;
	bool  bFunctionSuccessful;
	char * szDriveString = "A";


	//
	//  [ CODE ]
	//

	//  Set the drive letter in a string form in case we have to
	//  display an error for this drive.
	//
	szDriveString[0] = (char)ucDriveLetter;

	//  Determine the drive number based on a drive letter.  The drive
	//  number for the abs read function starts with 0.
	//
	iDriveNumber = ucDriveLetter - 'A';

	//  Perform an abs read using the WIN95 interface for
	//  FAT32 drives
	//
	bFunctionSuccessful = dlAbsRead32( (iDriveNumber + 1),
			uNumberOfSectors, STARTING_SECTOR_OF_BOOT_SECTOR,
			fpBuffer );

	//  Check if there was an error
	//
	if( !bFunctionSuccessful )
	{
		//  Get the error code for the function
		//
		uErrorCode = __dlCheckAbsRdWrError();

		uErrorCode &= 0xff;
		switch( uErrorCode )
		{
			case ABS_ERROR_UNKNOWN_UNIT:
 				ErrorExit( ERR_BADDRIVE, szDriveString, NULL );

			case ABS_ERROR_DRIVE_NOT_READY:
				ErrorExit( ERR_DRVNOTREADY, szDriveString, NULL );

			case ABS_ERROR_INVALID_CHANGE:
				ErrorExit( ERR_CHANGE, NULL, NULL );

			default:
				ErrorExit( ERR_DRVSPECS, NULL, NULL );
		}
	}
}

#endif // #if USING_FAT32

//  Check command line options
//
void ParseCommandLineOptions( int iArgumentCount,
		char * pszArgumentString[], uchar * pucDestinationDriveLetter,
		CLSwitchType * psCLSwitch )
{
	//  [ LOCALS ]
	//
	char *   szCurrentCharacter;
	unsigned uIndex;
	uchar    ucSwitchIndicator;
	bool     bFunctionSuccessful;
	bool     bIsRemote;
	uchar    ucDriveNumber;
#if USING_FAT32
#else
	bool		bIsAFAT32Drive;
#endif
//	bool		bSuccessful;


	//
	//  [ CODE ]
	//

	//  Determine which switch is being used for parmeters.  This
	//  is for international support.
	//
	bFunctionSuccessful = dlGetSwitchCharacter( &ucSwitchIndicator );
	if( !bFunctionSuccessful )
	{
		ucSwitchIndicator = '/';
	}

	//  Check if there are any command line parameters
	//
	if( iArgumentCount < 2)
	{
		//  Display the usage screen
		//
		DisplayUsage();

		//  Exit with an error code for no command line parameters
		//
		ErrorExit( ERR_NO_PARM, NULL, NULL );
	}

	//  Process each command line string checking for multiple
	//  parameters within each string.
	//
	for( uIndex=1; uIndex < iArgumentCount; ++uIndex )
	{

		//  Set a character pointer to the beginning of the
		//  string.
		//
		szCurrentCharacter = pszArgumentString[uIndex];

		//
		//
		while( szCurrentCharacter[0] != CHAR_NULL )
		{
			//  Find a character other that a space or tab
			//
			while(( szCurrentCharacter[0] == CHAR_SPACE ) ||
					( szCurrentCharacter[0] == CHAR_TAB   ))
			{
				szCurrentCharacter++;
			}

			//  Determine if this is a command line switch or a drive
			//  letter.
			//
			if( szCurrentCharacter[0] == ucSwitchIndicator )
			{
				//  Change the character to upper case
				//
				szCurrentCharacter[1] = toupper( szCurrentCharacter[1] );

				//  Set the appropriate boolean switch
				//
				switch( szCurrentCharacter[1] )
				{

					//  User does not want us to hide the system files
					//
					case 'H':
						psCLSwitch->bHideSystemFiles = FALSE;
						break;

					//  User does not want to confirm in order to overwrite
					//  system files.
					//
					case 'C':
						psCLSwitch->bConfirmDelete = FALSE;
						break;

					//  User does not want to see signon message
					//
					case '[':
						psCLSwitch->bShowSignOn = FALSE;
						break;

					//  User is requesting the usage screen
					//
					case '?':
						DisplayUsage();
						exit( NO_ERROR );

					//  User entered a bad command line switch
					//
					default:
						printf( bad_opt, szCurrentCharacter[1] );
						DisplayUsage();
						exit( 255 );
				}
			}
			else
			{
				//  This should be the drive letter
				//
				if( szCurrentCharacter[1] != CHAR_COLON )
				{
					//  User has entered unknown stuff on the
					//  command line.
					//
					printf( bad_opt, szCurrentCharacter[0] );
					DisplayUsage();
					exit( 255 );
				}

				//  Do we already have a drive letter
				//
				if( *pucDestinationDriveLetter != INVALID_DRIVE )
				{
					//  User gave multiple drive letters to sys
					//
					DisplayUsage();
					ErrorExit( ERR_MULTIPLE, NULL, NULL );
				}

				//  Make this the drive letter
				//
				*pucDestinationDriveLetter = toupper( szCurrentCharacter[0] );
			}

			//  Go to the next parameter
			//
			szCurrentCharacter += 2;
		}
	}

	//  Was a drive letter present on the command line?
	//
	if( *pucDestinationDriveLetter == INVALID_DRIVE )
	{
		//  User did not give a drive letter
		//
		DisplayUsage();
		ErrorExit( ERR_NO_DRIVE, NULL, NULL );
	}

	//  Determine if this is a valid drive letter
	//
	if(( *pucDestinationDriveLetter < 'A' ) || ( *pucDestinationDriveLetter > 'Z' ))
	{
		//  User gave an invalid drive letter
		//
		DisplayUsage();
		ErrorExit( ERR_BADDRIVE, NULL, *pucDestinationDriveLetter );
	}

	//  Determine if this is a network drive
	//
	ucDriveNumber = *pucDestinationDriveLetter - 'A';
	bFunctionSuccessful = dlIsDriveRemote( ucDriveNumber+1, &bIsRemote );
	if( !bFunctionSuccessful )
	{
		//  User gave an invalid drive letter
		//
		ErrorExit( ERR_BADDRIVE, NULL, *pucDestinationDriveLetter );
	}

	if( bIsRemote )
	{
		//  User gave an drive letter of a network drive
		//
		ErrorExit( ERR_NETWORK, NULL, *pucDestinationDriveLetter );
	}

#if USING_FAT32
#else
	bIsAFAT32Drive = IsThisAFAT32Drive( ucDriveNumber );
	if( bIsAFAT32Drive )
	{
		ErrorExit( ERR_FAT_TYPE, NULL, NULL );
	}
#endif
}


//  Display the program usage screen
//
void DisplayUsage( void )
{
	//  [ LOCALS ]
	//


	//
	//  [ CODE ]
	//

	//  Display the program usage string
	//
	DisplaySignOn( sCLSwitch.bShowSignOn );
	printf(USAGE);
}


//  Display signon message
//
void DisplaySignOn( bool bShowSignOn )
{
	//  [ LOCALS ]
	//
	char * szSignOnMessage = "SYS %s %s\n%s\n\n";


	//
	//  [ CODE ]
	//

	//  Display the program signon message if we are
	//  supposed to.
	//
	if( bShowSignOn )
	{
		if( fIsFat32World )
		{
			printf( szSignOnMessage,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf( szSignOnMessage,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}
