//
//   $Id: NEWSYS.H 1.6 2002/10/16 22:02:52Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  NEWSYS.H  created 02/02/1999 by Brandon Thomas
//
//
//  Copyright (c) 1989-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//
//  This is the header file for the new version SYS.COM
//
//
//  HISTORY
//
//  02-05-1999 bt  Added a constant for the two bytes at the end of a
//                 boot sector (BOOT_SECTOR_MARKER)
//

#ifdef SYS

// ------------ <<< ERROR MESSAGES >>> --------------
//
//  These are indexes into the message array
//
#define ERR_NONE		0		/* system transfered */
#define ERR_WRITEPROTECT 1 // Write protect error
#define ERR_NO_PARM  2     // No command line parameters on command line
#define ERR_ABORT		3		/* user abort */
#define ERR_NOMEMORY	4 		/* not enough memory */
#define ERR_RESIZE   5     // Error resizing memory
#define ERR_RELEASE  6     // Error releasing memory
#define ERR_NO_DRIVE 7     // No drive letter specified
#define ERR_OPEN     8     // Error opening file
#define ERR_READ		9  	/* generic read error */
#define ERR_WRITE		10 	/* write error */
#define ERR_CLOSING  11    // Error closing file
#define ERR_DATETIME 12    // Error getting file date and time
#define ERR_BADDRIVE	13 	/* invalid drive, like N: */
#define ERR_CREATING 14    // Error deleting previous system files
#define ERR_DRVNOTREADY	15	/* drive door is open */
#define ERR_DELETING 16    // Error deleting previous system files
#define ERR_LOCK		17		// Error executing LOCK command
#define ERR_UNLOCK	18		// Error executing UNLOCK command
#define ERR_NOROOM	19 	/* not enough room for ROM-DOS or COMMAND */
#define ERR_BADBPB	20 	/* disk contains invalid BPB */
#define ERR_NODIR		21 	/* need 2 free entries in the root dir */
#define ERR_NETWORK	22 	/* can't SYS a networked drive */
#define ERR_NOBOOT	23		/* cannot create boot sector */
#define ERR_DRVSPECS 24		/* could not retrieve the drive specifications */
#define ERR_NOCMD		25		// no command.com in the current directory
#define ERR_CHANGE   26    // invalid change of disk
#define ERR_MULTIPLE	27    // Multiple drive letters on command line
#define ERR_FAT_TYPE	28    // only support FAT 12 or FAT 16 drives
#define ERR_SECSIZE  29    // Not 512 byte sectors
#define ERR_SYSWIN   30		// Don't sys a windows boot drive while in windows

#endif // #ifdef SYS

//  Error codes for ABSREAD/ABSWRITE functions.  AH assumed to be zero.
//
#define ABS_ERROR_WRITE_PROTECT		0
#define ABS_ERROR_UNKNOWN_UNIT		1
#define ABS_ERROR_DRIVE_NOT_READY	2
#define ABS_ERROR_UNKNOWN_MEDIA		7
#define ABS_ERROR_INVALID_CHANGE		0xf

//  Initial value for command line switch variables
//
#define DISPLAY_SIGNON_MESSAGE		TRUE
#define HIDE_SYSTEM_FILES				TRUE
#define CONFIRM_OLD_SYSTEM_DELETE	TRUE

//  Initial value for whether the BPB that was read
//  was FAT32 or FAT12/FAT16
//
#define DONT_SYS_FAT32					FALSE

//  These are constants for different character values
//
#define CHAR_SPACE						0x20
#define CHAR_TAB							0x08
#define CHAR_COLON						0x3a
#define CHAR_BACKSLASH					0x5c
#define CHAR_NULL							0x00
#define CHAR_RETURN						0x0d

//  These are constants for file names and path sizes
//
#define MAX_NAME_AND_PATH				128

//  This is the maximum size the IBMDOS.COM, which only
//  contains the romdos version and copywrite information,
//  will ever be.
//
#define MAX_DATALIGHT_DATA_SIZE		100

//  This value is used to when there is no segment value
//  for a block of memory.
//
#define NO_ALLOCATED_SEGMENT			0

//  This constant represents the offset for an allocated
//  block of memory.
//
#define INITIAL_BUFFER_OFFSET			0

//  This is the number of paragraphs you allocate when
//  you are trying to determine the available amount
//  of conventional memory.
//
#define MAX_PARAGRAPH_SIZE				0xffff

//  This is the byte size of read and write operations
//
#define READ_AND_WRITE_SIZE			0x8000

//  This constant is for clearity, it represents the
//  number of byte read when no bytes were read.
//
#define NO_BYTES_READ					0

//  This is the size of a paragraph of conventional
//  memory in bytes.
//
#define PARAGRAPH_SIZE					16

//  This value is used to represent an invalid drive letter
//
#define INVALID_DRIVE					0xff

//  This is an index into the boot sector which
//  is where the BPB starts
//
#define BPB_START_IN_BOOT_SECTOR		0x0b

//  Return code for the program when there were no errors
//
#define NO_ERROR		0

//  These are error codes returned from functions
//
#define ERR_INSUFFICIENT_MEMORY		0x0008
#define FUNCTION_NOT_SUPPORTED		0x0001

//  These are constants for boot sector information
//
#define SECTOR_SIZE										512
#define NUMBER_OF_SECTORS_IN_FAT16_BOOT_SECTOR	1
#define NUMBER_OF_SECTORS_IN_FAT32_BOOT_SECTOR	3
#define MAX_NUMBER_OF_SECTORS_IN_BOOT_SECTOR		3
#define STARTING_SECTOR_OF_BOOT_SECTOR				0
#define MAX_BOOT_SECTOR_SIZE		MAX_NUMBER_OF_SECTORS_IN_BOOT_SECTOR * 1024
#define BOOT_SECTOR_MARKER								0xaa55


//
// [ STRUCTURES ]
//

//  Structure to organize the file information for the
//  the system files
//
typedef struct{
	char szFileName[MAX_NAME_AND_PATH];
	unsigned uMemorySegment;
	ulong ulFileSize;
	DOSFileDate sDate;
	DOSFileTime sTime;
	unsigned  uNumberOfClusters;
} FileInfoType;

#if 0
//  This is the structure for a FAT32 BPB.  At the time this was
//  written the structurs to support FAT32 were not in dostruct.h
//
typedef struct{
    unsigned uBytesPerSector;
    uchar ucSectorsPerCluster;
    unsigned uReservedSectors;
    uchar ucNumberOfFATs;
    unsigned uRootEntries;
    unsigned uTotalSectors;
    uchar ucMediaDescriptor;
    unsigned uSectorsPerFAT;
    unsigned uSectorsPerTrack;
    unsigned uHeads;
    ulong ulHiddenSectors;
    ulong ulBigTotalSectors;
    ulong ulBigSectorsPerFat;
    unsigned uExtFlags;
    unsigned uFS_Version;
    ulong ulRootDirStrtClus;
    unsigned uFSInfoSec;
    unsigned uBkUpBootSec;
    uchar ucReserved[12];
} FAT32BPBType;
#endif // #if 0

//  Structure to organize command line switches
//
typedef struct{
	bool  bShowSignOn;
	bool  bHideSystemFiles;
	bool  bConfirmDelete;
} CLSwitchType;
