//
//   $Id: SYSCOMN.C 1.14 2003/07/03 17:09:44Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  SYSCOMN.C  created 02/02/199 by Brandon Thomas
//
//
//  Copyright (c) 1989-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//
//  THIS FILE IS NOT A STAND ALONE UNIT.
//
//  This file contains the part of sys code which is used by
//  format to sys a disk.
//
//
//  HISTORY
//
//  06-15-2000 td
//  - Added LFN support
//  05-10-1999 BT
//  - Made sure the the ROMDOS25 project and the ROMDOS7 project have
//    the same SYS code.
//  - Now execute the LOCK command for FAT16 drives when there is support
//    for FAT32 functions.
//  04-27-1999 BT
//  - Sys no longer writes the FSINFO sector.  We were writing a stale copy
//    SPR/ETR# 2065
//  - We now execute the LOCK and UNLOCK commands so we don't halt MSDOS.
//    SPR/ETR# 2088
//  02-08-1999 BT
//  - Put constants(more discriptive) in case statements for clearity in
//    the WriteBootSector code.
//  02-08-1999 BT
//  - Made a check to see if we have used all the memory
//  02-08-1999 BT
//  - No longer Pass the pointer to the structure for sCLSwitch, now
//    pass the boolean switches individually.
//  02-08-1999 BT
//  - Made WriteBootSector determine the address of the FAT12/FAT16 or
//    FAT32 buffer.  Also made it determine the number sectors to write
//    for the boot code.
//  02-08-1999 BT
//  - Changed all the (*x).y type dereferencing to x->y dereferencing
//  02-08-1999 BT
//  - Changed all the hungarian type notation of "ui" to "u"
//  02-08-1999 bt
//  - Created a table of system file names so that there was only one copy
//    of the name.
//  02-05-1999 bt  SPR 1914
//  02-05-1999 bt
//  - Split SetSystemFilePath into to functions. One which searched for
//    the romdos kernel and one that searched for command.com.
//  01/27/99  BT
//  -  Rewrote sys.  Created a boot sector which only needs the
//     BPB updated and does not require that the sytem files be
//     contigious or that the system files be the first entries
//     in the root directory.  This sys program just reads the
//     BPB, and validates it, puts in in the new boot sector code.
//     Then copies the system files and writes a new boot sector.
//     This code supports SYSing a FAT32 drive however we are still
//     developing FAT32.  FAT32 support has been turned off.
//

// [ INCLUDES ]
//
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#include <string.h>
//#include <mem.h>
#include <dir.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"

#include "fatworld.h"
#include "dosver.h"

#define SYS 1
#include "newsys.h"

//
// [ TYPES ]
//

//
// [ STRUCTURES ]
//

//
//  [ CONSTANTS ]
//

#define PREPEND_BOOT_CODE				TRUE
#define DONT_PREPEND_BOOT_CODE		FALSE
#define WATCH_FOR_BOOT_CODE			TRUE
#define DONT_WATCH_FOR_BOOT_CODE		FALSE

#define FIRST_SECTOR_ONLY	1
#define THIRD_SECTOR_ONLY	1
#define BYTE_OFFSET_OF_THIRD_SECTOR		1024
#define DISTANCE_IN_SECTORS_TO_THIRD_SECTOR		2

//  These permissions are set the same as MSDOS 7.1 SYS.COM
//
#define LOCK_LEVEL			0
#define DRIVE_PERMISSIONS	0

//
//  [ EXTERNAL VARIABLES ]
//

extern char sysboot_getsys[];
extern char * err_msg[];
extern bool fIsFat32World;
extern uchar rgucBootSector[];
extern uchar rgucBootSectorEnd[];
extern uchar rgucBootSector32[];
extern uchar rgucBootSector32End[];
extern char warn_olddos[];

extern bool fIsFat32World;

//
//  [ VARIABLES ]
//

//  These are structures used to hold information
//  about the system files such as the file date, time,
//  and size as well as where in conventional memory
//  it was loaded.
//
FileInfoType sRomdos;
FileInfoType sCommand;

//  This is a flag which indicate whether the romdos kernel system file
//  is a FAT32 kernel
//
bool fIsFat32Binary = FALSE;

//  This is a list of file names for other OS.  It has been
//  placed here so that additions can be made easily and
//  so that only one copy of it exists.
//
char * szOtherSystemFile[] = {
	"IBMBIO.COM",
	"IBMDOS.COM",
	"COMMAND.COM",
	"MSDOS.SYS",
	"IO.SYS",
	"DBLSPACE.BIN",
	"DRVSPACE.BIN",
};

//  This is an array for system file names and indexes into the
//  array.
//
#define DEFAULT_KERNEL_NAME			0
#define KERNAL_SYSTEM_FILE_NAME  	1
#define EXTRA_SYSTEM_FILE_NAME   	2
#define COMMAND_SYSTEM_FILE_NAME		3
char * szSystemFileNames[] = {
	"ROM-DOS.SYS",		//  First kernel name
	"IBMBIO.COM",		//  Another possible kernal name - system file on disk
	"IBMDOS.COM",		//  System file on disk
	"COMMAND.COM",		//  System file on disk
};

//  This is the number of strings which are in the
//  pszOtherSystemFile array.
//
#define FILE_SEARCH_ARRAY_SIZE	( (sizeof(szOtherSystemFile))/(sizeof(szOtherSystemFile[0])) )


//
//  [ FUNCTION PROTOTYPES ]
//
#include "syscomn.pt"

//  (BT) 2-5-1999 SPR 1914
//
//  Changed this function which was previously SetSystemFilePath and was
//  used for both command.com and rom-dos.sys to only be used for searching
//  for the romdos kernel.  Made a seperate function for checking for the
//  command.com path.  Also made this function look for the alternate romdos
//  kernel name IBMBIO.COM.
//
//  The system file search order:
//
//  1) look for ROM-DOS.SYS in current directory
//  2) look for IBMBIO.COM in root directory, current drive
//  3) look for IBMBIO.COM in root directory, boot drive
//  4) look for ROM-DOS.SYS in the execution path for SYS/FORMAT
//  5) ask user where ROM-DOS.SYS can be found
//
void SetRomdosSystemFilePath( char * szFileName, FileInfoType * psFileInfo,
		char * szSysNameAndPath, char * szPathWhereRomdosWasFound )
{
	//  [ LOCALS ]
	//
	static char szCurrentPath[LFN_MAX_PATH];
	char *   szAlternateKernel = szSystemFileNames[KERNAL_SYSTEM_FILE_NAME];
	char *   szDriveString = "A:\\";
	char *   szBootDriveString = "A:\\";
	char *   szRootPath = &szDriveString[2];
	uchar    ucCurrentDriveLetter;
	bool     bFunctionSuccessful;
	int      iDriveLetter;
	unsigned uAlternatePathSize;
	uchar    ucBackslashIndex;
	bool     bFileDoesExist;
	bool     bRunningRomdos;
	bool     bIsOurIBMDOS;
	bool     bSuccessful;
	bool     bDoneSearchingForRomdos = FALSE;
	unsigned uNameSpaceSize;


	//
	//  [ CODE ]
	//

	//  Get the current drive
	//
	bFunctionSuccessful = dlGetCurrentDrive( &iDriveLetter );
	if( !bFunctionSuccessful )
	{
		ErrorExit( ERR_DRVSPECS, NULL, NULL );
	}

	//  Save the current drive letter
	//
	ucCurrentDriveLetter = (uchar)(iDriveLetter + 'A');

	//  Get the current directory
	//
	bFunctionSuccessful = SmartGetCurrentDirectory( iDriveLetter+1,
					szCurrentPath );
	if( !bFunctionSuccessful )
	{
		ErrorExit( ERR_DRVSPECS, NULL, NULL );
	}

	//  Set the drive letter
	//
	szDriveString[0] = ucCurrentDriveLetter;

	//  Build the name and path for the current directory
	//
	//  BT 2-8-1999 Walk through bug fix 1
	//
	//  Make sure we return the path that we found romdos at
	//
	strcpy( szPathWhereRomdosWasFound, szDriveString );
	strcat( szPathWhereRomdosWasFound, szCurrentPath );
	if( szCurrentPath[0] != 0 )
	{
		strcat( szPathWhereRomdosWasFound, szRootPath );
	}
	strcpy( psFileInfo->szFileName, szPathWhereRomdosWasFound );
	strcat( psFileInfo->szFileName, szFileName );

	//  Check if the file exists
	//
	if( bFileExist( psFileInfo->szFileName ) )
	{
		return;
	}

	//  Check for IBMBIO.COM in the current directory of the current drive
	//  if we are running romdos.
	//
	bRunningRomdos = dlIsRunningRomdos();
	if( bRunningRomdos )
	{
		//  Build the name and path for the current directory of
		//  the current drive
		//
		//  BT 2-8-1999 Walk through bug fix 1
		//
		//  Make sure we return the path that we found romdos at
		//
		strcpy( szPathWhereRomdosWasFound, szDriveString );
		strcat( szPathWhereRomdosWasFound, szCurrentPath );
		if( szCurrentPath[0] != 0 )
		{
			strcat( szPathWhereRomdosWasFound, szRootPath );
		}
		strcpy( psFileInfo->szFileName, szPathWhereRomdosWasFound );
		strcat( psFileInfo->szFileName, szAlternateKernel );

		//  Check if the file exists
		//
		bFileDoesExist = bFileExist( psFileInfo->szFileName );
		if( bFileDoesExist )
		{
			//  Check that this is our ibmdos.com file
			//
			bIsOurIBMDOS = CheckForOurIBMDOS( szPathWhereRomdosWasFound );
			if( bIsOurIBMDOS )
			{
				return;
			}
		}
	}

	//  Check for IBMBIO.COM in the root directory of the current boot
	//  drive if we are running romdos.
	//
	if( bRunningRomdos )
	{
		//  Get the boot drive
		//
		bSuccessful = dlGetBootDriveLetter(
				(uchar far *)(&szBootDriveString[0]) );
		if( bSuccessful )
		{
			//  Build the name and path for the root of the boot drive
			//
			//  BT 2-8-1999 Walk through bug fix 1
			//
			//  Make sure we return the path that we found romdos at
			//
			strcpy( szPathWhereRomdosWasFound, szBootDriveString );
			strcpy( psFileInfo->szFileName, szPathWhereRomdosWasFound );
			strcat( psFileInfo->szFileName, szAlternateKernel );

			//  Check if the file exists
			//
			bFileDoesExist = bFileExist( psFileInfo->szFileName );
			if( bFileDoesExist )
			{
				//  Check that this is our ibmdos.com file
				//
				bIsOurIBMDOS = CheckForOurIBMDOS( szPathWhereRomdosWasFound );
				if( bIsOurIBMDOS )
				{
					return;
				}
			}
		}
	}

	//  Get the path to were SYS.COM was executed
	//
	ucBackslashIndex = strlen( szSysNameAndPath );
	while( szSysNameAndPath[ ucBackslashIndex ] != CHAR_BACKSLASH )
	{
		ucBackslashIndex--;
	}
	szSysNameAndPath[ ucBackslashIndex + 1 ] = CHAR_NULL;

	//  Build the name and path for the same directory that
	//  SYS.COM (or FORMAT.COM) was ran from
	//
	//  BT 2-8-1999 Walk through bug fix 1
	//
	//  Make sure we return the path that we found romdos at
	//
	strcpy( szPathWhereRomdosWasFound, szSysNameAndPath );
	strcpy( psFileInfo->szFileName, szPathWhereRomdosWasFound );
	strcat( psFileInfo->szFileName, szFileName );

	//  Check if the file exists
	//
	if( bFileExist( psFileInfo->szFileName ) )
	{
		return;
	}

	//  Ask the user for the location of ROM-DOS.SYS
	//
	while( !bDoneSearchingForRomdos )
	{
		//  Prompt for an alternate path
		//
		printf(sysboot_getsys,szFileName);
		uNameSpaceSize = LFN_MAX_PATH - 2;
		if( uNameSpaceSize > 255 )
		{
			szPathWhereRomdosWasFound[0] = 255;
		}
		else
		{
			szPathWhereRomdosWasFound[0] = (uchar)uNameSpaceSize;
		}
		dlGetLine( (uchar *)(szPathWhereRomdosWasFound) );
		szPathWhereRomdosWasFound[szPathWhereRomdosWasFound[1] + 2] = 0;
		memmove( szPathWhereRomdosWasFound, &szPathWhereRomdosWasFound[2],
				szPathWhereRomdosWasFound[1] + 1 );
		LFNStripArgQuotes( szPathWhereRomdosWasFound );

		//  User typed in nothing, interpret this as an abort
		//
		uAlternatePathSize = strlen(szPathWhereRomdosWasFound);
		if( !uAlternatePathSize )
		{
			ErrorExit( ERR_ABORT, NULL, NULL );
		}

		// Append the name onto path
		//
		if(( szPathWhereRomdosWasFound[uAlternatePathSize-1] != CHAR_COLON ) &&
				( szPathWhereRomdosWasFound[uAlternatePathSize-1] != CHAR_BACKSLASH ))
		{
			strcat( szPathWhereRomdosWasFound, szRootPath );
		}

		//  Build the name and path for the root directory
		//
		strcpy( psFileInfo->szFileName, szPathWhereRomdosWasFound );
		strcat( psFileInfo->szFileName, szFileName );

		//  Check if romdos exists at this path
		//
		bDoneSearchingForRomdos = bFileExist( psFileInfo->szFileName );
	}

	return;
}

//  (BT) 2-5-1999 SPR 1914
//
//  Changed this function which was previously SetSystemFilePath and was
//  used for both command.com and rom-dos.sys to only be used for searching
//  for the romdos kernel.  Made a seperate function for checking for the
//  command.com path.
//
void SetCommandSystemFilePath( char * szFileName, FileInfoType * psFileInfo,
		char * szPathToCommandFile )
{
	//  [ LOCALS ]
	//
//	uchar ucCurrentDriveLetter;
	char * szRootPath  = "\\";
//	bool  bFunctionSuccessful;
//	int  iDriveLetter;
	unsigned uAlternatePathSize;
//	uchar ucBackslashIndex;
//	bool   bFileDoesExist;
	bool   bDoneLookingForCommand = FALSE;
	unsigned uNameSpaceSize;


	//
	//  [ CODE ]
	//

	//  Ask the user for the location of command.com if we don't already
	//  have it.
	//
	while( !bDoneLookingForCommand )
	{
		//  Build the name and path for the root directory
		//
		strcpy( psFileInfo->szFileName, szPathToCommandFile );
		strcat( psFileInfo->szFileName, szFileName );

		//  Check for command.com at this path
		//
		if( bFileExist( psFileInfo->szFileName ) )
		{
			break;
		}

		//  Prompt for an alternate path
		//
		printf(sysboot_getsys, szFileName );
		uNameSpaceSize = LFN_MAX_PATH - 2;
		if( uNameSpaceSize > 255 )
		{
			szPathToCommandFile[0] = 255;
		}
		else
		{
			szPathToCommandFile[0] = (uchar)uNameSpaceSize;
		}
		dlGetLine( (uchar *)(szPathToCommandFile) );
		szPathToCommandFile[szPathToCommandFile[1] + 2] = 0;
		memmove( szPathToCommandFile, &szPathToCommandFile[2],
				szPathToCommandFile[1] + 1 );
		LFNStripArgQuotes( szPathToCommandFile );

		//  User typed in nothing, interpret this as an abort
		//
		uAlternatePathSize = strlen(szPathToCommandFile);
		if( !uAlternatePathSize )
		{
			ErrorExit( ERR_ABORT, NULL, NULL );
		}

		// Append the name onto path
		//
		if(( szPathToCommandFile[uAlternatePathSize-1] != CHAR_COLON ) &&
				( szPathToCommandFile[uAlternatePathSize-1] != CHAR_BACKSLASH ))
		{
			strcat( szPathToCommandFile, szRootPath );
		}
	}
	return;
}

//  Read the first couple byte of the IBMDOS.COM file, if it exists, and
//  check to make sure it says romdos.
//
bool CheckForOurIBMDOS( char * szPath )
{
	//  [LOCALS]
	//
	static char szFileName[LFN_MAX_PATH];
	char * szIBMDOS = "IBMDOS.COM";
	char * szRomdosString = "ROM-DOS";
	uchar rgucBuffer[8];
	int  iHandle, iActionTaken;
	unsigned uReadBytes;
	bool  bFunctionSuccessful;
	unsigned uReturnValue;

	//
	//  [ CODE ]
	//

	//  Create the file name and path
	//
	strcpy( szFileName, szPath );
	strcat( szFileName, szIBMDOS );

	//  Open the file
	//
   bFunctionSuccessful = SmartCreateOpenFile(OPEN_ACCESS_READONLY, A_NORMAL,
				FILE_OPEN, szFileName,	&iHandle, &iActionTaken);
	if( !bFunctionSuccessful )
	{
		return FALSE;
	}

	//  Read what should be the romdos string from the IBMDOS.COM file
	//
	uReadBytes = 7;
	bFunctionSuccessful = dlReadFile( iHandle, &uReadBytes,
			(char *)(rgucBuffer) );
	if( !bFunctionSuccessful )
	{
		dlCloseFile( iHandle );
		ErrorExit( ERR_READ, NULL, szPath[0] );
	}

	//  Close the file
	//
	bFunctionSuccessful = dlCloseFile( iHandle );
	if( !bFunctionSuccessful )
	{
		dlCloseFile( iHandle );
		ErrorExit( ERR_CLOSING, NULL, NULL );
	}

	//  Check if this is romdos's IBMDOS.COM
	//
	rgucBuffer[7] = 0;
	uReturnValue = strcmp( (char *)(rgucBuffer), szRomdosString );
	if( uReturnValue == 0 )
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

//  Check if a file exists
//
bool bFileExist( char * szFileName )
{
	//  [ LOCALS ]
	//
	bool  bFunctionSuccessful;
	unsigned uFileAttributes;


	//
	//  [ CODE ]
	//

	//  If we were not able to get the file attributes then the file
	//  does not exist
	//
	bFunctionSuccessful = SmartGetFileAttributes( szFileName,
			&uFileAttributes);

	//  Return if the file exists
	//
	return bFunctionSuccessful;
}


//  Copy the system files to the destination drive
//
void CopySystemFiles( uchar ucDestinationDriveLetter,	bool bConfirmDelete,
		bool bHideSystemFiles, bool bSysFAT32Drive )
{
	//  [ LOCALS ]
	//

	//
	//  [ CODE ]
	//

	//  Check if there is enough room for the files
	//  on the destination drive
	//
	CheckDriveSpace( ucDestinationDriveLetter, bSysFAT32Drive,
			bConfirmDelete );

	//  Write the system files to the destination drive
	//
	WriteSystemFiles( ucDestinationDriveLetter, bHideSystemFiles,
			bSysFAT32Drive );

   //  To save repeated alloc & dealloc from FORMAT,
   //  free the memory at the same code depth it was allocated.
   //
#if 0
   //  Release the memory used to hold the system files in
   //  memory.
   //
   FreeAllMemory();
#endif
}


//  Write the system files to the destination drive
//
void WriteSystemFiles( uchar ucDestinationDriveLetter, bool bHideSystemFiles,
		bool bSysFAT32Drive )
{
	//  [ LOCALS ]
	//
	static    char szFileName[LFN_MAX_PATH];
	static    char szDatalightData[MAX_DATALIGHT_DATA_SIZE];
	char * szVersionString = "ROM-DOS %s %s\r\n%s\r\n";
	char * szIBMBIO  = szSystemFileNames[KERNAL_SYSTEM_FILE_NAME];
	char * szIBMDOS  = szSystemFileNames[EXTRA_SYSTEM_FILE_NAME];
	char * szCOMMAND = szSystemFileNames[COMMAND_SYSTEM_FILE_NAME];
	ulong     ulDatalightDataSize;
	char * szDriveString = "A:\\";
	unsigned  uSystemFileAttributes;
	unsigned  uCommandAttributes;


	//
	//  [ CODE ]
	//

	//  Set the attribute for the system files
	//
	if( bHideSystemFiles )
	{
		//  By default we hide the system files
		//
		uSystemFileAttributes = A_SYSTEM | A_HIDDEN | A_READONLY | A_ARCHIVE;
	}
	else
	{
		//  User does not want to hide the system files
		//
		uSystemFileAttributes = A_ARCHIVE;
	}
	uCommandAttributes = A_ARCHIVE;

	//  Set the destination drive letter
	//
	szDriveString[0] = ucDestinationDriveLetter;

	//  Create the IBMBIO.COM file
	//
	strcpy( szFileName, szDriveString );
	strcat( szFileName, szIBMBIO );
	WriteFile( szFileName, uSystemFileAttributes, sRomdos.ulFileSize,
			sRomdos.uMemorySegment, INITIAL_BUFFER_OFFSET,
			&sRomdos.sDate, &sRomdos.sTime, !bSysFAT32Drive );

	//  Create the IBMDOS.COM file
	//
	strcpy( szFileName, szDriveString );
	strcat( szFileName, szIBMDOS );
	if( fIsFat32World )
	{
		sprintf(szDatalightData, szVersionString,SHORTVERSION,
				BUILDVERSION,COPYRIGHT);
	}
	else
	{
		sprintf(szDatalightData, szVersionString,SHORTVERSION622,
				BUILDVERSION,COPYRIGHT);
	}
	ulDatalightDataSize = strlen( szDatalightData );
	WriteFile( szFileName, uSystemFileAttributes, ulDatalightDataSize,
			FP_SEG(((void far *)szDatalightData)),
			FP_OFF(((void far *)szDatalightData)),
			&sRomdos.sDate, &sRomdos.sTime,	DONT_PREPEND_BOOT_CODE );

	//  Create the COMMAND.COM file
	//
	strcpy( szFileName, szDriveString );
	strcat( szFileName, szCOMMAND );
	WriteFile( szFileName, uCommandAttributes, sCommand.ulFileSize,
			sCommand.uMemorySegment, INITIAL_BUFFER_OFFSET,
			&sCommand.sDate, &sCommand.sTime, DONT_PREPEND_BOOT_CODE );
}


//  Create a file.  Write the data to a file.  Close the file.  Set the
//  files date and time.
//
void WriteFile( char * szFileName, unsigned uAttributes, ulong ulFileSize,
		unsigned uBufferSegment, unsigned uBufferOffset, DOSFileDate * psDate,
		DOSFileTime * psTime, bool bPrependBootCode )
{
	//  [ LOCALS ]
	//
	int  iHandle, iActionTaken;
	unsigned uWriteBytes;
	bool  bFunctionSuccessful;
	ulong ulTotalBytesWritten;
	bool  bDone = FALSE;


	//
	//  [ CODE ]
	//

	//  Create the file
	//
	bFunctionSuccessful = SmartCreateOpenFile(OPEN_ACCESS_WRITEONLY, uAttributes,
	      FILE_TRUNCATE | FILE_CREATE, szFileName, &iHandle, &iActionTaken);
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		ErrorExit( ERR_CREATING, szFileName, NULL );
	}

	//  Prepend the boot code if we are supposed to
	//
	if( bPrependBootCode )
	{
		uWriteBytes = ((unsigned)rgucBootSectorEnd -
				(unsigned)rgucBootSector) - SECTOR_SIZE;
		bFunctionSuccessful = dlWriteFile( iHandle, &uWriteBytes,
				(char far *)(&rgucBootSector[512]));
		if( !bFunctionSuccessful )
		{
			NoErrorFreeAllMemory();
			dlCloseFile( iHandle );
			ErrorExit( ERR_READ, NULL, NULL );
		}
	}

	//  Write the file
	//
	ulTotalBytesWritten = 0;
	while( !bDone )
	{
		//  Determine how many bytes we are going to write
		//
		if( (ulTotalBytesWritten + READ_AND_WRITE_SIZE) >= ulFileSize )
		{
			uWriteBytes = (unsigned)(ulFileSize - ulTotalBytesWritten);
			bDone = TRUE;
		}
		else
		{
			uWriteBytes = READ_AND_WRITE_SIZE;
		}

		bFunctionSuccessful = dlWriteFile( iHandle, &uWriteBytes,
				(char far *)MK_FP( uBufferSegment, uBufferOffset ));
		if( !bFunctionSuccessful )
		{
			NoErrorFreeAllMemory();
			dlCloseFile( iHandle );
			ErrorExit( ERR_WRITE, NULL, szFileName[0] );
		}

		//  Check if the buffer segment overflowed and point to the
		//  next chunk of data we are going to write.
		//
		if( ((unsigned)(uBufferOffset + uWriteBytes)) < uBufferOffset )
		{
			uBufferSegment += 0x1000;
		}
		uBufferOffset += uWriteBytes;
		ulTotalBytesWritten += uWriteBytes;
	}

	//  Set the files date and time
	//
	bFunctionSuccessful = dlSetFileDateAndTime( iHandle, psTime, psDate );
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		dlCloseFile( iHandle );
		ErrorExit( ERR_DATETIME, NULL, NULL );
	}

	//  Close the file
	//
	bFunctionSuccessful = dlCloseFile( iHandle );
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		dlCloseFile( iHandle );
		ErrorExit( ERR_CLOSING, NULL, NULL );
	}
}


//  Free all allocated memory with no error checking.  This is called
//  when memory has been allocated and there is an error.
//
void NoErrorFreeAllMemory( void )
{
	//  Free the memory used by command if memory was allocated to
	//  store it in memory.
	//
	if( sCommand.uMemorySegment != NO_ALLOCATED_SEGMENT )
	{
		dlFreeMem( sCommand.uMemorySegment );
	}

	//  Free the memory used by romdos if memory was allocated to
	//  store it in memory.
	//
	if( sRomdos.uMemorySegment != NO_ALLOCATED_SEGMENT )
	{
		dlFreeMem( sRomdos.uMemorySegment );
	}
}


//  Free all allocated memory with error checking.  This is called
//  when memory has been allocated and the program is about to
//  exit.
//
void FreeAllMemory( void )
{
	//  [ LOCALS ]
	//
	bool bFunctionSuccessful;


	//
	//  [ CODE ]
	//

	//  Free the memory used by command if memory was allocated to
	//  store it in memory.
	//
	if( sCommand.uMemorySegment != NO_ALLOCATED_SEGMENT )
	{
		bFunctionSuccessful = dlFreeMem( sCommand.uMemorySegment );
		if( !bFunctionSuccessful )
		{
			NoErrorFreeAllMemory();
			ErrorExit( ERR_RELEASE, NULL, NULL );
		}
	}

	//  Free the memory used by romdos if memory was allocated to
	//  store it in memory.
	//
	if( sRomdos.uMemorySegment != NO_ALLOCATED_SEGMENT )
	{
		dlFreeMem( sRomdos.uMemorySegment );
		if( !bFunctionSuccessful )
		{
			NoErrorFreeAllMemory();
			ErrorExit( ERR_RELEASE, NULL, NULL );
		}
	}
}


//  Check if the destination drive has enough space to copy
//  the system files.
//
#pragma argsused
void CheckDriveSpace( uchar ucDestinationDriveLetter, bool bSysFAT32Drive,
		bool bConfirmDelete )
{
	//  [ LOCALS ]
	//
#if USING_FAT32
	char * szDriveString = "A:\\";
#endif
	ulong ulClusterSize;
	unsigned uTotalClustersForSystemFiles;
	unsigned uClustersUsedByOtherSystem;
	unsigned uSectorsPerCluster;
	unsigned uFreeClusters;
	unsigned uBytesPerSector;
	unsigned uTotalClusters;
	bool  bFunctionSuccessful;
	uchar ucKey = 0;
#if USING_FAT32
	GDF32	 sFreeSpace;
#endif
	ulong ulTotalFreeClusters;


	//
	//  [ CODE ]
	//

#if USING_FAT32

	//  Determine if this is a FAT32 drive
	//
	if( bSysFAT32Drive )
	{
		//  Create a drive string which contains the
		//  proper drive letter
		//
		szDriveString[0] = (char)ucDestinationDriveLetter;

		//  Get the FAT32 drive specifications
		//
		bFunctionSuccessful = dlGetDriveFreeSpace32( szDriveString,
				&sFreeSpace );
		if( !bFunctionSuccessful )
		{
			NoErrorFreeAllMemory();
			ErrorExit( ERR_DRVSPECS, NULL, NULL );
		}

		//  Calculate the cluster size in bytes
		//
		ulClusterSize = sFreeSpace.ulBytesPerSector *
				sFreeSpace.ulSectorsPerCluster;

		//  Determine how many clusters it would take for
		//  the romdos sytem file
		//
		sRomdos.uNumberOfClusters =
				ComputeFileSizeInClusters( sRomdos.ulFileSize,
				ulClusterSize );

		//  Determine how many clusters it would take for
		//  the command sytem file
		//
		sCommand.uNumberOfClusters =
				ComputeFileSizeInClusters( sCommand.ulFileSize,
				ulClusterSize );

		//  Set the number of free clusters in a different
		//  variable.
		//
		ulTotalFreeClusters = sFreeSpace.ulAvailableClusters;
	}
	else

#endif // #if USING_FAT32

	{
		//  Get the FAT12/FAT16 drive specifications
		//
		bFunctionSuccessful = __dlGetFreeSpaceOnDrive(
					(ucDestinationDriveLetter - 'A' + 1), &uSectorsPerCluster,
					&uFreeClusters, &uBytesPerSector, &uTotalClusters );
		if( !bFunctionSuccessful )
		{
			NoErrorFreeAllMemory();
			ErrorExit( ERR_DRVSPECS, NULL, NULL );
		}

		//  Calculate the cluster size in bytes
		//
		ulClusterSize = (ulong)uSectorsPerCluster * (ulong)uBytesPerSector;

		//  Determine how many clusters it would take for
		//  the romdos sytem file including an extra sector
		//  for the second part of the boot loader program.
		//
		sRomdos.uNumberOfClusters =
				ComputeFileSizeInClusters( (sRomdos.ulFileSize + SECTOR_SIZE),
				ulClusterSize );

		//  Determine how many clusters it would take for
		//  the command sytem file
		//
		sCommand.uNumberOfClusters =
				ComputeFileSizeInClusters( sCommand.ulFileSize,
				ulClusterSize );

		//  Store the free clusters in a larger variable
		//
		ulTotalFreeClusters = uFreeClusters;
	}

	uTotalClustersForSystemFiles = sCommand.uNumberOfClusters +
			sRomdos.uNumberOfClusters;

	//  Get the number of clusters used by system files that
	//  are already on the drive.
	//
	uClustersUsedByOtherSystem = GetClustersForOtherSystem( ulClusterSize,
			ucDestinationDriveLetter );

	//  Determine if there is enough drive space for the system files
	//
	ulTotalFreeClusters += uClustersUsedByOtherSystem;
	if( ulTotalFreeClusters < uTotalClustersForSystemFiles )
	{
		//  Insufficient disk space
		//
		NoErrorFreeAllMemory();
		ErrorExit( ERR_NOROOM, NULL, NULL );
	}

	//  Give user a chance to abort if the disk has
	//  already been sysed with an OS.
	//
	if( uClustersUsedByOtherSystem )
	{
		//  Check if the user wants to confirm the removal of other
		//  system files
		//
		if( bConfirmDelete )
		{
			//  Prompt to delete other system files
			//
			printf( warn_olddos );
			while( (ucKey != 'Y') && (ucKey != 'y') && (ucKey != 'N') &&
					(ucKey != 'n') && (ucKey != CHAR_RETURN) )
			{
				dlFlushAndGetKeyWithEcho( &ucKey );
			}
			printf("\n\n");

			//  Check if the user aborted
			//
			if( (ucKey == 'N') || (ucKey == 'n') || (ucKey == CHAR_RETURN) )
			{
				NoErrorFreeAllMemory();
				ErrorExit( ERR_ABORT, NULL, NULL );
			}
		}

		//  Delete other system files
		//
		DeletePreviousSystemFiles( ucDestinationDriveLetter );
	}
}

//  Delete the previous system files on the destination disk
//
void DeletePreviousSystemFiles( uchar ucDestinationDriveLetter )
{
	//  [ LOCALS ]
	//
	static char szFileName[LFN_MAX_PATH];
	char * szDriveString = "A:\\";
	bool  bFunctionSuccessful;
	uchar ucIndex;
//	unsigned uAttributes;


	//
	//  [ CODE ]
	//

	//  Set the destination drive letter in the file name string
	//
	szDriveString[0] = ucDestinationDriveLetter;

	//  Look for other system files and determine how many clusters
	//  are being used by those files.
	//
	for( ucIndex = 0; ucIndex < FILE_SEARCH_ARRAY_SIZE; ucIndex++ )
	{
		//  Create a string which not only has the file name but includes
		//  a drive path.
		//
		strcpy( szFileName, szDriveString );
		strcat( szFileName, szOtherSystemFile[ucIndex] );

		//  Check if the file exists
		//
		bFunctionSuccessful = bFileExist( szFileName );
		if( bFunctionSuccessful )
		{
			//  Clear the attributes so we can delete the file
			//
			bFunctionSuccessful = dlSetFileAttributes( A_NORMAL, szFileName );
			if( !bFunctionSuccessful )
			{
				NoErrorFreeAllMemory();
				ErrorExit( ERR_DELETING, NULL, NULL );
			}

			//  Delete the file
			//
			bFunctionSuccessful = dlDeleteFile( szFileName );
			if( !bFunctionSuccessful )
			{
				NoErrorFreeAllMemory();
				ErrorExit( ERR_DELETING, NULL, NULL );
			}
		}
	}
}


//  Get the total number of clusters used by system files on the
//  destination drive.
//
unsigned GetClustersForOtherSystem( ulong ulClusterSize,
		uchar ucDestinationDriveLetter )
{
	//  [ LOCALS ]
	//
	static char szFileName[LFN_MAX_PATH];
	char * szDriveString = "A:\\";
	bool  bFunctionSuccessful;
	uchar ucIndex;
	ulong ulFileSize;
	unsigned uTotalUsedClusters = 0;


	//
	//  [ CODE ]
	//

	//  Set the destination drive letter in the file name string
	//
	szDriveString[0] = ucDestinationDriveLetter;

	//  Look for other system files and determine how many clusters
	//  are being used by those files.
	//
	for( ucIndex = 0; ucIndex < FILE_SEARCH_ARRAY_SIZE; ucIndex++ )
	{
		//  Create a string which not only has the file name but includes
		//  a drive path.
		//
		strcpy( szFileName, szDriveString );
		strcat( szFileName, szOtherSystemFile[ucIndex] );

		//  Get the size of the file if it exists
		//
		bFunctionSuccessful = GetFileSize( szFileName, &ulFileSize );
		if( bFunctionSuccessful )
		{
			uTotalUsedClusters += ComputeFileSizeInClusters(
					ulFileSize, ulClusterSize );
		}
	}

	//  Return the total number of clusters used by the
	//  system files already on the disk.
	//
	return uTotalUsedClusters;
}


//  Determine the number of clusters it would take to accomidate
//  a file of a particular size.
//
unsigned ComputeFileSizeInClusters( ulong ulFileSize, ulong ulClusterSize )
{
	//  [ LOCALS ]
	//
	unsigned uNumberOfClusters;


	//
	//  [ CODE ]
	//

	//  Determine how many clusters it would take for a file
	//  of this size.
	//
	uNumberOfClusters = (unsigned)(ulFileSize / ulClusterSize);
	if( (ulFileSize % ulClusterSize) != 0 )
	{
		uNumberOfClusters++;
	}

	//  Return the size of the file in clusters
	//
	return uNumberOfClusters;
}


//  Get the size in bytes of a file.  If the file exists, the
//  function returns a boolean TRUE and the file size is filled
//  in.
//
bool GetFileSize( char * szFileName, ulong * pulFileSize )
{
	//  [ LOCALS ]
	//
	static FIND sFileFind;
	bool  bFunctionSuccessful;
	void far * fpOldDTA;
	unsigned uAttributes;


	//
	//  [ CODE ]
	//

	//  Save the current DTA address
	//
	dlGetDTA( &fpOldDTA );

	//  Set a new DTA address (this function can not error)
	//
	dlSetDTA( &sFileFind );

	//  Perform a find first on the file
	//
	uAttributes = A_SYSTEM | A_HIDDEN | A_READONLY | A_ARCHIVE;
	bFunctionSuccessful = dlFindFirstFile( uAttributes, szFileName );
	if( bFunctionSuccessful )
	{
		//  The file was found, save the file size
		//
		*pulFileSize = sFileFind.size;
	}

	//  Restore the previous DTA address
	//
	dlSetDTA( fpOldDTA );

	//  Return the size of the file
	//
	return bFunctionSuccessful;
}


//  Write the boot sector to the destination drive
//
void WriteBootSector( uchar ucDriveLetter, bool bSysFAT32Drive )
{
	//  [ LOCALS ]
	//
	bool bIsFAT32FunctionSupport;
	bool fSuccessful;
	int  iDriveNumber;
	uchar ucReturnCode;

	//
	//  [ CODE ]
	//


	//  Determine the drive number based on a drive letter.  The drive
	//  number for the FAT32 abs read function starts with 1.
	//
	iDriveNumber = ucDriveLetter - 'A' + 1;

	bIsFAT32FunctionSupport = IsFAT32FunctionSupport( ucDriveLetter );

	if( bIsFAT32FunctionSupport )
	{
		//  Execute the lock command so that MSDOS 7.1 will not halt the
		//  system.
		//
		fSuccessful = dlLockLogicalVolume( iDriveNumber, LOCK_LEVEL,
				DRIVE_PERMISSIONS );
		if( !fSuccessful )
		{
			ErrorExit( ERR_LOCK, NULL, NULL );
		}
	}

	if( bSysFAT32Drive )
	{
#if USING_FAT32
		ucReturnCode = WriteFAT32BootSector( ucDriveLetter );
#else
		ErrorExit( ERR_FAT_TYPE, NULL, NULL );
#endif
	}
	else
	{
		ucReturnCode = WriteFAT16BootSector( ucDriveLetter );
	}
	if( ucReturnCode )
	{
		//  abs write says this is an unknown unit
		//
		ErrorExit( ucReturnCode, NULL, ucDriveLetter );
	}

	if( bIsFAT32FunctionSupport )
	{
		//  Execute the unlock command so that other applications will
		//  not have direct disk access after sys is done.
		//
		fSuccessful = dlUnlockLogicalVolume( iDriveNumber );
		if( !fSuccessful )
		{
			ErrorExit( ERR_UNLOCK, NULL, NULL );
		}
	}
}

#if USING_FAT32

//  Write the FAT32 boot sector to the destination drive
//
uchar WriteFAT32BootSector( uchar ucDriveLetter )
{
	//  [ LOCALS ]
	//
	int  iDriveNumber;
	unsigned uErrorCode;
	bool  bFunctionSuccessful;
//	bool  fSuccessful;
	char far * fpBuffer;
	unsigned uNumberOfSectors;
	bool bDoneWritingBootSector = FALSE;
	lpBOOTSECTOR lpBootSector;
	unsigned uSectorForBootSector;

	//
	//  [ CODE ]
	//


	//  Determine the drive number based on a drive letter.  The drive
	//  number for the FAT32 abs read function starts with 1.
	//
	iDriveNumber = ucDriveLetter - 'A' + 1;

	//  Set the address of the boot sector code
	//
	fpBuffer = (char far *)(&rgucBootSector32[0]);

	//  Set the address of the BPB
	//
	lpBootSector = (lpBOOTSECTOR)&rgucBootSector32[0];
	uSectorForBootSector =
			lpBootSector->BootSectorInfo.BootInfo32.BootSectorBPB.uBackUpBootSector;

	//  Set the number of sectors to write
	//
	uNumberOfSectors = FIRST_SECTOR_ONLY;

	while( bDoneWritingBootSector == FALSE )
	{
		//  Perform an abs write using the WIN95 interface for
		//  FAT32 drives
		//
		bFunctionSuccessful = dlAbsWrite32( iDriveNumber,
				uNumberOfSectors, uSectorForBootSector,
				fpBuffer, ABS32WRITE_UNKNOWN );

		//  If there was no error then write out the third sector of the
		//  boot sector
		//
		if( bFunctionSuccessful )
		{
			//  Set the address of the boot sector code
			//
			fpBuffer =
					(char far *)(&rgucBootSector32[BYTE_OFFSET_OF_THIRD_SECTOR]);

			//  Set the address of the BPB
			//
			uSectorForBootSector += DISTANCE_IN_SECTORS_TO_THIRD_SECTOR;

			//  Set the number of sectors to write
			//
			uNumberOfSectors = THIRD_SECTOR_ONLY;

			//  Perform an abs write using the WIN95 interface for
			//  FAT32 drives
			//
			bFunctionSuccessful = dlAbsWrite32( iDriveNumber,
					uNumberOfSectors, uSectorForBootSector,
					fpBuffer, ABS32WRITE_UNKNOWN );

		}

		//  Check if there was an error
		//
		if( !bFunctionSuccessful )
		{
			//  Get the error code
			//
			uErrorCode = __dlCheckAbsRdWrError();

			//  Execute the unlock command so that other applications will
			//  not have direct disk access after sys is done.
			//
			dlUnlockLogicalVolume( iDriveNumber );

			//  Determine how to handle this error
			//
			switch( uErrorCode )
			{
				//  abs write says this is an unknown unit
				//
				case ABS_ERROR_UNKNOWN_UNIT:
					return ERR_BADDRIVE;

				//  There was an error reading from the drive
				//
				default:
					uErrorCode &= 0xff;
					switch( uErrorCode )
					{
						case ABS_ERROR_DRIVE_NOT_READY:
							return ERR_DRVNOTREADY;

						case ABS_ERROR_INVALID_CHANGE:
							return ERR_CHANGE;

						default:
							return ERR_DRVSPECS;
					}
			}
		}

		if( uSectorForBootSector == 2 )
		{
			bDoneWritingBootSector = TRUE;
		}
		else
		{
			uSectorForBootSector = 0;

			//  Set the address of the boot sector code
			//
			fpBuffer = (char far *)(&rgucBootSector32[0]);
		}
	}

	return 0;
}

#endif // #if USING_FAT32

//  Write the FAT16 boot sector to the destination drive
//
uchar WriteFAT16BootSector( uchar ucDriveLetter )
{
	//  [ LOCALS ]
	//
	int  iDriveNumber;
	unsigned uErrorCode;
	bool  bFunctionSuccessful;
//	bool  fSuccessful;
	char far * fpBuffer;
	unsigned uNumberOfSectors;

	//
	//  [ CODE ]
	//


	//  Set the address of the boot sector code
	//
	fpBuffer = (char far *)(&rgucBootSector[0]);

	//  Set the number of sectors to write
	//
	uNumberOfSectors = NUMBER_OF_SECTORS_IN_FAT16_BOOT_SECTOR;

	//  Determine the drive number based on a drive letter.  The drive
	//  number for the abs read function starts with 0.
	//
	iDriveNumber = ucDriveLetter - 'A';

	//  Write the boot sector
	//
	bFunctionSuccessful = dlAbsWrite( TRUE, iDriveNumber,
			uNumberOfSectors, STARTING_SECTOR_OF_BOOT_SECTOR, fpBuffer );

	//  Check if there was an error
	//
	if( !bFunctionSuccessful )
	{
		//  Get the error code
		//
		uErrorCode = __dlCheckAbsRdWrError();

		if( IsFAT32FunctionSupport( ucDriveLetter ) )
		{
			//  Execute the unlock command so that other applications will
			//  not have direct disk access after sys is done.
			//
			dlUnlockLogicalVolume( iDriveNumber );
		}

		//  Determine how to handle this error
		//
		switch( uErrorCode )
		{
			//  We could be trying to access a FAT32 drive while running
			//  WIN95 OSR2 or above.  Try using the WIN95 abs write function
			//  to verify this.
			//
			case ABS_ERROR_UNKNOWN_UNIT:

				//  abs write says this is an unknown unit
				//
				return ERR_BADDRIVE;

			//  There was an error reading from the drive
			//
			default:
				uErrorCode &= 0xff;
				switch( uErrorCode )
				{
					case ABS_ERROR_WRITE_PROTECT:

						//  abs write says the disk is write protected
						//
						return ERR_WRITEPROTECT;

					case ABS_ERROR_DRIVE_NOT_READY:
						return ERR_DRVNOTREADY;

					case ABS_ERROR_INVALID_CHANGE:
						return ERR_CHANGE;

					default:
						return ERR_DRVSPECS;
				}

		}
	}

	return 0;
}


//  Display an error message and exit
//
void ErrorExit( unsigned uErrorCode, char * szString, uchar ucDriveLetter )
{
	//  [ LOCALS ]
	//


	//
	//  [ CODE ]
	//

	//  Display the error message
	//
	if( szString == NULL )
	{
		printf( err_msg[ uErrorCode ], ucDriveLetter );
	}
	else
	{
		printf( err_msg[ uErrorCode ], szString );
	}

	//  Exit with the error code
	//
	exit( uErrorCode );
}


//  Read in the COMMAND.COM and ROM-DOS.SYS files
//
void GetSystemFiles( char * szSysNameAndPath )
{
	//  [ LOCALS ]
	//
	static char szPathWhereRomdosWasFound[LFN_MAX_PATH];
	char * szRomdosFileName  = szSystemFileNames[DEFAULT_KERNEL_NAME];
	char * szCommandFileName = szSystemFileNames[COMMAND_SYSTEM_FILE_NAME];

	//
	//  [ CODE ]
	//

	//  Set the paths to the system files
	//
	//  (BT) 2-5-1999 SPR 1914
	//
	//  No we use seperate functions to search for the romdos kernel
	//  and command.com.  This is so we can look for alternate names
	//  for the romdos kernel. Such as IBMBIO.COM.
	//
	SetRomdosSystemFilePath( szRomdosFileName, &sRomdos, szSysNameAndPath,
			szPathWhereRomdosWasFound );
	SetCommandSystemFilePath( szCommandFileName, &sCommand,
			szPathWhereRomdosWasFound );

	//  Read the files into memory
	//
	//  (BT) 2-5-1999 SPR 1914
	//
	//  We are now telling LoadFile whether or not it is loading
	//  the romdos kernel.  This is so we can check for prepended
	//  boot sector code if the kernel is IBMBIO.COM.
	//
	LoadFile( &sRomdos,  WATCH_FOR_BOOT_CODE );
	LoadFile( &sCommand, DONT_WATCH_FOR_BOOT_CODE );

	//  Set a flag indicating whether the romdos kernel is a FAT32 kernel
	//
	SetKernelTypeFlag( MK_FP( sRomdos.uMemorySegment, 0 ) );
}


//  Set a flag indicating whether the romdos kernel is a FAT32 kernel
//
void SetKernelTypeFlag( char far * pBinaryImage )
{
	fIsFat32Binary = (   (_fstrcmp( &pBinaryImage[8],   "KFAT32" ) == 0)
	                  || (_fstrcmp( &pBinaryImage[493], "KFAT32" ) == 0)
	                 );
}


//  Load the file into conventional memory
//
//  02-08-1999 BT
//  - Made a check to see if we have used all the memory
//
void LoadFile( FileInfoType * psFileInfo, bool bRomdosKernel )
{
	//  [ LOCALS ]
	//
	unsigned   uLargestBlockSize;
	unsigned   uUsedSize;
	int        iHandle, iActionTaken;
	bool       bFunctionSuccessful;
	unsigned   uReadBytes;
	unsigned   uBufferSegment;
	unsigned   uBufferOffset;
	ulong      ulTotalSize;
	bool       bDone = FALSE;
	bool       bDoneLookingForBootCode;
	ulong      ulBytesOfAllocatedMemory;
	unsigned far * fpuBootMarker;


	//
	//  [ CODE ]
	//

	//  Set the variable which is used to tell if we are filtering
	//  out boot code from the kernel.
	//
	bDoneLookingForBootCode = !bRomdosKernel;

	//  Get the size of the largest block of conventional memory
	//
	dlAllocMem( MAX_PARAGRAPH_SIZE, &uLargestBlockSize );

	//  Allocate all the conventional memory
	//
	bFunctionSuccessful = dlAllocMem( uLargestBlockSize,
			&psFileInfo->uMemorySegment );
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		ErrorExit( ERR_NOMEMORY, NULL, NULL );
	}

	//  Determine the number of bytes we allocated
	//
	ulBytesOfAllocatedMemory = (ulong)uLargestBlockSize * PARAGRAPH_SIZE;

	//  Open the file
	//
	bFunctionSuccessful = SmartCreateOpenFile(OPEN_ACCESS_READONLY, A_NORMAL,
	   FILE_OPEN, psFileInfo->szFileName, &iHandle, &iActionTaken);
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		ErrorExit( ERR_OPEN, psFileInfo->szFileName, NULL );
	}

	//  Get the files date and time
	//
	bFunctionSuccessful = dlGetFileDateAndTime( iHandle, &psFileInfo->sTime,
			&psFileInfo->sDate );
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		dlCloseFile( iHandle );
		ErrorExit( ERR_DATETIME, NULL, NULL );
	}

	//  Read the file into conventional memory
	//
	ulTotalSize = 0;
	uBufferSegment = psFileInfo->uMemorySegment;
	uBufferOffset  = INITIAL_BUFFER_OFFSET;
	while( !bDone )
	{
		//  (BT) 2-5-1999 SPR 1914
		//
		//  We will check if the first couple of sectors of the romdos
		//  kernel has boot code prepended to it.  If so we don't want
		//  to keep that as part of the file.
		//
		if( bRomdosKernel && (!bDoneLookingForBootCode) )
		{
			uReadBytes = SECTOR_SIZE;
		}
		else
		{
			uReadBytes = READ_AND_WRITE_SIZE;
		}

		//  Make sure we still have enough memory
		//
		if( ulBytesOfAllocatedMemory < (uReadBytes + ulTotalSize) )
		{
			NoErrorFreeAllMemory();
			ErrorExit( ERR_NOMEMORY, NULL, NULL );
		}

		bFunctionSuccessful = dlReadFile( iHandle, &uReadBytes,
				(char far *)MK_FP( uBufferSegment, uBufferOffset ));
		if( !bFunctionSuccessful )
		{
			NoErrorFreeAllMemory();
			dlCloseFile( iHandle );
			ErrorExit( ERR_READ, NULL, NULL );
		}

		//  (BT) 2-5-1999 SPR 1914
		//
		//  Check if the first couple of sectors of the romdos kernel
		//  has boot code prepended to it.  If so we don't want to keep
		//  that as part of the file.
		//
		if( bRomdosKernel && (!bDoneLookingForBootCode) )
		{
			fpuBootMarker = (unsigned far *)MK_FP( uBufferSegment,
					uBufferOffset );
			if( fpuBootMarker[255] == BOOT_SECTOR_MARKER )
			{
				uReadBytes = 0;
			}
			else
			{
				bDoneLookingForBootCode = TRUE;
			}
		}

		if( ((unsigned)(uBufferOffset + uReadBytes)) < uBufferOffset )
		{
			uBufferSegment += 0x1000;
		}
		uBufferOffset += uReadBytes;
		ulTotalSize += uReadBytes;

		//  (BT) 2-5-1999 SPR 1914
		//
		//  Since we are not accknowledging that we read in kernel code
		//  that had a boot sector prepended to it, it is possible to
		//  be here with zero bytes read and not be done reading the file.
		//  This is only true in the case where we not done looking for
		//  prepended boot code.
		//
		if( (uReadBytes == NO_BYTES_READ) && bDoneLookingForBootCode )
		{
			bDone = TRUE;
		}
	}

	//  Save the size of the file
	//
	psFileInfo->ulFileSize = ulTotalSize;

	//  Close the file
	//
	bFunctionSuccessful = dlCloseFile( iHandle );
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		dlCloseFile( iHandle );
		ErrorExit( ERR_CLOSING, NULL, NULL );
	}

	//  Determine the amount of conventional memory used
	//
	uUsedSize = (unsigned)(ulTotalSize / PARAGRAPH_SIZE);
	if( (ulTotalSize % PARAGRAPH_SIZE) != 0 )
	{
		uUsedSize++;
	}

	//  Release the unused conventional memory
	//
	bFunctionSuccessful = dlResizeMem( uUsedSize, psFileInfo->uMemorySegment,
			&uLargestBlockSize );
	if( !bFunctionSuccessful )
	{
		NoErrorFreeAllMemory();
		ErrorExit( ERR_RESIZE, NULL, NULL );
	}
}

//  Perform a MSDOS 7.1 call to determine whether we are enabling FAT32
//  function call support.
//
#pragma argsused
bool IsFAT32FunctionSupport( uchar ucDriveLetter )
{
	bool  fIsFAT32World;

	fIsFAT32World = dlIsFat32World();

	return fIsFAT32World;
}







