//
//   $Id: TREE.C 1.9 2002/10/16 22:04:26Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  TREE.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//  05-05-2000  bt  Created
//


#define NEWTREE 1


//  Enable debug information
//
#define TREE_DEBUG		FALSE


//  Allow TREE to display the LFN names when is LFN mode
//
#define DISPLAY_LFNS		TRUE


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dosver.h"
#include "dosmsg.c"
#include "fatworld.h"
#include "dlapp.h"
#include "dllfn.h"
#include "stdc.h"
#include "dlpath.h"


//
//  [ TREE HEADER FILES ]
//
#include "tree.h"


//
//  [ PROTOTYPES ]
//
#include "tree.pt"


//
//  [ VARAIABLES ]
//

bool fIsFat32World = FALSE;

//  Flags for the command line switches
//  (accessed through macros in TREE.H)
//
bool fSwitches[MAX_BOOLEAN_SWITCHES];


//  Local copy of graphical characters array
//
char cGraphCharSet[LINE_COUNT];


//  Start of TREE
//
uchar main( int argc, char *argv[] )
{
	uchar ucExitCode = ERR_NONE;


#if TREE_DEBUG

	//  Hide the signon for output comparison with MS output
	//
	fSwitches[HIDE_SIGNON] = 1;

#endif


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	//  Parse the command line
	//
	DoParseCommandLine( argc-1, &argv[1] );


	//  Determine if we should display the signon message and if
	//  so display it
	//
	DisplaySignonMessage();


	//  Display the initial volume information and display
	//  graphically the directory tree
	//
	ucExitCode = Tree();


	//  Set the return code and exit
	//
	return ucExitCode;
}


//  Display the initial volume information and display
//  graphically the directory tree
//
uchar Tree( void )
{
	char * szPath;
	char * szOrgPath;
	uchar ucReturnCode;
	bool  fNoSubdirs;


	//  Get the variable which contains the starting path
	//
	szPath = GetPathVar();
	szOrgPath = GetOrgPathVar();


	//  Display the volume label and and serial number
	//
	DisplayVolumeInfo();
//	printf( szVolumeStr, "" );
//	printf( szSerialStr,  "" );


	//  Display the starting directory
	//
	printf( szPathStr, szOrgPath );


	//  Display graphically the directory tree
	//
	ucReturnCode = DisplayTree( szPath, &fNoSubdirs );
	if( fNoSubdirs )
	{
		//  Display a message indicating that there were no
		//  subdirectories found
		//
		printf( szNoSubdirMsg );


		//  if there is no exit code set, set the one indicating
		//  that no subdirectories were found
		//
		if( !ucReturnCode )
		{
			ucReturnCode = ERR_NO_FILES;
		}
	}


	//  Return the exit code
	//
	return ucReturnCode;
}


//  Display graphically the directory tree
//
uchar DisplayTree( char * szPath, bool * pfNoSubdirs )
{
	uchar ucReturnCode = ERR_NONE;
	static char * szOrgPath;
	static bool fFirstCall = TRUE;
	static bool fTemp;
	static bool fMultipleDirs;
	static bool fAtLeast1Dir;
	bool fSuccessful;
	static bool fSuccessful2;
	static char cGraphics[1024];
	static unsigned uGraphicsIndex = 0;
	static FIND Find;
	static char szLFNName[LFN_MAX_PATH];
	static char szLFNPathName[LFN_MAX_PATH];
	FIND TempFind;


	//  Set the initial setting for the return variable indicating
	//  if any subdirectories were found
	//
	*pfNoSubdirs = TRUE;


	//  Set the DTA to our find structure, so that we can
	//  perform some find firsts
	//
	Int_dlSetDTA( &TempFind );


	//  Determine if this is the first time this function has been called
	//
	if( fFirstCall )
	{
		//  Determine if there are any wildcards in the path
		//
		if( ContainsWildCards(szPath) )
		{
			//  Get the original user path
			//
			szOrgPath = GetOrgPathVar();


			//  Display a message indicating that the path was
			//  invalid
			//
			printf( szInvalidPath, szOrgPath );


			//  Set the exit code
			//
			ucReturnCode = ERR_INVALID_PATH;


			//  Set the flag indicating that no subdirectories were
			//  found (because we did not look for any)
			//
			*pfNoSubdirs = TRUE;


			//  Return the results
			//
			return ucReturnCode;
		}
		else
		{
			//  Get the original user path
			//
			szOrgPath = GetOrgPathVar();


			//  Make sure that the path is valid
			//
			szLFNPathName[0] = szOrgPath[0];
			szLFNPathName[1] = szOrgPath[1];
			szLFNPathName[2] = '\\';
			szLFNPathName[3] = 0;
			SmartGetCurrentDirectory( szOrgPath[0] - 'A' + 1,
					&(szLFNPathName[3]) );
			fSuccessful = SmartChangeDirectory( szOrgPath );
			SmartChangeDirectory( szLFNPathName );
			if( !fSuccessful )
			{
				//  Display a message indicating that the path was
				//  invalid
				//
				printf( szInvalidPath, szOrgPath );


				//  Set the exit code
				//
				ucReturnCode = ERR_INVALID_PATH;


				//  Set the flag indicating that no subdirectories were
				//  found (because we did not look for any)
				//
				*pfNoSubdirs = TRUE;


				//  Return the results
				//
				return ucReturnCode;
			}
		}
	}


	//  Determine if we will be showing files
	//
	if( DoShowFiles() )
	{
		//  Make sure that there is at least 1 subdirectory.  If
		//  there is not at least one subdirectory, MS does not
		//  show file output
		//
		fMultipleDirs = FALSE;
		Int_AddName( szPath, "*.*" );
		fAtLeast1Dir = Int_dlFindFirstFile( A_SUBDIR, szPath );
		Int_RemoveName( szPath );
		if( !fAtLeast1Dir )
		{
			if( fFirstCall )
			{
//				fSwitches[SHOW_FILES] = FALSE;
			}
		}
		else
		{
			//  Get the next file
			//
			fMultipleDirs = Int_dlFindNextFile( A_SUBDIR );
		}
	}


	//  Determine if we will be showing files
	//
	if( DoShowFiles() )
	{
		//  Display the files in this directory
		//
		Int_AddName( szPath, "*.*" );
		fSuccessful = Int_dlFindFirstFile( A_READONLY | A_HIDDEN | A_SYSTEM,
				szPath );
		Int_RemoveName( szPath );


		//  So long as files were found ...
		//
		if( fSuccessful )
		{
			//  Format the graphics output
			//
			if( fAtLeast1Dir || fMultipleDirs )
			{
				cGraphics[uGraphicsIndex] = cGraphCharSet[LINE_UD];
			}
			else
			{
				cGraphics[uGraphicsIndex] = cGraphCharSet[LINE_WHITE];
			}
			cGraphics[uGraphicsIndex + 1] = cGraphCharSet[LINE_WHITE];
			cGraphics[uGraphicsIndex + 2] = cGraphCharSet[LINE_WHITE];
			cGraphics[uGraphicsIndex + 3] = cGraphCharSet[LINE_WHITE];


			//  Continue retrieving entries for files until there
			//  are no more
			//
			while( fSuccessful )
			{
				//  Format the graphics output and display the string
				//
				cGraphics[uGraphicsIndex + 4] = cGraphCharSet[LINE_TERM];
#if DISPLAY_LFNS
				if( DriveSupportsLFNs(szPath[0] - 'A' + 1) )
				{
					Int_AddName( szPath, TempFind.name );
					fSuccessful = LFNGetFullPath( FULLPATH_LONGNAME,szPath,
							szLFNPathName );
					if( fSuccessful )
					{
						CopyNamePart( szLFNName, szLFNPathName );
						strcat( cGraphics, szLFNName );
					}
					else
					{
						strcat( cGraphics, TempFind.name );
					}
					Int_RemoveName( szPath );
				}
				else
				{
					strcat( cGraphics, TempFind.name );
				}
#else
				strcat( cGraphics, TempFind.name );
#endif
				strcat( cGraphics, "\n" );
				printf( cGraphics );


				//  Get the next file
				//
				fSuccessful = Int_dlFindNextFile(
						A_READONLY | A_HIDDEN | A_SYSTEM );
			}


			//  Format the graphics output and display a blank line
			//
			cGraphics[uGraphicsIndex + 4] = cGraphCharSet[LINE_TERM];
			strcat( cGraphics, "\n" );
			printf( cGraphics );
		}
	}


	//  Perform the find first for the subdirectories
	//
	Int_AddName( szPath, "*.*" );
	fSuccessful = Int_dlFindFirstFile( A_SUBDIR, szPath );
	Int_RemoveName( szPath );


	//  So long as a subdirectory was found ...
	//
	if( fSuccessful )
	{
		//  Clear the flag indicating that this is the first
		//  time this function has been called
		//
		fFirstCall = FALSE;


		//  Continue retrieving and recursing to process the subdirectories
		//
		while( fSuccessful )
		{
			//  Copy the find structure so that we can use it again
			//  to look at the next entry.
			//
			memcpy( &Find, &TempFind, sizeof(FIND) );


			//  Get the next entry on this level
			//
			Int_dlSetDTA( &TempFind );
			fSuccessful = Int_dlFindNextFile( A_SUBDIR );


			//  Display the graphical representation of this directory
			//
			if( fSuccessful )
			{
				cGraphics[uGraphicsIndex] = cGraphCharSet[LINE_UDR];
			}
			else
			{
				cGraphics[uGraphicsIndex] = cGraphCharSet[LINE_UR];
			}
			cGraphics[uGraphicsIndex + 1] = cGraphCharSet[LINE_LR];
			cGraphics[uGraphicsIndex + 2] = cGraphCharSet[LINE_LR];
			cGraphics[uGraphicsIndex + 3] = cGraphCharSet[LINE_LR];
			cGraphics[uGraphicsIndex + 4] = cGraphCharSet[LINE_TERM];
#if DISPLAY_LFNS
			if( DriveSupportsLFNs(szPath[0] - 'A' + 1) )
			{
				Int_AddName( szPath, Find.name );
				fSuccessful2 = LFNGetFullPath( FULLPATH_LONGNAME,szPath,
						szLFNPathName );
				if( fSuccessful2 )
				{
					CopyNamePart( szLFNName, szLFNPathName );
					strcat( cGraphics, szLFNName );
				}
				else
				{
					strcat( cGraphics, Find.name );
				}
				Int_RemoveName( szPath );
			}
			else
			{
				strcat( cGraphics, Find.name );
			}
#else
			strcat( cGraphics, Find.name );
#endif
			strcat( cGraphics, "\n" );
			printf( cGraphics );


			//  Set the flag indicating that subdirectories were found
			//
			*pfNoSubdirs = FALSE;


			//  Reset the graphical display for the next entry
			//
			if( fSuccessful )
			{
				cGraphics[uGraphicsIndex]     = cGraphCharSet[LINE_UD];
			}
			else
			{
				cGraphics[uGraphicsIndex]     = cGraphCharSet[LINE_WHITE];
			}
			cGraphics[uGraphicsIndex + 1] = cGraphCharSet[LINE_WHITE];
			cGraphics[uGraphicsIndex + 2] = cGraphCharSet[LINE_WHITE];
			cGraphics[uGraphicsIndex + 3] = cGraphCharSet[LINE_WHITE];
			cGraphics[uGraphicsIndex + 4] = cGraphCharSet[LINE_WHITE];
			cGraphics[uGraphicsIndex + 5] = cGraphCharSet[LINE_TERM];


			//  Set the next location in the graphics display
			//
			uGraphicsIndex += LINE_INDENT;


			//  Recurse to display the next level of subdirectories
			//
			Int_AddName( szPath, Find.name );
			DisplayTree( szPath, &fTemp );
			Int_RemoveName( szPath );


			//  Revert back to our previous location the graphics display
			//
			uGraphicsIndex -= LINE_INDENT;


			//  Reset the graphical display for the next entry
			//
//			cGraphics[uGraphicsIndex]     = cGraphCharSet[LINE_UD];
//			cGraphics[uGraphicsIndex + 1] = cGraphCharSet[LINE_TERM];
		}
	}


	//  Return whether the function was successful
	//
	return ucReturnCode;
}


//  Determine if there are any wildcards in the path
//
bool ContainsWildCards( char * szPath )
{
	while( *szPath )
	{
		if( (*szPath == '*') || (*szPath == '?') )
		{
			return TRUE;
		}
		szPath++;
	}

	return FALSE;
}


//  Process all the command line arguments
//
void DoParseCommandLine( int argc, char *argv[] )
{
	char * szPath;
	char * szOrgPath;
	bool fSuccessful;
	int  iDrive;
	static ARGDEF TreeOpts[] =
	{
		{ BOOL_SWITCH, &fSwitches[HIDE_SIGNON],      '[', 0 },
		{ BOOL_SWITCH, &fSwitches[ASCII_OUTPUT],     'a', 0 },
		{ BOOL_SWITCH, &fSwitches[SHOW_FILES],       'f', 0 },
		{ 0 }
	};


	//  Process all the boolean switches in the command line
	//
	SetParseFlags(PARSEF_NOARGSOK);
	argc = ParseCmdline(argc, &argv[0], TreeOpts);


	//  We should not have more than one path
	//
	if(argc > 1)
	{
		OptError( argv[1] );
//		Usage();
	}


	//  Set the path for tree
	//
	szPath    = GetPathVar();
	szOrgPath = GetOrgPathVar();
	if( argc == 1 )
	{
		//  Determine if the supplied path contains a drive letter
		//
		szPath[0] = 0;
		if( argv[0][1] != ':' )
		{
			//  Get the current drive
			//
			dlGetCurrentDrive( &iDrive );
			szPath[0] = 'A' + (uchar)iDrive;
			szPath[1] = ':';
			szPath[2] = 0;
		}
		else
		{
			if( (szPath[0] >= 'a') && (szPath[0] >= 'z') )
			{
				szPath[0] -= 'a' - 'A';
			}
		}


		//  Copy the user path
		//
		strcat( szPath, argv[0] );
		szPath[0] = toupper( szPath[0] );


		//  Save the original path specified on the command line
		//
		strcpy( szOrgPath, szPath );
		if( szOrgPath[2] == 0 )
		{
			strcat( szOrgPath, "." );
		}
	}
	else
	{
		//  Get the current drive and directory
		//
		SmartExpandPath( 0, ".", szPath );


		//  Get the current drive
		//
		dlGetCurrentDrive( &iDrive );


		//  Save the original path specified on the command line as
		//  the current drive and directory
		//
		szOrgPath[0] = 'A' + (uchar)iDrive;
		szOrgPath[1] = ':';
		szOrgPath[2] = '.';
		szOrgPath[3] = 0;
	}


	//  Initialize the graphics character set
	//
	if( DoAsciiOutput() )
	{
		//  Set the graphics characters for non-extended ascii
		//
		cGraphCharSet[LINE_UD]  = '|';
		cGraphCharSet[LINE_UDR] = '+';
		cGraphCharSet[LINE_LR]  = '-';
		cGraphCharSet[LINE_UR]  = '\\';
	}
	else
	{
		//  Set the graphics characters for extended ascii
		//
		cGraphCharSet[LINE_UD]  = (char)0xb3;
		cGraphCharSet[LINE_UDR] = (char)0xc3;
		cGraphCharSet[LINE_LR]  = (char)0xc4;
		cGraphCharSet[LINE_UR]  = (char)0xc0;
	}
	cGraphCharSet[LINE_WHITE] = ' ';
	cGraphCharSet[LINE_TERM]  = 0;


	//  Convert the original path specified on the command line to uppercase
	//
	while( *szOrgPath )
	{
		*szOrgPath = toupper( *szOrgPath );
		szOrgPath++;
	}
}


//  Display the usage screen and exit
//
void Usage( void )
{
	DisplaySignonMessage();
	printf( szUsage );
	exit(ERR_USAGE);
}


//  Determine if we should display the signon message and if
//  so display it
//
void DisplaySignonMessage( void )
{
	char * szSignOnMessage = "TREE %s %s\n%s\n\n";

	if( !DoHideSignon() )
	{
		if( fIsFat32World )
		{
			printf(szSignOnMessage,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(szSignOnMessage,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}


//  Return the address of the variable which contains the
//  starting/root path
//
char * GetPathVar( void )
{
	static char szPath[LFN_MAX_PATH];
	return szPath;
}


//  Return the address of the variable which contains the
//  original path specified on the command line
//
char * GetOrgPathVar( void )
{
	static char szOrgPath[LFN_MAX_PATH];
	return szOrgPath;
}


//  Return the address of the variable which contains the
//  address of the DTA
//
PFIND * GetFindVar( void )
{
	static PFIND pFind;
	return &pFind;
}


//  Add a name to the current path
//
//  NOTE: This is different that AddName() because AddName when passed
//        a path of just t: will put a backslash after the colon before
//        adding the name.
//
bool Int_AddName( char * szPath, const char * szName)
{
	bool fReturn = TRUE;


	if( (szPath[1] == ':') && (szPath[2] == 0) && (szPath[0]) )
	{
		strcat( szPath, szName );
	}
	else
	{
		fReturn = AddName( szPath, szName );
	}


	return fReturn;
}


//  Remove a name from the path
//
bool Int_RemoveName(char * szPath)
{
	char * szName;


	if( !RemoveName(szPath) )
	{
		szName = LastPart( szPath );
		if( *szName )
		{
			*szName = 0;
			return TRUE;
		}
	}

	return TRUE;
}


//  Record the address that the DTA is being set to and set the DTA
//
bool PASCAL Int_dlSetDTA(PFIND pNewFind)
{
	PFIND * ppFind;


	ppFind = GetFindVar();
	*ppFind = pNewFind;
	return dlSetDTA( pNewFind );
}


//  Perform a find first and only return the entries which match
//  the search mask.  Also filter the '.' and '..' entries.
//
bool PASCAL Int_dlFindFirstFile(uchar ucAttributes, char * szPathAndName)
{
	PFIND pFind;
	bool fSuccessful;


	//  Perform the find first
	//
	fSuccessful = dlFindFirstFile( ucAttributes, szPathAndName );
	if( fSuccessful )
	{
		//  Get the find structure address
		//
		pFind = *( GetFindVar() );


		//  Determine if this entry matches the search parameters
		//
		while( fSuccessful )
		{
			//  Determine if these attributes are what was expected
			//
			if( (ucAttributes & A_SUBDIR) == (pFind->attr & A_SUBDIR) )
			{
				//  Make sure that this is not the '.' and '..' entries
				//
				if( pFind->name[0] != '.' )
				{
					break;
				}
			}


			//  Find the next occurence
			//
			fSuccessful = dlFindNextFile();
		}
	}


	//  Return whether any entries were found matching the search
	//  mask
	//
	return fSuccessful;
}


//  Find the next entry which matches the search mask and is not
//  the '.' and '..' entries
//
bool PASCAL Int_dlFindNextFile( uchar ucAttributes )
{
	PFIND pFind;
	bool fSuccessful;


	//  Get the find structure address
	//
	pFind = *( GetFindVar() );


	//  Determine if this entry matches the search parameters
	//
	do
	{
		//  Find the next occurence
		//
		fSuccessful = dlFindNextFile();


		//  Determine if these attributes are what was expected
		//
		if( (ucAttributes & A_SUBDIR) == (pFind->attr & A_SUBDIR) )
		{
			//  Make sure that this is not the '.' and '..' entries
			//
			if( pFind->name[0] != '.' )
			{
				break;
			}
		}

	}while( fSuccessful );


	//  Return whether any entries were found matching the search
	//  mask
	//
	return fSuccessful;
}



//  Display information about the disk like the disk label
//  and serial number.
//
void DisplayVolumeInfo( void )
{
	static DISK_INFO DiskInfo;
	static XFCB XFCB;
	static char szVolume[12];
	bool fSuccessful;
	char * szTemp;
	uchar ucDrive;
	char * szPath;


	//  Get the path which contains the drive letter
	//
	szPath = GetPathVar();


	//  Determine the drive number
	//
	ucDrive = szPath[0] - 'A';


	XFCB.uXFCBIndicator = 0xff;
	XFCB.uAttributes = A_VOLUME;
	XFCB.drive = ucDrive + 1;
	memset( &XFCB.name, '?', 11 );

	dlSetDTA( &XFCB );
	fSuccessful = dlFcbFindFirst( (char *)(&XFCB));
	if( fSuccessful )
	{
		memcpy( szVolume, &XFCB.name, 11 );
		szVolume[11] = 0;

		//  remove any white space from the end of the volume label
		//
		szTemp = &(szVolume[10]);
		while( szTemp != szVolume )
		{
			if( *szTemp != ' ' )
			{
				break;
			}
			*szTemp = 0;
			szTemp--;
		}


		printf( szVolumeStr, szVolume );
	}


	//  Get the serial number
	//
	if( fIsFat32World )
	{
		dlGenericIOCTLForBlockDevices( ucDrive+1,
				0x48, 0x66,	&DiskInfo );
	}
	else
	{
		dlGenericIOCTLForBlockDevices( ucDrive+1,
				0x08, 0x66,	&DiskInfo );
	}



	//  Display the volume serial number
	//
	printf( szSerialStr,
			(unsigned)((DiskInfo.ulSerialNumber & 0xFFFF0000L) >> 16),
			(unsigned)((DiskInfo.ulSerialNumber & 0x0000FFFFL)) );
}


//  Display invalid switch message and exit
//
void OptError(char * szMsg)
{
	printf( szInvalidSwitch, szMsg );
	exit( ERR_INVALID_SWITCH );
}
