//
//   $Id: XCFILE.C 1.24 2003/01/28 19:16:00Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*--------------------------------------------------------

   XCFILE.C  created 03/04/99  by John Burnham

	Copyright (c) 1999-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This module contains the file routines for XCOPY.

   routines contained:
      void StoreFile(char *pSrcFilename, char *pDestFilename,
                     unsigned char ucAttr, unsigned uDate, unsigned uTime);
      void FlushFileStore(void);
      long DetermineFilesize(FILE *fp);
      bool ShouldCopyFile(char *pDest, bool exists);
      char GetAnswer(char *validResponse);
      void HandleWriteError(unsigned uError, char *pFile);
      bool bFileExist( char * szFileName )

---------------------------------------------------------*/
#define __FILENUM__ 5
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <dir.h>
#include <dos.h>
#include <ctype.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"

#include "dlmem.h"
#include "dlpath.h"
#include "dlerrcod.h"
#include "xcopy.h"
#include "dlgetkey.h"

// prototype includes
#include "xcfile.pt"
#include "utilerr.pt"

// external data
extern bool copyArchiveAndKeep;
extern bool copyArchiveAndClear;
extern unsigned copyAfterDate;
extern int iSrcRootLength;
extern char szFullSourcePath[];
extern char szSourceMask[];
extern bool fAutoOverwrite;
extern char askOverwrite[];
extern char overwriteOptStr[];
extern char overwriteOptions[];
extern bool promptCreateFile;
extern bool fPreserveReadOnlyAttribute;
extern bool fQuiteMode;
extern bool fShowFullPath;
extern char ask_create[];
extern char createOptStr[];
extern char createOptions[];
extern unsigned uFileCount;
extern char reading_str[];
#if !ENABLE_DIRECT_MEMSTO_COPIES
extern char *pDiskIOBuff;
#endif
extern bool fIsFat32World;


extern char szTempName[];
extern uchar ucSrcDOSDriveNumber;
extern uchar ucDestDOSDriveNumber;

extern int iSrcRootLength;				// length of source path (up to filename)
extern int iDestRootLength;			// length of destination path (up to filename)

#if CHANGE_DIRECTORY_TO_COPY

#else

extern char szCompleteName[LFN_MAX_PATH];

#endif

extern char *pCompleteFile;         // pointer to full filename+path

// global data for this file
bool fFilePrompt = FALSE;


bool fFlushedFileStore = FALSE;


/*
	Actually readin and store the given file.  This routine is a part of
	a state-machine that knows about disks, destinations, and memory store.

	This MAY need to flush data to disk(s) via FlushFileStore() if
	this file exceeds the in-memory store.
*/
bool StoreFile(char *pSrcFilename, char *pDstFilename, unsigned char ucAttr,
					unsigned uDate, unsigned uTime, unsigned long ulSize)
{
	ulong ulCurFilePos;
	struct FileRecord frCurFileRecord;		// file record for this file
	ulong ulRecordSize, ulCurRecordPos;			// size of data in record
	unsigned uSrcNameSize;								// part of record header
	unsigned uDstNameSize;								// part of record header
	unsigned uLen, uLenRequested;
	int iHandle;									// handle to open the file
   int iActionTaken;                      // return value
	ulong ulTemp;


#if 0
   printf("<%s>",pSrcFilename);
#endif

	// validate parameters
	ASSERT(pSrcFilename && *pSrcFilename);	// source must exist
	ASSERT(pDstFilename && *pDstFilename);	// destination must also exist
	ASSERT(!(attr & NOTFILE))					// don't copy dirs or labels

	// only copy those files with the archive attribute set
	if ((copyArchiveAndKeep || copyArchiveAndClear) &&
		 !(ucAttr & A_ARCHIVE))
	{
		return FALSE;
	}


	// only copy those files after date
	if (uDate < copyAfterDate)
	{
		return FALSE;
	}


	//  If the file store was just flushed ...
	//
	if( fFlushedFileStore )
	{
		//  Display a message that we are reading files
		//
		printf( reading_str );


		//  Clear the flag indicating that the file store was flushed
		//
		fFlushedFileStore = FALSE;
	}

#if CHANGE_DIRECTORY_TO_COPY
   pCompleteFile = pSrcFilename;
#else
   // tack on the full pathname for the upcoming open call
   //
   strcpy( szCompleteName, szFullSourcePath );
   if(strlen(szCompleteName) > 3) strcat( szCompleteName, "\\" );
   strcat( szCompleteName, pSrcFilename );
   pCompleteFile = szCompleteName;
#endif

#if 0
   printf("[store: %s]",pCompleteFile);
#endif

	// open the file
	if (SmartCreateOpenFile( O_RDONLY, A_NORMAL | A_READONLY | A_HIDDEN |
			A_SYSTEM, ACT_EXIST_OPEN, pCompleteFile, &iHandle,
			&iActionTaken) == FALSE)
   {
		error(ERR_FILENOTFOUND,pSrcFilename);
   }

	// get a copy of the actual file name, not the "root" of
	// the name, just the path part.
	uSrcNameSize = strlen(pSrcFilename) + 1;
	uDstNameSize = strlen(pDstFilename) + 1;

	frCurFileRecord.sig = FILE_RECORD_SIG;
	frCurFileRecord.attr = ucAttr;
	frCurFileRecord.date = uDate;
	frCurFileRecord.time = uTime;
	frCurFileRecord.size = ulSize;
	frCurFileRecord.uSrcNameSize = uSrcNameSize;
	frCurFileRecord.uDstNameSize = uDstNameSize;

	// start at the beginning and store the whole file
	ulCurFilePos = 0L;

	do
	{
		// remember where this file block start
		// (may not be entire file, as the file could be bigger
		// than the memory store).
		frCurFileRecord.startOffset = ulCurFilePos;

		// not enought room for header and some data?  flush file store
		if( MemStoreLeft() < MIN_RECORD_SIZE )
		{
			FlushFileStore();
		}

		// determine the maximum size of write
		// it will be a minimum of:
		//		1. file size
		//		2. bytes left in memory store
		ulRecordSize = MemStoreLeft() - sizeof(frCurFileRecord) -
				uSrcNameSize - uDstNameSize ;
		ulTemp = ulSize - ulCurFilePos;
		if( ulTemp < ulRecordSize )
		{
			ulRecordSize = ulTemp;
		}

		// write the file record to the memory store
		frCurFileRecord.next = MemStoreTell() + sizeof(frCurFileRecord) +
				uSrcNameSize + uDstNameSize + ulRecordSize;
		MemStoreWrite(&frCurFileRecord, sizeof(frCurFileRecord) );
		MemStoreWrite(pSrcFilename, uSrcNameSize);
		MemStoreWrite(pDstFilename, uDstNameSize);

		// write data to the store
		ulCurRecordPos = 0L;
		while (ulCurRecordPos < ulRecordSize)
		{
			// read in as much as we can from the file
			if (ulRecordSize - ulCurRecordPos < DISKIO_BUFFSIZE)
				uLen = (unsigned)(ulRecordSize - ulCurRecordPos);
			else
				uLen = DISKIO_BUFFSIZE;

			// read some bytes from the file
			uLenRequested = uLen;
#if !ENABLE_DIRECT_MEMSTO_COPIES
			dlReadFile(iHandle, &uLen, pDiskIOBuff);
#else
			dlReadFile(iHandle, &uLen, GetCurrentStoreAddress() );
			MemStoreSeek(uLen, SEEK_CUR);
#endif
			// if cannot read all that was requested, we cannot read from
			// file, so exit...
			if (uLen != uLenRequested)
				error(ERR_ERRORREADINGFILE);

#if !ENABLE_DIRECT_MEMSTO_COPIES
			// write them to the memory store
			MemStoreWrite(pDiskIOBuff, uLen);
#endif

			// we wrote that part of the file, on to the next
			ulCurRecordPos += uLen;
		}

		// update file position
		ulCurFilePos += ulRecordSize;

	} while (ulCurFilePos < ulSize);

	// done with file
	dlCloseFile(iHandle);

	return TRUE;
}


/*
	Flush in-memory file store to disk.

	Called during an XCOPY when in-memory store is full.  Also called
	at the of XCOPY to flish the final in-memory store.
*/
void FlushFileStore(void)
{
	static bool fFileOpened = FALSE;			// determines if we are finishing
														// a write to an existing file
	static int iHandle;							// handle to file to write
	struct FileRecord frCurFileRecord;		// file record for a file
	static char szSrcFilename[LFN_MAX_PATH]; // source filename
	static char szDstFilename[LFN_MAX_PATH]; // destination filename
	ulong ulMemTop;								// the memory position of file data end
	long lCurMemPos;								// current position in memory block
	long lCurStopPos;								// position in memory block to stop
	ulong ulCurFilePos;							// current position in file
	unsigned uWriteAmount;						// the amount of data to write
	unsigned uWriteAmountRequested;			// the amount of data wanted to write
	unsigned long ulDiskSpace;					// the amount of free disk space
	unsigned ucSrcAttr;							// attributes of source file
	static bool fCopyFile = TRUE;				// should this file be copied?
   int iActionTaken;                      // return values
	ulong ulBytesToWrite;
	bool Temp_promptCreateFile;
	bool fSuccessful;
	extern bool fSameDrive;
	char * szName;
	extern char szFullDestPath[];
	bool fIsDevice = FALSE;
	unsigned uDevAttributes;
	bool fAttrSuccessful;
	unsigned uAttributes;


	//  Get the used size of the memory store
	//
	ulMemTop = MemStoreTell();


	//  start at the beginning of memory
	//  and keep writing files until we have emptied the
	//  memory store.
	//
	MemStoreSeek(0L, SEEK_SET);


	//  while not at top of memory store
	//
	while( MemStoreTell() < ulMemTop )
	{
		//  read the header
		//
		MemStoreRead(&frCurFileRecord, sizeof(frCurFileRecord) );


		//  validate the header
		//
		SHIPASSERT(frCurFileRecord.sig == FILE_RECORD_SIG);


		//  read the filenames
		//
		MemStoreRead(szSrcFilename, frCurFileRecord.uSrcNameSize);
		MemStoreRead(szDstFilename, frCurFileRecord.uDstNameSize);


		//  get current position in memory store
		//
		lCurMemPos = MemStoreTell();


		//  get stop position in memory store
		//
		lCurStopPos = frCurFileRecord.next;


		//  get current file position
		//
		ulCurFilePos = frCurFileRecord.startOffset;


		//  Determine if the file is already open
		//
		if( (!fFileOpened) && (ulCurFilePos == 0) )
		{
			// if promting before creating files
			if( promptCreateFile == TRUE )
			{
				fCopyFile = ShouldCopyFile( szDstFilename, FALSE );
			}


			//  Determine if the user wants to create this file
			//
			if( fCopyFile )
			{
				//  Determine the file attributes of the destination file
				//
				ucSrcAttr = frCurFileRecord.attr;
				if(!fPreserveReadOnlyAttribute)
				{
					//  Clear read-only bit if not told otherwise
					//
					ucSrcAttr = ucSrcAttr & ~A_READONLY;
				}


				//  Set a flag indicating that the file store was flushed
				//
				fFlushedFileStore = TRUE;


				//  Create the file
				//
				if( fSameDrive )
				{
					AddName( szFullDestPath, szDstFilename);
					szName = szFullDestPath;
				}
				else
				{
#if CHANGE_DIRECTORY_TO_COPY
	   			szTempName[0] = ucDestDOSDriveNumber + 'A';
	   			szTempName[1] = ':';
					strcpy( &szTempName[2], szDstFilename );
					szName = szTempName;
#else
					AddName( szFullDestPath, szDstFilename);
					szName = szFullDestPath;
#endif
				}

#if 0
   printf("[flush: %s]",szName);
#endif
				fSuccessful = SmartCreateOpenFile(O_WRONLY, A_NORMAL,
	               ACT_NEXIST_CREATE | ACT_EXIST_OPEN, szName,
						&iHandle, &iActionTaken );
				if( fSuccessful )
				{
					// check to make sure that the file is not a reserved device
					// pb
					fIsDevice = dlGetDeviceAttributes( iHandle, &uDevAttributes );
					if( (fIsDevice) && (uDevAttributes & DEVA_QUERY) )
					{
						error(ERR_TORESERVEDDEVICE);
					}


					//  Set a flag indicating that the file is opened (or created)
					//
					fFileOpened = TRUE;


					//  Determine if the file already existed
					//
					if( iActionTaken == ACTION_OPENED )
					{
						//  We don't want to overwrite files with special
						//  attributes
						//
						fAttrSuccessful = SmartGetFileAttributes( szName,
								&uAttributes );
						if( fAttrSuccessful )
						{
							if( uAttributes & (A_SYSTEM|A_READONLY|A_HIDDEN) )
							{
								error(ERR_ACCESSDENIED);
							}
						}


						//  Determine if the user wants to overwrite the file
						//
 						Temp_promptCreateFile = promptCreateFile;
						promptCreateFile = FALSE;
						fCopyFile = ShouldCopyFile( szDstFilename,
								TRUE );
						promptCreateFile = Temp_promptCreateFile;
						if( !fCopyFile )
						{
							//  Close the file
							//
							dlCloseFile(iHandle);


							//  Clear the flag that indicates that a file is open
							//
							fFileOpened = FALSE;
						}
						else
						{
							//  Truncate the file
							//
							//  Note: Buffer address is irrelevent for a
							//        zero byte write
							//
							uWriteAmount = 0;
#if !ENABLE_DIRECT_MEMSTO_COPIES
							dlWriteFile(iHandle, &uWriteAmount, pDiskIOBuff);
#else
							dlWriteFile(iHandle, &uWriteAmount,
									(char far *)(&iHandle) );
#endif

							//  Set attributes
							//
			            SmartSetFileAttributes( szName, ucSrcAttr);
						}
					}
				}
				else
				{
					//  Check for an access denied error
					//
					if( dlCheckDOSError() == ERR_DENIED )
					{
						error(ERR_ACCESSDENIED);
					}
					else
					{
						error(ERR_CANNOTCREATE, szDstFilename);
					}
	         }


#if CHANGE_DIRECTORY_TO_COPY
				if( fSameDrive )
#endif
				{
					szFullDestPath[iDestRootLength] = 0;
				}

			}
		}


		//  Determine if the user wants to copy this file
		//
		if( !fCopyFile )
		{
			// move to next file
			MemStoreSeek(frCurFileRecord.next, SEEK_SET);

			// 11/18/2002 PB
			// IR 4118
			//
			// Reset the copy file flag for the next file.
			// This was not being done therefore when we
			// would go to copy the next file the flag said
			// we didn't want to copy the next file. This would
			// happen for all the files in the file store.
			// Now we set the fCopyFile flag back to true for each
			// file.
			//
			fCopyFile = TRUE;

			continue;
		}
		else
		{
         if( ulCurFilePos == 0 )
         {
            if( !fQuiteMode )
            {
               // build source name
               AddName(szFullSourcePath, szSrcFilename);

               // print the filename
               printf("%s\n", szFullSourcePath );

               // restore source path to original settings
               szFullSourcePath[iSrcRootLength] = '\0';
            }
         }
		}


		//  Determine the number of bytes to write
		//
		ulBytesToWrite = lCurStopPos - lCurMemPos;


		//  Write the file
		//
		while( ulBytesToWrite )
		{
			//  Determine the write size
			//
			if( ulBytesToWrite < DISKIO_BUFFSIZE )
			{
				uWriteAmount = (unsigned)ulBytesToWrite;
			}
			else
			{
				uWriteAmount = DISKIO_BUFFSIZE;
			}


#if !ENABLE_DIRECT_MEMSTO_COPIES
			//  read from memory store
			//
			MemStoreRead(pDiskIOBuff, uWriteAmount);
#endif


			//  write it to disk
			//
			uWriteAmountRequested = uWriteAmount;
#if !ENABLE_DIRECT_MEMSTO_COPIES
			fSuccessful = dlWriteFile(iHandle, &uWriteAmount, pDiskIOBuff);
#else
			fSuccessful = dlWriteFile(iHandle, &uWriteAmount,	GetCurrentStoreAddress() );
			MemStoreSeek(uWriteAmount, SEEK_CUR);
#endif


			//  Determine if the requested amount was written
			//
			if( (fSuccessful) && (uWriteAmountRequested == uWriteAmount) )
			{
 				ulCurFilePos += uWriteAmount;
				ulBytesToWrite -= uWriteAmount;
			}
			else
			{
				//  Determine the amount of space available on the DOS drive
				//
				dlGetFreeSpaceOnDrive( ucDestDOSDriveNumber + 1,
						&ulDiskSpace);


				//  Determine if the file size exceeds the available disk
				//  space
				//
				if( frCurFileRecord.size > ulDiskSpace )
				{
					dlSetCurrentDrive( (int)ucDestDOSDriveNumber );
					dlDeleteFile( szDstFilename );

					error(ERR_DISKSPACE);
				}
				else
				{
					AddName(szFullSourcePath, szSrcFilename);
					printf("Error copying file %s to %s\n", szFullSourcePath,
									szDstFilename );
//					RemoveName( szFullSourcePath );
					szFullSourcePath[iSrcRootLength] = '\0';
					HandleWriteError( dlCheckDOSError(),
							szDstFilename );
				}
			}
		}


		//  once the entire file is written, then we can count it...
		//
		if( ulCurFilePos == frCurFileRecord.size )
		{
			//  Update the file date and time
			//
			SetFileDate(iHandle, frCurFileRecord.date, frCurFileRecord.time);


			//  Update the file count
			//
			uFileCount++;


			//  Close the file
			//
			dlCloseFile(iHandle);


			//  Clear the flag that indicates that a file is open
			//
			fFileOpened = FALSE;


			// clear the archive bit of the source file if requested
			//
			if( copyArchiveAndClear )
			{
				//  get attributes, clear ARCHIVE bit, and set attributes back
				//
				ucSrcAttr = frCurFileRecord.attr;
 				ucSrcAttr = ucSrcAttr & ~A_ARCHIVE;
				SmartSetFileAttributes( szSrcFilename, ucSrcAttr);
			}
		}
	}


	//  reset memory store to the begining of memory after flush...
	//
	MemStoreSeek(0L, SEEK_SET);
}


/*
	determines if this file should be copied to the destination.
*/
bool ShouldCopyFile(char *pDest, bool exists)
{
	char cUserResponse;
	bool fDoCopy = TRUE;

	// validate params
	ASSERT(pSrc);
	ASSERT(pDest);

	if (promptCreateFile == TRUE)	// if promting before creating files
	{
		fFilePrompt = TRUE;
		printf(ask_create, pDest, createOptStr);	// ask user
		cUserResponse = GetAnswer(createOptions);
		if (cUserResponse == 'Y')
			fDoCopy = TRUE;
		else
			fDoCopy = FALSE;
	}
	else
		fFilePrompt = FALSE;

	if( fDoCopy )
	{
		if (exists == TRUE)			// if the destination file already exists
		{
			if (fAutoOverwrite == TRUE)
				fFilePrompt = FALSE;
			else
			{
				fFilePrompt = TRUE;
				printf(askOverwrite, pDest, overwriteOptStr);	// ask user
				cUserResponse = GetAnswer(overwriteOptions);

				switch (cUserResponse)
				{
					case 'Y':
						fDoCopy = TRUE;
						break;
					case 'N':
						fDoCopy = FALSE;
						break;
					default:
						fDoCopy = TRUE;
						fAutoOverwrite = TRUE;
						break;
				}
			}
		}
		else
		{
			fDoCopy = TRUE;			// the file doesn't exist
		}
	}


	return fDoCopy;
}

/*
	takes a list of valid user response chars (e.g., "YN", "YNA") and
	waits until the user enters one of those chars.  Returns the
	validated user response.
*/
char GetAnswer(char *validResponse)
{
	unsigned char c;

	//validate params
	ASSERT(validResponse);
	while(1)
	{
		c = StdinGetKey();
		printf("%c", c);
		c = toupper( c );
		if (strchr(validResponse, c) == 0)
			printf("%c %c", 0x08, 0x08);
		else
		{
			printf("\n");
			break;
		}
	}
	return c;
}

/*
	HandleWriteError will display a particular error message depending
	on the DOS errorcode passed to the routine.

	This routine DOES NOT return.
*/
void HandleWriteError(unsigned uError, char *pFile)
{
	unsigned long ulDiskSpace;

	// validate parameters
	ASSERT(pFile);

	switch (uError)
	{
		case 2:
			error(ERR_FILENOTFOUND, pFile);
			break;

		case 3:
			error(ERR_INVALIDDESTINATION);
			break;

		case 5:
  			error(ERR_ACCESSDENIED);
			break;

		case 8:
			error(ERR_INSUFFICENTMEMORY);
			break;

		default:
			dlGetFreeSpaceOnDrive( *pFile-'A'+1, &ulDiskSpace);
			if (ulDiskSpace == 0)
				error(ERR_DISKSPACE);
			else
				error(ERR_IOERROR);
			break;
	}
}
