//
//   $Id: XCOPY.C 1.45 2003/01/28 19:16:46Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*--------------------------------------------------------------

	XCOPY.C Created 02-24-99 John Burnham

	Copyright (c) 1989-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary

	This is the main source file for the XCOPY.EXE utility.


--------------------------------------------------------------*/

#define __FILENUM__ 1
#define	XCOPY 1

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"

#include "dosver.h"
#include "dlapp.h"
#include "dlpath.h"
#include "dlerrcod.h"
#include "dldate.h"
#include "fatworld.h"
#include "dlgetkey.h"
#include "stdc.h"
#include "dlmem.h"
#include "dosmsg.c"

#include "xcopy.h"

// prototype includes
#include "xcopy.pt"
#include "xcstore.pt"
#include "xcfile.pt"
#include "utilerr.pt"
#include "xcutil.pt"


unsigned _stklen = 2048;


bool copyArchiveAndKeep  = FALSE;// Copy only those files with the Archive
											// attribute set. Keep the Archive attribute
											// set after the copy.  Defaults to FALSE
bool copyArchiveAndClear = FALSE;// Copy only those files with the Archive
											// attribute set.  Clear the Archive attribute
											// after the copy.  Defaults to FALSE.
unsigned copyAfterDate = 0;		// copy files created on or after date
bool promptCreateFile = FALSE;	// prompt for creation of each file copied.
bool copyDirs = FALSE;				// create dirs and subdirs if NOT empty and
											// copy files to correct directories.
bool copyEmptyDirs = FALSE;		// create dirs and subdirs even if empty and
											// copy files to correct directories.
bool verifyCopy = FALSE;			// verify if copy was successful.
bool waitBeforeCopy = FALSE;		// force user to press a key to begin copy
bool fAutoOverwrite= FALSE;		// this is turned on by the /Y switch and
											// back off with /-Y.  May also be alter by
											// the value of the COPYCMD env_var.
bool fCopyHiddenSystem = FALSE;	// Copy hidden and system files also
bool fPreserveReadOnlyAttribute = FALSE;  // Don't clear the read-only attribute
bool fQuiteMode = FALSE;			//  Don't display the names of copied files
bool fShowFullPath = FALSE;		//  Display full path of source and destination
bool fHideSignon = FALSE;			// Hide the program signon message

#if !ENABLE_DIRECT_MEMSTO_COPIES
char *pDiskIOBuff;					// pointer to disk buffer
#endif
int verify_flag;						// previous verify setting

unsigned uFileCount = 0;			// number of files copied!

char fileParam[LFN_MAX_PATH];		      // source and dest specs
char szFullSourcePath[LFN_MAX_PATH];	// full source path
char szFullDestPath[LFN_MAX_PATH];	   // full destination path
char szSourceMask[LFN_MAX_PATH];		   // source copy mask
char szDestMask[LFN_MAX_PATH];			// destination copy mask

char szTempName[LFN_MAX_PATH];

#if CHANGE_DIRECTORY_TO_COPY

char szSourceOrgCWD[PATHLEN+3];
char szDestOrgCWD[PATHLEN+3];
uchar ucSrcDOSDriveNumber;

#else

char szCompleteName[LFN_MAX_PATH];

#endif

char *pCompleteFile;                   // pointer to the full filename

uchar ucDestDOSDriveNumber;
bool fSameDrive = FALSE;

// Use a flag of 0xFF to determine if this value has been initialized.
//
int iOriginalDriveNumber = 0xFF;

char szLFNDirName[LFN_MAX_PATH];

int iSrcRootLength = 0;				// length of source path (up to filename)
int iDestRootLength = 0;			// length of destination path (up to filename)

bool fDstIsDir = FALSE;				// destination is a file or dir flag

// The argument definitions
//
ARGDEF xcopyOpts[] =
{
	{ BOOL_SWITCH, &copyArchiveAndKeep,   				'A'    },
	{ BOOL_SWITCH, &copyArchiveAndClear,  				'M'    },
	{ DATE_SWITCH, &copyAfterDate,        				'D'    },
	{ BOOL_SWITCH, &promptCreateFile,     				'P'    },
	{ BOOL_SWITCH, &copyDirs,             				'S'    },
	{ BOOL_SWITCH, &copyEmptyDirs,        				'E'    },
	{ BOOL_SWITCH, &fCopyHiddenSystem,    				'H'    },
	{ BOOL_SWITCH, &fPreserveReadOnlyAttribute,		'K'    },
	{ BOOL_SWITCH, &fQuiteMode,           				'Q'    },
	{ BOOL_SWITCH, &fShowFullPath,        				'F'    },
	{ BOOL_SWITCH, &verifyCopy,           				'V'    },
	{ BOOL_SWITCH, &waitBeforeCopy,       				'W'    },
	{ BOOL_SWITCH, &fAutoOverwrite,       				'Y'    },
	{ BOOL_SWITCH, &fDstIsDir,            				'I'    },
	{ BOOL_SWITCH, &fHideSignon,          				'['    },
	{ 0 }
};

ARGDEF FileParamOpts[] =
{
	{ LFN_STRING_ARG, &fileParam, 0, LFN_MAX_PATH},
	{ 0 }
};


bool fIsFat32World;
bool fIsLFNWorld;


/*
	displays a signon message
*/
void SignOn(void)
{
	char * szSignOn = "XCOPY %s %s\n%s\n\n";


	if( !fHideSignon )
	{
		if( fIsFat32World )
		{
			printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}

//  Display invalid switch message and exit
//
void OptError(char * szMsg)
{
	unsigned uErrorIndex = ERR_INVALIDSWITCH;
	while( *szMsg == ' ' )
	{
		szMsg++;
	}

	if( (szMsg[0] == '/') && (toupper(szMsg[1]) == 'D') )
	{
		error(ERR_INVALIDDATE);
	}

	if( fIsLFNWorld )
	{
		uErrorIndex = ERR_LFNINVALIDSWITCH;
	}

	printf( msgs[uErrorIndex], szMsg );
   printf( "\n" );
	exit(ERR_PARAMETER_MISSING);
}

/*
	displays the program usage
*/
void Usage(void)
{
	// Display copyright and version info then process user options.
	SignOn();

	printf(usageMsg);
	exit(ERR_PARAMETER_MISSING);
}

/*
	displays an internal error message
*/
void InternalError(int fileNum, int lineNum)
{
	printf( msgs[ERR_INTERNALERROR], fileNum, lineNum);
   printf( "\n" );
	exit(ERR_INTERNAL);
}


/*
	xcopy entry point
*/
int main(void)
{
	char far * lpCopyCmd;
	char szCommandLine[LFN_MAX_PATH];
   char szSourceNoSubst[LFN_MAX_PATH]; // full source path w/out subst
   char szDestNoSubst[LFN_MAX_PATH];   // full destination path w/out subst
   static char * szCommandArgs[2];
	char * szTemp;
	char * szTemp2;
	bool fSuccessful = TRUE;
	char cTemp;
#if CHANGE_DIRECTORY_TO_COPY
	unsigned uIndex;
#endif
	bool fSamePaths;
	bool fIsRemote;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	//  Determine if LFNs are supported
	//
	fIsLFNWorld = LFNPresent();


	// When XCOPY is run from inside a .BAT file, the default behavior
	// is to overwrite files.
	//
	if(AreInBatch())

	{
		fAutoOverwrite = TRUE;
	}

	// see if there is an overwrite switch setting in the COPYCMD env-variable
	//
	lpCopyCmd = GetEnvEntry("COPYCMD");		// pointer to environment copy command
	if( lpCopyCmd )
	{
		while(*lpCopyCmd)
		{
			if(_fstrncmpi(lpCopyCmd, "/y", strlen("/y")) == 0)
			{
				fAutoOverwrite = TRUE;
				break;
			}
			else if(_fstrncmpi(lpCopyCmd, "/-y", strlen("/-y")) == 0)
			{
				fAutoOverwrite = FALSE;		// prompt, even in a .BAT file
				break;
			}
			++lpCopyCmd;
		}
	}

	// get the command line from the CMDLINE environment variable
	// or the PSP...
	GetLFNCommandLine(szCommandLine, sizeof(szCommandLine));

//  DEBUG
//
//  Enable if tracing in turbo debugger (TD) to remove the
//  additional XCOPY argument from the command line
//
#if 0
	if( strnicmp( szCommandLine, "xcopy ", 6) == 0 )
	{
		memmove( szCommandLine, &szCommandLine[6], strlen(&szCommandLine[6]) + 1 );
	}
	printf("szCommandLine: <%s>\n", szCommandLine );
#endif


	// setup the commandline
	szCommandArgs[0] = &szCommandLine[0];


	//  This is for the command line parsing library function.  It
	//  indicates that no command line options are ok
	//
	//  We don't want to allow conflicting switches like /y and /-y
	//  in the same command line
	//
	SetParseFlags( PARSEF_NOARGSOK | PARSEF_NOSAMEARGS );


	// break down the switches
	ParseCmdline(1, &szCommandArgs[0], xcopyOpts);


	if (fDstIsDir && fIsFat32World)
   {
      OptError ("/I");
   }


	// Display copyright and version info then process user options.
	SignOn();

	//  Display a message like MS XCOPY if there are no options
	//  on the command line
	//
	szTemp2 = szCommandArgs[0];
	while( *szTemp2 == ' ' )
	{
		szTemp2++;
	}
	if( *szTemp2 == 0 )
	{
      // error(ERR_INVALIDPARAMETERS); with errorlevel 255...
		printf( msgs[ERR_INVALIDPARAMETERS] );
      printf( "\n" );
		printf( file_count_str, 0 );
		exit(255);
	}


	// initialize country specific information
	InitCountryInfo();


#if DEBUG
	printf("szCommandArgs: <%s>\n", szCommandArgs[0] );
#endif
	// get the SOURCE path spec, and setup the various parameters
   //
	ParseCmdline(1, &szCommandArgs[0], FileParamOpts);


	//  Make sure this is not a device
	//
	if( IsDeviceName(fileParam) )
	{
		error(ERR_FROMRESERVEDDEVICE);
	}


	// see if the source is a directory
	if( IsDir(fileParam) )
   {
		if (LastChar(fileParam) != '\\' && LastChar(fileParam) != ':')
			strcat(fileParam, "\\");
		strcat(fileParam, "*.*");
   }
   CopyNamePart(szSourceMask, fileParam);
   if( strchr(fileParam,'\\') )
   {
   	RemoveName(fileParam);
      if( !strlen(fileParam) )
      {
         fileParam[0]='\\';
         fileParam[1]=0;
      }
   }
   else if( fileParam[1] == ':' )
   {
      fileParam[2]=0;
   }
   else
   {
      fileParam[0]='.';
      fileParam[1]=0;
   }

#if 0
	printf("fileParam: <%s>\n", fileParam );
#endif

	// Now expand the source path, removing "subst" entries.
   //
   if( !SmartExpandPath( FULLPATH_DEFNAME, fileParam, szFullSourcePath))
	{
		if( (fileParam[0] != 0) && (fileParam[1] == ':') )
		{
			if( !dlIsDriveRemote(toupper(fileParam[0]) - 'A', &fIsRemote) )
			{
				if( dlCheckDOSError() == ERR_NOTDRIVE )
				{
					error(ERR_INVALIDDRIVE);
				}
			}
		}

		error(ERR_INVALIDSOURCE);
   }
   if( !SmartExpandPath(_FULLPATH_DEFNAME, fileParam, szSourceNoSubst))
	{
		error(ERR_INVALIDSOURCE);
   }
	iSrcRootLength = strlen(szFullSourcePath);
	*fileParam = 0;

#if 0
	printf("szFullSourcePath: <%s>\n", szFullSourcePath );
#endif


	// get the DESTINATION path spec, and setup the various parameters
   //
	ParseCmdline(1, &szCommandArgs[0], FileParamOpts);

	// is the destination blank?
	//
	dlGetCurrentDrive( &iOriginalDriveNumber );
	if (strlen(fileParam) == 0)
	{
		strcpy(fileParam, "A:");
		*fileParam = iOriginalDriveNumber + 'A';
	}


	// see if the destination is a directory
	if (IsDir(fileParam))
	{
		if (LastChar(fileParam) != '\\' && LastChar(fileParam) != ':')
			strcat(fileParam, "\\");
		strcat(fileParam, "*.*");
		CopyNamePart(szDestMask, fileParam );
	}
	else
	{
		// get the "name" part of the destination spec
		CopyNamePart(szDestMask, fileParam);

		// ask user if it is a subdirectory.
		if(IsToBeSubDir(szDestMask))
		{
			strcat(fileParam, "\\*.*");
			CopyNamePart(szDestMask, fileParam );
		}
	}
   if( strchr(fileParam,'\\') )
   {
   	RemoveName(fileParam);
      if( !strlen(fileParam) )
      {
         fileParam[0]='\\';
         fileParam[1]=0;
      }
   }
   else if( fileParam[1] == ':' )
   {
      fileParam[2]=0;
   }
   else
   {
      fileParam[0]='.';
      fileParam[1]=0;
   }

#if 0
	printf("fileParam: <%s>\n", fileParam );
#endif

	// Now expand the destination path, removing "subst" entries.
   //
   if( !SmartExpandPath( FULLPATH_DEFNAME, fileParam, szFullDestPath))
	{
		if( (fileParam[0] != 0) && (fileParam[1] == ':') )
		{
			if( !dlIsDriveRemote(toupper(fileParam[0]) - 'A', &fIsRemote) )
			{
				if( dlCheckDOSError() == ERR_NOTDRIVE )
				{
					error(ERR_INVALIDDRIVE);
				}
			}
		}

		error(ERR_INVALIDDESTINATION);
	}
#if 0
	printf("szFullDestPath: <%s>\n", szFullDestPath );
#endif

#if 0
	printf("fileParam: <%s>\n", fileParam );
#endif
	if( !SmartExpandPath(_FULLPATH_DEFNAME, fileParam, szDestNoSubst))
	{
		error(ERR_INVALIDDESTINATION);
	}
	iDestRootLength = strlen(szFullDestPath);

	// check dest drive for validity
	if (!IsValidDrive(toupper(*szFullDestPath)-'A'))
	{
		error(ERR_INVALIDDESTINATION);
	}


	// are we copying into the same tree we are copying from?
   //
#if 0
	printf("szSourceNoSubst: <%s>\n", szSourceNoSubst );
	printf("szFullSourcePath: <%s>\n", szFullSourcePath );
	printf("szDestNoSubst: <%s>\n", szDestNoSubst );
	printf("szFullDestPath: <%s>\n", szFullDestPath );
#endif
	if (CyclicalCopy(szSourceNoSubst, szDestNoSubst))
	{
		error(ERR_CYCLICCOPY);
	}
	fSamePaths = (strcmp(szSourceNoSubst, szDestNoSubst) == 0);


	// check for verification of file writes
   //
	if (verifyCopy == TRUE)
	{
		verify_flag = GetVerify();
		SetVerify(TRUE);
	}


	// user wants to pause before copying...
   //
	if (waitBeforeCopy == TRUE)
	{
		printf(press_a_key);
		dlGetKey( &cTemp );
	}


   // Exclusive flags - last one given has priority
   //
   if (copyArchiveAndKeep && copyArchiveAndClear)
   {
      // Done with szCommandLine?  GetLFNCommandLine() will trash it
      //
   	GetLFNCommandLine(szCommandLine, sizeof(szCommandLine));

      // Convenient reuse of szTemp (/M) and szTemp2 (/A) --
      //   we need them to handle lower case.
      //
		szTemp  = GetSwitch( 'M', szCommandLine );
		szTemp2 = GetSwitch( 'A', szCommandLine );
      if (szTemp2 > szTemp)
         // A after M, reset copyArchiveAndClear (/M)
         copyArchiveAndClear = FALSE;
      else
         // M after A, reset copyArchiveAndKeep (/A)
         copyArchiveAndKeep = FALSE;
   }


#if CHANGE_DIRECTORY_TO_COPY
	//  Save the source and destination drive numbers
	//
	ucSrcDOSDriveNumber  = toupper(*szFullSourcePath)-'A';
	ucDestDOSDriveNumber = toupper(*szFullDestPath)-'A';


	//  Save the current working directories for the source
	//  and destination drives
	//
	for( uIndex = 0; uIndex < 3; uIndex++ )
	{
		szSourceOrgCWD[uIndex] = szFullSourcePath[uIndex];
		szDestOrgCWD[uIndex] = szFullDestPath[uIndex];
	}
	dlGetCurrentDirectory( ucSrcDOSDriveNumber + 1, &szSourceOrgCWD[3] );
	dlGetCurrentDirectory( ucDestDOSDriveNumber + 1, &szDestOrgCWD[3] );
#else
#endif


	//  create the root destination path
	//
	MakeDirTree(szFullDestPath);

#if 0
	printf("szFullSourcePath: <%s>\n", szFullSourcePath );
#endif

#if CHANGE_DIRECTORY_TO_COPY
	//  Change to the source and destination directories
	//
	if( !SmartChangeDirectory(szFullSourcePath) )
	{
		error(ERR_INVALIDSOURCE);
	}
	fSameDrive = (ucSrcDOSDriveNumber == ucDestDOSDriveNumber);
	if( !fSameDrive )
	{
		SmartChangeDirectory(szFullDestPath);
	}
	dlSetCurrentDrive( (int)ucSrcDOSDriveNumber );
#else
#endif


	// initialize file storage system.  can only fail if
	// there is not enough memory.
   //
	if(!InitMemStore())
		error(ERR_INSUFFICENTMEMORY);
	else
	{
#if !ENABLE_DIRECT_MEMSTO_COPIES
		// setup the pointer the file transfer buffer
		pDiskIOBuff = FileStoreBuff();
		SHIPASSERT(pDiskIOBuff);
#endif
	}

	SmartExpandPath( FULLPATH_LONGNAME, szFullDestPath, szFullDestPath );
	SmartExpandPath( FULLPATH_LONGNAME, szFullSourcePath, szFullSourcePath );

	//  start copying from source to destination....
	//
#if 0
	printf("szFullSourcePath: <%s>\n", szFullSourcePath );
	printf("szSourceMask: <%s>\n", szSourceMask );
	printf("szFullDestPath: <%s>\n", szFullDestPath );
	printf("szDestMask: <%s>\n", szDestMask );
#endif
	XCopy( &fSuccessful, fSamePaths );


	// flush one more time
	FlushFileStore();

	// clean up
	KillMemStore();


#if CHANGE_DIRECTORY_TO_COPY
	//  Restore the original current working directories
	//
	dlSetCurrentDirectory( szSourceOrgCWD );
	dlSetCurrentDirectory( szDestOrgCWD );


	// restore "current drive" to user's setting
   //
	dlSetCurrentDrive( iOriginalDriveNumber );
#else
#endif


	// see if any files were copied
	if( uFileCount == 0 )
	{
		error(ERR_FILENOTFOUND, szSourceMask);
	}

	// show count of files copied
	printf("        %u File(s) copied\n", uFileCount);

	// restore the verify flag
	if (verifyCopy == TRUE)
		SetVerify(verify_flag);

	return ERR_NONE;
}


char * GetSwitch( char cSwitchChar, char * szCommandLine )
{
	char * pResult;


	while( TRUE )
	{
		pResult = strchr( szCommandLine, '/' );
		if( pResult )
		{
			if( toupper(pResult[1]) == toupper(cSwitchChar) )
			{
				return pResult;
			}
		}
		else
		{
			break;
		}
		szCommandLine++;
	}


	return pResult;
}


bool SmarterExpandPath( int iLFNExpandFlags, char * szPath,
		char * szExpandedPath )
{
	char szLFNPath[LFN_MAX_PATH];
	char * szTemp;


	// get the destination's absolute path (don't check existence, yet...)
	//
	if(!SmartExpandPath( iLFNExpandFlags, szPath, szExpandedPath ))
	{
		return FALSE;
	}

	strcpy( szLFNPath, szExpandedPath );
	while( TRUE )
	{
		szTemp = strrchr( szLFNPath, '\\' );
		if( szTemp )
		{
			*szTemp = 0;
		}
		else
		{
			break;
		}

		if( SmartExpandPath(_FULLPATH_LONGNAME, szLFNPath, szLFNPath) )
		{
			if( LastChar(szLFNPath) == '\\' )
			{
				szTemp++;
			}
			strcat( szLFNPath, szExpandedPath + (FP_DIST(szTemp,szLFNPath)) );
			strcpy( szFullDestPath, szLFNPath );
			break;
		}
	}

	return TRUE;
}


unsigned XCopy( bool * pfSuccessful, bool fSamePaths )
{
	bool fSuccessful;
	static int level=0;
	unsigned uCount = 0;
	FIND ffSearchBlock;
	unsigned uCheckCount;
   char * szSourceName;
	unsigned uSearchAttributes = A_HIDDEN | A_SYSTEM | A_READONLY;
	unsigned uHiddenSystemAttributes = A_HIDDEN | A_SYSTEM;
	unsigned uDirSearchAttributes = A_SUBDIR | A_READONLY;
	unsigned uDirAttr;
	bool fCopyFile;
	char * szSmartFindLFNName;


	//  validate parameters
	//
	ASSERT(szFullSourcePath);
	ASSERT(szFullDestPath);


	//  Search for files or directories
	//
#if CHANGE_DIRECTORY_TO_COPY
	fSuccessful = SmartFindFirst( uSearchAttributes, szSourceMask,
			&ffSearchBlock );
#else
   strcpy( szTempName, szFullSourcePath );
   if(strlen(szTempName) > 3)  strcat( szTempName, "\\" );
   strcat( szTempName, szSourceMask );

	fSuccessful = SmartFindFirst( uSearchAttributes, szTempName,
			&ffSearchBlock );
#endif

#if 0
   printf("[xcopy: %s]", szTempName);
#endif

	//  Since the xcopy() is recursive, the success of this find first
	//  should be "or'ed" in so that a previous success setting is not
	//  cleared with an "and'ed" or assignment.
	//
	*pfSuccessful |= fSuccessful;


	while( fSuccessful )
	{
		//  Determine if we want to copy this file if it
		//  is hidden or readonly
		//
		fCopyFile = TRUE;
		if( (!fCopyHiddenSystem) && (ffSearchBlock.attr &
				uHiddenSystemAttributes) )
		{
			fCopyFile = FALSE;
		}


		//  Determine if we want to copy this file
		//
		if( fCopyFile )
		{
			//  Make sure the file isn't being copied to itself
			//
			if( fSamePaths )
			{
				if( stricmp( MaskToName(szDestMask, ffSearchBlock.name),
						ffSearchBlock.name) == 0 )
				{
					error(ERR_COPIEDTOITSELF);
				}
			}


			//  Determine if LFNs are supported
			//
			if( fIsLFNWorld )
			{
				//  Get the long file name
				//
				szTempName[0] = 0;

#if CHANGE_DIRECTORY_TO_COPY
            pCompleteFile = ffSearchBlock.name;
#else
            strcpy( szCompleteName, szFullSourcePath );
            if(strlen(szCompleteName) > 3) strcat( szCompleteName, "\\" );
            strcat( szCompleteName, ffSearchBlock.name );
            pCompleteFile = szCompleteName;
#endif

#if 0
   printf("[xcopy lfn: %s]",pCompleteFile);
#endif

            SmartExpandPath( FULLPATH_LONGNAME, pCompleteFile, szTempName );
			   szSourceName = strrchr( szTempName, '\\' );
			   szSourceName++;


				//  Make sure the file isn't being copied to itself
				//
				if( fSamePaths )
				{
					if( stricmp( MaskToName(szDestMask, szSourceName),
							ffSearchBlock.name) == 0 )
					{
						error(ERR_COPIEDTOITSELF);
					}
				}
			}
			else
			{
				szSourceName = ffSearchBlock.name;
			}


			//  Read in the file and store it in the memory store
			//
			if( StoreFile( szSourceName, MaskToName(szDestMask, szSourceName),
					ffSearchBlock.attr, ffSearchBlock.date, ffSearchBlock.time,
					ffSearchBlock.size))
			{
				uCount++;
			}
		}


		//  Search for more files or directories
		//
		fSuccessful = SmartFindNext(&ffSearchBlock);
	}


	//  Close the search handle
	//
	SmartFindClose( &ffSearchBlock );


	//  only handle subdirectories if the user says so
	//
	if( (!copyDirs) && (!copyEmptyDirs) )
	{
		return uCount;
	}


	//  look for all directories
	//
	AddName(szFullSourcePath, "*.*");


	//  any subdirectories to recurse into?
	//
	if( fCopyHiddenSystem )
	{
		uDirSearchAttributes |= uHiddenSystemAttributes;
	}


	//  Search for subdirectories
	//
#if CHANGE_DIRECTORY_TO_COPY
	fSuccessful = SmartFindFirst( uDirSearchAttributes, "*.*",
			&ffSearchBlock );
#else
   // szFullSourcePath just had "*.*" tacked on to it, use that!
   //
	fSuccessful = SmartFindFirst( uDirSearchAttributes, szFullSourcePath,
			&ffSearchBlock );
#endif

	GetSmartFindLFNAddress( &szSmartFindLFNName );
	szFullSourcePath[iSrcRootLength] = 0;
	while( fSuccessful )
	{
		//  is this a directory?  (only recurse into subdirs)
		//  ignore the . and .. directory entries.
		//
		if( ffSearchBlock.attr & A_SUBDIR )
		{
			//  Make sure that this is not the '.' or '..'
			//  directory entries
			//
			if( ffSearchBlock.name[0] != '.' )
			{
				if ((level+1) >= MAXDEEP)
				{
					SmartFindClose( &ffSearchBlock );
					ErrorMsg(ERR_PATHTOODEEP,4);
				}


				//  Copy the long file name from the GLOBAL
				//  LFN find structure in the Datalight
				//  libraries
				//
				if( fIsLFNWorld )
				{
					strcpy( szLFNDirName, szSmartFindLFNName );
				}
				else
				{
					strcpy( szLFNDirName, ffSearchBlock.name );
				}


				//  one more level deep
				//
				level++;


				FlushFileStore();


				//  add that subdirectory to the pathname
				//
				AddName(szFullSourcePath, ffSearchBlock.name );
				AddName(szFullDestPath, szLFNDirName );


#if CHANGE_DIRECTORY_TO_COPY
				//  make the destination directory
				//
         	if( !fSameDrive )
            {
   				memcpy( szTempName, szFullDestPath, 2 );
   				strcpy( &szTempName[2], szLFNDirName );
            }
            else
            {
   				strcpy( szTempName, szFullDestPath );
            }
            SmartMakeDirectory( szTempName );

				dlSetCurrentDirectory( ffSearchBlock.name );
         	if( !fSameDrive )
            {
   				SmartChangeDirectory( szTempName );
            }
#else
            SmartMakeDirectory( szFullDestPath );
#endif

				//  determine length of root path for the source and destination
				//
				iSrcRootLength  = strlen(szFullSourcePath);
				iDestRootLength = strlen(szFullDestPath);


				//  now xcopy that directory
				//
				uCheckCount = XCopy( pfSuccessful, fSamePaths );
				uCount += uCheckCount;


				FlushFileStore();


#if CHANGE_DIRECTORY_TO_COPY
				dlSetCurrentDirectory( ".." );
         	if( !fSameDrive )
            {
   				memcpy( szTempName, szFullDestPath, 2 );
   				strcpy( &szTempName[2], ".." );
   				dlSetCurrentDirectory( szTempName );
            }
#else
#endif

				//  remove empty directories if no files were copied.
				//
				if (uCheckCount == 0 && copyEmptyDirs == FALSE)
				{
					SmartRemoveDirectory(szFullDestPath);
				}
				else
				{
					uDirAttr = ffSearchBlock.attr;
					if(!fPreserveReadOnlyAttribute)
					{
						uDirAttr = uDirAttr & ~A_READONLY;
					}
					uDirAttr = uDirAttr & ~A_SUBDIR;
					if( uDirAttr & ~A_ARCHIVE )
					{
						SmartSetFileAttributes( szFullDestPath, uDirAttr );
					}
				}


				//  back out again
				//
				RemoveName(szFullSourcePath);
				RemoveName(szFullDestPath);


				//  determine length of root path for the source and destination
				//
				iSrcRootLength  = strlen(szFullSourcePath);
				iDestRootLength = strlen(szFullDestPath);


				//  one level up
				//
				level--;
				ASSERT(level >= 0);
			}
		}


		//  Search for more subdirectories
		//
		fSuccessful = SmartFindNext(&ffSearchBlock);
	}


	//  Close the search handle
	//
	SmartFindClose( &ffSearchBlock );


	//  Return the number of files copied
	//
	return uCount;
}


bool CyclicalCopy(char szSourcePath[], char szDestPath[])
{
	char c;
   int iSrcPathLen, iDstPathLen;

	// check parameters
	ASSERT(szSourcePath);
	ASSERT(szDestPath);

	// copy directories?  If not, no chance of cyclical copy
	//
	if (!copyDirs && !copyEmptyDirs)
	{
		return FALSE;
	}


	//  Remove ending backslashes
	//
	//  ex.  C:\TEST\
	//              ^
	//
	iSrcPathLen = strlen( szSourcePath );
	if( (iSrcPathLen) && (szSourcePath[iSrcPathLen-1] == '\\' ) )
	{
		szSourcePath[iSrcPathLen-1] = 0;
	}
	iDstPathLen = strlen( szDestPath );
	if( (iDstPathLen) && (szDestPath[iDstPathLen-1] == '\\' ) )
	{
		szDestPath[iDstPathLen-1] = 0;
	}


	//  ETR 3947 BT 2-25-02
	//
	//  The below code was causing the cyclic copy error message for
	//  the non-cyclical case:
	//
	//  XCOPY /S/E AAA\BBB\*.* AAA\CCC\BBB\*.*
	//
#if 0
	// We are always dealing with expanded paths here, so there
   // will always be a backslash - strrchr() to find the length.
	//
   iSrcPathLen = strrchr(szSourcePath, '\\') - szSourcePath;
   iDstPathLen = strrchr(szDestPath, '\\') - szDestPath;
#else
   iSrcPathLen = strlen(szSourcePath);
   iDstPathLen = strlen(szDestPath);
#endif


   // IR 3837 TD 10/22/01
   // Cyclic copy will only occur if the destination path is longer
   // than the source path, and if all the characters match.
   //
	if( iSrcPathLen < iDstPathLen )
   {
   	// do source and dest match
   	if (strnicmp(szDestPath, szSourcePath, iSrcPathLen) == 0)
   	{
   		c = szDestPath[iSrcPathLen];

   		// was dest at the end or a directory break?
   		if (c == '\\' || c == '\0')
   			// cyclical copy is occurring...
   			return TRUE;
   	}
   }

	return FALSE;
}

int GetVerify(void)
{
	uchar ucTemp;


	asm	mov	ah, 54h
	asm	int	21h
	asm	mov	ucTemp, al
	return ucTemp;
}

void SetVerify(unsigned flag)
{
	uchar ucTemp;


	ucTemp = (uchar)flag;
	asm	mov	ah, 2eh
	asm	mov	al, ucTemp;
	asm	mov	dh, 0
	asm	int	21h
}

bool IsToBeSubDir(char *pStr)
{
	int c;
   bool retval = FALSE;

 	// validate parameters
	ASSERT(pStr);

   if (fDstIsDir)          // filled by /I command line param
   {
      retval = TRUE;
   }
	else if ((strchr(pStr, '*') == NULL) && (strchr(pStr, '?') == NULL))
	{
again:
		printf(specify_file, pStr);
		c = StdinGetKey();
		printf("%c\n", c);
		c = toupper(c);
		if (c == 'F')
         retval = FALSE;
		else if (c== 'D')
         retval = TRUE;
		else
			goto again;
	}
	else if ((strchr(pStr, '*') != NULL) || (strchr(pStr, '?') != NULL))
   {
      retval = FALSE;
   }
	else
   {
      retval = TRUE;
   }
   return retval;
}

bool IsDir(char *pPath)
{
	char szOrgDir[LFN_MAX_PATH];
	char c1, c2;
   bool result = FALSE;
	int iTemp;
	char * szRoot = "A:\\";


	// validate parameters
	ASSERT(pPath);

	// if the path is
	//		"." or "\"
	//	it is a directory!
	if (strlen(pPath) == 1 &&
			(*pPath == '.' || *pPath == '\\'))
		return TRUE;

	// if the path ends in
	//	":"
	// "\"
	//	".."
	// ":."
	// "\."
	// it is a directory
	else
	{
		c1 = pPath[strlen(pPath)-1];
		c2 = pPath[strlen(pPath)-2];

		if (c1 == ':' || c1 == '\\' ||
		    (c1 == '.' && (c2 == '.' || c2 == ':' || c2 == '\\')))
			return TRUE;
	}

	// can we change into this directory?
	if (pPath[1] == ':')
   {
		iTemp = toupper(pPath[0])-'A';
   }
	else
	{
		dlGetCurrentDrive( &iTemp );
	}
	dlGetCurrentDirectory( iTemp+1, &szOrgDir[3] );

	result = SmartChangeDirectory( pPath );
   szRoot[0] = iTemp + 'A';
	szOrgDir[0] = szRoot[0];
	szOrgDir[1] = szRoot[1];
	szOrgDir[2] = szRoot[2];
   SetCurrentDir( szOrgDir );
   return result;
}


//  Determine if this is a device
//
bool IsDeviceName( char * fileParam )
{
	int iActionTaken;
	int iHandle;
	unsigned uDevAttributes;
	bool fIsDevice = FALSE;
	bool fDevSuccessful;


	fDevSuccessful = SmartCreateOpenFile(O_WRONLY, A_NORMAL,
		   ACT_EXIST_OPEN, fileParam,	&iHandle, &iActionTaken );
	if( fDevSuccessful )
	{
		// check to make sure that the file is not a reserved device
		// pb
		fIsDevice = dlGetDeviceAttributes( iHandle, &uDevAttributes );
		if( (fIsDevice) && (uDevAttributes & DEVA_QUERY) )
		{
			//  fIsDevice is already set to TRUE
		}
		else
		{
			fIsDevice = FALSE;
		}


		//  Close the file
		//
		dlCloseFile(iHandle);
	}


	return fIsDevice;
}
