//
//   $Id: XCUTIL.C 1.9 2002/10/16 22:04:26Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*--------------------------------------------------------

   XCUTIL.C  created 03/04/99  by John Burnham

	Copyright (c) 1999-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This module contains the utility routines for XCOPY.

	routines contained:

	char *MaskToName(char szMask[], char szName[]);
	void NamePack(char szEName[]);
	char *NameExpand(char szEName[], char szName[]);
	char *NameFillTo(char szEName[], char szName[], int nNum);
	void NameChange(char szEName[], char szEMask[]);

---------------------------------------------------------*/
#define __FILENUM__ 3

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dir.h>
#include <dos.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlcntry.h"

#include "dlapp.h"
#include "xcopy.h"

// prototype includes
#include "xcutil.pt"

/*
	Takes a specified filename and modifies it according to the mask:

	mask=BR*.*, name=FILE.TXT, result=BRLE.TXT
*/
char * MaskToName(char szMask[], char szName[])
{
   static char szDst[LFN_MAX_PATH];

	MatchCopyName( szDst, szMask, szName );

   return (szDst);
}

/*
	$DOCMODULE:MatchCopyName

	Merge a wildcard name into a destination name.

	On Entry
		szMask		a wildcard mask
		szSrc			a legal filename (already tested to exist)
	On Exit
		szDst			contains merged name

	Examples with mask "ab*.h":
   src name     dst name
	.c           ab.h
	1.c          ab.h
	2long.c      abong.h
	3company     abompany.h
*/
void MatchCopyName(char *szDst, char *szMask, char *szSrc)
{
	char *szExtMask;
	char *szExtSrc;

	// find the extensions of mask and src name
	szExtMask = ExtPart(szMask);
	szExtSrc  = ExtPart(szSrc);

#if DEBUG
	printf ("S: %s\nM: %s\nEM:%s\nES:%s\n\n",szSrc,szMask,szExtMask,szExtSrc);
#endif

	// build the dst (name part only) 1 char at a time
	szDst = BuildName(szDst, szMask, FP_DIST(szExtMask,szMask), szSrc,
			FP_DIST(szExtSrc,szSrc));

	// no mask extension? we're done
	if(*szExtMask != '.')
		return;

	// mask has extension, so dest will too
	++szExtMask;
	*szDst++ = '.';

	// skip past '.' on src extension
	if(*szExtSrc == '.')
		++szExtSrc;

	// build the dst extension 1 char at a time
	BuildName(szDst, szExtMask, strlen(szExtMask), szExtSrc, strlen(szExtSrc));
}

/*
	$DOCPROC:BuildName

	Create a wildcard name based on a mask and source.
*/
char * BuildName(char *szDst, char *szMask, int iMaskLen,
						char *szSrc, int iSrcLen)
{
	while(iMaskLen)
	{
		// match to end of name part
		if(*szMask == '*')
		{
			memmove(szDst, szSrc, iSrcLen);
			szDst += iSrcLen;
   		++szMask;
			break;
		}

		// just use the mask char
		if(*szMask != '?')
			*szDst++ = *szMask;

		// match single character wildcards too
		else if(iSrcLen)
			*szDst++ = *szSrc;

		// on to next character
		++szMask;
		--iMaskLen;

		if(iSrcLen)
		{
			++szSrc;
			--iSrcLen;
		}
	}
	while(*szMask)
   {
      if( *szMask == '.' )
      {
         break;
      }
		if( (*szMask != '*') && (*szMask != '?') )
      {
			*szDst++ = *szMask;
      }
		szMask++;
   }

#if DEBUG
	printf ("\tremaining mask [%s]\n",szMask);
#endif

	// done, null terminate the string
	*szDst = 0;

	// return end of dest
	return szDst;
}

/*
	Find the filename extension. If no file extension, returns ptr to the
	terminating 0.

   Per MS LFN functionality, the "extension" must start somewhere
   after the first character, e.g.  ".ExtremeCase" has no extension.
*/
char * PASCAL ExtPart(char *szPathName)
{
	char *szExt;

	// look for an extension in the name part
//	szExt = intl_strrchr(NamePart(szPathName), '.');
	szExt = strrchr(szPathName, '.');
	if( (!szExt) || (szExt == szPathName) )
		szExt = szPathName + strlen(szPathName);
	return szExt;
}
